#!/usr/bin/perl
#
# compare_ref_vs_rec_Dp_indel_cts.hs.pl:
#	This script calculates the indel counts in the Dollo parsimonious indel history,
#	instead of the average over the set of genuinely parsimonious indel histories.
#	[CREATION STARTED on July 18 (Mon), 2016, ENDED on July dd (www), 2016.]
#
# compare_ref_vs_rec_indel_cts.hs.pl:
#	This script takes account of the homology structure when calculating the log-likelihood regarding indels.
#	It was created by modifying 'compare_ref_vs_rec_indel_cts.pl'.
#	(CREATED on June 29 (Mon), 2015)
#
# compare_ref_vs_rec_indel_cts.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the number of insertions as well as the number of deletions, averaged with weighting by the complete likelihood score.
#	The script was created by modifying 'compare_ref_vs_rec_compllimment.pl'.
#	(CREATED on Oct 6 (Mon), 2014.)
#
# compare_ref_vs_rec_compllimment.pl:
#	This is a slightly modified version of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
#	The major difference from the predecessor is that it now accepts four input file paths,
#	instead of the path to an input directory as the predecessor did.
#	(CREATION STARTED on Aug 8 (Fri), 2014, ENDED on Aug 9 (Sat), 2014)
#
# compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl:
#	This revised version of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' uses the Perl package, "MyTreeMap_indels_spt_odr.pm,"
#	instead of "MyTreeMap_indels.pm" as the predecessor was using.
#	And it also underwent some additional minor modifications,
#	which were mainly imported from "br_infer_psm_indel_events_spt_odr_dawg.v2.pl"
#	in the "Proof_of_Principle2" session for the framework to enable indel likelihood analyses.
#	(CREATED on Aug 5 (Tue), 2014)
#
#	[REVISED on Aug 6 (Wed), 2014]
#	For the indel analyses (enumerating parsimonious indel scenarios
#	plus calculating the approximate probabilities of the gap configurations of the segments),
#	the script now uses the input tree as it is, 
#	even if it has a bifurcated root.
#
# compare_ref_vs_rec_abs_scoring.PhyML.pl:
#	This Perl script compares the reference MSA and a reconstructed MSA
#	in terms of the 'absolute score', which is the summation
#	of the log-likelihood of gap pattern and that of the residue pattern.
#	It was created by importing some parts from the script 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	

use strict;
use warnings;

use MyArrayManipu_num qw( log_sum_exp );
use MyArrayManipu qw( binsearch_string );
use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices
	);

use MyMalnIO qw(
	extract_fasta
	extract_alignment
	pr_ODEN_input
	pr_phylip_aln
	);
	## pr_clustal_like_lite_aa
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file output_tree
	);

	# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	);

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );

=for the predecessor 'compare_ml_msa.bch.extJC.6t.PhyML.pl' 
use MyBrief_NJtree_extJC qw( brief_nj_tree_extJC );
=cut

use MyPAML_parser qw( retrieve_aa_freqs );
	##prepare_paml_ctrl 

use MyApBayes_Score_MSA_ExtJC qw(
	readin_set_trees3 log_akaike_weights
	);
	## apbayes_scoring_msa_extJC_f 

	### The following three modules are for calculating log-likelihoods of substitution patterns. ###

=for the predecessor 'compare_ml_msa.bch.extJC.6t.PhyML.pl'
use MyMalnManipu_UH2 qw( 
	parse_altaln_header extract_aln_mtrx3
	fetch_nonred_clms_in_msas_thr
	);
	##fetch_nonred_clms_in_msas_mdst
use MyTreeMap_MLf qw(
	msa_likelihood_extJC_fnr
	);
	## msa_likelihood_extJC_f
=cut
use MyPhyML_parser qw(
	est_resfreqs_frm_msa print_custom_sbst_model
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	);
	## extract_phyml_statistics


	### For enumerating parsimonious indel scenarios. ###

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
	# A MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). #
use MyTreeMap_indels_spt_odr qw(
	enum_psm_indel_scns_for_msa
	);
=cut

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'.
use MyTreeMap_indels qw(
	enum_psm_indel_scns_for_msa
	);
	## extract_cls_gblocks msa_rep_gpatterns
=cut

	### For log-likelihoods of indel scenarios. ###

		## IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. ##
use MyReadAlnProbIngredients qw(
	readin_delta_del
	assign_br2log_mfacs
	);

		## IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. ##
use MyTreeMap_indels_ML_hs qw(
	apprx1_tot_lnlk_gpattern_msa2
	);
	## cal_delta_exp_del_5p2
	## cal_log_normal_facs_indel_distri
	## cal_lnlk1_indel_scn_cls2


=for the predecessor 'compare_ref_vs_rec_indel_cts.pl'.
use MyTreeMap_indels_ML qw(
	cal_log_normal_facs_indel_distri
	cal_delta_exp_del_5p
	cal_lnlk1_indel_scn_cls
	apprx1_tot_lnlk_gpattern_msa
	);
=cut

		## IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl', in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. ##
use MyTreeManipu_III qw( make_node2abs_depth  make_node2ct_exoffs );

use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	create_initial_set_indels_spt_odr
	);
	## enum_psm_indel_scns_for_msa

use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
		## END of "IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl', in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'." ##

=set aside for future use
use MyBin_Oper qw( format_bin_pattern select_by_bin_value );
use MyMalnManipu_UH qw( div_msa_by_gappatterns cls_blcs_by_patterns );
=cut

sub readin_list_segs2 (\$\@); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_list_segs (\$\%);
sub readin_simul_params (\$);
sub inspect_indel_events (\@\@$$$);

sub readin_dawg_tree2 (\$); # ADDED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##sub readin_dawg_tree (\$); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 
sub ln_ml_sgl_seq2 (\$\%); # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011 #
sub add_dummy_seq (\@\@\%\%\%\%\%\@); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 12 (Tue), 2011 #
####sub even_brlen_2seq_tree (\%$);
####sub parse_paml_main_output (\$);
##sub fetch_clmwise_lnlks (\$$);

sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@); # IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl' in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
##sub cal_init_psm_cands_for_twin_msas (\%\%\%\%\%\@\@\@$$);

$| = 1; # Flush the buffer of standard output everytime data flows into it.

my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 


	## File names storing important parameters. ##
	## (IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'.) ##

our $FILENAME_DELTA_DEL = 'delta_del.txt'; # The name of the file storing '$delta_del', which is important for calculating the exit rate of the entire sequence. #

our @FILENAMES_LOG_MFACS = ( # Names of the files storing logarithmic multiplication factors, which are in $indir_mfacs. #

		# In the bulk #
	'lmfacs_asites_vs_dgaps.txt', # For the cases where there are some ancestral sites and no descendant sites in between preserved ancestral sites (PASs). #
	'lmfacs_agaps_vs_dsites.txt', # For the cases where there are no ancestral sites and some descendant sites in between PASs.
	'lmfacs_asites_nhom_dsites.txt', # For the cases where there are some ancestral sites and some descendant sites in between PASs.

		# On either boundary (i.e., either PWA end). "bd_$filename" corresponds to $filename above. #
	'bd_lmfacs_asites_vs_dgaps.txt',
	'bd_lmfacs_agaps_vs_dsites.txt',
	'bd_lmfacs_asites_nhom_dsites.txt'
	);


	### Fixed Parameters ###

	# INPORTED from 'ml_scoring_msa.bch.pl' on Mar 15 (Tue), 2011 #
(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;
our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

##our $GAP = '-'; # Character representing a gap
our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

	# (ADDED on Nov 29 (Mon), 2010) #
our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	# (ADDED on Mar 7 (Mon), 2011) #
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.

our %SBSTMODEL2FILENAME = ( # $sbst_model_name (for PhyML) => $dat_name (for PAML)  (relevant only when $seqtype eq 'AA')
	'LG' => 'lg.dat',
	'WAG' => 'wag.dat',
	'JTT' => 'jones.dat',
	'MtREV' => 'mtREV24.dat',
	'Dayhoff' => 'dayhoff.dat',
	'CpREV' => 'cpREV.dat',
	'MtMam' => 'mtmam.dat',
	'MtArt' => 'mtArt.dat',
	);

our %SUPPORTED_SBSTMODELS = ( # $sbst_model_name (for PhyML) => 1 (revelant only when $seqtype eq 'NT')
	'HKY85' => 1, # default #
	'JC69' => 1,
	'K80' => 1, 
	'F81' => 1,
	# 'F84' => 1,
	# 'TN93' => 1,
	);

	## For calculating log-likelihoods of indel scenarios. ###

our %INDEL_MODEL2DISTRI = ('PL' => 'P'); # = (indel model => distribution type, ...) # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##our %INDEL_MODEL2DISTRI = ('zipf' => 'P'); # = (indel model => distribution type, ...)


	### Parameters to be adjusted ###

	# The level of tests to examine the behavior of the script (ADDED on Sep 1 (Sun), 2012):
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.
our $TEST_LEVEL = 1;
#our $TEST_LEVEL = 3;

		# ADDED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
our $TEST_LEVEL2 = 1;
#our $TEST_LEVEL2 = 5;

	# (IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl', in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.)

		# For the subroutine, 'cal_init_psm_cands_for_sgl_msa':
our $TEST_LEVEL_cipc2 = 0;
##our $TEST_LEVEL_cipc2 = 5;

		# For the subroutine, 'cal_init_psm_cands_for_twin_msas':
#our $TEST_LEVEL_cipc = 0;
##our $TEST_LEVEL_cipc = 5;

	# END of "(IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl', in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.)"


	# The level of debugging output for 'extract_phyml_statistics2' (ADDED on Aug 26 (Sun), 2012)
our $DBG_LEVEL_EXPHYML = 0;

our $FLG_RMV_TMPOUT = 1; ## For actual (large-scale) analyses
#our $FLG_RMV_TMPOUT = 0; ## For TESTING the script ##
	# = 1 if you want to remove temporary directories and files under them,
	# = 0 if you want to keep them.

our $MIN_FREQ = 1E-06; # If the frequency of a residue is zero in the anchor MSA, this nominal frequency is used instead.

	# Essential for calculating the log-likelihoods of indel scenarios. #
	# $OPT_BD_INDELS = 0, fixed length; = 1, sum of rates; = 2, exact calculation.
#our $OPT_BD_INDELS = 0;
#our $OPT_BD_INDELS = 1;
our $OPT_BD_INDELS = 2;


	# ADDED in creation of 'compare_ml_msa.bch.extJC.6t.I.pl' (on Mar 22 (Tue), 2011) #
our $SBST_MODEL_AA = 'JTT'; # Substitution model for amino acid sequences
our $SBST_MODEL_NT = 'HKY85'; # Substitution model for nucleotide sequences
##our $SBST_MODEL_NT = 'HKY'; # Substitution model for nucleotide sequences --- a BUG fixed on Apr 20 (Wed), 2011
our $CT_RCATE = 1; # #{substitution rate categories}
our $ALPHA =0; # Gamma shape parameter for the rate variation across sites ('': to be determined; 0: uniform rate model)
our $TOPO_SEARCH = 'NNI'; # Tree topology search method. ('NNI', 'SPR', or 'BEST')

	# ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' (on Mar 24 (Thu), 2011) #
	# REVISED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 13 (Wed), 2011
our $RESFREQ_AA = 'm'; # (To be changed to 'e' in the near future)
our $RESFREQ_NT = 'e'; # (To be changed to 'e' in the near future)
##our $RESFREQ = 'm'; # (To be changed to 'e' in the near future)
	# Option on what to do with residue frequencies.
	# = 'e' if you want to use the residue frequencies in the input anchor MSA;
	# = 'm' if you want to use the ML estimates of the nucleotide frequencies, or the amino-acid frequencies defined by the substitution model.

	### The following are IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'
	### in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ###

	# In this analysis of an MSA simulated by Dawg, null columns will play important roles.
$MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS = 0;

	# Null-clusters (=clusters consisting only of a null-block each) will be analysed.
$MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS = 1;

	# This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	#

our $UPPER_BOUND_ANAL_GAPLEN = 3000; # MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
###our $UPPER_BOUND_ANAL_GAPLEN = 150;
#our $UPPER_BOUND_ANAL_GAPLEN = 120;
$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

$MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 5000; # MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
###our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;
#our $UPPER_BOUND_ANAL_LEN_GBCLS = 150;
$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;

$MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;

	### END of "The following are IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014)." ###


		### ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING ###

	## (MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' on Aug 5 (Tue), 2014.) ##

	# Control the level of verboseness of the test message for 'MyTreeMap_indels_spt_odr.pm'.
	# (Basically, believe in the behaviours of the main subroutine 'enum_psm_indel_scns_for_msa',
	#  as well as of its satellite subroutines.)
$MyTreeMap_indels_spt_odr::TEST_LEVEL = 1;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1hf = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2 = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2c = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2inb = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL3 = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL3d = 0;

=for DEBUGGING
$MyTreeMap_indels_spt_odr::TEST_LEVEL = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1hf = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2 = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 5;
=cut

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'. #
	# Control the level of verboseness of the test message for 'MyTreeMap_indels.pm'.
	# (Basically, believe in the behaviours of the main subroutine 'enum_psm_indel_scns_for_msa',
	#  as well as of its satellite subroutines.)
$MyTreeMap_indels::TEST_LEVEL = 1;
$MyTreeMap_indels::TEST_LEVEL1 = 0;
$MyTreeMap_indels::TEST_LEVEL1hf = 0;
$MyTreeMap_indels::TEST_LEVEL2 = 0;
$MyTreeMap_indels::TEST_LEVEL2i = 0;
$MyTreeMap_indels::TEST_LEVEL2c = 0;
$MyTreeMap_indels::TEST_LEVEL2d = 0;
$MyTreeMap_indels::TEST_LEVEL3 = 0;
$MyTreeMap_indels::TEST_LEVEL3d = 0;
##=for DEBUGGING
$MyTreeMap_indels::TEST_LEVEL = 5;
$MyTreeMap_indels::TEST_LEVEL1 = 5;
$MyTreeMap_indels::TEST_LEVEL1hf = 5;
$MyTreeMap_indels::TEST_LEVEL2 = 5;
$MyTreeMap_indels::TEST_LEVEL2i = 5;
##=cut
=cut
	## END of "(MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' on Aug 5 (Tue), 2014.)" ##

	# Control the level of verboseness of the test message for 'MyTreeMap_indels_ML.pm'.
	# (Basically, believe in the behaviours of the main subroutine 'apprx1_tot_lnlk_gpattern_msa'
	#  as well as of its satellite subroutines.)
$MyTreeMap_indels_ML_hs::TEST_LEVEL = 1;
$MyTreeMap_indels_ML_hs::TEST_LEVEL1 = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL1a = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2a = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2b = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2c = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2s = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL3 = 0;

		### END of "ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING" ###




	## IMPORTED from 'compare_ml_msa.bch.extJC.6t.PhyML.pl' (on Aug 26 (Sun), 2012) ##
	# Now, $main_bin_tree and $main_bin_clmlk are identical, renamed as $main_bin.
	# $src_ctrl and $src_dat are now optional.
	# Added another option "--resfreq", which specifies what are used for the residue frequencies.
	# Added another option "--sv_ratio", which specifies the transition/transversion ratio (on Apr 8 (Fri), 2011).

	# Further MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #

unless (@ARGV > 9) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} {out_summ_file} {log_dir} [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio={transition/transversion ratio (default: undef)} --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)} ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $outtopdir, $out_summ, $logdir) = @ARGV[0 .. 9];

=For the predecessor 'compare_ref_vs_rec_indel_cts.hs.pl'.
	# MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.hs.pl' (on June 29 (Mon), 2015). #
unless (@ARGV > 10) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {indir_mfac_data} {out_top_dir} {out_summ_file} {log_dir} [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio={transition/transversion ratio (default: undef)} --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)} ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $indir_mfacs, $outtopdir, $out_summ, $logdir) = @ARGV[0 .. 10];
=cut

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
		# IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #
my $infile_delta_del = "$indir_mfacs/$FILENAME_DELTA_DEL"; # This file stores '$delta_del', which is important for calculating the exit rate of the entire sequence. #
=cut

=for the predecessor 'compare_ref_vs_rec_indel_cts.pl'.
	# MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #
unless (@ARGV > 9) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {inaln_file_ref} {inaln_file_rec} {inlist_segs} {inlist_params} {out_top_dir} {out_summ_file} {log_dir} [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio={transition/transversion ratio (default: undef)} --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)} ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $outtopdir, $out_summ, $logdir) = @ARGV[0 .. 9];
=cut

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
unless (@ARGV > 6) {
	print STDERR "\nUsage:  [perl] $program_name {main_binary} {sequence_type ('AA' for amino acid sequences, 'NT' for  nucleotide sequences)} {anchor_msa_file} {in_top_dir} {out_top_dir} {out_summ_file} {log_dir} [ --sbst_model={substitution model name (default: $SBST_MODEL_AA for amino acids, $SBST_MODEL_NT for nucleotides)} --ct_rcate=\#{substitution rate categories (default: $CT_RCATE) }}  --alpha={gamma shape parameter for rate variation across sites (default: $ALPHA) } --topo_search={tree topology search method (default: $TOPO_SEARCH)} --resfreq={what to do with residue frequencies (default: $RESFREQ_AA for amino acids, $RESFREQ_NT for nucleotides)} --sv_ratio={transition/transversion ratio (default: undef)} --ctrl={path to the control file (default: undef)} --sbst_mtrx={path to the file for substitution model (MANDATORY for amino-acids)} --intree={input_tree_file (default: undef)} ]\n"; 
	die "$! .\n\n";
}
my ($main_bin, $seqtype, $anchor_msa_file, $intopdir, $outtopdir, $out_summ, $logdir) = @ARGV[0 .. 6];
=cut

my $main_bin_tree = my $main_bin_clmlk = $main_bin;

$seqtype = uc $seqtype;
unless (($seqtype eq 'AA') or ($seqtype eq 'NT')) {
	print STDOUT "[ERROR '$program_name'] ILLEGAL INPUT: The sequence type '$seqtype' is not expected. Please specify either 'AA' or 'NT' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $sbst_model = ($seqtype eq 'AA') ? $SBST_MODEL_AA : $SBST_MODEL_NT;
my $ct_rcate = $CT_RCATE;
my $alpha = $ALPHA;
my $topo_search = $TOPO_SEARCH;


my $sv_ratio ;

my $resfreq = ($seqtype eq 'AA') ? $RESFREQ_AA : $RESFREQ_NT;

my ($intree, $src_ctrl, $src_dat); # Input tree file, control file specifying detailed options, and data file specifying the substitution model
for (my $k=10; $k<@ARGV; $k++) { # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #
##for (my $k=7; $k<@ARGV; $k++) {...}

	my $arg = $ARGV[$k];
	unless ($arg =~ s/^\-\-?//) {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	    next;
	}
	if ($arg =~ s/^intree\=// ) {
	    ($arg =~ m/\S/) and ($intree = $arg);

	} elsif ($arg =~ s/^sv_ratio\=//) {
	    ($arg =~ m/\S/) and ($sv_ratio = $arg);

	} elsif ($arg =~ s/^ctrl\=//) {
	    ($arg =~ m/\S/) and ($src_ctrl = $arg);
	} elsif ($arg =~ s/^sbst_mtrx\=//) {
	    ($arg =~ m/\S/) and ($src_dat = $arg);
	} elsif ($arg =~ s/^resfreq\=//) {
	     if (($arg =~ m/\S/ ) and ($arg ne $resfreq)) {
		print STDOUT "[INFO '$program_name'] The argument \$resfreq = '$resfreq' will be changed to '$arg' ...\n\n";
		$resfreq = $arg;
	     }

	} elsif ($arg =~ s/^sbst_model\=//) {
	     if (($arg =~ m/\S/ ) and ($arg ne $sbst_model)) {
		print STDOUT "[INFO '$program_name'] The argument \$sbst_model = '$sbst_model' will be changed to '$arg' ...\n\n";
		$sbst_model = $arg;
	     }
	} elsif ($arg =~ s/^ct_rcate\=// ) {
	    if (($arg =~ m/\S/ ) and ($arg ne $ct_rcate)) {
		print STDOUT "[INFO '$program_name'] The argument \$ct_rcate = '$ct_rcate' will be changed to '$arg' ...\n\n";
		$ct_rcate = $arg;
		##$sbst_model = $arg; # A BUG FIXED on March 23 (Wed), 2011 #
	    }
	} elsif ($arg =~ s/^alpha\=// ) {
	    if ($arg =~ m/\S/ ) {
		if ($arg ne $alpha) {
		   print STDOUT "[INFO '$program_name'] The argument \$alpha = '$alpha' will be changed to '$arg' ...\n\n";
		   $alpha = $arg;
		   ##$sbst_model = $arg; # A BUG FIXED on March 23 (Wed), 2011 #
		}
	    } else {
		print STDOUT "[INFO '$program_name'] I found \$alpha = '', I will regard this as an order to UNDEFINE it (, which means optimizing it) ...\n\n";
		$alpha = undef;
	    }
	} elsif ($arg =~ s/^topo_search\=// ) {
	    if (($arg =~ m/\S/ ) and ($arg ne $topo_search)) {
		print STDOUT "[INFO '$program_name'] The argument \$topo_search = '$topo_search' will be changed to '$arg' ...\n\n";
		$topo_search = $arg;
	    }
	} else {
	    print STDOUT "[WARNING '$program_name'] Found an UNEXPECTED command-line ARGUMENT '--$arg': $! !\n";
	    print STDOUT "                 ---  I will just skip it ...\n\n";
	}
}


	## Check the consistency of the command-line arguments #

if ($seqtype eq 'AA') {
    unless (($sbst_model eq 'custom') or (defined $SBSTMODEL2FILENAME{$sbst_model})) {
	print STDOUT "[ERROR '$program_name'] Sorry, the amino-acid substitution model '$sbst_model' CANNOT BE USED for this version !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

} else {

    unless (defined $SUPPORTED_SBSTMODELS{$sbst_model}) {
	print STDOUT "[ERROR '$program_name'] Sorry, the nucleotide substitution model '$sbst_model' is NOT SUPPORTED by this version !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if ($seqtype eq 'AA') { # DEBUGGED on Apr 20 (Wed), 2011 #
##if (($seqtype eq 'AA') and ($resfreq eq 'e')) {
    unless (defined $src_dat) {
	print STDOUT "[ERROR '$program_name'] The PATH to the FILE for the amino-acid substitution model '$sbst_model' is UNSPECIFIED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (($resfreq ne 'm') and ($resfreq ne 'e') and ($seqtype ne 'NT')) {
	print STDOUT "[ERROR '$program_name'] The residue frequency option '$resfreq' is NOT COMPATIBLE with the sequence type '$seqtype' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
if ((defined $intree) and ($intree =~ m/\S/) and ($seqtype eq 'NT') and ($resfreq eq 'm')) {
	print STDOUT "[ERROR '$program_name'] The option '-seqtype=NT -resfreq=m' is NOT COMPATIBLE with the use of an input tree ($intree), please use a different option for -resfreq !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
if ( (defined $sv_ratio) and (($sbst_model eq 'JC69') or ($sbst_model eq 'F81')) ) {
	print STDOUT "[WARNING '$program_name'] The option '--sv_ratio=$sv_ratio' is NOT NECESSARY for the substitution model '$sbst_model' !\n";
	print STDOUT "          --- I will UNDEFINE it ...\n\n";
	$sv_ratio = undef;
}
if (($sbst_model eq 'JC69') or ($sbst_model eq 'K80')) {
	print STDOUT "[WARNING '$program_name'] The substitution model '$sbst_model' has uniform equilibrium frequencies (fA = fC = fG = fT = 0.25), so I will SET \$resfreq = '' !\n";
	$resfreq = '';
}

if ((defined $intree) and !(defined $alpha)) {
	print STDOUT "[ERROR '$program_name'] Now \$alpha = undef, but I canNOT ML-estimate the shape parameter when an input tree file is specified (as '$intree') !\n";
	print STDOUT "[ERROR '$program_name'] $!\n\n";
	die ;
}


	# Out of command-line arguments,
	# build up options shared by the tree inference step and
	# the column-wise likelihood calculation step.

my $seqtype_lc = lc $seqtype;

our $OPTIONS_SHARED = " -d $seqtype_lc -c $ct_rcate ";
## (($ct_rcate > 1) and (defined $alpha) and ($alpha > 0)) and ($OPTIONS_SHARED .= " -a $alpha");



	### Input/Output files/directories ###

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
my $inaln_file_ref = "$intopdir/dawg.fasta";
my $inaln_file_rec = "$intopdir/mafft.fasta";
my $inlist_segs = "$intopdir/segs.txt";
my $inlist_params = "$intopdir/dawg_inp.txt";
=cut

	## Interim files ##

my $name_interim_trees = "tree_whole_full.txt";
my $interim_trees = "$outtopdir/$name_interim_trees";

my $name_log_interim_trees = "log_$name_interim_trees";
my $log_interim_trees = "$logdir/$name_log_interim_trees";

my $interim_aln_name = 'input_aln.fa'; # $interim_aln = "$outdir/$interim_aln_name"
my $tmpfile_name = 'lnl_msa.txt';

	# $int_clmlk_file_ref = "$top_clmlk_dir_ref/$pref_clmlk\{seg_id\}$suff_clmlk".
	# $int_clmlk_file_rec = "$top_clmlk_dir_rec/$pref_clmlk\{seg_id\}$suff_clmlk".
my $top_clmlk_dir_ref = "$outtopdir/Clm_lnlk/Ref";
my $top_clmlk_dir_rec = "$outtopdir/Clm_lnlk/Rec";
my $pref_clmlk = 'lnlk_clm.';
my $suff_clmlk = '.txt';

	# Top Temporary directory for PAML #
	# temporary directory (reference) = "$top_tmpdir_ref/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
	# temporary directory (reconstructed) = "$top_tmpdir_rec/s{seg_ID}/g{indx_gpattern}/t{indx_tree}"
my $top_tmpdir_ref = "$outtopdir/Tmp/Ref";
my $top_tmpdir_rec = "$outtopdir/Tmp/Rec";


	# Temporary directory for phylogenetic tree reconstruction #
	# $inaln_for_tree = "$tmpdir_phyml/$inaln_for_tree_name";
my $tmpdir_phyml = "$outtopdir/Infer_tree";

	# Master file for a custom substitution model (only for $seqtype eq 'AA' and $resfreq eq 'e') #
	# $master_sbst_file = "$outtopdir/$dat_name";
	# $master_ctrl_file = "$outtopdir/$ctrl_name";
my $master_sbst_file;
my $master_ctrl_file;


	## Check Necessary Files/Directories ##

foreach my $binary ($main_bin_tree, $main_bin_clmlk) {
    unless (( -f $binary ) and ( -x $binary )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND an executable PROGRAM '$binary' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (defined $src_ctrl) {
   unless (( -f $src_ctrl ) and ( -r $src_ctrl )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable PAML CONTROL FILE '$src_ctrl' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
   }
}
if (defined $src_dat) {
   unless (( -f $src_dat ) and ( -r $src_dat )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable PAML CONTROL FILE '$src_dat' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
   }
}

my ($ctrl_name) = (defined $src_ctrl) ? ($src_ctrl =~ m!([^/\s]+)\s*$!) : ('ctrl_phyml.txt');
my ($dat_name) = (defined $src_dat) ? ($src_dat =~ m!([^/\s]+)\s*$!) : ('undef');
my $log_name = (defined $src_ctrl) ? "log_$ctrl_name.txt" : "log_PhyML.txt";


if (($seqtype eq 'AA') and (defined $dat_name)) {
    my $dat_name0 = $SBSTMODEL2FILENAME{$sbst_model};
    unless (($dat_name eq 'custom') or ($dat_name eq $dat_name0)) {
	print STDOUT "[ERROR '$program_name'] The DATA FILE NAME for the amino-acid substitution model '$sbst_model' is NOT '$dat_name', as you specified, BUT '$dat_name0' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (($seqtype eq 'AA') and ($resfreq eq 'e')) {
	$dat_name =~ s/\.dat$//;
	##$dat_name =~ s/\.dat^//; # A BUG Fixed on Mar 29 (Tue), 2011 #

	$dat_name .= '_custom.dat';
	$master_sbst_file = "$outtopdir/$dat_name";
	$master_ctrl_file = "$outtopdir/$ctrl_name";
}


foreach my $file ($intree) {
    (defined $file) or next; ## ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) ##

    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
	# ADDED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #
foreach my $indir ($indir_mfacs) {
    unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT DIRECTORY '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}
=cut

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

##foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params, $infile_delta_del) { # MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. # # DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. # ...}
foreach my $infile ($inaln_file_ref, $inaln_file_rec, $inlist_segs, $inlist_params) { # RESURRECTED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
    unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ; 
    }
}

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'
unless (( -d $intopdir ) and ( -r $intopdir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable  INPUT DIRECTORY '$intopdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
=cut

foreach my $outdir ($outtopdir, $logdir, $top_clmlk_dir_ref, $top_clmlk_dir_rec, $top_tmpdir_ref, $top_tmpdir_rec, $tmpdir_phyml) {
    unless ( -d $outdir ) {
	unless (0 == system ("mkdir -p $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create a TOP OUTPUT DIRECTORY '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
    unless (( -r $outdir ) and ( -w $outdir )) {
	unless (0 == system ("chmod u+rw $outdir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE the permission MODES of the top output directory '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<<<< Program name: $program_name (path: $program_path) >>>>>>>>\n\n";

print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

print STDOUT "Binary programs:\n";
print STDOUT "     (for tree inference)  $main_bin_tree ,\n";
print STDOUT "     (for column-wise likelihoods) $main_bin_clmlk .\n\n";

print STDOUT "Source control file (for PhyML): ", ((defined $src_ctrl) ? $src_ctrl : 'Not used'), " .\n";
print STDOUT "Source file of substitution model (for PhyML): ", ((defined $src_dat) ?  $src_dat : 'Not used'), " .\n\n";

print STDOUT "Sequence type: ", (($seqtype eq 'AA') ? 'Amino Acid' : 'Nucleotide'), " .\n\n";

print STDOUT "Input files:\n";

if (defined $intree) { 
    print STDOUT "     (Tree)      $intree .\n\n";
}
print STDOUT "    (anchor MSA) $anchor_msa_file .\n\n";


	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 8 (Fri), 2014). #

print STDOUT "   (simulation parameters} $inlist_params,\n";
print STDOUT "   (reference MSA)         $inaln_file_ref,\n";
print STDOUT "   (reconstructed MSA)     $inaln_file_rec,\n";
print STDOUT "   (list of segments)      $inlist_segs .\n\n";

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
print STDOUT "   (simulation parameters} {in_top_dir}/dawg_inp.txt,\n";
print STDOUT "   (reference MSA)         {in_top_dir}/dawg.fasta,\n";
print STDOUT "   (reconstructed MSA)     {in_top_dir}/mafft.fasta,\n";
print STDOUT "   (list of segments)      {in_top_dir}/segs.txt.\n";
print STDOUT "        where {in_top_dir} = $intopdir .\n\n";
=cut

print STDOUT "Directory for interim output files for tree inference: $tmpdir_phyml .\n\n";

print STDOUT "Interim output list of candidate trees:  {topoutdir}/$name_interim_trees .\n";

if (defined $master_sbst_file) {
    print STDOUT "Master file for a custom  substitution model: {topoutdir}/$dat_name .\n";
}
if (defined $master_ctrl_file) {
    print STDOUT "Master control file for PhyML: {topoutdir}/$ctrl_name .\n";
}

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
	## IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. ##
print STDOUT "Input file recording ", '$delta_del', ": $infile_delta_del .\n\n";
print STDOUT "Input directory storing (log-)multiplication factors: $indir_mfacs ,\n\n";
print STDOUT "   which holds the following input files:\n\n";
print STDOUT '     ', join (', ', @FILENAMES_LOG_MFACS), " .\n\n\n";
=cut


=DISABLED because they do not appear to be used effectively.
print STDOUT "Interim output list of column-wise log-likelihoods:  {top_clmlk_dir}/$pref_clmlk\{seg_id\}$suff_clmlk,\n";
print STDOUT "   where {top_clmlk_dir} = $top_clmlk_dir_ref for the reference MSA,\n";
print STDOUT "                         = $top_clmlk_dir_rec for the reconstructed MSA .\n\n";
=cut

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
print STDOUT "Temporary directories for PAML:  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n";
print STDOUT "   where {top_tmpdir} = $top_tmpdir_ref for the reference MSA,\n";
print STDOUT "                      = $top_tmpdir_rec for the reconstructed MSA.\n\n";
=cut

print STDOUT "Summary output: $out_summ .\n\n";
print STDOUT "Directory for the log files: {log_dir} = $logdir .\n\n";
print STDOUT "Log files:\n";

print STDOUT "   (for candidate tree enumeration) {log_dir}/$name_log_interim_trees .\n\n";

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.
print STDOUT "   (for scoring the sbst pattern of the \{reference,reconstructd\} MSA for each segment) {log_dir}/log_apbayes_score.\{ref,rec\}.\{seg_id\}.txt .\n\n";
=cut

print STDOUT "   (for scoring the gap-pattern of the \{reference,reconstructed\} MSA for each segment} {log_dir}/log_lnlk_gpattern.\{ref,rec\}.\{seg_id\}.txt .\n\n";

if ($FLG_RMV_TMPOUT) {
	print STDOUT "* All the temporary outputs will be removed .\n\n";
} else {
	print STDOUT "* All the temporary outputs will be kept .\n\n";
}


print STDOUT "Parameters for tree inference (via PhyML):\n\n";
print STDOUT "   Substitution model:  $sbst_model .\n";
print STDOUT "   #{Substitution rate categories}: $ct_rcate .\n";
print STDOUT "   Gamma shape parameter for the rate variation across sites: ", ((defined $alpha) ? $alpha : 'undef'), " .\n";
print STDOUT "   Tree topology search method: $topo_search .\n";

print STDOUT "   Option on what to do with the residue frequencies: ", (($resfreq =~ m/\S/) ? $resfreq : "uniform (default for $sbst_model)"), " .\n\n\n";

if ($seqtype eq 'NT') {
    print STDOUT "   Transition/transversion ratio: ", ((defined $sv_ratio) ? $sv_ratio : 'undef'), " .\n";
}

print STDOUT "\nOptions for PhyML shared by both tree inference and column-wise likelihood calculation:\n\n";
##print STDOUT "     '$OPTIONS_SHARED' .\n\n";
print STDOUT '   $OPTIONS_SHARED = ', "'$OPTIONS_SHARED' .\n\n";

print STDOUT "\nOption for the calculation of log-likelihoods of indel scenarios:\n\n";
print STDOUT '   $OPT_BD_INDELS  = ', "'$OPT_BD_INDELS' .\n\n";

print STDOUT "\nLevels of tests for debugging or behavior tests:\n\n";

print STDOUT '   $TEST_LEVEL        =  ', $TEST_LEVEL, " ,\n";
print STDOUT '   $TEST_LEVEL2       =  ', $TEST_LEVEL2, " ,\n"; # ADDED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
print STDOUT '   $TEST_LEVEL_cipc2  =  ', $TEST_LEVEL_cipc2, " ,\n"; # ADDED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
print STDOUT '   $DBG_LEVEL_EXPHYML =  ', $DBG_LEVEL_EXPHYML, " ,\n";
print STDOUT '   $FLG_RMV_TMPOUT    =  ', $FLG_RMV_TMPOUT, " .\n\n";

	## IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##

print STDOUT "\nParameters in the Perl package 'MyTreeMap_indels_spt_odr.pm' (and 'MyTreeMap_indels_ML_hs.pm').\n\n";

print STDOUT '$MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS =' , $MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS, " ,\n";

print STDOUT '$MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS = ', $MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS, " ,\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_GAPLEN = ', $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN, " .\n\n";

print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " .\n\n";

print STDOUT '$MyTreeMap_indels_spt_odr::$TEST_LEVEL_GEN_NULL_BLK = ', $MyTreeMap_indels_spt_odr::TEST_LEVEL_GEN_NULL_BLK, " ,\n";

print STDOUT '$MyTreeMap_indels_spt_odr::$TEST_LEVEL_SCN  = ', $MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN , " .\n\n";

	## END of "IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014)." ##

	### Read in the list of segments. ###

print STDOUT "\n<<<<< Now, read in the list of segments >>>>>\n\n";

our @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##our @LABELS_SEG_ATTR = ('seg_type', 'ref_len', 'rec_len', '#len_grp', 'alt_state');

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
our %LABEL_SATTR2INDX;
for (my $l=0; $l<@LABELS_SEG_ATTR; $l++) {
	$LABEL_SATTR2INDX{$LABELS_SEG_ATTR[$l]} = $l+1;
}
=cut

# @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.
#
my $list_segments = readin_list_segs2 ($inlist_segs, @LABELS_SEG_ATTR);# MODIFIED  in creation of 'compare_ref_vs_rec_compllimment.pl'. #

# @{$list_segments[$s]} = ($seg_ID, @attrs) gives information on the $s th segment.
#
##my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
unless (defined $list_segments) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_segs2' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
##my $inlist_segs = "$intopdir/segs.txt";

my $ct_segments = @{$list_segments};

print STDOUT "\n#{segments} = $ct_segments\n\n";
print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR[1..$#LABELS_SEG_ATTR]), "\n\n"; # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
##print STDOUT join ("\t", 'Seg_ID', @LABELS_SEG_ATTR), "\n\n";

for (my $s=0; $s<$ct_segments; $s++) { print STDOUT join ("\t",  @{$list_segments->[$s]}), "\n"; }
print STDOUT "\n\n";

=set aside for later use
my $NEW_INDX_DIAG; # It points to the element of 'Diagnosis' in the segment information. #
for (my $i=0; $i<@LABELS_SEG_ATTR; $i++) { # Preliminary for-loop (over field labels). #
    ($LABELS_SEG_ATTR[$i] eq 'Diagnosis') or next;
    $NEW_INDX_DIAG = $i;
    last;
}				# End of the preliminary for-loop (over field labels). #
=cut


	### Read in the simulation parameters ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the parameters used for the simulation >>>>>\n\n";

# %{$hash_simul_params} = ($label => $value, ...) for the simulation parameters.
#
my $hash_simul_params = readin_simul_params ($inlist_params);
unless (defined $hash_simul_params) {
	print STDOUT "[ERROR '$program_name'] THe subroutine 'readin_simul_params' FAILED on the input flie '$inlist_params' !\n";
	print STDERR "[ERROR $program_name'] $!\n\n";
	die ;
}
##my $inlist_params = "$intopdir/dawg_inp.txt";

print STDOUT "\n    ... Finished the subroutine 'readin_simul_params' ...\n\n";

=for TESTING the subroutine 'readin_simul_params'
print STDOUT join ("\t", 'Label', 'Value'), "\n\n";
foreach my $label (sort keys %{$hash_simul_params}) { print STDOUT join ("\t", $label, $hash_simul_params->{$label}), "\n"; }
print STDOUT "\n\n";
=cut

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
my $indel_model = $hash_simul_params->{'GapModel'};
##my $indel_model = $hash_simul_params->{'Indel.Model'};
unless ((defined $indel_model) and ($INDEL_MODEL2DISTRI{$indel_model})) {
	(defined $indel_model) or ($indel_model = 'undef');
	print STDOUT "[ERROR '$program_name'] The indel model ($indel_model) does NOT CORRESPOND to any distribution defined in \%INDEL_MODEL2DISTRI !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
my $rel_indel_rate = $hash_simul_params->{'Lambda'};
my $indel_params = $hash_simul_params->{'GapParams'};
##my $rel_indel_rate = $hash_simul_params->{'Indel.Rate'};
##my $indel_params = $hash_simul_params->{'Indel.Params'};
unless ((defined $rel_indel_rate) and (defined $indel_params)) { 
	(defined $rel_indel_rate) or ($rel_indel_rate = 'undef');
	(defined $indel_params) or ($indel_params = 'undef');
	print STDOUT "[ERROR '$program_name'] Some parameters fetched via 'readin_simul_params' are UNDEFINED: (\$rel_indel_rate, \$indel_params) = ($rel_indel_rate, $indel_params) !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my ($exponent, $cutoff_len) = split /\,/, $indel_params;
unless (defined $cutoff_len) {
	print STDOUT "[ERROR '$program_name'] ILLEGAL OUTPUT of 'readin_simul_params': \$indel_params = $indel_params: !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print STDOUT "{Indel_model} = $indel_model .\n";
print STDOUT "{Relative_indel_rate} = $rel_indel_rate .\n";
print STDOUT "{Exponent for the power-law}  = $exponent .\n";
print STDOUT "{Cutoff length} = $cutoff_len .\n\n";

	### Set Parameters for the log-likelihoods of indel scenarios. ###

our $TOTRATE_INS = our $TOTRATE_DEL = $rel_indel_rate/2;
our $DISTRI = $INDEL_MODEL2DISTRI{$indel_model};
our $PARAM_INS = our $PARAM_DEL = $exponent;


	# IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'
	# and used for MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'.

my $gaplen_cutoff = $cutoff_len; # ADDED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #

	# Cut-off insertion length.
our $CUTOFF_INS_LEN = $gaplen_cutoff ;
	# Cut-off deletion length.
our $CUTOFF_DEL_LEN = $gaplen_cutoff ;

=for the predecessor 'compare_ref_vs_rec_indel_cts.pl'
our $CUTOFF_3P = our $CUTOFF_5P = $cutoff_len;
=cut

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL);

print STDOUT "\nConstructed Input Parameters for Log-likelihoods of Gap Patterns:\n\n";

print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.

	## IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl',
	## and used for MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'.

print STDOUT "\n<<< Now, as another preliminary process, read in the multiplication factor '\$delta_del' for a 'universal' term of log-lilekihood ... >>>\n\n";

my $delta_del = readin_delta_del ($infile_delta_del);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_delta_del' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}

print STDOUT '$delta_del = ', $delta_del, " .\n\n";

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'."

=for the predecessor 'compare_ref_vs_rec_indel_cts.pl'
print STDOUT "\n<<< Now, as another preliminary process, calculate the normalization factors for the indel distributions >>>\n\n";

my $log_nfacs = cal_log_normal_facs_indel_distri (@PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P);
unless (defined $log_nfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'cal_log_normal_facs_indel_distri' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}

our @LABELS_NFACS = ("Insertions(5p)", "Deletions(5p)", "Insertions(3p)", "Deletions(3p)");

print STDOUT "\n<< Outputs of 'cal_log_normal_facs_indel_distri' >>\n\n";
print STDOUT join ("\t", 'Category', 'log(norm_fac)', 'norm_fac'), "\n";

for (my $i=0; $i<4; $i++) { # Preliminary for-loop #
    my $label = $LABELS_NFACS[$i];
    my $log_nfac = $log_nfacs->[$i];
    my $nfac = (defined $log_nfac) ? exp ($log_nfac) : 'undef';
    (defined $log_nfac) or ($log_nfac = 'undef');
    print STDOUT join ("\t", $label, $log_nfac, $nfac), "\n";
}		# End of the preliminary for-loop #

print STDOUT "\n\n";

my $log_nfac_del_5p = $log_nfacs->[1];
print STDOUT "\n<<< Now, calculate the multiplicative factor '\$delta_del_5p' for a common term of log-likelihood (with \$log_nfac_del_5p = $log_nfac_del_5p). >>>\n\n";

my $delta_del_5p = cal_delta_exp_del_5p (@PARAMS_DEL, $CUTOFF_5P, $log_nfac_del_5p); 
unless (defined $delta_del_5p) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'cal_delta_exp_del_5p' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT "\n<< Output of 'cal_delta_exp_del_5p' >>\n\n";
print STDOUT '$delta_del_5p = ', $delta_del_5p, " .\n\n";
=cut


	### Read in the reference MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reference MSA >>>>>\n\n";

# $inaln_file_ref = "$intopdir/dawg.fasta";

(my $ref_msa, my $sequence_ids, undef) = extract_alignment ($inaln_file_ref); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $ref_msa, my $sequence_ids, undef, undef) = extract_fasta ($inaln_file_ref);
unless (defined $ref_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reference MSA '$inaln_file_ref' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_sequences = @{$sequence_ids};
unless ($ct_sequences > 1) {
	print STDOUT "[ERROR '$program_name'] The number of sequences ($ct_sequences) is TOO SMALL ! I will stop here ...\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_refaln = length ${$ref_msa->[0]};

print STDOUT "#{sequences in MSA} =  $ct_sequences .\n";
print STDOUT "#{columns in MSA}   = $totlen_refaln .\n\n";

	### Read in the reconstructed MSA ###
	### (ADDED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the reconstructed MSA >>>>>\n\n";

# $inaln_file_rec = "$intopdir/mafft.fasta";

(my $rec_msa, my $seq_ids2, undef) = extract_alignment ($inaln_file_rec); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
##(my $rec_msa, my $seq_ids2, undef, undef) = extract_fasta ($inaln_file_rec);
unless (defined $rec_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the reconstructed MSA '$inaln_file_rec' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


unless ($ct_sequences == @{$seq_ids2}) {
	print STDOUT "[ERROR '$program_name'] THe number of sequences in the core alignment (", scalar (@{$seq_ids2}), ") CLASHES with that of the reference alignment ($ct_sequences) !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
for (my $k=0; $k<$ct_sequences; $k++) {
	my $seqid = $sequence_ids->[$k];
	my $seqid2 = $seq_ids2->[$k];
	unless ($seqid eq $seqid2) {
	    print STDOUT "[ERROR '$program_name'] The ID of the $k th sequence in the reconstructed MSA ($seqid2) does NOT MATCH that in the reference MSA ($seqid) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
}

my $totlen_recaln = length ${$rec_msa->[0]};

print STDOUT "#{columns in MSA} = $totlen_recaln .\n\n"; 


	### Extract a list of sequence names (IDs), as well as concatenated anchor MSAs, from the core alignment ###
	### (MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring.PhyML.pl' on Aug 26 (Sun), 2012) ###

print STDOUT "\n<<<<< Now, read in the core alignment (to be fed into tree reconstruction, if necessary) >>>>>\n\n";

(my $anchor_msa, my $sequence_ids_core, undef, undef) = extract_alignment ($anchor_msa_file); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). # 
##(my $anchor_msa, my $sequence_ids_core, undef, undef) = extract_fasta ($anchor_msa_file);
unless (defined $sequence_ids_core) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the core alignment '$anchor_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
unless ($ct_sequences == @{$sequence_ids_core}) {
	print STDOUT "[ERROR '$program_name'] THe number of sequences in the core alignment (", scalar (@{$sequence_ids_core}), ") CLASHES with that of the reference alignment ($ct_sequences) !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
for (my $k=0; $k<$ct_sequences; $k++) {
	my $seqid = $sequence_ids->[$k];
	my $seqid2 = $sequence_ids_core->[$k];
	unless ($seqid eq $seqid2) {
	    print STDOUT "[ERROR '$program_name'] The ID of the $k th sequence in the core alignment ($seqid2) does NOT MATCH that in the reference MSA ($seqid) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
}

my $ancaln_len = length ${$anchor_msa->[0]};

print STDOUT "#{columns in MSA} = $ancaln_len .\n\n"; 


	### Give nominal frequencies to the residues. ###

print STDOUT "\n<<<<< Now, give nominal frequencies to the residues >>>>>\n\n";

our %RES2FREQ; # = ($residue => $equil_freq, ...)
our @RESIDUES = ($seqtype eq 'NT') ? qw( A C G T ) :
	@MyPAML_parser::AMINO_ACIDS ;
our @RES_FREQS = ();

if ($resfreq =~ m/\S/) {
  if ($resfreq eq 'm') {
    if ($seqtype eq 'AA') {

      unless ( retrieve_aa_freqs ($src_dat, %RES2FREQ) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'retrieve_aa_freq' FAILED on the input data file '$src_dat' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      print STDOUT "\n --- Retrieved the Amino-acid frequencies from the substitution model ...\n\n";

    } else {
	# Readin the ML estimates of nucleotide frequencies

      print STDOUT "\n --- The nucleotide frequencies will be estimated via ML later ...\n\n";
    }

  } elsif ($resfreq eq 'e') {

	# Calculate the residue frequencies in the core alignment #

      my $fl_caseignore = 1; # = 1 to convert lower-case letters to upper-case letters, = 0 to keep lower-case letters as they are.
      my $fl_relative = 1; # = 1 to compute the relative frequencies, = 0 to compute the absolute frequencies (i.e. the counts), of the residues.
	
      my $totct_res = est_resfreqs_frm_msa (@{$anchor_msa}, $GAP, %RES2FREQ, $fl_caseignore, $fl_relative);
      unless (defined $totct_res) {
	print STDOUT "[ERROR '$program_name'] 'est_resfreqs_frm_msa' FAILED on the anchor MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      unless ($totct_res > 0) {
	print STDOUT "[ERROR '$program_name'] NO RESIDUES other than gaps ($GAP) in the anchor MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      print STDERR "\n --- estimated the residue frequencies out of the anchor MSA (total residue number = $totct_res) ...\n\n";

  } elsif ($seqtype eq 'NT') { # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 8 (Fri), 2011 #

      my @freqs = split /\,/, $resfreq;
      unless (@freqs==4) {
	print STDOUT "[ERROR '$program_name'] The number (", scalar (@freqs), ") of user-specified frequencies = (", join (', ', @freqs), ") CLASHES with 4, which is the number of residues (", join (', ', @RESIDUES), ") !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
      }
      for (my $r=0; $r<4; $r++) { $RES2FREQ{$RESIDUES[$r]} = $freqs[$r]; }
  }

} elsif (($sbst_model eq 'JC69') or ($sbst_model eq 'K80')) {
      for (my $r=0; $r<4; $r++) { $RES2FREQ{$RESIDUES[$r]} = 1/4; }
} else {
	print STDOUT "[ERROR '$program_name'] I do NOT SUPPORT the option \$resfreq = '$resfreq' for the substitution model '$sbst_model' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

	## Normalize the frequencies. ##

unless (($seqtype eq 'NT') and ($resfreq eq 'm')) {

    my $tot_resfreq = 0; # ADDED on Mar 29 (Tue), 2011 #
    foreach my $res (@RESIDUES) { # The 1st foreach-loop over residues #
	my $freq = $RES2FREQ{$res};
	unless (defined $freq) {
	   if ($resfreq eq 'm') {
	 	print STDERR "[ERROR '$program_name'] UNDEFINED FREQUENCY for the residue '$res' !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    } else {
		print STDOUT "[WARNING '$program_name'] The frequency for the residue '$res' is UNDEFINED : $! !\n";
		print STDOUT "        --- I will set it to $MIN_FREQ ...\n\n";
		##$freq = $RES2FREQ{$res} = 0;
		$freq = $RES2FREQ{$res} = $MIN_FREQ;
	    }
	}
	$tot_resfreq += $freq; # ADDED on Mar 29 (Tue), 2011 #
    }			# End of the 1st foreach-loop over residues #
    
		# Normalize the frequency (ADDED on Mar 29 (Tue), 2011) #

    foreach my $res (@RESIDUES) { # The 2nd foreach-loop over residues #
 	my $freq = $RES2FREQ{$res} = sprintf ("%9.7f", $RES2FREQ{$res} /$tot_resfreq );
	push @RES_FREQS, $freq;
    }			# End of the foreach-loop over residues #

    if ($resfreq eq 'm') {
	print STDOUT "\n\n<< Residue frequencies defined in the model >>\n\n";
    } else {
	print STDOUT "\n\n<< Residue frequencies estimated from the core alignment >>\n\n";
    }
    print STDOUT join ("\t", 'Residue', 'Frequency'), "\n\n";

    foreach my $res (@RESIDUES) { print STDOUT join ("\t", $res, $RES2FREQ{$res}), "\n"; }
    print STDOUT "\n\n";
}


	### Make a master file for a custom amino-acid substitution model ###

if (($seqtype eq 'AA') and ($resfreq eq 'e')) {

    print STDOUT "\n<<<<< Now, make master files for a custom amino-acid substitution model >>>>>\n\n";

    my $resct_in_line = 7; # Maximum number of residues whose frequencies are recorded in a line.
    my $endmark = ''; # A string that marks the end of the substitution model in the file #
    unless ( print_custom_sbst_model ($src_dat, $master_sbst_file, @RES_FREQS, $resct_in_line, $endmark) ) {
	print STDOUT "[ERROR '$program_name'] 'print_custom_sbst_model' FAILED to print out the pre-substitution matrix '$src_dat' combined with the empirical amino-acid frequencies:\n";
	print STDOUT join (', ', @RES_FREQS), " !!\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    my $fh_otmp = open_output_file ($master_ctrl_file, 0);
    unless (defined $fh_otmp) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the master control file '$master_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print $fh_otmp $dat_name, "\n";
    close ($fh_otmp);

    foreach my $master_file ($master_sbst_file, $master_ctrl_file) {
	unless (( -f $master_file ) and ( -r $master_file )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a READABLE master file '$master_file' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
    }

    print STDOUT "(Master file of the model) $master_sbst_file ,\n";
    print STDOUT "(Master control file)      $master_ctrl_file .\n\n"; 
}


	### Either retrieve or reconstruct a phylogenetic tree that our main process is based on. ###

my @std_set_cand_trees = ();
our %STATS_PHYML;
	# = ($key => scalar_value ($key is the string between '.' and ':', which ' ' replaced with '_'),
	# 'discrete_gamma' => [$ncate, $alpha] (or undef),
	# 'nuc_freqs' => [$fA, $fC, $fG, $fT] (or undef),
	# 'rel_rates' => [$r_AC, $r_AG, $r_AT, $r_CG, $r_CT, $r_GT] (or undef),
	# 'rate_mtrx' => \@mtrx, with $mtrx[$i]->[$j] giving the instantaneous rate from the $i-th to the $j-th nucleotide (in the order: A, C, G, T) (or undef)
	# );
our @REL_RATES = ();

if ((defined $intree) and ($intree =~ m/\S/)) {

	## Read in the input tree (ESSENTIAL for the 'beta' version) ##

    print STDOUT "\n<<<<< Reading-in the Input Tree >>>>>\n\n";

    # @{$info_dawg_tree} = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr)
    #
    my $info_dawg_tree = readin_dawg_tree2 ($intree); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). #
    ##my $info_dawg_tree = readin_dawg_tree ($intree);
    unless (defined $info_dawg_tree) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'readin_dawg_tree2' FAILED on the input tree file '$intree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @std_set_cand_trees = (['NA', 'NA', 'NA', 0, 0, @{$info_dawg_tree}]);

} else {

	## Infer a phylogenetic tree from the core alignment ##

		# Prepare an input alignment #

    ## my $tmpdir_phyml = "$outtopdir/Infer_tree";
    my $inaln_for_tree_name = 'in_phyml.seq';
    my $inaln_for_tree = "$tmpdir_phyml/$inaln_for_tree_name";

    my $fh_tmpaln = open_output_file ($inaln_for_tree, 0);
    unless (defined $fh_tmpaln) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the input alignment file for the tree inference '$inaln_for_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print $fh_tmpaln "$ct_sequences $ancaln_len\n";

    unless ( pr_phylip_aln (@{$sequence_ids}, @{$anchor_msa}, $DIS_LEN, *${$fh_tmpaln}) ) {
	print STDOUT "[ERROR '$program_name'] 'pr_phylip_aln' FAILED to output the input alignment for the tree inference into the file '$inaln_for_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    close ($fh_tmpaln);

		# Put together command-line options #

    ## my $sbst_model = ($seqtype eq 'AA') ? $SBST_MODEL_AA : $SBST_MODEL_NT;
    ## my $ct_rcate = $CT_RCATE;
    ## my $alpha = $ALPHA;
    ## my $topo_search = $TOPO_SEARCH;

    my $ct_bootstrap = 0; # 0: Neither approximate likelihood ratio test nor bootstrap values are computed.
			 # -1: Approximate likelihood ratio test returning aLRT statistics.
    my $opt_params = 'tlr'; # Tree topology (t), branch lengths (l), and substitution rate parameters (r) are optimized.


    my $options = " -i $inaln_for_tree_name $OPTIONS_SHARED -b $ct_bootstrap -s $topo_search -o $opt_params";
    (($ct_rcate > 1) and (defined $alpha) and ($alpha > 0)) and ($options .= " -a $alpha");

    (($seqtype eq 'NT') and (defined $sv_ratio)) and ($options .= " -t $sv_ratio");

		# Paths to the output/log files # 

    my $name_outtree = "${inaln_for_tree_name}_phyml_tree.txt";
    my $out_tree = "$tmpdir_phyml/$name_outtree";

	# A BUG Fixed on Mar 30 (Wed), 2011 #
    my $name_outstat = "${inaln_for_tree_name}_phyml_stats.txt"; # ADDED on Mar 29 (Tue), 2011 #
    ##my $name_outstat = "${inaln_for_tree_name}_phyml_stat.txt"; # ADDED on Mar 29 (Tue), 2011 #

    my $out_stat = "$tmpdir_phyml/$name_outstat"; # ADDED on Mar 29 (Tue), 2011 #
    my $name_log_tree = "log_phyml.txt";
    my $log_tree = "$tmpdir_phyml/$name_log_tree";

    if (($seqtype eq 'AA') and ($resfreq eq  'e')) {

	my $tmp_sbst_file = "$tmpdir_phyml/$dat_name";
	my $tmp_ctrl_file = "$tmpdir_phyml/$ctrl_name";

	unless (0 == system ("cp -p $master_sbst_file $tmp_sbst_file") ) {
	    print STDOUT "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master file for the custom substitution model '$master_sbst_file' to the temporary directory '$tmpdir_phyml' : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	unless (0 == system ("cp -p $master_ctrl_file $tmp_ctrl_file") ) {
	    print STDOUT "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master control file for PhyML '$master_ctrl_file' to the temporary directory '$tmpdir_phyml' : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	
	$options .= " -m custom -f m < $ctrl_name";

    } else {
		# This may have to be revised for $seqtype eq 'NT' #

	    # REVISED on Apr 13 (Wed), 2011 #
	$options .= " -m $sbst_model";
	($resfreq =~ m/\S/) and ($options .= " -f $resfreq");
    }

    print STDOUT "\nOptions for PhyML tree inference:\n";
    print STDOUT "   '$options'\n\n";

   ### exit 0; ### For DEBUGGING the script ###

		# Execute the PhyML program #

    print STDOUT "\n<<< Start executing the PhyML program to infer the phylogenetic tree >>>\n\n";

    unless ( chdir $tmpdir_phyml ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$tmpdir_phyml' to infer the phylogenetic tree via PhyML !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless (0 == system ("$main_bin_tree $options > $name_log_tree 2>&1") ) {
	print STDOUT "[ERROR '$program_name'] FAILED 'system ('$main_bin_tree ...')' FAILED to infer the phylogenetic tree out of the core alignment '$inaln_for_tree': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to MOVE BACK to the original working directory '$CURR_WORK_DIR' after inferring the phylogenetic tree out of the core alignment: $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

		# Read in the inferred tree #

    print STDOUT "\n ... Now that PhyML finished, read in the resulting output files...\n\n";

    unless (( -f $out_tree ) and ( -r $out_tree )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable OUTPUT TREE FILE '$out_tree' of PhyML for the tree inferred from the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

	# Hashes to store tree information (see 'MyTreeManipu_II.pm' for their formats) #
    my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;

    unless ( extract_sgltree_from_file ($out_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, $COMMENT) ) {
	print STDOUT "[ERROR '$program_name'] 'extract_sgltree_from_file' FAILED to extract a tree from the file '$out_tree' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @std_set_cand_trees = (['NA', 'NA', 'NA', 0, 0, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr]);

		# Read in statistics file

    unless (( -f $out_stat ) and ( -r $out_stat )) {
	print STDERR "[ERROR '$program_name'] CanNOT FIND a readable OUTPUT STATISTICS FILE '$out_stat' of PhyML for the ML tree inference from the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

	# REVISED on Apr 8 (Fri), 2011 #
	# => BUG fixed on Aug 26 (Sun), 2012. #
    unless (  extract_phyml_statistics2 ($out_stat, %STATS_PHYML, $DBG_LEVEL_EXPHYML) ) {
    ##unless (  extract_phyml_statistics2 ($out_stat, %STATS_PHYML) ) {
    ##unless (  extract_phyml_statistics ($out_stat, %STATS_PHYML) ) {
	print STDOUT "[ERROR '$program_name'] 'extract_phyml_statistics2' FAILED on the PhyML statistics file '$out_stat' concerning the core alignment !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}


	## If necessary, store the Gamma shape parameter for use in later analyses ##

if (($ct_rcate > 1) and !(defined $alpha)) {

    print STDOUT "\n\n<< Estimate of Gamma shape parameter for substitution rate variation >>\n\n";
    my $para_gamma = $STATS_PHYML{'discrete_gamma'};
    unless (defined $para_gamma) {
	print STDOUT "[ERROR '$program_name'] UNDEFINED STATISTICS for discrete gamma model !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    my $ref = ref $para_gamma;
    unless ((defined $ref) and ($ref eq 'ARRAY')) {
	print STDOUT "[ERROR '$program_name'] UNEXPECTED VALUE for the key 'discrete_gamma' of hash ", '%STATS_PHYML', " (= $para_gamma) !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless (@{$para_gamma} == 2) {
	print STDOUT "[ERROR '$program_name'] The UNEXMECTED NUMBER (", scalar (@{$para_gamma}), ") of output statistics parameters for discrete gamma model, which was supposed to be 2 !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    my ($ct_rcate1, $alpha1) = @{$para_gamma};
    unless ($ct_rcate1 == $ct_rcate) {
	print STDOUT "[ERROR '$program_name'] The OUTPUT number of discrete categoryes ($ct_rcate1) CLASHES with the INPUT NUMBER ($ct_rcate) !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    print STDOUT "Alpha = $alpha1\n\n";

    $alpha = $alpha1;
}

	## Store the ML estimates of nucleotide frequencies for use in later analyses ##

if (($seqtype eq 'NT') and ($resfreq eq 'm')) {

    print STDOUT "\n\n<< ML estimates of Residue frequencies >>\n\n";

    my $nfreqs = $STATS_PHYML{'nuc_freqs'};
    unless (defined $nfreqs) {
	print STDOUT "[ERROR '$program_name'] UNDEFINED ML FREQUENCIES of nucleotides !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    unless (@{$nfreqs} == 4) {
	print STDOUT "[ERROR '$program_name'] The number (", scalar (@{$nfreqs}), ") of ML frequencies (", join (', ', @{$nfreqs}), ") is NOT the EXPECTED value, which is 4 !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    @RES_FREQS = @{$nfreqs};
    for (my $r=0; $r<4; $r++) { $RES2FREQ{$RESIDUES[$r]} = $nfreqs->[$r]; }

    print STDERR join ("\t", 'Residue', 'Frequency'), "\n\n";

    foreach my $res (@RESIDUES) { print STDOUT join ("\t", $res, $RES2FREQ{$res}), "\n"; }
    print STDOUT "\n\n";
}

	### If necessary, store the Transition/transversion ratio for use in later analysis ###

if (($seqtype eq 'NT') and !(defined $sv_ratio) and ($sbst_model ne 'JC69') and ($sbst_model ne 'F81')) {

    print STDOUT "\n\n<< ML estimate of the transition/transversion ratio >>\n\n";

    my $sv_ratio0 = $STATS_PHYML{'Transition/transversion_ratio'};
    unless (defined $sv_ratio0) {
	print STDOUT "[ERROR '$program_name'] UNDEFINED ML TRANSISION/TRANSVERSION RATIO !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    ($sv_ratio) = ($sv_ratio0 =~ m/^(\S+)/);

    print STDOUT "{ML estimate of the transition/transversion ratio} = $sv_ratio .\n\n";
}

	## Output the input tree to an interim file ##

print STDOUT "\n<<<<< Now, create an interim file recording the input trees >>>>>\n\n";
##print STDOUT "\n<<<<< Now, create an interim file recording the input tree >>>>>\n\n";

my $ct_cand_trees = @std_set_cand_trees;

my $fh_otree0  = open_output_file ($interim_trees, 0);
unless (defined $fh_otree0) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'open_output_file' FAILED on the interim tree file '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = $ct_cand_trees) >>\n\n";
##print $fh_otree0 "<< Candidate trees that are considerably likely (#{candidates} = 1) >>\n\n";

for (my $t=0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

	# REVISED in creation of 'compare_ml_msa.bch.extJC.6t.PAML.pl' (on Mar 22 (Tue), 2011) #
	# => Further REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	my ($ln_ml_msa, $node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[$t]}[1,5..9];
	##my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$std_set_cand_trees[0]}[5..9];
	####my ($node2pa0, $node2ch0, $name2id0, $id2name0, $br_attr0) = @{$info_dawg_tree};
	my $topnode_id0 = $name2id0->{$TOP_NODE};

	print $fh_otree0 "< The $t th candidate >\n\n";
	##print $fh_otree0 "< The 0 th candidate >\n\n";
	print $fh_otree0 "{ Phylogenetic Tree (top node ID: $topnode_id0) }\n\n";

	unless ( output_tree ($topnode_id0, %{$node2ch0}, %{$id2name0}, %{$br_attr0}, *${$fh_otree0}, 0, 0) ) {
	    print STDOUT "[ERROR '$program_name'] Subroutine 'output_tree' FAILED to output the dawg guide-tree onto the interim tree file '$interim_trees' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		# FIXED on Mar 24 (Thu), 2011 #
		# => REVISED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'.
	print $fh_otree0 "\n{ Log-likelihood of the tree for the MSA: $ln_ml_msa }\n\n\n";
	##print $fh_otree0 "\n{ Log-likelihood of the MSA under the tree: 0 }\n\n";

}	# Preliminary for-loop (over candidate trees). # ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014). #

close ($fh_otree0);


	## Calculate the Akaike weights of the candidate trees

print STDOUT "\n<<<<< Now, calculate the Akaike weights of the candidate trees and necessary inputs  >>>>>\n\n";

# $ln_akaike_wgts->[$c] is the logarithmic Akaike weight factor of the $c th candidate tree.
#
my $ln_akaike_wgts = log_akaike_weights (@std_set_cand_trees, 3);
unless (defined $ln_akaike_wgts) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'log_akaike_weights' FAILED on the tree list extracted from '$interim_trees' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=for TESTING the script (to see if the final results correctly take into account the effects of the Akaike weights ) (probably, it will NOT influence the comparison of the scores among alternatives)
for (my $t=0; $t<$ct_cand_trees; $t++) { $ln_akaike_wgts->[$t]++; }
=cut

##=for TESTING the script
for (my $t = 0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop #

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};
	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	print STDOUT "\< The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) >\n\n";
	print STDOUT "Sum (-log(Pv)) = $sum_neg_log_pv\n";
	print STDOUT "log (Prob(MSA|Tree)_ML) = $ln_ml_msa\n";
	print STDOUT "#{free parameters} = $ct_free_params\n";
	print STDOUT "AIC = $aic\n";
	print STDOUT "BIC = $bic\n\n";

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *STDOUT, $fl_bootlabel, $fl_nodeid) ) {
	    print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n";

	##print STDERR "\n\nNode_ID --> Parent_ID\n\n";
	##foreach my $nodeid (sort keys %{$node2pa}) { print STDERR "$nodeid  --> ", $node2pa->{$nodeid}, "\n"; }
	##print STDERR "\n\n";

	print STDOUT join ("\t", 'Branch_ID', 'length', 'P-value', 'parent_node_ID'), "\n\n";
	foreach my $brid (sort keys %{$br_attr}) { # inner foreach-loop #
	    my $attr = $br_attr->{$brid};
	    my ($brlen, $pvalue) = @{$attr}[$INDX_LENGTH, $INDX_BOOT];
	    (defined $brlen) or ($brlen = 'NA');
	    (defined $pvalue) or ($pvalue = 'NA');
	    my $pa = $node2pa->{$brid};
	    print STDOUT join ("\t", $brid, $brlen, $pvalue, $pa), "\n";
	}				# End of the inner foreach-loop #
	print STDOUT "\n\n";

}			# end of the preliminary for-loop #
print STDOUT "\n";

### exit 0; ### For TESTING the Upper part of the script
##=cut

	### Construct a shared part of the command-line options
	### for calculating column-wise likelihoods

print STDOUT "\n<<<<<  Now, specify the shared part of the command-line options. >>>>>\n\n";

my $ct_bootstrap_clmlk = 0; # Neither approximate likelihood ratio test nor bootstrap values are computed.
my $opt_params_clmlk = 'n'; # no parameter is optimized.

	# A BUG fixed on Apr 11 (Mon), 2011 #
our $OPTIONS_CLMLK = "--print_site_lnl $OPTIONS_SHARED -o $opt_params_clmlk -b $ct_bootstrap_clmlk";
##our $OPTIONS_CLMLK = "---print_site_lnl $OPTIONS_SHARED -o $opt_params_clmlk -b $ct_bootstrap_clmlk";

(($ct_rcate > 1) and (defined $alpha) and ($alpha > 0)) and ($OPTIONS_CLMLK .= " -a $alpha");

if ($seqtype eq 'AA') { # For an amino-acid alignment #
    if ($resfreq eq 'e') {
	$OPTIONS_CLMLK .= " -m custom -f m";

	# NOTE: you will have to feed the name of a file for the substitution model through a control file at 'Step 2' of the main process #

    } else {
	$OPTIONS_CLMLK .= " -m $sbst_model -f m";
    }
} else { # For a nucleotide alignment #

    my $cnct_basefreqs = join (',', @RES_FREQS); # = join (',', $fA , $fC, $fG, $fT) for nucleotides
    $OPTIONS_CLMLK .= " -m $sbst_model";
    (($sbst_model ne 'J69') and ($sbst_model ne 'K80')) and ($OPTIONS_CLMLK .= " -f $cnct_basefreqs");

    (defined $sv_ratio) and ($OPTIONS_CLMLK .= " -t $sv_ratio");

	# NOTE: you will have to feed the rate parameters through a control file at 'Step 2' of the main process.
	#  (==> THis is NOT NECESSARY any longer.) #
}

print STDOUT "<< Shared options for calculating column-wise likelihoods >>\n\n";
print STDOUT "  '$OPTIONS_CLMLK' .\n\n";


### exit 0; ### For TESTING the part added for the script 'compare_ml_msa.bch.extJC.6t.PhyML.pl'

	## Create a copy of each candidate tree, which will be used in the main indel analyses.
	## (ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014).) ##

my @std_set_cand_trees_indels = (); # element = [\%cp_node2pa, \%cp_node2ch, \%cp_name2id, \%cp_id2name, \%cp_br_attr]

		## IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl' in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. ##
my @set_node2depth = (); # %{$set_node2depth[$t]} = ($node_id => $depth, ...) for all nodes in the $t th candidate tree.
my @set_node2ct_exoffs = (); # %{$set_node2ct_exoffs[$t]} = ($node_id => #{exteior offsprings}, ...) for all nodes in the $t th candidate tree.
my @set_node2seq_indx = (); # %{$set_node2seq_indx[$t]} = ($node_id => {index in @{$sequence_ids}}, ...) for all exterior nodes for the $t th candidate tree. # (ADDED on Oct 28 (Tue), 2014.) #
		## END of "IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl' in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'." ##

		# MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
print STDOUT "<<<<< Now, prepare candidate trees for indel analyses, and assign the depth and the #{external offsprings} to each node of each tree >>>>>\n\n";

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (0a)
print STDOUT "\n<<<<< Now, assign the logarithmic multiplication factors to each branche, based on the files in '$indir_mfacs' ... >>>>>\n\n"; # ADDED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #

my @set_br2log_mfacs = (); # %{$set_br2log_mfacs[$t]} = ($br_id => \@set_log_mfacs, ...) for all branches.
=cut

for (my $t = 0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop (over candidate trees). #

	my ($node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]}[5..9];

	my %cp_node2pa;
	foreach my $node (keys %{$node2pa}) { $cp_node2pa{$node} = $node2pa->{$node}; }

 	my %cp_node2ch;
	foreach my $node (keys %{$node2ch}) { my @cp_ch = @{$node2ch->{$node}}; $cp_node2ch{$node} = \@cp_ch; }

	my %cp_name2id;
	foreach my $name (keys %{$name2id}) { $cp_name2id{$name} = $name2id->{$name}; }

	my %cp_id2name;
	foreach my $id (keys %{$id2name}) { $cp_id2name{$id} = $id2name->{$id}; }

	my %cp_br_attr;
	foreach my $br (keys %{$br_attr}) { my @attr = @{$br_attr->{$br}}; $cp_br_attr{$br} = \@attr; }

	$std_set_cand_trees_indels[$t] = [\%cp_node2pa, \%cp_node2ch, \%cp_name2id, \%cp_id2name, \%cp_br_attr];

	
		## IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl',
		## in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.

		# Assign depths to the nodes. #

	my $topnode_id = $name2id->{$TOP_NODE};

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2abs_depth' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2abs_depth' Ended ...\n\n";

   if ($TEST_LEVEL2>2) {
	print STDOUT join ("\t", 'Node_ID', 'Depth'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}), "\n"; }
	print STDOUT "\n\n\n";
   }

	$set_node2depth[$t] = $node2depth;


		# Assign #{exterior offsprings} to the nodes. #

	my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
	unless (defined $node2ct_exoffs) {
	    print STDOUT "[ERROR '$program_name'] 'make_node2ct_exoffs' FAILED on the $t th candidate tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n ... The subroutine 'make_node2ct_exoffs' Ended ...\n\n";

	$set_node2ct_exoffs[$t] = $node2ct_exoffs;


   if ($TEST_LEVEL2>2) {
	print STDOUT join ("\t", 'Node_ID', 'Depth', '#{exteior_offsprings}'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}, $node2ct_exoffs->{$node}), "\n"; }
	print STDOUT "\n\n\n";
   }


		# Assign indices (for @{$sequence_ids}) to the exterior nodes. #

   (print STDOUT join ("\t", 'Seq_indx', 'Seqname', 'Node_ID') , "\n\n") if ($TEST_LEVEL2>2);

	my %node2seq_indx;

	for (my $k=0; $k<$ct_sequences; $k++) { # 5th inner for-loop (over aligned sequences). #

	    my $seqname = $sequence_ids->[$k];
	    my $nodeid = $name2id->{$seqname};
	    unless (defined $nodeid) {
		print STDOUT "[ERROR '$program_name'] The $k th sequence (name: $seqname) has NO CORRESPONDING NODE in the $t th candidate tree !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    $node2seq_indx{$nodeid} = $k;

   (print STDOUT join ("\t", $k, $seqname, $nodeid), "\n") if ($TEST_LEVEL2>2);

	}			# End of the 5th inner for-loop (over aligned sequences). #

   (print STDOUT "\n\n\n") if ($TEST_LEVEL2>2);

	$set_node2seq_indx[$t] = \%node2seq_indx;


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (0a)

	### Read in the logarithmic multiplication factors. ###
	### (IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'
	### in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'.) ###

	my $br2log_mfacs = assign_br2log_mfacs (%{$br_attr}, $TOTRATE_INS + $TOTRATE_DEL, $indir_mfacs, @FILENAMES_LOG_MFACS);
	unless (defined $br2log_mfacs) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'assign_br2log_mfacs' FAILED !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	$set_br2log_mfacs[$t] = $br2log_mfacs; # ADDED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (0a)"

}		# End of the preliminary for-loop (over candidate trees). #

	## END of "(ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014).)" ##


	### Estimate and Compare the ML values of the Candidate MSAs ###

print STDOUT "\n<<<<<<< The main analyses >>>>>>>\n\n";


my $fh_out = open_output_file ($out_summ, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_summ' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_anom_segs = my $ct_in_segs = my $ct_succ_segs = my $ct_fail_segs = my $ct_skipped_segs = 0;
my $ct_skipped_segs2 = 0; # ADDED in DEBUGGING on Aug 11 (Mon), 2014. #

#my $list_segments = readin_list_segs ($inlist_segs, %LABEL_SATTR2INDX);
# my $ct_segments = @{$list_segments};

	# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
my $last_ref_coord_end = my $last_rec_coord_end = 0;
##my $ref_coord_beg = my $rec_coord_beg = 0; # The 'beginning' coordinates of the segment in the reference MSA and reconstructed MSA, respectively. #

for (my $s = 0; $s<$ct_segments; $s++) { # Main outer for-loop (over MSA segments) #

		## MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014). ##

	my ($segid, $ref_coord_beg, $ref_coord_end, $rec_coord_beg, $rec_coord_end, $diagnosis) = @{$list_segments->[$s]};

	    # @LABELS_SEG_ATTR = ('Indx_spacer', 'Start_in_Ref', 'End_in_ref', 'Start_in_Rec', 'End_in_Rec', 'Diagnosis'); # MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    # @{$list_segments[$s]} = @attrs gives information on the $s th segment, including its ID.

	unless ( uc ($diagnosis) =~ m/DIFFERENT$/ ) { # Skip segments NOT diagnosed as having reconstructed MSA 'DIFFERENT'  from the reference.

   (print STDOUT "\n[INFORMATION] The $s th segment (ID: $segid) will be SKIPPED, because its reference MSA (in [$ref_coord_beg\, $ref_coord_end\) ) and reconstructed MSA (in [$rec_coord_beg\, $rec_coord_end\) ) are diagnosed as '$diagnosis' !\n\n") if ($TEST_LEVEL>0);

	    $ct_skipped_segs++;

	    next;
	}

	print STDOUT "<<<<<< Segment ID: $segid  (Diagnosis: $diagnosis) >>>>>>\n\n";


		# Remove the right flanking regions, if at all. #
	if ($last_ref_coord_end < $ref_coord_beg) {
	    my $len_rflnk = $ref_coord_beg - $last_ref_coord_end;
	    for (my $k=0; $k < $ct_sequences; $k++) { substr (${$ref_msa->[$k]}, 0, $len_rflnk, ''); }
	}

	if ($last_rec_coord_end < $rec_coord_beg) {
	    my $len_rflnk = $rec_coord_beg - $last_rec_coord_end;
	    for (my $k=0; $k < $ct_sequences; $k++) { substr (${$rec_msa->[$k]}, 0, $len_rflnk, ''); }
	}

	print STDOUT "<<<< Prepare for segments of reference/reconstructed MSAs >>>>\n\n";

	my $ref_len = $ref_coord_end - $ref_coord_beg;
	my $rec_len = $rec_coord_end - $rec_coord_beg;

	my @ref_seg_msa = my @rec_seg_msa = ();

	for (my $k=0; $k<$ct_sequences; $k++) { #  1st middle for-loop #

	    my $seg_ref = substr (${$ref_msa->[$k]}, 0, $ref_len, '');
	    my $seg_rec = substr (${$rec_msa->[$k]}, 0, $rec_len, '');

	    push @ref_seg_msa, \$seg_ref;
	    push @rec_seg_msa, \$seg_rec;

	}	# 2nd of the 1st middle for-loop #

	print STDOUT "(ref_len, rec_len) = ($ref_len, $rec_len)\n\n";


		# Update the last end-coordinates. #

	$last_ref_coord_end = $ref_coord_end;
	$last_rec_coord_end = $rec_coord_end;

		## END of "MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl' (on Aug 9 (Sat), 2014)." ##

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'

	my ($segid, $seg_type, $ref_len, $rec_len, $ct_lgrps, $alt_state) = @{$list_segments->[$s]};
# our @LABELS_SEG_ATTR = ('seg_type', 'ref_len', 'rec_len', '#len_grp', 'alt_state');

	print STDOUT "<<<<<< Segment ID: $segid >>>>>>\n\n";

# (my $ref_msa, my $sequence_ids, undef, undef) = extract_fasta ($inaln_file_ref);
# (my $rec_msa, my $seq_ids2, undef, undef) = extract_fasta ($inaln_file_rec);


	print STDOUT "<<<< Prepare for segments of reference/reconstructed MSAs >>>>\n\n";

	my @ref_seg_msa = my @rec_seg_msa = ();

	for (my $k=0; $k<$ct_sequences; $k++) { #  1st middle for-loop #

	    my $seg_ref = substr (${$ref_msa->[$k]}, 0, $ref_len, '');
	    my $seg_rec = substr (${$rec_msa->[$k]}, 0, $rec_len, '');

	    push @ref_seg_msa, \$seg_ref;
	    push @rec_seg_msa, \$seg_rec;

	}	# 2nd of the 1st middle for-loop #

	print STDOUT "seg_type = $seg_type\n";
	print STDOUT "(ref_len, rec_len) = ($ref_len, $rec_len)\n\n";

	if ($seg_type == 0) {
	    $ref_coord_beg += $ref_len;
	    $rec_coord_beg += $rec_len;
	    $ct_skipped_segs++;
	    print STDOUT "\n --- Sgment with seg_type = 0 is skipped ---\n\n";
	    next;
	}
=cut

=for targetted DEBUGGING (on Sep 22 (Sat) & 23 (Sun), 2012)
	##unless ($segid eq '0040') {
	unless ($segid eq '0014') {
	    $ref_coord_beg += $ref_len;
	    $rec_coord_beg += $rec_len;
	    $ct_skipped_segs++;
	    ##print STDOUT "\n --- Sgment with ID ne '0040' is skipped ---\n\n";
	    print STDOUT "\n --- Sgment with ID ne '0014' is skipped ---\n\n";
	    next;
	}
=cut


		## Prepare for the files necessary for analyzing the segment. ##

# print STDERR "   (for scoring the sbst pattern of the \{reference,reconstructd\} MSA for each segment) {log_dir}/log_apbayes_score.\{ref,rec\}.\{seg_id\}.txt .\n\n";
# print STDERR "   (for scoring the gap-pattern of the \{reference,reconstructed\} MSA for each segment} {log_dir}/log_lnlk_gpattern.\{ref,rec\}.\{seg_id\}.txt .\n\n";

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-1)
	my $logfile_sbst_ref = "$logdir/log_apbayes_score.ref.$segid.txt";
	my $fh_log_sref = open_output_file ($logfile_sbst_ref, 0);
	unless (defined $fh_log_sref) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_sbst_ref' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	my $logfile_sbst_rec = "$logdir/log_apbayes_score.rec.$segid.txt";
	my $fh_log_srec = open_output_file ($logfile_sbst_rec, 0);
	unless (defined $fh_log_srec) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_sbst_rec' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-1)"

	my $logfile_gap_ref = "$logdir/log_lnlk_gpattern.ref.$segid.txt";
	my $fh_log_gref = open_output_file ($logfile_gap_ref, 0);
	unless (defined $fh_log_gref) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_gap_ref' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	my $logfile_gap_rec = "$logdir/log_lnlk_gpattern.rec.$segid.txt";
	my $fh_log_grec = open_output_file ($logfile_gap_rec, 0);
	unless (defined $fh_log_grec) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the log file '$logfile_gap_rec' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		### MOVED from below, within the loop over gap patterns (on Sep 1 (Sat), 2012) ###
    if ($TEST_LEVEL>2) {
	foreach my $fh_log ($fh_log_gref) { # MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
	##foreach my $fh_log ($fh_log_sref, $fh_log_gref) {...}
	    print $fh_log "Original Segment of the Reference Alignment:\n\n";
	    unless ( pr_ODEN_input (@{$sequence_ids}, @ref_seg_msa, 60, *${$fh_log}) ) {
		##print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA for the $indx_gp th gap pattern !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	}

	foreach my $fh_log ($fh_log_grec) { # MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
	##foreach my $fh_log ($fh_log_srec, $fh_log_grec) {...}
	    print $fh_log "Original Segment of the Reconstructed Alignment:\n\n";
	    unless ( pr_ODEN_input (@{$sequence_ids}, @rec_seg_msa, 60, *${$fh_log}) ) {
		##print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA for the $indx_gp th gap pattern !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reconstructed MSA !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	}
    }
		### END of "MOVED from below, within the loop over gap patterns" ###


		## ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014). ##

	unless (($ref_len>0) and ($rec_len>0)) {

	    print STDOUT "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the ComplLiMMent scores for this segment pair... \n\n";

	    foreach my $fh_log ($fh_log_gref, $fh_log_grec) { # MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
	    ##foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) { ...}
		print $fh_log "\n[INFORMATION] Either reference or reconstructed sub-MSA (or both) is NULL! ... I will skip the calculation of the complLiMMent scores for this segment pair (ID: $segid) ... \n\n\n";
	    }

	    print $fh_out "< Segment ID: $segid >\n\n";

		# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	    print $fh_out join ("\t", 'Ref/Rec', 'Total_Score',  'AvAv#{insertions}', 'AvAv#{deletions}', 'Av#{null_columns}'), "\n\n";
	    ##print $fh_out join ("\t", 'Ref/Rec', 'Total_Score', 'Gap_Score', 'Av#{indel_events}', 'Av#{null_columns}'), "\n\n";

	    print $fh_out join ("\t", 'Ref', 'N/A', 'N/A', 'N/A', 'N/A'), "\n";
	    print $fh_out join ("\t", 'Rec', 'N/A', 'N/A', 'N/A', 'N/A'), "\n";
	    
	    print $fh_out "\n\\\\\n\n";

	    $ct_skipped_segs2++;

	    next;
	}
		## END of "ADDED for DEBUGGING the script (on Aug 11 (Mon), 2014)." ##

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-3)

# print STDERR "Temporary directories for PAML:  {top_tmpdir}/s{seg_id}/g{indx_gpattern}/t{indx_tree}/,\n";

		# The grand-parent directory of temporary working directories #

	my $grpa_tmp_workdir_ref = sprintf ("%s/s%s", $top_tmpdir_ref, $segid);
	my $grpa_tmp_workdir_rec = sprintf ("%s/s%s", $top_tmpdir_rec, $segid);

	foreach my $grpa_tmpdir ($grpa_tmp_workdir_ref, $grpa_tmp_workdir_rec) { # Preliminary middle foreach-loop #
	    unless ( -d $grpa_tmpdir) {
		unless (0 == system ("mkdir -p $grpa_tmpdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to create a grand-parent of temporary working directories '$grpa_tmpdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	    unless (( -r $grpa_tmpdir ) and ( -w $grpa_tmpdir )) {
		unless (0 == system ("chmod -r u+rw $grpa_tmpdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the grand-parent of temporary working directories '$grpa_tmpdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}			# End of the preliminary middle foreach-loop #

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-3)"

			# MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #

	print STDOUT "<<<<< Step (ii) Main Process A: Calculating the inferred counts of indels that can explain the reference & reconstructed MSA segments, under the Dollo parsimony principle.   >>>>>\n\n";
	foreach my $fh_log ($fh_log_gref, $fh_log_grec) {
	    print $fh_log "<<<<< Step (ii) Main Process A: Calculating the inferred counts of indels that can explain the reference & reconstructed MSA segments, under the Dollo parsimony principle.   >>>>>\n\n";
	}

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-5)

		## (ii) Main Process A:
		## Calculate the approximate ML scores of the substitution & gap patterns
		## of the columns in reference & reconstructed MSAs
		##  under each of the candidate trees.

	print STDOUT "<<<<< Step (ii) Main Process A: Calculating apML scores of the reference & reconstructed MSA segments >>>>>\n\n";
	foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {
	    print $fh_log "<<<<< Step (ii) Main Process A: Calculating apML scores of the reference & reconstructed MSA segments >>>>>\n\n";
	}

	my @setset_refseg_scores = my @setset_recseg_scores = ();
		# $setset_{ref,rec}seg_scores[$indx_gpattern]->[$t] is the score
		# of the $indx_gpattern set in the reference/reconstructed MSA segments 
		# under the $t th tree candidate.

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-5)"

=DISABLED because they didn't appear working as they are supposed to be.
	my $int_clmlk_file_ref = "$top_clmlk_dir_ref/$pref_clmlk$segid$suff_clmlk";
	my $fh_clmlk_ref = open_output_file ($int_clmlk_file_ref, 0);
	unless (defined $fh_clmlk_ref) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the interim output file '$int_clmlk_file_ref' for the column-wise log-likelihood of the reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $int_clmlk_file_rec = "$top_clmlk_dir_rec/$pref_clmlk$segid$suff_clmlk";
	my $fh_clmlk_rec = open_output_file ($int_clmlk_file_rec, 0);
	unless (defined $fh_clmlk_rec) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the interim output file '$int_clmlk_file_rec' for the column-wise log-likelihood of the reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
=cut


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-6)
	my @set_refseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reference segment under the $t th tree candidate.
	my @set_recseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reconstructed segment under the $t th tree candidate.
	my @set_refseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reference segment under the $t th tree candidate.
	my @set_recseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reconstructed segment under the $t th tree candidate.
=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-6)"


		# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	my @set_avct_ins_ref = my @set_avct_ins_rec = (); # $set_avct_ins_{ref,rec}[$t] is the average number of insertion events under the $t th tree candidate.
	my @set_avct_dels_ref = my @set_avct_dels_rec = (); # $set_avct_dels_{ref,rec}[$t] is the average number of deletion events under the $t th tree candidate.
	##my @set_ct_indels_ref = my @set_ct_indels_rec = (); # $set_ct_indels_{ref,rec}[$t] is the average number of indel events under the $t th tree candidate.

	my @set_ct_null_clms_ref = my @set_ct_null_clms_rec = (); # $set_ct_null_clms_{ref,rec}[$t] is the number of 'null' columns in the original {reference,reconstructed} MSA segment under the $t th candidate tree.

	for (1 .. $ct_cand_trees) {
=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ia-7)
	   push @set_refseg_scores_sb, 0;
	   push @set_recseg_scores_sb, 0;
	   push @set_refseg_scores_gp, 0;
	   push @set_recseg_scores_gp, 0;
=cut

		# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	   push @set_avct_ins_ref, undef;
	   push @set_avct_ins_rec, undef;
	   push @set_avct_dels_ref, undef;
	   push @set_avct_dels_rec, undef;
	   ##push @set_ct_indels_ref, undef;
	   ##push @set_ct_indels_rec, undef;
	}

	##my @set_clm_scores = (); # $set_clm_scores[$indx_clm]->[$t] is the ML score of the $indx_clm th nr-column under the $t th tree candidate.
	##for (1 .. $ct_nr_clms) { # 2nd interim auxiliary middle for-loop (over non-redundant columns)  #
	##    my @tmp = ();
	##    $tmp[$ct_cand_trees - 1] = undef;
	##    push @set_clm_scores, \@tmp;
	##} # End of the 2nd interim auxiliary middle for-loop (over non-redundant columns) #

  for (my $t = 0; $t<$ct_cand_trees; $t++) { # 1st outer-middle for-loop (over candidate trees) #

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ib)

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	foreach my $fh_log ($fh_log_sref, $fh_log_srec) { # MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'. #
	##foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {...}
		# Fixed a very minor bug. #
	    print $fh_log "<<\{\{ The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) \}\}>>\n\n";
	    ##print $fh_log_sref "<<\{\{ The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) \}\}>>\n\n";
	}


		# Handle Exceptional Case #

	if ((@{$node2ch->{$topnode_id}} == 2) and (@{$sequence_ids} >2)) {
	    foreach my $fh_log ($fh_log_sref, $fh_log_srec) { # MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'. #
	    ##foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {...}
		print $fh_log "[CAUTION '$program_name'] The candidate tree is BIFURCATED --- I will make it multifurcated ...\n\n";
	    }
	    unless ( bi2multifurcated_top (%{$node2pa}, %{$node2ch}, %{$name2id}, %{$br_attr}, @INDICES_ATTR) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'bi2multifurcated_top' FAILED on the $t th candidate tree !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    $topnode_id = $name2id->{$TOP_NODE};
	}

	foreach my $fh_log ($fh_log_sref, $fh_log_srec) { # MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'. #
	##foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) {...}
	    print $fh_log "\n[ Input Tree (top_node_ID = $topnode_id) ]\n\n";
	
	    my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	    my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	    unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the $t th candidate tree (for substitution analyses): $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	}

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ib)"


		## Retrieve a tree for the indel analyses. ##
		## (ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014).) ##

	my ($node2pa_id, $node2ch_id, $name2id_id, $id2name_id, $br_attr_id) = @{$std_set_cand_trees_indels[$t]};
	my $topnode_id_id = $name2id_id->{$TOP_NODE};

	foreach my $fh_log ($fh_log_gref, $fh_log_grec) {
	    print $fh_log "<<\{\{ The $t th candidate tree (topnode ID: $topnode_id_id) \}\}>>\n\n";
	    ##print $fh_log "<<\{\{ The $t th candidate tree (topnode ID: $topnode_id_id\; Log_Akaike_weight = $ln_awgt) \}\}>>\n\n";
	}

	foreach my $fh_log ($fh_log_gref, $fh_log_grec) {

	    print $fh_log "\n[ Input Tree (top_node_ID = $topnode_id_id) ]\n\n";
	
	    my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	    my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	    unless ( output_tree ($topnode_id_id, %{$node2ch_id}, %{$id2name_id}, %{$br_attr_id}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the $t th candidate tree (for indel analyses): $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	}

		## END of "(ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 6 (Wed), 2014).)" ##

 		# ADDED in creation of 'classify_msa_errors_via_mblks.2.pl'. #

	my ($node2pa, $node2ch, $name2id, $id2name, $br_attr, $topnode_id)
		= ($node2pa_id, $node2ch_id, $name2id_id, $id2name_id, $br_attr_id, $topnode_id_id);

	my $node2depth = $set_node2depth[$t];
	my $node2ct_exoffs = $set_node2ct_exoffs[$t];
	my $node2seq_indx = $set_node2seq_indx[$t];


	print STDOUT "\n<<<< Sub-Step (1): Infer indel histories that can result in the reference & reconstructed MSA segments, based on the Dollo parsimony criteriion, and count the inferred insertions/deletions >>>>\n\n";

 		# END of "ADDED in creation of 'classify_msa_errors_via_mblks.2.pl'." #


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ic)

		### Calculate log-likelihoods of the indel scenarios
		### that result in reference & reconstructed MSA segments.

	print STDOUT "\n<<<< Sub-Step (1): Calculating log-likelihoods of the indel scenarios of the reference & reconstructed MSA segments >>>>\n\n";

##our $CUTOFF_3P = our $CUTOFF_5P = $cutoff_len;
##our @PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS);
##our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL);
##our @LABELS_NFACS = ("Insertions(5p)", "Deletions(5p)", "Insertions(3p)", "Deletions(3p)");
##my $log_nfacs = cal_log_normal_facs_indel_distri (@PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P);
##my $delta_del_5p = cal_delta_exp_del_5p (@PARAMS_DEL, $CUTOFF_5P, $log_nfac_del_5p); 

	##$ref_coord_beg += $ref_len;
	##$rec_coord_beg += $rec_len;
	##push @ref_seg_msa, \$seg_ref;
	##push @rec_seg_msa, \$seg_rec;
	##@set_refseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reference segment under the $t th tree candidate.
	##@set_recseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reconstructed segment under the $t th tree candidate.
	##@set_ct_indels_ref = my @set_ct_indels_rec = (); # $set_ct_indels_{ref,rec}[$t] is the average number of indel events under the $t th tree candidate.
	##@set_ct_null_clms_ref = my @set_ct_null_clms_rec = (); # $set_ct_null_clms_{ref,rec}[$t] is the number of 'null' columns in the original {reference,reconstructed} MSA segment under the $t th candidate tree.
	##my $totlen_refaln = length ${$ref_msa->[0]};

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (Ic)"

		# ADDED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
	my @info_gpatterns = ([], [], []);
	my @set_refseg_scores_gp = my @set_recseg_scores_gp = (); # Dummy arrays. #
		# END of "ADDED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'." #

	foreach my $set_info ( ## MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. ##
		##['Refefence', \@ref_seg_msa, $ref_coord_beg, $ref_len, $totlen_refaln, \@set_refseg_scores_gp, \@set_avct_ins_ref, \@set_avct_dels_ref, $fh_log_gref], # A BUG fixed on Nov 17 (Tue), 2015. #
		['Reference', \@ref_seg_msa, $ref_coord_beg, $ref_len, $totlen_refaln, \@set_refseg_scores_gp, \@set_avct_ins_ref, \@set_avct_dels_ref, $fh_log_gref],
		['Reconstructed', \@rec_seg_msa, $rec_coord_beg, $rec_len, $totlen_recaln, \@set_recseg_scores_gp, \@set_avct_ins_rec, \@set_avct_dels_rec, $fh_log_grec]) { ## 0th middle for-loop (over reference/reconstructed MSAs) ##

	    my ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_avct_ins, $set_avct_dels, $fh_log_g) = @{$set_info};


	    print STDOUT "<<< For '$msa_type' (coord_beg = $coord_beg, seg_len = $seg_len; totlen_aln = $totlen_aln ) >>>\n\n";
	    print $fh_log_g "\n(coord_seg_beg, seg_length) = ($coord_beg, $seg_len)\n\n";
	    print $fh_log_g "(cf. total alignment length = $totlen_aln)\n\n";

		# Copy the MSA #
	    my @cp_seg_msa = ();
	    for (my $k=0; $k< $ct_sequences; $k++) { my $cp_seg = ${$seg_msa->[$k]}; push @cp_seg_msa, \$cp_seg; }


			### MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl',
			### by IMPORTING some blocks from 'Ingredients/classify_msa_errors_via_mblks.2.pl'.

	my $fh_log = $fh_log_g;

	print STDOUT "<< (i): Infer an indel history in each gapped segment in the segmental MSA ($msa_type), based on the Dollo parsimony principle. >>\n\n\n";
	print $fh_log "<< (i): Infer an indel history in each gapped segment in the segmental MSA ($msa_type), based on the Dollo parsimony principle. >>\n\n\n";

		# @{$info_init_psm_scn} stores information on the initial parsimonious scenario candidate for each cluster (gapped segment) in the intermediate MSA;
		# @{$info_gpatterns_updated} stores information on the gap-patterns, as well as their Dollo parsimony scenarios, that are found in the intermediate MSA. 
		#
		# For the contents of these variables, see below.
		#
		my ($info_init_psm_scn, $info_gpatterns_updated) = cal_init_psm_cands_for_sgl_msa (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$node2ct_exoffs}, %{$name2id}, @{$sequence_ids}, @cp_seg_msa, $CHAR_ABS, $CHAR_PRES, @info_gpatterns); # MODIFIED on Oct 14 (Wed), 2015.

		unless (defined $info_init_psm_scn) {
		    print STDOUT "[ERROR '$program_name'] Subroutine 'cal_init_psm_cands_for_sgl_msa' FAILED for the '$msa_type' MSA in the $s th segment !\n\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}

		my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};
		my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns_updated};

		@info_gpatterns = @{$info_gpatterns_updated};

	    my $ct_cls = @{$set_cls_gblocks};

   (print $fh_log "\n#{Gapped segments involved} = $ct_cls .\n\n") if ($TEST_LEVEL2>1);

	    my $totct_ins = my $totct_dels = 0;

	    for (my $cl=0; $cl<$ct_cls; $cl++) { # 2nd main outer for-loop (over the gap-block clusters). #

		my $gbs_in_cl = $set_cls_gblocks->[$cl];
		my $beg_cl = $gbs_in_cl->[0][2];
		my $end_cl = $gbs_in_cl->[$#{$gbs_in_cl}][3];

		my $ct_gbs = @{$gbs_in_cl};

   (print $fh_log "   <  The $cl th gapped segment \; [beg, end] = [$beg_cl, $end_cl] \; #{gap-blocks} = $ct_gbs >\n\n") if ($TEST_LEVEL2>2);

   if ($TEST_LEVEL2>3) {
	print $fh_log "   [ Gap pattern ]\n\n";
	print $fh_log '   ', join ("\t", 'Indx_gb', 'beg', 'end', 'len', "'Gap_pattern'"), "\n\n";

	for (my $gb=0; $gb <$ct_gbs; $gb++) {
	    my ($indx_gp, $len_gb, $beg_gb, $end_gb) = @{$gbs_in_cl->[$gb]};
	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print $fh_log '   ', join ("\t", $gb, $beg_gb, $end_gb, $len_gb, "'$gpattern'"), "\n";
	}
	print $fh_log "\n\n";
   }


		my $init_psm_scn = $init_psm_scn_for_cls->[$cl];
		my ($br2indels, $node2bw_gstat) = @{$init_psm_scn};

    (print $fh_log "   [ Indel history ]\n\n") if ($TEST_LEVEL2>3);

    (print $fh_log '   ', join ("\t", 'Br_ID', 'state_ue', 'state_le', 'indices,involved,blocks'), "\n\n") if ($TEST_LEVEL2>3);

		my $ct_ins = my $ct_dels = 0;
		foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # Inner foreach-loop (over branches suffering indels). #
			foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop (over indels along $br). #
			    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
			    (defined $stat_ue) or next; # Probably, this entry is empty.
			    ($stat_ue eq $CHAR_ABS) ? ($ct_ins++) : ($ct_dels++);

			    ##my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef'; # Set aside for later use.

    (print $fh_log '   ', join ("\t", $br, $stat_ue, $stat_le, join (',', @{$indices_blk}) ), "\n") if ($TEST_LEVEL2>3);

			}			# End of the innermost foreach-loop (over indels along $br). #
		}				# End of the inner foreach-loop (over branches suffering indels). #

   (print $fh_log "\n   (#{ins} = $ct_ins, #{del} = $ct_dels) .\n\n\n") if ($TEST_LEVEL2>2); 

		$totct_ins += $ct_ins;
		$totct_dels += $ct_dels;

	    }			# End of the 2nd main outer for-loop (over the gap-block clusters). #

		## Important Output (iii) ###

	    $set_avct_ins->[$t]  = $totct_ins;
	    $set_avct_dels->[$t] = $totct_dels;

   if ($TEST_LEVEL2>1) {
	print $fh_log "\n";
	print $fh_log "Total#{insertions} = $totct_ins ,\n";
	print $fh_log "Total#{deletions}  = $totct_dels .\n\n\n";
   }


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (II)

	    print STDOUT "\n\n<< (i) Enumerate the parsimonious indel scenarios, each of which realizes the segment MSA >>\n\n";
	    print $fh_log_g "\n\n<< (i) Enumerate the parsimonious indel scenarios, each of which realizes the segment MSA >>\n\n";

	    my ($set_cls_gblocks, $min_totcts_indel_for_cls, $sets_psm_scns_for_cls, $set_gpatterns, $set_node2dparsstat, $set_br2change)
		  = enum_psm_indel_scns_for_msa (%{$node2ch_id}, %{$node2pa_id}, %{$name2id_id}, @{$sequence_ids}, @cp_seg_msa, $CHAR_ABS, $CHAR_PRES); # MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'. #
		  ##= enum_psm_indel_scns_for_msa (%{$node2ch}, %{$node2pa}, %{$name2id}, @{$sequence_ids}, @cp_seg_msa, $CHAR_ABS, $CHAR_PRES);
	    unless (defined $set_cls_gblocks) {
		print STDOUT "[ERROR '$program_name'] The preliminary subroutine 'enum_psm_indel_scns_for_msa' FAILED !\n";
		print STDERR "[ERROR '$program_name'] $! !\n\n";
		die ;
	    }
	    print STDOUT "\n--- The preliminary subroutine 'enum_psm_indel_scns_for_msa' ENDED ...\n\n";

	    print $fh_log_g "\n[[ Secondary Inputs (i): Outputs of 'enum_psm_indel_scns_for_msa' ]]\n\n";

	    my $ct_gp = @{$set_gpatterns};

    if ($TEST_LEVEL>1) {
	    print $fh_log_g "[1] List of gap patterns encountered (#{patterns} = $ct_gp):\n\n";
	    print $fh_log_g join ("\t", 'Index', "'gap_patterns(horizontal_representation)'"), "\n\n";
	    for (my $indx_gp1=0; $indx_gp1<$ct_gp; $indx_gp1++) { # 1st interim outer for-loop (over the gap patterns) #

		my $gpattern = ${$set_gpatterns->[$indx_gp1]};
		print $fh_log_g join ("\t", $indx_gp1, "'$gpattern'"), "\n";

	    }	# End of the 1st interim outer for-loop (over the gap patterns) #
	    print $fh_log_g "\n\n";
    } ##  END of "if ($TEST_LEVEL>1)" 

    if ($TEST_LEVEL>2) {
	    print $fh_log_g "[2] Dollo parsimonious scenarios realizing the gap patterns:\n\n";
	    for (my $indx_gp1=0; $indx_gp1<$ct_gp; $indx_gp1++) { # 2nd interim outer for-loop (over the gap patterns) #
    
		my $gpattern = ${$set_gpatterns->[$indx_gp1]};
		my $node2dparsstat = $set_node2dparsstat->[$indx_gp1];
		my $br2change = $set_br2change->[$indx_gp1];

		print $fh_log_g "\n{ The $indx_gp1 th gap pattern }\n\n";

		print $fh_log_g "Pattern:  '$gpattern'\n\n";
		print $fh_log_g "( Map of Dollo parsimonious states onto the nodes )\n\n";
		print $fh_log_g join ("\t", 'Node_ID', 'state', '#{child nodes with #{exterior descendants with \'absence\' state}>0}'), "\n\n";
		foreach my $node (sort {$a <=> $b} keys %{$node2dparsstat}) { print $fh_log_g join ("\t", $node, @{$node2dparsstat->{$node}}), "\n"; }
		print $fh_log_g "\n\n";
		print $fh_log_g "( Map of Dollo parsimonious indel events onto the branches )\n\n";
		print $fh_log_g join ("\t", 'Branch_ID', 'state(upper-end)', 'state(lower-end)'), "\n\n";
		foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print $fh_log_g join ("\t", $br, @{$br2change->{$br}}), "\n"; }
		print $fh_log_g "\n\n";
    
	    }			# End of the 2nd interim outer for-loop (over the gap patterns) #
    } ##  END of "if ($TEST_LEVEL>2)"


	    my $ct_cls = @{$set_cls_gblocks};
	    my $min_totct_indels = 0;


	    print $fh_log_g "\n[3] Clusters of gap-pattern-blocks (#{clusters} = $ct_cls) and Genuinely Parsimonious Scenarios for each of the clusters:\n\n\n";

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # 3rd interim outer for-loop (over the clusters) #

		my $cls_gblks = $set_cls_gblocks->[$cl];
		my $ct_gblks = @{$cls_gblks};

		print $fh_log_g "< The $cl th cluster (#{gap-blocks} = $ct_gblks) >\n\n";

		print $fh_log_g "[ Gap-blocks in the cluster ]\n\n";
		print $fh_log_g join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

		for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st middle for-loop (over the gap-blocks) #

			## Shift the start/end coordinates to the left by $coord_beg. ##
			## Actually, this operation is very important when calculating the log-likelihoods of indel scenarios. ##
		    my $info_gblock = $cls_gblks->[$bl];
		    $info_gblock->[2] += $coord_beg;
		    $info_gblock->[3] += $coord_beg; 

		    my ($indx_gp, $len_gb, $start, $end) = @{$info_gblock};
	
		    my $gpattern = ${$set_gpatterns->[$indx_gp]};
		    print $fh_log_g join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
		}			# End of the 1st middle for-loop (over the gap-blocks) #
		print $fh_log_g "\n\n";

		my $min_sbtotct_indels = $min_totcts_indel_for_cls->[$cl];

			## MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ## 
		unless ($min_totct_indels eq 'N/A') {
		    ($min_sbtotct_indels eq 'N/A') ? ($min_totct_indels = 'N/A') : ($min_totct_indels += $min_sbtotct_indels);
		}
		##$min_totct_indels += $min_sbtotct_indels;

		my $set_psm_scns = $sets_psm_scns_for_cls->[$cl];
		my $ct_psm_scns = @{$set_psm_scns};

		print $fh_log_g "[ Parsimonious Scenarios (Min(Total#{indel events}) = $min_sbtotct_indels\; #{scenarios} = $ct_psm_scns) ]\n\n";

		for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # 2nd middle for-loop (over the parsimonious scenarios) #

		    print $fh_log_g ' ' x 3, "{ The $ps th scenario }\n\n";

		    my ($br2indels, $node2bw_gstats) = @{$set_psm_scns->[$ps]};

    if ($TEST_LEVEL>2) {
		    print $fh_log_g ' ' x 3, "( Map of block-wise gap states onto the nodes )\n\n";
		    print $fh_log_g ' ' x 3, join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";
		    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats}) { # 1st inner foreach-loop #
			my $cnct_bw_gstats = join ('', @{$node2bw_gstats->{$node}});
			print $fh_log_g ' ' x 3, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
		    }				# End of the 1st inner foreach-loop #
		    print $fh_log_g "\n\n";
    } ## END of "if ($TEST_LEVEL>2)"

		    print $fh_log_g ' ' x 3, "( Map of indel events onto the branches )\n\n";
		    print $fh_log_g ' ' x 3, join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
		    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner foreach-loop #
			my @out_indel_info = ();
			foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop #
			    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
			    (defined $stat_ue) or ($stat_ue = 'undef');
			    (defined $stat_le) or ($stat_le = 'undef');
			    my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
			    push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
			}			# End of the innermost foreach-loop #
			print $fh_log_g ' ' x 3, join ("\t", $br, @out_indel_info), "\n";
		    }				# End of the 2nd inner foreach-loop #
		    print $fh_log_g "\n\n";

		}			# End of the 2nd middle for-loop (over the parsimonious scenarios) #

	    }		# End of the 3rd interim outer for-loop (over the clusters) #

	    print $fh_log_g "\n\n.....Finished displaying the Secondary Inputs (i).....\n\n";


	    print $fh_log_g "\n(For reference) Minimum Total #{indels} = $min_totct_indels .\n\n";
	    ##print $fh_log_g "\n(Main!!) Minimum Total #{indels} = $min_totct_indels .\n\n";



		## Important Output (i) ##

	    ##$set_ct_indels->[$t] = $min_totct_indels; # DISABLED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

	    ## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info};


		## ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##

	    if ($min_totct_indels eq 'N/A') {
		print STDOUT "\n\n[WARNING '$program_name'] The $msa_type MSA has minimum total #{indels} NOT AVAILABLE, which means that search for its parsimonious indel scenarios HALTED PREMATURELY!!\n";
		print STDOUT ' ' x 5, "... Thus, it's no use going ahead with calculating the likelihood for indel scenarios. Instead, I will just set it to be 'undef' ...\n\n";

		print $fh_log_g "\n\n[WARNING '$program_name'] The $msa_type MSA has minimum total #{indels} NOT AVAILABLE, which means that search for its parsimonious indel scenarios HALTED PREMATURELY!!\n";
		print $fh_log_g ' ' x 5, "... Thus, it's no use to go ahead with calculating the likelihood for indel scenarios...\n\n";

		print $fh_log_g '               $tot_lnlk_gp = undef', " .\n\n\n";

		    ## Important Output (ii) (Exceptional) ##
	        $set_seg_scores_gp->[$t] = undef; # RESURRECTED because it turned out to be important.
	        ##$set_seg_scores_gp->[$t] = undef; # DISABLED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

			# MODIFIED via IMPORT from 'compare_ref_vs_rec_compllimment.hs.pl' for BUG-PREVENTION on Nov 17 (Tue), 2015. #

		if ($msa_type  eq 'Reference') {

		    $set_recseg_scores_gp[$t] = undef;
		    ##$set_ct_indels_rec[$t] = $min_totct_indels; # DISABLED when imported to 'compare_ref_vs_rec_indel_cts.hs.pl'. #

		    print STDOUT "[WARNING '$program_name'] This analysis of the reconstructed MSA in segment was SKIPPED, because the analysis of the reference MSA FAILED !!\n\n";
		    print $fh_log_grec "[WARNING '$program_name'] This analysis of the reconstructed MSA in segment was SKIPPED, because the analysis of the reference MSA FAILED !!\n\n";
		}
		last;
		##next;
			# END of "MODIFIED via IMPORT from 'compare_ref_vs_rec_compllimment.hs.pl' for BUG-PREVENTION on Nov 17 (Tue), 2015." #

	    ## [ ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info}; ] --- for the predecessor 'compare_ref_vs_rec_compllimment.hs.pl'. #
	    ## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_avct_ins, $set_avct_dels, $fh_log_g) = @{$set_info};
	     # ['Reconstructed', \@rec_seg_msa, $rec_coord_beg, $rec_len, $totlen_recaln, \@set_recseg_scores_gp, \@set_avct_ins_rec, \@set_avct_dels_rec, $fh_log_grec]) { ## 0th middle for-loop (over reference/reconstructed MSAs) ## ...}

	    }

		## END of "ADDED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014)." ##



			## Before beginning to calculate the log-likelihoods of the parsimonious indel scenarios,
			## check whether all indels are shorter than or equal to the specified cut-off length.
			## (ADDED on Sep 29 (Sat), 2012)

		# IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl',
		# and used for MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. 
	    my $min_cutoff_len_ins = $CUTOFF_INS_LEN;
	    my $min_cutoff_len_del = $CUTOFF_DEL_LEN;
	    ##my $min_cutoff_len_ins = my $min_cutoff_len_del = ($CUTOFF_5P < $CUTOFF_3P) ? $CUTOFF_5P : $CUTOFF_3P; 

	    print STDOUT "\n\n<< Examine whether indel events are shorter than or equal to the cut-off length (= $min_cutoff_len_ins for insertions, = $min_cutoff_len_del for deletions) >>\n\n";
	    print $fh_log_g "\n\n<< Examine whether indel events are shorter than or equal to the cut-off length (= $min_cutoff_len_ins for insertions, = $min_cutoff_len_del for deletions) >>\n\n";

	    my $illegal_indels = inspect_indel_events (@{$set_cls_gblocks}, @{$sets_psm_scns_for_cls}, $min_cutoff_len_ins, $min_cutoff_len_del, $CHAR_ABS);
	    unless (defined $illegal_indels) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'inspect_indel_events' FAILED !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    if (@{$illegal_indels}>0) {
		my $ct_illegal = @{$illegal_indels};
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with

		print STDOUT "[WARNING '$program_name'] The $msa_type MSA has $ct_illegal illegal indels:\n\n";
		print $fh_log_g "[WARNING '$program_name'] The $msa_type MSA has $ct_illegal illegal indels:\n\n";
		print STDOUT ' ' x 3, join ("\t", 'No.', 'ID_cls', 'ID_psm_scn', 'Br_ID', 'ID_indel', 'Length'), "\n\n";
		print $fh_log_g ' ' x 3, join ("\t", 'No.', 'ID_cls', 'ID_psm_scn', 'Br_ID', 'ID_indel', 'Length'), "\n\n";
		for (my $il=0; $il<$ct_illegal; $il++) {
		    print STDOUT ' ' x 3, join ("\t", $il, @{$illegal_indels->[$il]}), "\n";
		    print $fh_log_g ' ' x 3, join ("\t", $il, @{$illegal_indels->[$il]}), "\n";
		}
		print STDOUT "\n\n", ' ' x 3, "... I will refrain from calculating the likelihood of indel scenarios. Instead, I will just set it to be 'undef' ...\n\n";
		print $fh_log_g "\n\n", ' ' x 3, "... I will refrain from calculating the likelihood of indel scenarios...\n\n";
		print $fh_log_g '               $tot_lnlk_gp = undef', " .\n\n\n";

		    ## Important Output (ii) (Exceptional) ##
	        $set_seg_scores_gp->[$t] = undef; # RESURRECTED because it turned out to be important.
	        ##$set_seg_scores_gp->[$t] = undef; # DISABLED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #


			# MODIFIED via IMPORT from 'compare_ref_vs_rec_compllimment.hs.pl' for BUG-PREVENTION on Nov 17 (Tue), 2015. #

		if ($msa_type  eq 'Reference') {

		    $set_recseg_scores_gp[$t] = undef;
		    ##$set_ct_indels_rec[$t] = $min_totct_indels; # DISABLED on import to 'compare_ref_vs_rec_indel_cts.hs.pl'. #

		    print STDOUT "[WARNING '$program_name'] This analysis of the reconstructed MSA in segment was SKIPPED, because the analysis of the reference MSA FAILED !!\n\n";
		    print $fh_log_grec "[WARNING '$program_name'] This analysis of the reconstructed MSA in segment was SKIPPED, because the analysis of the reference MSA FAILED !!\n\n";
		}
		last;
		##next;
			# END of "MODIFIED via IMPORT from 'compare_ref_vs_rec_compllimment.hs.pl' for BUG-PREVENTION on Nov 17 (Tue), 2015." #

	    ## [ ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info}; ] --- for the predecessor 'compare_ref_vs_rec_compllimment.hs.pl'. #
	    ## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_avct_ins, $set_avct_dels, $fh_log_g) = @{$set_info};
	     # ['Reconstructed', \@rec_seg_msa, $rec_coord_beg, $rec_len, $totlen_recaln, \@set_recseg_scores_gp, \@set_avct_ins_rec, \@set_avct_dels_rec, $fh_log_grec]) { ## 0th middle for-loop (over reference/reconstructed MSAs) ## ...}

	    }

	    my @interim_in_sets_psm_scns_for_cls = @{$sets_psm_scns_for_cls}; # IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #

		# ADDED in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'. #
	    my $br2log_mfacs = $set_br2log_mfacs[$t];
			# %{$set_br2log_mfacs[$t]} = ($br_id => \@set_log_mfacs, ...) for all branches.


	    print STDOUT "\n\n<< (ii) Calculate the log-likelihoods of the parsimonious indel scenarios, and therefore the log-likelihood of the gap-pattern of this MSA segment >>\n\n";
	    print $fh_log_g "\n\n<< (ii) Calculate the log-likelihoods of the parsimonious indel scenarios, and therefore the log-likelihood of the gap-pattern of this MSA segment >>\n\n";

	    my ($tot_lnlk_gp0, $shared_fac1, $shared_fac2, $set_clwise_tot_lnlk_gp, $tot_div_tree, $totlen_ng, $setset_lnlk_gp_cls_scn)
		= apprx1_tot_lnlk_gpattern_msa2 (%{$br_attr_id}, %{$node2pa_id}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @interim_in_sets_psm_scns_for_cls, $TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, %{$br2log_mfacs}); # IMPORTED from 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl', and used for MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.hs.pl'.

#=for the predecessor 'compare_ref_vs_rec_indel_cts.pl'.
	     = apprx1_tot_lnlk_gpattern_msa (%{$br_attr_id}, %{$node2pa_id}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @{$sets_psm_scns_for_cls}, @PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P, @{$log_nfacs}, $delta_del_5p, $OPT_BD_INDELS); # MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'. #
	     ##= apprx1_tot_lnlk_gpattern_msa (%{$br_attr}, %{$node2pa}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @{$sets_psm_scns_for_cls}, @PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P, @{$log_nfacs}, $delta_del_5p, $OPT_BD_INDELS);
#=cut
	    unless (defined $tot_lnlk_gp0) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'apprx1_tot_lnlk_gpattern_msa2' FAILED: $! !\n\n";
		print STDERR "[ERROR '$program_name'] $! !\n\n";
		die ;
	    }
	    unless (@{$set_clwise_tot_lnlk_gp} == $ct_cls) {
		print STDOUT "[ERROR '$program_name'] The output ", '@{$set_clwise_tot_lnlk_gp} ', "has the size ", scalar (@{$set_clwise_tot_lnlk_gp}), ", which CLASHES with $ct_cls, the number of gap-block clusters !\n";
		print STDOUT "[ERROR '$program_name'] $!\n";
		print STDERR "[ERROR '$program_name'] $! !\n\n";
		die ;
	    }
	    unless (@{$setset_lnlk_gp_cls_scn} == $ct_cls) {
		print STDOUT "[ERROR '$program_name'] The output ", '@{$setset_lnlk_gp_cls_scn} ', "has the size ", scalar (@{$setset_lnlk_gp_cls_scn}), ", which CLASHES with $ct_cls, the number of gap-block clusters !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }
	    print STDOUT "\n--- The main subroutine 'apprx1_tot_lnlk_gpattern_msa2' ENDED ...\n\n";


		## An Important Adjustment, from 'MSA-wide' log-likelihoods to segment-wise log-likelihoods. ##

	    my $tot_lnlk_gp = $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len); 


	    print $fh_log_g "\n[[ Outputs of 'apprx1_tot_lnlk_gpattern_msa2', to be fed into the Final Output ]]\n\n";

	    print $fh_log_g "[1] Scalar Outputs:\n\n";
	    print $fh_log_g '($tot_div_tree, $totlen_ng) = ', "($tot_div_tree, $totlen_ng) ,\n";
	    print $fh_log_g '($shared_fac1, $shared_fac2) = ', "($shared_fac1, $shared_fac2) .\n\n";
	    print $fh_log_g 'RAW $tot_lnlk_gp = ', $tot_lnlk_gp0, " .\n\n";

	    print $fh_log_g "\n[2] Cluster-wise log-likelihoods (for the $ct_cls gap-block clusters):\n\n";
	    print $fh_log_g join ("\t", 'Indx_gb_cluster', 'clwise_tot_lnlk_gp', '#{psm_scenarios}', 'lnlk_gp_cls(scn1)', '[lnlk_gp_slc(scn2)', '...]'), "\n\n";

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over the gap-block clusters) #
    
		my $cw_tot_lnlk_gp = $set_clwise_tot_lnlk_gp->[$cl];
		my $ct_psm_scn = @{$sets_psm_scns_for_cls->[$cl]};
		my $set_lnlk_gp_scn = $setset_lnlk_gp_cls_scn->[$cl];

		unless (@{$set_lnlk_gp_scn} == $ct_psm_scn) {
		    print $fh_log_g "\n[CAUTION '$program_name'] In the $cl th cluster, the number of log-likelihoods of parsimonious scenarios (=", scalar (@{$set_lnlk_gp_scn}), ") CLASHES with the number of scenarios themselves (=$ct_psm_scn) ... !!\n\n";
		}
		print $fh_log_g join ("\t", $cl, $cw_tot_lnlk_gp, $ct_psm_scn, @{$set_lnlk_gp_scn}), "\n";

	    }		# End of the main outer for-loop (over the clusters) #
	    
	    print $fh_log_g "\n\n.....Finished displaying the 'Final' Outputs (ii) ....\n\n";

	    print $fh_log_g '(Main!!) FINAL $tot_lnlk_gp (= $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len)):', "\n";
	    print $fh_log_g '               $tot_lnlk_gp = ', $tot_lnlk_gp, " .\n\n\n";

		## Important Output (ii) ##
	    $set_seg_scores_gp->[$t] = $tot_lnlk_gp; # RESURRECTED because it turned out to be important.
	    ##$set_seg_scores_gp->[$t] = $tot_lnlk_gp; # DISABLED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

	    ## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info};


		## ADDED as a MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.pl'. ##

	    my $avct_ins = my $avct_dels = 0;

	    for (my $cl=0; $cl<$ct_cls; $cl++) { # 2nd main outer for-loop (over the gap-block clusters). #
		my $cw_tot_lnlk_gp = $set_clwise_tot_lnlk_gp->[$cl];
		(defined $cw_tot_lnlk_gp) or next; # In this case, the likelihood of the cluster's gap-pattern is regarded as ZERO.

		if ($cw_tot_lnlk_gp eq 'N/A') { # In this case, we cannot calculate the average counts of insertions or deletions.
		    $avct_ins = $avct_dels = undef;
		    last;
		}

		my $set_psm_scns = $sets_psm_scns_for_cls->[$cl];
		my $ct_psm_scns = @{$set_psm_scns};
		my $set_lnlk_gp_scn = $setset_lnlk_gp_cls_scn->[$cl];

		my $av_sbct_ins = my $av_sbct_dels = 0;

		for (my $scn=0; $scn<$ct_psm_scns; $scn++) { # Middle for-loop (over parsimonious scenarios). #

		    my $lnlk_gp_scn = $set_lnlk_gp_scn->[$scn];
		    (defined $lnlk_gp_scn) or next; # In this caes, the likelihood of the scenario is regarded as ZERO.

		    my ($br2indels, $node2bw_gstats) = @{$set_psm_scns->[$scn]};

		    my $ct_ins = my $ct_dels = 0;
		    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # Inner foreach-loop (over branches suffering indels). #
			foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop (over indels along $br). #
			    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
			    (defined $stat_ue) or next; # Probably, this entry is empty.
			    ($stat_ue eq $CHAR_ABS) ? ($ct_ins++) : ($ct_dels++);

			    ##my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef'; # Set aside for later use.

			}			# End of the innermost foreach-loop (over indels along $br). #
		    }				# End of the inner foreach-loop (over branches suffering indels). #

		    my $wgt = exp ($lnlk_gp_scn - $cw_tot_lnlk_gp);
		    $av_sbct_ins += $wgt * $ct_ins;
		    $av_sbct_dels += $wgt * $ct_dels;

		}			# End of the middle for-loop (over parsimonious scenarios). #

		$avct_ins += $av_sbct_ins;
		$avct_dels += $av_sbct_dels;

	    }			# End of the 2nd main outer for-loop (over the gap-block clusters). #

		## Important Output (iii) ###

	    $set_avct_ins->[$t] = $avct_ins;
	    $set_avct_dels->[$t] = $avct_dels;

		## END of "ADDED as a MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.pl'." ##

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (II)"


	}			## End of the 0th middle for-loop (over reference/reconstructed MSAs) ##


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (III)

	print STDOUT "\n<<<< Sub-Step (2): Calculating log-likelihoods of substitution patterns in the reference & reconstructed MSA segments >>>>\n\n";


		### MOVED from above (outside of the for-loop over candidate trees) on Sep 1 (Sat), 2012 ###
	my $ct_gpats = 1; # A dummy number of gap patterns. #

    for (my $indx_gp = 0; $indx_gp < $ct_gpats; $indx_gp++) { # The 1st middle for-loop (over gap patterns) #

	my @ng_indices = (0 .. $ct_sequences - 1); # A dummy set of indices of non-gapped sequences
	my @std_ng_indices = sort {$a <=> $b} @ng_indices;

	my $sbct_seqs = @std_ng_indices;

	   # Prepare a 'reduced' alignment by keeping only sequences with residues #

	my @red_seqnames = (); # element = the name of a non-gapped sequence
	my @red_vmsa_ref = my @red_vmsa_rec  = (); # element = \${a non-gapped sequence embedded in the virtual MSA}
	foreach my $k (@std_ng_indices) { # Preliminary middle foreach-loop #
	    push @red_seqnames, $sequence_ids->[$k];
	    my $seg_ref = ${$ref_seg_msa[$k]};
	    my $seg_rec = ${$rec_seg_msa[$k]};

	    push @red_vmsa_ref, \$seg_ref;
	    push @red_vmsa_rec, \$seg_rec;
	}			# End of the preliminary middle foreach-loop #

	foreach my $fh_log ($fh_log_sref, $fh_log_srec) {
	    print $fh_log "\n<< The $indx_gp th dummy gap pattern (full) >>\n\n";
	}

	my $red_vmsa_ref = rm_gapcolumn2 (@red_vmsa_ref, $GAP);
	my $red_vmsa_rec = rm_gapcolumn2 (@red_vmsa_rec, $GAP);

	my $ref_len1 = length ${$red_vmsa_ref->[0]};
	my $rec_len1 = length ${$red_vmsa_rec->[0]};

	my $ct_null_clms_ref = $ref_len - $ref_len1;
	my $ct_null_clms_rec = $rec_len - $rec_len1;

	$set_ct_null_clms_ref[$t] = $ct_null_clms_ref;
	$set_ct_null_clms_rec[$t] = $ct_null_clms_rec;

	if ($ct_null_clms_ref>0) {
	    print STDOUT "\nCAUTION: REMOVED $ct_null_clms_ref 'null' columns from the reference MSA segment (original length $ref_len) ...\n\n";
	    print $fh_log_sref "\nCAUTION: REMOVED $ct_null_clms_ref 'null' columns from the reference MSA segment (original length $ref_len) ...\n\n";
	}
	if ($ct_null_clms_rec>0) {
	    print STDOUT "\nCAUTION: REMOVED $ct_null_clms_rec 'null' columns from the reconstructed MSA segment (original length $rec_len) ...\n\n";
	    print $fh_log_sref "\nCAUTION: REMOVED $ct_null_clms_rec 'null' columns from the reconstructed MSA segment (original length $rec_len) ...\n\n";
	}

	@red_vmsa_ref = @{$red_vmsa_ref};
	@red_vmsa_rec = @{$red_vmsa_rec};

	##foreach my $fh_clmlk ($fh_clmlk_ref, $fh_clmlk_rec) {
	##    print $fh_clmlk "\n<< The $indx_gp th dummy gap pattern (full) >>\n\n";
	##
	##    print $fh_clmlk "\n< Column-wise log-likelihoods under $ct_cand_trees candidate trees >\n\n";
	##    print $fh_clmlk join ("\t", '#Indx_clm', 'lnlk_clm(t0)', 'lnlk_clm(t1)', '...'), "\n\n";
	##}
		### END of "MOVED from above (outside of the for-loop over candidate trees)" ###


		### MOVED from above (outside of the for-loop over candidate trees) on Sep 1 (Sat), 2012 ###
    if ($TEST_LEVEL>2) {
	print $fh_log_sref "<< Reduced Segment of the Reference Alignment (#{involved sequences} = $sbct_seqs, #{columns} = $ref_len1) >>\n\n";
	unless ( pr_ODEN_input (@red_seqnames, @red_vmsa_ref, 60, *${$fh_log_sref}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the reduced segment of the reference MSA for the $indx_gp th gap pattern !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_log_sref "\n\n";

	print $fh_log_srec "<< Reduced Segment of the Reconstructed Alignment (#{involved sequences} = $sbct_seqs, #{columns} = $rec_len1) >>\n\n";
	unless ( pr_ODEN_input (@red_seqnames, @red_vmsa_rec, 60, *${$fh_log_srec}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the reduced segment of the reconstructed MSA for the $indx_gp th gap pattern !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print $fh_log_srec "\n\n";
     }
		### END of "MOVED from above (outside of the for-loop over candidate trees)" ###


	   ## Prepare a 'reduced' phylogenetic tree by excluding sequences containing only gaps ##

	my %sb_name2id;
	my %sb_id2name;

	my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);

	if ($sbct_seqs == $ct_sequences) {
		$sb_topnode_id = $topnode_id;
		$sb_node2ch = $node2ch;
		$sb_node2pa = $node2pa;
		$sb_br_attr = $br_attr;
		%sb_name2id = %{$name2id};
		%sb_id2name = %{$id2name};
	} else {

	  my @sub_nodes = ();
	  foreach my $seqname (@red_seqnames) { # Preliminary inner foreach-loop (over non-gapped sequences) #
	    my $nodeid = $name2id->{$seqname};
	    unless (defined $nodeid) {
		print STDOUT "[ERROR '$program_name'] For the $indx_gp th gap pattern, the sequence name '$seqname' has NO node ID in the $t th candidate tree !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    push @sub_nodes, $nodeid;

		## A BUG fixed on June 30 (Wed), 2010 ##
	    $sb_name2id{$seqname} = $nodeid;
	    ##$sb_name2id{$seqname} = $name2id{$nodeid};

	    $sb_id2name{$nodeid} = $seqname;

	  }			# End of the preliminary inner foreach-loop (over non-gapped sequences) #

	  (my $sb_topnode_ids, $sb_node2ch, $sb_node2pa, $sb_br_attr)
		= construct_neat_subtree (%{$node2pa}, @sub_nodes, %{$br_attr});

	  unless (defined $sb_topnode_ids) {
	    print STDOUT "[ERROR '$program_name'] FAILED to reduce the $t th candidate tree for the $indx_gp th gap pattern !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	  }
	  unless (1 == @{$sb_topnode_ids}) {
	    print STDOUT "[ERROR '$program_name'] Found MULTIPLE TOP NODES for the reduced tree of the $t th tree for the $indx_gp th gap pattern: ", join (', ', @{$sb_topnode_ids}), " !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	  }
	  $sb_topnode_id = $sb_topnode_ids->[0];
	  $sb_name2id{$TOP_NODE} = $sb_topnode_id;
	}

		# Handle Exceptional Case -- Part I + 1/2 -- #

	if (@red_seqnames == 2) {
	    foreach my $fh_log ($fh_log_sref, $fh_log_srec) {
		print $fh_log "\n[CAUTION '$program_name'] PhyML does NOT accept a tree and an MSA consisting of two sequences !\n";
		print $fh_log "                So I will add a dummy sequence and make a 'virtual' 3-sequence tree and MSA ...\n\n";
	    }

	    unless ( add_dummy_seq (@red_seqnames, @red_vmsa_ref, %{$sb_node2pa}, %{$sb_node2ch}, %sb_name2id, %sb_id2name, %{$sb_br_attr}, @INDICES_ATTR) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'add_dummy' FAILED on the $indx_gp th gap pattern !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    my $dupl_rec2 = ${$red_vmsa_rec[1]};
	    $red_vmsa_rec[2] = \$dupl_rec2;
	}

		# Handle Exceptional Cases  -- Part II --- #

	if ((@{$sb_node2ch->{$sb_topnode_id}} == 2) and (@red_seqnames >2)) {
	    foreach my $fh_log ($fh_log_sref, $fh_log_srec) {
		print $fh_log "[CAUTION '$program_name'] The reduced tree for the $indx_gp th gap pattern  is BIFURCATED --- I will make it multifurcated ...\n\n";
	    }
	    unless ( bi2multifurcated_top (%{$sb_node2pa}, %{$sb_node2ch}, %sb_name2id, %{$sb_br_attr}, @INDICES_ATTR) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'bi2multifurcated_top' FAILED on the $indx_gp th gap pattern !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    $sb_topnode_id = $sb_name2id{$TOP_NODE};
	}

	foreach my $fh_log ($fh_log_sref, $fh_log_srec) {
	    print $fh_log "\n[ Reduced Tree (top_node_ID = $sb_topnode_id) ]\n\n";
	
	    my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	    my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	    unless ( output_tree ($sb_topnode_id, %{$sb_node2ch}, %sb_id2name, %{$sb_br_attr}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the reduced tree of the $t th tree for the $indx_gp th gap pattern: $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    print $fh_log "\n\n";
	}

	    ## Calculate the log-likelihoods of substitution patterns in the reference/reconstructed MSA segments ##

		# Temporary working directory #

	my $tmp_workdir_ref = sprintf ("%s/g%03d/t%03d", $grpa_tmp_workdir_ref, $indx_gp, $t);
	my $tmp_workdir_rec = sprintf ("%s/g%03d/t%03d", $grpa_tmp_workdir_rec, $indx_gp, $t);

	foreach my $tmp_workdir ($tmp_workdir_ref, $tmp_workdir_rec) {
	    unless ( -d $tmp_workdir ) {
		unless (0 == system ("mkdir -p $tmp_workdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to create a temporary working directory '$tmp_workdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	    unless (( -r $tmp_workdir ) and ( -w $tmp_workdir )) {
		unless (0 == system ("chmod -r u+rw $tmp_workdir" ) ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the temporary working directory '$tmp_workdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}

		# Temporary input files #

	my $tmp_inaln_name = 'input.aln';
	my $tmp_intree_name = 'intree.ph';
	## my ($ctrl_name) = ($src_ctrl =~ m!([^/\s]+)\s*$! );
	## my ($dat_name) = ($src_dat =~ m!([^/\s]+)\s*$! );

	my $tmp_inaln_ref = "$tmp_workdir_ref/$tmp_inaln_name";
	my $tmp_inaln_rec = "$tmp_workdir_rec/$tmp_inaln_name";
	my $tmp_intree_ref = "$tmp_workdir_ref/$tmp_intree_name";
	my $tmp_intree_rec = "$tmp_workdir_rec/$tmp_intree_name";
	my $tmp_ctrl_ref = "$tmp_workdir_ref/$ctrl_name";
	my $tmp_ctrl_rec = "$tmp_workdir_rec/$ctrl_name";
	my $tmp_dat_ref = "$tmp_workdir_ref/$dat_name";
	my $tmp_dat_rec = "$tmp_workdir_rec/$dat_name";

	my $tmp_out_name = "${tmp_inaln_name}_phyml_lk.txt";
	my $tmp_out_ref = "$tmp_workdir_ref/$tmp_out_name";
	my $tmp_out_rec = "$tmp_workdir_rec/$tmp_out_name";


		# Rename temporary files if they already exist #

	foreach my $tmpfile ($tmp_inaln_ref, $tmp_inaln_rec, $tmp_intree_ref, $tmp_intree_rec, $tmp_ctrl_ref, $tmp_ctrl_rec, $tmp_dat_ref, $tmp_dat_rec, $tmp_out_ref, $tmp_out_rec) { # foreach-loop over temporary files #
	    if ( -f $tmpfile ) {
		print STDOUT "[CAUTION 'ml_scoring_msa.pl'] The temporary file '$tmpfile' ALREADY EXISTS !\n";
		print STDOUT "   --- I will rename the existing file ...\n\n";
		unless (0 == system ("mv -f $tmpfile $tmpfile.OLD") ) {
		    print STDOUT "[ERROR '$program_name'] 'system ('mv ...') FAILED on the temporary file '$tmpfile': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n"; 
		    die ;
		}
	    }
	}		# End of the foreach-loop over temporary files #


		# Prepare a temporary input (virtual) alignment file #

	my $fh_otmp_ref = open_output_file ($tmp_inaln_ref, 0);
	unless (defined $fh_otmp_ref) {
	    print STDOUT "[ERROR '$program_name'] FAILED to open a temporary input (virtual) alignment file '$tmp_inaln_ref' for the $indx_gp th gap pattern under the $t th tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	my $fh_otmp_rec = open_output_file ($tmp_inaln_rec, 0);
	unless (defined $fh_otmp_rec) {
	    print STDOUT "[ERROR '$program_name'] FAILED to open a temporary input (virtual) alignment file '$tmp_inaln_rec' for the $indx_gp th gap pattern under the $t th tree !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	   # REVISED to handle a 2-sequence alignment via PhyML (on April 12 (Tue), 2011) #
	my $sbct_vseqs = ($sbct_seqs == 2) ? 3 : $sbct_seqs;

	##print $fh_otmp "$sbct_vseqs $ct_nr_clms\n";
	print $fh_otmp_ref "$sbct_vseqs $ref_len1\n"; # DEBUGGED on Aug 10 (Sun), 2014. #
	##print $fh_otmp_ref "$sbct_vseqs $ref_len\n";

	unless ( pr_phylip_aln (@red_seqnames, @red_vmsa_ref, $DIS_LEN, *${$fh_otmp_ref}) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_phylip_aln' FAILED to output a temporary input (virtual) alignment for the $indx_gp th gap pattern of the reference MSA segment under the $t th tree !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}
	close ($fh_otmp_ref);

	print $fh_otmp_rec "$sbct_vseqs $rec_len1\n"; # DEBUGGED on Aug 10 (Sun), 2014. #
	##print $fh_otmp_rec "$sbct_vseqs $rec_len\n";

	unless ( pr_phylip_aln (@red_seqnames, @red_vmsa_rec, $DIS_LEN, *${$fh_otmp_rec}) ) {
	    print STDERR "[ERROR '$program_name'] The subroutine 'pr_phylip_aln' FAILED to output a temporary input (virtual) alignment for the $indx_gp th gap pattern of the reconstructed MSA segment under the $t th tree !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}
	close ($fh_otmp_rec);


		# Prepare a temporary input phylogenetic tree file #

	my $fl_o_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_o_nodeid = 0; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	foreach my $info ([$tmp_intree_ref, 'reference'], [$tmp_intree_rec, 'reconstructed']) {
	    my ($tmp_intree, $type_msa) = @{$info};

	    my $fh_otmp2 = open_output_file ($tmp_intree, 0);
	    unless (defined $fh_otmp2) {
		print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the temporary input tree file '$tmp_intree' !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    unless ( output_tree ($sb_topnode_id, %{$sb_node2ch}, %sb_id2name, %{$sb_br_attr}, *${$fh_otmp2}, $fl_o_bootlabel, $fl_o_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the reduced tree of the $t th tree for the $indx_gp th gap pattern of the $type_msa MSA !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    close ($fh_otmp2);
	}


		# Prepare a file for a custom substitution model and a control file

	if (($seqtype eq 'AA') and ($resfreq eq  'e')) {

	    foreach my $tmp_info ([$tmp_dat_ref, $tmp_ctrl_ref, $tmp_workdir_ref], [$tmp_dat_rec, $tmp_ctrl_rec, $tmp_workdir_rec]) { # Inner foreach-loop (over alignment types) #

		my ($tmp_dat, $tmp_ctrl, $tmp_workdir) = @{$tmp_info};

		unless (0 == system ("cp -p $master_sbst_file $tmp_dat") ) {
	            print STDOUT "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master file for the custom substitution model '$master_sbst_file' to the temporary working directory '$tmp_workdir' : $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
		unless (0 == system ("cp -p $master_ctrl_file $tmp_ctrl") ) {
		   print STDERR "[ERROR '$program_name'] 'system (cp ...)' FAILED to copy the master control file for PhyML '$master_ctrl_file' to the temporary working directory '$tmp_workdir' : $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }		# End of the inner foreach-loop (over alignment types) # 

	} # End of "if (($seqtype eq 'AA') and ($resfreq eq  'e'))"

	##if ($seqtype eq 'NT') {
	# TO BE REVISED in the future for $seqtype eq 'NT' # -- Or it is not necessary any longer ?
	##}

		## Finish making the command-line options. ##

	## $OPTIONS_CLMLK = "--print_site_lnl $OPTIONS_SHARED -o $opt_params_clmlk -b $ct_bootstrap_clmlk"; # A BUG fixed
	#### $OPTIONS_CLMLK = "---print_site_lnl $OPTIONS_SHARED -o $opt_params_clmlk -b $ct_bootstrap_clmlk";
	## $OPTIONS_CLMLK .= " -m custom -f m";
	my $options_clmlk = $OPTIONS_CLMLK;
	(($seqtype eq 'AA') and ($resfreq eq  'e')) and ($options_clmlk .= " < $ctrl_name");

    if ($TEST_LEVEL>2) {
	print STDOUT "\n{ Segment: $segid; Gap pattern index: $indx_gp; Tree no.: $t }\n\n";
	print STDOUT "PhyML options for column-wise likelihood estimation:\n\n";
	print STDOUT "   '$options_clmlk'.\n\n";
    } # End of "if ($TEST_LEVEL>2)"

	    # $set_clmwise_lnlk2_ref/rec[$c] is the log-likelihood of the $c th column (in exactly the same order as in the virtual alignment).
	    # It is undef if the likelihood is zero.
	my @set_clmwise_lnlk2_ref = my @set_clmwise_lnlk2_rec = ();


		# Actually perform the program that calculates the likelihoods of substitution patterns. #

	foreach my $info_alntype ([$tmp_workdir_ref, $tmp_out_ref, $fh_log_sref, \@set_clmwise_lnlk2_ref, \@red_vmsa_ref, 'reference'], [$tmp_workdir_rec, $tmp_out_rec, $fh_log_srec, \@set_clmwise_lnlk2_rec, \@red_vmsa_rec, 'reconstructed']) {  # Big inner foreach-loop (over alignment types) #

	    my ($tmp_workdir, $tmp_out, $fh_log, $ref_set_clmwise_lnlk2, $red_vmsa, $msa_type) = @{$info_alntype};
	    my $ct_nr_clms = length ${$red_vmsa->[0]};

	    unless ($ct_nr_clms>0) {## ADDED for DEBUGGING on Aug 11 (Mon), 2014. ##

		print STDOUT "\n[INFORMATION] NO COLUMN was found in the reduced MSA for the $indx_gp th gap pattern of the $msa_type MSA under the $t th tree ... Thus I will skip performing '$main_bin_clmlk' to calculate column-wise log-likelihoods ...\n\n";
		next;
	    }

	    unless ( chdir $tmp_workdir ) {
		print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$tmp_workdir' to process the $indx_gp th gap pattern of the $msa_type MSA under the $t th tree !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

###=DISABLED for TESTING the input and options of $main_bin_clmlk (Part I)
	    unless (0 == system ("$main_bin_clmlk -i $tmp_inaln_name -u $tmp_intree_name $options_clmlk > $log_name 2>&1") ) {
		print STDOUT "[ERROR '$program_name'] FAILED 'system ('$main_bin_clmlk ...')' FAILED for the $indx_gp th gap pattern of the $msa_type MSA under the tree $t: $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
###=cut ## End of "DISABLED for TESTING the input and options of $main_bin_clmlk (Part I)"

	    unless ( chdir $CURR_WORK_DIR ) {
		print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to MOVE BACK to the original working directory '$CURR_WORK_DIR' when processing the $indx_gp th gap pattern of the $msa_type MSA under the tree $t: !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

###=DISABLED for TESTING the input and options of $main_bin_clmlk (Part II)
	    unless (( -f $tmp_out ) and ( -r $tmp_out )) {
		print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable OUTPUT FILE '$tmp_out' of PAML  for the $indx_gp th gap pattern of the $msa_type MSA under the tree $t !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    # Extract log-likelihoods of the non-redundant columns from the temporary output file,
	    # Then store  them into a 2-dim array.

	    my $fl_log = 1; # = 1 if you want LOG-likelihoods, = 0 if you want RAW likelihoods.
	    # $set_clmwise_lnlk2->[$c] is the log-likelihood of the $c th column (in exactly the same order as in the virtual alignment).
	    # It is undef if the likelihood is zero.
	    #
	    my ($tmpset_clmwise_lnlk2) = fetch_clmwise_likelihoods_phyml ($tmp_out, $fl_log);
	    unless (defined $tmpset_clmwise_lnlk2) {
		print STDOUT "[ERROR '$program_name'] 'fetch_clmwise_likelihoods_phyml' FAILED on the temporary output file '$tmp_out' for the $indx_gp th gap pattern of the $msa_type MSA (under the $t th tree) !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    unless ($ct_nr_clms == @{$tmpset_clmwise_lnlk2}) {
		print STDOUT "[ERROR '$program_name'] The OUTPUT ARRAY of 'fetch_clmwise_likelihoods_phyml' has the SIZE (", scalar (@{$tmpset_clmwise_lnlk2}), ") that CLASHES with the number of non-redundant columns ($ct_nr_clms) for the $indx_gp th gap pattern of the $msa_type MSA (under the $t th tree): $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

	    @{$ref_set_clmwise_lnlk2} = @{$tmpset_clmwise_lnlk2}; # Copy the temporary set of column-wise log-likelihoods to an array outside of the loop. #

	}		# End of the big inner foreach-loop (over alignment types) #

##=for TESTING 'fetch_clmwise_likelihoods_phyml'
	foreach my $info_alntype ([$tmp_workdir_ref, $tmp_out_ref, $fh_log_sref, \@set_clmwise_lnlk2_ref, \@red_vmsa_ref, 'reference'], [$tmp_workdir_rec, $tmp_out_rec, $fh_log_srec, \@set_clmwise_lnlk2_rec, \@red_vmsa_rec, 'reconstructed']) {  # Medium-size inner foreach-loop (over alignment types) #

	    my ($tmp_workdir, $tmp_out, $fh_log, $ref_set_clmwise_lnlk2, $red_vmsa, $msa_type) = @{$info_alntype};
	    my $ct_nr_clms = length ${$red_vmsa->[0]};

	    print $fh_log "\n[ Log likelihoods of the columns with the $indx_gp th gap pattern (full) of the $msa_type MSA under the $t th candidate tree (Total: NA, #{nr columns} = $ct_nr_clms) ]\n\n";
	    print $fh_log join ("\t", '#Indx_clm', 'ln_lk', 'column'), "\n\n";

	    my @cp_vmsa = ();
	    foreach my $seq (@{$red_vmsa}) { my $cp_seq = ${$seq}; push @cp_vmsa, \$cp_seq; }

	    for (my $l = 0; $l<$ct_nr_clms; $l++) { # Auxiliary for-loop #
		my $ln_lk = $ref_set_clmwise_lnlk2->[$l];
		##my $column = ${$red_nr_clms[$l]};
		my $column = '';
		foreach my $seq (@cp_vmsa) { $column .= substr (${$seq}, 0, 1, ''); }

		print $fh_log join ("\t", $l, $ln_lk, "'$column'"), "\n";
	    }			# End of the auxiliary for-loop #
	    print $fh_log "\n\n";

	}		# End of the medium-size inner foreach-loop (over alignment types) #
##=cut


	##my @set_refseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reference segment under the $t th tree candidate.
	##my @set_recseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reconstructed segment under the $t th tree candidate.

	my $sb_refseg_score = my $sb_recseg_score = 0;

	my $ct_nr_clms_ref = @set_clmwise_lnlk2_ref;
	for (my $indx_clm=0; $indx_clm<$ct_nr_clms_ref; $indx_clm++) {
	    my $clm_lnlk = $set_clmwise_lnlk2_ref[$indx_clm];
	    unless (defined $clm_lnlk) {
		$sb_refseg_score = undef;
		last;
	    }
	    $sb_refseg_score += $clm_lnlk;
	}
	if (defined $sb_refseg_score) {
	    (defined $set_refseg_scores_sb[$t]) and ($set_refseg_scores_sb[$t] += $sb_refseg_score);
	} else {
	    $set_refseg_scores_sb[$t] = undef;
	}

	my $ct_nr_clms_rec = @set_clmwise_lnlk2_rec;
	for (my $indx_clm=0; $indx_clm<$ct_nr_clms_rec; $indx_clm++) {
	    my $clm_lnlk = $set_clmwise_lnlk2_rec[$indx_clm];
	    unless (defined $clm_lnlk) {
		$sb_recseg_score = undef;
		last;
	    }
	    $sb_recseg_score += $clm_lnlk;
	}
	if (defined $sb_recseg_score) {
	    (defined $set_recseg_scores_sb[$t]) and ($set_recseg_scores_sb[$t] += $sb_recseg_score);
	} else {
	    $set_recseg_scores_sb[$t] = undef;
	}

###=cut ### END of "DISABLED for TESTING the input and options of $main_bin_clmlk (Part II)"

		# Remove the temporary working directory if you want #

	if ($FLG_RMV_TMPOUT) {
	    foreach my $tmp_workdir ($tmp_workdir_ref, $tmp_workdir_rec) {
		unless (0 == system ("rm -rf $tmp_workdir") ) {
		    print STDOUT "[ERROR '$program_name'] 'system ('rm -rf {outdir}')' FAILED to remove the temporary working directory '$tmp_workdir': $? !\n";
		    print STDOUT "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}

    ## END of "Calculate the log-likelihoods of gap patterns and substitution patterns in the reference/reconstructed alignment segments"

     }			# End of the 1st middle for-loop (over gap patterns) #

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (III)"

  }		# End of the 1st outer-middle for-loop (over candidate trees) #


=DISABLED because they didn't appear working as they are supposed to be.
	foreach my $fh_int ($fh_clmlk_ref, $fh_clmlk_rec) { close ($fh_int); }
=cut

			# MODIFIED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #
	foreach my $fh_log ($fh_log_gref, $fh_log_grec) { close ($fh_log); }
	##foreach my $fh_log ($fh_log_sref, $fh_log_srec, $fh_log_gref, $fh_log_grec) { close ($fh_log); }

	### next; ### For TESTING the Step (ii) Main Process A ###

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (V)

		# REMOVE the grand-parent directory of temporary working directories #
	if ($FLG_RMV_TMPOUT) {
	    foreach my $grpa_tmp_workdir ($grpa_tmp_workdir_ref, $grpa_tmp_workdir_rec) {
		unless (0 == system ("rm -rf $grpa_tmp_workdir") ) {
		    print STDOUT "[ERROR '$program_name'] 'system ('rm -rf {outdir}')' FAILED to remove the grand-parent of temporary working directories '$grpa_tmp_workdir': $? !\n";
		    print STDERR "[ERROR '$program_name'] $!\n\n";
		    die ;
		}
	    }
	}
=cut


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VI)

		### Step (iii) Main Process B:
		### Calculate the total ML scores for the reference & reconstructed MSA segments,
		### under each fixed candidate tree as well as averaged over candidate trees.


	## @set_refseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reference segment under the $t th tree candidate.
	## @set_recseg_scores_sb = (); # $set_seg_scores_sb[$t] is the ML score of the substitution pattern of the reconstructed segment under the $t th tree candidate.
	## @set_refseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reference segment under the $t th tree candidate.
	## @set_recseg_scores_gp = (); # $set_seg_scores_gp[$t] is the ML score of the gap pattern of the reconstructed segment under the $t th tree candidate.

	my @set_refseg_def_gpscores = (); # lists only the defined (weighted) gap scores of the reference MSA segment under particular candidate trees.
	my @set_recseg_def_gpscores = (); # lists only the defined (weighted) gap scores of the reconstructed MSA segment under particular candidate trees.
	my @totscores_ud_fxd_trees_ref = (); # lists the defined (weighted) total scores (= gap score + substitution score) of the reference MSA segment under particular  candidate trees.
	my @totscores_ud_fxd_trees_rec = (); # lists the defined (weighted) total score (= gap score + substitution score) of the reference MSA segment under particular candidate trees.

	my $avct_indels_ref = my $avct_indels_rec = 0;
	my $avct_null_clms_ref = my $avct_null_clms_rec = 0;

    if ($TEST_LEVEL>1) {
	print STDOUT "\n<<< Scores under each of the $ct_cand_trees candidate trees: >>>\n\n";

		# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	print STDOUT join ("\t", '#Indx_tree', 'Ref/Rec', 'ln(Akaike_wgt)', 'tot_score(wgtd)', 'gp_score(wgtd)', 'tot_score(unwgtd)', 'gp_score(unwgtd)', 'sb_score(unwgtd)'), "\n\n";
	##print STDOUT join ("\t", '#Indx_tree', 'Ref/Rec', 'ln(Akaike_wgt)', '#{indels}(unwgtd)', 'tot_score(wgtd)', 'gp_score(wgtd)', 'tot_score(unwgtd)', 'gp_score(unwgtd)', 'sb_score(unwgtd)'), "\n\n";
    }

		## ADDED on Sep 29 (Sat), 2012. ##
	my $ct_undef_tot_score_ref = my $ct_undef_tot_score_rec = my $ct_undef_gp_score_ref = my $ct_undef_gp_score_rec = 0;

	my @indices_defined_totscore_ref = my @indices_defined_totscore_rec = () ; # ADDED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

	for (my $t=0; $t<$ct_cand_trees; $t++) { # 2nd outer-middle for-loop (over candidate trees) #

	    my $ln_awgt = $ln_akaike_wgts->[$t];
	    my $sb_score_ref = $set_refseg_scores_sb[$t];
	    my $sb_score_rec = $set_recseg_scores_sb[$t];
	    my $gp_score_ref = $set_refseg_scores_gp[$t];
	    my $gp_score_rec = $set_recseg_scores_gp[$t];

#=for the predecessor 'compare_ref_vs_rec_compllimment.pl'.
	    my $ct_indels_ref = $set_ct_indels_ref[$t];
	    my $ct_indels_rec = $set_ct_indels_rec[$t];
#=cut

	    my $ct_null_clms_ref = $set_ct_null_clms_ref[$t];
	    my $ct_null_clms_rec = $set_ct_null_clms_rec[$t];

	    my $tot_score_ref = ((defined $sb_score_ref) and (defined $gp_score_ref)) ? $sb_score_ref + $gp_score_ref : 'undef' ;
	    my $tot_score_rec = ((defined $sb_score_rec) and (defined $gp_score_rec)) ? $sb_score_rec + $gp_score_rec : 'undef' ;

		# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	    foreach my $score (\$ln_awgt, \$sb_score_ref, \$sb_score_rec, \$gp_score_ref, \$gp_score_rec) { (defined ${$score}) or (${$score} = 'undef'); }
	    ##foreach my $score (\$ln_awgt, \$sb_score_ref, \$sb_score_rec, \$gp_score_ref, \$gp_score_rec, \$ct_indels_ref, \$ct_indels_rec) { (defined ${$score}) or (${$score} = 'undef'); }

	    if ($ln_awgt eq 'undef') {
    if ($TEST_LEVEL>1) {
			# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
		print STDOUT join ("\t", $t, 'Ref', $ln_awgt, 'undef', 'undef', $tot_score_ref, $gp_score_ref, $sb_score_ref), "\n";
		print STDOUT join ("\t", $t, 'Ref', $ln_awgt, 'undef', 'undef', $tot_score_rec, $gp_score_rec, $sb_score_rec), "\n";
		##print STDOUT join ("\t", $t, 'Ref', $ln_awgt, $ct_indels_ref, 'undef', 'undef', $tot_score_ref, $gp_score_ref, $sb_score_ref), "\n";
		##print STDOUT join ("\t", $t, 'Ref', $ln_awgt, $ct_indels_rec, 'undef', 'undef', $tot_score_rec, $gp_score_rec, $sb_score_rec), "\n";
    } ## END of "if ($TEST_LEVEL>1)"
		next;
	    }

		## ADDED on Sep 29 (Sat), 2012. ##
	    ($tot_score_ref eq 'undef') and ($ct_undef_tot_score_ref++);
	    ($tot_score_rec eq 'undef') and ($ct_undef_tot_score_rec++);
	    ($gp_score_ref eq 'undef') and ($ct_undef_gp_score_ref++);
	    ($gp_score_rec eq 'undef') and ($ct_undef_gp_score_rec++);

	    my $wgtd_tot_score_ref = ($tot_score_ref eq 'undef') ? 'undef' : $tot_score_ref + $ln_awgt;
	    my $wgtd_tot_score_rec = ($tot_score_rec eq 'undef') ? 'undef' : $tot_score_rec + $ln_awgt;
	    my $wgtd_gp_score_ref = ($gp_score_ref eq 'undef') ? 'undef' : $gp_score_ref + $ln_awgt;
	    my $wgtd_gp_score_rec = ($gp_score_rec eq 'undef') ? 'undef' : $gp_score_rec + $ln_awgt;

    if ($TEST_LEVEL>1) {
			# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	    print STDOUT join ("\t", $t, 'Ref', $ln_awgt, $wgtd_tot_score_ref, $wgtd_gp_score_ref, $tot_score_ref, $gp_score_ref, $sb_score_ref), "\n";
	    print STDOUT join ("\t", $t, 'Rec', $ln_awgt, $wgtd_tot_score_rec, $wgtd_gp_score_rec, $tot_score_rec, $gp_score_rec, $sb_score_rec), "\n";
	    ##print STDOUT join ("\t", $t, 'Ref', $ln_awgt, $ct_indels_ref, $wgtd_tot_score_ref, $wgtd_gp_score_ref, $tot_score_ref, $gp_score_ref, $sb_score_ref), "\n";
	    ##print STDOUT join ("\t", $t, 'Rec', $ln_awgt, $ct_indels_rec, $wgtd_tot_score_rec, $wgtd_gp_score_rec, $tot_score_rec, $gp_score_rec, $sb_score_rec), "\n";
    } ## END of "if ($TEST_LEVEL>1)"

	    ($wgtd_gp_score_ref eq 'undef')  or (push @set_refseg_def_gpscores, $wgtd_gp_score_ref);
	    ($wgtd_gp_score_rec eq 'undef')  or (push @set_recseg_def_gpscores, $wgtd_gp_score_rec);
	    ($wgtd_tot_score_ref eq 'undef') or (push @totscores_ud_fxd_trees_ref, $wgtd_tot_score_ref);
	    ($wgtd_tot_score_rec eq 'undef') or (push @totscores_ud_fxd_trees_rec, $wgtd_tot_score_rec);

			# ADDED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	    ($wgtd_tot_score_ref eq 'undef') or (push @indices_defined_totscore_ref, $t);
	    ($wgtd_tot_score_rec eq 'undef') or (push @indices_defined_totscore_rec, $t);

	    my $awgt = exp ($ln_awgt);

#=for the predecessor 'compare_ref_vs_rec_compllimment.pl'.
		# MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). #
	    unless ($avct_indels_ref eq 'N/A') {
		(($ct_indels_ref eq 'undef') or ($ct_indels_ref eq 'N/A')) ? ($avct_indels_ref = 'N/A') : ($avct_indels_ref += $ct_indels_ref * $awgt);
	    }
	    unless ($avct_indels_rec eq 'N/A') {
		(($ct_indels_rec eq 'undef') or ($ct_indels_rec eq 'N/A')) ? ($avct_indels_rec = 'N/A') : ($avct_indels_rec += $ct_indels_rec * $awgt);
	    }
	    ##($ct_indels_ref eq 'undef') or ($avct_indels_ref += $ct_indels_ref * $awgt);
	    ##($ct_indels_rec eq 'undef') or ($avct_indels_rec += $ct_indels_rec * $awgt);
#=cut

	    $avct_null_clms_ref += $ct_null_clms_ref * $awgt;
	    $avct_null_clms_rec += $ct_null_clms_rec * $awgt;

	}		# End of the 2nd outer-middle for-loop (over candidate trees) #

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VI)"


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VII)

		# Calculating average ML scores (i.e. the approximate Bayesian score) #
		# for the total, as well as for the gap pattern only,
		# of the reference and reconstructed MSA segments.

	my $apprx_bayes_totscore_ref = my $apprx_bayes_totscore_rec = my $apprx_bayes_gpscore_ref = my $apprx_bayes_gpscore_rec = 'undef';

		## REVISED on Sep 29 (Sat), 2012. ##
		## => Further MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##
	if ((@totscores_ud_fxd_trees_ref>0) and ($ct_undef_tot_score_ref==0) and ($avct_indels_ref ne 'N/A') and ($avct_indels_rec ne 'N/A') ) {
	##if ((@totscores_ud_fxd_trees_ref>0) and ($ct_undef_tot_score_ref==0)) {...}
	####if (@totscores_ud_fxd_trees_ref>0) {...}
	    $apprx_bayes_totscore_ref = log_sum_exp (@totscores_ud_fxd_trees_ref);
	    unless (defined $apprx_bayes_totscore_ref) {
		print STDOUT "\n[ERROR '$program_name'] 'log_sum_exp' FAILED to calculate the approximate Bayesian total score of the reference MSA segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}

		## REVISED on Sep 29 (Sat), 2012. ##
		## => Further MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##
	if ((@totscores_ud_fxd_trees_rec>0) and ($ct_undef_tot_score_rec==0) and ($avct_indels_ref ne 'N/A') and ($avct_indels_rec ne 'N/A') ) {
	##if ((@totscores_ud_fxd_trees_rec>0) and ($ct_undef_tot_score_rec==0)) {...}
	####if (@totscores_ud_fxd_trees_rec>0) {...}
	    $apprx_bayes_totscore_rec = log_sum_exp (@totscores_ud_fxd_trees_rec);
	    unless (defined $apprx_bayes_totscore_rec) {
		print STDOUT "\n[ERROR '$program_name'] 'log_sum_exp' FAILED to calculate the approximate Bayesian total score of the reconstructed MSA segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VII)"


=For the predecessor 'compare_ref_vs_rec_compllimment.pl'.

		## REVISED on Sep 29 (Sat), 2012. ##
		## => Further MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##
	if ((@set_refseg_def_gpscores>0) and ($ct_undef_gp_score_ref==0) and ($avct_indels_ref ne 'N/A') and ($avct_indels_rec ne 'N/A') ) {
	##if ((@set_refseg_def_gpscores>0) and ($ct_undef_gp_score_ref==0)) {...}
	####if (@set_refseg_def_gpscores>0) {...}
	    $apprx_bayes_gpscore_ref = log_sum_exp (@set_refseg_def_gpscores);
	    unless (defined $apprx_bayes_gpscore_ref) {
		print STDOUT "\n[ERROR '$program_name'] 'log_sum_exp' FAILED to calculate the approximate Bayesian gap score of the reference MSA segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}

		## REVISED on Sep 29 (Sat), 2012. ##
		## => Further MODIFIED in creation of 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl' (on Aug 5 (Tue), 2014). ##
	if ((@set_recseg_def_gpscores>0) and ($ct_undef_gp_score_rec==0) and ($avct_indels_ref ne 'N/A') and ($avct_indels_rec ne 'N/A') ) {
	##if ((@set_recseg_def_gpscores>0) and ($ct_undef_gp_score_rec==0)) {...}
	####if (@set_recseg_def_gpscores>0) {...}
	    $apprx_bayes_gpscore_rec = log_sum_exp (@set_recseg_def_gpscores);
	    unless (defined $apprx_bayes_gpscore_rec) {
		print STDOUT "\n[ERROR '$program_name'] 'log_sum_exp' FAILED to calculate the approximate Bayesian gap score of the reconstructed MSA segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}

=cut # END of "For the predecessor 'compare_ref_vs_rec_compllimment.pl'."

		## Another MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. ##

	my $avtotct_ins_ref = my $avtotct_ins_rec = my $avtotct_dels_ref = my $avtotct_dels_rec = 0;
	my $ct_defined_ref = my $ct_defined_rec = 0;

		### RESTART FROM HERE (II) ##

	for (my $t = 0; $t<$ct_cand_trees; $t++) { # Trailing outer-middle for-loop (over candidate trees). #

	    my $totct_ins_ref = $set_avct_ins_ref[$t];
	    my $totct_ins_rec = $set_avct_ins_rec[$t];
	    my $totct_dels_ref = $set_avct_dels_ref[$t];
	    my $totct_dels_rec = $set_avct_dels_rec[$t];

	    if (defined $totct_ins_ref) {
		$ct_defined_ref++;
		$avtotct_ins_ref += $totct_ins_ref;
		$avtotct_dels_ref += $totct_dels_ref;
	    }

	    if (defined $totct_ins_rec) {
		$ct_defined_rec++;
		$avtotct_ins_rec += $totct_ins_rec;
		$avtotct_dels_rec += $totct_dels_rec;
	    }

	}				# End of the trailing outer-middle for-loop (over candidate trees). #

	if ($ct_defined_ref>0) {
	    $avtotct_ins_ref /= $ct_defined_ref;
	    $avtotct_dels_ref /= $ct_defined_ref;
	} else {
	    $avtotct_ins_ref = $avtotct_dels_ref = 'N/A';
	}

	if ($ct_defined_rec>0) {
	    $avtotct_ins_rec /= $ct_defined_rec;
	    $avtotct_dels_rec /= $ct_defined_rec;
	} else {
	    $avtotct_ins_rec = $avtotct_dels_rec = 'N/A';
	}


=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VIII)

		## ADDED as a MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.pl'. ##

	my $avavct_ins_ref = my $avavct_ins_rec = my $avavct_dels_ref = my $avavct_dels_rec = 'N/A';

	if ($apprx_bayes_totscore_ref ne 'undef') {

	    $avavct_ins_ref = $avavct_dels_ref = 0;

	    for (my $i=0; $i<@totscores_ud_fxd_trees_ref; $i++) { # 1st trailing for-loop (over trees with defined total score of reference segment).
		my $wgt = exp ($totscores_ud_fxd_trees_ref[$i] - $apprx_bayes_totscore_ref);
		my $indx = $indices_defined_totscore_ref[$i];

		my $avct_ins_ref = $set_avct_ins_ref[$indx];
		if (defined $avct_ins_ref) { # This means that the average count of insertions can be calculated under this tree.
		   ($avavct_ins_ref ne 'N/A') and ($avavct_ins_ref += $wgt * $avct_ins_ref);
		} else {
		    $avavct_ins_ref = 'N/A';
		}

		my $avct_dels_ref = $set_avct_dels_ref[$indx];
		if (defined $avct_dels_ref) { # This means that the average count of deletions can be calculated under this tree.
		    ($avavct_dels_ref ne 'N/A') and ($avavct_dels_ref += $wgt * $avct_dels_ref);
		} else {
		    $avavct_dels_ref = 'N/A';
		}

	    }				# End of the 1st trailing for-loop (over trees with defined total score of reference segment). #

	}

	if ($apprx_bayes_totscore_rec ne 'undef') {

	    $avavct_ins_rec = $avavct_dels_rec = 0;

	    for (my $i=0; $i<@totscores_ud_fxd_trees_rec; $i++) { # 2nd trailing for-loop (over trees with defined total score of reconstructed segment).
		my $wgt = exp ($totscores_ud_fxd_trees_rec[$i] - $apprx_bayes_totscore_rec);
		my $indx = $indices_defined_totscore_rec[$i];

		my $avct_ins_rec = $set_avct_ins_rec[$indx];
		if (defined $avct_ins_rec) { # This means that the average count of insertions can be calculated under this tree.
		   ($avavct_ins_rec ne 'N/A') and ($avavct_ins_rec += $wgt * $avct_ins_rec);
		} else {
		    $avavct_ins_rec = 'N/A';
		}

		my $avct_dels_rec = $set_avct_dels_rec[$indx];
		if (defined $avct_dels_rec) { # This means that the average count of deletions can be calculated under this tree.
		    ($avavct_dels_rec ne 'N/A') and ($avavct_dels_rec += $wgt * $avct_dels_rec);
		} else {
		    $avavct_dels_rec = 'N/A';
		}
	    }				# End of the 2nd trailing for-loop (over trees with defined total score of reconstructed segment). #

	}

	# @indices_defined_totscore_ref = my @indices_defined_totscore_rec = () ; # ADDED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

		## END of "ADDED as a MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.pl'." ##


    if ($TEST_LEVEL>1) {
	print STDERR "\n{Approximate Bayes Scores}\n\n";
	print STDOUT "{Total}_ref = $apprx_bayes_totscore_ref ,\n";
	print STDOUT "{Total}_rec = $apprx_bayes_totscore_rec .\n\n";

			# MODIFIED in creation of 'compare_ref_vs_rec_indel_cts.pl'. #
	print STDOUT "AvAv#{insertions}_ref = $avavct_ins_ref ,\n";
	print STDOUT "AvAv#{insertions}_rec = $avavct_ins_rec .\n\n";
	print STDOUT "AvAv#{deletions}_ref  = $avavct_dels_ref ,\n";
	print STDOUT "AvAv#{deletions}_rec  = $avavct_dels_rec .\n\n";

#=for the predecessor 'compare_ref_vs_rec_compllimment.pl'.
	print STDOUT "{Gap-pattern}_ref = $apprx_bayes_gpscore_ref ,\n";
	print STDOUT "{Gap-pattern}_rec = $apprx_bayes_gpscore_rec .\n\n";
	print STDOUT "Av#{indels}_ref = $avct_indels_ref ,\n";
	print STDOUT "Av#{indels}_rec = $avct_indels_rec .\n\n";
#=cut
	print STDOUT "Av#{null columns}_ref = $avct_null_clms_ref ,\n";
	print STDOUT "Av#{null columns}_rec = $avct_null_clms_rec .\n\n";
    } # END of "if ($TEST_LEVEL>1)"

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (VIII)"

	### RESTART FROM HERE ###

	print $fh_out "< Segment ID: $segid >\n\n";

	print $fh_out join ("\t", 'Ref/Rec', 'Tot#{insertions}_Dpars', 'Tot#{deletions}_Dpars'), "\n\n";

	print $fh_out join ("\t", 'Ref', $avtotct_ins_ref, $avtotct_dels_ref), "\n" ;
	print $fh_out join ("\t", 'Rec', $avtotct_ins_rec, $avtotct_dels_rec), "\n" ;

=DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (IX)

			# A MAJOR MODIFICATION in creation of 'compare_ref_vs_rec_indel_cts.pl'. #

	print $fh_out join ("\t", 'Ref/Rec', 'Total_Score',  'AvAv#{insertions}', 'AvAv#{deletions}', 'Av#{null_columns}'), "\n\n";

	print $fh_out join ("\t", 'Ref', $apprx_bayes_totscore_ref, $avavct_ins_ref, $avavct_dels_ref, $avct_null_clms_ref), "\n";
	print $fh_out join ("\t", 'Rec', $apprx_bayes_totscore_rec, $avavct_ins_rec, $avavct_dels_rec, $avct_null_clms_rec), "\n";

=cut # END of "DISABLED in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. (IX)"

=for the predecessor 'compare_ref_vs_rec_compllimment.pl'.
	print $fh_out join ("\t", 'Ref/Rec', 'Total_Score', 'Gap_Score', 'Av#{indel_events}', 'Av#{null_columns}'), "\n\n";

	print $fh_out join ("\t", 'Ref', $apprx_bayes_totscore_ref, $apprx_bayes_gpscore_ref, $avct_indels_ref, $avct_null_clms_ref), "\n";
	print $fh_out join ("\t", 'Rec', $apprx_bayes_totscore_rec, $apprx_bayes_gpscore_rec, $avct_indels_rec, $avct_null_clms_rec), "\n";
=cut
	
	print $fh_out "\n\\\\\n\n";

	print STDOUT "\n --- Finished analysis for the segment $segid ! ---\n\n";

	## close ($fh_log);

	$ct_succ_segs++;

=for the predecessor 'compare_ref_vs_rec_abs_scoring_spt_odr.PhyML.pl'
	$ref_coord_beg += $ref_len;
	$rec_coord_beg += $rec_len;
=cut

}			# End of the main outer for-loop (over MSA segments) #

close ($fh_out);

print STDOUT "\n\n === All the Analyses FINISHED !! ===\n\n";

print STDOUT "#{Segments skipped} = $ct_skipped_segs .\n";
print STDOUT "#{Segments whose ComplLiMMent calculations skipped} = $ct_skipped_segs2 .\n";
print STDOUT "#{Anomalous segments} = $ct_anom_segs .\n";
print STDOUT "#{Segments used as inputs} = $ct_in_segs .\n";
print STDOUT "#{Segments successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Segments failed to be processed} = $ct_fail_segs .\n\n";

unless (0 == $ct_fail_segs) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail_segs segments !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


	# DEFINITIONS of the subroutines added in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'. #

# cal_init_psm_cands_for_sgl_msa:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of a single input MSA.
#	It also outputs the (non-redundant) list of gap-patterns and the information on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSA.
#	It was created by modifying the subroutine 'cal_init_psm_cands_for_twin_msas'.
#	[CREATED on Oct 9 (Fri), 2015]
#	[=> IMPORTED from 'Ingredients/classify_msa_errors_via_mblks.2.pl', in creation of 'compare_ref_vs_rec_Dp_indel_cts.hs.pl'.]
#
# cal_init_psm_cands_for_twin_msas:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of the reference and reconstructed MSAs,
#	it also outputs the (non-redundant) list of gap-patterns and the informaion on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSAs.
#	(CREATED on Oct 19 (Sun), 2014)
#
# INPUT: (\%node2ch, \%node2pa, \%node2depth, \%node2ct_exoffs, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres, \@info_gpatterns0), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#
#	%node2depth = ($node_id => $absolute_depth, ...) is for all the nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) is for all the nodes;
#
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA (reference/reconstructed)
#	(NOTE that these names must exactly match those used in %name2id);
#
#	${$in_msa$k]} is the $k th sequence embedded in the input MSA;
#
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#	@info_gpatterns0 = (\@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0) gives information on the gap-patterns found in the reference & reconstructed MSAs, and the Dollo parsimonious scenarios yielding these gap-patterns, where
#	${$set_gpatterns0[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
#	%{$set_node2dparsstat0[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
#	%{$set_br2change0[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
#
# OUTPUT: (\@info_init_psm_scn, \@info_gpatterns), where
#
#	@info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$init_totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@info_gpatterns =  (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change) gives information on the gap-patterns found in the input MSA, and the Dollo parsimonious scenarios yielding these gap-patterns, (on top of those initially given,) where
#
#	@set_gpatterns follows the same format as @info_gpatterns0 in the input,
#	@set_node2dparsstat follows the same format as @set_node2dparsstat0 in the input,
#	@set_br2change follows the same format as @set_br2change0 in the input.
#
##sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$) { ....}
##	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;
#
#		... MODIFIED on Oct 14 (Wed), 2015...
#
sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@) {

	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres, $info_gpatterns0) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};

    (print STDOUT "<< Subroutine 'cal_init_psm_cands_for_sgl_msa' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL_cipc2>0);

		# Main Process 1: Extract gapped segments and their constituent gap-pattern blocks from the input MSA. #
	

			# MODIFIED on Oct 14 (Mon), 2015. #
	my ($set_gpatterns0, $set_node2dparsstat0, $set_br2change0) = @{$info_gpatterns0}; # Initial information of the gap-patterns and their Dollo parsimonious scenarios. #

	my %gpattern2indx;
	my $ct_gpatterns = my $ct_gpatterns0 = @{$set_gpatterns0};
	for (my $i=0; $i<$ct_gpatterns0; $i++) { $gpattern2indx{${$set_gpatterns0->[$i]}} = $i; }
	##my $ct_gpatterns = 0;

	my $set_cls_gblocks = extract_cls_gblocks (@{$in_msa}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'extract_cls_gblocks' FAILED: $! !\n\n";
	    return (undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended (#{gap clusters} = $ct_cls) ...\n\n") if ($TEST_LEVEL_cipc2>1);

	# substr (${$sec_msa[$k]}, $indx_gp, 1) is the presence/absence state of the $k th sequence in the $indx_gp th gap pattern.
	# ${$set_gpatterns[$indx_gp]} is the string made by concatenating the states in the $indx_gp th gap pattern.
	#
	my ($sec_msa, $set_gpatterns) = msa_rep_gpatterns (%gpattern2indx, length ($char_abs), $ct_seqs);
	unless (defined $sec_msa) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'msa_rep_gpatterns' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n --- 'msa_rep_gpatterns' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);


		# Scan the information of gapped segments for screening those with gapped segment(s) and/or gap(s) that are TOO LONG. #

	foreach my $cls_gblk (@{$set_cls_gblocks}) { # Screening foreach-loop (over gapped segments in the input MSA). #

	    my $ct_gblks = @{$cls_gblk};
	    my $start_cls = $cls_gblk->[0][2];
	    my $end_cls = $cls_gblk->[$ct_gblks-1][3];
	    my $cl_len = $end_cls - $start_cls+1;

	    unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found a cluster of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this input MSA: $! !\n\n";
		return ([], [], []);
	    }

	    my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gblk}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	    unless (defined $list_long_gaps) {
		print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'scan_gbcls_for_long_gaps' FAILED: $! !\n\n";
		return (undef, undef, undef);
	    }
	    unless (@{$list_long_gaps} == 0) {
		my $ct_long_gaps = @{$list_long_gaps};
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this input MSA: $! !\n\n"; 
		return ([], [], []);
	    }

	}		# End of the screening foreach-loop (over gapped segments in reference & reconstructed MSAs). #


		## Main Process 2: Calculate the Dollo parsimonious indel scenario for each gap-pattern. ##

 	my ($set_node2dparsstat, $set_br2change) = ([], []); # ADDED on Oct 14 (Wed), 2015. #
    if ($ct_gpatterns>$ct_gpatterns0) { # ADDED on Oct 14 (Wed), 2015. #

		# MODIFIED on Oct 14 (Wed), 2015. #

	foreach my $sec_seq (@{$sec_msa}) { substr (${$sec_seq}, 0, $ct_gpatterns0 * length ($char_abs), ''); } # Remove the first $ct_gpatterns0 gap-patterns, whose Dollo parsimonious scenarios were already calculated.

	my ($set_node2dparsstat1, $set_br2change1) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	##my ($set_node2dparsstat, $set_br2change) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	unless (defined $set_node2dparsstat1) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'map_parsim_states_dollo_msa' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT " --- 'map_parsim_states_dollo_msa' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);

		# ADDED on Oct 14 (Wed), 2015. #

	@{$set_node2dparsstat} = (@{$set_node2dparsstat0}, @{$set_node2dparsstat1});
	@{$set_br2change} = (@{$set_br2change0}, @{$set_br2change1});

   } else {
		# In this case, the initial (input) set exhausts the gap-patterns in the MSA. #
		# (ADDED on Oct 14 (Wed), 2015.) #

	$set_node2dparsstat = $set_node2dparsstat0;
	$set_br2change = $set_br2change0;

   } # END of "if ($ct_gpatterns>$ct_gpatterns0) {...} else {...}" # ADDED on Oct 14 (Wed), 2015. #


	my @info_gpatterns = ($set_gpatterns, $set_node2dparsstat, $set_br2change);


		## Main Process 3: Calculate the initial candidate of the parsimonious indel scenario resulting in the gap-configurations of each gapped segment in the input MSA. ##

			# Create a set of all nodes.
	my @set_all_nodes = sort keys %{$node2pa};
	unshift @set_all_nodes, $topnode_id;

			# Initialize the output. #
	my @info_init_psm_scn = ($set_cls_gblocks, [], []);

		# @info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the input MSA gap configuration, where

	my $indx_indx_gp = 0; # Index to specify where the index of the gap-pattern is stored.

	foreach my $info_in_out (
		['input MSA', \@info_init_psm_scn]) { # Main outer foreach-loop (for the input MSA). #

	    my ($title_msa, $info_init_psm_scn) = @{$info_in_out};

	    my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};

	    my $ct_cls = @{$set_cls_gblocks};

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

		my $blocks_in_cls = $set_cls_gblocks->[$cl];

		my ($ct_indels0, $node2blwise_gstat0, $br2indels0, $list_blank_blks) = create_initial_set_indels_spt_odr ($topnode_id, %{$node2pa}, %{$node2ch}, @set_all_nodes, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change});
		unless (defined $ct_indels0) {
		    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'create_initial_set_indels' FAILED on the $cl th cluster of '$title_msa': $! !\n\n";
		    return (undef, undef, undef);
		}
    (print STDOUT "\n --- Subroutine 'create_initial_set_indels_spt_odr' Ended (with #{initial indels} = $ct_indels0) ...\n\n") if ($TEST_LEVEL_cipc2>1);

		$init_totcts_indels_for_cls->[$cl] = $ct_indels0;
		$init_psm_scn_for_cls->[$cl] = [$br2indels0, $node2blwise_gstat0];

	# @{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 

	    }				# End of the main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

	}					# End of the main outer foreach-loop (for the input MSA). #

    (print STDOUT "<< END of the subroutine 'cal_init_psm_cand_for_sgl_msa' />>\n\n") if ($TEST_LEVEL_cipc2>0);

	return (\@info_init_psm_scn, \@info_gpatterns);

}	# End of the subroutine 'cal_init_psm_cands_for_sgl_msa'. #



	### DEFINITIONS OF THE NEW SUBROUTINES (ADDED since Aug 26 (Sun), 2012) ###

# readin_list_segs:
#	This subroutine reads in a list of segments.
#
# INPUT: (\$inlist, \%label_sattr2indx), where
#	$inlist is the path to the input file.
#	%label_sattr2indx = ($label_seg_attr => $indx, ...)
#	with $indx being the index in the output list.
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = ($seg_ID, @attributes)
#	is the information on the $s th segment, with
#	$seg_ID being the ID, and
#	@attributes storing the attributes of the segment in the same order as in @LABELS_ATTR.
#
sub readin_list_segs (\$\%) {

	my ($inlist, $label_sattr2indx) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head1 = <$fh_tmpin>;
	my ($ct_seqs)  = ($head1 =~ m/^\#Sequences\s+(\d+)/);
	my $head2 = <$fh_tmpin>;
	my ($ct_segs) = ($head2 =~ m/^\#Segments\s+(\d+)/);
	unless ((defined $ct_seqs) and (defined $ct_segs)) {
	    (defined $ct_seqs) or ($ct_seqs = 'undef');
	    (defined $ct_segs) or ($ct_segs = 'undef');
	    print STDOUT "[ERROR 'readin_list_segs'] ILLEGAL HEADER: (\$ct_seqs, \$ct_segs) = ($ct_seqs, $ct_segs): $! !\n";
	    return undef;
	}

	my @list_segments = ();

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    my @flds = split /\s+/, $in;   

	    my $segid = shift @flds;
	    $segid =~ s/^seg\#//;
	    my @seg_info = ($segid);

	    while (my $label = shift @flds) { # Inner while-loop #
		$label =~ s/\:$//;
		my $indx = $label_sattr2indx->{$label};
		my $attr = shift @flds;
	 	unless (defined $indx) {
		    print STDERR "[CAUTION 'readin_list_segs'] NO INDEX corresponds to the attribute-label '$label' (with value '$attr') in the segment '$segid': $! !\n";
		    next;
		}
		$seg_info[$indx] = $attr; 
	    }		# End of the inner while-loop #

	    push @list_segments, \@seg_info;

	}		# End of the main while-loop #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs' #


# readin_list_segs2:
#	This subroutine reads in a list of segments in a format different from the input for 'readin_list_segs'.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# INPUT: (\$inlist, \@labels_seg_attr), where
#	$inlist is the path to the input file.
#	@labels_seg_attr stores the labels of the fields for the segment attributes (including the ID (or index)).
#
# OUTPUT: \@list_segments, where
#	@{$list_segments[$s]} = @attributes
#	is the information on the $s th segment, with
#	@attributes storing the attributes of the segment in the same order as in @labels_seg_attr.
#
sub readin_list_segs2 (\$\@) {

	my ($inlist, $labels_seg_attr) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	   print STDOUT "[ERROR 'readin_list_segs2'] 'open_input_file2' FAILED on the input file '${$inlist}' $! !\n";
	    return undef;
	}

	my $head = <$fh_tmpin>;

	my $indices_attr = tfm_labels2indices (@{$labels_seg_attr}, $head);
	unless (defined $indices_attr) {
	    $head =~ s/\s+$//;
	    print STDOUT "[ERROR 'readin_list_segs2'] 'tfm_labels2indices' FAILED on\n";
	    print STDOUT "      the labels  = (", join (', ', @{$labels_seg_attr}), " ),\n";
	    print STDOUT "  and the header  = '$head' : $! !\n\n";
	    return undef;
	}

=set aside for later use
	my $new_indx_diag;
	for (my $i=0; $i<@{$labels_seg_attr}; $i++) { # Preliminary for-loop (over field labels). #
	    ($labels_seg_attr->[$i] eq 'Diagnosis') or next;
	    $new_indx_diag = $i;
	    last;
	}				# End of the preliminary for-loop (over field labels). #
=cut

	my $max_indx_fld = 0;
	foreach my $indx (@{$indices_attr}) { ($max_indx_fld < $indx) and ($max_indx_fld = $indx); }

	my @list_segments = ();

	while (my $tmpin = <$fh_tmpin>) { # Main while-loop (over input lines). #

	    $tmpin =~ s/\s+$//;
	    my @fields = split /\t/, $tmpin;

	    (@fields > $max_indx_fld) or next; # Skip lines with an insufficient number of fields.

	    my @attributes = @fields[@{$indices_attr}];

	    push @list_segments, \@attributes;

	}		# End of the main while-loop (over input lines). #

	close ($fh_tmpin);

	return \@list_segments;

}	# End of the subroutine 'readin_list_segs2' #


# readin_simul_params:
#	This subroutine reads simulation parameters,
#	except the input tree and those in the comment line(s),
#	out of an input list.
#
# INPUT: (\$inlist), where
#	$inlist is the path to an input list file.
#
# OUTPUT: \%hash_simul_params, where
#	%hash_simul_params = ($label => $value, ...) for the simulation parameters.
#
sub readin_simul_params (\$) {

	my ($inlist) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$inlist});
	unless (defined $fh_tmpin) {
	    print STDOUT "[ERROR 'readin_simul_params'] 'open_input_file2' FAILED on the input list '${$inlist}': $! !\n";
	    return undef;
	}

	my %hash_simul_params;

	while (my $in = <$fh_tmpin>) { # Main while-loop #

	    ($in =~ m/^\s*\#/) and next; # Skip a comment line. #

		# MODIFIED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    ($in =~ m/^Tree\s/) and next; # Skip the input tree. #
	    ##($in =~ m/^Tree\.Tree/) and next; # Skip the input tree. #

	    ($in =~ s/\s+$//);

	    my ($label, $value) = split /\s*\=\s*/, $in;
	    unless ((defined $value) and ($value =~ m/\S/)) {
		print STDOUT "[CAUTION 'readin_simul_params'] NO VALUE in the input line:\n";
		print STDOUT "'$in': $! !\n\n";
		next;
	    }

		# ADDED in creation of 'compare_ref_vs_rec_compllimment.pl'. #
	    if ($value =~ s/^\"//) { ($value =~ s/\"$//); }
	    if ($label eq 'GapParams') { $value =~ s/^\{//; $value =~ s/\}$//; }

	    $hash_simul_params{$label} = $value;

	}			# End of the main while-loop #
	close ($fh_tmpin);

	return \%hash_simul_params;

}		# End of the subroutine 'readin_simul_params' #


# inspect_indel_events:
#	This subroutine examines whether all the indel events are shorter than or equal to the cutoff-lengths.
#
# INPUT: (\@set_cls_gblocks, \@sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
#	is for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordiate of the block (in the 0-based convention);
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) if the map of gap states onto the interior nodes,
#	with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#	
#	$cutoff_len_ins/del is the cutoff length for the insertion/deletion events;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \@illegal_indels, where
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with
#	$cl specifying a cluster,
#	$ps specifying a parsimonious scenario in the cluster,
#	$br_id specifying a branch,
#	$id specifying an illegal event in the node,
#	and $length is the length of the event.
#
#	The subroutine returns undef if it fails.
#
sub inspect_indel_events (\@\@$$$) {

	my ($set_cls_gblocks, $sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'inspect_indel_events' >\n\n") if ($TEST_LEVEL>1);

	my $ct_cls = @{$set_cls_gblocks};
	unless (@{$sets_psm_scns_for_cls} == $ct_cls) {
	    print STDOUT "[ERROR 'inspect_indel_events'] ILLEGAL INPUT: The size of ", '@{$sets_psm_scns_for_cls} = (', scalar (@{$sets_psm_scns_for_cls}), ") CLASHES with the number of gap-block clusters (=$ct_cls): $! !\n";
	    return undef;
	}

	my @illegal_indels = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer for-loop (over clusters) #

	    my $cls_gblock = $set_cls_gblocks->[$cl];
	    my $psm_scns_for_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_for_cls};

    (print STDOUT ' ' x 3, "[ The $cl th cluster has $ct_psm_scns parsimonious scenarios of indel events ]\n\n") if ($TEST_LEVEL>2);

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Middle for-loop (over parsiminous scenarios) #

    (print STDOUT ' ' x 5, '( The $ps th parsimonious scenario )', "\n\n") if ($TEST_LEVEL>3);

		my ($br2indels, $node2bw_gstat) = @{$psm_scns_for_cls->[$ps]};

		foreach my $br_id (sort {$a <=> $b} keys %{$br2indels}) { # Inner-middle foreach-loop (over branches) #
		    my $indel_events = $br2indels->{$br_id};
		    my $ct_events = @{$indel_events};

    (print STDOUT ' ' x 7, "( The branch '$br_id' has $ct_events nominal indel events )\n") if ($TEST_LEVEL>3);

		    for (my $id=0; $id<$ct_events; $id++) { # Inner for-loop (over indel events) #
			my ($stat_pa, $stat_node, $indices_blk) = (defined $indel_events->[$id]) ? @{$indel_events->[$id]} : ();
			unless (defined $stat_pa) {
			    print STDOUT "[WARNING 'inspect_indel_events'] The $id th event on the branch '$br_id' is EMPTY  for the $ps th indel scenario for the $cl th cluster: $! !\n";
			    next;
			}
			my $cutoff_len = ($stat_pa eq $char_abs) ? $cutoff_len_ins : $cutoff_len_del;
			my $indel_len = 0;
			foreach my $indx_blk (@{$indices_blk}) { $indel_len += $cls_gblock->[$indx_blk][1] ; }

			unless ($indel_len <= $cutoff_len) {
			    print ' ' x 7, "\n[WARNING 'inspect_indel_events'] The $id th indel event along the $br_id th branch is EXCEEDINGLY LONG (length = $indel_len)  for the $ps th scenario in the $cl th gap-block cluster: $! !\n\n";
			    push @illegal_indels, [$cl, $ps, $br_id, $id, $indel_len];
			}

		    }				# End of the inner for-loop (over indel events) #

		} 		# End of the inner-middle foreach-loop (over branches) #

	    }			# End of the middle for-loop (over parsimonious scenarios) #

	}			# End of the outer for-loop (over clusters) #

    (print STDOUT ' ' x 3, "--- End of the Subroutine 'inspect_indel_events'  ---\n\n") if ($TEST_LEVEL>1);

	return \@illegal_indels;

}	# End of the subroutine 'inspect_indel_events' #


# readin_dawg_tree2:
#	This modified version of 'readin_dawg_tree' copes with the new format
#	of the Dawg control file.
#	(CREATED in creation of 'compare_ref_vs_rec_compllimment.pl' on Aug 9 (Sat), 2014)
#
# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree2 (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ s/^Tree\s+\=\s+//) or next; # MODIFIED in creation of 'readin_dawg_tree2'. #
	    ##($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree2'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		## MODIFIED in creation of 'readin_dawg_tree2'. ##

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr; my %name2id_intnode; my $topnode_name;

	unless ( input_tree_file_dawg ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file_dawg' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

		# (Just in case,) remove the interior node names (or bootstrap supports) from the list of branch attributes. #

	foreach my $br (keys %br_attr) {
	    my $attr = $br_attr{$br};
	    my @new_attr = ();
	    $new_attr[$INDX_LENGTH] = $attr->[$INDX_LENGTH];
	    $br_attr{$br} = \@new_attr;
	}

=for the predecessor 'readin_dawg_tree'
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree2'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
=cut

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree2' #


	### DEFINITIONS of the SUBROUTINES in the predecessor,
	### "../../../2nd_Stage.DNA/ML_ScoringF6t.PhyML/Scripts/compare_ml_msa.bch.extJC.6t.PhyML.pl".

# readin_dawg_tree:
#	This subroutine reads in the input tree used to guide
#	a simulation of sequence evolution via DAWG.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011) 
#
# INPUT: \$infile, where
#	$infile is the file recording input parameters of DAWG.
#
# OUTPUT: \@info_dawg_tree, where
#	@info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr).
#	See e.g. 'MyTreeManipu_II.pm' for the format of the output hash.
#
sub readin_dawg_tree (\$) {

	my ($infile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$infile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'open_input_file2' FAILED on the input file '${$infile}': $! \n";
	    return undef;
	}

	my $in_tree;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #
	    ($tmpin =~ s/^Tree\.Tree\s+\=\s+//) or next;
	    $in_tree = $tmpin;
	    last;
	}		# End of the main while-loop # 
	close ($fh_tmpin);

	unless ((defined $in_tree) and ($in_tree =~ m/\(/)) {
	    (defined $in_tree) or ($in_tree = 'undef');
	    $in_tree =~ s/\s+$//;
	    print STDERR "[ERROR 'readin_dawg_tree'] NO TREE was extracted from the input file '${$infile}': $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}
		# Process the input tree a bit #
	($in_tree =~ s/[^\)\;]+$//);
	($in_tree =~ m/\;$/) or ($in_tree .= ';');

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file ($in_tree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'readin_dawg_tree'] 'input_tree_file' FAILED on the input tree: $! !\n";
	    print STDERR "\$in_tree = '$in_tree' .\n\n";
	    return undef;
	}

	my @info_dawg_tree = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr);

	return \@info_dawg_tree;

}	# End of the subroutine 'readin_dawg_tree' #


# ln_ml_sgl_seq2:
#	This revised subroutine of 'ln_ml_sgl_seq' also outputs 'column-wise' log ML values.
#	(IMPORTED from 'MyApBayes_Score_MSA_ExtJC.pm' and MODIFIED in creation of
#	'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011)
#
# ln_ml_sgl_seq:
#	This subroutine estimates the log-likelihood of obtaining a single sequence.
#	The log-likelihood is simply calculated as:
#	$lnl = Sum_$r log ($freq[$r]) ,
#	where $freq[$r] is the equilibrium frequency of the $r th amino acid residue.
#
# INPUT: (\$seq, \%aa2freq), where
#	$seq is the sequence;
#	%aa2freq = ($amino_acid (a single capital letter) => $equilibrium_frequency, ...).
#
# OUTPUT: ($lnl, \@clmwise_lnl), where
#	$lnl is the log-likelihood of the whole alignment;
#	$clmwise_lnl[$l] is the log-likelihood of the $l th column.
#
#	The subroutine returns (undef, undef) if it failed.
#	
sub ln_ml_sgl_seq2 (\$\%) {

	my ($seq, $aa2freq) = @_;

	my $cp_seq = uc ${$seq};
	my $len = length $cp_seq;

	my @clmwise_lnl = ();
	my $lnl = 0;
	for (my $l=0; $l<$len; $l++) { # For-loop #
	    my $res = substr ($cp_seq, 0, 1, '');
	    my $frq = $aa2freq->{$res};
	    unless (defined $frq) {
		print STDERR "[ERROR 'ln_ml_sgl_seq2'] The frequency is UNDEFINED for the residue '$res' at the $l th site: $! !\n\n";
		return (undef, undef);
	    }
	    my $clmwise = log $frq;
	    $lnl += $clmwise;
	    $clmwise_lnl[$l] = $clmwise; # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' (on Mar 7 (Mon), 2011) #
	}		# End of the for-loop #

	return ($lnl, \@clmwise_lnl);

}	# End of the subroutine 'ln_ml_sgl_seq2' #


# add_dummy_seq:
#	This subroutine adds a 'dummy' sequence to a tree and an MSA consisting of two sequences,
#	and makes a 'virtual' 3-sequence tree and MSA.
#	The dummy is the exact copy of the 2nd sequence, which 'diverge'd $eps above its exterior node,
#	where $eps is the minimum length permitted by PhyML.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 11 (Mon) & 12 (Tue), 2011)
#
# INPUT: (\@seqnames, \@msa, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr, \@indices_attr), where
#	$seqnames[$k] is the name of the $k th sequence;
#	${$msa[$k]} is the $k th sequence embedded in the alignment;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	%name2id = ($name => $node_id, ...) for exterior nodes and the top node;
#	%id2name = ($node_id => $name, ...) for exterior nodes;
#	%br_attr = ($node_id => \@branch_attributes, ...) for the nodes immediately downstream of the branches,
#	with @branch_attributes[@indices_attr] = ($branch_length, $bootstrap_support, Total#{bootstrap resamplings}).
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub add_dummy_seq (\@\@\%\%\%\%\%\@) {

	my ($seqnames, $msa, $node2pa, $node2ch, $name2id, $id2name, $br_attr, $indices_attr) = @_;

	unless ((@{$seqnames} == 2) and (@{$msa} == 2)) {
	    print STDERR "[ERROR 'add_dummy_seq'] ILLEGAL INPUT: Either the size of MSA (", scalar (@{$msa}), ") or the number of sequence names (", scalar (@{$seqnames}), ") is NOT 2, which they are supposed to be: $! !\n";
	    return 0;
	}

		## Manipulate the input tree ##

	my $eps = 0.0000000001; # Minimum branch length permitted in PhyML #
	my ($indx_len, $indx_boot, $tot_boot) = @{$indices_attr};

	my $max_nodeid = 0;
	foreach my $nodeid (keys %{$node2ch}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }
	foreach my $nodeid (keys %{$id2name}) { ($nodeid > $max_nodeid) and ($max_nodeid = $nodeid); }

		# Change the identity of the second sequence #
	my $name_seq2 = $seqnames->[1];
	my $nid_seq2 = $name2id->{$name_seq2};
	delete $id2name->{$nid_seq2}; # $nid_seq2 is no longer an exterior node

		# Create two new exterior nodes, one real and the other virtual #
	my $nid_sbl1 = $max_nodeid + 1;
	my $nid_sbl2 = $max_nodeid + 2;

	##my $name_sbl1 = $name_seq2;
	my $name_sbl2 = 'dummy';

	$name2id->{$name_seq2} = $nid_sbl1;
	$name2id->{$name_sbl2} = $nid_sbl2;

	$id2name->{$nid_sbl1} = $name_seq2;
	$id2name->{$nid_sbl2} = $name_sbl2;

		# Establish the relationships between seq2, and the new siblings #
	$node2ch->{$nid_seq2} = [$nid_sbl1, $nid_sbl2];
	$node2pa->{$nid_sbl1} = $nid_seq2;
	$node2pa->{$nid_sbl2} = $nid_seq2;

		# Slightly change the length of the branch above seq2, which is now the 'virtual' parent node #
	my $attr_br2 = $br_attr->{$nid_seq2};
	my $len2 = $attr_br2->[$indx_len];
	$attr_br2->[$indx_len] = ($len2 > $eps) ? $len2 - $eps : 0 ;

		# Fix the lengths of the 'virtual' child branches as $eps #
	foreach my $nid ($nid_sbl1, $nid_sbl2) { # foreach-loop #
	    my @attr = ();
	    $attr[$indx_len] = $eps; 
	    $br_attr->{$nid} = \@attr;
	}			# End of the foreach-loop #

		## Manipulate the input alignment ##
	my $seq2 = ${$msa->[1]};
	$msa->[2] = \$seq2;
	$seqnames->[2] = $name_sbl2;

	return 1;

}	# End of the subroutine 'add_dummy_seq' #


=set aside for later use
# even_brlen_2seq_tree:
#	This subroutine receives a 2-node tree,
#	and evenly distributes the branch length
#	into the 2 branches, to conform to the molecular clock.
#	(IMPORTED from 'MyApBayes_Score_MSA_ExtJC.pm' on Mar 7 (Mon), 2011)
#
# INPUT: (\%br_attr, $indx_brlen), where
#	%br_attr = (ID of the node immediately below the branch => \@attr, ...),
#	with $attr[$indx_brlen] being the branch length.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub even_brlen_2seq_tree (\%$) {

	my ($br_attr, $indx_brlen) = @_;

	my @nodes = keys %{$br_attr};
	unless (@nodes == 2) {
	    print STDERR "[ERROR 'even_brlen_2seq_tree'] The input tree has ", scalar (@nodes), " NODES, instead of 2 nodes : $! !\n";
	    return 0;
	}

	my $tot_brlen = 0;
	foreach my $node (@nodes) { # The 1st foreach-loop #
	    my $attr = $br_attr->{$node};
	    unless (defined $attr) {
		print STDERR "[ERROR 'even_brlen_2seq_tree'] The node '$node' does NOT HAVE AN ATTRIBUTE for the upstream branch: $! !\n";
		return 0;
	    }
	    my $brlen = $attr->[$indx_brlen];
	    unless (defined $brlen) {
		print STDERR "[ERROR 'even_brlen_2seq_tree'] The node '$node' does NOT HAVE A BRANCH LENGTH for the upstream branch: $! !\n";
		return 0;
	    }
	    $tot_brlen += $brlen;
	}		# End of the 1st foreach-loop #

	foreach my $node (@nodes) { # The 2nd foreach-loop #
	    $br_attr->{$node}[$indx_brlen] = $tot_brlen / 2;
	}		# End of the 2nd foreach-loop #

	return 1;

}	# End of the subroutine 'even_brlen_2seq_tree' #
=cut


=the following are for old versions (set aside for potential use)
# parse_paml_main_output:
#	This subroutine parses the main output file (the 'mlc' file by default).
#	(IMPORTED from 'ml_scoring_msa.bch.pl' on Mar 16 (Wed), 2011)
#
# INPUT: (\$infile), where
#	$infile is the path to the main PAML output file.
#
# OUTPUT: (\%output), where
#	%output = ('lnL' => [$lnl_1, $lnl_2, ...],
#		   'se_lnL' => [$se_lnl_1, $se_lnl_2, ...], ...),
#	with $lnl_k being the logarithmic ML value for the tree # k.
#
#	The subroutine returns undef if it failed.
#
sub parse_paml_main_output (\$) {

	my ($infile) = @_;
	local $/ = ''; # Use blank lines as input-record-separators

	my $fh_tmp = open_input_file2 (${$infile});
	unless (defined $fh_tmp) {
	     print STDERR "[ERROR 'parse_paml_main_output'] The subroutine 'open_output_file2' FAILED on the input file '${$infile}': $! !\n\n";
	     return undef;
	}

	my %output;

	while (my $in = <$fh_tmp>) { # Outer while-loop #

	     my ($k) = ($in =~ m/^TREE\s+\#\s+(\d+)\:/ );
	     (defined $k)  or next;

	    my ($lnl, $se) ;
	    my @lines = split /\n/, $in;
	    while (my $line = shift @lines) { # Inner while-loop #
		my ($tmp_lnl, $tmp_se) = ($line =~ m/^lnL\([^\)]+\)\:\s+(\S+)\s+(\S+)/ );
		(defined $tmp_lnl) or next;
		$lnl = $tmp_lnl;
		$se = $tmp_se;
		last;
	    }			# End of the inner while-loop #
	    unless (defined $lnl) {
		print STDERR "[ERROR 'parse_paml_main_output'] FAILED to extract the log likelihood for the $k th tree: $! !\n\n";
		return undef;
	    }

	    (defined $output{'lnL'}) or ($output{'lnL'} = []);
	    (defined $output{'se_lnL'}) or ($output{'se_lnL'} = []);
	    $output{'lnL'}->[$k-1] = $lnl; 
	    $output{'se_lnL'}->[$k-1] = $se;

	}		# End of the outer while-loop #
	close ($fh_tmp);

	return \%output;

}	# End of the subroutine 'parse_paml_main_output' #


# fetch_clmwise_lnlks:
#	This subroutine reads in the 'lnf' file output by the 'codeml' or 'baseml'
#	program of the PAML package,
#	and makes a list of column-wise log-likelihoods.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 16 (Wed), 2011)
#
# INPUT: (\$tmpin_file, $gap), where
#	$tmpin_file specifies the path to a 'lnf' file you want to read;
#	$gap is a character representing a gap.
#
# OUTPUT: \@set_clmwise_lnlk, where
#	@{$set_clmwise_lnlk[$c]} = (\$string_clm, $lnlk_clm)
#	is for the $c th column (sorted in lexical order),
#	$string_clm is the horozontal representation of the column
#	(including gaps and/or stop codons in general),
#	and $lnlk_clm is the log likelihood of the column.
#
sub fetch_clmwise_lnlks (\$$) {

	my ($tmpin_file, $gap) = @_;
	local $/ = "\n";

	my @set_clmwise_lnlk = ();

	my $fh_tmpin = open_input_file2 (${$tmpin_file});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'fetch_clmwise_lnlks'] 'open_input_file2' FAILED on the input file '${$tmpin_file}': $! !\n";
	    return undef;
	}

	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #

	    ($tmpin =~ m/[\*a-zA-Z$gap]\s*$/o) or next;
	    my @flds = split ' ', $tmpin;
	    my ($lnlk_clm, $string_clm) = @flds[2,5];
	    unless ((defined $string_clm) and ($string_clm =~ m/[\*a-zA-Z$gap]/o)) {
		(defined $string_clm) or ($string_clm = 'undef');
		print STDERR "[ERROR 'fetch_clmwise_lnlks'] ANOMALOUS STRING REPRESENTATION of the column ('$string_clm') was found: $! !\n";
		return undef;
	    }

	    push @set_clmwise_lnlk, [\$string_clm, $lnlk_clm];

	}		# End of the main while-loop #

	close ($fh_tmpin);

	my @std_set_clmwise_lnlk = sort { ${$a->[0]} cmp ${$b->[0]} } @set_clmwise_lnlk;

	return \@std_set_clmwise_lnlk;

}	# End of the subroutine 'fetch_clmwise_lnlks' #

=cut 	### End of "the following are for old versions (set aside for potential use)"

