###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#	Version 0.7: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# anex_for_sgl_wd.ver0.7.pl:
#	This script is the UPGRADED version of 'anex_for_sgl_wd.ver0.5.pl'.
#	The following MAJOR REVISIONS were made:
#
#	(1) The script can now take full advantage of the main outputs of 'lastpiece.xxx.pl' in the program package, "LASTPIECE_P.verNNN",
#	and thus the computed MSA probabilities can now be MUCH MORE ACCURATE than computed by the predecessor (ver 0.5);
#	To enable this revision,
#	  (1-i) the block for checking files of multiplication factors can now accept symbolic links as well, in addition to normal files as the predecessor (ver 0.5) does;
#
#	  (1-ii) the subroutines, 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm") (& their satellite subroutines) & 'compt_log_msa_prob_plus' & 'nonrec_mlt_shifts_gapblocks' (the latter two are in "MyANEX_Main.pm"), were substantially REVISED
#		(& renamed the 1st one as 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm")),
#		especially to take advantage of multiplication factors for case-(i) "gapped segment"s;
#
#	   (1-iii) the subroutine, 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm"), was REVISED, so that it can now read in the input file(s) recording multiplication factors for case-(i) "gapped segment"s;
#
#	(2) REVISED the rules (in the subroutine 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm") and its satellite subroutines) to "extrapolate" the multiplication factors (for case-(iv) and case-(i)) to those w/ the time-lapses smaller than the minimum time-lapse recorded in the input file (of the relevant multiplication factors).
#
#	[MAJOR REVISION STARTED on 2020/04/24; ENDED on 2020/04/dd.]
#
#
# anex_for_sgl_wd.ver0.5.pl:
#	This Perl script is the 1st prototype (written in Perl) of the MAIN ENGINE of 'anex.ver0.5.pl'.
#	It explores the neighborhood of an input MSA within a window,
#	by first performing the pure multiple-"shift"s the gap-blocks,
#	then by performing other elementary changes of the MSA, followed by multiple-"shift"s.
#	It was created by putting together useful portions of a number of scripts.
#	More precisely, the main sources are:
#	(1) "anex.ver0.5.pl", especially its 1st half;
#	(2) "detect_purge_cands.ver0.5.pl", especially its 1st half;
#	(3) The main & some other parts of 'test_nonrec_mlt_shifts_gapblocks.pl' (in ./Ingredients/);
#	(5) The main & some other parts of the other test scripts for the "move"s other than pure "shift"s (named 'test_xxxxx.pl') (also in ./Ingredients/);
#
#	Then, other necessary parts were either imported from other scripts
#	or created de novo.
#	[CREATION STARTED on 2019/06/24, ENDED on 2019/mm/dd.]
#
#	NOTE: This script ASSUMES the following structure of the working directory:
#
#	{Top_Working_Directory}/ .
#
#	For the main window:
#	{Main_Input_Directory}/ = "{Top_Working_Directory}/Inputs/" .
#	{Input_file_for_MSA_in_the_window} = "{Main_Input_Directory}/msa_in_wd.txt" ,
#	{Input_file_for_pcands} = "{Main_Input_Directory}/list_rlv_pcands.txt" , # Added on 2019/07/03.
#	{Input_file_for_initial_set_of_columns} = "{Main_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Main_Input_directory}/list_gblocks0.txt" .
#
#	For the sub-windows belonging to the window:
#	{Sub_Working_Directory}/ = "{Top_Working_Directory}/Sbwd{sub-window-ID}/" .
#	{Sub_Input_Directory}/   = "{Sub_Working_Directory}/Inputs/" .
#	{Input_file_for_pcands} = "{Sub_Input_Directory}/list_rlv_pcands.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Sub_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Sub_Input_Directory}/list_gblocks0.txt" .
#
#	NOTE2 (added on 2019/07/04):
#	This script does NOT perform "Reverse-Purge" (aka, "Ex-Nihilo") + Multiple-"Shift"s;
#	such moves are performed via the script, 'anex_rev_purge_1wd.verXXX.pl'.
#
#	[MAJOR REVISION on 2019/07/07]:
#	MOVED (nearly) ALL the subroutines within this script to the package, "MyANEX_IO.pm".
#
#	[MAJOR REVISION on 2019/07/18-19]:
#	The 4 subroutines, 'print_in_nr_mshifts_glks', 'bch_print_out_other_gblk_attributes', 'nonrec_mlt_shifts_gapblocks' & 'print_out_nr_mshifts_gblks',
#	were BUNDLED together into a single subroutine, 'perform_mshifts_and_print_inputs_outputs'.
#
#	[REVISED on 2019/07/21]:
#	Now, 'purge_gblock_pair' & 'cmerge_gblock_pair' invokes 'mk_orgs_bds_gb_coords' (in " MyANEX_Supple.pm") from its own inside.
#
#	[REVISED on 2019/08/01]:
#	Now, a new global variable, $XLEVEL, is available,
#	for specifying the LEVEL of the MSA Space Exploration.
#
#	[REVISED on 2019/10/18-19]:
#	Now, TWO new global variables, $IF_TEST_W_COMPACT_MSS & $HF_WDTH_SHIFT_4_CMSS, are available,
#	for enabling the test with a "compact" multiple-shift space (whose half-width is given by $HF_WDTH_SHIFT_4_CMSS).
#
#	[REVISED on 2020/04/20-21]:
#	Now, two subroutines, 'find_maxvalue_in_multidim_storage' & 'cmpt_log_total_qnt_in_multidim_storage',
#	in the Perl package, "MyANEX_IO2.pm", are incorporated.
#	And the script now outputs a new "synopsis" file,
#	which summarizes the results of ALL runs of multiple-"shift"s performed (& their inputs).
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyMalnIO qw(
	extract_alignment
	pr_ODEN_input
	);
	# extract_fasta
	# pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyTreeManipu_II qw(
	input_tree_file output_tree
	br_output_tree
	);
	# extract_sgltree_from_file
	# make_node2parent

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	br_output_tree_dawg
		) ;
	# extract_sgltree_from_file_dawg


use MyTreeMap_indels_spt_odr2 qw(
	cal_init_psm_cands_for_sgl_msa
	bch_br_parsimonize_dollo_indel_histories
	);

use MyTreeMap_indels_spt_odr_finer qw(
	enum_psm_indel_hsts_for_msa_finer_PRE
	);
	# partit_into_gblocks_seqblocks
	# if_idb_includes_root
	# partit_into_eff_indep_indel_blocks
	# find_invlvd_in_eff_indep_indel_blocks
	# mk_node2blwise_gstat

use MyReadAlnProbIngredients qw(
	readin_delta_del
	assign_br2log_mfacs
	);

use MyANEX_Supple qw(
	prepare_sliding_windows
	mk_orgs_bds_gb_coords
	);
	# Added 'mk_orgs_bds_gb_coords' in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).

use MyArrayManipu_num qw(
	diff_nmarray
	);

use MyTreeMap_indels_spt_odr_hs qw(
	enum_psm_indel_scns_for_msa_hs_mrounds
	);
	## enum_psm_indel_scns_for_msa_hs

use MyTreeMap_indels_ML_hs_hs qw(
	apprx1_tot_lnlk_gpattern_msa2_hs
	);
	## cal_delta_exp_del_5p2
	## cal_log_normal_facs_indel_distri
	## cal_lnlk1_indel_scn_cls2

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
use MyTreeMap_sppl qw(
	wrapper_mk_node2exoffs_msa_rows
	mk_set_msa_rows2br_up_or_lw
	mk_set_msa_rows2br_up_or_lw_dp
	);
	# Added 'mk_set_msa_rows2br_up_or_lw_dp' on 2019/07/03.

	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27). #
use MyLikelihoods_respattern_msa_dna qw(
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);
	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29). #
use MyGapPatternTopology qw(
	mk_inter_block_relations
	mk_list_gblock_sizes
	mk_lists_gblocks_w_spec_bounds
	cts_vincluding_vincluded_onx_gblks
	);
	# Added 'mk_list_gblock_sizes' & 'mk_lists_gblocks_w_spec_bounds' & 'cts_vincluding_vincluded_onx_gblks'  on 2019/07/16.

	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02). #
use MyANEX_Main qw(
	printout_multidim_storage
	nonrec_mlt_shifts_gapblocks
	compt_log_msa_prob_plus

	list_nn_purge_and_cmerge_candidates
	list_nn_smerge_candidates
	list_nn_cvmerge_candidates
	list_smerge_split_candidates
	list_nn_cmerge_split_pre_candidates
	list_cmerge_split_candidates

	list_ssplit_candidates
	list_csplit_candidates
	list_nn_csplit_candidates
	list_nn_gen_split_candidates
	list_vsplit_candidates
	list_cvsplit_candidates

	list_nn_vsplit_plus_smerge_candidates
	list_nn_cvsplit_plus_smerge_candidates
	list_nn_ssplit_plus_vmerge_candidates
	
	purge_gblock_pair
	cmerge_gblock_pair
	smerge_gblock_pair
	cvmerge_gblock_pair

	ssplit_gblock
	csplit_gblock
	vsplit_gblock
	cvsplit_gblock

	smerge_split_gblock_pair
	icmerge_gblock_pair
	cmerge_split_grow_gblock_pair
	vsplit_plus_smerge_gblock_pair
	cvsplit_plus_smerge_gblock_pair
	ssplit_plus_vmerge_gblock_pair
	);
	# Added 'compt_log_msa_prob_plus' on 2019/07/04.
	# Added 'list_nn_purge_and_cmerge_candidates' & 'purge_gblock_pair' on 2019/07/16.
	# Added 'cmerge_gblock_pair' on 2019/07/20.
	# Added 'list_nn_smerge_candidates' & 'smerge_gblock_pair' on 2019/07/22.
	# Added 'list_nn_cvmerge_candidates' & 'cvmerge_gblock_pair' on 2019/07/29.
	# Added 'list_smerge_split_candidates' & 'smerge_split_gblock_pair' on 2019/08/01.
	# Added 'list_nn_cmerge_split_pre_candidates' & 'list_cmerge_split_candidates' & 'icmerge_gblock_pair' & 'cmerge_split_grow_gblock_pair' on 2019/08/10.
	# Added 'list_ssplit_candidates' & 'ssplit_gblock' on 2019/08/13.
	# Added 'list_csplit_candidates' & 'list_nn_csplit_candidates' & 'list_nn_gen_split_candidates' & 'list_vsplit_candidates' & 'csplit_gblock' & 'vsplit_gblock' on 2019/08/25.
	# Added 'list_cvsplit_candidates' & 'cvsplit_gblock' on 2019/08/30.
	# Added 'list_nn_vsplit_plus_smerge_candidates' & 'vsplit_plus_smerge_gblock_pair' on 2019/09/07.
	# Added 'list_nn_cvsplit_plus_smerge_candidates' & 'cvsplit_plus_smerge_gblock_pair' on 2019/09/12.
	# Added 'list_nn_ssplit_plus_vmerge_candidates' & 'ssplit_plus_vmerge_gblock_pair' on 2019/10/17.


	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/09/30). #
use MyANEX_MainX qw(
	list_cii_candidates
	filter_cii_candidates
	reverse_spec_cii_in_wd
	);
	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/09/30)." #

=Set Asite for Later Use. (PM)
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_III qw(
	make_node2abs_depth
	make_node2ct_exoffs
	);

use MyTreeManipu_IV qw(
	bi2multifurcated_top
	construct_neat_subtree
	construct_skeleton_subtree
	);

##use MyTreeMap_indels qw(
##	enum_psm_indel_scns_for_msa
##	);
##	## extract_cls_gblocks msa_rep_gpatterns

use MyTreeMap_indels_sppl qw(
	create_br2eqbr
	create_br2nn
	);

use MyDetect_cmplx_error_cands qw(
	wrapper_detect_cmplx_error_cand_acls0
	);
	# mk_self_nn_events

=cut # End of "Set Asite for Later Use. (PM)"

	### ADDED in MAJOR REVISION on 2019/07/07. ###
use MyANEX_IO qw(
	read_dawg_ctrl_file_sbst 
	read_dawg_ctrl_file3 
	readin_list_mgd_pcands 
	readin_set_clms 
	readin_list_gblks 
	revise_list_gblocks 

	print_out_input_tree 
	print_out_br2log_mfacs 
	print_out_tree_profile 
	print_out_set_clms 

	print_out_readin_list_gblks 

	print_out_wrapper_mk_node2exoffs_msa_rows 
	print_out_mk_set_msa_rows2br_up_or_lw 
	print_out_br2tr_mtrx 
	print_out_node2ntfreqs 
	print_out_node2set_sw_cond_probs 
	print_out_br2set_sw_jnt_probs 
	print_out_set_clm_probs 
	print_out_orgs_bds_gb_coords 
	print_out_row_clm_info_gblocks 
	print_out_inter_block_relations 
	print_out_list_mgd_pcands 
	print_out_features_gblks 
	print_out_x_bin_hrels 
	print_in_nr_mshifts_glks 
	print_out_ln_prb_plus_msa 
	print_out_coord_frame 
	print_out_nr_mshifts_gblks 
	print_out_list_gblock_sizes

	print_out_purge_or_merge_cands
	print_out_merge_split_cands
	print_out_split_cands
	print_out_vsplit_cands

	print_out_list_gpblocks
	print_out_list_cii_cands

	print_out_gblks_w_spec_bds
	print_out_cts_vincluding_vincluded_onx_gblks
	bch_print_out_other_gblk_attributes

	print_out_info_mvd_sbj
	print_out_info_chngs_by_purge
	print_out_info_chngs_by_cmerge
	print_out_info_chngs_by_smerge
	print_out_info_chngs_by_cvmerge
	print_out_info_chngs_by_merge_split
	print_out_info_chngs_by_cmerge_split_grow
	print_out_info_chngs_by_split
	print_out_info_chngs_by_vsplit
	print_out_info_chngs_by_vsplit_plus_smerge
	print_out_info_chngs_by_ssplit_plus_vmerge

	print_out_sbsdry_outputs_rev_spec_cii_in_wd

	perform_mshifts_and_print_inputs_outputs

	relative_coords_for_list_mgd_pcands 
	);
	# Added 'print_out_list_gblock_sizes' & 'print_out_purge_or_merge_cands' on 2019/07/16.
	# Added 'print_out_info_mvd_sb' & 'print_out_info_chngs_by_purge' on 2019/07/17.
	# Added 'print_out_info_chngs_by_cmerge' on 2019/07/19.
	# Added 'print_out_gblks_w_spec_bds' & 'print_out_cts_vincluding_vincluded_onx_gblks' & 'bch_print_out_other_gblk_attributes' & 'print_out_info_mvd_sbj' & 'print_out_info_chngs_by_purge' on 2019/07/17.
	# Added 'perform_mshifts_and_print_inputs_outputs' in MAJOR REVISION (on 2019/07/18). #
	# Added 'print_out_info_chngs_by_smerge' on 2019/07/22.
	# Added 'print_out_info_chngs_by_cvmerge' on 2019/07/29.
	# Added 'print_out_merge_split_cands' & 'print_out_info_chngs_by_merge_split' on 2019/08/01.
	# Added 'print_out_info_chngs_by_cmerge_split_grow' on 2019/08/11.
	# Added 'print_out_split_cands' on 2019/08/13.
	# Added 'print_out_info_chngs_by_split' on 2019/08/14.
	# Added 'print_out_vsplit_cands' on 2019/08/24.
	# Added 'print_out_info_chngs_by_vsplit' on 2019/08/25.
	# Added 'print_out_info_chngs_by_vsplit_plus_smerge' on 2019/09/08.
	# Added 'print_out_list_gpblocks' & 'print_out_list_cii_cands' & 'print_out_sbsdry_outputs_rev_spec_cii_in_wd' on 2019/10/01. #
	# Added 'print_out_info_chngs_by_ssplit_plus_vmerge' on 2019/10/17.

	### End of "ADDED in MAJOR REVISION on 2019/07/07." ###

	### ADDED in REVISION on 2020/04/20. (US) ##
use MyANEX_IO2 qw(
	find_maxvalue_in_multidim_storage
	cmpt_log_total_qnt_in_multidim_storage
	);
	### End of "ADDED in REVISION on 2020/04/20. (US)" ##


$| = 1; # Flush the input buffer as soon as it receives data.

my $plscript_path = $0;
my ($plscript_name) = ($plscript_path =~ m/([^\/]+)$/); 
my $program_name = $plscript_name; # Added on 2019/06/10, to reduce the number of error messages.

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

my $exec_perl = "/usr/bin/perl"; # Perl executable to be used. (Added on 2019/06/19.) 


	### File names storing important parameters. ###

our $FILENAME_DELTA_DEL = 'delta_del.txt'; # The name of the file storing '$delta_del', which is important for calculating the exit rate of the entire sequence. #

our @FILENAMES_LOG_MFACS = ( # Names of the files storing logarithmic multiplication factors, which are in $indir_mfacs. #

		# In the bulk #
	'lmfacs_asites_vs_dgaps.txt', # For the cases where there are some ancestral sites and no descendant sites in between preserved ancestral sites (PASs). #
	'lmfacs_agaps_vs_dsites.txt', # For the cases where there are no ancestral sites and some descendant sites in between PASs.
	'lmfacs_asites_nhom_dsites.txt', # For the cases where there are some ancestral sites and some descendant sites in between PASs.

		# On either boundary (i.e., either PWA end). "bd_$filename" corresponds to $filename above. #
	'bd_lmfacs_asites_vs_dgaps.txt',
	'bd_lmfacs_agaps_vs_dsites.txt',
	'bd_lmfacs_asites_nhom_dsites.txt',

		# For case-(i) "gapped segment"s. #
		# (ADDED in creation of 'anex_for_sgl_wd.ver0.7.pl' on 2020/04/24). (GL1) #
	'lmfacs_nosites.txt',
	'bd_lmfacs_nosites.txt',
		# End of "(ADDED in creation of 'anex_for_sgl_wd.ver0.7.pl' on 2020/04/24). (GL1)" #
	);

	### End of "File names storing important parameters." ###

	##### Glogal parameter setting #####

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

our $STRING_UNDEF = 'n/a'; # This is used for printing out some elements or quantities that are undefined for some reasons. (Added on 2019/07/02.)


			## Indexes in the terminal array of a multi-dimensional storage (for multiple-"shift"s). ##
			## (Added in REVISION on 2020/04/20.) (GL) ##
our $INDX_DIFF_LN_PRB_MSA = 3; # For the difference in the log-probability of the MSA (under a model w/ both substitutions & indels.)
our $INDX_DGN = 1;  # For the degree of degeneracy of the MSA.
			## End of "(Added in REVISION on 2020/04/20.) (GL)" ##

		# Added in revision (on 2019/06/18). #
#our $DIS_LEN = 50; # The maximum length of the (sub)sequence displayed in a line.
our $DIS_LEN = 60; # The maximum length of the (sub)sequence displayed in a line.
		# End of "Added in revision (on 2019/06/18)." #

$MyANEX_IO::DIS_LEN = $DIS_LEN; # Added in REVISION (on 2019/07/07). #



		# ADDED on 2019/08/01. (GL) #
	# The level of MSA space exploration.
	# (1: performing only pure multiple-"shift"s;
	#  2: also performing ;
	#  3: also performing ;
	#  5: performing all the programmed moves.)
#our $XLEVEL = 1; 
our $XLEVEL = 3; 
#our $XLEVEL = 5; 
		# END of "ADDED on 2019/08/01. (GL)" #

	# Minimum and maximum sizes of windows used for detecting purge-candidate regions. #
our $MIN_WD_SIZE_PCAND = 3;
our $MAX_WD_SIZE_PCAND = 6;

	# Each gap-block can "shift" up to (and including) $HF_WDTH_SHIFT columns in each direction. # (Added on 2019/06/29.)
our $HF_WDTH_SHIFT = 10;

		# ADDED on 2019/10/18. (GL) #
our $IF_TEST_W_COMPACT_MSS = 0; # Whether to perform tests with "compact" multiple-shift spaces(=1) or not(=0).
our $HF_WDTH_SHIFT_4_CMSS  = 2;
		# END of "ADDED on 2019/10/18. (GL)" #


our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

=Set Aside for Later Use. (GL 1)
	# Characters indicating that the site is missing or that its orthology is questionable. 
our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);
=cut # END of "Set Aside for Later Use. (GL 1)"

	# Parameters for the relative probabilities of Indel Histories. #
our %INDEL_MODEL2DISTRI = ( # = (indel model => distribution type, ...)
	'zipf' => 'P',
	'"PL"' => 'P',
	'"NB"' => 'NB', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	'"US"' => 'US', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	);


our @RESIDUES = our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
##our @RESIDUES = our @NUCLEOTIDES = @MyPRANK_Scores::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our $SV_RATIO_R = 1; # Ratio of purine-type transition rate to transversion rate.
our $SV_RATIO_Y = 1; # Ratio of pyrimidine-type transition rate to transversion rate.

	# Equilibrium residue frequencies. #
my @root_resfreqs = (0.25, 0.25, 0.25, 0.25);
##my @root_resfreqs = (0.1, 0.2, 0.3, 0.4);

my @labels_resfreqs = ();
foreach my $res (@NUCLEOTIDES) { push @labels_resfreqs, "f($res)"; }


	### ReSet parameters defined in some Perl packages. ###

$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS = 1; # Ignore null columns(, if at all).

$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = 0; # NOT analyze null-clusters (=clusters consisting only of null-columns), if at all.

	# This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = 120;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 150;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;



	### ReSet Parameters to adjust the verboseness of the test outputs. ###
	### (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)

	# For this script. #
our $TEST_LEVEL = 5;
##our $TEST_LEVEL = 1;

	# Testing the inputs/outputs of the processes going through this MAIN program.
our $TEST_LEVEL_MAIN = 5;

$MyANEX_IO::TEST_LEVEL_MAIN = $TEST_LEVEL_MAIN; # Added in REVISION (on 2019/07/07). #

	# For some subroutines defined in this script. #
	# (e.g., 'print_in_nr_mshifts_glks', 'print_out_nr_mshifts_gblks'). # (Added on 2019/07/02.)
our $TEST_LEVEL_SUB = 2;

$MyANEX_IO::TEST_LEVEL_SUB = $TEST_LEVEL_SUB; # Added in REVISION (on 2019/07/07). #


	# Testing some primary satellite subroutines of 'readin_delta_del' & 'assign_br2log_mfacs'. #
$MyReadAlnProbIngredients::TEST_LEVEL1  = 0;


        # For 'calc_prob_descresstats_cond_ancstat2'. #
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 5;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 6;

	# For 'calc_joint_prob_ancresstats_and_ancstat2'. #
	# (ADDED in creation of 'test_calc_jnt_prob_ancresstats2.pl'.)
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 6;

	# For 'enum_psm_indel_hsts_for_msa_finer_PRE'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 0;
#$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 3;

	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_spt_odr_hs.pm" for details.)
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL = 3;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1 = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1hf = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL2 = 0;

	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_ML_hs_hs.pm" for details.)
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL = 1;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1 = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1a = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL2a = 0;

=Set Aside for Later Use. (GL 2) 
	# For 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_1ST = 0;

	# For 'br_parsimonize_dollo_indel_history'. 
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = 0;

	# For 'bch_br_parsimonize_dollo_indel_histories'.
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = 0;

=cut # End of "Set Aside for Later Use. (GL 2)"

	##### END of "Glogal parameter setting" #####


	####### Take in the Command-Line Arguments. #######

unless (@ARGV > 2) {
	print STDERR "\nUsage:  [perl] $program_name {top_working_directory} {dawg_ctrl_file (incl. input tree)} {dir_mfac_data} [ options ] ,\n";
	print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";
	print STDERR "            --MIN_WD_SIZE=integer      for the minimum window size for detecting purge-candidates (default=$MIN_WD_SIZE_PCAND);\n";
	## print STDERR "            --MAX_WD_SIZE=integer      for the maximum window size for detecting purge-candidates (default=$MAX_WD_SIZE_PCAND);\n";
	print STDERR "           --HF_WDTH_SHIFT=integer     for the maximum number of \"shift\"s of each gap-block in one direction (default=$HF_WDTH_SHIFT);\n";  # Added on 2019/06/29.
	print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # ADDED on 2019/08/01.

			# ADDED in REViSION (on 2019/10/18). (ARGV1) #
	print STDERR "           --IF_TEST_W_COMPACT_MSS=boolean     for whether tests with \"compact\" multiple-shift spaces are performed (=1) or not (=0) (default=$IF_TEST_W_COMPACT_MSS);\n";
	print STDERR "           --HF_WDTH_SHIFT_4_CMSS=integer      for the maximum number of \"shift\"s of each gap-block in one direction WHEN \$IF_TEST_W_COMPACT_MSS=1  (default=$HF_WDTH_SHIFT_4_CMSS);\n";
			# END of "ADDED in REViSION (on 2019/10/18). (ARGV1)" #

	print STDERR "\n"; # ADDED in REVISION (on 2019/10/18). (ARGV2) #

	die "$!\n\n";
}
my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	# NOTE: The input/output file system will be specified & prepared below. #

	####### END of "Take in the Command-Line Arguments." #######


	##### Check the Command-Line Options. #####

for (my $k=3; $k <= $#ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'MIN_WD_SIZE') {
	    $MIN_WD_SIZE_PCAND = $value;

	##} elsif ($option eq 'MAX_WD_SIZE') {
	##    $MAX_WD_SIZE_PCAND = $value;

		# Added on 2019/06/29. (OP) #
	} elsif ($option eq 'HF_WDTH_SHIFT') {
	    $HF_WDTH_SHIFT = $value;
		# End of "Added on 2019/06/29. (OP)" #

		# Added on 2019/08/01. (OP) #
	} elsif ($option eq 'XLEVEL') {
	    $XLEVEL = $value;
		# End of "Added on 2019/08/01. (OP)" #

		# Added on 2019/10/18. (OP) #
	} elsif ($option eq 'IF_TEST_W_COMPACT_MSS') {
	    $IF_TEST_W_COMPACT_MSS = $value;

	} elsif ($option eq 'HF_WDTH_SHIFT_4_CMSS') {
	    $HF_WDTH_SHIFT_4_CMSS = $value;
		# End of "Added on 2019/10/18. (OP)" #

	} else {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

}	# End of the preliminary for-loop (over additional command-line arguments. #

	##### End of "Check the Command-Line Options." #####`


	##### Specify & Check Input/Output Files/Directories. #####

## my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	### Sub-working directories. ###

my $pref_sbworkdir = 'Sbwd';
my $suff_sbworkdir = '';

my $sbwd_IDs = list_sessions_ext ($top_work_dir, $pref_sbworkdir, $suff_sbworkdir, 'd');
my $ct_sbwds = @{$sbwd_IDs};
my @sub_work_dirs = ();
for (my $sw=0; $sw <$ct_sbwds; $sw++) { # 1st preliminary for-loop (over sub-windows).
	my $dir = join ('', $top_work_dir, '/', $pref_sbworkdir, $sbwd_IDs->[$sw], $suff_sbworkdir);
	$sub_work_dirs[$sw] = \$dir;
} # End of the 1st preliminary for-loop (over sub-windows). #


	### Input files/directories. ###

my $intree_file = $dawg_ctrl_file;
my $infile_delta_del = "$indir_mfacs/$FILENAME_DELTA_DEL"; # This file stores '$delta_del', which is important for calculating the exit rate of the entire sequence. #

my $maindir_in = "$top_work_dir/Inputs";

my $infile_msa_in_wd = "$maindir_in/msa_in_wd.txt";
my $infile_set_clms0 = "$maindir_in/set_columns0.txt";
my $infile_list_gblks0  = "$maindir_in/list_gblocks0.txt";
my $infile_pcands0 = "$maindir_in/list_rlv_pcands.txt"; # Added in REVISION (on 2019/07/03). #

#	{Main_Input_Directory}/ = "{Top_Working_Directory}/Inputs/" .
#	{Input_file_for_MSA_in_the_window} = "{Main_Input_Directory}/msa_in_wd.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Main_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Main_Input_directory}/list_gblocks0.txt" .

my @subdirs_in = ();
my @infiles_pcands = my @infiles_sbset_clms0 = my @infiles_sblist_gblks0 = ();
for (my $sw=0; $sw < $ct_sbwds; $sw++) { # 2nd preliminary for-loop (over sub-windows). #
	my $sbwork_dir = ${$sub_work_dirs[$sw]};
	my $sbdir_in = "$sbwork_dir/Inputs";
	$subdirs_in[$sw] = \$sbdir_in;

	my $infile_pcands = "$sbdir_in/list_rlv_pcands.txt";
	$infiles_pcands[$sw] = \$infile_pcands;

	my $infile_sbset_clms0 = "$sbdir_in/set_columns0.txt";
	$infiles_sbset_clms0[$sw] = \$infile_sbset_clms0;

	my $infile_sblist_gblks0 = "$sbdir_in/list_gblocks0.txt";
	$infiles_sblist_gblks0[$sw] = \$infile_sblist_gblks0;

} # End of the 2nd preliminary for-loop (over sub-windows). #

#	{Sub_Working_Directory}/ = "{Top_Working_Directory}/Sbwd{sub-window-ID}/" .
#	{Sub_Input_Directory}/   = "{Sub_Working_Directory}/Inputs/" .
#	{Input_file_for_pcands} = "{Sub_Input_Directory}/list_rlv_pcands.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Sub_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Sub_Input_Directory}/list_gblocks0.txt" .

	# Check input files/directories. #

unless (( -d $indir_mfacs ) and ( -r $indir_mfacs )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY, '$indir_mfacs', that stores data files for the multiplication factors !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

foreach my $filename (@FILENAMES_LOG_MFACS) { # 1st preliminary foreach-loop (over input data files for multiplication factors). #
    my $file = "$indir_mfacs/$filename";
    ##unless (( -f $file ) and ( -r $file )) {
    unless ( (( -f $file ) or ( -l $file )) and ( -r $file )) { # MODIFIED in creation of 'anex_for_sgl_wd.ver0.7.pl' (on 2020/04/24). (CHK) #
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 1st preliminary foreach-loop (over input data files for multiplication factors). #

foreach my $file ($dawg_ctrl_file, $intree_file, $infile_delta_del) { # 2nd preliminary foreach-loop (over some key input files). #
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 2nd preliminary foreach-loop (over some key input files). #

foreach my $rf_dir (\$top_work_dir, \$maindir_in, @subdirs_in) { # 3rd preliminary foraech-loop (over some key directories). #
    my $dir = ${$rf_dir};
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 3rd preliminary foreach-loop (over some key directories). #

foreach my $rf_file (\$infile_msa_in_wd, \$infile_set_clms0, \$infile_list_gblks0, \$infile_pcands0,
		@infiles_pcands, @infiles_sbset_clms0, @infiles_sblist_gblks0) { # 5th preliminary foreach-loop (over some input files). #
		# Added \$infile_pcands0 on 2019/07/03.

    my $file = ${$rf_file};
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 5th preliminary foreach-loop (over some input files). #

	### Output directories. ###

my $maindir_in_chk = "$top_work_dir/Check_Inputs";
my $maindir_interim = "$top_work_dir/Interim_Data";
my $maindir_out = "$top_work_dir/Outputs";

my @subdirs_in_chk = my @subdirs_interim = my @subdirs_out = ();
for (my $sw=0; $sw < $ct_sbwds; $sw++) { # 3rd preliminary for-loop (over sub-windows). #
	my $sbwork_dir = ${$sub_work_dirs[$sw]};

	my $sbdir_in_chk = "$sbwork_dir/Check_Inputs";
	$subdirs_in_chk[$sw] = \$sbdir_in_chk;

	my $sbdir_interim = "$sbwork_dir/Interim_Data";
	$subdirs_interim[$sw] = \$sbdir_interim;

	my $sbdir_out = "$sbwork_dir/Outputs";
	$subdirs_out[$sw] = \$sbdir_out;

} # End of the 3rd preliminary for-loop (over sub-windows). #

foreach my $rf_dir (\$maindir_in_chk, \$maindir_interim, \$maindir_out, @subdirs_in_chk, @subdirs_interim, @subdirs_out) { # 6th preliminary foreach-loop (over putput directories). #
    my $dir = ${$rf_dir};
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the 6th preliminary foreach-loop (over output directories). #


		## "Synopsis" Output File, which summarizes the results of ALL multiple-"shift"s performed. ##
		## (ADDED in REVISION on 2020/04/20.) (I/O-F/D) ##

my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";

		## End of "(ADDED in REVISION on 2020/04/20.) (I/O-F/D)" ##


	##### End of "Specify & Check Input/Output Files/Directories." #####


	##### Declare the Beginning of the Script #####

if ($TEST_LEVEL_MAIN>0) { # (1) #

	print STDOUT "\n<<<< Program name: $plscript_name (path: $plscript_path) >>>>\n\n";

  if ($TEST_LEVEL_MAIN>1) {

	print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

	print STDOUT "There are $ct_sbwds sub-windows: '", join ("', '", @{$sbwd_IDs}), "' .\n\n";

	print STDOUT "<< Input/Output Files/Directories >>\n\n";

## my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	print STDOUT "< Key Input Files: >\n\n";

	print STDOUT "Control file: $dawg_ctrl_file .\n\n";

	print STDOUT "Input tree and indel parameters: $intree_file .\n\n";

	print STDOUT "Input file recording ", '$delta_del', ": $infile_delta_del .\n\n";

	print STDOUT "Input directory storing (log-)multiplication factors: $indir_mfacs ,\n\n";
	print STDOUT "   which holds the following input files:\n\n";
	print STDOUT '     ', join (', ', @FILENAMES_LOG_MFACS), " .\n";
	print STDOUT "\n\n";

	print STDOUT "< Working File System: >\n\n";

	print STDOUT "{Top Working Directory}: $top_work_dir .\n";
	print STDOUT "{Working Sub-directories}:\n";
	for (my $sw=0; $sw <$ct_sbwds; $sw++) { # 1st in/out for-loop (over sub-windows). #
	    print STDOUT "\t\t", ${$sub_work_dirs[$sw]}, (($sw < $ct_sbwds-1) ? " ,\n" : " .\n\n");
	} # End of the 1st in/out for-loop (over sub-windows). #

	print STDOUT "< Input Files/Directories: >\n\n";

	print STDOUT "[ For the full Window: ]\n\n";
	print STDOUT "{Main Input Directory}:       $maindir_in .\n";
	print STDOUT "{Input MSA in the window}:    $infile_msa_in_wd ,\n";
	print STDOUT "{Initial set of columns}:     $infile_set_clms0 ,\n";
	print STDOUT "{Initial list of gap-blocks}: $infile_list_gblks0 ,\n";
	print STDOUT "{List of purge-candidates}:   $infile_pcands0 .\n\n"; # Added in REVISION (on 2019/07/03). #

	for (my $sw=0; $sw <$ct_sbwds; $sw++) { # 2nd in/out for-loop (over sub-windows). #
	    my $sbwd_id = $sbwd_IDs->[$sw];
	    print STDOUT ' ' x 3, "[ For the sub-window, \"$sbwd_id\": ]\n\n";
	    print STDOUT ' ' x 3, '{Input sub-directory}:            ', ${$sub_work_dirs[$sw]}, " .\n";
	    print STDOUT ' ' x 3, '{List of purge-candidates}:       ', ${$infiles_pcands[$sw]}, " ,\n";
	    print STDOUT ' ' x 3, '{Initial sub-set of columns}:     ', ${$infiles_sbset_clms0[$sw]}, " ,\n";
	    print STDOUT ' ' x 3, '{Initial sub-list of gap-blocks}: ', ${$infiles_sblist_gblks0[$sw]}, " .\n\n";
	} # End of the 2nd in/out for-loop (over sub-windows). #
	print STDOUT "\n";

				# (ADDED in REVISION on 2020/04/20.) (DC2) #
	print STDOUT "< \"Synopsis\" Output File: >\n\n";

	print STDOUT "$outfile_synopsis .\n\n\n";

# my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";
				# End of "(ADDED in REVISION on 2020/04/20.) (DC2)" #

	print STDOUT "< Output Directories: >\n\n";

	print STDOUT "[ For the full Window: ]\n\n";
	print STDOUT "{Directory to Check the Inputs}: $maindir_in_chk ,\n";
	print STDOUT "{Directory for Interim Data}:    $maindir_interim ,\n";
	print STDOUT "{Directory for the Outputs}:     $maindir_out .\n\n";

	for (my $sw=0; $sw <$ct_sbwds; $sw++) { # 3rd in/out for-loop (over sub-windows). #
	    my $sbwd_id = $sbwd_IDs->[$sw];
	    print STDOUT ' ' x 3, "[ For the sub-window, \"$sbwd_id\": ]\n\n";
	    print STDOUT ' ' x 3, '{Sub-directory to check the inputs}: ', ${$subdirs_in_chk[$sw]}, " ,\n";
	    print STDOUT ' ' x 3, '{Sub-directory for interim data}:    ', ${$subdirs_interim[$sw]}, " ,\n";
	    print STDOUT ' ' x 3, '{Sub-directory for the Outputs}:     ', ${$subdirs_out[$sw]}, " .\n\n";
	} # End of the 3rd in/out for-loop (over sub-windows). #

	print STDOUT "\n\n";


	print STDOUT "<< User-specified Parameters >>\n\n";

	print STDOUT '$STRING_UNDEF = ', $STRING_UNDEF, " .\n\n"; # Added on 2019/07/04. #

	print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	print STDOUT '$CHAR_ABS  =  $GAP  = ', $GAP,  " ,\n";
	print STDOUT '$CHAR_PRES =  $NGAP = ', $NGAP, " .\n\n";

	print STDOUT "{A displayed alignment will be chopped into chunks of length $DIS_LEN, & a shorter remainder.}\n\n";
	print STDOUT '$MyANEX_IO::DIS_LEN = ', $MyANEX_IO::DIS_LEN, " .\n\n"; # Added in REVISION (on 2019/07/07). #

=Set Aside for Later Use. (PR 1)
	print STDOUT "Characters indicating that the site is missing or that its orthology is questionable:\n\n";
	print STDOUT ' ' x 5, join (', ', sort keys %CHAR_MISSING), " .\n";
	print STDOUT "\n\n";
=cut # END of "Set Aside for Later Use. (PR 1)"

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS =' , $MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS, " ,\n";
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = ', $MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS, " ,\n\n";
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_GAPLEN = ', $UPPER_BOUND_ANAL_GAPLEN, " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " .\n";
	print STDOUT "\n\n";

			# Added on 2019/08/01. (DC) #
	print STDOUT "< The LEVEL of MSA Space Exploration. >\n\n";
	print STDOUT '$XLEVEL = ', $XLEVEL, " .\n";
	print STDOUT "\n";
			# End of "Added on 2019/08/01. (DC)" #

	print STDOUT "< Range of sizes of windows for detecting purge-candidates. >\n\n";
	print STDOUT '($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) = ', "($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) .\n";
	print STDOUT "\n";

			# Added on 2019/06/29. (PR 3) #
	print STDOUT "< Half-width of the coordinate of each gap-block when performing multiple-\"shift\"s. >\n\n";
	print STDOUT '$HF_WDTH_SHIFT = ', "$HF_WDTH_SHIFT .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/06/29. (PR 3)" #

			# Added on 2019/10/18. (DC) #
	print STDOUT "< Options regarding the tests with \"compact\" multiple-shift spaces. >\n\n";
	print STDOUT '$IF_TEST_W_COMPACT_MSS = ', $IF_TEST_W_COMPACT_MSS, " .\n";
	if ($IF_TEST_W_COMPACT_MSS) {
	    print STDOUT '$HF_WDTH_SHIFT_4_CMSS  = ', $HF_WDTH_SHIFT_4_CMSS, " .\n";
	}
	print STDOUT "\n\n";
			# End of "Added on 2019/10/18. (DC)" #

	print STDOUT "< Parameters specifying How verbose the Test Outputs will be: >\n\n";

	print STDOUT '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

	print STDOUT '$TEST_LEVEL_MAIN = ', $TEST_LEVEL_MAIN, " ,n";
	print STDOUT '$MyANEX_IO::TEST_LEVEL_MAIN = ', $MyANEX_IO::TEST_LEVEL_MAIN, " .\n\n"; # Added in REVISION (on 2019/07/07). #

	print STDOUT '$TEST_LEVEL_SUB  = ', $TEST_LEVEL_SUB , " ,\n"; # Added on 2019/07/02.
	print STDOUT '$MyANEX_IO::TEST_LEVEL_SUB = ', $MyANEX_IO::TEST_LEVEL_SUB, " .\n\n"; # Added in REVISION (on 2019/07/07). #

	print STDOUT '$MyReadAlnProbIngredients::TEST_LEVEL1 = ', $MyReadAlnProbIngredients::TEST_LEVEL1, " .\n\n";

	print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca, " ,\n\n";
	print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa, " .\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = ', $MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE, " .\n\n";

=Set Aside for Later Use. (PR 2)
	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs , " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs, " ,\n\n";
=cut # END of "Set Aside for Later Use. (PR 2)"

	print STDOUT "\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) {...}"

} # END of "if ($TEST_LEVEL_MAIN>0) { # (1) # ...}"

	##### End of "Declare the Beginning of the Script" #####


	##### Read in the Control file & Set Some Model Parameters. #####

print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for substitution model parameters. >>>\n\n\n";

my ($model_sbst, $nt_freqs, $sbst_params) = read_dawg_ctrl_file_sbst ($dawg_ctrl_file);
unless (defined $model_sbst) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $MODEL_SBST = $model_sbst;

our @NT_FREQS = @{$nt_freqs}[0,2,1,3]; # Re-arrange the frequencies: from (ACGT) in Dawg to (AGCT) in this package.
##our @NT_FREQS = @{$nt_freqs};

@root_resfreqs = @NT_FREQS; # ADDED on 2019/06/27.

our @SBST_PARAMS = @{$sbst_params};

if ($TEST_LEVEL_MAIN>2) { # (2A) #

	## Print out the Parameters for Substitution Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Substitution Probabilities ]]\n\n";

	print STDOUT '$MODEL_SBST = ', $MODEL_SBST, " .\n\n";

	##print STDOUT '@NT_FREQS = (', join (', ', @NT_FREQS), ") .\n\n";
	print STDOUT '@NT_FREQS (=@root_resfreqs) = (', join (', ', @NT_FREQS), ") .\n\n"; # MODIFIED on 2019/06/27.

	print STDOUT '@SBST_PARAMS = (', join (', ', @SBST_PARAMS), ") .\n\n";

	print STDOUT "\n";

} # End of "if ($TEST_LEVEL_MAIN>2) { # (2A) # ...}"


	# Translate the @SBST_PARAMS into the Parameters for the Tamura-Nei model. #

our ($ALPHA1, $ALPHA2, $BETA);
	# $ALPHA1 = rate for a transitional substitution (A <-> G); 
	# $ALPHA2 = rate for a transitional substitution (C <-> T);
	# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).

if (($MODEL_SBST eq 'JC') or ($MODEL_SBST eq 'F81')) {

	($ALPHA1, $ALPHA2, $BETA) = (1, 1, 1);

} elsif (($MODEL_SBST eq 'K2P') or ($MODEL_SBST eq 'HKY')) {

	($ALPHA1, $BETA) = @SBST_PARAMS;
	 $ALPHA2 = $ALPHA1;
	
} elsif ($MODEL_SBST eq 'TN') {

	($ALPHA1, $ALPHA2, $BETA) = @SBST_PARAMS;
} else {
	print STDOUT "[ERROR '$program_name'] The current version does NOT HANDLE the substitution model, '$MODEL_SBST'! ... USE either 'JC', 'K2P', 'HKY', 'F81' or 'TN'!!\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL_MAIN>1) { # (2A') #
	print STDOUT '[ Translation of @SBST_PARAMS into Parameters of Tamura-Nei model ]', "\n\n";
	print STDOUT '($ALPHA1, $ALPHA2, $BETA) = ', "($ALPHA1, $ALPHA2, $BETA) .\n\n";
	print STDOUT "# NOTE:\n";
	print STDOUT '# $ALPHA1 = rate for a transitional substitution (A <-> G);', "\n";
	print STDOUT '# $ALPHA2 = rate for a transitional substitution (C <-> T);', "\n";
	print STDOUT '# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).', "\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (2A') # ...}"


print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for indel model parameters and phylogenetic tree. >>>\n\n\n";

my ($tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del, $gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate)
	= read_dawg_ctrl_file3 ($dawg_ctrl_file);
unless (defined $tree_string) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $INS_MODEL = $gapmodel_ins;
our $DEL_MODEL = $gapmodel_del;
our $DISTRI_INS = $INDEL_MODEL2DISTRI{$INS_MODEL};
our $DISTRI_DEL = $INDEL_MODEL2DISTRI{$DEL_MODEL};

	# Total rates of insertions & deletions at a site relative to the substitution rate.
our $TOTRATE_INS = $ins_rate;
our $TOTRATE_DEL = $del_rate;

 	# Exponents for the power-law distributions of insertion/deletion lengths (when $INDEL_MODEL = 'zipf').
our $PARAM_INS = $gap_param_ins;
our $PARAM_DEL = $gap_param_del;

	# Cut-off insertion/deletion lengths.
our $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
our $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL); 


if ($TEST_LEVEL_MAIN>1) { # (2B) #

	## Print out the Parameters for Indel Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Indel Probabilities ]]\n\n";

	print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
	print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";

	print STDOUT '$CUTOFF_INS_LEN = ', "$CUTOFF_INS_LEN ,\n";
	print STDOUT '$CUTOFF_DEL_LEN = ', "$CUTOFF_DEL_LEN .\n";
	print STDOUT "\n\n";

	## End of "Print out the Parameters for Indel Probabilities." ##

} # End of "if ($TEST_LEVEL_MAIN>1) { # (2B) # ...}"

	##### End of "Read in the Control file." #####


	##### Read in the $delta_del. #####

(print STDOUT "\n<<< Read in the multiplication factor '\$delta_del' for a 'universal' term of log-lilekihood ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $delta_del = readin_delta_del ($infile_delta_del);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_delta_del' FAILED on the input file, '$infile_delta_del' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}
(print STDOUT "\n ... The subroutine, 'readin_delta_del' Ended (w/ \$delta_del = $delta_del) ...\n\n") if ($TEST_LEVEL_MAIN> 0);

	##### End of "Read in the $delta_del." #####


	##### Read in the INPUT TREE. #####

(print STDOUT "\n<<< Read in the Input Tree >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
# [my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches. ] --- for the predecessor 'br_infer_psm_indel_events.pl'. #
my %br_attr; # = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches.
my %name2id_intnode; # = ($node_name => $node_id, ...) for all interior nodes. #
my $topnode_name; # The name of the top-node (root).

unless ( input_tree_file_dawg (${$tree_string}, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $topnode_id = $name2id{$TOP_NODE};


if (1) { # (3) # MODIFIED on 2019/07/17. # 
##if ($TEST_LEVEL_MAIN>2) { # (3) #

	print STDOUT "\n ... To check the input tree, perform the subroutine, 'print_out_input_tree' ...\n\n";

	unless ( print_out_input_tree ($maindir_in_chk, $topnode_id, %node2ch, %node2pa, %id2name, %br_attr, $topnode_name) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_input_tree' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_input_tree' Ended ...\n";
	print STDOUT "\n\n";

##} # End of "if ($TEST_LEVEL_MAIN>2) { # (3) # ...}"
} # End of "if (1) { # (3) # MODIFIED on 2019/07/17. # ...}"

	##### End of "Read in the INPUT TREE." #####


	##### Assign Log-Multiplication Factors to Each Branch. #####

(print STDOUT "\n<<< Assign the logarithmic multiplication factors to each branch, based on the files in '$indir_mfacs' ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $br2log_mfacs = assign_br2log_mfacs (%br_attr, $TOTRATE_INS + $TOTRATE_DEL, $indir_mfacs, @FILENAMES_LOG_MFACS);
unless (defined $br2log_mfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'assign_br2log_mfacs' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'assign_br2log_mfacs', Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN >= 5) { # (5) #

	print STDOUT "\n ... To check the outputs of 'assign_br2log_mfacs', Run the subroutine, 'print_out_br2log_mfacs' ...\n\n";

# $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
# $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

	unless ( print_out_br2log_mfacs ($maindir_in_chk, %{$br2log_mfacs}, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2log_mfacs' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_br2log_mfacs' Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN >= 5) { # (5) # ...}"

	##### End of "Assign Log-Multiplication Factors to Each Branch." #####


	##### Read in the INPUT MSA. #####

(print STDOUT "\n<<< Read in the Input MSA, via the subroutine, 'extract_alignment' >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my ($in_msa, $seqnames, $seq_lens) = extract_alignment ($infile_msa_in_wd);
unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from the file '$infile_msa_in_wd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_seqs = @{$seqnames};

unless ($ct_seqs>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] NO SEQUENCES in the input MSA from the file '$infile_msa_in_wd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_aln = length ${$in_msa->[0]} ;

unless ($totlen_aln>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] ZERO-LENGTH SEQUENCES in the input MSA from the file '$infile_msa_in_wd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=SET ASIDE for potential use.
	# REMOVE the Ancestral Sequences from the Input MSA. #
	# (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my @msa_extseqs = my @ext_seqnames = ();
for (my $i=0; $i<$ct_seqs; $i++) { # Preliminary for-loop (over sequences in the input MSA). #

	my $seqid = $seqnames->[$i];
	($seqid  =~ m/^anc/) and next;

	push @ext_seqnames, $seqid;
	push @msa_extseqs, $in_msa->[$i];

}			# End of the preliminary for-loop (over sequences in the input MSA). #

@{$in_msa} = @msa_extseqs;
@{$seqnames} = @ext_seqnames;
$ct_seqs = @{$seqnames};

=cut # End of "SET ASIDE for potential use."

(print STDOUT "\n ... The subroutine, 'extract_alignment', Ended (w/ #{(external) sequences}=$ct_seqs, #{columns}=$totlen_aln) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (6) #

	my $chk_imsa_file = "$maindir_in_chk/in_msa.txt";
	my $fh_ci_msa = open_output_file ($chk_imsa_file, 0);
	unless (defined $fh_ci_msa) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}

	print STDOUT "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";
	print $fh_ci_msa "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";

	unless ( pr_ODEN_input (@{$seqnames}, @{$in_msa}, $DIS_LEN, *${$fh_ci_msa}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_ODEN_input'  FAILED to print out the input MSA into \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print $fh_ci_msa "\n";
	print STDOUT "\n ... the input MSA has been written into the file, \"$chk_imsa_file\" ...\n\n";

	close ($fh_ci_msa);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (6) # ...}"

	##### End of "Read in the INPUT MSA." #####


	##### Read in the Initial Set of Columns. #####

## my $infile_set_clms0 = "$maindir_in/set_columns0.txt";

my ($beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, $set_clms0_in_wd)
	= readin_set_clms ($infile_set_clms0);
unless (defined $set_clms0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_set_clms', FAILED on the input file, \"$infile_set_clms0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
my $ct_clms0_in_wd = @{$set_clms0_in_wd};

(print STDOUT "\n ... The subroutine, 'readin_set_clms' Ended (w/ \$ct_clms0_in_wd = $ct_clms0_in_wd, [\$beg_wd, \$end_wd] = [$beg_wd, $end_wd], [\$if_on_le_msa, \$if_on_re_msa] = [$if_on_le_msa, $if_on_re_msa]) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if ($TEST_LEVEL_MAIN>1) { # (7) #

	my $fname_set_clms0 = 'set_columns0.txt';

	unless ( print_out_set_clms ($maindir_in_chk, $fname_set_clms0, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING, the subroutine, 'print_out_set_clms', Ended RECORDING information (on the main window) into the output file, \"$maindir_in_chk/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (7) # ...}"

	##### End of "Read in the Initial Set of Columns." #####


	##### Read in the Initial List of Gap-Blocks. #####

## my $infile_list_gblks0  = "$maindir_in/list_gblocks0.txt";

my ($list_gblks0_in_wd) = readin_list_gblks ($infile_list_gblks0);
unless (defined $list_gblks0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_list_gblks', FAILED on the input file, \"$infile_list_gblks0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
my $ct_gblks0_in_wd = @{$list_gblks0_in_wd};

(print STDOUT "\n ... The subroutine, 'readin_list_gblks' Ended (w/ \$ct_gblks0_in_wd = $ct_gblks0_in_wd) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (8) #

	my $fname_list_gblks0 = 'list_gblocks0_org.txt';

	unless ( print_out_readin_list_gblks ($maindir_in_chk, $fname_list_gblks0, @{$list_gblks0_in_wd}, 'org') ) { # REVISED on 2019/06/27.
	##unless ( print_out_readin_list_gblks ($maindir_in_chk, $fname_list_gblks0, @{$list_gblks0_in_wd}) ) { ...}
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING 'readin_list_gblks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'original' gap-blocks in the window) into the output file, \"$maindir_in_chk/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (8) # ...}"


	## NOTE: Because the node (branch) ID assignments may have changed from the environment where $infile_list_gblks0 was output to the current environment,
	##	we MUST examine the node (branch) IDs, and, if necessary, REVISE them in "(Pre 3)"!!

	##### End of "Read in the Initial List of Gap-Blocks." #####



	####### Perform some Preliminary Processes #######

(print STDOUT "\n<<<<< Perform some Preliminary Processes >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


	## (Pre 1) Assign indices of MSA rows under each node to the node. ##

(print STDOUT '<<< Preliminary 1: Compute %{$node2exoffs_msa_rows}, %{$node2depth2}, @{$nodeids_in_msa} & %{$nodeid2msa_row}, ', "by performing 'wrapper_mk_node2exoffs_msa_rows' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my ($node2exoffs_msa_rows, $node2depth2, $nodeids_in_msa, $nodeid2msa_row)
	= wrapper_mk_node2exoffs_msa_rows ($topnode_id, %node2ch, @{$seqnames}, %name2id);
unless (defined $node2exoffs_msa_rows) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'wrapper_mk_node2exoffs_msa_rows', FAILED on the input tree !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'wrapper_mk_node2exoffs_msa_rows' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if (1) { # (Pre 1) # MODIFIED on 2019/07/17. #
##if ($TEST_LEVEL_MAIN>3) { # (Pre 1) #

	my $ofile_name = 'out_wrapper_mk_node2exoffs_msa_rows.txt';

	print STDOUT "\n ... Check the outputs of 'wrapper_mk_node2exoffs_msa_rows', by performing 'print_out_wrapper_mk_node2exoffs_msa_rows' ...\n\n";
	unless ( print_out_wrapper_mk_node2exoffs_msa_rows ($maindir_interim, $ofile_name, %{$node2exoffs_msa_rows}, %{$node2depth2}, @{$nodeids_in_msa}, %{$nodeid2msa_row}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_wrapper_mk_node2exoffs_msa_rows' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	##print STDOUT "\n ... The subroutine, 'print_out_wrapper_mk_node2exoffs_msa_rows', Ended ...\n\n";
	print STDOUT "\n ... The subroutine, 'print_out_wrapper_mk_node2exoffs_msa_rows', Ended  (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n"; # Modified on 2020/05/08.

##} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 1) # ...}"
} # End of "if (1) { # (Pre 1) # MODIFIED on 2019/07/17. # ...}"

	## End of "(Pre 1) Assign indices of MSA rows under each node to the node." ##


	## (Pre 2) Using %{$node2exoffs_msa_rows}, assosiate some sets of MSA rows to the upper- & lower-sides of the branches. ## (Added on 2019/06/25.)

(print STDOUT '<<< Preliminary 2: Compute %{$set_msa_rows2br_up_or_lw}, ', "by performing 'mk_set_msa_rows2br_up_or_lw_dp' >>>\n\n") if ($TEST_LEVEL_MAIN>1); # Modified on 2019/07/03.
##(print STDOUT '<<< Preliminary 2: Compute %{$set_msa_rows2br_up_or_lw}, ', "by performing 'mk_set_msa_rows2br_up_or_lw' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my @set_all_msa_rows = (0 .. $ct_seqs -1);

my ($set_msa_rows2br_up_or_lw) = mk_set_msa_rows2br_up_or_lw_dp (%{$node2exoffs_msa_rows}, @set_all_msa_rows, %{$node2depth2}); # Modified on 2019/07/03.
##my ($set_msa_rows2br_up_or_lw) = mk_set_msa_rows2br_up_or_lw (%{$node2exoffs_msa_rows}, @set_all_msa_rows);
unless (defined $set_msa_rows2br_up_or_lw) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_set_msa_rows2br_up_or_lw_dp' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'mk_set_msa_rows2br_up_or_lw_dp' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (Pre 2) #

	my $ofile_name = 'out_mk_set_msa_rows2br_up_or_lw.txt';

	print STDOUT "\n ... Check the outputs of 'mk_set_msa_rows2br_up_or_lw_dp', by performing 'print_out_mk_set_msa_rows2br_up_or_lw' ...\n\n"; # Modified on 2019/07/03.
	##print STDOUT "\n ... Check the outputs of 'mk_set_msa_rows2br_up_or_lw', by performing 'print_out_mk_set_msa_rows2br_up_or_lw' ...\n\n";
	unless ( print_out_mk_set_msa_rows2br_up_or_lw ($maindir_interim, $ofile_name, %{$set_msa_rows2br_up_or_lw}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_mk_set_msa_rows2br_up_or_lw' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_mk_set_msa_rows2br_up_or_lw', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 2) # ...}"


	## End of "(Pre 2) Using %{$node2exoffs_msa_rows}, assosiate some sets of MSA rows to the upper- & lower-sides of the branches." ##


	## (Pre 3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary. ## (Added on 2019/06/26-27)

(print STDOUT '<<< Preliminary 3: Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary,', " via 'revise_list_gblocks' >>>\n\n") if ($TEST_LEVEL_MAIN>1);


unless ( revise_list_gblocks (@{$list_gblks0_in_wd}, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa}) ) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'revise_list_gblocks' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'revise_list_gblocks' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>1) { # (Pre 3) #

	my $fname_list_gblks0 = 'list_gblocks0_rvsd.txt';

	unless ( print_out_readin_list_gblks ($maindir_interim, $fname_list_gblks0, @{$list_gblks0_in_wd}, 'rvsd') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING 'revise_list_gblocks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'revised' gap-blocks in the window) into the output file, \"$maindir_interim/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 3) # ...}"

	## End of "(Pre 3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary." ##


	### (Pre 5)  Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, via 'enum_psm_indel_hsts_for_msa_finer_PRE'. ### (Added on 2019/06/25)

(print STDOUT '<<< Preliminary 5: Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, ', "by performing 'enum_psm_indel_hsts_for_msa_finer_PRE' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

(print STDOUT "\n (If you want to know what's going on, SWITCH ON the global variable, \$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE .)\n\n") if ($TEST_LEVEL_MAIN>1);

my ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes)
	= enum_psm_indel_hsts_for_msa_finer_PRE ($topnode_id, %node2ch, %node2pa);
unless (defined $br2nn) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (Pre 5) #

	print STDOUT "\n ... Check the outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE', by performing 'print_out_tree_profile' ...\n\n";

	unless ( print_out_tree_profile ($maindir_interim, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, %{$br2nn}, @{$set_all_nodes}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_tree_profile' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_tree_profile', Ended (w/ records under \"$maindir_interim/\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 5) # ...}"

		# Create a set of all external nodes. #
		# (ADDED on 2019/08/24.) #
my @set_exnodes = ();
foreach my $node (@{$set_all_nodes}) {
	my $ch = $node2ch{$node};
	unless ((defined $ch) and (@{$ch}>0)) { push @set_exnodes, $node; }
}
if ($TEST_LEVEL_MAIN>3) { # (Pre 5') #
	print STDOUT "\n", 'Manuall Set: @set_exnodes = (', join (', ', @set_exnodes), ") .\n\n\n";
} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 5') # ...}"

		# End of "(ADDED on 2019/08/24.)" #

	### End of "(Pre 5)  Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, via 'enum_psm_indel_hsts_for_msa_finer_PRE'." ###


	### (Pre 6) Compute %{$br2tr_mtrx}, via 'calc_br2tr_mtrx_TN'. ### (Added on 2019/06/27)

(print STDOUT "<<< Preliminary 6: Assign a substitution transition matrix to each branch, via 'calc_br2tr_mtrx_TN' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

# @root_resfreqs = @NT_FREQS; # ADDED on 2019/06/27.
# our ($ALPHA1, $ALPHA2, $BETA);

my $br2tr_mtrx = calc_br2tr_mtrx_TN (%br_attr, @root_resfreqs, $ALPHA1, $ALPHA2, $BETA);
unless (defined $br2tr_mtrx) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'calc_br2tr_mtrx_TN', FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_br2tr_mtrx_TN', Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);

if ($TEST_LEVEL_MAIN >2) { # (Pre 6) #

	my $ofile_name = 'br2tr_mtrx.txt';

	print STDOUT "\n ... Check the outputs of 'calc_br2tr_mtrx_TN', by performing 'print_out_br2tr_mtrx' ...\n\n";
	unless ( print_out_br2tr_mtrx ($maindir_interim, $ofile_name, %{$br2tr_mtrx}, @RESIDUES) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2tr_mtrx' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2tr_mtrx', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 6) # ...}"

	### End of "(Pre 6) Compute %{$br2tr_mtrx}, via 'calc_br2tr_mtrx_TN'." ### 


	### (Pre 6') Compute the residue frequences at ALL nodes. ### (Added on 2019/06/27)

(print STDOUT "<<< Preliminary 6': Compute the residue frequences at ALL nodes, via 'calc_node2ntfreqs_td'. >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my $node2ntfreqs = calc_node2ntfreqs_td (@root_resfreqs, %{$br2tr_mtrx}, %node2ch, $topnode_id);

unless (defined $node2ntfreqs) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine, 'calc_node2ntfreqs_td', FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_node2ntfreqs_td', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL_MAIN >2) { # (Pre 6') #

	my $ofile_name = 'node2ntfreqs.txt';

	print STDOUT "\n ... Check the outputs of 'calc_node2ntfreqs_td', by performing 'print_out_node2ntfreqs' ...\n\n";
	unless ( print_out_node2ntfreqs ($maindir_interim, $ofile_name, %{$node2ntfreqs}, @labels_resfreqs) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2ntfreqs' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2ntfreqs', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 6') # ...}"

	# End of "(Pre 6') Compute the residue frequences at ALL nodes." #


	### (Pre 7) Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0}, via 'calc_prob_descresstats_cond_ancstat2'. ### (Added on 2019/06/27.)
	# (That is, the probabilities of the residue states of extant sequences on the descendant side of each branch, conditioned on the residue state at either end of the branch.)

(print STDOUT '<<< Preliminary 7: Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0},', " via 'calc_prob_descresstats_cond_ancstat2'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($node2set_sw_cond_probs0, $node2set_sw_extcond_probs0)
 = calc_prob_descresstats_cond_ancstat2 (@{$in_msa}, @{$seqnames}, %name2id, %node2ch, %{$br2tr_mtrx}, $GAP);

unless (defined $node2set_sw_cond_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_prob_descresstats_cond_ancstat2', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);


if ($TEST_LEVEL_MAIN >3) { # (Pre 7) #

	my $ofile_name1 = 'node2set_sw_cond_probs0.txt';
	my $ofile_name2 = 'node2set_sw_extcond_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_prob_descresstats_cond_ancstat2', by performing 'print_out_node2set_sw_cond_probs' TWICE ...\n\n";
	unless ( print_out_node2set_sw_cond_probs ($maindir_interim, $ofile_name1, %{$node2set_sw_cond_probs0}, @NUCLEOTIDES, '%{$node2set_sw_cond_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2set_sw_cond_probs' FAILED for '", '%{$node2set_sw_cond_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2set_sw_cond_probs', Ended on '", '%{$node2set_sw_cond_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name1\") ...\n\n";

	unless ( print_out_node2set_sw_cond_probs ($maindir_interim, $ofile_name2, %{$node2set_sw_extcond_probs0}, @NUCLEOTIDES, '%{$node2set_sw_extcond_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2set_sw_cond_probs' FAILED for '", '%{$node2set_sw_extcond_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2set_sw_cond_probs', Ended on '", '%{$node2set_sw_extcond_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name2\") ...\n\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 7) # ...}"

	# END of "(Pre 7) Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0}, via 'calc_prob_descresstats_cond_ancstat2'." #

###=The Following Preliminary Processes May be UNnecessary; if so, DISABLE them.

	### (Pre 8) Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0}, via 'calc_joint_prob_ancresstats_and_ancstat2'. ### (Added on 2019/06/28.)
	### (That is, the joint probabilities of the residue state of the upper-end of each branch and the residue states of extant sequences on the ancestor side of the branch.)

(print STDOUT '<<< Preliminary 8: Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0},', " via 'calc_joint_prob_ancresstats_and_ancstat2'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($br2set_sw_jnt_probs0, $br2set_sw_extjnt_probs0)
    = calc_joint_prob_ancresstats_and_ancstat2 (@{$in_msa}, @{$seqnames}, %name2id, %node2ch, %{$br2tr_mtrx}, $GAP, @root_resfreqs, %{$node2set_sw_extcond_probs0}) ;

unless (defined $br2set_sw_jnt_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_joint_prob_ancresstats_and_ancstat2', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);


if ($TEST_LEVEL_MAIN >3) { # (Pre 8) #

	my $ofile_name1 = 'br2set_sw_jnt_probs0.txt';
	my $ofile_name2 = 'br2set_sw_extjnt_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_joint_prob_ancresstats_and_ancstat2', by performing 'print_out_br2set_sw_jnt_probs' TWICE ...\n\n";
	unless ( print_out_br2set_sw_jnt_probs ($maindir_interim, $ofile_name1, %{$br2set_sw_jnt_probs0}, @NUCLEOTIDES, '%{$br2set_sw_jnt_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2set_sw_jnt_probs' FAILED for '", '%{$br2set_sw_jnt_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2set_sw_jnt_probs', Ended on '", '%{$br2set_sw_jnt_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name1\") ...\n\n";

	unless ( print_out_br2set_sw_jnt_probs ($maindir_interim, $ofile_name2, %{$br2set_sw_extjnt_probs0}, @NUCLEOTIDES, '%{$br2set_sw_extjnt_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2set_sw_jnt_probs' FAILED for '", '%{$br2set_sw_extjnt_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2set_sw_jnt_probs', Ended on '", '%{$br2set_sw_extjnt_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name2\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 8) # ...}"

	### End of "(Pre 8) Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0}, via 'calc_joint_prob_ancresstats_and_ancstat2'." ###


	### (Pre 9) Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 9: Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my $set_clm_probs0 = calc_clm_probs (@root_resfreqs, %{$node2set_sw_cond_probs0}, $topnode_id);
unless (defined $set_clm_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_clm_probs', Ended ...\n\n") if ($TEST_LEVEL>0);

my $ct_clms0 = @{$set_clm_probs0};
unless ($ct_clms0 == $ct_clms0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The number of columns according to 'calc_clm_probs' (=$ct_clms0) CLASHES with that according to 'readin_set_clms' (=$ct_clms0_in_wd) !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}

if ($TEST_LEVEL_MAIN >3) { # (Pre 9) # 

	my $ofile_name = 'set_clm_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_clm_probs', by performing 'print_out_set_clm_probs' ...\n\n";
	unless ( print_out_set_clm_probs ($maindir_interim, $ofile_name, @{$set_clm_probs0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clm_probs' FAILED on ", '@{$set_clm_probs0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_set_clm_probs', Ended on '", '@{$set_clm_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 9) #  ...}"

	### End of "(Pre 9) Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'." ###

###=cut # END of "The Following Preliminary Processes May be UNnecessary; if so, DISABLE them."


	### (Pre 10) Prepare @{$inter_block_relations0}, via 'mk_inter_block_relations'. ### (Added on 2019/06/29.) 

(print STDOUT '<<< Preliminary 10: Compute @{$inter_block_relations0}, ', " via 'mk_inter_block_relations'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

# @{$list_gblks[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

		# MODIFIED on 2019/07/02. #
our ($INDX_BR, $INDX_U_OR_D) = (0, 1);
(print STDOUT 'Manually Set: ($INDX_BR, $INDX_U_OR_D) = ', "($INDX_BR, $INDX_U_OR_D) .\n\n") if ($TEST_LEVEL_MAIN>1);
##my ($indx_br, $indx_u_or_d) = (0, 1);
##(print STDOUT 'Manually Set: ($indx_br, $indx_u_or_d) = ', "($indx_br, $indx_u_or_d) .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($inter_block_relations0) = mk_inter_block_relations (@{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %node2pa, %node2ch, $topnode_id, %{$br2eq}); # Modified on 2019/07/02. #
##my ($inter_block_relations0) = mk_inter_block_relations (@{$list_gblks0_in_wd}, $indx_br, $indx_u_or_d, %node2pa, %node2ch, $topnode_id, %{$br2eq});
unless (defined $inter_block_relations0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_inter_block_relations' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_inter_block_relations', Ended ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (Pre 10) #

	my $ofile_name = 'inter_block_relations0.txt';

	print STDOUT "\n ... Check the outputs of 'mk_inter_block_relations', by performing 'print_out_inter_block_relations' ...\n\n";
	unless ( print_out_inter_block_relations ($maindir_interim, $ofile_name, @{$inter_block_relations0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_inter_block_relations' FAILED on ", '@{
$inter_block_relations0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_inter_block_relations', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 10) # ...}"

	### End of "(Pre 10) Prepare @{$inter_block_relations0}, via 'mk_inter_block_relations'." ###  


	### (Pre 11) MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd. ### (Added on 2019/06/29.)

(print STDOUT '<<< Preliminary 11: MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd >>>', "\n\n") if ($TEST_LEVEL_MAIN >0);

# @{$list_gblks[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

my @bds_gblocks0_in_wd = ();
my @rows_affd_by_blks0_in_wd = ();
for (my $gb=0; $gb <$ct_gblks0_in_wd; $gb++) {

	my ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, $set_affd_exnodes, $set_affd_msa_rows) = @{$list_gblks0_in_wd->[$gb]};

	$bds_gblocks0_in_wd[$gb] = [$lm_clm - $beg_wd, $rm_clm - $beg_wd];
	$rows_affd_by_blks0_in_wd[$gb] = $set_affd_msa_rows;
}

if ($TEST_LEVEL_MAIN>1) { # (Pre 11) #

	my $ofile_name = 'row_clm_info_gblocks0.txt';

	print STDOUT "\n", ' ... Check the contents of @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd, ', "by performing 'print_out_row_clm_info_gblocks' ...\n\n";
	unless ( print_out_row_clm_info_gblocks ($maindir_interim, $ofile_name, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_row_clm_info_gblocks' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_row_clm_info_gblocks', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 11) # ...}"

	### End of "(Pre 11) MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd." ### 


	### (Pre 12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 12: Prepare the coordinate frame for the multiple-\"shifts\" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($orgs_gb0_coords, $bds_gb0_coords, $gb0_coords_wd_frm)
	= mk_orgs_bds_gb_coords (@bds_gblocks0_in_wd, 0, $end_wd - $beg_wd, @{$inter_block_relations0}, $HF_WDTH_SHIFT);
unless (defined $gb0_coords_wd_frm) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_orgs_bds_gb_coords', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);

if ($TEST_LEVEL_MAIN>1) { # (Pre 12) #

	my $ofile_name = 'orgs_bds_gblocks0_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords', by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($maindir_interim, $ofile_name, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 12) # ...}"

	### End of "(Pre 12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'." ###


	### (Pre 13: IMPORTANT!!) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_wd}. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 13 (IMPORTANT!!): Manually SUBTRACT \$beg_wd (=$beg_wd) from ALL the site numbers in ", '@{$set_clms0_in_wd}', ". >>>\n\n") if ($TEST_LEVEL_MAIN >0);

for (my $c=0; $c <$ct_clms0_in_wd; $c++) { # Outer for-loop (over columns). #

	my $column = $set_clms0_in_wd->[$c];
	my $ct_rows = @{$column};

	for (my $r=0; $r <$ct_rows; $r++) { # Inner for-loop (over rows). #
	    unless ($column->[$r] eq $CHAR_ABS) { $column->[$r] -= $beg_wd; }
	} # End of the inner for-loop (over rows). #

} # End of the outer for-loop (over columns). #

if ($TEST_LEVEL_MAIN>1) { # (Pre 13) #

        my $fname_set_clms0 = 'set_columns0_af_coord_chg.txt';

        ##unless ( print_out_set_clms ($maindir_interim, $fname_set_clms0, @{$set_clms0_in_wd}, 0, $end_wd - $beg_wd, $if_on_le_msa, $if_on_re_msa) ) {
        unless ( print_out_set_clms ($maindir_interim, $fname_set_clms0, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) { # MODIFIED on 2020/05/14.
                print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the main window !\n";
                print STDERR "[ERROR '$plscript_name'] $! !\n";
                die ;
        }
    print STDOUT "\n ... For the purpose of CHECKING the result of  SUBTRACTING \$beg_wd (=$beg_wd) from the sites in the columns, the subroutine, 'print_out_set_clms', Ended RECORDING ", '@{$set_clms0_in_wd}', " (AFTER the SUBTRACTION) into the output file, \"$maindir_interim/$fname_set_clms0\" ... \n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 13) # ...}"

	### End of "(Pre 13) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_wd}." ###


	### (Pre 14) Read in the list of (merged) purge-candidate regions, and SUBTRACT $beg_wd from ALL positions. ### (ADDED on 2019/07/03.)

(print STDOUT "<<< Preliminary 14: Read in the list of (merged) purge-candidate regions, and manually SUBTRACT \$beg_wd (=$beg_wd) from ALL positions. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($list_mgd_pcands0, $maxspan_mgd_pcand0) = readin_list_mgd_pcands ($infile_pcands0);
unless (defined $list_mgd_pcands0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_mgd_pcands' FAILED on the input file, \"$infile_pcands0\" !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_mpcands0 = @{$list_mgd_pcands0};
(print STDOUT "\n ... The subroutine 'readin_list_mgd_pcands' Ended for the main window (w/ #{purge-candidates}=$ct_mpcands0, Max{span}=$maxspan_mgd_pcand0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


unless ( relative_coords_for_list_mgd_pcands (@{$list_mgd_pcands0}, $beg_wd) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'relative_coords_for_list_mgd_pcands' FAILED for ", '@{$list_mgd_pcands0}', " of the main window !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'relative_coords_for_list_mgd_pcands' Ended for ", '@{$list_mgd_pcands0}', " of the main window ...\n\n") if ($TEST_LEVEL_MAIN >1);


if (1) { # (Pre 14) # MODIFIED on 2019/07/17. #
##if ($TEST_LEVEL_MAIN >1) { # (Pre 14) #

	my $ofname_pcands = 'list_rlv_pcands.txt';

	unless ( print_out_list_mgd_pcands ($maindir_interim, $ofname_pcands, @{$list_mgd_pcands0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... For the purpose of CHECKING the result of SUBTRACTING \$beg_wd (=$beg_wd) from the positions regarding the purge-candidate regions, the subroutine, 'print_out_list_mgd_pcands', Ended RECORDING ", '@{$list_mgd_pcands0}', " (AFTER the SUBTRACTION) into the output file, \"$maindir_interim/$ofname_pcands\" ... \n\n";

##} # End of "if ($TEST_LEVEL_MAIN >1) { # (Pre 14) # ...}"
} # End of "if (1) { # (Pre 14) # MODIFIED on 2019/07/17. # ...}"

	### End of "(Pre 14) Read in the list of (merged) purge-candidate regions, and manually SUBTRACT $beg_wd from ALL positions." ###


	### (Pre 15) Prepare @{$gblock_sizes0}, via 'mk_list_gblock_sizes'. ### (ADDED on 2019/07/16.)

(print STDOUT '<<< Preliminary 15: Compute @{$gblock_sizes0}, ', " via 'mk_list_gblock_sizes'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

our ($INDX_LBD, $INDX_RBD) = (0, 1);
(print STDOUT 'Manually Set: ($INDX_LBD, $INDX_RBD) = ', "($INDX_LBD, $INDX_RBD) .\n\n") if ($TEST_LEVEL_MAIN>1); 

my ($gblock_sizes0) = mk_list_gblock_sizes (@bds_gblocks0_in_wd, $INDX_LBD, $INDX_RBD, @{$inter_block_relations0});
unless (defined $gblock_sizes0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_list_gblock_sizes' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_list_gblock_sizes', Ended for the main window ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN >2) { # (Pre 15) #

	my $ofname = 'gblock_sizes0.txt';

	unless ( print_out_list_gblock_sizes ($maindir_interim, $ofname, @{$gblock_sizes0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblock_sizes' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_list_gblock_sizes', Ended on the main window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 15) # ...}"

	### End of "(Pre 15) Prepare @{$gblock_sizes0}, via 'mk_list_gblock_sizes'." ###


	### (Pre 16) Prepare @{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}, via 'mk_lists_gblocks_w_spec_bounds'. ### (ADDED on 2019/07/16-17.) #

(print STDOUT "\n\n<<< Preliminary 16: Compute ", '@{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}', ", via 'mk_lists_gblocks_w_spec_bounds'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($le_wd, $re_wd) = (0, $ct_clms0_in_wd-1);

(print STDOUT 'Manually Set: ($le_wd, $re_wd) = ', "($le_wd, $re_wd) .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($gblks_w_spec_lb0, $gblks_w_spec_rb0) = mk_lists_gblocks_w_spec_bounds ($le_wd, $re_wd, @bds_gblocks0_in_wd);
unless (defined $gblks_w_spec_rb0) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n --- Subroutine 'mk_lists_gblocks_w_spec_bounds' Ended on the main window ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN >2) { # (Pre 16) #

	my $ofname = 'gblks_w_spec_bds0.txt';

	unless ( print_out_gblks_w_spec_bds ($maindir_interim, $ofname, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_gblks_w_spec_bds' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_gblks_w_spec_bds', Ended on the main window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 16) # ...}"

	### End of "(Pre 16) Prepare @{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}, via 'mk_lists_gblocks_w_spec_bounds'." ###


	### (Pre 17) Prepare @{$cts_vincluding0}, @{$cts_vincluded0} & @{$cts_onx0},', " via 'cts_vincluding_vincluded_onx_gblks'. ### (ADDED on 2019/07/16-17.) #

(print STDOUT "\n\n", '<<< Preliminary 17: Compute @{$cts_vincluding0}, @{$cts_vincluded0} & @{$cts_onx0},', " via 'cts_vincluding_vincluded_onx_gblks'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($cts_vincluding0, $cts_vincluded0, $cts_onx0) = cts_vincluding_vincluded_onx_gblks (@{$inter_block_relations0});
unless (defined $cts_onx0) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n --- Subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on the main window ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN >2) { # (Pre 17) #

	my $ofname = 'cts_vincluding_vincluded_onx_gblks0.txt';

	unless ( print_out_cts_vincluding_vincluded_onx_gblks ($maindir_interim, $ofname, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks', Ended on the main window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 17) # ...}"

	### End of "(Pre 17) Prepare @{$cts_vincluding}, @{$cts_vincluded} & @{$cts_onx},', " via 'cts_vincluding_vincluded_onx_gblks'." ###



	### (Pre-MAIN) Prepare the remaining input arguments for 'nonrec_mlt_shifts_gapblocks'. ### (Added on 2019/06/29.)

(print STDOUT "\n<<< Pre-MAIN: Prepare some input arguments for 'nonrec_mlt_shifts_gapblocks'. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

our @SET_INGREDIENTS_PRB_MSA_INDEL = ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, $br2log_mfacs);

my %clm2ln_prb_sbst;  # = ( join (':', @column_in_site_numbers) => $log_probability_under_substitution_model, ...) .
my @set_ingredients_indel_hsts = ({}, [], [], []); # = (\%gpattern2indx, \@set_gpatterns, \@set_node2dparsstat, \@set_br2change) .

if ($TEST_LEVEL_MAIN>1) { # (Pre-MAIN) #

	print STDOUT "[ The Input Arguments for 'nonrec_mlt_shifts_gapblocks' Manually Set Here: ]\n\n";
	print STDOUT '@SET_INGREDIENTS_PRB_MSA_INDEL (= ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, $br2log_mfacs))', "\n";
	print STDOUT ' = (', join (', ', @SET_INGREDIENTS_PRB_MSA_INDEL), ") .\n\n";
	print STDOUT '%clm2ln_prb_sbst = ()', " .\n\n";
	print STDOUT '@set_ingredients_indel_hsts = ({}, [], [], [])', " .\n\n\n";
	
} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre-MAIN) # ...}"

	### End of "(Pre-MAIN) Prepare the remaining input arguments for 'nonrec_mlt_shifts_gapblocks'." ###



	### (Pre-MAIN2) Open a file-handle for the "Synopsis" Output File ($outfile_synopsis). ### (Added in REVISION on 2020/04/20-21.)

# my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";

(print STDOUT "\n<<< Pre-MAIN2: Open a file-handle for the \"Synopsis\" Output File (=\"$outfile_synopsis\").  >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

my $fh_osyn = open_output_file ($outfile_synopsis, 0);
unless (defined $fh_osyn) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the \"Synopsis\" Output File, \"$outfile_synopsis\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}

print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

my $ln_prb_msa_init;

	# perform_mshifts_and_print_inputs_outputs

	### End of "(Pre-MAIN2) Open a file-handle for the "Synopsis" Output File ($outfile_synopsis)." ### 


	##### (MAIN I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}. #####

(print STDOUT "\n<<<<< MAIN (I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}. >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


	# (MAIN I-1) Perform Pure Multiple-"Shift"s. #

if (1) { # (MAIN I-1) #

	my $session = 'MAIN I-1';

   (print STDOUT "\n<<< MAIN I-1: Perform Pure Multiple-\"Shift\"s, via 'nonrec_mlt_shifts_gapblocks' (in 'perform_mshifts_and_print_inputs_outputs'). >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # MODIFIED on 2019/07/19.
   ##(print STDOUT "\n<<< MAIN I-1: Perform Pure Multiple-\"Shift\"s, via 'nonrec_mlt_shifts_gapblocks'. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

		# MOVED from below the Definition of CMSS. (on 2019/10/18) #
	my $subdir = "$maindir_out/I_1.Pure_MltShifts";
	##my $subdir_in = "$subdir/In";
	my $subdir_out0 = "$maindir_out/Initial_Status"; # REVISED on 2019/07/16.
	my $subdir_out = "$subdir/Out";
		# End of "MOVED from below the Definition of CMSS. (on 2019/10/18)" #


	my @list_excluded_bin_hrels0 = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels0 = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# ADDED in REVISION on 2019/10/18. #
	my ($orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm);
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... Define a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$set_clms0_in_wd});

	($orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@bds_gblocks0_in_wd, $le_new_wd, $re_new_wd, @{$inter_block_relations0}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $subdir_out4CMSS = "$subdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($subdir_out4CMSS, $ofile_name, @{$orgs_gb_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$subdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    } else {
	($orgs_gb_coords, $bds_gb_coords, $gb_coords_wd_frm) = ($orgs_gb0_coords, $bds_gb0_coords, $gb0_coords_wd_frm);
    }
		# END of "ADDED in REVISION on 2019/10/18." #


			### MODIFIED in MAJOR REVISION (on 2019/07/18-19). (MAIN I-1) ###

			# (i) & (ii) Print out Input Data & auxiliary information. #
			# & (iii) Perform the Multiple-"Shift"s. #
			# & (iv) Print out the Outputs of 'nonrec_mlt_shifts_gapblocks'. #

=MOVED to above the Definition of CMSS. (on 2019/10/18)
	my $subdir = "$maindir_out/I_1.Pure_MltShifts";
	##my $subdir_in = "$subdir/In";
	my $subdir_out0 = "$maindir_out/Initial_Status"; # REVISED on 2019/07/16.
	my $subdir_out = "$subdir/Out";
=cut # End of "MOVED to above the Definition of CMSS. (on 2019/10/18)"

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $subdir_out0, $subdir_out, @ofnames1, @ofnames2, @ofnames3,

	    ##@{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$orgs_gb_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES, # Modified in REVISION (on 2019/10/18). #

	    @{$set_clms0_in_wd}, @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd,
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels0,
	    @{$inter_block_relations0}, @{$gblock_sizes0}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0},
	    @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);

			# [NOTE added on 2019/07/19]
			# Before the MAJOR REVISION (on 2019/07/18), this block had consisted of 4 subroutines:
			# (i) 'print_in_nr_mshifts_glks' (in "MyANEX_IO.pm") to record the initial local MSA and main configurations of gap-blocks;
			# (ii) 'bch_print_out_other_gblk_attributes' (in "MyANEX_IO.pm") to record auxiliary configurations & attributes of gap-blocks;
			# (iii) 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") to perform the multiple-"shift"s;
			# and (iv) 'print_out_nr_mshifts_gblks' (in MyANEX_IO.pm") to record the results of the multiple-"shift"s.
			# In the  MAJOR REVISION,
			# these 4 subroutines were bundled together into a single subroutine,
			# 'perform_mshifts_and_print_inputs_outputs', and moved to "MyANEX_IO.pm".

			# End of "MODIFIED in MAJOR REVISION (on 2019/07/18-19). (MAIN I-1)" #


			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/20-21.) (MAIN I-1) ##

	$ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$orgs_gb_coords}, @{$bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$orgs_gb_coords}, @{$bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $subdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/20-21.) (MAIN I-1)" ##


} # End of "if (1) { # (MAIN I-1) # ...}"

	# End of "(MAIN I-1) Perform Pure Multiple-"Shift"s." #


my $subdir_tgt_cands = "$maindir_out/Target_Candidates";


	# (MAIN I-2) Perform "Purge" + Multiple-"Shift"s. #
	#  (Constructed from 2019/07/16 till 2019/07/dd; Mostly IMPORTED from 'test_purge_gblock_pair.pl'.)
	# (See SM-5 (ii) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-2: Perform \"Purge\" + Multiple-\"Shift\"s, via 'purge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

my $subdir_MI_2 = "$maindir_out/I_2.Purge_plus_MltShifts";


		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Complementary Gap-Block Pairs to be \"Purge\"d & to be \"Merge\"d , via 'list_nn_purge_and_cmerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_PURGE = $HF_WDTH_SHIFT;
our $THRSH_DIST_CMERGE = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: ($THRSH_DIST_PURGE, $THRSH_DIST_CMERGE) = ', "($THRSH_DIST_PURGE, $THRSH_DIST_CMERGE) .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_purged0, $to_be_cmerged0)
 = list_nn_purge_and_cmerge_candidates (@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_PURGE, $THRSH_DIST_CMERGE);
unless (defined $to_be_cmerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_purge_and_cmerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_purged0 = @{$to_be_purged0};
my $ct_to_be_cmerged0 = @{$to_be_cmerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_purge_and_cmerge_candidates' ENDED on the main window (w/ #{to_be_purged}=$ct_to_be_purged0, #{to_be_cmerged}=$ct_to_be_cmerged0) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if (1) { # (Main I-2 (i) #
##if ($TEST_LEVEL_MAIN >1) { # (Main I-2 (i) #

	my $ofile_name1 = 'to_be_purged.txt';
	my $ofile_name2 = 'to_be_cmerged.txt';
	my $label1 = "to be \"Purge\"d";
	my $label2 = "to be \"Merge\"d (complementary types)";

	print STDOUT "\n ... Record the outputs of 'list_nn_purge_and_cmerge_candidates', by performing 'print_out_purge_or_merge_cands' TWICE ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name1, @{$to_be_purged0}, $label1) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{to_be_purged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_purged0}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name1\") ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_cmerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{to_be_cmered0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_cmerged0}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

##} # End of "if ($TEST_LEVEL_MAIN >1) { # (Main I-2 (i) # ...}"
} # End of "if (1) { # (Main I-2 (i) # ...}"


		# (ii): Perform 'purge_gblock_pair'. 

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'purge_gblock_pair', on each of the $ct_to_be_purged0 gap-block pair to be \"purge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'purge_gblock_pair', on each of the $ct_to_be_purged0 gap-block pairs to be \"purge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

my @null_column = ();
for (1 .. $ct_seqs) { push @null_column, $GAP; }
my $cnct_null_clm = join (':', @null_column);

if ($TEST_LEVEL_MAIN>1) {
	print STDOUT "[ An Additional Input: ]\n\n";
	print STDOUT '$cnct_null_clm = ', $cnct_null_clm, " .\n";
	print STDOUT "\n\n";
} 


for (my $i=0; $i < $ct_to_be_purged0; $i++) { # Main Outer for-loop (over pairs to be "purge"d). #

	my ($gb1, $gb2, $dist) = @{$to_be_purged0->[$i]};

	my $session = 'MAIN I-2';

    (print STDOUT "\n< The $i th pair to be \"purge\"d: (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_2/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";
	

##my $subdir_MI_2 = "$maindir_out/I_2.Purge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_purged.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th pair to be \"purge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

		# (i) Actually "Purge" the Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'purge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($flag_end, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_indices_gbs, $old_indices_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = purge_gblock_pair (
	     $gb1, $gb2, @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

			# MODIFIED in REVISION on 2019/07/21. #
	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT
	     ##@{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'purge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'purge_gblock_pair' ENDED on the $i th pair for $session (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$ct_rmvd_null_clms = $ct_rmvd_null_clms) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

				# REVISED on 2019/08/01. #
	my @new_list_gblks_in_wd = ();
	for (my $ngb=0; $ngb <@{$old_indices_gbs}; $ngb++) {
	    my $ogb = $old_indices_gbs->[$ngb];
	    my @info_gb = @{$list_gblks0_in_wd->[$ogb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblks_in_wd[$ngb] = \@info_gb;
	}
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];
				# END of "REVISED on 2019/08/01." #


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Purge". #

	##my $sbsbdir_info = "$sbsbdir/Info";

	my @ofnames0 = ('info_serial_shifts.txt', 'old_new_indices_gblks.txt'); # ADDED on 2019/07/18.

	unless ( print_out_info_chngs_by_purge ($sbsbdir_info, @ofnames0, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, @{$new_indices_gbs}, @{$old_indices_gbs}) ) { # RIVISED on 2019/07/18.
	##unless ( print_out_info_chngs_by_purge ($sbsbdir_info, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, @{$new_indices_gbs}, @{$old_indices_gbs}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_purge' FAILED on the $i th pair to be \"purge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_purge', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);

		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Purge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

			# MODIFIED in MAJOR REVISION (on 2019/07/19). (MAIN I-2) #

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];
    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th pair had been \"purge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"purge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);

			# [NOTE added on 2019/07/19]
			# Before the MAJOR REVISION (on 2019/07/18), this block had consisted of 4 subroutines:
			# (i) 'print_in_nr_mshifts_glks' (in "MyANEX_IO.pm") to record the initial local MSA and main configurations of gap-blocks;
			# (ii) 'bch_print_out_other_gblk_attributes' (in "MyANEX_IO.pm") to record auxiliary configurations & attributes of gap-blocks;
			# (iii) 'nonrec_mlt_shifts_gapblocks' (in "MyANEX_Main.pm") to perform the multiple-"shift"s;
			# and (iv) 'print_out_nr_mshifts_gblks' (in MyANEX_IO.pm") to record the results of the multiple-"shift"s.
			# In the  MAJOR REVISION,
			# these 4 subroutines were bundled together into a single subroutine,
			# 'perform_mshifts_and_print_inputs_outputs', and moved to "MyANEX_IO.pm".

			# End of "MODIFIED in MAJOR REVISION (on 2019/07/19). (MAIN I-2)" #



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-2) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"purge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"purge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"purge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"purge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-2)" ##


} # End of the Main Outer for-loop (over pairs to be "purge"d). #

	# End of "(MAIN I-2) Perform "Purge" + Multiple-"Shift"s." #


	# (MAIN I-3a) Perform "Merge" (same-type) + Multiple-"Shift"s. # 
	# (See SM-5 (iii-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-3a: Perform \"Merge\" (same-type) + Multiple-\"Shift\"s, via 'smerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED on 2019/07/22. (MAIN I-3a) #

my $subdir_MI_3a = "$maindir_out/I_3a.SMerge_plus_MltShifts";


		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Same-Type Gap-Block Pairs to be \"Merge\"d , via 'list_nn_smerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_SMERGE = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_SMERGE = ', "$THRSH_DIST_SMERGE .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_smerged0)
 = list_nn_smerge_candidates (@bds_gblocks0_in_wd, @{$inter_block_relations0}, $THRSH_DIST_SMERGE);
unless (defined $to_be_smerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_smerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_smerged0 = @{$to_be_smerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_smerge_candidates' ENDED on the main window (w/ #{to_be_smerged}=$ct_to_be_smerged0) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if (1) { # (Main I-3a (i) #

	my $ofile_name2 = 'to_be_smerged.txt';
	my $label2 = "to be \"Merge\"d (same type)";

	print STDOUT "\n ... Record the outputs of 'list_nn_smerge_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_smerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{to_be_smerged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_smerged0}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-3a (i) # ...}"


		# (ii): Perform 'smerge_gblock_pair'. 

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'smerge_gblock_pair', on each of the $ct_to_be_smerged0 same-type gap-block pair to be \"merge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'smerge_gblock_pair', on each of the $ct_to_be_smerged0 same-type gap-block pairs to be \"merge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

for (my $i=0; $i < $ct_to_be_smerged0; $i++) { # Main Outer for-loop (over same-type pairs to be "merge"d). #

	my ($gb1, $gb2, $dist) = @{$to_be_smerged0->[$i]};

	my $session = 'MAIN I-3a';

    (print STDOUT "\n< The $i th same-type pair to be \"merge\"d: (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_3a/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_3a = "$maindir_out/I_3a.SMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_smerged.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th same-type pair to be \"merge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

		# (i) Actually "Merge" the Same-Type Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'smerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, $new_indices_gbs, $old_indices_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = smerge_gblock_pair (
	     $gb1, $gb2, @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT
		); # ADDED "$HF_WDTH_SHIFT" to adapt to the REVISION of 'smerge_gblock_pair' (on 2019/08/11). #
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'smerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'smerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$gb_absbd = $gb_absbd, \$gb_kpt = $gb_kpt) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);

		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

				# REVISED on 2019/08/01. #
	my @new_list_gblks_in_wd = ();
	for (my $ngb=0; $ngb <@{$old_indices_gbs}; $ngb++) {
	    my $ogb = $old_indices_gbs->[$ngb];
	    my @info_gb = @{$list_gblks0_in_wd->[$ogb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblks_in_wd[$ngb] = \@info_gb;
	}
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];
				# END of "REVISED on 2019/08/01." #

	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Merge (same type)". #

	##my $sbsbdir_info = "$sbsbdir/Info";

	my @ofnames0 = ('gbs_absbd_kpt.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt');

	unless ( print_out_info_chngs_by_smerge ($sbsbdir_info, @ofnames0, $gb_absbd, $gb_kpt, $bl_sbj, $incr_coord_sbj, @{$new_indices_gbs}, @{$old_indices_gbs}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_smerge' FAILED on the $i th same-type pair to be \"smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_smerge', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Smerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th same-type pair had been \"merge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-3a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"smerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"smerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-3a)" ##

} # End of the Main Outer for-loop (over same-type pairs to be "merge"d). #


		# END of "ADDED on 2019/07/22. (MAIN I-3a)" #

	# End of "(MAIN I-3a) Perform "Merge" (same-type) + Multiple-"Shift"s." #


	# (MAIN I-3b) Perform "Merge" (complementary-types) + Multiple-"Shift"s. #
	# (See SM-5 (iii-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

		# ADDED on 2019/07/20. (MAIN I-3b) #

my $subdir_MI_3b = "$maindir_out/I_3b.CMerge_plus_MltShifts";

##(print STDOUT "\n<<< MAIN I-3b: Perform \"Merge\" (complementary-types) + Multiple-\"Shift\"s, via '' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);
(print STDOUT "\n<<< MAIN I-3b: Perform \"Merge\" (complementary-types) + Multiple-\"Shift\"s, via 'cmerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # Completed on 2020/05/09.

for (my $i=0; $i < $ct_to_be_cmerged0; $i++) { # Main Outer for-loop (over complementary pairs to be "merge"d). #

	my ($gb1, $gb2, $dist) = @{$to_be_cmerged0->[$i]};

	my $session = 'MAIN I-3b';

    (print STDOUT "\n< The $i th complementary pair to be \"merge\"d: (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_3b/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_3b = "$maindir_out/I_3b.CMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_cmerged.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th complementary pair to be \"merge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

		# (i) Actually "Merge" the Complementary Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'cmerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, $ct_rmvd_null_clms, $new_indices_gbs, $old_indices_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = cmerge_gblock_pair (
	     $gb1, $gb2, @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

			# MODIFIED in REVISION on 2019/07/21. #
	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT
	     ##@{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'cmerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'cmerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$gb_rmvd = $gb_rmvd, \$gb_kpt = $gb_kpt, \$ct_rmvd_null_clms = $ct_rmvd_null_clms) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

				# REVISED on 2019/08/01. #
	my @new_list_gblks_in_wd = ();
	for (my $ngb=0; $ngb <@{$old_indices_gbs}; $ngb++) {
	    my $ogb = $old_indices_gbs->[$ngb];
	    my @info_gb = @{$list_gblks0_in_wd->[$ogb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblks_in_wd[$ngb] = \@info_gb;
	}
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];
				# END of "REVISED on 2019/08/01." #


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Merge (complementary types)". #

	##my $sbsbdir_info = "$sbsbdir/Info";

	my @ofnames0 = ('gbs_rmvd_kpt.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt'); # ADDED on 2019/07/18; REVISED on 2019/07/20. #

	unless ( print_out_info_chngs_by_cmerge ($sbsbdir_info, @ofnames0, $gb_rmvd, $gb_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, @{$new_indices_gbs}, @{$old_indices_gbs}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_cmerge' FAILED on the $i th complementary pair to be \"cmerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_cmerge', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);

		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Cmerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th complementary pair had been \"merge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"cmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-3b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"cmerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"cmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"cmerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"cmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-3b)" ##


} # End of the Main Outer for-loop (over complementary pairs to be "merge"d). #

		# ADDED on 2019/07/20. (MAIN I-3b) #

	# End of "(MAIN I-3b) Perform "Merge" (complementary-types) + Multiple-"Shift"s." #



	# (MAIN I-6a) Perform "(incomplete) Vertical-Merge" (sibling blocks) + Multiple-"Shift"s. #
	# (See SM-5 (vi-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-6a: Perform \"(incomplete) Vertical-Merge\" (sibling blocks) + Multiple-\"Shift\"s, via '' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

if ($TEST_LEVEL_MAIN>1) { # (MAIN I-6a) #

	print STDOUT "\n";
	print STDOUT " ... Actually, these moves are already incorporated in 'nonrec_mlt_shifts_gapblocks' for Pure Multiple-\"Shift\"s (in MAIN I-1)!\n";
	print STDOUT " ... For the moment, I will do nothing here ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (MAIN I-6a) # ...}"

	# End of "(MAIN I-6a) Perform "(incomplete) Vertical-Merge" (sibling blocks) + Multiple-"Shift"s." #



	# (MAIN I-6b) Perform "(incomplete) Vertical-Merge" (complementary-sibling blocks) + Multiple-"Shift"s. #
	# (See SM-5 (vi-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

##(print STDOUT "\n<<< MAIN I-6b: Perform \"(incomplete) Vertical-Merge\" (complementary-sibling blocks) + Multiple-\"Shift\"s, via '' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);
(print STDOUT "\n<<< MAIN I-6b: Perform \"(incomplete) Vertical-Merge\" (complementary-sibling blocks) + Multiple-\"Shift\"s, via 'cvmerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # Completed on 2020/05/09.

		# ADDED on 2019/07/29. (MAIN I-6b) #

my $subdir_MI_6b = "$maindir_out/I_6b.CVMerge_plus_MltShifts";

		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Complementary-Sibling Gap-Block Pairs to be \"(incomplete) Vertical-Merge\"d , via 'list_nn_cvmerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_CVMERGE = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_CVMERGE = ', "$THRSH_DIST_CVMERGE .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_cvmerged0)
 = list_nn_cvmerge_candidates (@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_CVMERGE);
unless (defined $to_be_cvmerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_cvmerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_cvmerged0 = @{$to_be_cvmerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_cvmerge_candidates' ENDED (w/ #{to_be_cvmerged}=$ct_to_be_cvmerged0) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if (1) { # (Main I-6b (i) #

	my $ofile_name2 = 'to_be_cvmerged.txt';
	my $label2 = "to be \"(incomplete) Vertical-Merge\"d (complementary-sibling)";

	print STDOUT "\n ... Record the outputs of 'list_nn_cvmerge_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_cvmerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{to_be_cvmerged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_cvmerged0}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-6b (i) # ...}"


		# (ii): Perform 'cvmerge_gblock_pair'. 

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'cvmerge_gblock_pair', on each of the $ct_to_be_cvmerged0 complementary-sibling gap-block pair to be \"(incomplete) vertical-merge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'cvmerge_gblock_pair', on each of the $ct_to_be_cvmerged0 complementary-sibling gap-block pairs to be \"(incomplete) vertical-merge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

for (my $i=0; $i < $ct_to_be_cvmerged0; $i++) { # Main Outer for-loop (over complementary-sibling pairs to be "(incomplete) vertical-merge"d). #

	my ($gb1, $gb2, $dist) = @{$to_be_cvmerged0->[$i]};

	my $session = 'MAIN I-6b';

    (print STDOUT "\n< The $i th complementary-sibling pair to be \"(incomplete) vertical-merge\"d: (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_6b/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_6b = "$maindir_out/I_6b.CVMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_cvmerged.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th complementary-sibling pair to be \"(incomplete) vertical-merge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

		# (i) Actually "(incomplete) Vertical-Merge" the Complementary-Sibling Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'cvmerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, $new_indices_old_gbs, $old_indices_new_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks # ADDED in DEBUGGING on 2020/05/13. #
		)
	  = cvmerge_gblock_pair (
	     $gb1, $gb2, @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0},  $HF_WDTH_SHIFT,
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'cvmerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'cvmerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$if_complete = $if_complete, \$gb_rplcd = $gb_rplcd, \$gb_rmvd_kpt = $gb_rmvd_kpt, \$ct_rmvd_null_clms = $ct_rmvd_null_clms) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "(incomplete) Vertical-Merge (complementary-siblings)". #

	##my $sbsbdir_info = "$sbsbdir/Info";

	my @ofnames0 = ('gbs_rplcd_rmvd_kpt.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt');

	##my ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, $new_indices_old_gbs, $old_indices_new_gbs,

	unless ( print_out_info_chngs_by_cvmerge ($sbsbdir_info, @ofnames0, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_cvmerge' FAILED on the $i th complementary-sibling pair to be \"cvmerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_cvmerge', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "CVmerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;

		# A BUG Fixed on 2020/05/13. #
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_new_gbs}];
	my @new_list_gblks = @{$new_list_gblocks};
		# End of "A BUG Fixed on 2020/05/13." #

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";

	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_new_gbs}]', " .\n";
	print STDOUT 'Manually Set:  @new_list_gblks = @{$new_list_gblocks}', " .\n"; # Modified along with the DEBUGGING on 2020/05/13. #
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th complementary-sibling pair had been \"(incomplete) vertical-merge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"cvmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-6b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"cvmerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"cvmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"cvmerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"cvmerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-6b)" ##

} # End of the Main Outer for-loop (over complementary-sibling pairs to be "(incomplete) vertical-merge"d). #

		# END of "ADDED on 2019/07/29. (MAIN I-6b)" #

	# End of "(MAIN I-6b) Perform "(incomplete) Vertical-Merge" (complementary-sibling blocks) + Multiple-"Shift"s." #



	# (MAIN-I-34a) Perform "(Horizontal) Merge + Split" (same-type) + Multiple-"Shift"s. #
	# (See SM-6 (iii-vi-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-34a: Perform \"(Horizontal) Merge + Split\" (same-type) + Multiple-\"Shift\"s, via 'smerge_split_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

			## ADDED on 2019/08/01. (MAIN I-34a). #

my $subdir_MI_34a = "$maindir_out/I_34a.SMerge_plus_Split_plus_MltShifts";

(print STDOUT "\n<< (i)  Search for Same-Type Gap-Block Pairs to be \"Merge + Split\" , via 'list_smerge_split_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $UPPER_BOUND_NRSIZE = ($XLEVEL < 3) ? $MIN_WD_SIZE_PCAND : -1 ;
(print STDOUT 'Manually Set: $UPPER_BOUND_NRSIZE = ', "$UPPER_BOUND_NRSIZE .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($to_be_smerge_split0) = list_smerge_split_candidates (@{$to_be_smerged0}, @{$gblock_sizes0}, $UPPER_BOUND_NRSIZE);
unless (defined $to_be_smerge_split0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_smerge_split_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_smerge_split0 = @{$to_be_smerge_split0};

(print STDOUT "\n --- The Subroutine 'list_smerge_split_candidates' ENDED (w/ #{to_be_smerge_split}=$ct_to_be_smerge_split0) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if (1) { # (Main I-34a (i) #

	my $ofile_name2 = 'to_be_smerge_split.txt';
	my $label2 = "to be \"Merge + Split (same type)\"";
	my $label_sizes = 'sizes,of,new,right,gap-blocks';

	print STDOUT "\n ... Record the outputs of 'list_smerge_split_candidates', by performing 'print_out_merge_split_cands' ...\n\n";

	unless ( print_out_merge_split_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_smerge_split0}, $label2, $label_sizes) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_merge_split_cands' FAILED on ", '@{$to_be_smerge_split0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_merge_split_cands', Ended on ", '@{$to_be_smerge_split0}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-34a (i) # ...}"


		# (ii): Perform 'smerge_split_gblock_pair'. 

(print STDOUT "\n<< (ii) Perform the Subroutine, 'smerge_split_gblock_pair', on each of the $ct_to_be_smerge_split0 same-type gap-block pairs to be \"merge + split\". >>\n\n") if ($TEST_LEVEL_MAIN >0);


for (my $i=0; $i < $ct_to_be_smerge_split0; $i++) { # Main Outer for-loop (over same-type pairs to be "merge + split"). #

	my ($gb1, $gb2, $dist, $new_sizes_right) = @{$to_be_smerge_split0->[$i]};
	my $cnct_nrsizes = join (',', @{$new_sizes_right});

	my $session = 'MAIN I-34a';

    (print STDOUT "\n< The $i th same-type pair to be \"merge + split\": (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist, ", '@{$new_sizes_right}', "=($cnct_nrsizes)) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_34a/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";

##my $subdir_MI_34a = "$maindir_out/I_34a.SMerge_plus_Split_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_smerge_split.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th same-type pair to be \"merge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (i) Actually "Merge + Split" the Same-Type Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'smerge_split_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

    foreach my $new_size_right (@{$new_sizes_right}) { # Middle foreach-loop (over $new_size_right). #

	(print STDOUT "\n{{ \$new_size_right = $new_size_right }}\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbdir_out = "$sbsbdir/Out/nrsize$new_size_right";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($flag_end, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $new_gblock_sizes)
	  = smerge_split_gblock_pair (
	     $gb1, $gb2, $new_size_right, @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0},  $HF_WDTH_SHIFT
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'smerge_split_gblock_pair' FAILED on the $i th gap-block pair for $session (w/ \$new_size_right = $new_size_right) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'smerge_split_gblock_pair' ENDED on the $i th pair (w/ \$new_size_right = $new_size_right) for $session (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$new_bl_sbj = $new_bl_sbj, \$incr_coord_new_sbj = $incr_coord_new_sbj) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$inter_block_relations0}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = ();
	for (my $gb=0; $gb <$ct_gblks0_in_wd; $gb++) { # Inner for-loop (over gap-blocks). #
	    my @info_gb = @{$list_gblks0_in_wd->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$gb]};
	    $new_list_gblks_in_wd[$gb] = \@info_gb;
	} # End of the Inner for-loop (over gap-blocks). #
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Merge + Split (same type)". #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbdir_info = "$sbsbdir_info/nrsize$new_size_right";

	my @ofnames0 = ('info_serial_shifts.txt');

	my $ct_rmvd_null_clms = 0; # Added in revision on 2019/08/11.
	unless ( print_out_info_chngs_by_merge_split ($sbsbsbdir_info, @ofnames0, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, $new_bl_sbj, $incr_coord_new_sbj) ) {
	##unless ( print_out_info_chngs_by_merge_split ($sbsbsbdir_info, @ofnames0, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_merge_split' FAILED on the $i th same-type pair to be \"smerge + split\" for $session (w/ \$new_size_right = $new_size_right) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_merge_split', Ended on the $i th pair for $session (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Smerge + Split". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

	my ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $rows_affd_by_new_blks)
		= ($inter_block_relations0, $cts_vincluding0, $cts_vincluded0, $cts_onx0, \@rows_affd_by_blks0_in_wd);

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT 'Manually Set: ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $rows_affd_by_new_blks)', "\n";       
	print STDOUT '            (= ($inter_block_relations0, $cts_vincluding0, $cts_vincluded0, $cts_onx0, $rows_affd_by_blks0_in_wd))', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th same-type pair had been \"merge + split\" (w/ \$new_size_right = $new_size_right) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"smerge + split\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-34a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"smerge + split\" (w/ \$new_size_right = $new_size_right): $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"smerge + split\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"smerge + split\" (w/ \$new_size_right = $new_size_right): $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"smerge + split\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-34a)" ##


    } # End of the Middle foreach-loop (over $new_size_right). #

} # End of the Main Outer for-loop (over same-type pairs to be "merge + split"). #


			## END of "ADDED on 2019/08/01. (MAIN I-34a)." #

	# End of "(MAIN-I-34a) Perform "(Horizontal) Merge + Split" (same-type) + Multiple-"Shift"s." #



	# (MAIN-I-34b) Perform "(Horizontal) Merge (or purge) + Split (or ex-nihilo)" (complementary-types) + Multiple-"Shift"s. #
	# (See SM-6 (iii-vi-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

##(print STDOUT "\n<<< MAIN I-34b: Perform \"(Horizontal) Merge (or purge) + Split (or ex-nihilo)\" (complementary-types) + Multiple-\"Shift\"s, via '' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);
(print STDOUT "\n<<< MAIN I-34b: Perform \"(Horizontal) Merge (or purge) + Split (or ex-nihilo)\" (complementary-types) + Multiple-\"Shift\"s, via ('icmerge_gblock_pair' & 'cmerge_split_grow_gblock_pair') & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # Completed on 2020/05/09.

			## ADDED on 2019/08/10. (MAIN I-34b) (1) ##
	
my $subdir_MI_34b = "$maindir_out/I_34b.CMerge_plus_Split_plus_MltShifts";

(print STDOUT "\n<< (i)  Search for Complementary-Type Gap-Block Pairs to be \"Merge + Split\" , via 'list_nn_cmerge_split_pre_candidates' & 'list_cmerge_split_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_CMERGE_SPLIT = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_CMERGE_SPLIT = ', "$THRSH_DIST_CMERGE_SPLIT .\n\n") if ($TEST_LEVEL_MAIN >1);


my ($pre_to_be_cmerge_split0)
 = list_nn_cmerge_split_pre_candidates (@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_CMERGE_SPLIT);
unless (defined $pre_to_be_cmerge_split0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_cmerge_split_pre_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_pre_to_be_cmerge_split0 = @{$pre_to_be_cmerge_split0};

(print STDOUT "\n --- The Subroutine 'list_nn_purge_and_cmerge_candidates' ENDED on the main window (w/ #{pre_to_be_cmerge_split0}=$ct_pre_to_be_cmerge_split0) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if (1) { # (Main I-34b (i) #

	my $ofile_name2 = 'to_be_cmerge_split_PRE.txt';
	my $label2 = "to be \"Merge + Split\" (complementary types) (PREliminary)";

	print STDOUT "\n ... Record the outputs of 'list_nn_cmerge_split_pre_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$pre_to_be_cmerge_split0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{$pre_to_be_cmerge_split0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$pre_to_be_cmerge_split0}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-34b (i) # ...}"


our $LBOUND_DSIZE = ($XLEVEL < 3) ? - ($MIN_WD_SIZE_PCAND -1) : - $HF_WDTH_SHIFT ;
our $UBOUND_DSIZE = ($XLEVEL < 3) ?  $MIN_WD_SIZE_PCAND -1 :  $HF_WDTH_SHIFT ;

(print STDOUT 'Manually Set: ($LBOUND_DSIZE, $UBOUND_DSIZE) = ', "($LBOUND_DSIZE, $UBOUND_DSIZE) .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($to_be_cmerge_split0)
  = list_cmerge_split_candidates (@{$pre_to_be_cmerge_split0}, @{$gblock_sizes0}, @bds_gblocks0_in_wd, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$inter_block_relations0}, $le_wd, $re_wd, $LBOUND_DSIZE, $UBOUND_DSIZE);
unless (defined $to_be_cmerge_split0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_cmerge_split_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_cmerge_split0 = @{$to_be_cmerge_split0};

(print STDOUT "\n --- The Subroutine 'list_cmerge_split_candidates' ENDED (w/ #{to_be_cmerge_split}=$ct_to_be_cmerge_split0) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if (1) { # (Main I-34b (ii) #

	my $ofile_name2 = 'to_be_cmerge_split.txt';
	my $label2 = "to be \"Merge + Split (complementary types)\"";
	my $label_sizes = 'gap-block,size,increments';

	print STDOUT "\n ... Record the outputs of 'list_cmerge_split_candidates', by performing 'print_out_merge_split_cands' ...\n\n";

	unless ( print_out_merge_split_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_cmerge_split0}, $label2, $label_sizes) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_merge_split_cands' FAILED on ", '@{to_be_cmerge_split0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_merge_split_cands', Ended on ", '@{$to_be_cmerge_split0}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-34b (ii) # ...}"


		# (ii): Perform 'icmerge_gblock_pair' & 'cmerge_split_grow_gblock_pair'. . 

(print STDOUT "\n<< (ii) Perform the Subroutines, 'icmerge_gblock_pair' & 'cmerge_split_grow_gblock_pair', on each of the $ct_to_be_cmerge_split0 complementary-type gap-block pairs to be \"merge + split\". >>\n\n") if ($TEST_LEVEL_MAIN >0);

			## END of "ADDED on 2019/08/10. (MAIN I-34b) (1)" ##


			## ADDED on 2019/08/11. (MAIN I-34b) (2) ##

for (my $i=0; $i < $ct_to_be_cmerge_split0; $i++) { # Main Outer for-loop (over complementary-type pairs to be "merge + split"). #

	my ($gb1, $gb2, $dist, $incr_gb_sizes) = @{$to_be_cmerge_split0->[$i]};
	my $cnct_incr_gb_sizes = join (',', @{$incr_gb_sizes});

	my $session = 'MAIN I-34b';

    (print STDOUT "\n< The $i th complementary-type pair to be \"merge + split\": (indx_gb1, indx_gb2) = ($gb1, $gb2) ; (distance=$dist, ", '@{$incr_gb_sizes}', "=($cnct_incr_gb_sizes)) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_34b/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";

##my $subdir_MI_34b = "$maindir_out/I_34b.CMerge_plus_Split_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_cmerge_split.txt';

	my @subjects = ($gb1, $gb2);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th same-type pair to be \"merge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (i) Actually "Merge + Split" the Complementary-Type Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutines, 'icmerge_gblock_pair' & 'cmerge_split_grow_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);


    foreach my $incr_gb_size (@{$incr_gb_sizes}) { # Middle foreach-loop (over $incr_gb_size). #

	(print STDOUT "\n{{ \$incr_gb_size = $incr_gb_size }}\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbdir_out = ($incr_gb_size >0) ? "$sbsbdir/Out/igsize_pos$incr_gb_size" : sprintf ("$sbsbdir/Out/igsize_neg%d", -$incr_gb_size) ;
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

	my @curr_gb_coords = @{$orgs_gb0_coords};

	my ($new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $new_gblock_sizes); # The shared (main) outputs. #

			# MOVED from (ii) (on 2019/08/19). #
	my ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $rows_affd_by_new_blks)
		= ($inter_block_relations0, $cts_vincluding0, $cts_vincluded0, $cts_onx0, \@rows_affd_by_blks0_in_wd);
			# END of "MOVED from (ii) (on 2019/08/19)." #

	my ($new_indices_old_gbs, $old_indices_new_gbs, $new_list_gblocks); # Outputs specifically for 'cmerge_split_grow_gblock_pair'. (ADDED in REVISION on 2019/08/19.) #

	my $flag_end; # The flag shared by both subroutines.
	my ($bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2); # The auxiliary outputs of 'icmerge_gblock_pair'.
	my ($gb_left, $gb_right, $records_sessions, $gb2incr_bl_coord_undo); # The auxiliary outputs of 'cmerge_split_grow_gblock_pair'.

     if ($incr_gb_size < 0) { # (1) #

	my $ct_to_be_merged = - $incr_gb_size;

	($flag_end, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $new_gblock_sizes)
	  = icmerge_gblock_pair (
	     $gb1, $gb2, $ct_to_be_merged,
	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,
	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0},  $HF_WDTH_SHIFT
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'icmerge_gblock_pair' FAILED on the $i th gap-block pair for $session (w/ \$ct_to_be_merged = $ct_to_be_merged) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The Subroutine 'icmerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$ct_to_be_merged = $ct_to_be_merged))  (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj1 = $incr_coord_sbj1, \$ct_rmvd_null_clms = $ct_rmvd_null_clms, \$incr_coord_sbj2 = $incr_coord_sbj2) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);

      } else { # if ($incr_gb_size > 0)

	 ($flag_end, $gb_left, $gb_right, $records_sessions, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
			# MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (1) #
	    ## [ $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $new_gblock_sizes ]... OBSOLETE as of 2019/08/19. ##
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_indices_old_gbs, $old_indices_new_gbs, $new_list_gblocks,
			# END of "MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (1)" #
		)
	  = cmerge_split_grow_gblock_pair (
	     $gb1, $gb2, $incr_gb_size, $GAP,
	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,
	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0},  $HF_WDTH_SHIFT,
			# MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (2) #
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq} 
			# END of "MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (2)" #
		);
	unless ($flag_end) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'cmerge_split_grow_gblock_pair' FAILED FAILED on the $i th gap-block pair for $session (w/ \$incr_gb_size = $incr_gb_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_sessions = @{$records_sessions}; 
   (print STDOUT "\n --- The Subroutine 'cmerge_split_grow_gblock_pair' ENDED on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size)  (w/ \$flag_end = $flag_end, \$incr_gb_size = $incr_gb_size)  (w/ \$gb_left = $gb_left, \$gb_right = $gb_right, #{sessions}=$ct_sessions) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);

	unless ($flag_end > 0) {
	     print STDOUT "[WARNING '$plscript_name'] The subroutine 'cmerge_split_grow_gblock_pair' Ended WITH \$flag_end = $flag_end  on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size) !!\n\n";
	}

      } # End of "if ($incr_gb_size < 0) { # (1) # ...} else {...}"


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = ();

   if ($incr_gb_size < 0) { # (1.5) # ADDED the condition on 2019/08/19. #
	for (my $gb=0; $gb <$ct_gblks0_in_wd; $gb++) { # Inner for-loop (over gap-blocks). #
	    my @info_gb = @{$list_gblks0_in_wd->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$gb]};
	    $new_list_gblks_in_wd[$gb] = \@info_gb;
	} # End of the Inner for-loop (over gap-blocks). #
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

   } else { # ADDED the block on 2019/08/19. #

	@new_list_gblks_in_wd = @{$new_list_gblocks};

   } # END of "if ($incr_gb_size < 0) { # (1.5) # ...} else {...}".


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Merge + Split (same type)". #

     if ($incr_gb_size < 0) { # (2) #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbdir_info = sprintf ("$sbsbdir_info/igsize_neg%d", -$incr_gb_size) ;

	my @ofnames0 = ('info_serial_shifts.txt');

	# ($bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2); # The auxiliary outputs of 'icmerge_gblock_pair'.

	unless ( print_out_info_chngs_by_merge_split ($sbsbsbdir_info, @ofnames0, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $bl_sbj, $incr_coord_sbj2) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_merge_split' FAILED on the $i th complementary-type pair to be \"cmerge + split\" for $session (w/ \$incr_gb_size = $incr_gb_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_merge_split', Ended on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_merge_split', Ended on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size) (, writing to \"$sbsbsbdir_info/", $ofnames0[0], "\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

      } else { # if ($incr_gb_size > 0)

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbdir_info = sprintf ("$sbsbdir_info/igsize_pos%d", $incr_gb_size) ;

	my @ofnames0 = ('basic_info.txt', 'rec_cmerge_split.txt', 'rec_undo_hovlps.txt', 'old_new_indices_gblks.txt'); # MODIFIED in revision of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (1) #
	##my @ofnames0 = ('basic_info.txt', 'rec_cmerge_split.txt', 'rec_undo_hovlps.txt');

	unless ( print_out_info_chngs_by_cmerge_split_grow ($sbsbsbdir_info, @ofnames0, $flag_end, $gb_left, $gb_right, @{$records_sessions}, %{$gb2incr_bl_coord_undo}, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}) ) { # MODIFIED in revision of 'cmerge_split_grow_gblock_pair' (on 2019/08/19). (2) #
	## unless ( print_out_info_chngs_by_cmerge_split_grow ($sbsbsbdir_info, @ofnames0, $flag_end, $gb_left, $gb_right, @{$records_sessions}, %{$gb2incr_bl_coord_undo}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_cmerge_split_grow' FAILED on the $i th complementary-type pair to be \"cmerge + split\" for $session (w/ \$incr_gb_size = $incr_gb_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_cmerge_split_grow', Ended on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_cmerge_split_grow', Ended on the $i th pair for $session (w/ \$incr_gb_size = $incr_gb_size) (, writing to several files under \"$sbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

	# ($gb_left, $gb_right, $records_sessions, $gb2incr_bl_coord_undo); # The auxiliary outputs of 'cmerge_split_grow_gblock_pair'.

      } # End of "if ($incr_gb_size < 0) { # (2) # ...} else {...}"


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Cmerge + CSplit". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

			# MOVED to above the main subroutines (on 2019/08/19). #
	##my ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $rows_affd_by_new_blks)
	##	= ($inter_block_relations0, $cts_vincluding0, $cts_vincluded0, $cts_onx0, \@rows_affd_by_blks0_in_wd);

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT 'Manually Set: ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $rows_affd_by_new_blks)', "\n";       
	print STDOUT '            (= ($inter_block_relations0, $cts_vincluding0, $cts_vincluded0, $cts_onx0, $rows_affd_by_blks0_in_wd))', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th complementary-type pair had been \"merge + split\" (w/ \$incr_gb_size = $incr_gb_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"cmerge + split\" (w/ \$incr_gb_size = $incr_gb_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-34b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"cmerge + split\" (w/ \$incr_gb_size = $incr_gb_size): $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"cmerge + split\" (w/ \$incr_gb_size = $incr_gb_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"cmerge + split\" (w/ \$incr_gb_size = $incr_gb_size) : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"cmerge + split\" (w/ \$incr_gb_size = $incr_gb_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-34b)" ##

    } # End of the Middle foreach-loop (over $incr_gb_size). #

} # End of the Main Outer for-loop (over complementary-type pairs to be "merge + split"). #

			## END of "ADDED on 2019/08/11. (MAIN I-34b) (2)" ##

	# End of "(MAIN-I-34b) Perform "(Horizontal) Merge (or purge) + Split (or ex-nihilo)" (complementary-types) + Multiple-"Shift"s." #



	# (MAIN-I-37a) Perform "Vertical-Split (into sibling blocks) + Horizontal Merge (same-type)" + Multiple-"Shift"s. #
	# (See SM-6 (iii-vii-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-37a: Perform \" Vertical-Split (into sibling blocks) + Horizontal Merge (same-type)\" + Multiple-\"Shift\"s, via 'vsplit_plus_smerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # MODIFIED on 2019/09/07. #
##(print STDOUT "\n<<< MAIN I-37a: Perform \"Horizontal Merge + (incomplete) Vertical-Split\" (into sibling blocks) + Multiple-\"Shift\"s, via '' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

			## ADDED on 2019/09/07. (MAIN I-37a) (1) ##

my $subdir_MI_37a = "$maindir_out/I_37a.VSplit_plus_SMerge_plus_MltShifts";


		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Effective Parent-Child Gap-Block Pairs to be \"VSplit + SMerge\"d , via 'list_nn_vsplit_plus_smerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_VSSM = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_VSSM = ', "$THRSH_DIST_VSSM .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_vsplit_plus_smerged0)
 = list_nn_vsplit_plus_smerge_candidates (
	@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_VSSM);
unless (defined $to_be_vsplit_plus_smerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_vsplit_plus_smerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_vssm0 = @{$to_be_vsplit_plus_smerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_vsplit_plus_smerge_candidates' ENDED on the main window (w/ #{to_be_vsplit+smerged}=$ct_to_be_vssm0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


if (1) { # (Main I-37a (i) #

	my $ofile_name2 = 'to_be_vsplit_plus_smerged.txt';
	my $label2 = "to be \"Vertically Split\" + \"Merge\"d (same type)";

	print STDOUT "\n ... Record the outputs of 'list_nn_vsplit_plus_smerge_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_vsplit_plus_smerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{$to_be_vsplit_plus_smerged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_vsplit_plus_smerged0}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-37a (i) # ...}"


		# (ii): Perform 'vsplit_plus_smerge_gblock_pair'.

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'vsplit_plus_smerge_gblock_pair', on each of the $ct_to_be_vssm0 effective parent-child gap-block pair to be \"vsplit + smerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'vsplit_plus_smerge_gblock_pair', on each of the $ct_to_be_vssm0 effective parent-child gap-block pairs to be \"vsplit + smerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

			## END of "ADDED on 2019/09/07. (MAIN I-37a) (1)" ##

			## ADDED on 2019/09/08. (MAIN I-37a) (2) ##

for (my $i=0; $i < $ct_to_be_vssm0; $i++) { # Main Outer for-loop (over effective parent-child pairs to be "vsplit + smerge"d). #

	my $info_cand = $to_be_vsplit_plus_smerged0->[$i];
	my ($gb_sbj_pa, $gb_sbj_ch, $dist) = @{$info_cand};

	my $session = 'MAIN I-37a';

    (print STDOUT "\n< The $i th effective parent-child pair to be \"vsplit + smerge\"d: (indx_gb1, indx_gb2) = ($gb_sbj_pa, $gb_sbj_ch) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_37a/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_37a = "$maindir_out/I_37a.VSplit_plus_SMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_vsplit_plus_smerged.txt';

	my @subjects = ($gb_sbj_pa, $gb_sbj_ch);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th effective parent-child pair to be \"vsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (i) Actually "VSplit + SMerge" the Effective Parent-Child Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'vsplit_plus_smerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

			# Modified for 'vsplit_plus_smerge_gblock_pair'. #
	## [ my ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, $new_indices_gbs, $old_indices_gbs, ] ... for the predecessor, 'smerge_gblock_pair'. ##
	my ($flag_vssm, $gb_ch_left_NEW, $gb_kpt_NEW, 
	    $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	    $gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	    $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
			# End of "Modified for 'vsplit_plus_smerge_gblock_pair'." #

	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks, $new_list_excluded_bin_hrels # Added for 'vsplit_plus_smerge_gblock_pair'. #
	    )
	  = vsplit_plus_smerge_gblock_pair ( $gb_sbj_pa, $gb_sbj_ch,
	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT,
			# Added for 'vsplit_plus_smerge_gblock_pair'. #
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw},  @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq},
	    @set_exnodes, %{$nodeid2msa_row}
			# End of "Added for 'vsplit_plus_smerge_gblock_pair'." #
		);
	unless ($flag_vssm) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'vsplit_plus_smerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	unless ($flag_vssm >0) {
	    print STDOUT "[CAUTION '$plscript_name'] The MAIN Subroutine 'vsplit_plus_smerge_gblock_pair' FAILED because some overlaps canNOT be UNDONE, on the $i th gap-block pair for $session ! ... I will SKIP this pair ... \n";
	    next;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'vsplit_plus_smerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$gb_ch_left_NEW = $gb_ch_left_NEW, \$gb_kpt_NEW = $gb_kpt_NEW) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = @{$new_list_gblocks};

	my @list_excluded_bin_hrels = @{$new_list_excluded_bin_hrels};

	my %gb2incr_coord_1way_dsplc_bf_vspl = ('N.A' => 1); # A dummy hash. (ADDED in revision on 2019/08/30.)
  (print STDOUT 'Manually Set: %gb2incr_coord_1way_dsplc_bf_vspl = ', "('N.A' => 1) .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Vertical Split + Merge (same type)". #

	##my $sbsbdir_info = "$sbsbdir/Info";


	##my @ofnames0 = ('basic_info.txt', 'info_1way_displacements.txt', 'info_serial_shifts_af_vspl.txt', 'rec_undo_hovlps_af_vspl.txt', 'info_smerge_af_vspl.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_smg.txt'); # MODIFIED in revision (on 2019/08/30). #
	my @ofnames0 = ('basic_info.txt', 'info_1way_displacements_bf_vspl.txt', 'info_serial_shifts_af_vspl.txt', 'rec_undo_hovlps_af_vspl.txt', 'info_smerge_af_vspl.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_smg.txt'); # => Further MODIFIED (on 2019/10/17). #

	# $flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	unless ( print_out_info_chngs_by_vsplit_plus_smerge ($sbsbdir_info, @ofnames0, 
		$flag_vssm, $gb_sbj_pa, $gb_sbj_ch, $gb_ch_left_NEW, $gb_kpt_NEW, 
		%gb2incr_coord_1way_dsplc_bf_vspl, 
		$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, %{$gb2incr_bl_coord_undo_ovlps_af_vspl},
		$gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
		@{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$intm_indices_old_gbs}, @{$new_indices_intm_gbs},
		%{$gb2incr_coord_undo_af_smg}
			) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_vsplit_plus_smerge' FAILED on the $i th effective parent-child pair to be \"vsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge', Ended on the $i th pair for $session  ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge', Ended on the $i th pair for $session  (, writing to several files under \"$sbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.



		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Vsplit + Smerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	##print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th effective parent-child pair had been \"vsplit + smerge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-37a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"vsplit + smerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"vsplit + smerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-37a)" ##


} # End of the Main Outer for-loop (over effective parent-child pairs to be "vsplit + smerge"d). #

			## END of "ADDED on 2019/09/08. (MAIN I-37a) (2)" ##

	# End of "(MAIN-I-37a) Perform "Vertical-Split (into sibling blocks) + Horizontal Merge (same-type)" + Multiple-"Shift"s." #



	# (MAIN-I-37b) Perform "Vertical-Split (into complementary-sibling blocks) + Horizontal Merge (same-type)" + Multiple-"Shift"s. #
	# (See SM-6 (iii-vii-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

(print STDOUT "\n<<< MAIN I-37b: Perform \" Vertical-Split (into complementary-sibling blocks) + Horizontal Merge (same-type)\" + Multiple-\"Shift\"s, via 'cvsplit_plus_smerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # MODIFIED on 2019/09/12. # => CORRECTED on 2020/04/20.
##(print STDOUT "\n<<< MAIN I-37a: Perform \" Vertical-Split (into complementary-sibling blocks) + Horizontal Merge (same-type)\" + Multiple-\"Shift\"s, via 'cvsplit_plus_smerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # MODIFIED on 2019/09/12. #
####(print STDOUT "\n<<< MAIN I-37b: Perform \"Horizontal Merge + (incomplete) Vertical-Split\" (into complementary-sibling blocks) + Multiple-\"Shift\"s, via 'cvsplit_plus_smerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

			## ADDED on 2019/09/12. (MAIN I-37b) ##

my $subdir_MI_37b = "$maindir_out/I_37b.CVSplit_plus_SMerge_plus_MltShifts";


		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Effective Child-Parent Gap-Block Pairs to be \"CVSplit + SMerge\"d , via 'list_nn_cvsplit_plus_smerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_CVSSM = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_CVSSM = ', "$THRSH_DIST_CVSSM .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_cvsplit_plus_smerged0)
 = list_nn_cvsplit_plus_smerge_candidates (
	@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_CVSSM);
unless (defined $to_be_cvsplit_plus_smerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_cvsplit_plus_smerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_cvssm0 = @{$to_be_cvsplit_plus_smerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_vsplit_plus_smerge_candidates' ENDED on the main window (w/ #{to_be_cvsplit+smerged}=$ct_to_be_cvssm0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


if (1) { # (Main I-37b (i) #

	my $ofile_name2 = 'to_be_cvsplit_plus_smerged.txt';
	my $label2 = "to be \"Complementarily-Vertically Split\" + \"Merge\"d (same type)";

	print STDOUT "\n ... Record the outputs of 'list_nn_cvsplit_plus_smerge_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_cvsplit_plus_smerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{$to_be_cvsplit_plus_smerged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_cvsplit_plus_smerged0}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-37b (i) # ...}"


		# (ii): Perform 'cvsplit_plus_smerge_gblock_pair'.

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'cvsplit_plus_smerge_gblock_pair', on each of the $ct_to_be_cvssm0 effective child-parent gap-block pair to be \"cvsplit + smerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'cvsplit_plus_smerge_gblock_pair', on each of the $ct_to_be_cvssm0 effective child-parent gap-block pairs to be \"cvsplit + smerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.

for (my $i=0; $i < $ct_to_be_cvssm0; $i++) { # Main Outer for-loop (over effective child-parent pairs to be "cvsplit + smerge"d). #

	my $info_cand = $to_be_cvsplit_plus_smerged0->[$i];
	my ($gb_sbj_ch, $gb_sbj_pa, $dist) = @{$info_cand};

	my $session = 'MAIN I-37b';

    (print STDOUT "\n< The $i th effective child-parent pair to be \"cvsplit + smerge\"d: (indx_gb1, indx_gb2) = ($gb_sbj_ch, $gb_sbj_pa) ; (distance=$dist) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_37b/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_37b = "$maindir_out/I_37b.CVSplit_plus_SMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_cvsplit_plus_smerged.txt';

	my @subjects = ($gb_sbj_ch, $gb_sbj_pa);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th effective child-parent pair to be \"cvsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (i) Actually "CVSplit + SMerge" the Effective Child-Parent Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'cvsplit_plus_smerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

			# Modified for 'vsplit_plus_smerge_gblock_pair'. #
	## [ my ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, $new_indices_gbs, $old_indices_gbs, ] ... for the predecessor, 'smerge_gblock_pair'. ##
			# Modified for 'cvsplit_plus_smerge_gblock_pair'. (1) #
=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (1)
	my ($flag_vssm, $gb_ch_left_NEW, $gb_kpt_NEW, 
	    $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	    $gb_ch_2B_mgd_intm,
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (1)"
	my ($flag_cvssm, $gb_cch_left_NEW, $gb_kpt_NEW, 
	    $gb2incr_coord_1way_dsplc_bf_cvspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_cvspl,
	    $gb_cch_2B_mgd_intm, 
			# End of "Modified for 'cvsplit_plus_smerge_gblock_pair'. (1)" #

	    $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	    $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
			# End of "Modified for 'vsplit_plus_smerge_gblock_pair'." #

	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    ## [ $new_list_gblocks, $new_list_excluded_bin_hrels # Added for 'vsplit_plus_smerge_gblock_pair'. # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (2)
	    $new_list_gblocks # Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (2)
	    )
	  = cvsplit_plus_smerge_gblock_pair ( $gb_sbj_ch, $gb_sbj_pa, $CHAR_ABS, # Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (3)
	  ## [ = vsplit_plus_smerge_gblock_pair ( $gb_sbj_pa, $gb_sbj_ch, ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (3)
	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT,
			# Added for 'vsplit_plus_smerge_gblock_pair'. #
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw},  @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq},
	    @set_exnodes, %{$nodeid2msa_row}
			# End of "Added for 'vsplit_plus_smerge_gblock_pair'." #
		);
	unless ($flag_cvssm) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	unless ($flag_cvssm >0) {
	    print STDOUT "[CAUTION '$plscript_name'] The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' FAILED because some overlaps canNOT be UNDONE, on the $i th gap-block pair for $session ! ... I will SKIP this pair ... \n";
	    next;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$gb_cch_left_NEW = $gb_cch_left_NEW, \$gb_kpt_NEW = $gb_kpt_NEW) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = @{$new_list_gblocks};

	##my @list_excluded_bin_hrels = @{$new_list_excluded_bin_hrels};
	my @list_excluded_bin_hrels = ();
   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);

	my %gb2incr_coord_1way_dsplc_bf_cvspl = %{$gb2incr_coord_1way_dsplc_bf_cvspl};
   (print STDOUT 'Set %gb2incr_coord_1way_dsplc_bf_cvspl  according to an output of ', "'cvsplit_plus_smerge_gblock_pair' .\n\n") if ($TEST_LEVEL_MAIN>1);

	my ($gb_shftd_af_cvspl_intm, $incr_coord_shftd_af_cvspl) = ('N.A.', 'N.A.');
   (print STDOUT 'Manually Set: ($gb_shftd_af_cvspl_intm, $incr_coord_shftd_af_cvspl) = ', "($gb_shftd_af_cvspl_intm, $incr_coord_shftd_af_cvspl) .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Vertical Split + Merge (same type)". #

	##my $sbsbdir_info = "$sbsbdir/Info";


	##my @ofnames0 = ('basic_info.txt', 'info_1way_displacements.txt', 'info_serial_shifts_af_vspl.txt', 'rec_undo_hovlps_af_vspl.txt', 'info_smerge_af_vspl.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_smg.txt'); # MODIFIED in revision (on 2019/08/30). #
	my @ofnames0 = ('basic_info.txt', 'info_1way_displacements_bf_cvspl.txt', 'info_serial_shifts_af_cvspl.txt', 'rec_undo_hovlps_af_cvspl.txt', 'info_smerge_af_cvspl.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_smg.txt'); # => Further MODIFIED (on 2019/10/17). #

	# $flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	unless ( print_out_info_chngs_by_vsplit_plus_smerge ($sbsbdir_info, @ofnames0, 
		$flag_cvssm, $gb_sbj_ch, $gb_sbj_pa, $gb_cch_left_NEW, $gb_kpt_NEW, 
		%gb2incr_coord_1way_dsplc_bf_cvspl, 
		$gb_shftd_af_cvspl_intm, $incr_coord_shftd_af_cvspl, %{$gb2incr_bl_coord_undo_ovlps_af_cvspl},
		$gb_cch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
		@{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$intm_indices_old_gbs}, @{$new_indices_intm_gbs},
		%{$gb2incr_coord_undo_af_smg}
			) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_vsplit_plus_smerge' FAILED on the $i th effective child-parent pair to be \"cvsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge', Ended on the $i th pair for $session  ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge', Ended on the $i th pair for $session  (, writing to several files under \"$sbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0);



		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "CVSplit + Smerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	##print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th effective child-parent pair had been \"cvsplit + smerge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"cvsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-37b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"cvsplit + smerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"cvsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"cvsplit + smerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"cvsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-37b)" ##


} # End of the Main Outer for-loop (over effective child-parent pairs to be "cvsplit + smerge"d). #

			## END of "ADDED on 2019/09/12. (MAIN I-37b)" ##

	# End of "(MAIN-I-37b) Perform "Vertical-Split (into complementary-sibling blocks) + Horizontal Merge (same-type)" + Multiple-"Shift"s." #


	# (MAIN-I-46a) Perform "(Horizontal) Split (into same-type blocks) + (complete) Vertical-Merge (of sibling blocks)" + Multiple-"Shift"s. # (ADDED the item on 2019/09/30)

			## ADDED on 2019/10/17. (MAIN I-46a) ##

(print STDOUT "\n<<< MAIN I-46a: Perform \" (Horizontal-)Split (into same-type blocks) + (complete) Vertical-Merge (of effective sibling blocks)\" + Multiple-\"Shift\"s, via 'ssplit_plus_vmerge_gblock_pair' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

			## ADDED on 2019/09/07. (MAIN I-37a) (1) ##

my $subdir_MI_46a = "$maindir_out/I_46a.SSplit_plus_VMerge_plus_MltShifts";


		# (i) Search for candidate gap-block pairs. #

(print STDOUT "\n<< (i)  Search for Effective Sibling Gap-Block Pairs to be \"SSplit + VMerge\"d , via 'list_nn_ssplit_plus_vmerge_candidates'.  >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $THRSH_DIST_SSVM = $HF_WDTH_SHIFT;

(print STDOUT 'Manually Set: $THRSH_DIST_SSVM = ', "$THRSH_DIST_SSVM .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($to_be_ssplit_plus_vmerged0)
 = list_nn_ssplit_plus_vmerge_candidates (
	@bds_gblocks0_in_wd, @{$gblock_sizes0}, @{$inter_block_relations0}, $THRSH_DIST_SSVM);
unless (defined $to_be_ssplit_plus_vmerged0) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_ssplit_plus_vmerge_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_to_be_ssvm0 = @{$to_be_ssplit_plus_vmerged0};

(print STDOUT "\n --- The Subroutine 'list_nn_ssplit_plus_vmerge_candidates' ENDED on the main window (w/ #{to_be_ssplit+vmerged}=$ct_to_be_ssvm0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


if (1) { # (Main I-46a (i) #

	my $ofile_name2 = 'to_be_ssplit_plus_vmerged.txt';
	my $label2 = "to be \"(Horizontally) Split (same type)\" + \"Vertically Merge\"d";

	print STDOUT "\n ... Record the outputs of 'list_nn_ssplit_plus_vmerge_candidates', by performing 'print_out_purge_or_merge_cands' ...\n\n";

	unless ( print_out_purge_or_merge_cands ($subdir_tgt_cands, $ofile_name2, @{$to_be_ssplit_plus_vmerged0}, $label2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_purge_or_merge_cands' FAILED on ", '@{$to_be_ssplit_plus_vmerged0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_purge_or_merge_cands', Ended on ", '@{$to_be_ssplit_plus_vmerged0}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-46a (i) # ...}"


		# (ii): Perform 'ssplit_plus_vmerge_gblock_pair'.

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'ssplit_plus_vmerge_gblock_pair', on each of the $ct_to_be_vssm0 effective sibling gap-block pair to be \"ssplit + vmerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'ssplit_plus_vmerge_gblock_pair', on each of the $ct_to_be_vssm0 effective sibling gap-block pairs to be \"ssplit + vmerge\"d. >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.


for (my $i=0; $i < $ct_to_be_ssvm0; $i++) { # Main Outer for-loop (over effective sibling pairs to be "ssplit + vmerge"d). #

	my $info_cand = $to_be_ssplit_plus_vmerged0->[$i];
	my ($gb_sbj1, $gb_sbj2, $dist) = @{$info_cand};

	my $session = 'MAIN I-46a';

			# IMPORTED (w/ modifications) from 'test_ssplit_plus_vmerge_gblock_pair.pl' (on 2019/10/17). #
	my ($len1, $len2) = @{$gblock_sizes0}[$gb_sbj1, $gb_sbj2];

	my ($gb_sbj_L, $gb_sbj_S) = ($len1 > $len2) ? ($gb_sbj1, $gb_sbj2) : ($gb_sbj2, $gb_sbj1) ;
	#
	# $gb_sbj_L is larger, $gb_sbj_S is smaller.
			# END of "IMPORTED (w/ modifications) from 'test_ssplit_plus_vmerge_gblock_pair.pl' (on 2019/10/17)." #


    (print STDOUT "\n< The $i th effective sibling pair to be \"ssplit + vmerge\"d: (indx_gb1, indx_gb2) = ($gb_sbj1, $gb_sbj2) ; (distance=$dist) ; (\$gb_sbj_L, \$gb_sbj_S) = ($gb_sbj_L, $gb_sbj_S). >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbdir = "$subdir_MI_46a/Pair_$i";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_46a = "$maindir_out/I_46a.SSplit_plus_VMerge_plus_MltShifts";

	my $ofname_info1 = 'gblks_to_be_ssplit_plus_vmerged.txt';

	my @subjects = ($gb_sbj_L, $gb_sbj_S);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th effective sibling pair to be \"ssplit + vmerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session (, writing to \"$sbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (i) Actually "SSplit + VMerge" the Effective Sibling Pair of Gap-Blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'ssplit_plus_vmerge_gblock_pair' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my @curr_gb_coords = @{$orgs_gb0_coords};

			# Modified for 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/17). (I-46a-A) #
	## [[
	## my ($flag_vssm, $gb_ch_left_NEW, $gb_kpt_NEW, 
	##    $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	##    $gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	##    $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
	## ]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'."

	my ($flag_ssvm, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
	    $gb_sbj1_left_intm, $gb_sbj1_right_intm, $intm_indices_old_gbs, $gb_intm2incr_bl_coord_undo_ovlps_af_sspl,
	    $if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm,
	    $incr_coord_sbj_shftd_B4_vmg, $gb_new2incr_coord_1way_dsplc_B4_vmg, $new_indices_intm_gbs,
	    $new_indices_old_gbs, $old_indices_new_gbs, $gb_new2incr_coord_undo_af_vmg,

			# End of "Modified for 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/17). (I-46a-A)" #
 #
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks, $new_list_excluded_bin_hrels # Added for 'vsplit_plus_smerge_gblock_pair'. #
	    )
	## [  = vsplit_plus_smerge_gblock_pair ( $gb_sbj_pa, $gb_sbj_ch, ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. ##
	 = ssplit_plus_vmerge_gblock_pair ( $gb_sbj_L, $gb_sbj_S,  # Modified for 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/17). (I-46a-B) #

	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT,

			# Added for 'vsplit_plus_smerge_gblock_pair'. #
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw},  @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}

	## , @set_exnodes, %{$nodeid2msa_row} # DISABLED for 'ssplit_plus_vmerge_gblock_pair'. ##
			# End of "Added for 'vsplit_plus_smerge_gblock_pair'." #
		);
	unless ($flag_ssvm) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'ssplit_plus_vmerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	unless ($flag_ssvm >0) {
	    print STDOUT "[CAUTION '$plscript_name'] The MAIN Subroutine 'ssplit_plus_vmerge_gblock_pair' FAILED because some overlaps canNOT be UNDONE, on the $i th gap-block pair for $session ! ... I will SKIP this pair ... \n";
	    next;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'ssplit_plus_vmerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$gb_sbj1_unvmgd_NEW = $gb_sbj1_unvmgd_NEW, \$gb_mgd_NEW = $gb_mgd_NEW) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = @{$new_list_gblocks};

	my @list_excluded_bin_hrels = @{$new_list_excluded_bin_hrels};

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (A)
	my %gb2incr_coord_1way_dsplc_bf_vspl = ('N.A' => 1); # A dummy hash. (ADDED in revision on 2019/08/30.)
  (print STDOUT 'Manually Set: %gb2incr_coord_1way_dsplc_bf_vspl = ', "('N.A' => 1) .\n\n") if ($TEST_LEVEL_MAIN>1);
=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (A)"


		# (ii-0) Print out the Changes Resulting from the "(Horizontal) Split (same type) + Vertical-Merge". #

	##my $sbsbdir_info = "$sbsbdir/Info";

			# MODIFIED for 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/17). (I-46a-C) #

	my @ofnames0 = ('basic_info.txt', 'info_sspl.txt', 'rec_undo_hovlps_af_sspl.txt', 'info_vmerge_af_sspl.txt', 'info_1way_displacements_bf_vmg.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_vmg.txt');

		### RESTART FROM HERE (on 2019/10/17). (I-reserve) ###

	unless ( print_out_info_chngs_by_ssplit_plus_vmerge ($sbsbdir_info, @ofnames0,
	    $flag_ssvm, $gb_sbj_L, $gb_sbj_S, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
	    $gb_sbj1_left_intm, $gb_sbj1_right_intm, 
	    %{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl},
	    $if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm, $incr_coord_sbj_shftd_B4_vmg, 
	    %{$gb_new2incr_coord_1way_dsplc_B4_vmg}, 
	    @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$intm_indices_old_gbs}, @{$new_indices_intm_gbs},
	    %{$gb_new2incr_coord_undo_af_vmg}
		    ) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_ssplit_plus_vmerge' FAILED on the $i th effective parent-child pair to be \"ssplit + vmerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_ssplit_plus_vmerge', Ended on the $i th pair for $session  ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_ssplit_plus_vmerge', Ended on the $i th pair for $session  (, writing to several files under \"$sbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (B)
	my @ofnames0 = ('basic_info.txt', 'info_1way_displacements.txt', 'info_serial_shifts_af_vspl.txt', 'rec_undo_hovlps_af_vspl.txt', 'info_smerge_af_vspl.txt', 'old_intm_new_indices_gblks.txt', 'rec_undo_hovlps_af_smg.txt'); # MODIFIED in revision (on 2019/08/30). #

	# $flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	unless ( print_out_info_chngs_by_vsplit_plus_smerge ($sbsbdir_info, @ofnames0, 
		$flag_vssm, $gb_sbj_pa, $gb_sbj_ch, $gb_ch_left_NEW, $gb_kpt_NEW, 
		%gb2incr_coord_1way_dsplc_bf_vspl, 
		$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, %{$gb2incr_bl_coord_undo_ovlps_af_vspl},
		$gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, 
		@{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$intm_indices_old_gbs}, @{$new_indices_intm_gbs},
		%{$gb2incr_coord_undo_af_smg}
			) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_vsplit_plus_smerge' FAILED on the $i th effective parent-child pair to be \"vsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit_plus_smerge', Ended on the $i th pair for $session  ...\n\n") if ($TEST_LEVEL_MAIN >0);
=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (B)"

			# END of "MODIFIED for 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/17). (I-46a-C)" #


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Ssplit + Vmerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	##print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $i th effective parent-child pair had been \"vsplit + smerge\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $i th pair had been \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN I-46a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th pair \"vsplit + smerge\"d: $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th pair \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th pair \"vsplit + smerge\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th pair \"vsplit + smerge\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN I-46a)" ##


} # End of the Main Outer for-loop (over effective sibling pairs to be "ssplit + vmerge"d). #

			## END of "ADDED on 2019/10/17. (MAIN I-46a)" ##

	# END of "(MAIN-I-46a) Perform "(Horizontal) Split (into same-type blocks) + (complete) Vertical-Merge (of sibling blocks)" + Multiple-"Shift"s." # (ADDED the item on 2019/09/30)



	# (MAIN-I-8) Perform "Reverse-(i)CII ((incomplete) collapse of independent insertions)" + Multiple-"Shift"s. # (ADDED the item on 2019/09/12)
	# (See SM-7 (viii) of "suppl5_blueprint1_ANEX.draft10.odt" (to be constructed) as a reference.)

(print STDOUT "\n<<< MAIN I-8: Perform \"Reverse-(i)CII\" ((incomplete) collapse of independent insertions) + Multiple-\"Shift\"s, via 'reverse_spec_cii_in_wd' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

			## ADDED on 2019/09/30. (MAIN-I-8) (1) ##

my $subdir_MI_8 = "$maindir_out/I_8.RevCII_plus_MltShifts";


		# (i) Search for CII-candidate regions. #

(print STDOUT "\n<< (i-1)  Search for unrefined CII-candidate regions, via 'list_cii_candidates' >>\n\n") if ($TEST_LEVEL_MAIN>0) ;

our $MAXCT_MONOPHYL_SETS = 3;

(print STDOUT 'Manually Set: $MAXCT_MONOPHYL_SETS = ', $MAXCT_MONOPHYL_SETS, " .\n\n") if ($TEST_LEVEL_MAIN >1);

my ($list_cii_cands, $set_gpblocks) = list_cii_candidates (@{$set_clms0_in_wd}, %{$node2exoffs_msa_rows}, %{$set_msa_rows2br_up_or_lw}, $CHAR_ABS, %node2pa, %node2ch, $topnode_id, %{$br2eq}, $MAXCT_MONOPHYL_SETS);
unless (defined $list_cii_cands) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_cii_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_gpbs_in_wd = @{$set_gpblocks};
my $ct_cii_cands0 = @{$list_cii_cands};


(print STDOUT "\n --- The Subroutine 'list_cii_candidates', ENDED on the main window (w/ #{gap-pattern-blocks in the window}=$ct_gpbs_in_wd, #{CII-candidate-regions}_unrefined=$ct_cii_cands0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


if (1) { # (Main I-8 (i-1) #

	my $ofile_name1 = 'list_gap_pattern_blocks_in_wd.txt';
	my $ofile_name2 = 'cii_cands_unrefined.txt';
	my $cand_class = 'UNrefined'; #  ADDED on 2019/10/01. #

	print STDOUT "\n ... Record the outputs of 'list_cii_candidates', by performing 'print_out_list_gpblocks' & 'print_out_list_cii_cands' ...\n\n";

	unless ( print_out_list_gpblocks ($subdir_tgt_cands, $ofile_name1, @{$set_gpblocks}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gpblocks' FAILED on ", '@{$set_gpblocks}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_list_gpblocks', Ended on ", '@{$set_gpblocks}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name1\") ...\n\n\n";

	unless ( print_out_list_cii_cands ($subdir_tgt_cands, $ofile_name2, @{$list_cii_cands}, $cand_class) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_cii_cands' FAILED on ", '@{$list_cii_cands}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_list_cii_cands', Ended on ", '@{$list_cii_cands}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-8 (i-1) # ...}"


(print STDOUT "\n<< (i-2)  Weed out UNlikely candidate regions of CII, via 'filter_cii_candidates'. >>\n\n") if ($TEST_LEVEL_MAIN>0);

my ($selected_cii_cands) = filter_cii_candidates (@{$list_cii_cands}, @bds_gblocks0_in_wd, @{$inter_block_relations0});
unless (defined $selected_cii_cands) {
	print STDOUT "[ERROR '$plscript_name'] The Subroutine 'filter_cii_candidates' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_cii_cands = @{$selected_cii_cands};

print STDOUT "\n --- The Subroutine 'filter_cii_candidates', ENDED on the main window (w/ #{CII-candidate-regions}_selected=$ct_cii_cands) ...\n\n";


if (1) { # (Main I-8 (i-2) #

	my $ofile_name2 = 'cii_cands_selected.txt';
	my $cand_class = 'Selected'; # ADDED on 2019/10/01. #

	print STDOUT "\n ... Record the outputs of 'filter_cii_candidates', by performing 'print_out_list_cii_cands' ...\n\n";

	unless ( print_out_list_cii_cands ($subdir_tgt_cands, $ofile_name2, @{$selected_cii_cands}, $cand_class) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_cii_cands' FAILED on ", '@{$selected_cii_cands}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_list_cii_cands', Ended on ", '@{$selected_cii_cands}', " in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name2\") ...\n\n\n";

} # End of "if (1) { # (Main I-8 (i-2) # ...}"

			## END of "ADDED on 2019/09/30. (MAIN-I-8) (1)" ##

			## ADDED on 2019/10/01. (MAIN-I-8) (2) ##

		# (ii): Perform 'reverse_spec_cii_in_wd'. #

##(print STDOUT "\n<< (ii) Perform the Subroutine, 'reverse_spec_cii_in_wd', on each of the $ct_cii_cands LIKELY CII-candidate region.  >>\n\n") if ($TEST_LEVEL_MAIN >0);
(print STDOUT "\n<< (ii) Perform the Subroutine, 'reverse_spec_cii_in_wd', on each of the $ct_cii_cands LIKELY CII-candidate regions.  >>\n\n") if ($TEST_LEVEL_MAIN >0); # Corrected on 2020/05/09.


for (my $cc=0; $cc < $ct_cii_cands; $cc++) { # Main Outer for-loop (over CII-candidate regions).

	my ($le_cii_clm, $re_cii_clm, $set_mnp_sets_pres_rows, $set_br_up_or_lw, $le_cii_gpb, $re_cii_gpb) = @{$selected_cii_cands->[$cc]};

    (print STDOUT "\n< The $cc th likely CII-candidate region: (\$le_cii_clm, \$re_cii_clm) = ($le_cii_clm, $re_cii_clm) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $session = 'MAIN I-8';

	my $sbsbdir = "$subdir_MI_8/CCand_$cc";
	my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbdir_out = "$sbsbdir/Out";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

##my $subdir_MI_8 = "$maindir_out/I_8.RevCII_plus_MltShifts";

=For the predecessor, 'cvsplit_plus_smerge_gblock_pair'.
	my $ofname_info1 = 'gblks_to_be_cvsplit_plus_smerged.txt';

	my @subjects = ($gb_sbj_ch, $gb_sbj_pa);

	unless ( print_out_info_mvd_sbj ($sbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @bds_gblocks0_in_wd, @{$gblock_sizes0}, @rows_affd_by_blks0_in_wd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th effective child-parent pair to be \"cvsplit + smerge\"d for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th pair for $session ...\n\n") if ($TEST_LEVEL_MAIN >0);
=cut # END of "For the predecessor, 'cvsplit_plus_smerge_gblock_pair'."


		# (i) Actually "Reverse" the CII-candidate. #

	##my @@set_ingredients_indel_hsts0 = @set_ingredients_indel_hsts;
	my @set_ingredients_indel_hsts0 = @set_ingredients_indel_hsts; # CORRECTED on 2019/10/05. #

   (print STDOUT "\n", 'Manually Set: @@set_ingredients_indel_hsts0 = @set_ingredients_indel_hsts', " .\n\n") if ($TEST_LEVEL_MAIN >2);

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'reverse_spec_cii_in_wd' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my ($flag_revcii, 
	    $new_set_cls_gpbs, $sets_gpbs_at_spec_places, $new_set_gpatterns,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks)

	  = reverse_spec_cii_in_wd (
	     @{$set_clms0_in_wd}, $CHAR_ABS, $CHAR_PRES, 
	     $le_cii_clm, $re_cii_clm, @{$set_mnp_sets_pres_rows},
	     @set_ingredients_indel_hsts0,

	     ##%node2ch, %node2pa, $topnode_id, %name2id, @seqids, @set_exnodes,
	     %node2ch, %node2pa, $topnode_id, %name2id, @{$seqnames}, @set_exnodes, # CORRECTED on 2019/10/05. #

	     %{$br2eq}, %{$node2depth}, %{$node2ct_exoffs},
	     %{$set_msa_rows2br_up_or_lw}, %{$nodeid2msa_row},

	     ##@{$nodeids_in_msa}, $MAXCT_1WAY_SHIFTS);
	     @{$nodeids_in_msa}, $HF_WDTH_SHIFT); # CORRECTED on 2019/10/05. #

	unless ($flag_revcii) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'reverse_spec_cii_in_wd' FAILED on the $cc th likely CII-candidate region for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

	my $ct_new_gsegs = @{$new_set_cls_gpbs};
	my $ct_new_gpbs = 0;
	foreach my $gseg (@{$new_set_cls_gpbs}) { $ct_new_gpbs += @{$gseg}; }
	my $new_ct_clms = @{$new_curr_set_clms};

    ##(print STDOUT "\n --- The MAIN Subroutine 'reverse_spec_cii_in_wd' ENDED  on the $cc th likely CII-candidate region for $session (w/ \$flag_revcii = $flag_revcii, #{new gapped-segments}=$ct_new_gsegs, #{new gap-pattern blocks}=$ct_new_gpbs, #{columns}_new=$new_ct_clms) ...\n\n\n") if ($TEsT_LEVEL_MAIN >0);
    (print STDOUT "\n --- The MAIN Subroutine 'reverse_spec_cii_in_wd' ENDED  on the $cc th likely CII-candidate region for $session (w/ \$flag_revcii = $flag_revcii, #{new gapped-segments}=$ct_new_gsegs, #{new gap-pattern blocks}=$ct_new_gpbs, #{columns}_new=$new_ct_clms) ...\n\n\n") if ($TEST_LEVEL_MAIN >0); # CORRECTED on 2019/10/05. #


=For the predecessor, 'cvsplit_plus_smerge_gblock_pair'.
	my ($flag_cvssm, $gb_cch_left_NEW, $gb_kpt_NEW, 
	    $gb2incr_coord_1way_dsplc_bf_cvspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_cvspl,
	    $gb_cch_2B_mgd_intm, 
			# End of "Modified for 'cvsplit_plus_smerge_gblock_pair'. (1)" #

	    $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	    $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
			# End of "Modified for 'vsplit_plus_smerge_gblock_pair'." #

	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    ## [ $new_list_gblocks, $new_list_excluded_bin_hrels # Added for 'vsplit_plus_smerge_gblock_pair'. # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (2)
	    $new_list_gblocks # Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (2)
	    )
	  = cvsplit_plus_smerge_gblock_pair ( $gb_sbj_ch, $gb_sbj_pa, $CHAR_ABS, # Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (3)
	  ## [ = vsplit_plus_smerge_gblock_pair ( $gb_sbj_pa, $gb_sbj_ch, ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (3)
	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,

	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0}, $HF_WDTH_SHIFT,
			# Added for 'vsplit_plus_smerge_gblock_pair'. #
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw},  @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq},
	    @set_exnodes, %{$nodeid2msa_row}
			# End of "Added for 'vsplit_plus_smerge_gblock_pair'." #
		);
	unless ($flag_cvssm) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' FAILED on the $i th gap-block pair for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	unless ($flag_cvssm >0) {
	    print STDOUT "[CAUTION '$plscript_name'] The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' FAILED because some overlaps canNOT be UNDONE, on the $i th gap-block pair for $session ! ... I will SKIP this pair ... \n";
	    next;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'cvsplit_plus_smerge_gblock_pair' ENDED on the $i th pair for $session (w/ \$gb_cch_left_NEW = $gb_cch_left_NEW, \$gb_kpt_NEW = $gb_kpt_NEW) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);
=cut # END of "For the predecessor, 'cvsplit_plus_smerge_gblock_pair'."


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_wd = @{$new_list_gblocks};

	my @list_excluded_bin_hrels = ();
   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Subsidiary Outputs of 'reverse_spec_cii_in_wd'. #

	##my $sbsbdir_info = "$sbsbdir/Info";

	my $sets_old_indices_new_gpbs = $sets_gpbs_at_spec_places;

   (print STDOUT "\n", 'Manually Set: @{$sets_old_indices_new_gpbs} = @{$sets_gpbs_at_spec_places}', " .\n\n") if ($TEST_LEVEL_MAIN >1);

	my @ofnames0 = ('new_set_cls_gpbs.txt', 'sets_old_indices_for_new_gpbs.txt');

	unless ( print_out_sbsdry_outputs_rev_spec_cii_in_wd ($sbsbdir_info, @ofnames0, 
		   @{$new_set_cls_gpbs}, @{$sets_old_indices_new_gpbs}, @{$new_set_gpatterns},
			) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_sbsid_outputs_rev_spec_cii_in_wd' FAILED on the $cc th likely CII-candidate region for $session !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Subroutine, 'print_out_sbsid_outputs_rev_spec_cii_in_wd', Ended on the $cc th likely CII-candidate region for $session  ...\n\n") if ($TEST_LEVEL_MAIN >0);


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "CVSplit + Smerge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	##print STDOUT 'Manually Set:  @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks}, @{$rows_affd_by_new_blks},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts0, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA after the $cc th likely CII-candidate region had been \"Reverse\"d !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session after the $cc th likely CII-candidate region had been \"Reverse\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN-I-8) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $cc th CII-candidate \"Reverse\"d  : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the  w/ the $cc th CII-candidate \"Reverse\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the  w/ the $cc th CII-candidate \"Reverse\"d : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the  w/ the $cc th CII-candidate \"Reverse\"d ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'Main', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN-I-8)" ##


} # End of the Main Outer for-loop (over CII-candidate regions). #

			## END of "ADDED on 2019/10/01. (MAIN-I-8) (2)" ##

	# END of "(MAIN-I-8) Perform "Reverse-(i)CII ((incomplete) collapse of independent insertions)" + Multiple-"Shift"s." # (ADDED the item on 2019/09/12)

	##### END of "(MAIN I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}." #####



	##### (MAIN II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}. #####

(print STDOUT "\n<<<<< MAIN (II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}. >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

for (my $sw=0; $sw <$ct_sbwds; $sw++) { # OUTER for-loop for MAIN (II) (over the relevant sub-windows). #

				# Added on 2019/07/03. #

    my $sbwd_ID = $sbwd_IDs->[$sw];

  (print STDOUT "\n{{{ Sub-window: '$sbwd_ID' }}}\n\n") if ($TEST_LEVEL_MAIN>0);

    ##my $sbwork_dir = ${$sub_work_dirs[$sw]};
    ##my $sbdir_in = ${$subdirs_in[$sw]};

		# Input files. #
    my $infile_sbset_clms0 = ${$infiles_sbset_clms0[$sw]};
    my $infile_sblist_gblks0 = ${$infiles_sblist_gblks0[$sw]};
    my $infile_pcands = ${$infiles_pcands[$sw]};

		# Directories for outputs. #
    my $sbdir_in_chk = ${$subdirs_in_chk[$sw]};
    my $sbdir_interim = ${$subdirs_interim[$sw]};
    my $sbdir_out = ${$subdirs_out[$sw]};


	### Read in the Initial Sub-Set of Columns. ###

    my ($beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2, $set_clms0_in_sbwd)
	= readin_set_clms ($infile_sbset_clms0);
unless (defined $set_clms0_in_sbwd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_set_clms', FAILED on the input file, \"$infile_sbset_clms0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    my $ct_clms0_in_sbwd = @{$set_clms0_in_sbwd};

  (print STDOUT "\n ... The subroutine, 'readin_set_clms' Ended (w/ \$ct_clms0_in_sbwd = $ct_clms0_in_sbwd, [\$beg_sbwd, \$end_sbwd] = [$beg_sbwd, $end_sbwd], [\$if_on_le_msa2, \$if_on_re_msa2] = [$if_on_le_msa2, $if_on_re_msa2]) ...\n\n") if ($TEST_LEVEL_MAIN>0);

  if ($TEST_LEVEL_MAIN>1) { # (S7) #

	my $fname_set_clms0 = 'set_columns0.txt';

	unless ( print_out_set_clms ($sbdir_in_chk, $fname_set_clms0, @{$set_clms0_in_sbwd}, $beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING, the subroutine, 'print_out_set_clms', Ended RECORDING information (on the sub-window '$sbwd_ID') into the output file, \"$sbdir_in_chk/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (S7) # ...}"

	##### End of "Read in the Initial Sub-Set of Columns." #####


	### Read in the Initial Sub-List of Gap-Blocks. ###

    # $infile_sblist_gblks0 = ${$infiles_sblist_gblks0[$sw]};

    my ($list_gblks0_in_sbwd) = readin_list_gblks ($infile_sblist_gblks0);
    unless (defined $list_gblks0_in_sbwd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_list_gblks', FAILED on the input file, \"$infile_sblist_gblks0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    my $ct_gblks0_in_sbwd = @{$list_gblks0_in_sbwd};

  (print STDOUT "\n ... The subroutine, 'readin_list_gblks' Ended for the sub-window '$sbwd_ID' (w/ \$ct_gblks0_in_sbwd = $ct_gblks0_in_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);


 if ($TEST_LEVEL_MAIN>1) { # (S8) #

	my $fname_list_gblks0 = 'list_gblocks0_org.txt';

	unless ( print_out_readin_list_gblks ($sbdir_in_chk, $fname_list_gblks0, @{$list_gblks0_in_sbwd}, 'org') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING 'readin_list_gblks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'original' gap-blocks in the sub-window '$sbwd_ID') into the output file, \"$sbdir_in_chk/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (S8) # ...}"


	## NOTE: Because the node (branch) ID assignments may have changed from the environment where $infile_list_gblks0 was output to the current environment,
	##      we MUST examine the node (branch) IDs, and, if necessary, REVISE them in "(Pre S3)"!!

	### End of "Read in the Initial Sub-List of Gap-Blocks." ###


        ## (Pre S3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_sbwd}, if necessary. ## 

   (print STDOUT '<<< Preliminary S3: Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_sbwd}, if necessary,', " via 'revise_list_gblocks' >>>\n\n") if ($TEST_LEVEL_MAIN>1);


    unless ( revise_list_gblocks (@{$list_gblks0_in_sbwd}, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa}) ) {
        print STDOUT "[ERROR '$plscript_name'] The subroutine 'revise_list_gblocks' FAILED on the input tree & for the sub-window '$sbwd_ID' !\n";
        print STDERR "[ERROR '$plscript_name'] $! !\n";
        die ;
    }
  (print STDOUT "\n ... The subroutine 'revise_list_gblocks' Ended for the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN>1);

  if ($TEST_LEVEL_MAIN>1) { # (Pre S3) #

        my $fname_list_gblks0 = 'list_gblocks0_rvsd.txt';

        unless ( print_out_readin_list_gblks ($sbdir_interim, $fname_list_gblks0, @{$list_gblks0_in_sbwd}, 'rvsd') ) {
                print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the sub-window '$sbwd_ID' !\n";
                print STDERR "[ERROR '$plscript_name'] $! !\n";
                die ;
        }
    (print STDOUT "\n ... For the purpose of CHECKING 'revise_list_gblocks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'revised' gap-blocks in the sub-window '$sbwd_ID') into the output file, \"$sbdir_interim/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre S3) # ...}"

        ## End of "(Pre S3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_sbwd}, if necessary." ## 
				# End of "Added on 2019/07/03." #

				# Added on 2019/07/04. #

	# NOTE on (Pre S7): We will Use %{$node2set_sw_extcond_probs0} as it is.
	#		TO do so, subtract $beg_wd (instead of $beg_sbwd) from the site numbers in @{$set_columns0_in_sbwd} (in Pre S13).

	### (Pre S10) Prepare @{$inter_block_relations0_sbwd}, via 'mk_inter_block_relations'. ###

   (print STDOUT '<<< Preliminary S10: Compute @{$inter_block_relations0_sbwd}, ', " via 'mk_inter_block_relations'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

# @{$list_gblks[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

    my ($inter_block_relations0_sbwd) = mk_inter_block_relations (@{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, %node2pa, %node2ch, $topnode_id, %{$br2eq}); # Modified on 2019/07/02. #
    ##my ($inter_block_relations0_sbwd) = mk_inter_block_relations (@{$list_gblks0_in_sbwd}, $indx_br, $indx_u_or_d, %node2pa, %node2ch, $topnode_id, %{$br2eq});
    unless (defined $inter_block_relations0_sbwd) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_inter_block_relations' FAILED on the input MSA for the sub-window '$sbwd_ID' !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
    }
(print STDOUT " ... Subroutine, 'mk_inter_block_relations', Ended for the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN>0);


 if ($TEST_LEVEL_MAIN>1) { # (Pre S10) #

	my $ofile_name = 'inter_block_relations0.txt';

	print STDOUT "\n ... Check the outputs of 'mk_inter_block_relations', by performing 'print_out_inter_block_relations' ...\n\n";
	unless ( print_out_inter_block_relations ($sbdir_interim, $ofile_name, @{$inter_block_relations0_sbwd}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_inter_block_relations' FAILED on ", '@{$inter_block_relations0_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_inter_block_relations', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre S10) # ...}"

	### End of "(Pre S10) Prepare @{$inter_block_relations0_sbwd}, via 'mk_inter_block_relations'." ###  


	### (Pre S11) MANUALLY prepare @bds_gblocks0_in_sbwd & @rows_affd_by_blks0_in_sbwd. ###

  (print STDOUT '<<< Preliminary S11: MANUALLY prepare @bds_gblocks0_in_sbwd & @rows_affd_by_blks0_in_sbwd >>>', "\n\n") if ($TEST_LEVEL_MAIN >0);

  # @{$list_gblks0_in_sbwd[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

    my @bds_gblocks0_in_sbwd = ();
    my @rows_affd_by_blks0_in_sbwd = ();
    for (my $gb=0; $gb <$ct_gblks0_in_sbwd; $gb++) {

	my ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, $set_affd_exnodes, $set_affd_msa_rows) = @{$list_gblks0_in_sbwd->[$gb]};

		# NOTE: Because these boundaries specify the positions within the local alignment within the sub-window,  
		# subtract $beg_sbwd (instead of $beg_wd) from the raw values.

	$bds_gblocks0_in_sbwd[$gb] = [$lm_clm - $beg_sbwd, $rm_clm - $beg_sbwd];
	$rows_affd_by_blks0_in_sbwd[$gb] = $set_affd_msa_rows;
    }

  if ($TEST_LEVEL_MAIN>1) { # (Pre S11) #

	my $ofile_name = 'row_clm_info_gblocks0.txt';

	print STDOUT "\n", ' ... Check the contents of @bds_gblocks0_in_sbwd & @rows_affd_by_blks0_in_sbwd, ', "by performing 'print_out_row_clm_info_gblocks' ...\n\n";
	unless ( print_out_row_clm_info_gblocks ($sbdir_interim, $ofile_name, @bds_gblocks0_in_sbwd, @rows_affd_by_blks0_in_sbwd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_row_clm_info_gblocks' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_row_clm_info_gblocks', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre S11) # ...}"

	### End of "(Pre S11) MANUALLY prepare @bds_gblocks0_in_sbwd & @rows_affd_by_blks0_in_sbwd." ### 


	### (Pre S12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment in the sub-window, via 'mk_orgs_bds_gb_coords'. ###

  (print STDOUT "<<< Preliminary S12: Prepare the coordinate frame for the multiple-\"shifts\" from the initial alignment of the sub-window, via 'mk_orgs_bds_gb_coords'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

    my ($orgs_gb0_coords_sbwd, $bds_gb0_coords_sbwd, $gb0_coords_sbwd_frm)
	= mk_orgs_bds_gb_coords (@bds_gblocks0_in_sbwd, 0, $end_sbwd - $beg_sbwd, @{$inter_block_relations0_sbwd}, $HF_WDTH_SHIFT);
unless (defined $gb0_coords_sbwd_frm) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED on the input MSA in the sub-window '$sbwd_ID' !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
    }
  (print STDOUT " ... Subroutine, 'mk_orgs_bds_gb_coords', Ended for the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);

  if ($TEST_LEVEL_MAIN>1) { # (Pre S12) #

	my $ofile_name = 'orgs_bds_gblocks0_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords', by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbdir_interim, $ofile_name, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre S12) # ...}"

	### End of "(Pre S12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment of the sub-window, via 'mk_orgs_bds_gb_coords'." ###


        ### (Pre S13: IMPORTANT!!) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_sbwd}. ###

	# NOTE: As explained on (Pre S7) above, We will Use %{$node2set_sw_extcond_probs0} as it is.
	#   Therefore, we need to subtract $beg_wd (instead of $beg_sbwd) from the site numbers in @{$set_columns0_in_sbwd}.

  (print STDOUT "<<< Preliminary S13 (IMPORTANT!!): Manually SUBTRACT \$beg_wd (=$beg_wd) from ALL the site numbers in ", '@{$set_clms0_in_sbwd}', ". >>>\n\n") if ($TEST_LEVEL_MAIN >0);

    for (my $c=0; $c <$ct_clms0_in_sbwd; $c++) { # Outer for-loop (over columns). #

        my $column = $set_clms0_in_sbwd->[$c];
        my $ct_rows = @{$column};

        for (my $r=0; $r <$ct_rows; $r++) { # Inner for-loop (over rows). #
            unless ($column->[$r] eq $CHAR_ABS) { $column->[$r] -= $beg_wd; }
        } # End of the inner for-loop (over rows). #

    } # End of the outer for-loop (over columns). #

  if ($TEST_LEVEL_MAIN>1) { # (Pre S13) #

        my $fname_set_clms0 = 'set_columns0_af_coord_chg.txt';

        unless ( print_out_set_clms ($sbdir_interim, $fname_set_clms0, @{$set_clms0_in_sbwd}, $beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2) ) { # CORRECTED on 2019/08/14.
        ##unless ( print_out_set_clms ($sbdir_interim, $fname_set_clms0, @{$set_clms0_in_sbwd}, $beg_sbwd - $beg_wd, $end_sbwd - $beg_wd, $if_on_le_msa, $if_on_re_msa) ) {
                print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the sub-window '$sbwd_ID' !\n";
                print STDERR "[ERROR '$plscript_name'] $! !\n";
                die ;
        }
    print STDOUT "\n ... For the purpose of CHECKING the result of  SUBTRACTING \$beg_wd (=$beg_wd) from the sites in the columns, the subroutine, 'print_out_set_clms', Ended RECORDING ", '@{$set_clms0_in_sbwd}', " (AFTER the SUBTRACTION) into the output file, \"$sbdir_interim/$fname_set_clms0\" ... \n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre S13) # ...}"

        ### End of "(Pre S13) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_sbwd}." ###


	### (Pre S14) Read in the list of (merged) purge-candidate regions, and SUBTRACT $beg_sbwd from ALL positions. ###
	# NOTE: In this case, $beg_sbwd (instead of $beg_wd) needs to be subtracted, because what we need are the positions within the initial alignment for the sub-window.

(print STDOUT "<<< Preliminary S14: Read in the list of (merged) purge-candidate regions, and manually SUBTRACT \$beg_sbwd (=$beg_sbwd) from ALL positions. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

    # $infile_pcands = ${$infiles_pcands[$sw]};

    my ($list_mgd_pcands0_sbwd, $maxspan_mgd_pcand0_sbwd) = readin_list_mgd_pcands ($infile_pcands);
    unless (defined $list_mgd_pcands0_sbwd) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_mgd_pcands' FAILED on the input file, \"$infile_pcands\" !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
    }
    my $ct_mpcands0_sbwd = @{$list_mgd_pcands0_sbwd};
 (print STDOUT "\n ... The subroutine 'readin_list_mgd_pcands' Ended for the sub-window '$sbwd_ID' (w/ #{purge-candidates}=$ct_mpcands0_sbwd, Max{span}=$maxspan_mgd_pcand0_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN >1);

    unless ( relative_coords_for_list_mgd_pcands (@{$list_mgd_pcands0_sbwd}, $beg_sbwd) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'relative_coords_for_list_mgd_pcands' FAILED for ", '@{$list_mgd_pcands0_sbwd}', " of the sub-window '$sbwd_ID' !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
    }
  (print STDOUT "\n ... The subroutine 'relative_coords_for_list_mgd_pcands' Ended for ", '@{$list_mgd_pcands0_sbwd}', " of the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >1);


  if (1) { # (Pre S14) # MODIFIED on 2019/07/17. #
  ##if ($TEST_LEVEL_MAIN >1) { # (Pre S14) #

	my $ofname_pcands = 'list_rlv_pcands.txt';

	unless ( print_out_list_mgd_pcands ($sbdir_interim, $ofname_pcands, @{$list_mgd_pcands0_sbwd}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... For the purpose of CHECKING the result of SUBTRACTING \$beg_sbwd (=$beg_sbwd) from the positions regarding the purge-candidate regions, the subroutine, 'print_out_list_mgd_pcands', Ended RECORDING ", '@{$list_mgd_pcands0_sbwd}', " (AFTER the SUBTRACTION) into the output file, \"$sbdir_interim/$ofname_pcands\" ... \n\n";

  ##} # End of "if ($TEST_LEVEL_MAIN >1) { # (Pre S14) # ...}"
  } # End of "if (1) { # (Pre S14) # MODIFIED on 2019/07/17. # ...}"

	### End of "(Pre S14) Read in the list of (merged) purge-candidate regions, and manually SUBTRACT $beg_sbwd from ALL positions." ###


	### (Pre S15) Prepare @{$gblock_sizes0_sbwd}, via 'mk_list_gblock_sizes'. ### (ADDED on 2019/07/16.)

  (print STDOUT '<<< Preliminary S15: Compute @{$gblock_sizes0_sbwd}, ', " via 'mk_list_gblock_sizes'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

    my ($gblock_sizes0_sbwd) = mk_list_gblock_sizes (@bds_gblocks0_in_sbwd, $INDX_LBD, $INDX_RBD, @{$inter_block_relations0_sbwd});
    unless (defined $gblock_sizes0_sbwd) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_list_gblock_sizes' FAILED on the sub-window '$sbwd_ID' !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
    }
  (print STDOUT " ... Subroutine, 'mk_list_gblock_sizes', Ended for the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN>0);


  if ($TEST_LEVEL_MAIN >1) { # (Pre S15) #

	my $ofname = 'gblock_sizes0.txt';

	unless ( print_out_list_gblock_sizes ($sbdir_interim, $ofname, @{$gblock_sizes0_sbwd}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblock_sizes' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_list_gblock_sizes', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofname\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN >1) { # (Pre S15) # ...}"

	### End of "(Pre S15) Prepare @{$gblock_sizes0_sbwd}, via 'mk_list_gblock_sizes'." ###


	### (Pre S16) Prepare @{$gblks_w_spec_lb0_sbwd} & @{$gblks_w_spec_rb0_sbwd}, via 'mk_lists_gblocks_w_spec_bounds'. ### (ADDED on 2019/07/17.) #

  (print STDOUT "\n\n<<< Preliminary S16: Compute ", '@{$gblks_w_spec_lb0_sbwd} & @{$gblks_w_spec_rb0_sbwd}', ", via 'mk_lists_gblocks_w_spec_bounds'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_sbwd, $re_sbwd) = (0, $ct_clms0_in_sbwd-1);

  (print STDOUT 'Manually Set: ($le_sbwd, $re_sbwd) = ', "($le_sbwd, $re_sbwd) .\n\n") if ($TEST_LEVEL_MAIN>1);

	my ($gblks_w_spec_lb0_sbwd, $gblks_w_spec_rb0_sbwd) = mk_lists_gblocks_w_spec_bounds ($le_sbwd, $re_sbwd, @bds_gblocks0_in_sbwd);
	unless (defined $gblks_w_spec_rb0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
  (print STDOUT "\n --- Subroutine 'mk_lists_gblocks_w_spec_bounds' Ended on the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN>0);


  if ($TEST_LEVEL_MAIN >2) { # (Pre S16) #

	my $ofname = 'gblks_w_spec_bds0.txt';

	unless ( print_out_gblks_w_spec_bds ($sbdir_interim, $ofname, @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_gblks_w_spec_bds' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    ##print STDOUT "\n ... The subroutine 'print_out_gblks_w_spec_bds', Ended on the sub-window '$sbwd_ID' (w/ records in \"$maindir_interim/$ofname\") ...\n\n";
    print STDOUT "\n ... The subroutine 'print_out_gblks_w_spec_bds', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofname\") ...\n\n"; # CORRECTED on 2020/05/11.


  } # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre S16) # ...}"

	### End of "(Pre S16) Prepare @{$gblks_w_spec_lb0_sbwd} & @{$gblks_w_spec_rb0_sbwd}, via 'mk_lists_gblocks_w_spec_bounds'." ###


	### (Pre S17) Prepare @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd} & @{$cts_onx0_sbwd},', " via 'cts_vincluding_vincluded_onx_gblks'. ### (ADDED on 2019/07/17.) #

  (print STDOUT "\n\n", '<<< Preliminary S17: Compute @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd} & @{$cts_onx0_sbwd},', " via 'cts_vincluding_vincluded_onx_gblks'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($cts_vincluding0_sbwd, $cts_vincluded0_sbwd, $cts_onx0_sbwd) = cts_vincluding_vincluded_onx_gblks (@{$inter_block_relations0_sbwd});
	unless (defined $cts_onx0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
  (print STDOUT "\n --- Subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN>0);


  if ($TEST_LEVEL_MAIN >2) { # (Pre S17) #

	my $ofname = 'cts_vincluding_vincluded_onx_gblks0.txt';

	unless ( print_out_cts_vincluding_vincluded_onx_gblks ($sbdir_interim, $ofname, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks' FAILED for the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbdir_interim/$ofname\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre S17) # ...}"

	### End of "(Pre S17) Prepare @{$cts_vincluding}, @{$cts_vincluded} & @{$cts_onx},', " via 'cts_vincluding_vincluded_onx_gblks'." ###



	### (Pre-MAIN_sbwd): The question is whether or not we shuould re-set
	### %clm2ln_prb_sbst (=()) & @set_ingredients_indel_hsts (= ({}, [], [], [])).
	### For the moment, I will NOT reset them, because I consider it unnecessary.
	### However, the decision may change in the future.


	### Declare the Beginning of the Main Processes (in the sub-window). ###

  (print STDOUT "\n === NOW that the Preliminary Processes ENDED, Go Ahead to the MAIN Processes, where various moves + multiple-\"shift\"s are actually performed ===\n\n\n") if ($TEST_LEVEL_MAIN>0);


	# (Main II-S0) Compute the Log-Probabilities & some other quantities of the Initial MSA. 

   if (1) { # (Main II-S0) #

  (print STDOUT "\n<<< Main II-S0: Compute the Log-Probabilities & some other quantities of the Initial MSA. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

	my $session = 'MAIN II-S0';

	my @list_excluded_bin_hrels0 = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels0 = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);

			# (i) Print out specific Input Data. #

	my $subsbdir = "$sbdir_out/II_S0.Initial_MSA";
	my $subsbdir_in = "$subsbdir/In";

	my @ofnames = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.

	unless ( print_in_nr_mshifts_glks ($subsbdir_in, @ofnames, @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_sbwd, @rows_affd_by_blks0_in_sbwd, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @{$set_clms0_in_sbwd}, $beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2, @list_excluded_bin_hrels0) ) { # REVISED on 2019/07/18.
	##unless ( print_in_nr_mshifts_glks ($subsbdir_in, @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_sbwd, @rows_affd_by_blks0_in_sbwd, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @{$set_clms0_in_sbwd}, $beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2, @list_excluded_bin_hrels0) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_in_nr_mshifts_gblks' FAILED for $session on the input MSA of the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'print_in_nr_mshifts_gblks', Ended for $session on the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);

			# ADDED on 2019/07/17. (MII-S0) #

	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.

	unless ( bch_print_out_other_gblk_attributes ($subsbdir_in, @ofnames2, @{$inter_block_relations0_sbwd}, @{$gblock_sizes0_sbwd}, @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}) ) { # REVISED on 2019/07/18.
	##unless ( bch_print_out_other_gblk_attributes ($subsbdir_in, @{$inter_block_relations0_sbwd}, @{$gblock_sizes0_sbwd}, @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_other_gblk_attributes' FAILED for $session on the initial gap-blocks in the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'bch_print_out_other_gblk_attributes', Ended for $session on the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);

			# END of "ADDED on 2019/07/17. (MII-S0)" #


			# (ii) Perform 'compt_log_msa_prob_plus'. #

	my ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $grand_code_topology0)
	= compt_log_msa_prob_plus (
	    @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$set_clms0_in_sbwd}, @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_sbwd, @rows_affd_by_blks0_in_sbwd,
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 

            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa2, $if_on_re_msa2, # CORRECTED on 2019/08/14.
            ##@set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels0
		);
	unless ($if_success) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'compt_log_msa_prob_plus' FAILED for $session on the input MSA of the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'compt_log_msa_prob_plus', Ended for $session on the sub-window '$sbwd_ID' (w/ \$grand_code_topology0 = '$grand_code_topology0') ...\n\n") if ($TEST_LEVEL_MAIN >0);


			# (iii) Print out the Output Data. #

	my $subsbdir_out = "$subsbdir/Out";
	my $ofname_S0 = 'ln_prb_plus_msa.txt';

	unless ( print_out_ln_prb_plus_msa ($subsbdir_out, $ofname_S0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $STRING_UNDEF) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_ln_prb_plus_msa' FAILED for $session on the input MSA of the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT " ... Subroutine, 'print_out_ln_prb_plus_msa', Ended for $session on the sub-window '$sbwd_ID' ...\n\n\n") if ($TEST_LEVEL_MAIN >0);
   ##(print STDOUT " ... Subroutine, 'print_out_ln_prb_plus_msa', Ended for $session on the sub-window '$sbwd_ID' (, writing to \"$sbsbdir_out/$ofname_S0") ...\n\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/10.
   (print STDOUT " ... Subroutine, 'print_out_ln_prb_plus_msa', Ended for $session on the sub-window '$sbwd_ID' (, writing to \"$subsbdir_out/$ofname_S0\") ...\n\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/10. => CORRECTED on 2020/05/11.

	 $ln_prb_msa_init = $ln_prb_msa0; # ADDED in REVISION on 2020/04/21. #

   } # End of "if (1) { # (Main II-S0) # ...}"

	# End of "(Main II-S0) Compute the Log-Probabilities & some other quantities of the Initial MSA." 

				# End of "Added on 2019/07/04." #


    my $sbsubdir_tgt_cands = "$sbdir_out/Target_Candidates"; # Added on 2019/08/13.



	# (Main II-S4a) Perform "Split" (into blocks of the same-type) + Multiple-"Shift"s. #
	# (See SM-5 (iv-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

   (print STDOUT "\n<<< Main II-S4a: Perform \"Split\" (into blocks of the same-type) + Multiple-\"Shift\"s, via 'ssplit_gblock' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # ADDED on 2020/05/09. #

	my $sbsubdir_MII_S4a = "$sbdir_out/II_S4a.SSplit_plus_MltShifts"; # Added on 2019/08/13.

			# Added on 2019/08/11. (Main II-S4a) (1) #

	my $THRSH_DIST_PC = $HF_WDTH_SHIFT;
	my $UPPER_BOUND_NRSIZE2 = ($XLEVEL < 3) ? $MIN_WD_SIZE_PCAND : -1 ;

	my @INDICES_PC = my ($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC) = (0, 1, 3, 4);

   if ($TEST_LEVEL_MAIN>1) {
	print STDOUT "Manually Set:\n\n";
	print STDOUT '$THRSH_DIST_PC  = ', $THRSH_DIST_PC , " ;\n\n";
	print STDOUT '$UPPER_BOUND_NRSIZE2 = ', $UPPER_BOUND_NRSIZE2, " ;\n\n";

	print STDOUT '@INDICES_PC (=($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC))', "\n";
	print STDOUT '             = (', join (', ', @INDICES_PC), ") .\n\n";

   } # End of "if ($TEST_LEVEL_MAIN>1) {...}"

	my ($to_be_ssplit0_sbwd) = list_ssplit_candidates (@bds_gblocks0_in_sbwd, @{$gblock_sizes0_sbwd}, @{$inter_block_relations0_sbwd}, @rows_affd_by_blks0_in_sbwd, @{$list_mgd_pcands0_sbwd}, @INDICES_PC, $ct_seqs, $THRSH_DIST_PC, $UPPER_BOUND_NRSIZE2);
	unless (defined $to_be_ssplit0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_ssplit_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_to_be_ssplit0_sbwd = @{$to_be_ssplit0_sbwd};

   (print STDOUT "\n --- The Subroutine 'list_ssplit_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_ssplit}=$ct_to_be_ssplit0_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);
 
			# End of "Added on 2019/08/11. (Main II-S4a) (1)" #

			# Added on 2019/08/13. (Main II-S4a) (2) #

  if (1) { # (Main II-S4a) #

	my $ofile_name2 = 'to_be_ssplit.txt';
	my $label2 = "to be \"Split (same type)\"";
	my $label_sizes = 'sizes,of,new,right,gap-blocks';

	(print STDOUT "\n ... Record the outputs of 'list_ssplit_candidates', by performing 'print_out_split_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_split_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$to_be_ssplit0_sbwd}, $label2, $label_sizes) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{to_be_ssplit0_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_split_cands', Ended on ", '@{$to_be_ssplit0_sbwd}', " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S4a) # ...}"

			# End of "Added on 2019/08/13. (Main II-S4a) (2)" #

			# Added on 2019/08/14. (Main II-S4a) (3) #

    (print STDOUT "\n<< (ii) Perform the Subroutine, 'ssplit_gblock', on each of the $ct_to_be_ssplit0_sbwd gap-blocks to be \"split (same type)\". >>\n\n") if ($TEST_LEVEL_MAIN >0);


  for (my $i=0; $i < $ct_to_be_ssplit0_sbwd; $i++) { # Main Outer for-loop (over gap-blocks to be "split (same type)"). #

	my ($gb_sbj, $dist, $new_sizes_right, $indices_rlv_pcs, $indx_closest_rlv_pc) = @{$to_be_ssplit0_sbwd->[$i]};
	my $cnct_nrsizes = join (',', @{$new_sizes_right});

	my $session = 'MAIN II-S4a';

    (print STDOUT "\n< The $i th \"subject\" gap-block to be \"merge + split\": indx_gb_sbj = $gb_sbj ; (distance=$dist, ", '@{$new_sizes_right}', "=($cnct_nrsizes)) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbsbdir = "$sbsubdir_MII_S4a/Gblock_sbj$i";
	my $sbsbsbdir_info = "$sbsbsbdir/Info";

	## my $sbsubdir_MII_S4a = "$sbdir_out/II_S4a.SSplit_plus_MltShifts"; # Added on 2019/08/13.

	my $ofname_info1 = 'gblk_to_be_ssplit.txt';

	my @subjects = ($gb_sbj);

	unless ( print_out_info_mvd_sbj ($sbsbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @bds_gblocks0_in_sbwd, @{$gblock_sizes0_sbwd}, @rows_affd_by_blks0_in_sbwd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th \"subject\" gap-block to be \"split (same type)\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to \"$sbsbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09. #


		# (i) Actually "Split" the subject Gap-Block into same-type blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'ssplit_gblock' ...\n\n") if ($TEST_LEVEL_MAIN>0);

    foreach my $new_size_right (@{$new_sizes_right}) { # Middle foreach-loop (over $new_size_right). #

	(print STDOUT "\n{{ \$new_size_right = $new_size_right }}\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbsbdir_out = "$sbsbsbdir/Out/nrsize$new_size_right";
	my $sbsbsbdir_out0 = "$sbsbsbdir_out/Origin";
	my $sbsbsbdir_out_ms = "$sbsbsbdir_out/MShifts";

	my @curr_gb_coords_sbwd = @{$orgs_gb0_coords_sbwd};

	my ($flag_sspl, $new_gb_left, $new_gb_right, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)
	  = ssplit_gblock (
	     $gb_sbj, $new_size_right,
	     @{$set_clms0_in_sbwd}, @bds_gblocks0_in_sbwd, @curr_gb_coords_sbwd, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm},
	     @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @rows_affd_by_blks0_in_sbwd, $cnct_null_clm,
	     @{$inter_block_relations0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}, @{$gblock_sizes0_sbwd},  $HF_WDTH_SHIFT
		);
	unless ($flag_sspl) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'ssplit_gblock' FAILED on the $i th \"subject\" gap-block for $session (w/ \$new_size_right = $new_size_right) with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'ssplit_gblock' ENDED on the $i th \"subject\" gap-block (w/ \$new_size_right = $new_size_right) for $session with the sub-window '$sbwd_ID' (w/ \$flag_sspl = $flag_sspl, \$new_gb_left = $new_gb_left, \$new_gb_right = $new_gb_right) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_sbwd, $re_new_sbwd) = (0, $#{$new_curr_set_clms_in_sbwd});

	($new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks_in_sbwd}, $le_new_sbwd, $re_new_sbwd, @{$new_inter_block_relations_sbwd}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_sbwd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session with the sub-window 
$sbwd_ID' on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbsbdir_out4CMSS = "$sbsbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbsbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my @new_list_gblks_in_sbwd = ();
	my $ct_out_gbs = @{$new_curr_bds_gblks_in_sbwd};
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # Inner for-loop (over gap-blocks). #
	    my $ogb = $old_indices_new_gbs->[$ngb];
	    unless (defined $ogb) { $ogb = $gb_sbj; }
	    my @info_gb = @{$list_gblks0_in_sbwd->[$ogb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks_in_sbwd->[$ngb]};
	    $new_list_gblks_in_sbwd[$ngb] = \@info_gb;
	} # End of the Inner for-loop (over gap-blocks). #
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded (EXCEPT those specified by default).

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Split (same type)". #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbsbdir_info = "$sbsbsbdir_info/nrsize$new_size_right";

	my @ofnames0 = ('basic_info.txt', 'old_new_indices_gblks.txt', 'rec_split_sessions.txt', 'rec_undo_hovlps.txt');
	##my @ofnames0 = ('basic_info.txt', 'old_new_indices_gblks.txt', 'rec_ssplit.txt', 'rec_undo_hovlps.txt');

	# $flag_sspl, $new_gb_left, $new_gb_right, $new_inices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

			# ADDED in incorporating (MAIN II-S4b) (on 2019/08/20). #
	my $gb_sbj_NEW = $new_indices_old_gbs->[$gb_sbj];
	my $gb_cmpl_NEW = 'N.A.';
	my $med_indices_new_gbs = ['N.A'];
			# End of "ADDED in incorporating (MAIN II-S4b) (on 2019/08/20)." #

	my @split_sessions = (); # A dummy list.

	unless ( print_out_info_chngs_by_split ($sbsbsbsbdir_info, @ofnames0, $flag_sspl, $gb_sbj, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$med_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) { # MODIFIED in incorporating (MAIN II-S4b) (on 2019/08/20). #
	##unless ( print_out_info_chngs_by_split ($sbsbsbsbdir_info, @ofnames0, $flag_sspl, $gb_sbj, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_split' FAILED on the $i th \"subject\" gap-block to be \"ssplit\" for $session with the sub-window '$sbwd_ID' (w/ \$new_size_right = $new_size_right) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$new_size_right = $new_size_right) (, writing to several files under \"$sbsbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Ssplit". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords_sbwd = $new_curr_bl_coords_sbwd;
	my @new_list_gblks = @new_list_gblks_in_sbwd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords_sbwd} (=@{$new_curr_bl_coords_sbwd}) = (', join (', ', @{$new_orgs_gb_coords_sbwd}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_sbwd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	    ## $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    ## $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    ## $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbsbdir_out0, $sbsbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms_in_sbwd}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks_in_sbwd}, @{$rows_affd_by_new_blks_in_sbwd},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa2, $if_on_re_msa2,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations_sbwd}, @{$new_gblock_sizes_sbwd}, @{$new_curr_gblks_w_spec_lb_sbwd}, @{$new_curr_gblks_w_spec_rb_sbwd},
	    @{$new_cts_vincluding_sbwd}, @{$new_cts_vincluded_sbwd}, @{$new_cts_onx_sbwd}, $STRING_UNDEF, $beg_sbwd, $end_sbwd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA of the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"split\" into same-type blocks (w/ \$new_size_right = $new_size_right) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session with the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"ssplit\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN II-S4a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"ssplit\" (w/ \$new_size_right = $new_size_right): $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"ssplit\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"ssplit\" (w/ \$new_size_right = $new_size_right) : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"ssplit\" (w/ \$new_size_right = $new_size_right) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", "Sub($sbwd_ID)", $session_out, $sbsbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN II-S4a)" ##


    } # End of the Middle foreach-loop (over $new_size_right). #

  } # End of the Main Outer for-loop (over gap-blocks to be "split (same type)"). #

			# End of "Added on 2019/08/14. (Main II-S4a) (3)" #

	# End of "(Main II-S4a) Perform "Split" (into blocks of the same-type) + Multiple-"Shift"s." #


	# (Main II-S4b) Perform "Split" (into blocks of the complementary-types) + Multiple-"Shift"s. #
	# (See SM-5 (iv-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

   (print STDOUT "\n<<< Main II-S4b: Perform \"Split\" (into blocks of the complementary-types) + Multiple-\"Shift\"s, via 'csplit_gblock' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # ADDED on 2020/05/09. #

			# Added on 2019/08/20. (Main II-S4b) #

	my $sbsubdir_MII_S4b = "$sbdir_out/II_S4b.CSplit_plus_MltShifts";


	##my $THRSH_DIST_PC = $HF_WDTH_SHIFT;
	my $UPPER_BOUND_NCSIZE = ($XLEVEL < 3) ? $MIN_WD_SIZE_PCAND : -1 ;

	##my @INDICES_PC = my ($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC) = (0, 1, 3, 4);

   if ($TEST_LEVEL_MAIN>1) {
	print STDOUT "Manually Set:\n\n";
	##print STDOUT '$THRSH_DIST_PC  = ', $THRSH_DIST_PC , " ;\n\n";
	print STDOUT '$UPPER_BOUND_NCSIZE = ', $UPPER_BOUND_NCSIZE, " ;\n\n";

	##print STDOUT '@INDICES_PC (=($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC))', "\n";
	##print STDOUT '             = (', join (', ', @INDICES_PC), ") .\n\n";

   } # End of "if ($TEST_LEVEL_MAIN>1) {...}"


	my ($to_be_csplit0_sbwd)
	    = list_csplit_candidates (@bds_gblocks0_in_sbwd,
		@{$orgs_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, $le_sbwd, $re_sbwd, 
		@{$gblock_sizes0_sbwd}, @{$inter_block_relations0_sbwd}, @rows_affd_by_blks0_in_sbwd, @{$list_mgd_pcands0_sbwd}, @INDICES_PC, $ct_seqs, $THRSH_DIST_PC, $UPPER_BOUND_NCSIZE);
	unless (defined $to_be_csplit0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_csplit_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_to_be_csplit0_sbwd = @{$to_be_csplit0_sbwd};

   (print STDOUT "\n --- The Subroutine 'list_csplit_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_csplit}=$ct_to_be_csplit0_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);

  if (1) { # (Main II-S4b) # (1) #

	my $ofile_name2 = 'to_be_csplit_Qual.txt';
	my $label2 = "to be \"Split (complementary type)\" (Qualified)";
	my $label_sizes = 'sizes,of,new,complementary,gap-blocks';

	(print STDOUT "\n ... Record the outputs of 'list_csplit_candidates', by performing 'print_out_split_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_split_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$to_be_csplit0_sbwd}, $label2, $label_sizes) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{to_be_csplit0_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_split_cands', Ended on ", '@{$to_be_csplit0_sbwd}', " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S4b) # (1) # ...}"


			## ADDED Specifically for (Main II-S4b) (on 2019/08/20). ##

	my ($list_nn_csplit_cands_sbwd) = list_nn_csplit_candidates (@{$to_be_csplit0_sbwd});
	unless (defined $list_nn_csplit_cands_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_csplit_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_nn_cscands_sbwd = @{$list_nn_csplit_cands_sbwd};
 
   (print STDOUT "\n --- The Subroutine 'list_nn_csplit_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_csplit}_nearest_neighboring=$ct_nn_cscands_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);

  if (1) { # (Main II-S4b) # (2) #

	my $ofile_name2 = 'to_be_csplit_NN.txt';
	my $label2 = "to be \"Split (complementary type)\" (Nearest-Neighboring)";
	my $label_sizes = 'sizes,of,new,complementary,gap-blocks';

	(print STDOUT "\n ... Record the outputs of 'list_nn_csplit_candidates', by performing 'print_out_split_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_split_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$list_nn_csplit_cands_sbwd}, $label2, $label_sizes) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{$list_nn_csplit_cands_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_split_cands', Ended on ", '@{$list_nn_csplit_cands_sbwd}' , " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S4b) # (2) # ...}"

			## END of "ADDED Specifically for (Main II-S4b) (on 2019/08/20)." ##


    (print STDOUT "\n<< (ii) Perform the Subroutine, 'csplit_gblock', on each of the $ct_nn_cscands_sbwd \"nearest-neighboring\" gap-blocks to be \"split (complementary types)\". >>\n\n") if ($TEST_LEVEL_MAIN >0);


  for (my $i=0; $i < $ct_nn_cscands_sbwd; $i++) { # Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "split (complementary types)"). #

	my $info_cand = $list_nn_csplit_cands_sbwd->[$i];
	my ($gb_sbj, $dist, $new_sizes_cmpl, $indices_rlv_pcs, $indx_closest_rlv_pc, $pot_avail_sites) = @{$info_cand};
	my $cnct_ncsizes = join (',', @{$new_sizes_cmpl});

	my $session = 'MAIN II-S4b';

    (print STDOUT "\n< The $i th \"subject\" gap-block to be \"split (complementary types)\": indx_gb_sbj = $gb_sbj ; (distance=$dist, ", '@{$new_sizes_cmpl}', "=($cnct_ncsizes)) >\n\n") if ($TEST_LEVEL_MAIN >0);

	my $sbsbsbdir = "$sbsubdir_MII_S4b/Gblock_sbj$i";
	my $sbsbsbdir_info = "$sbsbsbdir/Info";

	## my $sbsubdir_MII_S4b = "$sbdir_out/II_S4b.CSplit_plus_MltShifts";

	my $ofname_info1 = 'gblk_to_be_csplit.txt';

	my @subjects = ($gb_sbj);

	unless ( print_out_info_mvd_sbj ($sbsbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @bds_gblocks0_in_sbwd, @{$gblock_sizes0_sbwd}, @rows_affd_by_blks0_in_sbwd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th \"subject\" gap-block to be \"split (complementary types)\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to \"$sbsbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09. #


		# (i) Actually "Split" the subject Gap-Block into complementary-type blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'csplit_gblock' ...\n\n") if ($TEST_LEVEL_MAIN>0);

    foreach my $new_size_cmpl (@{$new_sizes_cmpl}) { # Middle foreach-loop (over $new_size_cmpl). #

	(print STDOUT "\n{{ \$new_size_cmpl = $new_size_cmpl }}\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbsbdir_out = "$sbsbsbdir/Out/ncsize$new_size_cmpl";
	my $sbsbsbdir_out0 = "$sbsbsbdir_out/Origin";
	my $sbsbsbdir_out_ms = "$sbsbsbdir_out/MShifts";

	my @curr_gb_coords_sbwd = @{$orgs_gb0_coords_sbwd};

	my ($flag_cspl, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, $new_indices_old_gbs, $old_indices_new_gbs, $records_sessions, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd,
	    $new_list_gblocks_in_sbwd, $med_indices_new_gbs)
	  = csplit_gblock (
	     $gb_sbj, $new_size_cmpl, @{$pot_avail_sites}, $CHAR_ABS,
	     @{$set_clms0_in_sbwd}, @bds_gblocks0_in_sbwd, @curr_gb_coords_sbwd, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm},
	     @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @rows_affd_by_blks0_in_sbwd, $cnct_null_clm,
	     @{$inter_block_relations0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}, @{$gblock_sizes0_sbwd},  $HF_WDTH_SHIFT,
	     @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}
		);
	unless ($flag_cspl) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'csplit_gblock' FAILED on the $i th \"subject\" gap-block for $session (w/ \$new_size_cmpl = $new_size_cmpl) with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'csplit_gblock' ENDED on the $i th \"subject\" gap-block (w/ \$new_size_cmpl = $new_size_cmpl) for $session with the sub-window '$sbwd_ID' (w/ \$flag_cspl = $flag_cspl, \$gb_sbj_NEW = $gb_sbj_NEW, \$gb_cmpl_NEW = $gb_cmpl_NEW, \$new_gb_left = $new_gb_left, \$new_gb_right = $new_gb_right) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_sbwd, $re_new_sbwd) = (0, $#{$new_curr_set_clms_in_sbwd});

	($new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks_in_sbwd}, $le_new_sbwd, $re_new_sbwd, @{$new_inter_block_relations_sbwd}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_sbwd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session with the sub-window 
$sbwd_ID' on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbsbdir_out4CMSS = "$sbsbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbsbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my $ct_out_gbs = @{$new_curr_bds_gblks_in_sbwd};

	my @new_list_gblks_in_sbwd = @{$new_list_gblocks_in_sbwd};
=For (Main II-S4a)
	my @new_list_gblks_in_sbwd = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # Inner for-loop (over gap-blocks). #
	    my $ogb = $old_indices_new_gbs->[$ngb];
	    unless (defined $ogb) { $ogb = $gb_sbj; }
	    my @info_gb = @{$list_gblks0_in_sbwd->[$ogb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks_in_sbwd->[$ngb]};
	    $new_list_gblks_in_sbwd[$ngb] = \@info_gb;
	} # End of the Inner for-loop (over gap-blocks). #
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];
=cut # END of "For (Main II-S4a)"


	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded (EXCEPT those specified by default).

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Split (complementary types)". #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbsbdir_info = "$sbsbsbdir_info/ncsize$new_size_cmpl";

	my @ofnames0 = ('basic_info.txt', 'old_new_indices_gblks.txt', 'rec_split_sessions.txt', 'rec_undo_hovlps.txt');

	# $flag_sspl, $new_gb_left, $new_gb_right, $new_inices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	my @split_sessions = @{$records_sessions};
	##my @split_sessions = (); # A dummy list.

	unless ( print_out_info_chngs_by_split ($sbsbsbsbdir_info, @ofnames0, $flag_cspl, $gb_sbj, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$med_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) { # MODIFIED on 2019/08/20.
	## unless ( print_out_info_chngs_by_split ($sbsbsbsbdir_info, @ofnames0, $flag_sspl, $gb_sbj, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) { 
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_split' FAILED on the $i th \"subject\" gap-block to be \"csplit\" for $session with the sub-window '$sbwd_ID' (w/ \$new_size_cmpl = $new_size_cmpl) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$new_size_cmpl = $new_size_cmpl) ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$new_size_cmpl = $new_size_cmpl) (, writing to several files under \"$sbsbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Split (complementary types)". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords_sbwd = $new_curr_bl_coords_sbwd;
	my @new_list_gblks = @new_list_gblks_in_sbwd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords_sbwd} (=@{$new_curr_bl_coords_sbwd}) = (', join (', ', @{$new_orgs_gb_coords_sbwd}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_sbwd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	    ## $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    ## $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    ## $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbsbdir_out0, $sbsbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms_in_sbwd}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks_in_sbwd}, @{$rows_affd_by_new_blks_in_sbwd},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa2, $if_on_re_msa2,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations_sbwd}, @{$new_gblock_sizes_sbwd}, @{$new_curr_gblks_w_spec_lb_sbwd}, @{$new_curr_gblks_w_spec_rb_sbwd},
	    @{$new_cts_vincluding_sbwd}, @{$new_cts_vincluded_sbwd}, @{$new_cts_onx_sbwd}, $STRING_UNDEF, $beg_sbwd, $end_sbwd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA of the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"split\" into complementary-type blocks (w/ \$new_size_cmpl = $new_size_cmpl) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session with the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"csplit\" (w/ \$new_size_cmpl = $new_size_cmpl) ...\n\n") if ($TEST_LEVEL_MAIN >0);


			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN II-S4b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"csplit\" (w/ \$new_size_cmpl = $new_size_cmpl): $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"csplit\" (w/ \$new_size_cmpl = $new_size_cmpl) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"ssplit\" (w/ \$new_size_cmpl = $new_size_cmpl) : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"csplit\" (w/ \$new_size_cmpl = $new_size_cmpl) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", "Sub($sbwd_ID)", $session_out, $sbsbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN II-S4b)" ##


    } # End of the Middle foreach-loop (over $new_size_cmpl). #

  } # End of the Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "split (complementary types)"). #

			# End of "Added on 2019/08/20. (Main II-S4b)" #


	# End of "(Main II-S4b) Perform "Split" (into blocks of the complementary-types) + Multiple-"Shift"s." #



	# (Main II-S7a) Perform "Vertical-Split" (into sibling blocks) + Multiple-"Shift"s. #
	# (See SM-5 (vii-a) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

   (print STDOUT "\n<<< Main II-S7a: Perform \"Vertical-Split\" (into sibling blocks) + Multiple-\"Shift\"s, via 'vsplit_gblock' & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # ADDED on 2020/05/09. #

			# Added on 2019/08/24. (Main II-S7a) #

	my $sbsubdir_MII_S7a = "$sbdir_out/II_S7a.VSplit_plus_MltShifts";


	##my $THRSH_DIST_PC = $HF_WDTH_SHIFT;
	##my @INDICES_PC = my ($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC) = (0, 1, 3, 4);

   if ($TEST_LEVEL_MAIN>1) {
	##print STDOUT "Manually Set:\n\n";
	##print STDOUT '$THRSH_DIST_PC  = ', $THRSH_DIST_PC , " ;\n\n";
	##print STDOUT '@INDICES_PC (=($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC))', "\n";
	##print STDOUT '             = (', join (', ', @INDICES_PC), ") .\n\n";

   } # End of "if ($TEST_LEVEL_MAIN>1) {...}"


	my ($to_be_vsplit0_sbwd)
	    = list_vsplit_candidates (
		@{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, 
		%node2pa, %node2ch, $topnode_id, %{$br2eq}, @set_exnodes,
		%{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw},
		@bds_gblocks0_in_sbwd, @{$inter_block_relations0_sbwd}, 
		@{$list_mgd_pcands0_sbwd}, @INDICES_PC, $ct_seqs, $THRSH_DIST_PC);
	unless (defined $to_be_vsplit0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_vsplit_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_to_be_vsplit0_sbwd = @{$to_be_vsplit0_sbwd};

   (print STDOUT "\n --- The Subroutine 'list_vsplit_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_vsplit}=$ct_to_be_vsplit0_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);


  if (1) { # (Main II-S7a) # (1) #

	my $ofile_name2 = 'to_be_vsplit_Qual.txt';
	my $label2 = "to be \"Vertically Split\" into Effective-Child-Blocks (Qualified)";
	my $suffix_new_sbj = 'eff_ch';


	(print STDOUT "\n ... Record the outputs of 'list_vsplit_candidates', by performing 'print_out_vsplit_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_vsplit_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$to_be_vsplit0_sbwd}, $label2, $suffix_new_sbj) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{to_be_vsplit0_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_vsplit_cands', Ended on ", '@{$to_be_vsplit0_sbwd}', " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S7a) # (1) # ...}"


	my ($indx_dist, $indx_indx_clsst_pc) = (1, 3);
   (print STDOUT 'Manually Set: ($indx_dist, $indx_indx_clsst_pc) = ', "($indx_dist, $indx_indx_clsst_pc) .\n\n") if ($TEST_LEVEL_MAIN >3);

	my ($list_nn_vsplit_cands_sbwd) = list_nn_gen_split_candidates (@{$to_be_vsplit0_sbwd}, $indx_dist, $indx_indx_clsst_pc);
	unless (defined $list_nn_vsplit_cands_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_gen_split_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_nn_vscands_sbwd = @{$list_nn_vsplit_cands_sbwd};
 
   (print STDOUT "\n --- The Subroutine 'list_nn_gen_split_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_vsplit}_nearest_neighboring=$ct_nn_vscands_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);

  if (1) { # (Main II-S7a) # (2) #

	my $ofile_name2 = 'to_be_vsplit_NN.txt';
	my $label2 = "to be \"Vertically Split\" into Effective-Child-Blocks (Nearest-Neighboring)";
	my $suffix_new_sbj = 'eff_ch';

	(print STDOUT "\n ... Record the outputs of 'list_nn_gen_split_candidates', by performing 'print_out_vsplit_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_vsplit_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$list_nn_vsplit_cands_sbwd}, $label2, $suffix_new_sbj) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{$list_nn_vsplit_cands_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_vsplit_cands', Ended on ", '@{$list_nn_vsplit_cands_sbwd}' , " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S7a) # (2) # ...}"


    (print STDOUT "\n<< (ii) Perform the Subroutine, 'vsplit_gblock', on each of the $ct_nn_vscands_sbwd \"nearest-neighboring\" gap-blocks to be \"vertically split\" into effective child blocks. >>\n\n") if ($TEST_LEVEL_MAIN >0);


  for (my $i=0; $i < $ct_nn_vscands_sbwd; $i++) { # Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "vertically split" into effective child blocks). #

	my $info_cand = $list_nn_vsplit_cands_sbwd->[$i];
	my ($gb_sbj, $dist, $indices_rlv_pcs, $indx_closest_rlv_pc, $info_effch1, $info_effch2) = @{$info_cand};
	my ($br1, $up_or_lw1) = @{$info_effch1}[0,1];
	my ($br2, $up_or_lw2) = @{$info_effch2}[0,1];

	my $session = 'MAIN II-S7a';

    (print STDOUT "\n< The $i th \"subject\" gap-block to be \"vertically split\" into effective child blocks: indx_gb_sbj = $gb_sbj ; (distance=$dist, {closest purge-candidate}=$indx_closest_rlv_pc, [br, up_or_lw]_effch1 = [$br1, $up_or_lw1], [br, up_or_lw]_effch2 = [$br2, $up_or_lw2]) >\n\n") if ($TEST_LEVEL_MAIN >0);


	my $sbsbsbdir = "$sbsubdir_MII_S7a/Gblock_sbj$i";
	my $sbsbsbdir_info = "$sbsbsbdir/Info";

	## my $sbsubdir_MII_S7a = "$sbdir_out/II_S7a.VSplit_plus_MltShifts";

	my $ofname_info1 = 'gblk_to_be_vsplit.txt';

	my @subjects = ($gb_sbj);

	unless ( print_out_info_mvd_sbj ($sbsbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @bds_gblocks0_in_sbwd, @{$gblock_sizes0_sbwd}, @rows_affd_by_blks0_in_sbwd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th \"subject\" gap-block to be \"vertically split (info effective child-blocks)\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to \"$sbsbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09. #


		# (i) Actually "Vertically Split" the subject Gap-Block into effective child blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'vsplit_gblock' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbsbdir_out = "$sbsbsbdir/Out";
	my $sbsbsbdir_out0 = "$sbsbsbdir_out/Origin";
	my $sbsbsbdir_out_ms = "$sbsbsbdir_out/MShifts";

	my @curr_gb_coords_sbwd = @{$orgs_gb0_coords_sbwd};

	my ($flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd,
	    $new_list_gblocks_in_sbwd, $new_list_excluded_bin_hrels)
	  = vsplit_gblock (
	     $gb_sbj, @{$info_effch1}, @{$info_effch2},
	     @{$set_clms0_in_sbwd}, @bds_gblocks0_in_sbwd, @curr_gb_coords_sbwd, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm},
	     @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @rows_affd_by_blks0_in_sbwd, $cnct_null_clm,
	     @{$inter_block_relations0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}, @{$gblock_sizes0_sbwd},  $HF_WDTH_SHIFT,
	     @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}
		);
	unless ($flag_vspl) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'vsplit_gblock' FAILED on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'vsplit_gblock' ENDED on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$flag_vspl = $flag_vspl, \$gb_ch1_NEW = $gb_ch1_NEW, \$gb_ch2_NEW = $gb_ch2_NEW, \$gb_shftd = $gb_shftd, \$incr_coord_shftd = $incr_coord_shftd) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_sbwd, $re_new_sbwd) = (0, $#{$new_curr_set_clms_in_sbwd});

	($new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks_in_sbwd}, $le_new_sbwd, $re_new_sbwd, @{$new_inter_block_relations_sbwd}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_sbwd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session with the sub-window 
$sbwd_ID' on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbsbdir_out4CMSS = "$sbsbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbsbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my $ct_out_gbs = @{$new_curr_bds_gblks_in_sbwd};

	my @new_list_gblks_in_sbwd = @{$new_list_gblocks_in_sbwd};
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];

	my %gb2incr_coord_1way_dsplc = ('N.A' => 1); # A dummy hash. (ADDED in revision on 2019/08/30.)
  (print STDOUT 'Manually Set: %gb2incr_coord_1way_dsplc = ', "('N.A' => 1) .\n\n") if ($TEST_LEVEL_MAIN>1);


			# Use an output of 'vsplit_gblock'. #
	my @list_excluded_bin_hrels = @{$new_list_excluded_bin_hrels};

  (print STDOUT 'Set  @list_excluded_bin_hrels according to an output of ', "'vsplit_gblock' .\n\n") if ($TEST_LEVEL_MAIN>1);
  ## (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Vertical Split" (into effective child blocks). #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbsbdir_info = "$sbsbsbdir_info/Out";

	my @ofnames0 = ('basic_info.txt', 'info_1way_displacements.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt', 'rec_undo_hovlps.txt'); # MODIFIED in revision (on 2019/08/30). #
	##my @ofnames0 = ('basic_info.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt', 'rec_undo_hovlps.txt');

	# $flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	unless ( print_out_info_chngs_by_vsplit ($sbsbsbsbdir_info, @ofnames0, $flag_vspl, $gb_sbj, $gb_ch1_NEW, $gb_ch2_NEW, %gb2incr_coord_1way_dsplc, $gb_shftd, $incr_coord_shftd, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, %{$gb2incr_bl_coord_undo}) ) { # REVISED on 2019/08/30.
	##unless ( print_out_info_chngs_by_vsplit ($sbsbsbsbdir_info, @ofnames0, $flag_vspl, $gb_sbj, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, %{$gb2incr_bl_coord_undo}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_vsplit' FAILED on the $i th \"subject\" gap-block to be \"vsplit\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to several files under \"$sbsbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Vertical Split (into effective child gap-blocks)".
		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords_sbwd = $new_curr_bl_coords_sbwd;
	my @new_list_gblks = @new_list_gblks_in_sbwd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords_sbwd} (=@{$new_curr_bl_coords_sbwd}) = (', join (', ', @{$new_orgs_gb_coords_sbwd}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_sbwd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	    ## $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    ## $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    ## $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbsbdir_out0, $sbsbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms_in_sbwd}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks_in_sbwd}, @{$rows_affd_by_new_blks_in_sbwd},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa2, $if_on_re_msa2,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations_sbwd}, @{$new_gblock_sizes_sbwd}, @{$new_curr_gblks_w_spec_lb_sbwd}, @{$new_curr_gblks_w_spec_rb_sbwd},
	    @{$new_cts_vincluding_sbwd}, @{$new_cts_vincluded_sbwd}, @{$new_cts_onx_sbwd}, $STRING_UNDEF, $beg_sbwd, $end_sbwd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA of the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"vertically split\" into effective child blocks !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session with the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"vsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN II-S7a) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"vsplit\" : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"vsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"vsplit\" : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"vsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", "Sub($sbwd_ID)", $session_out, $sbsbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN II-S7a)" ##


  } # End of the Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "vertically split" into effective child blocks). #

			# End of "Added on 2019/08/24. (Main II-S7a)" #

	# End of "(Main II-S7a) Perform "Vertical-Split" (into sibling blocks) + Multiple-"Shift"s." #


	# (Main II-S7b) Perform "Vertical-Split" (into COMPLEMENTARY-sibling blocks) + Multiple-"Shift"s. #
	# (See SM-5 (vii-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

   (print STDOUT "\n<<< Main II-S7b: Perform \"Vertical-Split\" (into COMPLEMENTARY-sibling blocks) + Multiple-\"Shift\"s, via 'cvsplit_gblock'  & 'nonrec_mlt_shifts_gapblocks', etc. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0); # ADDED on 2020/05/09. #


			# Added on 2019/08/30. (Main II-S7b) #

	my $sbsubdir_MII_S7b = "$sbdir_out/II_S7b.CVSplit_plus_MltShifts";

	##my $THRSH_DIST_PC = $HF_WDTH_SHIFT;
	##my @INDICES_PC = my ($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC) = (0, 1, 3, 4);

   if ($TEST_LEVEL_MAIN>1) {
	##print STDOUT "Manually Set:\n\n";
	##print STDOUT '$THRSH_DIST_PC  = ', $THRSH_DIST_PC , " ;\n\n";
	##print STDOUT '@INDICES_PC (=($INDX_BEG_PC, $INDX_END_PC, $INDX_ROWS_UNDER_BR_PC, $INDX_CNSTTNT_SITES_PC))', "\n";
	##print STDOUT '             = (', join (', ', @INDICES_PC), ") .\n\n";

   } # End of "if ($TEST_LEVEL_MAIN>1) {...}"


	my ($to_be_cvsplit0_sbwd)
	    = list_cvsplit_candidates (
		@{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, 
		%node2pa, %node2ch, $topnode_id, %{$br2eq}, @set_exnodes,
		%{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw},
		@bds_gblocks0_in_sbwd, @{$inter_block_relations0_sbwd}, 
		@{$list_mgd_pcands0_sbwd}, @INDICES_PC, $ct_seqs, $THRSH_DIST_PC);
	unless (defined $to_be_cvsplit0_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_cvsplit_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_to_be_cvsplit0_sbwd = @{$to_be_cvsplit0_sbwd};

   (print STDOUT "\n --- The Subroutine 'list_cvsplit_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_cvsplit}=$ct_to_be_cvsplit0_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);


  if (1) { # (Main II-S7b) # (1) #

	my $ofile_name2 = 'to_be_cvsplit_Qual.txt';
	my $label2 = "to be \"Vertically Split\" into COMPLEMENTARY-Effective Child-Blocks (Qualified)";
	my $suffix_new_sbj = 'ceff_ch';


	(print STDOUT "\n ... Record the outputs of 'list_cvsplit_candidates', by performing 'print_out_vsplit_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	unless ( print_out_vsplit_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$to_be_cvsplit0_sbwd}, $label2, $suffix_new_sbj) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_split_cands' FAILED on ", '@{to_be_cvsplit0_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_vsplit_cands', Ended on ", '@{$to_be_cvsplit0_sbwd}', " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S7b) # (1) # ...}"


	my ($indx_dist2, $indx_indx_clsst_pc2) = (1, 3);
   (print STDOUT 'Manually Set: ($indx_dist2, $indx_indx_clsst_pc2) = ', "($indx_dist2, $indx_indx_clsst_pc2) .\n\n") if ($TEST_LEVEL_MAIN >3);

	my ($list_nn_cvsplit_cands_sbwd) = list_nn_gen_split_candidates (@{$to_be_cvsplit0_sbwd}, $indx_dist2, $indx_indx_clsst_pc2);
	unless (defined $list_nn_cvsplit_cands_sbwd) {
	    print STDOUT "[ERROR '$plscript_name'] The Subroutine 'list_nn_gen_split_candidates' FAILED on the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
	my $ct_nn_cvscands_sbwd = @{$list_nn_cvsplit_cands_sbwd};
 
   (print STDOUT "\n --- The Subroutine 'list_nn_gen_split_candidates' ENDED on the sub-window '$sbwd_ID' (w/ #{to_be_cvsplit}_nearest_neighboring=$ct_nn_cvscands_sbwd) ...\n\n") if ($TEST_LEVEL_MAIN>0);

  if (1) { # (Main II-S7b) # (2) #

	my $ofile_name2 = 'to_be_cvsplit_NN.txt';
	my $label2 = "to be \"Vertically Split\" into COMPLEMENTARY-Effective-Child-Blocks (Nearest-Neighboring)";
	my $suffix_new_sbj = 'ceff_ch';

	(print STDOUT "\n ... Record the outputs of 'list_nn_gen_split_candidates', by performing 'print_out_vsplit_cands' ...\n\n") if ($TEST_LEVEL_MAIN>1);

	##unless ( print_out_vsplit_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$list_nn_vsplit_cands_sbwd}, $label2, $suffix_new_sbj) ) {
	unless ( print_out_vsplit_cands ($sbsubdir_tgt_cands, $ofile_name2, @{$list_nn_cvsplit_cands_sbwd}, $label2, $suffix_new_sbj) ) { # CORRECTED on 2020/05/16.
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_vsplit_cands' FAILED on ", '@{$list_nn_vsplit_cands_sbwd}', " in the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_vsplit_cands', Ended on ", '@{$list_nn_vsplit_cands_sbwd}' , " in the sub-window '$sbwd_ID' (w/ records in \"$sbsubdir_tgt_cands/$ofile_name2\") ...\n\n\n") if ($TEST_LEVEL_MAIN>1);

  } # End of "if (1) { # (Main II-S7b) # (2) # ...}"


    (print STDOUT "\n<< (ii) Perform the Subroutine, 'cvsplit_gblock', on each of the $ct_nn_cvscands_sbwd \"nearest-neighboring\" gap-blocks to be \"vertically split\" into COMPLEMENTARY-effective child blocks. >>\n\n") if ($TEST_LEVEL_MAIN >0);


  for (my $i=0; $i < $ct_nn_cvscands_sbwd; $i++) { # Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "vertically split" into COMPLEMENTARY-effective child blocks). #

	my $info_cand = $list_nn_cvsplit_cands_sbwd->[$i];
	my ($gb_sbj, $dist, $indices_rlv_pcs, $indx_closest_rlv_pc, $info_ceffch1, $info_ceffch2) = @{$info_cand};
	my ($br1, $up_or_lw1) = @{$info_ceffch1}[0,1];
	my ($br2, $up_or_lw2) = @{$info_ceffch2}[0,1];

	my $session = 'MAIN II-S7b';

    (print STDOUT "\n< The $i th \"subject\" gap-block to be \"vertically split\" into COMPLEMENTARY-effective child blocks: indx_gb_sbj = $gb_sbj ; (distance=$dist, {closest purge-candidate}=$indx_closest_rlv_pc, [br, up_or_lw]_ceffch1 = [$br1, $up_or_lw1], [br, up_or_lw]_ceffch2 = [$br2, $up_or_lw2]) >\n\n") if ($TEST_LEVEL_MAIN >0);


	my $sbsbsbdir = "$sbsubdir_MII_S7b/Gblock_sbj$i";
	my $sbsbsbdir_info = "$sbsbsbdir/Info";

	## my $sbsubdir_MII_S7b = "$sbdir_out/II_S7b.CVSplit_plus_MltShifts";

	my $ofname_info1 = 'gblk_to_be_vsplit.txt';

	my @subjects = ($gb_sbj);

	unless ( print_out_info_mvd_sbj ($sbsbsbdir_info, $ofname_info1, @subjects, @{$orgs_gb0_coords_sbwd}, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm}, @bds_gblocks0_in_sbwd, @{$gblock_sizes0_sbwd}, @rows_affd_by_blks0_in_sbwd) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_mvd_sbj' FAILED on the $i th \"subject\" gap-block to be \"vertically split (info COMPLEMENTARY-effective child-blocks)\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_mvd_sbj', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to \"$sbsbsbdir_info/$ofname_info1\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09. #


		# (i) Actually "Vertically Split" the subject Gap-Block into COMPLEMENTARY-effective child blocks. #

   (print STDOUT " --- Now, Perform the MAIN Subroutine, 'cvsplit_gblock' ...\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbsbdir_out = "$sbsbsbdir/Out";
	my $sbsbsbdir_out0 = "$sbsbsbdir_out/Origin";
	my $sbsbsbdir_out_ms = "$sbsbsbdir_out/MShifts";


	my @curr_gb_coords_sbwd = @{$orgs_gb0_coords_sbwd};

	my ($flag_cvspl, $gb_cch1_NEW, $gb_cch2_NEW, $gb2incr_coord_1way_dsplc, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd,
	    $new_list_gblocks_in_sbwd)
	    ##$new_list_gblocks_in_sbwd, $new_list_excluded_bin_hrels)
	  = cvsplit_gblock (
	     $gb_sbj, @{$info_ceffch1}, @{$info_ceffch2}, $CHAR_ABS, 
	     @{$set_clms0_in_sbwd}, @bds_gblocks0_in_sbwd, @curr_gb_coords_sbwd, @{$bds_gb0_coords_sbwd}, @{$gb0_coords_sbwd_frm},
	     @{$gblks_w_spec_lb0_sbwd}, @{$gblks_w_spec_rb0_sbwd}, @rows_affd_by_blks0_in_sbwd, $cnct_null_clm,
	     @{$inter_block_relations0_sbwd}, @{$cts_vincluding0_sbwd}, @{$cts_vincluded0_sbwd}, @{$cts_onx0_sbwd}, @{$gblock_sizes0_sbwd},  $HF_WDTH_SHIFT,
	     @{$list_gblks0_in_sbwd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}
		);
	unless ($flag_cvspl) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'cvsplit_gblock' FAILED on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'cvsplit_gblock' ENDED on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (w/ \$flag_cvspl = $flag_cvspl, \$gb_cch1_NEW = $gb_cch1_NEW, \$gb_cch2_NEW = $gb_cch2_NEW, \$gb_shftd = 'N.A.', \$incr_coord_shftd = 'N.A') ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# ADDED in REVISION on 2019/10/19. #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_sbwd, $re_new_sbwd) = (0, $#{$new_curr_set_clms_in_sbwd});

	($new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks_in_sbwd}, $le_new_sbwd, $re_new_sbwd, @{$new_inter_block_relations_sbwd}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_sbwd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED for $session with the sub-window 
$sbwd_ID' on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbsbdir_out4CMSS = "$sbsbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the sub-window '$sbwd_ID' !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the sub-window '$sbwd_ID' (w/ records in \"$sbsbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    }
		# END of "ADDED in REVISION on 2019/10/19." #


		# (i') Prepare Some Additional Data-Sets. #

	my $ct_out_gbs = @{$new_curr_bds_gblks_in_sbwd};

	my @new_list_gblks_in_sbwd = @{$new_list_gblocks_in_sbwd};
	##my @new_list_gblks_in_wd = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


	my ($gb_shftd, $incr_coord_shftd) = ('N.A.', 'N.A.');
  (print STDOUT 'Manually Set: ($gb_shftd, $incr_coord_shftd) = ', "($gb_shftd, $incr_coord_shftd) .\n\n") if ($TEST_LEVEL_MAIN>1);

			# Use an output of 'vsplit_gblock'. #
	my %gb2incr_coord_1way_dsplc = %{$gb2incr_coord_1way_dsplc};
  (print STDOUT 'Set %gb2incr_coord_1way_dsplc  according to an output of ', "'cvsplit_gblock' .\n\n") if ($TEST_LEVEL_MAIN>1);


	my @list_excluded_bin_hrels = ();

  (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Vertical Split" (into COMPLEMENTARY-effective child blocks). #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbsbdir_info = "$sbsbsbdir_info/Out";

	my @ofnames0 = ('basic_info.txt', 'info_1way_displacements.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt', 'rec_undo_hovlps.txt'); # MODIFIED in revision (on 2019/08/30). #
	##my @ofnames0 = ('basic_info.txt', 'info_serial_shifts.txt', 'old_new_indices_gblks.txt', 'rec_undo_hovlps.txt');

	# $flag_vspl, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, $new_indices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	unless ( print_out_info_chngs_by_vsplit ($sbsbsbsbdir_info, @ofnames0, $flag_cvspl, $gb_sbj, $gb_cch1_NEW, $gb_cch2_NEW, %gb2incr_coord_1way_dsplc, $gb_shftd, $incr_coord_shftd, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, %{$gb2incr_bl_coord_undo}) ) { # REVISED on 2019/08/30.
	##unless ( print_out_info_chngs_by_vsplit ($sbsbsbsbdir_info, @ofnames0, $flag_vspl, $gb_sbj, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, %{$gb2incr_bl_coord_undo}) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_vsplit' FAILED on the $i th \"subject\" gap-block to be \"cvsplit\" for $session with the sub-window '$sbwd_ID' !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_vsplit', Ended on the $i th \"subject\" gap-block for $session with the sub-window '$sbwd_ID' (, writing to several files under \"$sbsbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/09.


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Complementary-Vertical Split (into effective child residue-blocks)".
		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords_sbwd = $new_curr_bl_coords_sbwd;
	my @new_list_gblks = @new_list_gblks_in_sbwd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords_sbwd} (=@{$new_curr_bl_coords_sbwd}) = (', join (', ', @{$new_orgs_gb_coords_sbwd}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_sbwd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	    ## $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    ## $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    ## $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbsbdir_out0, $sbsbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$new_gb_coords_sbwd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms_in_sbwd}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks_in_sbwd}, @{$rows_affd_by_new_blks_in_sbwd},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa2, $if_on_re_msa2,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations_sbwd}, @{$new_gblock_sizes_sbwd}, @{$new_curr_gblks_w_spec_lb_sbwd}, @{$new_curr_gblks_w_spec_rb_sbwd},
	    @{$new_cts_vincluding_sbwd}, @{$new_cts_vincluded_sbwd}, @{$new_cts_onx_sbwd}, $STRING_UNDEF, $beg_sbwd, $end_sbwd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA of the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"vertically split\" into COMPLEMENTARY-effective child blocks !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session with the sub-window '$sbwd_ID', after the $i th \"subject\" gap-block had been \"cvsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/21.) (MAIN II-S7b) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"cvsplit\" : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"cvsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords_sbwd}, @{$new_bds_gb_coords_sbwd}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"cvsplit\" : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session with the sub-window '$sbwd_ID', w/ the $i th subject gap-block \"cvsplit\" ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	unless (defined $total_wgts) { $total_wgts = $STRING_UNDEF; } # Added in minor DEBUGGING on 2020/05/09. #

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", "Sub($sbwd_ID)", $session_out, $sbsbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/21.) (MAIN II-S7b)" ##


  } # End of the Main Outer for-loop (over "nearest-neighboring" gap-blocks to be "vertically split" into effective COMPLEMENTARY-child blocks). #

			# End of "Added on 2019/08/30. (Main II-S7b)" #

	# End of "(Main II-S7b) Perform "Vertical-Split" (into complementary-sibling blocks) + Multiple-"Shift"s." #


  (print STDOUT " === ALL Processes Ended for the sub-window, '$sbwd_ID' ===\n\n") if ($TEST_LEVEL_MAIN>0);

} # END of the "OUTER for-loop for MAIN (II) (over the relevant sub-windows)." #

	##### END of "(MAIN II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}." #####


close ($fh_osyn); # (Added in REVISION on 2020/04/21.) (Pre-END) #


(print STDOUT "\n ===== ALL THE PROCESSED ENDED !!! =====\n\n") if ($TEST_LEVEL_MAIN>0);


	####### DEFINITIONS of the SUBROUTINES #######


	 ### NOTE: The Subroutines that had been here were MOVED to "MyANEX_IO.pm" in MAJOR REVISION (on 2019/07/07). ###


