package MyANEX_Main;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#	Version 0.7: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
# 
#	[NOTE 1 (on 2019/09/10)]:
#	Come to think of it now, most of the subroutines named 'list_nn(_xsplit(_plus))_ymerge_candidates'
#	could have been merged into a single subroutine, (named, e.g., as 'list_nn_gen_merge_candidates'),
#	provided that an input argument, $rel4tgt (the relation that the targetted gap-block pair must be in), is added,
#	and that we always search [ 0<= $gb1  <$ct_gbs, 0 <= $gb2 < $ct_gbs, $gb1 != $gb2].
#	[Hope to implement this idea ... some day in the future...]
#
#	[NOTE 2 (on 2019/09/10)]:
#	Actually, some of 'xsplit_gblock', 'xmerge_gblock_pair', and 'xsplit_ymerge_gblock_pair', may also be merged into a single subroutine, (named, e.g., as 'gen_split_gblock', etc.),
#	provided that an input argument, \@rows_2B_moved (listing the indices of the MSA rows that are to be moved in the process of the desired 'split', 'merge', or ...).
#	[This is a bit ambitious goal; its implementation may require some devising & works.]
#
# MyANEX_Main.pm:
#	This package is a collection of subroutines
#	that perform the main processes of ANEX.
#	[CREATION STARTED on 2019/02/09, ENDED on 2019/mm/dd.]
#
#	[REVISED on 2019/06/27]:
#	Changed the source of the subroutine, 'associate_nodeids_w_msa_rows',
#	from "MyANEX_Supple.pm" to "MyTreeMap_sppl.pm".
#
#	[REVISED on 2019/07/04]:
#	Added a new subroutine, 'compt_log_msa_prob_plus'.
#
#	[REVISED on 2019/07/08-09]:
#	Added new subroutines, 'serial_shifts_sgl_gblock' & 'serial_shifts_point2point'.
#
#	[MAJOR REVISION from 2019/07/11 till 2019/mm/dd]:
#	ADDED some subroutines for a wide variety of moves of gap-blocks
#	(as described, e.g., in "suppl5_blueprint1_ANEX.xxx.odt").
#
#	[REVISED on 2019/07/21]:
#	Now, some subroutines invokes 'mk_orgs_bds_gb_coords' (in "MyANEX_Supple.pm")
#	for computing new boundaries of gap-block coordinates.
#
#	[REVISED on 2019/08/06]:
#	Created a NEW VERSION of 'cmerge_split_grow_gblock_pair',
#	and RENAMED the old version as 'cmerge_split_grow_gblock_pair_OLD'.
#
#	[MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30)]:
#	Now, the subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm")
#	has been REPLACED with 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm").
#	In consequence,
#	this package can NOW use log-multiplication factors of case-(i) "gapped segments",
#	so that it can take FULL advantage of the main outputs of the program package, "LASTPIECE".
#	Therefore, it should NOW be able to compute the probabilities of MSAs
#	MUCH MORE ACCURATELY than before the revision.
#
#
# MyANEX_Supple.pm:
#	This package is a collection of subroutines
#	that supplement the main processes of ANEX.
#	[CREATION STARTED on 2019/02/09, ENDED on 2019/mm/dd.]
#
# MyDetect_cmplx_error_cands.pm:
#	This package is a collection of subroutines
#	necessary for detecting the candidates of regions containing complex errors.
#	[CREATION STARTED on December 10th (Sat), 2016, ENDED on Mmm dd (www), yyyy.]
#

use strict;
use warnings;

	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (1) #


use MyTreeManipu_II qw( output_tree ); # Modified on 2019/05/28. #
##use MyTreeManipu_II qw( );
##	# extract_sgltree_from_file output_tree 

use MyTaxonomy qw(
	fetch_common_ancestors
	);
	# if_ancestor fetch_ancestry calculate_abs_depth
	# fetch_ext_offsprings_nr fetch_all_offsprings_nr

use MyLikelihoods_respattern_msa_dna qw(
	br_calc_sbst_prob_sgl_clm
	); # 'br_calc_sbst_prob_sgl_clm' was ADDED on 2019/02/23.
	# calc_node2ntfreqs_td
	# calc_prob_descresstats_cond_ancstat2
	# calc_joint_prob_ancresstats_and_ancstat2
	# calc_clm_probs

=Set aside for potential use (I)

use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw( binomial_cum_via_betai );

=cut # END of "Set aside for potential use (I)" #

	# END of "IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (1)" #

use MyHash_2 qw(
	cp_comp_array
	cp_comp_hash
	); # Added on 2019/02/18.


	# ADDED in MAJOR REVISION (on 2019/05/21). (0a) #

use MyGapPatternTopology qw(
	mk_inter_block_relations
	extensive_exam_inter_block_relations

	enum_sites_contribute_to_gblk
	mk_list_gblock_sizes
	search_for_coll_cmpl_gblks
	compt_aln_degeneracy

	ct_necessary_shifts
	compt_distance_bw_gblk_and_pt
	compt_inter_gblock_distance

	mk_lists_gblocks_w_spec_bounds

	cts_vincluding_vincluded_onx_gblks

	vextend_right_end_sbj_gblk
	vextend_left_end_sbj_gblk
	vextend_rear_ends_rflnk_vincluded_gblks
	vextend_rear_ends_lflnk_vincluded_gblks

	vretract_left_end_sbj_gblk
	vretract_right_end_sbj_gblk
	vretract_right_ends_vincluded_gblks
	vretract_left_ends_vincluded_gblks

	mk_sets_vert_equiv_gblks
	mk_sets_swappable_gblks
	chk_ovlp_bw_vequiv_gblks
	chk_pos_orders_amg_swpbl_gblks

	chk_chng_gptrn_topology_psm_R
	chk_chng_gptrn_topology_psm_L
	encode_gptrn_topology_psm

	smlt2_extend_left_end_to_left
	smlt2_extend_right_end_to_right

	examine_if_gbs_on_msa_ends
	chk_excluded_bin_hrels

	expand_list_x_bhrels
	); # ADDED 'enum_sites_contribute_to_gblk' in creation of 'smerge_split_gblock_pair' (on 2019/07/31). #
	   # ADDED 'expand_list_x_bhrels' (for improved exclusion of redundant MSAs) (on 2020/05/17). #
	#
	# mk_list_interfering_blocks
	# mk_br2up_down2blocks
	# mk_br2up_down2blocksets
	# mk_list_interfering_blocksets
	# mk_list_cmpl_interfering_blocksets
	# compt_size_gblock
	# three_smallest_sizes
	# mk_list_coll_cmpl_gblocks
	# mk_gblock2coll_cmpl
	# cmp_nmarrays
	# smlt_extend_left_end_to_left
	# smlt_extend_right_end_to_right

use MyTreeMap_indels_spt_odr_hs qw(
	enum_psm_indel_scns_for_msa_hs_mrounds
	);

			## MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (US) ##
##use MyTreeMap_indels_ML_hs_hs qw(
##	apprx1_tot_lnlk_gpattern_msa2_hs
##	);
##
use MyTreeMap_indels_ML_hs_hs_wLP qw(
	apprx1_wlp_tot_lnlk_gpattern_msa2_hs
	);
			## END of "MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (US)" ##

	# END of "ADDED in MAJOR REVISION (on 2019/05/21). (0a)" #

use MyArrayManipu qw (
	multiple_sort
	); # Added in MAJOR REVISION (on 2019/07/11). #

use MyANEX_Supple qw(
	mk_orgs_bds_gb_coords
	); # Added in REVISION (on 2019/07/21). #

use MyTreeMap_indels_spt_odr2 qw(
	cmp_gblocks2
	cmp_gblocks3
	); # Added in creation of 'cvmerge_gblock_pair' on 2019/07/26.
	   # Added 'cmp_gblocks3' in revision of 'cvmerge_gblock_pair' on 2019/07/28.

use MyTreeMap_sppl qw(
	vsplit_into_eff_child_gblocks
	vsplit_into_cmpl_eff_child_rblocks
	); # Added 'vsplit_into_eff_child_gblocks' in creation of 'vsplit_gblock' (on 2019/08/23). #
	   # Added 'vsplit_into_cmpl_eff_child_rblocks' in creation of 'cvsplit_gblock' (on 2019/08/23). #

use MyArrayManipu_num qw(
	diff_nmarray
	union_nmarray
	); # Added 'union_nmarray' in creation of 'vmerge_gblock_pair' (on 2019/10/11). #


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	nonrec_explore_hyperrect
	initialize_multidim_storage

	shift_bl_and_compt_prob_incr_brief
	shift_bl_and_compt_prob_incr
	shift_gblock
	nonrec_mlt_shifts_noninf_gapblocks
	nonrec_mlt_shifts_gapblocks
	compt_log_msa_prob_plus
	serial_shifts_sgl_gblock
	serial_shifts_point2point

	displace_including_and_onx_to_left
	displace_including_and_onx_to_right
	displace_vincluding_veq_and_onx_to_right_end

	list_nn_purge_and_cmerge_candidates
	list_nn_smerge_candidates
	list_nn_cvmerge_candidates

	list_smerge_split_candidates
	list_nn_cmerge_split_pre_candidates
	enum_sites_avail_for_csplit
	enum_sites_avail_for_csplit_sgl_gb
	list_cmerge_split_candidates

	list_ssplit_candidates
	list_csplit_candidates
	list_nn_csplit_candidates
	list_nn_gen_split_candidates
	list_vsplit_candidates
	list_cvsplit_candidates

	list_nn_vsplit_plus_smerge_candidates
	list_nn_cvsplit_plus_smerge_candidates
	list_nn_ssplit_plus_vmerge_candidates

	list_revpurge_cand_sizes

	purge_gblock_pair
	cmerge_gblock_pair
	smerge_gblock_pair
	cvmerge_gblock_pair
	vmerge_gblock_pair

	ssplit_gblock
	initial_csplit_gblock
	csplit_gblock
	vsplit_gblock
	cvsplit_gblock

	smerge_split_gblock_pair
	icmerge_gblock_pair
	undo_hovlp_vcmpl_gbsets
	undo_hovlp_vcmpl_gbsets_1sbj
	cmerge_split_grow_gblock_pair

	vsplit_plus_smerge_gblock_pair
	cvsplit_plus_smerge_gblock_pair
	ssplit_plus_vmerge_gblock_pair

	initial_revpurge_gblock_pair
	reverse_purge_gblock_pair

	printout_multidim_storage
	print_out_gblock_configs
	print_out_remng_attr_gblks
	br_print_out_remng_attr_gblks
	);
	# Added 'nonrec_mlt_shifts_gapblocks' on 2019/05/30.
	# Added 'compt_log_msa_prob_plus' on 2019/07/04.

sub nonrec_explore_hyperrect (\@\@);
sub initialize_multidim_storage (\@);

sub shift_bl_and_compt_prob_incr_brief (\@\@\@$$\@$\%\@$$\%\%\@\%\%); # RENAMED on 2019/03/27.
##sub shift_bl_and_compt_prob_incr (\@\@\@$$\@$\%\@$$\%\%\@\%\%);

sub shift_bl_and_compt_prob_incr (\@\@\@\@\@$$\@$\%\@$$\%\%\@\%\%); # ADDED on 2019/03/27.

sub shift_gblock (\@\@\@\@\@$$\@$); # ADDED on 2019/03/28.

##sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\%$\%\%\@\%\%); # OBSOLETE as of 2019/02/23.
sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\@\%$\%\%\%\@\%\%); # Modified on 2019/02/23.
sub nonrec_mlt_shifts_gapblocks (\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@); # ADDED on 2019/05/30.

sub serial_shifts_sgl_gblock (\@\@\@\@\@\@$$\@$\@\@\@\@\@); # ADDED on 2019/07/09.
sub serial_shifts_point2point (\@\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/09.
##sub displace_including_and_onx_to_left ($$\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/25.
sub displace_including_and_onx_to_left ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%); # => REVISED on 2019/07/26.
##sub displace_including_and_onx_to_right ($$\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/25.
sub displace_including_and_onx_to_right ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%); # => REVISED on 2019/07/26.
sub displace_vincluding_veq_and_onx_to_right_end ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%); # ADDED on 2019/08/03.

sub list_nn_purge_and_cmerge_candidates (\@\@\@$$); # ADDED on 2019/07/11.
sub list_nn_smerge_candidates (\@\@$); # ADDED on 2019/07/21.
sub list_nn_cvmerge_candidates (\@\@\@$); # ADDED on 2019/07/23.
sub list_smerge_split_candidates (\@\@$); # ADDED on 2019/08/01.
sub list_nn_cmerge_split_pre_candidates (\@\@\@$); # ADDED on 2019/08/04.
sub enum_sites_avail_for_csplit ($$\@\@\@\@$$); # ADDED on 2019/08/02.
sub enum_sites_avail_for_csplit_sgl_gb ($\@\@\@\@$$); # ADDED on 2019/08/15.
sub list_cmerge_split_candidates (\@\@\@\@\@\@$$$$); # ADDED on 2019/08/02.
sub list_ssplit_candidates (\@\@\@\@\@\@$$$); # ADDED on 2019/08/11.
sub list_csplit_candidates (\@\@\@$$\@\@\@\@\@$$$); # ADDED on 2019/08/15.
sub list_nn_csplit_candidates (\@); # ADDED on 2019/08/20.
sub list_nn_gen_split_candidates (\@$$); # ADDED on 2019/08/21. #
sub list_vsplit_candidates (\@$$\%\%$\%\@\%\%\@\@\@\@$$); # ADDED on 2019/08/21. #
sub list_cvsplit_candidates (\@$$\%\%$\%\@\%\%\@\@\@\@$$); # ADDED on 2019/08/27. #
sub list_nn_vsplit_plus_smerge_candidates (\@\@\@$); # ADDED on 2019/09/02. #
sub list_nn_cvsplit_plus_smerge_candidates (\@\@\@$); # ADDED on 2019/09/10. #
sub list_nn_ssplit_plus_vmerge_candidates (\@\@\@$); # ADDED on 2019/10/05. #

sub list_revpurge_cand_sizes (\@\@$$$); # ADDED on 2019/09/13. #

####sub purge_gblock_pair ($$\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/14.
##sub purge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@); # => REVISED on 2019/07/14.
sub purge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # => Further REVISED on 2019/07/21.

##sub cmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/19.
sub cmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # => REVISED on 2019/07/19.

##sub smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/07/21.
sub smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # => REVISED on 2019/08/07.

sub cvmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/07/27.
sub vmerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/10/12.

sub ssplit_gblock ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # ADDED on 2019/08/12.
sub initial_csplit_gblock ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/08/16.
sub csplit_gblock ($$\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/08/16.
sub vsplit_gblock ($\@\@\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/08/23.
sub cvsplit_gblock ($\@\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/08/26.

sub smerge_split_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # ADDED on 2019/07/30.
sub icmerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # ADDED on 2019/08/02.

sub undo_hovlp_vcmpl_gbsets ($$\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/08/08.
sub undo_hovlp_vcmpl_gbsets_1sbj ($\@\@\@\@\@\@\@$\@\@\@\@\@); # ADDED on 2019/08/22.

sub cmerge_split_grow_gblock_pair_OLD ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # ADDED on 2019/08/04. # => RENAMED on 2019/08/06.
####sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$); # ADDED on 2019/08/06.
##sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\%$\%); # => REVISED on 2019/08/18.
sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\@\%\%$\%); # => Further REVISED on 2019/08/19.

sub vsplit_plus_smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%); # ADDED on 2019/09/04.
sub cvsplit_plus_smerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%); # ADDED on 2019/09/11.
##sub ssplit_plus_vmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%); # ADDED on 2019/10/13.
sub ssplit_plus_vmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # => REVISED on 2019/10/14. #

sub initial_revpurge_gblock_pair (\@$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/09/15.
sub reverse_purge_gblock_pair (\@$\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%); # ADDED on 2019/09/15.


sub printout_multidim_storage (\@\@\@$\*$); # Added on 2019/03/04.
sub print_out_gblock_configs (\@\@\@\@\@\*); # ADDED on 2019/07/09.
sub print_out_remng_attr_gblks (\@\@\@\@\@\@\@\@\*); # ADDED on 2019/07/14.
sub br_print_out_remng_attr_gblks (\@\@\@\*); # ADDED on 2019/07/31.


	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2) #

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;


	# END of "IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2)" #


	# This specifies the level of tests on the MAIN subroutine, 'xxx'.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# For a MAIN subroutine, 'nonrec_mlt_shifts_noninf_gapblocks'. #
our $TEST_LEVEL_nr_msht_ninf_gblks = 0;
#our $TEST_LEVEL_nr_msht_ninf_gblks = 1;
#our $TEST_LEVEL_nr_msht_ninf_gblks = 5;

	# For printing out the ingredients of alignment probabilities in 'nonrec_mlt_shifts_noninf_gapblocks'. #
our $TEST_LEVEL_nr_msht_ninf_gblks2 = 0;
#our $TEST_LEVEL_nr_msht_ninf_gblks2 = 3;

	# For a MAIN subroutine, 'nonrec_mlt_shifts_gapblocks'. # (ADDED on 2019/05/19.)
our $TEST_LEVEL_nr_msht_gblks = 0;
#our $TEST_LEVEL_nr_msht_gblks = 1;
#our $TEST_LEVEL_nr_msht_gblks = 5;

	# For printing out the ingredients of alignment probabilities in 'nonrec_mlt_shifts_gapblocks'. #
our $TEST_LEVEL_nr_msht_gblks2 = 0;
##our $TEST_LEVEL_nr_msht_gblks2 = 3;

	# For the subroutine, 'compt_log_msa_prob_plus'. # (ADDED on 2019/07/04.)
our $TEST_LEVEL_cmpt_lmpp = 0;
#our $TEST_LEVEL_cmpt_lmpp = 5;

	# For printing out the ingredients of alignment probabilities in 'compt_log_msa_prob_plus'. # (ADDED on 2019/07/04.)
our $TEST_LEVEL_cmpt_lmpp2 = 0;
#our $TEST_LEVEL_cmpt_lmpp2 = 5;

	# For 'nonrec_explore_hyperrect'.
our $TEST_LEVEL_nr_exp_hprct = 0;
#our $TEST_LEVEL_nr_exp_hprct = 5;

	# For 'initialize_multidim_storage'.
our $TEST_LEVEL_init_md_strg = 0;
#our $TEST_LEVEL_init_md_strg = 5;

	# For 'serial_shifts_sgl_gblock'. #
our $TEST_LEVEL_ssh_sgblk = 0;
#our $TEST_LEVEL_ssh_sgblk = 5;

	# For 'serial_shifts_point2point'. #
our $TEST_LEVEL_ssh_p2p = 0;
#our $TEST_LEVEL_ssh_p2p = 5;

	# For 'displace_including_and_onx_to_left'. #
our $TEST_LEVEL_diotl = 0;
#our $TEST_LEVEL_diotl = 6;

	# For 'displace_including_and_onx_to_right'. #
our $TEST_LEVEL_diotr = 0;
#our $TEST_LEVEL_diotr = 6;

	# For 'displace_vincluding_veq_and_onx_to_right_end'.
our $TEST_LEVEL_dvieotre = 0;
#our $TEST_LEVEL_dvieotre = 6;

	# For 'purge_gblock_pair'. #
our $TEST_LEVEL_prg_gbp = 0;
#our $TEST_LEVEL_prg_gbp = 6;

	# For 'cmerge_gblock_pair'. # (Added on 2019/07/19.)
our $TEST_LEVEL_cmrg_gbp = 0;
#our $TEST_LEVEL_cmrg_gbp = 5;
 
	# For 'smerge_gblock_pair'. # (Added on 2019/07/21.)
our $TEST_LEVEL_smrg_gbp = 0;
#our $TEST_LEVEL_smrg_gbp = 6;

	# For 'cvmerge_gblock_pair'. # (Added on 2019/07/27.)
our $TEST_LEVEL_cvmrg_gbp = 0;
#our $TEST_LEVEL_cvmrg_gbp = 0;

	# For 'vmerge_gblock_pair'. # (ADDED on 2019/10/10.) #
our $TEST_LEVEL_vmg_gbp = 0;
#our $TEST_LEVEL_vmg_gbp = 6;

	# For 'ssplit_gblock'. # (ADDED on 2019/08/12.)
our $TEST_LEVEL_sspl_gb = 0;
##our $TEST_LEVEL_sspl_gb = 6;

	# For 'initial_csplit_gblock'. # (ADDED on 2019/08/15; Moved to here on 2019/08/17.)
our $TEST_LEVEL_init_cspl_gb = 0;
##our $TEST_LEVEL_init_cspl_gb = 6;

	# For 'csplit_gblock'. # (ADDED on 2019/08/16; Moved to here on 2019/08/17.)
our $TEST_LEVEL_cspl_gb = 0;
##our $TEST_LEVEL_cspl_gb = 6;

	# For 'vsplit_gblock'. # (ADDED on 2019/08/23.)
our $TEST_LEVEL_vspl_gb = 0;
##our $TEST_LEVEL_vspl_gb = 6;

	# For 'cvsplit_gblock'. # (ADDED on 2019/08/25.)
our $TEST_LEVEL_cvspl_gb = 0;
##our $TEST_LEVEL_cvspl_gb = 6;

	# For 'smerge_split_gblock_pair'. # (Added on 2019/07/30.)
our $TEST_LEVEL_smrg_spl_gbp = 0;
##our $TEST_LEVEL_smrg_spl_gbp = 6;

	# For 'icmerge_gblock_pair'. # (ADDED on 2019/08/02.)
our $TEST_LEVEL_icmrg_gbp = 0;
##our $TEST_LEVEL_icmrg_gbp = 6;

	# For 'undo_hovlp_vcmpl_gbsets'. #
our $TEST_LEVEL_undo_ho_vcp_sets = 0;
##our $TEST_LEVEL_undo_ho_vcp_sets = 6;

	# For 'cmerge_split_grow_gblock_pair'. # (ADDED on 2019/08/04; MOVED to here on 2019/08/05.)
our $TEST_LEVEL_cmrg_spl_grw_gbp = 0;
##our $TEST_LEVEL_cmrg_spl_grw_gbp = 6;

	# For 'vsplit_smerge_gblock_pair'. # (ADDED on 2019/09/02, MOVED on 2019/09/04.)
our $TEST_LEVEL_vssm_gbp = 0;
##our $TEST_LEVEL_vssm_gbp = 6;

	# For 'cvsplit_plus_smerge_gblock_pair'. # (ADDED on 2019/09/10.) #
our $TEST_LEVEL_cvssm_gbp = 0;
#our $TEST_LEVEL_cvssm_gbp = 6;

	# For 'ssplit_plus_vmerge_gblock_pair'. # (ADDED on 2019/10/05.) #
our $TEST_LEVEL_ssvm_gbp = 0;
#our $TEST_LEVEL_ssvm_gbp = 6;

	# For 'initial_revpurge_gblock_pair'. # (ADDED on 2019/09/15) #
our $TEST_LEVEL_init_rvprg_gbp = 0;
#our $TEST_LEVEL_init_rvprg_gbp = 6;

	# For 'reverse_purge_gblock_pair'. # (ADDED on 2019/09/13; MOVED to here on 2019/09/15)
our $TEST_LEVEL_rvprg_gbp = 0;
#our $TEST_LEVEL_rvprg_gbp = 6;


	# For 'printout_multidim_storage'.
our $TEST_LEVEL_po_md_strg = 0;
#our $TEST_LEVEL_po_md_strg = 5;


	# For a sub-main subroutine, 'shift_bl_and_compt_prob_incr' (& 'shift_bl_and_compt_prob_incr_brief'). #
our $TEST_LEVEL_sh_bl_cmpt_p_incr = 0;
##our $TEST_LEVEL_sh_bl_cmpt_p_incr = 1;
##our $TEST_LEVEL_sh_bl_cmpt_p_incr = 5;

	# For a derivative subroutine, 'shift_gblock'. #
our $TEST_LEVEL_sh_gblk = 0;
##our $TEST_LEVEL_sh_gblk = 1;
##our $TEST_LEVEL_sh_gblk = 5;



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


# nonrec_explore_hyperrect:
#	This pilot subroutine of the main subroutine, 'nonrec_mlt_shifts_gapblocks',
#	implements a non-recursive method to explore a hyperrectangle
#	that is defined as a Cartesian product of a number of intervals, each of a coordinate.
#	[CREATION STARTED on 2019/02/18, ENDED on 2019/02/19.]
#
#	[REVISED on 2020/04/15]:
#	Fixed a potential BUG that causes an abnormal behavior of the subroutine when #{coordinates} = 0.
#
#	How to explore?:
#	Starting with the "origin", 
#	(i) it first moves to the right along each coordinate axis;
#	(ii) after reaching the "right-end", it returns to the "origin",
#	and then moves to the left along the same coordinate axis;
#	(iii) after examining the entire interval of the coordinate,
#	it then moves the "parent" coordinate by one;
#	(vi) repeat (i)-(iii) until the entire hyperrectangle is explored.
#
#	(See SM-2 of "suppl3_blueprint1_ANEX.draft9.odt" for details.)
#
# INPUTS: (\@org_coords, \@bds_coords), where
#	$org_coords[$k] gives the "origin" of the $k th coordinate;
#	($left_bd, $right_bd) = @{$bds_coords[$k]} gives the left and right-boundaries
#	of the $k th coordinate;
#	(NOTE that the boundaries are also included in the interval to be explored.)
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   undef if it failed.
#
sub nonrec_explore_hyperrect (\@\@) {

	my ($org_coords, $bds_coords) = @_;

	my $ct_coords = @{$org_coords};

    (print STDOUT "< Subtoutine 'nonrec_explore_hyperrect' (#{coordinates} = $ct_coords) >\n\n") if ($TEST_LEVEL_nr_exp_hprct >0);

		# 1st consistency check. #

	if (@{$bds_coords} != $ct_coords) {
	    my $ct_bd_pairs = @{$bds_coords};
	    print STDERR "[ERROR 'nonrec_explore_hyperrect'] The number of coordinate boundary-pairs (=$ct_bd_pairs) CLASHES with the number of coordinate origins (=$ct_coords); $! !\n\n";
	    return undef;
	}

    if ($TEST_LEVEL_nr_exp_hprct >2) {
	print STDOUT "\n[ Input variables ]\n\n";
	print STDOUT join ("\t", 'Indx_coord', 'org', 'left_bd', 'right_bd'), "\n\n";
	for (my $k = 0; $k <$ct_coords; $k++) {
	    print STDOUT join ("\t", $k, $org_coords->[$k], @{$bds_coords->[$k]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_exp_hprct >2) {...}"

	if ($ct_coords == 0) {
			# (ADDED to fix a possible BUG (on 2020/04/15).)
	    print STDOUT "[INFORMATION 'nonrec_explore_hyperrect'] The number of coordinates is ZERO!! ... I will end this subroutine withOUT doing anything ...\n\n";
	    return 1;
	}

		# 2nd consistency check. #
	my @ill_coords = ();
	for (my $k=0; $k <$ct_coords; $k++) { # Preliminary for-loop (over coordinates). #
	    my $org = $org_coords->[$k];
	    my ($lbd, $rbd) = @{$bds_coords->[$k]};
	    if (($org < $lbd) or ($rbd < $org)) { push @ill_coords, $k; }
	} # End of the preliminary for-loop (over coordinates). #

	if (@ill_coords>0) {
	    print STDERR "[ERROR 'nonrec_explore_hyperrect'] In the coordinates, (", join (', ', @ill_coords), "), the 'origin's are NOT within the intervals specified by the boundaries: $! !\n\n";
	    return undef;
	}

		# Initialize some important variables. #

	my @incr_coords = (); # The $k th element specify the "next" increment of the $k th coordinate. (= +1 if the coordinate increases, -1 if it decreases.)
	for (1 .. $ct_coords) { push @incr_coords, 1; }

	my @curr_coords0 = @{$org_coords}; # Initial values of the "current" coordinates.
	my @layer0 = (\@curr_coords0);

	my @stack = ();
	for (0 .. $ct_coords) {
	    my $cp_layer0 = cp_comp_array (@layer0);
	    push @stack, $cp_layer0;
	}
	my $ct_layers = @stack;

    if ($TEST_LEVEL_nr_exp_hprct >2) {

	print STDOUT "[ Initialized variables ]\n\n";

	print STDOUT '@incr_coords    = (', join (', ', @incr_coords), "),\n\n";
	print STDOUT '@curr_coords0 = (', join (', ', @curr_coords0), ").\n\n";

	if ($TEST_LEVEL_nr_exp_hprct >3) {

	    print STDOUT "\n";
	    print STDOUT '{ Contents of @stack ', "(#{layers) = $ct_layers) }:\n\n";

	    for (my $ly = 0; $ly < $ct_layers; $ly++) {
		print STDOUT "| The $ly th layer |\n\n";
		my $layer = $stack[$ly];
		my ($coords_ly) = @{$layer};
		print STDOUT '@{$coords_ly} = (', join (', ', @{$coords_ly}), "),\n\n";
		print STDOUT "\n\n";
	    }

	} # End of "if ($TEST_LEVEL_nr_exp_hprct >3) {...}"

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_nr_exp_hprct >2) {...}"


		### The Main Process ###

    (print STDOUT "[[ The Main Process ]]\n\n") if ($TEST_LEVEL_nr_exp_hprct >1);

	my $wrk_layer = pop @stack;

	my $indx_sbj_crd = $ct_coords -1;

	my $counter = 0;
	my $ct_moves = 0; # Added on 2019/02/19.

# $TEST_LEVEL_nr_exp_hprct

	while (1) { # The MAIN outer while-loop (for exploring the entire hyper-rectangle). #

	    $counter++;

    (print STDOUT "( The $counter th point visited; ", '$indx_sbj_crd =', " $indx_sbj_crd )\n\n") if ($TEST_LEVEL_nr_exp_hprct >1);

	    my ($curr_coords) = @{$wrk_layer};
	    my $incr = $incr_coords[$indx_sbj_crd];

	    my $sbj_coord = $curr_coords->[$indx_sbj_crd];
	    my $new_sbj_coord = $sbj_coord + $incr;
	    my ($lbd_sbj_coord, $rbd_sbj_coord) = @{$bds_coords->[$indx_sbj_crd]};

    if ($TEST_LEVEL_nr_exp_hprct > 2) {

	my $size_stack = @stack;

	print STDOUT '@{$curr_coords} = (', join (', ', @{$curr_coords}), ").\n\n";
	print STDOUT "{Next increment} = $incr ,\n";
	print STDOUT "{Stack size} = $size_stack .\n\n";

    } # End of "if ($TEST_LEVEL_nr_exp_hprct > 2) {...}"

    if ($TEST_LEVEL_nr_exp_hprct >= 5) {

	print STDOUT '$sbj_coord     = ', $sbj_coord, " ,\n";
	print STDOUT '$new_sbj_coord = ', $new_sbj_coord, " ,\n";
	print STDOUT '($lbd_sbj_coord, $rbd_sbj_coord) = ', "($lbd_sbj_coord, $rbd_sbj_coord) .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_exp_hprct >= 5) {...}"

	    if ($rbd_sbj_coord < $new_sbj_coord) {

		## Stepping over the right-boundary.

		my $copy = cp_comp_array (@{$stack[$#stack]});
		$wrk_layer = $copy;

		$incr_coords[$indx_sbj_crd] = -1;

    (print STDOUT " ... It will go beyond the right-boundary !! ... Move back to the origin...\n\n\n") if ($TEST_LEVEL_nr_exp_hprct > 2); # Added on 2019/02/19.

		next;

	    } elsif ($new_sbj_coord < $lbd_sbj_coord) {

		## Stepping over the left-boundary.

		($indx_sbj_crd == 0) and last; # End the exploration. #

		$wrk_layer = pop @stack;

		$incr_coords[$indx_sbj_crd] = +1;
		$indx_sbj_crd--;

    (print STDOUT " ... It will go beyond the left-boundary !! ... Move back to the origin, and shift the 'parent' coordinate...\n\n\n") if ($TEST_LEVEL_nr_exp_hprct > 2); # Added on 2019/02/19.

		next;
	    }

		## Within the range. ##

	    $ct_moves++; # Added on 2019/02/19.

    (print STDOUT " ... The $ct_moves th Move has been executed !!\n\n\n") if ($TEST_LEVEL_nr_exp_hprct > 1); # Added on 2019/02/19.

		# Actually update the current coordinate. #

	    $curr_coords->[$indx_sbj_crd] = $new_sbj_coord;

		# Raise $indx_sbj_crd all the way to ($ct_coords -1).

	    my $ct_cps = 0;
	    while ($indx_sbj_crd < $ct_coords -1) { # 1st inner while-loop. #
		$indx_sbj_crd++;

		$ct_cps++;
	    }  # End of the 1st inner while-loop. #

		# Stuff the stack with copies of the newly computed layer.

	    while ($ct_cps>0) { # 2nd inner while-loop. #
		$ct_cps--;
		my $cp_layer = cp_comp_array (@{$wrk_layer});
		push @stack, $cp_layer;
	    }  # End of the 2nd inner while-loop. #

	} # End of the MAIN outer while-loop (for exploring the entire hyper-rectangle). #

# $TEST_LEVEL_nr_exp_hprct

    (print STDOUT "< Subtoutine 'nonrec_explore_hyperrect' Ended (#{visits} = $counter, #{moves} = $ct_moves) />\n\n") if ($TEST_LEVEL_nr_exp_hprct >0); # Modified on 2019/02/19.
    ## (print STDOUT "< Subtoutine 'nonrec_explore_hyperrect' Ended />\n\n") if ($TEST_LEVEL_nr_exp_hprct >0); # OBSOLETE as of 2019/02/19. #

	return 1;

}	# End of the subroutine 'nonrec_explore_hyperrect'. #


# initialize_multidim_storage:
#	This satellite subroutine of 'nonrec_mlt_shifts(_noninf)_gapblocks'
#	initializes the storage for some information (e.g., the increment of the log-probability
#	of the alternative alignment compared to the initial alignment)
#	on each point in the coordinate space.
#	[CREATED on 2019/02/21]
#	[REVISED on 2020/04/15]:
#	Fixed a potentail BUG that causes the subroutine to behave abnormally if (#{coordinates} == 0).
#
# INPUT: (\@bds_bl_coords), where
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds.]
#
# OUTPUT: (\@mltdim_storage), where
#
#	$mltdim_storage[$k_0]->...->[$k_{B-1}] = [] (where B = scalar (@bds_bl_coords))
#	is the storage for some information on the point with the coordinate,
#	($k_0, ..., $k_{B-1}) within the hyperrectangle in question.
#
sub initialize_multidim_storage (\@) {

	my ($bds_bl_coords) = @_;

	my $ct_blks = @{$bds_bl_coords};

    (print STDOUT ' ' x 2, "< Subroutine 'initialize_multidim_storage' (#{coordinates} = $ct_blks) >\n\n") if ($TEST_LEVEL_init_md_strg>0);

	if ($ct_blks == 0) {
			# (ADDED to fix a Potential BUG (on 2020/04/15).) #
	    print STDOUT "[INFORMATION 'initialize_multidim_storage'] The number of coordinates is ZERO!! ... I will end this subroutine withOUT doing anything ...\n\n";
	    return [];
	}

		# Initialize the current point. #
	my @curr_point = ();
	for (my $k=0; $k < $ct_blks; $k++) { $curr_point[$k] = $bds_bl_coords->[$k][0]; }

	my @mltdim_storage = ();

		# Initialize the output storage. #
		# & Initialize the stack (as a nested series of single anonymous arrays).

	my $sbj_storage0 = \@mltdim_storage;
	my @stack = ($sbj_storage0);
	for (my $k=0; $k < $ct_blks; $k++) { # 2nd for-loop (over blocks).
	    my $coord = $curr_point[$k];
	    my $child_storage = $sbj_storage0->[$coord] = [];
	    push @stack, $child_storage;
	    $sbj_storage0 = $child_storage;
	} # End of the 2nd for-loop (over blocks).
	my $term_storage0 = pop @stack;

    if ($TEST_LEVEL_init_md_strg > 2) {
	push @{$term_storage0}, @curr_point;
    }

		# Main: Construct the entire output storage. #

	my $indx_curr_crd = $ct_blks -1;

	while (1) { # Main while-loop (to construct the output).

	    my $curr_sbj_coord = $curr_point[$indx_curr_crd];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$indx_curr_crd]};

	    if ($curr_sbj_coord == $rbd) {
		if ($indx_curr_crd == 0) { last; }

		$curr_point[$indx_curr_crd] = $lbd;
		$indx_curr_crd--;
		pop @stack;
		next;
	    }

		# A BUG fixed (on 2019/02/22).
	    $curr_point[$indx_curr_crd]++;
	    $curr_sbj_coord = $curr_point[$indx_curr_crd];
	    ##$curr_sbj_coord = $curr_point[$indx_curr_crd]++;

	    my $parent_storage = $stack[$#stack];

	    my $curr_storage = $parent_storage->[$curr_sbj_coord] = [];

	    while ($indx_curr_crd < $ct_blks - 1) { # Inner while-loop.
		push @stack, $curr_storage;
		$indx_curr_crd++;
		$curr_sbj_coord = $curr_point[$indx_curr_crd];
		$curr_storage = $curr_storage->[$curr_sbj_coord] = [];

	    } # End of the inner while-loop.

    if ($TEST_LEVEL_init_md_strg > 2) {
	push @{$curr_storage}, @curr_point;
    }

	} # End of the main while-loop (to construct the output).


    (print STDOUT ' ' x 2, "< Subroutine 'initialize_multidim_storage' Ended />\n\n") if ($TEST_LEVEL_init_md_strg>0);

	return \@mltdim_storage;

}	# End of the subroutine 'initialize_multidim_storage'. #



# shift_bl_and_compt_prob_incr_brief:
#	This satellite subroutine of 'nonrec_mlt_shifts_noninf_gapblocks'
#	actually performs a shift of a specified gap-block,
#	updates @curr_set_clms, @curr_bds_blks, and @curr_bl_coords accordingly,
#	and computes the resulting increments of the log-probability and the count of null columns.
#	[CREATION STARTED on 2019/02/24, ENDED on 2019/03/01.]
#	[On 2019/03/02, some major test outputs were added.]
#	[On 2019/03/27, RENAMED from 'shift_bl_and_compt_prob_incr' to 'shift_bl_and_compt_prob_incr_brief'.]
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, $bl, $sh, \@rows_affd_by_gblks, $cnct_null_clm, \%clm2ln_prob_sbst, \@exnodes_in_msa, $char_abs, $topnode_id, \%node2ch, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where
#
#	@curr_set_clms is the current set of alignment columns.
#	@column = @{$curr_set_clms[$c]} specifies the $c th column. More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#
#	@{$curr_bds_blks[$bl]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $bl th gap-block;
#
#	$curr_bl_coords[$bl] gives the current coordinate of the $bl th gap-block;
#
#	$bl specifies the subject gap-block;
#
#	$sh specifies the direction of the "shift". More precisely,
#	$sh = +1 if the block will be "shift"ed to the right,
#	    = -1 if it will be "shift"ed to the left;
#
#	@{$rows_affd_by_gblks[$bl]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $bl th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column);
#
#	%clm2ln_prob_sbst = ( join (':', @column) => $log_prob_sbst, ...)
#	where $log_prob_sbst is the log-probability of the column @column under the given substitution model;
#
#	$exnodes_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
#
# OUTPUTS: ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms), where
#
#	$flag = 0 if all the process ended successfully,
#	      = -1 if the subject block stepped over the left-end of the local MSA,
#	      = +1 if the subject block stepped over the right-end of the local MSA,
#	      = undef if another type of error occurs;
#
#	$incr_ln_prb_sbst is the increment of the log-probability (under the given substitution model) resulting from the shift (= undef if an error occurs);
#
#	$incr_ct_null_clms is the increment of the count of null-columns resulting from the shift (= undef if an error occurs).
#
sub shift_bl_and_compt_prob_incr_brief (\@\@\@$$\@$\%\@$$\%\%\@\%\%) {

	my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bl, $sh, $rows_affd_by_blks, $cnct_null_clm, $clm2ln_prob_sbst, $exnodes_in_msa, $char_abs, $topnode_id, $node2ch, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_;

	unless (($sh == +1) or ($sh == -1)) { # Added on 2019/03/01. #
	    print STDERR "[ERROR 'shift_bl_and_compt_prob_incr_brief'] ILLEGAL INPUT: ", '$sh = ', "$sh: $! !\n\n";
	    return (undef, undef, undef);
	}

	my $ct_clms = @{$curr_set_clms};

	my $prev_bl_coord = $curr_bl_coords->[$bl]; # Keep the previous coordinate.

    (print STDOUT "< Subroutine 'shift_bl_and_compt_prob_incr_brief' (#{columns} = $ct_clms, ", '$sh = ', $sh, ', $bl = ', $bl, ", {block coordinate}_bf = $prev_bl_coord) >\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 0); # Added on 2019/03/02.


	my $bds_block_sbj = $curr_bds_blks->[$bl];
	my ($prev_lbd, $prev_rbd) = @{$bds_block_sbj}; # Keep the boundaries of the subject block.


		# Actually shift the $bl th block on column in the $sh direction.

	my $curr_bl_coord = ($curr_bl_coords->[$bl] += $sh); # Shift the current coordinate.
	my $curr_lbd = ($bds_block_sbj->[0] += $sh); # Shift the left boundary.
	my $curr_rbd = ($bds_block_sbj->[1] += $sh); # Shift the right boundary.

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (1) (added on 2019/03/02)
	print STDOUT "\n[ Information on the subject block (index = $bl) ]\n\n";
	print STDOUT "{block-coordinate}_bf = $prev_bl_coord ,\n";
	print STDOUT "{block-coordinate}_af = $curr_bl_coord .\n\n";
	print STDOUT "(left_bound, right-bound)_bf = ($prev_lbd, $prev_rbd) ,\n";
	print STDOUT "(left_bound, right-bound)_af = ($curr_lbd, $curr_rbd) .\n\n";
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (1) ...}"


	if ($curr_lbd < 0) {
		# The subjet block stepped over the left-end of the local MSA.
	    return (-1, undef, undef);

	} elsif ($curr_rbd >= $ct_clms) {
		# The subjet block stepped over the right-end of the local MSA.
	    return (+1, undef, undef);
	}


		# Initialize the main outputs. #

	my $incr_ln_prb_sbst = my $incr_ct_null_clms = 0;

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3) { # (Added on 2019/03/02.)

	print STDOUT "\n[ Initial Values of the Main Outputs ]\n\n";
	print STDOUT '$incr_ln_prb_sbst = ', $incr_ln_prb_sbst, " ,\n";
	print STDOUT '$incr_ct_null_clms = ', $incr_ct_null_clms, " .\n\n";

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3) {...}"

		# Extract the columns that are to be modified. #

	my @rlv_clm_ids = ($sh>0) ? ($prev_lbd, $curr_rbd) : ($curr_lbd, $prev_rbd);
	my ($rlv_clm1, $rlv_clm2) = my @rlv_clms = @{$curr_set_clms}[@rlv_clm_ids];

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) (added on 2019/03/02)

	print STDOUT "\n[ The columns to be modified (BEFORE the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) ...}"

		# Compute the contributions from the columns BEFORE being modified. #

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (1) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (1): Compute the contributions from the columns BEFORE the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', '$incr_ct_null_clms', '$ln_cw_prb', '$incr_ln_prb_sbst'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (1) ...}"


	foreach my $clm (@rlv_clms) { # 1st foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms--;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, 0, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }
	    my $ln_cw_prb = $clm2ln_prob_sbst->{$cnct_clm};
	    if (defined $ln_cw_prb) {
		$incr_ln_prb_sbst -= $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next;
	    }

	    my $cw_prb = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$clm}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb) {
		print STDERR "[ERROR 'shift_bl_and_compt_prob_incr_brief'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED for the column BEFORE the \"shift\" (concatenation of the pre-shift column is: '$cnct_clm'): $! !\n\n";
		return (undef, undef, undef);
	    }

	    	# $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

	    $ln_cw_prb = log ($cw_prb);

	    $incr_ln_prb_sbst -= $ln_cw_prb;
	    $clm2ln_prob_sbst->{$cnct_clm} = $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

	} # End of the 1st foreach-loop (over relevant columns).


		# MODIFY @{$curr_set_columns}.
		# More precisely, SWAP the portions of the relevant columns affected by the subject gap-block.
	my $affd_rows = $rows_affd_by_blks->[$bl];

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (2): MODIFY the columns as a result of the \"shift\". (rows affected: ", join (',', @{$affd_rows}), ") ]\n\n";

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) ...}"



	foreach my $row (@{$affd_rows}) { # 2nd foreach-loop (over affected rows).
	    my $tmp = $rlv_clm1->[$row];
	    $rlv_clm1->[$row] = $rlv_clm2->[$row];
	    $rlv_clm2->[$row] = $tmp;
	} # End of the 2nd foreach-loop (over affected rows).


   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2)

	print STDOUT "\n[ The columns modified (AFTER the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) ...}"


		# Compute the contributions from the columns AFTER being modified. #

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (3): Compute the contributions from the columns AFTER the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', '$incr_ct_null_clms', '$ln_cw_prb', '$incr_ln_prb_sbst'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) ...}"


	foreach my $clm (@rlv_clms) { # 3rd foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms++;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, 0, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }
	    my $ln_cw_prb = $clm2ln_prob_sbst->{$cnct_clm};
	    if (defined $ln_cw_prb) {
		$incr_ln_prb_sbst += $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next;
	    }

	    my $cw_prb = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$clm}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb) {
		print STDERR "[ERROR 'shift_bl_and_compt_prob_incr_brief'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED for the column AFTER the \"shift\" (concatenation of the resulting column is: '$cnct_clm'): $! !\n\n";
		return (undef, undef, undef);
	    }

	    	# $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

	    $ln_cw_prb = log ($cw_prb);

	    $incr_ln_prb_sbst += $ln_cw_prb;
	    $clm2ln_prob_sbst->{$cnct_clm} = $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

	} # End of the 3rd foreach-loop (over relevant columns).


	#  ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

    (print STDOUT "\n< Subroutine 'shift_bl_and_compt_prob_incr_brief' Ended ", 'w/ ($incr_ln_prb_sbst, $incr_ct_null_clms) = ', "($incr_ln_prb_sbst, $incr_ct_null_clms) />\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 0); # Added on 2019/03/02.

	return (0, $incr_ln_prb_sbst, $incr_ct_null_clms);

}	# END of the subroutine 'shift_bl_and_compt_prob_incr_brief'. #



# shift_bl_and_compt_prob_incr:
#	This satellite subroutine of 'nonrec_mlt_shifts_noninf_gapblocks'
#	actually performs a shift of a specified gap-block,
#	updates @curr_set_clms, @curr_bds_blks, and @curr_bl_coords accordingly,
#	and computes the resulting increments of the log-probability and the count of null columns.
#	[CREATION STARTED on 2019/02/24, ENDED on 2019/03/01; REVISED On 2019/03/27.]
#	[On 2019/03/02, some major test outputs were added.]
#
# [ INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, $bl, $sh, \@rows_affd_by_gblks, $cnct_null_clm, \%clm2ln_prob_sbst, \@exnodes_in_msa, $char_abs, $topnode_id, \%node2ch, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where ] ... OBSOLETE as of 2019/03/27.
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gblks_w_spec_lb, \@gblks_w_spec_rb, $bl, $sh, \@rows_affd_by_gblks, $cnct_null_clm, \%clm2ln_prob_sbst, \@exnodes_in_msa, $char_abs, $topnode_id, \%node2ch, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where
#
#	@curr_set_clms is the current set of alignment columns.
#	@column = @{$curr_set_clms[$c]} specifies the $c th column. More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#
#	@{$curr_bds_blks[$bl]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $bl th gap-block;
#
#	$curr_bl_coords[$bl] gives the current coordinate of the $bl th gap-block;
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	$bl specifies the subject gap-block;
#
#	$sh specifies the direction of the "shift". More precisely,
#	$sh = +1 if the block will be "shift"ed to the right,
#	    = -1 if it will be "shift"ed to the left;
#
#	@{$rows_affd_by_gblks[$bl]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $bl th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column);
#
#	%clm2ln_prob_sbst = ( join (':', @column) => $log_prob_sbst, ...)
#	where $log_prob_sbst is the log-probability of the column @column under the given substitution model;
#
#	$exnodes_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
#
# OUTPUTS: ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms), where
#
#	$flag = 0 if all the process ended successfully,
#	      = -1 if the subject block stepped over the left-end of the local MSA,
#	      = +1 if the subject block stepped over the right-end of the local MSA,
#	      = undef if another type of error occurs;
#
#	$incr_ln_prb_sbst is the increment of the log-probability (under the given substitution model) resulting from the shift (= undef if an error occurs);
#
#	$incr_ct_null_clms is the increment of the count of null-columns resulting from the shift (= undef if an error occurs).
#
sub shift_bl_and_compt_prob_incr (\@\@\@\@\@$$\@$\%\@$$\%\%\@\%\%) {

	my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gblks_w_spec_lb, $gblks_w_spec_rb, $bl, $sh, $rows_affd_by_blks, $cnct_null_clm, $clm2ln_prob_sbst, $exnodes_in_msa, $char_abs, $topnode_id, $node2ch, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_;

	unless (($sh == +1) or ($sh == -1)) { # Added on 2019/03/01. #
	    print STDERR "[ERROR 'shift_bl_and_compt_prob_incr'] ILLEGAL INPUT: ", '$sh = ', "$sh: $! !\n\n";
	    return (undef, undef, undef);
	}

	my $ct_clms = @{$curr_set_clms};

	my $prev_bl_coord = $curr_bl_coords->[$bl]; # Keep the previous coordinate.

    (print STDOUT "< Subroutine 'shift_bl_and_compt_prob_incr' (#{columns} = $ct_clms, ", '$sh = ', $sh, ', $bl = ', $bl, ", {block coordinate}_bf = $prev_bl_coord) >\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 0); # Added on 2019/03/02.


	my $bds_block_sbj = $curr_bds_blks->[$bl];
	my ($prev_lbd, $prev_rbd) = @{$bds_block_sbj}; # Keep the boundaries of the subject block.


		# Actually shift the $bl th block on column in the $sh direction.

	my $curr_bl_coord = ($curr_bl_coords->[$bl] += $sh); # Shift the current coordinate.
	my $curr_lbd = ($bds_block_sbj->[0] += $sh); # Shift the left boundary.
	my $curr_rbd = ($bds_block_sbj->[1] += $sh); # Shift the right boundary.

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (1) (added on 2019/03/02)
	print STDOUT "\n[ Information on the subject block (index = $bl) ]\n\n";
	print STDOUT "{block-coordinate}_bf = $prev_bl_coord ,\n";
	print STDOUT "{block-coordinate}_af = $curr_bl_coord .\n\n";
	print STDOUT "(left_bound, right-bound)_bf = ($prev_lbd, $prev_rbd) ,\n";
	print STDOUT "(left_bound, right-bound)_af = ($curr_lbd, $curr_rbd) .\n\n";
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (1) ...}"


	if ($curr_lbd < 0) {
		# The subjet block stepped over the left-end of the local MSA.
	    return (-1, undef, undef);

	} elsif ($curr_rbd >= $ct_clms) {
		# The subjet block stepped over the right-end of the local MSA.
	    return (+1, undef, undef);
	}

		## ADDED in REVISION (on 2019/03/27). ##

		# Update @{$gblks_w_spec_lb}. #

	my @new_gblks_w_prev_lbd = ();
	foreach my $bl2 (@{$gblks_w_spec_lb->[$prev_lbd]}) {
	    if ($bl2 != $bl) { push @new_gblks_w_prev_lbd, $bl2; }
	}
	$gblks_w_spec_lb->[$prev_lbd] = \@new_gblks_w_prev_lbd;

	push @{$gblks_w_spec_lb->[$curr_lbd]}, $bl;

		# Update @{$gblks_w_spec_rb}. #
	my @new_gblks_w_prev_rbd = ();
	foreach my $bl2 (@{$gblks_w_spec_rb->[$prev_rbd]}) {
	    if ($bl2 != $bl) { push @new_gblks_w_prev_rbd, $bl2; }
	}
	$gblks_w_spec_rb->[$prev_rbd] = \@new_gblks_w_prev_rbd;

	push @{$gblks_w_spec_rb->[$curr_rbd]}, $bl;

		## END of "ADDED in REVISION (on 2019/03/27)." ##

		# Initialize the main outputs. #

	my $incr_ln_prb_sbst = my $incr_ct_null_clms = 0;

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3) { # (Added on 2019/03/02.)

	print STDOUT "\n[ Initial Values of the Main Outputs ]\n\n";
	print STDOUT '$incr_ln_prb_sbst = ', $incr_ln_prb_sbst, " ,\n";
	print STDOUT '$incr_ct_null_clms = ', $incr_ct_null_clms, " .\n\n";

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3) {...}"

		# Extract the columns that are to be modified. #

	my @rlv_clm_ids = ($sh>0) ? ($prev_lbd, $curr_rbd) : ($curr_lbd, $prev_rbd);
	my ($rlv_clm1, $rlv_clm2) = my @rlv_clms = @{$curr_set_clms}[@rlv_clm_ids];

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) (added on 2019/03/02)

	print STDOUT "\n[ The columns to be modified (BEFORE the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) ...}"

		# Compute the contributions from the columns BEFORE being modified. #

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (1) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (1): Compute the contributions from the columns BEFORE the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', '$incr_ct_null_clms', '$ln_cw_prb', '$incr_ln_prb_sbst'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (1) ...}"


	foreach my $clm (@rlv_clms) { # 1st foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms--;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, 0, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }
	    my $ln_cw_prb = $clm2ln_prob_sbst->{$cnct_clm};
	    if (defined $ln_cw_prb) {
		$incr_ln_prb_sbst -= $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next;
	    }

	    my $cw_prb = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$clm}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb) {
		print STDERR "[ERROR 'shift_bl_and_compt_prob_incr'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED for the column BEFORE the \"shift\" (concatenation of the pre-shift column is: '$cnct_clm'): $! !\n\n";
		return (undef, undef, undef);
	    }

	    	# $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

	    $ln_cw_prb = log ($cw_prb);

	    $incr_ln_prb_sbst -= $ln_cw_prb;
	    $clm2ln_prob_sbst->{$cnct_clm} = $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

	} # End of the 1st foreach-loop (over relevant columns).


		# MODIFY @{$curr_set_columns}.
		# More precisely, SWAP the portions of the relevant columns affected by the subject gap-block.
	my $affd_rows = $rows_affd_by_blks->[$bl];

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (2): MODIFY the columns as a result of the \"shift\". (rows affected: ", join (',', @{$affd_rows}), ") ]\n\n";

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) ...}"



	foreach my $row (@{$affd_rows}) { # 2nd foreach-loop (over affected rows).
	    my $tmp = $rlv_clm1->[$row];
	    $rlv_clm1->[$row] = $rlv_clm2->[$row];
	    $rlv_clm2->[$row] = $tmp;
	} # End of the 2nd foreach-loop (over affected rows).


   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2)

	print STDOUT "\n[ The columns modified (AFTER the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 2) { # (2) ...}"


		# Compute the contributions from the columns AFTER being modified. #

   if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) (added on 2019/03/02)

	print STDOUT "\n[ MAIN Process (3): Compute the contributions from the columns AFTER the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', '$incr_ct_null_clms', '$ln_cw_prb', '$incr_ln_prb_sbst'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 1) { # (3) ...}"


	foreach my $clm (@rlv_clms) { # 3rd foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms++;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, 0, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }
	    my $ln_cw_prb = $clm2ln_prob_sbst->{$cnct_clm};
	    if (defined $ln_cw_prb) {
		$incr_ln_prb_sbst += $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

		next;
	    }

	    my $cw_prb = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$clm}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb) {
		print STDERR "[ERROR 'shift_bl_and_compt_prob_incr'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED for the column AFTER the \"shift\" (concatenation of the resulting column is: '$cnct_clm'): $! !\n\n";
		return (undef, undef, undef);
	    }

	    	# $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

	    $ln_cw_prb = log ($cw_prb);

	    $incr_ln_prb_sbst += $ln_cw_prb;
	    $clm2ln_prob_sbst->{$cnct_clm} = $ln_cw_prb;

    (print STDOUT join ("\t", $cnct_clm, $incr_ct_null_clms, $ln_cw_prb, $incr_ln_prb_sbst), "\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 3);

	} # End of the 3rd foreach-loop (over relevant columns).


	#  ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

    (print STDOUT "\n< Subroutine 'shift_bl_and_compt_prob_incr' Ended ", 'w/ ($incr_ln_prb_sbst, $incr_ct_null_clms) = ', "($incr_ln_prb_sbst, $incr_ct_null_clms) />\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr > 0); # Added on 2019/03/02.

	return (0, $incr_ln_prb_sbst, $incr_ct_null_clms);

}	# END of the subroutine 'shift_bl_and_compt_prob_incr'. #


# shift_gblock:
#	This subroutine merely "shift"s a specified gap-block
#	(, and computes the increment of the count of null columns).
#	It was made from 'shift_bl_and_compt_prob_incr' above,
#	by DISABLING the operations concerning the computation of log-probabilities.
#	[CREATED on 2019/03/28.]
#
#	[[ Header of the predecessor:
# shift_bl_and_compt_prob_incr:
#	This satellite subroutine of 'nonrec_mlt_shifts_noninf_gapblocks'
#	actually performs a shift of a specified gap-block,
#	updates @curr_set_clms, @curr_bds_blks, and @curr_bl_coords accordingly,
#	and computes the resulting increments of the log-probability and the count of null columns.
#	[CREATION STARTED on 2019/02/24, ENDED on 2019/03/01; REVISED On 2019/03/27.]
#	[On 2019/03/02, some major test outputs were added.]
#	]] ... END of the "Header of the predecessor."
#
# [ INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gblks_w_spec_lb, \@gblks_w_spec_rb, $bl, $sh, \@rows_affd_by_gblks, $cnct_null_clm, \%clm2ln_prob_sbst, \@exnodes_in_msa, $char_abs, $topnode_id, \%node2ch, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where ] ... For the predecessor, 'shift_bl_and_compt_prob_incr'.
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gblks_w_spec_lb, \@gblks_w_spec_rb, $bl, $sh, \@rows_affd_by_gblks, $cnct_null_clm), where
#
#	@curr_set_clms is the current set of alignment columns.
#	@column = @{$curr_set_clms[$c]} specifies the $c th column. More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#
#	@{$curr_bds_blks[$bl]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $bl th gap-block;
#
#	$curr_bl_coords[$bl] gives the current coordinate of the $bl th gap-block;
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	$bl specifies the subject gap-block;
#
#	$sh specifies the direction of the "shift". More precisely,
#	$sh = +1 if the block will be "shift"ed to the right,
#	    = -1 if it will be "shift"ed to the left;
#
#	@{$rows_affd_by_gblks[$bl]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $bl th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column);
#
# OUTPUTS: ($flag, $incr_ct_null_clms), where
#
#	$flag = 0 if all the process ended successfully,
#	      = -1 if the subject block stepped over the left-end of the local MSA,
#	      = +1 if the subject block stepped over the right-end of the local MSA,
#	      = undef if another type of error occurs;
#
#	$incr_ct_null_clms is the increment of the count of null-columns resulting from the shift (= undef if an error occurs).
#
sub shift_gblock (\@\@\@\@\@$$\@$) {

	my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gblks_w_spec_lb, $gblks_w_spec_rb, $bl, $sh, $rows_affd_by_blks, $cnct_null_clm) = @_;

	unless (($sh == +1) or ($sh == -1)) { # Added on 2019/03/01. #
	    print STDERR "[ERROR 'shift_gblock'] ILLEGAL INPUT: ", '$sh = ', "$sh: $! !\n\n";
	    return (undef, undef);
	}

	my $ct_clms = @{$curr_set_clms};

	my $prev_bl_coord = $curr_bl_coords->[$bl]; # Keep the previous coordinate.

    (print STDOUT "< Subroutine 'shift_gblock' (#{columns} = $ct_clms, ", '$sh = ', $sh, ', $bl = ', $bl, ", {block coordinate}_bf = $prev_bl_coord) >\n\n") if ($TEST_LEVEL_sh_gblk > 0);


	my $bds_block_sbj = $curr_bds_blks->[$bl];
	my ($prev_lbd, $prev_rbd) = @{$bds_block_sbj}; # Keep the boundaries of the subject block.


		# Actually shift the $bl th block on column in the $sh direction.

	my $curr_bl_coord = ($curr_bl_coords->[$bl] += $sh); # Shift the current coordinate.
	my $curr_lbd = ($bds_block_sbj->[0] += $sh); # Shift the left boundary.
	my $curr_rbd = ($bds_block_sbj->[1] += $sh); # Shift the right boundary.

   if ($TEST_LEVEL_sh_gblk > 2) { # (1) 

	print STDOUT "\n[ Information on the subject block (index = $bl) ]\n\n";
	print STDOUT "{block-coordinate}_bf = $prev_bl_coord ,\n";
	print STDOUT "{block-coordinate}_af = $curr_bl_coord .\n\n";
	print STDOUT "(left_bound, right-bound)_bf = ($prev_lbd, $prev_rbd) ,\n";
	print STDOUT "(left_bound, right-bound)_af = ($curr_lbd, $curr_rbd) .\n\n";
   } # End of "if ($TEST_LEVEL_sh_gblk > 2) { # (1) ...}"


	if ($curr_lbd < 0) {
		# The subjet block stepped over the left-end of the local MSA.
	    return (-1, undef);

	} elsif ($curr_rbd >= $ct_clms) {
		# The subjet block stepped over the right-end of the local MSA.
	    return (+1, undef);
	}

		## ADDED in REVISION (on 2019/03/27). ##

		# Update @{$gblks_w_spec_lb}. #

	my @new_gblks_w_prev_lbd = ();
	foreach my $bl2 (@{$gblks_w_spec_lb->[$prev_lbd]}) {
	    if ($bl2 != $bl) { push @new_gblks_w_prev_lbd, $bl2; }
	}
	$gblks_w_spec_lb->[$prev_lbd] = \@new_gblks_w_prev_lbd;

	push @{$gblks_w_spec_lb->[$curr_lbd]}, $bl;

		# Update @{$gblks_w_spec_rb}. #
	my @new_gblks_w_prev_rbd = ();
	foreach my $bl2 (@{$gblks_w_spec_rb->[$prev_rbd]}) {
	    if ($bl2 != $bl) { push @new_gblks_w_prev_rbd, $bl2; }
	}
	$gblks_w_spec_rb->[$prev_rbd] = \@new_gblks_w_prev_rbd;

	push @{$gblks_w_spec_rb->[$curr_rbd]}, $bl;

		## END of "ADDED in REVISION (on 2019/03/27)." ##

		# Initialize the main outputs. #

	my $incr_ln_prb_sbst = my $incr_ct_null_clms = 0;

   if ($TEST_LEVEL_sh_gblk > 3) {

	print STDOUT "\n[ Initial Values of the Main Outputs ]\n\n";
	print STDOUT '$incr_ln_prb_sbst = ', $incr_ln_prb_sbst, " ,\n";
	print STDOUT '$incr_ct_null_clms = ', $incr_ct_null_clms, " .\n\n";

   } # End of "if ($TEST_LEVEL_sh_gblk > 3) {...}"

		# Extract the columns that are to be modified. #

	my @rlv_clm_ids = ($sh>0) ? ($prev_lbd, $curr_rbd) : ($curr_lbd, $prev_rbd);
	my ($rlv_clm1, $rlv_clm2) = my @rlv_clms = @{$curr_set_clms}[@rlv_clm_ids];

   if ($TEST_LEVEL_sh_gblk > 2) { # (2) 

	print STDOUT "\n[ The columns to be modified (BEFORE the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_gblk > 2) { # (2) ...}"

		# Compute the contributions from the columns BEFORE being modified. #

   if ($TEST_LEVEL_sh_gblk > 1) { # (1) 

	print STDOUT "\n[ MAIN Process (1): Compute the contributions from the columns BEFORE the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', 'null(Y)_or_not(N)', '$incr_ct_null_clms'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_gblk > 1) { # (1) ...}"


	foreach my $clm (@rlv_clms) { # 1st foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms--;

    (print STDOUT join ("\t", $cnct_clm, 'Y', $incr_ct_null_clms), "\n") if ($TEST_LEVEL_sh_gblk > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }

    (print STDOUT join ("\t", $cnct_clm, 'N', $incr_ct_null_clms), "\n") if ($TEST_LEVEL_sh_gblk > 3);

	} # End of the 1st foreach-loop (over relevant columns).


		# MODIFY @{$curr_set_columns}.
		# More precisely, SWAP the portions of the relevant columns affected by the subject gap-block.
	my $affd_rows = $rows_affd_by_blks->[$bl];

   if ($TEST_LEVEL_sh_gblk > 1) { # (3) 

	print STDOUT "\n[ MAIN Process (2): MODIFY the columns as a result of the \"shift\". (rows affected: ", join (',', @{$affd_rows}), ") ]\n\n";

   } # End of "if ($TEST_LEVEL_sh_gblk > 1) { # (3) ...}"



	foreach my $row (@{$affd_rows}) { # 2nd foreach-loop (over affected rows).
	    my $tmp = $rlv_clm1->[$row];
	    $rlv_clm1->[$row] = $rlv_clm2->[$row];
	    $rlv_clm2->[$row] = $tmp;
	} # End of the 2nd foreach-loop (over affected rows).


   if ($TEST_LEVEL_sh_gblk > 2) { # (2)

	print STDOUT "\n[ The columns modified (AFTER the \"shift\") ]\n\n";
	print STDOUT 'ID_clm', "\t", 'rows:constituting:the:column', "\n\n";
	for (my $k = 0; $k < 2; $k++) {
	    my $cnct_clm = join (':', @{$rlv_clms[$k]});
	    print STDOUT $rlv_clm_ids[$k], "\t", $cnct_clm, "\n";
	}
	print STDOUT "\n\n";
	 
   } # End of "if ($TEST_LEVEL_sh_gblk > 2) { # (2) ...}"


		# Compute the contributions from the columns AFTER being modified. #

   if ($TEST_LEVEL_sh_gblk > 1) { # (3) 

	print STDOUT "\n[ MAIN Process (3): Compute the contributions from the columns AFTER the \"shift\". ]\n\n";

	(print STDOUT join ("\t", 'rows:constituting:the:column', 'null(Y)_or_not(N)', '$incr_ct_null_clms'), "\n\n") if ($TEST_LEVEL_sh_bl_cmpt_p_incr> 3);

   } # End of "if ($TEST_LEVEL_sh_gblk > 1) { # (3) ...}"


	foreach my $clm (@rlv_clms) { # 3rd foreach-loop (over relevant columns).

	    my $cnct_clm = join (':', @{$clm});
	    if ($cnct_clm eq $cnct_null_clm) {
		$incr_ct_null_clms++;

    (print STDOUT join ("\t", $cnct_clm, 'Y', $incr_ct_null_clms), "\n") if ($TEST_LEVEL_sh_gblk > 3);

		next; # NOTE: the log-probability of a null-column is always 0 (ZERO) under any substitution model (in which each column evolves independently).
	    }

    (print STDOUT join ("\t", $cnct_clm, 'N', $incr_ct_null_clms), "\n") if ($TEST_LEVEL_sh_gblk > 3);

	} # End of the 3rd foreach-loop (over relevant columns).


	#  ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

    (print STDOUT "\n< Subroutine 'shift_gblock' Ended ", 'w/ ($incr_ct_null_clms = ', " $incr_ct_null_clms) />\n\n") if ($TEST_LEVEL_sh_gblk > 0);

	return (0, $incr_ct_null_clms);

}	# END of the subroutine 'shift_gblock'. #



# nonrec_mlt_shifts_noninf_gapblocks:
#	This simple version of the main subroutine, 'nonrec_mlt_shifts_gapblocks'.
#	performs multiple "shift"s of gap-blocks that are NOT interfering with any other.
#	It was created by exploiting the framework of the pilot subroutine, 'nonrec_explore_hyperrect'.
#	[CREATION STARTED on 2019/02/19, ENDED on 2019/03/06.]
#
#	[REVISED on 2020/04/15]:
#	Fixed a potentail BUG that causes the subroutine to behave abnormally if (#{coordinates} == 0).
#
#	NOTE: Because this subroutine deals only with NON-INTERFERING gap-blocks,
#	the topology of the gap-pattern will NEVER change.
#	Thus, the subroutine does NOT check topological changes.
#
# [Header of the pilot subroutine:
#
#  nonrec_explore_hyperrect:
#	This pilot subroutine of the main subroutine, 'nonrec_mlt_shifts_gapblocks',
#	implements a non-recursive method to explore a hyperrectangle
#	that is defined as a Cartesian product of a number of intervals, each of a coordinate.
#	[CREATION STARTED on 2019/02/18, ENDED on 2019/02/19.]
#
#	How to explore?:
#	Starting with the "origin", 
#	(i) it first moves to the right along each coordinate axis;
#	(ii) after reaching the "right-end", it returns to the "origin",
#	and then moves to the left along the same coordinate axis;
#	(iii) after examining the entire interval of the coordinate,
#	it then moves the "parent" coordinate by one;
#	(vi) repeat (i)-(iii) until the entire hyperrectangle is explored.
#
#	(See SM-2 of "suppl3_blueprint1_ANEX.draft9.odt" for details.)
#  ]
#
# [ INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@exnode_ids_in_msa, $char_abs, \@set_columns0, \@list_gblocks0, \%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%node2pa, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where ] ... OBSOLETE as of 2019/02/23.
# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@exnodeids_in_msa, $char_abs, \@set_columns0, \@bds_gblocks0, \@rows_affd_by_gblks, \%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%node2pa, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where
#
#	$org_bl_coords[$bl] gives the "origin" of the coordinate of the $bl th gap-block;
#	($left_bd, $right_bd) = @{$bds_bl_coords[$bl]} gives the left and right-boundaries
#	of the coordinate of the $bl th gap-block;
#	(NOTE that the boundaries are also included in the interval to be explored.)
#
#	$exnodeids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#	[ [Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyANEX_Supple.pm".] ] ... OBSOLETE as of 2019/06/27.
#	[Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyTreeMap_sppl.pm".] # Revised on 2019/06/27.
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#
#	@column = @{$set_columns0[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_columns0 is an input of, e.g., 'mk_init_set_columns0' in "MyANEX_Supple.pm". (... ADDED on 2019/02/27)]
#
#	[ @list_gblocks0 stores information on the gap-blocks. more precisely,
#	@{$list_gblocks0[$gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) are for the $gb th gap-block. Here,
#	$br is the ID of the branch separating the block,
#	$up_or_lw = 'U'/'L' if the block affects (i.e., puts gaps in) the upper/lower-side of $br,
#	$lm/rm_clm specifies the left/right-most column of the block,
#	@invlvd_gpbs stores indices (within the gapped-segment) of the gap-pattern blocks involved in the gap-block,
#	@affected_exnodes stores IDs of the "external" nodes afffected (given gaps) by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment that the gap-block belongs to,
#	$gb_in_gs is the index of the gap-block within the $gs th gapped segment;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm".] ] ... ADDED & OBSOLETE as of 2019/02/23.
#
#	@{$bds_gblocks0[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[\@bds_gblocks0 and \@rows_affd_by_gblks are outputs of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	%clm2ln_prob_sbst = ( join (':', @column) => $log_prob_sbst, ...)
#	where $log_prob_sbst is the log-probability of the column @column under the given substitution model;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#	%node2pa = ($node_id => $parent_id, ...) for the nodes other than the root;
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
#
# [ OUTPUTS: ($ln_prb_sbst0, \@diff_ln_prbs_sbst), where ] ... OBSOLETE as of 2019/02/24.
# OUTPUTS: ($ln_prb_sbst0, \@diffs_ln_prbs_sbst), where
#
#	$ln_prb_sbst0 is the log-probability of the initial local alignment under the given substitution model;
#
#	[ $diff_ln_prbs_sbst[$k_0]->...->[$k_B] = [$incr_ln_prob], where
#	$incr_ln_prob is the difference (from the initial alignment)    ] ... OBSOLETE as of 2019/02/24.
#	$diffs_ln_prbs_sbst[$k_0]->...->[$k_B] = [$diff_ln_prob, $ct_null_clms], where
#	$diff_ln_prob is the difference (from the initial alignment)
#	in the log-probability of the local alignment resulting from the
#	shifts of the 0 th gap-block by ($k_0 - $org_bl_coords[0]) columns,
#	..., and the B (= #{blocks} -1) by ($k_B - $org_bl_coords[B]) columns;
#	$ct_null_clms is the number of null-columns in the alignment.
#
#[ sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\%$\%\%\@\%\%) {
#	my ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $list_gblocks0, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; ] ... OBSOLETE as of 2019/02/23.
#
sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\@\%$\%\%\%\@\%\%) { # Modified on 2019/02/23.

	my ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, 
	$clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, 
	$base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

	my $ct_blks = @{$org_bl_coords};
	my $ct_rows = @{$exnodes_in_msa};
	my $ct_clms = @{$set_columns0};

    (print STDOUT "<< Subtoutine 'nonrec_mlt_shifts_noninf_gapblocks' (#{gap-blocks} = $ct_blks, #{rows} = $ct_rows, #{columns} = $ct_clms) >>\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks >0);

		# 1st consistency check. #

	if (@{$bds_bl_coords} != $ct_blks) {
	    my $ct_bd_pairs = @{$bds_bl_coords};
	    print STDERR "[ERROR 'nonrec_mlt_shifts_noninf_gapblocks'] The number of coordinate boundary-pairs (=$ct_bd_pairs) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (undef, undef);
	}

    if ($TEST_LEVEL_nr_msht_ninf_gblks >2) {
	print STDOUT "\n[ Input variables ]\n\n";
	print STDOUT join ("\t", 'Indx_bl', 'org', 'left_bd', 'right_bd'), "\n\n";
	for (my $k = 0; $k <$ct_blks; $k++) {
	    print STDOUT join ("\t", $k, $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks >2) {...}"


		# 2nd consistency check. #
	my @ill_blocks1 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 1st Preliminary for-loop (over blocks). #
	    if ($bds_bl_coords->[$k][0] < 0) { push @ill_blocks1, $k; }
	} # End of the 1st preliminary for-loop (over blocks). #
	if (@ill_blocks1>0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_noninf_gapblocks'] For the blocks, (", join (', ', @ill_blocks1), "), the left-ends are NEGATIVE: $! !\n\n";
	    return (undef, undef);
	}

		# 3rd consistency check. #
	my @ill_blocks2 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 2nd Preliminary for-loop (over blocks). #
	    my $org = $org_bl_coords->[$k];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$k]};
	    if (($org < $lbd) or ($rbd < $org)) { push @ill_blocks2, $k; }
	} # End of the 2nd preliminary for-loop (over blocks). #

	if (@ill_blocks2>0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_noninf_gapblocks'] For the blocks, (", join (', ', @ill_blocks2), "), the 'origin's are NOT within the intervals specified by the boundaries: $! !\n\n";
	    return (undef, undef);
	}

		## Preliminary Process (1): Prepare some important variables. ##

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }
	my $cnct_null_clm = join (':', @null_clm);

	unless (defined $clm2ln_prob_sbst->{$cnct_null_clm}) { $clm2ln_prob_sbst->{$cnct_null_clm} = 0; }
		# NOTE: The probability of a null column is always 1 (unity) under any substitution model (in which each column evolves independently).

	my $ct_null_clms0 = 0;
	my $ln_prb_sbst0 = 0;

	for (my $c = 0; $c < $ct_clms; $c++) { # The 3rd preliminary for-loop (over columns).

	    my $column0 = $set_columns0->[$c];
	    my $cnct_clm0 = join (':', @{$column0});
	    if ($cnct_clm0 eq $cnct_null_clm) {
		$ct_null_clms0++;
		next; # The log-probability of the null column is 0 (ZERO).
	    }

		# Examine whether the probability of the column is already computed or not.

	    my $ln_cw_prb0 = $clm2ln_prob_sbst->{$cnct_clm0};
	    if (defined $ln_cw_prb0) {
		$ln_prb_sbst0 += $ln_cw_prb0;
		next;
	    }

		# Newly compute the probability of the column. #

	    my $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb0) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_noninf_gapblocks'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED on the $c th column, whose concatenation is '$cnct_clm0': $! !\n\n";
		return (undef, undef);
	    }

	    $ln_cw_prb0 = log ($cw_prb0);

	    $clm2ln_prob_sbst->{$cnct_clm0} = $ln_cw_prb0;
	    $ln_prb_sbst0 += $ln_cw_prb0;

	} # End of the 3rd preliminary for-loop (over columns).
	

		## Preliminary Process (2): Initialize some important variables. ##

	my @bl_shifts = (); # The $k th element specifies the "next" shift (the direction and the amount) of the $k th block. (= +1 if to the right, -1 if to the left.)
	for (1 .. $ct_blks) { push @bl_shifts, 1; }

	my @bl_coords0 = @{$org_bl_coords}; # Initial values of the block coordinates.

		# MODIFIED on 2019/02/24. (1a) #
		# => A BUG Fixed on 2019/03/06.
	my @layer0 = ($set_columns0, $bds_blocks0, \@bl_coords0, 0, $ct_null_clms0);
	##my @layer0 = ($set_columns0, $bds_blocks0, \@bl_coords0, $ln_prb_sbst0, $ct_null_clms0);
	##my @layer0 = (\@bl_coords0);

	my @stack = ();
	for (0 .. $ct_blks) { # The 5th preliminary for-loop (to construct the stack).

		# MODIFIED on 2019/02/24. (1b) #

		# A BUG fixed on 2019/03/06.
	    my ($set_clms0, $bds_blks0, $bl_coords0, $diff_ln_prb_sbst, $ct_null_clms0) = @layer0;
	    ##my ($set_clms0, $bds_blks0, $bl_coords0, $ln_prb_sbst0, $ct_null_clms0) = @layer0;
	    my @cp_set_clms0 = ();
	    foreach my $clm (@{$set_clms0}) {
		my @cp_clm = @{$clm};
		push @cp_set_clms0, \@cp_clm;
	    }
	    my @cp_bds_blks0 = ();
	    foreach my $bds_blk0 (@{$bds_blks0}) {
		my @cp_bds_blk0 = @{$bds_blk0};
		push @cp_bds_blks0, \@cp_bds_blk0;
	    }
	    my @cp_bl_coords0 = @{$bl_coords0};

		# A BUG fixed on 2019/03/06.
	    my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $diff_ln_prb_sbst, $ct_null_clms0);
	    ##my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $ln_prb_sbst0, $ct_null_clms0);

	    ##my ($bl_coords0) = @layer0;
	    ##my @cp_bl_coords0 = @{$bl_coords0};
	    ##my @cp_layer0 = (\@cp_bl_coords0);

	    push @stack, \@cp_layer0;
	} # End of the 5th preliminary for-loop (to construct the stack).

	my $ct_layers = @stack;


    if ($TEST_LEVEL_nr_msht_ninf_gblks >2) {

	print STDOUT "[[ 1st Output ]]\n\n";

	print STDOUT '$ln_prb_sbst0 = ', $ln_prb_sbst0, " .\n\n\n";

	print STDOUT "[[ Initialized variables ]]\n\n";

	print STDOUT '$ct_null_clms0 = ', $ct_null_clms0 , " .\n\n";

	print STDOUT '@bl_shifts  = (', join (', ', @bl_shifts), "),\n\n";
	print STDOUT '@bl_coords0 = (', join (', ', @bl_coords0), ").\n\n";

	if ($TEST_LEVEL_nr_msht_ninf_gblks >3) {

	    print STDOUT "\n";
	    print STDOUT '{ Contents of @stack ', "(#{layers) = $ct_layers) }:\n\n";

	    for (my $ly = 0; $ly < $ct_layers; $ly++) {
		print STDOUT "| The $ly th layer |\n\n";
		my $layer = $stack[$ly];
		my ($set_clms_ly, $bds_blks_ly, $coords_ly, $ln_prb_sbst_ly, $null_clms_ly) = @{$layer}; # Modified on 2019/02/24.
		##my ($coords_ly) = @{$layer};
		print STDOUT '@{$coords_ly} = (', join (', ', @{$coords_ly}), "),\n\n";
		print STDOUT "\n\n";
	    }

	} # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks >3) {...}"

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks >2) {...}"

    if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {

	print STDOUT '(( Contents of %{$clm2ln_prob_sbst} BEFORE ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'concatenated:site:numbers:within:column', 'ln_prob_sbst'), "\n\n";
	foreach my $clm (sort keys %{$clm2ln_prob_sbst}) {
		print STDOUT join ("\t", $clm, $clm2ln_prob_sbst->{$clm}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {...}"


	my $diffs_ln_prbs_sbst = initialize_multidim_storage (@{$bds_bl_coords});

		# Added on 2019/02/24. #
	my $storage_pnt0 = $diffs_ln_prbs_sbst;
	for (my $bl = 0; $bl < $ct_blks; $bl++) {
	    my $coord = $bl_coords0[$bl];
	    $storage_pnt0 = $storage_pnt0->[$coord];
	}
		# Further modified on 2019/02/26.
	$storage_pnt0->[0] = 0;
	$storage_pnt0->[1] = $ct_null_clms0;
	#push @{$storage_pnt0}, 0, $ct_null_clms0; # MODIFIED on 2019/02/24.
	##push @{$storage_pnt0}, $ln_prb_sbst0, $ct_null_clms0;
		# END of "Added on 2019/02/24." #

	if ($ct_blks == 0) {
			# ADDED to fix a potential BUG on 2020/04/15. #
	    print STDOUT "[INFORMATION 'nonrec_mlt_shifts_noninf_gapblocks'] The number of gap-blocks is ZERO!! ... I will end this subroutine withOUT going into the MAIN outer while-loop ...\n\n";
	    return ($ln_prb_sbst0, $diffs_ln_prbs_sbst);
	}

		### The Main Process ###

    (print STDOUT "[[ The Main Process ]]\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks >1);

	my $wrk_layer = pop @stack;

	my $bl = $ct_blks -1;

	my $counter = 0;
	my $ct_moves = 0; # Added on 2019/02/19.


# $TEST_LEVEL_nr_msht_ninf_gblks

	#  ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

	while (1) { # The MAIN outer while-loop (for exploring the entire hyper-rectangle). #

	    $counter++;

    (print STDOUT "(( The $counter th point visited; ", '$bl =', " $bl ))\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks >1);

		# MODIFIED on 2019/02/24. (2) #
	    my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms) = @{$wrk_layer};
	    # my ($curr_bl_coords) = @{$wrk_layer};
	    my $sh = $bl_shifts[$bl];

	    my $bl_coord = $curr_bl_coords->[$bl];
	    my $new_bl_coord = $bl_coord + $sh;
	    my ($lbd_bl_coord, $rbd_bl_coord) = @{$bds_bl_coords->[$bl]};


    if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {

	print STDOUT "\n[ Data sets BEFORE the \"shift\": ]\n\n";

	print STDOUT '$curr_diff_ln_prb_sbst = ', $curr_diff_ln_prb_sbst, " ,\n";
	print STDOUT '$curr_ct_null_clms = ', $curr_ct_null_clms, " .\n\n";

	my $size_stack = @stack;

	print STDOUT '@{$curr_bl_coords} = (', join (', ', @{$curr_bl_coords}), ").\n\n";
	print STDOUT "{Next increment} = $sh ,\n";
	print STDOUT "{Stack size} = $size_stack .\n\n";

	print STDOUT '$bl_coord     = ', $bl_coord, " ,\n";
	print STDOUT '$new_bl_coord = ', $new_bl_coord, " ,\n";
	print STDOUT '($lbd_bl_coord, $rbd_bl_coord) = ', "($lbd_bl_coord, $rbd_bl_coord) .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {...}"

    if ($TEST_LEVEL_nr_msht_ninf_gblks >= 5) {

	print STDOUT '( Contents of @{$curr_bds_blks}: )', "\n\n";
	print STDOUT join ("\t", 'Indx_bl', 'left-end', 'right-end'), "\n\n";
	for (my $k = 0; $k < $ct_blks; $k++) {
	    print STDOUT join ("\t", $k , @{$curr_bds_blks->[$k]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks >= 5) {...}"


	    if ($rbd_bl_coord < $new_bl_coord) {

		## Stepping over the right-boundary.

			# MODIFIED on 2019/02/22.
			# => Further MODIFIED on 2019/02/24.

		my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms) = @{$stack[$#stack]};
		my @cp_set_clms = ();
		foreach my $clm (@{$src_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$src_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$src_bl_coords};
		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms);
		#my $cp_ln_prb_sbst = $src_ln_prb_sbst;
		#my $cp_ct_null_clms = $src_ct_null_clms;
		#my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $cp_ln_prb_sbst, $cp_ct_null_clms);
		##my ($src_bl_coords) = @{$stack[$#stack]};
		##my @cp_bl_coords = @{$src_bl_coords};
		##my @cp_layer = (\@cp_bl_coords);
		$wrk_layer = \@cp_layer;

		###my $copy = cp_comp_array (@{$stack[$#stack]});
		###$wrk_layer = $copy;

		$bl_shifts[$bl] = -1;

    (print STDOUT " ... It will go beyond the right-boundary of its coordinate !! ... Move back to the origin...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 2); # Added on 2019/02/19.

		next;

	    } elsif ($new_bl_coord < $lbd_bl_coord) {

		## Stepping over the left-boundary of the block coordinate.

		($bl == 0) and last; # End the exploration. #

		$wrk_layer = pop @stack;

		$bl_shifts[$bl] = +1;
		$bl--;

    (print STDOUT " ... It will go beyond the left-boundary of its coordinate !! ... Move back to the origin, and shift our forcus onto the 'parent' coordinate...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 2); # Added on 2019/02/19.

		next;
	    }


		## Within the range. ##

	    $ct_moves++; # Added on 2019/02/19.
    (print STDOUT " ... The $ct_moves th Move is likely WITHIN the INTENDED RANGE ... It will be attempted ...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 1); # Added on 2019/02/19, MODIFIED on 2019/03/03.


    if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {

	print STDOUT '[ Contents of @{$curr_set_clms} BEFORE the \"shift\": ]', "\n\n";

	print STDOUT join ("\t", 'Column/Row:', (0 .. $ct_rows -1) ), "\n\n";
	for (my $c = 0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$curr_set_clms->[$c]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {...}"

		## MODIFIED on 2019/02/24. (3) ##

		# Actually Update @{$curr_set_clms}, @{$curr_bds_blks}, and @{$curr_bl_coords},
		# & compute the increments of $curr_diff_ln_prb_sbst and $curr_ct_null_clms.

			# The subroutine was RENAMED (on 2019/03/27). #
	    my ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = shift_bl_and_compt_prob_incr_brief (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, $bl, $sh, @{$rows_affd_by_blks}, $cnct_null_clm, %{$clm2ln_prob_sbst}, @{$exnodes_in_msa}, $char_abs, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    ##my ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = shift_bl_and_compt_prob_incr (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, $bl, $sh, @{$rows_affd_by_blks}, $cnct_null_clm, %{$clm2ln_prob_sbst}, @{$exnodes_in_msa}, $char_abs, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

	    unless (defined $flag) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_noninf_gapblocks'] The subroutine 'shift_bl_and_compt_prob_incr_brief' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", while shifting the $bl th block in the $sh direction: $! !\n\n";
		return (undef, undef);
	    }
    (print STDOUT "\n .... The subroutine 'shift_bl_and_compt_prob_incr_brief' ended ...\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 2);
			# END of "The subroutine was RENAMED (on 2019/03/27)." #

    if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {

	print STDOUT "[ Outputs of 'shift_bl_and_compt_prob_incr_brief': ]\n\n"; # The subroutine was RENAMED (on 2019/03/27).
	##print STDOUT "[ Outputs of 'shift_bl_and_compt_prob_incr': ]\n\n";
	my @printed_out = ();
	foreach my $out ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) {
	    push @printed_out, (defined $out) ? $out : 'undef';
	}
	print STDOUT '($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = (', join (', ', @printed_out), ") .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {...}"


		# Added on 2019/02/26. (1) #

	    if ($flag > 0) { # The block stepped over the right-end of the window.

		# DO the same thing as the "if ($rbd_bl_coord < $new_bl_coord) {...}" block above. #

		my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms) = @{$stack[$#stack]};
		my @cp_set_clms = ();
		foreach my $clm (@{$src_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$src_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$src_bl_coords};
		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms);
		$wrk_layer = \@cp_layer;

		$bl_shifts[$bl] = -1;

    (print STDOUT " ... It stepped over the right-end of the window !! ... Move back to the origin...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 2);

		next;

	    } elsif ($flag < 0) { # The block stepped over the left-end of the window.

		# DO the same thing as the "elsif ($new_bl_coord < $lbd_bl_coord) {...}" block above. #

		($bl == 0) and last; # End the exploration. #

		$wrk_layer = pop @stack;

		$bl_shifts[$bl] = +1;
		$bl--;

    (print STDOUT " ... It stepped over the left-end of the window!! ... Move back to the origin, and shift our forcus onto the 'parent' coordinate...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 2);

		next;
	    }

    (print STDOUT " ... The increments for the $ct_moves th Move has been SUCCESSFULLY calculated !!\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 1);

		# End of "Added on 2019/02/26. (1)" #

		# Now that the main 3 arrays has been successfully updated,
		# update the log-probability and the count of null columns as well.

	    $curr_diff_ln_prb_sbst = ($wrk_layer->[3] += $incr_ln_prb_sbst);
	    $curr_ct_null_clms = ($wrk_layer->[4] += $incr_ct_null_clms);


		# Added on 2019/02/26. (2) #

	    my $storage_pnt = $diffs_ln_prbs_sbst;
	    for (my $k = 0; $k < $ct_blks; $k++) {
		my $coord = $curr_bl_coords->[$k];
		$storage_pnt = $storage_pnt->[$coord];
	    }
	    $storage_pnt->[0] = $curr_diff_ln_prb_sbst;
	    $storage_pnt->[1] = $curr_ct_null_clms;

		# End of "Added on 2019/02/26. (2)" #


	    # ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms) = @{$wrk_layer};

	    ##	# Actually update the current coordinate. #
	    ##
	    ## $curr_bl_coords->[$bl] = $new_bl_coord;

		## END of "MODIFIED on 2019/02/24. (3)" ##

    if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {

	print STDOUT "\n[ Data sets AFTER the \"shift\": ]\n\n";

	print STDOUT '$curr_diff_ln_prb_sbst = ', $curr_diff_ln_prb_sbst, " ,\n";
	print STDOUT '$curr_ct_null_clms = ', $curr_ct_null_clms, " .\n\n";

	print STDOUT '@{$curr_bl_coords} = (', join (', ', @{$curr_bl_coords}), ").\n\n\n";
    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 3) {...}"

    if ($TEST_LEVEL_nr_msht_ninf_gblks >= 5) {

	print STDOUT '( Contents of @{$curr_bds_blks}: )', "\n\n";
	print STDOUT join ("\t", 'Indx_bl', 'left-end', 'right-end'), "\n\n";
	for (my $k = 0; $k < $ct_blks; $k++) {
	    print STDOUT join ("\t", $k , @{$curr_bds_blks->[$k]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks >= 5) {...}"

    if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {

	print STDOUT '[ Contents of @{$curr_set_clms} AFTER the \"shift\": ]', "\n\n";

	print STDOUT join ("\t", 'Column/Row:', (0 .. $ct_rows -1) ), "\n\n";
	for (my $c = 0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$curr_set_clms->[$c]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {...}"


		# Raise $bl all the way to ($ct_blks -1).

	    my $ct_cps = 0;
	    while ($bl < $ct_blks -1) { # 1st inner while-loop. #
		$bl++;

		$ct_cps++;
	    }  # End of the 1st inner while-loop. #

		# Stuff the stack with copies of the newly computed layer.

	    while ($ct_cps>0) { # 2nd inner while-loop. #
		$ct_cps--;

			# MODIFIED on 2019/02/24. (5) #
		##my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_ln_prb_sbst, $src_ct_null_clms) = @{$wrk_layer}; # This is probably unnecessary!!
		my @cp_set_clms = ();
		foreach my $clm (@{$curr_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$curr_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$curr_bl_coords};
		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms);
		#my $cp_ln_prb_sbst = $src_ln_prb_sbst;
		#my $cp_ct_null_clms = $src_ct_null_clms;
		#my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $cp_ln_prb_sbst, $cp_ct_null_clms);
		##my ($src_bl_coords) = @{$wrk_layer};
		##my @cp_bl_coords = @{$src_bl_coords};
		##my @cp_layer = (\@cp_bl_coords);

			# END of "MODIFIED on 2019/02/24. (5)" #

		push @stack, \@cp_layer;

		#### my $cp_layer = cp_comp_array (@{$wrk_layer});
		#### push @stack, $cp_layer;

	    }  # End of the 2nd inner while-loop. #


	} # End of the MAIN outer while-loop (for exploring the entire hyper-rectangle). #

    (print STDOUT "\n ... ALL the \"shift\"s Attempted ...\n\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks > 1);

    if ($TEST_LEVEL_nr_msht_ninf_gblks2 > 1) { # Modified on 2019/06/06. #
    ##if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {

	print STDOUT '(( Contents of %{$clm2ln_prob_sbst} AFTER ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'concatenated:site:numbers:within:column', 'ln_prob_sbst'), "\n\n";
	foreach my $clm (sort keys %{$clm2ln_prob_sbst}) {
		print STDOUT join ("\t", $clm, $clm2ln_prob_sbst->{$clm}), "\n";
	}
	print STDOUT "\n\n";

    ##} # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks > 5) {...}"
    } # End of "if ($TEST_LEVEL_nr_msht_ninf_gblks2 > 1) { # Modified on 2019/06/06. # ...}"


    (print STDOUT "<< Subtoutine 'nonrec_mlt_shifts_noninf_gapblocks' Ended (#{visits} = $counter, #{moves} = $ct_moves) />>\n\n") if ($TEST_LEVEL_nr_msht_ninf_gblks >0); # Modified on 2019/02/19.


	return ($ln_prb_sbst0, $diffs_ln_prbs_sbst);

}	# End of the subroutine 'nonrec_mlt_shifts_noninf_gapblocks'. #



# nonrec_mlt_shifts_gapblocks:
#	This MAIN subroutine performs multiple "shift"s of gap-blocks,
#	which may in general interfere with one another.
#	It was created by substantially extending the simple version, 'nonrec_mlt_shifts_noninf_gapblocks',
#	with some additional (and advanced) functions.
#
#	MAJOR DIFFERENCES from 'nonrec_mlt_shifts_noninf_gapblocks':
#	(1) Before and after the "shift" of the subject gap-block,
#		this extended version attempts to extend or retract the boundaries of
#		the subject and/or interfering blocks, whenever necessary;
#	(2) This extended version computes the GAP-COMPOMENT of the probability of the alternative alignments;
#	(3) To keep the number of computations of the gap-component as small as possible,
#	    it examines whether the gap-pattern topology changed or not in accompany with each "shift";
#	(5) ...
#
#	[CREATION STARTED on 2019/05/19, ENDED on 2019/05/29.]
#
#	[MAJOR REVISION on 2019/05/27]:
#	The subroutine NOW examines whether or not each examined alignment 
#	contains some binary horizontal relationships that are PROHIBITED
#	(, probably to avoid redundantly counting the alignment).
#	
#	[REVISED on 2020/04/15]:
#	Fixed a potentail BUG that causes the subroutine to behave abnormally if (#{coordinates} == 0).
#
#	[MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30)]:
#	Now, the satellite subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm")
#	has been REPLACED with 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm").
#	In consequence,
#	this subroutine can NOW use log-multiplication factors of case-(i) "gapped segments",
#	so that it can take FULL advantage of the main outputs of the program package, "LASTPIECE".
#	Therefore, it should NOW be able to compute the probabilities of MSAs
#	MUCH MORE ACCURATELY than before the revision.
#
#	[REVISED on 2020/05/17]:
#	Now, before computing the log-probabilities,
#	the subroutine attempts to expand the list of "excluded" binary horizontal relationships
#	(via 'expand_list_x_bhrels' in "MyGapPatternTopology.pm"),
#	so that all the swappable equivalents of the relevant gap-blocks will be monitored.
#
#
# nonrec_mlt_shifts_noninf_gapblocks:
#	This simple version of the main subroutine, 'nonrec_mlt_shifts_gapblocks'.
#	performs multiple "shift"s of gap-blocks that are NOT interfering with any other.
#	It was created by exploiting the framework of the pilot subroutine, 'nonrec_explore_hyperrect'.
#	[CREATION STARTED on 2019/02/19, ENDED on 2019/03/06.]
#
#	NOTE: Because this subroutine deals only with NON-INTERFERING gap-blocks,
#	the topology of the gap-pattern will NEVER change.
#	Thus, the subroutine does NOT check topological changes.
#
# [Header of the pilot subroutine:
#
#  nonrec_explore_hyperrect:
#	This pilot subroutine of the main subroutine, 'nonrec_mlt_shifts_gapblocks',
#	implements a non-recursive method to explore a hyperrectangle
#	that is defined as a Cartesian product of a number of intervals, each of a coordinate.
#	[CREATION STARTED on 2019/02/18, ENDED on 2019/02/19.]
#
#	How to explore?:
#	Starting with the "origin", 
#	(i) it first moves to the right along each coordinate axis;
#	(ii) after reaching the "right-end", it returns to the "origin",
#	and then moves to the left along the same coordinate axis;
#	(iii) after examining the entire interval of the coordinate,
#	it then moves the "parent" coordinate by one;
#	(vi) repeat (i)-(iii) until the entire hyperrectangle is explored.
#
#	(See SM-2 of "suppl3_blueprint1_ANEX.draft9.odt" for details.)
#  ]
#
# [ INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@exnode_ids_in_msa, $char_abs, \@set_columns0, \@list_gblocks0, 
#	\%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%node2pa,
#	 \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where ] ... OBSOLETE as of 2019/02/23.
#
# [ INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@exnodeids_in_msa, $char_abs, \@set_columns0, \@bds_gblocks0,
#	 \@rows_affd_by_gblks, \%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%node2pa, \%node2depth, 
#	\@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where ] ... for the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'.
#
# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@gb_coords_wd_frm, \@exnodeids_in_msa, $char_abs, $char_pres, \@set_columns0, 
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \@bds_gblocks0, \@rows_affd_by_gblks, 
#	\%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%br2attr, \%node2pa, \%node2depth, \%node2ct_exoffs, \%br2eq, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx, 
#	\@set_ingredients_indel_hsts, @set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
#	\@list_excluded_bin_hrels), where # MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (from 2019/05/19 till 2019/05/30). (A)
#
#	$org_bl_coords[$bl] gives the "origin" of the coordinate of the $bl th gap-block;
#	($left_bd, $right_bd) = @{$bds_bl_coords[$bl]} gives the left and right-boundaries
#	of the coordinate of the $bl th gap-block;
#	(NOTE that the boundaries are also included in the interval to be explored.)
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks'. (0a) #
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window.
#	NOTE 2: $gb_coord_le_wd <= $lbd_coord <= $rbd_coord <= $gb_coord_re_wd MUST always hold.
#		(However, $gb_coord_le_wd may be negative.) 
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks'. (0a)" #
#
#	$exnodeids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#	[ [Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyANEX_Supple.pm".] ] ... OBSOLETE as of 2019/06/27.
#	[Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyTreeMap_sppl.pm".] # Revised on 2019/06/27.
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#	$char_pres is the character that represents a residue (of any type), or a "presence" state;
#
#	@column = @{$set_columns0[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_columns0 is an input of, e.g., 'mk_init_set_columns0' in "MyANEX_Supple.pm". (... ADDED on 2019/02/27)]
#
#	[
#	@list_gblocks0 stores information on the gap-blocks. more precisely,
#	@{$list_gblocks0[$gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) are for the $gb th gap-block. Here,
#	$br is the ID of the branch separating the block,
#	$up_or_lw = 'U'/'L' if the block affects (i.e., puts gaps in) the upper/lower-side of $br,
#	$lm/rm_clm specifies the left/right-most column of the block,
#	@invlvd_gpbs stores indices (within the gapped-segment) of the gap-pattern blocks involved in the gap-block,
#	@affected_exnodes stores IDs of the "external" nodes afffected (given gaps) by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment that the gap-block belongs to,
#	$gb_in_gs is the index of the gap-block within the $gs th gapped segment;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm".]
#	] ... ADDED & OBSOLETE as of 2019/02/23.
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0a) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0a)" #
#
#	@{$bds_gblocks0[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[\@bds_gblocks0 and \@rows_affd_by_gblks are outputs of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	%clm2ln_prob_sbst = ( join (':', @column) => $log_prob_sbst, ...)
#	where $log_prob_sbst is the log-probability of the column @column under the given substitution model;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#	%node2pa = ($node_id => $parent_id, ...) for the nodes other than the root;
#	%br2attr = ($br_id => \@attr, ...) for all branches, with 
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0b) #
#	%node2ct_exoffs = ($node_id => #{external offsprings}, ...) for ALL nodes;
#	[\%node2ct_exoffs is an output of, e.g., 'make_node2ct_exoffs' in "MyTreeManipu_III.pm"];
#
#	%br2eq = ($br_id => $equivalent_branch_id, ...) for branches with equivalent ones;
#	[\%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm"];
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0b)" #
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0c) #
#
#	(\%gpattern2indx, \@set_gpatterns, \@set_node2dparsstat, \@set_br2change) = @set_ingredients_indel_hsts, 
#	where
#	%gpattern2indx, @set_gpatterns, @set_node2dparsstat and @set_br2change, 
#	each grow from a null array (or hash) 
#	(, which means that, initially, ({}, [], [], []) = @set_ingredients_indel_hsts,)
#	by gradually incoporating newly encountered gap-patterns and the dollo parsimonious indel histories that can result in the patterns.
#	Their formats are as follows:
#
#	%gpattern2indx = ($gpattern => $indx_gp, ...) for the gap-patterns encountered thus far,
#	with $indx_gp assigned in the order of occurrences;
#	
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern in the input MSA, created by concatenating $char_pres/abs in the sequences;
#
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0c)" #
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23). (0d) #
#
#	($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs) = @set_ingredients_prb_msa_indel, where
#	$totrate_ins/del is the total rate of insertions/deletions (relative to the substitution rate),
#
#	$cutoff_ins/del is the cut-off length for insertions/deletions, which is common along the MSA;
#
#	$delta_del = Sum_{$l=1}^{$cutoff_del}($l-1)*distr_del ($l)/$nfac_del,
#	which can be calculated via 'cal_delta_exp_del_5p', is the excessive exponential factor
#	due to deletions extending out of the MSA boundaries;
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	$if_on_le/re_msa = 1 if the window we are dealing with is on the left/right-end of the input MSA,
#			 = 0 otherwise;
#
#		# END of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23). (0d)" #
#
		# ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (0f) #
#	@list_excluded_bin_hrels lists the PROHIBITED binary horizontal relationships,
#	which is a main input of 'chk_excluded_bin_hrels';
#	More precisely,
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined;
#	Currently, the following codes can possibly be in the list:
#	'-', indicating that the gap-blocks are "connected" (overlapping or touching);
#	'>', indicating that $gb1 horizontally includes $gb2;
#	'<', indicating that $gb1 is horizontally included in $gb2;
#	'=', indicating that $gb1 is horizontally identical to $gb2;
#	(NOTE that the relationship assignment is the opposite of that for the vertical relationships);
#	(See the header of 'encode_gptrn_topology_psm' for detailed descriptions of all codes);
		# END of "ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (0f)" #
#
# [ OUTPUTS: ($ln_prb_sbst0, \@diff_ln_prbs_sbst), where ] ... OBSOLETE as of 2019/02/24.
#
# [ # (For the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'.)
# OUTPUTS: ($ln_prb_sbst0, \@diffs_ln_prbs_sbst), where 
#	$ln_prb_sbst0 is the log-probability of the initial local alignment under the given substitution model;
#	$diff_ln_prbs_sbst[$k_0]->...->[$k_B] = [$incr_ln_prob], where
#	$incr_ln_prob is the difference (from the initial alignment)    ] ... OBSOLETE as of 2019/02/24.
#	[ $diffs_ln_prbs_sbst[$k_0]->...->[$k_B] = [$diff_ln_prob, $ct_null_clms], where
#	$diff_ln_prob is the difference (from the initial alignment)
#	in the log-probability of the local alignment resulting from the
#	shifts of the 0 th gap-block by ($k_0 - $org_bl_coords[0]) columns,
#	..., and the B (= #{blocks} -1) by ($k_B - $org_bl_coords[B]) columns;
#	$ct_null_clms is the number of null-columns in the alignment.
#
# ] # End of "(For the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'.)"
#
#		# MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26-27). (0e) #
#
# OUTPUTS: ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, \@diffs_ln_prbs_plus), where
#
#	$if_success	= 1 if the subroutine ended "successfully", 
#			= 0 otherwise;
#	$ct_null_clms0 is the number of null columns in the initial local alignment;
#	$degeneracy0 is the degree of degeneracy of the initial local alignment;
#	$minct_indels0 is the minimum number of indel events required to create the initial local alignment;
#	$ln_prb_msa0 (= $ln_prb_sbst0 + $ln_prb_indel0) is the total log-probability
#	of the initial local alignment under the given evolution model (with the degeneracy factor NOT subtracted);
#	$ln_prb_sbst0 is the log-probability of the initial local alignment under the given substitution model;
#	$ln_prb_indel0 is the log-probability of the initial local alignment under the given insertion/deletion model;
#
#
#	$diffs_ln_prbs_plus[$k_0]->...->[$k_B] = [$ct_null_clms, $degeneracy, $diff_minct_indels, $diff_ln_prb_msa, $diff_ln_prb_sbst, $diff_ln_prb_indel]
#	provide information on the local alignment resulting from the
#	shifts of the 0 th gap-block by ($k_0 - $org_bl_coords[0]) columns,
#	..., and the B (= #{blocks} -1) by ($k_B - $org_bl_coords[B]) columns.
#	Here,
#	$ct_null_clms is the number of null columns in the alignment,
#	$degeneracy is the degree of degeneracy of the alignment,
#	$diff_minct_indels is the difference of the minimum number of indel events (compared to $minct_indels0),
#	$diff_ln_prb_msa (= $dff_ln_prb_sbst + $diff_ln_prb_indel) is the difference of the total log-probability
#	of the alignment (with the degeneracy factor NOT subtracted) compared to $ln_prb_msa0,
#	$diff_ln_prb_sbst is the difference of the log-probability of the alignment under the given substitution model (compared to $ln_prb_sbst0),
#	$diff_ln_prb_indel is the difference of the log-probability of the alignment under the given indel model (compared to $ln_prb_indel0).
#
#		# END of "MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26-27). (0e)" #
#
#[ sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\%$\%\%\@\%\%) {
#	my ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $list_gblocks0,
#	$clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, 
#	$base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; ] ... OBSOLETE as of 2019/02/23.
#
# [ sub nonrec_mlt_shifts_noninf_gapblocks (\@\@\@$\@\@\@\%$\%\%\%\@\%\%) { # Modified on 2019/02/23.
#	my ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, 
#	$bds_blocks0, $rows_affd_by_blks, 
#	$clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, 
#	$base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23. ] ... for the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'.
#

sub nonrec_mlt_shifts_gapblocks (\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@) { # MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (from 2019/05/19 till 2019/05/27). (B)

	my ($org_bl_coords, $bds_bl_coords, $gb_coords_wd_frm, $exnodes_in_msa, $char_abs, $char_pres, $set_columns0,
	 $list_gblocks0, $indx_br, $indx_u_or_d, $bds_blocks0, $rows_affd_by_blks,
	 $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $br2attr, $node2depth, $node2ct_exoffs, $br2eq, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx, 
	$set_ingredients_indel_hsts, $set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
	$list_excluded_bin_hrels) = @_; # MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (from 2019/05/19 till 2019/05/27). (C)

	my $ct_blks = @{$org_bl_coords};
	my $ct_rows = @{$exnodes_in_msa};
	my $ct_clms = @{$set_columns0};

	my $ct_x_bhrels = @{$list_excluded_bin_hrels}; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). #

    (print STDOUT "<< Subtoutine 'nonrec_mlt_shifts_gapblocks' (#{gap-blocks} = $ct_blks, #{rows} = $ct_rows, #{columns} = $ct_clms, #{excluded_binary_relations}=$ct_x_bhrels) >>\n\n") if ($TEST_LEVEL_nr_msht_gblks >0); # Modified on 2019/05/27.
    ##(print STDOUT "<< Subtoutine 'nonrec_mlt_shifts_gapblocks' (#{gap-blocks} = $ct_blks, #{rows} = $ct_rows, #{columns} = $ct_clms) >>\n\n") if ($TEST_LEVEL_nr_msht_gblks >0);

		# 1st consistency check. #

	if (@{$bds_bl_coords} != $ct_blks) {
	    my $ct_bd_pairs = @{$bds_bl_coords};
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The number of coordinate boundary-pairs (=$ct_bd_pairs) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}


	if (@{$gb_coords_wd_frm} != $ct_blks) {
	    my $ct_gb_coords = @{$gb_coords_wd_frm};
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The number of coordinate pairs corresponding to the window frame (=$ct_gb_coords) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

	if (@{$list_gblocks0} != $ct_blks) { # Added on 2019/05/22.
	    my $ct_blks2 = @{$list_gblocks0};
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The size of the list of gap-blocks (=$ct_blks2) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

    if ($TEST_LEVEL_nr_msht_gblks >2) {

	print STDOUT "\n[ Input variables: Contents of ", '@{$list_gblocks0}, @{$bds_blocks0},  @{$org_bl_coords}, @{$bds_bl_coords}, @{$gb_coords_wd_frm} & @{$rows_affd_by_blks}', " ]\n\n";
	##print STDOUT "\n[ Input variables ]\n\n";  # MODIFIED on 2019/06/02.

	print STDOUT join ("\t", 'Indx_bl', 'br', 'up(U)_or_lw(L)', 'lb_blk', 'rb_blk', 'org_gcoord', 'lb_gcoord', 'rb_gcoord', 'gb_coord_lbd_wd', 'gb_coord_rbd_wd', '[rows,affected]'), "\n\n"; # Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/06/02). #
	##print STDOUT join ("\t", 'Indx_bl', 'br', 'up(U)_or_lw(L)', 'org', 'left_bd', 'right_bd', 'gb_coord_lbd_wd', 'gb_coord_rbd_wd', '[rows,affected]'), "\n\n"; # Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28). #
	##print STDOUT join ("\t", 'Indx_bl', 'br', 'up(U)_or_lw(L)', 'org', 'left_bd', 'right_bd', 'gb_coord_lbd_wd', 'gb_coord_rbd_wd'), "\n\n"; # Modified in creation of 'nonrec_mlt_shifts_gapblocks'. #
	####print STDOUT join ("\t", 'Indx_bl', 'org', 'left_bd', 'right_bd', 'gb_coord_lbd_wd', 'gb_coord_rbd_wd'), "\n\n"; # Modified in creation of 'nonrec_mlt_shifts_gapblocks'. #
	######print STDOUT join ("\t", 'Indx_bl', 'org', 'left_bd', 'right_bd'), "\n\n";
	for (my $k = 0; $k <$ct_blks; $k++) {
	    my $cnct_rows_affd = join (',', @{$rows_affd_by_blks->[$k]});
	    print STDOUT join ("\t", $k, @{$list_gblocks0->[$k]}[$indx_br, $indx_u_or_d], @{$bds_blocks0->[$k]}, $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}, @{$gb_coords_wd_frm->[$k]}, "[$cnct_rows_affd]" ), "\n"; # Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/06/02). #
	    ##print STDOUT join ("\t", $k, @{$list_gblocks0->[$k]}[$indx_br, $indx_u_or_d], $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}, @{$gb_coords_wd_frm->[$k]}, "[$cnct_rows_affd]" ), "\n"; # Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28). #
	    ####print STDOUT join ("\t", $k, @{$list_gblocks0->[$k]}[$indx_br, $indx_u_or_d], $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}, @{$gb_coords_wd_frm->[$k]} ), "\n"; # Modified in creation of 'nonrec_mlt_shifts_gapblocks'. #
	    ####print STDOUT join ("\t", $k, $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}, @{$gb_coords_wd_frm->[$k]} ), "\n"; # Modified in creation of 'nonrec_mlt_shifts_gapblocks''. #
	    ######print STDOUT join ("\t", $k, $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}), "\n";
	}
	print STDOUT "\n\n";

		# Added on 2019/05/28. (T1) #

      if ($TEST_LEVEL_nr_msht_gblks >3) { # (T1a)

	print STDOUT '( Contents of @{$exnodes_in_msa} )', "\n\n";
	print STDOUT join ("\t", 'Indx_row', 'node_id'), "\n\n";
	for (my $r=0; $r<$ct_rows; $r++) {
	    print STDOUT join ("\t", $r, $exnodes_in_msa->[$r]), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$set_columns0} )', "\n\n";
	print STDOUT join ("\t", 'Column\Row', (0 .. $ct_rows-1)), "\n\n";
	for (my $c=0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$set_columns0->[$c]}), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_nr_msht_gblks >3) { # (T1a) ...}"

	print STDOUT "( Input Tree )\n\n";

	my %nodeid2rowid;
	for (my $r=0; $r<$ct_rows; $r++) {
	    my $nodeid = $exnodes_in_msa->[$r];
	    $nodeid2rowid{$nodeid} = $r;
	}

	unless ( output_tree ($topnode_id, %{$node2ch}, %nodeid2rowid, %{$br2attr}, *STDOUT, 1, 1) ) {
	    print STDOUT "[CAUTION 'nonrec_mlt_shifts_gapblocks'] The subroutine 'output_tree' FAILED on the input tree: $! !\n\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_nr_msht_gblks >3) { # (T1b)

	print STDOUT '( Contents of %{$node2depth}, %{$node2ct_exoffs} & %{$br2eq} )', "\n\n";
	print STDOUT join ("\t", 'Node_ID(=br_ID)', 'depth', '#{external_offsprings}', 'equiv_br_ID'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) {
	    my $depth = $node2depth->{$node};
	    my $ct_exoffs = $node2ct_exoffs->{$node};
	    my $eqbr = $br2eq->{$node};
	    unless (defined $eqbr) { $eqbr = 'n/a'; }
	    print STDOUT join ("\t", $node, $depth, $ct_exoffs, $eqbr), "\n";
	}

	print STDOUT "\n\n";

	print STDOUT '@{$base_freqs_root} = (', join (', ', @{$base_freqs_root}), ") .\n\n";

      } # End of "if ($TEST_LEVEL_nr_msht_gblks >3) { # (T1b) ...}"

	print STDOUT '( Contents of @{$set_ingredients_prb_msa_indel} )', "\n\n";

	print STDOUT '@{$set_ingredients_prb_msa_indel}', "\n";
	print STDOUT ' = ($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs)', "\n";
	print STDOUT '  = (', join (', ', @{$set_ingredients_prb_msa_indel}), ") .\n";
	print STDOUT "\n\n";

	print STDOUT '($if_on_le_msa, $if_on_re_msa) = ', "($if_on_le_msa, $if_on_re_msa) .\n";
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$list_excluded_bin_hrels} ', "(#{excld_bin_hrels}=$ct_x_bhrels) )\n\n";
	print STDOUT join ("\t", 'No.', 'gb1', 'gb2', 'hrel'), "\n\n";

	for (my $k=0; $k<$ct_x_bhrels; $k++) {
	    print STDOUT join ("\t", $k, @{$list_excluded_bin_hrels->[$k]}), "\n";
	}
	print STDOUT "\n\n\n";

		# End of "Added on 2019/05/28. (T1)" #

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >2) {...}"

		# 2nd consistency check. #
	my @ill_blocks1 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 1st Preliminary for-loop (over blocks). #
	    if ($bds_bl_coords->[$k][0] < 0) { push @ill_blocks1, $k; }
	} # End of the 1st preliminary for-loop (over blocks). #
	if (@ill_blocks1>0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] For the blocks, (", join (', ', @ill_blocks1), "), the left-ends are NEGATIVE: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

		# 3rd consistency check. #
	my @ill_blocks2 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 2nd Preliminary for-loop (over blocks). #
	    my $org = $org_bl_coords->[$k];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$k]};
	    if (($org < $lbd) or ($rbd < $org)) { push @ill_blocks2, $k; }
	} # End of the 2nd preliminary for-loop (over blocks). #

	if (@ill_blocks2>0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] For the blocks, (", join (', ', @ill_blocks2), "), the 'origin's are NOT within the intervals specified by the boundaries: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks'' (on 2019/05/19). #

		# 3rd consistench check. (B) #
	my @ill_blocks3 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 2nd Preliminary for-loop (over blocks). (B) #
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$k]};
	    my ($gcd_lbd_wd, $gcd_rbd_wd) = @{$gb_coords_wd_frm->[$k]};
	    if (($lbd < $gcd_lbd_wd) or ($gcd_rbd_wd < $rbd)) { push @ill_blocks3, $k; }
	} # End of the 2nd preliminary for-loop (over blocks). (B) #

	if (@ill_blocks3>0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] For the blocks, (", join (', ', @ill_blocks3), "), either or both of the block-coordinate boundaries STICK OUT of the window frame: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks'' (on 2019/05/19)." #

   (print STDOUT "\n ... The input data PASSED some brief consistency checks ... NOW, we will perform some Preliminary Processes ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 1); # Added on 2019/05/27.


		## Preliminary Process (0.5): Prepare some important variables. ##
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks', from 2019/05/20 till 2019/05/dd.)

    (print STDOUT "<  Preliminary Process (0.5): Prepare some important variables. >\n\n") if ($TEST_LEVEL_nr_msht_gblks >1); # Added on 2019/05/28.

	my ($le_wd, $re_wd) = (0, $ct_clms-1);
	my ($gblks_w_spec_lb0, $gblks_w_spec_rb0) = mk_lists_gblocks_w_spec_bounds ($le_wd, $re_wd, @{$bds_blocks0});
	unless (defined $gblks_w_spec_lb0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_lists_gblocks_w_spec_bounds' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28 (T2) #
	print STDOUT "[ Outputs of 'mk_lists_gblocks_w_spec_bounds': Contents of ", '@{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}', " ]\n\n";
	print STDOUT join ("\t", 'Indx_cl(=c)', '[gap,blocks,with,left-bound0=c]', '[gap,blocks,with,right-bound0=c]'), "\n\n";
	for (my $c=0; $c<$ct_clms; $c++) {
			# A BUG fixed on 2019/06/02. #
	     my $cnct_gblks_w_lb0_eq_c = join (',', @{$gblks_w_spec_lb0->[$c]});
	     my $cnct_gblks_w_rb0_eq_c = join (',', @{$gblks_w_spec_rb0->[$c]});
	     ##my $cnct_gblks_w_lb0_eq_c = join (',', @{$gblks_w_spec_lb0});
	     ##my $cnct_gblks_w_rb0_eq_c = join (',', @{$gblks_w_spec_rb0});
	     print STDOUT join ("\t", $c, "[$cnct_gblks_w_lb0_eq_c]", "[$cnct_gblks_w_rb0_eq_c]"), "\n";
	}
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28 (T2) # ...}"
    


	my ($inter_block_relations, $interfering_blocks, $interfering_blocksets, $cmpl_interfering_blocksets, $br2up_down2blocksets) = extensive_exam_inter_block_relations (@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, %{$node2depth});
	unless (defined $cmpl_interfering_blocksets) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'extensive_exam_inter_block_relations' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'extensive_exam_inter_block_relations' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.


	my $totct_interferences = 0;
		# A BUG fixed (on 2019/06/02). #
	foreach my $intrfring (@{$interfering_blocks}) {
	    if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}) /2; } # MODIFIED to take account of the double-counting (on 2019/06/02).
	    ## if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}) ; }
	} 
	foreach my $intrfring (@{$interfering_blocksets}) {
	    if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}); } 
	}
	####foreach my $intrfring (@{$interfering_blocks}) { $totct_interferences += @{$intrfring->[0]}; }
	##foreach my $intrfring (@{$interfering_blocksets}) { $totct_interferences += scalar (@{$intrfring->[0]}); }
		# END of "A BUG fixed (on 2019/06/02)." #
	
    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T3) #

	print STDOUT "[ Outputs of 'extensive_exam_inter_block_relations' (Total#{interferences}=$totct_interferences) ]\n\n";

	print STDOUT '( Contents of @{$inter_block_relations} )', "\n\n";
	print STDOUT join ("\t", 'Gb1\Gb2:', (0 .. $ct_blks-1)), "\n\n";
	for (my $gb1 = 0; $gb1 <$ct_blks; $gb1++) {
	    my @out_rels_w_gb1 = @{$inter_block_relations->[$gb1]};
	    for (my $gb2=0; $gb2<$ct_blks; $gb2++) {
		unless (defined $out_rels_w_gb1[$gb2]) { $out_rels_w_gb1[$gb2] = 'undef'; }
	    }
	    print STDOUT join ("\t", $gb1, @out_rels_w_gb1), "\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT '( Contents of @{$interfering_blocks} )', "\n\n";

	print STDOUT join ("\t", 'Gb', 'intrfring_block1(rel1)', 'intrfring_block2(rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($blocks, $rels) = @{$interfering_blocks->[$gb]};
	    unless (defined $blocks) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_intrfr_blks = @{$blocks};
	    my @out_blocks = ();
	    for (my $k=0; $k<$ct_intrfr_blks; $k++) {
		my $block = $blocks->[$k];
		my $rel = $rels->[$k];
		push @out_blocks, "$block($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_blocks), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$interfering_blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Gb', '[blocks,in,intrfring,set,1](rel1)', '[blocks,in,intrfring,set,2](rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($blocksets, $rels) = @{$interfering_blocksets->[$gb]};
	    unless (defined $blocksets) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_sets = @{$blocksets};
	    my @out_blocksets = ();
	    for (my $k=0; $k<$ct_sets; $k++) {
		my $blockset = $blocksets->[$k];
		my $cnct_blockset = join (',', @{$blockset});
		my $rel = $rels->[$k];
		push @out_blocksets, "[$cnct_blockset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_blocksets), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"


      if ($TEST_LEVEL_nr_msht_gblks > 5) {

	print STDOUT '( Contents of @{$cmpl_interfering_blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Gb', '[cmpl,blocks,in,intrfring,set,1](rel1)', '[cmpl,blocks,in,intrfring,set,2](rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($cblocksets, $rels) = @{$cmpl_interfering_blocksets->[$gb]};
	    unless (defined $cblocksets) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_csets = @{$cblocksets};
	    my @out_cblocksets = ();
	    for (my $k=0; $k<$ct_csets; $k++) {
		my $cblockset = $cblocksets->[$k];
		my $cnct_cblockset = join (',', @{$cblockset});
		my $rel = $rels->[$k];
		push @out_cblocksets, "[$cnct_cblockset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_cblocksets), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of %{$br2up_down2blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Br_ID', 'up(U)_or_down(L)', '[blocks,in,set,1]', '[blocks,in,set,2]', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocksets}) {
	    my $up_down2blocksets = $br2up_down2blocksets->{$br};
	    foreach my $u_or_d (sort keys %{$up_down2blocksets}) {
		my @out_blsets = ();
		foreach my $blset (@{$up_down2blocksets->{$u_or_d}}) {
		    my $cnct_blset =  join (',', @{$blset});
		    push @out_blsets, "[$cnct_blset]";
		}
		print STDOUT join ("\t", $br, $u_or_d, @out_blsets), "\n";
	    }
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_nr_msht_gblks > 5) {...}"

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T3) # ...}"


	my $gblock_sizes = mk_list_gblock_sizes (@{$bds_blocks0}, 0, 1, @{$inter_block_relations});
	unless ($gblock_sizes) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'mk_list_gblock_sizes' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_list_gblock_sizes' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5a) #

	print STDOUT "[ Output of 'mk_list_gblock_sizes' ]\n\n";
	print STDOUT '@{$gblock_sizes} = (', join (', ', @{$gblock_sizes}), ") .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5a) # ...}"


	my ($br2up_down2blocks, $coll_cmpl_gblocks, $gblock2coll_cmpl) = search_for_coll_cmpl_gblks (@{$list_gblocks0}, $indx_br, $indx_u_or_d, $topnode_id, %{$node2ch}, %{$node2depth}, @{$gblock_sizes});  
	unless (defined $gblock2coll_cmpl) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'search_for_coll_cmpl_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_coll_cmpl_gblocks = @{$coll_cmpl_gblocks};

    (print STDOUT " ... The subroutine 'search_for_coll_cmpl_gblks' Ended (w/ #{coll-cmpl-gblocks}=$ct_coll_cmpl_gblocks) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5b) #

	print STDOUT "[ Outputs of 'search_for_coll_cmpl_gblks' ]\n\n";

	print STDOUT '( Contents of %{$br2up_down2blocks} )', "\n\n";
	print STDOUT join ("\t", 'Br_ID', 'up(U)_or_down(L)', '[blocks,delimited,as,specified]'), "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocks}) {
	    my $up_down2blocks = $br2up_down2blocks->{$br};
	    foreach my $u_or_d (sort keys %{$up_down2blocks}) {
		my $cnct_blocks = join (',', @{$up_down2blocks->{$u_or_d}});
		print STDOUT join ("\t", $br, $u_or_d, "[$cnct_blocks]"), "\n";
	    }
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$coll_cmpl_gblocks} ', "(w/ #{elements}=$ct_coll_cmpl_gblocks) )\n\n";
	print STDOUT join ("\t", 'Indx_ccgb', '[gap,blocks,constituting,the,set]'), "\n\n";
	for (my $k=0; $k<$ct_coll_cmpl_gblocks; $k++) {
	     my $cnct_ccgb_set = join (',', @{$coll_cmpl_gblocks->[$k]});
	     print STDOUT join ("\t", $k, "[$cnct_ccgb_set]"), "\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT '( Contents of @{$gblock2coll_cmpl} )', "\n\n";
	print STDOUT join ("\t", 'Gb', '(indices,of,ccgbs,that,Gb,belongs,to)' ), "\n\n";

	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my $cnct_ccgbs = join (',', @{$gblock2coll_cmpl->[$gb]});
	    print STDOUT join ("\t", $gb, "($cnct_ccgbs)"), "\n";
	}
	print STDOUT "\n\n";
	
      } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5b) # ...}"

	my ($cts_vincluding, $cts_vincluded, $cts_onx) = cts_vincluding_vincluded_onx_gblks (@{$inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5c) #

	print STDOUT "[ Outputs of 'cts_vincluding_vincluded_onx_gblks' ]\n\n";
	print STDOUT '( Contents of @{$cts_vincluding}, @{$cts_vincluded} & @{$cts_onx} )', "\n\n";
	print STDOUT join ("\t", 'Gb', '#{vincluding}', '#{vincluded}', '#{onx}'), "\n\n";

	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    print STDOUT join ("\t", $gb, $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb]), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5c) # ...}"


	my $sets_vequiv_gblks = mk_sets_vert_equiv_gblks (%{$br2up_down2blocks}, %{$br2eq});
	unless (defined $sets_vequiv_gblks) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'mk_sets_vert_equiv_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_sets_vequiv_gblks = @{$sets_vequiv_gblks};

    (print STDOUT " ... The subroutine 'mk_sets_vert_equiv_gblks' Ended (w/ #{sets of v-equiv g-blocks}=$ct_sets_vequiv_gblks) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5d) #

	print STDOUT "[ Outputs of 'mk_sets_vert_equiv_gblks' ]\n\n";
	print STDOUT '( Contents of @{$sets_vequiv_gblks} )', "\n\n";
	print STDOUT join ("\t", 'Indx_set', '[v-equiv,gap,blocks,constituting,the,set]'), "\n\n";
	for (my $i=0; $i<$ct_sets_vequiv_gblks; $i++) {
	    my $cnct_set = join (',', @{$sets_vequiv_gblks->[$i]});
	    print STDOUT join ("\t", $i, "[$cnct_set]"), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5d) # ...}"


	my $sets_swappable_gblks = mk_sets_swappable_gblks (@{$sets_vequiv_gblks}, @{$gblock_sizes});
	unless (defined $sets_swappable_gblks) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'mk_sets_swappable_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_sets_swappable_gblks = @{$sets_swappable_gblks};

    (print STDOUT " ... The subroutine 'mk_sets_swappable_gblks' Ended (w/ #{sets of swappable g-blocks}=$ct_sets_swappable_gblks) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5e) #

	print STDOUT "[ Outputs of 'mk_sets_swappable_gblks' ]\n\n";
	print STDOUT '( Contents of @{$sets_swappable_gblks} )', "\n\n";
	print STDOUT join ("\t", 'Indx_set', '[swappable,gap,blocks,constituting,the,set]'), "\n\n";
	for (my $i=0; $i<$ct_sets_swappable_gblks; $i++) {
	    my $cnct_set = join (',', @{$sets_swappable_gblks->[$i]});
	    print STDOUT join ("\t", $i, "[$cnct_set]"), "\n";
	}
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5e) # ...}"


			# (ADDED to rectify a REDUNDANCY problem (on 2020/05/17).) #

	if (($ct_sets_swappable_gblks>0) and ($ct_x_bhrels>0)) {

			# Attempt to expand the list of excluded binary horizontal relationships. #

	    my $exp_list_excluded_bin_hrels = expand_list_x_bhrels (@{$list_excluded_bin_hrels}, @{$sets_swappable_gblks});
	    unless (defined $exp_list_excluded_bin_hrels) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'expand_list_x_bhrels' FAILED: $! !\n\n";
		return (0, undef, undef, undef, undef, undef, undef, undef);
	    }

	    my $exp_ct_x_bhrels = @{$exp_list_excluded_bin_hrels};

    (print STDOUT " ... The subroutine 'expand_list_x_bhrels' Ended (w/ (#{excld_bin_hrels}_AFTER_EXPANSION=$exp_ct_x_bhrels) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ;

	    if ($exp_ct_x_bhrels > $ct_x_bhrels) {

		##print STDOUT "[INFORMATION 'nonrec_mlt_shifts_gapblocks'] The list of binary horizontal relationships WILL BE EXPANDED (w/ #{excld_bin_hrels}_BEFORE_EXPANSION=$ct_x_bhrels,  #{excld_bin_hrels}_AFTER_EXPANSION=$exp_ct_x_bhrels)!!\n\n\ns";
		print STDOUT "[INFORMATION 'nonrec_mlt_shifts_gapblocks'] The list of binary horizontal relationships WILL BE EXPANDED (w/ #{excld_bin_hrels}_BEFORE_EXPANSION=$ct_x_bhrels,  #{excld_bin_hrels}_AFTER_EXPANSION=$exp_ct_x_bhrels)!!\n\n\n"; # CORRECTED on 2020/05/19.

			# Update the list. #

		$list_excluded_bin_hrels = $exp_list_excluded_bin_hrels;
		$ct_x_bhrels = @{$list_excluded_bin_hrels};

    if ($TEST_LEVEL_nr_msht_gblks >2) {

	print STDOUT '( Contents of EXPANDED @{$list_excluded_bin_hrels} ', "(#{excld_bin_hrels}_af_exp=$ct_x_bhrels) )\n\n";
	print STDOUT join ("\t", 'No.', 'gb1', 'gb2', 'hrel'), "\n\n";

	for (my $k=0; $k<$ct_x_bhrels; $k++) {
	    print STDOUT join ("\t", $k, @{$list_excluded_bin_hrels->[$k]}), "\n";
	}
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >2) {...}"

	    } # End of "if ($exp_ct_x_bhrels > $ct_x_bhrels) {

	} # END of "if (($ct_sets_swappable_gblks>0) and ($ct_x_bhrels>0)) {...}"

			# END of "(ADDED to rectify a REDUNDANCY problem (on 2020/05/17).)" #


	my @set_all_nodes = sort {$a <=> $b} keys %{$node2depth};

	my ($gpattern2indx, $set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$set_ingredients_indel_hsts};

	my ($set_cls_gblocks0, $min_totcts_indels_for_cls0, $sets_psm_hsts_for_cls0, $cts_indels0_for_cls0, $set_psm_hsts0_for_cls0)
	   = enum_psm_indel_scns_for_msa_hs_mrounds (%{$node2ch}, %{$node2pa}, $topnode_id, @{$exnodes_in_msa}, @{$set_columns0}, $char_abs, $char_pres, @set_all_nodes, %{$node2depth}, %{$node2ct_exoffs}, %{$gpattern2indx}, @{$set_gpatterns}, @{$set_node2dparsstat}, @{$set_br2change});
	unless (defined $set_cls_gblocks0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks0};

    (print STDOUT " ... The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' Ended (w/ #{gapped segments}=$ct_cls) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

		# Added in creation of ''nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	my $minct_indels0 = 0;
	foreach my $totct (@{$min_totcts_indels_for_cls0}) { $minct_indels0 += $totct; }
		# End of "Added in creation of ''nonrec_mlt_shifts_gapblocks' (on 2019/05/26)." #

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5f) #


      if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T5f') #

	my $ct_gpatterns0=@{$set_gpatterns};

	print STDOUT '[ INITIAL Contents of @{$set_gpatterns},', " one of the Main Effective Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds' (#{gap-patterns}=$ct_gpatterns0) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', 'gap-pattern'), "\n\n";

	for (my $i=0; $i<$ct_gpatterns0; $i++) {
	    print STDOUT join ("\t", $i, ${$set_gpatterns->[$i]}), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T5f') # ...}"

	print STDOUT "[ Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds' (#{gapped segments}=$ct_cls; Min[Tot#{indels}]=$minct_indels0) ]\n\n";

	print STDOUT "# NOTE: ... Here, we will OMIT displaying some main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel histories),\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL\"!!\n\n\n";

     if ($TEST_LEVEL_nr_msht_gblks >= 5) {
	print STDOUT '(( Contents of @{$set_cls_gblocks0}, @{$min_totcts_indels_for_cls0}, @{$sets_psm_hsts_for_cls0}, @{$cts_indels0_for_cls0}, & @{$set_psm_hsts0_for_cls0} ))', "\n\n";
     } else {
	print STDOUT '(( Contents of @{$set_cls_gblocks0}, @{$min_totcts_indels_for_cls0} & @{$cts_indels0_for_cls0} ))', "\n\n";
     } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) {...} else {...}".

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Test outer for-loop (over gapped segments).

	    my $set_gpbs = $set_cls_gblocks0->[$cl];
	    my $ct_gpbs = @{$set_gpbs};
	    
	    print STDOUT "||  The $cl th gapped segment (#{gap-pattern-blocks}=$ct_gpbs) ||\n\n";
	    print STDOUT '( Gap-pattern structure )', "\n\n";
	    print STDOUT join ("\t", 'Indx_block', 'indx_gap_pattern', 'length', 'beg', 'end'), "\n\n";
	    for (my $gpb=0; $gpb<$ct_gpbs; $gpb++) {
		print STDOUT join ("\t", $gpb, @{$set_gpbs->[$gpb]}), "\n";
	    }
	    print STDOUT "\n\n";

	    my $min_totct_indels = $min_totcts_indels_for_cls0->[$cl];
	    my $ct_indels0 = $cts_indels0_for_cls0->[$cl];

	    my $psm_hst0_for_cls = $set_psm_hsts0_for_cls0->[$cl];
	    my $psm_hsts_for_cls = $sets_psm_hsts_for_cls0->[$cl];
	    my $ct_psm_hsts = @{$psm_hsts_for_cls};

	    print STDOUT 'Min#{indels}   = ', $min_totct_indels, " ,\n";
	    print STDOUT '#{indels}_init = ', $ct_indels0, " ,\n";
	    print STDOUT '#{parsimonious indel histories} = ', $ct_psm_hsts, " .\n\n";


=DISABLED

		# NOTE: If you want to know details on the MAIN outputs (storing information on the indel histories), 
		# SWITCH ON "$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL"!!

     if ($TEST_LEVEL_nr_msht_gblks >= 5) {

	    my ($br2indels0, $node2blwise_gstat0) =  @{$psm_hst0_for_cls};

	    print STDOUT '| Initial candidate for the parsimonious indel history |', "\n\n";

	    print STDOUT '( Contents of %{$br2indels0} )', "\n\n";
	    print STDOUT join ("\t", 'Br_ID', '[stat_ue/stat_le/involved,blocks]_1', '[stat_ue/stat_le/involved,blocks]_2', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) {
		my $info_indels = $br2indels0->{$br};
		unless (defined $info_indels) { next; }
		my @out_indels = ();
		foreach my $info_indel (@{$info_indels}) {
		    my ($stat_ue, $stat_le, $invlvd_blks) = @{$info_indel};
		    unless (defined $stat_ue) { next; }
		    my $cnct_invlvd_blks = join (',', @{$invlvd_blks});
		    push @out_indels, "[$stat_ue/$stat_le/$cnct_invlvd_blks]";
		}
		print STDOUT join ("\t", $br, @out_indels), "\n";
	    }
	    print STDOUT "\n\n";

     } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) {...}"
=cut # END of "DISABLED"

	    print STDOUT "\n";

	} # End of the test outer for-loop (over gapped segments).


    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5f) # ...}"


	my ($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, $br2log_mfacs) = @{$set_ingredients_prb_msa_indel}; # INCORPORATED INTO THE INPUT SET of this subroutine itself (on 2019/05/23).
	my $size_wd = $ct_clms;

				# MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (1) #
	## [ my ($tot_ln_prob_gpattern0, $shared_fac10, $shared_fac20, $set_clwise_tot_ln_prb_gp0, $tot_div_tree0, $ct_ng_clms0, $setset_ln_prb_gp_hsts0, $set_rlv_indel_hst2corr_fac0) 
	##  = apprx1_tot_lnlk_gpattern_msa2_hs ] ... before the revision on 2020/04/30.
	##
	my ($tot_ln_prob_gpattern0, $shared_fac10, $shared_fac20, $set_clwise_tot_ln_prb_gp0, $tot_div_tree0, $ct_ng_clms0, $setset_ln_prb_gp_hsts0, $set_rlv_indel_hst2corr_fac0, $shared_fac30, $totct_inter_gapless_clm_positions0, $ct_ends_flnk_gapless_clms0) 
	= apprx1_wlp_tot_lnlk_gpattern_msa2_hs ($topnode_id, %{$br2attr}, %{$node2pa}, %{$node2ch}, $size_wd, $char_abs, @{$set_cls_gblocks0}, @{$set_psm_hsts0_for_cls0}, @{$sets_psm_hsts_for_cls0},
	   $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, %{$br2log_mfacs}, $if_on_le_msa, $if_on_re_msa); # ... MUST REVISE the INPUT SET of 'apprx1_tot_lnlk_gpattern_msa2_hs'!! (... DONE!!)
				# End of "MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (1)" #
	unless (defined $tot_ln_prob_gpattern0) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' FAILED: $! !\n\n";
	    return (0, undef, undef, $minct_indels0, undef, undef, undef, undef);
	}

	my $ln_prb_indel0 = $tot_ln_prob_gpattern0; # Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #

    (print STDOUT " ... The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' Ended (w/ \$ln_prb_indel0 = $ln_prb_indel0) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28. => Modified on 2020/04/30.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5g) #

	##print STDOUT "[ Outputs of 'apprx1_tot_lnlk_gpattern_msa2_hs' ]\n\n";
	print STDOUT "[ Outputs of 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' ]\n\n"; # Modified on 2020/04/30. #

	print STDOUT "# NOTE: ... Here, we will OMIT displaying some main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel probabilities),\n";
	##print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL\"!!\n\n\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs_wLP::TEST_LEVEL\"!!\n\n\n"; # Modified on 2020/04/30. #

	print STDOUT '$tot_ln_prob_gpattern0 = ', $tot_ln_prob_gpattern0, " .\n\n";

	print STDOUT '( Contents of @{$set_clwise_tot_ln_prb_gp0} )', "\n\n";

	print STDOUT join ("\t", 'Indx_cl', 'tot_ln_prb_gp'), "\n\n";
	for (my $cl=0; $cl<$ct_cls; $cl++) {
	    my $tot_ln_prb_gp = $set_clwise_tot_ln_prb_gp0->[$cl];
	    unless (defined $tot_ln_prb_gp) { $tot_ln_prb_gp = 'undef'; }
	    print STDOUT join ("\t", $cl, $tot_ln_prb_gp), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT "( Secondary outputs )\n\n";

	print STDOUT '$shared_fac10  = ', $shared_fac10, " ,\n";
	print STDOUT '$shared_fac20  = ', $shared_fac20, " ,\n";
	print STDOUT '$tot_div_tree0 = ', $tot_div_tree0, " ,\n";
	print STDOUT '$ct_ng_clms0   = ', $ct_ng_clms0, " .\n\n";

			# (ADDED on 2020/04/30.) #
	print STDOUT '$shared_fac30 = ', $shared_fac30, " ,\n";
	print STDOUT '$totct_inter_gapless_clm_positions0 = ', $totct_inter_gapless_clm_positions0, " ,\n";
	print STDOUT '$ct_ends_flnk_gapless_clms0  = ', $ct_ends_flnk_gapless_clms0 , " .\n\n";
			# End of "(ADDED on 2020/04/30.)" #

	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5g) # ...}"



		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks', from 2019/05/20 till 2019/05/dd.)"


		## Preliminary Process (1): Prepare some important variables. ##

    (print STDOUT "<  Preliminary Process (1): Prepare some important variables. >\n\n") if ($TEST_LEVEL_nr_msht_gblks >1); # Added on 2019/05/28.

   (print STDOUT "| First, compute the substitution component of the log-probability of the initial alignment |\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2);

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }
	my $cnct_null_clm = join (':', @null_clm);

	unless (defined $clm2ln_prob_sbst->{$cnct_null_clm}) { $clm2ln_prob_sbst->{$cnct_null_clm} = 0; }
		# NOTE: The probability of a null column is always 1 (unity) under any substitution model (in which each column evolves independently).

	my $ct_null_clms0 = 0;
	my $ln_prb_sbst0 = 0;

	for (my $c = 0; $c < $ct_clms; $c++) { # The 3rd preliminary for-loop (over columns).

	    my $column0 = $set_columns0->[$c];
	    my $cnct_clm0 = join (':', @{$column0});
	    if ($cnct_clm0 eq $cnct_null_clm) {
		$ct_null_clms0++;
		next; # The log-probability of the null column is 0 (ZERO).
	    }

		# Examine whether the probability of the column is already computed or not.

	    my $ln_cw_prb0 = $clm2ln_prob_sbst->{$cnct_clm0};
	    if (defined $ln_cw_prb0) {
		$ln_prb_sbst0 += $ln_cw_prb0;
		next;
	    }

		# Newly compute the probability of the column. #

	    my $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb0) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED on the $c th column, whose concatenation is '$cnct_clm0': $! !\n\n";
		return (0, undef, undef, $minct_indels0, undef, undef, $ln_prb_indel0, undef);
	    }

	    $ln_cw_prb0 = log ($cw_prb0);

    (print STDOUT " ... The subroutine 'br_calc_sbst_prob_sgl_clm' Ended on the $c th column (w/ \$ln_cw_prb0 = $ln_cw_prb0) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2) ; # Added on 2019/05/28.

	    $clm2ln_prob_sbst->{$cnct_clm0} = $ln_cw_prb0;
	    $ln_prb_sbst0 += $ln_cw_prb0;

	} # End of the 3rd preliminary for-loop (over columns).

	my $ln_prb_msa0 = $ln_prb_sbst0 + $ln_prb_indel0; # Added on 2019/05/27.
	
    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5g-h) #
	print STDOUT "\n";
	print STDOUT '$ct_null_clms0  = ', $ct_null_clms0, " ,\n";
	print STDOUT '$ln_prb_sbst0   = ', $ln_prb_sbst0, " ,\n";
	print STDOUT '($ln_prb_indel0 = ', $ln_prb_indel0, ") ,\n";
	print STDOUT '$ln_prb_msa0    = ', $ln_prb_msa0, " .\n";
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5g-h) # ...}"

		## Preliminary Process (2): Initialize some important variables. ##

	my @bl_shifts = (); # The $k th element specifies the "next" shift (the direction and the amount) of the $k th block. (= +1 if to the right, -1 if to the left.)
	for (1 .. $ct_blks) { push @bl_shifts, 1; }

	my @bl_coords0 = @{$org_bl_coords}; # Initial values of the block coordinates.


			## Compute some topology-related codes. ##
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23).) ##

	my ($degeneracy0, $cnct_sets_ccgbs_cs_dgnrcy0) = compt_aln_degeneracy (@{$coll_cmpl_gblocks}, @{$bds_blocks0}, @bl_coords0, @{$bds_bl_coords}, @{$inter_block_relations}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0});
	unless (defined $cnct_sets_ccgbs_cs_dgnrcy0) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'compt_aln_degeneracy' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, undef, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy0 = $degeneracy0)...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5h) #

	print STDOUT "[ Outputs of 'compt_aln_degeneracy' ]\n\n";
	print STDOUT '$degeneracy0 = ', $degeneracy0, " ,\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy0 = ', $cnct_sets_ccgbs_cs_dgnrcy0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5h) # ...}"


	my $code_topology0 = encode_gptrn_topology_psm (@{$bds_blocks0}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0});
	unless (defined $code_topology0) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'encode_gptrn_topology_psm' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'encode_gptrn_topology_psm' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5i) #

	print STDOUT "[ Output of 'encode_gptrn_topology_psm' ]\n\n";
	print STDOUT '$code_topology0 = ', $code_topology0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5i) # ...}"


	my @no_gbs_on_msa_ends = (); 
	for (my $k=0; $k<$ct_blks; $k++) { $no_gbs_on_msa_ends[$k] = 'N'; }
	my $cnct_no_gbs_on_msa_ends = join (',', @no_gbs_on_msa_ends); # Code indicating NO gap-blocks are on either MSA-end.

	my $if_gbs_on_msa_ends0 = ($if_on_le_msa or $if_on_re_msa) ?
	  examine_if_gbs_on_msa_ends (@bl_coords0, @{$gb_coords_wd_frm}, $if_on_le_msa, $if_on_re_msa) :
		 \@no_gbs_on_msa_ends ;
	unless (defined $if_gbs_on_msa_ends0) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'examine_if_gbs_on_msa_ends' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'examine_if_gbs_on_msa_ends' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

	my $cnct_if_gbs_on_msa_ends0 = join (',', @{$if_gbs_on_msa_ends0});

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5j) #

	print STDOUT "[ Output of 'examine_if_gbs_on_msa_ends' ]\n\n";
	print STDOUT '$cnct_if_gbs_on_msa_ends0 = ', $cnct_if_gbs_on_msa_ends0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5j) # ...}"

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23).)" ##

			## Initialize the hash storing pre-computed indel components of the log-probabilities of the alignment.
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) ##

	my $grand_code_topology0 = join ('::', $code_topology0, $cnct_sets_ccgbs_cs_dgnrcy0, $cnct_if_gbs_on_msa_ends0);
	my %gcode_topo2diff_ln_prb_indel_plus = ($grand_code_topology0 => [0, 0]); # [$diff_minct_indels, $diff_ln_prb_indel]
	##my %gcode_topo2diff_ln_prb_indel = ($grand_code_topology0 => 0);

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).)" ##

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5k) #

	print STDOUT '[ Initial key for  %gcode_topo2diff_ln_prb_indel_plus ]', "\n\n";
	print STDOUT '$grand_code_topology0  = ', $grand_code_topology0 , " .\n";
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5k) # ...}"

			## Re-assign ($degeneracy0 = 0), if the initial alignment satisfies some conditions. ##
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28).) ##

	if ($degeneracy0 >0) {

	    if ($ct_null_clms0>0) {
		# If the alignment contains some null-columns,
		# it MUST be excluded from the analysis.

    (print STDOUT " ... Because (\$ct_null_clms0 == 0), Set \$degeneracy = 0 !!\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.
		$degeneracy0 = 0;

	    } elsif ($ct_x_bhrels>0) {

		# NOTE: "(!($if_ovlp_bw_veq0) and $if_right_order_swpbl0)" MUST usually be satisfied.
		# 	Thus, we will NOT perform 'chk_ovlp_bw_vequiv_gblks' or 'chk_pos_orders_amg_swpbl_gblks'. # 

		my $if_excluded0 = chk_excluded_bin_hrels (@{$list_excluded_bin_hrels}, @{$bds_blocks0}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @{$inter_block_relations});
		unless (defined $if_excluded0) {
		    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'chk_excluded_bin_hrels' FAILED on the initial alignment: $! !\n\n";
		    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
		}
    (print STDOUT " ... The subroutine 'chk_excluded_bin_hrels' Ended (w/ \$if_excluded0 = $if_excluded0) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

		if ($if_excluded0) { $degeneracy0 = 0; } # The gap-block configuration indicated that the initial alignment MUST be excluded from the analysis.

	    } # End of "if ($ct_null_clms0>0) {...} elsif ($ct_x_bhrels>0) {...}".

	} # End of "if ($degeneracy0 >0) {...}".

    (print STDOUT "\n", '{FINAL VALUE of $degeneracy0} = ', $degeneracy0, " .\n\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/05/28. #

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28).)" ##


		# MODIFIED on 2019/02/24. (1a) #
		# => A BUG Fixed on 2019/03/06.
		# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		# => Yet Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	my @layer0 = ($set_columns0, $bds_blocks0, \@bl_coords0, 0, 0, 0, $ct_null_clms0, $gblks_w_spec_lb0, $gblks_w_spec_rb0,
		 $degeneracy0, $code_topology0, $cnct_sets_ccgbs_cs_dgnrcy0, $cnct_if_gbs_on_msa_ends0 );
	##my @layer0 = ($set_columns0, $bds_blocks0, \@bl_coords0, 0, 0, $ct_null_clms0, $gblks_w_spec_lb0, $gblks_w_spec_rb0);
	####my @layer0 = ($set_columns0, $bds_blocks0, \@bl_coords0, $ln_prb_sbst0, $ct_null_clms0);
	######my @layer0 = (\@bl_coords0);


	my @stack = ();
	for (0 .. $ct_blks) { # The 5th preliminary for-loop (to construct the stack).

		# MODIFIED on 2019/02/24. (1b) #

		# A BUG fixed on 2019/03/06.
		# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		# => Yet Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	    my ($set_clms0, $bds_blks0, $bl_coords0, $diff_ln_prb_sbst, $diff_ln_prb_indel, $diff_minct_indels, $ct_null_clms0, $gblks_w_spec_lb0, $gblks_w_spec_rb0, 
		$degeneracy, $code_topo, $cnct_sets_ccgbs_cs_dgnrcy, $cnct_if_gbs_on_msa_ends) = @layer0;
	    ##my ($set_clms0, $bds_blks0, $bl_coords0, $diff_ln_prb_sbst, $diff_ln_prb_indel, $ct_null_clms0, $gblks_w_spec_lb0, $gblks_w_spec_rb0) = @layer0;
	    ####my ($set_clms0, $bds_blks0, $bl_coords0, $diff_ln_prb_sbst, $ct_null_clms0) = @layer0;
	    ######my ($set_clms0, $bds_blks0, $bl_coords0, $ln_prb_sbst0, $ct_null_clms0) = @layer0;
	    my @cp_set_clms0 = ();
	    foreach my $clm (@{$set_clms0}) {
		my @cp_clm = @{$clm};
		push @cp_set_clms0, \@cp_clm;
	    }
	    my @cp_bds_blks0 = ();
	    foreach my $bds_blk0 (@{$bds_blks0}) {
		my @cp_bds_blk0 = @{$bds_blk0};
		push @cp_bds_blks0, \@cp_bds_blk0;
	    }
	    my @cp_bl_coords0 = @{$bl_coords0};

		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
	    my @cp_gblks_w_spec_lb0 = my @cp_gblks_w_spec_rb0 = ();
	    foreach my $gblks (@{$gblks_w_spec_lb0}) {
		my @cp_gblks = @{$gblks};
		push @cp_gblks_w_spec_lb0, \@cp_gblks;
	    }
	    foreach my $gblks (@{$gblks_w_spec_rb0}) {
		my @cp_gblks = @{$gblks};
		push @cp_gblks_w_spec_rb0, \@cp_gblks;
	    }
		# END of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20)." #

		# A BUG fixed on 2019/03/06.
		# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		# => Yet Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	    my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $diff_ln_prb_sbst, $diff_ln_prb_indel, $diff_minct_indels, $ct_null_clms0, \@cp_gblks_w_spec_lb0, \@cp_gblks_w_spec_rb0, 
		$degeneracy, $code_topo, $cnct_sets_ccgbs_cs_dgnrcy, $cnct_if_gbs_on_msa_ends);
	    ##my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $diff_ln_prb_sbst, $ct_null_clms0, \@cp_gblks_w_spec_lb0, \@cp_gblks_w_spec_rb0);
	    ####my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $diff_ln_prb_sbst, $ct_null_clms0);
	    ####my @cp_layer0 = (\@cp_set_clms0, \@cp_bds_blks0, \@cp_bl_coords0, $ln_prb_sbst0, $ct_null_clms0);

	    ##my ($bl_coords0) = @layer0;
	    ##my @cp_bl_coords0 = @{$bl_coords0};
	    ##my @cp_layer0 = (\@cp_bl_coords0);

	    push @stack, \@cp_layer0;

	} # End of the 5th preliminary for-loop (to construct the stack).

	my $ct_layers = @stack;

		# Newly Added variables: @{$gblks_w_spec_lb}, @{$gblks_w_spec_rb}, 


    if ($TEST_LEVEL_nr_msht_gblks >2) {

	print STDOUT "[[ 1st Output ]]\n\n";

	print STDOUT '$ln_prb_sbst0 = ', $ln_prb_sbst0, " .\n\n\n";

	print STDOUT "[[ Initialized variables ]]\n\n";

	print STDOUT '$ct_null_clms0 = ', $ct_null_clms0 , " .\n\n";

	print STDOUT '@bl_shifts  = (', join (', ', @bl_shifts), "),\n\n";
	print STDOUT '@bl_coords0 = (', join (', ', @bl_coords0), ").\n\n";

	if ($TEST_LEVEL_nr_msht_gblks >3) {

	   print STDOUT "\n";
	   print STDOUT '{ Contents of @stack ', "(#{layers) = $ct_layers) }:\n\n";

	   for (my $ly = 0; $ly < $ct_layers; $ly++) {

		print STDOUT "| The $ly th layer |\n\n";

		my $layer = $stack[$ly];

 			# Modified on 2019/02/24.
			# => Further Modified on 2019/05/27.
		my ($set_clms_ly, $bds_blks_ly, $bl_coords_ly, $diff_ln_prb_sbst_ly, $diff_ln_prb_indel_ly, 
			$diff_minct_indels_ly, $ct_null_clms_ly, $gblks_w_spec_lb_ly, $gblks_w_spec_rb_ly, 
		$degeneracy_ly, $code_topo_ly, $cnct_sets_ccgbs_cs_dgncry_ly, $cnct_if_gbs_on_msa_ends_ly) = @{$layer};
		##my ($set_clms_ly, $bds_blks_ly, $coords_ly, $ln_prb_sbst_ly, $null_clms_ly) = @{$layer};
		####my ($coords_ly) = @{$layer};

	     if (($TEST_LEVEL_nr_msht_gblks >5) or ($ly == 0)) { # (Modified on 2019/05/28) #
	     ##if ($TEST_LEVEL_nr_msht_gblks >5) {

		print STDOUT '@{$set_clms_ly} = ( # [residue,numbers,or,gaps,in,the,msa,rows]', "\n";
		foreach my $clm (@{$set_clms_ly}) { print STDOUT "\t[", join (',', @{$clm}), "],\n"; }
		print STDOUT "\t);\n\n";

		print STDOUT '@{$bds_blks_ly} = ( # [left_bound, right_bound]', "\n";
		foreach my $bds (@{$bds_blks_ly}) { print STDOUT "\t[", join (', ', @{$bds}), "],\n"; }
		print STDOUT "\t);\n\n";

		print STDOUT '@{$bl_coords_ly} = (', join (', ', @{$bl_coords_ly}), ");\n\n\n";
		print STDOUT '($diff_ln_prb_sbst_ly, $diff_ln_prb_indel_ly, $diff_minct_indels_ly, $ct_null_clms_ly) = ', "($diff_ln_prb_sbst_ly, $diff_ln_prb_indel_ly, $diff_minct_indels_ly, $ct_null_clms_ly);\n\n";

		print STDOUT '@{$gblks_w_spec_lb_ly} = ( # [gap,blocks,with,the,specific,left,bound]', "\n";
		foreach my $gblks (@{$gblks_w_spec_lb_ly}) { print STDOUT "\t[", join (',', @{$gblks}), "],\n"; }
		print STDOUT "\t);\n\n";

		print STDOUT '@{$gblks_w_spec_rb_ly} = ( # [gap,blocks,with,the,specific,right,bound]', "\n";
		foreach my $gblks (@{$gblks_w_spec_rb_ly}) { print STDOUT "\t[", join (',', @{$gblks}), "],\n"; }
		print STDOUT "\t);\n\n";

		print STDOUT '$degeneracy_ly                = ', $degeneracy_ly, " ,\n";
		print STDOUT '$code_topo_ly                 = ',  $code_topo_ly,  " ,\n";
		print STDOUT '$cnct_sets_ccgbs_cs_dgncry_ly = ',  $cnct_sets_ccgbs_cs_dgncry_ly, " ,\n";
		print STDOUT '$cnct_if_gbs_on_msa_ends_ly   = ', $cnct_if_gbs_on_msa_ends_ly, " .\n\n";

	     ##} # End of "if ($TEST_LEVEL_nr_msht_gblks >5) {...}"

	     } else { # (Added on 2019/05/28.) #

		print STDOUT "\n# ... The contents are identical to those for the 0 th layer ...\n\n"; # (Added on 2019/05/28.) #

	     } # End of "if (($TEST_LEVEL_nr_msht_gblks >5) or ($ly == 0)) { # (Modified on 2019/05/28) # ...} else {...}".

		##print STDOUT '@{$coords_ly} = (', join (', ', @{$coords_ly}), "),\n\n";

		print STDOUT "\n";
		##print STDOUT "\n\n";
	   }

	} # End of "if ($TEST_LEVEL_nr_msht_gblks >3) {...}"

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >2) {...}"

    if ($TEST_LEVEL_nr_msht_gblks > 5) {

	print STDOUT '(( Contents of %{$clm2ln_prob_sbst} BEFORE ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'concatenated:site:numbers:within:column', 'ln_prob_sbst'), "\n\n";
	foreach my $clm (sort keys %{$clm2ln_prob_sbst}) {
		print STDOUT join ("\t", $clm, $clm2ln_prob_sbst->{$clm}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_nr_msht_gblks > 5) {...}"


	my $diffs_ln_prbs_plus = initialize_multidim_storage (@{$bds_bl_coords}); # Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).
	##my $diffs_ln_prbs_sbst = initialize_multidim_storage (@{$bds_bl_coords});

	unless (defined $diffs_ln_prbs_plus) {
	    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'initialize_multidim_storage ' FAILED on the initial alignment: $! !\n\n";
	    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'initialize_multidim_storage ' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

		# Added on 2019/02/24. #
	my $storage_pnt0 = $diffs_ln_prbs_plus; # Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).
	##my $storage_pnt0 = $diffs_ln_prbs_sbst;
	for (my $bl = 0; $bl < $ct_blks; $bl++) {
	    my $coord = $bl_coords0[$bl];
	    $storage_pnt0 = $storage_pnt0->[$coord];
	}
		# Further modified on 2019/02/26.
		# => Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (5a) #
	$storage_pnt0->[0] = $ct_null_clms0;
	$storage_pnt0->[1] = $degeneracy0;
	$storage_pnt0->[2] = 0; # = $minct_indels0 - $minct_indels0;
	$storage_pnt0->[3] = 0; # = $ln_prb_msa0 - $ln_prb_msa0;
	$storage_pnt0->[4] = 0; # = $ln_prb_sbst0 - $ln_prb_sbst0;
	$storage_pnt0->[5] = 0; # = $ln_prb_indel0 - $ln_prb_indel0;
	#$storage_pnt0->[0] = 0; # $ln_prb_sbst0 - $ln_prb_sbst0;
	#$storage_pnt0->[1] = $ct_null_clms0;
		# End of "=> Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (5a)" #
	##push @{$storage_pnt0}, 0, $ct_null_clms0; # MODIFIED on 2019/02/24.
	###push @{$storage_pnt0}, $ln_prb_sbst0, $ct_null_clms0;
		# END of "Added on 2019/02/24." #

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T6) #

	print STDOUT '[ The Initialized value of @{$diffs_ln_prbs_plus}, ', "the output of 'initialize_multidim_storage' ]\n\n";
	print STDOUT '{ The (', join (', ', @bl_coords0), ')-th component of @{$diffs_ln_prbs_plus} }', "\n";
	print STDOUT '= @{$storage_pnt0} (=($ct_null_clms0, $degeneracy0, $diff_minct_indels0, $diff_ln_prb_msa0, $diff_ln_prb_sbst0, $diff_ln_prb_indel0))', "\n";
	print STDOUT ' = (', join (', ', @{$storage_pnt0}), ") .\n";
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T6) # ...}"


	if ($ct_blks == 0) {
		# (ADDED to fix a potential BUG on 2020/04/15.) #

	    print STDOUT "[INFORMATION ''] The number of gap-blocks is ZERO!! ... I will end this subroutine withOUT going into the MAIN outer while-loop ...\n\n";

	    return (1, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
	}


		### The MAIN PROCESS ###

    (print STDOUT "[[ The MAIN PROCESS ]]\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >1);

	my $wrk_layer = pop @stack;

	my $bl = $ct_blks -1;

	my $counter = 0;
	my $ct_moves = 0; # Added on 2019/02/19.


# $TEST_LEVEL_nr_msht_gblks

	#  ($org_bl_coords, $bds_bl_coords, $exnodes_in_msa, $char_abs, $set_columns0, $bds_blocks0, $rows_affd_by_blks, $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $node2depth, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx) = @_; # Modified on 2019/02/23.

	while (1) { # The MAIN outer while-loop (for exploring the entire hyper-rectangle). #

	    $counter++;

    (print STDOUT "(( The $counter th point visited; ", '$bl =', " $bl ))\n\n") if ($TEST_LEVEL_nr_msht_gblks >1);

		# MODIFIED on 2019/02/24. (2) #
		# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		# => Yet Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #

	    my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
		 $curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends) = @{$wrk_layer};
	    #my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb) = @{$wrk_layer};
	    ## my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms) = @{$wrk_layer};
	    ### my ($curr_bl_coords) = @{$wrk_layer};
	    my $sh = $bl_shifts[$bl];

	    my $bl_coord = $curr_bl_coords->[$bl];
	    my $prev_bl_coord = $bl_coord; # Added on 2019/05/26.
	    my $new_bl_coord = $bl_coord + $sh;
	    my ($lbd_bl_coord, $rbd_bl_coord) = @{$bds_bl_coords->[$bl]};


    if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT "\n[ Data sets BEFORE the \"shift\": ]\n\n";

	print STDOUT '$curr_diff_ln_prb_sbst = ', $curr_diff_ln_prb_sbst, " ,\n";

		# BUGs Fixed on 2019/06/02. (c) #
	my $out_cd_ln_prb_indel = (defined $curr_diff_ln_prb_indel) ? $curr_diff_ln_prb_indel : 'undef';
	my $out_cd_minct_indels = (defined $curr_diff_minct_indels) ? $curr_diff_minct_indels : 'undef';
	print STDOUT '$curr_diff_ln_prb_indel = ', $out_cd_ln_prb_indel, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
	print STDOUT '$curr_diff_minct_indels = ', $out_cd_minct_indels, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	##print STDOUT '$curr_diff_ln_prb_indel = ', $curr_diff_ln_prb_indel, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
	##print STDOUT '$curr_diff_minct_indels = ', $curr_diff_minct_indels, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
		# END of "BUGs Fixed on 2019/06/02. (c)" #

	print STDOUT '$curr_ct_null_clms = ', $curr_ct_null_clms, " .\n\n";

		# Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	print STDOUT '$curr_degeneracy                = ', $curr_degeneracy, " ,\n";

		# BUGs Fixed on 2019/06/02. (d) #
	my $out_c_code_topo = (defined $curr_code_topology) ? $curr_code_topology : 'undef';
	my $out_cc_sets_ccgbs_cs_dg = (defined $curr_cnct_sets_ccgbs_cs_dgnrcy) ? $curr_cnct_sets_ccgbs_cs_dgnrcy : 'undef';
	my $out_cc_if_gbs_on_msa_ends = (defined $curr_cnct_if_gbs_on_msa_ends) ? $curr_cnct_if_gbs_on_msa_ends : 'undef';
	print STDOUT '$curr_code_topology             = ', $out_c_code_topo, " ,\n";
	print STDOUT '$curr_cnct_sets_ccgbs_cs_dgnrcy = ', $out_cc_sets_ccgbs_cs_dg, " ,\n";
	print STDOUT '$curr_cnct_if_gbs_on_msa_ends   = ', $out_cc_if_gbs_on_msa_ends, " .\n\n";
	##print STDOUT '$curr_code_topology             = ', $curr_code_topology, " ,\n";
	##print STDOUT '$curr_cnct_sets_ccgbs_cs_dgnrcy = ', $curr_cnct_sets_ccgbs_cs_dgnrcy, " ,\n";
	##print STDOUT '$curr_cnct_if_gbs_on_msa_ends   = ', $curr_cnct_if_gbs_on_msa_ends, " .\n\n";
		# End of "BUGs Fixed on 2019/06/02. (d)" #

		# End of "Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26)." #

	my $size_stack = @stack;

	print STDOUT '@{$curr_bl_coords} = (', join (', ', @{$curr_bl_coords}), ").\n\n";
	print STDOUT "{Next increment} = $sh ,\n";
	print STDOUT "{Stack size} = $size_stack .\n\n";

	print STDOUT '$bl_coord     = ', $bl_coord, " ,\n";
	print STDOUT '$new_bl_coord = ', $new_bl_coord, " ,\n";
	print STDOUT '($lbd_bl_coord, $rbd_bl_coord) = ', "($lbd_bl_coord, $rbd_bl_coord) .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"

    if ($TEST_LEVEL_nr_msht_gblks >= 5) {

	print STDOUT '( Contents of @{$curr_bds_blks}: )', "\n\n";
	print STDOUT join ("\t", 'Indx_bl', 'left-end', 'right-end'), "\n\n";
	for (my $k = 0; $k < $ct_blks; $k++) {
	    print STDOUT join ("\t", $k , @{$curr_bds_blks->[$k]}), "\n";
	}
	print STDOUT "\n\n";

		# Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #

	print STDOUT '( Contents of @{$curr_gblks_w_spec_lb} & @{$curr_gblks_w_spec_rb} )', "\n\n";
	print STDOUT join ("\t", 'Indx_clm(=c)', '[gap,blocks,w,lb=c]', '[gap,blocks,w,rb=c]'), "\n\n";
	for (my $c = $le_wd; $c <= $re_wd; $c++) {
	    my $cnct_gblks_w_lb_eq_c = join (',', @{$curr_gblks_w_spec_lb->[$c]});
	    my $cnct_gblks_w_rb_eq_c = join (',', @{$curr_gblks_w_spec_rb->[$c]});
	    print STDOUT join ("\t", $c, "[$cnct_gblks_w_lb_eq_c]", "[$cnct_gblks_w_rb_eq_c]"), "\n";
	}
	print STDOUT "\n\n";

		# End of "Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20)." #

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) {...}"


	    if ($rbd_bl_coord < $new_bl_coord) {

		## Stepping over the right-boundary.

			# MODIFIED on 2019/02/22.
			# => Further MODIFIED on 2019/02/24.
			# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
			# => Yet Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
		my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_diff_minct_indels, $src_ct_null_clms, $src_gblks_w_spec_lb, $src_gblks_w_spec_rb,
			$src_degeneracy, $src_code_topology, $src_cnct_sets_ccgbs_cs_dgnrcy, $src_cnct_if_gbs_on_msa_ends) = @{$stack[$#stack]};
		##my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_ct_null_clms, $src_gblks_w_spec_lb, $src_gblks_w_spec_rb) = @{$stack[$#stack]};
		####my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms) = @{$stack[$#stack]};
		my @cp_set_clms = ();
		foreach my $clm (@{$src_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$src_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$src_bl_coords};

			# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		my @cp_gblks_w_spec_lb = my @cp_gblks_w_spec_rb = ();
		foreach my $gblks (@{$src_gblks_w_spec_lb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_lb, \@cp_gblks;
		}
		foreach my $gblks (@{$src_gblks_w_spec_rb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_rb, \@cp_gblks;
		}
			# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20)." #

 			# MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
			# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #

		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_diff_minct_indels, $src_ct_null_clms, \@cp_gblks_w_spec_lb, \@cp_gblks_w_spec_rb,
			$src_degeneracy, $src_code_topology, $src_cnct_sets_ccgbs_cs_dgnrcy, $src_cnct_if_gbs_on_msa_ends);
		##my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_ct_null_clms, \@cp_gblks_w_spec_lb, \@cp_gblks_w_spec_rb);
		####my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms);
		#my $cp_ln_prb_sbst = $src_ln_prb_sbst;
		#my $cp_ct_null_clms = $src_ct_null_clms;
		#my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $cp_ln_prb_sbst, $cp_ct_null_clms);
		##my ($src_bl_coords) = @{$stack[$#stack]};
		##my @cp_bl_coords = @{$src_bl_coords};
		##my @cp_layer = (\@cp_bl_coords);
		$wrk_layer = \@cp_layer;

		###my $copy = cp_comp_array (@{$stack[$#stack]});
		###$wrk_layer = $copy;

		$bl_shifts[$bl] = -1;

	   ## ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb) = @{$wrk_layer};

    (print STDOUT " ... It will go beyond the right-boundary of its coordinate !! ... Move back to the origin...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/02/19.

		next;

	    } elsif ($new_bl_coord < $lbd_bl_coord) {

		## Stepping over the left-boundary of the block coordinate.

		($bl == 0) and last; # End the exploration. #

		$wrk_layer = pop @stack;

		$bl_shifts[$bl] = +1;
		$bl--;

    (print STDOUT " ... It will go beyond the left-boundary of its coordinate !! ... Move back to the origin, and shift our forcus onto the 'parent' coordinate...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/02/19.

		next;
	    }


		## Within the range. ##

	    $ct_moves++; # Added on 2019/02/19.
    (print STDOUT " ... The $ct_moves th Move is likely WITHIN the INTENDED RANGE ... It will be attempted ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 1); # Added on 2019/02/19, MODIFIED on 2019/03/03.


    if ($TEST_LEVEL_nr_msht_gblks > 6) { # Modified on 2019/05/29. #
    ##if ($TEST_LEVEL_nr_msht_gblks > 5) {

	print STDOUT '[ Contents of @{$curr_set_clms} BEFORE the \"shift\": ]', "\n\n";

	print STDOUT join ("\t", 'Column/Row:', (0 .. $ct_rows -1) ), "\n\n";
	for (my $c = 0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$curr_set_clms->[$c]}), "\n";
	}
	print STDOUT "\n\n";

    ##} # End of "if ($TEST_LEVEL_nr_msht_gblks > 5) {...}"
    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 6) { # Modified on 2019/05/29. # ...}"


		## Extend some block-boundaries, if possible.
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1a) ##

	    my ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; # Block-boundaries BEFORE the "shift".

    (print STDOUT "\n", 'Before v-extensions: ($lb_blk_bf, $rb_blk_bf) = ', "($lb_blk_bf, $rb_blk_bf) .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/05/29.

	    if ($sh>0) {
		if (($cts_vincluding->[$bl] + $cts_onx->[$bl] >0) and (@{$curr_gblks_w_spec_lb->[$rb_blk_bf+1]}>0)) {
		    my $ct_vextensions = vextend_right_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextensions) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vextend_right_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_right_end_sbj_gblk' ended (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/05/29.
		    if ($ct_vextensions>0) { ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; } # Update Block-boundaries BEFORE the "shift".
		}
		if (($cts_vincluded->[$bl]>0) and (@{$curr_gblks_w_spec_lb->[$rb_blk_bf+1]}>0)) {
		    my $ct_vextended = vextend_rear_ends_rflnk_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vextended) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vextend_rear_ends_rflnk_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_rear_ends_rflnk_vincluded_gblks' ended (w/ \$ct_vextended = $ct_vextended) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/05/29.
		    ##if ($ct_vextended>0) { ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; } # Update Block-boundaries BEFORE the "shift". # Probably superfluous. #
		}

	    } else { # if ($sh<0) 

		if (($cts_vincluding->[$bl] + $cts_onx->[$bl] >0) and (@{$curr_gblks_w_spec_rb->[$lb_blk_bf-1]}>0)) {
		    my $ct_vextensions = vextend_left_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextensions) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vextend_left_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_left_end_sbj_gblk' ended (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/05/29.
		    if ($ct_vextensions>0) { ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; } # Update Block-boundaries BEFORE the "shift".
		}
		if (($cts_vincluded->[$bl]>0) and (@{$curr_gblks_w_spec_rb->[$lb_blk_bf-1]}>0)) {
		    my $ct_vextended = vextend_rear_ends_lflnk_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextended) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vextend_rear_ends_lflnk_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_rear_ends_lflnk_vincluded_gblks' ended (w/ \$ct_vextended = $ct_vextended) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/05/29.
		    ##if ($ct_vextended>0) { ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; } # Update Block-boundaries BEFORE the "shift". # Probably superfluous. #
		}

	    } # End of "if ($sh>0) {...} else {...}".

	    ($lb_blk_bf, $rb_blk_bf) = @{$curr_bds_blks->[$bl]}; # Update Block-boundaries BEFORE the "shift".

    (print STDOUT "\n", 'After v-extensions: ($lb_blk_bf, $rb_blk_bf) = ', "($lb_blk_bf, $rb_blk_bf) .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/05/29.

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1a)" ##

			# Keep the two key values before the "shift".
			# (Added on 2019/05/29.)
	my $prev_diff_ln_prb_sbst = $curr_diff_ln_prb_sbst;
	my $prev_ct_null_clms = $curr_ct_null_clms;
			# End of "(Added on 2019/05/29.)"

		## MODIFIED on 2019/02/24. (3) ##

		# Actually Update @{$curr_set_clms}, @{$curr_bds_blks}, and @{$curr_bl_coords},
		# & compute the increments of $curr_diff_ln_prb_sbst and $curr_ct_null_clms.

			# MAJOR MODIFICATION in creation of 'nonrec_mlt_shifts_gapblocks'. (MAIN) (on 2019/05/20) #

	    my ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = shift_bl_and_compt_prob_incr (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, $bl, $sh, @{$rows_affd_by_blks}, $cnct_null_clm, %{$clm2ln_prob_sbst}, @{$exnodes_in_msa}, $char_abs, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});

=For the predecessor, 'nonrec_mlt_shifts_ninf_gapblocks'. (MAIN)
			# The subroutine was RENAMED (on 2019/03/27). #
	    my ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = shift_bl_and_compt_prob_incr_brief (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, $bl, $sh, @{$rows_affd_by_blks}, $cnct_null_clm, %{$clm2ln_prob_sbst}, @{$exnodes_in_msa}, $char_abs, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    ##my ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = shift_bl_and_compt_prob_incr (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, $bl, $sh, @{$rows_affd_by_blks}, $cnct_null_clm, %{$clm2ln_prob_sbst}, @{$exnodes_in_msa}, $char_abs, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
			# END of "The subroutine was RENAMED (on 2019/03/27)." #
=cut # End of "For the predecessor, 'nonrec_mlt_shifts_ninf_gapblocks'. (MAIN)"

			# END of "MAJOR MODIFICATION in creation of 'nonrec_mlt_shifts_gapblocks'. (MAIN) (on 2019/05/20)" #

	    unless (defined $flag) {
		print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'shift_bl_and_compt_prob_incr' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", while shifting the $bl th block in the $sh direction: $! !\n\n";
		return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
	    }
    (print STDOUT "\n .... The subroutine 'shift_bl_and_compt_prob_incr' ended (w/ \$flag = $flag) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/05/29.

    if ($TEST_LEVEL_nr_msht_gblks > 3) {

	##print STDOUT "[ Outputs of 'shift_bl_and_compt_prob_incr_brief': ]\n\n"; # The subroutine was RENAMED (on 2019/03/27).
	print STDOUT "[ Outputs of 'shift_bl_and_compt_prob_incr': ]\n\n"; # RESURRECTED (on 2019/05/20). #
	my @printed_out = ();
	foreach my $out ($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) {
	    push @printed_out, (defined $out) ? $out : 'undef';
	}
	print STDOUT '($flag, $incr_ln_prb_sbst, $incr_ct_null_clms) = (', join (', ', @printed_out), ") .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"


		# Added on 2019/02/26. (1) #

	    if ($flag > 0) { # The block stepped over the right-end of the window.

		# DO the same thing as the "if ($rbd_bl_coord < $new_bl_coord) {...}" block above. #

			# MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
			# => Further  MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #

		my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_diff_minct_indels, $src_ct_null_clms, $src_gblks_w_spec_lb, $src_gblks_w_spec_rb,
			$src_degeneracy, $src_code_topology, $src_cnct_sets_ccgbs_cs_dgnrcy, $src_cnct_if_gbs_on_msa_ends) = @{$stack[$#stack]};
		##my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_ct_null_clms, $src_gblks_w_spec_lb, $src_gblks_w_spec_rb) = @{$stack[$#stack]};
		####my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms) = @{$stack[$#stack]};

		my @cp_set_clms = ();
		foreach my $clm (@{$src_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$src_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$src_bl_coords};

			# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
		my @cp_gblks_w_spec_lb = my @cp_gblks_w_spec_rb = ();
		foreach my $gblks (@{$src_gblks_w_spec_lb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_lb, \@cp_gblks;
		}
		foreach my $gblks (@{$src_gblks_w_spec_rb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_rb, \@cp_gblks;
		}
			# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20)." #

 			# MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
 			# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_diff_minct_indels, $src_ct_null_clms, \@cp_gblks_w_spec_lb, \@cp_gblks_w_spec_rb,
			$src_degeneracy, $src_code_topology, $src_cnct_sets_ccgbs_cs_dgnrcy, $src_cnct_if_gbs_on_msa_ends);
		##my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_diff_ln_prb_indel, $src_ct_null_clms, \@cp_gblks_w_spec_lb, \@cp_gblks_w_spec_rb);
		####my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $src_diff_ln_prb_sbst, $src_ct_null_clms);
		$wrk_layer = \@cp_layer;

		$bl_shifts[$bl] = -1;

    (print STDOUT " ... It stepped over the right-end of the window !! ... Move back to the origin...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2);

		next;

	    } elsif ($flag < 0) { # The block stepped over the left-end of the window.

		# DO the same thing as the "elsif ($new_bl_coord < $lbd_bl_coord) {...}" block above. #

		($bl == 0) and last; # End the exploration. #

		$wrk_layer = pop @stack;

		$bl_shifts[$bl] = +1;
		$bl--;

    (print STDOUT " ... It stepped over the left-end of the window!! ... Move back to the origin, and shift our forcus onto the 'parent' coordinate...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2);

		next;
	    }

    (print STDOUT " ... The increments for the $ct_moves th Move has been SUCCESSFULLY calculated !!\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 1);

		# End of "Added on 2019/02/26. (1)" #

		# Now that the main 3 arrays has been successfully updated,
		# update the log-probability and the count of null columns as well.

	    $curr_diff_ln_prb_sbst = ($wrk_layer->[3] += $incr_ln_prb_sbst);

	    $curr_ct_null_clms = ($wrk_layer->[6] += $incr_ct_null_clms); # MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	    ##$curr_ct_null_clms = ($wrk_layer->[5] += $incr_ct_null_clms); # MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
	    ####$curr_ct_null_clms = ($wrk_layer->[4] += $incr_ct_null_clms);

   if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT "[ The Two key values before and after the 'shift' ]\n\n";

	print STDOUT '($prev_diff_ln_prb_sbst, $incr_ln_prb_sbst, $curr_diff_ln_prb_sbst) = ', "($prev_diff_ln_prb_sbst, $incr_ln_prb_sbst, $curr_diff_ln_prb_sbst) .\n";
	print STDOUT '($prev_ct_null_clms, $incr_ct_null_clms, $curr_ct_null_clms) = ', "($prev_ct_null_clms, $incr_ct_null_clms, $curr_ct_null_clms) .\n";
	print STDOUT "\n\n";
	
   } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"


		# Added on 2019/02/26. (2) #

	    my $storage_pnt = $diffs_ln_prbs_plus; # Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).
	    ##my $storage_pnt = $diffs_ln_prbs_sbst;
	    for (my $k = 0; $k < $ct_blks; $k++) {
		my $coord = $curr_bl_coords->[$k];
		$storage_pnt = $storage_pnt->[$coord];
	    }
			# MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (5b) #
			# => CORRECTED on 2019/05/29. #
	    $storage_pnt->[4] = $curr_diff_ln_prb_sbst;
	    $storage_pnt->[0] = $curr_ct_null_clms;
	    ##$storage_pnt0->[4] = $curr_diff_ln_prb_sbst;
	    ##$storage_pnt0->[0] = $curr_ct_null_clms;
	    ####$storage_pnt->[0] = $curr_diff_ln_prb_sbst;
	    ####$storage_pnt->[1] = $curr_ct_null_clms;
			# END of "MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (5b)" #

		# End of "Added on 2019/02/26. (2)" #


	    # ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms) = @{$wrk_layer};


	    ##	# Actually update the current coordinate. # ... OBSOLETE (as of 2019/02/24), because this function is included in the subroutine, 'shift_bl_and_compt_prob_incr(_brief)'!!
	    ##
	    ## $curr_bl_coords->[$bl] = $new_bl_coord;

		## END of "MODIFIED on 2019/02/24. (3)" ##


	    my ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; # Block-boundaries AFTER the "shift". # COPIED from below the retractions of block-boundaries (on 2019/05/27).


    (print STDOUT "\n", 'Before v-retractions: ($lb_blk_af, $rb_blk_af) = ', "($lb_blk_af, $rb_blk_af) .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/05/29.


		## Retract some block boundaries, if possible.
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1b) ##
		## (=> MOVED from below the computation of log-probabilities for indels, etc. (on 2019/05/27).) ##

	    if ($sh >0) {
		if (($cts_vincluded->[$bl] >0) and (@{$curr_gblks_w_spec_rb->[$rb_blk_af]}>1)) {
		    my $ct_vretracted = vretract_right_ends_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_right_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.
		}
		if (($cts_vincluding->[$bl]>0) and (@{$curr_gblks_w_spec_lb->[$lb_blk_af]}>1)) {
		    my $ct_vretracting = vretract_left_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_left_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_end_sbj_gblk' ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.

		    if ($ct_vretracting>0) { ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; } # Update the block-boundaries AFTER the "shift".
		
		}

	    } else { # if ($sh <0)
		if (($cts_vincluded->[$bl] >0) and (@{$curr_gblks_w_spec_lb->[$lb_blk_af]}>1)) {
		    my $ct_vretracted = vretract_left_ends_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_left_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.
		}
		if (($cts_vincluding->[$bl]>0) and (@{$curr_gblks_w_spec_rb->[$rb_blk_af]}>1)) {
		    my $ct_vretracting = vretract_right_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_right_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_end_sbj_gblk' (w/ \$ct_vretracting = $ct_vretracting) ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.

		    if ($ct_vretracting>0) { ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; } # Update the block-boundaries AFTER the "shift".
		}

	    } # End of "if ($sh >0) {...} else {...}".


		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1b)" ##
		## END of "(=> MOVED from below the computation of log-probabilities for indels, etc. (on 2019/05/27).)" ##


	    ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; # Update the Block-boundaries AFTER the "shift". (MODIFIED after MOVING the block-boundary retractions (on 2019/05/27).)
	    ##my ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; # Block-boundaries AFTER the "shift".

    (print STDOUT "\n", 'After v-retractions: ($lb_blk_af, $rb_blk_af) = ', "($lb_blk_af, $rb_blk_af) .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added on 2019/05/29.

	    my $curr_bl_coord = $curr_bl_coords->[$bl]; # (Added on 2019/05/26). #

    (print STDOUT "After the 'shift' (& associated v-extensions and/or v-retractions): \$curr_bl_coord = $curr_bl_coord .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # A BUG fixed on 2019/06/02. #
    ##(pirnt STDOUT "After the 'shift' (& associated v-extensions and/or v-retractions): \$curr_bl_coord = $curr_bl_coord .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2);

		# Keep some OLD "current" values as "PREVIOUS" ones. #
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (3c) ##

	    my $prev_diff_minct_indels = $curr_diff_minct_indels;
	    my $prev_diff_ln_prb_indel = $curr_diff_ln_prb_indel;
	    my $degeneracy_prev = $curr_degeneracy;
	    my $code_topology_prev = $curr_code_topology;
	    my $cnct_sets_ccgbs_cs_dgnrcy_prev = $curr_cnct_sets_ccgbs_cs_dgnrcy;
	    my $cnct_if_gbs_on_msa_ends_prev = $curr_cnct_if_gbs_on_msa_ends;

    if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T8a) #
	print STDOUT "[ BEFORE the computation: Key values concerning gap-pattern topology & indel component of probability ]\n\n";

	print STDOUT '$prev_diff_minct_indels = ', ( (defined $prev_diff_minct_indels) ? $prev_diff_minct_indels : 'undef'), " ,\n";
	print STDOUT '$prev_diff_ln_prb_indel = ', ( (defined $prev_diff_ln_prb_indel) ?  $prev_diff_ln_prb_indel : 'undef'), " ,\n";
	print STDOUT '$degeneracy_prev        = ', ( (defined $degeneracy_prev) ? $degeneracy_prev : 'undef'), " ,\n";
	print STDOUT '$code_topology_prev             = ', ( (defined $code_topology_prev) ? $code_topology_prev : 'undef'), " ,\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy_prev = ', ( (defined $cnct_sets_ccgbs_cs_dgnrcy_prev) ? $cnct_sets_ccgbs_cs_dgnrcy_prev : 'undef'), " ,\n";
	print STDOUT '$cnct_if_gbs_on_msa_ends_prev   = ', ( (defined $cnct_if_gbs_on_msa_ends_prev) ? $cnct_if_gbs_on_msa_ends_prev : 'undef'), " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T8a) # ...}"

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (3c)" ##


		## Examine whether the gap-pattern topology changed or not. ##
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/25-26).) (1c) ##

	    my $if_chng_topology = my $if_chng_bl_bd_status = 0; # (Modified on 2019/05/26.)
	    ##my $if_chng_topology = 0;

	    my ($if_ovlp_bw_veq, $if_right_order_swpbl, $if_excluded) = (0, 1, 0); # Modified in MAJOR REVISIOn of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). #
	    ##my ($if_ovlp_bw_veq, $if_right_order_swpbl) = (0, 1);



	    my ($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy);
	    my ($cnct_if_gbs_on_msa_ends, $code_topology);


	    unless ($curr_ct_null_clms >0) {

			# (i) Examine the positional relationships among vertically equivalent blocks (including swappable ones). #

		if ($ct_sets_vequiv_gblks>0) { # Added the condition on 2019/05/26. #

		    $if_ovlp_bw_veq = chk_ovlp_bw_vequiv_gblks (@{$sets_vequiv_gblks}, @{$curr_bds_blks}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		    unless (defined $if_ovlp_bw_veq) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'chk_ovlp_bw_vequiv_gblks' FAILED: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT " ... The subroutine 'chk_ovlp_bw_vequiv_gblks' Ended (w/ \$if_ovlp_bw_veq = $if_ovlp_bw_veq) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

		} # End of "if ($ct_sets_vequiv_gblks>0) { # Added the condition on 2019/05/26. # ...}"

		if (!($if_ovlp_bw_veq) and ($ct_sets_swappable_gblks>0)) { # Modified the condition on 2019/05/26. #
		##unless ($if_ovlp_bw_veq) { # Obsolete as of 2019/05/26. #... }

		    $if_right_order_swpbl = chk_pos_orders_amg_swpbl_gblks (@{$sets_swappable_gblks}, @{$curr_bds_blks});
		    unless (defined $if_right_order_swpbl) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'chk_pos_orders_amg_swpbl_gblks' FAILED: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT " ... The subroutine 'chk_pos_orders_amg_swpbl_gblks' Ended (w/ \$if_right_order_swpbl = $if_right_order_swpbl) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

		} # End of "if (!($if_ovlp_bw_veq) and ($ct_sets_swappable_gblks>0)) { # Modified the condition on 2019/05/26. # ...}".


			# (i-hf) Examine whether the alignment contains any of the PROHIBITED binary-relations. #
			# (ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).) 

		
		if (!($if_ovlp_bw_veq) and $if_right_order_swpbl and ($ct_x_bhrels>0)) {

		    $if_excluded = chk_excluded_bin_hrels (@{$list_excluded_bin_hrels}, @{$curr_bds_blks}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$inter_block_relations});
		    unless (defined $if_excluded) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'chk_excluded_bin_hrels' FAILED: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT " ... The subroutine 'chk_excluded_bin_hrels' Ended (w/ \$if_excluded = $if_excluded) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

		} # End of "if (!($if_ovlp_bw_veq) and $if_right_order_swpbl and ($ct_x_bhrels>0)) {...}".

			# End of "(ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).)" 


		if (!($if_ovlp_bw_veq) and $if_right_order_swpbl and !($if_excluded)) { # Modified in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). #
		##if (!($if_ovlp_bw_veq) and $if_right_order_swpbl) {

			# (ii) Compute the alignment degeneracy. #

				# (Modified on 2019/05/26.) #
		    ($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy) = ($ct_coll_cmpl_gblocks>0) ? compt_aln_degeneracy (@{$coll_cmpl_gblocks}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_bl_coords}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}) : (1, 'NONE') ;
		    ##($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy) = compt_aln_degeneracy (@{$coll_cmpl_gblocks}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_bl_coords}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
				# End of "(Modified on 2019/05/26.)" #
		    unless (defined $cnct_sets_ccgbs_cs_dgnrcy) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'compt_aln_degeneracy' FAILED: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT " ... The subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy = $degeneracy) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7a) #

	print STDOUT "[ Output of 'compt_aln_degeneracy' ]\n\n";
	print STDOUT '$degeneracy = ', $degeneracy, " .\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy = ', $cnct_sets_ccgbs_cs_dgnrcy, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7a) # ...}"


		  if ($degeneracy>0) { # (Added the condition on 2019/05/26).

			# (iii) Examine whether any gap-blocks are on the MSA-ends. #

    (print STDOUT "{ (iii) Examine whether any gap-blocks are on the MSA-ends. }\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2); # Added on 2019/05/29.

			# Before that, examine whether the boundary status of $bl changed or not. #
			# (Added on 2019/05/26. (2a))

		    my ($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm->[$bl]};

		    if ($sh>0) {
			if ( (($if_on_le_msa) and ($prev_bl_coord == $gb_coord_le_wd)) 
			   or (($if_on_re_msa) and ($curr_bl_coord == $gb_coord_re_wd)) ) {
			     $if_chng_bl_bd_status = 1;
			}
		    } else { # if ($sh<0)
			if ( (($if_on_le_msa) and ($curr_bl_coord == $gb_coord_le_wd))
			   or  (($if_on_re_msa) and ($prev_bl_coord == $gb_coord_re_wd)) ) {
			     $if_chng_bl_bd_status = 1;
			}
		    }
			# End of "(Added on 2019/05/26. (2a))"

		    unless (defined $cnct_if_gbs_on_msa_ends_prev) { $if_chng_bl_bd_status = 1; } # Added on 2019/05/26. #


    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7b) #

	print STDOUT "[ Inputs (\$bl = $bl) ]\n\n";
	print STDOUT '$sh = ', $sh, " .\n";
	print STDOUT '($prev_bl_coord, $curr_bl_coord) = ', "($prev_bl_coord, $curr_bl_coord) .\n";
	print STDOUT '($gb_coord_le_wd, $gb_coord_re_wd) = ', "($gb_coord_le_wd, $gb_coord_re_wd) .\n\n";
	print STDOUT '$cnct_if_gbs_on_msa_ends_prev = ', ((defined $cnct_if_gbs_on_msa_ends_prev) ? $cnct_if_gbs_on_msa_ends_prev : 'undef'), " .\n";

	print STDOUT "\n[ Output ]\n\n";
	print STDOUT '$if_chng_bl_bd_status = ', $if_chng_bl_bd_status, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7b) # ...}"


			# (Modified on 2019/05/26. (2b)) #
		    if ($if_chng_bl_bd_status) {

		    	##my $if_gbs_on_msa_ends = examine_if_gbs_on_msa_ends (@{$curr_bl_coords}, @{$gb_coords_wd_frm}, $if_on_le_msa, $if_on_re_msa) ; # DISABLED because of "unless (defined $cnct_if_gbs_on_msa_ends_prev) { $if_chng_bl_bd_status = 1; }".
			my $if_gbs_on_msa_ends = ($if_on_le_msa or $if_on_re_msa) ?
			   examine_if_gbs_on_msa_ends (@{$curr_bl_coords}, @{$gb_coords_wd_frm}, $if_on_le_msa, $if_on_re_msa) :
					 \@no_gbs_on_msa_ends ; # RESURRECTED because of "unless (defined $cnct_if_gbs_on_msa_ends_prev) { $if_chng_bl_bd_status = 1; }".

			# End of "(Modified on 2019/05/26. (2b))" #

			unless (defined $if_gbs_on_msa_ends) {
			    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'examine_if_gbs_on_msa_ends' FAILED: $! !\n\n";
			    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
			}
			$cnct_if_gbs_on_msa_ends = join (',', @{$if_gbs_on_msa_ends});

    (print STDOUT " ... The subroutine 'examine_if_gbs_on_msa_ends' Ended (w/ \$cnct_if_gbs_on_msa_ends = $cnct_if_gbs_on_msa_ends) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

			# (Added on 2019/05/26. (2c))
		    } else { # if (!($if_chng_bl_bd_status))

    (print STDOUT " ... \$if_chng_bl_bd_status is FALSE!! ... thus, \$if_chng_bl_bd_status (= $cnct_if_gbs_on_msa_ends_prev) must remain UNCHANGED through the 'shift'!!\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2) ; # Added on 2019/05/29.

			$cnct_if_gbs_on_msa_ends = $cnct_if_gbs_on_msa_ends_prev;
		    }
			# End of "(Added on 2019/05/26. (2c))"


			# (v) Examine whether the gap-pattern topology changed or not;
			#	if so, encode the topology.

		    if ($totct_interferences>0) { # Added the condition on 2019/05/26.

		 	if (($incr_ct_null_clms <0) or !($degeneracy_prev) or !(defined $code_topology_prev)) { # The "or ($if_ovlp_bw_veq_prev) or !($if_right_order_swappable_prev)" was REMOVED on 2019/05/26.
		 	##if (($incr_ct_null_clms <0) or ($if_ovlp_bw_veq_prev) or !($if_right_order_swappable_prev) or !($degeneracy_prev) or !(defined $code_topology_prev)) { # The "or !($degeneracy_prev) or !(defined $code_topology_prev)" was added on 2019/05/26. #

    (print STDOUT ' ... ($incr_ct_null_clms <0) or !($degeneracy_prev) or !(defined $code_topology_prev) is TRUE!! ... thus, set $if_chng_topology = 1 .', "\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2); # Added on 2019/05/29.

			    $if_chng_topology = 1;

			} elsif (($totct_interferences>0) and ((@{$interfering_blocks->[$bl]}>0) or (@{$interfering_blocksets->[$bl]}>0) or (@{$cmpl_interfering_blocksets->[$bl]}>0)) ) {

			    $if_chng_topology = ($sh>0) ?
			       chk_chng_gptrn_topology_psm_R ($bl, $lb_blk_bf, $rb_blk_bf, @{$curr_bds_blks}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$cmpl_interfering_blocksets}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}) :
			       chk_chng_gptrn_topology_psm_L ($bl, $lb_blk_bf, $rb_blk_bf, @{$curr_bds_blks}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$cmpl_interfering_blocksets}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}) ;
			    unless (defined $if_chng_topology) {
				print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'chk_chng_gptrn_topology_psm_R/L' FAILED: $! !\n\n";
				return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
			    }
    (print STDOUT " ... The subroutine 'chk_chng_gptrn_topology_psm_R/L' Ended (w/ \$if_chng_topology = $if_chng_topology) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

			}
		    } # End of "if ($totct_interferences>0) { # Added the condition on 2019/05/26. ...}"

		    if ($if_chng_topology) { 

			$code_topology = ($totct_interferences>0) ? encode_gptrn_topology_psm (@{$curr_bds_blks}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}) : $code_topology0; # Modified on 2019/05/26.
			##$code_topology = encode_gptrn_topology_psm (@{$curr_bds_blks}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
			unless (defined $code_topology) {
			    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'encode_gptrn_topology_psm' FAILED: $! !\n\n";
			    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
			}
    (print STDOUT " ... The subroutine 'encode_gptrn_topology_psm' Ended (w/ \$code_topology = $code_topology) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

		    } else {

    (print STDOUT ' ... $if_chng_topology is FALSE! ... thus, set $code_topology(=', $code_topology_prev, ") must remain UNCHANGED through the 'shift'!!\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2); # Added on 2019/05/29.

			$code_topology = $code_topology_prev;
		    }

		  } # End of "if ($degeneracy>0) { # (Added the condition on 2019/05/26.)...}."

		##} # End of "if (!($if_ovlp_bw_veq) and $if_right_order_swpbl) {...}".
		} # End of "if (!($if_ovlp_bw_veq) and $if_right_order_swpbl and !($if_excluded)) { # Modified in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). # ...}".

	    } # End of "unless ($curr_ct_null_clms >0) {...}"

    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7c) #

	print STDOUT '[ Results of examining anomalies in gap-pattern topology ]', "\n\n";
	print STDOUT '$curr_ct_null_clms    = ', $curr_ct_null_clms, " ,\n";
	print STDOUT '$if_ovlp_bw_veq       = ', ((defined $if_ovlp_bw_veq) ? $if_ovlp_bw_veq : 'undef'), " ,\n";
	print STDOUT '$if_right_order_swpbl = ', ((defined $if_right_order_swpbl) ? $if_right_order_swpbl : 'undef'), " ,\n";
	print STDOUT '$if_excluded          = ', ((defined $if_excluded) ? $if_excluded : 'undef'), " ,\n";
	print STDOUT '$if_chng_bl_bd_status = ', ((defined $if_chng_bl_bd_status) ? $if_chng_bl_bd_status : 'undef'), " ,\n";
	print STDOUT '$if_chng_topology     = ', ((defined $if_chng_topology) ? $if_chng_topology : 'undef'), " ,\n";
	print STDOUT "\n";
	print STDOUT '$code_topology                  = ', ((defined $code_topology) ? $code_topology : 'undef'), " ,\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy      = ', ((defined $cnct_sets_ccgbs_cs_dgnrcy) ? $cnct_sets_ccgbs_cs_dgnrcy : 'undef'), " ,\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy_prev = ', ((defined $cnct_sets_ccgbs_cs_dgnrcy_prev) ? $cnct_sets_ccgbs_cs_dgnrcy_prev : 'undef'), " ,\n";
	print STDOUT '$cnct_if_gbs_on_msa_ends        = ', ((defined $cnct_if_gbs_on_msa_ends) ? $cnct_if_gbs_on_msa_ends : 'undef'), " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7c) # ...}"

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/25-26?).) (1c)" ##


		## If necessary, compute the indel component of the log-probability of the current alignment (within the window).
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (1d) ##
		

	    my ($diff_minct_indels, $diff_ln_prb_indel); # Modified in Debugging (on 2019/05/27). #
	    ##my $diff_ln_prb_indel;

	    if (!($curr_ct_null_clms) and !($if_ovlp_bw_veq) and ($if_right_order_swpbl) and !($if_excluded) and ($degeneracy>0)) { # Modified in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). #
	    ##if (!($curr_ct_null_clms) and !($if_ovlp_bw_veq) and ($if_right_order_swpbl) and ($degeneracy>0)) {

		if (($if_chng_bl_bd_status) or ($if_chng_topology) or !(defined $cnct_sets_ccgbs_cs_dgnrcy_prev) or ($cnct_sets_ccgbs_cs_dgnrcy ne $cnct_sets_ccgbs_cs_dgnrcy_prev)) {

		    my $grand_code_topology = join ('::', $code_topology, $cnct_sets_ccgbs_cs_dgnrcy, $cnct_if_gbs_on_msa_ends);

    (print STDOUT '$grand_code_topology = ', $grand_code_topology, " .\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >3); # Added on 2019/05/29.

		    my $diff_ln_prb_indel_plus = $gcode_topo2diff_ln_prb_indel_plus{$grand_code_topology}; # Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28).
		    ##$diff_ln_prb_indel = $gcode_topo2diff_ln_prb_indel{$grand_code_topology};

			# %gcode_topo2diff_ln_prb_indel = ($grand_code_topology0 => 0);

		    ##unless (defined $diff_ln_prb_indel) {
		    if (defined $diff_ln_prb_indel_plus) {

			($diff_minct_indels, $diff_ln_prb_indel) = @{$diff_ln_prb_indel_plus};

    (print STDOUT ' ... $gcode_topo2diff_ln_prb_indel_plus{$grand_code_topology} is DEFINED!! ... Use the values stored in it...', "\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2);

		    } else {

    (print STDOUT ' ... $gcode_topo2diff_ln_prb_indel_plus{$grand_code_topology} is UNdefined!!', " ... Perform the MAIN subroutines, 'enum_psm_indel_scns_for_msa_hs_mrounds' and 'apprx1_tot_lnlk_gpattern_msa2_hs' ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2);

			# Fully compute the indel component of the log-probability. #

			my ($curr_set_cls_gblocks, $curr_min_totcts_indels_for_cls, $curr_sets_psm_hsts_for_cls, $curr_cts_indels0_for_cls, $curr_set_psm_hsts0_for_cls)
			   = enum_psm_indel_scns_for_msa_hs_mrounds (%{$node2ch}, %{$node2pa}, $topnode_id,
			     @{$exnodes_in_msa}, @{$curr_set_clms}, $char_abs, $char_pres, @set_all_nodes, 
			     %{$node2depth}, %{$node2ct_exoffs}, %{$gpattern2indx}, @{$set_gpatterns}, @{$set_node2dparsstat}, @{$set_br2change});
			unless (defined $curr_set_cls_gblocks) {
			    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' FAILED: $! !\n\n";
			    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
			}
			my $curr_minct_indels = 0;
			foreach my $min_totct (@{$curr_min_totcts_indels_for_cls}) { $curr_minct_indels += $min_totct; }
			my $curr_ct_cls = @{$curr_set_cls_gblocks}; # Added on 2019/05/29.

    (print STDOUT " ... The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' Ended (w/ \$curr_ct_cls = $curr_ct_cls, \$curr_minct_indels = $curr_minct_indels) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29.

    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7h) #

	print "[ Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds' (#{gapped segments} = $curr_ct_cls) ]\n\n";
	print STDOUT '@{$curr_min_totcts_indels_for_cls} = (', join (', ', @{$curr_min_totcts_indels_for_cls}), ") .\n\n";
	
	print STDOUT "# NOTE: ... Here, we will OMIT displaying the other main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel histories),\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL\"!!\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7h) # ...}"


					# MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (2) #
			## [ my ($curr_tot_ln_prob_gpattern, $shared_fac1C, $shared_fac2C, $curr_set_clwise_tot_ln_prb_gp, $tot_div_tree_C, $curr_ct_ng_clms, $curr_setset_ln_prb_gp_hsts, $curr_set_rlv_indel_hst2corr_fac) 
			##  = apprx1_tot_lnlk_gpattern_msa2_hs  ] ... before the revision on 2020/04/30.
			##
			my ($curr_tot_ln_prob_gpattern, $shared_fac1C, $shared_fac2C, $curr_set_clwise_tot_ln_prb_gp, $tot_div_tree_C, $curr_ct_ng_clms, $curr_setset_ln_prb_gp_hsts, $curr_set_rlv_indel_hst2corr_fac, $shared_fac3C, $totct_inter_gapless_clm_positions_C, $ct_ends_flnk_gapless_clms_C) 
			    = apprx1_wlp_tot_lnlk_gpattern_msa2_hs ($topnode_id, %{$br2attr}, %{$node2pa}, %{$node2ch}, 
			   $size_wd, $char_abs, @{$curr_set_cls_gblocks}, @{$curr_set_psm_hsts0_for_cls}, @{$curr_sets_psm_hsts_for_cls},
			   $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, %{$br2log_mfacs}, $if_on_le_msa, $if_on_re_msa);
					# End of "MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). (2)" #
			unless (defined $curr_tot_ln_prob_gpattern) {
			    ##print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' FAILED: $! !\n\n";
			    print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' FAILED: $! !\n\n";
			    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
			}
    (print STDOUT " ... The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' Ended (w/ \$curr_tot_ln_prob_gpattern = $curr_tot_ln_prob_gpattern) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2) ; # Added on 2019/05/29. => Modified on 2020/04/30.

    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7i) #

	##print STDOUT "[ Outputs of 'apprx1_tot_lnlk_gpattern_msa2_hs' ]\n\n";
	print STDOUT "[ Outputs of 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' ]\n\n"; # Modified on 2020/04/30.

	print STDOUT '$curr_tot_ln_prob_gpattern = ', $curr_tot_ln_prob_gpattern, " .\n\n";

	print STDOUT '( Contents of @{$curr_set_clwise_tot_ln_prb_gp} )', "\n\n";

	print STDOUT join ("\t", 'Indx_cl', 'tot_ln_prb_gp'), "\n\n";
	for (my $cl=0; $cl<$curr_ct_cls; $cl++) {
	    my $tot_ln_prb_gp = $curr_set_clwise_tot_ln_prb_gp->[$cl];
	    unless (defined $tot_ln_prb_gp) { $tot_ln_prb_gp = 'undef'; }
	    print STDOUT join ("\t", $cl, $tot_ln_prb_gp), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT "( Secondary outputs )\n\n";

	print STDOUT '$shared_fac1C    = ', $shared_fac1C, " ,\n";
	print STDOUT '$shared_fac2C    = ', $shared_fac2C, " ,\n";
	print STDOUT '$tot_div_tree_C  = ', $tot_div_tree_C, " ,\n";
	print STDOUT '$curr_ct_ng_clms = ', $curr_ct_ng_clms, " .\n";

			# (ADDED on 2020/04/30.) #
	print STDOUT "\n";
	print STDOUT '$shared_fac3C = ', $shared_fac3C, " ,\n";
	print STDOUT '$totct_inter_gapless_clm_positions_C = ', $totct_inter_gapless_clm_positions_C, " ,\n";
	print STDOUT '$ct_ends_flnk_gapless_clms_C = ', $ct_ends_flnk_gapless_clms_C, " .\n";
			# End of "(ADDED on 2020/04/30.)" #

	print STDOUT "\n\n";

	print STDOUT "# NOTE: ... Here, we will OMIT displaying some main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel probabilities),\n";
	##print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL\"!!\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs_wLP::TEST_LEVEL\"!!\n"; # Modified on 2020/04/30. #
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T7i) # ...}"


			$diff_minct_indels = $curr_minct_indels - $minct_indels0 ;
			$diff_ln_prb_indel = $curr_tot_ln_prob_gpattern - $tot_ln_prob_gpattern0 ;

			$gcode_topo2diff_ln_prb_indel_plus{$grand_code_topology} = [$diff_minct_indels, $diff_ln_prb_indel] ;
			##$gcode_topo2diff_ln_prb_indel{$grand_code_topology} = $diff_ln_prb_indel ;

		    } # End of "if (defined $diff_ln_prb_indel_plus) {...} else {...}".

		} else {

    (print STDOUT " ... There are NO SIGNs of changes in the gap-pattern topology (including the boundary status)!! ... Just use the Previous values ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks >2);

		    $diff_minct_indels = $prev_diff_minct_indels;
		    $diff_ln_prb_indel = $prev_diff_ln_prb_indel;

		} # End of "if (($if_chng_bl_bd_status) or ($if_chng_topology) or !(defined $cnct_sets_ccgbs_cs_dgnrcy_prev) or ($cnct_sets_ccgbs_cs_dgnrcy ne $cnct_sets_ccgbs_cs_dgnrcy_prev)) {...} else {...}"

	    } else {

    (print STDOUT " ... Some Key Flag(s) and/or Variable(s) indicate(s) that the current alignment MUST BE EXCLUDED from the probability distribution constructed here!! ... So, set \$degeneracy = 0 (if not yet) ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks>2);

		$degeneracy = 0; # This immediately indicates that the current alignment must be EXCLUDED from the probability distribution for some reason(s). #

	    ##} # End of "if (!($curr_ct_null_clms) and !($if_ovlp_bw_veq) and ($if_right_order_swpbl) and ($degeneracy>0)) {...} else {...}".
	    } # End of "if (!($curr_ct_null_clms) and !($if_ovlp_bw_veq) and ($if_right_order_swpbl) and !($if_excluded) and ($degeneracy>0)) { # Modified in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). # ...} else {...}".

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (1d)" ##


		# Update some "current" variables. 
		# (Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (3c) ##
	    $curr_diff_minct_indels = $diff_minct_indels;
	    $curr_diff_ln_prb_indel = $diff_ln_prb_indel;
	    $curr_degeneracy = $degeneracy;
	    $curr_code_topology = $code_topology;
	    $curr_cnct_sets_ccgbs_cs_dgnrcy = $cnct_sets_ccgbs_cs_dgnrcy;
	    $curr_cnct_if_gbs_on_msa_ends = $cnct_if_gbs_on_msa_ends;
		# End of "(Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) (3c)" ##

			# A BUG fixed on 2019/06/02.
	    my $curr_diff_ln_prb_msa;
	    if ((defined $curr_diff_ln_prb_sbst) and (defined $curr_diff_ln_prb_indel)) { $curr_diff_ln_prb_msa = $curr_diff_ln_prb_sbst + $curr_diff_ln_prb_indel; }
	    ##my $curr_diff_ln_prb_msa = $curr_diff_ln_prb_sbst + $curr_diff_ln_prb_indel; # Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (3c+)
			# End of "A BUG fixed on 2019/06/02."

	    ## ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, 
	    ## $curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends) = @{$wrk_layer};

    if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T8b) #

	print STDOUT "[ AFTER the computation: Key values concerning gap-pattern topology & indel component of probability ]\n\n";

	print STDOUT '$curr_diff_minct_indels = ', ( (defined $curr_diff_minct_indels) ? $curr_diff_minct_indels : 'undef'), " ,\n";
	print STDOUT '$curr_diff_ln_prb_indel = ', ( (defined $curr_diff_ln_prb_indel) ?  $curr_diff_ln_prb_indel : 'undef'), " ,\n";
	print STDOUT '$curr_degeneracy        = ', ( (defined $curr_degeneracy) ? $curr_degeneracy : 'undef'), " ,\n";
	print STDOUT '$curr_code_topology             = ', ( (defined $curr_code_topology) ? $curr_code_topology : 'undef'), " ,\n";
	print STDOUT '$curr_cnct_sets_ccgbs_cs_dgnrcy = ', ( (defined $curr_cnct_sets_ccgbs_cs_dgnrcy) ? $curr_cnct_sets_ccgbs_cs_dgnrcy : 'undef'), " ,\n";
	print STDOUT '$curr_cnct_if_gbs_on_msa_ends   = ', ( (defined $curr_cnct_if_gbs_on_msa_ends) ? $curr_cnct_if_gbs_on_msa_ends : 'undef'), " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T8b) # ...}"


		# Store some "current" variables into the storage. #
		# (Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).) (5c) #
		# => CORRECTED on 2019/05/29.
	    $storage_pnt->[1] = $curr_degeneracy;
	    $storage_pnt->[2] = $curr_diff_minct_indels;
	    $storage_pnt->[3] = $curr_diff_ln_prb_msa;
	    $storage_pnt->[5] = $curr_diff_ln_prb_indel;
	    ##$storage_pnt0->[1] = $curr_degeneracy;
	    ##$storage_pnt0->[2] = $curr_diff_minct_indels;
	    ##$storage_pnt0->[3] = $curr_diff_ln_prb_msa;
	    ##$storage_pnt0->[5] = $curr_diff_ln_prb_indel;
		# End of "(Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27).) (5c)" #

    if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T8c) #


	print STDOUT '{ The (', join (', ', @{$curr_bl_coords}), ')-th component of @{$diffs_ln_prbs_plus} }', "\n";
	print STDOUT '= @{$storage_pnt}', "\n";
	print STDOUT '(= ($curr_ct_null_clms, $curr_degeneracy, $curr_diff_minct_indels, $curr_diff_ln_prb_msa, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel)', "\n";

		# A BUG fixed on 2019/06/02. #
	my @out_storage_pnt = ();
	foreach my $x (@{$storage_pnt}) { push @out_storage_pnt, ((defined $x) ? $x : 'undef'); }
	print STDOUT ' = (', join (', ', @out_storage_pnt), ") .\n";
	##print STDOUT ' = (', join (', ', @{$storage_pnt}), ") .\n";
		# End of "A BUG fixed on 2019/06/02." #
	print STDOUT "\n\n\n";

	    # $storage_pnt->[4] = $curr_diff_ln_prb_sbst;
	    # $storage_pnt->[0] = $curr_ct_null_clms;

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) { # Added on 2019/05/29. (T8c) # ...}"


		## Retract some block boundaries, if possible.
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1b) ##
		## (=> MOVED to above the computation of log-probabilities for indels, etc. (on 2019/05/27).) ##

=MOVED to above the computation of log-probabilities for indels, etc.
	    if ($sh >0) {
		if (($cts_vincluded->[$bl] >0) and (@{$curr_gblks_w_spec_rb->[$rb_blk_af]}>1)) {
		    my $ct_vretracted = vretract_right_ends_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_right_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.
		}
		if (($cts_vincluding->[$bl]>0) and (@{$curr_gblks_w_spec_lb->[$lb_blk_af]}>1)) {
		    my $ct_vretracting = vretract_left_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_left_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_end_sbj_gblk' ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.

		    if ($ct_vretracting>0) { ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; } # Update the block-boundaries AFTER the "shift".
		
		}

	    } else { # if ($sh <0)
		if (($cts_vincluded->[$bl] >0) and (@{$curr_gblks_w_spec_lb->[$lb_blk_af]}>1)) {
		    my $ct_vretracted = vretract_left_ends_vincluded_gblks ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_left_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.
		}
		if (($cts_vincluding->[$bl]>0) and (@{$curr_gblks_w_spec_rb->[$rb_blk_af]}>1)) {
		    my $ct_vretracting = vretract_right_end_sbj_gblk ($bl, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'nonrec_mlt_shifts_gapblocks'] The subroutine 'vretract_right_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl th block in the $sh direction: $! !\n\n";
			return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_end_sbj_gblk' ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks > 2); # Added "(w/ ...)" on 2019/06/06.

		    if ($ct_vretracting>0) { ($lb_blk_af, $rb_blk_af) = @{$curr_bds_blks->[$bl]}; } # Update the block-boundaries AFTER the "shift".
		}

	    } # End of "if ($sh >0) {...} else {...}".
=cut # END of "MOVED to above the computation of log-probabilities for indels, etc."

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/24).) (1b)" ##
		## END of "(=> MOVED to above the computation of log-probabilities for indels, etc. (on 2019/05/27).)" ##


		# COPIED from "[ Data sets BEFORE the \"shift\": ]" in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/29). #

    if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT "\n[ Data sets AFTER the \"shift\" (and associated v-extensions and/or v-retractions): ]\n\n";

	print STDOUT '$curr_diff_ln_prb_sbst = ', $curr_diff_ln_prb_sbst, " ,\n";

		# BUGs Fixed on 2019/06/02. (a) #
	my $out_cd_ln_prb_indel = (defined $curr_diff_ln_prb_indel) ? $curr_diff_ln_prb_indel : 'undef';
	my $out_cd_mct_indels = (defined $curr_diff_minct_indels) ? $curr_diff_minct_indels : 'undef';
	print STDOUT '$curr_diff_ln_prb_indel = ', $out_cd_ln_prb_indel, " ,\n";
	print STDOUT '$curr_diff_minct_indels = ', $out_cd_mct_indels, " ,\n";
	##print STDOUT '$curr_diff_ln_prb_indel = ', $curr_diff_ln_prb_indel, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/20). #
	##print STDOUT '$curr_diff_minct_indels = ', $curr_diff_minct_indels, " ,\n"; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
		# End of "BUGs Fixed on 2019/06/02. (a)" #
	print STDOUT '$curr_ct_null_clms = ', $curr_ct_null_clms, " .\n\n";

		# Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	print STDOUT '$curr_degeneracy                = ', $curr_degeneracy, " ,\n";
		# BUGs Fixed on 2019/06/02. (b) #
	my $out_c_code_topo = (defined $curr_code_topology) ? $curr_code_topology : 'undef';
	my $out_c_cnct_sets_ccgbs_cs_dg = (defined $curr_cnct_sets_ccgbs_cs_dgnrcy) ? $curr_cnct_sets_ccgbs_cs_dgnrcy : 'undef';
	my $out_c_cnct_if_gbs_on_msa_es = (defined $curr_cnct_if_gbs_on_msa_ends) ? $curr_cnct_if_gbs_on_msa_ends : 'undef';
	print STDOUT '$curr_code_topology             = ', $out_c_code_topo, " ,\n";
	print STDOUT '$curr_cnct_sets_ccgbs_cs_dgnrcy = ', $out_c_cnct_sets_ccgbs_cs_dg, " ,\n";
	print STDOUT '$curr_cnct_if_gbs_on_msa_ends   = ', $out_c_cnct_if_gbs_on_msa_es, " .\n\n";
	##print STDOUT '$curr_code_topology             = ', $curr_code_topology, " ,\n";
	##print STDOUT '$curr_cnct_sets_ccgbs_cs_dgnrcy = ', $curr_cnct_sets_ccgbs_cs_dgnrcy, " ,\n";
	##print STDOUT '$curr_cnct_if_gbs_on_msa_ends   = ', $curr_cnct_if_gbs_on_msa_ends, " .\n\n";
		# End of "BUGs Fixed on 2019/06/02. (b)" #
		# End of "Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26)." #

	my $size_stack = @stack;

	print STDOUT '@{$curr_bl_coords} = (', join (', ', @{$curr_bl_coords}), ").\n\n";
	print STDOUT "{Current increment} = $sh ,\n";
	print STDOUT "{Stack size} = $size_stack .\n\n";

	print STDOUT '$prev_bl_coord     = ', $prev_bl_coord, " ,\n";
	print STDOUT '$curr_bl_coord = ', $curr_bl_coord, " .\n\n";
	print STDOUT '($lbd_bl_coord, $rbd_bl_coord) = ', "($lbd_bl_coord, $rbd_bl_coord) .\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"

		# END of "COPIED from "[ Data sets BEFORE the \"shift\": ]" in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/29)." #

=For the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'. (T)
    if ($TEST_LEVEL_nr_msht_gblks > 3) {

	print STDOUT "\n[ Data sets AFTER the 'shift' (and associated v-extensions and/or v-retractions): ]\n\n";

	print STDOUT '$curr_diff_ln_prb_sbst = ', $curr_diff_ln_prb_sbst, " ,\n";
	print STDOUT '$curr_ct_null_clms = ', $curr_ct_null_clms, " .\n\n";

	print STDOUT '@{$curr_bl_coords} = (', join (', ', @{$curr_bl_coords}), ").\n\n\n";
    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 3) {...}"
=cut # END of "For the predecessor, 'nonrec_mlt_shifts_noninf_gapblocks'. (T)"

    if ($TEST_LEVEL_nr_msht_gblks >= 5) {

	print STDOUT '( Contents of @{$curr_bds_blks}: )', "\n\n";
	print STDOUT join ("\t", 'Indx_bl', 'left-end', 'right-end'), "\n\n";
	for (my $k = 0; $k < $ct_blks; $k++) {
	    print STDOUT join ("\t", $k , @{$curr_bds_blks->[$k]}), "\n";
	}
	print STDOUT "\n\n";

		# Copied from "BEFORE the 'shift'" in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/29). #

	print STDOUT '( Contents of @{$curr_gblks_w_spec_lb} & @{$curr_gblks_w_spec_rb} )', "\n\n";
	print STDOUT join ("\t", 'Indx_clm(=c)', '[gap,blocks,w,lb=c]', '[gap,blocks,w,rb=c]'), "\n\n";
	for (my $c = $le_wd; $c <= $re_wd; $c++) {
	    my $cnct_gblks_w_lb_eq_c = join (',', @{$curr_gblks_w_spec_lb->[$c]});
	    my $cnct_gblks_w_rb_eq_c = join (',', @{$curr_gblks_w_spec_rb->[$c]});
	    print STDOUT join ("\t", $c, "[$cnct_gblks_w_lb_eq_c]", "[$cnct_gblks_w_rb_eq_c]"), "\n";
	}
	print STDOUT "\n\n";

		# End of "Copied from "BEFORE the 'shift'" in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/29)." #

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) {...}"


    if ($TEST_LEVEL_nr_msht_gblks > 6) { # Modified on 2019/05/29. #
    ##if ($TEST_LEVEL_nr_msht_gblks > 5) {

	print STDOUT '[ Contents of @{$curr_set_clms} AFTER the \"shift\": ]', "\n\n";

	print STDOUT join ("\t", 'Column/Row:', (0 .. $ct_rows -1) ), "\n\n";
	for (my $c = 0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$curr_set_clms->[$c]}), "\n";
	}
	print STDOUT "\n\n";

    ##} # End of "if ($TEST_LEVEL_nr_msht_gblks > 5) {...}"
    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 6) { # Modified on 2019/05/29. # ...}"

    (print STDOUT "\n") if ($TEST_LEVEL_nr_msht_gblks >3); # Added on 2019/05/29.

			## Update the scalar components of @{$wrk_layer}. ##
			## (ADDED in DEBUGGING of 'nonrec_mlt_shifts_gapblocks' on 2019/06/03.) ##

    (print STDOUT "\n", ' ... Now, Update the scalar components of @{$wrk_layer} ...', "\n\n") if ($TEST_LEVEL_nr_msht_gblks >2);

	    # ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	    #	 $curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends) = @{$wrk_layer};

	    @{$wrk_layer}[3,4,5,6] = ($curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms);
	    @{$wrk_layer}[9,10,11,12] = ($curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends);

    (print STDOUT "\n ... Done !!\n\n") if ($TEST_LEVEL_nr_msht_gblks >2);

			## END of "(ADDED in DEBUGGING of 'nonrec_mlt_shifts_gapblocks' on 2019/06/03.)" ##

    (print STDOUT "\n ... Finally, Re-Build the stack, if necessary ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2); # Added on 2019/05/29.

		# Raise $bl all the way to ($ct_blks -1).

	    my $ct_cps = 0;
	    while ($bl < $ct_blks -1) { # 1st inner while-loop. #
		$bl++;

		$ct_cps++;
	    }  # End of the 1st inner while-loop. #

		# Stuff the stack with copies of the newly computed layer.

	    while ($ct_cps>0) { # 2nd inner while-loop. #
		$ct_cps--;

			# MODIFIED on 2019/02/24. (5) #
		##my ($src_set_clms, $src_bds_blks, $src_bl_coords, $src_ln_prb_sbst, $src_ct_null_clms) = @{$wrk_layer}; # This is probably unnecessary!!

		my @cp_set_clms = ();
		foreach my $clm (@{$curr_set_clms}) {
		    my @cp_clm = @{$clm};
		    push @cp_set_clms, \@cp_clm;
		}
		my @cp_bds_blks = ();
		foreach my $bds_blk (@{$curr_bds_blks}) {
		    my @cp_bds_blk = @{$bds_blk};
		    push @cp_bds_blks, \@cp_bds_blk;
		}
		my @cp_bl_coords = @{$curr_bl_coords};

			# ADDED in DEBUGGING (on 2019/06/02). #
		my @cp_gblks_w_spec_lb = my @cp_gblks_w_spec_rb = ();
		foreach my $gblks (@{$curr_gblks_w_spec_lb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_lb, \@cp_gblks;
		}
		foreach my $gblks (@{$curr_gblks_w_spec_rb}) {
		    my @cp_gblks = @{$gblks};
		    push @cp_gblks_w_spec_rb, \@cp_gblks;
		}

			# END of "ADDED in DEBUGGING (on 2019/06/02)." #

	    # ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	    #	 $curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends) = @{$wrk_layer};

			# => Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
			# => A BUG Fixed on 2019/06/02. #
		my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms, \@cp_gblks_w_spec_lb, \@cp_gblks_w_spec_rb,
	    		$curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends);
		##my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $curr_diff_ln_prb_sbst, $curr_diff_ln_prb_indel, $curr_diff_minct_indels, $curr_ct_null_clms,
	    	##	$curr_degeneracy, $curr_code_topology, $curr_cnct_sets_ccgbs_cs_dgnrcy, $curr_cnct_if_gbs_on_msa_ends);
		####my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $curr_diff_ln_prb_sbst, $curr_ct_null_clms);
			# End of "=> Further MODIFIED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26)." #

		#my $cp_ln_prb_sbst = $src_ln_prb_sbst;
		#my $cp_ct_null_clms = $src_ct_null_clms;
		#my @cp_layer = (\@cp_set_clms, \@cp_bds_blks, \@cp_bl_coords, $cp_ln_prb_sbst, $cp_ct_null_clms);
		##my ($src_bl_coords) = @{$wrk_layer};
		##my @cp_bl_coords = @{$src_bl_coords};
		##my @cp_layer = (\@cp_bl_coords);

			# END of "MODIFIED on 2019/02/24. (5)" #

		push @stack, \@cp_layer;

		#### my $cp_layer = cp_comp_array (@{$wrk_layer});
		#### push @stack, $cp_layer;

	    }  # End of the 2nd inner while-loop. #

    (print STDOUT "\n ... Finished Re-Building the stack ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >2); # Added on 2019/05/29.

	} # End of the MAIN outer while-loop (for exploring the entire hyper-rectangle). #

    (print STDOUT "\n ... ALL the \"shift\"s Attempted ...\n\n\n") if ($TEST_LEVEL_nr_msht_gblks > 1);

    if ($TEST_LEVEL_nr_msht_gblks2 > 1) { # Modified on 2019/06/06. #
    ##if ($TEST_LEVEL_nr_msht_gblks > 5) {

	print STDOUT '(( Contents of %{$clm2ln_prob_sbst} AFTER ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'concatenated:site:numbers:within:column', 'ln_prob_sbst'), "\n\n";
	foreach my $clm (sort keys %{$clm2ln_prob_sbst}) {
		print STDOUT join ("\t", $clm, $clm2ln_prob_sbst->{$clm}), "\n";
	}
	print STDOUT "\n\n";

		# ADDED on 2019/06/04. #

	## my %gcode_topo2diff_ln_prb_indel_plus = ($grand_code_topology0 => [0, 0]); # [$diff_minct_indels, $diff_ln_prb_indel]
	## $grand_code_topology = join ('::', $code_topology, $cnct_sets_ccgbs_cs_dgnrcy, $cnct_if_gbs_on_msa_ends);

	print STDOUT '(( Contents of %gcode_topo2diff_ln_prb_indel_plus  AFTER ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'Code_topology::cnct_sets_ccgbs_cs_dgnrcy::cnct_if_gbs_on_msa_ends', 'Diff_Min#{indels}', 'Diff_ln_prb_indel'), "\n\n";
	foreach my $gcode_topo (sort keys %gcode_topo2diff_ln_prb_indel_plus) {
	    my $diff_ln_prb_indel_plus = $gcode_topo2diff_ln_prb_indel_plus{$gcode_topo};
	    my ($diff_minct_indels, $diff_ln_prb_indel) =
		(defined $diff_ln_prb_indel_plus) ? @{$diff_ln_prb_indel_plus} : ('undef', 'undef') ;
	    print STDOUT join ("\t", $gcode_topo, $diff_minct_indels, $diff_ln_prb_indel), "\n" ;
	}
	print STDOUT "\n\n";

		# END of "ADDED on 2019/06/04." #

   ##} # End of "if ($TEST_LEVEL_nr_msht_gblks > 5) {...}"
   } # End of "if ($TEST_LEVEL_nr_msht_gblks2 > 1) { # Modified on 2019/06/06. # ...}"

    if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/29. (T9) #

	my $ct_gpatterns_fnl=@{$set_gpatterns};

	print STDOUT '[ Contents of @{$set_gpatterns},', " one of the Main Effective Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds', AFTER all the \"shift\"s (#{gap-patterns}=$ct_gpatterns_fnl) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', 'gap-pattern'), "\n\n";

	for (my $i=0; $i<$ct_gpatterns_fnl; $i++) {
	    print STDOUT join ("\t", $i, ${$set_gpatterns->[$i]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) { # Added on 2019/05/28. (T9) #...}"


    (print STDOUT "<< Subtoutine 'nonrec_mlt_shifts_gapblocks' Ended (#{visits} = $counter, #{moves} = $ct_moves) />>\n\n") if ($TEST_LEVEL_nr_msht_gblks >0); # Modified on 2019/02/19.


		## MAJOR MODIFICATION in creation of 'nonrec_mlt_shifts_gapblocks'. (OUT) ##
	return (1, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
	##return ($ln_prb_sbst0, $diffs_ln_prbs_sbst); 

}	# End of the subroutine 'nonrec_mlt_shifts_gapblocks'. #


# compt_log_msa_prob_plus:
#	This subroutine computes the log-probability, as well as its substitution & indel components,
#	of an input MSA (expressed in terms of a set of columns).
#	It was produced mainly from the part of 'nonrec_mlt_shifts_gapblocks'
#	before the while-loop performing non-recursive multiple-"shift"s.
#	[CREATED on 2019/07/04.]
#
#	[MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30)]:
#	Now, the satellite subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm")
#	has been REPLACED with 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm").
#	In consequence,
#	this subroutine can NOW use log-multiplication factors of case-(i) "gapped segments",
#	so that it can take FULL advantage of the main outputs of the program package, "LASTPIECE".
#	Therefore, it should NOW be able to compute the probabilities of MSAs
#	MUCH MORE ACCURATELY than before the revision.
#
# nonrec_mlt_shifts_gapblocks:
#	This MAIN subroutine performs multiple "shift"s of gap-blocks,
#	which may in general interfere with one another.
#	It was created by substantially extending the simple version, 'nonrec_mlt_shifts_noninf_gapblocks',
#	with some additional (and advanced) functions.
#
#	MAJOR DIFFERENCES from 'nonrec_mlt_shifts_noninf_gapblocks':
#	[... OMITTED the descriptions; see the header of 'nonrec_mlt_shifts_gapblocks' if interested ...]
#
#
# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@gb_coords_wd_frm, \@exnodeids_in_msa, $char_abs, $char_pres, \@set_columns0, 
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \@bds_gblocks0, \@rows_affd_by_gblks, 
#	\%clm2ln_prob_sbst, $topnode_id, \%node2ch, \%br2attr, \%node2pa, \%node2depth, \%node2ct_exoffs, \%br2eq, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx, 
#	\@set_ingredients_indel_hsts, @set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
#	\@list_excluded_bin_hrels), where
#
#	$org_bl_coords[$bl] gives the "origin" of the coordinate of the $bl th gap-block;
#	($left_bd, $right_bd) = @{$bds_bl_coords[$bl]} gives the left and right-boundaries
#	of the coordinate of the $bl th gap-block;
#	(NOTE that the boundaries are also included in the interval to be explored.)
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window.
#	NOTE 2: $gb_coord_le_wd <= $lbd_coord <= $rbd_coord <= $gb_coord_re_wd MUST always hold.
#		(However, $gb_coord_le_wd may be negative.)
#
#	$exnodeids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#	[Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyTreeMap_sppl.pm".]
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#	$char_pres is the character that represents a residue (of any type), or a "presence" state;
#
#	@column = @{$set_columns0[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_columns0 is an input of, e.g., 'mk_init_set_columns0' in "MyANEX_Supple.pm". (... ADDED on 2019/02/27)]
#
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	@{$bds_gblocks0[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[\@bds_gblocks0 and \@rows_affd_by_gblks are outputs of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	%clm2ln_prob_sbst = ( join (':', @column) => $log_prob_sbst, ...)
#	where $log_prob_sbst is the log-probability of the column @column under the given substitution model;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#	%node2pa = ($node_id => $parent_id, ...) for the nodes other than the root;
#	%br2attr = ($br_id => \@attr, ...) for all branches, with 
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0b) #
#	%node2ct_exoffs = ($node_id => #{external offsprings}, ...) for ALL nodes;
#	[\%node2ct_exoffs is an output of, e.g., 'make_node2ct_exoffs' in "MyTreeManipu_III.pm"];
#
#	%br2eq = ($br_id => $equivalent_branch_id, ...) for branches with equivalent ones;
#	[\%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm"];
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0b)" #
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0c) #
#
#	(\%gpattern2indx, \@set_gpatterns, \@set_node2dparsstat, \@set_br2change) = @set_ingredients_indel_hsts, 
#	where
#	%gpattern2indx, @set_gpatterns, @set_node2dparsstat and @set_br2change, 
#	each grow from a null array (or hash) 
#	(, which means that, initially, ({}, [], [], []) = @set_ingredients_indel_hsts,)
#	by gradually incoporating newly encountered gap-patterns and the dollo parsimonious indel histories that can result in the patterns.
#	Their formats are as follows:
#
#	%gpattern2indx = ($gpattern => $indx_gp, ...) for the gap-patterns encountered thus far,
#	with $indx_gp assigned in the order of occurrences;
#	
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern in the input MSA, created by concatenating $char_pres/abs in the sequences;
#
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#
#		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/22). (0c)" #
#
#		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23). (0d) #
#
#	($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs) = @set_ingredients_prb_msa_indel, where
#	$totrate_ins/del is the total rate of insertions/deletions (relative to the substitution rate),
#
#	$cutoff_ins/del is the cut-off length for insertions/deletions, which is common along the MSA;
#
#	$delta_del = Sum_{$l=1}^{$cutoff_del}($l-1)*distr_del ($l)/$nfac_del,
#	which can be calculated via 'cal_delta_exp_del_5p', is the excessive exponential factor
#	due to deletions extending out of the MSA boundaries;
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	$if_on_le/re_msa = 1 if the window we are dealing with is on the left/right-end of the input MSA,
#			 = 0 otherwise;
#
#		# END of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23). (0d)" #
#
#		# ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (0f) #
#	@list_excluded_bin_hrels lists the PROHIBITED binary horizontal relationships,
#	which is a main input of 'chk_excluded_bin_hrels';
#	More precisely,
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined;
#	Currently, the following codes can possibly be in the list:
#	'-', indicating that the gap-blocks are "connected" (overlapping or touching);
#	'>', indicating that $gb1 horizontally includes $gb2;
#	'<', indicating that $gb1 is horizontally included in $gb2;
#	'=', indicating that $gb1 is horizontally identical to $gb2;
#	(NOTE that the relationship assignment is the opposite of that for the vertical relationships);
#	(See the header of 'encode_gptrn_topology_psm' for detailed descriptions of all codes);
#		# END of "ADDED in MAJOR REVISION of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). (0f)" #
#
# [ OUTPUTS: ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, \@diffs_ln_prbs_plus, where ] ... for the predecessor, 'nonrec_mlt_shifts_gapblocks'.
#
# OUTPUTS: ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $grand_code_topology0), where
#
#	$if_success	= 1 if the subroutine ended "successfully", 
#			= 0 otherwise;
#	$ct_null_clms0 is the number of null columns in the initial local alignment;
#	$degeneracy0 is the degree of degeneracy of the initial local alignment;
#	$minct_indels0 is the minimum number of indel events required to create the initial local alignment;
#	$ln_prb_msa0 (= $ln_prb_sbst0 + $ln_prb_indel0) is the total log-probability
#	of the initial local alignment under the given evolution model (with the degeneracy factor NOT subtracted);
#	$ln_prb_sbst0 is the log-probability of the initial local alignment under the given substitution model;
#	$ln_prb_indel0 is the log-probability of the initial local alignment under the given insertion/deletion model;
#
#	$grand_code_topology0 = join ('::', $code_topology0, $cnct_sets_ccgbs_cs_dgnrcy0, $cnct_if_gbs_on_msa_ends0) is a string that codes the topology of the gap-block configuration & boundary status of the initial local alignment. # Added in creation of 'compt_log_msa_prob_plus' (on 2019/07/04.)
#
#	[
#	$diffs_ln_prbs_plus[$k_0]->...->[$k_B] = [$ct_null_clms, $degeneracy, $diff_minct_indels, $diff_ln_prb_msa, $diff_ln_prb_sbst, $diff_ln_prb_indel]
#	provide information on the local alignment resulting from the
#	shifts of the 0 th gap-block by ($k_0 - $org_bl_coords[0]) columns,
#	..., and the B (= #{blocks} -1) by ($k_B - $org_bl_coords[B]) columns.
#	Here,
#	$ct_null_clms is the number of null columns in the alignment,
#	$degeneracy is the degree of degeneracy of the alignment,
#	$diff_minct_indels is the difference of the minimum number of indel events (compared to $minct_indels0),
#	$diff_ln_prb_msa (= $dff_ln_prb_sbst + $diff_ln_prb_indel) is the difference of the total log-probability
#	of the alignment (with the degeneracy factor NOT subtracted) compared to $ln_prb_msa0,
#	$diff_ln_prb_sbst is the difference of the log-probability of the alignment under the given substitution model (compared to $ln_prb_sbst0),
#	$diff_ln_prb_indel is the difference of the log-probability of the alignment under the given indel model (compared to $ln_prb_indel0).
#	] ... for the predecessor, 'nonrec_mlt_shifts_gapblocks'.
#
#
# [ sub nonrec_mlt_shifts_gapblocks (\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@) {  ] ... for the predecessor, 'nonrec_mlt_shifts_gapblocks'.
#
sub compt_log_msa_prob_plus (\@\@\@\@$$\@\@$$\@\@\%$\%\%\%\%\%\%\@\%\%\@\@$$\@) {

	my ($org_bl_coords, $bds_bl_coords, $gb_coords_wd_frm, $exnodes_in_msa, $char_abs, $char_pres, $set_columns0,
	 $list_gblocks0, $indx_br, $indx_u_or_d, $bds_blocks0, $rows_affd_by_blks,
	 $clm2ln_prob_sbst, $topnode_id, $node2ch, $node2pa, $br2attr, $node2depth, $node2ct_exoffs, $br2eq, $base_freqs_root, $node2set_sw_extcond_probs0, $br2tr_mtrx, 
	$set_ingredients_indel_hsts, $set_ingredients_prb_msa_indel, $if_on_le_msa, $if_on_re_msa,
	$list_excluded_bin_hrels) = @_;

	my $ct_blks = @{$org_bl_coords};
	my $ct_rows = @{$exnodes_in_msa};
	my $ct_clms = @{$set_columns0};

	my $ct_x_bhrels = @{$list_excluded_bin_hrels}; # ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/27). #

    (print STDOUT "<< Subtoutine 'compt_log_msa_prob_plus' (#{gap-blocks} = $ct_blks, #{rows} = $ct_rows, #{columns} = $ct_clms, #{excluded_binary_relations}=$ct_x_bhrels) >>\n\n") if ($TEST_LEVEL_cmpt_lmpp >0);

		# 1st consistency check. #

	if (@{$bds_bl_coords} != $ct_blks) {
	    my $ct_bd_pairs = @{$bds_bl_coords};
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The number of coordinate boundary-pairs (=$ct_bd_pairs) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}


	if (@{$gb_coords_wd_frm} != $ct_blks) {
	    my $ct_gb_coords = @{$gb_coords_wd_frm};
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The number of coordinate pairs corresponding to the window frame (=$ct_gb_coords) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

	if (@{$list_gblocks0} != $ct_blks) { # Added on 2019/05/22.
	    my $ct_blks2 = @{$list_gblocks0};
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The size of the list of gap-blocks (=$ct_blks2) CLASHES with the number of coordinate origins (=$ct_blks); $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

    if ($TEST_LEVEL_cmpt_lmpp >2) {

	print STDOUT "\n[ Input variables: Contents of ", '@{$list_gblocks0}, @{$bds_blocks0},  @{$org_bl_coords}, @{$bds_bl_coords}, @{$gb_coords_wd_frm} & @{$rows_affd_by_blks}', " ]\n\n";

	print STDOUT join ("\t", 'Indx_bl', 'br', 'up(U)_or_lw(L)', 'lb_blk', 'rb_blk', 'org_gcoord', 'lb_gcoord', 'rb_gcoord', 'gb_coord_lbd_wd', 'gb_coord_rbd_wd', '[rows,affected]'), "\n\n"; # Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/06/02). #
	for (my $k = 0; $k <$ct_blks; $k++) {
	    my $cnct_rows_affd = join (',', @{$rows_affd_by_blks->[$k]});
	    print STDOUT join ("\t", $k, @{$list_gblocks0->[$k]}[$indx_br, $indx_u_or_d], @{$bds_blocks0->[$k]}, $org_bl_coords->[$k], @{$bds_bl_coords->[$k]}, @{$gb_coords_wd_frm->[$k]}, "[$cnct_rows_affd]" ), "\n"; # Yet Further Modified in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/06/02). #
	}
	print STDOUT "\n\n";

		# Added on 2019/05/28. (T1) #

      if ($TEST_LEVEL_cmpt_lmpp >3) { # (T1a)

	print STDOUT '( Contents of @{$exnodes_in_msa} )', "\n\n";
	print STDOUT join ("\t", 'Indx_row', 'node_id'), "\n\n";
	for (my $r=0; $r<$ct_rows; $r++) {
	    print STDOUT join ("\t", $r, $exnodes_in_msa->[$r]), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$set_columns0} )', "\n\n";
	print STDOUT join ("\t", 'Column\Row', (0 .. $ct_rows-1)), "\n\n";
	for (my $c=0; $c < $ct_clms; $c++) {
	    print STDOUT join ("\t", $c, @{$set_columns0->[$c]}), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_cmpt_lmpp >3) { # (T1a) ...}"

	print STDOUT "( Input Tree )\n\n";

	my %nodeid2rowid;
	for (my $r=0; $r<$ct_rows; $r++) {
	    my $nodeid = $exnodes_in_msa->[$r];
	    $nodeid2rowid{$nodeid} = $r;
	}

	unless ( output_tree ($topnode_id, %{$node2ch}, %nodeid2rowid, %{$br2attr}, *STDOUT, 1, 1) ) {
	    print STDOUT "[CAUTION 'compt_log_msa_prob_plus'] The subroutine 'output_tree' FAILED on the input tree: $! !\n\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_cmpt_lmpp >3) { # (T1b)

	print STDOUT '( Contents of %{$node2depth}, %{$node2ct_exoffs} & %{$br2eq} )', "\n\n";
	print STDOUT join ("\t", 'Node_ID(=br_ID)', 'depth', '#{external_offsprings}', 'equiv_br_ID'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) {
	    my $depth = $node2depth->{$node};
	    my $ct_exoffs = $node2ct_exoffs->{$node};
	    my $eqbr = $br2eq->{$node};
	    unless (defined $eqbr) { $eqbr = 'n/a'; }
	    print STDOUT join ("\t", $node, $depth, $ct_exoffs, $eqbr), "\n";
	}

	print STDOUT "\n\n";

	print STDOUT '@{$base_freqs_root} = (', join (', ', @{$base_freqs_root}), ") .\n\n";

      } # End of "if ($TEST_LEVEL_cmpt_lmpp >3) { # (T1b) ...}"

	print STDOUT '( Contents of @{$set_ingredients_prb_msa_indel} )', "\n\n";

	print STDOUT '@{$set_ingredients_prb_msa_indel}', "\n";
	print STDOUT ' = ($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs)', "\n";
	print STDOUT '  = (', join (', ', @{$set_ingredients_prb_msa_indel}), ") .\n";
	print STDOUT "\n\n";

	print STDOUT '($if_on_le_msa, $if_on_re_msa) = ', "($if_on_le_msa, $if_on_re_msa) .\n";
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$list_excluded_bin_hrels} ', "(#{excld_bin_hrels}=$ct_x_bhrels) )\n\n";
	print STDOUT join ("\t", 'No.', 'gb1', 'gb2', 'hrel'), "\n\n";

	for (my $k=0; $k<$ct_x_bhrels; $k++) {
	    print STDOUT join ("\t", $k, @{$list_excluded_bin_hrels->[$k]}), "\n";
	}
	print STDOUT "\n\n\n";

		# End of "Added on 2019/05/28. (T1)" #

    } # End of "if ($TEST_LEVEL_cmpt_lmpp >2) {...}"

		# 2nd consistency check. #
	my @ill_blocks1 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 1st Preliminary for-loop (over blocks). #
	    if ($bds_bl_coords->[$k][0] < 0) { push @ill_blocks1, $k; }
	} # End of the 1st preliminary for-loop (over blocks). #
	if (@ill_blocks1>0) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] For the blocks, (", join (', ', @ill_blocks1), "), the left-ends are NEGATIVE: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

		# 3rd consistency check. #
	my @ill_blocks2 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 2nd Preliminary for-loop (over blocks). #
	    my $org = $org_bl_coords->[$k];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$k]};
	    if (($org < $lbd) or ($rbd < $org)) { push @ill_blocks2, $k; }
	} # End of the 2nd preliminary for-loop (over blocks). #

	if (@ill_blocks2>0) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] For the blocks, (", join (', ', @ill_blocks2), "), the 'origin's are NOT within the intervals specified by the boundaries: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}

		# ADDED in creation of 'nonrec_mlt_shifts_gapblocks'' (on 2019/05/19). #

		# 3rd consistench check. (B) #
	my @ill_blocks3 = ();
	for (my $k=0; $k <$ct_blks; $k++) { # 2nd Preliminary for-loop (over blocks). (B) #
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$k]};
	    my ($gcd_lbd_wd, $gcd_rbd_wd) = @{$gb_coords_wd_frm->[$k]};
	    if (($lbd < $gcd_lbd_wd) or ($gcd_rbd_wd < $rbd)) { push @ill_blocks3, $k; }
	} # End of the 2nd preliminary for-loop (over blocks). (B) #

	if (@ill_blocks3>0) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] For the blocks, (", join (', ', @ill_blocks3), "), either or both of the block-coordinate boundaries STICK OUT of the window frame: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
		# End of "ADDED in creation of 'nonrec_mlt_shifts_gapblocks'' (on 2019/05/19)." #

   (print STDOUT "\n ... The input data PASSED some brief consistency checks ... NOW, we will perform some Preliminary Processes ...\n\n\n") if ($TEST_LEVEL_cmpt_lmpp > 1); # Added on 2019/05/27.


		## Process (0.5): Prepare some important variables. ##
		## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks', from 2019/05/20 till 2019/05/dd.)

    (print STDOUT "<  Process (0.5): Prepare some important variables. >\n\n") if ($TEST_LEVEL_cmpt_lmpp >1); # Added on 2019/05/28.

	my ($le_wd, $re_wd) = (0, $ct_clms-1);
	my ($gblks_w_spec_lb0, $gblks_w_spec_rb0) = mk_lists_gblocks_w_spec_bounds ($le_wd, $re_wd, @{$bds_blocks0});
	unless (defined $gblks_w_spec_lb0) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_lists_gblocks_w_spec_bounds' Ended ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28 (T2) #
	print STDOUT "[ Outputs of 'mk_lists_gblocks_w_spec_bounds': Contents of ", '@{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}', " ]\n\n";
	print STDOUT join ("\t", 'Indx_cl(=c)', '[gap,blocks,with,left-bound0=c]', '[gap,blocks,with,right-bound0=c]'), "\n\n";
	for (my $c=0; $c<$ct_clms; $c++) {
			# A BUG fixed on 2019/06/02. #
	     my $cnct_gblks_w_lb0_eq_c = join (',', @{$gblks_w_spec_lb0->[$c]});
	     my $cnct_gblks_w_rb0_eq_c = join (',', @{$gblks_w_spec_rb0->[$c]});
	     ##my $cnct_gblks_w_lb0_eq_c = join (',', @{$gblks_w_spec_lb0});
	     ##my $cnct_gblks_w_rb0_eq_c = join (',', @{$gblks_w_spec_rb0});
	     print STDOUT join ("\t", $c, "[$cnct_gblks_w_lb0_eq_c]", "[$cnct_gblks_w_rb0_eq_c]"), "\n";
	}
	print STDOUT "\n\n";
    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28 (T2) # ...}"
    


	my ($inter_block_relations, $interfering_blocks, $interfering_blocksets, $cmpl_interfering_blocksets, $br2up_down2blocksets) = extensive_exam_inter_block_relations (@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, %{$node2depth});
	unless (defined $cmpl_interfering_blocksets) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'extensive_exam_inter_block_relations' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'extensive_exam_inter_block_relations' Ended ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.


	my $totct_interferences = 0;
		# A BUG fixed (on 2019/06/02). #
	foreach my $intrfring (@{$interfering_blocks}) {
	    if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}) /2; } # MODIFIED to take account of the double-counting (on 2019/06/02).
	    ## if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}) ; }
	} 
	foreach my $intrfring (@{$interfering_blocksets}) {
	    if (@{$intrfring}>0) { $totct_interferences += scalar (@{$intrfring->[0]}); } 
	}
	####foreach my $intrfring (@{$interfering_blocks}) { $totct_interferences += @{$intrfring->[0]}; }
	##foreach my $intrfring (@{$interfering_blocksets}) { $totct_interferences += scalar (@{$intrfring->[0]}); }
		# END of "A BUG fixed (on 2019/06/02)." #
	
    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T3) #

	print STDOUT "[ Outputs of 'extensive_exam_inter_block_relations' (Total#{interferences}=$totct_interferences) ]\n\n";

	print STDOUT '( Contents of @{$inter_block_relations} )', "\n\n";
	print STDOUT join ("\t", 'Gb1\Gb2:', (0 .. $ct_blks-1)), "\n\n";
	for (my $gb1 = 0; $gb1 <$ct_blks; $gb1++) {
	    my @out_rels_w_gb1 = @{$inter_block_relations->[$gb1]};
	    for (my $gb2=0; $gb2<$ct_blks; $gb2++) {
		unless (defined $out_rels_w_gb1[$gb2]) { $out_rels_w_gb1[$gb2] = 'undef'; }
	    }
	    print STDOUT join ("\t", $gb1, @out_rels_w_gb1), "\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_cmpt_lmpp > 3) {

	print STDOUT '( Contents of @{$interfering_blocks} )', "\n\n";

	print STDOUT join ("\t", 'Gb', 'intrfring_block1(rel1)', 'intrfring_block2(rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($blocks, $rels) = @{$interfering_blocks->[$gb]};
	    unless (defined $blocks) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_intrfr_blks = @{$blocks};
	    my @out_blocks = ();
	    for (my $k=0; $k<$ct_intrfr_blks; $k++) {
		my $block = $blocks->[$k];
		my $rel = $rels->[$k];
		push @out_blocks, "$block($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_blocks), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$interfering_blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Gb', '[blocks,in,intrfring,set,1](rel1)', '[blocks,in,intrfring,set,2](rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($blocksets, $rels) = @{$interfering_blocksets->[$gb]};
	    unless (defined $blocksets) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_sets = @{$blocksets};
	    my @out_blocksets = ();
	    for (my $k=0; $k<$ct_sets; $k++) {
		my $blockset = $blocksets->[$k];
		my $cnct_blockset = join (',', @{$blockset});
		my $rel = $rels->[$k];
		push @out_blocksets, "[$cnct_blockset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_blocksets), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_cmpt_lmpp > 3) {...}"


      if ($TEST_LEVEL_cmpt_lmpp > 5) {

	print STDOUT '( Contents of @{$cmpl_interfering_blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Gb', '[cmpl,blocks,in,intrfring,set,1](rel1)', '[cmpl,blocks,in,intrfring,set,2](rel2)', '...'), "\n\n";
	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my ($cblocksets, $rels) = @{$cmpl_interfering_blocksets->[$gb]};
	    unless (defined $cblocksets) {
		print STDOUT join ("\t", $gb, 'none'), "\n";
		next;
	    }
	    my $ct_csets = @{$cblocksets};
	    my @out_cblocksets = ();
	    for (my $k=0; $k<$ct_csets; $k++) {
		my $cblockset = $cblocksets->[$k];
		my $cnct_cblockset = join (',', @{$cblockset});
		my $rel = $rels->[$k];
		push @out_cblocksets, "[$cnct_cblockset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out_cblocksets), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of %{$br2up_down2blocksets} )', "\n\n";

	print STDOUT join ("\t", 'Br_ID', 'up(U)_or_down(L)', '[blocks,in,set,1]', '[blocks,in,set,2]', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocksets}) {
	    my $up_down2blocksets = $br2up_down2blocksets->{$br};
	    foreach my $u_or_d (sort keys %{$up_down2blocksets}) {
		my @out_blsets = ();
		foreach my $blset (@{$up_down2blocksets->{$u_or_d}}) {
		    my $cnct_blset =  join (',', @{$blset});
		    push @out_blsets, "[$cnct_blset]";
		}
		print STDOUT join ("\t", $br, $u_or_d, @out_blsets), "\n";
	    }
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_cmpt_lmpp > 5) {...}"

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T3) # ...}"


	my $gblock_sizes = mk_list_gblock_sizes (@{$bds_blocks0}, 0, 1, @{$inter_block_relations});
	unless ($gblock_sizes) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'mk_list_gblock_sizes' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_list_gblock_sizes' Ended ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5a) #

	print STDOUT "[ Output of 'mk_list_gblock_sizes' ]\n\n";
	print STDOUT '@{$gblock_sizes} = (', join (', ', @{$gblock_sizes}), ") .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5a) # ...}"


	my ($br2up_down2blocks, $coll_cmpl_gblocks, $gblock2coll_cmpl) = search_for_coll_cmpl_gblks (@{$list_gblocks0}, $indx_br, $indx_u_or_d, $topnode_id, %{$node2ch}, %{$node2depth}, @{$gblock_sizes});  
	unless (defined $gblock2coll_cmpl) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'search_for_coll_cmpl_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_coll_cmpl_gblocks = @{$coll_cmpl_gblocks};

    (print STDOUT " ... The subroutine 'search_for_coll_cmpl_gblks' Ended (w/ #{coll-cmpl-gblocks}=$ct_coll_cmpl_gblocks) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5b) #

	print STDOUT "[ Outputs of 'search_for_coll_cmpl_gblks' ]\n\n";

	print STDOUT '( Contents of %{$br2up_down2blocks} )', "\n\n";
	print STDOUT join ("\t", 'Br_ID', 'up(U)_or_down(L)', '[blocks,delimited,as,specified]'), "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocks}) {
	    my $up_down2blocks = $br2up_down2blocks->{$br};
	    foreach my $u_or_d (sort keys %{$up_down2blocks}) {
		my $cnct_blocks = join (',', @{$up_down2blocks->{$u_or_d}});
		print STDOUT join ("\t", $br, $u_or_d, "[$cnct_blocks]"), "\n";
	    }
	}
	print STDOUT "\n\n";

	print STDOUT '( Contents of @{$coll_cmpl_gblocks} ', "(w/ #{elements}=$ct_coll_cmpl_gblocks) )\n\n";
	print STDOUT join ("\t", 'Indx_ccgb', '[gap,blocks,constituting,the,set]'), "\n\n";
	for (my $k=0; $k<$ct_coll_cmpl_gblocks; $k++) {
	     my $cnct_ccgb_set = join (',', @{$coll_cmpl_gblocks->[$k]});
	     print STDOUT join ("\t", $k, "[$cnct_ccgb_set]"), "\n";
	}
	print STDOUT "\n\n";

      if ($TEST_LEVEL_cmpt_lmpp > 3) {

	print STDOUT '( Contents of @{$gblock2coll_cmpl} )', "\n\n";
	print STDOUT join ("\t", 'Gb', '(indices,of,ccgbs,that,Gb,belongs,to)' ), "\n\n";

	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    my $cnct_ccgbs = join (',', @{$gblock2coll_cmpl->[$gb]});
	    print STDOUT join ("\t", $gb, "($cnct_ccgbs)"), "\n";
	}
	print STDOUT "\n\n";
	
      } # End of "if ($TEST_LEVEL_cmpt_lmpp > 3) {...}"

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5b) # ...}"


=DISABLED because these are no longer necessary. (2)

	my ($cts_vincluding, $cts_vincluded, $cts_onx) = cts_vincluding_vincluded_onx_gblks (@{$inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5c) #

	print STDOUT "[ Outputs of 'cts_vincluding_vincluded_onx_gblks' ]\n\n";
	print STDOUT '( Contents of @{$cts_vincluding}, @{$cts_vincluded} & @{$cts_onx} )', "\n\n";
	print STDOUT join ("\t", 'Gb', '#{vincluding}', '#{vincluded}', '#{onx}'), "\n\n";

	for (my $gb=0; $gb<$ct_blks; $gb++) {
	    print STDOUT join ("\t", $gb, $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb]), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5c) # ...}"


	my $sets_vequiv_gblks = mk_sets_vert_equiv_gblks (%{$br2up_down2blocks}, %{$br2eq});
	unless (defined $sets_vequiv_gblks) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'mk_sets_vert_equiv_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_sets_vequiv_gblks = @{$sets_vequiv_gblks};

    (print STDOUT " ... The subroutine 'mk_sets_vert_equiv_gblks' Ended (w/ #{sets of v-equiv g-blocks}=$ct_sets_vequiv_gblks) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5d) #

	print STDOUT "[ Outputs of 'mk_sets_vert_equiv_gblks' ]\n\n";
	print STDOUT '( Contents of @{$sets_vequiv_gblks} )', "\n\n";
	print STDOUT join ("\t", 'Indx_set', '[v-equiv,gap,blocks,constituting,the,set]'), "\n\n";
	for (my $i=0; $i<$ct_sets_vequiv_gblks; $i++) {
	    my $cnct_set = join (',', @{$sets_vequiv_gblks->[$i]});
	    print STDOUT join ("\t", $i, "[$cnct_set]"), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5d) # ...}"


	my $sets_swappable_gblks = mk_sets_swappable_gblks (@{$sets_vequiv_gblks}, @{$gblock_sizes});
	unless (defined $sets_swappable_gblks) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'mk_sets_swappable_gblks' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_sets_swappable_gblks = @{$sets_swappable_gblks};

    (print STDOUT " ... The subroutine 'mk_sets_swappable_gblks' Ended (w/ #{sets of swappable g-blocks}=$ct_sets_swappable_gblks) ...\n\n") if ($TEST_LEVEL_nr_msht_gblks >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5e) #

	print STDOUT "[ Outputs of 'mk_sets_swappable_gblks' ]\n\n";
	print STDOUT '( Contents of @{$sets_swappable_gblks} )', "\n\n";
	print STDOUT join ("\t", 'Indx_set', '[swappable,gap,blocks,constituting,the,set]'), "\n\n";
	for (my $i=0; $i<$ct_sets_swappable_gblks; $i++) {
	    my $cnct_set = join (',', @{$sets_swappable_gblks->[$i]});
	    print STDOUT join ("\t", $i, "[$cnct_set]"), "\n";
	}
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_nr_msht_gblks > 2) { # Added on 2019/05/28. (T5e) # ...}"

=cut # END of "DISABLED because these are no longer necessary. (2)"


	my @set_all_nodes = sort {$a <=> $b} keys %{$node2depth};

	my ($gpattern2indx, $set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$set_ingredients_indel_hsts};

	my ($set_cls_gblocks0, $min_totcts_indels_for_cls0, $sets_psm_hsts_for_cls0, $cts_indels0_for_cls0, $set_psm_hsts0_for_cls0)
	   = enum_psm_indel_scns_for_msa_hs_mrounds (%{$node2ch}, %{$node2pa}, $topnode_id, @{$exnodes_in_msa}, @{$set_columns0}, $char_abs, $char_pres, @set_all_nodes, %{$node2depth}, %{$node2ct_exoffs}, %{$gpattern2indx}, @{$set_gpatterns}, @{$set_node2dparsstat}, @{$set_br2change});
	unless (defined $set_cls_gblocks0) {
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' FAILED: $! !\n\n";
	    return (0, undef, undef, undef, undef, undef, undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks0};

    (print STDOUT " ... The subroutine 'enum_psm_indel_scns_for_msa_hs_mrounds' Ended (w/ #{gapped segments}=$ct_cls) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

		# Added in creation of ''nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #
	my $minct_indels0 = 0;
	foreach my $totct (@{$min_totcts_indels_for_cls0}) { $minct_indels0 += $totct; }
		# End of "Added in creation of ''nonrec_mlt_shifts_gapblocks' (on 2019/05/26)." #

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5f) #


      if ($TEST_LEVEL_cmpt_lmpp >= 5) { # Added on 2019/05/29. (T5f') #

	my $ct_gpatterns0=@{$set_gpatterns};

	print STDOUT '[ INITIAL Contents of @{$set_gpatterns},', " one of the Main Effective Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds' (#{gap-patterns}=$ct_gpatterns0) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', 'gap-pattern'), "\n\n";

	for (my $i=0; $i<$ct_gpatterns0; $i++) {
	    print STDOUT join ("\t", $i, ${$set_gpatterns->[$i]}), "\n";
	}
	print STDOUT "\n\n";

      } # End of "if ($TEST_LEVEL_cmpt_lmpp >= 5) { # Added on 2019/05/29. (T5f') # ...}"

	print STDOUT "[ Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds' (#{gapped segments}=$ct_cls; Min[Tot#{indels}]=$minct_indels0) ]\n\n";

	print STDOUT "# NOTE: ... Here, we will OMIT displaying some main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel histories),\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL\"!!\n\n\n";

     if ($TEST_LEVEL_cmpt_lmpp >= 5) {
	print STDOUT '(( Contents of @{$set_cls_gblocks0}, @{$min_totcts_indels_for_cls0}, @{$sets_psm_hsts_for_cls0}, @{$cts_indels0_for_cls0}, & @{$set_psm_hsts0_for_cls0} ))', "\n\n";
     } else {
	print STDOUT '(( Contents of @{$set_cls_gblocks0}, @{$min_totcts_indels_for_cls0} & @{$cts_indels0_for_cls0} ))', "\n\n";
     } # End of "if ($TEST_LEVEL_cmpt_lmpp >= 5) {...} else {...}".

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Test outer for-loop (over gapped segments).

	    my $set_gpbs = $set_cls_gblocks0->[$cl];
	    my $ct_gpbs = @{$set_gpbs};
	    
	    print STDOUT "||  The $cl th gapped segment (#{gap-pattern-blocks}=$ct_gpbs) ||\n\n";
	    print STDOUT '( Gap-pattern structure )', "\n\n";
	    print STDOUT join ("\t", 'Indx_block', 'indx_gap_pattern', 'length', 'beg', 'end'), "\n\n";
	    for (my $gpb=0; $gpb<$ct_gpbs; $gpb++) {
		print STDOUT join ("\t", $gpb, @{$set_gpbs->[$gpb]}), "\n";
	    }
	    print STDOUT "\n\n";

	    my $min_totct_indels = $min_totcts_indels_for_cls0->[$cl];
	    my $ct_indels0 = $cts_indels0_for_cls0->[$cl];

	    my $psm_hst0_for_cls = $set_psm_hsts0_for_cls0->[$cl];
	    my $psm_hsts_for_cls = $sets_psm_hsts_for_cls0->[$cl];
	    my $ct_psm_hsts = @{$psm_hsts_for_cls};

	    print STDOUT 'Min#{indels}   = ', $min_totct_indels, " ,\n";
	    print STDOUT '#{indels}_init = ', $ct_indels0, " ,\n";
	    print STDOUT '#{parsimonious indel histories} = ', $ct_psm_hsts, " .\n\n";


=DISABLED

		# NOTE: If you want to know details on the MAIN outputs (storing information on the indel histories), 
		# SWITCH ON "$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL"!!

     if ($TEST_LEVEL_nr_msht_gblks >= 5) {

	    my ($br2indels0, $node2blwise_gstat0) =  @{$psm_hst0_for_cls};

	    print STDOUT '| Initial candidate for the parsimonious indel history |', "\n\n";

	    print STDOUT '( Contents of %{$br2indels0} )', "\n\n";
	    print STDOUT join ("\t", 'Br_ID', '[stat_ue/stat_le/involved,blocks]_1', '[stat_ue/stat_le/involved,blocks]_2', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) {
		my $info_indels = $br2indels0->{$br};
		unless (defined $info_indels) { next; }
		my @out_indels = ();
		foreach my $info_indel (@{$info_indels}) {
		    my ($stat_ue, $stat_le, $invlvd_blks) = @{$info_indel};
		    unless (defined $stat_ue) { next; }
		    my $cnct_invlvd_blks = join (',', @{$invlvd_blks});
		    push @out_indels, "[$stat_ue/$stat_le/$cnct_invlvd_blks]";
		}
		print STDOUT join ("\t", $br, @out_indels), "\n";
	    }
	    print STDOUT "\n\n";

     } # End of "if ($TEST_LEVEL_nr_msht_gblks >= 5) {...}"
=cut # END of "DISABLED"

	    print STDOUT "\n";

	} # End of the test outer for-loop (over gapped segments).


    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5f) # ...}"


	my ($totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, $br2log_mfacs) = @{$set_ingredients_prb_msa_indel}; # INCORPORATED INTO THE INPUT SET of this subroutine itself (on 2019/05/23).
	my $size_wd = $ct_clms;

			# MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30). #
	## [ my ($tot_ln_prob_gpattern0, $shared_fac10, $shared_fac20, $set_clwise_tot_ln_prb_gp0, $tot_div_tree0, $ct_ng_clms0, $setset_ln_prb_gp_hsts0, $set_rlv_indel_hst2corr_fac0) 
	## = apprx1_tot_lnlk_gpattern_msa2_hs ] ... before the revision on 2020/04/30.
	##
	my ($tot_ln_prob_gpattern0, $shared_fac10, $shared_fac20, $set_clwise_tot_ln_prb_gp0, $tot_div_tree0, $ct_ng_clms0, $setset_ln_prb_gp_hsts0, $set_rlv_indel_hst2corr_fac0, $shared_fac30, $totct_inter_gapless_clm_positions0, $ct_ends_flnk_gapless_clms0) 
	= apprx1_wlp_tot_lnlk_gpattern_msa2_hs ($topnode_id, %{$br2attr}, %{$node2pa}, %{$node2ch}, $size_wd, $char_abs, @{$set_cls_gblocks0}, @{$set_psm_hsts0_for_cls0}, @{$sets_psm_hsts_for_cls0},
	   $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, %{$br2log_mfacs}, $if_on_le_msa, $if_on_re_msa); # ... MUST REVISE the INPUT SET of 'apprx1_tot_lnlk_gpattern_msa2_hs'!! (... DONE!!)
			# END of "MAJOR REVISION along with the creation of 'ANEX_P.ver0.7' (on 2020/04/30)." #

	unless (defined $tot_ln_prob_gpattern0) {
	    ##print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' FAILED: $! !\n\n";
	    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' FAILED: $! !\n\n"; # Modified on 2020/04/30.
	    return (0, undef, undef, $minct_indels0, undef, undef, undef, undef);
	}

	my $ln_prb_indel0 = $tot_ln_prob_gpattern0; # Added in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26). #

    ##(print STDOUT " ... The subroutine 'apprx1_tot_lnlk_gpattern_msa2_hs' Ended (w/ \$ln_prb_indel0 = $ln_prb_indel0) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.
    (print STDOUT " ... The subroutine 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' Ended (w/ \$ln_prb_indel0 = $ln_prb_indel0) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # => Modified on 2020/04/30.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5g) #

	##print STDOUT "[ Outputs of 'apprx1_tot_lnlk_gpattern_msa2_hs' ]\n\n";
	print STDOUT "[ Outputs of 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' ]\n\n";

	print STDOUT "# NOTE: ... Here, we will OMIT displaying some main outputs ...\n";
	print STDOUT "# If you want to know further DETAILS on the MAIN outputs (storing information on the indel probabilities),\n";
	##print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL\"!!\n\n\n";
	print STDOUT "# SWITCH ON \"\$MyTreeMap_indels_ML_hs_hs_wLP::TEST_LEVEL\"!!\n\n\n"; # Modified on 2020/04/30.

	print STDOUT '$tot_ln_prob_gpattern0 = ', $tot_ln_prob_gpattern0, " .\n\n";

	print STDOUT '( Contents of @{$set_clwise_tot_ln_prb_gp0} )', "\n\n";

	print STDOUT join ("\t", 'Indx_cl', 'tot_ln_prb_gp'), "\n\n";
	for (my $cl=0; $cl<$ct_cls; $cl++) {
	    my $tot_ln_prb_gp = $set_clwise_tot_ln_prb_gp0->[$cl];
	    unless (defined $tot_ln_prb_gp) { $tot_ln_prb_gp = 'undef'; }
	    print STDOUT join ("\t", $cl, $tot_ln_prb_gp), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT "( Secondary outputs )\n\n";

	print STDOUT '$shared_fac10  = ', $shared_fac10, " ,\n";
	print STDOUT '$shared_fac20  = ', $shared_fac20, " ,\n";
	print STDOUT '$tot_div_tree0 = ', $tot_div_tree0, " ,\n";
	print STDOUT '$ct_ng_clms0   = ', $ct_ng_clms0, " .\n\n";

			# (ADDED on 2020/04/30.) #
	print STDOUT '$shared_fac30  = ', $shared_fac30 , " ,\n";
	print STDOUT '$totct_inter_gapless_clm_positions0 = ', $totct_inter_gapless_clm_positions0, " ,\n";
	print STDOUT '$ct_ends_flnk_gapless_clms0 = ', $ct_ends_flnk_gapless_clms0, " .\n\n";
			# End of "(ADDED on 2020/04/30.)" #

	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5g) # ...}"

		## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks', from 2019/05/20 till 2019/05/dd.)"


		##  Process (1): Prepare some important variables. ##

    (print STDOUT "< Process (1): Prepare some important variables. >\n\n") if ($TEST_LEVEL_cmpt_lmpp >1); # Added on 2019/05/28.

   (print STDOUT "| First, compute the substitution component of the log-probability of the initial alignment |\n\n") if ($TEST_LEVEL_cmpt_lmpp > 2);

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }
	my $cnct_null_clm = join (':', @null_clm);

	unless (defined $clm2ln_prob_sbst->{$cnct_null_clm}) { $clm2ln_prob_sbst->{$cnct_null_clm} = 0; }
		# NOTE: The probability of a null column is always 1 (unity) under any substitution model (in which each column evolves independently).

	my $ct_null_clms0 = 0;
	my $ln_prb_sbst0 = 0;

	for (my $c = 0; $c < $ct_clms; $c++) { # The 3rd preliminary for-loop (over columns).

	    my $column0 = $set_columns0->[$c];
	    my $cnct_clm0 = join (':', @{$column0});
	    if ($cnct_clm0 eq $cnct_null_clm) {
		$ct_null_clms0++;
		next; # The log-probability of the null column is 0 (ZERO).
	    }

		# Examine whether the probability of the column is already computed or not.

	    my $ln_cw_prb0 = $clm2ln_prob_sbst->{$cnct_clm0};
	    if (defined $ln_cw_prb0) {
		$ln_prb_sbst0 += $ln_cw_prb0;
		next;
	    }

		# Newly compute the probability of the column. #

	    my $cw_prb0 = br_calc_sbst_prob_sgl_clm (@{$exnodes_in_msa}, $char_abs, @{$column0}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$base_freqs_root}, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx});
	    unless (defined $cw_prb0) {
		print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'br_calc_sbst_prob_sgl_clm' FAILED on the $c th column, whose concatenation is '$cnct_clm0': $! !\n\n";
		return (0, undef, undef, $minct_indels0, undef, undef, $ln_prb_indel0, undef);
	    }

	    $ln_cw_prb0 = log ($cw_prb0);

    (print STDOUT " ... The subroutine 'br_calc_sbst_prob_sgl_clm' Ended on the $c th column (w/ \$ln_cw_prb0 = $ln_cw_prb0) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp > 2) ; # Added on 2019/05/28.

	    $clm2ln_prob_sbst->{$cnct_clm0} = $ln_cw_prb0;
	    $ln_prb_sbst0 += $ln_cw_prb0;

	} # End of the 3rd preliminary for-loop (over columns).

	my $ln_prb_msa0 = $ln_prb_sbst0 + $ln_prb_indel0; # Added on 2019/05/27.
	
    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5g-h) #
	print STDOUT "\n";
	print STDOUT '$ct_null_clms0  = ', $ct_null_clms0, " ,\n";
	print STDOUT '$ln_prb_sbst0   = ', $ln_prb_sbst0, " ,\n";
	print STDOUT '($ln_prb_indel0 = ', $ln_prb_indel0, ") ,\n";
	print STDOUT '$ln_prb_msa0    = ', $ln_prb_msa0, " .\n";
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5g-h) # ...}"


		## Process (2): Initialize some important variables. ##

	my @bl_shifts = (); # The $k th element specifies the "next" shift (the direction and the amount) of the $k th block. (= +1 if to the right, -1 if to the left.)
	for (1 .. $ct_blks) { push @bl_shifts, 1; }

	my @bl_coords0 = @{$org_bl_coords}; # Initial values of the block coordinates.


			## Compute some topology-related codes. ##
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23).) ##

	my ($degeneracy0, $cnct_sets_ccgbs_cs_dgnrcy0) = compt_aln_degeneracy (@{$coll_cmpl_gblocks}, @{$bds_blocks0}, @bl_coords0, @{$bds_bl_coords}, @{$inter_block_relations}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0});
	unless (defined $cnct_sets_ccgbs_cs_dgnrcy0) {
		print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'compt_aln_degeneracy' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, undef, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy0 = $degeneracy0)...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5h) #

	print STDOUT "[ Outputs of 'compt_aln_degeneracy' ]\n\n";
	print STDOUT '$degeneracy0 = ', $degeneracy0, " ,\n";
	print STDOUT '$cnct_sets_ccgbs_cs_dgnrcy0 = ', $cnct_sets_ccgbs_cs_dgnrcy0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5h) # ...}"


	my $code_topology0 = encode_gptrn_topology_psm (@{$bds_blocks0}, @{$inter_block_relations}, @{$interfering_blocks}, @{$interfering_blocksets}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0});
	unless (defined $code_topology0) {
		print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'encode_gptrn_topology_psm' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'encode_gptrn_topology_psm' Ended ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5i) #

	print STDOUT "[ Output of 'encode_gptrn_topology_psm' ]\n\n";
	print STDOUT '$code_topology0 = ', $code_topology0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5i) # ...}"


	my @no_gbs_on_msa_ends = (); 
	for (my $k=0; $k<$ct_blks; $k++) { $no_gbs_on_msa_ends[$k] = 'N'; }
	my $cnct_no_gbs_on_msa_ends = join (',', @no_gbs_on_msa_ends); # Code indicating NO gap-blocks are on either MSA-end.

	my $if_gbs_on_msa_ends0 = ($if_on_le_msa or $if_on_re_msa) ?
	  examine_if_gbs_on_msa_ends (@bl_coords0, @{$gb_coords_wd_frm}, $if_on_le_msa, $if_on_re_msa) :
		 \@no_gbs_on_msa_ends ;
	unless (defined $if_gbs_on_msa_ends0) {
		print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'examine_if_gbs_on_msa_ends' FAILED: $! !\n\n";
		return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
	}
    (print STDOUT " ... The subroutine 'examine_if_gbs_on_msa_ends' Ended ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

	my $cnct_if_gbs_on_msa_ends0 = join (',', @{$if_gbs_on_msa_ends0});

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5j) #

	print STDOUT "[ Output of 'examine_if_gbs_on_msa_ends' ]\n\n";
	print STDOUT '$cnct_if_gbs_on_msa_ends0 = ', $cnct_if_gbs_on_msa_ends0, " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5j) # ...}"

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/23).)" ##

			## Initialize the hash storing pre-computed indel components of the log-probabilities of the alignment.
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).) ##

	my $grand_code_topology0 = join ('::', $code_topology0, $cnct_sets_ccgbs_cs_dgnrcy0, $cnct_if_gbs_on_msa_ends0);

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/26).)" ##

    if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5k) #

	print STDOUT '[ Initial key for  %gcode_topo2diff_ln_prb_indel_plus ]', "\n\n";
	print STDOUT '$grand_code_topology0  = ', $grand_code_topology0 , " .\n";
	print STDOUT "\n\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp > 2) { # Added on 2019/05/28. (T5k) # ...}"

			## Re-assign ($degeneracy0 = 0), if the initial alignment satisfies some conditions. ##
			## (ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28).) ##

	if ($degeneracy0 >0) {

	    if ($ct_null_clms0>0) {
		# If the alignment contains some null-columns,
		# it MUST be excluded from the analysis.

    (print STDOUT " ... Because (\$ct_null_clms0 == 0), Set \$degeneracy = 0 !!\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.
		$degeneracy0 = 0;

	    } elsif ($ct_x_bhrels>0) {

		# NOTE: "(!($if_ovlp_bw_veq0) and $if_right_order_swpbl0)" MUST usually be satisfied.
		# 	Thus, we will NOT perform 'chk_ovlp_bw_vequiv_gblks' or 'chk_pos_orders_amg_swpbl_gblks'. # 

		my $if_excluded0 = chk_excluded_bin_hrels (@{$list_excluded_bin_hrels}, @{$bds_blocks0}, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @{$inter_block_relations});
		unless (defined $if_excluded0) {
		    print STDERR "[ERROR 'compt_log_msa_prob_plus'] The subroutine 'chk_excluded_bin_hrels' FAILED on the initial alignment: $! !\n\n";
		    return (0, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, undef);
		}
    (print STDOUT " ... The subroutine 'chk_excluded_bin_hrels' Ended (w/ \$if_excluded0 = $if_excluded0) ...\n\n") if ($TEST_LEVEL_cmpt_lmpp >1) ; # Added on 2019/05/28.

		if ($if_excluded0) { $degeneracy0 = 0; } # The gap-block configuration indicated that the initial alignment MUST be excluded from the analysis.

	    } # End of "if ($ct_null_clms0>0) {...} elsif ($ct_x_bhrels>0) {...}".

	} # End of "if ($degeneracy0 >0) {...}".

    (print STDOUT "\n", '{FINAL VALUE of $degeneracy0} = ', $degeneracy0, " .\n\n\n\n") if ($TEST_LEVEL_cmpt_lmpp > 2); # Added on 2019/05/28. #

			## END of "(ADDED in creation of 'nonrec_mlt_shifts_gapblocks' (on 2019/05/28).)" ##

    if ($TEST_LEVEL_cmpt_lmpp2 > 1) {

	print STDOUT '(( Contents of %{$clm2ln_prob_sbst} AFTER ALL the', " \"shift\"s:  ))\n\n";
	print STDOUT join ("\t", 'concatenated:site:numbers:within:column', 'ln_prob_sbst'), "\n\n";
	foreach my $clm (sort keys %{$clm2ln_prob_sbst}) {
		print STDOUT join ("\t", $clm, $clm2ln_prob_sbst->{$clm}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmpt_lmpp2 > 1) {...}"

    if ($TEST_LEVEL_cmpt_lmpp >= 5) { # Added on 2019/05/29. (T9) #

	my $ct_gpatterns_fnl=@{$set_gpatterns};

	print STDOUT '[ Contents of @{$set_gpatterns},', " one of the Main Effective Outputs of 'enum_psm_indel_scns_for_msa_hs_mrounds', AFTER all the \"shift\"s (#{gap-patterns}=$ct_gpatterns_fnl) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', 'gap-pattern'), "\n\n";

	for (my $i=0; $i<$ct_gpatterns_fnl; $i++) {
	    print STDOUT join ("\t", $i, ${$set_gpatterns->[$i]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmpt_lmpp >= 5) { # Added on 2019/05/28. (T9) #...}"


    (print STDOUT "<< Subtoutine 'compt_log_msa_prob_plus' Ended  />>\n\n") if ($TEST_LEVEL_cmpt_lmpp >0);


		## MAJOR MODIFICATION in creation of 'compt_log_msa_prob_plus'. (OUT) ##
	return (1, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $grand_code_topology0);
	##	## MAJOR MODIFICATION in creation of 'nonrec_mlt_shifts_gapblocks'. (OUT) ##
	##return (1, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $diffs_ln_prbs_plus);
	####return ($ln_prb_sbst0, $diffs_ln_prbs_sbst); 

}	# End of the subroutine, 'compt_log_msa_prob_plus'. #


# serial_shifts_sgl_gblock:
#	This is the MAIN satellite subroutine of 'serial_shifts_point2point';
#	it "shift"s a specified gap-block multiple times SERIALLY,
#	while also virtually extending and/or retracting the specified and/or other gap-blocks, if necessary.
#	[CREATION STARTED on 2019/07/08, ENDED on 2019/07/09.]
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb,
#	$bl_sbj, $incr_coord_sbj, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes), where
#
#	@curr_set_clms is the current set of alignment columns, that is,
#	@column = @{$curr_set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@curr_set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @gb_coords_wd_frm below];
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]}
#	give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@gb_coords_wd_frm is an output of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	@{$curr_gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are currently at the $lb th column;
#	@{$curr_gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are currently at the $rb th column;
#	[Typically, \@curr_gblks_w_spec_lb and \@curr_gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	$bl_sbj specifies the subject gap-block;
#
#	$incr_coord_sbj is an integer providing the increment of the coordinate of the $bl_sbj th gap-block;
#	$incr_coord_sbj > 0 means that the block will be "shift"ed to the right,
#			 < 0 means that the block will be "shift"ed to the left;
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column), with @null_column consisting of as many "absence" states as the MSA rows;
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];

#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[\@cts_vincluding, \@cts_vincluded & \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#
# OUTPUTS: ($flag, $incr_ct_null_clms), where
#
#	$flag = 0 if all the process ended successfully,
#	      = -1 if the subject block stepped over the left-end of the local MSA,
#	      = +1 if the subject block stepped over the right-end of the local MSA,
#	      = undef if another type of error occurs;
#
#	$incr_ct_null_clms is the increment of the count of null-columns resulting from the series of "shift"s (= undef if an error occurs).
#
sub serial_shifts_sgl_gblock (\@\@\@\@\@\@$$\@$\@\@\@\@\@) {

	my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gb_coords_wd_frm, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	$bl_sbj, $incr_coord_sbj, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes) = @_;

##our $TEST_LEVEL_ssh_sgblk = 0;

	my $ct_clms = @{$curr_set_clms};

	my $bl_coord_bf = $curr_bl_coords->[$bl_sbj]; # Keep the coordinate before the serial-"shift"s.

	my $bds_block_sbj = $curr_bds_blks->[$bl_sbj];
	my ($lbd_bf, $rbd_bf) = @{$bds_block_sbj}; # Keep the boundaries of the subject block before the serial-"shift"s.


    (print STDOUT "< Subroutine 'serial_shifts_sgl_gblock' (#{columns} = $ct_clms, ", ', $bl_sbj = ', $bl_sbj, ', $incr_coord_sbj = ', $incr_coord_sbj, ", {block coordinate}_bf = $bl_coord_bf, [lbd, rbd]_bf = [$lbd_bf, $rbd_bf]) >\n\n") if ($TEST_LEVEL_ssh_sgblk > 0);

	my $bl_coord_af = $bl_coord_bf + $incr_coord_sbj;

	my ($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm->[$bl_sbj]};

	unless ($gb_coord_le_wd <= $bl_coord_af) {
		# The subjet block WILL step over the left-end of the window (i.e., local MSA).

	    print STDOUT "[ERROR 'serial_shifts_sgl_gblock'] The coordinate of the subject ($bl_sbj th block) AFTER the serial shift (=$bl_coord_af) is ON THE LEFT of the left-end of the window (=$gb_coord_le_wd): $! !\n\n";
	    return (-1, undef);
	}

	unless ($bl_coord_af <= $gb_coord_re_wd) {
		# The subjet block WILL step over the right-end of the window (i.e., local MSA).

	    print STDOUT "[ERROR 'serial_shifts_sgl_gblock'] The coordinate of the subject ($bl_sbj th block) AFTER the serial shift (=$bl_coord_af) is ON THE RIGHT of the right-end of the window (=$gb_coord_re_wd): $! !\n\n";
	    return (+1, undef);
	}

	if ($incr_coord_sbj == 0) {
	    print STDOUT "[INFORMATION 'serial_shifts_sgl_gblock'] \$incr_coord_sbj is ZERO, thus NOTHING needs to be DONE...\n\n";
	    return (0, 0);
	}

	my $sh = ($incr_coord_sbj>0) ? +1 : -1; # Determine the "direction" of the shift.
	my $ct_shifts = ($incr_coord_sbj>0) ? $incr_coord_sbj : - $incr_coord_sbj ;

	my $incr_ct_null_clms = 0;
	foreach my $time (1 .. $ct_shifts) { # The main foreach-loop (over the times of the "shift"s). #

	    my $bl_coord_curr = $curr_bl_coords->[$bl_sbj];

   (print STDOUT "{ The $time th \"shift\" (w/ current-coordinate = $bl_coord_curr) }\n\n") if ($TEST_LEVEL_ssh_sgblk>1);

			### (Pre-Processes) Virtually extend the gap-blocks if necessary. ###

		## (IMPORTED with MODIFICATIONS from 'nonrec_mlt_shifts_gapblocks' (on 2019/07/08).) (1a) ##

	    my ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; # Block-boundaries BEFORE the "shift".

    (print STDOUT "\n", 'Before v-extensions: ($lb_sbj_bf, $rb_sbj_bf) = ', "($lb_sbj_bf, $rb_sbj_bf) .\n\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

	    if ($sh>0) {
		if (($cts_vincluding->[$bl_sbj] + $cts_onx->[$bl_sbj] >0) and (@{$curr_gblks_w_spec_lb->[$rb_sbj_bf+1]}>0)) {
		    my $ct_vextensions = vextend_right_end_sbj_gblk ($bl_sbj, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextensions) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vextend_right_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_right_end_sbj_gblk' ended (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_ssh_sgblk > 2);
		    if ($ct_vextensions>0) { ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; } # Update Block-boundaries BEFORE the "shift".
		}
		if (($cts_vincluded->[$bl_sbj]>0) and (@{$curr_gblks_w_spec_lb->[$rb_sbj_bf+1]}>0)) {
		    my $ct_vextended = vextend_rear_ends_rflnk_vincluded_gblks ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vextended) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vextend_rear_ends_rflnk_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_rear_ends_rflnk_vincluded_gblks' ended (w/ \$ct_vextended = $ct_vextended) ...\n\n") if ($TEST_LEVEL_ssh_sgblk > 2);
		    ##if ($ct_vextended>0) { ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; } # Update Block-boundaries BEFORE the "shift". # Probably superfluous. #
		}

	    } else { # if ($sh<0) 

		if (($cts_vincluding->[$bl_sbj] + $cts_onx->[$bl_sbj] >0) and (@{$curr_gblks_w_spec_rb->[$lb_sbj_bf-1]}>0)) {
		    my $ct_vextensions = vextend_left_end_sbj_gblk ($bl_sbj, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextensions) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vextend_left_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_left_end_sbj_gblk' ended (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_ssh_sgblk > 2);
		    if ($ct_vextensions>0) { ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; } # Update Block-boundaries BEFORE the "shift".
		}
		if (($cts_vincluded->[$bl_sbj]>0) and (@{$curr_gblks_w_spec_rb->[$lb_sbj_bf-1]}>0)) {
		    my $ct_vextended = vextend_rear_ends_lflnk_vincluded_gblks ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vextended) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vextend_rear_ends_lflnk_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", before shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vextend_rear_ends_lflnk_vincluded_gblks' ended (w/ \$ct_vextended = $ct_vextended) ...\n\n") if ($TEST_LEVEL_ssh_sgblk > 2);
		    ##if ($ct_vextended>0) { ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; } # Update Block-boundaries BEFORE the "shift". # Probably superfluous. #
		}

	    } # End of "if ($sh>0) {...} else {...}".

	    ($lb_sbj_bf, $rb_sbj_bf) = @{$curr_bds_blks->[$bl_sbj]}; # Update Block-boundaries BEFORE the "shift".

    (print STDOUT "\n", 'After v-extensions: ($lb_sbj_bf, $rb_sbj_bf) = ', "($lb_sbj_bf, $rb_sbj_bf) .\n\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

		## END of "(IMPORTED with MODIFICATIONS from 'nonrec_mlt_shifts_gapblocks' (on 2019/07/08).) (1a)" ##


			### (CORE) Actually perform the "shift". ###

	    my ($flag, $sb_incr_ct_null_clms) = shift_gblock (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, $bl_sbj, $sh, @{$rows_affd_by_blks}, $cnct_null_clm);
	    unless ((defined $flag) and ($flag == 0)) {
		my $flag_out = (defined $flag) ? $flag : 'undef' ;
		print STDOUT "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'shift_gblock' FAILED (w/ \$flag = $flag_out): $! !\n\n";
		return ($flag, undef);
	    }

	    $incr_ct_null_clms += $sb_incr_ct_null_clms;


			### (Post-Processes) Virtually retract the gap-blocks if necessary. ###

		## (IMPORTED with MODIFICATIONS from 'nonrec_mlt_shifts_gapblocks' (on 2019/07/08).) (1b) ##

	    my ($lb_sbj_af, $rb_sbj_af) = @{$curr_bds_blks->[$bl_sbj]}; # Block-boundaries AFTER the "shift". #

    (print STDOUT "\n", 'Before v-retractions: ($lb_sbj_af, $rb_sbj_af) = ', "($lb_sbj_af, $rb_sbj_af) .\n\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

	    if ($sh >0) {
		if (($cts_vincluded->[$bl_sbj] >0) and (@{$curr_gblks_w_spec_rb->[$rb_sbj_af]}>1)) {
		    my $ct_vretracted = vretract_right_ends_vincluded_gblks ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vretract_right_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_ssh_sgblk > 2);
		}
		if (($cts_vincluding->[$bl_sbj]>0) and (@{$curr_gblks_w_spec_lb->[$lb_sbj_af]}>1)) {
		    my $ct_vretracting = vretract_left_end_sbj_gblk ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vretract_left_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_end_sbj_gblk' ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

		    if ($ct_vretracting>0) { ($lb_sbj_af, $rb_sbj_af) = @{$curr_bds_blks->[$bl_sbj]}; } # Update the block-boundaries AFTER the "shift".
		}

	    } else { # if ($sh <0)
		if (($cts_vincluded->[$bl_sbj] >0) and (@{$curr_gblks_w_spec_lb->[$lb_sbj_af]}>1)) {
		    my $ct_vretracted = vretract_left_ends_vincluded_gblks ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		    unless (defined $ct_vretracted) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vretract_left_ends_vincluded_gblks' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_left_ends_vincluded_gblks' ended (w/ \$ct_vretracted = $ct_vretracted) ...\n\n") if ($TEST_LEVEL_ssh_sgblk >2);
		}
		if (($cts_vincluding->[$bl_sbj]>0) and (@{$curr_gblks_w_spec_rb->[$rb_sbj_af]}>1)) {
		    my $ct_vretracting = vretract_right_end_sbj_gblk ($bl_sbj, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		    unless (defined $ct_vretracting) {
			print STDERR "[ERROR 'serial_shifts_sgl_gblock'] The subroutine 'vretract_right_end_sbj_gblk' FAILED at the coordinate, (", join (', ', @{$curr_bl_coords}), ", after shifting the $bl_sbj th block in the $sh direction: $! !\n\n";
			return (undef, undef);
		    }
    (print STDOUT "\n .... The subroutine 'vretract_right_end_sbj_gblk' (w/ \$ct_vretracting = $ct_vretracting) ended ...\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

		    if ($ct_vretracting>0) { ($lb_sbj_af, $rb_sbj_af) = @{$curr_bds_blks->[$bl_sbj]}; } # Update the block-boundaries AFTER the "shift".
		}

	    } # End of "if ($sh >0) {...} else {...}".

	    ($lb_sbj_af, $rb_sbj_af) = @{$curr_bds_blks->[$bl_sbj]}; # Update the Block-boundaries AFTER the "shift".

    (print STDOUT "\n", 'After v-retractions: ($lb_sbj_af, $rb_sbj_af) = ', "($lb_sbj_af, $rb_sbj_af) .\n\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

	    my $curr_bl_coord_sbj = $curr_bl_coords->[$bl_sbj];

    (print STDOUT "After the 'shift' (& associated v-extensions and/or v-retractions): \$curr_bl_coord_sbj = $curr_bl_coord_sbj .\n\n\n") if ($TEST_LEVEL_ssh_sgblk >2);

		## END of "(IMPORTED with MODIFICATIONS from 'nonrec_mlt_shifts_gapblocks' (on 2019/07/08).) (1b)" ##

   if ($TEST_LEVEL_ssh_sgblk > 2) {

	print STDOUT "\n[ Outputs of the $time th \"shift\": ]\n\n";

	print STDOUT '$sb_incr_ct_null_clms = ', $sb_incr_ct_null_clms, " ,\n";
	print STDOUT '$incr_ct_null_clms   = ', $incr_ct_null_clms, " .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssh_sgblk > 2) {...}"

	} # End of the main foreach-loop (over the times of the "shift"s). #


   if ($TEST_LEVEL_ssh_sgblk >1) {

	my $bl_coord_fin = $curr_bl_coords->[$bl_sbj];
	my ($lbd_fin, $rbd_fin) = @{$bds_block_sbj}; # Keep the boundaries of the subject block before the serial-"shift"s.

	print STDOUT "\n[ FINAL coordinate & boundaries of the $bl_sbj th gap-block: ]\n\n";

	print STDOUT "\$bl_coord_fin = $bl_coord_fin .\n\n";
	print STDOUT '($lbd_fin, $rbd_fin) = ', "($lbd_fin, $rbd_fin) .\n\n";

   } # End of "if ($TEST_LEVEL_ssh_sgblk >1) {...}"

   if ($TEST_LEVEL_ssh_sgblk >3) {

	print STDOUT "\n[ Gap-block configurations AFTER the $incr_coord_sbj \"shift\"s of the $bl_sbj th gap-block ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'serial_shifts_sgl_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'serial_shifts_sgl_gblock' ...\n\n";
	}

   } # End of "if ($TEST_LEVEL_ssh_sgblk >3) {...}"

   (print STDOUT "\n< Subroutine 'serial_shifts_sgl_gblock' Ended ", 'w/ ($incr_ct_null_clms = ', " $incr_ct_null_clms) />\n\n") if ($TEST_LEVEL_ssh_sgblk > 0);


	return (0, $incr_ct_null_clms);

}	# End of the subroutine, 'serial_shifts_sgl_gblock'. #


# serial_shifts_point2point:
#	This subroutine performs serial "shift"s from a point to another point, both in the coordinate-frame of gap-blocks within a window.
#	What it actually does is to perform the rounds of 'serial_shifts_sgl_gblock' on different gap-blocks whose coordinates will change.
#	[CREATED on 2019/07/09.]
#	[REVISED on 2019/07/11]:
#	Now, the subroutine avoids calling 'serial_shifts_sgl_gblock' for a gap-block if its coordinates are identical at the two points.
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@bl_coords_frm, \@bl_coords_to, \@gb_coords_wd_frm,
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes), where
#
#	@curr_set_clms is the current set of alignment columns, that is,
#	@column = @{$curr_set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@curr_set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$bl_coords_frm[$gb] gives the coordinate of the $gb th gap-block BEFORE all the serial "shift"s;
#	$bl_coords_to[$gb] gives the coordinate of the $gb th gap-block at the DESTINATION point (i.e., AFTER the serial "shift"s);
#	[@bl_coords_frm and @bl_coords_to MUST be within @gb_coords_wd_frm below];
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]}
#	give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@gb_coords_wd_frm is an output of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	@{$curr_gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are currently at the $lb th column;
#	@{$curr_gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are currently at the $rb th column;
#	[Typically, \@curr_gblks_w_spec_lb and \@curr_gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column), with @null_column consisting of as many "absence" states as the MSA rows;
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];

#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[\@cts_vincluding, \@cts_vincluded & \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#
# OUTPUTS: ($flag, $tot_incr_ct_null_clms), where
#
#	$flag = 0 if all the process ended successfully,
#	      = -1 if any of the gap-blocks stepped over the left-end of the local MSA,
#	      = +1 if any of the gap-blocks stepped over the right-end of the local MSA,
#	      = undef if another type of error occurs;
#
#	$tot_incr_ct_null_clms is the total increment of the count of null-columns resulting from the series of "shift"s (= undef if an error occurs).
#
sub serial_shifts_point2point (\@\@\@\@\@\@\@\@$\@\@\@\@\@) {

	my ($curr_set_clms, $curr_bds_blks, $bl_coords_frm, $bl_coords_to, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes) = @_;

	my $ct_clms = @{$curr_set_clms};
	my $ct_gblks = @{$curr_bds_blks};


    (print STDOUT "< Subroutine 'serial_shifts_point2point' (w/ #{columns} = $ct_clms, #{gap-blocks}=$ct_gblks) >\n\n") if ($TEST_LEVEL_ssh_p2p >0);

   if ($TEST_LEVEL_ssh_sgblk >2) {

	print STDOUT "\n[ Relevant Points & Boundaries in the Gap-Block Coordinate Space ]\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'coord_frm', 'coord_to', 'coord_le_wd', 'coord_re_wd'), "\n\n";

	for (my $gb = 0; $gb <$ct_gblks; $gb++) { # Preceding for-loop (over gap-blocks). #

	    print STDOUT join ("\t", $gb, $bl_coords_frm->[$gb], $bl_coords_to->[$gb], @{$gb_coords_wd_frm->[$gb]}), "\n";

	} # End of the preceding for-loop (over gap-blocks).

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssh_sgblk >2) {...}"

			# Examine if the coordinates are within the specified boundaries. #
	my @anom1_gblks = my @anom2_gblks = ();
	for (my $gb = 0; $gb <$ct_gblks; $gb++) { # Testing for-loop (over gap-blocks). #

	    my $coord_frm = $bl_coords_frm->[$gb];
	    my $coord_to  = $bl_coords_to->[$gb];
	    my ($coord_le_wd, $coord_re_wd) = @{$gb_coords_wd_frm->[$gb]};

	    if (($coord_frm < $coord_le_wd) or ($coord_re_wd < $coord_frm)) { push @anom1_gblks, $gb; }
	    if (($coord_to < $coord_le_wd) or ($coord_re_wd < $coord_to)) { push @anom2_gblks, $gb; }

	} # End of the testing for-loop (over gap-blocks).

	if ((@anom1_gblks>0) or (@anom2_gblks>0)) {
	    print STDOUT "[ERROR 'serial_shifts_point2point'] Found Anomalous Gap-Blocks: (", join (',', @anom1_gblks), ') have @{$bl_coords_frm} stick out; (', join (',', @anom2_gblks), ') have @{$bl_coords_to} stick out', ": $! !\n\n";
	    return (undef, undef);
	}
			# End of "Examine if the coordinates are within the specified boundaries." #


	my @curr_bl_coords = @{$bl_coords_frm}; # Initialize the current set of block-coordinates, which will be updated as gap-blocks are "shift"ed. #
	my $tot_incr_ct_null_clms = 0;

   if ($TEST_LEVEL_ssh_sgblk >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @curr_bl_coords, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'serial_shifts_sgl_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'serial_shifts_sgl_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssh_sgblk >3) {...}"


   (print STDOUT "\n ... Now, the Serial \"Shift\"s Begin ...\n\n") if ($TEST_LEVEL_ssh_p2p >1);

	for (my $bl_sbj =0; $bl_sbj < $ct_gblks; $bl_sbj++) { # Main for-loop (over the gap-blocks). #

	    my $sbj_coord_frm = $bl_coords_frm->[$bl_sbj];
	    my $sbj_coord_to = $bl_coords_to->[$bl_sbj];

	    my $incr_coord_sbj = $sbj_coord_to - $sbj_coord_frm;

   (print STDOUT "\n{{ \$bl_sbj = $bl_sbj  (w/ \$sbj_coord_frm = $sbj_coord_frm, \$sbj_coord_to = $sbj_coord_to, \$incr_coord_sbj = $incr_coord_sbj) }}\n\n") if ($TEST_LEVEL_ssh_p2p >1);

	    if ($incr_coord_sbj == 0) { # Added to AVOID an UNNECESSARY CALL of subroutine 'serial_shifts_sgl_gblock' (on 2019/07/11). #

   (print STDOUT "[INFORMATION 'serial_shifts_point2point'] \$incr_coord_sbj is ZERO, thus SKIP the subroutine 'serial_shifts_sgl_gblock' ...\n\n") if ($TEST_LEVEL_ssh_p2p >1);

   (print STDOUT "  {Current \$tot_incr_ct_null_clms} = $tot_incr_ct_null_clms .\n\n") if ($TEST_LEVEL_ssh_p2p >2);

		next;
	    } # END of "if ($incr_coord_sbj == 0) {...}"


	    my ($flag, $incr_ct_null_clms)
	     = serial_shifts_sgl_gblock (
		@{$curr_set_clms}, @{$curr_bds_blks}, @curr_bl_coords, @{$gb_coords_wd_frm}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb},
		$bl_sbj, $incr_coord_sbj, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes});
	    unless ((defined $flag) and ($flag == 0)) {
		my $flag_out = (defined $flag) ? $flag : 'undef';
		print STDOUT "[ERROR 'serial_shifts_point2point'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

		return ($flag, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_ssh_p2p >1);

	    $tot_incr_ct_null_clms += $incr_ct_null_clms;

   (print STDOUT "  {Current \$tot_incr_ct_null_clms} = $tot_incr_ct_null_clms .\n\n") if ($TEST_LEVEL_ssh_p2p >2);


	} # End of the main for-loop (over the gap-blocks). #

   (print STDOUT "\n ... Now, the Whole Serial \"Shift\"s have ENDED ...\n\n") if ($TEST_LEVEL_ssh_p2p >1);

   if ($TEST_LEVEL_ssh_sgblk >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @curr_bl_coords, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'serial_shifts_sgl_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'serial_shifts_sgl_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssh_sgblk >3) {...}"


    (print STDOUT "\n< Subroutine 'serial_shifts_point2point' Ended ", 'w/ ($tot_incr_ct_null_clms = ', " $tot_incr_ct_null_clms) />\n\n") if ($TEST_LEVEL_ssh_p2p >0);

	return (0, $tot_incr_ct_null_clms);

}	# End of the subroutines, 'serial_shifts_point2point'. #


# displace_including_and_onx_to_left:
#	This subroutine performs multiple rounds of 'serial_shifts_sgl_gblock'
#	to "displace" the gap-blocks that include or are in the 'onx' relationships with 
#	a specified "subject" gap-block,
#	that is, "shift" the gap-blocks to the immediate left of the region
#	specified by the left-enf of the subject and a fixed right-limit point.
#	[CREATED on 2019/07/24-25.]
#	[REVISED on 2019/07/26]:
#	ADDED %ignored to the input arguments on 2019/07/26.
#	[DEBUGGED on 2019/07/28]:
#	CORRECTED the positional condition when $gb and $gb_sbj are in the 'onx' relationships.
#
# INPUTS: ($gb_sbj, $right_limit,
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb,
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes,
#	\%ignored # Added \%ignored on 2019/07/26.
#	), where
#
#	$gb_sbj is the specified "subject" gap-block providing the region encompassing the gap-blocks to be "displace"d;
#	$right_limit is the index of the column providing the right-limit (inclusive) of the encompassing region;
#
#	The other arguments are also among the input arguments of 'serial_shifts_sgl_gblock' above;
#	[For details, see the header of 'serial_shifts_sgl_gblock' above.]
#
#	%ignored = ($gb => 1, ...) for the gap-blocks NOT to be displaced; # ADDED on 2019/07/26. #
#
# OUTPUTS: (\%gb2incr_coord_left, $totct_nec_shifts), where
#
#	%gb2incr_coord_left = ($indx_gb => $incr_coord_left, ...)
#	for the gap-blocks "displaced" in this round, with
#	$incr_coord_left being the increment of the coordinate of the $indx_gb th gap-block resulting from the serial-"shift"s;
#	(\%gb2incr_coord_left is undef if the subroutine failed);
#
#	$totct_nec_shifts is the total number of necessary "shift"s;
#
# [ sub displace_including_and_onx_to_left ($$\@\@\@\@\@\@\@$\@\@\@\@\@) { ] ... OBSOLETE as of 2019/07/26. #
#
sub displace_including_and_onx_to_left ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%) {

	my ($gb_sbj, $right_limit,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gb_coords_wd_frm, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	$rows_affd_by_gblks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes,
	$ignored # Added $ignored in REVISION on 2019/07/26.
	) = @_;

	my $ct_gblks = @{$curr_bds_blks};


   (print STDOUT "< Subroutine 'displace_including_and_onx_to_left' (#{gap-blocks}=$ct_gblks; \$gb_sbj = \$gb_sbj, \$right_limit = \$right_limit) >\n\n") if ($TEST_LEVEL_diotl>0);

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my %gb2incr_coord_left;
	my $totct_nec_shifts = 0;

	for (my $gb=0; $gb<$ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    if ($gb == $gb_sbj) { next; }
	    if (defined $ignored->{$gb}) { next; } # ADDED in REVISION on 2019/07/26.

	    my ($le_sbj, $re_sbj) = @{$curr_bds_blks->[$gb_sbj]}; # The current boundaries of the subject.

	    my ($le_gb, $re_gb) = @{$curr_bds_blks->[$gb]};

	    unless (($le_sbj <= $le_gb) and ($re_gb <= $right_limit)) { next; } # $gb is NOT within the specified region. #

	    my $rel = $rels_w_sbj->[$gb];
	    unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb < $gb_sbj)) or ($rel eq 'ONN') or ($rel eq 'ONCS')) { next; } # $gb does NEITHER include, NOR is in the 'onx' relationship with, $gb_sbj. # ADDED "(($rel eq '=') and ($gb < $gb_sbj))" on 2019/07/27. #

			# The gap-blocks "survived" up to here WILL BE DISPLACED to the left. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($le_sbj-1, $gb, @{$curr_bds_blks}, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $le_sbj-1;
		print STDOUT "[ERROR 'displace_including_and_onx_to_left'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $gb th gap-block: $! !\n\n";
		return (undef, undef);
	    }

	    my $incr_coord_left = - $ct_nec_shifts;
	    $totct_nec_shifts += $ct_nec_shifts;

	    $gb2incr_coord_left{$gb} = $incr_coord_left;


	    my ($flag, $incr_ct_null_clms)
	     = ($ct_nec_shifts == 0) ?  (0, 0) :
		serial_shifts_sgl_gblock (
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$gb_coords_wd_frm}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb},
		$gb, $incr_coord_left, @{$rows_affd_by_gblks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
	    	);
	    unless ((defined $flag) and ($flag == 0)) {
		my $flag_out = (defined $flag) ? $flag : 'undef';
		print STDOUT "[ERROR 'displace_including_and_onx_to_left'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $gb th gap-block (w/ \$flag = $flag_out): $! !\n\n";

		return (undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $gb th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_diotl >1);


	    if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb_sbj][0] < $curr_bds_blks->[$gb][0]) ) { # A BUG fixed on 2019/07/28.
	    ##if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb_sbj][0] > $curr_bds_blks->[$gb][0]) ) {

			# => swap the positions of $gb and $gb_sbj, if not yet. 

		my $ct_vextensions = vextend_left_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		unless (defined $ct_vextensions) {
		    print STDOUT "[ERROR 'displace_including_and_onx_to_left'] The subroutine, 'vextend_left_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vextend_left_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_diotl >1);

		my $ct_vretracting = vretract_left_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_lb});
		unless (defined $ct_vretracting) {
		    print STDOUT "[ERROR 'displace_including_and_onx_to_left'] The subroutine, 'vretract_left_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vretract_left_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_diotl >1);

	    } # End of "if ($ct_nec_shifts == 0) {...}".

	} # End of the main for-loop (over gap-blocks). #

   (print STDOUT "< Subroutine 'displace_including_and_onx_to_left' Ended (w/ \$totct_nec_shifts = $totct_nec_shifts) />\n\n") if ($TEST_LEVEL_diotl>0);

	return (\%gb2incr_coord_left, $totct_nec_shifts);

}	# End of the subroutine, 'displace_including_and_onx_to_left'. #


# displace_including_and_onx_to_right:
#	This subroutine does almost the same things with the predecessor, 'displace_including_and_onx_to_left',
#	EXCEPT that it "displaces" the gap-blocks TO THE RIGHT of the region
#	specified by a fixed left-limit point and the right-end of the subject gap-block.
#	It was created by modifying 'displace_including_and_onx_to_left' appropriately.
#	[CREATED on 2019/07/25.]
#	[REVISED on 2019/07/26]:
#	Now, the gap-blocks are examined from the lowest-ranked to the highest-ranked,
#	in order to keep the horozontal order among vertically identical gap-blocks.
#
#	[REVISED on 2019/07/26]:
#	ADDED %ignored to the input arguments.
#	[DEBUGGED on 2019/07/28]:
#	CORRECTED the positional condition when $gb and $gb_sbj are in the 'onx' relationships.
#
# displace_including_and_onx_to_left:
#	This subroutine performs multiple rounds of 'serial_shifts_sgl_gblock'
#	to "displace" the gap-blocks that include or are in the 'onx' relationships with 
#	a specified "subject" gap-block,
#	that is, "shift" the gap-blocks to the immediate left of the region
#	specified by the left-enf of the subject and a fixed right-limit point.
#	[CREATED on 2019/07/24-25.]
#
# INPUTS: ($gb_sbj, $left_limit,
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb,
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes,
#	\%ignored # Added on 2019/07/26.
#	), where
#
#	$gb_sbj is the specified "subject" gap-block providing the region encompassing the gap-blocks to be "displace"d;
#	$left_limit is the index of the column providing the left-limit (inclusive) of the encompassing region;
#
#	The other arguments are also among the input arguments of 'serial_shifts_sgl_gblock' above;
#	[For details, see the header of 'serial_shifts_sgl_gblock' above.]
#
#	%ignored = ($gb => 1, ...) for the gap-blocks NOT to be displaced; # ADDED on 2019/07/26. #
#
# OUTPUTS: (\%gb2incr_coord_right, $totct_nec_shifts), where
#
#	%gb2incr_coord_right = ($indx_gb => $incr_coord_right, ...)
#	for the gap-blocks "displaced" in this round, with
#	$incr_coord_left being the increment of the coordinate of the $indx_gb th gap-block resulting from the serial-"shift"s;
#	(\%gb2incr_coord_right is undef if the subroutine failed);
#
#	$totct_nec_shifts is the total number of necessary "shift"s;
#
# [ sub displace_including_and_onx_to_right ($$\@\@\@\@\@\@\@$\@\@\@\@\@) { ] ... OBSOLETE as of 2019/07/26.
#
sub displace_including_and_onx_to_right ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%) {

	my ($gb_sbj, $left_limit,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gb_coords_wd_frm, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	$rows_affd_by_gblks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes,
	$ignored # Added $ignored in REVISION on 2019/07/26.
	) = @_;

	my $ct_gblks = @{$curr_bds_blks};


   (print STDOUT "< Subroutine 'displace_including_and_onx_to_right' (#{gap-blocks}=$ct_gblks; \$gb_sbj = \$gb_sbj, \$left_limit = \$left_limit) >\n\n") if ($TEST_LEVEL_diotr>0);

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my %gb2incr_coord_right;
	my $totct_nec_shifts = 0;

	for (my $gb= $ct_gblks-1; $gb >=0; $gb--) { # Main for-loop (over gap-blocks). # REVISED on 2019/07/26.
	##for (my $gb=0; $gb<$ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    if ($gb == $gb_sbj) { next; }
	    if (defined $ignored->{$gb}) { next; } # ADDED in REVISION on 2019/07/26. #

	    my ($le_sbj, $re_sbj) = @{$curr_bds_blks->[$gb_sbj]}; # The current boundaries of the subject.

	    my ($le_gb, $re_gb) = @{$curr_bds_blks->[$gb]};

	    unless (($left_limit <= $le_gb) and ($re_gb <= $re_sbj)) { next; } # $gb is NOT within the specified region. #

	    my $rel = $rels_w_sbj->[$gb];
	    unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb < $gb_sbj)) or ($rel eq 'ONN') or ($rel eq 'ONCS')) { next; } # $gb does NEITHER include, NOR is in the 'onx' relationship with, $gb_sbj. # ADDED "(($rel eq '=') and ($gb < $gb_sbj))" on 2019/07/27.

			# The gap-blocks "survived" up to here WILL BE DISPLACED to the right. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($re_sbj+1, $gb, @{$curr_bds_blks}, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $re_sbj+1;
		print STDOUT "[ERROR 'displace_including_and_onx_to_right'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $gb th gap-block: $! !\n\n";
		return (undef, undef);
	    }

	    my $incr_coord_right = $ct_nec_shifts;
	    $totct_nec_shifts += $ct_nec_shifts;

	    $gb2incr_coord_right{$gb} = $incr_coord_right;


	    my ($flag, $incr_ct_null_clms)
	     = ($ct_nec_shifts == 0) ?  (0, 0) :
		serial_shifts_sgl_gblock (
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$gb_coords_wd_frm}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb},
		$gb, $incr_coord_right, @{$rows_affd_by_gblks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
	    	);
	    unless ((defined $flag) and ($flag == 0)) {
		my $flag_out = (defined $flag) ? $flag : 'undef';
		print STDOUT "[ERROR 'displace_including_and_onx_to_right'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $gb th gap-block (w/ \$flag = $flag_out): $! !\n\n";

		return (undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $gb th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_diotr >1);


	    if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb][1] < $curr_bds_blks->[$gb_sbj][1]) ) { # DEBUGGED on 2019/07/28.
	    ##if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb][1] > $curr_bds_blks->[$gb_sbj][1]) ) {

			# => swap the positions of $gb and $gb_sbj, if not yet. 

		my $ct_vextensions = vextend_right_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		unless (defined $ct_vextensions) {
		    print STDOUT "[ERROR 'displace_including_and_onx_to_right'] The subroutine, 'vextend_right_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vextend_right_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_diotr >1);

		my $ct_vretracting = vretract_right_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		unless (defined $ct_vretracting) {
		    print STDOUT "[ERROR 'displace_including_and_onx_to_right'] The subroutine, 'vretract_right_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vretract_right_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_diotr >1);

	    } # End of "if ($ct_nec_shifts == 0) {...}".

	} # End of the main for-loop (over gap-blocks). #

   (print STDOUT "< Subroutine 'displace_including_and_onx_to_right' Ended (w/ \$totct_nec_shifts = $totct_nec_shifts) />\n\n") if ($TEST_LEVEL_diotr>0);

	return (\%gb2incr_coord_right, $totct_nec_shifts);

}	# End of the subroutine, 'displace_including_and_onx_to_right'. #


# displace_vincluding_veq_and_onx_to_right_end:
#	This subroutine performs multiple rounds of 'serial_shifts_sgl_gblock'
#	to "displace" the gap-blocks, that is, effectively "shift" them, to the ritht-end of the window we are dealing with,
#	if they are on the right of a specified left-boundary,
#	and if they are vertically including, equivalent, or in the 'onx' relationships with 
#	a specified "subject" gap-block,
#	[CREATED on 2019/08/03.]
#	[REVISED on 2019/08/04]:
#	Now, the subroutine imposes the condition on the RIGHT-BOUNDs of the gap-blocks,
#	instead of on their left-bounds as before the revision.
#
# displace_including_and_onx_to_right:
#	This subroutine does almost the same things with the predecessor, 'displace_including_and_onx_to_left',
#	EXCEPT that it "displaces" the gap-blocks TO THE RIGHT of the region
#	specified by a fixed left-limit point and the right-end of the subject gap-block.
#	It was created by modifying 'displace_including_and_onx_to_left' appropriately.
#	[CREATED on 2019/07/25.]
#	[REVISED on 2019/07/26]:
#	Now, the gap-blocks are examined from the lowest-ranked to the highest-ranked,
#	in order to keep the horozontal order among vertically identical gap-blocks.
#
#	[REVISED on 2019/07/26]:
#	ADDED %ignored to the input arguments.
#	[DEBUGGED on 2019/07/28]:
#	CORRECTED the positional condition when $gb and $gb_sbj are in the 'onx' relationships.
#
# displace_including_and_onx_to_left:
#	This subroutine performs multiple rounds of 'serial_shifts_sgl_gblock'
#	to "displace" the gap-blocks that include or are in the 'onx' relationships with 
#	a specified "subject" gap-block,
#	that is, "shift" the gap-blocks to the immediate left of the region
#	specified by the left-enf of the subject and a fixed right-limit point.
#	[CREATED on 2019/07/24-25.]
#
# INPUTS: ($gb_sbj, $left_bound,
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb,
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes,
#	\%ignored # Added on 2019/07/26.
#	), where
#
#	$gb_sbj is the specified "subject" gap-block providing the region encompassing the gap-blocks to be "displace"d;
#	$left_bound is the index of the column providing the left-boundary,
#	more precisely, the subroutine deals ONLY with the gap-blocks on the right of $left_bound (inclusive);
#
#	The other arguments are also among the input arguments of 'serial_shifts_sgl_gblock' above;
#	[For details, see the header of 'serial_shifts_sgl_gblock' above.]
#
#	%ignored = ($gb => 1, ...) for the gap-blocks NOT to be displaced; # ADDED on 2019/07/26. #
#
# OUTPUTS: (\%gb2incr_coord, $totct_nec_shifts, $le_rlv_gbs), where
#
#	%gb2incr_coord = ($indx_gb => $incr_coord, ...)
#	for the gap-blocks that are dealt with by this subroutine, with
#	$incr_coord being the increment of the coordinate of the $indx_gb th gap-block resulting from the serial-"shift"s;
#	(\%gb2incr_coord is undef if the subroutine failed);
#
#	$totct_nec_shifts is the total number of necessary "shift"s;
#
#	$le_rlv_gbs is the leftmost-end of the relevant gap-blocks;
#
sub displace_vincluding_veq_and_onx_to_right_end ($$\@\@\@\@\@\@\@$\@\@\@\@\@\%) {

	my ($gb_sbj, $left_bound,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $gb_coords_wd_frm, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb,
	$rows_affd_by_gblks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes,
	$ignored
	) = @_; # Added $ignored in REVISION on 2019/07/26.

	my $ct_gblks = @{$curr_bds_blks};


   (print STDOUT "< Subroutine 'displace_vincluding_veq_and_onx_to_right_end' (#{gap-blocks}=$ct_gblks; \$gb_sbj = \$gb_sbj, \$left_bound = \$left_bound) >\n\n") if ($TEST_LEVEL_dvieotre>0);

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my %gb2incr_coord;
	my $totct_nec_shifts = 0;

	my $le_rlv_gbs = scalar (@{$curr_set_clms}) - 1;

	for (my $gb= $ct_gblks-1; $gb >=0; $gb--) { # Main for-loop (over gap-blocks). # REVISED on 2019/07/26.
	    if ($gb == $gb_sbj) { next; }
	    if (defined $ignored->{$gb}) { next; }

=For the predecessor 'displace_including_and_onx_to_right'. (a)
	    my ($le_sbj, $re_sbj) = @{$curr_bds_blks->[$gb_sbj]}; # The current boundaries of the subject.
=cut # End of "For the predecessor 'displace_including_and_onx_to_right'. (a)"

	    my ($le_gb, $re_gb) = @{$curr_bds_blks->[$gb]};

	    unless ($left_bound <= $re_gb) { next; } # $gb does NOT AT ALL occupy the right-side of $left_bound (inclusive). (Modified in REVISION of 'displace_including_and_onx_to_right' (on 2019/08/04).)
	    ##unless ($left_bound <= $le_gb) { next; } # $gb is NOT on the right of $left_bound. (Modified in creation of 'displace_including_and_onx_to_right'. (1))

	    my $rel = $rels_w_sbj->[$gb];

	    unless (($rel eq '<') or ($rel eq '<(pa)') or ($rel eq '=') or ($rel eq 'ONN') or ($rel eq 'ONCS')) { next; } # $gb does NEITHER include, NOR is in the 'onx' relationship with, $gb_sbj. # CHANGED "(($rel eq '=') and ($gb < $gb_sbj))" to "($rel eq '=')" in creation of 'displace_vincluding_veq_and_onx_to_right_end'.

=For the predecessor 'displace_including_and_onx_to_right'. (b)
	    unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb < $gb_sbj)) or ($rel eq 'ONN') or ($rel eq 'ONCS')) { next; } # $gb does NEITHER include, NOR is in the 'onx' relationship with, $gb_sbj. # ADDED "(($rel eq '=') and ($gb < $gb_sbj))" on 2019/07/27.
=cut # End of "For the predecessor 'displace_including_and_onx_to_right'. (b)"


			# The gap-blocks "survived" up to here WILL BE DISPLACED to the right-end. #

			# MODIFIED in creation of 'displace_vincluding_veq_and_onx_to_right_end'. (2) #

	    my $incr_coord = $gb_coords_wd_frm->[$gb][1] - $curr_bl_coords->[$gb];

	    my $ct_nec_shifts = ($incr_coord>0) ? $incr_coord : - $incr_coord;

=For the predecessor 'displace_including_and_onx_to_right'. (c)
	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($re_sbj+1, $gb, @{$curr_bds_blks}, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $re_sbj+1;
		print STDOUT "[ERROR 'displace_vincluding_veq_and_onx_to_right_end'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $gb th gap-block: $! !\n\n";
		return (undef, undef);
	    }

	    my $incr_coord_right = $ct_nec_shifts;
=cut # End of "For the predecessor 'displace_including_and_onx_to_right'. (c)"

			# End of "MODIFIED in creation of 'displace_vincluding_veq_and_onx_to_right_end'. (2)" #

	    $totct_nec_shifts += $ct_nec_shifts;

	    $gb2incr_coord{$gb} = $incr_coord;
	    ##$gb2incr_coord_right{$gb} = $incr_coord_right;


	    my ($flag, $incr_ct_null_clms)
	     = ($ct_nec_shifts == 0) ?  (0, 0) :
		serial_shifts_sgl_gblock (
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$gb_coords_wd_frm}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb},
		$gb, $incr_coord, @{$rows_affd_by_gblks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
	    	);
	    unless ((defined $flag) and ($flag == 0)) {
		my $flag_out = (defined $flag) ? $flag : 'undef';
		print STDOUT "[ERROR 'displace_vincluding_veq_and_onx_to_right_end'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $gb th gap-block (w/ \$flag = $flag_out): $! !\n\n";

		return (undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $gb th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_dvieotre >1);


=For the predecessor 'displace_including_and_onx_to_right'. (d)
	    if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb][1] < $curr_bds_blks->[$gb_sbj][1]) ) { # DEBUGGED on 2019/07/28.
	    ##if ( (($rel eq 'ONN') or ($rel eq 'ONCS')) and ($curr_bds_blks->[$gb][1] > $curr_bds_blks->[$gb_sbj][1]) ) {

			# => swap the positions of $gb and $gb_sbj, if not yet. 

		my $ct_vextensions = vextend_right_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$curr_set_clms}, @{$inter_block_relations}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb});
		unless (defined $ct_vextensions) {
		    print STDOUT "[ERROR 'displace_vincluding_veq_and_onx_to_right_end'] The subroutine, 'vextend_right_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vextend_right_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_dvieotre >1);

		my $ct_vretracting = vretract_right_end_sbj_gblk ($gb, @{$curr_bds_blks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$curr_gblks_w_spec_rb});
		unless (defined $ct_vretracting) {
		    print STDOUT "[ERROR 'displace_vincluding_veq_and_onx_to_right_end'] The subroutine, 'vretract_right_end_sbj_gblk', FAILED on the $gb th gap-block: $! !\n\n";
		    return (undef, undef);
		}
   (print STDOUT "\n ... The subroutine, 'vretract_right_end_sbj_gblk', Ended on the $gb th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_dvieotre >1);

	    } # End of "if ($ct_nec_shifts == 0) {...}".
=cut # End of "For the predecessor 'displace_including_and_onx_to_right'. (d)"

	    my $le_gb_af = $curr_bds_blks->[$gb][0];
	    if ($le_gb_af < $le_rlv_gbs) { $le_rlv_gbs = $le_gb_af; }

	} # End of the main for-loop (over gap-blocks). #


   (print STDOUT "< Subroutine 'displace_vincluding_veq_and_onx_to_right_end' Ended (w/ \$totct_nec_shifts = $totct_nec_shifts, \$le_rlv_gbs = $le_rlv_gbs) />\n\n") if ($TEST_LEVEL_dvieotre>0);

	return (\%gb2incr_coord, $totct_nec_shifts, $le_rlv_gbs);

}	# End of the subroutine, 'displace_vincluding_veq_and_onx_to_right_end'. #


# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# [ INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, \@rows_affd_by_gblks, $thrsh_dist_purge, $thrsh_dist_cmerge), where ] ... SET ASIDE for later use.
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_purge, $thrsh_dist_cmerge), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_purge is the threshold for the distance between gap-blocks to be "purge"d;
#	$thrsh_dist_cmerge is the threshold for the distance between gap-blocks (of complementary types) to be "merge"d;
#
#
# OUTPUTS: (\@to_be_purged, \@to_be_cmerged), where
#
#	@{$to_be_purged[$pg]} = ($gb1, $gb2, $dist) provides information on the $pg th gap-block pair to be "purge"d, where
#	$gb1 and $gb2 are the indices of the gap-blocks,
#	$dist is the distance of $gb2 from $gb1;
#
#	@{$to_be_cmerged[$cmg]} = ($gb1, $gb2, $dist) provides information on the $cmg th complementary gap-block pair to be "merge"d,
#	which follows the same format as @{$to_be_purged[$pg]} above.
#
# [ sub list_nn_purge_and_cmerge_candidates (\@\@\@\@$$) {
#	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $rows_affd_by_gblks, $thrsh_dist_purge, $thrsh_dist_cmerge) = @_; ] ... SET ASIDE for later use.
#
sub list_nn_purge_and_cmerge_candidates (\@\@\@$$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_purge, $thrsh_dist_cmerge) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' complementary pairs for each $gb1. ###

	my %gb2_2pg_cands;
	my %gb2_2cmg_cands;

	for (my $gb1 =0; $gb1 < $ct_gblks-1; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

	    my @pg_cands = ();
	    my $mindist_pgc;
	    my @cmg_cands = ();
	    my $mindist_cmgc;

	    for (my $gb2 = $gb1+1; $gb2 < $ct_gblks; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq 'Cp') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_purge_and_cmerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef, undef);
		}

		my $dist = $dist2;

		if ($size_gb1 == $size_gb2) {

		    if ($dist > $thrsh_dist_purge) { next; }

		    if (!(defined $mindist_pgc) or ($dist < $mindist_pgc)) {
			@pg_cands = ([$gb1, $gb2, $dist]);
			$mindist_pgc = $dist;

		    } elsif ($dist == $mindist_pgc) {
			push @pg_cands, [$gb1, $gb2, $dist];
		    }

		} else {

		    if ($dist > $thrsh_dist_cmerge) { next; }

		    if (!(defined $mindist_cmgc) or ($dist < $mindist_cmgc)) {
			@cmg_cands = ([$gb1, $gb2, $dist]);
			$mindist_cmgc = $dist;

		    } elsif ($dist == $mindist_cmgc) {
			push @cmg_cands, [$gb1, $gb2, $dist];
		    }

		} # End of "if ($size_gb1 == $size_gb2) {...} else {...}" #

	    } # End of the inner for-loop (over $gb2). #

	    foreach my $pair (@pg_cands) {
		my $gb2 = $pair->[1];
		my $pg_cands = $gb2_2pg_cands{$gb2};
		unless (defined $pg_cands) { $pg_cands = $gb2_2pg_cands{$gb2} = []; }
		push @{$pg_cands}, $pair;
	    }

	    foreach my $pair (@cmg_cands) {
		my $gb2 = $pair->[1];
		my $cmg_cands = $gb2_2cmg_cands{$gb2};
		unless (defined $cmg_cands) {$cmg_cands = $gb2_2cmg_cands{$gb2} = []; }
		push @{$cmg_cands}, $pair;
	    }

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

	my @to_be_purged = ();
	foreach my $gb2 (keys %gb2_2pg_cands) { # 1st trailing foreach-loop (over $gb2). #

	    my $pg_cands = $gb2_2pg_cands{$gb2};
	    my $ct_pg_cands = @{$pg_cands};
	    if ($ct_pg_cands < 2) {
		push @to_be_purged, @{$pg_cands};
		next;
	    }
	    my $mindist = $pg_cands->[0][2];
	    for (my $i=1; $i <$ct_pg_cands; $i++) {
		my $dist = $pg_cands->[$i][2];
		if ($dist < $mindist) { $mindist = $dist; }
	    }
	    foreach my $pg_cand (@{$pg_cands}) {
		if ($pg_cand->[2] == $mindist) { push @to_be_purged, $pg_cand; }
	    }

	} # End of the 1st trailing foreach-loop (over $gb2). #

	my @to_be_cmerged = ();
	foreach my $gb2 (keys %gb2_2cmg_cands) { # 2nd trailing foreach-loop (over $gb2). #

	    my $cmg_cands = $gb2_2cmg_cands{$gb2};
	    my $ct_cmg_cands = @{$cmg_cands};
	    if ($ct_cmg_cands < 2) {
		push @to_be_cmerged, @{$cmg_cands};
		next;
	    }
	    my $mindist = $cmg_cands->[0][2];
	    for (my $i=1; $i <$ct_cmg_cands; $i++) {
		my $dist = $cmg_cands->[$i][2];
		if ($dist < $mindist) { $mindist = $dist; }
	    }
	    foreach my $cmg_cand (@{$cmg_cands}) { 
		if ($cmg_cand->[2] == $mindist) { push @to_be_cmerged, $cmg_cand; }
	    }

	} # End of the 2nd trailing foreach-loop (over $gb2). #


		### (3) Finally, sort the candidates in each of the sets. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

	my $std_to_be_purged = multiple_sort (@to_be_purged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_purged) {
	    print STDOUT "[ERROR 'list_nn_purge_and_cmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of gap-block pairs to be \"purge\"d: $! !\n\n";
	    return (undef, undef);
	}

	my $std_to_be_cmerged = multiple_sort (@to_be_cmerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_cmerged) {
	    print STDOUT "[ERROR 'list_nn_purge_and_cmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of complementary gap-block pairs to be \"merge\"d: $! !\n\n";
	    return (undef, undef);
	}
	
	return ($std_to_be_purged, $std_to_be_cmerged);

}	# End of the subroutine, 'list_nn_purge_and_cmerge_candidates'. #


# list_nn_smerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of the same type to be "merge"d.
#	[CREATED on 2019/07/21.]
#
# [ INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_smerge), where ] ... Set Aside for Later Use.
# INPUTS: (\@bds_gblocks0, \@inter_block_relations, $thrsh_dist_smerge), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	[[
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#	]] ... Set Aside for Later Use.
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	$thrsh_dist_smerge is the threshold for the distance between gap-blocks (of the same type) to be "merge"d;
#
#
# OUTPUTS: (\@to_be_smerged), where
#
#	@{$to_be_smerged[$smg]} = ($gb1, $gb2, $dist) provides information on the $smg th same-type gap-block pair to be "merge"d,
#	which follows the same format as @{$to_be_purged[$pg]} above.
#
sub list_nn_smerge_candidates (\@\@$) {

	my ($bds_gblocks0, $inter_block_relations, $thrsh_dist_smerge) = @_;

	my $ct_gblks = @{$inter_block_relations};

	my @to_be_smerged = ();

	for (my $gb2 = 1; $gb2 < $ct_gblks; $gb2++) { # Main for-loop (over the nearest-neighboring pairs of gap-blocks). #

	    my $gb1 = $gb2-1;

	    my $rel = $inter_block_relations->[$gb1][$gb2];
	    unless ((defined $rel) and ($rel eq '=')) { next; }

	    my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
	    unless (defined $dist2) {
		print STDOUT "[ERROR 'list_nn_smerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		return (undef);
	    }

	    my $dist = $dist2;

	    if ($dist <= $thrsh_dist_smerge) { push @to_be_smerged, [$gb1, $gb2, $dist]; }

	} # End of the main for-loop (over the nearest-neighboring pairs of gap-blocks). #

	return \@to_be_smerged;

}	# End of the subroutine, 'list_nn_smerge_candidates'. #


# list_nn_cvmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	complementary-sibling gap-blocks to be "(incomplete) vertical-merge"d.
#	It was created by taking advantage of the framework of 'list_nn_purge_and_cmerge_candidates'.
#	[CREATED on 2019/07/23.]
#	[REVISED on 2019/07/27]:
#	Now, when identifying "nearest-neighboring" pairs,
#	this subroutine only compares gap-blocks that are vertically identical to one another.
#
# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_cvmerge), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_cvmerge is the threshold for the distance between complementary-sibling gap-blocks to be "(incomplete) vertical-merge"d;
#
#
# OUTPUTS: (\@to_be_cvmerged), where
#
#	@{$to_be_cvmerged[$cvm]} = ($gb1, $gb2, $dist) provides information on the $cmv th complementary-sibling gap-block pair to be "(incomplete) vertical-merge"d,
#	$gb1 and $gb2 are the indices of the gap-blocks,
#	$dist is the distance of $gb2 from $gb1;
#
sub list_nn_cvmerge_candidates (\@\@\@$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_cvmerge) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' complementary pairs for each $gb1. ###

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (1) #
	my %gb2_2cvm_cands;
	##my %gb2_2pg_cands;
	##my %gb2_2cmg_cands;

	for (my $gb1 =0; $gb1 < $ct_gblks-1; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (2) #
	    my @cvm_cands = ();
	    my $mindist_cvmc;
	    ##my @pg_cands = ();
	    ##my $mindist_pgc;
	    ##my @cmg_cands = ();
	    ##my $mindist_cmgc;

	    for (my $gb2 = $gb1+1; $gb2 < $ct_gblks; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq 'ONCS') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_cvmerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef);
		}

		my $dist = $dist2;

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (3) #

		if (1) {
		    if ($dist > $thrsh_dist_cvmerge) { next; }


				# MODIFIED in revision (on 2019/07/27). (1) #

		    my $rels_w_gb2 = $inter_block_relations->[$gb2];

		    my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		    foreach my $pair (@cvm_cands) {
			my ($gb1c, $gb3, $dist3) = @{$pair};
			unless ($rels_w_gb2->[$gb3] eq '=') { next; }
			if ($dist < $dist3) {
			   $result_cmp = -1;
			} elsif ($dist > $dist3) {
			   $result_cmp = +1;
			} 
			last;
		    }

		    if ($result_cmp >0) {

			next; # Exclude the current pair. #

		    } elsif ($result_cmp < 0) {

			    # Remove the pairs of the same type. # 

			my @new_cvm_cands = ();
			foreach my $pair (@cvm_cands) {
			    my ($gb1c, $gb3, $dist3) = @{$pair};
			    if ($rels_w_gb2->[$gb3] eq '=') { next; }
			    push @new_cvm_cands, $pair;
			}
			@cvm_cands = @new_cvm_cands;
		    }

		    push @cvm_cands, [$gb1, $gb2, $dist];

=OBSOLETE as of 2019/07/27. (1)
		    if (!(defined $mindist_cvmc) or ($dist < $mindist_cvmc)) {
			@cvm_cands = ([$gb1, $gb2, $dist]);
			$mindist_cvmc = $dist;

		    } elsif ($dist == $mindist_cvmc) {
			push @cvm_cands, [$gb1, $gb2, $dist];
		    }
=cut # END of "OBSOLETE as of 2019/07/27. (1)"
				# End of "MODIFIED in revision (on 2019/07/27). (1)" #

		}

=For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (1)
		if ($size_gb1 == $size_gb2) {

		    if ($dist > $thrsh_dist_purge) { next; }

		    if (!(defined $mindist_pgc) or ($dist < $mindist_pgc)) {
			@pg_cands = ([$gb1, $gb2, $dist]);
			$mindist_pgc = $dist;

		    } elsif ($dist == $mindist_pgc) {
			push @pg_cands, [$gb1, $gb2, $dist];
		    }

		} else {

		    if ($dist > $thrsh_dist_cmerge) { next; }

		    if (!(defined $mindist_cmgc) or ($dist < $mindist_cmgc)) {
			@cmg_cands = ([$gb1, $gb2, $dist]);
			$mindist_cmgc = $dist;

		    } elsif ($dist == $mindist_cmgc) {
			push @cmg_cands, [$gb1, $gb2, $dist];
		    }

		} # End of "if ($size_gb1 == $size_gb2) {...} else {...}" #
=cut # End of "For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (1)"


	    } # End of the inner for-loop (over $gb2). #


		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (5) #

	    foreach my $pair (@cvm_cands) {
		my $gb2 = $pair->[1];
		my $cvm_cands = $gb2_2cvm_cands{$gb2};
		unless (defined $cvm_cands) {$cvm_cands = $gb2_2cvm_cands{$gb2} = []; }
		push @{$cvm_cands}, $pair;
	    }

=For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (2)
	    foreach my $pair (@pg_cands) {
		my $gb2 = $pair->[1];
		my $pg_cands = $gb2_2pg_cands{$gb2};
		unless (defined $pg_cands) { $pg_cands = $gb2_2pg_cands{$gb2} = []; }
		push @{$pg_cands}, $pair;
	    }

	    foreach my $pair (@cmg_cands) {
		my $gb2 = $pair->[1];
		my $cmg_cands = $gb2_2cmg_cands{$gb2};
		unless (defined $cmg_cands) {$cmg_cands = $gb2_2cmg_cands{$gb2} = []; }
		push @{$cmg_cands}, $pair;
	    }
=cut # End of "For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (2)"

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (6) #

	my @to_be_cvmerged = ();
	foreach my $gb2 (keys %gb2_2cvm_cands) { # Trailing foreach-loop (over $gb2). #

	    my $cvm_cands = $gb2_2cvm_cands{$gb2};
	    my $ct_cvm_cands = @{$cvm_cands};
	    if ($ct_cvm_cands < 2) {
		push @to_be_cvmerged, @{$cvm_cands};
		next;
	    }

			# MODIFIED in revision (on 2019/07/27). (2) #

	    my @nn_cvm_cands = ($cvm_cands->[0]);

	    for (my $i=1; $i <$ct_cvm_cands; $i++) {
		my ($gb1, $gb2c, $dist) = @{$cvm_cands->[$i]};

		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		foreach my $pair (@nn_cvm_cands) {
		    my ($gb1c, $gb2c, $distc) = @{$pair}; 
		    unless ($rels_w_gb1->[$gb1c] eq '=') { next; }
		    if ($dist < $distc) {
			$result_cmp = -1;
		    } elsif ($dist > $distc) {
			$result_cmp = +1;
		    }
		    last;
		} # End of "foreach my $pair (@nn_cvm_cands) {...}"

		if ($result_cmp > 0) {

		    next; # Exclude the current pair. #

		} elsif ($result_cmp < 0) {
			# Remove the pairs of the same type. #
		    my @new_nn_cvm_cands = ();
		    foreach my $pair (@nn_cvm_cands) {
			my ($gb1c, $gb2cc, $distc) = @{$pair}; 
			if ($rels_w_gb1->[$gb1c] eq '=') { next; }
			push @new_nn_cvm_cands, $pair;
		    }
		    @nn_cvm_cands = @new_nn_cvm_cands;
		}

		push @nn_cvm_cands, $cvm_cands->[$i];

	    } # End of "for (my $i=1; $i <$ct_cvm_cands; $i++) {...}"

	    push @to_be_cvmerged, @nn_cvm_cands;

=OBSOLETE as of 2019/07/27. (2)
	    my $mindist = $cvm_cands->[0][2];
	    for (my $i=1; $i <$ct_cvm_cands; $i++) {
		my $dist = $cvm_cands->[$i][2];
		if ($dist < $mindist) { $mindist = $dist; }
	    }
	    foreach my $cvm_cand (@{$cvm_cands}) { 
		if ($cvm_cand->[2] == $mindist) { push @to_be_cvmerged, $cvm_cand; }
	    }
=cut # END of "OBSOLETE as of 2019/07/27. (2)"
			# End of "MODIFIED in revision (on 2019/07/27). (2)" #

	} # End of the trailing foreach-loop (over $gb2). #

=For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (3)
	my @to_be_purged = ();
	foreach my $gb2 (keys %gb2_2pg_cands) { # 1st trailing foreach-loop (over $gb2). #

	    my $pg_cands = $gb2_2pg_cands{$gb2};
	    my $ct_pg_cands = @{$pg_cands};
	    if ($ct_pg_cands < 2) {
		push @to_be_purged, @{$pg_cands};
		next;
	    }
	    my $mindist = $pg_cands->[0][2];
	    for (my $i=1; $i <$ct_pg_cands; $i++) {
		my $dist = $pg_cands->[$i][2];
		if ($dist < $mindist) { $mindist = $dist; }
	    }
	    foreach my $pg_cand (@{$pg_cands}) {
		if ($pg_cand->[2] == $mindist) { push @to_be_purged, $pg_cand; }
	    }

	} # End of the 1st trailing foreach-loop (over $gb2). #

	my @to_be_cmerged = ();
	foreach my $gb2 (keys %gb2_2cmg_cands) { # 2nd trailing foreach-loop (over $gb2). #

	    my $cmg_cands = $gb2_2cmg_cands{$gb2};
	    my $ct_cmg_cands = @{$cmg_cands};
	    if ($ct_cmg_cands < 2) {
		push @to_be_cmerged, @{$cmg_cands};
		next;
	    }
	    my $mindist = $cmg_cands->[0][2];
	    for (my $i=1; $i <$ct_cmg_cands; $i++) {
		my $dist = $cmg_cands->[$i][2];
		if ($dist < $mindist) { $mindist = $dist; }
	    }
	    foreach my $cmg_cand (@{$cmg_cands}) { 
		if ($cmg_cand->[2] == $mindist) { push @to_be_cmerged, $cmg_cand; }
	    }

	} # End of the 2nd trailing foreach-loop (over $gb2). #
=cut # End of "For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (3)"


		### (3) Finally, sort the candidates in each of the sets. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (6) #

	my $std_to_be_cvmerged = multiple_sort (@to_be_cvmerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_cvmerged) {
	    print STDOUT "[ERROR 'list_nn_cvmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of complementary-sibling gap-block pairs to be \"(incomplete) vertical-merge\"d: $! !\n\n";
	    return (undef, undef);
	}

=For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (5)
	my $std_to_be_purged = multiple_sort (@to_be_purged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_purged) {
	    print STDOUT "[ERROR 'list_nn_cvmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of gap-block pairs to be \"purge\"d: $! !\n\n";
	    return (undef, undef);
	}

	my $std_to_be_cmerged = multiple_sort (@to_be_cmerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_cmerged) {
	    print STDOUT "[ERROR 'list_nn_cvmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of complementary gap-block pairs to be \"merge\"d: $! !\n\n";
	    return (undef, undef);
	}
=cut # End of "For the predecessor, 'list_nn_purge_and_cmerge_candidates'. (5)"

		# MODIFIED in creation of 'list_nn_cvmerge_candidates'. (7) #
	return ($std_to_be_cvmerged);
	##return ($std_to_be_purged, $std_to_be_cmerged);

}	# End of the subroutine, 'list_nn_cvmerge_candidates'. #


# list_smerge_split_candidates:
#	This subroutine creates a list of candidate pairs of 
#	gap-blocks of the same type to be "merge + split",
#	out of a list of candidates to be "merge"d.
#	[CREATED on 2019/08/01.]
#
# INPUTS: (\@to_be_smerged, \@gblock_sizes, $upper_bound_nrsize), where
#
#	@{$to_be_smerged[$smg]} = ($gb1, $gb2, $dist) provides information on the $smg th same-type gap-block pair to be "merge"d,
#	which follows the same format as @{$to_be_purged[$pg]} above;
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$upper_bound_nrsize is the upper-bound of the size of the newly created gap-block on the right; it is NEGATIVE if you want to impose NO aritficial upper-bounds;
#
# OUTPUTS: (\@to_be_smerge_split), where
#
#	@{$to_be_smerge_split[$ss]} = ($gb1, $gb2, $dist, \@new_sizes_right)
#	provides informaiton on the $ss th same-type gap-block pairs to be "merge + split",
#	where @new_sizes_right stores the sizes for the newly created gap-block on the right;
#
sub list_smerge_split_candidates (\@\@$) {

	my ($to_be_smerged, $gblock_sizes, $upper_bound_nrsize) = @_;

	my @to_be_smerge_split = ();

	foreach my $info_pair (@{$to_be_smerged}) { # Main foreach-loop (over pairs). #

	    my ($gb1, $gb2, $dist) = @{$info_pair};

	    my ($size1, $size2) = @{$gblock_sizes}[$gb1, $gb2];

	    my $max_nrsize = int (($size1 + $size2)/2) ;
	    if (($upper_bound_nrsize>0) and ($upper_bound_nrsize < $max_nrsize)) {
		$max_nrsize = $upper_bound_nrsize;
	    }

	    my @new_sizes_right = ();
	    foreach my $size (1 .. $max_nrsize) {
		if (($size == $size1) or ($size == $size2)) { next; }
		push @new_sizes_right, $size;
	    }
	    unless (@new_sizes_right > 0) { next; } # Skip if there are no sizes for the new gap-block on the right.

	    push @to_be_smerge_split, [$gb1, $gb2, $dist, \@new_sizes_right];
	    
	} # End of the main foreach-loop (over pairs). #

	return \@to_be_smerge_split;

}	# End of the subroutine, 'list_smerge_split_candidates'. #


# list_nn_cmerge_split_pre_candidates:
#	This subroutine creates a list of preliminary candidate pairs of "nearest-neighboring"
#	complementary gap-blocks to be "merge + split".
#	It was created by modifying 'list_nn_cvmerge_candidates'.
#	[CREATED on 2019/08/04.]
#
# list_nn_cvmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	complementary-sibling gap-blocks to be "(incomplete) vertical-merge"d.
#	It was created by taking advantage of the framework of 'list_nn_purge_and_cmerge_candidates'.
#	[CREATED on 2019/07/23.]
#	[REVISED on 2019/07/27]:
#	Now, when identifying "nearest-neighboring" pairs,
#	this subroutine only compares gap-blocks that are vertically identical to one another.
#
# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_cmerge_split), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_cmerge_split is the threshold for the distance between complementary-type gap-blocks to be "merge + split";
#
#
# OUTPUTS: (\@pre_to_be_cmerge_split), where
#
#	@{$pre_to_be_cmerge_split[$cms]} = ($gb1, $gb2, $dist) provides information on the $cms th complementary-type gap-block pair to be "merge + split",
#	$gb1 and $gb2 are the indices of the gap-blocks,
#	$dist is the distance of $gb2 from $gb1;
#
sub list_nn_cmerge_split_pre_candidates (\@\@\@$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_cmerge_split) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' complementary pairs for each $gb1. ###

	my %gb2_2cms_pre_cands;

	for (my $gb1 =0; $gb1 < $ct_gblks-1; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

	    my @cms_pre_cands = ();
	    my $mindist_cmspc;

	    for (my $gb2 = $gb1+1; $gb2 < $ct_gblks; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq 'Cp') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_cmerge_split_pre_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef);
		}

		my $dist = $dist2;

		if (1) {
		    if ($dist > $thrsh_dist_cmerge_split) { next; }

		    my $rels_w_gb2 = $inter_block_relations->[$gb2];

		    my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		    foreach my $pair (@cms_pre_cands) {
			my ($gb1c, $gb3, $dist3) = @{$pair};
			unless ($rels_w_gb2->[$gb3] eq '=') { next; }
			if ($dist < $dist3) {
			   $result_cmp = -1;
			} elsif ($dist > $dist3) {
			   $result_cmp = +1;
			} 
			last;
		    }

		    if ($result_cmp >0) {

			next; # Exclude the current pair. #

		    } elsif ($result_cmp < 0) {

			    # Remove the pairs of the same type. # 

			my @new_cms_pre_cands = ();
			foreach my $pair (@cms_pre_cands) {
			    my ($gb1c, $gb3, $dist3) = @{$pair};
			    if ($rels_w_gb2->[$gb3] eq '=') { next; }
			    push @new_cms_pre_cands, $pair;
			}
			@cms_pre_cands = @new_cms_pre_cands;
		    }

		    push @cms_pre_cands, [$gb1, $gb2, $dist];
		}

	    } # End of the inner for-loop (over $gb2). #


	    foreach my $pair (@cms_pre_cands) {
		my $gb2 = $pair->[1];
		my $cms_pre_cands = $gb2_2cms_pre_cands{$gb2};
		unless (defined $cms_pre_cands) {$cms_pre_cands = $gb2_2cms_pre_cands{$gb2} = []; }
		push @{$cms_pre_cands}, $pair;
	    }

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

	my @pre_to_be_cmerge_split = ();
	foreach my $gb2 (keys %gb2_2cms_pre_cands) { # Trailing foreach-loop (over $gb2). #

	    my $cms_pre_cands = $gb2_2cms_pre_cands{$gb2};
	    my $ct_cms_pre_cands = @{$cms_pre_cands};
	    if ($ct_cms_pre_cands < 2) {
		push @pre_to_be_cmerge_split, @{$cms_pre_cands};
		next;
	    }

	    my @nn_cms_pre_cands = ($cms_pre_cands->[0]);

	    for (my $i=1; $i <$ct_cms_pre_cands; $i++) {
		my ($gb1, $gb2c, $dist) = @{$cms_pre_cands->[$i]};

		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		foreach my $pair (@nn_cms_pre_cands) {
		    my ($gb1c, $gb2c, $distc) = @{$pair}; 
		    unless ($rels_w_gb1->[$gb1c] eq '=') { next; }
		    if ($dist < $distc) {
			$result_cmp = -1;
		    } elsif ($dist > $distc) {
			$result_cmp = +1;
		    }
		    last;
		} # End of "foreach my $pair (@nn_cms_pre_cands) {...}"

		if ($result_cmp > 0) {

		    next; # Exclude the current pair. #

		} elsif ($result_cmp < 0) {
			# Remove the pairs of the same type. #
		    my @new_nn_cms_pre_cands = ();
		    foreach my $pair (@nn_cms_pre_cands) {
			my ($gb1c, $gb2cc, $distc) = @{$pair}; 
			if ($rels_w_gb1->[$gb1c] eq '=') { next; }
			push @new_nn_cms_pre_cands, $pair;
		    }
		    @nn_cms_pre_cands = @new_nn_cms_pre_cands;
		}

		push @nn_cms_pre_cands, $cms_pre_cands->[$i];

	    } # End of "for (my $i=1; $i <$ct_cms_pre_cands; $i++) {...}"

	    push @pre_to_be_cmerge_split, @nn_cms_pre_cands;

	} # End of the trailing foreach-loop (over $gb2). #


		### (3) Finally, sort the pre-candidates in the set. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');


	my $std_pre_to_be_cmerge_split = multiple_sort (@pre_to_be_cmerge_split, @indices_srt, @methods_srt);
	unless (defined $std_pre_to_be_cmerge_split) {
	    print STDOUT "[ERROR 'list_nn_cmerge_split_pre_candidates'] The subroutine 'multiple_sort' FAILED on the list of complementary-type gap-block pairs to be \"merge + split\": $! !\n\n";
	    return (undef, undef);
	}

	return ($std_pre_to_be_cmerge_split);

}	# End of the subroutine, 'list_nn_cmerge_split_pre_candidates'. #


# enum_sites_avail_for_csplit:
#	This satelite subroutine of 'list_cmerge_split_candidates' and 'cmerge_split_grow_gblock_pair'
#	enumerates the sites (i.e., columns) available for the size-growing "cmerge + split" of complementary-types.
#	[CREATED on 2019/08/01.]
#
# INPUTS: ($gb1, $gb2, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@inter_block_relations, $le_wd, $re_wd), where
#
#	$gb1 and $gb2 are the gap-blocks that will be "cmerge + split";
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @bds_gb_coords below];
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	[Typically, \@bds_gb_coords is an output of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	$le_wd (usually zero) and $re_wd are the indices of the columns at the left-end and right-end, respectively, of the window we are dealing with;
#
# OUTPUTS: (\@avail_sites_left, \@avail_sites_in_btwn, \@avail_sites_right), where
#
#	Each of @avail_sites_left, @avail_sites_in_btwn, and @avail_sites_right
#	stores the indices of the sites (i.e., columns) available for the "cmerge + split";
#	They store the sites on the left of the left gap-block,
#	those in between the gap-blocks,
#	and those on the right of the right gap-block, respectively.
#
sub enum_sites_avail_for_csplit ($$\@\@\@\@$$) {

	my ($gb1, $gb2, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd) = @_;

	my $ct_gbs = @{$curr_bds_blks};
	my $rels_w_gb1 = $inter_block_relations->[$gb1];
	my $rels_w_gb2 = $inter_block_relations->[$gb2];

	my %clm2avoided_by_gb1;
	my %clm2avoided_by_gb2;
	my %clm2gbs_eq_ch;

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # 1st outer for-loop (over gap-blocks). #

	    if (($gb == $gb1) or ($gb == $gb2)) { next; }

	    my $rel1 = $rels_w_gb1->[$gb];
	    my $rel2 = $rels_w_gb2->[$gb];

	    my @relevant = ();
	    if (($rel1 eq '<') or ($rel1 eq '<(pa)') or (($rel1 eq '=') and ($gb < $gb1)) or ($rel1 eq 'ONN') or ($rel1 eq 'ONCS')) {
		push @relevant, \%clm2avoided_by_gb1;
	    }

	    if (($rel2 eq '<') or ($rel2 eq '<(pa)') or (($rel2 eq '=') and ($gb < $gb2)) or ($rel2 eq 'ONN') or ($rel2 eq 'ONCS')) {
		push @relevant, \%clm2avoided_by_gb2;
	    }

	    if (($rel1 eq '>(ch)') or (($rel1 eq '=') and ($gb > $gb1)) or ($rel2 eq '>(ch)') or (($rel2 eq '=') and ($gb > $gb2)) ) {
		push @relevant, \%clm2gbs_eq_ch;
	    }

	    my ($lbd, $rbd) = @{$curr_bds_blks->[$gb]};

	    foreach my $clm2gbs_rlv (@relevant) { # Middle foreach-loop.

		for (my $c=$lbd; $c <= $rbd; $c++) { # Inner for-loop (over columns in $gb). #
		    my $gbs_rlv = $clm2gbs_rlv->{$c};
		    unless (defined $gbs_rlv) { $gbs_rlv = $clm2gbs_rlv->{$c} = []; }
		    push @{$gbs_rlv}, $gb;
		} # End of the inner for-loop (over columns in $gb). #
	    } # End of the middle foreach-loop.

	} # End of the 1st outer for-loop (over gap-blocks). #


	my ($gb_left, $gb_right) = ($curr_bds_blks->[$gb1][0] < $curr_bds_blks->[$gb2][0]) ? ($gb1, $gb2) : ($gb2, $gb1);

	my ($lbd_left,  $rbd_left)  = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right, $rbd_right) = @{$curr_bds_blks->[$gb_right]};

	my $ct_left_margin = $curr_bl_coords->[$gb_left] - $bds_gb_coords->[$gb_left][0];
	my $ct_right_margin = $bds_gb_coords->[$gb_right][1] - $curr_bl_coords->[$gb_right];

	## my ($gb1, $gb2, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd) = @_;

	##my %clm2avoided_by_gb1;
	##my %clm2avoided_by_gb2;
	##my %clm2gbs_eq_ch;

	my @avail_sites_in_btwn = ();
	for (my $c=$rbd_left+1; $c <$lbd_right; $c++) { # 2nd for-loop (over sites in between). #
	    if (defined $clm2avoided_by_gb1{$c}) { next; }
	    if (defined $clm2avoided_by_gb2{$c}) { next; }
	    if (defined $clm2gbs_eq_ch{$c}) { next; }
	    push @avail_sites_in_btwn, $c;
	} # End of the 2nd for-loop (over sites in between). #

	my ($clm2avoided_by_left, $clm2avoided_by_right) = ($gb1 == $gb_left) ?
		(\%clm2avoided_by_gb1, \%clm2avoided_by_gb2) :
		(\%clm2avoided_by_gb2, \%clm2avoided_by_gb1) ;

	my @avail_sites_right = ();
	my $incr_coord_r=0;
	for (my $c=$rbd_right+1; $c <= $re_wd; $c++) { # 3rd for-loop (over sites on the right). #
	    if (defined $clm2avoided_by_right->{$c}) { next; }
	    $incr_coord_r++;
	    if ($incr_coord_r > $ct_right_margin) { last; }
	    if (defined $clm2avoided_by_left->{$c}) { next; }
	    if (defined $clm2gbs_eq_ch{$c}) { next; }
	    push @avail_sites_right, $c;
	} # End of the 3rd for-loop (over sites on the right). #

	my @avail_sites_left = ();
	my $incr_coord_l=0;
	for (my $c=$lbd_left-1; $c >= $le_wd; $c--) { # 5th for-loop (over sites on the left). #
	    if (defined $clm2avoided_by_left->{$c}) { next; }
	    $incr_coord_l++;
	    if ($incr_coord_l > $ct_left_margin) { last; }
	    if (defined $clm2avoided_by_right->{$c}) { next; }
	    if (defined $clm2gbs_eq_ch{$c}) { next; }
	    unshift @avail_sites_left, $c;

	} # End of the 5th for-loop (over sites on the left). #


	return (\@avail_sites_left, \@avail_sites_in_btwn, \@avail_sites_right);

}	# End of the subroutine, 'enum_sites_avail_for_csplit'. #


# enum_sites_avail_for_csplit_sgl_gb:
#	This satelite subroutine of 'list_csplit_candidates' and possibly 'csplit_gblock'
#	enumerates the sites (i.e., columns) available for the  "split" of the subject gap-block into a pair of complementary-type blocks.
#	It was created by modifying 'enum_sites_avail_for_csplit'.
#	[CREATED on 2019/08/15.]
#
# enum_sites_avail_for_csplit:
#	This satelite subroutine of 'list_cmerge_split_candidates' and 'cmerge_split_grow_gblock_pair'
#	enumerates the sites (i.e., columns) available for the size-growing "cmerge + split" of complementary-types.
#	[CREATED on 2019/08/01.]
#
# [ INPUTS: ($gb1, $gb2, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@inter_block_relations, $le_wd, $re_wd), where ] ... for the predecessor 'enum_sites_avail_for_csplit'.
#
# INPUTS: ($gb_sbj, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@inter_block_relations, $le_wd, $re_wd), where
#
#	[[
#	$gb1 and $gb2 are the gap-blocks that will be "cmerge + split";
#	]] ... for the predecessor 'enum_sites_avail_for_csplit'.
#
#	$gb_sbj is the gap-block that will be "csplit";
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @bds_gb_coords below];
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	[Typically, \@bds_gb_coords is an output of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	$le_wd (usually zero) and $re_wd are the indices of the columns at the left-end and right-end, respectively, of the window we are dealing with;
#
# [ OUTPUTS: (\@avail_sites_left, \@avail_sites_in_btwn, \@avail_sites_right), where ] ... for the predecessor 'enum_sites_avail_for_csplit'.
#
# OUTPUTS: (\@avail_sites_left, \@avail_sites_right), where
#
#	Each of @avail_sites_left and @avail_sites_right
#	stores the indices of the sites (i.e., columns) available for the "cmerge + split";
#	They store the sites on the left of the subject gap-block,
#	and those on the right of the subject gap-block, respectively.
#
sub enum_sites_avail_for_csplit_sgl_gb ($\@\@\@\@$$) {

	my ($gb_sbj, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd) = @_;

	my $ct_gbs = @{$curr_bds_blks};
	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

			# MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (1) #
	my %clm2avoided_by_sbj;
	my %clm2gbs_eq_cmpl_ch;
	##my %clm2avoided_by_gb1;
	##my %clm2avoided_by_gb2;
	##my %clm2gbs_eq_ch;
			# End of "MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (1)" #

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # 1st outer for-loop (over gap-blocks). #

	    if ($gb == $gb_sbj) { next; }

	    my $rel = $rels_w_sbj->[$gb];

	    my @relevant = ();
	    if (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb < $gb_sbj)) or ($rel eq 'ONN') or ($rel eq 'ONCS')) {
		push @relevant, \%clm2avoided_by_sbj;
	    }

			# MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (2) #
	    if (($rel eq '>(ch)') or (($rel eq '=') and ($gb > $gb_sbj)) or ($rel eq 'Cp') or ($rel eq 'S') ) {
		push @relevant, \%clm2gbs_eq_cmpl_ch;
	    }

=For the predecessor 'enum_sites_avail_for_csplit'. (1)
	    if (($rel2 eq '<') or ($rel2 eq '<(pa)') or (($rel2 eq '=') and ($gb < $gb2)) or ($rel2 eq 'ONN') or ($rel2 eq 'ONCS')) {
		push @relevant, \%clm2avoided_by_gb2;
	    }

	    if (($rel1 eq '>(ch)') or (($rel1 eq '=') and ($gb > $gb1)) or ($rel2 eq '>(ch)') or (($rel2 eq '=') and ($gb > $gb2)) ) {
		push @relevant, \%clm2gbs_eq_ch;
	    }
=cut # END of "For the predecessor 'enum_sites_avail_for_csplit'. (1)"
			# End of "MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (2)" #

	    my ($lbd, $rbd) = @{$curr_bds_blks->[$gb]};

	    foreach my $clm2gbs_rlv (@relevant) { # Middle foreach-loop.

		for (my $c=$lbd; $c <= $rbd; $c++) { # Inner for-loop (over columns in $gb). #
		    my $gbs_rlv = $clm2gbs_rlv->{$c};
		    unless (defined $gbs_rlv) { $gbs_rlv = $clm2gbs_rlv->{$c} = []; }
		    push @{$gbs_rlv}, $gb;
		} # End of the inner for-loop (over columns in $gb). #
	    } # End of the middle foreach-loop.

	} # End of the 1st outer for-loop (over gap-blocks). #


			# MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (3) #
	my ($lbd_sbj,  $rbd_sbj)  = @{$curr_bds_blks->[$gb_sbj]};

	my $ct_left_margin = $curr_bl_coords->[$gb_sbj] - $bds_gb_coords->[$gb_sbj][0];
	my $ct_right_margin = $bds_gb_coords->[$gb_sbj][1] - $curr_bl_coords->[$gb_sbj];

=For the predecessor 'enum_sites_avail_for_csplit'. (2)
	my ($gb_left, $gb_right) = ($curr_bds_blks->[$gb1][0] < $curr_bds_blks->[$gb2][0]) ? ($gb1, $gb2) : ($gb2, $gb1);

	my ($lbd_left,  $rbd_left)  = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right, $rbd_right) = @{$curr_bds_blks->[$gb_right]};

	my $ct_left_margin = $curr_bl_coords->[$gb_left] - $bds_gb_coords->[$gb_left][0];
	my $ct_right_margin = $bds_gb_coords->[$gb_right][1] - $curr_bl_coords->[$gb_right];
=cut # End of "For the predecessor 'enum_sites_avail_for_csplit'. (2)"
			# End of "MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (3)" #

=For the predecessor 'enum_sites_avail_for_csplit'. (3)
	## my ($gb1, $gb2, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd) = @_;

	##my %clm2avoided_by_gb1;
	##my %clm2avoided_by_gb2;
	##my %clm2gbs_eq_ch;

	my @avail_sites_in_btwn = ();
	for (my $c=$rbd_left+1; $c <$lbd_right; $c++) { # 2nd for-loop (over sites in between). #
	    if (defined $clm2avoided_by_gb1{$c}) { next; }
	    if (defined $clm2avoided_by_gb2{$c}) { next; }
	    if (defined $clm2gbs_eq_ch{$c}) { next; }
	    push @avail_sites_in_btwn, $c;
	} # End of the 2nd for-loop (over sites in between). #

	my ($clm2avoided_by_left, $clm2avoided_by_right) = ($gb1 == $gb_left) ?
		(\%clm2avoided_by_gb1, \%clm2avoided_by_gb2) :
		(\%clm2avoided_by_gb2, \%clm2avoided_by_gb1) ;
=cut # END of "For the predecessor 'enum_sites_avail_for_csplit'. (3)"

	my @avail_sites_right = ();
	my $incr_coord_r=0;
	for (my $c=$rbd_sbj+1; $c <= $re_wd; $c++) { # 3rd for-loop (over sites on the right). #
	    if (defined $clm2avoided_by_sbj{$c}) { next; }
	    ##if (defined $clm2avoided_by_right->{$c}) { next; }
	    $incr_coord_r++;
	    if ($incr_coord_r > $ct_right_margin) { last; }
	    ##if (defined $clm2avoided_by_left->{$c}) { next; }
	    if (defined $clm2gbs_eq_cmpl_ch{$c}) { next; }
	    ##if (defined $clm2gbs_eq_ch{$c}) { next; }
	    push @avail_sites_right, $c;
	} # End of the 3rd for-loop (over sites on the right). #

	my @avail_sites_left = ();
	my $incr_coord_l=0;
	for (my $c=$lbd_sbj-1; $c >= $le_wd; $c--) { # 5th for-loop (over sites on the left). #
	    if (defined $clm2avoided_by_sbj{$c}) { next; }
	    ##if (defined $clm2avoided_by_left->{$c}) { next; }
	    $incr_coord_l++;
	    if ($incr_coord_l > $ct_left_margin) { last; }
	    ##if (defined $clm2avoided_by_right->{$c}) { next; }
	    if (defined $clm2gbs_eq_cmpl_ch{$c}) { next; }
	    ##if (defined $clm2gbs_eq_ch{$c}) { next; }
	    unshift @avail_sites_left, $c;

	} # End of the 5th for-loop (over sites on the left). #


			# MODIFIED in creation of 'enum_sites_avail_for_csplit_sgl_gb'. (5) #
	return (\@avail_sites_left, \@avail_sites_right);
	##return (\@avail_sites_left, \@avail_sites_in_btwn, \@avail_sites_right);

}	# End of the subroutine, 'enum_sites_avail_for_csplit_sgl_gb'. #


# list_cmerge_split_candidates:
#	This subroutine creates a list of candidate pairs of 
#	gap-blocks of complementary types to be "merge + split",
#	out of a list of candidates to be "merge"d.
#	[CREATED on 2019/08/01.]
#	[REVISED on 2019/08/05]:
#	Now, it discards preliminary-candidates if NO size increments, either negative or positive, are available.
#
# [ INPUTS: (\@to_be_cmerged, \@gblock_sizes, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@inter_block_relations, $le_wd, $re_wd, $lower_bound_dsize, $upper_bound_dsize), where ] ... OBSOLETE as of 2019/08/04.
# INPUTS: (\@pre_to_be_cmerge_split, \@gblock_sizes, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@inter_block_relations, $le_wd, $re_wd, $lower_bound_dsize, $upper_bound_dsize), where
#
#		# Modified on 2019/08/04.
#	@{$pre_to_be_cmerge_split[$cms]} = ($gb1, $gb2, $dist) provides information on the $cms th preliminary candidate for the complementary-type gap-block pairs to be "merge + split",
#	which follows the same format as @{$to_be_purged[$pg]} above;
#	[[
#	@{$to_be_cmerged[$cmg]} = ($gb1, $gb2, $dist) provides information on the $cmg th complementary-type gap-block pair to be "merge"d,
#	which follows the same format as @{$to_be_purged[$pg]} above;
#	]] ... OBSOLETE as of 2019/08/04.
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @bds_gb_coords below];
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	[Typically, \@bds_gb_coords is an output of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	$le_wd (usually zero) and $re_wd are the indices of the columns at the left-end and right-end, respectively, of the window we are dealing with;
#
#	$lower_bound_dsize (usually negative) and $upper_bound_dsize (usually positive) are the lower-bound and upper-bound, respectively, of the size differences that will result from the "cmerge + split";
#
# OUTPUTS: (\@to_be_cmerge_split), where
#
#	@{$to_be_cmerge_split[$cms]} = ($gb1, $gb2, $dist, \@incr_gb_sizes)
#	provides informaiton on the $cms th complementary-type gap-block pairs to be "merge + split", where
#	@incr_gb_sizes stores the gap-block-size increments that will result from the "cmerge + split";
#
sub list_cmerge_split_candidates (\@\@\@\@\@\@$$$$) {

	## [ my ($to_be_cmerged, $gblock_sizes, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd, $lower_bound_dsize, $upper_bound_dsize) = @_; ] ... OBSOLETE as of 2019/08/04.
	my ($pre_to_be_cmerge_split, $gblock_sizes, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $inter_block_relations, $le_wd, $re_wd, $lower_bound_dsize, $upper_bound_dsize) = @_;

	my @to_be_cmerge_split = ();

	foreach my $info_pair (@{$pre_to_be_cmerge_split}) { # Main foreach-loop (over pairs). # MODIFIED in revision on 2019/08/04.
	##foreach my $info_pair (@{$to_be_cmerged}) { # Main foreach-loop (over pairs). #

	    my ($gb1, $gb2, $dist) = @{$info_pair};

	    my ($size1, $size2) = @{$gblock_sizes}[$gb1, $gb2];

	    my $minsize = ($size1 < $size2) ? $size1 : $size2 ;

	    my $min_dsize = - ($minsize -1);
	    if ($min_dsize < $lower_bound_dsize) { $min_dsize = $lower_bound_dsize; }

	    my @incr_gb_sizes = ($min_dsize .. -1);

	    my ($avail_sites_left, $avail_sites_in_btwn, $avail_sites_right)
		= enum_sites_avail_for_csplit ($gb1, $gb2, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$inter_block_relations}, $le_wd, $re_wd);
	    unless (defined $avail_sites_right) {
		print STDOUT "[ERROR 'list_cmerge_split_candidates'] The subroutine 'enum_sites_avail_for_csplit' FAILED on the gap-block pair, ($gb1, $gb2): $! !\n\n";
		return undef;
	    }

	    my $max_dsize = scalar (@{$avail_sites_left}) + scalar (@{$avail_sites_in_btwn}) + scalar(@{$avail_sites_right});
	    if ($upper_bound_dsize < $max_dsize) { $max_dsize = $upper_bound_dsize; }

	    push @incr_gb_sizes, (1 .. $max_dsize);

	    unless (@incr_gb_sizes>0) { next; } # Skip if NO size-increments are stored. (ADDED in revision on 2019/08/05.)

	    push @to_be_cmerge_split, [$gb1, $gb2, $dist, \@incr_gb_sizes];
	    
	} # End of the main foreach-loop (over pairs). #

	return \@to_be_cmerge_split;

}	# End of the subroutine, 'list_cmerge_split_candidates'. #


# list_ssplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	that are to be "split" into two blocks of the same type.
#	It was created by taking advantage of the framework of 'list_smerge_split_candidates'.
#	[CREATED on 2019/08/11.]
#
#	[Outstanding Problem]:
#	When there are two or more mutually vertically equivalent gap-blocks,
#	some DEGENERACIES could result from "split"ting different (but vertically equivalent) gap-blocks.
#	In the future, this subroutine MUST take account of such degeneracies as well.
#
# list_smerge_split_candidates:
#	This subroutine creates a list of candidate pairs of 
#	gap-blocks of the same type to be "merge + split",
#	out of a list of candidates to be "merge"d.
#	[CREATED on 2019/08/01.]
#
# INPUTS: (\@curr_bds_gblocks, \@gblock_sizes, \@inter_block_relations, \@rows_affd_by_blks, \@list_pcands, \@indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_nrsize), where
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$upper_bound_nrsize is the upper-bound of the size of the newly created gap-block on the right; it is NEGATIVE if you want to impose NO aritficial upper-bounds; # INHERITED from 'list_smerge_split_candidates'.
#
#	@{$list_pcands[$pc]}[@indices_pc] = ($beg_pc, $end_pc, \@rows_under_br, \@constituent_sites)
#	provide information on the $pc th purge-candidate region, where
#	$beg/end_pc is the index of the column at the beginning/end of the region,
#	@rows_under_br stores the indices of the MSA rows under the branch delimiting the $pc th purge-candidate,
#	@constituent_sites stores the indices of the sites constituting the purge-candidate region;
#	[Typically, \@list_pcands is the main output of 'readin_list_mgd_pcands' (in "MyANEX_IO.pm")];
#
# OUTPUTS: (\@to_be_ssplit), where
#
#	[ @{$to_be_ssplit[$ss]} = ($indx_gb, $dist, \@new_sizes_right, \@indices_rlv_pcs, $indx_closest_rlv_pc) ] ... OBSOLETE as of 2019/08/13.
#	@{$to_be_ssplit[$ss]} = ($indx_gb, $dist, \@new_sizes_right, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@available_sites)
#	provides informaiton on the $ss th gap-block to be "split" into two blocks of the same type,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	@new_sizes_right stores the sizes for the newly created gap-block on the right,
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#
#	@available_sites stores the sites available for the complementary split (to be used, e.g., in 'csplit_gblock'); it is empty in this subroutine; # ADDED on 2019/08/13.
#
sub list_ssplit_candidates (\@\@\@\@\@\@$$$) {

	my ($curr_bds_gblocks, $gblock_sizes, $inter_block_relations, $rows_affd_by_blks, 
	    $list_pcands, $indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_nrsize) = @_;

	my $ct_gbs = @{$curr_bds_gblocks};
	my $ct_pcs = @{$list_pcands};

	my @to_be_ssplit = ();

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_gblocks->[$gb]};
	    my $rows_affd_by_gb = $rows_affd_by_blks->[$gb];
	    my $ct_rows_affd_by_gb = @{$rows_affd_by_gb};

	    my $size_gb = $gblock_sizes->[$gb];

	    if ($size_gb < 2) { next; } # If the size is less than 2, the block can NEVER be "split". #

	    my @indices_rlv_pcs = ();
	    my @closest_pc = (); # = ($indx_pc, $dist, $ct_cnst_sites)

	    for (my $pc=0; $pc <$ct_pcs; $pc++) { # Inner for-loop (over purge-candidate regions). #

		my ($beg_pc, $end_pc, $rows_under_br, $constituent_sites) = @{$list_pcands->[$pc]}[@{$indices_pc}];
		my $ct_rows_ud_br = @{$rows_under_br};
		unless (($ct_rows_affd_by_gb == $ct_rows_ud_br) or ($ct_rows_affd_by_gb + $ct_rows_ud_br == $ct_seqs)) {
		    next;
		}
		my ($common, $ex_gb, $ex_pc) = diff_nmarray (@{$rows_affd_by_gb}, @{$rows_under_br});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_ssplit_candidates'] Subroutine 'diff_nmarray' FAILED between the $gb th gap-block and the $pc th purege-candidate region: $! !\n\n";
		    return undef;
		}

		my $if_relv =0;
		if ((@{$ex_gb} == 0) and (@{$ex_pc} == 0)) {
		    $if_relv = 1; # The two sets of rows are identical. #

		} elsif ((@{$common} == 0) and ($ct_rows_affd_by_gb + $ct_rows_ud_br == $ct_seqs)) {
		    $if_relv = 1; # The two sets of rows are complementary. #
		}
		unless ($if_relv) { next; }

		my $dist;

		if ($rbd_gb < $beg_pc) {
		    $dist = compt_distance_bw_gblk_and_pt ($beg_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_ssplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $beg_pc, which is the left-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }

		} elsif ($end_pc < $lbd_gb) {
		    $dist = compt_distance_bw_gblk_and_pt ($end_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_ssplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $end_pc, which is the right-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }
		} else {
		    $dist = -1;
		}

		if ($thrsh_dist_pc < $dist) { next; }

		push @indices_rlv_pcs, $pc;

		my $ct_cnst_sites = @{$constituent_sites};

		if (@closest_pc > 0) {
		    if ($dist > $closest_pc[1]) {
			next;
		    } elsif ($dist == $closest_pc[1]) {
			if ($ct_cnst_sites <= $closest_pc[2]) { next; }
		    }
		}

		@closest_pc = ($pc, $dist, $ct_cnst_sites);

	    } # End of the inner for-loop (over purge-candidate regions). #

	    unless (@indices_rlv_pcs>0) { next; } # Skip if no purge-candidates share the delimiting branch with $gb. #
	    my ($indx_closest_rlv_pc, $dist_closest, $ct_cnst_sites_closest) = @closest_pc;

	    my $max_nrsize = int ($size_gb /2) ;
	    if ($ct_cnst_sites_closest < $max_nrsize) { $max_nrsize = $ct_cnst_sites_closest; }

	    if (($upper_bound_nrsize>0) and ($upper_bound_nrsize < $max_nrsize)) {
		$max_nrsize = $upper_bound_nrsize;
	    }

	    my @new_sizes_right = (1 .. $max_nrsize);

	    push @to_be_ssplit, [$gb, $dist_closest, \@new_sizes_right, \@indices_rlv_pcs, $indx_closest_rlv_pc, []]; # REVISED on 2019/08/13. #
	    ##push @to_be_ssplit, [$gb, $dist_closest, \@new_sizes_right, \@indices_rlv_pcs, $indx_closest_rlv_pc];

	} # End of the main for-loop (over gap-blocks). #

	return \@to_be_ssplit;

}	# End of the subroutine, 'list_ssplit_candidates'. #


# list_csplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	that are to be "split" into two blocks of complementary types.
#	It was created by modifying 'list_ssplit_candidates'.
#	[CREATED on 2019/08/14.]
#
# list_ssplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	that are to be "split" into two blocks of the same type.
#	It was created by taking advantage of the framework of 'list_smerge_split_candidates'.
#	[CREATED on 2019/08/11.]
#
#	[Outstanding Problem]:
#	When there are two or more mutually vertically equivalent gap-blocks,
#	some DEGENERACIES could result from "split"ting different (but vertically equivalent) gap-blocks.
#	In the future, this subroutine MUST take account of such degeneracies as well.
#
# list_smerge_split_candidates:
#	This subroutine creates a list of candidate pairs of 
#	gap-blocks of the same type to be "merge + split",
#	out of a list of candidates to be "merge"d.
#	[CREATED on 2019/08/01.]
#
# [ INPUTS: (\@curr_bds_gblocks, \@gblock_sizes, \@inter_block_relations, \@rows_affd_by_blks, \@list_pcands, \@indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_nrsize), where ] ... for the predecessor, 'list_ssplit_candidates'.
#
# INPUTS: (\@curr_bds_gblocks, 
#	\@curr_bl_coords, \@gb_coords_wd_frm, $le_wd, $re_wd,
#	\@gblock_sizes, \@inter_block_relations, \@rows_affd_by_blks, \@list_pcands, \@indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_ncsize), where
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#		# ADDED in creation of 'list_csplit_candidates'. (HD) #
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @gb_coords_wd_frm below];
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@gb_coords_wd_frm is an output of 'mk_orgs_bds_gb_coords' (in "MyANEX_Supple.pm")];
#
#	$le_wd (usually zero) and $re_wd are the indices of the columns at the left-end and right-end, respectively, of the window we are dealing with;
#		# END of "ADDED in creation of 'list_csplit_candidates'. (HD)" #
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$upper_bound_ncsize is the upper-bound of the size of the newly created gap-block vertically complementary to the subject; it is NEGATIVE if you want to impose NO aritficial upper-bounds; # INHERITED from 'list_smerge_split_candidates'. => Modified in creation of 'list_csplit_candidates'.
#
#	@{$list_pcands[$pc]}[@indices_pc] = ($beg_pc, $end_pc, \@rows_under_br, \@constituent_sites)
#	provide information on the $pc th purge-candidate region, where
#	$beg/end_pc is the index of the column at the beginning/end of the region,
#	@rows_under_br stores the indices of the MSA rows under the branch delimiting the $pc th purge-candidate,
#	@constituent_sites stores the indices of the sites constituting the purge-candidate region;
#	[Typically, \@list_pcands is the main output of 'readin_list_mgd_pcands' (in "MyANEX_IO.pm")];
#
# [ OUTPUTS: (\@to_be_ssplit), where ] ... for the predecessor 'list_ssplit_candidates'.
#
# OUTPUTS: (\@to_be_csplit), where
#
#	@{$to_be_csplit[$cs]} = ($indx_gb, $dist, \@new_sizes_cmpl, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@available_sites)
#	provides informaiton on the $cs th gap-block to be "split" into two blocks of complementary types,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	[ @new_sizes_right stores the sizes for the newly created gap-block on the right, ] ... for the predecessor 'list_ssplit_candidates'.
#	@new_sizes_cmpl stores the sizes for the newly created gap-block of the complementary type, # MODIFIED in creation of 'list_csplit_candidates'. #
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#
#	@available_sites stores the sites available for the complementary split (to be used, e.g., in 'csplit_gblock');
#
# [ sub list_ssplit_candidates (\@\@\@\@\@\@$$$) {
#	my ($curr_bds_gblocks, $gblock_sizes, $inter_block_relations, $rows_affd_by_blks, 
#	    $list_pcands, $indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_nrsize) = @_; ] ... for the predecessor 'list_ssplit_candidates'.
#
sub list_csplit_candidates (\@\@\@$$\@\@\@\@\@$$$) {

	my ($curr_bds_gblocks, 
	    $curr_bl_coords, $gb_coords_wd_frm, $le_wd, $re_wd,
	    $gblock_sizes, $inter_block_relations, $rows_affd_by_blks, 
	    $list_pcands, $indices_pc, $ct_seqs, $thrsh_dist_pc, $upper_bound_ncsize) = @_;


	my $ct_gbs = @{$curr_bds_gblocks};
	my $ct_pcs = @{$list_pcands};

	my @to_be_csplit = ();

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_gblocks->[$gb]};
	    my $rows_affd_by_gb = $rows_affd_by_blks->[$gb];
	    my $ct_rows_affd_by_gb = @{$rows_affd_by_gb};

			# MODIFIED in creation of 'list_csplit_candidates'. (1) #
	    my ($pre_avail_sites_left, $pre_avail_sites_right)
		 = enum_sites_avail_for_csplit_sgl_gb ($gb, @{$curr_bds_gblocks}, @{$curr_bl_coords}, @{$gb_coords_wd_frm}, @{$inter_block_relations}, $le_wd, $re_wd);
	    unless (defined $pre_avail_sites_left) {
		print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'enum_sites_avail_for_csplit_sgl_gb' FAILED on the $gb th gap-block: $! !\n\n";
		return undef;
	    }

	    unless ((@{$pre_avail_sites_left}>0) or (@{$pre_avail_sites_right}>0)) { next; } # Skip if NO sites are available for the "csplit" of $gb. 

=For the predecessor 'list_ssplit_candidates'. (1)
	    my $size_gb = $gblock_sizes->[$gb];

	    if ($size_gb < 2) { next; } # If the size is less than 2, the block can NEVER be "split". #
=cut # END of "For the predecessor 'list_ssplit_candidates'. (1)"
			# End of "MODIFIED in creation of 'list_csplit_candidates'. (1)" #

	    my @indices_rlv_pcs = ();
	    my @closest_pc = (); # = ($indx_pc, $dist, $ct_cnst_sites)

	    for (my $pc=0; $pc <$ct_pcs; $pc++) { # Inner for-loop (over purge-candidate regions). #

		my ($beg_pc, $end_pc, $rows_under_br, $constituent_sites) = @{$list_pcands->[$pc]}[@{$indices_pc}];
		my $ct_rows_ud_br = @{$rows_under_br};
		unless (($ct_rows_affd_by_gb == $ct_rows_ud_br) or ($ct_rows_affd_by_gb + $ct_rows_ud_br == $ct_seqs)) {
		    next;
		}
		my ($common, $ex_gb, $ex_pc) = diff_nmarray (@{$rows_affd_by_gb}, @{$rows_under_br});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'diff_nmarray' FAILED between the $gb th gap-block and the $pc th purege-candidate region: $! !\n\n";
		    return undef;
		}

		my $if_relv =0;
		if ((@{$ex_gb} == 0) and (@{$ex_pc} == 0)) {
		    $if_relv = 1; # The two sets of rows are identical. #

		} elsif ((@{$common} == 0) and ($ct_rows_affd_by_gb + $ct_rows_ud_br == $ct_seqs)) {
		    $if_relv = 1; # The two sets of rows are complementary. #
		}
		unless ($if_relv) { next; }

		my $dist;

		if ($rbd_gb < $beg_pc) {
		    $dist = compt_distance_bw_gblk_and_pt ($beg_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $beg_pc, which is the left-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }

		} elsif ($end_pc < $lbd_gb) {
		    $dist = compt_distance_bw_gblk_and_pt ($end_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $end_pc, which is the right-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }
		} else {
		    $dist = -1;
		}

		if ($thrsh_dist_pc < $dist) { next; }

		push @indices_rlv_pcs, $pc;

		my $ct_cnst_sites = @{$constituent_sites};

		if (@closest_pc > 0) {
		    if ($dist > $closest_pc[1]) {
			next;
		    } elsif ($dist == $closest_pc[1]) {
			if ($ct_cnst_sites <= $closest_pc[2]) { next; }
		    }
		}

		@closest_pc = ($pc, $dist, $ct_cnst_sites);

	    } # End of the inner for-loop (over purge-candidate regions). #

	    unless (@indices_rlv_pcs>0) { next; } # Skip if no purge-candidates share the delimiting branch with $gb. #
	    my ($indx_closest_rlv_pc, $dist_closest, $ct_cnst_sites_closest) = @closest_pc;

			# MODIFIED in creation of 'list_csplit_candidates'. (2) #

	    my ($beg_clsst_pc, $end_clsst_pc, $rows_under_br_clsst, $constituent_sites_clsst) = @{$list_pcands->[$indx_closest_rlv_pc]}[@{$indices_pc}];

	    # unless ((@{$pre_avail_sites_left}>0) or (@{$pre_avail_sites_right}>0)) { next; } # Skip if NO sites are available for the "csplit" of $gb. 

	    my ($avail_sites_left, $ex_avleft, $ex_pc_clsst_l) = diff_nmarray (@{$pre_avail_sites_left}, @{$constituent_sites_clsst});
	    unless (defined $avail_sites_left) {
		print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'diff_nmarray' FAILED for the available sites on the left of the $gb th gap-block: $! !\n\n";
		return undef;
	    }
	    my ($avail_sites_right, $ex_avright, $ex_pc_clsst_r) = diff_nmarray (@{$pre_avail_sites_right}, @{$constituent_sites_clsst});
	    unless (defined $avail_sites_right) {
		print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'diff_nmarray' FAILED for the available sites on the left of the $gb th gap-block: $! !\n\n";
		return undef;
	    }

	    my @prot_avail_sites = (@{$avail_sites_right}, reverse  (@{$avail_sites_left}) );
	    my @avail_sites;
	    if (($upper_bound_ncsize>0) and ($upper_bound_ncsize < @prot_avail_sites)) {
		@avail_sites = @prot_avail_sites[0 .. $upper_bound_ncsize-1];
	    } else {
		@avail_sites = @prot_avail_sites;
	    }

	    my $max_size_cmpl = @avail_sites;

=For the predecessor 'list_ssplit_candidates'. (2)
	    my $max_nrsize = int ($size_gb /2) ;
	    if ($ct_cnst_sites_closest < $max_nrsize) { $max_nrsize = $ct_cnst_sites_closest; }

	    if (($upper_bound_nrsize>0) and ($upper_bound_nrsize < $max_nrsize)) {
		$max_nrsize = $upper_bound_nrsize;
	    }
	    my @new_sizes_right = (1 .. $max_nrsize);
=cut # END of "For the predecessor 'list_ssplit_candidates'. (2)"

			# END of "MODIFIED in creation of 'list_csplit_candidates'. (2)" #

	    my @new_sizes_cmpl = (1 .. $max_size_cmpl);

			# MODIFIED in creation of 'list_csplit_candidates'. (3) #
	    push @to_be_csplit, [$gb, $dist_closest, \@new_sizes_cmpl, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@avail_sites];
	    ## push @to_be_ssplit, [$gb, $dist_closest, \@new_sizes_right, \@indices_rlv_pcs, $indx_closest_rlv_pc, []]; # REVISED on 2019/08/13. #
			# END of "MODIFIED in creation of 'list_csplit_candidates'. (3)" #

	} # End of the main for-loop (over gap-blocks). #

	return \@to_be_csplit;

}	# End of the subroutine, 'list_csplit_candidates'. #


# list_nn_csplit_candidates:
#	This subroutine receives a list of "nearest-neighboring" candidate gap-blocks,
#	which are to be "split" into two blocks of complementary types,
#	and selects only candidates that are the closest to (or with non-positive distances to) the 'closest' purge-candidate regions.
#	(The input list was typically output by 'list_csplit_candidates'.)
#	[CREATED on 2019/08/20.]
#
# INPUTS: (\@to_be_csplit), where
#
#	@{$to_be_csplit[$cs]} = ($indx_gb, $dist, \@new_sizes_cmpl, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@available_sites)
#	provides informaiton on the $cs th gap-block to be "split" into two blocks of complementary types,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	[ @new_sizes_right stores the sizes for the newly created gap-block on the right, ] ... for the predecessor 'list_ssplit_candidates'.
#	@new_sizes_cmpl stores the sizes for the newly created gap-block of the complementary type, # MODIFIED in creation of 'list_csplit_candidates'. #
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	@available_sites stores the sites available for the complementary split (to be used, e.g., in 'csplit_gblock');
#	[Typically, \@to_be_csplit is the output of 'list_csplit_candidates'];
#
# OUTPUT: (\@list_nn_csplit_cands), where
#
#	@{$list_nn_csplit_cands[$nnc]} = ($indx_gb, $dist, \@new_sizes_cmpl, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@available_sites)
#	provides informaiton on the $nnc th gap-block to be "split" into two blocks of complementary types;
#	Its format is the same as that for @{$to_be_csplit[$cs]} in the input.
#
sub list_nn_csplit_candidates (\@) {

	my ($to_be_csplit) = @_;

	my $ct_2Bcss = @{$to_be_csplit};

	my %clsst_pc2indices_2Bcs; # = ($indx_closest_rlv_pc => \@indices_to_be_csplit, ...)

	for (my $cs=0; $cs < $ct_2Bcss; $cs++) { # Preliminary for-loop. #

	    my ($indx_gb, $dist, $new_sizes_cmpl, $indices_rlv_pcs, $indx_closest_rlv_pc, $available_sites) = @{$to_be_csplit->[$cs]};
	    my $indices_2Bcs = $clsst_pc2indices_2Bcs{$indx_closest_rlv_pc};
	    unless (defined $indices_2Bcs) { $indices_2Bcs = $clsst_pc2indices_2Bcs{$indx_closest_rlv_pc} = []; }
	    push @{$indices_2Bcs}, $cs;

	} # End of the preliminary for-loop.

	my @indices_nn_cs_cands = ();

	foreach my $clsst_pc (sort {$a <=> $b} keys %clsst_pc2indices_2Bcs) { # MAIN foreach-loop. #
	    my $indices_2Bcs = $clsst_pc2indices_2Bcs{$clsst_pc};
	    if (@{$indices_2Bcs} <2) {
		push @indices_nn_cs_cands, @{$indices_2Bcs};
		next;
	    }
	    my $min_dist;
	    foreach my $cs (@{$indices_2Bcs}) { # 1st inner foreach-loop. #
		my ($indx_gb, $dist, $new_sizes_cmpl, $indices_rlv_pcs, $indx_closest_rlv_pc, $available_sites) = @{$to_be_csplit->[$cs]};
		unless ((defined $min_dist) and ($min_dist <= $dist)) { $min_dist = $dist; }
	    } # End of the 1st inner foreach-loop. #

	    if ($min_dist < 0) { $min_dist = 0; }

	    foreach my $cs (@{$indices_2Bcs}) { # 2nd inner foreach-loop. #
		my ($indx_gb, $dist, $new_sizes_cmpl, $indices_rlv_pcs, $indx_closest_rlv_pc, $available_sites) = @{$to_be_csplit->[$cs]};
		if ($dist <= $min_dist) { push @indices_nn_cs_cands, $cs; }
	    } # End of the 2nd inner foreach-loop. #

	} # End of the MAIN foreach-loop. #

	my @list_nn_csplit_cands = ();

	foreach my $cs (sort {$a <=> $b} @indices_nn_cs_cands) { # Finishing foreach-loop. #

	    push @list_nn_csplit_cands, $to_be_csplit->[$cs];
	    
	} # End of the finishing foreach-loop. #

	return \@list_nn_csplit_cands ;

}	# End of the subroutine, 'list_nn_csplit_candidates'. #


# list_nn_gen_split_candidates:
#	This subroutine is a generalization of the subroutine, 'list_nn_csplit_candidates'.
#	It selects only candidates that are the closest (or with non-positive distance to) the "closest' purge-candidate regions.
#	[CTEATED on 2019/08/21.]
#
# list_nn_csplit_candidates:
#	This subroutine receives a list of "nearest-neighboring" candidate gap-blocks,
#	which are to be "split" into two blocks of complementary types,
#	and selects only candidates that are the closest to (or with non-positive distances to) the 'closest' purge-candidate regions.
#	(The input list was typically output by 'list_csplit_candidates'.)
#	[CREATED on 2019/08/20.]
#
# INPUTS: (\@to_be_split, $indx_dist, $indx_indx_clsst_pc), where
#
#	@{$to_be_split[$sp]} provides informaiton on the $sp th gap-block to be "split" into two blocks of some specified type(s);
#	more precisely,
#	@{$to_be_split[$sp]}[$indx_dist, $indx_indx_clsst_pc] = ($dist, $indx_closest_rlv_pc)
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	[Typically, \@to_be_csplit is the output of 'list_csplit_candidates', 'list_vsplit_candidates', etc.];
#
# OUTPUT: (\@list_nn_split_cands), where
#
#	@{$list_nn_split_cands[$nns]} 
#	provides informaiton on the $nns th gap-block to be "split" into two blocks of some specified type(s);
#	Its format is the same as that for @{$to_be_split[$sp]} in the input.
#
sub list_nn_gen_split_candidates (\@$$) {

	my ($to_be_split, $indx_dist, $indx_indx_clsst_pc) = @_;

	my $ct_2Bsps = @{$to_be_split};

	my %clsst_pc2indices_2Bsp; # = ($indx_closest_rlv_pc => \@indices_to_be_split, ...)

	for (my $sp=0; $sp < $ct_2Bsps; $sp++) { # Preliminary for-loop. #

	    my ($dist, $indx_closest_rlv_pc) = @{$to_be_split->[$sp]}[$indx_dist, $indx_indx_clsst_pc];
	    my $indices_2Bsp = $clsst_pc2indices_2Bsp{$indx_closest_rlv_pc};
	    unless (defined $indices_2Bsp) { $indices_2Bsp = $clsst_pc2indices_2Bsp{$indx_closest_rlv_pc} = []; }
	    push @{$indices_2Bsp}, $sp;

	} # End of the preliminary for-loop.

	my @indices_nn_sp_cands = ();

	foreach my $clsst_pc (sort {$a <=> $b} keys %clsst_pc2indices_2Bsp) { # MAIN foreach-loop. #

	    my $indices_2Bsp = $clsst_pc2indices_2Bsp{$clsst_pc};
	    if (@{$indices_2Bsp} <2) {
		push @indices_nn_sp_cands, @{$indices_2Bsp};
		next;
	    }

	    my $min_dist;
	    foreach my $sp (@{$indices_2Bsp}) { # 1st inner foreach-loop. #
	        my ($dist, $indx_closest_rlv_pc) = @{$to_be_split->[$sp]}[$indx_dist, $indx_indx_clsst_pc];
		unless ((defined $min_dist) and ($min_dist <= $dist)) { $min_dist = $dist; }
	    } # End of the 1st inner foreach-loop. #

	    if ($min_dist < 0) { $min_dist = 0; }

	    foreach my $sp (@{$indices_2Bsp}) { # 2nd inner foreach-loop. #
	        my ($dist, $indx_closest_rlv_pc) = @{$to_be_split->[$sp]}[$indx_dist, $indx_indx_clsst_pc];
		if ($dist <= $min_dist) { push @indices_nn_sp_cands, $sp; }
	    } # End of the 2nd inner foreach-loop. #

	} # End of the MAIN foreach-loop. #

	my @list_nn_split_cands = ();

	foreach my $sp (sort {$a <=> $b} @indices_nn_sp_cands) { # Finishing foreach-loop. #

	    push @list_nn_split_cands, $to_be_split->[$sp];
	    
	} # End of the finishing foreach-loop. #

	return \@list_nn_split_cands ;

}	# End of the subroutine, 'list_nn_gen_split_candidates'. #


# list_vsplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	that are to be "vertically split" into two blocks that are effective children of the original block.
#	It was created by taking advantage of some parts of 'list_csplit_candidates' above.
#	[CREATED on 2019/08/21.]
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq, \@set_all_exnodes,
#	\%nodeid2msa_row, \%set_msa_rows2br_up_or_lw,
#	\@curr_bds_gblocks, \@inter_block_relations, 
#	\@list_pcands, \@indices_pc, $ct_seqs, $thrsh_dist_pc), where
#
#		# Necessary for the satellite subroutine, 'vsplit_into_eff_child_gblocks' (in "MyTreeMap_sppl.pm"). #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	@set_all_exnodes stores the indices of all external nodes;
#	
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
#	[Typcally, \%set_msa_rows2br_up_or_lw is an output of 'mk_set_msa_rows2br_up_or_lw_dp' above];
#		# End of "Necessary for the satellite subroutine, 'vsplit_into_eff_child_gblocks' (in "MyTreeMap_sppl.pm")." #
# 
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	@{$list_pcands[$pc]}[@indices_pc] = ($beg_pc, $end_pc, \@rows_under_br, \@constituent_sites)
#	provide information on the $pc th purge-candidate region, where
#	$beg/end_pc is the index of the column at the beginning/end of the region,
#	@rows_under_br stores the indices of the MSA rows under the branch delimiting the $pc th purge-candidate,
#	@constituent_sites stores the indices of the sites constituting the purge-candidate region;
#	[Typically, \@list_pcands is the main output of 'readin_list_mgd_pcands' (in "MyANEX_IO.pm")];
#
# OUTPUTS: (\@to_be_vsplit), where
#
#	@{$to_be_vsplit[$vs]} = ($indx_gb, $dist, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@info_effch1, \@info_effch2)
#	provides informaiton on the $vs th gap-block to be "vertically split" into two blocks that are its effective children,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	@info_effch1/2 = ($br_ch, $up_or_lw_ch, \@exnodes_affd_by_ch, \@rows_affd_by_ch)
#	provides information on the 1st/2nd "effective child" of $gb_sbj, where
#	the child is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@exnodes_affd_by_ch stores the indices of external nodes affected by the child,
#	and @rows_affd_by_ch stores the indices of the MSA rows affected by the child.
#
sub list_vsplit_candidates (\@$$\%\%$\%\@\%\%\@\@\@\@$$) {

	my ($list_gblocks, $indx_br, $indx_u_or_d,
	    $node2pa, $node2ch, $topnode_id, $br2eq, $set_all_exnodes,
	    $nodeid2msa_row, $set_msa_rows2br_up_or_lw,
	    $curr_bds_gblocks, $inter_block_relations, 
	    $list_pcands, $indices_pc, $ct_seqs, $thrsh_dist_pc) = @_;

	my $ct_gbs = @{$list_gblocks};
	my $ct_pcs = @{$list_pcands};

	my @to_be_vsplit = ();

	for (my $gb=0; $gb < $ct_gbs; $gb++) { # Main outer for-loop (over gap-blocks). #

	    my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks->[$gb]}[$indx_br, $indx_u_or_d]; # A BUG fixed on 2019/09/02. #
	    ##my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks->[$gb]};

	    if ($up_or_lw_sbj eq 'L') {
		my $children = $node2ch->{$br_sbj};
		unless ((defined $children) and (@{$children}>0)) { next; } # Immediately SKIP if the gap-block explicitly affects only a single external node. #
	    }

	    my ($set_eff_child_gblks) = vsplit_into_eff_child_gblocks ($br_sbj, $up_or_lw_sbj,
				%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, @{$set_all_exnodes},
				%{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw}) ;
	    unless (defined $set_eff_child_gblks) {
		print STDOUT "[ERROR 'list_vsplit_candidates'] The subroutine 'vsplit_into_eff_child_gblocks' FAILED on (\$br_sbj, \$up_or_lw_sbj) = ($br_sbj, $up_or_lw_sbj): $! !\n\n";
		return undef;
	    }

	    unless (@{$set_eff_child_gblks} == 2) { next; } # Skip if the gap-block will NOT be split into TWO.

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_gblocks->[$gb]};

	    my @indices_rlv_pcs = ();
	    my @closest_pc = (); # = ($indx_pc, $dist, $ct_cnst_sites)

	  foreach my $info_effch (@{$set_eff_child_gblks}) { # Middle for-loop (over effective children). #

	     my ($br_ch, $up_or_lw_ch, $exnodes_affd_by_ch, $rows_affd_by_ch) = @{$info_effch};
	     my $ct_rows_affd_by_ch = @{$rows_affd_by_ch};

	      for (my $pc=0; $pc <$ct_pcs; $pc++) { # Inner for-loop (over purge-candidate regions). #

		my ($beg_pc, $end_pc, $rows_under_br, $constituent_sites) = @{$list_pcands->[$pc]}[@{$indices_pc}];
		my $ct_rows_ud_br = @{$rows_under_br};
		unless (($ct_rows_affd_by_ch == $ct_rows_ud_br) or ($ct_rows_affd_by_ch + $ct_rows_ud_br == $ct_seqs)) {
		    next;
		}
		my ($common, $ex_ch, $ex_pc) = diff_nmarray (@{$rows_affd_by_ch}, @{$rows_under_br});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'diff_nmarray' FAILED between the $gb th gap-block and the $pc th purege-candidate region: $! !\n\n";
		    return undef;
		}

		my $if_relv =0;
		if ((@{$ex_ch} == 0) and (@{$ex_pc} == 0)) {
		    $if_relv = 1; # The two sets of rows are identical. #

		} elsif ((@{$common} == 0) and ($ct_rows_affd_by_ch + $ct_rows_ud_br == $ct_seqs)) {
		    $if_relv = 1; # The two sets of rows are complementary. #
		}
		unless ($if_relv) { next; }


			## Compute the distance from the purge-candidate. ##

		my $dist;

			# NOTE: Although it is better to compute the distance between the effective child and the purge-candidate, to do so is actually quite cumbersome.
			# Thus, we settle for computing the distance between $gb and the purge-candidate.
			# (Hopefully in some future, we will attempt to actually compute the distance from the effective child.)

		if ($rbd_gb < $beg_pc) {
		    $dist = compt_distance_bw_gblk_and_pt ($beg_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $beg_pc, which is the left-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }

		} elsif ($end_pc < $lbd_gb) {
		    $dist = compt_distance_bw_gblk_and_pt ($end_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $end_pc, which is the right-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }
		} else {
		    $dist = -1;
		}

		if ($thrsh_dist_pc < $dist) { next; }

		push @indices_rlv_pcs, $pc;

		my $ct_cnst_sites = @{$constituent_sites};

		if (@closest_pc > 0) {
		    if ($dist > $closest_pc[1]) {
			next;
		    } elsif ($dist == $closest_pc[1]) {
			if ($ct_cnst_sites <= $closest_pc[2]) { next; }
		    }
		}

		@closest_pc = ($pc, $dist, $ct_cnst_sites);

	      } # End of the inner for-loop (over purge-candidate regions). #
	  } # End of the Middle for-loop (over effective children). #

	    unless (@indices_rlv_pcs>0) { next; } # Skip if no purge-candidates share the delimiting branch with $gb. #

	    my ($indx_closest_rlv_pc, $dist_closest, $ct_cnst_sites_closest) = @closest_pc;

	    my ($info_effch1, $info_effch2) = @{$set_eff_child_gblks};

	    push @to_be_vsplit, [$gb, $dist_closest, \@indices_rlv_pcs, $indx_closest_rlv_pc, $info_effch1, $info_effch2];

	} # End of the Main outer for-loop (over gap-blocks). #

	return \@to_be_vsplit;

}	# End of the subroutine, 'list_vsplit_candidates'. #


# list_cvsplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	each of whose complementary residue-block is to be "vertically split" into two residue-blocks that are effective children of the original residue-block.
#	It was created by modifying 'list_csplit_candidates' above.
#	[CREATED on 2019/08/25.]
#
# list_vsplit_candidates:
#	This subroutine creates a list of candidate gap-blocks
#	that are to be "vertically split" into two blocks that are effective children of the original block.
#	It was created by taking advantage of some parts of 'list_csplit_candidates' above.
#	[CREATED on 2019/08/21.]
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq, \@set_all_exnodes,
#	\%nodeid2msa_row, \%set_msa_rows2br_up_or_lw,
#	\@curr_bds_gblocks, \@inter_block_relations, 
#	\@list_pcands, \@indices_pc, $ct_seqs, $thrsh_dist_pc), where
#
#		# Necessary for the satellite subroutine, 'vsplit_into_eff_child_gblocks' (in "MyTreeMap_sppl.pm"). #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	@set_all_exnodes stores the indices of all external nodes;
#	
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
#	[Typcally, \%set_msa_rows2br_up_or_lw is an output of 'mk_set_msa_rows2br_up_or_lw_dp' above];
#		# End of "Necessary for the satellite subroutine, 'vsplit_into_eff_child_gblocks' (in "MyTreeMap_sppl.pm")." #
# 
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	@{$list_pcands[$pc]}[@indices_pc] = ($beg_pc, $end_pc, \@rows_under_br, \@constituent_sites)
#	provide information on the $pc th purge-candidate region, where
#	$beg/end_pc is the index of the column at the beginning/end of the region,
#	@rows_under_br stores the indices of the MSA rows under the branch delimiting the $pc th purge-candidate,
#	@constituent_sites stores the indices of the sites constituting the purge-candidate region;
#	[Typically, \@list_pcands is the main output of 'readin_list_mgd_pcands' (in "MyANEX_IO.pm")];
#
# [ OUTPUTS: (\@to_be_vsplit), where ] ... for the predecessor, 'list_vsplit_candidates'.
# OUTPUTS: (\@to_be_cvsplit), where
#
#	@{$to_be_cvsplit[$cvs]} = ($indx_gb, $dist, \@indices_rlv_pcs, $indx_closest_rlv_pc, \@info_ceffch1, \@info_ceffch2)
#	provides informaiton on the $cvs th gap-block whose complementary residue-block will be "vertically split" into two residue-blocks that are its effective children,
#	where
#	$dist is the distance between the gap-block and the nearest (and relevant) purge-candidate region,
#	@indices_rlv_pcs stores the indices of the relevant purge-candidate regions,
#	$indx_closest_rlv_pc is the index of the closest relevant purge-candidate region,
#	@info_ceffch1/2 = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)
#	provides information on the 1st/2nd "complementary-effective child" of $gb_sbj, where
#	the "complementary-child" residue-block's complementary GAP-block is on the $up_or_lw_cch side of the branch with ID $br_cch ,
#	@cexnodes_affd_by_cch stores the indices of external nodes affected by the complementary GAP-block of the "complementary-child" residue-block,
#	and @crows_affd_by_cch stores the indices of the MSA rows affected by the complementary GAP-block of the "complementary-child" residue-block.
#
# [ sub list_vsplit_candidates (\@$$\%\%$\%\@\%\%\@\@\@\@$$) { ] ... for the predecessor, 'list_vsplit_candidates'.
#
sub list_cvsplit_candidates (\@$$\%\%$\%\@\%\%\@\@\@\@$$) {

	my ($list_gblocks, $indx_br, $indx_u_or_d,
	    $node2pa, $node2ch, $topnode_id, $br2eq, $set_all_exnodes,
	    $nodeid2msa_row, $set_msa_rows2br_up_or_lw,
	    $curr_bds_gblocks, $inter_block_relations, 
	    $list_pcands, $indices_pc, $ct_seqs, $thrsh_dist_pc) = @_;

	my $ct_gbs = @{$list_gblocks};
	my $ct_pcs = @{$list_pcands};

	my @to_be_cvsplit = ();

	for (my $gb=0; $gb < $ct_gbs; $gb++) { # Main outer for-loop (over gap-blocks). #

	    my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks->[$gb]}[$indx_br, $indx_u_or_d]; # A BUG Fixed on 2019/09/02. #
	    ##my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks->[$gb]};

	    if ($up_or_lw_sbj eq 'U') {
		my $children = $node2ch->{$br_sbj};
		unless ((defined $children) and (@{$children}>0)) { next; } # Immediately SKIP if the residue-block complementary to $gb  explicitly affects only a single external node. #
	    }

	    my ($set_cmpl_eff_child_rblks) = vsplit_into_cmpl_eff_child_rblocks ($br_sbj, $up_or_lw_sbj,
				%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, @{$set_all_exnodes},
				%{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw}) ;
	    unless (defined $set_cmpl_eff_child_rblks) {
		print STDOUT "[ERROR 'list_cvsplit_candidates'] The subroutine 'vsplit_into_cmpl_eff_child_gblocks' FAILED on (\$br_sbj, \$up_or_lw_sbj) = ($br_sbj, $up_or_lw_sbj): $! !\n\n";
		return undef;
	    }

	    unless (@{$set_cmpl_eff_child_rblks} == 2) { next; } # Skip if the residue-block complementary to $gb will NOT be split into TWO.

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_gblocks->[$gb]};

	    my @indices_rlv_pcs = ();
	    my @closest_pc = (); # = ($indx_pc, $dist, $ct_cnst_sites)

	  foreach my $info_ceffch (@{$set_cmpl_eff_child_rblks}) { # Middle for-loop (over "complementary-effective children"). #

	     my ($br_cch, $up_or_lw_cch, $cexnodes_affd_by_cch, $crows_affd_by_cch) = @{$info_ceffch};
	     my $ct_crows_affd_by_cch = @{$crows_affd_by_cch};

	      for (my $pc=0; $pc <$ct_pcs; $pc++) { # Inner for-loop (over purge-candidate regions). #

		my ($beg_pc, $end_pc, $rows_under_br, $constituent_sites) = @{$list_pcands->[$pc]}[@{$indices_pc}];
		my $ct_rows_ud_br = @{$rows_under_br};
		unless (($ct_crows_affd_by_cch == $ct_rows_ud_br) or ($ct_crows_affd_by_cch + $ct_rows_ud_br == $ct_seqs)) {
		    next;
		}
		my ($common, $ex_cch, $ex_pc) = diff_nmarray (@{$crows_affd_by_cch}, @{$rows_under_br});
		unless (defined $common) {
		    print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'diff_nmarray' FAILED between the $gb th gap-block and the $pc th purege-candidate region: $! !\n\n";
		    return undef;
		}

		my $if_relv =0;
		if ((@{$ex_cch} == 0) and (@{$ex_pc} == 0)) {
		    $if_relv = 1; # The two sets of rows are identical. #

		} elsif ((@{$common} == 0) and ($ct_crows_affd_by_cch + $ct_rows_ud_br == $ct_seqs)) {
		    $if_relv = 1; # The two sets of rows are complementary. #
		}
		unless ($if_relv) { next; }


			## Compute the distance from the purge-candidate. ##

		my $dist;

			# NOTE: Although it is better to compute the distance between the effective child and the purge-candidate, to do so is actually quite cumbersome.
			# Thus, we settle for computing the distance between $gb and the purge-candidate.
			# (Hopefully in some future, we will attempt to actually compute the distance from the effective child.)

		if ($rbd_gb < $beg_pc) {
		    $dist = compt_distance_bw_gblk_and_pt ($beg_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $beg_pc, which is the left-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }

		} elsif ($end_pc < $lbd_gb) {
		    $dist = compt_distance_bw_gblk_and_pt ($end_pc, $gb, @{$curr_bds_gblocks}, @{$inter_block_relations});
		    unless (defined $dist) {
			print STDOUT "[ERROR 'list_csplit_candidates'] Subroutine 'compt_distance_bw_gblk_and_pt' FAILED between the $gb th gap-block and the point $end_pc, which is the right-end of the $pc th purege-candidate region: $! !\n\n";
			return undef;
		    }
		} else {
		    $dist = -1;
		}

		if ($thrsh_dist_pc < $dist) { next; }

		push @indices_rlv_pcs, $pc;

		my $ct_cnst_sites = @{$constituent_sites};

		if (@closest_pc > 0) {
		    if ($dist > $closest_pc[1]) {
			next;
		    } elsif ($dist == $closest_pc[1]) {
			if ($ct_cnst_sites <= $closest_pc[2]) { next; }
		    }
		}

		@closest_pc = ($pc, $dist, $ct_cnst_sites);

	      } # End of the inner for-loop (over purge-candidate regions). #
	  } # End of the Middle for-loop (over effective children). #

	    unless (@indices_rlv_pcs>0) { next; } # Skip if no purge-candidates share the delimiting branch with $gb. #

	    my ($indx_closest_rlv_pc, $dist_closest, $ct_cnst_sites_closest) = @closest_pc;

	    my ($info_ceffch1, $info_ceffch2) = @{$set_cmpl_eff_child_rblks};

	    push @to_be_cvsplit, [$gb, $dist_closest, \@indices_rlv_pcs, $indx_closest_rlv_pc, $info_ceffch1, $info_ceffch2];

	} # End of the Main outer for-loop (over gap-blocks). #

	return \@to_be_cvsplit;

}	# End of the subroutine, 'list_cvsplit_candidates'. #


# list_nn_vsplit_plus_smerge_candidates:
#	This subroutine creates a list of "nearest-neighboring" candidates of
#	effective parent-child pairs to be "vertically split + (same-type) merged".
#	It was created by taking advantage of the framework of 'list_nn_cvmerge_candidates'.
#	[CREATED on 2019/09/02.]
#	
# list_nn_cvmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	complementary-sibling gap-blocks to be "(incomplete) vertical-merge"d.
#	It was created by taking advantage of the framework of 'list_nn_purge_and_cmerge_candidates'.
#	[CREATED on 2019/07/23.]
#	[REVISED on 2019/07/27]:
#	Now, when identifying "nearest-neighboring" pairs,
#	this subroutine only compares gap-blocks that are vertically identical to one another.
#
# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_vssm), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_vssm is the threshold for the distance between an "effective parent-child" pair of gap-blocks to be "vertically split + (horizontally) same-type merge"d;
#
# OUTPUTS: (\@to_be_vsplit_plus_smerged), where
#
#	@{$to_be_vsplit_plus_smerged[$vsm]} = ($gb1, $gb2, $dist) provides information on the $vms th candidate for the "nearest-neighboring" effective parent-chind pair to be "vertically split + (same-type) merged",
#	$gb1 and $gb2 are the indices of the gap-blocks that are the effective parent and effective child, respectively, to each other,
#	$dist is the distance of $gb2 from $gb1;
#
sub list_nn_vsplit_plus_smerge_candidates (\@\@\@$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_vssm) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' effective parent-child pairs for each $gb1. ###

	my %gb2_2vssm_cands;

	for (my $gb1 =0; $gb1 < $ct_gblks-1; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

	    my @vssm_cands = ();
	    my $mindist_vssmc;

	    for (my $gb2 = $gb1+1; $gb2 < $ct_gblks; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq '>(ch)') { next; }
		##unless ($rel eq 'ONCS') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_vsplit_plus_smerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef);
		}

		my $dist = $dist2;

		if (1) {
		    if ($dist > $thrsh_dist_vssm) { next; }


		    my $rels_w_gb2 = $inter_block_relations->[$gb2];

		    my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		    foreach my $pair (@vssm_cands) {
			my ($gb1c, $gb3, $dist3) = @{$pair};
			unless ($rels_w_gb2->[$gb3] eq '=') { next; }
			if ($dist < $dist3) {
			   $result_cmp = -1;
			} elsif ($dist > $dist3) {
			   $result_cmp = +1;
			} 
			last;
		    }

		    if ($result_cmp >0) {

			next; # Exclude the current pair. #

		    } elsif ($result_cmp < 0) {

			    # Remove the pairs of the same type. # 

			my @new_vssm_cands = ();
			foreach my $pair (@vssm_cands) {
			    my ($gb1c, $gb3, $dist3) = @{$pair};
			    if ($rels_w_gb2->[$gb3] eq '=') { next; }
			    push @new_vssm_cands, $pair;
			}
			@vssm_cands = @new_vssm_cands;
		    }

		    push @vssm_cands, [$gb1, $gb2, $dist];

		} # End of "if (1) {...}"

	    } # End of the inner for-loop (over $gb2). #


	    foreach my $pair (@vssm_cands) {
		my $gb2 = $pair->[1];
		my $vssm_cands = $gb2_2vssm_cands{$gb2};
		unless (defined $vssm_cands) {$vssm_cands = $gb2_2vssm_cands{$gb2} = []; }
		push @{$vssm_cands}, $pair;
	    }

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

	my @to_be_vsplit_plus_smerged = ();
	foreach my $gb2 (keys %gb2_2vssm_cands) { # Trailing foreach-loop (over $gb2). #

	    my $vssm_cands = $gb2_2vssm_cands{$gb2};
	    my $ct_vssm_cands = @{$vssm_cands};
	    if ($ct_vssm_cands < 2) {
		push @to_be_vsplit_plus_smerged, @{$vssm_cands};
		next;
	    }

	    my @nn_vssm_cands = ($vssm_cands->[0]);

	    for (my $i=1; $i <$ct_vssm_cands; $i++) {
		my ($gb1, $gb2c, $dist) = @{$vssm_cands->[$i]};

		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		foreach my $pair (@nn_vssm_cands) {
		    my ($gb1c, $gb2c, $distc) = @{$pair}; 
		    unless ($rels_w_gb1->[$gb1c] eq '=') { next; }
		    if ($dist < $distc) {
			$result_cmp = -1;
		    } elsif ($dist > $distc) {
			$result_cmp = +1;
		    }
		    last;
		} # End of "foreach my $pair (@nn_vssm_cands) {...}"

		if ($result_cmp > 0) {

		    next; # Exclude the current pair. #

		} elsif ($result_cmp < 0) {
			# Remove the pairs of the same type. #
		    my @new_nn_vssm_cands = ();
		    foreach my $pair (@nn_vssm_cands) {
			my ($gb1c, $gb2cc, $distc) = @{$pair}; 
			if ($rels_w_gb1->[$gb1c] eq '=') { next; }
			push @new_nn_vssm_cands, $pair;
		    }
		    @nn_vssm_cands = @new_nn_vssm_cands;
		}

		push @nn_vssm_cands, $vssm_cands->[$i];

	    } # End of "for (my $i=1; $i <$ct_vssm_cands; $i++) {...}"

	    push @to_be_vsplit_plus_smerged, @nn_vssm_cands;

	} # End of the trailing foreach-loop (over $gb2). #


		### (3) Finally, sort the candidates in each of the sets. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

	my $std_to_be_vsplit_plus_smerged = multiple_sort (@to_be_vsplit_plus_smerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_vsplit_plus_smerged) {
	    print STDOUT "[ERROR 'list_nn_vsplit_plus_smerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of nearest-neighboring candidates of effective parent-child pairs to be \"vertically split + (same-type) merge\"d: $! !\n\n";
	    return (undef, undef);
	}

	return ($std_to_be_vsplit_plus_smerged);

}	# End of the subroutine, 'list_nn_vsplit_plus_smerge_candidates'. #


# list_nn_cvsplit_plus_smerge_candidates:
#	This subroutine creates a list of "nearest-neighboring" candidates of
#	effective child-parent pairs to be "complementarily vertically split + (same-type) merged".
#	It was created by modifying 'list_nn_vsplit_plus_smerge_candidates'.
#	[CREATED on 2019/09/10.]
#
# list_nn_vsplit_plus_smerge_candidates:
#	This subroutine creates a list of "nearest-neighboring" candidates of
#	effective parent-child pairs to be "vertically split + (same-type) merged".
#	It was created by taking advantage of the framework of 'list_nn_cvmerge_candidates'.
#	[CREATED on 2019/09/02.]
#	
# list_nn_cvmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	complementary-sibling gap-blocks to be "(incomplete) vertical-merge"d.
#	It was created by taking advantage of the framework of 'list_nn_purge_and_cmerge_candidates'.
#	[CREATED on 2019/07/23.]
#	[REVISED on 2019/07/27]:
#	Now, when identifying "nearest-neighboring" pairs,
#	this subroutine only compares gap-blocks that are vertically identical to one another.
#
# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_cvssm), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_cvssm is the threshold for the distance between an "effective child-parent" pair of gap-blocks to be "complementarily vertically split + (horizontally) same-type merge"d;
#
# OUTPUTS: (\@to_be_cvsplit_plus_smerged), where
#
#	@{$to_be_cvsplit_plus_smerged[$vsm]} = ($gb1, $gb2, $dist) provides information on the $vms th candidate for the "nearest-neighboring" effective child-parent pair to be "complementarily vertically split + (same-type) merged",
#	$gb1 and $gb2 are the indices of the gap-blocks that are the effective child and effective parent, respectively, to each other,
#	$dist is the distance of $gb2 from $gb1;
#
# [ sub list_nn_vsplit_plus_smerge_candidates (\@\@\@$) { ] ... for the predecessor, 'list_nn_vsplit_plus_smerge_candidates'.
#
sub list_nn_cvsplit_plus_smerge_candidates (\@\@\@$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_cvssm) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' effective parent-child pairs for each $gb1. ###

	my %gb2_2cvssm_cands;

	for (my $gb1 =1; $gb1 < $ct_gblks; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

	    my @cvssm_cands = ();
	    my $mindist_cvssmc;

	    for (my $gb2 = 0; $gb2 < $gb1; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq '<(pa)') { next; }
		##unless ($rel eq '>(ch)') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_cvsplit_plus_smerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef);
		}

		my $dist = $dist2;

		if (1) {
		    if ($dist > $thrsh_dist_cvssm) { next; }


		    my $rels_w_gb2 = $inter_block_relations->[$gb2];

		    my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		    foreach my $pair (@cvssm_cands) {
			my ($gb1c, $gb3, $dist3) = @{$pair};
			unless ($rels_w_gb2->[$gb3] eq '=') { next; }
			if ($dist < $dist3) {
			   $result_cmp = -1;
			} elsif ($dist > $dist3) {
			   $result_cmp = +1;
			} 
			last;
		    }

		    if ($result_cmp >0) {

			next; # Exclude the current pair. #

		    } elsif ($result_cmp < 0) {

			    # Remove the pairs of the same type. # 

			my @new_cvssm_cands = ();
			foreach my $pair (@cvssm_cands) {
			    my ($gb1c, $gb3, $dist3) = @{$pair};
			    if ($rels_w_gb2->[$gb3] eq '=') { next; }
			    push @new_cvssm_cands, $pair;
			}
			@cvssm_cands = @new_cvssm_cands;
		    }

		    push @cvssm_cands, [$gb1, $gb2, $dist];

		} # End of "if (1) {...}"

	    } # End of the inner for-loop (over $gb2). #


	    foreach my $pair (@cvssm_cands) {
		my $gb2 = $pair->[1];
		my $cvssm_cands = $gb2_2cvssm_cands{$gb2};
		unless (defined $cvssm_cands) {$cvssm_cands = $gb2_2cvssm_cands{$gb2} = []; }
		push @{$cvssm_cands}, $pair;
	    }

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

	my @to_be_cvsplit_plus_smerged = ();
	foreach my $gb2 (keys %gb2_2cvssm_cands) { # Trailing foreach-loop (over $gb2). #

	    my $cvssm_cands = $gb2_2cvssm_cands{$gb2};
	    my $ct_cvssm_cands = @{$cvssm_cands};
	    if ($ct_cvssm_cands < 2) {
		push @to_be_cvsplit_plus_smerged, @{$cvssm_cands};
		next;
	    }

	    my @nn_cvssm_cands = ($cvssm_cands->[0]);

	    for (my $i=1; $i <$ct_cvssm_cands; $i++) {
		my ($gb1, $gb2c, $dist) = @{$cvssm_cands->[$i]};

		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		foreach my $pair (@nn_cvssm_cands) {
		    my ($gb1c, $gb2c, $distc) = @{$pair}; 
		    unless ($rels_w_gb1->[$gb1c] eq '=') { next; }
		    if ($dist < $distc) {
			$result_cmp = -1;
		    } elsif ($dist > $distc) {
			$result_cmp = +1;
		    }
		    last;
		} # End of "foreach my $pair (@nn_cvssm_cands) {...}"

		if ($result_cmp > 0) {

		    next; # Exclude the current pair. #

		} elsif ($result_cmp < 0) {
			# Remove the pairs of the same type. #
		    my @new_nn_cvssm_cands = ();
		    foreach my $pair (@nn_cvssm_cands) {
			my ($gb1c, $gb2cc, $distc) = @{$pair}; 
			if ($rels_w_gb1->[$gb1c] eq '=') { next; }
			push @new_nn_cvssm_cands, $pair;
		    }
		    @nn_cvssm_cands = @new_nn_cvssm_cands;
		}

		push @nn_cvssm_cands, $cvssm_cands->[$i];

	    } # End of "for (my $i=1; $i <$ct_cvssm_cands; $i++) {...}"

	    push @to_be_cvsplit_plus_smerged, @nn_cvssm_cands;

	} # End of the trailing foreach-loop (over $gb2). #


		### (3) Finally, sort the candidates in each of the sets. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

	my $std_to_be_cvsplit_plus_smerged = multiple_sort (@to_be_cvsplit_plus_smerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_cvsplit_plus_smerged) {
	    print STDOUT "[ERROR 'list_nn_cvsplit_plus_smerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of nearest-neighboring candidates of effective child-parent pairs to be \"complementarily vertically split + (same-type) merge\"d: $! !\n\n";
	    return (undef, undef);
	}

	return ($std_to_be_cvsplit_plus_smerged);

}	# End of the subroutine, 'list_nn_cvsplit_plus_smerge_candidates'. #


# list_nn_ssplit_plus_vmerge_candidates:
#	This subroutine creates a list of "nearest-neighboring" candidates of
#	effective sibling pairs to be "(horizontally) split (into same-type blocks) + vertically merge"d.
#	It was created by modifying 'list_nn_cvsplit_plus_smerge_candidates'.
#	[CREATED on 2019/10/05.]
#
# list_nn_cvsplit_plus_smerge_candidates:
#	This subroutine creates a list of "nearest-neighboring" candidates of
#	effective child-parent pairs to be "complementarily vertically split + (same-type) merged".
#	It was created by modifying 'list_nn_vsplit_plus_smerge_candidates'.
#	[CREATED on 2019/09/10.]
#
#	... OMITTED ...
#
# list_nn_purge_and_cmerge_candidates:
#	This subroutine creates a list of candidate pairs of "nearest-neighboring"
#	gap-blocks of complementary types to be "purge"d and/ or to be "merge"d.
#	[CREATED on 2019/07/11.]
#
#	NOTE: This subroutine attempts to find pairs of "nearest-neighboring" gap-blocks in each category in the following manner:
#	(i) For each "higher-ranked" gap-block, pick only the complementary one closest to it;
#	(if there are two closest complementary gap-blocks, keep both ones for the moment);
#	(iii) If some pairs share the same complementary gap-block, choose the pair with the smallest distance;
#	(if there are two pairs with the smallest distance, keep both ones).
#
#	(NOTE that this selection scheme could leave some gap-blocks UNpaired with ANY of their complementary ones, even if they exist.)
#
#
# INPUTS: (\@bds_gblocks0, \@gblock_sizes, \@inter_block_relations, $thrsh_dist_ssvm), where
#
#	@{$bds_gblocks0[$gb]} = ($lbd0_gb, $rbd0_gb) gives the INITIAL positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	[[
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... SET ASIDE for later use.
#
#	$thrsh_dist_ssvm is the threshold for the distance between an "effective child-parent" pair of gap-blocks to be "complementarily vertically split + (horizontally) same-type merge"d;
#
# OUTPUTS: (\@to_be_ssplit_plus_vmerged), where
#
#	@{$to_be_ssplit_plus_vmerged[$ssvm]} = ($gb1, $gb2, $dist) provides information on the $ssvm th candidate for the "nearest-neighboring" effective sibling pair to be "(horizontally) (same-type) split + vertically merg"ed,
#	$gb1 and $gb2 are the indices of the gap-blocks that are effective siblings to each other,
#	$dist is the distance of $gb2 from $gb1;
#
# [ sub list_nn_cvsplit_plus_smerge_candidates (\@\@\@$) { ] ... for the predecessor, 'list_nn_cvsplit_plus_smerge_candidates'.
#
sub list_nn_ssplit_plus_vmerge_candidates (\@\@\@$) {

	my ($bds_gblocks0, $gblock_sizes, $inter_block_relations, $thrsh_dist_ssvm) = @_;

	my $ct_gblks = @{$inter_block_relations};


		### (1) Select the 'closest' effective sibling pairs for each $gb1. ###

	my %gb2_2ssvm_cands;

	for (my $gb1 =1; $gb1 < $ct_gblks; $gb1++) { # Main outer for-loop (over $gb1). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my $size_gb1 = $gblock_sizes->[$gb1];

	    my @ssvm_cands = ();
	    my $mindist_ssvmc;

	    for (my $gb2 = 0; $gb2 < $gb1; $gb2++) { # Inner for-loop (over $gb2). #

		my $rel = $rels_w_gb1->[$gb2];
		unless ($rel eq 'S') { next; } # The MAJOR REVISION in creation of 'list_nn_ssplit_plus_vmerge_candidates' (on 2019/10/05). #
		##unless ($rel eq '<(pa)') { next; }
		####unless ($rel eq '>(ch)') { next; }

		my $size_gb2 = $gblock_sizes->[$gb2];

		if ($size_gb1 == $size_gb2) { next; } # Skip if $gb1 and $gb2 have the same size. # (ADDED in creation of 'list_nn_ssplit_plus_vmerge_candidates' (on 2019/10/05).) #

		my ($dist1, $dist2) = compt_inter_gblock_distance ($gb1, $gb2, @{$bds_gblocks0}, @{$inter_block_relations});
		unless (defined $dist2) {
		    print STDOUT "[ERROR 'list_nn_ssplit_plus_vmerge_candidates'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, ($gb1, $gb2): $! !\n\n";
		    return (undef);
		}

		my $dist = $dist2;

		if (1) {
		    if ($dist > $thrsh_dist_ssvm) { next; }


		    my $rels_w_gb2 = $inter_block_relations->[$gb2];

		    my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		    foreach my $pair (@ssvm_cands) {
			my ($gb1c, $gb3, $dist3) = @{$pair};
			unless ($rels_w_gb2->[$gb3] eq '=') { next; }
			if ($dist < $dist3) {
			   $result_cmp = -1;
			} elsif ($dist > $dist3) {
			   $result_cmp = +1;
			} 
			last;
		    }

		    if ($result_cmp >0) {

			next; # Exclude the current pair. #

		    } elsif ($result_cmp < 0) {

			    # Remove the pairs of the same type. # 

			my @new_ssvm_cands = ();
			foreach my $pair (@ssvm_cands) {
			    my ($gb1c, $gb3, $dist3) = @{$pair};
			    if ($rels_w_gb2->[$gb3] eq '=') { next; }
			    push @new_ssvm_cands, $pair;
			}
			@ssvm_cands = @new_ssvm_cands;
		    }

		    push @ssvm_cands, [$gb1, $gb2, $dist];

		} # End of "if (1) {...}"

	    } # End of the inner for-loop (over $gb2). #


	    foreach my $pair (@ssvm_cands) {
		my $gb2 = $pair->[1];
		my $ssvm_cands = $gb2_2ssvm_cands{$gb2};
		unless (defined $ssvm_cands) {$ssvm_cands = $gb2_2ssvm_cands{$gb2} = []; }
		push @{$ssvm_cands}, $pair;
	    }

	} # End of the Main outer for-loop (over $gb1). #


		### (2) From the selected pairs, choose the 'closest' pairs for each $gb2. ###

	my @to_be_ssplit_plus_vmerged = ();
	foreach my $gb2 (keys %gb2_2ssvm_cands) { # Trailing foreach-loop (over $gb2). #

	    my $ssvm_cands = $gb2_2ssvm_cands{$gb2};
	    my $ct_ssvm_cands = @{$ssvm_cands};
	    if ($ct_ssvm_cands < 2) {
		push @to_be_ssplit_plus_vmerged, @{$ssvm_cands};
		next;
	    }

	    my @nn_ssvm_cands = ($ssvm_cands->[0]);

	    for (my $i=1; $i <$ct_ssvm_cands; $i++) {
		my ($gb1, $gb2c, $dist) = @{$ssvm_cands->[$i]};

		my $rels_w_gb1 = $inter_block_relations->[$gb1];

		my $result_cmp = 0; # = -1 if $dist is smaller than the distance stored, = +1 if it is larger, = 0 if it equals (or there were NO pairs of the same type).
		foreach my $pair (@nn_ssvm_cands) {
		    my ($gb1c, $gb2c, $distc) = @{$pair}; 
		    unless ($rels_w_gb1->[$gb1c] eq '=') { next; }
		    if ($dist < $distc) {
			$result_cmp = -1;
		    } elsif ($dist > $distc) {
			$result_cmp = +1;
		    }
		    last;
		} # End of "foreach my $pair (@nn_ssvm_cands) {...}"

		if ($result_cmp > 0) {

		    next; # Exclude the current pair. #

		} elsif ($result_cmp < 0) {
			# Remove the pairs of the same type. #
		    my @new_nn_ssvm_cands = ();
		    foreach my $pair (@nn_ssvm_cands) {
			my ($gb1c, $gb2cc, $distc) = @{$pair}; 
			if ($rels_w_gb1->[$gb1c] eq '=') { next; }
			push @new_nn_ssvm_cands, $pair;
		    }
		    @nn_ssvm_cands = @new_nn_ssvm_cands;
		}

		push @nn_ssvm_cands, $ssvm_cands->[$i];

	    } # End of "for (my $i=1; $i <$ct_ssvm_cands; $i++) {...}"

	    push @to_be_ssplit_plus_vmerged, @nn_ssvm_cands;

	} # End of the trailing foreach-loop (over $gb2). #


		### (3) Finally, sort the candidates in each of the sets. ###

	my @indices_srt = (0, 1); my @methods_srt = ('+', '+');

	my $std_to_be_ssplit_plus_vmerged = multiple_sort (@to_be_ssplit_plus_vmerged, @indices_srt, @methods_srt);
	unless (defined $std_to_be_ssplit_plus_vmerged) {
	    print STDOUT "[ERROR 'list_nn_ssplit_plus_vmerge_candidates'] The subroutine 'multiple_sort' FAILED on the list of nearest-neighboring candidates of effective sibling pairs to be \"horizontally  (same-type) split + vertically merge\"d: $! !\n\n";
	    return (undef, undef);
	}

	return ($std_to_be_ssplit_plus_vmerged);

}	# End of the subroutine, 'list_nn_ssplit_plus_vmerge_candidates'. #


# list_revpurge_cand_sizes:
#	This subroutine creates a list of sizes of the candidate regions that will be actually "reverse-purge"d.
#	[CREATED on 2019/09/13; DEBUGGED on 2020/05/21]
#
#	RULES:
#	(1) When $xlevel < 2, return  [];
#	(2) When $xlevel == 2, return [1 .. $maxsize_for_xlevel2];
#	(3) When $xlevel == 3, return [1 .. Min(size{sgn_wd}];
#	(3.9) When 3.5 <= $xlevel < 5, return [1 .. Min(size(sgn_wd)-1, {sizes of sgn_wds}];
#	(5) When $xlevel == 5 (and if mpc is larger than the largest sgn_wd),
#		return [1 .. Min(size(sgn_wd)-1, {sizes of sgn_wds}, {size of mpc}];
#	(6) When $xlevel == 6, return [1 .. {size of mpc}];
#
# INPUTS: (\@sites_in_mpc, \@sgn_wds_included, $indx_sites_in_sgn_wd, $xlevel, $maxsize_for_xlevel2), where
#
#	@sites_in_mpc stores the sites that constitute the subject (merged) "purge"-candidate;
#	@sgn_wds_included lists the significant windows included in the subject (merged) "purge"-candidate;
#	More specifically, @{$sgn_wds_included[$k]->[$indx_sites_in_sgn_wd]}, or @{$sgn_wds_included[$k]} itself if $indx_sites_in_sgn_wd is undefined,  stores the sites that constitute the $k th significant window;
#	[Typically, \@constituent_sites_in_mpc and \@sgn_wds_included are components of each element of @list_mgd_pcands, which is a main output of 'readin_list_mgd_pcands' in "MyANEX_IO.pm"];
#
#	$xlevel is an integer specifying the level of MSA-space exploration;
#
#	$maxsize_for_xlevel2 is the maximum size of the "reverse-purge"d regions attempted when $xlevel = 2;
#
# OUTPUT: (\@rvprg_cand_sizes), where	
#
#	@rvprg_cand_sizes stores the sizes of the candidate regions to be "reverse-purge"d;
#
sub list_revpurge_cand_sizes (\@\@$$$) {

	my ($sites_in_mpc, $sgn_wds_included, $indx_sites_in_sgn_wd, $xlevel, $maxsize_for_xlevel2) = @_;

	unless ((defined $sites_in_mpc) and (@{$sites_in_mpc}>0)) {
	    print STDOUT "[ERROR 'list_revpurge_cand_sizes'] ILLEGAL INPUT: ", '@{$sites_in_mpc}', " is UNDEFINED or EMPTY: $! !\n\n";
	    return undef;
	}

			# MODIFIED in DEBUGGING on 2020/05/21. #
	unless (defined $sgn_wds_included) {
	    print STDOUT "[ERROR 'list_revpurge_cand_sizes'] ILLEGAL INPUT: ", '@{$sgn_wds_included}', " is UNDEFINED: $! !\n\n";
	    return undef;
	}
=A BUG fixed on 2020/05/21.
	unless ((defined $sgn_wds_included) and (@{$sgn_wds_included})) {
	    print STDOUT "[ERROR 'list_revpurge_cand_sizes'] ILLEGAL INPUT: ", '@{$sgn_wds_included}', " is UNDEFINED or EMPTY: $! !\n\n";
	    return undef;
	}
=cut # End of "A BUG fixed on 2020/05/21."
			# End of "MODIFIED in DEBUGGING on 2020/05/21." #

	if ($xlevel < 2) {
	    return []; # When $xlevel < 2, ANEX should perform ONLY Multiple-"Shift"s (from the original input MSA).
	}

	if ($xlevel < 3) {
	    my @rvprg_cand_sizes = (1 .. $maxsize_for_xlevel2);
	    return \@rvprg_cand_sizes ;
	}

	my $size_mpc = @{$sites_in_mpc};

	if ($xlevel >= 6) {
	    my @rvprg_cand_sizes = (1 .. $size_mpc);
	    return \@rvprg_cand_sizes ;
	}


	my $minsize_sgn_wd = $size_mpc;
	my $maxsize_sgn_wd = 0;
	my @nr_sizes_sgn_wds = ();
	my %already; # = ($size => 1, ...) for the sizes already occurred.

	foreach my $sgn_wd (@{$sgn_wds_included}) {
	    my $sites_in_sgn_wd = (defined $indx_sites_in_sgn_wd) ? $sgn_wd->[$indx_sites_in_sgn_wd] : $sgn_wd ;
	    my $size = @{$sites_in_sgn_wd};

	    unless ($already{$size}) {
		if ($size < $minsize_sgn_wd) { $minsize_sgn_wd = $size; }
		if ($maxsize_sgn_wd < $size) { $maxsize_sgn_wd = $size; }
		push @nr_sizes_sgn_wds, $size;
		$already{$size} = 1;
	    }
	}

	my @rvprg_cand_sizes = (1 .. $minsize_sgn_wd);

	if ($xlevel < 3.5) {
	    return \@rvprg_cand_sizes;
	}

	my @std_sizes_sgn_wds = sort {$a <=> $b} @nr_sizes_sgn_wds;

	if ($xlevel >= 5) {
	    if ($maxsize_sgn_wd < $size_mpc) { push @std_sizes_sgn_wds, $size_mpc; }
	}

	pop @rvprg_cand_sizes; # For the moment, remove $minsize_sgn_wd from the output.

	@rvprg_cand_sizes = (@rvprg_cand_sizes, @std_sizes_sgn_wds);

	return \@rvprg_cand_sizes;

}	# End of the subroutine, 'list_revpurge_cand_sizes'. #


# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# [ INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes), where ] ... OBSOLETE as of 2019/07/21.
#
# INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "purge"d;
#	They MUST be (vertically) complementary to each other, and MUST have the identical size;
#
#	@curr_set_clms is the current set of alignment columns, that is,
#	@column = @{$curr_set_clms[$c]} specifies the $c th column;
#	More precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input local alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@curr_set_clms is an output of 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#
#	@{$curr_bds_blks[$gb]} = ($lbd_gb, $rbd_gb) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	[Typically, \@curr_bds_blks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$curr_bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#	[@curr_bl_coords MUST be given within @gb_coords_wd_frm below];
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]}
#	give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@bds_gb_coords and \@gb_coords_wd_frm are outputs of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	@{$curr_gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are currently at the $lb th column;
#	@{$curr_gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are currently at the $rb th column;
#	[Typically, \@curr_gblks_w_spec_lb and \@curr_gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
#	$cnct_null_clm = join (':', @null_column), with @null_column consisting of as many "absence" states as the MSA rows;
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];

#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[\@cts_vincluding, \@cts_vincluded & \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#		# Added in revision (on 2019/07/21). #
#	$maxct_1way_shifts  specifies the upper-bound of the number of shifts in 1 way, i.e., either to the left or to the right;
#		# End of "Added in revision (on 2019/07/21)." #
#
#
# [ OUTPUTS: ($flag_end, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs, ] ... OBSOLETE as of 2019/07/15.
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes), where
#
#	$flag = 1 if the subroutine ended without any remarkable problems,
#	      = 0 if it failed;
#
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#		# END of "ADDED in revision on 2019/07/15." #
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "purge" (= undef if an error occurs);
#
#	@new_indices_gbs stores the new indices assigned to the gap-blocks AFTER the "purge";
#	More precisely, the gap-block that had been the $gb th before the "purge"
#	has been assigned $new_indices_gbs[$gb] after the "purge"
#	(, and $new_indices_gbs[$gb1] and $new_indices_gbs[$gb2] are undefined);
#
#	@old_indices_gbs stores the old indices assigned to the gap-blocks BEFORE the "purge";
#	More precisely, the gap-block that is the $gb th after the "purge"
#	had been assigned $old_indices_gbs[$gb] before the "purge";
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "purge";
#
# [ sub purge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@) { ] ... OBSOLETE as of 2019/07/21.
#
sub purge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {

	my ($gb1, $gb2, $curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_; # MODIFIED in revision on 2019/07/21.
	##$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'purge_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];
	unless ($size1 == $size2) {
	    print STDOUT "[ERROR 'purge_gblock_pair'] The gap-block pair, ($gb1, $gb2), have DIFFERENT SIZES ($size1 and $size2, respectively): $! ! ... they SHOULD be fed into 'cmerge_gblock_pair' ... \n\n"; # REVISED on 2019/07/19.
	    ##print STDOUT "[ERROR 'purge_gblock_pair'] The gap-block pair, ($gb1, $gb2), have DIFFERENT SIZES ($size1 and $size2, respectively): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};


    (print STDOUT "< Subroutine 'purge_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; block-pair to be \"purge\"d: (gb1, gb2) = ($gb1, $gb2); {block-size}=$size1; {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf)) >\n\n") if ($TEST_LEVEL_prg_gbp > 0);

		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_prg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp > 5) {...}"


   if ($TEST_LEVEL_prg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb2; # Always move the 2nd gap-block in the pair. #
	my $incr_coord_sbj;

	if ($lbd1_bf < $lbd2_bf) {

		# $gb1 is on the left of $gb2. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($lbd1_bf-1, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $lbd1_bf-1;
		print STDOUT "[ERROR 'purge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = - $ct_nec_shifts; # "Shift" $gb2 to the left.

	} else { # if ($lbd2_bf < $lbd1_bf) 

		# $gb1 is on the right of $gb2. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($rbd1_bf+1, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $rbd1_bf+1;
		print STDOUT "[ERROR 'purge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = $ct_nec_shifts; # "Shift" $gb2 to the right.

	} # End of "if ($lbd1_bf < $lbd2_bf) {...} else {...}"


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are "purge"d. ###

   (print STDOUT "\n ... Perform 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj) ...\n\n") if ($TEST_LEVEL_prg_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'purge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_prg_gbp >1);


   if ($TEST_LEVEL_prg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp >3) {...}"


		### MAIN 2: Finish the "purge" by removing the generated null-columns,
		###   and by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.

	my @new_indices_gbs = my @old_indices_gbs = ();

	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) { # 1st main for-loop (over the old gap-blocks). #

	    if (($ogb == $gb1) or ($ogb == $gb2)) {
		$new_indices_gbs[$ogb] = undef;
		next;
	    }

	    my $ngb;
	    if ($ogb < $gb1) {
		$ngb = $ogb;
	    } elsif ($ogb < $gb2) {
		$ngb = $ogb-1;
	    } else {
		$ngb = $ogb-2;
	    }

	    $new_indices_gbs[$ogb] = $ngb;
	    $old_indices_gbs[$ngb] = $ogb;

	} # End of the 1st main for-loop (over the old gap-blocks). #

	my $ct_out_gbs = @old_indices_gbs;

   if ($TEST_LEVEL_prg_gbp >2) {

	print STDOUT "\n[ Relationships between new and old indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

			# A BUG Fixed on 2019/07/15. #
	my @out_new_indices_gbs = my @out_old_indices_gbs = ();
	foreach my $indx (@new_indices_gbs) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
	foreach my $indx (@old_indices_gbs) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	print STDOUT '@new_indices_gbs = (', join (', ', @out_new_indices_gbs), ") .\n";
	print STDOUT '@old_indices_gbs = (', join (', ', @out_old_indices_gbs), ") .\n";
	##print STDOUT '@new_indices_gbs = (', join (', ', @new_indices_gbs), ") .\n";
	##print STDOUT '@old_indices_gbs = (', join (', ', @old_indices_gbs), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_prg_gbp >2) {...}"


	##my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs]; ## DISABLED in revision on 2019/07/21.
	my @new_curr_bds_blks = ();
	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) { $new_curr_bds_blks[$ngb] = []; }

	my ($lbd1_af, $rbd1_af) = @{$cp_curr_bds_blks[$gb1]};
	my ($lbd2_af, $rbd2_af) = @{$cp_curr_bds_blks[$gb2]};
	unless (($lbd1_af == $lbd2_af) and ($rbd1_af == $rbd2_af)) {
	    print STDOUT "[ERROR 'purge_gblock_pair'] The BOUNDARIES of \$gb1(=$gb1), [$lbd1_af, $rbd1_af], CLASHES with those of \$gb2(=$gb2), [$lbd2_af, $rbd2_af]: $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj, undef);
	}

	my $ct_rmvd_null_clms = 0;

	my @new_curr_set_clms = ();
 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $new_indx_clm = -1;
	for (my $c = 0; $c < $ct_in_clms; $c++) { # 2nd main for-loop (over the columns). #

	    if (($lbd1_af <= $c) and ($c <= $rbd1_af)) {
		my $cnct_clm = join (':', @{$cp_curr_set_clms[$c]}) ;
		if ($cnct_clm eq $cnct_null_clm) {
		    $ct_rmvd_null_clms++;
		    next;
		}
	    }

	    $new_indx_clm++;
	    $new_curr_set_clms[$new_indx_clm] = $cp_curr_set_clms[$c];

	    my @ngbs_w_old_lb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 1st inner foreach-loop (over gap-blocks with old left-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_lb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[0] = $new_indx_clm;
		}
	    } # End of the 1st inner foreach-loop. #

	    $new_curr_gblks_w_spec_lb[$new_indx_clm] = \@ngbs_w_old_lb_eq_c;

	    my @ngbs_w_old_rb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 2nd inner foreach-loop (over gap-blocks with old right-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_rb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[1] = $new_indx_clm;
		}
	    } # End of the 2nd inner foreach-loop. #

	    $new_curr_gblks_w_spec_rb[$new_indx_clm] = \@ngbs_w_old_rb_eq_c;

	} # End of the 2nd main for-loop (over the columns). #


=MOVED to below MAIN 3 in revision on 2019/07/21.
   if ($TEST_LEVEL_prg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp >3) {...}"
=cut # END of "MOVED to below MAIN 3 in revision on 2019/07/21."


		### MAIN 3: Create the remaining output data-sets. ###

	my @rows_affd_by_new_blks = @cp_rows_affd_by_blks[@old_indices_gbs];

=DISABLED in revision on 2019/07/21.
	my @new_bds_gb_coords = @cp_bds_gb_coords[@old_indices_gbs];
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm[@old_indices_gbs];
=cut # END of "DISABLED in revision on 2019/07/21."

	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #

		# ADDED in revision on 2019/07/21. #

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);
	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'purge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_prg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		# END of "ADDED in revision on 2019/07/21." #


	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'purge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_prg_gbp >1);


		# MOVED from above MAIN 3 in revision on 2019/07/21. #
   if ($TEST_LEVEL_prg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in revision on 2019/07/21." #

   if ($TEST_LEVEL_prg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_prg_gbp > 3) {...}"


    (print STDOUT "\n< Subroutine 'purge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, #{removed null-columns}=$ct_rmvd_null_clms). >\n\n") if ($TEST_LEVEL_prg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes);

}	# End of the subroutine, 'purge_gblock_pair'. #


# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes), where ] ... OBSOLETE as of 2019/07/21.
# INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#		# END of "ADDED in revision on 2019/07/15." #
#
#		# ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
#	$gb_rmvd and $gb_kpt are the (old) indices of the gap-blocks that were removed and kept, respectively;
#		# End of "ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19)." #
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "merge" (= undef if an error occurs);
#
#	@new_indices_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $gb th before the "merge"
#	has been assigned $new_indices_gbs[$gb] after the "merge"
#	(, and $new_indices_gbs[$gb1] and $new_indices_gbs[$gb2] are undefined);
#
#	@old_indices_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $gb th after the "merge"
#	had been assigned $old_indices_gbs[$gb] before the "merge";
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
# [ sub cmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@) { ] ... OBSOLETE as of 2019/07/21.
#
sub cmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {

	my ($gb1, $gb2, $curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_; # Modified in REVISION on 2019/07/21.
	##$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes) = @_; # OBSOLETE as of 2019/07/21.

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}


	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];
	unless ($size1 != $size2) {
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), have the IDENTICAL SIZE (=$size1): $! ! ... They SHOULD be fed into 'purge_gblock_pair' ...\n\n";
	    return (0, undef, undef, undef);
	}

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};


    (print STDOUT "< Subroutine 'cmerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf)) >\n\n") if ($TEST_LEVEL_cmrg_gbp > 0);

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #

	my ($gb_rmvd, $gb_kpt) = ($size1 < $size2) ? ($gb1, $gb2) : ($gb2, $gb1);
	my ($size_rmvd, $size_kpt) = @{$gblock_sizes}[$gb_rmvd, $gb_kpt];

   if ($TEST_LEVEL_cmrg_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_rmvd, $gb_kpt) = ', "($gb_rmvd, $gb_kpt) .\n";
	print STDOUT '($size_rmvd, $size_kpt) = ', "($size_rmvd, $size_kpt) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cmrg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Merge (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp > 5) {...}"


   if ($TEST_LEVEL_cmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb2; # Always move the 2nd gap-block in the pair. #
	my $incr_coord_sbj;

	if ($lbd1_bf < $lbd2_bf) {

		# $gb1 is on the left of $gb2. #

	    my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($rbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts1) {
		my $pt = $rbd1_bf;
		print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = - ($ct_nec_shifts1 + $size_rmvd); # "Shift" $gb2 to the left.

	} else { # if ($lbd2_bf < $lbd1_bf) 

		# $gb1 is on the right of $gb2. #

	    my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($lbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts1) {
		my $pt = $lbd1_bf;
		print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = $ct_nec_shifts1 + $size_rmvd; # "Shift" $gb2 to the right.

	} # End of "if ($lbd1_bf < $lbd2_bf) {...} else {...}"


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are completely "merge"d. ###

   (print STDOUT "\n ... Perform 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj) ...\n\n") if ($TEST_LEVEL_cmrg_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_cmrg_gbp >1);


   if ($TEST_LEVEL_cmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >3) {...}"


		### MAIN 2: Finish the "merge (complementary types)" by removing the generated null-columns,
		###   and by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.

	my @new_indices_gbs = my @old_indices_gbs = ();

	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) { # 1st main for-loop (over the old gap-blocks). #

	    if ($ogb == $gb_rmvd) {
		$new_indices_gbs[$ogb] = undef;
		next;
	    }

	    my $ngb;
	    if ($ogb < $gb_rmvd) {
		$ngb = $ogb;
	    } else { # if ($gb_rmvd < $ogb)
		$ngb = $ogb-1;
	    }

	    $new_indices_gbs[$ogb] = $ngb;
	    $old_indices_gbs[$ngb] = $ogb;

	} # End of the 1st main for-loop (over the old gap-blocks). #

	my $ct_out_gbs = @old_indices_gbs;

   if ($TEST_LEVEL_cmrg_gbp >2) {

	print STDOUT "\n[ Relationships between new and old indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

			# A BUG Fixed on 2019/07/15. #
	my @out_new_indices_gbs = my @out_old_indices_gbs = ();
	foreach my $indx (@new_indices_gbs) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
	foreach my $indx (@old_indices_gbs) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	print STDOUT '@new_indices_gbs = (', join (', ', @out_new_indices_gbs), ") .\n";
	print STDOUT '@old_indices_gbs = (', join (', ', @out_old_indices_gbs), ") .\n";
	##print STDOUT '@new_indices_gbs = (', join (', ', @new_indices_gbs), ") .\n";
	##print STDOUT '@old_indices_gbs = (', join (', ', @old_indices_gbs), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >2) {...}"


	##my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs]; # DISABLED in revision on 2019/07/21.

	my @new_curr_bds_blks = ();
	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) { $new_curr_bds_blks[$ngb] = []; }

	my ($lbd_rmvd_af, $rbd_rmvd_af) = @{$cp_curr_bds_blks[$gb_rmvd]};
	my ($lbd_kpt_af, $rbd_kpt_af) = @{$cp_curr_bds_blks[$gb_kpt]};
	##my ($lbd1_af, $rbd1_af) = @{$cp_curr_bds_blks[$gb1]};
	##my ($lbd2_af, $rbd2_af) = @{$cp_curr_bds_blks[$gb2]};

	unless (($lbd_kpt_af <= $lbd_rmvd_af) and ($rbd_rmvd_af <= $rbd_kpt_af)) {
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The BOUNDARIES of \$gb_kpt(=$gb_kpt), [$lbd_kpt_af, $rbd_kpt_af], does NOT ENCOMPASS those of \$gb_rmvd(=$gb_rmvd), [$lbd_rmvd_af, $rbd_rmvd_af]: $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, undef);
	}

	my $ct_rmvd_null_clms = 0;

	my @new_curr_set_clms = ();
 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $new_indx_clm = -1;
	my $pos_1st_rmvd_null_clm;

	for (my $c = 0; $c < $ct_in_clms; $c++) { # 2nd main for-loop (over the columns). #

	    if (($lbd_rmvd_af <= $c) and ($c <= $rbd_rmvd_af)) {
		my $cnct_clm = join (':', @{$cp_curr_set_clms[$c]}) ;
		if ($cnct_clm eq $cnct_null_clm) {
		    $ct_rmvd_null_clms++;
		    unless (defined $pos_1st_rmvd_null_clm) { $pos_1st_rmvd_null_clm = $c; }
		    next;
		}
	    }

	    $new_indx_clm++;
	    $new_curr_set_clms[$new_indx_clm] = $cp_curr_set_clms[$c];

	    my @ngbs_w_old_lb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 1st inner foreach-loop (over gap-blocks with old left-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_lb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[0] = $new_indx_clm;
		}
	    } # End of the 1st inner foreach-loop. #

	    $new_curr_gblks_w_spec_lb[$new_indx_clm] = \@ngbs_w_old_lb_eq_c;

	    my @ngbs_w_old_rb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 2nd inner foreach-loop (over gap-blocks with old right-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_rb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[1] = $new_indx_clm;
		}
	    } # End of the 2nd inner foreach-loop. #

	    $new_curr_gblks_w_spec_rb[$new_indx_clm] = \@ngbs_w_old_rb_eq_c;

	} # End of the 2nd main for-loop (over the columns). #


			# MAIN 3pre: Create @new_gblock_sizes & @new_inter_block_relations.
			# (MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.) #

	my $ngb_kpt = $new_indices_gbs[$gb_kpt]; # The new index of the gap-block to be kept.

	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];
	$new_gblock_sizes[$ngb_kpt] = $size_kpt - $size_rmvd;

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #

			# End of "(MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.)" #


			# MAIN 2post: Recover the removed boundary of $gb_kpt, #
			#	then, Retract it if necessary.
			# (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (2)) #

	my $if_left_part; # = 1 if the left-part of $gb_kpt will remain,
			  # = 0 if its right-part will remain.

	if ($incr_coord_sbj > 0) { # $gb2 was "shift"ed to the right.

	   if ($size1 < $size2) { # ($gb_kpt == $gb2)
		$if_left_part = 1;
	   } else { # ($gb_kpt == $gb1)
		$if_left_part = 0;
	   }
	} else { # $gb2 was "shift"ed to the left.

	   if ($size1 < $size2) { # ($gb_kpt == $gb2)
		$if_left_part = 0;
	   } else { # ($gb_kpt == $gb1)
		$if_left_part = 1;
	   }
	} # End of "if ($incr_coord_sbj > 0) {...} else {...}".

	if ($if_left_part) { # The left-part of $gb_kpt remains. => Recover its new right-bound.

	    my $new_rb_kpt = $pos_1st_rmvd_null_clm-1; # The right-boundary SHOULD be immediately on the left of the removed null columns.
	    push @{$new_curr_gblks_w_spec_rb[$new_rb_kpt]}, $ngb_kpt;
	    $new_curr_bds_blks[$ngb_kpt]->[1] = $new_rb_kpt;

	    my $ct_vretracting = vretract_right_end_sbj_gblk ($ngb_kpt, @new_curr_bds_blks, @new_inter_block_relations, @new_gblock_sizes, @new_curr_gblks_w_spec_rb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk' FAILED on the new gap-block to be kept, $ngb_kpt (or $gb_kpt in the old gap-block set): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, undef);
	    }
   (print STDOUT "\n ... Subroutine 'vretract_right_end_sbj_gblk' Ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_cmrg_gbp >1);

	} else { # The right-part of $gb_kpt remains. => Recover its new left-bound.

	    my $new_lb_kpt = $pos_1st_rmvd_null_clm; # After removing the null-columns, the new left-boundary SHOULD be positioned like this.
	    push @{$new_curr_gblks_w_spec_lb[$new_lb_kpt]}, $ngb_kpt;
	    $new_curr_bds_blks[$ngb_kpt]->[0] = $new_lb_kpt;

	    my $ct_vretracting = vretract_left_end_sbj_gblk ($ngb_kpt, @new_curr_bds_blks, @new_inter_block_relations, @new_gblock_sizes, @new_curr_gblks_w_spec_lb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk' FAILED on the new gap-block to be kept, $ngb_kpt (or $gb_kpt in the old gap-block set): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, undef);
	    }
   (print STDOUT "\n ... Subroutine 'vretract_left_end_sbj_gblk' Ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_cmrg_gbp >1);

	} # END of "if ($if_left_part) {...} else {...}"

			# End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (2))" #


=MOVED to below MAIN 3 in revision on 2019/07/21.
   if ($TEST_LEVEL_cmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge (complelemtary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >3) {...}"
=cut # End of "MOVED to below MAIN 3 in revision on 2019/07/21."


		### MAIN 3: Create the remaining output data-sets. ###

	my @rows_affd_by_new_blks = @cp_rows_affd_by_blks[@old_indices_gbs];

=DISABLED in revision on 2019/07/21.
	my @new_bds_gb_coords = @cp_bds_gb_coords[@old_indices_gbs];
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm[@old_indices_gbs];
=cut # End of "DISABLED in revision on 2019/07/21."

		# ADDED in revision on 2019/07/21. #

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);
	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_cmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		# END of "ADDED in revision on 2019/07/21." #

=MOVED to "MAIN 3pre" above (on 2019/07/19).
	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #
=cut # End of "MOVED to "MAIN 3pre" above (on 2019/07/19)."

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'cmerge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_cmrg_gbp >1);


		# MOVED from above MAIN 3 in revision on 2019/07/21. #
   if ($TEST_LEVEL_cmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in revision on 2019/07/21." #


   if ($TEST_LEVEL_cmrg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Merge (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_gbp > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'cmerge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$gb_rmvd = $gb_rmvd, \$gb_kpt = $gb_kpt, #{removed null-columns}=$ct_rmvd_null_clms). >\n\n") if ($TEST_LEVEL_cmrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes);

}	# End of the subroutine, 'cmerge_gblock_pair'. #


# smerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks of the SAME type.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/07/21.]
#	[DEBUGGED on 2019/07/22]:
#	Now, the subroutine can correctly handle a gap-block pair with ZERO distance
#	that flank another gap-block in the 'onx' relationship.
#	[DEBUGGED on 2019/08/07]:
#	Now, the subroutine can handle the cases where $gb1 +1 < $gb2.
#	[REVISED on 2019/08/07]:
#	Now, the subroutine invokes 'mk_orgs_bds_gb_coords' from inside.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where 
#	[ \@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes), where ] ... OBSOLETE as of 2019/08/07.
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be of the same vertical type;
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, \@new_indices_gbs, \@old_indices_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes), where
#
#	$flag = 1 if the subroutine ended without any remarkable problems,
#	      = 0 if it failed;
#
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	$gb_absbd and $gb_kpt are the indices of the gap-blocks that were absorbed and kept (actually, absorbing), respectively;
#
#	@new_indices_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $gb th before the "merge"
#	has been assigned $new_indices_gbs[$gb] after the "merge"
#	(, and $new_indices_gbs[$gb_absbd] is undefined);
#
#	@old_indices_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $gb th after the "merge"
#	had been assigned $old_indices_gbs[$gb] before the "merge";
#
#	$maxct_1way_shifts is the maximum count of "shift"s permitted; # ADDED on 2019/08/07. #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
# [ sub smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@) { ] ... OBSOLETE as of 2019/08/07.
#
sub smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) { # REVISED on 2019/08/07. #

	my ($gb1, $gb2, $curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts
		) = @_; # ADDED $maxct_1way_shifts on 2019/08/07. #
	## [ $inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes) = @_; ] ... OBSOLETE as of 2019/08/07.

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq '=') {
	    print STDOUT "[ERROR 'smerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};


    (print STDOUT "< Subroutine 'smerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; same-type block-pair to be \"merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf)) >\n\n") if ($TEST_LEVEL_smrg_gbp > 0);

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'smerge_gblock_pair' on 2019/07/21. (1)) #

	my ($gb_absbd, $gb_kpt) = ($gb2, $gb1); # Always keep $gb1. (MODIFIED in creation of 'smerge_gblock_pair'.)
	##my ($gb_rmvd, $gb_kpt) = ($size1 < $size2) ? ($gb1, $gb2) : ($gb2, $gb1);

	my ($size_absbd, $size_kpt) = @{$gblock_sizes}[$gb_absbd, $gb_kpt];

   if ($TEST_LEVEL_smrg_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_absbd, $gb_kpt) = ', "($gb_absbd, $gb_kpt) .\n";
	print STDOUT '($size_absbd, $size_kpt) = ', "($size_absbd, $size_kpt) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'smerge_gblock_pair' on 2019/07/21. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_smrg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Merge (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp > 5) {...}"


   if ($TEST_LEVEL_smrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb2; # Always move the 2nd gap-block in the pair. #
	my $incr_coord_sbj;
	my $if_left = 0; # Added for DEBUGGING (on 2019/07/22). (1) #

	if ($lbd1_bf < $lbd2_bf) {

		# $gb1 is on the left of $gb2. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($rbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $rbd1_bf;
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    ##$incr_coord_sbj = - ($ct_nec_shifts1 + $size_rmvd); # "Shift" $gb2 to the left.
	    $incr_coord_sbj = - $ct_nec_shifts; # => MODIFIED in creation of 'smerge_gblock_pair'.

	    $if_left = 1; # Added for DEBUGGING (on 2019/07/22). (2) #

	} else { # if ($lbd2_bf < $lbd1_bf) 

		# $gb1 is on the right of $gb2. #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($lbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $lbd1_bf;
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    ##$incr_coord_sbj = $ct_nec_shifts1 + $size_rmvd; # "Shift" $gb2 to the right.
	    $incr_coord_sbj = $ct_nec_shifts; # => MODIFIED in creation of 'smerge_gblock_pair'. 

	} # End of "if ($lbd1_bf < $lbd2_bf) {...} else {...}"


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are completely "merge"d. ###

   (print STDOUT "\n ... Perform 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj) ...\n\n") if ($TEST_LEVEL_smrg_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);


		# MOVE the gap-blocks in the 'ONX' relationships with $gb1 & $gb2, if any,
		# TO the outside of THE REAR-END of $bl_sbj. #
		# (ADDED in creation of 'smerge_gblock_pair' (on 2019/07/21). (MAIN 1)) #

		# MODIFIED for DEBUGGING (on 2019/07/22). (3) #
	if ($if_left) { # $gb1 is on the left of $gb2(=$bl_sbj).
	##if ($incr_coord_sbj < 0) { # $bl_sbj has just been "shift"ed to the left.
		# End of "MODIFIED for DEBUGGING (on 2019/07/22). (3)" #

	    my $ct_vextensions = vextend_left_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @cp_curr_set_clms, @{$inter_block_relations}, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vextensions) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vextend_left_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vextend_left_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	    my $ct_vretracting = vretract_left_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations}, @{$gblock_sizes}, @cp_curr_gblks_w_spec_lb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vretract_left_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	} else { # if ($incr_coord_sbj > 0) # $bl_sbj has just been "shift"ed to the right. #

	    my $ct_vextensions = vextend_right_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @cp_curr_set_clms, @{$inter_block_relations}, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vextensions) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vextend_right_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vextend_right_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	    my $ct_vretracting = vretract_right_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations}, @{$gblock_sizes}, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vretract_right_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	} # End of "if ($incr_coord_sbj < 0) {...} else {...}". #

		# END of "(ADDED in creation of 'smerge_gblock_pair' (on 2019/07/21). (MAIN 1))" #

   if ($TEST_LEVEL_smrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >3) {...}"


		### MAIN 2: Finish the "merge (same type)" 
		###   by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.

	my @new_indices_gbs = my @old_indices_gbs = ();

	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) { # 1st main for-loop (over the old gap-blocks). #

	    if ($ogb == $gb_absbd) {
		$new_indices_gbs[$ogb] = undef;
		next;
	    }

	    my $ngb;
	    if ($ogb < $gb_absbd) {
		$ngb = $ogb;
	    } else { # if ($gb_absbd < $ogb)
		$ngb = $ogb-1;
	    }

	    $new_indices_gbs[$ogb] = $ngb;
	    $old_indices_gbs[$ngb] = $ogb;

	} # End of the 1st main for-loop (over the old gap-blocks). #

	my $ct_out_gbs = @old_indices_gbs;

   if ($TEST_LEVEL_smrg_gbp >2) {

	print STDOUT "\n[ Relationships between new and old indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

			# A BUG Fixed on 2019/07/15. #
	my @out_new_indices_gbs = my @out_old_indices_gbs = ();
	foreach my $indx (@new_indices_gbs) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
	foreach my $indx (@old_indices_gbs) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	print STDOUT '@new_indices_gbs = (', join (', ', @out_new_indices_gbs), ") .\n";
	print STDOUT '@old_indices_gbs = (', join (', ', @out_old_indices_gbs), ") .\n";
	##print STDOUT '@new_indices_gbs = (', join (', ', @new_indices_gbs), ") .\n";
	##print STDOUT '@old_indices_gbs = (', join (', ', @old_indices_gbs), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >2) {...}"

		# MOVED from below the 2nd main for-loop (in creation of 'smerge_gblock_pair'). #
	my $ngb_kpt = $new_indices_gbs[$gb_kpt]; # The new index of the gap-block to be kept.
		# End of "MOVED from below the 2nd main for-loop (in creation of 'smerge_gblock_pair')." #

	###my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs]; # DISABLED in REVISION on 2019/08/07. #

=DISABLED in creation of 'smerge_gblock_pair'. (2a) #
	my @new_curr_bds_blks = ();
	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) { $new_curr_bds_blks[$ngb] = []; }
=cut # END of "DISABLED in creation of 'smerge_gblock_pair'. (2a)" #

	my ($lbd_absbd_af, $rbd_absbd_af) = @{$cp_curr_bds_blks[$gb_absbd]};
	my ($lbd_kpt_af, $rbd_kpt_af) = @{$cp_curr_bds_blks[$gb_kpt]};
	##my ($lbd1_af, $rbd1_af) = @{$cp_curr_bds_blks[$gb1]};
	##my ($lbd2_af, $rbd2_af) = @{$cp_curr_bds_blks[$gb2]};

	unless (($rbd_kpt_af < $lbd_absbd_af) or ($rbd_absbd_af < $lbd_kpt_af)) {
	    print STDOUT "[ERROR 'smerge_gblock_pair'] The BOUNDARIES of \$gb_kpt(=$gb_kpt), [$lbd_kpt_af, $rbd_kpt_af], does OVERLAP those of \$gb_absbd(=$gb_absbd), [$lbd_absbd_af, $rbd_absbd_af]: $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, undef);
	}

		# ADDED in creation of 'smerge_gblock_pair'. #

	my @new_curr_bds_blks = @cp_curr_bds_blks[@old_indices_gbs];

	if ($rbd_kpt_af < $lbd_absbd_af) {
		# Replace the right-end of the gap-block to be kept. #
	    $new_curr_bds_blks[$ngb_kpt]->[1] = $rbd_absbd_af;

	} else { # if ($rbd_absbd_af < $lbd_kpt_af) 
		# Replace the left-end of the gap-block to be kept.
	    $new_curr_bds_blks[$ngb_kpt]->[0] = $lbd_absbd_af;

	}
		# End of "ADDED in creation of 'smerge_gblock_pair'." #

		# MODIFIED in creation of 'smerge_gblock_pair'. (2a) #

	my @new_curr_set_clms = @cp_curr_set_clms; # The set of columns does NOT change as the result of the "merge (same type)", since the end of the serial "shift"s. #

 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # Preliminary main for-loop (over the columns). #
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	} # End of the preliminary main for-loop (over the columns).

	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 2nd main for-loop (over the new gap-blocks). #
	    my ($lb_ngb, $rb_ngb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb_ngb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb_ngb]}, $ngb;
	} # End of the 2nd main for-loop (over the new gap-blocks). #

=DISABLED in creation of 'smerge_gblock_pair'. (2b) #
	my @new_curr_set_clms = ();
 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $new_indx_clm = -1;
	my $pos_1st_rmvd_null_clm;

	for (my $c = 0; $c < $ct_in_clms; $c++) { # 2nd main for-loop (over the columns). #

	    $new_indx_clm++;
	    $new_curr_set_clms[$new_indx_clm] = $cp_curr_set_clms[$c];

	    my @ngbs_w_old_lb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 1st inner foreach-loop (over gap-blocks with old left-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_lb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[0] = $new_indx_clm;
		}
	    } # End of the 1st inner foreach-loop. #

	    $new_curr_gblks_w_spec_lb[$new_indx_clm] = \@ngbs_w_old_lb_eq_c;

	    my @ngbs_w_old_rb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 2nd inner foreach-loop (over gap-blocks with old right-bound = $c). #
		my $ngb = $new_indices_gbs[$ogb];
		if (defined $ngb) {
		    push @ngbs_w_old_rb_eq_c, $ngb;
		    $new_curr_bds_blks[$ngb]->[1] = $new_indx_clm;
		}
	    } # End of the 2nd inner foreach-loop. #

	    $new_curr_gblks_w_spec_rb[$new_indx_clm] = \@ngbs_w_old_rb_eq_c;

	} # End of the 2nd main for-loop (over the columns). #
=cut # End of "DISABLED in creation of 'smerge_gblock_pair'. (2b)" #

		# END of "MODIFIED in creation of 'smerge_gblock_pair'. (2a)" #


			# MAIN 3pre: Create @new_gblock_sizes & @new_inter_block_relations.
			# (MOVED from "MAIN 3" below, in creation of 'smerge_gblock_pair' on 2019/07/19.) #

=MOVED to above the 2nd main for-loop.
	my $ngb_kpt = $new_indices_gbs[$gb_kpt]; # The new index of the gap-block to be kept.
=cut # END of "MOVED to above the 2nd main for-loop."


	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];
	##$new_gblock_sizes[$ngb_kpt] = $size_kpt - $size_rmvd;
	$new_gblock_sizes[$ngb_kpt] = $size_kpt + $size_absbd; # => MODIFIED in creation of 'smerge_gblock_pair' (on 2019/07/21). (3pre) #

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #

			# End of "(MOVED from "MAIN 3" below, in creation of 'smerge_gblock_pair' on 2019/07/19.)" #


=DISABLED in creation of 'smerge_gblock_pair'. (2post)
			# MAIN 2post: Recover the removed boundary of $gb_kpt, #
			#	then, Retract it if necessary.
			# (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (2)) #

	my $if_left_part; # = 1 if the left-part of $gb_kpt will remain,
			  # = 0 if its right-part will remain.

	if ($incr_coord_sbj > 0) { # $gb2 was "shift"ed to the right.

	   if ($size1 < $size2) { # ($gb_kpt == $gb2)
		$if_left_part = 1;
	   } else { # ($gb_kpt == $gb1)
		$if_left_part = 0;
	   }
	} else { # $gb2 was "shift"ed to the left.

	   if ($size1 < $size2) { # ($gb_kpt == $gb2)
		$if_left_part = 0;
	   } else { # ($gb_kpt == $gb1)
		$if_left_part = 1;
	   }
	} # End of "if ($incr_coord_sbj > 0) {...} else {...}".

	if ($if_left_part) { # The left-part of $gb_kpt remains. => Recover its new right-bound.

	    my $new_rb_kpt = $pos_1st_rmvd_null_clm-1; # The right-boundary SHOULD be immediately on the left of the removed null columns.
	    push @{$new_curr_gblks_w_spec_rb[$new_rb_kpt]}, $ngb_kpt;
	    $new_curr_bds_blks[$ngb_kpt]->[1] = $new_rb_kpt;

	    my $ct_vretracting = vretract_right_end_sbj_gblk ($ngb_kpt, @new_curr_bds_blks, @new_inter_block_relations, @new_gblock_sizes, @new_curr_gblks_w_spec_rb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk' FAILED on the new gap-block to be kept, $ngb_kpt (or $gb_kpt in the old gap-block set): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, undef);
	    }
   (print STDOUT "\n ... Subroutine 'vretract_right_end_sbj_gblk' Ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	} else { # The right-part of $gb_kpt remains. => Recover its new left-bound.

	    my $new_lb_kpt = $pos_1st_rmvd_null_clm; # After removing the null-columns, the new left-boundary SHOULD be positioned like this.
	    push @{$new_curr_gblks_w_spec_lb[$new_lb_kpt]}, $ngb_kpt;
	    $new_curr_bds_blks[$ngb_kpt]->[0] = $new_lb_kpt;

	    my $ct_vretracting = vretract_left_end_sbj_gblk ($ngb_kpt, @new_curr_bds_blks, @new_inter_block_relations, @new_gblock_sizes, @new_curr_gblks_w_spec_lb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk' FAILED on the new gap-block to be kept, $ngb_kpt (or $gb_kpt in the old gap-block set): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, undef);
	    }
   (print STDOUT "\n ... Subroutine 'vretract_left_end_sbj_gblk' Ended (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	} # END of "if ($if_left_part) {...} else {...}"

			# End of "(ADDED in creation of 'smerge_gblock_pair' on 2019/07/19. (2))" #
=cut # END of "DISABLED in creation of 'smerge_gblock_pair'. (2post)"

	
=MOVED to below MAIN 3 in creation of 'smerge_gblock_pair'.
   if ($TEST_LEVEL_smrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge (complelemtary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >3) {...}"
=cut # END of "MOVED to below MAIN 3 in creation of 'smerge_gblock_pair'."


		### MAIN 3: Create the remaining output data-sets. ###

	my @rows_affd_by_new_blks = @cp_rows_affd_by_blks[@old_indices_gbs];

		#  REVISED on 2019/08/07. (0) #
	my $new_ct_clms = @new_curr_set_clms;
	my ($new_le_wd, $new_re_wd) = (0, $new_ct_clms -1);
	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_orgs_gb_coords) {
		print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the new set of gap-blocks: $! !\n\n";
		return (0, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, undef);
	}
   (print STDOUT "\n ... Subroutine 'mk_orgs_bds_gb_coords' Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};
	##my @new_bds_gb_coords = @cp_bds_gb_coords[@old_indices_gbs];
	##my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm[@old_indices_gbs];
		#  END of "REVISED on 2019/08/07. (0)" #


		# ADDED in creation of 'smerge_gblock_pair'. (MAIN 3) #

	my ($coord_bl_absbd, $coord_bl_kpt) = @cp_curr_bl_coords[$gb_absbd, $gb_kpt];
	my ($lb_coord_absbd, $rb_coord_absbd) = @{$cp_bds_gb_coords[$gb_absbd]};
	my ($lb_coord_kpt, $rb_coord_kpt) = @{$cp_bds_gb_coords[$gb_kpt]};
	my ($coord_absbd_le_wd, $coord_absbd_re_wd) = @{$cp_gb_coords_wd_frm[$gb_absbd]};
	my ($coord_kpt_le_wd, $coord_kpt_re_wd) = @{$cp_gb_coords_wd_frm[$gb_kpt]};

	my $rels_w_kpt = $inter_block_relations->[$gb_kpt]; # ADDED on 2019/08/07. (1) #

	if ($rbd_kpt_af < $lbd_absbd_af) {

		# $gb_kpt is on the left of $gb_absbd. #

	    my $ct_left_movable = $coord_bl_kpt - $lb_coord_kpt;
	    my $ct_right_movable = $rb_coord_absbd - $coord_bl_absbd;

	    for (my $gb3=$gb_kpt +1; $gb3 < $gb_absbd; $gb3++) { # ADDED on 2019/08/07. (2a) #
		unless ($rels_w_kpt->[$gb3] eq '=') { next; }
		my ($lbd3, $rbd3) = @{$cp_curr_bds_blks[$gb3]};
		if ($rbd_absbd_af <$lbd3) { $ct_right_movable += $gblock_sizes->[$gb3]; }
	    }

	    my $new_rb_coord_kpt = $ct_left_movable + $ct_right_movable;

	    $new_curr_bl_coords[$ngb_kpt] = $ct_left_movable;
	    $new_bds_gb_coords[$ngb_kpt] = [0, $new_rb_coord_kpt];

	    my $new_left_margin = $coord_kpt_le_wd - $lb_coord_kpt;
	    my $new_right_margin = $coord_absbd_re_wd - $rb_coord_absbd;

=DISABLED because it seemed superfluous. (a)
	    for (my $gb3=$gb_kpt +1; $gb3 < $gb_absbd; $gb3++) { # ADDED on 2019/08/07. (3a) #
		unless ($rels_w_kpt->[$gb3] eq '=') { next; }
		my ($lbd3, $rbd3) = @{$cp_curr_bds_blks[$gb3]};
		if ($rbd_absbd_af <$lbd3) { $new_right_margin += $gblock_sizes->[$gb3]; }
	    }
=cut # END of "DISABLED because it seemed superfluous. (a)"

	    $new_gb_coords_wd_frm[$ngb_kpt] = [ $new_left_margin, $new_rb_coord_kpt + $new_right_margin];

	} else { # if ($rbd_absbd_af < $lbd_kpt_af) 

		# $gb_absbd is on the left of $gb_kpt. #

	    my $ct_left_movable = $coord_bl_absbd - $lb_coord_absbd;
	    my $ct_right_movable = $rb_coord_kpt - $coord_bl_kpt;

	    for (my $gb3=$gb_kpt +1; $gb3 < $gb_absbd; $gb3++) { # ADDED on 2019/08/07. (2b) #
		unless ($rels_w_kpt->[$gb3] eq '=') { next; }
		my ($lbd3, $rbd3) = @{$cp_curr_bds_blks[$gb3]};
		if ($rbd3 < $lbd_absbd_af) { $ct_left_movable += $gblock_sizes->[$gb3]; }
	    }

	    my $new_rb_coord_kpt = $ct_left_movable + $ct_right_movable;

	    $new_curr_bl_coords[$ngb_kpt] = $ct_left_movable;
	    $new_bds_gb_coords[$ngb_kpt] = [0, $new_rb_coord_kpt];

	    my $new_left_margin = $coord_absbd_le_wd - $lb_coord_absbd;
	    my $new_right_margin = $coord_kpt_re_wd - $rb_coord_kpt;

=DISABLED because it seemed superfluous. (b)
	    for (my $gb3=$gb_kpt +1; $gb3 < $gb_absbd; $gb3++) { # ADDED on 2019/08/07. (3b) #
		unless ($rels_w_kpt->[$gb3] eq '=') { next; }
		my ($lbd3, $rbd3) = @{$cp_curr_bds_blks[$gb3]};
		if ($rbd3 < $lbd_absbd_af) { $new_left_margin += $gblock_sizes->[$gb3]; }
	    }
=cut # END of "DISABLED because it seemed superfluous. (b)"

	    $new_gb_coords_wd_frm[$ngb_kpt] = [ $new_left_margin, $new_rb_coord_kpt + $new_right_margin];

	}
		# END of "ADDED in creation of 'smerge_gblock_pair'. (MAIN 3)" #


=MOVED to "MAIN 3pre" above (on 2019/07/19).
	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #
=cut # End of "MOVED to "MAIN 3pre" above (on 2019/07/19)."

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'smerge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);


		# MOVED from above MAIN 3 in creation of 'smerge_gblock_pair'. #
   if ($TEST_LEVEL_smrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge (same type)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in creation of 'smerge_gblock_pair'." #


   if ($TEST_LEVEL_smrg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Merge (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'smerge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$gb_absbd = $gb_absbd, \$gb_kpt = $gb_kpt). >\n\n") if ($TEST_LEVEL_smrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, \@new_indices_gbs, \@old_indices_gbs,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes);

}	# End of the subroutine, 'smerge_gblock_pair'. #


# cvmerge_gblock_pair:
#	This subroutine actually performs the "(incomplete) vertical-merge" of a specified pair of
#	complementary-sibling gap-blocks.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_gblock_pair'.
#	[CREATION STARTED on 2019/07/23, ENDED on 2019/07/27.]
#	[REVISED on 2019/07/29]:
#	SWAPPED the subroutines 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right' in order,
#	and post-processed the output of 'displace_including_and_onx_to_right',
#	in order to make the final result as close as possible to the desired result.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 1)" #
#	), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "(incomplete) vertical-merge"d;
#	They MUST be (vertically) complementary-siblings.
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#			# (Added on 2019/07/26.) #
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#			# End of "(Added on 2019/07/26.)" #
#
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 2)" #
#	
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 3) #
#	\@new_list_gblocks
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 3)" #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#		# END of "ADDED in revision on 2019/07/15." #
#
#		# ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
#	$if_complete = 1 if the "vertical-merge" is complete,
#		     = 0 if it is incomplete;
#		# End of "ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24)." #
#
#		# ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
		# => MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
#	$gb_rplcd, $gb_rmvd_kpt, respectively, are the old indices of the gap-blocks that was replaced (with the "vmerged" blocks) and that was either kept (when $if_complete = 0) or removed (when $if_complete = 1);
#		# End of "ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
		# => MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24)." #
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "(incomplete) vertical-merge" (= undef if an error occurs);
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "(incomplete) vertical-merge";
#	More precisely, the gap-block that had been the $gb th before the "(incomplete) vertical-merge"
#	has been assigned $new_indices_old_gbs[$gb] after the "(incomplete) vertical-merge"
#	(, and $new_indices_old_gbs[$gb_rmvd_kpt] is undefined if ($if_complete = 1));
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "(incomplete) vertical-merge";
#	More precisely, the gap-block that is the $gb th after the "(incomplete) vertical-merge"
#	had been assigned $old_indices_new_gbs[$gb] before the "(incomplete) vertical-merge";
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "(incomplete) vertical-merge";
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 5) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 5)" #
#
sub cvmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb1, $gb2, $curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,

		# Added in creation of 'cvmerge_gblock_pair'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq # (Added this line on 2019/07/26.)
		# End of "Added in creation of 'cvmerge_gblock_pair'. (IN)" #
	) = @_; 

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'ONCS') {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}


	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

=For the predecessor, 'cmerge_gblock_pair'. (1)
	unless ($size1 != $size2) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), have the IDENTICAL SIZE (=$size1): $! ! ... They SHOULD be fed into 'purge_gblock_pair' ...\n\n";
	    return (0, undef, undef, undef);
	}
=cut # End of "For the predecessor, 'cmerge_gblock_pair'. (1)"

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};


    (print STDOUT "< Subroutine 'cvmerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; complementary-sibling block-pair to be \"merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf)) >\n\n") if ($TEST_LEVEL_cvmrg_gbp > 0);


		# (Preliminary 0): Determine which will be replaced and which will be removed/kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'cvmerge_gblock_pair' on 2019/07/24. (1)) # 

	my ($if_complete, $gb_rplcd, $gb_rmvd_kpt);
	if ($size1 < $size2) {
	    $if_complete = 0;
	    ($gb_rplcd, $gb_rmvd_kpt) = ($gb1, $gb2);

	} elsif ($size1 > $size2) {
	    $if_complete = 0;
	    ($gb_rplcd, $gb_rmvd_kpt) = ($gb2, $gb1);
	} else {
	    $if_complete = 1;
	    ($gb_rplcd, $gb_rmvd_kpt) = ($gb1, $gb2);
	}
	my ($size_rplcd, $size_rmvd_kpt) = @{$gblock_sizes}[$gb_rplcd, $gb_rmvd_kpt];
	##my ($size_rmvd, $size_kpt) = @{$gblock_sizes}[$gb_rmvd, $gb_kpt];
	##my ($gb_rmvd, $gb_kpt) = ($size1 < $size2) ? ($gb1, $gb2) : ($gb2, $gb1);
	##my ($size_rmvd, $size_kpt) = @{$gblock_sizes}[$gb_rmvd, $gb_kpt];

   if ($TEST_LEVEL_cvmrg_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '$if_complete = ', $if_complete, " .\n";
	print STDOUT '($gb_rplcd, $gb_rmvd_kpt) = ', "($gb_rplcd, $gb_rmvd_kpt) .\n";
	print STDOUT '($size_rplcd, $size_rmvd_kpt) = ', "($size_rplcd, $size_rmvd_kpt) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'cvmerge_gblock_pair' on 2019/07/24. (1))" # 


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cvmrg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Merge (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp > 5) {...}"



   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb2; # Always move the 2nd gap-block in the pair. #
	my $incr_coord_sbj;
	my $if_left; # = 1 if $bl_sbj (=$gb2) will be moved to the left, = 0 if to the right. # ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). (3a) #

	if ($lbd1_bf < $lbd2_bf) {

		# $gb1 is on the left of $gb2. #

	    my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($rbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts1) {
		my $pt = $rbd1_bf;
		print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = - $ct_nec_shifts1 ; # "Shift" $gb2 to the left. # MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
	    ##$incr_coord_sbj = - ($ct_nec_shifts1 + $size_rmvd); # "Shift" $gb2 to the left.

	    $if_left = 1; # Added in creation of 'cvmerge_gblock_pair' (on 2019/07/24). (3b) #

	} else { # if ($lbd2_bf < $lbd1_bf) 

		# $gb1 is on the right of $gb2. #

	    my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($lbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts1) {
		my $pt = $lbd1_bf;
		print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    $incr_coord_sbj = $ct_nec_shifts1; # "Shift" $gb2 to the right. # MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
	    ##$incr_coord_sbj = $ct_nec_shifts1 + $size_rmvd; # "Shift" $gb2 to the right.

	    $if_left = 0; # Added in creation of 'cvmerge_gblock_pair' (on 2019/07/24). (3c) #

	} # End of "if ($lbd1_bf < $lbd2_bf) {...} else {...}"


		### [ MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are completely "merge"d. ### ] ... for the predecessor, 'cmerge_gblock_pair'.
		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' (on $bl_sbj) until the two gap-blocks effectively "touch" each other. ### (REVISED in creation of 'cvmerge_gblock_pair' (on 2019/07/24).) # 

   (print STDOUT "\n ... Perform 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);


   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s of \$bl_sbj (=$bl_sbj) ]\n\n"; # Modified in creation of 'cvmerge_gblock_pair'. #
	##print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s  ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"


		### MAIN 2: Displace the gap-blocks either vertically including $gb1 or $gb2 
		### or in the 'ONX' relationships with $gb1 or $gb2 from
		### the regions spanned by $gb1 and $gb2 and that in between them.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###


	## my $if_left; # = 1 if $bl_sbj (=$gb2) will be moved to the left, = 0 if to the right. # ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). (3a) #

	my ($gb_left, $gb_right) = ($if_left) ? ($gb1, $gb2) : ($gb2, $gb1); # CORRECTED on 2019/07/25.
	##my ($gb_left, $gb_right) = ($if_left) ? ($gb2, $gb1) : ($gb1, $gb2);

	my %ignored = ($gb1 => 1, $gb2 => 1); # ADDED in revision on 2019/07/26. #

			## SWAPPED 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right', in REVISION of 'cvmerge_gblock_pair' (on 2019/07/29). ###

	my ($le_left, $re_left) = @{$cp_curr_bds_blks[$gb_left]};
	my $left_limit = $re_left+1;

	my ($gb2incr_coord_right, $totct_nec_shifts_right)
	   = displace_including_and_onx_to_right ($gb_right, $left_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored # Added %ignored to the input arguments on 2019/07/26.
		);
	unless (defined $gb2incr_coord_right) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_right', FAILED on the \$gb_right (=$gb_right) th gap-block and \$left_limit = $left_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_right', Ended on the \$gb_right (=$gb_right) th gap-block and \$left_limit = $left_limit (w/ \$totct_nec_shifts_right = $totct_nec_shifts_right) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);


   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_right} (before post-processing): ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_right} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_right->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"

		# Post-process %{$gb2incr_coord_right}. #
		# (ADDED in REVISION on (2019/07/29).) #

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    my $incr = $gb2incr_coord_right->{$gb};
	    if ($incr > $size_rplcd) {
		$gb2incr_coord_right->{$gb} = $incr - $size_rplcd;
	    } else {
		$gb2incr_coord_right->{$gb} = 0;
	    }
	} # End of "foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {...}"

   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_right} (AFTER post-processing): ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_right} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_right->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"

		# END of "(ADDED in REVISION on (2019/07/29).)" #


   if ($TEST_LEVEL_cvmrg_gbp >5) { # ADDED in revision on 2019/07/29. #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_right' on \$gb_right (=$gb_right) and \$left_limit = $left_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >5) { # ADDED in revision on 2019/07/29. # ...}"


	my ($le_right, $re_right) = @{$cp_curr_bds_blks[$gb_right]};
	my $right_limit = $le_right-1;

	my ($gb2incr_coord_left, $totct_nec_shifts_left)
	   = displace_including_and_onx_to_left ($gb_left, $right_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored # Added %ignored to the input arguments on 2019/07/26.
		);
	unless (defined $gb2incr_coord_left) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_left', FAILED on the \$gb_left (=$gb_left) th gap-block and \$right_limit = $right_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_left', Ended on the \$gb_left (=$gb_left) th gap-block and \$right_limit = $right_limit (w/ \$totct_nec_shifts_left = $totct_nec_shifts_left) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);


   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_left}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_left} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_left}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_left->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"

   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_left' on \$gb_left (=$gb_left) and \$right_limit = $right_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"

			## END of "SWAPPED 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right', in REVISION of 'cvmerge_gblock_pair' (on 2019/07/29)." ###


		# Post-MAIN 2: Check whether ALL the gap-blocks in between $gb1 and $gb2 are DISPLACED. #
		# (Added on 2019/07/25.) (Post M2) #

	my ($le_left_bf_cvm, $re_left_bf_cvm) = @{$cp_curr_bds_blks[$gb_left]};
	my ($le_right_bf_cvm, $re_right_bf_cvm) = @{$cp_curr_bds_blks[$gb_right]};

	unless ($re_left_bf_cvm + 1 == $le_right_bf_cvm) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] IMPOSSIBLE!!: Somehow, \$gb_left (=$gb_left) and \$gb_right (=$gb_right) do NOT TOUCH each other, EVEN AFTER performing 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right' (w/ \$re_left_bf_cvm = $re_left_bf_cvm, \$le_right_bf_cvm = $le_right_bf_cvm): $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# End of "(Added on 2019/07/25.) (Post M2)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)"  ###


		### MAIN 3: Prepare for the "(incomplete) vertical-merge" by creating information on the NEW gap-block resulting from the move,
		###         as well as the complements of the set of MSA rows affected by $gb_right. ###
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (6))  ###

	my ($rows_affd_by_gb1, $rows_affd_by_gb2) = @{$rows_affd_by_blks}[$gb1, $gb2];

	my ($rows_affd_by_new_gb, $cmpl_rows_affd_by_gb2, $cmpl_rows_affd_by_gb1) = diff_nmarray (@{$rows_affd_by_gb1}, @{$rows_affd_by_gb2});
	unless (defined $cmpl_rows_affd_by_gb1) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine 'diff_nmarray' FAILED on the rows affected by \$gb1 (=$gb1) and \$gb2 (=$gb2): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $cnct_rows_affd_by_new_gb = join (',', @{$rows_affd_by_new_gb});

	my $br_up_or_lw_new = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_new_gb};
	unless (defined $br_up_or_lw_new) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] IMPOSSIBLE!!: NO branch & upper/lower information is ASSOCIATED with the set of MSA rows, ($cnct_rows_affd_by_new_gb), that is supposed to be affected by the gap-block resulting from the \"vertical merge\" of $gb1 and $gb2: $! !\n\n";
	    return (0, undef, undef, undef);
	}
	my ($br_new, $up_or_lw_new) = @{$br_up_or_lw_new};

	my @exnodes_affd_by_new_gb = ();
	foreach my $row (@{$rows_affd_by_new_gb}) { push @exnodes_affd_by_new_gb, $nodeids_in_msa->[$row]; }
	if (@exnodes_affd_by_new_gb >1) { @exnodes_affd_by_new_gb = sort {$a <=> $b} @exnodes_affd_by_new_gb; }
	my $cnct_exnodes_affd_by_new_gb = join (',', @exnodes_affd_by_new_gb);

	my $re_new_gb = $re_left_bf_cvm; # In any case, the right-end of the new gap-block MUST BE identical to that of $gb_left.
	my $le_new_gb = $re_new_gb - $size_rplcd  +1;

	my @info_new_gb = ($br_new, $up_or_lw_new, $le_new_gb, $re_new_gb, $cnct_exnodes_affd_by_new_gb, $cnct_rows_affd_by_new_gb, \@exnodes_affd_by_new_gb, $rows_affd_by_new_gb);

    
   if ($TEST_LEVEL_cvmrg_gbp >3) {
	print STDOUT "\n[ Complements of the sets of MSA rows affected by each of \$gb1 (=$gb1) and \$gb2 (=$gb2): ]\n\n";
	print STDOUT '@{$cmpl_rows_affd_by_gb1} = (', join (',', @{$cmpl_rows_affd_by_gb1}), ") .\n";
	print STDOUT '@{$cmpl_rows_affd_by_gb2} = (', join (',', @{$cmpl_rows_affd_by_gb2}), ") .\n";
	print STDOUT "\n\n";

	print STDOUT "[ Information on the NEW Gap-Block supposed to result from the \"(incomplete) vertical-merge\": ]\n\n";
	print STDOUT '($br_new, $up_or_lw_new) = ', "($br_new, $up_or_lw_new) ,\n";
	print STDOUT '($le_new_gb, $re_new_gb) = ', "($le_new_gb, $re_new_gb) ,\n";
	print STDOUT '@exnodes_affd_by_new_gb = ', "($cnct_exnodes_affd_by_new_gb) ,\n";
	print STDOUT '@{$rows_affd_by_new_gb} = ', "($cnct_rows_affd_by_new_gb) .\n";
	print STDOUT "\n\n";
	
   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"


		# MAIN 3': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". #
		# (ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).) #


	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_cvmrg_gbp > 3) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp > 3) {...}"

		# End of "(ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (6))"  ###


		### MAIN 5: Perform the "(incomplete) vertical-merge",
		### by moving the complement (residue-block) of $gb_right to the left ###
		### (but the association of index vs gap-blocks remains unchanged).
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (7))  ###

	my $cmpl_rows_affd_by_right = ($gb_right == $gb1) ?
			 $cmpl_rows_affd_by_gb1 : $cmpl_rows_affd_by_gb2 ;

	my @new_curr_set_clms = ();
	my @old_gblks_w_spec_lb_af_cvm = ();
	my @old_gblks_w_spec_rb_af_cvm = ();
	my $indx_new_clm = -1;

	for (my $c=0; $c <$le_new_gb; $c++) {
	    $indx_new_clm++;

	    $new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];
	}

	my $ct_rmvd_null_clms = 0;
	my $indx_mgd_clm = $le_right_bf_cvm -1;
	for (my $k=0; $k <$size_rplcd; $k++) {
	    $indx_new_clm++;
	    $indx_mgd_clm++;

	    my @new_clm = @{$cp_curr_set_clms[$indx_new_clm]};
	    my $mgd_clm = $cp_curr_set_clms[$indx_mgd_clm];
	    foreach my $row (@{$cmpl_rows_affd_by_right}) { $new_clm[$row] = $mgd_clm->[$row]; }
	    $new_curr_set_clms[$indx_new_clm] = \@new_clm;

	    my @mgd_gblks_w_spec_lb = (@{$cp_curr_gblks_w_spec_lb[$indx_new_clm]}, @{$cp_curr_gblks_w_spec_lb[$indx_mgd_clm]});
	    my @mgd_gblks_w_spec_rb = (@{$cp_curr_gblks_w_spec_rb[$indx_new_clm]}, @{$cp_curr_gblks_w_spec_rb[$indx_mgd_clm]});
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = \@mgd_gblks_w_spec_lb;
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = \@mgd_gblks_w_spec_rb;

	    $ct_rmvd_null_clms++;
	}

	for (my $c=$indx_mgd_clm+1; $c < $ct_in_clms; $c++) {
	    $indx_new_clm++;

	    $new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];
	}

	my $ct_out_clms = @new_curr_set_clms;

	my @bds_old_gblks_af_cvm = ();
	for (my $gb=0; $gb <$ct_in_gbs; $gb++) { $bds_old_gblks_af_cvm[$gb] = []; }
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    foreach my $gb (@{$old_gblks_w_spec_lb_af_cvm[$c]}) { $bds_old_gblks_af_cvm[$gb]->[0] = $c; }
	    foreach my $gb (@{$old_gblks_w_spec_rb_af_cvm[$c]}) { $bds_old_gblks_af_cvm[$gb]->[1] = $c; }
	}


		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (7))"  ###


		### MAIN 6: Create intermediate lists of boundaries & other properties
		###	of gap-blocks AFTER the "(incomplete) vertical merge",
		###	& the correspondence between old and intermediate indices of gap-blocks. ###
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25-26. (8))  ###

	my @bds_gblks_af_cvm = @bds_old_gblks_af_cvm;
	$bds_gblks_af_cvm[$gb_rmvd_kpt] = ($gb_rmvd_kpt == $gb_left) ?
		 [$le_left_bf_cvm, $re_left_bf_cvm - $size_rplcd] : [$le_right_bf_cvm, $re_right_bf_cvm - $size_rplcd] ;
	##my $re_new_gb = $re_left_bf_cvm; # In any case, the right-end of the new gap-block MUST BE identical to that of $gb_left.
	##my $le_new_gb = $re_new_gb - $size_rplcd  +1;

	my @list_gblocks_af_cvm = @list_old_gblocks;
	$list_gblocks_af_cvm[$gb_rplcd] = \@info_new_gb;
	for (my $gb = 0; $gb < $ct_in_gbs; $gb++) { @{$list_gblocks_af_cvm[$gb]}[2,3] = @{$bds_gblks_af_cvm[$gb]}; }
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	my @gblock_sizes_af_cvm = @{$gblock_sizes};
	$gblock_sizes_af_cvm[$gb_rmvd_kpt] -= $size_rplcd;

   if ($TEST_LEVEL_cvmrg_gbp > 5) {

	print STDOUT "\n", '@gblock_sizes_af_cvm = (', join (', ', @gblock_sizes_af_cvm), ") .\n";
	print STDOUT "\n\n";

	print STDOUT '( Contents of @list_gblocks_af_cvm (, including @bds_gblks_af_cvm): )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_gblocks_af_cvm[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp > 5) {...}"



	my @med_indices_old_gbs = my @old_indices_med_gbs = ();
	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) {

	    if ($ogb == $gb_rmvd_kpt) {
		if ($if_complete) {
		    $med_indices_old_gbs[$ogb] = undef;
		} else {
		    my $mgb = $ogb;
		    $med_indices_old_gbs[$ogb] = $mgb;
		    $old_indices_med_gbs[$mgb] = $ogb;
		}
		next;
	    }

	    my $mgb;
	    if ($ogb < $gb_rmvd_kpt) {
		$mgb = $ogb;
	    } else { # if ($gb_rmvd_kpt < $ogb)
		$mgb = ($if_complete) ? $ogb-1 : $ogb ;
	    }

	    $med_indices_old_gbs[$ogb] = $mgb;
	    $old_indices_med_gbs[$mgb] = $ogb;
	}

	my $ct_out_gbs = @old_indices_med_gbs;

	my @bds_med_gblks = @bds_gblks_af_cvm[@old_indices_med_gbs];
	my @list_med_gblocks = @list_gblocks_af_cvm[@old_indices_med_gbs];
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25-26. (8))"  ###


		### MAIN 7: Create the correspondence between the old, intermediate & new indices of the gap-blocks,
		###    by Re-Sorting the gap-blocks according to their affected MSA rows and their horizontal positions. #
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (9))  ###

	my @med_indices_new_gbs = sort { cmp_gblocks3 (@{$list_med_gblocks[$a]}, @{$list_med_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1); # REVISED for a BETTER ordering (on 2019/07/26). #
	##my @med_indices_new_gbs = sort { cmp_gblocks2 (@{$list_med_gblocks[$a]}, @{$list_med_gblocks[$b]}, 2, 6) } (0 .. $ct_out_gbs -1);

	my @new_indices_med_gbs = my @new_indices_old_gbs = my @old_indices_new_gbs = ();
	for (my $ngb=0; $ngb < $ct_out_gbs; $ngb++) {
	    my $mgb = $med_indices_new_gbs[$ngb];
	    my $ogb = $old_indices_med_gbs[$mgb];

	    $new_indices_med_gbs[$mgb] = $ngb;
	    $new_indices_old_gbs[$ogb] = $ngb;
	    $old_indices_new_gbs[$ngb] = $ogb;
	}
	if ($if_complete) { $new_indices_old_gbs[$gb_rmvd_kpt] = undef; }


   if ($TEST_LEVEL_cvmrg_gbp >2) {

	print STDOUT "\n[ Relationships between old, intermediate, and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_med_indices_old_gbs = my @out_old_indices_med_gbs = ();
	my @out_new_indices_med_gbs = my @out_med_indices_new_gbs = ();
	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ([\@med_indices_old_gbs, \@out_med_indices_old_gbs],
			  [\@old_indices_med_gbs, \@out_old_indices_med_gbs],
			  [\@new_indices_med_gbs, \@out_new_indices_med_gbs],
			  [\@med_indices_new_gbs, \@out_med_indices_new_gbs],
			  [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@med_indices_old_gbs = (', join (', ', @out_med_indices_old_gbs), ") ,\n";
	print STDOUT '@new_indices_med_gbs = (', join (', ', @out_new_indices_med_gbs), ") ,\n";
	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@med_indices_new_gbs = (', join (', ', @out_med_indices_new_gbs), ") ,\n";
	print STDOUT '@old_indices_med_gbs = (', join (', ', @out_old_indices_med_gbs), ") ,\n";
	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >2) {...}"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (9))"  ###


		### MAIN 8: Create new lists of boundaries & other properties
		###	of gap-blocks AFTER the "(incomplete) vertical merge",
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (10))  ###

	my @new_curr_bds_blks = @bds_gblks_af_cvm[@old_indices_new_gbs];
	my @new_gblock_sizes = @gblock_sizes_af_cvm[@old_indices_new_gbs];
	my @new_list_gblocks = @list_gblocks_af_cvm[@old_indices_new_gbs];
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c < $ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$gb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $gb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $gb;
	}

	my ($new_inter_block_relations) = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine 'mk_inter_block_relations' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

	my @new_inter_block_relations = @{$new_inter_block_relations};

	my @rows_affd_by_blks_af_cvm = @cp_rows_affd_by_blks;
	$rows_affd_by_blks_af_cvm[$gb_rplcd] = $rows_affd_by_new_gb;
	my @rows_affd_by_new_blks = @rows_affd_by_blks_af_cvm[@old_indices_new_gbs];

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (10))"  ###


		### MAIN 9: Create the new coordinate frame VIA 3 STEPS.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (11))  ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

		# (Step 1:) Create a coordinate frame using the current gap-block configuration. #

	my ($new_orgs_gb_coords0, $new_bds_gb_coords0, $new_gb_coords_wd_frm0)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm0) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks immediately after the \"(incomplete) vertical merge\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately after the \"(incomplete) vertical merge\" ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);


   if ($TEST_LEVEL_cvmrg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER \"(incomplete) Vertical-Merge\" and BEFORE UNDOING the displacements: ]\n\n";

	my @new_curr_bl_coords0 = @{$new_orgs_gb_coords0};
	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords0, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >5) {...}"


		# (Step 2:) Attempt to "undo" the displacements of the gap-blocks. #

	my @new_curr_bl_coords1 = @{$new_orgs_gb_coords0};

	my $ct_coord_chngs = 0;

	if (!($if_complete) and ($bl_sbj == $gb_rmvd_kpt)) {
	    my $new_indx_sbj = $new_indices_old_gbs[$bl_sbj];
	    my $new_coord_sbj0 = $new_curr_bl_coords1[$new_indx_sbj];
	    my $new_coord_sbj = $new_coord_sbj0 - $incr_coord_sbj;
	    my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_sbj]};
	    if ($new_coord_sbj < $coord_le_wd0) {
		$new_coord_sbj = $coord_le_wd0;
	    } elsif ($coord_re_wd0 < $new_coord_sbj) {
		$new_coord_sbj = $coord_re_wd0;
	    }
	    if ($new_coord_sbj0 != $new_coord_sbj) {
		$new_curr_bl_coords1[$new_indx_sbj] = $new_coord_sbj;
		$ct_coord_chngs++;
	    }
	} # End of "if (!($if_complete) and ($bl_sbj == $gb_rmvd_kpt)) {...}"

	foreach my $gb2incr_coord ($gb2incr_coord_left, $gb2incr_coord_right) {
	    foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord}) {
		my $incr_coord = $gb2incr_coord->{$gb};
		my $new_indx_gb = $new_indices_old_gbs[$gb];
		my $new_coord_gb0 = $new_curr_bl_coords1[$new_indx_gb];
		my $new_coord_gb = $new_coord_gb0 - $incr_coord;
		my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_gb]};
		if ($new_coord_gb < $coord_le_wd0) {
		    $new_coord_gb = $coord_le_wd0;
		} elsif ($coord_re_wd0 < $new_coord_gb) {
		    $new_coord_gb = $coord_re_wd0;
		}
		if ($new_coord_gb0 != $new_coord_gb) {
		    $new_curr_bl_coords1[$new_indx_gb] = $new_coord_gb;
		    $ct_coord_chngs++;
		}
	    } # End of "foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord}) {...}"

	} # End of "foreach my $gb2incr_coord ($gb2incr_coord_left, $gb2incr_coord_right) {...}"

    if ($TEST_LEVEL_cvmrg_gbp > 5) {

	print STDOUT "\n", '( Contents of @{$new_orgs_gb_coords0} (for "coord0"), @new_curr_bl_coords1 (for "coord1") & @{$new_gb_coords_wd_frm0} (for "coord_le0_wd" and "coord_re0_wd"): )', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'coord0', 'coord1', 'coord_le0_wd', 'coord_re0_wd'), "\n\n";

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $new_orgs_gb_coords0->[$gb], $new_curr_bl_coords1[$gb], @{$new_gb_coords_wd_frm0->[$gb]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cvmrg_gbp > 5) {...}"


	if ($ct_coord_chngs>0) {

	    my ($flag_ssh_p2p, $tot_incr_ct_null_clms_ssh_p2p)
	     = serial_shifts_point2point (
		@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_bl_coords1, @{$new_gb_coords_wd_frm0}, 
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
			);
	    unless ((defined $flag_ssh_p2p) and ($flag_ssh_p2p == 0)) { # A BUG fixed on 2019/07/27. #
		my $out_flag = (defined $flag_ssh_p2p) ? $flag_ssh_p2p : 'undef';
	    ##unless ($flag_ssh_p2p) {
		print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine 'serial_shifts_point2point' FAILED on the new set of gap-blocks (w/ \$flag_ssh_p2p = $out_flag): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_point2point', Ended on the set of new gap-blocks (w/ \$tot_incr_ct_null_clms_ssh_p2p = $tot_incr_ct_null_clms_ssh_p2p) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

	} else {

   (print STDOUT "\n[INFORMATION 'cvmerge_gblock_pair'] The 'serial_shifts_point2point' will NOT be performed, because \$ct_coord_chngs = 0 !!...\n\n") if ($TEST_LEVEL_cvmrg_gbp > 1);

	} # End of "if ($ct_coord_chngs>0) {...} else {...}"


		# (Step 3:) Create the FINAL coordinate frame. #

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cvmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER UNDOING the displacements of the new gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER UNDOING the displacements of the new gap-blocks ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (11))"  ###


   if ($TEST_LEVEL_cvmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"(incomplete) Vertical-Merge\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp >3) {...}"


   if ($TEST_LEVEL_cvmrg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"(incomplete) Vertical-Merge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp > 3) {...}"


		### MAIN 10: Finally Update  @new_list_gblocks. #
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    @{$new_list_gblocks[$gb]}[2,3] = @{$new_curr_bds_blks[$gb]};
		 # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}

   if ($TEST_LEVEL_cvmrg_gbp > 3) {

	print STDOUT '[ FINAL Contents of @new_list_gblocks: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvmrg_gbp > 3) {...}"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))"  ###

	
    (print STDOUT "\n< Subroutine 'cvmerge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$if_complete = $if_complete, \$gb_rplcd = $gb_rplcd, \$gb_rmvd_kpt = $gb_rmvd_kpt, #{removed null-columns}=$ct_rmvd_null_clms). >\n\n") if ($TEST_LEVEL_cvmrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	##@new_list_gblocks);
	\@new_list_gblocks); # A BUG fixed on 2020/05/13.

}	# End of the subroutine, 'cvmerge_gblock_pair'. #


# vmerge_gblock_pair:
#	This subroutine performs the "(Complete) vertical-merge" of a specified pair of
#	sibling gap-blocks of the same length.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cvsplit_gblock'.
#	[CREATION STARTED on 2019/10/10, ENDED on 2019/10/13.]
#
# cvsplit_gblock:
#	This subroutine "vertically split"s the COMPLEMENTARY RESIDUE-BLOCK of the subject gap-block into a pair of RESIDUE-BLOCKs that are the "effective children" of the subject's complementary residue-block.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' & 'vsplit_gblock', both are above.
#	[CREATION STARTED on 2019/08/25, ENDED on 2019/08/26.]
#
#	[NOTE] Before the "vertical-split", it DISPLACES gap-blocks that are "vertically including" the subject gap-block to the left of the subject.
#	Then, it attempts to UNDO the displacements after the "vertical-split".
#	HOWEVER, those gap-blocks that are vertically EQUIVALENT to the subject will remain displaced.
#
#	[REVISED on 2019/08/28 (1)]:
#	Now, the subroutine first DISPLACES "vertically including" gap-blocks TO THE RIGHT,
#	instead of to the left, as before the revision;
#	this is expected to "UNDO" some moves of the flanking residues more securely.
#	[=> This REVISION was UNDONE immediately!!]
#
#	[REVISED on 2019/08/28 (2)]:
#	Now, the subroutine does NOT ATTEMPT to UNDO the DISPLACEMENT
#	if the displaced gap-block is an effective parent of the subject gap-block. 
#	(This way, superfluous "shift"s may be suppressed when undoing the horizontal overlaps.)
#
# vsplit_gblock:
#	This subroutine "vertically split"s the subject gap-block into a pair of gap-blocks that are the "effective children" of the subject,
#	then, it "shift"s one of the resulting child blocks to that they will NOT be nested.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' above.
#	[CREATION STARTED on 2019/08/22, ENDED on 2019/08/23.]
#	
# [ INPUTS: ($gb_sbj, \@info_ceffch1, \@info_ceffch2, $char_abs, # ADDED $char_abs on 2019/08/26. ] ... for the predecessor, 'cvsplit_gblock'.
#
# INPUTS: ($gb_sbj_fxd, $gb_sbj_shftd, $if_shift_to_right, # ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/10). #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#	), where
#
#		# MODIFIED in creation of 'vmerge_gblock_pair'. (HD1b) #
#	$gb_sbj_fxd and $gb_sbj_shftd are the (old) indices of the "subject" gap-blocks to be 'vmerge'd;
#	The coordinate of the former will be fixed,
#	whereas the latter will be "shift"ed before performing the 'vmerge';
#
#	$if_shift_to_right = 1 if $gb_sbj_shftd will be "shift"ed to right,
#			   = 0 if it will be "shift"ed to left;
#
#		# END of "MODIFIED in creation of 'vmerge_gblock_pair'. (HD1b)" #
#
#	[[
#		# MODIFIED in creation of 'vsplit_gblock'. (HD1b) #
#	$gb_sbj is the index of the gap-block to be "vertically split";
#
#		# ENd of "MODIFIED in creation of 'vsplit_gblock'. (HD1b)" #
#
#		# MODIFIED in creation of 'cvsplit_gblock'. (HD1b) #
#	@info_ceffch1/2 = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)
#	provides information on the 1st/2nd "complementary-effective child" of $gb_sbj, where
#	the "complementary-child" residue-block's complementary GAP-block is on the $up_or_lw_cch side of the branch with ID $br_cch ,
#	@cexnodes_affd_by_cch stores the indices of external nodes affected by the complementary GAP-block of the "complementary-child" residue-block,
#	and @crows_affd_by_cch stores the indices of the MSA rows affected by the complementary GAP-block of the "complementary-child" residue-block.
#	[Typically, \@info_ceffch1 and \@info_ceffch2 are parts of the output of 'list_cvsplit_candidates' above];
#		# END of "MODIFIED in creation of 'cvsplit_gblock'. (HD1b)" #
#	]] ... for the predecessor, 'cvsplit_gblock'.
#
#	$char_abs is the character resresenting an "absence" state; # (Added on 2019/08/26). #
#
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#			# (Added on 2019/07/26.) #
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#			# End of "(Added on 2019/07/26.)" #
#
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 2)" #
#	
#
# [OUTPUTS: ($flag_end, $gb_cch1_NEW, $gb_cch2_NEW, \%gb2incr_coord_1way_dsplc, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2a) # ] ... for the predecessor, 'cvsplit_gblock'. #
#
# OUTPUTS: ($flag_end, $incr_coord_sbj_shftd, \%gb2incr_coord_1way_dsplc, $gb_mgd_NEW, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, # MODIFIED in creation of 'vmerge_gblock_pair' (on 2019/10/10). (HD2a) #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; # IMPORTED from the header of 'csplit_gblock'. #
#
#			# MODIFIED in creation of 'vmerge_gblock_pair' (on 2019/010/10). (HD2b) #
#	$incr_coord_sbj_shftd is the increment of the coordinate of $gb_sbj_shftd
#	necessary for its 'vmerge' with $gb_sbj_fxd;
#
#	$gb_mgd_NEW is the NEW index of the newly created gap-block resulting from the 'vmerge';
#
#			# MODIFIED in creation of 'vmerge_gblock_pair' (on 2019/010/10). (HD2b) #
#
#	[[
#			# MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2b) #
#	$gb_cch1_NEW and $gb_cch2_NEW are the NEW indices of the COMPLEMENTARY GAP-BLOCKS of the 1st and 2nd child, respectively, of the RESIDUE-BLOCK COMPLEMENTARY to $gb_sbj;
#			# End of "MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2b)" #
#	]] ... for the predecessor, 'cvsplit_gblock'.
#
#			# MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2c) #
#	%gb2incr_coord_1way_dsplc = ($indx_gb => {increment of the block coordinate resulting from the DISPLACEMENT prior to the "complementary vertical split"}, ...) for the gap-blocks whose displacement canNOT be undone; (Such gap-blocks MUST BE vertically identical to $gb_sbj); $indx_gb SHOULD BE the index in the NEW gap-block set;
#			# End of "MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2c)" #
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vertical-split";
#	More precisely, the gap-block that had been the $gb th before the "vertical-split"
#	has been assigned $new_indices_old_gbs[$gb] after the "vertical-split"
#	(, and $new_indices_old_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vertical-split";
#	More precisely, the gap-block that is the $gb th after the "vertical-split"
#	had been assigned $old_indices_new_gbs[$gb] before the "vertical-split";
#
#			# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c) #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "vertical-split";
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 5) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 5)" #
#
#
# [ sub cvsplit_gblock ($\@\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb_sbj, $info_ceffch1, $info_ceffch2, $char_abs, ] ... for the predecessor, 'cvsplit_gblock'.
#
sub vmerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb_sbj_fxd, $gb_sbj_shftd, $if_shift_to_right, # Modified in creation of 'vmerge_gblock_pair' (on 2019/10/10). #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq
	) = @_;  # Added $char_abs on 2019/08/26. # => Removed in creation of 'vmerge_gblock_pair' (on 2019/10/10).

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	unless (($ct_in_clms>0) and ($ct_in_gbs>0)) { # Added on 2019/08/26.
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] ILLEGAL INPUT: (\$ct_in_clms, \$ct_in_gbs) = ($ct_in_clms, $ct_in_gbs): $! !\n";
	    return (0, undef, undef, undef);
	}

	my $ct_rows = @{$curr_set_clms->[0]}; # Added in creation of 'cvsplit_gblock' (on 2019/08/26). #


			## MODIFIED in creation of 'vmerge_gblock_pair' (on 2019/10/11). (Top) ###

	my $size_sbj_fxd = $gblock_sizes->[$gb_sbj_fxd];
	my $size_sbj_shftd = $gblock_sizes->[$gb_sbj_shftd];

	unless ($size_sbj_fxd == $size_sbj_shftd) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] ILLEGAL INPUT: The size of \$gb_sbj_fxd(=$gb_sbj_fxd), $size_sbj_fxd, DIFFERS from that of \$gb_sbj_shftd(=$gb_sbj_shftd), $size_sbj_shftd: $! !\n";
	    return (0, undef, undef, undef);
	}

	my $rel_shftd2fxd = $inter_block_relations->[$gb_sbj_fxd][$gb_sbj_shftd];
	unless ($rel_shftd2fxd eq 'S') {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] ILLEGAL INPUT: \$gb_sbj_fxd(=$gb_sbj_fxd) and \$gb_sbj_shftd(=$gb_sbj_shftd) are NOT Siblings to each other, but in '$rel_shftd2fxd' relationship: $! !\n";
	    return (0, undef, undef, undef);
	}

	my $gb_coord_fxd = $curr_bl_coords->[$gb_sbj_fxd];
	my $gb_coord_shftd = $curr_bl_coords->[$gb_sbj_shftd];

	my ($lbd_fxd, $rbd_fxd) = @{$curr_bds_blks->[$gb_sbj_fxd]};
	my ($lbd_shftd, $rbd_shftd) = @{$curr_bds_blks->[$gb_sbj_shftd]};

	my ($br_sbj_fxd, $up_or_lw_sbj_fxd) = @{$list_gblocks0->[$gb_sbj_fxd]}[$indx_br, $indx_u_or_d];
	my ($br_sbj_shftd, $up_or_lw_sbj_shftd) = @{$list_gblocks0->[$gb_sbj_shftd]}[$indx_br, $indx_u_or_d];

	my $rows_affd_by_fxd = $rows_affd_by_blks->[$gb_sbj_fxd];
	my $rows_affd_by_shftd = $rows_affd_by_blks->[$gb_sbj_shftd];

	my $cnct_rows_affd_by_fxd = join (',', @{$rows_affd_by_fxd});
	my $cnct_rows_affd_by_shftd = join (',', @{$rows_affd_by_shftd});

			## END of "MODIFIED in creation of 'vmerge_gblock_pair' (on 2019/10/11). (Top)" ###

    (print STDOUT "< Subroutine 'vmerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; The \"Subject\" Gap-Blocks to be \"Vertically Merge\"d: (\$gb_sbj_fxd, \$gb_sbj_shftd) = ($gb_sbj_fxd, $gb_sbj_shftd); \$if_shift_to_right = $if_shift_to_right) >\n\n") if ($TEST_LEVEL_vmg_gbp > 0);


   if ($TEST_LEVEL_vmg_gbp >1) { # (Top) # ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/11). #

	print STDOUT '[ More Information on the subject gap-blocks: ]', "\n\n";

	print STDOUT '($gb_coord_fxd, $gb_coord_shftd) = ', "($gb_coord_fxd, $gb_coord_shftd) .\n\n";
	print STDOUT "(lbd, rbd)_fxd   = ($lbd_fxd, $rbd_fxd) ,\n";
	print STDOUT "(lbd, rbd)_shftd = ($lbd_shftd, $rbd_shftd) .\n\n";

	print STDOUT "(br, up_or_lw)_fxd   = ($br_sbj_fxd, $up_or_lw_sbj_fxd) ,\n";
	print STDOUT "(br, up_or_lw)_shftd = ($br_sbj_shftd, $up_or_lw_sbj_shftd) .\n\n";

	print STDOUT '{MSA rows affected by $gb_sbj_fxd}   = ', "($cnct_rows_affd_by_fxd) ,\n";
	print STDOUT '{MSA rows affected by $gb_sbj_shftd} = ', "($cnct_rows_affd_by_shftd) .\n\n";

	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >1) { # (Top) # ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/11). # ...}"


=For the predecessor, 'cvsplit_gblock'. (Top)
	my ($br_cch1, $up_or_lw_cch1, $cexnodes_affd_by_cch1, $crows_affd_by_cch1) = @{$info_ceffch1};
	my ($br_cch2, $up_or_lw_cch2, $cexnodes_affd_by_cch2, $crows_affd_by_cch2) = @{$info_ceffch2};

	# @info_ceffch1/2 = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)

   if ($TEST_LEVEL_vmg_gbp >1) { # (Top) #

	my $cnct_crows_cch1 = join (',', @{$crows_affd_by_cch1});
	my $cnct_crows_cch2 = join (',', @{$crows_affd_by_cch2});

	my $cnct_cexnodes_cch1 = join (',', @{$cexnodes_affd_by_cch1});
	my $cnct_cexnodes_cch2 = join (',', @{$cexnodes_affd_by_cch2});

	print STDOUT "\n[ Information on the Child Blocks to be Created: ]\n\n";

	print STDOUT join ("\t", 'Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(external,nodes,affected)', '[MSA,rows,affected]'), "\n\n";

	print STDOUT join ("\t", 'Ch_1', $br_cch1, $up_or_lw_cch1, "($cnct_cexnodes_cch1)", "[$cnct_crows_cch1]"), "\n";
	print STDOUT join ("\t", 'Ch_2', $br_cch2, $up_or_lw_cch2, "($cnct_cexnodes_cch2)", "[$cnct_crows_cch2]"), "\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >1) { # (Top) # ...}"

=cut # END of "For the predecessor, 'cvsplit_gblock'. (Top)"


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_vmg_gbp > 5) {

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Compelemntary-Vertical Split\" ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Vertical Merge\" ]\n\n"; # Modified in creation of 'vmerge_gblock_pair'. #

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"



   if ($TEST_LEVEL_vmg_gbp >3) {

	##print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Complementary-Vertical Split\" ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Vertical Merge\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >3) {...}"


		### MAIN 1: Create information on the newly created gap-block as a result of the 'vertical merge'.
		### (ADDED in creation of 'vmerge_gblock_pair' on 2019/10/11.) (M1) ###

	my $rows_affd_by_mgd = union_nmarray (@{$rows_affd_by_fxd}, @{$rows_affd_by_shftd});
	unless (defined $rows_affd_by_mgd) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'union_nmarray' FAILED on the MSA rows affected by \$gb_sbj_fxd(=$gb_sbj_fxd) and \$gb_sbj_shftd(=$gb_sbj_shftd): $! !\n";
	    return (0, undef, undef, undef);
	}

	my @exnodes_affd_by_mgd = ();
	foreach my $row (@{$rows_affd_by_mgd}) { push @exnodes_affd_by_mgd, $nodeids_in_msa->[$row]; }
	if (@exnodes_affd_by_mgd >1) {
	    @exnodes_affd_by_mgd = sort {$a <=> $b} @exnodes_affd_by_mgd;
	}

	my $cnct_rows_affd_by_mgd = join (',', @{$rows_affd_by_mgd});
	my $cnct_exnodes_affd_by_mgd = join (',', @exnodes_affd_by_mgd);

	my $br_up_or_lw_mgd = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_mgd};
	unless (defined $br_up_or_lw_mgd) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] NO branch-up_or_lw pairs are ASSOCIATED with the set of MSA rows for the \"vertical-merge\"d gap-block, i.e., ($cnct_rows_affd_by_mgd): $! !\n";
	    return (0, undef, undef, undef);
	}

	my ($br_mgd, $up_or_lw_mgd) = @{$br_up_or_lw_mgd};

   if ($TEST_LEVEL_vmg_gbp >1) { # (M1) #

	print STDOUT "[ Main Attributes of the Gap-Block that will result from the \"Vertical Merge\": ]\n\n";
	print STDOUT '@{$rows_affd_by_mgd} = ', "($cnct_rows_affd_by_mgd) .\n\n";
	print STDOUT '@exnodes_affd_by_mgd = ', "($cnct_exnodes_affd_by_mgd) .\n\n";

	print STDOUT '($br_mgd, $up_or_lw_mgd) = ', "($br_mgd, $up_or_lw_mgd) .\n\n";

	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >1) { # (M1) # ...}"

		### END of "(ADDED in creation of 'vmerge_gblock_pair' on 2019/10/11.) (M1)" ###


		### [ MAIN 2: Displace the gap-blocks either vertically including $gb_sbj
		### or in the 'ONX' relationships with $gb_sbj from
		### the region spanned by $gb_sbj to its left.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###
		### (=> IMPORTED (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 2) ### ] ... for the predecessor, 'cvsplit_gblock'.

		### MAIN 2: Displace the gap-blocks either vertically including $gb_sbj_fxd
		### or in the 'ONX' relationships with $gb_sbj_fxd from
		### the region spanned by $gb_sbj_fxd to its left/right (depending on $if_shift_to_right).
		### (MODIFIED in creation of 'vmerge_gblock_pair' on 2019/10/11.)  (M2)  ###

	##my %ignored = (); # MOVED from above and MODIFIED in creation of 'cvsplit_gblock' (on 2019/0826). #
	my %ignored = ($gb_sbj_fxd => 1, $gb_sbj_shftd => 1); # MODIFIED in creation of 'vmerge_gblock_pair'. #

	my ($gb2incr_coord_dsplc1, $totct_nec_shifts_dsplc1);

	if ($if_shift_to_right) { # (M2) # $gb_sbj_shftd will be shifted to the right.

	    my $left_limit = $lbd_fxd;

	    ($gb2incr_coord_dsplc1, $totct_nec_shifts_dsplc1)
	    = displace_including_and_onx_to_right ($gb_sbj_fxd, $left_limit, 
		@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		@cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_dsplc1) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_right', FAILED on the \$gb_sbj_fxd (=$gb_sbj_fxd) th gap-block and \$left_limit = $left_limit: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_right', Ended on the \$gb_sbj_fxd (=$gb_sbj_fxd) th gap-block and \$left_limit = $left_limit (w/ \$totct_nec_shifts_dsplc1 = $totct_nec_shifts_dsplc1) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_right' on \$gb_sbj_fxd (=$gb_sbj_fxd) and \$left_limit = $left_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

	} else { # $gb_sbj_shftd will be shifted to the left.

	    my $right_limit = $rbd_fxd;

	    ($gb2incr_coord_dsplc1, $totct_nec_shifts_dsplc1)
	    = displace_including_and_onx_to_left ($gb_sbj_fxd, $right_limit, 
		@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		@cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_dsplc1) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_left', FAILED on the \$gb_sbj_fxd (=$gb_sbj_fxd) th gap-block and \$right_limit = $right_limit: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_left', Ended on the \$gb_sbj_fxd (=$gb_sbj_fxd) th gap-block and \$right_limit = $right_limit (w/ \$totct_nec_shifts_dsplc1 = $totct_nec_shifts_dsplc1) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_left' on \$gb_sbj_fxd (=$gb_sbj_fxd) and \$right_limit = $right_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

	} # End of "if ($if_shift_to_right) { # (M2) # ...} else {...}". #


	($lbd_fxd, $rbd_fxd) = @{$cp_curr_bds_blks[$gb_sbj_fxd]}; # Update the boundaries of $gb_sbj_fxd. 

   if ($TEST_LEVEL_vmg_gbp >2) { # (M2-A) #

	print STDOUT "\n";
	print STDOUT "Updated boundaries of \$gb_sbj_fxd (=$gb_sbj_fxd): (\$lbd_fxd, \$rbd_fxd) = ($lbd_fxd, $rbd_fxd) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) { # (M2-A) # ...}"


   if ($TEST_LEVEL_vmg_gbp >3) { # (M2-B) #

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_dsplc1}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_dsplc1} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_dsplc1}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_dsplc1->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >3) { # (M2-B) # ...}"

		### (MODIFIED in creation of 'vmerge_gblock_pair' on 2019/10/11.)  (M2)  ###


		### (ADDED in creation of 'vmerge_gblock_pair' on 2019/10/11.) (M3 & 5) ###

		### MAIN 3: "Shift" $gb_sbj_shftd so that its left/right-end will be aligned with that of $gb_sbj_fxd. ###

	my %skipped = (); # A dummy. (ADDED in DEBUGGING on 2019/10/13.)

	my $incr_coord_sbj_shftd;

	if ($if_shift_to_right) { # (M3) # $gb_sbj_shftd will be shifted to the right.

	    my $bd_curr = $lbd_shftd;
	    my $bd_goal = $lbd_fxd;

	    my $ct_shifts = ct_necessary_shifts ($bd_curr, $bd_goal, $gb_sbj_shftd, @cp_curr_bds_blks, @{$inter_block_relations}, %skipped);
	    unless (defined $ct_shifts) {
		print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'ct_necessary_shifts', FAILED on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and (\$bd_curr, \$bd_goal) = ($bd_curr, $bd_goal): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'ct_necessary_shifts', Ended on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and (\$bd_curr, \$bd_goal) = ($bd_curr, $bd_goal)  (w/ \$ct_shifts = $ct_shifts) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	    $incr_coord_sbj_shftd = $ct_shifts;

	} else { # $gb_sbj_shftd will be shifted to the left.

	    my $bd_curr = $rbd_shftd;
	    my $bd_goal = $rbd_fxd;

	    my $ct_shifts = ct_necessary_shifts ($bd_curr, $bd_goal, $gb_sbj_shftd, @cp_curr_bds_blks, @{$inter_block_relations}, %skipped);
	    unless (defined $ct_shifts) {
		print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'ct_necessary_shifts', FAILED on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and (\$bd_curr, \$bd_goal) = ($bd_curr, $bd_goal): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'ct_necessary_shifts', Ended on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and (\$bd_curr, \$bd_goal) = ($bd_curr, $bd_goal)  (w/ \$ct_shifts = $ct_shifts) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	    $incr_coord_sbj_shftd = - $ct_shifts;

	} # End of "if ($if_shift_to_right) { # (M3) # ...} else {...}". #

   if ($TEST_LEVEL_vmg_gbp >2) { # (M3-A) #

	print STDOUT "\n", '$incr_coord_sbj_shftd = ', $incr_coord_sbj_shftd, " .\n\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) { # (M3-A) # ...}"


	my ($flag_ssh, $incr_ct_null_clms_ssh)
	   = serial_shifts_sgl_gblock (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		$gb_sbj_shftd, $incr_coord_sbj_shftd, @cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes});
	unless ((defined $flag_ssh) and ($flag_ssh == 0)) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock' FAILED on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and with \$incr_coord_sbj_shftd = $incr_coord_sbj_shftd: $! !\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and \$incr_coord_sbj_shftd = $incr_coord_sbj_shftd (w/ \$incr_ct_null_clms_ssh = $incr_ct_null_clms_ssh) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


	($lbd_shftd, $rbd_shftd) = @{$cp_curr_bds_blks[$gb_sbj_shftd]}; # Update the boundaries of $gb_sbj_shftd. (1)

   if ($TEST_LEVEL_vmg_gbp >2) { # (M3-B) #

	print STDOUT "\n";
	print STDOUT "Updated boundaries of \$gb_sbj_shftd (=$gb_sbj_shftd) (1): (\$lbd_shftd, \$rbd_shftd) = ($lbd_shftd, $rbd_shftd) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) { # (M3-B) # ...}"


   if ($TEST_LEVEL_vmg_gbp >5) { # (M3-C) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'serial_shifts_sgl_gblock' on \$gb_sbj_shftd (=$gb_sbj_shftd) and \$incr_coord_sbj_shftd (=$incr_coord_sbj_shftd) ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) { # (M3-C) # ...}"


		### MAIN 5: Displace the gap-blocks either vertically including $gb_sbj_shftd
		### or in the 'ONX' relationships with $gb_sbj_shftd from
		### the region spanned by $gb_sbj_shftd to its left/right (depending on $if_shift_to_right).

	my ($gb2incr_coord_dsplc2, $totct_nec_shifts_dsplc2);

	if ($if_shift_to_right) { # (M5) # $gb_sbj_shftd has been shifted to the right.

	    my $left_limit = $lbd_fxd;

	    ($gb2incr_coord_dsplc2, $totct_nec_shifts_dsplc2)
	    = displace_including_and_onx_to_right ($gb_sbj_shftd, $left_limit, 
		@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		@cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_dsplc2) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_right', FAILED on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and \$left_limit = $left_limit: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_right', Ended on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and \$left_limit = $left_limit (w/ \$totct_nec_shifts_dsplc2 = $totct_nec_shifts_dsplc2) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	} else { # $gb_sbj_shftd has been shifted to the left.

	    my $right_limit = $rbd_fxd;

	    ($gb2incr_coord_dsplc2, $totct_nec_shifts_dsplc2)
	    = displace_including_and_onx_to_left ($gb_sbj_shftd, $right_limit, 
		@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		@cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_dsplc2) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'displace_including_and_onx_to_left', FAILED on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and \$right_limit = $right_limit: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_left', Ended on the \$gb_sbj_shftd (=$gb_sbj_shftd) th gap-block and \$right_limit = $right_limit (w/ \$totct_nec_shifts_dsplc2 = $totct_nec_shifts_dsplc2) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	} # End of "if ($if_shift_to_right) { # (M5) # ...} else {...}". #


	($lbd_shftd, $rbd_shftd) = @{$cp_curr_bds_blks[$gb_sbj_shftd]}; # Update the boundaries of $gb_sbj_shftd. (2)

   if ($TEST_LEVEL_vmg_gbp >2) { # (M5-A) #

	print STDOUT "\n";
	print STDOUT "Updated boundaries of \$gb_sbj_shftd (=$gb_sbj_shftd) (2): (\$lbd_shftd, \$rbd_shftd) = ($lbd_shftd, $rbd_shftd) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) { # (M5-A) # ...}"


   if ($TEST_LEVEL_vmg_gbp >3) { # (M5-B) #

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_dsplc2}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_dsplc2} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_dsplc2}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_dsplc2->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >3) { # (M5-B) # ...}"


   if ($TEST_LEVEL_vmg_gbp >5) { # (M5-C) # (Added on 2019/10/12) #

	print STDOUT "\n[ Gap-Block Configurations AFTER ALL the Physical Preparations for the \"Vertical Merge\" of \$gb_sbj_fxd (=$gb_sbj_fxd) and \$gb_sbj_shftd (=$gb_sbj_shftd): ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) { # (M5-C) # (Added on 2019/10/12) # ...}"

		### END of "(ADDED in creation of 'vmerge_gblock_pair' on 2019/10/11.) (M3 & 5)" ### 


		# MAIN 5post: Check whether the displacements produced the desired result. # 
		# (ADDED & MOVED (w/ modifications) from below in creation of 'vmerge_gblock_pair' (on 2019/10/11). (M5post) #

		# (A) Check the coordinate consistency between $gb_sbj_fxd and $gb_sbj_shftd. #

	unless (($lbd_fxd == $lbd_shftd) and ($rbd_fxd == $rbd_shftd)) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] IMPOSSIBLE!!: Somehow, the boundaries of the \$gb_sbj_fxd(=$gb_sbj_fxd) th gap-block, given by ($lbd_fxd, $rbd_fxd), does NOT COINCIDE with thouse of the \$gb_sbj_shftd(=$gb_sbj_shftd) th gap-block, given by ($lbd_shftd, $rbd_shftd), EVEN AFTER 'displace_including_and_onx_to_left': $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# (B) Check the block-size and the actual span of $gb_sbj_fxd. #

	my $curr_span_fxd = $rbd_fxd - $lbd_fxd +1;
	unless ($curr_span_fxd == $size_sbj_fxd) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] IMPOSSIBLE!!: Somehow, 'displace_including_and_onx_to_left' did NOT displace ENOUGH gap-blocks to EQUATE the SIZE, \$size_sbj_fxd (=$size_sbj_fxd), of the \$gb_sbj_fxd (=$gb_sbj_fxd) th gap-block to its SPAN, \$rbd_fxd - \$lbd_fxd +1 (=$curr_span_fxd) : $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# END of "(ADDED & MOVED (w/ modifications) from below in creation of 'vmerge_gblock_pair' (on 2019/10/11). (M5post)" #


		### MAIN 6: Formally Perform the "Vertical Merge" of $gb_sbj_fxd and $gb_sbj_shftd,
		###	by just Replacing the entries of $gb_sbj_fxd & $gb_sbj_shftd
		###	with the entry of the "vertical-merge"d gap-blocks already characterized in MAIN 1.
		### (ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M6) ###

		# (A) Create the correspondence between the gap-blocks before and after the 'vmerge'. #

	my @old_indices_intm_gbs = my @intm_indices_old_gbs = ();
	my $igb = 0;
	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) {
	    if (($ogb == $gb_sbj_fxd) or ($ogb == $gb_sbj_shftd)) {
		$intm_indices_old_gbs[$ogb] = undef;
		next;
	    }
	    $old_indices_intm_gbs[$igb] = $ogb;
	    $intm_indices_old_gbs[$ogb] = $igb;
	    $igb++;
	}

=DISABLED on 2019/10/11.
	##$old_indices_intm_gbs[$igb] = $ct_in_gbs;
	##$intm_indices_old_gbs[$ct_in_gbs] = $igb;

	##my $ct_intm_gbs = @old_indices_intm_gbs;
=cut # END of "DISABLED on 2019/10/11."


		# (B) Create an old list of gap-blocks. #
		# (IMPORTED (w/ modifications) from "MAIN 3" for 'cvsplit_gblock' (on 2019/10/12).) #

	##my @list_old_gblocks_plus_mgd = ();
	my @list_old_gblocks = (); # Modified on 2019/10/13.
	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) {
	    my ($br, $up_or_lw) = @{$list_gblocks0->[$ogb]}[$indx_br, $indx_u_or_d];
	    my ($lbd, $rbd) = @{$cp_curr_bds_blks[$ogb]};

	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$ogb]};

	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes >1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);

	    ##$list_old_gblocks_plus_mgd[$ogb] = [$br, $up_or_lw, $lbd, $rbd, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];
	    $list_old_gblocks[$ogb] = [$br, $up_or_lw, $lbd, $rbd, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows]; # Modified on 2019/10/13.
	}

	##$list_old_gblocks_plus_mgd[$ct_in_gbs] = \@info_gb_mgd; # DISABLED on 2019/10/13.

   if ($TEST_LEVEL_vmg_gbp > 3) {
	##print STDOUT "\n", '( Contents of @list_old_gblocks_plus_mgd: )', "\n\n";
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	##for (my $gb=0; $gb<=$ct_in_gbs; $gb++) {
	for (my $gb=0; $gb< $ct_in_gbs; $gb++) {
	    ##my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks_plus_mgd[$gb]}[0 .. 5];    
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 3" for 'cvsplit_gblock' (on 2019/10/12).)" #


		# (C) Create an entry for the new gap-block resulting from the 'vertical merge'. #

	my @info_gb_mgd = ($br_mgd, $up_or_lw_mgd, $lbd_fxd, $rbd_fxd, $cnct_exnodes_affd_by_mgd, $cnct_rows_affd_by_mgd, \@exnodes_affd_by_mgd, $rows_affd_by_mgd);

			# ADDED in DEBUGGING (on 2019/10/13). #
   if ($TEST_LEVEL_vmg_gbp > 3) {
	print STDOUT "\n", '( Contents of @info_gb_mgd: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	print STDOUT join ("\t", '[$gb_mgd_intm]', @info_gb_mgd[0 .. 5]), "\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"
			# END of "ADDED in DEBUGGING (on 2019/10/13)." #


		# (D) Perform the 'vertical-merge' while keeping the order among the existing gap-blocks. #
		# (IMPORTED (w/ modifications) from "MAIN 5" for 'cvsplit_gblock' (on 2019/10/12).) #

	my $gb_mgd_intm = @old_indices_intm_gbs;

	##my @intm_list_gblocks = @list_old_gblocks_plus_mgd[@old_indices_intm_gbs];
	my @intm_list_gblocks = @list_old_gblocks[@old_indices_intm_gbs]; # Modified in revision (on 2019/10/13). #
	$intm_list_gblocks[$gb_mgd_intm] = \@info_gb_mgd;

	my $ct_intm_gbs = @intm_list_gblocks;

	my @intm_rows_affd_by_blks = my @intm_curr_bds_blks = ();
	for (my $igb=0; $igb < $ct_intm_gbs; $igb++) {
	    my ($lbd, $rbd, $rows_affd) = @{$intm_list_gblocks[$igb]}[2,3,7];
	    $intm_rows_affd_by_blks[$igb] = $rows_affd;
	    $intm_curr_bds_blks[$igb] = [$lbd, $rbd];
	}

	$old_indices_intm_gbs[$gb_mgd_intm] = undef; # Finish the @old_indices_intm_gbs. #

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n";

	print STDOUT "\$ct_intm_gbs = $ct_intm_gbs .\n";
	print STDOUT "\$gb_mgd_intm = $gb_mgd_intm .\n\n";

	print STDOUT '( Contents of @intm_list_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_intm_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 5" for 'cvsplit_gblock' (on 2019/10/12).)" #


		# (IMPORTED (w/ modifications) from "MAIN 6" for 'cvsplit_gblock' (on 2019/10/12).) #
   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Relationships between old and intermediate indices of gap-blocks (#{gap-blocks}_intm=$ct_intm_gbs) : ]\n\n";

	my @out_intm_indices_old_gbs = my @out_old_indices_intm_gbs = ();

	foreach my $pair ( [\@intm_indices_old_gbs, \@out_intm_indices_old_gbs],
			  [\@old_indices_intm_gbs, \@out_old_indices_intm_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@intm_indices_old_gbs = (', join (', ', @out_intm_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_intm_gbs = (', join (', ', @out_old_indices_intm_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 6" for 'cvsplit_gblock' (on 2019/10/12).)" #

		### END of "(ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M6)" ###


		### MAIN 7: Create the New set of Gap-BLocks,
		###	   by Rearranging the gap-blocks according to the prescribed rules.
		### (ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M7) ###

		# (A) Create the correspondences between the old, intermediate, and new indices of gap-blocks. #
		# (IMPORTED (w/ modifications) from "MAIN 6" for 'cvsplit_gblock' (on 2019/10/12).) #

	my @intm_indices_new_gbs = sort { cmp_gblocks3 (@{$intm_list_gblocks[$a]}, @{$intm_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_intm_gbs -1) ;

	my $ct_out_gbs = $ct_intm_gbs;
	##my $ct_out_gbs = @intm_indices_new_gbs;

	my @new_indices_intm_gbs = ();
	my @new_indices_old_gbs = my @old_indices_new_gbs = ();

	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) {
	    my $igb = $intm_indices_new_gbs[$ngb];
	    $new_indices_intm_gbs[$igb] = $ngb;
	    my $ogb = $old_indices_intm_gbs[$igb];
	    $old_indices_new_gbs[$ngb] = $ogb;
	    if (defined $ogb) { $new_indices_old_gbs[$ogb] = $ngb; }
	}
	$new_indices_old_gbs[$gb_sbj_fxd] = undef;
	$new_indices_old_gbs[$gb_sbj_shftd] = undef;

	my $gb_mgd_NEW = $new_indices_intm_gbs[$gb_mgd_intm];

   if ($TEST_LEVEL_vmg_gbp >2) {

	print STDOUT "\n\$gb_mgd_new = $gb_mgd_NEW .\n\n";

	print STDOUT "\n[ Relationships between old (including intermediate) and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();
	my @out_new_indices_intm_gbs = my @out_intm_indices_new_gbs = ();

	foreach my $pair ([\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs],
			  [\@new_indices_intm_gbs, \@out_new_indices_intm_gbs],
			  [\@intm_indices_new_gbs, \@out_intm_indices_new_gbs]
		) {
	    my ($indices_gbs, $out_indices_gbs) = @{$pair};
	    foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_intm_gbs = (', join (', ', @out_new_indices_intm_gbs), ") .\n\n";
	print STDOUT '@intm_indices_new_gbs = (', join (', ', @out_intm_indices_new_gbs), ") .\n\n";

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";
	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 6" for 'cvsplit_gblock' (on 2019/10/12).)" #


		# (B) Create the data-sets characterizing the new set of gap-blocks. #
		# (IMPORTED (w/ modifications) from "MAIN 8" for 'cvsplit_gblock' (on 2019/10/12).) #

	my @new_curr_set_clms = @cp_curr_set_clms;

	my $ct_out_clms = @new_curr_set_clms;

	my @new_list_gblocks = @intm_list_gblocks[@intm_indices_new_gbs];

   if ($TEST_LEVEL_vmg_gbp > 3) { # (IMPORTED (w/ modifications) from "MAIN 12" for 'cvsplit_gblock' (on 2019/10/12).) #

	print STDOUT '[ Semi-FINAL Contents of @new_list_gblocks: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) { # (IMPORTED (w/ modifications) from "MAIN 12" for 'cvsplit_gblock' (on 2019/10/12).) # ...}"


	my @new_curr_bds_blks = @intm_curr_bds_blks[@intm_indices_new_gbs];

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c =0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $ngb;
	}

	my @rows_affd_by_new_blks = @intm_rows_affd_by_blks[@intm_indices_new_gbs];

	my ($new_inter_block_relations) = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'mk_inter_block_relations' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_inter_block_relations = @{$new_inter_block_relations};

	my ($new_gblock_sizes) = mk_list_gblock_sizes (@new_curr_bds_blks, 0, 1, @new_inter_block_relations) ;
	unless (defined $new_gblock_sizes) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'mk_list_gblock_sizes' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_list_gblock_sizes', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_gblock_sizes = @{$new_gblock_sizes};

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

		# End of "(IMPORTED (w/ modifications) from "MAIN 8" for 'cvsplit_gblock' (on 2019/10/12).)" #

		# (C) "Tentatively" create the coordinate frame for the new set of gap-blocks. #
		# (IMPORTED (w/ modifications) from "MAIN 9" for 'cvsplit_gblock' (on 2019/10/12).) #

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords0, $new_bds_gb_coords0, $new_gb_coords_wd_frm0)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm0) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new, SORTED gap-blocks AFTER the \"Vertical Merge\" of \$gb_sbj_fxd (=$gb_sbj_fxd) and \$gb_sbj_shftd (=$gb_sbj_shftd): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new, SORTED gap-blocks AFTER the \"Vertical Merge\" of \$gb_sbj_fxd (=$gb_sbj_fxd) and \$gb_sbj_shftd (=$gb_sbj_shftd) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER \"Vertical Merge\" and BEFORE UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER \"Vertical Merge\" and BEFORE UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @{$new_bds_gb_coords0}, @{$new_gb_coords_wd_frm0}, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 9" for 'cvsplit_gblock' (on 2019/10/12).)" #

		### END of "(ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M7)" ###


		### MAIN 8: UNDO the displacements (performed in MAIN 2 & MAIN 5), if possible.
		### (ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M8) ###

		# (A) Sort the coordinate increments due to the displacements into
		#	"UNDO"-able ones and UN-"UNDO"-able ones.
		# (IMPORTED (w/ modifications) from "MAIN 2" for 'cvsplit_gblock' (on 2019/10/12).) (M8) #

		# End of "(IMPORTED (w/ modifications) from "MAIN 2" for 'cvsplit_gblock' (on 2019/10/12).) (M8)" #

	my $rels_w_mgd = $new_inter_block_relations[$gb_mgd_NEW];

	my %ngb2incr_coord_1way_dsplc;
	my %ngb2incr_coord_round_trip_dsplc;

	foreach my $ogb2incr_coord_dsplc ($gb2incr_coord_dsplc1, $gb2incr_coord_dsplc2) {

	    foreach my $ogb (sort {$a <=> $b} keys %{$ogb2incr_coord_dsplc}) {

		##my $incr_coord = $ogb2incr_coord_left_dsplc->{$ogb};
		my $incr_coord = $ogb2incr_coord_dsplc->{$ogb}; # CORRECTED in DEBUGGING (on 2019/10/13).

		my $ngb = $new_indices_old_gbs[$ogb];

		my $rel = $rels_w_mgd->[$ngb];
		if (($rel eq '<') or ($rel eq '<(pa)')) {
		
		    if (defined $ngb2incr_coord_round_trip_dsplc{$ngb}) {
			$ngb2incr_coord_round_trip_dsplc{$ngb} += $incr_coord;
		    } else {
			$ngb2incr_coord_round_trip_dsplc{$ngb} = $incr_coord;
		    }

		} else {
		    if (defined $ngb2incr_coord_1way_dsplc{$ngb}) {
			$ngb2incr_coord_1way_dsplc{$ngb} += $incr_coord;
		    } else {
			$ngb2incr_coord_1way_dsplc{$ngb} = $incr_coord;
		    }
		}

	    } # End of "foreach my $ogb (sort {$a <=> $b} keys %{$ogb2incr_coord_dsplc}) {...}"

	} # End of "foreach my $gb2incr_coord_dsplc ($gb2incr_coord_dsplc1, $gb2incr_coord_dsplc2) {...}"


   if ($TEST_LEVEL_vmg_gbp >3) {

	print STDOUT "\n", '[ Contents of NEWLY CREATED %ngb2incr_coord_1way_dsplc & %ngb2incr_coord_round_trip_dsplc: ]', "\n\n";

	print STDOUT '%ngb2incr_coord_1way_dsplc = (', "\n";
	foreach my $ngb (sort {$a <=> $b} keys %ngb2incr_coord_1way_dsplc) {
	    print STDOUT "\t", $ngb, ' => ', $ngb2incr_coord_1way_dsplc{$ngb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

	print STDOUT '%ngb2incr_coord_round_trip_dsplc = (', "\n";
	foreach my $ngb (sort {$a <=> $b} keys %ngb2incr_coord_round_trip_dsplc) {
	    print STDOUT "\t", $ngb, ' => ', $ngb2incr_coord_round_trip_dsplc{$ngb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >3) {...}"


		# (IMPORTED (w/ modifications) from "MAIN 10" for 'cvsplit_gblock' (on 2019/10/12).) (M8) #

		# (B) Attempt to "undo" the displacements of the gap-blocks. #

	my @new_curr_bl_coords1 = @{$new_orgs_gb_coords0};

	my $ct_coord_chngs = 0;

	foreach my $ngb (sort {$a <=> $b} keys %ngb2incr_coord_round_trip_dsplc) {
		my $incr_coord = $ngb2incr_coord_round_trip_dsplc{$ngb};

		##my $new_indx_gb = $new_indices_old_gbs[$gb];
		##my $new_coord_gb0 = $new_curr_bl_coords1[$new_indx_gb];
		my $new_coord_gb0 = $new_curr_bl_coords1[$ngb]; # CORRECTED in REVISION (on 2019/10/13). #

		my $new_coord_gb = $new_coord_gb0 - $incr_coord;

		##my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_gb]};
		my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$ngb]}; # CORRECTED in REVISION (on 2019/10/13). #

		if ($new_coord_gb < $coord_le_wd0) {
		    $new_coord_gb = $coord_le_wd0;
		} elsif ($coord_re_wd0 < $new_coord_gb) {
		    $new_coord_gb = $coord_re_wd0;
		}
		if ($new_coord_gb0 != $new_coord_gb) {
		    ##$new_curr_bl_coords1[$new_indx_gb] = $new_coord_gb;
		    $new_curr_bl_coords1[$ngb] = $new_coord_gb; # CORRECTED in REVISION (on 2019/10/13). #

		    $ct_coord_chngs++;
		}
	} # End of "foreach my $ngb (sort {$a <=> $b} keys %ngb2incr_coord_round_trip_dsplc) {...}"


    if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n", '( Contents of @{$new_orgs_gb_coords0} (for "coord0"), @new_curr_bl_coords1 (for "coord1") & @{$new_gb_coords_wd_frm0} (for "coord_le0_wd" and "coord_re0_wd"): )', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'coord0', 'coord1', 'coord_le0_wd', 'coord_re0_wd'), "\n\n";

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $new_orgs_gb_coords0->[$gb], $new_curr_bl_coords1[$gb], @{$new_gb_coords_wd_frm0->[$gb]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"


	if ($ct_coord_chngs>0) {

	    my ($flag_ssh_p2p, $tot_incr_ct_null_clms_ssh_p2p)
	     = serial_shifts_point2point (
		@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_bl_coords1, @{$new_gb_coords_wd_frm0}, 
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
			);
	    unless ((defined $flag_ssh_p2p) and ($flag_ssh_p2p == 0)) { # A BUG fixed on 2019/07/27. #
		my $out_flag = (defined $flag_ssh_p2p) ? $flag_ssh_p2p : 'undef';
	    ##unless ($flag_ssh_p2p) {
		print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'serial_shifts_point2point' FAILED on the new set of gap-blocks (w/ \$flag_ssh_p2p = $out_flag): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_point2point', Ended on the set of new gap-blocks (w/ \$tot_incr_ct_null_clms_ssh_p2p = $tot_incr_ct_null_clms_ssh_p2p) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	} else {

   (print STDOUT "\n[INFORMATION 'vmerge_gblock_pair'] The 'serial_shifts_point2point' will NOT be performed, because \$ct_coord_chngs = 0 !!...\n\n") if ($TEST_LEVEL_vmg_gbp > 1);

	} # End of "if ($ct_coord_chngs>0) {...} else {...}"


		# (C) Create the FINAL coordinate frame. #

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cvspl_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER UNDOING the displacements of the new gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER UNDOING the displacements of the new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};


   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		# End of "(IMPORTED (w/ modifications) from "MAIN 10" for 'cvsplit_gblock' (on 2019/10/12).) (M8) #

		### END of "(ADDED in creation of 'vmerge_gblock_pair' (on 2019/10/12).) (M8)" ###


		### MAIN 9: UNDO the horizontal overlaps among those gap-blocks vertically equivalent or complementary to $gb_mgd_NEW.
		### (IMPORTED (w/ modifications) from "MAIN 11" for 'cvsplit_gblock' (on 2019/10/12).) (M9) ###

	my ($fl_undo, $ngb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_mgd_NEW,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)) on the cofiguration of gap-blocks after \"Vertical Merge\" of \$gb_sbj_fxd (=$gb_sbj_fxd) (OLD) and \$gb_sbj_shftd (=$gb_sbj_shftd) (OLD) into \$gb_mgd_NEW (=$gb_mgd_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)), Ended (w/ \$fl_undo = $fl_undo) on the configuration of gap-blocks after the \"Vertical Merge\" of \$gb_sbj_fxd (=$gb_sbj_fxd) (OLD) and \$gb_sbj_shftd (=$gb_sbj_shftd) (OLD) into \$gb_mgd_NEW (=$gb_mgd_NEW) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


   if ($TEST_LEVEL_vmg_gbp > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)): ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$ngb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $ngb (sort {$a <=> $b} keys %{$ngb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $ngb, ' => ', $ngb2incr_bl_coord_undo->{$ngb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

	if ($fl_undo < 0) {
	    print STDOUT "[CAUTION 'vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

	my %ngb2incr_bl_coord_undo = %{$ngb2incr_bl_coord_undo};

		### End of "(IMPORTED (w/ modifications) from "MAIN 11" for 'cvsplit_gblock' (on 2019/10/12).) (M9)" ###


=For the predecessor, 'cvsplit_gblock'. (M2post)
	    # Update the boundaries of $gb_sbj. # (Added in creation of 'vmerge_gblock_pair' (on 2019/08/26).) #
	($lbd_sbj, $rbd_sbj) = @{$cp_curr_bds_blks[$gb_sbj]};


   if ($TEST_LEVEL_vmg_gbp > 3) {
	print STDOUT "\nUpdated Boundaries of \$gb_sbj (=$gb_sbj): (\$lbd_sbj, \$rbd_sbj) = ($lbd_sbj, $rbd_sbj) .\n\n\n";
   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"


		# MAIN 2post: Check whether the displacement produced the desired result. # 
		# (MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2post)) #

	my $curr_span_sbj = $rbd_sbj - $lbd_sbj +1;
	unless ($curr_span_sbj == $size_sbj) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] IMPOSSIBLE!!: Somehow, 'displace_including_and_onx_to_left' did NOT displace ENOUGH gap-blocks to EQUATE the SIZE, \$size_sbj (=$size_sbj), of the \$gb_sbj (=$gb_sbj) th gap-block to its SPAN, \$rbd_sbj - \$lbd_sbj +1 (=$curr_span_sbj) : $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# End of "(MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2post))" #
=cut # END of "For the predecessor, 'cvsplit_gblock'. (M2post)"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###
		### (=> IMPORTED (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 2)" ###


=For the predecessor, 'cvsplit_gblock'. (M3)

		# MAIN 3: Create the list of gap-blocks BEFORE the "complementary vertical split". #
		# [ MAIN 1: Create the list of gap-blocks BEFORE the "vertical split". # ] ... for the predecessor, 'vsplit_gblock'. #
		# [ MAIN 3': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". # ] ... for the predecessor, 'cvmerge_gblock_pair'. #
		# (ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).) #


	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_vmg_gbp > 3) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

		# End of "(ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (6))"  ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M3)"


=For the predecessor, 'cvsplit_gblock'. (M5 & M5post)
		### MAIN 5: Create intermediate lists of properties
		###   of gap-blocks (before &) AFTER the "complementary vertical split".
		### [ MAIN 2: Create intermediate lists of properties
		###   of gap-blocks (before &) AFTER the "vertical split". ] ... for the predecessor 'vsplit_gblock'.
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2) ###
		### (=> Carried over (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 5) ###

	my @indices_cch_intm = ($ct_in_gbs, $ct_in_gbs+1);

	my @set_info_cchildren = ();
	foreach my $src_info_cch ($info_ceffch1, $info_ceffch2) {
	    my  ($br_cch, $up_or_lw_cch, $cexnodes_affd_by_cch, $crows_affd_by_cch) = @{$src_info_cch};
	    my $cnct_cexnodes_cch = join (',', @{$cexnodes_affd_by_cch});
	    my $cnct_crows_cch = join (',', @{$crows_affd_by_cch});

	    push @set_info_cchildren, [$br_cch, $up_or_lw_cch, $lbd_sbj, $rbd_sbj, $cnct_cexnodes_cch, $cnct_crows_cch, $cexnodes_affd_by_cch, $crows_affd_by_cch];
	}

	my @intm_list_gblocks = @list_old_gblocks;
	@intm_list_gblocks[@indices_cch_intm] = @set_info_cchildren;

	my @intm_rows_affd_by_blks = @cp_rows_affd_by_blks;
	@intm_rows_affd_by_blks[@indices_cch_intm] = ($crows_affd_by_cch1, $crows_affd_by_cch2);

	my $ct_intm_gbs = @intm_list_gblocks; # Modified in creation of 'cvsplit_gblock' (MAIN 5)


   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n";

	print STDOUT '@indices_cch_intm = (', join (', ', @indices_cch_intm), ") .\n\n";


	print STDOUT '( Contents of @intm_list_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_intm_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2) ###
		### (=> Carried over (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 5)" ###

		### MAIN 5post: Create inter-block relations between the gap-blocks in the intermediate set. #
		### (ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 5post) ###

	my ($intm_inter_block_relations) = mk_inter_block_relations (@intm_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $intm_inter_block_relations) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'mk_inter_block_relations' FAILED on the intermediate set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the intermediate set of  gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my ($intm_rels_w_cch1, $intm_rels_w_cch2) = @{$intm_inter_block_relations}[@indices_cch_intm];
	my @old_rels_w_cch1 = @{$intm_rels_w_cch1}[0 .. $ct_in_gbs-1];
	my @old_rels_w_cch2 = @{$intm_rels_w_cch2}[0 .. $ct_in_gbs-1];

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n[ Relations of the old gap-blocks with complementary-effective children 1 & 2: ]\n\n";

	print STDOUT '@old_rels_w_cch1 = (', join (', ', @old_rels_w_cch1), ") ;\n";
	print STDOUT '@old_rels_w_cch2 = (', join (', ', @old_rels_w_cch2), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		### End of "(ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 5post)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M5 & M5post)"

=For the predecessor, 'cvsplit_gblock'. (M6)

		### [ MAIN 3: Create the correspondence between the old (including intermediate) & new indices of the gap-blocks, ] ... for the predecessor, 'vsplit_gblock'.
		### MAIN 6: Create the correspondence between the old (including intermediate) & new indices of the gap-blocks,
		###    by Re-Sorting the gap-blocks according to their affected MSA rows and their horizontal positions. #
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3) ###
		### (=> Carried over to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 6) ###

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$intm_list_gblocks[$a]}, @{$intm_list_gblocks[$b]}, 2, 3, 6) } (0 .. $gb_sbj-1, $gb_sbj+1 .. $ct_intm_gbs -1) ;

	my $ct_out_gbs = @old_indices_new_gbs;

	my @new_indices_old_gbs = ();
	$new_indices_old_gbs[$gb_sbj] = undef;

	for (my $ngb=0; $ngb < $ct_out_gbs; $ngb++) {

	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	}

   if ($TEST_LEVEL_vmg_gbp >2) {

	print STDOUT "\n[ Relationships between old (including intermediate) and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >2) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3) ###
		### (=> Carried over to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 6)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M6)"

=For the predecessor, 'cvsplit_gblock'. (M7)

		### MAIN 7: Actually Perform the "Complementary-Vertical Split",
		###	    while keeping the OLD gap-block indices.
		### (ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 7) ###

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

	my ($gb_cch1_OLD, $gb_cch2_OLD) = @indices_cch_intm;

	my ($crows_affd_by_both, $ex_affd_by_cch1, $ex_affd_by_cch2) = diff_nmarray (@{$crows_affd_by_cch1}, @{$crows_affd_by_cch2});
	unless (defined $crows_affd_by_both) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'diff_nmarray' FAILED on ", '@{$crows_affd_by_cch1} & @{$crows_affd_by_cch2}', ": $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my @set_clms_in_cch1 = my @set_clms_in_cch2 = ();
	my @gblks_w_lb_in_cch1 = my @gblks_w_rb_in_cch1 = ();
	my @gblks_w_lb_in_cch2 = my @gblks_w_rb_in_cch2 = ();

	for (my $c = $lbd_sbj; $c <= $rbd_sbj; $c++) { # Main for-loop (over columns within the boundaries of $gb_sbj). #

	    my @clm_in_cch1 = @{$cp_curr_set_clms[$c]};
	    my @clm_in_cch2 = @null_clm;

	    foreach my $row (@{$ex_affd_by_cch1}) {
		my $tmp = $clm_in_cch1[$row];
		$clm_in_cch1[$row] = $clm_in_cch2[$row];
		$clm_in_cch2[$row] = $tmp;
	    } # End of "foreach my $row (...) {...}"

	    push @set_clms_in_cch1, \@clm_in_cch1;
	    push @set_clms_in_cch2, \@clm_in_cch2;

	    my @gblks_w_lb_eq_c_in_cch1 = my @gblks_w_lb_eq_c_in_cch2 = ();
	    my @gblks_w_rb_eq_c_in_cch1 = my @gblks_w_rb_eq_c_in_cch2 = ();

	    foreach my $trio (
		[$cp_curr_gblks_w_spec_lb[$c], \@gblks_w_lb_eq_c_in_cch1, \@gblks_w_lb_eq_c_in_cch2],
		[$cp_curr_gblks_w_spec_rb[$c], \@gblks_w_rb_eq_c_in_cch1, \@gblks_w_rb_eq_c_in_cch2],
		) {
		my ($src_gblks, $bds_into_cch1, $bds_into_cch2) = @{$trio};
		foreach my $gb (@{$src_gblks}) {
		    if ($gb == $gb_sbj) {
			push @{$bds_into_cch1}, $gb_cch1_OLD;
			push @{$bds_into_cch2}, $gb_cch2_OLD;
			next;
		    }
		    my $rel1 = $old_rels_w_cch1[$gb];
		    my $rel2 = $old_rels_w_cch2[$gb];
		    if (($rel1 eq '>') or ($rel1 eq '>(ch)')) {
			push @{$bds_into_cch2}, $gb;

		    } elsif (($rel2 eq '>') or ($rel2 eq '>(ch)')) {
			push @{$bds_into_cch1}, $gb;
		    } else {
			print STDOUT "[ERROR 'vmerge_gblock_pair'] The \$gb (=$gb) th gap-block (in the old set), whose boundary (=$c) is within the boundaries of \$gb_sbj (=$gb_sbj), is NOT INCLUDED in EITHER \$gb_cch1_OLD (=$gb_cch1_OLD) OR \$gb_cch2_OLD (=$gb_cch2_OLD): $! !\n\n";
			return (0, undef, undef, undef);
		    }
		}
	    } # End of "foreach my $trio (...) {...}"

	    push @gblks_w_lb_in_cch1, \@gblks_w_lb_eq_c_in_cch1;
	    push @gblks_w_lb_in_cch2, \@gblks_w_lb_eq_c_in_cch2;
	    push @gblks_w_rb_in_cch1, \@gblks_w_rb_eq_c_in_cch1;
	    push @gblks_w_rb_in_cch2, \@gblks_w_rb_eq_c_in_cch2;

	} # End of the Main for-loop (over columns within the boundaries of $gb_sbj). #

	# my @set_clms_in_cch1 = my @set_clms_in_cch2 = ();
	# my @gblks_w_lb_in_cch1 = my @gblks_w_rb_in_cch1 = ();
	# my @gblks_w_lb_in_cch2 = my @gblks_w_rb_in_cch2 = ();

	my @new_curr_set_clms = (
		@cp_curr_set_clms[0 .. $lbd_sbj-1],
		@set_clms_in_cch1,
		@set_clms_in_cch2,
		@cp_curr_set_clms[$rbd_sbj+1 .. $ct_in_clms-1]
		);
	my @intm_gblks_w_spec_lb_af_cvs = (
		@cp_curr_gblks_w_spec_lb[0 .. $lbd_sbj-1],
		@gblks_w_lb_in_cch1,
		@gblks_w_lb_in_cch2,
		@cp_curr_gblks_w_spec_lb[$rbd_sbj+1 .. $ct_in_clms-1]
		);
	my @intm_gblks_w_spec_rb_af_cvs = (
		@cp_curr_gblks_w_spec_rb[0 .. $lbd_sbj-1],
		@gblks_w_rb_in_cch1,
		@gblks_w_rb_in_cch2,
		@cp_curr_gblks_w_spec_rb[$rbd_sbj+1 .. $ct_in_clms-1]
		);

	my $ct_out_clms = @new_curr_set_clms;

	my @intm_curr_bds_blks = ();
	for (my $gb=0; $gb <$ct_intm_gbs; $gb++) { $intm_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$ct_out_clms; $c++) {
	    foreach my $gb (@{$intm_gblks_w_spec_lb_af_cvs[$c]}) {
		$intm_curr_bds_blks[$gb]->[0] = $c;
	    }
	    foreach my $gb (@{$intm_gblks_w_spec_rb_af_cvs[$c]}) {
		$intm_curr_bds_blks[$gb]->[1] = $c;
	    }
	}

		### END of "(ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 7)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M7)"


=For the predecessor, 'cvsplit_gblock'. (M8)

		### MAIN 8: Create the New Data-Sets, by using the new indices of the gap-blocks.
		### [ MAIN 5: Create new lists of boundaries & other properties
		###	of gap-blocks AFTER the "Vertical Split",  ] ... for the 
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 8) ###

	my ($gb_cch1_NEW, $gb_cch2_NEW) = @new_indices_old_gbs[@indices_cch_intm]; # MODIFIED in creation of 'cvsplit_gblock'. (MAIN 8a) #
	##my ($gb_ch1_NEW, $gb_ch2_NEW) = @new_indices_old_gbs[@indices_ch_intm];

	##my @new_curr_set_clms = @cp_curr_set_clms;
	##my $ct_out_clms = @new_curr_set_clms;

	my @new_curr_bds_blks = @intm_curr_bds_blks[@old_indices_new_gbs];

			# Modified in creation of 'cvsplit_gblock'. (MAIN 8b) #
	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs[$ngb];
	    my @info_ngb = @{$intm_list_gblocks[$ogb]};
	    @info_ngb[2,3] = @{$new_curr_bds_blks[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_ngb;
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}
	##my @new_list_gblocks = @intm_list_gblocks[@old_indices_new_gbs];
			# End of "Modified in creation of 'cvsplit_gblock'. (MAIN 8b)" #

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c =0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $ngb;
	}

	my @rows_affd_by_new_blks = @intm_rows_affd_by_blks[@old_indices_new_gbs];

	my ($new_inter_block_relations) = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'mk_inter_block_relations' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_inter_block_relations = @{$new_inter_block_relations};

	my ($new_gblock_sizes) = mk_list_gblock_sizes (@new_curr_bds_blks, 0, 1, @new_inter_block_relations) ;
	unless (defined $new_gblock_sizes) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'mk_list_gblock_sizes' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_list_gblock_sizes', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_gblock_sizes = @{$new_gblock_sizes};

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 8)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M8)"


=For the predecessor, 'cvsplit_gblock'. (M9)
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 6 & 7 & 8) ###

		### MAIN 9: Create the new (penultimate) coordinate frame.
		### [ MAIN 6: Create the new coordinate frame. ]
		### (CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 9) ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords0, $new_bds_gb_coords0, $new_gb_coords_wd_frm0)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm0) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER the \"Complementary-Vertical Split\" of \$gb_sbj (=$gb_sbj): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER the \"Complementary-Vertical Split\" of \$gb_sbj (=$gb_sbj) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER \"Complementary-Vertical-Split\" and BEFORE UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER \"Complementary-Vertical-Split\" and BEFORE UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @{$new_bds_gb_coords0}, @{$new_gb_coords_wd_frm0}, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		### END of "(CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 9)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M9)"

=For the predecessor, 'cvsplit_gblock'. (M10)

		### MAIN 10: UNDO the Displacement. ###
		### (IMPORTED (w/ modifications) from 'cvmerge_gblock_pair' in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 10) ###

		# (Step 1:) Attempt to "undo" the displacements of the gap-blocks. #

	my @new_curr_bl_coords1 = @{$new_orgs_gb_coords0};

	my $ct_coord_chngs = 0;

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_round_trip_dsplc) {
		my $incr_coord = $gb2incr_coord_round_trip_dsplc{$gb};
		my $new_indx_gb = $new_indices_old_gbs[$gb];
		my $new_coord_gb0 = $new_curr_bl_coords1[$new_indx_gb];
		my $new_coord_gb = $new_coord_gb0 - $incr_coord;
		my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_gb]};
		if ($new_coord_gb < $coord_le_wd0) {
		    $new_coord_gb = $coord_le_wd0;
		} elsif ($coord_re_wd0 < $new_coord_gb) {
		    $new_coord_gb = $coord_re_wd0;
		}
		if ($new_coord_gb0 != $new_coord_gb) {
		    $new_curr_bl_coords1[$new_indx_gb] = $new_coord_gb;
		    $ct_coord_chngs++;
		}
	} # End of "foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_round_trip_dsplc) {...}"


    if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n", '( Contents of @{$new_orgs_gb_coords0} (for "coord0"), @new_curr_bl_coords1 (for "coord1") & @{$new_gb_coords_wd_frm0} (for "coord_le0_wd" and "coord_re0_wd"): )', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'coord0', 'coord1', 'coord_le0_wd', 'coord_re0_wd'), "\n\n";

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $new_orgs_gb_coords0->[$gb], $new_curr_bl_coords1[$gb], @{$new_gb_coords_wd_frm0->[$gb]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"


	if ($ct_coord_chngs>0) {

	    my ($flag_ssh_p2p, $tot_incr_ct_null_clms_ssh_p2p)
	     = serial_shifts_point2point (
		@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_bl_coords1, @{$new_gb_coords_wd_frm0}, 
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
			);
	    unless ((defined $flag_ssh_p2p) and ($flag_ssh_p2p == 0)) { # A BUG fixed on 2019/07/27. #
		my $out_flag = (defined $flag_ssh_p2p) ? $flag_ssh_p2p : 'undef';
	    ##unless ($flag_ssh_p2p) {
		print STDOUT "[ERROR 'vmerge_gblock_pair'] The subroutine 'serial_shifts_point2point' FAILED on the new set of gap-blocks (w/ \$flag_ssh_p2p = $out_flag): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_point2point', Ended on the set of new gap-blocks (w/ \$tot_incr_ct_null_clms_ssh_p2p = $tot_incr_ct_null_clms_ssh_p2p) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	} else {

   (print STDOUT "\n[INFORMATION 'cvspl_gblock'] The 'serial_shifts_point2point' will NOT be performed, because \$ct_coord_chngs = 0 !!...\n\n") if ($TEST_LEVEL_vmg_gbp > 1);

	} # End of "if ($ct_coord_chngs>0) {...} else {...}"


		# (Step 2:) Create the FINAL coordinate frame. #

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cvspl_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER UNDOING the displacements of the new gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER UNDOING the displacements of the new gap-blocks ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};


   if ($TEST_LEVEL_vmg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >5) {...}"

   if ($TEST_LEVEL_vmg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 5) {...}"

		### (IMPORTED (w/ modifications) from 'cvmerge_gblock_pair' in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 10) ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M10)"


=For the predecessor, 'cvsplit_gblock'. (M11)

		### MAIN 11: UNDO the horizontal overlaps among the vertically identical & complementary gap-blocks.
		### [ MAIN 9: UNDO the horizontal overlaps among the vertically identical & complementary gap-blocks. ] ... for the predecessor 'vsplit_gblock'. 
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 11) ###

			# The 1st Round (w.r.t. $gb_cch1_NEW) #

	my ($fl_undo1, $gb2incr_bl_coord_undo1) 
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_cch1_NEW, # Modified in creation of 'cvsplit_gblock'. #
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo1) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)) on the cofiguration of gap-blocks after \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW and \$gb_cch2_NEW (=$gb_cch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)), Ended (w/ \$fl_undo1 = $fl_undo1) on the configuration of gap-blocks after the \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW) and \$gb_cch2_NEW (=$gb_cch2_NEW) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


   if ($TEST_LEVEL_vmg_gbp > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)): ]\n\n";

	print STDOUT "\$fl_undo1 = $fl_undo1 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo1} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo1}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo1->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

	if ($fl_undo1 < 0) {
	    print STDOUT "[CAUTION 'vmerge_gblock_pair'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}


			# The 2nd Round (w.r.t. $gb_cch2_NEW) #

	my ($fl_undo2, $gb2incr_bl_coord_undo2) 
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW, # A BUG Fixed on 2019/08/23. #
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch2_NEW,
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_cch2_NEW, # MODIFIED in creation of 'cvsplit_gblock'. #
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo2) {
	    print STDOUT "[ERROR 'vmerge_gblock_pair'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)) on the cofiguration of gap-blocks after \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW and \$gb_cch2_NEW (=$gb_cch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)), Ended (w/ \$fl_undo2 = $fl_undo2) on the configuration of gap-blocks after the \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW) and \$gb_cch2_NEW (=$gb_cch2_NEW) ...\n\n") if ($TEST_LEVEL_vmg_gbp >1);


   if ($TEST_LEVEL_vmg_gbp > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)): ]\n\n";

	print STDOUT "\$fl_undo2 = $fl_undo2 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo2} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo2}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo2->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

	if ($fl_undo2 < 0) {
	    print STDOUT "[CAUTION 'vmerge_gblock_pair'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

			# Combine the outputs of the TWO Rounds into an output of this subroutine. #
			# (There should be NO Problem with it, b/c the two sets of keys SHOULD NOT overlap.)

	my %gb2incr_bl_coord_undo = (%{$gb2incr_bl_coord_undo1}, %{$gb2incr_bl_coord_undo2});


		### End of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 11)" ###

=cut # END of "For the predecessor, 'cvsplit_gblock'. (M11)"


			### Test-Output the Final Results. ###

   if ($TEST_LEVEL_vmg_gbp >3) {

	##print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Complementary-Vertical-Split\" ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Vertical Merge\" ]\n\n"; # Modified in creation of 'vmerge_gblock_pair' (on 2019/10/12). #

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp >3) {...}"


   if ($TEST_LEVEL_vmg_gbp > 3) {

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Complementary-Vertical-Split\" ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Vertical Merge\" ]\n\n"; # Modified in creation of 'vmerge_gblock_pair' (on 2019/10/12). #

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"


		### MAIN 10: Finally Update  @new_list_gblocks. #
		### [ MAIN 12: Finally Update  @new_list_gblocks. # ] ... for the predecessor ''. #
		### [ MAIN 10: Finally Update  @new_list_gblocks. # ] ... for the predecessor 'vsplit_gblock'. #
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###
		### (=> Carried over to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 12) ###
		### (=> Further Carried over to 'vmerge_gblock_pair' (on 2019/10/12).) (MAIN 10) ###

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    @{$new_list_gblocks[$gb]}[2,3] = @{$new_curr_bds_blks[$gb]};
		 # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}

   if ($TEST_LEVEL_vmg_gbp > 3) {

	print STDOUT '[ FINAL Contents of @new_list_gblocks: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vmg_gbp > 3) {...}"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###
		### (=> Carried over to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 12) ###
		### (=> Further Carried over to 'vmerge_gblock_pair' (on 2019/10/12).) (MAIN 10)" ###


			# MODIFIED in creation of 'vmerge_gblock_pair'. (FIN & OUT) #

=For the predecessor, 'cvsplit_gblock' (FIN & OUT)
    (print STDOUT "\n< Subroutine 'cvsplit_gblock' Ended (w/ \$fl_end = 1, (\$gb_cch1_NEW, \$gb_cch2_NEW) = ($gb_cch1_NEW, $gb_cch2_NEW)). >\n\n") if ($TEST_LEVEL_cvspl_gb > 0); # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (FIN)

	return (1, $gb_cch1_NEW, $gb_cch2_NEW, \%gb2incr_coord_1way_dsplc, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (OUT) #
=cut # END of "For the predecessor, 'cvsplit_gblock' (FIN & OUT)"

    (print STDOUT "\n< Subroutine 'vmerge_gblock_pair' Ended (w/ \$fl_end = 1, \$incr_coord_sbj_shftd = \$incr_coord_sbj_shftd, \$gb_mgd_NEW = $gb_mgd_NEW). >\n\n") if ($TEST_LEVEL_vmg_gbp > 0);

	return (1, $incr_coord_sbj_shftd, \%ngb2incr_coord_1way_dsplc, $gb_mgd_NEW,
	\@new_indices_old_gbs, \@old_indices_new_gbs, \%ngb2incr_bl_coord_undo,

			# END of "MODIFIED in creation of 'vmerge_gblock_pair'. (FIN & OUT)" #

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks
	);

			### RESTART FROM HERE (on 2019/10/12). (2) ###

}	# End of the subroutine, 'vmerge_gblock_pair'. #


# ssplit_gblock:
#	This subroutine "split"s the subject gap-block into a pair of gap-blocks of the same type.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'smerge_split_gblock_pair' below.
#	[CREATED on 2019/08/12.]
#	[REVISED on 2919/08/12]:
#	Now, the subroutine performs 'undo_hovlp_vcmpl_gbsets', instead of 'serial_shifts_sgl_gblock',
#	to separate the two gap-blocks resulting from the "split".
#
# smerge_split_gblock_pair:
#	This subroutine first "merge" a specified pair of gap-blocks of the SAME type,
#	then, it immediately "split" the resulting gap-block,
#	creating a pair of gap-blocks of the identical type but with (hopefully) different sizes.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the main part of 'smerge_gblock_pair'.
#	[CREATION STARTED on 2019/07/30, ENDED on 2019/07/31.]
#	[DEBUGGED 1st on 2019/07/31.]
#	[REVISED on 2019/08/01]:
#	Now, the subroutine makes Utmose Attempts to separate the resulting gap-blocks from each other; It it canNOT, it returns an EXCEPTIONAL VALUE (= -1) for $fl_end.
#
# smerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks of the SAME type.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/07/21.]
#	[DEBUGGED on 2019/07/22]:
#	Now, the subroutine can correctly handle a gap-block pair with ZERO distance
#	that flank another gap-block in the 'onx' relationship.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
#
# [ INPUTS: ($gb1, $gb2, $new_size_right, ] ... for the predecessor 'smerge_split_gblock_pair'.
# INPUTS: ($gb_sbj, $new_size_right,
#	 \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm,
#	 \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	 \@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb_sbj is the index of the gap-block to be "split"; # ADDED in creation of 'ssplit_gblock_pair'.
#
#	[[
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be of the same vertical type;
#	]] ... for the predecessor 'smerge_split_gblock_pair'.
#
#	$new_size_right is the size of the gap-block on the right of the newly created pair;
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj, ] ... for the predecessor 'smerge_split_gblock_pair'.
#
# [ OUTPUTS: ($flag_end, $new_gb_left, $new_gb_right, $new_gb_shftd,  $incr_coord_via_shft, @new_indices_old_gbs, @old_indices_new_gbs, ] ... OBSOLETE as of 2019/08/12 (after the REVISION).
#
# OUTPUTS: ($flag_end, $new_gb_left, $new_gb_right, @new_indices_old_gbs, @old_indices_new_gbs, \%gb2incr_bl_coord_undo,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	[ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where ] ... for the predecessor 'smerge_split_gblock_pair'.
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed,
#	 	  = -1 if the resulting gap-blocks will be UNABLE to separate from each other (ADDED on 2019/08/01);
#
#	[[
#			# Added in creation of 'ssplit_gblock' (on 2019/08/12). #
#	$new_gb_left/right is the new index of the gap-block that is the left/right member of the newly created pair; 
#	$new_gb_shftd is the new index of the gap-block that was "shift"ed after the "split";
#	$incr_coord_via_shft is the increment of the coordinate of $new_gb_shftd resulting from the "shift" after the "split";
#			# End of "Added in creation of 'ssplit_gblock' (on 2019/08/12)." #
#	]] ... OBSOLETE as of 2019/08/12 (after the REVISION). 
#
#	[[
#	$bl_sbj is the index for the gap-block on which the 1st round of 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 1st round of 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	$new_bl_sbj is the index for the gap-block on which the 2nd round of 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_new_sbj is the increment of the coordinate of $new_bl_sbj that resulted from the 2nd round of 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#	]] ... for the predecessor 'smerge_split_gblock_pair'.
#
#	[[
#	$gb_absbd and $gb_kpt are the indices of the gap-blocks that were absorbed and kept (actually, absorbing), respectively;
#	]] ... for the predecesor, 'smerge_gblock_pair'.
#
#			# Resurrected (with modification) in creation of 'ssplit_gblock' (on 2019/08/12). #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# End of "Resurrected (with modification) in creation of 'ssplit_gblock' (on 2019/08/12)." #
#
#			# ADDED in REVISION of 'ssplit_gblock' (on 2019/08/12). #
#       %gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#       (if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#       [\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# END of "ADDED in REVISION of 'ssplit_gblock' (on 2019/08/12)." #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
sub ssplit_gblock ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {

	my ($gb_sbj, $new_size_right, 
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $size_sbj = $gblock_sizes->[$gb_sbj];
	## my $sum_size = $size1 + $size2;

	unless (($new_size_right>0) and ($new_size_right < $size_sbj)) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The \$new_size_right (=$new_size_right) is NOT within the range, (0, \$size_sbj (=$size_sbj)): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

	my $gb_sbj_coord_bf = $curr_bl_coords->[$gb_sbj];

	my ($lbd_coord_sbj, $rbd_coord_sbj) = @{$bds_gb_coords->[$gb_sbj]};

	my ($lbd_sbj_bf, $rbd_sbj_bf) = @{$curr_bds_blks->[$gb_sbj]};


    (print STDOUT "< Subroutine 'ssplit_gblock' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; gap-block to be \"split\": \$gb_sbj = $gb_sbj; {block-size}=$size_sbj; {coordinate}=$gb_sbj_coord_bf; {coordinate_boundaries}=($lbd_coord_sbj, $rbd_coord_sbj); {physical boundaries}=($lbd_sbj_bf, $rbd_sbj_bf); \$new_size_right = $new_size_right) >\n\n") if ($TEST_LEVEL_sspl_gb > 0);

		# ADDED in creation of 'ssplit_gblock'. (Pre 0) #

	unless (($lbd_coord_sbj <= $gb_sbj_coord_bf) and ($gb_sbj_coord_bf <= $rbd_coord_sbj)) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The \$gb_sbj_coord_bf (=$gb_sbj_coord_bf) is NOT within the boundaries: [$lbd_coord_sbj, $rbd_coord_sbj]: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

	if (($gb_sbj_coord_bf == $lbd_coord_sbj) and ($gb_sbj_coord_bf == $rbd_coord_sbj)) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The \$gb_sbj (=$gb_sbj) th gap-block canNOT be \"Shift\"ed at all: $! !\n\n";
	    return (-1, undef, undef, undef, undef);
	}
		# End of "ADDED in creation of 'ssplit_gblock'. (Pre 0)" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_sspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Split (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb > 5) {...}"


   if ($TEST_LEVEL_sspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Split (same type)\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >3) {...}"



		### REVISED in creation of 'ssplit_gblock' (on 2019/08/12). (MAIN 2) ###

		### MAIN 2: Assign the Indices & Boundaries of $new_gb_left and  $new_gb_right.
		###		& create some important data-sets.

	my ($new_gb_left, $new_gb_right) = ($gb_sbj, $gb_sbj+1);

	my $sites_cntr_to_sbj = enum_sites_contribute_to_gblk ($gb_sbj, @cp_curr_bds_blks, 0, 1, @{$inter_block_relations});
	unless (defined $sites_cntr_to_sbj) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The subroutine, 'enum_sites_contribute_to_gblk', FAILED on the \$gb_sbj (=$gb_sbj) th gap-block: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
	my $ct_sites_cntr_to_sbj = @{$sites_cntr_to_sbj};
   (print STDOUT "\n ... The subroutine, 'enum_sites_contribute_to_gblk', Ended on the \$gb_sbj (=$gb_sbj) th gap-block (w/ #{sites contributing to \$gb_sbj}=$ct_sites_cntr_to_sbj) ...\n\n") if ($TEST_LEVEL_sspl_gb >1);

	unless ($size_sbj == $ct_sites_cntr_to_sbj) {
	    print STDOUT "[ERROR 'ssplit_gblock'] IMPOSSIBLE!!: The numbers of sites contributing to \$gb_sbj (=$gb_sbj, $ct_sites_cntr_to_sbj, does NOT EQUAL its size, \$size_sbj = $size_sbj: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

   if ($TEST_LEVEL_sspl_gb >3) {

	print STDOUT "\n[ Output of 'enum_sites_contribute_to_gblk': ]\n\n";

	print STDOUT '@{$sites_cntr_to_sbj} = (', join (', ', @{$sites_cntr_to_sbj}), ") .\n\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >3) {...}"


	my @new_curr_set_clms = @cp_curr_set_clms;

	my $new_size_left = $size_sbj - $new_size_right;
	my ($new_lbd_left, $new_rbd_right) = @{$sites_cntr_to_sbj}[0, $size_sbj-1];
	my ($new_rbd_left, $new_lbd_right) = @{$sites_cntr_to_sbj}[$new_size_left -1, $new_size_left];

   if ($TEST_LEVEL_sspl_gb >2) {

	print STDOUT "\n[ (IMPORTANT!!) Boundaries of \$new_gb_left(=$new_gb_left) and \$new_gb_right(=$new_gb_right), BEFORE the final \"shift\": ]\n\n";

	print STDOUT "For \$new_gb_left:   [$new_lbd_left, $new_rbd_left] ,\n";
	print STDOUT "For \$new_gb_right:  [$new_lbd_right, $new_rbd_right] .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >2) {...}"


	my @new_indices_old_gbs = my @old_indices_new_gbs = ();
	$new_indices_old_gbs[$gb_sbj] = undef;
	$old_indices_new_gbs[$new_gb_left] = $old_indices_new_gbs[$new_gb_right] = undef;

	my @new_curr_bds_blks = (); # Modify @cp_curr_bds_blks;
	my @new_gblock_sizes = (); # Modify @{$gblock_sizes};
	my @rows_affd_by_new_blks = (); # Modify @ @cp_rows_affd_by_blks;
	my $new_indx_gb = -1;
	for (my $ogb=0; $ogb < $gb_sbj; $ogb++) {
	    $new_indx_gb++;
	    $new_indices_old_gbs[$ogb] = $new_indx_gb;
	    $old_indices_new_gbs[$new_indx_gb] = $ogb;
	    $new_curr_bds_blks[$new_indx_gb] = $cp_curr_bds_blks[$ogb];
	    $new_gblock_sizes[$new_indx_gb] = $gblock_sizes->[$ogb];
	    $rows_affd_by_new_blks[$new_indx_gb] = $cp_rows_affd_by_blks[$ogb];
	}
	$new_curr_bds_blks[$new_gb_left]  = [$new_lbd_left, $new_rbd_left];
	$new_curr_bds_blks[$new_gb_right] = [$new_lbd_right, $new_rbd_right];

	$new_gblock_sizes[$new_gb_left]  = $new_size_left;
	$new_gblock_sizes[$new_gb_right] = $new_size_right;

	$rows_affd_by_new_blks[$new_gb_left] = $cp_rows_affd_by_blks[$gb_sbj];
	$rows_affd_by_new_blks[$new_gb_right] = $cp_rows_affd_by_blks[$gb_sbj];

	$new_indx_gb += 2;

	for (my $ogb = $gb_sbj+1; $ogb < $ct_in_gbs; $ogb++) {
	    $new_indx_gb++;
	    $new_indices_old_gbs[$ogb] = $new_indx_gb;
	    $old_indices_new_gbs[$new_indx_gb] = $ogb;
	    $new_curr_bds_blks[$new_indx_gb] = $cp_curr_bds_blks[$ogb];
	    $new_gblock_sizes[$new_indx_gb] = $gblock_sizes->[$ogb];
	    $rows_affd_by_new_blks[$new_indx_gb] = $cp_rows_affd_by_blks[$ogb];
	}

	my $ct_out_clms = @new_curr_set_clms;
	my $ct_out_gbs = @new_curr_bds_blks;

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($lbd, $rbd) = @{$new_curr_bds_blks[$gb]};
	    push @{$new_curr_gblks_w_spec_lb[$lbd]}, $gb;
	    push @{$new_curr_gblks_w_spec_rb[$rbd]}, $gb;
	}


   if ($TEST_LEVEL_sspl_gb >2) { # IMPORTED with modifications from 'cvmerge_gblock_pair' (on 2019/08/12). #

	print STDOUT "\n[ Relationships between old, and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	##my @out_med_indices_old_gbs = my @out_old_indices_med_gbs = ();
	##my @out_new_indices_med_gbs = my @out_med_indices_new_gbs = ();
	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ( ##[\@med_indices_old_gbs, \@out_med_indices_old_gbs],
			   ##[\@old_indices_med_gbs, \@out_old_indices_med_gbs],
			   ##[\@new_indices_med_gbs, \@out_new_indices_med_gbs],
			   ##[\@med_indices_new_gbs, \@out_med_indices_new_gbs],
			  [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	##print STDOUT '@med_indices_old_gbs = (', join (', ', @out_med_indices_old_gbs), ") ,\n";
	##print STDOUT '@new_indices_med_gbs = (', join (', ', @out_new_indices_med_gbs), ") ,\n";
	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	##print STDOUT '@med_indices_new_gbs = (', join (', ', @out_med_indices_new_gbs), ") ,\n";
	##print STDOUT '@old_indices_med_gbs = (', join (', ', @out_old_indices_med_gbs), ") ,\n";
	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >2) {...}"

		### END of "REVISED in creation of 'ssplit_gblock' (on 2019/08/12). (MAIN 2)" ###


		### MAIN 2.5: Create the new set of inter-block relationships & its derivative data-sets. ###
		### (ADDED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 2.5) ###

	my @old_indices_new_gbs_ext = @old_indices_new_gbs;
	$old_indices_new_gbs_ext[$new_gb_left] = $gb_sbj;
	$old_indices_new_gbs_ext[$new_gb_right] = $gb_sbj;

	my @new_inter_block_relations = ();
	for (my $ngb =0; $ngb < $ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs_ext[$ngb];
	    my @rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_new_gbs_ext];
	    $new_inter_block_relations[$ngb] = \@rels_w_ngb;
	}
	$new_inter_block_relations[$new_gb_left]->[$new_gb_right] = '=';
	$new_inter_block_relations[$new_gb_right]->[$new_gb_left] = '=';

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_sspl_gb >1);

		### END of "(ADDED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 2.5)" ###


		### MAIN 3: Create the new coordinate frame. ###
		###	=> If possible, extend the ranges for $gb1 and $gb2. ###

		### (=> MODIFIED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 3) ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts); # MODIFIED in creation of 'ssplit_gblock' (on 2019/08/12). #
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER the \"Split\" of \$gb_sbj(=$gb_sbj) into (\$new_gb_left, \$new_gb_right) (=($new_gb_left, $new_gb_right)): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER the \"Split\" of \$gb_sbj(=$gb_sbj) into (\$new_gb_left, \$new_gb_right) (=($new_gb_left, $new_gb_right)) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

			# If possible, extend the ranges for $gb1 and $gb2. #

	my $left_margin = $gb_sbj_coord_bf - $lbd_coord_sbj;
	my $right_margin = $rbd_coord_sbj - $gb_sbj_coord_bf;

	my $org_new_coord = $left_margin;
	my ($lbd_new_coord, $rbd_new_coord) = (0, $left_margin + $right_margin);

	foreach my $gb ($new_gb_left, $new_gb_right) {
	    my $org_new_coord0 = $new_orgs_gb_coords->[$gb];
	    my $offset = $org_new_coord - $org_new_coord0;

	    $new_curr_bl_coords[$gb] = $org_new_coord;
	    $new_bds_gb_coords[$gb] = [$lbd_new_coord, $rbd_new_coord];
	    $new_gb_coords_wd_frm[$gb]->[0] += $offset;
	    $new_gb_coords_wd_frm[$gb]->[1] += $offset;

	} # End of "foreach my $gb ($gb1, $gb2) {...}"


			# Further DEBUGGED (on 2019/07/31). #
	if ($new_bds_gb_coords[$new_gb_left]->[1] + $new_size_right <= $new_gb_coords_wd_frm[$new_gb_left]->[1]) { # SWITCHED on in DEBUGGING (on 2019/07/31). #
	    $new_bds_gb_coords[$new_gb_left]->[1] += $new_size_right; # SWITCH ON this command, if you want $new_gb_left to be reachable to the right-end of the frame of $new_gb_right.
	}

		### END of "(=> MODIFIED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 3)" ###

		### End of "ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/30). (MAIN 2&3)" ###


		### ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/31). (MAIN 3post) ###

   if ($TEST_LEVEL_sspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split\" of \$gb_sbj (=$gb_sbj) ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >5) {...}"

   if ($TEST_LEVEL_sspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split\" of \$gb_sbj (=$gb_sbj) ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb > 5) {...}"



		### MAIN 3post: UNDO the overlaps between the gap-blocks
		###		vertically equivalent to and/or complementary to $new_gb_left and $new_gb_right. ###
		### (REVISED on 2019/08/12. (MAIN 3post)) ###

	my ($fl_undo, $gb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets ($new_gb_left, $new_gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'ssplit_gblock'] The subroutine, 'undo_hovlp_vcmpl_gbsets' FAILED on the cofiguration of gap-blocks after the \"split\" of \$gb_sbj (=$gb_sbj) into \$new_gb_left (=$new_gb_left) and \$new_gb_right (=$new_gb_right): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The the subroutine, 'undo_hovlp_vcmpl_gbsets', Ended on the configuration of gap-blocks immediately after the \"split\" of \$gb_sbj (=$gb_sbj) into \$new_gb_left (=$new_gb_left) and \$new_gb_right (=$new_gb_right) ...\n\n") if ($TEST_LEVEL_sspl_gb >1);


   if ($TEST_LEVEL_sspl_gb > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets': ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_sspl_gb > 3) {...}"


=DISABLED in Further REVISION on 2019/08/12. (MAIN 3post)
		### MAIN 3post: \"Shift" one member of the pair, to separate the blocks. ###
		### (REVISED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 3post) ###

	my ($gb_shftd, $incr_coord_via_shft) = ($right_margin>0) ? ($new_gb_right, +1) : ($new_gb_left, -1);

   (print STDOUT "\n ... Perform 'serial_shifts_sgl_gblock' (w/ \$gb_shftd = $gb_shftd, \$incr_coord_via_shft = $incr_coord_via_shft), to SEPARATE the newlycreated \$new_gb_left (=$new_gb_left) and \$new_gb_right (=$new_gb_right) ...\n\n") if ($TEST_LEVEL_sspl_gb > 2);

	my ($flag1, $incr_ct_null_clms1)
	 = serial_shifts_sgl_gblock (
	    @new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_gb_coords_wd_frm, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb,
	    $gb_shftd, $incr_coord_via_shft, @rows_affd_by_new_blks, $cnct_null_clm,
	    @new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ((defined $flag1) and ($flag1 == 0)) {
	    my $flag1_out = (defined $flag1) ? $flag1 : 'undef';
	    print STDOUT "[ERROR 'ssplit_gblock'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the \$gb_shftd (=$gb_shftd) th gap-block (w/ \$flag1 = $flag1_out): $! !\n\n";

	    return (0, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_shftd (=$gb_shftd) th gap-block (w/ \$incr_ct_null_clms1 = $incr_ct_null_clms1) ...\n\n") if ($TEST_LEVEL_sspl_gb >1);

=cut # END of "DISABLED in Further REVISION on 2019/08/12. (MAIN 3post)"

		### END of "(REVISED in creation of 'ssplit_gblock' (on 2019/08/12).) (MAIN 3post)" ###


   if ($TEST_LEVEL_sspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Split (same type)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb >3) {...}"

   if ($TEST_LEVEL_sspl_gb > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Split (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_sspl_gb > 3) {...}"

		### End of "ADDED in creation of 'smerge_split_grow_gblock_pair' (on 2019/07/31). (MAIN 3post)" ###


=For the predecessor 'smerge_split_gblock_pair'. (FIN)
    (print STDOUT "\n< Subroutine 'smerge_split_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$new_bl_sbj = $new_bl_sbj, \$incr_coord_new_sbj = $incr_coord_new_sbj). >\n\n") if ($TEST_LEVEL_smrg_spl_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes);
=cut # END of "For the predecessor 'smerge_split_gblock_pair'. (FIN)"

			# RESURRECTED with MODIFICATIONs in creation of 'ssplit_gblock' (on 2019/08/12). (FIN)) #

	my $fl_end = $fl_undo;

    (print STDOUT "\n< Subroutine 'ssplit_gblock' Ended (w/ \$fl_end = $fl_end, \$new_gb_left = $new_gb_left, \$new_gb_right = $new_gb_right). >\n\n") if ($TEST_LEVEL_sspl_gb > 0);

	return ($fl_end, $new_gb_left, $new_gb_right, \@new_indices_old_gbs, \@old_indices_new_gbs, $gb2incr_bl_coord_undo,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes);

			# END of "RESURRECTED with MODIFICATIONs in creation of 'ssplit_gblock' (on 2019/08/12). (FIN))" #

}	# End of the subroutine, 'ssplit_gblock'. #


# initial_csplit_gblock:
#	This satellite subroutine of 'csplit_gblock' performs the INITIAL "split (complementary types)"
#	of the subject gap-block.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of a part of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by importing the functions to create a gap-block of a NEW TYPE, from 'cvmerge_gblock_pair'.
#	[CREATED on 2019/08/16.]
#
# csplit_gblock:
#	This subroutine "split"s the subject gap-block into a pair of gap-blocks of complementary types.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by incoporating a NEW satellite subroutine, 'initial_csplit_gblock'.
#	[CREATION STARTED on 2019/08/15, ENDED on 2019/08/16.]
#	
#
# cmerge_split_grow_gblock_pair (NEW VERSION):
#	This new version does essentially the same job as its old version (now renamed as 'cmerge_split_grow_gblock_pair_OLD'),
#	but UNDER a DIFFERENT STRATEGY.
#	NOTE that, accompanying this revision, the portfolio of the OUTPUT data has also CHANGED.
#	[CREATION STARTED on 2019/08/06, ENDED on 2019/08/07.]
#
#	[STRATEGY]:
#	UNLIKE the old version,
#	it attempts to grow the subject pair of gap-blocks
#	while leaving the existing gap-blocks as unmoved as possible,
#	by (i) enumerating the columns available for the "split",
#	and by (ii) actually "split"ting the column one by one,
#	and "merg"ing the resulting gap-blocks to the subject gap-blocks each time a column is "split".
#
#	[REVISED on 2019/08/08]:
#	Now, the subroutine performs 'undo_hovlp_vcmpl_gbsets' before returning the results,
#	in order to undo the horizontal overlaps resulting from the main process of "merge + split".
#	[REVisEd on 2019/08/09]:
#	Now, the subroutine also outputs \%gb2incr_bl_coord_undo,
#	which is one of the main outputs of 'undo_hovlp_vcmpl_gbsets'.
#
# cmerge_split_grow_gblock_pair:
#	This subroutine performs the "merge + split" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, and grows their sizes by a specified number of columns.
#	It also prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'icmerge_gblock_pair'.
#	[CREATED on 2019/08/03.]
#	[DEBUGGED on 2019/08/05]:
#	(i) Now, it should be able to CORRECTLY separate the gap-block boundaries
#	belonging to the columns to be "split".
#	(ii) CORRECTED the use of the output, %{$gb2sbincr_coord}, of 'displace_vincluding_veq_and_onx_to_right_end'.
#	[REVISED on 2019/08/05]:
#	The rules for assigning the coordinate increments for the 2nd (UNDOing) serial "shift"s,
#	so that the final result will look closer to the input alignment.
#
# icmerge_gblock_pair:
#	This subroutine performs the "INcomplete merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, by a specified number of columns.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/08/02.]
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, $incr_gb_size, $char_abs, 
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where ] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (NEW VERSION).
#
# INPUTS: ($gb_sbj, $avail_site, $char_abs, 
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm,
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1)" #

#	), where
#
#			# ADDED in creation of 'initial_csplit_gblock'. (HD1) #
#	$gb_sbj is the index of the gap-block to be "split" into a pair of vertically complementary blocks;
#
#	$avail_site is the index of the site (i.e., column) available for the "Split (complementary types)";
#			# End of "ADDED in creation of 'initial_csplit_gblock'. (HD1)" #
#
#	[[
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#	$incr_gb_size is the number of columns by which the sizes of $gb1 and $gb2 will increase;
#	]] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (NEW VERSION).
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2)" #
#	
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# [ OUTPUTS: ($flag_end, $gb_left, $gb_right, \@records_sessions, \%gb2incr_bl_coord_undo, # REVISED on 2019/08/09. # ] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#
# OUTPUTS: ($flag_end, $gb_sbj_NEW, $gb_cmpl_NEW, \@new_indices_old_gbs, \@old_indices_new_gbs, \@record_session, 
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	[ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes ] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#			# IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#			# End of "IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD1)" #

#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_list_gblocks
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1)" #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#			# MODIFIED in creation of 'csplit_gblock'. (HD) #
#	$gb_sbj_NEW is the new index for the subject gap-block(, which grew in size during this subroutine);
#	$gb_cmpl_NEW is the new index for the vertically complementary gap-block;
#
#	[[
#	$gb_left and $gb_right are the indices of the gap-blocks that are on the left and on the right, respectively, each chosen between $gb1 and $gb2;
#	]] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#			# END of "MODIFIED in creation of 'csplit_gblock'. (HD)" #
#
#			# IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# End of "IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#			# ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#			# 	=> MODIFIED in creation of 'initial_csplit_gblock'. #
#	@record_session = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	[[
#	@{$records_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2 (='n/a') )
#	provides the record of the $k th session, where
#	]] ... for the predecessor 'cmerge_split_grow_gblock_pair' (New Version).
#	$session_ID (=1) is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_sbj th gap-block,
#	and $incr_coord_sbj_smg2 = 'n/a' always holds in this subroutine;
#			# End of "ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#			#	 => MODIFIED in creation of 'initial_csplit_gblock'." #
#
#	[[			
#			# ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09). #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09)." #
#	]] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (New Version).
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
# [ sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {
#	my ($gb1, $gb2, $incr_gb_size, $char_abs,
#	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
#	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
#	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_; ] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (New Version).
#
sub initial_csplit_gblock ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb_sbj, $avail_site, $char_abs,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,

		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq
		# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock'. (IN)" #
	) = @_;

# INPUTS: ($gb_sbj, $size_cmpl_gb, \@pot_avail_sites, $char_abs, 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

			# MODIFIED in creation of 'initial_csplit_gblock'. (BEG) #

	my $size_sbj_bf = $gblock_sizes->[$gb_sbj];
	my $sbj_coord_bf = $curr_bl_coords->[$gb_sbj];

	my ($lbd_sbj_bf, $rbd_sbj_bf) = @{$curr_bds_blks->[$gb_sbj]};

	my $sbjct_site = $avail_site; # ADDED in CORRECTION on 2019/08/17.


    (print STDOUT "< Subroutine 'initial_csplit_gblock' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; gap-block to be \"Split\": \$gb_sbj = $gb_sbj; {subject_block_size}_bf = $size_sbj_bf; {subject_coordinate}_bf = $sbj_coord_bf; {boundaries}_sbj_bf = [$lbd_sbj_bf, $rbd_sbj_bf]; {available site} (=\$avail_site = \$sbjct_site) = $avail_site) >\n\n") if ($TEST_LEVEL_init_cspl_gb > 0);

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (BEG)
	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless ($incr_gb_size>0) { # Added on 2019/08/03. #
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The \$incr_gb_size (=$incr_gb_size) is NOT POSITIVE: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};

    (print STDOUT "< Subroutine 'cmerge_split_grow_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"INcomplete-merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$incr_gb_size = $incr_gb_size) >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (BEG)"

			# END of "MODIFIED in creation of 'initial_csplit_gblock'. (BEG)" #


	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 0)

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #

	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1);
	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};

   if ($TEST_LEVEL_init_cspl_gb >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 0)"


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_init_cspl_gb >3) {

	##print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of INITIAL \"Split (complementary types)\" ]\n\n"; # CORRECTED on 2019/09/15.

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >3) {...}"

   if ($TEST_LEVEL_init_cspl_gb > 5) {

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of INITIAL \"Split (complementary types)\" ]\n\n"; # CORRECTED on 2019/09/15.

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 5) {...}"


		# (Preliminary 2)': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". #
		# (IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock' (on 2019/08/16).) #

	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_init_cspl_gb > 5) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 5) {...}"

		# END of "(IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock' (on 2019/08/16).)" #


		### Preliminary (Out): ###

	my ($gb_cmpl_old, $gb_incr_sbj) = ($ct_in_gbs, $ct_in_gbs+1);
	##my ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1); # For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).

	my $intm1_ct_gbs = $ct_in_gbs +2;
	my $ct_out_gbs = $ct_in_gbs +1;
	##my $intm2_ct_gbs = $ct_in_gbs +1; # For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).

		# Create @intm1_gblock_sizes & @intm1_list_gblocks & @intm1_inter_block_relations. #

	my @cp_curr_gblock_sizes = @{$gblock_sizes};
	my @intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);

				## MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) ##

			# (NOTE: At this point, we don't care whether $lm_clm and $rm_clm are correct or not.)
	my @intm1_list_gblocks = @list_old_gblocks;

	my @info_gb_incr_sbj = @{$list_old_gblocks[$gb_sbj]};
	@info_gb_incr_sbj[2,3] = ($avail_site+1, $avail_site+1);
	$intm1_list_gblocks[$gb_incr_sbj] = \@info_gb_incr_sbj;

	my $rows_affd_by_sbj = $rows_affd_by_blks->[$gb_sbj];
	my @all_rows = (0 .. $ct_rows-1);
	my ($common, $ex_sbj, $rows_affd_by_cmpl) = diff_nmarray (@{$rows_affd_by_sbj}, @all_rows);
	unless (defined $rows_affd_by_cmpl) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'diff_nmarray', FAILED to create the set of rows affected by the complementary block: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
	my @exnodes_affd_by_cmpl = ();
	foreach my $row (@{$rows_affd_by_cmpl}) { push @exnodes_affd_by_cmpl, $nodeids_in_msa->[$row]; }
	if (@exnodes_affd_by_cmpl > 1) { @exnodes_affd_by_cmpl = sort {$a <=> $b} @exnodes_affd_by_cmpl; }

	my $cnct_rows_affd_by_cmpl = join (',', @{$rows_affd_by_cmpl});
	my $cnct_exnodes_affd_by_cmpl = join (',', @exnodes_affd_by_cmpl);

	my $br_up_or_lw_cmpl = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_cmpl};
	unless (defined $br_up_or_lw_cmpl) {
	     print STDOUT "[ERROR 'initial_csplit_gblock'] IMPOSSIBLE!!: NO branch & upper/lower information is ASSOCIATED with the set of MSA rows, ($cnct_rows_affd_by_cmpl), that is supposed to be affected by the gap-block complementary to \$gb_sbj (=$gb_sbj): $! !\n\n";
	     return (0, undef, undef, undef);
	}
	my ($br_cmpl, $up_or_lw_cmpl) = @{$br_up_or_lw_cmpl};

	$intm1_list_gblocks[$gb_cmpl_old] = [$br_cmpl, $up_or_lw_cmpl, $avail_site, $avail_site, $cnct_exnodes_affd_by_cmpl, $cnct_rows_affd_by_cmpl, \@exnodes_affd_by_cmpl, $rows_affd_by_cmpl];


   if ($TEST_LEVEL_init_cspl_gb > 3) {

	print STDOUT "\n", '( Contents of @intm1_list_gblocks (maybe with incorrect lm_clm and rm_clm): )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$intm1_ct_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm1_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 3) {...}"

	my ($intm1_inter_block_relations)
	  = mk_inter_block_relations (@intm1_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $intm1_inter_block_relations) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@intm1_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@intm1_list_gblocks', " ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1) ;

	my @intm1_inter_block_relations = @{$intm1_inter_block_relations};

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre (Out))
	my @intm1_inter_block_relations = my @intm2_inter_block_relations = ();
	for (my $gb=0; $gb <$ct_in_gbs; $gb++) {
	    my @intm1_rels_w_gb = @{$inter_block_relations->[$gb]};
	    ##my @intm2_rels_w_gb = @{$inter_block_relations->[$gb]};

	    if ($gb == $gb_left) {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('=', 'Cp');
		##$intm2_rels_w_gb[$ct_in_gbs] = '=';
	    } elsif ($gb == $gb_right) {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('Cp', '=');
		##$intm2_rels_w_gb[$ct_in_gbs] = 'Cp';
	    } else {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1]
		 = @intm1_rels_w_gb[$gb_left, $gb_right];
		##$intm2_rels_w_gb[$ct_in_gbs] = $intm2_rels_w_gb[$gb_left];
	    }
	    $intm1_inter_block_relations[$gb] = \@intm1_rels_w_gb;
	    ##$intm2_inter_block_relations[$gb] = \@intm2_rels_w_gb;
	}
	my @intm1_rels_w_incr_left = @{$intm1_inter_block_relations[$gb_left]};
	@intm1_rels_w_incr_left[$gb_left, $ct_in_gbs] 
		= @{$intm1_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];

	my @intm1_rels_w_incr_right = @{$intm1_inter_block_relations[$gb_right]};
	@intm1_rels_w_incr_right[$gb_right, $ct_in_gbs+1]
		= @{$intm1_inter_block_relations[$gb_right]}[$ct_in_gbs+1, $gb_right];

	$intm1_inter_block_relations[$ct_in_gbs] = \@intm1_rels_w_incr_left;
	$intm1_inter_block_relations[$ct_in_gbs+1] = \@intm1_rels_w_incr_right;
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre (Out))"

				## END of "MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out))" ##

=Set aside for Later Use. # A relic of the predecessor 'cmerge_split_grow_gblock_pair' (New Version). #
	my @intm2_rels_w_incr_left = @{$intm2_inter_block_relations[$gb_left]};
	@intm2_rels_w_incr_left[$gb_left, $ct_in_gbs] 
		= @{$intm2_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];
	$intm2_inter_block_relations[$ct_in_gbs] = \@intm2_rels_w_incr_left;
=cut # END of "Set aside for Later Use."


		# Create $intm1_cts_vincluding, etc. #

	my ($intm1_cts_vincluding, $intm1_cts_vincluded, $intm1_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm1_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'initial_csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm1_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm1_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1) ;


=Set aside for Later Use. # A relic of the predecessor 'cmerge_split_grow_gblock_pair' (New Version). #
	my ($intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm2_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'initial_csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm2_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm2_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1) ;
=cut # END of "Set aside for Later Use."


		# Create @intm1_rows_affd_by_blks.

				## MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) (2) ##
	my @intm1_rows_affd_by_blks = @cp_rows_affd_by_blks;
	$intm1_rows_affd_by_blks[$gb_cmpl_old] = $rows_affd_by_cmpl;
	$intm1_rows_affd_by_blks[$gb_incr_sbj] = $rows_affd_by_sbj;
	
	##my @intm1_rows_affd_by_blks = (@cp_rows_affd_by_blks, @cp_rows_affd_by_blks[$gb_left, $gb_right]);
	####my @intm2_rows_affd_by_blks = (@cp_rows_affd_by_blks, $cp_rows_affd_by_blks[$gb_left]);
				## END of "MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) (2)" ##

		### MAIN 1: "Split" the $avail_site th column.
		### & MAIN 2: "Merge" the resulting $gb_incr_sbj with $gb_sbj.

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $curr_ct_clms = $ct_in_clms;

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN I & 2 (Front))
		### MAIN 1 & 2: Iteratively Perform the single-column-growing "Merge + Split"s. ###

	my ($rows_affd_by_left, $rows_affd_by_right) = @cp_rows_affd_by_blks[$gb_left, $gb_right];
	my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];

	my $curr_incr_gb_size = $incr_gb_size +1;
	my $curr_ct_clms = $ct_in_clms;

	my @records_sessions = (); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

     foreach my $session (1 .. $incr_gb_size) { # Outer main foreach-loop (over the sessions). #

	$curr_incr_gb_size--;
	my ($le_wd, $re_wd) = (0, $curr_ct_clms-1);

    (print STDOUT "\n{{ Session $session:  (\$curr_incr_gb_size = $curr_incr_gb_size, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_init_cspl_gb > 1);

		# Preliminary (In):  Enumerate the columns available for the "(size-growing) Merge + Split".

	my ($avail_sites_left, $avail_sites_in_btwn, $avail_sites_right)
	    = enum_sites_avail_for_csplit ($gb1, $gb2, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_bds_gb_coords, @{$inter_block_relations}, $le_wd, $re_wd);
	unless (defined $avail_sites_right) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'enum_sites_avail_for_csplit', FAILED on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'enum_sites_avail_for_csplit', Ended on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);

	my @avail_sites = (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left}) );
	my $ct_avail_sites = @avail_sites;

	unless ($ct_avail_sites >= $curr_incr_gb_size) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'enum_sites_avail_for_csplit' (for session $session), output ONLY $ct_avail_sites (< \$curr_incr_gb_size (=$curr_incr_gb_size)) sites available for the \"(size-growing) merge + split\": $! !\n\n";
	    ##return (0, undef, undef, undef);
	    return (-1, undef, undef, undef); # REVISED on 2019/08/07. #
	}

	my $sbjct_site = $avail_sites[0];

   if ($TEST_LEVEL_init_cspl_gb >3) {

	print STDOUT "\n[ Secondary Outputs of 'enum_sites_avail_for_csplit' (#{available sites}=$ct_avail_sites): ]\n\n";

	print STDOUT '@avail_sites (= (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left})) )', "\n";
	print STDOUT ' = (', join (', ', @avail_sites), ") .\n\n";
	print STDOUT '$sbjct_site = ', "$sbjct_site .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >3) {...}"

=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN I & 2 (Front))"


		## MAIN I: Create a new pair of single-column gap-blocks, by "split"ting the $sbjct_site th site. ##

	my @clm_gb_cmpl = @null_clm;
	my @clm_incr_sbj = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_sbj}) { # 1st middle foreach.
	    my $tmp = $clm_incr_sbj[$row];
	    $clm_incr_sbj[$row] = $clm_gb_cmpl[$row];
	    $clm_gb_cmpl[$row] = $tmp;
	} # End of the 1st middle foreach.

	my @gblks_w_lb_into_sbj = my @gblks_w_lb_into_cmpl = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$sbjct_site]}) { # 2nd middle foreach.
	    my $rel = $rels_w_sbj->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_lb_into_cmpl, $gb;
	    } else {
		push @gblks_w_lb_into_sbj, $gb;
	    }
	} # End of the 2nd middle foreach.
	push @gblks_w_lb_into_sbj, $gb_incr_sbj;
	push @gblks_w_lb_into_cmpl, $gb_cmpl_old;

	my @gblks_w_rb_into_sbj = my @gblks_w_rb_into_cmpl = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$sbjct_site]}) { # 3rd middle foreach.
	    my $rel = $rels_w_sbj->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_rb_into_cmpl, $gb;
	    } else {
		push @gblks_w_rb_into_sbj, $gb;
	    }
	} # End of the 3rd middle foreach.
	push @gblks_w_rb_into_sbj, $gb_incr_sbj;
	push @gblks_w_rb_into_cmpl, $gb_cmpl_old;

	my @intm1_curr_set_clms = ();
	my @intm1_curr_gblks_w_spec_lb = my @intm1_curr_gblks_w_spec_rb = ();

	for (my $c=0; $c < $sbjct_site; $c++) { # 1st Middle MAIN for-loop (over sites on the left of $sbjct_site). #
	    $intm1_curr_set_clms[$c] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 1st Middle MAIN for-loop. #

	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@clm_incr_sbj, \@clm_gb_cmpl) : (\@clm_gb_cmpl, \@clm_incr_sbj) ;
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@gblks_w_lb_into_sbj, \@gblks_w_lb_into_cmpl) : (\@gblks_w_lb_into_cmpl, \@gblks_w_lb_into_sbj) ;
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@gblks_w_rb_into_sbj, \@gblks_w_rb_into_cmpl) : (\@gblks_w_rb_into_cmpl, \@gblks_w_rb_into_sbj) ;;

	for (my $c = $sbjct_site+1; $c <$curr_ct_clms; $c++) { # 2nd Middle MAIN for-loop (over sites on the right of $sbjct_site). #
	    $intm1_curr_set_clms[$c+1] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c+1] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c+1] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 2nd Middle MAIN for-loop. #


	my $intm1_ct_clms = @intm1_curr_set_clms;


		# Create @intm1_curr_bds_blks. #

	my @intm1_curr_bds_blks = ();
	for (my $gb=0; $gb < $intm1_ct_gbs; $gb++) { $intm1_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$intm1_ct_clms; $c++) { # 2nd trailing for-loop for Main 1-(i). #
            foreach my $gb (@{$intm1_curr_gblks_w_spec_lb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[0] = $c;
            }
            foreach my $gb (@{$intm1_curr_gblks_w_spec_rb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[1] = $c;
            }
	} # End of the 2nd trailing for-loop for Main 1-(i). #


		## Preliminary (In 2): Create the coordinate frame for the 1st intermediate gap-block configuration. #

	my ($intm1_le_wd, $intm1_re_wd) = (0, $intm1_ct_clms-1);

	my ($intm1_orgs_gb_coords, $intm1_bds_gb_coords, $intm1_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@intm1_curr_bds_blks, $intm1_le_wd, $intm1_re_wd, @intm1_inter_block_relations, $maxct_1way_shifts);
	unless (defined $intm1_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column : $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);

	my @intm1_curr_bl_coords = @{$intm1_orgs_gb_coords};


			# (Added on 2019/08/07. (Main II)) #

   if ($TEST_LEVEL_init_cspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_gblock_configs (@intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >5) {...}"

   if ($TEST_LEVEL_init_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_remng_attr_gblks (@intm1_inter_block_relations, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm}, @intm1_gblock_sizes, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 5) {...}"


		## MAIN II-(i): Merge the $gb_sbj and $gb_incr_sbj th gap-block, via 'smerge_gblock_pair'. ##

	## ($gb_cmpl_old, $gb_incr_sbj) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the subroutine 'smerge_gblock_pair', to merge the \$gb_sbj (=$gb_sbj) th and \$gb_incr_sbj (=$gb_incr_sbj) th gap-blocks ...\n\n") if ($TEST_LEVEL_init_cspl_gb > 1);

	my ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_sbj, $gb_incr_sbj, @intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm},
		@intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, @intm1_rows_affd_by_blks, $cnct_null_clm,
		@intm1_inter_block_relations, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_gblock_sizes, $maxct_1way_shifts
		);
	unless ($flag_end_smg1) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);



   if ($TEST_LEVEL_init_cspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'smerge_gblock_pair' (to \"merge\" \$gb_sbj (=$gb_sbj) and \$gb_incr_sbj (=$gb_incr_sbj) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >5) {...}"

   if ($TEST_LEVEL_init_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'smerge_gblock_pair' for (to \"merge\" \$gb_sbj (=$gb_sbj) and \$gb_incr_sbj (=$gb_incr_sbj) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm2_inter_block_relations}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm}, @{$intm2_gblock_sizes}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 5) {...}"


=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN II-(ii))

		## MAIN II-(ii): Merge the $gb_left and $gb_incr_left th gap-block, via the 2nd Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 2nd Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_left (=$gb_left) th and \$gb_incr_left (=$gb_incr_left) th gap-blocks ...\n\n") if ($TEST_LEVEL_init_cspl_gb > 1);

	my ($flag_end_smg2, $bl_sbj_smg2, $incr_coord_sbj_smg2, $gb_absbd_smg2, $gb_kpt_smg2, $new_indices_gbs_smg2, $old_indices_gbs_smg2,
	    $intm3_curr_set_clms, $intm3_curr_bds_blks, $intm3_curr_bl_coords, $intm3_bds_gb_coords, $intm3_gb_coords_wd_frm,
	    $intm3_curr_gblks_w_spec_lb, $intm3_curr_gblks_w_spec_rb, $intm3_rows_affd_by_blks,
	    $intm3_inter_block_relations, $intm3_cts_vincluding, $intm3_cts_vincluded, $intm3_cts_onx, $intm3_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_left, $gb_incr_left, @{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm},
		@{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, @{$intm2_rows_affd_by_blks}, $cnct_null_clm,
		@{$intm2_inter_block_relations}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_end_smg2) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The 2nd round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);


   if ($TEST_LEVEL_init_cspl_gb >= 5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm3_curr_set_clms}, @{$intm3_curr_bds_blks}, @{$intm3_curr_bl_coords}, @{$intm3_curr_gblks_w_spec_lb}, @{$intm3_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >= 5) {...}"

   if ($TEST_LEVEL_init_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm3_inter_block_relations}, @{$intm3_bds_gb_coords}, @{$intm3_gb_coords_wd_frm}, @{$intm3_gblock_sizes}, @{$intm3_cts_vincluding}, @{$intm3_cts_vincluded}, @{$intm3_cts_onx}, @{$intm3_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 5) {...}"


		## MAIN IIpost: Update Important Data-Sets
		##	that characterize the Updated gap-block configuration. ##

			# Update the set of columns. #
	@cp_curr_set_clms = @{$intm3_curr_set_clms};
	$curr_ct_clms = @cp_curr_set_clms;

			# Update the set of gap-block boundaries & its "inverse-functions". #
	@cp_curr_bds_blks = @{$intm3_curr_bds_blks};
	@cp_curr_gblks_w_spec_lb = @{$intm3_curr_gblks_w_spec_lb};
	@cp_curr_gblks_w_spec_rb = @{$intm3_curr_gblks_w_spec_rb};

			# Update the coordinate frame. #
	@cp_curr_bl_coords = @{$intm3_curr_bl_coords};
	@cp_bds_gb_coords = @{$intm3_bds_gb_coords};
	@cp_gb_coords_wd_frm = @{$intm3_gb_coords_wd_frm};

			# Update the sets of gap-block sizes. #
	@cp_curr_gblock_sizes = @{$intm3_gblock_sizes};
	@intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##@intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);

=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN II-(ii))"


		## Record the Core information on this Session. ##

	my @record_session = (1, $sbjct_site, $incr_coord_sbj_smg1, 'n/a'); # MODIFIED in creation of 'initial_csplit_gblock'. #
	##push @records_sessions, [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

			# End of "(Added on 2019/08/07. (Main II))" #

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN I & 2 (Back))
    } # End of the outer MAIN foreach-loop (over the sessions). #


  (print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the $incr_gb_size Sessions !!---\n\n\n") if ($TEST_LEVEL_init_cspl_gb >1);

=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN I & 2 (Back))"


		### MAIN IIIpre:  Create the correspondence between the old & new indices of the gap-blocks. ###
		### (ADDED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN IIIpost) ##


	my @intm2_list_gblocks = ();
	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my @info_gb = @{$intm1_list_gblocks[$gb]};
	    @info_gb[2,3] = @{$intm2_curr_bds_blks->[$gb]};
	    $intm2_list_gblocks[$gb] = \@info_gb;
	}

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$intm2_list_gblocks[$a]}, @{$intm2_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1);
	my @new_indices_old_gbs = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	}

	my $gb_sbj_NEW = $new_indices_old_gbs[$gb_sbj];
	my $gb_cmpl_NEW = $new_indices_old_gbs[$gb_cmpl_old];

   if ($TEST_LEVEL_init_cspl_gb > 2) {

	print STDOUT "\n[ Relationships between old and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ([\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";
	print STDOUT "\n\n";

	print STDOUT '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 2) {...}"


		### END of "(ADDED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN IIIpost)" ##

		### MAIN III: Create the Final Output Data-Sets. ###
		### (Added on 2019/08/07. (MAIN III)) ###
		### (=> MODIFIED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN III) ###

	my @new_curr_set_clms = @{$intm2_curr_set_clms};
	my $ct_out_clms = @new_curr_set_clms;

	my @new_curr_bds_blks = @{$intm2_curr_bds_blks}[@old_indices_new_gbs];

	my @new_gblock_sizes = @{$intm2_gblock_sizes}[@old_indices_new_gbs];
	my @rows_affd_by_new_blks = @{$intm2_rows_affd_by_blks}[@old_indices_new_gbs];

	# ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	#    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	#    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	#    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lbd, $rbd) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lbd]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rbd]}, $ngb;
	}

			# Create  @new_list_gblocks & @{$new_inter_block_relations}. #

	my @new_list_gblocks = @intm2_list_gblocks[@old_indices_new_gbs];

	my ($new_inter_block_relations)
	  = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@new_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@new_list_gblocks', " ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1) ;

			# Create @{$new_cts_vincluding}, etc. #

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'initial_csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@{$new_inter_block_relations}', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1) ;


			# Create @new_curr_bl_coords  & @new_bds_gb_coords  & @new_gb_coords_wd_frm . #

	my ($new_le_wd, $new_re_wd) = (0, $ct_out_clms-1);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks after Whole Process of INITIAL \"Split (complementary types)\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks after the Whole Process of the INITIAL \"Split (complementary types)\" ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);

	my @new_curr_bl_coords    = @{$new_orgs_gb_coords} ;
	my @new_bds_gb_coords     = @{$new_bds_gb_coords} ;
	my @new_gb_coords_wd_frm  = @{$new_gb_coords_wd_frm} ;

	##my @new_curr_bl_coords = @{$intm2_curr_bl_coords}[@old_indices_new_gbs];
	##my @new_bds_gb_coords = @{$intm2_bds_gb_coords}[@old_indices_new_gbs];
	##my @new_gb_coords_wd_frm = @{$intm2_gb_coords_wd_frm}[@old_indices_new_gbs];

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN III)

	my @new_curr_set_clms = @cp_curr_set_clms;
	my @new_curr_bds_blks = @cp_curr_bds_blks;

	my @new_curr_bl_coords = @cp_curr_bl_coords;
	my @new_bds_gb_coords = @cp_bds_gb_coords;
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm;

	my @new_curr_gblks_w_spec_lb = @cp_curr_gblks_w_spec_lb;
	my @new_curr_gblks_w_spec_rb = @cp_curr_gblks_w_spec_rb;
	my @new_gblock_sizes = @cp_curr_gblock_sizes;

		### END of "(Added on 2019/08/07. (MAIN III))" ###

		### MAIN IIIpost: UNDO the overlaps between the gap-blocks
		###		vertically equivalent to and/or complementary to $gb1 and $gb2. ###
		### (ADDED on 2019/08/08. (MAIN IIIpost)) ###

	my ($fl_undo, $gb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets ($gb_left, $gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'initial_csplit_gblock'] The subroutine, 'undo_hovlp_vcmpl_gbsets' FAILED on the cofiguration of gap-blocks after the main foreach-loop for \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The the subroutine, 'undo_hovlp_vcmpl_gbsets', Ended on the configuration of gap-blocks immediately after the main foreach-loop for the \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right) ...\n\n") if ($TEST_LEVEL_init_cspl_gb >1);


   if ($TEST_LEVEL_init_cspl_gb > 3) { # ADDED on 2019/08/09. #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets': ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 3) {...}"

		### END of "(ADDED on 2019/08/08. (MAIN IIIpost))" ###

=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (MAIN III)"

		### END of "(=> MODIFIED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN III)" ###


   if ($TEST_LEVEL_init_cspl_gb >2) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of INITIAL \"Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_cspl_gb >2) {...}"

   if ($TEST_LEVEL_init_cspl_gb > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of INITIAL \"Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).
				# MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02). #
	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
				# END of "MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02)." #
	print STDOUT "\n";
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version)."

   } # End of "if ($TEST_LEVEL_init_cspl_gb > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'initial_csplit_gblock' Ended (w/ ", '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) . >\n\n") if ($TEST_LEVEL_init_cspl_gb > 0);


	return (1, $gb_sbj_NEW, $gb_cmpl_NEW, \@new_indices_old_gbs, \@old_indices_new_gbs, \@record_session,  
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks);

}	# End of the subroutine, 'initial_csplit_gblock'. #


# csplit_gblock:
#	This subroutine "split"s the subject gap-block into a pair of gap-blocks of complementary types.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by incoporating a NEW satellite subroutine, 'initial_csplit_gblock'.
#	[CREATION STARTED on 2019/08/15, ENDED on 2019/08/16.]
#	[REVISED on 2019/08/17-19]:
#	Corrected a BUG; as a result, the subroutine can now order the gap-blocks CORRECTLY
#	even if the size-growing "split + merge" process changes
#	the intrinsic ranks between some vertically equivalent gap-blocks.
#
# cmerge_split_grow_gblock_pair (NEW VERSION):
#	This new version does essentially the same job as its old version (now renamed as 'cmerge_split_grow_gblock_pair_OLD'),
#	but UNDER a DIFFERENT STRATEGY.
#	NOTE that, accompanying this revision, the portfolio of the OUTPUT data has also CHANGED.
#	[CREATION STARTED on 2019/08/06, ENDED on 2019/08/07.]
#
#	[STRATEGY]:
#	UNLIKE the old version,
#	it attempts to grow the subject pair of gap-blocks
#	while leaving the existing gap-blocks as unmoved as possible,
#	by (i) enumerating the columns available for the "split",
#	and by (ii) actually "split"ting the column one by one,
#	and "merg"ing the resulting gap-blocks to the subject gap-blocks each time a column is "split".
#
#	[REVISED on 2019/08/08]:
#	Now, the subroutine performs 'undo_hovlp_vcmpl_gbsets' before returning the results,
#	in order to undo the horizontal overlaps resulting from the main process of "merge + split".
#	[REVisEd on 2019/08/09]:
#	Now, the subroutine also outputs \%gb2incr_bl_coord_undo,
#	which is one of the main outputs of 'undo_hovlp_vcmpl_gbsets'.
#
# cmerge_split_grow_gblock_pair:
#	This subroutine performs the "merge + split" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, and grows their sizes by a specified number of columns.
#	It also prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'icmerge_gblock_pair'.
#	[CREATED on 2019/08/03.]
#	[DEBUGGED on 2019/08/05]:
#	(i) Now, it should be able to CORRECTLY separate the gap-block boundaries
#	belonging to the columns to be "split".
#	(ii) CORRECTED the use of the output, %{$gb2sbincr_coord}, of 'displace_vincluding_veq_and_onx_to_right_end'.
#	[REVISED on 2019/08/05]:
#	The rules for assigning the coordinate increments for the 2nd (UNDOing) serial "shift"s,
#	so that the final result will look closer to the input alignment.
#
# icmerge_gblock_pair:
#	This subroutine performs the "INcomplete merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, by a specified number of columns.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/08/02.]
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, $incr_gb_size, $char_abs, 
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where ] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (NEW VERSION).
#
# INPUTS: ($gb_sbj, $size_cmpl, \@pot_avail_sites, $char_abs, 
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm,
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1)" #

#	), where
#
#			# ADDED in creation of 'csplit_gblock'. (HD1) #
#	$gb_sbj is the index of the gap-block to be "split" into a pair of vertically complementary blocks;
#
#	$size_cmpl is the final size of the newly created block vertically complementary to $gb_sbj;
#
#	@pot_avail_sites stores the indices of the sites (i.e., columns) POTENTIALLY available for the "Split (complementary types)", in the order to be used;
#			# End of "ADDED in creation of 'csplit_gblock'. (HD1)" #
#
#	[[
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#	$incr_gb_size is the number of columns by which the sizes of $gb1 and $gb2 will increase;
#	]] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (NEW VERSION).
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2)" #
#	
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# [ OUTPUTS: ($flag_end, $gb_left, $gb_right, \@records_sessions, \%gb2incr_bl_coord_undo, # REVISED on 2019/08/09. # ] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#
# OUTPUTS: ($flag_end, $gb_sbj_NEW, $gb_cmpl_NEW, $gb_left, $gb_right, \@new_indices_old_gbs, \@old_indices_new_gbs,
#	   \@records_sessions, \%gb2incr_bl_coord_undo,
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	[ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes ] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#			# IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD1) #
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#			# End of "IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD1)" #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_list_gblocks,
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1)" #
#	\@med_indices_new_gbs # ADDED in revision on 2019/08/19.
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#			# ADDED in creation of 'csplit_gblock'. (HD) #
#	$gb_sbj_NEW is the new index for the subject gap-block(, which grew in size during this subroutine);
#	$gb_cmpl_NEW is the new index for the vertically complementary gap-block;
#			# END of "ADDED in creation of 'csplit_gblock'. (HD)" #
#
#	$gb_left and $gb_right are the indices of the gap-blocks that are on the left and on the right, respectively, each chosen between $gb_sbj_NEW and $gb_cmpl_NEW;
#
#			# IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# End of "IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD2)" #
#
#			# ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#	@{$records_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	provides the record of the $k th session, where
#	$session_ID is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1/2 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_right/$gb_left th gap-block;
#			# End of "ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version)." #
#
#			
#			# ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09). #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09)." #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#			# ADDED in REVISION (on 2019/08/19). #
#	@med_indices_new_gbs records the re-ordering of the new gap-blocks
#	resulting from growing the complementary gap-block pair by accruing the products of "split"ting the available sites;
#	Specifically, $med_indices_new_gbs[$ngb] gives the index that
#	the $ngb th gap-block (in the final output data-sets) had immediately after the subroutine 'initial_csplit_gblock';
#			# END of "ADDED in REVISION (on 2019/08/19)." #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
# [ sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {
#	my ($gb1, $gb2, $incr_gb_size, $char_abs,
#	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
#	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
#	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_; ] ... for the predecessor, 'cmerge_split_grow_gblock_pair' (New Version).
#
sub csplit_gblock ($$\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb_sbj, $size_cmpl, $pot_avail_sites, $char_abs,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,

		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq
		# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (IN)" #
	) = @_;

# INPUTS: ($gb_sbj, $size_cmpl_gb, \@pot_avail_sites, $char_abs, 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

			# MODIFIED in creation of 'csplit_gblock'. (BEG) #

	unless ($size_cmpl <= @{$pot_avail_sites}) {
	    my $ct = @{$pot_avail_sites};
	    print STDOUT "[ERROR 'csplit_gblock'] The number of potentially available sites (=$ct) is SMALLER THAN the final size of the vertically complementary gap-block (=$size_cmpl): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my @avail_sites = ($size_cmpl < @{$pot_avail_sites}) ? @{$pot_avail_sites}[0 .. $size_cmpl-1] : @{$pot_avail_sites};

	my $size_sbj_bf = $gblock_sizes->[$gb_sbj];
	my $sbj_coord_bf = $curr_bl_coords->[$gb_sbj];

	my ($lbd_sbj_bf, $rbd_sbj_bf) = @{$curr_bds_blks->[$gb_sbj]};


    (print STDOUT "< Subroutine 'csplit_gblock' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; gap-block to be \"Split\": \$gb_sbj = $gb_sbj; {subject_block_size}_bf = $size_sbj_bf; {subject_coordinate}_bf = $sbj_coord_bf; {boundaries}_sbj_bf = [$lbd_sbj_bf, $rbd_sbj_bf]; {complementary_block_size}_fin = $size_cmpl; {available sites} = (", join (',', @avail_sites), "). ) >\n\n") if ($TEST_LEVEL_cspl_gb > 0);


   if ($TEST_LEVEL_cspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >3) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"


=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (1)
	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless ($incr_gb_size>0) { # Added on 2019/08/03. #
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The \$incr_gb_size (=$incr_gb_size) is NOT POSITIVE: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};

    (print STDOUT "< Subroutine 'cmerge_split_grow_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"INcomplete-merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$incr_gb_size = $incr_gb_size) >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (1)"

			# END of "MODIFIED in creation of 'csplit_gblock'. (BEG)" #

		### (MAIN I): Perform the Initial "csplit", using the 1st available site.

	my $avail_site_1st = shift @avail_sites;
	my $ct_curr_avail_sites = @avail_sites;
	for (my $i=0; $i <$ct_curr_avail_sites; $i++) {
	    if ($avail_site_1st < $avail_sites[$i]) { $avail_sites[$i]++; }
	}

    (print STDOUT "\n ... First, Perform the 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st = $avail_site_1st) ...\n\n") if ($TEST_LEVEL_cspl_gb>1);

	my ($flag_init_cs, $gb_sbj_NEW, $gb_cmpl_NEW, $new_indices_old_gbs, $old_indices_new_gbs,
	   $record_init_session,
	$new1_curr_set_clms, $new1_curr_bds_blks, $new1_curr_bl_coords, $new1_bds_gb_coords, $new1_gb_coords_wd_frm,
	$new1_curr_gblks_w_spec_lb, $new1_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new1_gblock_sizes,
	$new_list_gblocks
			)
	    = initial_csplit_gblock (
		$gb_sbj, $avail_site_1st, $char_abs,
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, 
		@{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, $maxct_1way_shifts,

		@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
		%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
			);
	unless ($flag_init_cs) {
	    print STDERR "[ERROR 'csplit_gblock'] The subroutine 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st = $avail_site_1st) FAILED on the input gap-block configuration: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st  = $avail_site_1st) Ended on the input gap-block configuration (w/ \$flag_init_cs = $flag_init_cs, \$gb_sbj_NEW = $gb_sbj_NEW, \$gb_cmpl_NEW = $gb_cmpl_NEW) ...\n\n") if ($TEST_LEVEL_cspl_gb >1) ;

	my $ct_in2_clms = @{$new1_curr_set_clms};
	my $ct_in2_gbs = @{$new1_curr_bds_blks};


   if ($TEST_LEVEL_cspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations output by the subroutine, 'initial_csplit_gblock' ]\n\n";

	unless ( print_out_gblock_configs (@{$new1_curr_set_clms}, @{$new1_curr_bds_blks}, @{$new1_curr_bl_coords}, @{$new1_curr_gblks_w_spec_lb}, @{$new1_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >5) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks by the subroutine, 'initial_csplit_gblock' ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new1_bds_gb_coords}, @{$new1_gb_coords_wd_frm}, @{$new1_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"


		### (MAIN II): Grow the subject and the newly created v-complementary gap-blocks,
		###	by "csplit"ting the available sites one by one.

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		# (=> MODIFIED in DEBUGGING 'csplit_gblock' (on 2019/08/17). (Pre 0)) #

	my ($gb_left, $gb_right) = ($new1_curr_bds_blks->[$gb_sbj_NEW][0] < $new1_curr_bds_blks->[$gb_cmpl_NEW][0]) ? ($gb_sbj_NEW, $gb_cmpl_NEW) : ($gb_cmpl_NEW, $gb_sbj_NEW); # MODIFIED in creation of 'csplit_gblock'. #

	my ($size_left, $size_right) = @{$new1_gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$new1_curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$new1_curr_bds_blks->[$gb_right]};

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 0)
	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1);

	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 0)"

		# END of "(=> MODIFIED in DEBUGGING 'csplit_gblock' (on 2019/08/17). (Pre 0))" #


   if ($TEST_LEVEL_cspl_gb >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


			# (MODIFIED in creation of 'csplit_gblock'. (Pre 1)) #

		# (Preliminary 1): NOMINALLY Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = @{$new1_curr_set_clms};

	my @cp_curr_gblks_w_spec_lb = @{$new1_curr_gblks_w_spec_lb};
	my @cp_curr_gblks_w_spec_rb = @{$new1_curr_gblks_w_spec_rb};

	my @cp_curr_bds_blks = @{$new1_curr_bds_blks};
	my @cp_bds_gb_coords = @{$new1_bds_gb_coords};
	my @cp_gb_coords_wd_frm = @{$new1_gb_coords_wd_frm};
	my @cp_rows_affd_by_blks = @{$rows_affd_by_new_blks};

	my @cp_curr_bl_coords = @{$new1_curr_bl_coords};

	# $new1_curr_set_clms, $new1_curr_bds_blks, $new1_curr_bl_coords, $new1_bds_gb_coords, $new1_gb_coords_wd_frm,
	# $new1_curr_gblks_w_spec_lb, $new1_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	# $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new1_gblock_sizes


=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 1)
		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >3) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"

=cut # End of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre 1)"

			# END of "(MODIFIED in creation of 'csplit_gblock'. (Pre 1))" #


		### Preliminary (Out): ###

	my ($gb_incr_left, $gb_incr_right) = ($ct_in2_gbs, $ct_in2_gbs+1);
	##my ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

	my $intm1_ct_gbs = $ct_in2_gbs +2;
	my $intm2_ct_gbs = $ct_in2_gbs +1;
	##my $intm1_ct_gbs = $ct_in_gbs +2;
	##my $intm2_ct_gbs = $ct_in_gbs +1;

		# Create @intm1_gblock_sizes & @intm1_inter_block_relations. #

	my @cp_curr_gblock_sizes = @{$new1_gblock_sizes};
	##my @cp_curr_gblock_sizes = @{$gblock_sizes};
	my @intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##my @intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);
		# NOTE: The above 3 arrays need be UPDATED in every session. #

	my @intm1_inter_block_relations = my @intm2_inter_block_relations = ();
	for (my $gb=0; $gb <$ct_in2_gbs; $gb++) {
	##for (my $gb=0; $gb <$ct_in_gbs; $gb++) {
	    my @intm1_rels_w_gb = @{$new_inter_block_relations->[$gb]};
	    ##my @intm1_rels_w_gb = @{$inter_block_relations->[$gb]};
	    ####my @intm2_rels_w_gb = @{$inter_block_relations->[$gb]};

	    if ($gb == $gb_left) {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1] = ('=', 'Cp');
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('=', 'Cp');
		####$intm2_rels_w_gb[$ct_in_gbs] = '=';
	    } elsif ($gb == $gb_right) {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1] = ('Cp', '=');
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('Cp', '=');
		####$intm2_rels_w_gb[$ct_in_gbs] = 'Cp';
	    } else {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1]
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1]
		 = @intm1_rels_w_gb[$gb_left, $gb_right];
		####$intm2_rels_w_gb[$ct_in_gbs] = $intm2_rels_w_gb[$gb_left];
	    }
	    $intm1_inter_block_relations[$gb] = \@intm1_rels_w_gb;
	    ##$intm2_inter_block_relations[$gb] = \@intm2_rels_w_gb;
	}
	my @intm1_rels_w_incr_left = @{$intm1_inter_block_relations[$gb_left]};
	@intm1_rels_w_incr_left[$gb_left, $ct_in2_gbs] 
		= @{$intm1_inter_block_relations[$gb_left]}[$ct_in2_gbs, $gb_left];
	##@intm1_rels_w_incr_left[$gb_left, $ct_in_gbs] 
	##	= @{$intm1_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];

	my @intm1_rels_w_incr_right = @{$intm1_inter_block_relations[$gb_right]};
	@intm1_rels_w_incr_right[$gb_right, $ct_in2_gbs+1]
		= @{$intm1_inter_block_relations[$gb_right]}[$ct_in2_gbs+1, $gb_right];
	##@intm1_rels_w_incr_right[$gb_right, $ct_in_gbs+1]
	##	= @{$intm1_inter_block_relations[$gb_right]}[$ct_in_gbs+1, $gb_right];

	$intm1_inter_block_relations[$ct_in2_gbs] = \@intm1_rels_w_incr_left;
	##$intm1_inter_block_relations[$ct_in_gbs] = \@intm1_rels_w_incr_left;

	$intm1_inter_block_relations[$ct_in2_gbs+1] = \@intm1_rels_w_incr_right;
	##$intm1_inter_block_relations[$ct_in_gbs+1] = \@intm1_rels_w_incr_right;

=Set aside for Later Use.
	my @intm2_rels_w_incr_left = @{$intm2_inter_block_relations[$gb_left]};
	@intm2_rels_w_incr_left[$gb_left, $ct_in_gbs] 
		= @{$intm2_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];
	$intm2_inter_block_relations[$ct_in_gbs] = \@intm2_rels_w_incr_left;
=cut # END of "Set aside for Later Use."


		# Create $intm1_cts_vincluding, etc. #

	my ($intm1_cts_vincluding, $intm1_cts_vincluded, $intm1_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm1_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm1_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm1_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_cspl_gb >1) ;


=Set aside for Later Use.
	my ($intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm2_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm2_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm2_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_cspl_gb >1) ;
=cut # END of "Set aside for Later Use."



		# Create @intm1_rows_affd_by_blks.

	my @intm1_rows_affd_by_blks = (@cp_rows_affd_by_blks, @cp_rows_affd_by_blks[$gb_left, $gb_right]);
	##my @intm2_rows_affd_by_blks = (@cp_rows_affd_by_blks, $cp_rows_affd_by_blks[$gb_left]);



		### MAIN 1 & 2: Iteratively Perform the single-column-growing "Merge + Split"s. ###

	my ($rows_affd_by_left, $rows_affd_by_right) = @cp_rows_affd_by_blks[$gb_left, $gb_right];
	my ($rels_w_left, $rels_w_right) = @{$new_inter_block_relations}[$gb_left, $gb_right];
	##my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];

	my $curr_size_cmpl = 1;
	##my $curr_incr_gb_size = $incr_gb_size +1;

	my $curr_ct_clms = $ct_in2_clms;
	##my $curr_ct_clms = $ct_in_clms;

	my @records_sessions = ($record_init_session); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2]; # MODIFIED in creation of 'csplit_gblock'. #
	##my @records_sessions = (); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

     foreach my $session (2 .. $size_cmpl) { # Outer main foreach-loop (over the sessions). #
     ##foreach my $session (1 .. $incr_gb_size) { # Outer main foreach-loop (over the sessions). #

	##$curr_incr_gb_size--;
	my ($le_wd, $re_wd) = (0, $curr_ct_clms-1);

    (print STDOUT "\n{{ Session $session:  (\$curr_size_cmpl = $curr_size_cmpl, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_cspl_gb > 1); # MODIFIED in creation of 'csplit_gblock'. #
    ##(print STDOUT "\n{{ Session $session:  (\$curr_incr_gb_size = $curr_incr_gb_size, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_cspl_gb > 1);


			# MODIFIED in creation of 'csplit_gblock' (on 2019/08/16). (Pre (In)) #

		# Preliminary (In):  Extract the subject site, and update @avail_sites. 

	my $sbjct_site = shift @avail_sites;
	my $ct_curr_avail_sites = @avail_sites;
	for (my $i=0; $i <$ct_curr_avail_sites; $i++) {
	    if ($sbjct_site < $avail_sites[$i]) { $avail_sites[$i]++; }
	}

   if ($TEST_LEVEL_cspl_gb >3) {

	print STDOUT "[ Current sets of available sites: ]\n\n";

	print STDOUT "\$sbjct_site = $sbjct_site .\n\n";
	print STDOUT 'Current @avail_sites = (', join (', ', @avail_sites), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >3) {...}"

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre (In))

		# Preliminary (In):  Enumerate the columns available for the "(size-growing) Merge + Split".

	my ($avail_sites_left, $avail_sites_in_btwn, $avail_sites_right)
	    = enum_sites_avail_for_csplit ($gb1, $gb2, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_bds_gb_coords, @{$inter_block_relations}, $le_wd, $re_wd);
	unless (defined $avail_sites_right) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'enum_sites_avail_for_csplit', FAILED on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'enum_sites_avail_for_csplit', Ended on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session ...\n\n") if ($TEST_LEVEL_cspl_gb >1);

	my @avail_sites = (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left}) );
	my $ct_avail_sites = @avail_sites;

	unless ($ct_avail_sites >= $curr_incr_gb_size) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'enum_sites_avail_for_csplit' (for session $session), output ONLY $ct_avail_sites (< \$curr_incr_gb_size (=$curr_incr_gb_size)) sites available for the \"(size-growing) merge + split\": $! !\n\n";
	    ##return (0, undef, undef, undef);
	    return (-1, undef, undef, undef); # REVISED on 2019/08/07. #
	}

	my $sbjct_site = $avail_sites[0];

   if ($TEST_LEVEL_cspl_gb >3) {

	print STDOUT "\n[ Secondary Outputs of 'enum_sites_avail_for_csplit' (#{available sites}=$ct_avail_sites): ]\n\n";

	print STDOUT '@avail_sites (= (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left})) )', "\n";
	print STDOUT ' = (', join (', ', @avail_sites), ") .\n\n";
	print STDOUT '$sbjct_site = ', "$sbjct_site .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >3) {...}"

=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version). (Pre (In))"

			# END of "MODIFIED in creation of 'csplit_gblock' (on 2019/08/16). (Pre (In))" #


		## MAIN I: Create a new pair of single-column gap-blocks, by "split"ting the $sbjct_site th site. ##

	my @clm_incr_left = @null_clm;
	my @clm_incr_right = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_right}) { # 1st middle foreach.
	    my $tmp = $clm_incr_right[$row];
	    $clm_incr_right[$row] = $clm_incr_left[$row];
	    $clm_incr_left[$row] = $tmp;
	} # End of the 1st middle foreach.

	my @gblks_w_lb_into_left = my @gblks_w_lb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$sbjct_site]}) { # 2nd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_lb_into_left, $gb;
	    } else {
		push @gblks_w_lb_into_right, $gb;
	    }
	} # End of the 2nd middle foreach.
	push @gblks_w_lb_into_left, $ct_in2_gbs;
	push @gblks_w_lb_into_right, $ct_in2_gbs +1;
	##push @gblks_w_lb_into_left, $ct_in_gbs;
	##push @gblks_w_lb_into_right, $ct_in_gbs +1;

	my @gblks_w_rb_into_left = my @gblks_w_rb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$sbjct_site]}) { # 3rd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_rb_into_left, $gb;
	    } else {
		push @gblks_w_rb_into_right, $gb;
	    }
	} # End of the 3rd middle foreach.
	push @gblks_w_rb_into_left, $ct_in2_gbs;
	push @gblks_w_rb_into_right, $ct_in2_gbs +1;
	##push @gblks_w_rb_into_left, $ct_in_gbs;
	##push @gblks_w_rb_into_right, $ct_in_gbs +1;

	my @intm1_curr_set_clms = ();
	my @intm1_curr_gblks_w_spec_lb = my @intm1_curr_gblks_w_spec_rb = ();

	for (my $c=0; $c < $sbjct_site; $c++) { # 1st Middle MAIN for-loop (over sites on the left of $sbjct_site). #
	    $intm1_curr_set_clms[$c] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 1st Middle MAIN for-loop. #

	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = (\@clm_incr_left, \@clm_incr_right);
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_lb_into_left, \@gblks_w_lb_into_right);
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_rb_into_left, \@gblks_w_rb_into_right);

	for (my $c = $sbjct_site+1; $c <$curr_ct_clms; $c++) { # 2nd Middle MAIN for-loop (over sites on the right of $sbjct_site). #
	    $intm1_curr_set_clms[$c+1] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c+1] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c+1] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 2nd Middle MAIN for-loop. #


	my $intm1_ct_clms = @intm1_curr_set_clms;


		# Create @intm1_curr_bds_blks. #

	my @intm1_curr_bds_blks = ();
	for (my $gb=0; $gb < $intm1_ct_gbs; $gb++) { $intm1_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$intm1_ct_clms; $c++) { # 2nd trailing for-loop for Main 1-(i). #
            foreach my $gb (@{$intm1_curr_gblks_w_spec_lb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[0] = $c;
            }
            foreach my $gb (@{$intm1_curr_gblks_w_spec_rb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[1] = $c;
            }
	} # End of the 2nd trailing for-loop for Main 1-(i). #


		## Preliminary (In 2): Create the coordinate frame for the 1st intermediate gap-block configuration. #

	my ($intm1_le_wd, $intm1_re_wd) = (0, $intm1_ct_clms-1);

	my ($intm1_orgs_gb_coords, $intm1_bds_gb_coords, $intm1_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@intm1_curr_bds_blks, $intm1_le_wd, $intm1_re_wd, @intm1_inter_block_relations, $maxct_1way_shifts);
	unless (defined $intm1_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_cspl_gb >1);

	my @intm1_curr_bl_coords = @{$intm1_orgs_gb_coords};


			# (Added on 2019/08/07. (Main II)) #

   if ($TEST_LEVEL_cspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_gblock_configs (@intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >5) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_remng_attr_gblks (@intm1_inter_block_relations, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm}, @intm1_gblock_sizes, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"


		## MAIN II-(i): Merge the $gb_right and $gb_incr_right th gap-block, via the 1st Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 1st Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_right (=$gb_right) th and \$gb_incr_right (=$gb_incr_right) th gap-blocks ...\n\n") if ($TEST_LEVEL_cspl_gb > 1);

	my ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_right, $gb_incr_right, @intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm},
		@intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, @intm1_rows_affd_by_blks, $cnct_null_clm,
		@intm1_inter_block_relations, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_gblock_sizes, $maxct_1way_shifts
		);
	unless ($flag_end_smg1) {
	    print STDOUT "[ERROR 'csplit_gblock'] The 1st round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_cspl_gb >1);


   if ($TEST_LEVEL_cspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >5) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm2_inter_block_relations}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm}, @{$intm2_gblock_sizes}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"



		## MAIN II-(ii): Merge the $gb_left and $gb_incr_left th gap-block, via the 2nd Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 2nd Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_left (=$gb_left) th and \$gb_incr_left (=$gb_incr_left) th gap-blocks ...\n\n") if ($TEST_LEVEL_cspl_gb > 1);

	my ($flag_end_smg2, $bl_sbj_smg2, $incr_coord_sbj_smg2, $gb_absbd_smg2, $gb_kpt_smg2, $new_indices_gbs_smg2, $old_indices_gbs_smg2,
	    $intm3_curr_set_clms, $intm3_curr_bds_blks, $intm3_curr_bl_coords, $intm3_bds_gb_coords, $intm3_gb_coords_wd_frm,
	    $intm3_curr_gblks_w_spec_lb, $intm3_curr_gblks_w_spec_rb, $intm3_rows_affd_by_blks,
	    $intm3_inter_block_relations, $intm3_cts_vincluding, $intm3_cts_vincluded, $intm3_cts_onx, $intm3_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_left, $gb_incr_left, @{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm},
		@{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, @{$intm2_rows_affd_by_blks}, $cnct_null_clm,
		@{$intm2_inter_block_relations}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_end_smg2) {
	    print STDOUT "[ERROR 'csplit_gblock'] The 2nd round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session ...\n\n") if ($TEST_LEVEL_cspl_gb >1);


   if ($TEST_LEVEL_cspl_gb >= 5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm3_curr_set_clms}, @{$intm3_curr_bds_blks}, @{$intm3_curr_bl_coords}, @{$intm3_curr_gblks_w_spec_lb}, @{$intm3_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >= 5) {...}"

   if ($TEST_LEVEL_cspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm3_inter_block_relations}, @{$intm3_bds_gb_coords}, @{$intm3_gb_coords_wd_frm}, @{$intm3_gblock_sizes}, @{$intm3_cts_vincluding}, @{$intm3_cts_vincluded}, @{$intm3_cts_onx}, @{$intm3_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 5) {...}"


		## MAIN IIpost: Update Important Data-Sets
		##	that characterize the Updated gap-block configuration. ##

			# Update the set of columns. #
	@cp_curr_set_clms = @{$intm3_curr_set_clms};
	$curr_ct_clms = @cp_curr_set_clms;

			# Update the set of gap-block boundaries & its "inverse-functions". #
	@cp_curr_bds_blks = @{$intm3_curr_bds_blks};
	@cp_curr_gblks_w_spec_lb = @{$intm3_curr_gblks_w_spec_lb};
	@cp_curr_gblks_w_spec_rb = @{$intm3_curr_gblks_w_spec_rb};

			# Update the coordinate frame. #
	@cp_curr_bl_coords = @{$intm3_curr_bl_coords};
	@cp_bds_gb_coords = @{$intm3_bds_gb_coords};
	@cp_gb_coords_wd_frm = @{$intm3_gb_coords_wd_frm};

			# Update the sets of gap-block sizes. #
	@cp_curr_gblock_sizes = @{$intm3_gblock_sizes};
	@intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##@intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);


		## Record the Core information on this Session. ##

	push @records_sessions, [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

			# End of "(Added on 2019/08/07. (Main II))" #

    } # End of the outer MAIN foreach-loop (over the sessions). #


  (print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the Sessions,  2 .. $size_cmpl !!---\n\n\n") if ($TEST_LEVEL_cspl_gb >1);
  ##(print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the $incr_gb_size Sessions !!---\n\n\n") if ($TEST_LEVEL_cspl_gb >1);


		### MAIN III: Create the Final Output Data-Sets. ###
		### (Added on 2019/08/07. (MAIN III)) ###

	my @new_curr_set_clms = @cp_curr_set_clms;
	my @new_curr_bds_blks = @cp_curr_bds_blks;

	my @new_curr_bl_coords = @cp_curr_bl_coords;
	my @new_bds_gb_coords = @cp_bds_gb_coords;
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm;

	my @new_curr_gblks_w_spec_lb = @cp_curr_gblks_w_spec_lb;
	my @new_curr_gblks_w_spec_rb = @cp_curr_gblks_w_spec_rb;
	my @new_gblock_sizes = @cp_curr_gblock_sizes;

		### END of "(Added on 2019/08/07. (MAIN III))" ###


		### MAIN III': Re-Rank the resulting Gap-Blocks, if necessary. ###
		### (ADDED for DEBUGGING on 2019/08/18. (MAIN III')) ###

	##my @new_indices_old_gbs = @{$new_indices_old_gbs};
	##my @old_indices_new_gbs = @{$old_indices_new_gbs};

		# Fixed a Minor BUG. #
	my $ct_out_gbs = @new_curr_bds_blks;
	my $ct_out_clms = @new_curr_set_clms;
	##my $ct_out_gbs = $ct_in2_gbs;
	##my $ct_out_clms = $ct_in2_clms;

	my @tmp_new_list_gblocks = ();
	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my @info_gb = @{$new_list_gblocks->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};
	    $tmp_new_list_gblocks[$gb] = \@info_gb;
	}

	my @new_indices_new2_gbs = sort { cmp_gblocks3 (@{$tmp_new_list_gblocks[$a]}, @{$tmp_new_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1);

	my $if_reordered = 0;
	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {
	    if ($new_indices_new2_gbs[$n2gb] != $n2gb) {
		$if_reordered = 1;
		last;
	    }
	}

    if ($if_reordered) {

  (print STDOUT "[INRORMATION 'csplit_gblock'] AFTER the \"Split (complementary types)\", The the gap-blocks are RE-ORDERED (w/ ", '@new_indices_new2_gbs = (', join (',', @new_indices_new2_gbs), ") !!\n\n") if ($TEST_LEVEL_cspl_gb> 0);
  (print STDOUT " ... Accordingly, the output data-sets will be UPDATED ...\n\n") if ($TEST_LEVEL_cspl_gb >1);

	my @new2_indices_new_gbs = ();
	my @new2_indices_old_gbs = my @old_indices_new2_gbs = ();

	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {

	    my $ngb = $new_indices_new2_gbs[$n2gb];
	    $new2_indices_new_gbs[$ngb] = $n2gb;

	    my $ogb = $old_indices_new_gbs->[$ngb];
	    $new2_indices_old_gbs[$ogb] = $n2gb;
	    $old_indices_new2_gbs[$n2gb] = $ogb;
	}
	$new_indices_old_gbs = \@new2_indices_old_gbs;
	$old_indices_new_gbs = \@old_indices_new2_gbs;

    if ($TEST_LEVEL_cspl_gb >2) {

	print STDOUT 'Updated @{$new_indices_old_gbs} = (', join (', ', @{$new_indices_old_gbs}), ") ;\n";
	print STDOUT 'Updated @{$old_indices_new_gbs} = (', join (', ', @{$old_indices_new_gbs}), ") .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_cspl_gb >2) {...}"


			# Update the KEY gap-block indices. #

	$gb_sbj_NEW = $new2_indices_new_gbs[$gb_sbj_NEW];
	$gb_cmpl_NEW = $new2_indices_new_gbs[$gb_cmpl_NEW];
	$gb_left = $new2_indices_new_gbs[$gb_left];
	$gb_right = $new2_indices_new_gbs[$gb_right];

    if ($TEST_LEVEL_cspl_gb >2) {

	print STDOUT 'Updated ($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) ;\n";
	print STDOUT 'Updated ($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_cspl_gb >2) {...}"

			# Update the data-sets. #

	@new_curr_bds_blks = @cp_curr_bds_blks[@new_indices_new2_gbs];
	@new_curr_gblks_w_spec_lb = @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$n2gb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $n2gb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $n2gb;
	}


	@new_gblock_sizes = @cp_curr_gblock_sizes[@new_indices_new2_gbs];

	my @rows_affd_by_new2_blks = @{$rows_affd_by_new_blks}[@new_indices_new2_gbs];
	$rows_affd_by_new_blks = \@rows_affd_by_new2_blks;

			# Re-Create  @{$new_inter_block_relations}. #

	@tmp_new_list_gblocks = @tmp_new_list_gblocks[@new_indices_new2_gbs];
	$new_list_gblocks = \@tmp_new_list_gblocks;

	($new_inter_block_relations)
	  = mk_inter_block_relations (@tmp_new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@tmp_new_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@tmp_new_list_gblocks', " ...\n\n") if ($TEST_LEVEL_cspl_gb >1) ;

			# Re-Create @{$new_cts_vincluding}, etc. #

	($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'csplit_gblock'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@{$new_inter_block_relations}', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_cspl_gb >1) ;


			# Re-Create @new_curr_bl_coords  & @new_bds_gb_coords  & @new_gb_coords_wd_frm . #

	my ($new_le_wd, $new_re_wd) = (0, $ct_out_clms-1);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks after the \"Split (complementary types)\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks after the \"Split (complementary types)\" ...\n\n") if ($TEST_LEVEL_cspl_gb >1);

	@new_curr_bl_coords    = @{$new_orgs_gb_coords} ;
	@new_bds_gb_coords     = @{$new_bds_gb_coords} ;
	@new_gb_coords_wd_frm  = @{$new_gb_coords_wd_frm} ;

    } # END of "if ($if_reordered) {...}"

		### END of "(ADDED for DEBUGGING on 2019/08/18. (MAIN III'))" ###


		### MAIN IIIpost: UNDO the overlaps between the gap-blocks
		###		vertically equivalent to and/or complementary to $gb1 and $gb2. ###
		### (ADDED on 2019/08/08. (MAIN IIIpost)) ###

	my ($fl_undo, $gb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets ($gb_left, $gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @{$rows_affd_by_new_blks}, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'csplit_gblock'] The subroutine, 'undo_hovlp_vcmpl_gbsets' FAILED on the cofiguration of gap-blocks after the main foreach-loop for \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The the subroutine, 'undo_hovlp_vcmpl_gbsets', Ended on the configuration of gap-blocks immediately after the main foreach-loop for the \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right) ...\n\n") if ($TEST_LEVEL_cspl_gb >1);


   if ($TEST_LEVEL_cspl_gb > 3) { # ADDED on 2019/08/09. #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets': ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 3) {...}"


		### END of "(ADDED on 2019/08/08. (MAIN IIIpost))" ###

   if ($TEST_LEVEL_cspl_gb >2) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cspl_gb >2) {...}"

   if ($TEST_LEVEL_cspl_gb > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

=For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).
				# MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02). #
	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'csplit_gblock'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'csplit_gblock' ...\n\n";
	}
				# END of "MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02)." #
	print STDOUT "\n";
=cut # END of "For the predecessor 'cmerge_split_grow_gblock_pair' (New Version)."

   } # End of "if ($TEST_LEVEL_cspl_gb > 3) {...}"

			# (Added in creation of 'csplit_gblock'.) (FIN) #

		# Create @new_list_gblocks. #

	my @new_list_gblocks = ();
	for (my $gb=0; $gb <$ct_in2_gbs; $gb++) {
	    my @info_gb = @{$new_list_gblocks->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};
	    $new_list_gblocks[$gb] = \@info_gb;
	}

	my $med_indices_new_gbs = \@new_indices_new2_gbs; # ADDED in revision on 2019/08/19.

	my $fl_end = $fl_undo;
			# END of "(Added in creation of 'csplit_gblock'.) (FIN)" #

    (print STDOUT "\n< Subroutine 'csplit_gblock' Ended (w/ \$fl_end = $fl_end, ", '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW)", '($gb_left, $gb_right) = ', "($gb_left, $gb_right) . >\n\n") if ($TEST_LEVEL_cspl_gb > 0);

	return ($fl_end, $gb_sbj_NEW, $gb_cmpl_NEW, $gb_left, $gb_right, 
	$new_indices_old_gbs, $old_indices_new_gbs, \@records_sessions, $gb2incr_bl_coord_undo,

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks, $med_indices_new_gbs); # ADDED "$med_indices_new_gbs" on 2019/08/19.


}	# End of the subroutine, 'csplit_gblock'. #


# vsplit_gblock:
#	This subroutine "vertically split"s the subject gap-block into a pair of gap-blocks that are the "effective children" of the subject,
#	then, it "shift"s one of the resulting child blocks to that they will NOT be nested.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' above.
#	[CREATION STARTED on 2019/08/22, ENDED on 2019/08/23.]
#	
# cvmerge_gblock_pair:
#	This subroutine actually performs the "(incomplete) vertical-merge" of a specified pair of
#	complementary-sibling gap-blocks.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_gblock_pair'.
#	[CREATION STARTED on 2019/07/23, ENDED on 2019/07/27.]
#	[REVISED on 2019/07/29]:
#	SWAPPED the subroutines 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right' in order,
#	and post-processed the output of 'displace_including_and_onx_to_right',
#	in order to make the final result as close as possible to the desired result.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 1)" #
#	), where ] ... for the predecessor, 'cvmerge_gblock_pair'.
#
# INPUTS: ($gb_sbj, \@info_effch1, \@info_effch2, # MODIFIED in creation of 'vsplit_gblock'. (HD1a) #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#	), where
#
#		# MODIFIED in creation of 'vsplit_gblock'. (HD1b) #
#	$gb_sbj is the index of the gap-block to be "vertically split";
#
#	@info_effch1 and @info_effch2 provide information on the TWO effective children of $gb_sbj;
#	more specifically,
#	@info_effch1/2 = ($br_ch, $up_or_lw_ch, \@exnodes_affd_by_ch, \@rows_affd_by_ch), where
#	the 1st/2nd child is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@exnodes_affd_by_ch stores the indices of external nodes affected by the child,
#	and @rows_affd_by_ch stores the indices of the MSA rows affected by the child.
#	[Typically, \@info_effch1 and \@info_effch2 are parts of the output of 'list_vsplit_candidates' above];
#
#		# ENd of "MODIFIED in creation of 'vsplit_gblock'. (HD1b)" #
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#			# (Added on 2019/07/26.) #
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#			# End of "(Added on 2019/07/26.)" #
#
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 2)" #
#	
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs, ] ... for the predecessor, 'cvmerge_gblock_pair'.
#
# OUTPUTS: ($flag_end, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2a) #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	@new_list_excluded_bin_hrels # ADDED in creation of 'vsplit_gblock'. (HD1a) #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; # IMPORTED from the header of 'csplit_gblock'. #
#
#			# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2b) #
#
#	$gb_ch1_NEW and $gb_ch2_NEW are the NEW indices of the 1st and 2nd child, respectively, of $gb_sbj;
#
#	$gb_shftd is the index (in the NEW gap-block set) of the child gap-block that was "shift"ed,
#	to make the two children non-nested;
#
#	$incr_coord_shftd (= +1 or -1) is the increment of the coordinate of $gb_shftd via the aforementioned "shift";
#			# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2b)" #
#
#	[[
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#		# END of "ADDED in revision on 2019/07/15." #
#
#		# ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
#	$if_complete = 1 if the "vertical-merge" is complete,
#		     = 0 if it is incomplete;
#		# End of "ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24)." #
#
#		# ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
#		# => MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). #
#	$gb_rplcd, $gb_rmvd_kpt, respectively, are the old indices of the gap-blocks that was replaced (with the "vmerged" blocks) and that was either kept (when $if_complete = 0) or removed (when $if_complete = 1);
#		# End of "ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
#		# => MODIFIED in creation of 'cvmerge_gblock_pair' (on 2019/07/24)." #
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "(incomplete) vertical-merge" (= undef if an error occurs);
#	]] ... For the predecessor, 'cvmerge_gblock_pair'. 
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vertical-split";
#	More precisely, the gap-block that had been the $gb th before the "vertical-split"
#	has been assigned $new_indices_old_gbs[$gb] after the "vertical-split"
#	(, and $new_indices_old_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vertical-split";
#	More precisely, the gap-block that is the $gb th after the "vertical-split"
#	had been assigned $old_indices_new_gbs[$gb] before the "vertical-split";
#
#			# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c) #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "vertical-split";
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 5) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 5)" #
#
#		# ADDED in creation of 'vsplit_gblock'. (HD1b) #
#	@new_list_excluded_bin_hrels = ([$gb_ch1_NEW, $gb_ch2_NEW, '='])
#	is the NEW list of PROHIBITED binary horizontal relationships,
#	which conforms to the format prescribed in the header of 'nonrec_mlt_shifts_gapblocks' above;
#		# End of "ADDED in creation of 'vsplit_gblock'. (HD1b)" #
#
#sub cvmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb1, $gb2, 
#	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
#	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
#	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,
#		# Added in creation of 'cvmerge_gblock_pair'. (IN) #
#	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
#		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
#	$node2pa, $node2ch, $topnode_id, $br2eq # (Added this line on 2019/07/26.)
#		# End of "Added in creation of 'cvmerge_gblock_pair'. (IN)" #
#	) = @_; 
#
sub vsplit_gblock ($\@\@\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb_sbj, $info_effch1, $info_effch2, 
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq
	) = @_; 

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $gb_coord_sbj = $curr_bl_coords->[$gb_sbj];
	my ($lbd_sbj, $rbd_sbj) = @{$curr_bds_blks->[$gb_sbj]};

	my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks0->[$gb_sbj]}[$indx_br, $indx_u_or_d];
	my $cnct_rows_sbj = join (',', @{$rows_affd_by_blks->[$gb_sbj]});

    (print STDOUT "< Subroutine 'vsplit_gblock' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; The \"Subject\" Gap-Block to be \"vertically split\": \$gb_sbj = $gb_sbj, {coordinate} = $gb_coord_sbj, [lbd, rbd]_sbj = [$lbd_sbj, $rbd_sbj], \$br_sbj = $br_sbj, \$up_or_lw_sbj = $up_or_lw_sbj, [rows,affected]_sbj=[$cnct_rows_sbj]) >\n\n") if ($TEST_LEVEL_vspl_gb > 0);

	my ($br_ch1, $up_or_lw_ch1, $exnodes_affd_by_ch1, $rows_affd_by_ch1) = @{$info_effch1};
	my ($br_ch2, $up_or_lw_ch2, $exnodes_affd_by_ch2, $rows_affd_by_ch2) = @{$info_effch2};


   if ($TEST_LEVEL_vspl_gb >1) { # (Top) #

	my $cnct_rows_ch1 = join (',', @{$rows_affd_by_ch1});
	my $cnct_rows_ch2 = join (',', @{$rows_affd_by_ch2});

	my $cnct_exnodes_ch1 = join (',', @{$exnodes_affd_by_ch1});
	my $cnct_exnodes_ch2 = join (',', @{$exnodes_affd_by_ch2});

	print STDOUT "\n[ Information on the Child Blocks to be Created: ]\n\n";

	print STDOUT join ("\t", 'Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(external,nodes,affected)', '[MSA,rows,affected]'), "\n\n";

	print STDOUT join ("\t", 'Ch_1', $br_ch1, $up_or_lw_ch1, "($cnct_exnodes_ch1)", "[$cnct_rows_ch1]"), "\n";
	print STDOUT join ("\t", 'Ch_2', $br_ch2, $up_or_lw_ch2, "($cnct_exnodes_ch2)", "[$cnct_rows_ch2]"), "\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >1) { # (Top) # ...}"


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	##my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    ##my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    ##$cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    ##my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    ##$cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_vspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Vertical Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 5) {...}"



   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Vertical Split\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"



=For the predecessor 'cvmerge_gblock_pair'. (MAIN 2)
		### MAIN 2: Displace the gap-blocks either vertically including $gb1 or $gb2 
		### or in the 'ONX' relationships with $gb1 or $gb2 from
		### the regions spanned by $gb1 and $gb2 and that in between them.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###


	## my $if_left; # = 1 if $bl_sbj (=$gb2) will be moved to the left, = 0 if to the right. # ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). (3a) #

	my ($gb_left, $gb_right) = ($if_left) ? ($gb1, $gb2) : ($gb2, $gb1); # CORRECTED on 2019/07/25.
	##my ($gb_left, $gb_right) = ($if_left) ? ($gb2, $gb1) : ($gb1, $gb2);

	my %ignored = ($gb1 => 1, $gb2 => 1); # ADDED in revision on 2019/07/26. #

			## SWAPPED 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right', in REVISION of 'cvmerge_gblock_pair' (on 2019/07/29). ###

	my ($le_left, $re_left) = @{$cp_curr_bds_blks[$gb_left]};
	my $left_limit = $re_left+1;

	my ($gb2incr_coord_right, $totct_nec_shifts_right)
	   = displace_including_and_onx_to_right ($gb_right, $left_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored # Added %ignored to the input arguments on 2019/07/26.
		);
	unless (defined $gb2incr_coord_right) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine, 'displace_including_and_onx_to_right', FAILED on the \$gb_right (=$gb_right) th gap-block and \$left_limit = $left_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_right', Ended on the \$gb_right (=$gb_right) th gap-block and \$left_limit = $left_limit (w/ \$totct_nec_shifts_right = $totct_nec_shifts_right) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);


   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_right} (before post-processing): ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_right} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_right->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"

		# Post-process %{$gb2incr_coord_right}. #
		# (ADDED in REVISION on (2019/07/29).) #

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    my $incr = $gb2incr_coord_right->{$gb};
	    if ($incr > $size_rplcd) {
		$gb2incr_coord_right->{$gb} = $incr - $size_rplcd;
	    } else {
		$gb2incr_coord_right->{$gb} = 0;
	    }
	} # End of "foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {...}"

   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_right} (AFTER post-processing): ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_right} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_right->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"

		# END of "(ADDED in REVISION on (2019/07/29).)" #


   if ($TEST_LEVEL_vspl_gb >5) { # ADDED in revision on 2019/07/29. #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_right' on \$gb_right (=$gb_right) and \$left_limit = $left_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >5) { # ADDED in revision on 2019/07/29. # ...}"


	my ($le_right, $re_right) = @{$cp_curr_bds_blks[$gb_right]};
	my $right_limit = $le_right-1;

	my ($gb2incr_coord_left, $totct_nec_shifts_left)
	   = displace_including_and_onx_to_left ($gb_left, $right_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored # Added %ignored to the input arguments on 2019/07/26.
		);
	unless (defined $gb2incr_coord_left) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine, 'displace_including_and_onx_to_left', FAILED on the \$gb_left (=$gb_left) th gap-block and \$right_limit = $right_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_left', Ended on the \$gb_left (=$gb_left) th gap-block and \$right_limit = $right_limit (w/ \$totct_nec_shifts_left = $totct_nec_shifts_left) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);


   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_left}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_left} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_left}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_left->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"

   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_left' on \$gb_left (=$gb_left) and \$right_limit = $right_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"

			## END of "SWAPPED 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right', in REVISION of 'cvmerge_gblock_pair' (on 2019/07/29)." ###


		# Post-MAIN 2: Check whether ALL the gap-blocks in between $gb1 and $gb2 are DISPLACED. #
		# (Added on 2019/07/25.) (Post M2) #

	my ($le_left_bf_cvm, $re_left_bf_cvm) = @{$cp_curr_bds_blks[$gb_left]};
	my ($le_right_bf_cvm, $re_right_bf_cvm) = @{$cp_curr_bds_blks[$gb_right]};

	unless ($re_left_bf_cvm + 1 == $le_right_bf_cvm) {
	    print STDOUT "[ERROR 'vsplit_gblock'] IMPOSSIBLE!!: Somehow, \$gb_left (=$gb_left) and \$gb_right (=$gb_right) do NOT TOUCH each other, EVEN AFTER performing 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right' (w/ \$re_left_bf_cvm = $re_left_bf_cvm, \$le_right_bf_cvm = $le_right_bf_cvm): $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# End of "(Added on 2019/07/25.) (Post M2)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)"  ###

=cut # END of "For the predecessor 'cvmerge_gblock_pair'. (MAIN 2)"


		# MAIN 1: Create the list of gap-blocks BEFORE the "vertical split". #
		# [ MAIN 3': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". # ] ... for the predecessor, 'cvmerge_gblock_pair'. #
		# (ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).) #


	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_vspl_gb > 3) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 3) {...}"

		# End of "(ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (6))"  ###


=For the predecessor 'cvmerge_gblock_pair'. (MAIN 5)
		### MAIN 5: Perform the "(incomplete) vertical-merge",
		### by moving the complement (residue-block) of $gb_right to the left ###
		### (but the association of index vs gap-blocks remains unchanged).
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (7))  ###

	my $cmpl_rows_affd_by_right = ($gb_right == $gb1) ?
			 $cmpl_rows_affd_by_gb1 : $cmpl_rows_affd_by_gb2 ;

	my @new_curr_set_clms = ();
	my @old_gblks_w_spec_lb_af_cvm = ();
	my @old_gblks_w_spec_rb_af_cvm = ();
	my $indx_new_clm = -1;

	for (my $c=0; $c <$le_new_gb; $c++) {
	    $indx_new_clm++;

	    $new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];
	}

	my $ct_rmvd_null_clms = 0;
	my $indx_mgd_clm = $le_right_bf_cvm -1;
	for (my $k=0; $k <$size_rplcd; $k++) {
	    $indx_new_clm++;
	    $indx_mgd_clm++;

	    my @new_clm = @{$cp_curr_set_clms[$indx_new_clm]};
	    my $mgd_clm = $cp_curr_set_clms[$indx_mgd_clm];
	    foreach my $row (@{$cmpl_rows_affd_by_right}) { $new_clm[$row] = $mgd_clm->[$row]; }
	    $new_curr_set_clms[$indx_new_clm] = \@new_clm;

	    my @mgd_gblks_w_spec_lb = (@{$cp_curr_gblks_w_spec_lb[$indx_new_clm]}, @{$cp_curr_gblks_w_spec_lb[$indx_mgd_clm]});
	    my @mgd_gblks_w_spec_rb = (@{$cp_curr_gblks_w_spec_rb[$indx_new_clm]}, @{$cp_curr_gblks_w_spec_rb[$indx_mgd_clm]});
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = \@mgd_gblks_w_spec_lb;
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = \@mgd_gblks_w_spec_rb;

	    $ct_rmvd_null_clms++;
	}

	for (my $c=$indx_mgd_clm+1; $c < $ct_in_clms; $c++) {
	    $indx_new_clm++;

	    $new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
	    $old_gblks_w_spec_lb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
	    $old_gblks_w_spec_rb_af_cvm[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];
	}

	my $ct_out_clms = @new_curr_set_clms;

	my @bds_old_gblks_af_cvm = ();
	for (my $gb=0; $gb <$ct_in_gbs; $gb++) { $bds_old_gblks_af_cvm[$gb] = []; }
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    foreach my $gb (@{$old_gblks_w_spec_lb_af_cvm[$c]}) { $bds_old_gblks_af_cvm[$gb]->[0] = $c; }
	    foreach my $gb (@{$old_gblks_w_spec_rb_af_cvm[$c]}) { $bds_old_gblks_af_cvm[$gb]->[1] = $c; }
	}


		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (7))"  ###
=cut # END of "For the predecessor 'cvmerge_gblock_pair'. (MAIN 5)"


		### MAIN 2: Create intermediate lists of properties
		###   of gap-blocks (before &) AFTER the "vertical split".
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2) ###

	my @indices_ch_intm = ($ct_in_gbs, $ct_in_gbs+1);

	my @set_info_children = ();
	foreach my $src_info_ch ($info_effch1, $info_effch2) {
	    my  ($br_ch, $up_or_lw_ch, $exnodes_affd_by_ch, $rows_affd_by_ch) = @{$src_info_ch};
	    my $cnct_exnodes_ch = join (',', @{$exnodes_affd_by_ch});
	    my $cnct_rows_ch = join (',', @{$rows_affd_by_ch});

	    push @set_info_children, [$br_ch, $up_or_lw_ch, $lbd_sbj, $rbd_sbj, $cnct_exnodes_ch, $cnct_rows_ch, $exnodes_affd_by_ch, $rows_affd_by_ch];
	}

	my @intm_list_gblocks = @list_old_gblocks;
	@intm_list_gblocks[@indices_ch_intm] = @set_info_children;

	my @intm_rows_affd_by_blks = @cp_rows_affd_by_blks;
	@intm_rows_affd_by_blks[@indices_ch_intm] = ($rows_affd_by_ch1, $rows_affd_by_ch2);

	my @intm_curr_bds_blks = @cp_curr_bds_blks;
	@intm_curr_bds_blks[@indices_ch_intm] = ([$lbd_sbj, $rbd_sbj], [$lbd_sbj, $rbd_sbj]);

	my $ct_intm_gbs = @intm_curr_bds_blks;

   if ($TEST_LEVEL_vspl_gb > 5) {

	print STDOUT "\n";

	print STDOUT '@indices_ch_intm = (', join (', ', @indices_ch_intm), ") .\n\n";


	print STDOUT '( Contents of @intm_list_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_intm_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 5) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2)" ###


		### MAIN 3: Create the correspondence between the old (including intermediate) & new indices of the gap-blocks,
		###    by Re-Sorting the gap-blocks according to their affected MSA rows and their horizontal positions. #
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3) ###

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$intm_list_gblocks[$a]}, @{$intm_list_gblocks[$b]}, 2, 3, 6) } (0 .. $gb_sbj-1, $gb_sbj+1 .. $ct_intm_gbs -1) ;

	my $ct_out_gbs = @old_indices_new_gbs;

	my @new_indices_old_gbs = ();
	$new_indices_old_gbs[$gb_sbj] = undef;

	for (my $ngb=0; $ngb < $ct_out_gbs; $ngb++) {

	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	}

   if ($TEST_LEVEL_vspl_gb >2) {

	print STDOUT "\n[ Relationships between old (including intermediate) and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >2) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3)" ###


		### MAIN 5: Create new lists of boundaries & other properties
		###	of gap-blocks AFTER the "Vertical Split",
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5) ###

	my ($gb_ch1_NEW, $gb_ch2_NEW) = @new_indices_old_gbs[@indices_ch_intm];

	my @new_curr_set_clms = @cp_curr_set_clms;
	my $ct_out_clms = @new_curr_set_clms;

	my @new_curr_bds_blks = @intm_curr_bds_blks[@old_indices_new_gbs];
	my @new_list_gblocks = @intm_list_gblocks[@old_indices_new_gbs];
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c =0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $ngb;
	}

	my @rows_affd_by_new_blks = @intm_rows_affd_by_blks[@old_indices_new_gbs];

	my ($new_inter_block_relations) = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine 'mk_inter_block_relations' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vspl_gb >1);

	my @new_inter_block_relations = @{$new_inter_block_relations};

	my ($new_gblock_sizes) = mk_list_gblock_sizes (@new_curr_bds_blks, 0, 1, @new_inter_block_relations) ;
	unless (defined $new_gblock_sizes) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine 'mk_list_gblock_sizes' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_list_gblock_sizes', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_vspl_gb >1);

	my @new_gblock_sizes = @{$new_gblock_sizes};

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_vspl_gb >1);

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5)" ###


		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 6 & 7 & 8) ###

		### MAIN 6: Create the new coordinate frame.

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER the \"Vertical Split\" of \$gb_sbj (=$gb_sbj): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER the \"Vertical Split\" of \$gb_sbj (=$gb_sbj) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};


   if ($TEST_LEVEL_vspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER \"Vertical-Split\" and BEFORE the UN-NESTING \"Shift\": ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >5) {...}"

   if ($TEST_LEVEL_vspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER \"Vertical-Split\" and BEFORE the UN-NESTING \"Shift\": ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 5) {...}"


		### MAIN 7: Choose the child gap-block to be "Shift"ed.

	my ($gb_shftd, $incr_coord_shftd);

	my ($gb_ch_hg, $gb_ch_lw) = ($gb_ch1_NEW < $gb_ch2_NEW) ?
			($gb_ch1_NEW, $gb_ch2_NEW) : ($gb_ch2_NEW, $gb_ch1_NEW) ;

	my ($curr_coord_ch_hg, $curr_coord_ch_lw) = @new_curr_bl_coords[$gb_ch_hg, $gb_ch_lw];
	my ($lb_coord_ch_hg, $rb_coord_ch_hg) = @{$new_bds_gb_coords[$gb_ch_hg]};
	my ($lb_coord_ch_lw, $rb_coord_ch_lw) = @{$new_bds_gb_coords[$gb_ch_lw]};

	if ($curr_coord_ch_lw < $rb_coord_ch_lw) {
	    $gb_shftd = $gb_ch_lw;
	    $incr_coord_shftd = +1;

	} elsif ($lb_coord_ch_lw < $curr_coord_ch_lw) {
	    $gb_shftd = $gb_ch_lw;
	    $incr_coord_shftd = -1;
	
	} elsif ($curr_coord_ch_hg < $rb_coord_ch_hg) {
	    $gb_shftd = $gb_ch_hg;
	    $incr_coord_shftd = +1;

	} elsif ($lb_coord_ch_hg < $curr_coord_ch_hg) {
	    $gb_shftd = $gb_ch_hg;
	    $incr_coord_shftd = -1;
	} 

	unless (defined $gb_shftd) {
	    print STDOUT "[ERROR 'vsplit_gblock'] It turned out that the TWO child gap-blocks ($gb_ch1_NEW, $gb_ch2_NEW) canNOT get NON-NESTED!! ... I will stop here ...: $! !\n\n";
	    return (0, undef, undef, undef);
	}

   if ($TEST_LEVEL_vspl_gb >2) { # (M7) #

	print STDOUT "\n[ The Child Gap-Block to be \"Shift\"ed: ]\n\n";

	print STDOUT '($gb_shftd, $incr_coord_shftd) = ', "($gb_shftd, $incr_coord_shftd) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >2) { # (M7) # ...}"


		### MAIN 8: Perform the "Shift" to UN-NEST the children. ###

	my ($flag_shft, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_gb_coords_wd_frm, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb,
	    $gb_shftd, $incr_coord_shftd, @rows_affd_by_new_blks, $cnct_null_clm,
	    @new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ((defined $flag_shft) and ($flag_shft == 0)) {
	    my $flag_out = (defined $flag_shft) ? $flag_shft : 'undef';
	    print STDOUT "[ERROR 'vsplit_gblock'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the \$gb_shftd (=$gb_shftd) th gap-block (w/ \$flag_shft = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_shftd (=$gb_shftd) th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);

   if ($TEST_LEVEL_vspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the UN-NESTING \"Shift\": ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >5) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 6 & 7 & 8)" ###


		### MAIN 9: UNDO the horizontal overlaps among the vertically identical & complementary gap-blocks.
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9) ###

			# The 1st Round (w.r.t. $gb_ch1_NEW) #

	my ($fl_undo1, $gb2incr_bl_coord_undo1) 
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo1) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch1_NEW (=$gb_ch1_NEW)) on the cofiguration of gap-blocks after \"vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_ch1_NEW (=$gb_ch1_NEW and \$gb_ch2_NEW (=$gb_ch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch1_NEW (=$gb_ch1_NEW)), Ended (w/ \$fl_undo1 = $fl_undo1) on the configuration of gap-blocks after the \"vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_ch1_NEW (=$gb_ch1_NEW) and \$gb_ch2_NEW (=$gb_ch2_NEW) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);


   if ($TEST_LEVEL_vspl_gb > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch1_NEW (=$gb_ch1_NEW)): ]\n\n";

	print STDOUT "\$fl_undo1 = $fl_undo1 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo1} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo1}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo1->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 3) {...}"

	if ($fl_undo1 < 0) {
	    print STDOUT "[CAUTION 'vsplit_gblock'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch1_NEW (=$gb_ch1_NEW)), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}


			# The 2nd Round (w.r.t. $gb_ch2_NEW) #

	my ($fl_undo2, $gb2incr_bl_coord_undo2) 
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW, # A BUG Fixed on 2019/08/23. #
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch2_NEW,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo2) {
	    print STDOUT "[ERROR 'vsplit_gblock'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch2_NEW (=$gb_ch2_NEW)) on the cofiguration of gap-blocks after \"vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_ch1_NEW (=$gb_ch1_NEW and \$gb_ch2_NEW (=$gb_ch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch2_NEW (=$gb_ch2_NEW)), Ended (w/ \$fl_undo2 = $fl_undo2) on the configuration of gap-blocks after the \"vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_ch1_NEW (=$gb_ch1_NEW) and \$gb_ch2_NEW (=$gb_ch2_NEW) ...\n\n") if ($TEST_LEVEL_vspl_gb >1);


   if ($TEST_LEVEL_vspl_gb > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch2_NEW (=$gb_ch2_NEW)): ]\n\n";

	print STDOUT "\$fl_undo2 = $fl_undo2 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo2} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo2}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo2->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cspl_gb > 3) {...}"

	if ($fl_undo2 < 0) {
	    print STDOUT "[CAUTION 'vsplit_gblock'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch2_NEW (=$gb_ch2_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

			# Combine the outputs of the TWO Rounds into an output of this subroutine. #
			# (There should be NO Problem with it, b/c the two sets of keys SHOULD NOT overlap.)

	my %gb2incr_bl_coord_undo = (%{$gb2incr_bl_coord_undo1}, %{$gb2incr_bl_coord_undo2});


		### End of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9)" ###


			### Test-Output the Final Results. ###

   if ($TEST_LEVEL_vspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Vertical-Split\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb >3) {...}"


   if ($TEST_LEVEL_vspl_gb > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Vertical-Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 3) {...}"


		### MAIN 10: Finally Update  @new_list_gblocks. #
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    @{$new_list_gblocks[$gb]}[2,3] = @{$new_curr_bds_blks[$gb]};
		 # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}

   if ($TEST_LEVEL_vspl_gb > 3) {

	print STDOUT '[ FINAL Contents of @new_list_gblocks: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 3) {...}"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))"  ###

		### MAIN 11: Create @new_list_excluded_bin_hrels.
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (MAIN 11)) ###

	my @new_list_excluded_bin_hrels = ([$gb_ch1_NEW, $gb_ch2_NEW, '=']);

   if ($TEST_LEVEL_vspl_gb > 3) {

	print STDOUT '[ Contents of @new_list_excluded_bin_hrels: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";
	for (my $i=0; $i < @new_list_excluded_bin_hrels; $i++) {
	    print STDOUT join ("\t", $i, @{$new_list_excluded_bin_hrels[$i]}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vspl_gb > 3) {...}"

		### End of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (MAIN 11))" ###


    (print STDOUT "\n< Subroutine 'vsplit_gblock' Ended (w/ \$fl_end = 1, (\$gb_ch1_NEW, \$gb_ch2_NEW) = ($gb_ch1_NEW, $gb_ch2_NEW), \$gb_shftd = $gb_shftd, \$incr_coord_shftd = $incr_coord_shftd). >\n\n") if ($TEST_LEVEL_vspl_gb > 0);

		# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/23). (OUT) #
	## [return (1, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs, ] ... for the predecessor, 'cvmerge_gblock_pair'. #

	return (1,  $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo,
		# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/23). (OUT)" #

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks, 
	\@new_list_excluded_bin_hrels,  # ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (OUT) #
	);

}	# End of the subroutine, 'vsplit_gblock'. #


# cvsplit_gblock:
#	This subroutine "vertically split"s the COMPLEMENTARY RESIDUE-BLOCK of the subject gap-block into a pair of RESIDUE-BLOCKs that are the "effective children" of the subject's complementary residue-block.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' & 'vsplit_gblock', both are above.
#	[CREATION STARTED on 2019/08/25, ENDED on 2019/08/26.]
#
#	[NOTE] Before the "vertical-split", it DISPLACES gap-blocks that are "vertically including" the subject gap-block to the left of the subject.
#	Then, it attempts to UNDO the displacements after the "vertical-split".
#	HOWEVER, those gap-blocks that are vertically EQUIVALENT to the subject will remain displaced.
#
#	[REVISED on 2019/08/28 (1)]:
#	Now, the subroutine first DISPLACES "vertically including" gap-blocks TO THE RIGHT,
#	instead of to the left, as before the revision;
#	this is expected to "UNDO" some moves of the flanking residues more securely.
#	[=> This REVISION was UNDONE immediately!!]
#
#	[REVISED on 2019/08/28 (2)]:
#	Now, the subroutine does NOT ATTEMPT to UNDO the DISPLACEMENT
#	if the displaced gap-block is an effective parent of the subject gap-block. 
#	(This way, superfluous "shift"s may be suppressed when undoing the horizontal overlaps.)
#
# vsplit_gblock:
#	This subroutine "vertically split"s the subject gap-block into a pair of gap-blocks that are the "effective children" of the subject,
#	then, it "shift"s one of the resulting child blocks to that they will NOT be nested.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' above.
#	[CREATION STARTED on 2019/08/22, ENDED on 2019/08/23.]
#	
# cvmerge_gblock_pair:
#	This subroutine actually performs the "(incomplete) vertical-merge" of a specified pair of
#	complementary-sibling gap-blocks.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_gblock_pair'.
#	[CREATION STARTED on 2019/07/23, ENDED on 2019/07/27.]
#	[REVISED on 2019/07/29]:
#	SWAPPED the subroutines 'displace_including_and_onx_to_left' and 'displace_including_and_onx_to_right' in order,
#	and post-processed the output of 'displace_including_and_onx_to_right',
#	in order to make the final result as close as possible to the desired result.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 1)" #
#	), where ] ... for the predecessor, 'cvmerge_gblock_pair'.
#
# [ INPUTS: ($gb_sbj, \@info_effch1, \@info_effch2, # MODIFIED in creation of 'vsplit_gblock'. (HD1a) # ] ... For the predecessor, 'vsplit_gblock'.
#
# [ INPUTS: ($gb_sbj, \@info_ceffch1, \@info_ceffch2, # MODIFIED in creation of 'cvsplit_gblock'. (HD1a) # ] ... OBSOLETE as of 2019/08/26.
#
# INPUTS: ($gb_sbj, \@info_ceffch1, \@info_ceffch2, $char_abs, # ADDED $char_abs on 2019/08/26.
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#	), where
#
#		# MODIFIED in creation of 'vsplit_gblock'. (HD1b) #
#	$gb_sbj is the index of the gap-block to be "vertically split";
#
#	[[
#	@info_effch1 and @info_effch2 provide information on the TWO effective children of $gb_sbj;
#	more specifically,
#	@info_effch1/2 = ($br_ch, $up_or_lw_ch, \@exnodes_affd_by_ch, \@rows_affd_by_ch), where
#	the 1st/2nd child is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@exnodes_affd_by_ch stores the indices of external nodes affected by the child,
#	and @rows_affd_by_ch stores the indices of the MSA rows affected by the child.
#	[Typically, \@info_effch1 and \@info_effch2 are parts of the output of 'list_vsplit_candidates' above];
#	]] ... for the predecessor, 'vsplit_gblock'.
#
#		# ENd of "MODIFIED in creation of 'vsplit_gblock'. (HD1b)" #
#
#		# MODIFIED in creation of 'cvsplit_gblock'. (HD1b) #
#	@info_ceffch1/2 = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)
#	provides information on the 1st/2nd "complementary-effective child" of $gb_sbj, where
#	the "complementary-child" residue-block's complementary GAP-block is on the $up_or_lw_cch side of the branch with ID $br_cch ,
#	@cexnodes_affd_by_cch stores the indices of external nodes affected by the complementary GAP-block of the "complementary-child" residue-block,
#	and @crows_affd_by_cch stores the indices of the MSA rows affected by the complementary GAP-block of the "complementary-child" residue-block.
#	[Typically, \@info_ceffch1 and \@info_ceffch2 are parts of the output of 'list_cvsplit_candidates' above];
#		# END of "MODIFIED in creation of 'cvsplit_gblock'. (HD1b)" #
#
#	$char_abs is the character resresenting an "absence" state; # (Added on 2019/08/26). #
#
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#			# (Added on 2019/07/26.) #
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#			# End of "(Added on 2019/07/26.)" #
#
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 2)" #
#	
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs, ] ... for the predecessor, 'cvmerge_gblock_pair'.
#
# [ OUTPUTS: ($flag_end, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2a) # ] ... for the predecessor, 'vsplit_gblock'.
#
# OUTPUTS: ($flag_end, $gb_cch1_NEW, $gb_cch2_NEW, \%gb2incr_coord_1way_dsplc, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2a) #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	
#	## [ @new_list_excluded_bin_hrels # ADDED in creation of 'vsplit_gblock'. (HD1a) # ] ... for the predecessor, 'vsplit_gblock'. 
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; # IMPORTED from the header of 'csplit_gblock'. #
#
#
#			# MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2b) #
#	$gb_cch1_NEW and $gb_cch2_NEW are the NEW indices of the COMPLEMENTARY GAP-BLOCKS of the 1st and 2nd child, respectively, of the RESIDUE-BLOCK COMPLEMENTARY to $gb_sbj;
#			# End of "MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2b)" #
#
#	[[
#			# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2b) #
#	$gb_ch1_NEW and $gb_ch2_NEW are the NEW indices of the 1st and 2nd child, respectively, of $gb_sbj;
#	$gb_shftd is the index (in the NEW gap-block set) of the child gap-block that was "shift"ed,
#	to make the two children non-nested;
#	$incr_coord_shftd (= +1 or -1) is the increment of the coordinate of $gb_shftd via the aforementioned "shift";
#			# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2b)" #
#	]] ... for the predecessor, 'vsplit_gblock'.
#
#			# MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2c) #
#	%gb2incr_coord_1way_dsplc = ($indx_gb => {increment of the block coordinate resulting from the DISPLACEMENT prior to the "complementary vertical split"}, ...) for the gap-blocks whose displacement canNOT be undone; (Such gap-blocks MUST BE vertically identical to $gb_sbj); $indx_gb SHOULD BE the index in the NEW gap-block set;
#			# End of "MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/25). (HD2c)" #
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vertical-split";
#	More precisely, the gap-block that had been the $gb th before the "vertical-split"
#	has been assigned $new_indices_old_gbs[$gb] after the "vertical-split"
#	(, and $new_indices_old_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vertical-split";
#	More precisely, the gap-block that is the $gb th after the "vertical-split"
#	had been assigned $old_indices_new_gbs[$gb] before the "vertical-split";
#
#			# MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c) #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2c)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "vertical-split";
#
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 5) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#		# End of "ADDED in creation of 'cvmerge_gblock_pair'. (HD 5)" #
#
#	[[
#		# ADDED in creation of 'vsplit_gblock'. (HD1b) #
#	@new_list_excluded_bin_hrels = ([$gb_ch1_NEW, $gb_ch2_NEW, '='])
#	is the NEW list of PROHIBITED binary horizontal relationships,
#	which conforms to the format prescribed in the header of 'nonrec_mlt_shifts_gapblocks' above;
#		# End of "ADDED in creation of 'vsplit_gblock'. (HD1b)" #
#	]] ... for the predecessor, 'vsplit_gblock'. (This is UNNECESSARY for 'cvsplit_gblock', because the resulting two gap-blocks are in the 'ONCS' relationship, so that they NEVER OVERLAP since their creation.)
#
#sub cvmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb1, $gb2, 
#	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
#	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
#	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,
#		# Added in creation of 'cvmerge_gblock_pair'. (IN) #
#	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
#		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
#	$node2pa, $node2ch, $topnode_id, $br2eq # (Added this line on 2019/07/26.)
#		# End of "Added in creation of 'cvmerge_gblock_pair'. (IN)" #
#	) = @_; 
#
# [ sub vsplit_gblock ($\@\@\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb_sbj, $info_effch1, $info_effch2,  ] ... for the predecessor, 'vsplit_gblock'.
#
# [ sub cvsplit_gblock ($\@\@\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) { ] ... OBSOLETE as of 2019/08/26.
#
sub cvsplit_gblock ($\@\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

	my ($gb_sbj, $info_ceffch1, $info_ceffch2, $char_abs,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq
	) = @_;  # Added $char_abs on 2019/08/26.

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	unless (($ct_in_clms>0) and ($ct_in_gbs>0)) { # Added on 2019/08/26.
	    print STDOUT "[ERROR 'cvsplit_gblock'] ILLEGAL INPUT: (\$ct_in_clms, \$ct_in_gbs) = ($ct_in_clms, $ct_in_gbs): $! !\n";
	    return (0, undef, undef, undef);
	}

	my $ct_rows = @{$curr_set_clms->[0]}; # Added in creation of 'cvsplit_gblock' (on 2019/08/26). #

	my $gb_coord_sbj = $curr_bl_coords->[$gb_sbj];
	my ($lbd_sbj, $rbd_sbj) = @{$curr_bds_blks->[$gb_sbj]};

	my $size_sbj = $gblock_sizes->[$gb_sbj]; # Added in creation of 'cvsplit_gblock'. #

	my ($br_sbj, $up_or_lw_sbj) = @{$list_gblocks0->[$gb_sbj]}[$indx_br, $indx_u_or_d];
	my $cnct_rows_sbj = join (',', @{$rows_affd_by_blks->[$gb_sbj]});

    (print STDOUT "< Subroutine 'cvsplit_gblock' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; The \"Subject\" Gap-Block to be \"complementary-vertically split\": \$gb_sbj = $gb_sbj, {coordinate} = $gb_coord_sbj, [lbd, rbd]_sbj = [$lbd_sbj, $rbd_sbj], {size} = $size_sbj, \$br_sbj = $br_sbj, \$up_or_lw_sbj = $up_or_lw_sbj, [rows,affected]_sbj=[$cnct_rows_sbj]) >\n\n") if ($TEST_LEVEL_cvspl_gb > 0);


	my ($br_cch1, $up_or_lw_cch1, $cexnodes_affd_by_cch1, $crows_affd_by_cch1) = @{$info_ceffch1};
	my ($br_cch2, $up_or_lw_cch2, $cexnodes_affd_by_cch2, $crows_affd_by_cch2) = @{$info_ceffch2};

	# @info_ceffch1/2 = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)

   if ($TEST_LEVEL_cvspl_gb >1) { # (Top) #

	my $cnct_crows_cch1 = join (',', @{$crows_affd_by_cch1});
	my $cnct_crows_cch2 = join (',', @{$crows_affd_by_cch2});

	my $cnct_cexnodes_cch1 = join (',', @{$cexnodes_affd_by_cch1});
	my $cnct_cexnodes_cch2 = join (',', @{$cexnodes_affd_by_cch2});

	print STDOUT "\n[ Information on the Child Blocks to be Created: ]\n\n";

	print STDOUT join ("\t", 'Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(external,nodes,affected)', '[MSA,rows,affected]'), "\n\n";

	print STDOUT join ("\t", 'Ch_1', $br_cch1, $up_or_lw_cch1, "($cnct_cexnodes_cch1)", "[$cnct_crows_cch1]"), "\n";
	print STDOUT join ("\t", 'Ch_2', $br_cch2, $up_or_lw_cch2, "($cnct_cexnodes_cch2)", "[$cnct_crows_cch2]"), "\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >1) { # (Top) # ...}"


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Compelemntary-Vertical Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"



   if ($TEST_LEVEL_cvspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Complementary-Vertical Split\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"



		### MAIN 2: Displace the gap-blocks either vertically including $gb_sbj
		### or in the 'ONX' relationships with $gb_sbj from
		### the region spanned by $gb_sbj to its left.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###
		### (=> IMPORTED (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 2) ###


	my %ignored = (); # MOVED from above and MODIFIED in creation of 'cvsplit_gblock' (on 2019/0826). #

=UNDID the REVISION (on 2019/08/28). (1)
		# MODIFIED in REVISION (on 2019/08/28). (1) #

	my $left_limit = $lbd_sbj; # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (1) #

	my ($gb2incr_coord_right_dsplc, $totct_nec_shifts_right_dsplc)
	   = displace_including_and_onx_to_right ($gb_sbj, $left_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_right_dsplc) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine, 'displace_including_and_onx_to_right', FAILED on the \$gb_sbj (=$gb_sbj) th gap-block and \$left_limit = $left_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_right', Ended on the \$gb_sbj (=$gb_sbj) th gap-block and \$left_limit = $left_limit (w/ \$totct_nec_shifts_right_dsplc = $totct_nec_shifts_right_dsplc) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);


   if ($TEST_LEVEL_cvspl_gb >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_right_dsplc}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_right_dsplc} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right_dsplc}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_right_dsplc->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"
=cut # END of "UNDID the REVISION (on 2019/08/28). (1)"

	my $right_limit = $rbd_sbj; # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (1) #
	##my ($le_right, $re_right) = @{$cp_curr_bds_blks[$gb_right]};
	##my $right_limit = $le_right-1;

	my ($gb2incr_coord_left_dsplc, $totct_nec_shifts_left_dsplc)
	   = displace_including_and_onx_to_left ($gb_sbj, $right_limit,  # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (2) #
	   ##= displace_including_and_onx_to_left ($gb_left, $right_limit, 
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
		);
	unless (defined $gb2incr_coord_left_dsplc) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine, 'displace_including_and_onx_to_left', FAILED on the \$gb_sbj (=$gb_sbj) th gap-block and \$right_limit = $right_limit: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'displace_including_and_onx_to_left', Ended on the \$gb_sbj (=$gb_sbj) th gap-block and \$right_limit = $right_limit (w/ \$totct_nec_shifts_left_dsplc = $totct_nec_shifts_left_dsplc) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);


   if ($TEST_LEVEL_cvspl_gb >3) {

	print STDOUT "\n", '[ Contents of %{$gb2incr_coord_left_dsplc}: ]', "\n\n";
	print STDOUT '%{$gb2incr_coord_left_dsplc} = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_left_dsplc}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_left_dsplc->{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"

		# END of "MODIFIED in REVISION (on 2019/08/28). (1)" # => The REVISION was UNDONE immediately!!


			# Sort the hash  %{$gb2incr_coord_left_dsplc} into %gb2incr_coord_1way_dsplc and %gb2incr_coord_round_trip_dsplc.
			# (ADDED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2)) #

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my %gb2incr_coord_1way_dsplc;
	my %gb2incr_coord_round_trip_dsplc;


=UNDID the REVISION (on 2019/08/28). (2)
			# MODIFIED in REVISION (on 2019/08/28). (2) #
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_right_dsplc}) {
	    my $incr_coord = $gb2incr_coord_right_dsplc->{$gb};
=cut # END of "UNDID the REVISION (on 2019/08/28). (2)"
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_left_dsplc}) {
	    my $incr_coord = $gb2incr_coord_left_dsplc->{$gb};

			# END of "MODIFIED in REVISION (on 2019/08/28). (2)" # => The REVISION was UNDONE immediately!!

	    my $rel = $rels_w_sbj->[$gb];
	    if ($rel eq '<')  { # MODIFIED in REVISION (on 2019/08/28). #
	    ##if (($rel eq '<') or ($rel eq '<(pa)')) {
		$gb2incr_coord_round_trip_dsplc{$gb} = $incr_coord;

	    } else { # if (($rel eq '=') or ($rel eq 'ONN') or ($rel eq 'ONCS')) # or ($rel eq '<(pa)') #
		$gb2incr_coord_1way_dsplc{$gb} = $incr_coord;
	    }
	}

   if ($TEST_LEVEL_cvspl_gb >3) {

	print STDOUT "\n", '[ Contents of NEWLY CREATED %gb2incr_coord_1way_dsplc & %gb2incr_coord_round_trip_dsplc: ]', "\n\n";

	print STDOUT '%gb2incr_coord_1way_dsplc = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_1way_dsplc) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_1way_dsplc{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n";

	print STDOUT '%gb2incr_coord_round_trip_dsplc = (', "\n";
	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_round_trip_dsplc) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_round_trip_dsplc{$gb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"
			# End of "(ADDED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2))" #

   if ($TEST_LEVEL_cvspl_gb >3) {

	##print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_right' on \$gb_sbj (=$gb_sbj) and \$left_limit = $left_limit ]\n\n"; # MODIFIED in REVISION (on 2019/08/28). (3) # => UNDONE immediately!!
	print STDOUT "\n[ Gap-Block Configurations AFTER the 'displace_including_and_onx_to_left' on \$gb_sbj (=$gb_sbj) and \$right_limit = $right_limit ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"


	    # Update the boundaries of $gb_sbj. # (Added in creation of 'cvsplit_gblock' (on 2019/08/26).) #
	($lbd_sbj, $rbd_sbj) = @{$cp_curr_bds_blks[$gb_sbj]};


   if ($TEST_LEVEL_cvspl_gb > 3) {
	print STDOUT "\nUpdated Boundaries of \$gb_sbj (=$gb_sbj): (\$lbd_sbj, \$rbd_sbj) = ($lbd_sbj, $rbd_sbj) .\n\n\n";
   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"


		# MAIN 2post: Check whether the displacement produced the desired result. # 
		# (MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2post)) #

	my $curr_span_sbj = $rbd_sbj - $lbd_sbj +1;
	unless ($curr_span_sbj == $size_sbj) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] IMPOSSIBLE!!: Somehow, 'displace_including_and_onx_to_left' did NOT displace ENOUGH gap-blocks to EQUATE the SIZE, \$size_sbj (=$size_sbj), of the \$gb_sbj (=$gb_sbj) th gap-block to its SPAN, \$rbd_sbj - \$lbd_sbj +1 (=$curr_span_sbj) : $! !!\n\n";
	    return (0, undef, undef, undef);
	}

		# End of "(MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 2post))" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/24-25.)  (5)  ###
		### (=> IMPORTED (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 2)" ###


		# MAIN 3: Create the list of gap-blocks BEFORE the "complementary vertical split". #
		# [ MAIN 1: Create the list of gap-blocks BEFORE the "vertical split". # ] ... for the predecessor, 'vsplit_gblock'. #
		# [ MAIN 3': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". # ] ... for the predecessor, 'cvmerge_gblock_pair'. #
		# (ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).) #


	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_cvspl_gb > 3) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"

		# End of "(ADDED in creation of 'cvmerge_gblock_pair' (on 2019/07/24). => MOVED from "(Preliminary 1)" (on 2019/07/25).)" #

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/25. (6))"  ###



		### MAIN 5: Create intermediate lists of properties
		###   of gap-blocks (before &) AFTER the "complementary vertical split".
		### [ MAIN 2: Create intermediate lists of properties
		###   of gap-blocks (before &) AFTER the "vertical split". ] ... for the predecessor 'vsplit_gblock'.
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2) ###
		### (=> Carried over (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 5) ###

	my @indices_cch_intm = ($ct_in_gbs, $ct_in_gbs+1);

	my @set_info_cchildren = ();
	foreach my $src_info_cch ($info_ceffch1, $info_ceffch2) {
	    my  ($br_cch, $up_or_lw_cch, $cexnodes_affd_by_cch, $crows_affd_by_cch) = @{$src_info_cch};
	    my $cnct_cexnodes_cch = join (',', @{$cexnodes_affd_by_cch});
	    my $cnct_crows_cch = join (',', @{$crows_affd_by_cch});

	    push @set_info_cchildren, [$br_cch, $up_or_lw_cch, $lbd_sbj, $rbd_sbj, $cnct_cexnodes_cch, $cnct_crows_cch, $cexnodes_affd_by_cch, $crows_affd_by_cch];
	}

	my @intm_list_gblocks = @list_old_gblocks;
	@intm_list_gblocks[@indices_cch_intm] = @set_info_cchildren;

	my @intm_rows_affd_by_blks = @cp_rows_affd_by_blks;
	@intm_rows_affd_by_blks[@indices_cch_intm] = ($crows_affd_by_cch1, $crows_affd_by_cch2);

	my $ct_intm_gbs = @intm_list_gblocks; # Modified in creation of 'cvsplit_gblock' (MAIN 5)


   if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n";

	print STDOUT '@indices_cch_intm = (', join (', ', @indices_cch_intm), ") .\n\n";


	print STDOUT '( Contents of @intm_list_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_intm_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 2) ###
		### (=> Carried over (w/ modifications) to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 5)" ###

		### MAIN 5post: Create inter-block relations between the gap-blocks in the intermediate set. #
		### (ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 5post) ###

	my ($intm_inter_block_relations) = mk_inter_block_relations (@intm_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $intm_inter_block_relations) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine 'mk_inter_block_relations' FAILED on the intermediate set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the intermediate set of  gap-blocks ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

	my ($intm_rels_w_cch1, $intm_rels_w_cch2) = @{$intm_inter_block_relations}[@indices_cch_intm];
	my @old_rels_w_cch1 = @{$intm_rels_w_cch1}[0 .. $ct_in_gbs-1];
	my @old_rels_w_cch2 = @{$intm_rels_w_cch2}[0 .. $ct_in_gbs-1];

   if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n[ Relations of the old gap-blocks with complementary-effective children 1 & 2: ]\n\n";

	print STDOUT '@old_rels_w_cch1 = (', join (', ', @old_rels_w_cch1), ") ;\n";
	print STDOUT '@old_rels_w_cch2 = (', join (', ', @old_rels_w_cch2), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"

		### End of "(ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 5post)" ###


		### [ MAIN 3: Create the correspondence between the old (including intermediate) & new indices of the gap-blocks, ] ... for the predecessor, 'vsplit_gblock'.
		### MAIN 6: Create the correspondence between the old (including intermediate) & new indices of the gap-blocks,
		###    by Re-Sorting the gap-blocks according to their affected MSA rows and their horizontal positions. #
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3) ###
		### (=> Carried over to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 6) ###

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$intm_list_gblocks[$a]}, @{$intm_list_gblocks[$b]}, 2, 3, 6) } (0 .. $gb_sbj-1, $gb_sbj+1 .. $ct_intm_gbs -1) ;

	my $ct_out_gbs = @old_indices_new_gbs;

	my @new_indices_old_gbs = ();
	$new_indices_old_gbs[$gb_sbj] = undef;

	for (my $ngb=0; $ngb < $ct_out_gbs; $ngb++) {

	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	}

   if ($TEST_LEVEL_cvspl_gb >2) {

	print STDOUT "\n[ Relationships between old (including intermediate) and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >2) {...}"

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 3) ###
		### (=> Carried over to 'cvsplit_gblock' in its creation (on 2019/08/26).) (MAIN 6)" ###


		### MAIN 7: Actually Perform the "Complementary-Vertical Split",
		###	    while keeping the OLD gap-block indices.
		### (ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 7) ###

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

	my ($gb_cch1_OLD, $gb_cch2_OLD) = @indices_cch_intm;

	my ($crows_affd_by_both, $ex_affd_by_cch1, $ex_affd_by_cch2) = diff_nmarray (@{$crows_affd_by_cch1}, @{$crows_affd_by_cch2});
	unless (defined $crows_affd_by_both) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine 'diff_nmarray' FAILED on ", '@{$crows_affd_by_cch1} & @{$crows_affd_by_cch2}', ": $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my @set_clms_in_cch1 = my @set_clms_in_cch2 = ();
	my @gblks_w_lb_in_cch1 = my @gblks_w_rb_in_cch1 = ();
	my @gblks_w_lb_in_cch2 = my @gblks_w_rb_in_cch2 = ();

	for (my $c = $lbd_sbj; $c <= $rbd_sbj; $c++) { # Main for-loop (over columns within the boundaries of $gb_sbj). #

	    my @clm_in_cch1 = @{$cp_curr_set_clms[$c]};
	    my @clm_in_cch2 = @null_clm;

	    foreach my $row (@{$ex_affd_by_cch1}) {
		my $tmp = $clm_in_cch1[$row];
		$clm_in_cch1[$row] = $clm_in_cch2[$row];
		$clm_in_cch2[$row] = $tmp;
	    } # End of "foreach my $row (...) {...}"

	    push @set_clms_in_cch1, \@clm_in_cch1;
	    push @set_clms_in_cch2, \@clm_in_cch2;

	    my @gblks_w_lb_eq_c_in_cch1 = my @gblks_w_lb_eq_c_in_cch2 = ();
	    my @gblks_w_rb_eq_c_in_cch1 = my @gblks_w_rb_eq_c_in_cch2 = ();

	    foreach my $trio (
		[$cp_curr_gblks_w_spec_lb[$c], \@gblks_w_lb_eq_c_in_cch1, \@gblks_w_lb_eq_c_in_cch2],
		[$cp_curr_gblks_w_spec_rb[$c], \@gblks_w_rb_eq_c_in_cch1, \@gblks_w_rb_eq_c_in_cch2],
		) {
		my ($src_gblks, $bds_into_cch1, $bds_into_cch2) = @{$trio};
		foreach my $gb (@{$src_gblks}) {
		    if ($gb == $gb_sbj) {
			push @{$bds_into_cch1}, $gb_cch1_OLD;
			push @{$bds_into_cch2}, $gb_cch2_OLD;
			next;
		    }
		    my $rel1 = $old_rels_w_cch1[$gb];
		    my $rel2 = $old_rels_w_cch2[$gb];
		    if (($rel1 eq '>') or ($rel1 eq '>(ch)')) {
			push @{$bds_into_cch2}, $gb;

		    } elsif (($rel2 eq '>') or ($rel2 eq '>(ch)')) {
			push @{$bds_into_cch1}, $gb;
		    } else {
			print STDOUT "[ERROR 'cvsplit_gblock'] The \$gb (=$gb) th gap-block (in the old set), whose boundary (=$c) is within the boundaries of \$gb_sbj (=$gb_sbj), is NOT INCLUDED in EITHER \$gb_cch1_OLD (=$gb_cch1_OLD) OR \$gb_cch2_OLD (=$gb_cch2_OLD): $! !\n\n";
			return (0, undef, undef, undef);
		    }
		}
	    } # End of "foreach my $trio (...) {...}"

	    push @gblks_w_lb_in_cch1, \@gblks_w_lb_eq_c_in_cch1;
	    push @gblks_w_lb_in_cch2, \@gblks_w_lb_eq_c_in_cch2;
	    push @gblks_w_rb_in_cch1, \@gblks_w_rb_eq_c_in_cch1;
	    push @gblks_w_rb_in_cch2, \@gblks_w_rb_eq_c_in_cch2;

	} # End of the Main for-loop (over columns within the boundaries of $gb_sbj). #

	# my @set_clms_in_cch1 = my @set_clms_in_cch2 = ();
	# my @gblks_w_lb_in_cch1 = my @gblks_w_rb_in_cch1 = ();
	# my @gblks_w_lb_in_cch2 = my @gblks_w_rb_in_cch2 = ();

	my @new_curr_set_clms = (
		@cp_curr_set_clms[0 .. $lbd_sbj-1],
		@set_clms_in_cch1,
		@set_clms_in_cch2,
		@cp_curr_set_clms[$rbd_sbj+1 .. $ct_in_clms-1]
		);
	my @intm_gblks_w_spec_lb_af_cvs = (
		@cp_curr_gblks_w_spec_lb[0 .. $lbd_sbj-1],
		@gblks_w_lb_in_cch1,
		@gblks_w_lb_in_cch2,
		@cp_curr_gblks_w_spec_lb[$rbd_sbj+1 .. $ct_in_clms-1]
		);
	my @intm_gblks_w_spec_rb_af_cvs = (
		@cp_curr_gblks_w_spec_rb[0 .. $lbd_sbj-1],
		@gblks_w_rb_in_cch1,
		@gblks_w_rb_in_cch2,
		@cp_curr_gblks_w_spec_rb[$rbd_sbj+1 .. $ct_in_clms-1]
		);

	my $ct_out_clms = @new_curr_set_clms;

	my @intm_curr_bds_blks = ();
	for (my $gb=0; $gb <$ct_intm_gbs; $gb++) { $intm_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$ct_out_clms; $c++) {
	    foreach my $gb (@{$intm_gblks_w_spec_lb_af_cvs[$c]}) {
		$intm_curr_bds_blks[$gb]->[0] = $c;
	    }
	    foreach my $gb (@{$intm_gblks_w_spec_rb_af_cvs[$c]}) {
		$intm_curr_bds_blks[$gb]->[1] = $c;
	    }
	}

		### END of "(ADDED in creation of 'cvsplit_gblock' (on 2019/08/26).) (MAIN 7)" ###



		### MAIN 8: Create the New Data-Sets, by using the new indices of the gap-blocks.
		### [ MAIN 5: Create new lists of boundaries & other properties
		###	of gap-blocks AFTER the "Vertical Split",  ] ... for the 
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 8) ###

	my ($gb_cch1_NEW, $gb_cch2_NEW) = @new_indices_old_gbs[@indices_cch_intm]; # MODIFIED in creation of 'cvsplit_gblock'. (MAIN 8a) #
	##my ($gb_ch1_NEW, $gb_ch2_NEW) = @new_indices_old_gbs[@indices_ch_intm];

	##my @new_curr_set_clms = @cp_curr_set_clms;
	##my $ct_out_clms = @new_curr_set_clms;

	my @new_curr_bds_blks = @intm_curr_bds_blks[@old_indices_new_gbs];

			# Modified in creation of 'cvsplit_gblock'. (MAIN 8b) #
	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs[$ngb];
	    my @info_ngb = @{$intm_list_gblocks[$ogb]};
	    @info_ngb[2,3] = @{$new_curr_bds_blks[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_ngb;
		# element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}
	##my @new_list_gblocks = @intm_list_gblocks[@old_indices_new_gbs];
			# End of "Modified in creation of 'cvsplit_gblock'. (MAIN 8b)" #

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c =0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $ngb;
	}

	my @rows_affd_by_new_blks = @intm_rows_affd_by_blks[@old_indices_new_gbs];

	my ($new_inter_block_relations) = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine 'mk_inter_block_relations' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_inter_block_relations', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

	my @new_inter_block_relations = @{$new_inter_block_relations};

	my ($new_gblock_sizes) = mk_list_gblock_sizes (@new_curr_bds_blks, 0, 1, @new_inter_block_relations) ;
	unless (defined $new_gblock_sizes) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine 'mk_list_gblock_sizes' FAILED on the new set of gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_list_gblock_sizes', Ended on the set of new gap-blocks ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

	my @new_gblock_sizes = @{$new_gblock_sizes};

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

		### END of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 5) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 8)" ###


		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/22).) (MAIN 6 & 7 & 8) ###

		### MAIN 9: Create the new (penultimate) coordinate frame.
		### [ MAIN 6: Create the new coordinate frame. ]
		### (CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 9) ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords0, $new_bds_gb_coords0, $new_gb_coords_wd_frm0)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm0) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER the \"Complementary-Vertical Split\" of \$gb_sbj (=$gb_sbj): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER the \"Complementary-Vertical Split\" of \$gb_sbj (=$gb_sbj) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);


   if ($TEST_LEVEL_cvspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER \"Complementary-Vertical-Split\" and BEFORE UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >5) {...}"

   if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER \"Complementary-Vertical-Split\" and BEFORE UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @{$new_bds_gb_coords0}, @{$new_gb_coords_wd_frm0}, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"

		### END of "(CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 9)" ###


		### MAIN 10: UNDO the Displacement. ###
		### (IMPORTED (w/ modifications) from 'cvmerge_gblock_pair' in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 10) ###

		# (Step 1:) Attempt to "undo" the displacements of the gap-blocks. #

	my @new_curr_bl_coords1 = @{$new_orgs_gb_coords0};

	my $ct_coord_chngs = 0;

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_round_trip_dsplc) {
		my $incr_coord = $gb2incr_coord_round_trip_dsplc{$gb};
		my $new_indx_gb = $new_indices_old_gbs[$gb];
		my $new_coord_gb0 = $new_curr_bl_coords1[$new_indx_gb];
		my $new_coord_gb = $new_coord_gb0 - $incr_coord;
		my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_gb]};
		if ($new_coord_gb < $coord_le_wd0) {
		    $new_coord_gb = $coord_le_wd0;
		} elsif ($coord_re_wd0 < $new_coord_gb) {
		    $new_coord_gb = $coord_re_wd0;
		}
		if ($new_coord_gb0 != $new_coord_gb) {
		    $new_curr_bl_coords1[$new_indx_gb] = $new_coord_gb;
		    $ct_coord_chngs++;
		}
	} # End of "foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord_round_trip_dsplc) {...}"


    if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n", '( Contents of @{$new_orgs_gb_coords0} (for "coord0"), @new_curr_bl_coords1 (for "coord1") & @{$new_gb_coords_wd_frm0} (for "coord_le0_wd" and "coord_re0_wd"): )', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'coord0', 'coord1', 'coord_le0_wd', 'coord_re0_wd'), "\n\n";

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $new_orgs_gb_coords0->[$gb], $new_curr_bl_coords1[$gb], @{$new_gb_coords_wd_frm0->[$gb]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"


	if ($ct_coord_chngs>0) {

	    my ($flag_ssh_p2p, $tot_incr_ct_null_clms_ssh_p2p)
	     = serial_shifts_point2point (
		@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_bl_coords1, @{$new_gb_coords_wd_frm0}, 
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@new_inter_block_relations, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
			);
	    unless ((defined $flag_ssh_p2p) and ($flag_ssh_p2p == 0)) { # A BUG fixed on 2019/07/27. #
		my $out_flag = (defined $flag_ssh_p2p) ? $flag_ssh_p2p : 'undef';
	    ##unless ($flag_ssh_p2p) {
		print STDOUT "[ERROR 'cvsplit_gblock'] The subroutine 'serial_shifts_point2point' FAILED on the new set of gap-blocks (w/ \$flag_ssh_p2p = $out_flag): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_point2point', Ended on the set of new gap-blocks (w/ \$tot_incr_ct_null_clms_ssh_p2p = $tot_incr_ct_null_clms_ssh_p2p) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

	} else {

   (print STDOUT "\n[INFORMATION 'cvspl_gblock'] The 'serial_shifts_point2point' will NOT be performed, because \$ct_coord_chngs = 0 !!...\n\n") if ($TEST_LEVEL_cvspl_gb > 1);

	} # End of "if ($ct_coord_chngs>0) {...} else {...}"


		# (Step 2:) Create the FINAL coordinate frame. #

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @new_inter_block_relations, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cvspl_gblock'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_inter_block_relations', " & the cofiguration of new gap-blocks AFTER UNDOING the displacements of the new gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER UNDOING the displacements of the new gap-blocks ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};


   if ($TEST_LEVEL_cvspl_gb >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER UNDOing the Displacement: ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvmege_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >5) {...}"

   if ($TEST_LEVEL_cvspl_gb > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER UNDOING the Displacement: ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 5) {...}"

		### (IMPORTED (w/ modifications) from 'cvmerge_gblock_pair' in creation of 'cvsplit_gblock' (on 2019/08/26). (MAIN 10) ###



		### MAIN 11: UNDO the horizontal overlaps among the vertically identical & complementary gap-blocks.
		### [ MAIN 9: UNDO the horizontal overlaps among the vertically identical & complementary gap-blocks. ] ... for the predecessor 'vsplit_gblock'. 
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 11) ###

			# The 1st Round (w.r.t. $gb_cch1_NEW) #

	my ($fl_undo1, $gb2incr_bl_coord_undo1) 
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_cch1_NEW, # Modified in creation of 'cvsplit_gblock'. #
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo1) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)) on the cofiguration of gap-blocks after \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW and \$gb_cch2_NEW (=$gb_cch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)), Ended (w/ \$fl_undo1 = $fl_undo1) on the configuration of gap-blocks after the \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW) and \$gb_cch2_NEW (=$gb_cch2_NEW) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);


   if ($TEST_LEVEL_cvspl_gb > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)): ]\n\n";

	print STDOUT "\$fl_undo1 = $fl_undo1 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo1} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo1}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo1->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"

	if ($fl_undo1 < 0) {
	    print STDOUT "[CAUTION 'cvsplit_gblock'] The 1st Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch1_NEW (=$gb_cch1_NEW)), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}


			# The 2nd Round (w.r.t. $gb_cch2_NEW) #

	my ($fl_undo2, $gb2incr_bl_coord_undo2) 
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch1_NEW, # A BUG Fixed on 2019/08/23. #
	  ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch2_NEW,
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_cch2_NEW, # MODIFIED in creation of 'cvsplit_gblock'. #
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo2) {
	    print STDOUT "[ERROR 'cvsplit_gblock'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)) on the cofiguration of gap-blocks after \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW and \$gb_cch2_NEW (=$gb_cch2_NEW): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)), Ended (w/ \$fl_undo2 = $fl_undo2) on the configuration of gap-blocks after the \"complementary-vertical split\" of \$gb_sbj (=$gb_sbj) into \$gb_cch1_NEW (=$gb_cch1_NEW) and \$gb_cch2_NEW (=$gb_cch2_NEW) ...\n\n") if ($TEST_LEVEL_cvspl_gb >1);


   if ($TEST_LEVEL_cvspl_gb > 3) {

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW)): ]\n\n";

	print STDOUT "\$fl_undo2 = $fl_undo2 .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo2} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo2}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo2->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"

	if ($fl_undo2 < 0) {
	    print STDOUT "[CAUTION 'cvsplit_gblock'] The 2nd Round of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch2_NEW (=$gb_cch2_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

			# Combine the outputs of the TWO Rounds into an output of this subroutine. #
			# (There should be NO Problem with it, b/c the two sets of keys SHOULD NOT overlap.)

	my %gb2incr_bl_coord_undo = (%{$gb2incr_bl_coord_undo1}, %{$gb2incr_bl_coord_undo2});


		### End of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/23).) (MAIN 9) ###
		### (=> CARRIED OVER (w/ modifications) to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 11)" ###


			### Test-Output the Final Results. ###

   if ($TEST_LEVEL_cvspl_gb >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Complementary-Vertical-Split\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb >3) {...}"


   if ($TEST_LEVEL_cvspl_gb > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Complementary-Vertical-Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"


		### MAIN 12: Finally Update  @new_list_gblocks. #
		### [ MAIN 10: Finally Update  @new_list_gblocks. # ] ... for the predecessor 'vsplit_gblock'. #
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###
		### (=> Carried over to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 12) ###

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    @{$new_list_gblocks[$gb]}[2,3] = @{$new_curr_bds_blks[$gb]};
		 # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 
	}

   if ($TEST_LEVEL_cvspl_gb > 3) {

	print STDOUT '[ FINAL Contents of @new_list_gblocks: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_out_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (12))  ###
		### (=> Carried over to 'cvsplit_gblock' (on 2019/08/26).) (MAIN 12)" ###

=For the predecessor, 'vsplit_gblock'. (MAIN 11)
		### MAIN 11: Create @new_list_excluded_bin_hrels.
		### (ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (MAIN 11)) ###

	my @new_list_excluded_bin_hrels = ([$gb_ch1_NEW, $gb_ch2_NEW, '=']);

   if ($TEST_LEVEL_cvspl_gb > 3) {

	print STDOUT '[ Contents of @new_list_excluded_bin_hrels: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";
	for (my $i=0; $i < @new_list_excluded_bin_hrels; $i++) {
	    print STDOUT join ("\t", $i, @{$new_list_excluded_bin_hrels[$i]}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvspl_gb > 3) {...}"

		### End of "(ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (MAIN 11))" ###
=cut # END of "For the predecessor, 'vsplit_gblock'. (MAIN 11)"


=For the predecessor, 'vsplit_gblock'. (FIN)
    (print STDOUT "\n< Subroutine 'vsplit_gblock' Ended (w/ \$fl_end = 1, (\$gb_ch1_NEW, \$gb_ch2_NEW) = ($gb_ch1_NEW, $gb_ch2_NEW), \$gb_shftd = $gb_shftd, \$incr_coord_shftd = $incr_coord_shftd). >\n\n") if ($TEST_LEVEL_vspl_gb > 0);

	## [return (1, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs, ] ... for the predecessor, 'cvmerge_gblock_pair'. #
	return (1,  $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, # MODIFIED in creation of 'vsplit_gblock' (on 2019/08/23). (OUT) # 
=cut # END of "For the predecessor, 'vsplit_gblock'. (FIN)"

    (print STDOUT "\n< Subroutine 'cvsplit_gblock' Ended (w/ \$fl_end = 1, (\$gb_cch1_NEW, \$gb_cch2_NEW) = ($gb_cch1_NEW, $gb_cch2_NEW)). >\n\n") if ($TEST_LEVEL_cvspl_gb > 0); # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (FIN)

	return (1, $gb_cch1_NEW, $gb_cch2_NEW, \%gb2incr_coord_1way_dsplc, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, # MODIFIED in creation of 'cvsplit_gblock' (on 2019/08/26). (OUT) #

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks
	### [, \@new_list_excluded_bin_hrels,  # ADDED in creation of 'vsplit_gblock' (on 2019/08/23). (OUT) # ] ... For the predecessor, 'vsplit_gblock'. #
	);

}	# End of the subroutine, 'cvsplit_gblock'. #


# smerge_split_gblock_pair:
#	This subroutine first "merge" a specified pair of gap-blocks of the SAME type,
#	then, it immediately "split" the resulting gap-block,
#	creating a pair of gap-blocks of the identical type but with (hopefully) different sizes.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the main part of 'smerge_gblock_pair'.
#	[CREATION STARTED on 2019/07/30, ENDED on 2019/07/31.]
#	[DEBUGGED 1st on 2019/07/31.]
#	[REVISED on 2019/08/01]:
#	Now, the subroutine makes Utmose Attempts to separate the resulting gap-blocks from each other; It it canNOT, it returns an EXCEPTIONAL VALUE (= -1) for $fl_end.
#
# smerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks of the SAME type.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/07/21.]
#	[DEBUGGED on 2019/07/22]:
#	Now, the subroutine can correctly handle a gap-block pair with ZERO distance
#	that flank another gap-block in the 'onx' relationship.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# INPUTS: ($gb1, $gb2, $new_size_right, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be of the same vertical type;
#
#	$new_size_right is the size of the gap-block on the right of the newly created pair;
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#		  = 0 if it failed,
#	 	  = -1 if the resulting gap-blocks will be UNABLE to separate from each other (ADDED on 2019/08/01);
#
#	$bl_sbj is the index for the gap-block on which the 1st round of 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj is the increment of the coordinate of $bl_sbj that resulted from the 1st round of 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	$new_bl_sbj is the index for the gap-block on which the 2nd round of 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_new_sbj is the increment of the coordinate of $new_bl_sbj that resulted from the 2nd round of 'serial_shifts_sgl_gblock' (= undef if the subroutine failed);
#
#	[[
#	$gb_absbd and $gb_kpt are the indices of the gap-blocks that were absorbed and kept (actually, absorbing), respectively;
#
#	@new_indices_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $gb th before the "merge"
#	has been assigned $new_indices_gbs[$gb] after the "merge"
#	(, and $new_indices_gbs[$gb1] and $new_indices_gbs[$gb2] are undefined);
#
#	@old_indices_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $gb th after the "merge"
#	had been assigned $old_indices_gbs[$gb] before the "merge";
#	]] ... for the predecesor, 'smerge_gblock_pair'.
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
sub smerge_split_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {

	my ($gb1, $gb2, $new_size_right, $curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq '=') {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];
	my $sum_size = $size1 + $size2;

	unless (($new_size_right>0) and ($new_size_right < $sum_size)) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The \$new_size_right (=$new_size_right) is NOT within the range, (0, \$size1 + \$size2 (=$sum_size)): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};

    (print STDOUT "< Subroutine 'smerge_split_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; same-type block-pair to be \"merge\"d + \"split\": (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$new_size_right = $new_size_right) >\n\n") if ($TEST_LEVEL_smrg_spl_gbp > 0);


		# (Preliminary 0): Determine which will be "remove"d and which will be "kept". #
		#	(NOTE added on 2019/07/30: In 'smerge_split_gblock_pair', $gb_absbd and $gb_kpt are merely nominal classifications, because BOTH gap-blocks are actually kept (albeit with size changes); 
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'smerge_gblock_pair' on 2019/07/21. (1)) #

	my ($gb_absbd, $gb_kpt) = ($gb2, $gb1); # Always keep $gb1. (MODIFIED in creation of 'smerge_gblock_pair'.)
	##my ($gb_rmvd, $gb_kpt) = ($size1 < $size2) ? ($gb1, $gb2) : ($gb2, $gb1);

	my ($size_absbd, $size_kpt) = @{$gblock_sizes}[$gb_absbd, $gb_kpt];

			# ADDED in creation  of 'smerge_split_gblock_pair' (on 2019/07/30). #
	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1) ;
	##my $lbd_left_bf = ($lbd1_bf < $lbd2_bf) ? $lbd1_bf : $lbd2_bf ;
	##my $rbd_right_bf = ($rbd1_bf > $rbd2_bf) ? $rbd1_bf : $rbd2_bf ;

				# MODIFIED in DEBUGGING on 2019/07/31. #
				# => Further MODIFIED on 2019/08/07. #
	my ($le_coord_left, $re_coord_left) = @{$bds_gb_coords->[$gb_left]};
	my ($le_coord_right,$re_coord_right) = @{$bds_gb_coords->[$gb_right]};
	##my $le_coord_left = $bds_gb_coords->[$gb_left][0];
	##my $re_coord_right = $bds_gb_coords->[$gb_right][1];
	####my ($le_coord1, $re_coord1) = @{$bds_gb_coords->[$gb1]};
	####my ($le_coord2, $re_coord2) = @{$bds_gb_coords->[$gb2]};
	####my $le_coord_left = ($le_coord1 < $le_coord2) ? $le_coord1 : $le_coord2;
	####my $re_coord_right = ($re_coord1 > $re_coord2) ? $re_coord1 : $re_coord2;
			# End of "ADDED in creation  of 'smerge_split_gblock_pair' (on 2019/07/30)." #

   if ($TEST_LEVEL_smrg_spl_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_absbd, $gb_kpt) = ', "($gb_absbd, $gb_kpt) .\n";
	print STDOUT '($size_absbd, $size_kpt) = ', "($size_absbd, $size_kpt) .\n\n";

			# ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/30). #
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";

			# => MODIFIED on 2019/08/07. #
	print STDOUT '($le_coord_left,  $re_coord_left)  = ', "($le_coord_left, $re_coord_left) ,\n";
	print STDOUT '($le_coord_right, $re_coord_right) = ', "($le_coord_right, $re_coord_right) .\n";
	##print STDOUT '($le_coord_left, $re_coord_right) = ', "($le_coord_left, $re_coord_right) .\n";
	####print STDOUT '($lbd_left_bf, $rbd_right_bf) = ', "($lbd_left_bf, $rbd_right_bf) .\n";
			# End of "ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/30)." #
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		#  (=> MODIFIED in creation of 'smerge_gblock_pair' on 2019/07/21. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_smrg_spl_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Merge + Split (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp > 5) {...}"


   if ($TEST_LEVEL_smrg_spl_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the 1st Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb2; # Always move the 2nd gap-block in the pair. #
	my $incr_coord_sbj;
	my $if_left = 0; # Added for DEBUGGING (on 2019/07/22). (1) #

	if ($lbd1_bf < $lbd2_bf) {

		# $gb1 is on the left of $gb2 (=$bl_sbj). #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($rbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $rbd1_bf;
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }

	    ##$incr_coord_sbj = - ($ct_nec_shifts1 + $size_rmvd); # "Shift" $gb2 to the left.
	    $incr_coord_sbj = - $ct_nec_shifts; # => MODIFIED in creation of 'smerge_gblock_pair'.

	    $if_left = 1; # Added for DEBUGGING (on 2019/07/22). (2) #

	} else { # if ($lbd2_bf < $lbd1_bf) 

		# $gb1 is on the right of $gb2 (=$bl_sbj). #

	    my $ct_nec_shifts = compt_distance_bw_gblk_and_pt ($lbd1_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	    unless (defined $ct_nec_shifts) {
		my $pt = $lbd1_bf;
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }

	    ##$incr_coord_sbj = $ct_nec_shifts1 + $size_rmvd; # "Shift" $gb2 to the right.
	    $incr_coord_sbj = $ct_nec_shifts; # => MODIFIED in creation of 'smerge_gblock_pair'. 

	} # End of "if ($lbd1_bf < $lbd2_bf) {...} else {...}"


			## If it turns out that the resulting new gap-blocks ($gb1 & $gb2)
			## canNOT be separated from each other,
			## return an EXCEPTIONAL VALUE (= -1) for $fl_end.
			## (ADDED in REVISION (on 2019/08/01).) (1) ##

	if ($incr_coord_sbj == 0) {
	    my ($curr_coord_left, $curr_coord_right) = @cp_curr_bl_coords[$gb_left, $gb_right];
		# my $le_coord_left = $bds_gb_coords->[$gb_left][0];
		# my $re_coord_right = $bds_gb_coords->[$gb_right][1];

	    if (($le_coord_left == $curr_coord_left) and ($curr_coord_right == $re_coord_right)) {
		return (-1, undef, undef, undef, undef);
	    }
	} # End of "if ($incr_coord_sbj == 0) {...}"

			## END of "(ADDED in REVISION (on 2019/08/01).) (1)" ##


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are completely "merge"d. ###

   (print STDOUT "\n ... Perform the 1st round of 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The 1st round of subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);


		# MOVE the gap-blocks in the 'ONX' relationships with $gb1 & $gb2, if any,
		# TO the outside of THE REAR-END of $bl_sbj. #
		# (ADDED in creation of 'smerge_gblock_pair' (on 2019/07/21). (MAIN 1)) #

		# MODIFIED for DEBUGGING (on 2019/07/22). (3) #
	if ($if_left) { # $gb1 is on the left of $gb2(=$bl_sbj).
	##if ($incr_coord_sbj < 0) { # $bl_sbj has just been "shift"ed to the left.
		# End of "MODIFIED for DEBUGGING (on 2019/07/22). (3)" #

	    my $ct_vextensions = vextend_left_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @cp_curr_set_clms, @{$inter_block_relations}, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vextensions) {
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'vextend_left_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vextend_left_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);

	    my $ct_vretracting = vretract_left_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations}, @{$gblock_sizes}, @cp_curr_gblks_w_spec_lb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vretract_left_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);

	} else { # if ($incr_coord_sbj > 0) # $bl_sbj has just been "shift"ed to the right. #

	    my $ct_vextensions = vextend_right_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @cp_curr_set_clms, @{$inter_block_relations}, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vextensions) {
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'vextend_right_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vextend_right_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);

	    my $ct_vretracting = vretract_right_end_sbj_gblk ($bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations}, @{$gblock_sizes}, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk', FAILED on the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vretract_right_end_sbj_gblk', Ended on the $bl_sbj th gap-block (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);

	} # End of "if ($incr_coord_sbj < 0) {...} else {...}". #

		# END of "(ADDED in creation of 'smerge_gblock_pair' (on 2019/07/21). (MAIN 1))" #


   if ($TEST_LEVEL_smrg_spl_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 1st Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >3) {...}"



		### ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/30). (MAIN 2&3) ###

		### MAIN 2:  Re-assign the boundaries of $gb1 and $gb2. ###
		###	(NOTE: Here, we WILL always place $gb1 on the left of $gb2.) ###
		###	Then, create some output data-sets.

	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my $sites_cntr_to_left = enum_sites_contribute_to_gblk ($gb_left, @cp_curr_bds_blks, 0, 1, @{$inter_block_relations});
	unless (defined $sites_cntr_to_left) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'enum_sites_contribute_to_gblk', FAILED on the \$gb_left (=$gb_left) th gap-block: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
	my $ct_sites_cntr_to_left = @{$sites_cntr_to_left};
   (print STDOUT "\n ... The subroutine, 'enum_sites_contribute_to_gblk', Ended on the \$gb_left (=$gb_left) th gap-block (w/ #{sites contributing to \$gb_left}=$ct_sites_cntr_to_left) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);

	my $sites_cntr_to_right = enum_sites_contribute_to_gblk ($gb_right, @cp_curr_bds_blks, 0, 1, @{$inter_block_relations});
	unless (defined $sites_cntr_to_right) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'enum_sites_contribute_to_gblk', FAILED on the \$gb_right (=$gb_right) th gap-block: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
	my $ct_sites_cntr_to_right = @{$sites_cntr_to_right};
   (print STDOUT "\n ... The subroutine, 'enum_sites_contribute_to_gblk', Ended on the \$gb_right (=$gb_right) th gap-block (w/ #{sites contributing to \$gb_right}=$ct_sites_cntr_to_right) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);


	unless (($size_left == $ct_sites_cntr_to_left) and ($size_right == $ct_sites_cntr_to_right)) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] IMPOSSIBLE!!: The numbers of sites contributing to \$gb_left(=$gb_left) and \$gb_right(=$gb_right), ($ct_sites_cntr_to_left, $ct_sites_cntr_to_right), do NOT EQUAL their sizes, (\$size_left, \$size_right) = ($size_left, $size_right): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}

	my @sites_cntr_to_merged = (@{$sites_cntr_to_left}, @{$sites_cntr_to_right});

   if ($TEST_LEVEL_smrg_spl_gbp >3) {

	print STDOUT "\n[ Outputs of 'enum_sites_contribute_to_gblk': ]\n\n";

	print STDOUT '@{$sites_cntr_to_left} = (', join (', ', @{$sites_cntr_to_left}), ") ,\n\n";
	print STDOUT '@{$sites_cntr_to_right} = (', join (', ',  @{$sites_cntr_to_right}), ") ,\n\n";
	print STDOUT '@sites_cntr_to_merged = (', join (', ',  @sites_cntr_to_merged), ") .\n\n";
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >3) {...}"

	my @new_curr_set_clms = @cp_curr_set_clms;

	my $new_size_left = $sum_size - $new_size_right; # ADDED on 2019/07/31.
	my ($new_lbd1, $new_rbd2) = @sites_cntr_to_merged[0, $sum_size-1];
	my ($new_rbd1, $new_lbd2) = @sites_cntr_to_merged[$new_size_left -1, $new_size_left];

   if ($TEST_LEVEL_smrg_spl_gbp >2) {

	print STDOUT "\n[ (IMPORTANT!!) New boundaries of \$gb1(=$gb1) and \$gb2(=$gb2), BEFORE the final serial \"shift\"s: ]\n\n";

	print STDOUT "For \$gb1:  [$new_lbd1, $new_rbd1] ,\n";
	print STDOUT "For \$gb2:  [$new_lbd2, $new_rbd2] .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >2) {...}"


	my @new_curr_bds_blks = @cp_curr_bds_blks;
	$new_curr_bds_blks[$gb1] = [$new_lbd1, $new_rbd1];
	$new_curr_bds_blks[$gb2] = [$new_lbd2, $new_rbd2];

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($lbd, $rbd) = @{$new_curr_bds_blks[$gb]};
	    push @{$new_curr_gblks_w_spec_lb[$lbd]}, $gb;
	    push @{$new_curr_gblks_w_spec_rb[$rbd]}, $gb;
	}

			# ADDEd on 2019/07/31. #
	my @new_gblock_sizes = @{$gblock_sizes};
	$new_gblock_sizes[$gb1] = $new_size_left;
	$new_gblock_sizes[$gb2] = $new_size_right;
			# End of "ADDEd on 2019/07/31." #


		### MAIN 3: Create the new coordinate frame. ###
		###	=> If possible, extend the ranges for $gb1 and $gb2. ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @{$inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@{$nter_block_relations}', " & the cofiguration of new gap-blocks AFTER the \"Merge + Split\" of \$gb1(=$gb1) and \$gb2(=$gb2): $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER the \"Merge + Split\" of \$gb1(=$gb1) and \$gb2(=$gb2) ...\n\n") if ($TEST_LEVEL_cvmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

			# If possible, extend the ranges for $gb1 and $gb2. #

	##my $le_coord_left = $bds_gb_coords->[$gb_left][0];
	##my $re_coord_right = $bds_gb_coords->[$gb_right][1];
	## [my ($le_coord1, $re_coord1) = @{$bds_gb_coords->[$gb1]};
	##  my ($le_coord2, $re_coord2) = @{$bds_gb_coords->[$gb2]};
	##  my $le_coord_left = ($le_coord1 < $le_coord2) ? $le_coord1 : $le_coord2;
	##  my $re_coord_right = ($re_coord1 > $re_coord2) ? $re_coord1 : $re_coord2; ] ... OBSOLETE after the DEBUGGING!!

	my $left_margin = $cp_curr_bl_coords[$gb_left] - $le_coord_left;
	my $right_margin = $re_coord_right - $cp_curr_bl_coords[$gb_right];

	my $org_new_coord = $left_margin;
	my ($lbd_new_coord, $rbd_new_coord) = (0, $left_margin + $right_margin);

	foreach my $gb ($gb1, $gb2) {
	    my $org_new_coord0 = $new_orgs_gb_coords->[$gb];
	    my $offset = $org_new_coord - $org_new_coord0;

	    $new_curr_bl_coords[$gb] = $org_new_coord;
	    $new_bds_gb_coords[$gb] = [$lbd_new_coord, $rbd_new_coord];
	    $new_gb_coords_wd_frm[$gb]->[0] += $offset;
	    $new_gb_coords_wd_frm[$gb]->[1] += $offset;

	} # End of "foreach my $gb ($gb1, $gb2) {...}"


			# Further DEBUGGED (on 2019/07/31). #
	if ($new_bds_gb_coords[$gb1]->[1] + $new_size_right <= $new_gb_coords_wd_frm[$gb1]->[1]) { # SWITCHED on in DEBUGGING (on 2019/07/31). #
	    $new_bds_gb_coords[$gb1]->[1] += $new_size_right; # SWITCH ON this command, if you want $gb1 to be reachable to the right-end of the frame of $gb2.
	}
	##if ($new_bds_gb_coords->[$gb1][1] + $new_size_right <= $new_gb_coords_wd_frm[$gb1][1]) { # SWITCHED on in DEBUGGING (on 2019/07/31). #
	##    $new_bds_gb_coords->[$gb1][1] += $new_size_right; # SWITCH ON this command, if you want $gb1 to be reachable to the right-end of the frame of $gb2.
	##}


		### End of "ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/30). (MAIN 2&3)" ###


		### ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/31). (MAIN 3post) ###

   if ($TEST_LEVEL_smrg_spl_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations after the RE-ASSIGNMENT of the Gap-Block Indices ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >5) {...}"

   if ($TEST_LEVEL_smrg_spl_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the RE-ASSIGNMENT of the Gap-Block Indices ]\n\n";

	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp > 5) {...}"


		### MAIN 3post: UNDO the 1st Serial \"Shift"s. ###

	my $new_bl_sbj = ($if_left) ? $gb2 : $gb1 ; # It MUST point to the $bl_sbj BEFORE the re-assignment of gap-block indices. #
		## if ($if_left) { # OLD $gb1 WAS on the left of OLD $gb2(=$bl_sbj). ...}
	my $incr_coord_new_sbj = - $incr_coord_sbj;

		# Make Utmost Attemts to separate the new gap-blocks, $gb1 and $gb2. #
		# (ADDED in REVISION (on 2019/08/01). (2)) #

	if ($incr_coord_sbj ==0) {
	    if ($if_left) { # OLD $gb1 WAS on the left of OLD $gb2(=$bl_sbj); ($new_bl_sbj= NEW $gb2).

		if ($new_curr_bl_coords[$gb2] < $new_curr_bds_blks[$gb2]->[1]) {
		    $incr_coord_new_sbj = 1;
		} elsif ($new_curr_bds_blks[$gb1]->[0] < $new_curr_bl_coords[$gb1]) {
		    $new_bl_sbj = $gb1;
		    $incr_coord_new_sbj = -1;
		}

	    } else { # OLD $gb2(=$bl_sbj) WAS on the left of OLD $gb1; ($new_bl_sbj = NEW $gb1)

		if ($new_curr_bds_blks[$gb1]->[0] < $new_curr_bl_coords[$gb1]) {
		    $incr_coord_new_sbj = -1;
		} elsif ($new_curr_bl_coords[$gb2] < $new_curr_bds_blks[$gb2]->[1]) {
		    $new_bl_sbj = $gb2;
		    $incr_coord_new_sbj = +1;
		}


	    } # End of "if ($if_left) {...} else {...}"

	} # End of "if ($incr_coord_sbj ==0) {...}"

		# END of "(ADDED in REVISION (on 2019/08/01). (2))" #


   (print STDOUT "\n ... Perform the 2nd round of 'serial_shifts_sgl_gblock' (w/ \$new_bl_sbj = $new_bl_sbj, \$incr_coord_new_sbj = $incr_coord_new_sbj), to UNDO the 1st one ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp > 2);

	my ($flag2, $incr_ct_null_clms2)
	 = serial_shifts_sgl_gblock (
	    @new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_gb_coords_wd_frm, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb,
	    $new_bl_sbj, $incr_coord_new_sbj, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
		);
	unless ((defined $flag2) and ($flag2 == 0)) {
	    my $flag2_out = (defined $flag2) ? $flag2 : 'undef';
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The 2nd round of subroutine, 'serial_shifts_sgl_gblock', FAILED on the $new_bl_sbj th gap-block (w/ \$flag2 = $flag2_out): $! !\n\n";

	    return (0, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd round of subroutine, 'serial_shifts_sgl_gblock', Ended on the $new_bl_sbj th gap-block (w/ \$incr_ct_null_clms2 = $incr_ct_null_clms2) ...\n\n") if ($TEST_LEVEL_smrg_spl_gbp >1);


   if ($TEST_LEVEL_smrg_spl_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge + Split (same type)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp >3) {...}"

   if ($TEST_LEVEL_smrg_spl_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Merge + Split (same type)\" ]\n\n";

	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_spl_gbp > 3) {...}"

		### End of "ADDED in creation of 'smerge_split_gblock_pair' (on 2019/07/31). (MAIN 3post)" ###


=For the predecessor 'smerge_gblock_pair'. (MAIN 2&3)

		### MAIN 2: Finish the "merge (same type)" 
		###   by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.

	my @new_indices_gbs = my @old_indices_gbs = ();

	for (my $ogb=0; $ogb < $ct_in_gbs; $ogb++) { # 1st main for-loop (over the old gap-blocks). #

	    if ($ogb == $gb_absbd) {
		$new_indices_gbs[$ogb] = undef;
		next;
	    }

	    my $ngb;
	    if ($ogb < $gb_absbd) {
		$ngb = $ogb;
	    } else { # if ($gb_absbd < $ogb)
		$ngb = $ogb-1;
	    }

	    $new_indices_gbs[$ogb] = $ngb;
	    $old_indices_gbs[$ngb] = $ogb;

	} # End of the 1st main for-loop (over the old gap-blocks). #

	my $ct_out_gbs = @old_indices_gbs;

   if ($TEST_LEVEL_smrg_gbp >2) {

	print STDOUT "\n[ Relationships between new and old indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

			# A BUG Fixed on 2019/07/15. #
	my @out_new_indices_gbs = my @out_old_indices_gbs = ();
	foreach my $indx (@new_indices_gbs) { push @out_new_indices_gbs, ((defined $indx) ? $indx : 'undef'); }
	foreach my $indx (@old_indices_gbs) { push @out_old_indices_gbs, ((defined $indx) ? $indx : 'undef'); }

	print STDOUT '@new_indices_gbs = (', join (', ', @out_new_indices_gbs), ") .\n";
	print STDOUT '@old_indices_gbs = (', join (', ', @out_old_indices_gbs), ") .\n";
	##print STDOUT '@new_indices_gbs = (', join (', ', @new_indices_gbs), ") .\n";
	##print STDOUT '@old_indices_gbs = (', join (', ', @old_indices_gbs), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >2) {...}"

		# MOVED from below the 2nd main for-loop (in creation of 'smerge_gblock_pair'). #
	my $ngb_kpt = $new_indices_gbs[$gb_kpt]; # The new index of the gap-block to be kept.
		# End of "MOVED from below the 2nd main for-loop (in creation of 'smerge_gblock_pair')." #

	my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs];

	my ($lbd_absbd_af, $rbd_absbd_af) = @{$cp_curr_bds_blks[$gb_absbd]};
	my ($lbd_kpt_af, $rbd_kpt_af) = @{$cp_curr_bds_blks[$gb_kpt]};
	##my ($lbd1_af, $rbd1_af) = @{$cp_curr_bds_blks[$gb1]};
	##my ($lbd2_af, $rbd2_af) = @{$cp_curr_bds_blks[$gb2]};

	unless (($rbd_kpt_af < $lbd_absbd_af) or ($rbd_absbd_af < $lbd_kpt_af)) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The BOUNDARIES of \$gb_kpt(=$gb_kpt), [$lbd_kpt_af, $rbd_kpt_af], does OVERLAP those of \$gb_absbd(=$gb_absbd), [$lbd_absbd_af, $rbd_absbd_af]: $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, undef);
	}

		# ADDED in creation of 'smerge_gblock_pair'. #

	my @new_curr_bds_blks = @cp_curr_bds_blks[@old_indices_gbs];

	if ($rbd_kpt_af < $lbd_absbd_af) {
		# Replace the right-end of the gap-block to be kept. #
	    $new_curr_bds_blks[$ngb_kpt]->[1] = $rbd_absbd_af;

	} else { # if ($rbd_absbd_af < $lbd_kpt_af) 
		# Replace the left-end of the gap-block to be kept.
	    $new_curr_bds_blks[$ngb_kpt]->[0] = $lbd_absbd_af;

	}
		# End of "ADDED in creation of 'smerge_gblock_pair'." #

		# MODIFIED in creation of 'smerge_gblock_pair'. (2a) #

	my @new_curr_set_clms = @cp_curr_set_clms; # The set of columns does NOT change as the result of the "merge (same type)", since the end of the serial "shift"s. #

 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # Preliminary main for-loop (over the columns). #
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	} # End of the preliminary main for-loop (over the columns).

	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 2nd main for-loop (over the new gap-blocks). #
	    my ($lb_ngb, $rb_ngb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb_ngb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb_ngb]}, $ngb;
	} # End of the 2nd main for-loop (over the new gap-blocks). #

		# END of "MODIFIED in creation of 'smerge_gblock_pair'. (2a)" #


			# MAIN 3pre: Create @new_gblock_sizes & @new_inter_block_relations.
			# (MOVED from "MAIN 3" below, in creation of 'smerge_gblock_pair' on 2019/07/19.) #


	my @new_gblock_sizes = @{$gblock_sizes}[@old_indices_gbs];
	##$new_gblock_sizes[$ngb_kpt] = $size_kpt - $size_rmvd;
	$new_gblock_sizes[$ngb_kpt] = $size_kpt + $size_absbd; # => MODIFIED in creation of 'smerge_gblock_pair' (on 2019/07/21). (3pre) #

	my @new_inter_block_relations = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) { # 3rd main for-loop (over the new gap-blocks). #

	    my $ogb = $old_indices_gbs[$ngb];
	    my @new_rels_w_ngb = @{$inter_block_relations->[$ogb]}[@old_indices_gbs];
	    $new_inter_block_relations[$ngb] = \@new_rels_w_ngb;

	} # End of the 3rd main for-loop (over the new gap-blocks). #

			# End of "(MOVED from "MAIN 3" below, in creation of 'smerge_gblock_pair' on 2019/07/19.)" #


		### MAIN 3: Create the remaining output data-sets. ###

	my @rows_affd_by_new_blks = @cp_rows_affd_by_blks[@old_indices_gbs];
	my @new_bds_gb_coords = @cp_bds_gb_coords[@old_indices_gbs];
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm[@old_indices_gbs];

		# ADDED in creation of 'smerge_gblock_pair'. (MAIN 3) #

	my ($coord_bl_absbd, $coord_bl_kpt) = @cp_curr_bl_coords[$gb_absbd, $gb_kpt];
	my ($lb_coord_absbd, $rb_coord_absbd) = @{$cp_bds_gb_coords[$gb_absbd]};
	my ($lb_coord_kpt, $rb_coord_kpt) = @{$cp_bds_gb_coords[$gb_kpt]};
	my ($coord_absbd_le_wd, $coord_absbd_re_wd) = @{$cp_gb_coords_wd_frm[$gb_absbd]};
	my ($coord_kpt_le_wd, $coord_kpt_re_wd) = @{$cp_gb_coords_wd_frm[$gb_kpt]};

	if ($rbd_kpt_af < $lbd_absbd_af) {

		# $gb_kpt is on the left of $gb_absbd. #

	    my $ct_left_movable = $coord_bl_kpt - $lb_coord_kpt;
	    my $ct_right_movable = $rb_coord_absbd - $coord_bl_absbd;

	    my $new_rb_coord_kpt = $ct_left_movable + $ct_right_movable;

	    $new_curr_bl_coords[$ngb_kpt] = $ct_left_movable;
	    $new_bds_gb_coords[$ngb_kpt] = [0, $new_rb_coord_kpt];

	    my $new_left_margin = $coord_kpt_le_wd - $lb_coord_kpt;
	    my $new_right_margin = $coord_absbd_re_wd - $rb_coord_absbd;

	    $new_gb_coords_wd_frm[$ngb_kpt] = [ $new_left_margin, $new_rb_coord_kpt + $new_right_margin];

	} else { # if ($rbd_absbd_af < $lbd_kpt_af) 

		# $gb_absbd is on the left of $gb_kpt. #

	    my $ct_left_movable = $coord_bl_absbd - $lb_coord_absbd;
	    my $ct_right_movable = $rb_coord_kpt - $coord_bl_kpt;

	    my $new_rb_coord_kpt = $ct_left_movable + $ct_right_movable;

	    $new_curr_bl_coords[$ngb_kpt] = $ct_left_movable;
	    $new_bds_gb_coords[$ngb_kpt] = [0, $new_rb_coord_kpt];

	    my $new_left_margin = $coord_absbd_le_wd - $lb_coord_absbd;
	    my $new_right_margin = $coord_kpt_re_wd - $rb_coord_kpt;

	    $new_gb_coords_wd_frm[$ngb_kpt] = [ $new_left_margin, $new_rb_coord_kpt + $new_right_margin];

	}
		# END of "ADDED in creation of 'smerge_gblock_pair'. (MAIN 3)" #


	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	   = cts_vincluding_vincluded_onx_gblks (@new_inter_block_relations) ;
	unless (defined $new_cts_onx) {
	    print STDOUT "[ERROR 'smerge_split_gblock_pair'] The subroutine, 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@new_inter_block_relations', ": $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, undef);
	}
   (print STDOUT "\n ... The subroutine, 'cts_vincluding_vincluded_onx_gblks', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_smrg_gbp >1);


		# MOVED from above MAIN 3 in creation of 'smerge_gblock_pair'. #
   if ($TEST_LEVEL_smrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Merge (same type)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in creation of 'smerge_gblock_pair'." #


   if ($TEST_LEVEL_smrg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Merge (same type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@new_inter_block_relations, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'smerge_split_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'smerge_split_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_smrg_gbp > 3) {...}"

=cut # END of "For the predecessor 'smerge_gblock_pair'. (MAIN 2&3)"

		### MODIFIED in creation of 'smerge_split_gblock_pair' (on 2019/07/31). (FIN) ###

    (print STDOUT "\n< Subroutine 'smerge_split_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$new_bl_sbj = $new_bl_sbj, \$incr_coord_new_sbj = $incr_coord_new_sbj). >\n\n") if ($TEST_LEVEL_smrg_spl_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $new_bl_sbj, $incr_coord_new_sbj,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes);

=For the predecessor 'smerge_gblock_pair'. (FIN)
    (print STDOUT "\n< Subroutine 'smerge_split_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj = $incr_coord_sbj, \$gb_absbd = $gb_absbd, \$gb_kpt = $gb_kpt). >\n\n") if ($TEST_LEVEL_smrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj, $gb_absbd, $gb_kpt, \@new_indices_gbs, \@old_indices_gbs,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	\@new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes);
=cut # END of "For the predecessor 'smerge_gblock_pair'. (FIN)"

		### END of "MODIFIED in creation of 'smerge_split_gblock_pair' (on 2019/07/31). (FIN)" ###

}	# End of the subroutine, 'smerge_split_gblock_pair'. #


# icmerge_gblock_pair:
#	This subroutine performs the "INcomplete merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, by a specified number of columns.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/08/02.]
#	[REVISED on 2019/08/10]:
#	CHANGED the coordinate increment for the 2nd "serial_shifts_sgl_gblock'
#	so that the resulting gap-configuration will be closer to the input one.
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where ] ... for the predecessor, 'cmerge_gblock_pair'.
# INPUTS: ($gb1, $gb2, $ct_to_be_merged, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#
#	$ct_to_be_merged is the number of columns that are to be "merge"d via this subroutine;
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $gb_rmvd, $gb_kpt, $ct_rmvd_null_clms, \@new_indices_gbs, \@old_indices_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes), where ] ... for the predecessor 'cmerge_gblock_pair'.
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj1 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) BEFORE the "incomplete merge";
#		# END of "ADDED in revision on 2019/07/15." #
#
#	$incr_coord_sbj2 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) AFTER the "incomplete merge";
#
#	[[
#		# ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19). #
#	$gb_rmvd and $gb_kpt are the (old) indices of the gap-blocks that were removed and kept, respectively;
#		# End of "ADDED in creation of 'cmerge_gblock_pair' (on 2019/07/19)." #
#	]] ... for the predecessor 'cmerge_gblock_pair'. (a)
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "merge" (= undef if an error occurs);
#
#	[[
#	@new_indices_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $gb th before the "merge"
#	has been assigned $new_indices_gbs[$gb] after the "merge"
#	(, and $new_indices_gbs[$gb1] and $new_indices_gbs[$gb2] are undefined);
#
#	@old_indices_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $gb th after the "merge"
#	had been assigned $old_indices_gbs[$gb] before the "merge";
#	]] ... for the predecessor 'cmerge_gblock_pair'. (b)
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
sub icmerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {

	my ($gb1, $gb2, $ct_to_be_merged,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless ($ct_to_be_merged>0) { # Added on 2019/08/03. #
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The \$ct_to_be_merged (=$ct_to_be_merged) is NOT POSITIVE: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];
	unless (($ct_to_be_merged < $size1) and ($ct_to_be_merged < $size2)) {
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The \$ct_to_be_merged (=$ct_to_be_merged) is NOT SMALLER THAN BOTH the size of \$gb1(=$gb1), i.e., $size1, and the size of \$gb2(=$gb2), i.e., $size2: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};



    (print STDOUT "< Subroutine 'icmerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"INcomplete-merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$ct_to_be_merged = $ct_to_be_merged) >\n\n") if ($TEST_LEVEL_icmrg_gbp > 0);

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #

	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1);
	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};

   if ($TEST_LEVEL_icmrg_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_icmrg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Merge (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'icmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp > 5) {...}"


   if ($TEST_LEVEL_icmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'icmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp >3) {...}"


		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb_right; # Always move the gap-block on the right. #
	##my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	##my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};


	my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($rbd_left_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	unless (defined $ct_nec_shifts1) {
		my $pt = $rbd_left_bf;
		print STDOUT "[ERROR 'icmerge_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	}

	my $incr_coord_sbj1 = - ($ct_nec_shifts1 + $ct_to_be_merged); # "Shift" $gb2 to the left.

	my $incr_coord_sbj2 = - $incr_coord_sbj1; # MODIFIED in REVISION on 2019/08/10.
	##my $incr_coord_sbj2 = + $ct_nec_shifts1;


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are INcompletely "merge"d. ###

   (print STDOUT "\n ... Perform the 1st round of 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj1 = $incr_coord_sbj1) ...\n\n") if ($TEST_LEVEL_icmrg_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj1, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The 1st round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);


   if ($TEST_LEVEL_icmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'icmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp >3) {...}"


		### MAIN 2: Finish the "merge (complementary types)" by removing the generated null-columns,
		###   and by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.


	##my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs]; # DISABLED in revision on 2019/07/21.

	my $ct_out_gbs = $ct_in_gbs; # ADDED on 2019/08/02.

	my @new_curr_bds_blks = ();
	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) { $new_curr_bds_blks[$ngb] = []; }

	my ($lbd_left_af, $rbd_left_af) = @{$cp_curr_bds_blks[$gb_left]};
	my ($lbd_right_af, $rbd_right_af) = @{$cp_curr_bds_blks[$gb_right]};

	unless (($lbd_left_af < $lbd_right_af) and ($lbd_right_af <= $rbd_left_af) and ($rbd_left_af < $rbd_right_af)) {
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The BOUNDARIES of \$gb_left(=$gb_left), [$lbd_left_af, $rbd_left_af], does NOT NON-NESTEDLY OVERLAP those of \$gb_right(=$gb_right), [$lbd_right_af, $rbd_right_af]: $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}

	my $ct_rmvd_null_clms = 0;

	my @new_curr_set_clms = ();
 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $new_indx_clm = -1;
	my $pos_1st_rmvd_null_clm;

	for (my $c = 0; $c < $ct_in_clms; $c++) { # 2nd main for-loop (over the columns). #

	    if (($lbd_right_af <= $c) and ($c <= $rbd_left_af)) { # MODIFIED in creation of 'icmerge_gblock_pair'. #
	    ##if (($lbd_rmvd_af <= $c) and ($c <= $rbd_rmvd_af)) {
		my $cnct_clm = join (':', @{$cp_curr_set_clms[$c]}) ;
		if ($cnct_clm eq $cnct_null_clm) {
		    $ct_rmvd_null_clms++;
		    unless (defined $pos_1st_rmvd_null_clm) { $pos_1st_rmvd_null_clm = $c; }
		    next;
		}
	    }

	    $new_indx_clm++;
	    $new_curr_set_clms[$new_indx_clm] = $cp_curr_set_clms[$c];

	    my @ngbs_w_old_lb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 1st inner foreach-loop (over gap-blocks with old left-bound = $c). #
		my $ngb = $ogb; # MODIFIED in creation of 'icmerge_gblock_pair'. #
		push @ngbs_w_old_lb_eq_c, $ngb;
		$new_curr_bds_blks[$ngb]->[0] = $new_indx_clm;
	    } # End of the 1st inner foreach-loop. #

	    $new_curr_gblks_w_spec_lb[$new_indx_clm] = \@ngbs_w_old_lb_eq_c;

	    my @ngbs_w_old_rb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 2nd inner foreach-loop (over gap-blocks with old right-bound = $c). #
		my $ngb = $ogb; # MODIFIED in creation of 'icmerge_gblock_pair'. #
		push @ngbs_w_old_rb_eq_c, $ngb;
		$new_curr_bds_blks[$ngb]->[1] = $new_indx_clm;
	    } # End of the 2nd inner foreach-loop. #

	    $new_curr_gblks_w_spec_rb[$new_indx_clm] = \@ngbs_w_old_rb_eq_c;

	} # End of the 2nd main for-loop (over the columns). #

	unless ($ct_to_be_merged == $ct_rmvd_null_clms) {
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The BOUNDARIES of \$ct_rmvd_null_clms (=$ct_rmvd_null_clms) CLASHES with the INPUT, \$ct_to_be_merged (=$ct_to_be_merged): $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}


			# MAIN 3pre: Create @new_gblock_sizes & @new_inter_block_relations.
			# (MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.) #

	my @new_gblock_sizes = @{$gblock_sizes};
	$new_gblock_sizes[$gb_left] -= $ct_to_be_merged;
	$new_gblock_sizes[$gb_right] -= $ct_to_be_merged;

			# End of "(MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.)" #

			# MAIN 2post: Recover the removed boundaries of $gb_left and $gb_right,
			# then, Retract it if necessary.
			# (ADDED in creation of 'icmerge_gblock_pair' on 2019/08/02. (2)) #

	my $new_rbd_left = $pos_1st_rmvd_null_clm-1;
	push @{$new_curr_gblks_w_spec_rb[$new_rbd_left]}, $gb_left;
	$new_curr_bds_blks[$gb_left]->[1] = $new_rbd_left;

	my $new_lbd_right = $pos_1st_rmvd_null_clm;
	push @{$new_curr_gblks_w_spec_lb[$new_lbd_right]}, $gb_right;
	$new_curr_bds_blks[$gb_right]->[0] = $new_lbd_right;

			# Attempt to Retract the right-end of $gb_left. #

	my $ct_vretracting_r = vretract_right_end_sbj_gblk ($gb_left, @new_curr_bds_blks, @{$inter_block_relations}, @new_gblock_sizes, @new_curr_gblks_w_spec_rb);
	unless (defined $ct_vretracting_r) {
		print STDOUT "[ERROR 'icmerge_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk' FAILED on \$gb_left (=$gb_left): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'vretract_right_end_sbj_gblk' Ended on \$gb_left (=$gb_left) (w/ \$ct_vretracting_r = $ct_vretracting_r) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);


			# Attempt to Retract the left-end of $gb_right. #

	my $ct_vretracting_l = vretract_left_end_sbj_gblk ($gb_right, @new_curr_bds_blks, @{$inter_block_relations}, @new_gblock_sizes, @new_curr_gblks_w_spec_lb);
	unless (defined $ct_vretracting_l) {
		print STDOUT "[ERROR 'icmerge_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk' FAILED on \$gb_right (=$gb_right): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'vretract_left_end_sbj_gblk' Ended on \$gb_right (=$gb_right) (w/ \$ct_vretracting_l = $ct_vretracting_l) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);


   if ($TEST_LEVEL_icmrg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the  \"INcomplete Merge (complementary types)\" but BEFORE the 2nd Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @cp_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp >5) {...}"

			# END of "(ADDED in creation of 'icmerge_gblock_pair' on 2019/08/02. (2))" #


		### MAIN 3: Perform the 2nd Serial "Shift"s. ###
		### (ADDED in creation of 'icmerge_gblock_pair' (on 2019/08/02).) ### 

   (print STDOUT "\n ... Perform the 2nd round of 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj2 = $incr_coord_sbj2) ...\n\n") if ($TEST_LEVEL_icmrg_gbp > 2);

	my ($flag2, $incr_ct_null_clms2)
	 = serial_shifts_sgl_gblock (
	    @new_curr_set_clms, @new_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj2, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
		);
	unless ((defined $flag2) and ($flag2 == 0)) {
	    my $flag_out = (defined $flag2) ? $flag2 : 'undef';
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The 2nd round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag2 = $flag_out): $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms2 = $incr_ct_null_clms2) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);


		### END of "(ADDED in creation of 'icmerge_gblock_pair' (on 2019/08/02).)" ### 


		### [ MAIN 3: Create the remaining output data-sets. ### ] ... for the predecessor 'cmerge_gblock_pair'. #
		### MAIN 5: Create the remaining output data-sets. ###

		# ADDED in revision on 2019/07/21. #

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);
	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @{$inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'icmerge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_curr_bds_blks', " : $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		# END of "ADDED in revision on 2019/07/21." #


		# MOVED from above MAIN 3 in revision on 2019/07/21. #
   if ($TEST_LEVEL_icmrg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"INcomplete Merge (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in revision on 2019/07/21." #


   if ($TEST_LEVEL_icmrg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"INcomplete Merge (complementary types)\" ]\n\n";

				# MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02). #
	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'icmerge_gblock_pair'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'icmerge_gblock_pair' ...\n\n";
	}
				# END of "MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02)." #
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_icmrg_gbp > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'icmerge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj1 = $incr_coord_sbj1, #{removed null-columns}=$ct_rmvd_null_clms, \$incr_coord_sbj2 = $incr_coord_sbj2). >\n\n") if ($TEST_LEVEL_icmrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes);

}	# End of the subroutine, 'icmerge_gblock_pair'. #


# cmerge_split_grow_gblock_pair_OLD:
#	It turned out that building a perfect subroutine is extremely difficult
#	under this design.
#	Thus, this version of 'cmerge_split_grow_gblock_pair' was RENAMED as 'cmerge_split_grow_gblock_pair_OLD',
#	to remind the users that this is (nearly) obsolete.
#	Still, it was kept, with the hope that it may become useful on some occasions.
#	[RENAMED on 2019/08/06.]
#
# cmerge_split_grow_gblock_pair:
#	This subroutine performs the "merge + split" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, and grows their sizes by a specified number of columns.
#	It also prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'icmerge_gblock_pair'.
#	[CREATED on 2019/08/03.]
#	[DEBUGGED on 2019/08/05]:
#	(i) Now, it should be able to CORRECTLY separate the gap-block boundaries
#	belonging to the columns to be "split".
#	(ii) CORRECTED the use of the output, %{$gb2sbincr_coord}, of 'displace_vincluding_veq_and_onx_to_right_end'.
#	[REVISED on 2019/08/05]:
#	The rules for assigning the coordinate increments for the 2nd (UNDOing) serial "shift"s,
#	so that the final result will look closer to the input alignment.
#
# icmerge_gblock_pair:
#	This subroutine performs the "INcomplete merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, by a specified number of columns.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/08/02.]
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, $ct_to_be_merged, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where ] ... for the predecessor 'icmerge_gblock_pair'.
#
# INPUTS: ($gb1, $gb2, $incr_gb_size, $char_abs, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#
#	$incr_gb_size is the number of columns by which the sizes of $gb1 and $gb2 will increase;
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where ] ... for the predecessor 'icmerge_gblock_pair'.
#
# OUTPUTS: ($flag_end, $gb_left, $gb_right, $incr_coord_left1, $incr_coord_right1, $incr_coord_left2, $incr_coord_right2,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#
#	$gb_left and $gb_right are the indices of the gap-blocks that are on the left and on the right, respectively, each chosen between $gb1 and $gb2;
#
#	$incr_coord_left1 and $incr_coord_right1 are the coordinate increments
#	for the 1st rounds of serial "shift"s on $gb_left and $gb_right, respectively;
#
#	$incr_coord_left2 and $incr_coord_right2 are the coordinate increments
#	for the 2nd rounds of serial "shift"s on $gb_left and $gb_right, respectively;
#	(NOTE that the 2nd rounds are intended to UNDO the 1st rounds);
#
#	[[
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj1 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) BEFORE the "incomplete merge";
#		# END of "ADDED in revision on 2019/07/15." #
#
#	$incr_coord_sbj2 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) AFTER the "incomplete merge";
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "merge" (= undef if an error occurs);
#	]] ... for the predecessor 'icmerge_gblock_pair'.
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
#sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) {
#
sub cmerge_split_grow_gblock_pair_OLD ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) { # RENAMED on 2019/08/06.


	my ($gb1, $gb2, $incr_gb_size, $char_abs,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless ($incr_gb_size>0) { # Added on 2019/08/03. #
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The \$incr_gb_size (=$incr_gb_size) is NOT POSITIVE: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

=For the predecessor 'icmerge_gblock_pair'.
	unless (($ct_to_be_merged < $size1) and ($ct_to_be_merged < $size2)) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The \$ct_to_be_merged (=$ct_to_be_merged) is NOT SMALLER THAN BOTH the size of \$gb1(=$gb1), i.e., $size1, and the size of \$gb2(=$gb2), i.e., $size2: $! !\n\n";
	    return (0, undef, undef, undef);
	}
=cut # End of "For the predecessor 'icmerge_gblock_pair'."

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};

    (print STDOUT "< Subroutine 'cmerge_split_grow_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"INcomplete-merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$incr_gb_size = $incr_gb_size) >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #

	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1);
	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};

   if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {...}"

		# MAIN 1: Perform the 1st rounds of 'serial_shifts_sgl_gblock' on $gb_left and $gb_right,
		#        so that $gb_left will be EFFECTIVELY on the left-end of the window
		#	and so that $gb_right will EFFECTIVELY touch the right-end of $gb_left,
		# 	withOUT being intervened by ANY gap-blocks with 'onx' relationship with $gb_right.

		# (i) Perform 'serial_shifts_sgl_gblock' on $gb_left.

	my $incr_coord_left1 = $cp_gb_coords_wd_frm[$gb_left]->[0] - $cp_curr_bl_coords[$gb_left];
	my $incr_coord_left2 = - $incr_coord_left1;

   (print STDOUT "\n ... Perform the 1st round of 'serial_shifts_sgl_gblock' on \$gb_left (=$gb_left)  (w/  \$incr_coord_left1 = $incr_coord_left1) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 2);

	my ($flag_left1, $incr_ct_null_clms_left1)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $gb_left, $incr_coord_left1, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag_left1) and ($flag_left1 == 0)) {
	    my $flag_out = (defined $flag_left1) ? $flag_left1 : 'undef';
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 1st round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on \$gb_left (=$gb_left) (w/ \$flag_left1 = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'serial_shifts_sgl_gblock', Ended on \$gb_left (=$gb_left) (w/ \$incr_ct_null_clms_left1 = $incr_ct_null_clms_left1) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	my $rbd_left_af_1R = $cp_curr_bds_blks[$gb_left]->[1];


		# (ii) Perform 'serial_shifts_sgl_gblock' on $gb_right.

	my $ct_nec_shifts_right1 = compt_distance_bw_gblk_and_pt ($rbd_left_af_1R, $gb_right, @cp_curr_bds_blks, @{$inter_block_relations});
	unless (defined $ct_nec_shifts_right1) {
		my $pt = $rbd_left_af_1R;
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the \$gb_right (=$gb_right) th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	}

	my $incr_coord_right1 = - $ct_nec_shifts_right1;
	my $incr_coord_right2 = $ct_nec_shifts_right1; # REVISED on 2019/08/05.
	##my $incr_coord_right2 = ($ct_nec_shifts_right1>$incr_gb_size) ? $ct_nec_shifts_right1 - $incr_gb_size : 0;

   (print STDOUT "\n ... Perform the 1st round of 'serial_shifts_sgl_gblock' on \$gb_right (=$gb_right)  (w/  \$incr_coord_right1 = $incr_coord_right1) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 2);

	my ($flag_right1, $incr_ct_null_clms_right1)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $gb_right, $incr_coord_right1, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag_right1) and ($flag_right1 == 0)) {
	    my $flag_out = (defined $flag_right1) ? $flag_right1 : 'undef';
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 1st round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on \$gb_right (=$gb_right) (w/ \$flag_right1 = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'serial_shifts_sgl_gblock', Ended on \$gb_right (=$gb_right) (w/ \$incr_ct_null_clms_right1 = $incr_ct_null_clms_right1) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	my ($lbd_right_af_1R, $rbd_right_af_1R) = @{$cp_curr_bds_blks[$gb_right]};


		# (iii) If any, move gap-blocks with 'onx' relationship with $gb_right to the right of $gb_right.

	if (($rbd_left_af_1R + 1 < $lbd_right_af_1R) and ($cts_onx->[$gb_right]>0)) {

	    my $ct_vextensions = vextend_left_end_sbj_gblk ($gb_right, @cp_curr_bds_blks, @cp_curr_set_clms, @{$inter_block_relations}, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb);
	    unless (defined $ct_vextensions) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'vextend_left_end_sbj_gblk', FAILED on \$gb_right (=$gb_right): $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vextend_left_end_sbj_gblk', Ended on \$gb_right (=$gb_right)  (w/ \$ct_vextensions = $ct_vextensions) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	    my $ct_vretracting = vretract_left_end_sbj_gblk ($gb_right, @cp_curr_bds_blks, @{$inter_block_relations}, @{$gblock_sizes}, @cp_curr_gblks_w_spec_lb);
	    unless (defined $ct_vretracting) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk', FAILED on \$gb_right (=$gb_right): $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'vretract_left_end_sbj_gblk', Ended on \$gb_right (=$gb_right) (w/ \$ct_vretracting = $ct_vretracting) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	} # End of "if (($rbd_left_af_1R + 1 < $lbd_right_af_1R) and ($cts_onx->[$gb_right]>0)) {...}"

	($lbd_right_af_1R, $rbd_right_af_1R) = @{$cp_curr_bds_blks[$gb_right]};


	### MAIN 1.5:  DISPLACE the gap-blocks to the right-end of the window,
	###		if they are on the right of $gb_right
	###		AND if they are either including, vertically equivalent, or in 'onx' relationships with either $gb_left or $gb_right.

	my $left_bound = $rbd_right_af_1R +1;

	my %ignored = ($gb_left => 1, $gb_right => 1);

		# MODIFIED in revision on 2019/08/04.
	my %gb2incr_coord1 = ($gb_left => - $incr_coord_left2, $gb_right => - $incr_coord_right2);
	##my %gb2incr_coord1;
		# End of "MODIFIED in revision on 2019/08/04."

	my $totct_nec_shifts = 0;

	my $le_rlv_gbs = $ct_in_clms -1;

	foreach my $gb_sbj ($gb_left, $gb_right) {

	    my ($gb2sbincr_coord, $sbtotct_nec_shifts, $le_sbrlv_gbs)
		= displace_vincluding_veq_and_onx_to_right_end ($gb_sbj, $left_bound,
		  @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
		  @cp_rows_affd_by_blks, $cnct_null_clm,
		  @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, %ignored
			);
	    unless (defined $gb2sbincr_coord) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'displace_vincluding_veq_and_onx_to_right_end', FAILED on \$gb_sbj (=$gb_sbj) (w/ \$left_bound = $left_bound): $! !\n\n";
		return (0, undef, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'displace_vincluding_veq_and_onx_to_right_end', Ended on the \$gb_sbj (=$gb_sbj) th gap-block and \$left_bound = $left_bound (w/ \$sbtotct_nec_shifts = $sbtotct_nec_shifts, \$le_sbrlv_gbs = $le_sbrlv_gbs) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	    $totct_nec_shifts += $sbtotct_nec_shifts;

	    foreach my $gb (sort {$a <=> $b} keys %{$gb2sbincr_coord}) {

		my $sbincr_coord = $gb2sbincr_coord->{$gb}; # A BUG FIXED on 2019/08/05. #
		##my $sbincr_coord = $gb2sbincr_coord;
		if (defined $gb2incr_coord1{$gb}) {
		    $gb2incr_coord1{$gb} += $sbincr_coord;
		} else {
		    $gb2incr_coord1{$gb} = $sbincr_coord;
		}
	    }

	    ##if ($le_sbrlv_gbs < $le_rlv_gbs) { $le_rlv_gbs = $le_sbrlv_gbs; } # DISABLED in revision on 2019/08/04.

	} # End of "foreach my $gb_sbj ($gb_left, $gb_right) {...}"

			# ADDED in REVISION on 2019/08/04. #
	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord1) {
	     if (($gb == $gb_left) or ($gb == $gb_right)) { next; }
	     my $lbd_gb = $cp_curr_bds_blks[$gb]->[0];
	    if ($lbd_gb < $le_rlv_gbs) { $le_rlv_gbs = $lbd_gb; }
	}
			# End of "ADDED in REVISION on 2019/08/04." #

	my $effective_margin = $le_rlv_gbs - $left_bound;

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Intermediate Variables ]\n\n";

	print STDOUT '$totct_nec_shifts = ', $totct_nec_shifts, " .\n";
	print STDOUT "\n\n";

	print STDOUT '%gb2incr_coord1 = ($gb => $incr_coord,', "\n";
	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord1) {
		print STDOUT "\t", $gb , ' => ', $gb2incr_coord1{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

	print STDOUT '$le_rlv_gbs = ', $le_rlv_gbs, " .\n";
	print STDOUT '$effective_margin = ', $effective_margin, " .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"


	unless ($incr_gb_size <= $effective_margin) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The input argument, \$incr_gb_size (=$incr_gb_size), is LAGER THAN the size of the EFFECTIVE MARGIN (=$effective_margin) available after all the pre-processing \"Shift\"s: $! !\n\n";
	    return (0, undef, undef, undef);
	}

=For the predecessor 'icmerge_gap_blocks'. (Pre2 & Main1)
		# (Preliminary 2): Give the two missing arguments for 'serial_shifts_sgl_gblock'. #

	my $bl_sbj = $gb_right; # Always move the gap-block on the right. #
	##my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	##my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};


	my $ct_nec_shifts1 = compt_distance_bw_gblk_and_pt ($rbd_left_bf, $bl_sbj, @cp_curr_bds_blks, @{$inter_block_relations});
	unless (defined $ct_nec_shifts1) {
		my $pt = $rbd_left_bf;
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'compt_distance_bw_gblk_and_pt' FAILED to compute the distance between the point $pt and the $bl_sbj th gap-block: $! !\n\n";
		return (0, undef, undef, undef);
	}

	my $incr_coord_sbj1 = - ($ct_nec_shifts1 + $ct_to_be_merged); # "Shift" $gb2 to the left.
	my $incr_coord_sbj2 = + $ct_nec_shifts1;


		### MAIN 1: Perform the 'serial_shifts_sgl_gblock' until the two gap-blocks are INcompletely "merge"d. ###

   (print STDOUT "\n ... Perform the 1st round of 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj1 = $incr_coord_sbj1) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 2);

	my ($flag, $incr_ct_null_clms)
	 = serial_shifts_sgl_gblock (
	    @cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj1, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $flag) and ($flag == 0)) {
	    my $flag_out = (defined $flag) ? $flag : 'undef';
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 1st round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag = $flag_out): $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms = $incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);

=cut # End of "For the predecessor 'icmerge_gap_blocks'. (Pre2 & Main1)"


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 1st Rounds of Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {...}"


		### MAIN 2: Perform the size-growing "Merge + Split". ###

		# (i) Prepare the increments to $gb_left and $gb_right. #
		#     & new elements for @curr_gblks_w_spec_lb & @curr_gblks_w_spec_rb.

	my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];
	my ($rows_affd_by_left, $rows_affd_by_right) = @cp_rows_affd_by_blks[$gb_left, $gb_right];

	my @clms_in_incr_left = my @clms_in_incr_right = ();
	my @curr_gblks_w_lb_in_incr_left  = my @curr_gblks_w_lb_in_incr_right = ();
	my @curr_gblks_w_rb_in_incr_left  = my @curr_gblks_w_rb_in_incr_right = ();

	for (my $k=0; $k <$incr_gb_size; $k++) { # Preliminary for-loop for MAIN 2 (over columns to be "split"). #

	    my $c = $left_bound + $k;
	    my @clm_in_incr_left = @null_clm;
	    my @clm_in_incr_right = @{$cp_curr_set_clms[$c]};
	    foreach my $row (@{$rows_affd_by_right}) { # 1st inner foreach.
		my $tmp = $clm_in_incr_right[$row];
		$clm_in_incr_right[$row] = $clm_in_incr_left[$row];
		$clm_in_incr_left[$row] = $tmp;
	    } # End of the 1st inner foreach.
	    push @clms_in_incr_left, \@clm_in_incr_left;
	    push @clms_in_incr_right, \@clm_in_incr_right;

	    my @gblks_w_lb_into_left = my @gblks_w_lb_into_right = ();
	    foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 2nd inner foreach.
		my $rel = $rels_w_right->[$gb];

		if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
		##if ($rel eq '>') {
		    push @gblks_w_lb_into_left, $gb;
		} else {
		    push @gblks_w_lb_into_right, $gb;
		}
	    } # End of the 2nd inner foreach.
	    push @curr_gblks_w_lb_in_incr_left, \@gblks_w_lb_into_left;
	    push @curr_gblks_w_lb_in_incr_right, \@gblks_w_lb_into_right;

	    my @gblks_w_rb_into_left = my @gblks_w_rb_into_right = ();
	    foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 3rd inner foreach.
		my $rel = $rels_w_right->[$gb];

		if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
		##if ($rel eq '>') {
		    push @gblks_w_rb_into_left, $gb;
		} else {
		    push @gblks_w_rb_into_right, $gb;
		}
	    } # End of the 3rd inner foreach.
	    push @curr_gblks_w_rb_in_incr_left, \@gblks_w_rb_into_left;
	    push @curr_gblks_w_rb_in_incr_right, \@gblks_w_rb_into_right;


	} # End of the Preliminary for-loop for MAIN 2 (over columns to be "split"). #


		# (ii) Execute the size-growing "Merge + Split". #

	my @new_curr_set_clms = ();
	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $indx_new_clm= -1;
	for (my $c = 0; $c <= $rbd_left_af_1R; $c++) { # 1st main for-loop for MAIN 2. #
		$indx_new_clm++;

		$new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
		$new_curr_gblks_w_spec_lb[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
		$new_curr_gblks_w_spec_rb[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];

	} # End of the 1st main for-loop for MAIN 2. #

	##my @clms_in_incre_left = my @clms_in_incre_right = ();
	##my @curr_gblks_w_lb_in_incre_left  = my @curr_gblks_w_lb_in_incr_right = ();
	##my @curr_gblks_w_rb_in_incre_left  = my @curr_gblks_w_rb_in_incr_right = ();

	for (my $k=0; $k <$incr_gb_size; $k++) { # 2nd main for-loop for MAIN 2. #
		$indx_new_clm++;

		$new_curr_set_clms[$indx_new_clm] = $clms_in_incr_left[$k];
		$new_curr_gblks_w_spec_lb[$indx_new_clm] = $curr_gblks_w_lb_in_incr_left[$k];
		$new_curr_gblks_w_spec_rb[$indx_new_clm] = $curr_gblks_w_rb_in_incr_left[$k];

	} # End of the 2nd main for-loop for MAIN 2. #

	for (my $c=$rbd_left_af_1R +1; $c <= $rbd_right_af_1R; $c++) { # 3rd main for-loop for MAIN 2.
		$indx_new_clm++;

		$new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
		$new_curr_gblks_w_spec_lb[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
		$new_curr_gblks_w_spec_rb[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];

	} # End of the 3rd main for-loop for MAIN 2. #

	for (my $k=0; $k <$incr_gb_size; $k++) { # 5th main for-loop for MAIN 2. #
		$indx_new_clm++;

		$new_curr_set_clms[$indx_new_clm] = $clms_in_incr_right[$k];
		$new_curr_gblks_w_spec_lb[$indx_new_clm] = $curr_gblks_w_lb_in_incr_right[$k];
		$new_curr_gblks_w_spec_rb[$indx_new_clm] = $curr_gblks_w_rb_in_incr_right[$k];

	} # End of the 5th main for-loop for MAIN 2. #

	for (my $c=$rbd_right_af_1R +$incr_gb_size+1; $c < $ct_in_clms; $c++) { # 6th main for-loop for MAIN 2.  
		$indx_new_clm++;

		$new_curr_set_clms[$indx_new_clm] = $cp_curr_set_clms[$c];
		$new_curr_gblks_w_spec_lb[$indx_new_clm] = $cp_curr_gblks_w_spec_lb[$c];
		$new_curr_gblks_w_spec_rb[$indx_new_clm] = $cp_curr_gblks_w_spec_rb[$c];

	} # End of the 6th main for-loop for MAIN 2. #

	my $ct_out_clms = @new_curr_set_clms;
	unless ($ct_out_clms == $ct_in_clms + $incr_gb_size) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] Something is WRONG!!: The number of output columns (=$ct_out_clms) does NOT EQUAL the number of input columns (=$ct_in_clms) PLUS the size of the gap-block increment (=$incr_gb_size): $! !\n\n";
	    return (0, undef, undef, undef);
	}
	my $ct_out_gbs = $ct_in_gbs; # The number of gap-blocks does NOT change. # (ADDED in DEBUGGING on 2019/08/04.) #


		# Create @new_curr_bds_blks. #

	my @new_curr_bds_blks = ();
	for (my $gb=0; $gb < $ct_out_gbs; $gb++) { $new_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$ct_out_clms; $c++)  { # 2nd trailing for-loop for MAIN 2. #

	    foreach my $gb (@{$new_curr_gblks_w_spec_lb[$c]}) {
		$new_curr_bds_blks[$gb]->[0] = $c;
	    }
	    foreach my $gb (@{$new_curr_gblks_w_spec_rb[$c]}) {
		$new_curr_bds_blks[$gb]->[1] = $c;
	    }

	} # End of the 2nd trailing for-lop for MAIN 2. #


=For the predecessor 'icmerge_gap_blocks'. (Main2)
		### MAIN 2: Finish the "merge (complementary types)" by removing the generated null-columns,
		###   and by creating new data-sets to appropriately represent the resulting local MSA & gap-blocks.


	##my @new_curr_bl_coords = @cp_curr_bl_coords[@old_indices_gbs]; # DISABLED in revision on 2019/07/21.

	my $ct_out_gbs = $ct_in_gbs; # ADDED on 2019/08/02.

	my @new_curr_bds_blks = ();
	for (my $ngb = 0; $ngb < $ct_out_gbs; $ngb++) { $new_curr_bds_blks[$ngb] = []; }

	my ($lbd_left_af, $rbd_left_af) = @{$cp_curr_bds_blks[$gb_left]};
	my ($lbd_right_af, $rbd_right_af) = @{$cp_curr_bds_blks[$gb_right]};

	unless (($lbd_left_af < $lbd_right_af) and ($lbd_right_af <= $rbd_left_af) and ($rbd_left_af < $rbd_right_af)) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The BOUNDARIES of \$gb_left(=$gb_left), [$lbd_left_af, $rbd_left_af], does NOT NON-NESTEDLY OVERLAP those of \$gb_right(=$gb_right), [$lbd_right_af, $rbd_right_af]: $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}

	my $ct_rmvd_null_clms = 0;

	my @new_curr_set_clms = ();
 	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();

	my $new_indx_clm = -1;
	my $pos_1st_rmvd_null_clm;

	for (my $c = 0; $c < $ct_in_clms; $c++) { # 2nd main for-loop (over the columns). #

	    if (($lbd_right_af <= $c) and ($c <= $rbd_left_af)) { # MODIFIED in creation of 'icmerge_gblock_pair'. #
	    ##if (($lbd_rmvd_af <= $c) and ($c <= $rbd_rmvd_af)) {
		my $cnct_clm = join (':', @{$cp_curr_set_clms[$c]}) ;
		if ($cnct_clm eq $cnct_null_clm) {
		    $ct_rmvd_null_clms++;
		    unless (defined $pos_1st_rmvd_null_clm) { $pos_1st_rmvd_null_clm = $c; }
		    next;
		}
	    }

	    $new_indx_clm++;
	    $new_curr_set_clms[$new_indx_clm] = $cp_curr_set_clms[$c];

	    my @ngbs_w_old_lb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_lb[$c]}) { # 1st inner foreach-loop (over gap-blocks with old left-bound = $c). #
		my $ngb = $ogb; # MODIFIED in creation of 'icmerge_gblock_pair'. #
		push @ngbs_w_old_lb_eq_c, $ngb;
		$new_curr_bds_blks[$ngb]->[0] = $new_indx_clm;
	    } # End of the 1st inner foreach-loop. #

	    $new_curr_gblks_w_spec_lb[$new_indx_clm] = \@ngbs_w_old_lb_eq_c;

	    my @ngbs_w_old_rb_eq_c = ();
	    foreach my $ogb (@{$cp_curr_gblks_w_spec_rb[$c]}) { # 2nd inner foreach-loop (over gap-blocks with old right-bound = $c). #
		my $ngb = $ogb; # MODIFIED in creation of 'icmerge_gblock_pair'. #
		push @ngbs_w_old_rb_eq_c, $ngb;
		$new_curr_bds_blks[$ngb]->[1] = $new_indx_clm;
	    } # End of the 2nd inner foreach-loop. #

	    $new_curr_gblks_w_spec_rb[$new_indx_clm] = \@ngbs_w_old_rb_eq_c;

	} # End of the 2nd main for-loop (over the columns). #

	unless ($ct_to_be_merged == $ct_rmvd_null_clms) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The BOUNDARIES of \$ct_rmvd_null_clms (=$ct_rmvd_null_clms) CLASHES with the INPUT, \$ct_to_be_merged (=$ct_to_be_merged): $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}
=cut # END of "For the predecessor 'icmerge_gap_blocks'. (Main2)"


			# MAIN 3pre: Create @new_gblock_sizes & @new_inter_block_relations.
			# (MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.) #

	my @new_gblock_sizes = @{$gblock_sizes};
	$new_gblock_sizes[$gb_left] += $incr_gb_size;
	$new_gblock_sizes[$gb_right] += $incr_gb_size;

			# End of "(MOVED from "MAIN 3" below, in creation of 'cmerge_gblock_pair' on 2019/07/19.)" #

		### MAIN 2post: Correct the right-boundaries of $gb_left and $gb_right. #

	foreach my $gb ($gb_left, $gb_right) { # For-loop for MAIN 2post.

		my $old_rbd_gb = $new_curr_bds_blks[$gb]->[1];
		my $new_rbd_gb = $old_rbd_gb + $incr_gb_size;
		$new_curr_bds_blks[$gb]->[1] = $new_rbd_gb;

		my @new_gblks_w_rb_eq_old = ();
		foreach my $gb2 (@{$new_curr_gblks_w_spec_rb[$old_rbd_gb]}) {
		    unless ($gb2 == $gb) { push @new_gblks_w_rb_eq_old, $gb2; }
		}
		$new_curr_gblks_w_spec_rb[$old_rbd_gb] = \@new_gblks_w_rb_eq_old;

		push @{$new_curr_gblks_w_spec_rb[$new_rbd_gb]}, $gb;

	} # End of the for-loop for MAIN 2post.


=For the predecessor 'icmerge_gblock_pair'. (Main 2post)
			# MAIN 2post: Recover the removed boundaries of $gb_left and $gb_right,
			# then, Retract it if necessary.
			# (ADDED in creation of 'icmerge_gblock_pair' on 2019/08/02. (2)) #

	my $new_rbd_left = $pos_1st_rmvd_null_clm-1;
	push @{$new_curr_gblks_w_spec_rb[$new_rbd_left]}, $gb_left;
	$new_curr_bds_blks[$gb_left]->[1] = $new_rbd_left;

	my $new_lbd_right = $pos_1st_rmvd_null_clm;
	push @{$new_curr_gblks_w_spec_lb[$new_lbd_right]}, $gb_right;
	$new_curr_bds_blks[$gb_right]->[0] = $new_lbd_right;

			# Attempt to Retract the right-end of $gb_left. #

	my $ct_vretracting_r = vretract_right_end_sbj_gblk ($gb_left, @new_curr_bds_blks, @{$inter_block_relations}, @new_gblock_sizes, @new_curr_gblks_w_spec_rb);
	unless (defined $ct_vretracting_r) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'vretract_right_end_sbj_gblk' FAILED on \$gb_left (=$gb_left): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'vretract_right_end_sbj_gblk' Ended on \$gb_left (=$gb_left) (w/ \$ct_vretracting_r = $ct_vretracting_r) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);


			# Attempt to Retract the left-end of $gb_right. #

	my $ct_vretracting_l = vretract_left_end_sbj_gblk ($gb_right, @new_curr_bds_blks, @{$inter_block_relations}, @new_gblock_sizes, @new_curr_gblks_w_spec_lb);
	unless (defined $ct_vretracting_l) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'vretract_left_end_sbj_gblk' FAILED on \$gb_right (=$gb_right): $! !\n\n";

		return (0, $bl_sbj, $incr_coord_sbj1, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'vretract_left_end_sbj_gblk' Ended on \$gb_right (=$gb_right) (w/ \$ct_vretracting_l = $ct_vretracting_l) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the  \"(size-growing) Merge + Split (complementary types)\" but BEFORE UNDOING the 1st Round of Serial \"Shift\"s ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @cp_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {...}"

			# END of "(ADDED in creation of 'icmerge_gblock_pair' on 2019/08/02. (2))" #

=cut # END of "For the predecessor 'icmerge_gblock_pair'. (Main 2post)"


		### MAIN 3: UNDO the 1st Rounds of Serial "Shift"s &
		###		Create the new coordinate frame VIA 3 STEPS.
		### (ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (11)
		### => IMPORTED to 'cmerge_split_grow_gblock_pair' on 2019/08/04. (M3)) ###

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);

		# (Step 1:) Create a coordinate frame using the current gap-block configuration. #

	my ($new_orgs_gb_coords0, $new_bds_gb_coords0, $new_gb_coords_wd_frm0)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @{$inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm0) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of new gap-blocks immediately after the \"(size-growing) merge + split\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately after the \"(size-growing) merge + split\" ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations immediately AFTER \"(size-growing) Merge + Split\" and BEFORE UNDOING the 1st Rounds of Serial \"Shift\"s: ]\n\n";

	my @new_curr_bl_coords0 = @{$new_orgs_gb_coords0};
	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords0, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmege_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {...}"


		# (Step 2:) Attempt to "UNDO" the 1st Rounds of Serial "Shift"s (including the Displacements) of the gap-blocks. #

	my @new_curr_bl_coords1 = @{$new_orgs_gb_coords0};

	my $ct_coord_chngs = 0;

=For the predecessor 'cvmerge_gblock_pair'.
	if (!($if_complete) and ($bl_sbj == $gb_rmvd_kpt)) {
	    my $new_indx_sbj = $new_indices_old_gbs[$bl_sbj];
	    my $new_coord_sbj0 = $new_curr_bl_coords1[$new_indx_sbj];
	    my $new_coord_sbj = $new_coord_sbj0 - $incr_coord_sbj;
	    my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_sbj]};
	    if ($new_coord_sbj < $coord_le_wd0) {
		$new_coord_sbj = $coord_le_wd0;
	    } elsif ($coord_re_wd0 < $new_coord_sbj) {
		$new_coord_sbj = $coord_re_wd0;
	    }
	    if ($new_coord_sbj0 != $new_coord_sbj) {
		$new_curr_bl_coords1[$new_indx_sbj] = $new_coord_sbj;
		$ct_coord_chngs++;
	    }
	} # End of "if (!($if_complete) and ($bl_sbj == $gb_rmvd_kpt)) {...}"
=cut # End of "For the predecessor 'cvmerge_gblock_pair'."

	## my %gb2incr_coord1 = ($gb_left => - $incr_coord_left2, $gb_right => - $incr_coord_right2);
	## my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];

			# MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (a-L) #
	foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord1) { 
	##foreach my $gb2incr_coord ($gb2incr_coord_left, $gb2incr_coord_right) {
	##    foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord}) {
			# End of "MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (a-L)" #


			# MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (b) #
		my $incr_coord1 = $gb2incr_coord1{$gb};
		my $incr_coord = $incr_coord1;


			# => Further MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/05). #
		my $rel_r = $rels_w_right->[$gb];
		my $rel_l = $rels_w_left->[$gb];
		my $if_lengthened =0;
		if ((defined $rel_l) and (($rel_l eq '<') or ($rel_l eq '<(pa)') or (($rel_l eq '=') and ($gb < $gb_left)) ) ) {
			if ($curr_bds_blks->[$gb][1] <= $curr_bds_blks->[$gb_left][1]) { $if_lengthened = 1; }
		} elsif ((defined $rel_r) and (($rel_r eq '<') or ($rel_r eq '<(pa)') or (($rel_r eq '=') and ($gb < $gb_right)) ) ) {
			if ($curr_bds_blks->[$gb][1] <= $curr_bds_blks->[$gb_right][1]) { $if_lengthened = 1; }
		}
		if ($if_lengthened) { $incr_coord += $incr_gb_size; }
		##my $rel_r = $rels_w_right->[$gb];
		##if ( (defined $rel_r) and (($rel_r eq 'ONN') or ($rel_r eq 'ONCS')) ) {
		##    if ($incr_coord > $incr_gb_size) {
		##	$incr_coord -= $incr_gb_size;
		##    } else {
		##	$incr_coord = 0;
		##    }
		##}
		##my $incr_coord = $gb2incr_coord->{$gb};
			# End of "=> Further MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/05)." #
			# End of "MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (b)" #

		my $new_indx_gb = $gb; # MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (c) #
		##my $new_indx_gb = $new_indices_old_gbs[$gb];

		my $new_coord_gb0 = $new_curr_bl_coords1[$new_indx_gb];
		my $new_coord_gb = $new_coord_gb0 - $incr_coord;
		my ($coord_le_wd0, $coord_re_wd0) = @{$new_gb_coords_wd_frm0->[$new_indx_gb]};
		if ($new_coord_gb < $coord_le_wd0) {
		    $new_coord_gb = $coord_le_wd0;
		} elsif ($coord_re_wd0 < $new_coord_gb) {
		    $new_coord_gb = $coord_re_wd0;
		}
		if ($new_coord_gb0 != $new_coord_gb) {
		    $new_curr_bl_coords1[$new_indx_gb] = $new_coord_gb;
		    $ct_coord_chngs++;
		}
			# MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (a-R) #
	##    } # End of "foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord}) {...}"
	##} # End of "foreach my $gb2incr_coord ($gb2incr_coord_left, $gb2incr_coord_right) {...}"
	} # End of "foreach my $gb (sort {$a <=> $b} keys %gb2incr_coord1) {...}"
			# End of "MODIFIED in creation of 'cmerge_split_grow_gblock_pair' (on 2019/08/04). (a-R)" #

    if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n", '( Contents of @{$new_orgs_gb_coords0} (for "coord0"), @new_curr_bl_coords1 (for "coord1") & @{$new_gb_coords_wd_frm0} (for "coord_le0_wd" and "coord_re0_wd"): )', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'coord0', 'coord1', 'coord_le0_wd', 'coord_re0_wd'), "\n\n";

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $new_orgs_gb_coords0->[$gb], $new_curr_bl_coords1[$gb], @{$new_gb_coords_wd_frm0->[$gb]}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"


	if ($ct_coord_chngs>0) {

	    my ($flag_ssh_p2p, $tot_incr_ct_null_clms_ssh_p2p)
	     = serial_shifts_point2point (
		@new_curr_set_clms, @new_curr_bds_blks, @{$new_orgs_gb_coords0}, @new_curr_bl_coords1, @{$new_gb_coords_wd_frm0}, 
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @cp_rows_affd_by_blks, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
			);
	    unless ((defined $flag_ssh_p2p) and ($flag_ssh_p2p == 0)) { # A BUG fixed on 2019/07/27. #
		my $out_flag = (defined $flag_ssh_p2p) ? $flag_ssh_p2p : 'undef';
	    ##unless ($flag_ssh_p2p) {
		print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'serial_shifts_point2point' FAILED on the UPDATED set of gap-block-related data (w/ \$flag_ssh_p2p = $out_flag): $! !\n\n";
		return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_point2point', Ended on the set of UPDATED set of gap-block-related data (w/ \$tot_incr_ct_null_clms_ssh_p2p = $tot_incr_ct_null_clms_ssh_p2p) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	} else {

   (print STDOUT "\n[INFORMATION 'cmerge_split_grow_gblock_pair'] The 'serial_shifts_point2point' will NOT be performed, because \$ct_coord_chngs = 0 !!...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 1);

	} # End of "if ($ct_coord_chngs>0) {...} else {...}"


		# (Step 3:) Create the FINAL coordinate frame. #

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @{$inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of new gap-blocks AFTER UNDOING the serial \"shift\"s (including the displacements) of the gap-blocks: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of new gap-blocks immediately AFTER UNDOING the serial \"shift\"s (including displacements) of the gap-blocks ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		### END of "(ADDED in creation of 'cvmerge_gblock_pair' on 2019/07/26. (11)    ###
		###        => IMPORTED to 'cmerge_split_grow_gblock_pair' on 2019/08/04. (M3))" ###


=For the predecessor 'icmerge_gblock_pair'. (Main 3 & 5)
		### MAIN 3: Perform the 2nd Serial "Shift"s. ###
		### (ADDED in creation of 'icmerge_gblock_pair' (on 2019/08/02).) ### 

   (print STDOUT "\n ... Perform the 2nd round of 'serial_shifts_sgl_gblock' (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj2 = $incr_coord_sbj2) ...\n\n") if ($TEST_LEVEL_icmrg_gbp > 2);

	my ($flag2, $incr_ct_null_clms2)
	 = serial_shifts_sgl_gblock (
	    @new_curr_set_clms, @new_curr_bds_blks, @cp_curr_bl_coords, @cp_gb_coords_wd_frm, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb,
	    $bl_sbj, $incr_coord_sbj2, @cp_rows_affd_by_blks, $cnct_null_clm,
	    @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
		);
	unless ((defined $flag2) and ($flag2 == 0)) {
	    my $flag_out = (defined $flag2) ? $flag2 : 'undef';
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 2nd round of the subroutine, 'serial_shifts_sgl_gblock', FAILED on the $bl_sbj th gap-block (w/ \$flag2 = $flag_out): $! !\n\n";

	    return (0, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'serial_shifts_sgl_gblock', Ended on the $bl_sbj th gap-block (w/ \$incr_ct_null_clms2 = $incr_ct_null_clms2) ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);

		### END of "(ADDED in creation of 'icmerge_gblock_pair' (on 2019/08/02).)" ### 


		### [ MAIN 3: Create the remaining output data-sets. ### ] ... for the predecessor 'cmerge_gblock_pair'. #
		### MAIN 5: Create the remaining output data-sets. ###

		# ADDED in revision on 2019/07/21. #

	my ($le_new_wd, $re_new_wd) = (0, $#new_curr_set_clms);
	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $le_new_wd, $re_new_wd, @{$inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on ", '@new_curr_bds_blks', " : $! !\n\n";
	    return (0, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the new set of gap-blocks ...\n\n") if ($TEST_LEVEL_icmrg_gbp >1);

	my @new_curr_bl_coords = @{$new_orgs_gb_coords};
	my @new_bds_gb_coords = @{$new_bds_gb_coords};
	my @new_gb_coords_wd_frm = @{$new_gb_coords_wd_frm};

		# END of "ADDED in revision on 2019/07/21." #
=cut # END of "For the predecessor 'icmerge_gblock_pair'. (Main 3 & 5)"


		# MOVED from above MAIN 3 in revision on 2019/07/21. #
   if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {...}"
		# END of "MOVED from above MAIN 3 in revision on 2019/07/21." #


   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

				# MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02). #
	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
				# END of "MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02)." #
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'cmerge_split_grow_gblock_pair' Ended (w/ ", '($gb_left, $gb_right) = ', "($gb_left, $gb_right), \$incr_coord_left1 = $incr_coord_left1, \$incr_coord_right1 = $incr_coord_right1, \$incr_coord_left2 = $incr_coord_left2, \$incr_coord_right2 = $incr_coord_right2). >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);

	return (1, $gb_left, $gb_right, $incr_coord_left1, $incr_coord_right1, $incr_coord_left2, $incr_coord_right2,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes);

=For the predecessor, 'icmerge_gblock_pair'. (FIN)
    (print STDOUT "\n< Subroutine 'icmerge_gblock_pair' Ended (w/ \$bl_sbj = $bl_sbj, \$incr_coord_sbj1 = $incr_coord_sbj1, #{removed null-columns}=$ct_rmvd_null_clms, \$incr_coord_sbj2 = $incr_coord_sbj2). >\n\n") if ($TEST_LEVEL_icmrg_gbp > 0);

	return (1, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes);
=cut # END of "For the predecessor, 'icmerge_gblock_pair'. (FIN)"

##}	# End of the subroutine, 'cmerge_split_grow_gblock_pair'. #
}	# End of the subroutine, 'cmerge_split_grow_gblock_pair_OLD'. #


# undo_hovlp_vcmpl_gbsets_1sbj:
#	This modified version of the predecessor, 'undo_hovlp_vcmpl_gbsets', uses only ONE subject gap-block,
#	instead of TWO as done by the predecessor.
#	[CREATED on 2019/08/22.]
#
#	NOTE that this subroutine carries over the predecessor's "overshooting" problem noted on 2019/08/23;
#	The problem MUST be rectified as soon as possible.
#
# undo_hovlp_vcmpl_gbsets:
#	This satellite subroutine of 'cmerge_split_grow_gblock_pair'
#	UNDOes the overlaps between vertically equivalent and/or complementary gap-blocks
#	resulting from the main process of \"merge + split\" of a pair of gap-blocks.
#	[CREATED on 2019/08/08.]
#	[DEBUGGED on 2019/08/09-10]:
#	A BUG was found in the 3rd main for-loop;
#	it prevented the "undo"ing of the overlap(s) via the "shift"s to the left.
#	Then, the BUG was fixed.
#	[REVISED on 2019/08/12]:
#	Now, the subroutine can also accept gap-block pairs that are vertical identical to each other.
#
#	[NOTE added on 2019/08/23]:
#	It seems that the current version sometimes "overshoot"s,
#	especially when some vertically identical blocks are nested.
#	In the near future, this problem must be rectified.
#
#	[STRATEGY]
#	(1) First, attempt to undo the overlap(s) by moving the gap-blocks on the right of the overlap(s) further to the left;
#	(2), If the overlap(s) still remain(s), move the gap-blocks on the left of the overlap(s) further to the right;
#	(3) If the overlap(s) still remain(s), return -1.
#
# [ INPUTS: ($gb_left, $gb_right, ] ... for the predecessor, 'undo_hovlp_vcmpl_gbsets'.
# INPUTS: ($gb_sbj, # MODIFIED in creation of 'undo_hovlp_vcmpl_gbsets_1sbj'. (HD1a) #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_bds_gb_coords, \@new_curr_bl_coords,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_blks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@new_gblock_sizes
#	), where
#
#	$gb_sbj is the index of the gap-block concerning which the overlaps will be UNDONE; # MODIFIED in creation of 'undo_hovlp_vcmpl_gbsets_1sbj'. (HD1b) #
#
#	[[
#	$gb_left and $gb_right are the indices of the gap-blocks on the left and on the right, respectively;
#	They MUST be vertically complementary to each other;
#	]] ... for the predecessor, 'undo_hovlp_vcmpl_gbsets'.
#
#	Other input arguments are the same as those used for 'serial_shifts_point2point',
#	which will be invoked within this subroutine;
#	[For details, see the header of 'serial_shifts_point2point'];
#
# OUTPUTS: ($fl_end, \%gb2incr_bl_coord_undo), where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed,
#		= -1 if the overlap(s) canNOT be removed completely; (in this case, the input gap-block configuration remains intact.)
#
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block-coordinate necessary for 'UNDO'ing the overlaps (incompletely if $fl_end = -1)}, ...) for the gap-blocks serially "shift"ed (or intended to be done so when $fl_end = -1) within this subroutine;
#
sub undo_hovlp_vcmpl_gbsets_1sbj ($\@\@\@\@\@\@\@$\@\@\@\@\@) {

	my ($gb_sbj,
	$curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes
	) = @_;

	my $ct_gbs = @{$curr_bds_blks};

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];
	my ($lbd_sbj, $rbd_sbj) = @{$curr_bds_blks->[$gb_sbj]};


   (print STDOUT "\n< Subroutine 'undo_hovlp_vcmpl_gbsets_1sbj' (w/ \$gb_sbj = $gb_sbj, #{gap-blocks}=$ct_gbs, [lbd, rbd]_sbj = [$lbd_sbj, $rbd_sbj]) >\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets > 0); # ADDED on 2019/08/09. #

	##my ($rel_left_w_left, $rel_right_w_left) = ('=', $rel_lr); # A potential BUG fixed on 2019/08/13.
	#my ($rel_left_w_left, $rel_right_w_left) = ('=', 'Cp');

	my @relv_gbs_left = my @relv_gbs_in_btwn = my @relv_gbs_right = (); # element = [$indx_gb, $lbd_gb, $rbd_gb, $rel_w_left]

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Preliminary for-loop (over gap-blocks). #

	    if ($gb == $gb_sbj) { next; } # A BUG Fixed on 2019/08/09. #
	    ##if (($gb == $gb_left) or ($gb == $gb_right)) { next; } # A BUG Fixed on 2019/08/09. #
	    ##unless (($gb == $gb_left) or ($gb == $gb_right)) { next; }

	    my $rel_w_sbj = $rels_w_sbj->[$gb];
	    unless (($rel_w_sbj eq '=') or ($rel_w_sbj eq 'Cp')) { next; }

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_blks->[$gb]};

	    if ($lbd_gb < $lbd_sbj) {
		push @relv_gbs_left, [$gb, $lbd_gb, $rbd_gb, $rel_w_sbj];
	    } else {
		push @relv_gbs_right, [$gb, $lbd_gb, $rbd_gb, $rel_w_sbj];
	    }
=For the predecessor 'undo_hovlp_vcmpl_gbsets'. (1)
	    if ($lbd_gb < $lbd_left) {
		push @relv_gbs_left, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
		next;
	    } elsif ($rbd_right < $rbd_gb) {
		push @relv_gbs_right, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
		next;
	    }

	    if ($rbd_gb <$rbd_left) {
		push @relv_gbs_left, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    } elsif ($lbd_right < $lbd_gb) {
		push @relv_gbs_right, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    } else {
		push @relv_gbs_in_btwn, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    }
=cut # END of "For the predecessor 'undo_hovlp_vcmpl_gbsets'. (1)"

	} # End of the preliminary for-loop (over gap-blocks). #

	if (@relv_gbs_left >1) {
	    my $prprev_rel_w_sbj = '=';
	    my @std_relv_gbs_left = ();
	    foreach my $element (sort {$b->[2] <=> $a->[2]} @relv_gbs_left) { # 2nd preliminary foreach-loop.
		if (@std_relv_gbs_left <1) {
		    unshift @std_relv_gbs_left, $element;
		    next;
		}
		my ($lbd_prev, $rbd_prev, $prev_rel_w_sbj) = @{$std_relv_gbs_left[0]}[1,2,3];
		my ($lbd_curr, $rbd_curr, $curr_rel_w_sbj) = @{$element}[1,2,3];

		if ($rbd_curr < $rbd_prev) {
		    unshift @std_relv_gbs_left, $element;

		} elsif (($prev_rel_w_sbj eq $prprev_rel_w_sbj) and ($curr_rel_w_sbj ne $prprev_rel_w_sbj)) {  # Attempt to alternate the types of gap-blocks as much as possible.
		    my $prev_element = shift @std_relv_gbs_left;
		    unshift @std_relv_gbs_left, $element;
		    unshift @std_relv_gbs_left, $prev_element;

		} else {
		    unshift @std_relv_gbs_left, $element;
		}
		$prprev_rel_w_sbj = $prev_rel_w_sbj;

	    } # End of the 2nd preliminary foreach-loop.

	    @relv_gbs_left = @std_relv_gbs_left;

	} # End of "if (@relv_gbs_left >1) {...}"



	##my $rf_relv_gbs_in_btwn = \@relv_gbs_in_btwn;
	my $rf_relv_gbs_right = \@relv_gbs_right;

	if (@relv_gbs_right >1) {

	    my $prprev_rel_w_sbj = '=';
	    my @std_relv_gbs = ();
	    foreach my $element (sort {$a->[1] <=> $a->[1]} @relv_gbs_right) { # Inner preliminary foreach-loop.
		if (@std_relv_gbs <1) {
		    push @std_relv_gbs, $element;
		    next;
		}
		my ($lbd_prev, $rbd_prev, $prev_rel_w_sbj) = @{$std_relv_gbs[$#std_relv_gbs]}[1,2,3];
		my ($lbd_curr, $rbd_curr, $curr_rel_w_sbj) = @{$element}[1,2,3];

		if ($lbd_prev < $lbd_curr) {
		    push @std_relv_gbs, $element;

		} elsif (($prev_rel_w_sbj eq $prprev_rel_w_sbj) and ($curr_rel_w_sbj ne $prprev_rel_w_sbj)) {  # Attempt to alternate the types of gap-blocks as much as possible.
		    my $prev_element = pop @std_relv_gbs;
		    push @std_relv_gbs, $element;
		    push @std_relv_gbs, $prev_element;

		} else {
		    push @std_relv_gbs, $element;
		}
		$prprev_rel_w_sbj = $prev_rel_w_sbj;

	    } # End of the inner preliminary foreach-loop.

	    @relv_gbs_right = @std_relv_gbs; # Update the set of relevant gap-blocks. #

	} # End of "if (@relv_gbs_right >1) {...}"


	my @relv_gbs = (@relv_gbs_left, [$gb_sbj, $lbd_sbj, $rbd_sbj, '='], @relv_gbs_right); # A BUG Fixed (on 2019/08/13), after the Revision on 2019/08/12. #

=For the predecessor 'undo_hovlp_vcmpl_gbsets'. (2)
	@relv_gbs_in_btwn = @{$rf_relv_gbs_in_btwn};
	@relv_gbs_right = @{$rf_relv_gbs_right};

	my @relv_gbs = (@relv_gbs_left, [$gb_left, $lbd_left, $rbd_left, '='],
		 @relv_gbs_in_btwn, [$gb_right, $lbd_right, $rbd_right, $rel_lr], @relv_gbs_right); # A BUG Fixed (on 2019/08/13), after the Revision on 2019/08/12. #
		 ## @relv_gbs_in_btwn, [$gb_right, $lbd_right, $rbd_right, 'Cp'], @relv_gbs_right);
=cut # END of "For the predecessor 'undo_hovlp_vcmpl_gbsets'. (2)"

	my $ct_relv_gbs = @relv_gbs;
	my $relv_lm_gb = $relv_gbs[0]->[0];
	my $relv_rm_gb = $relv_gbs[$ct_relv_gbs-1]->[0];

   if ($TEST_LEVEL_undo_ho_vcp_sets > 2) { # ADDED in DEBUGGING (on 2019/08/09).

	print STDOUT "\n[ Outputs of the Preliminary Process (#{relevant-blocks}=$ct_relv_gbs): ]\n\n";

	print STDOUT '@relv_gbs_left = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_sbj], ', "\n";
	foreach my $info_gb (@relv_gbs_left) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";

=For the predecessor 'undo_hovlp_vcmpl_gbsets'. (3)
	print STDOUT '@relv_gbs_in_btwn = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_left], ', "\n";
	foreach my $info_gb (@relv_gbs_in_btwn) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";
=cut # END of "For the predecessor 'undo_hovlp_vcmpl_gbsets'. (3)"

	print STDOUT '@relv_gbs_right = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_sbj], ', "\n";
	foreach my $info_gb (@relv_gbs_right) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";
	print STDOUT "\n";

	print STDOUT '@relv_gbs = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_sbj], ', "\n";
	foreach my $info_gb (@relv_gbs) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";
	print STDOUT "\n";

	##print STDOUT '@relv_gbs_left = (', join (', ', @relv_gbs_left), ") ;\n";
	##print STDOUT '@relv_gbs_in_btwn = (', join (', ', @relv_gbs_in_btwn), ") ;\n";
	##print STDOUT '@relv_gbs_right = (', join (', ', @relv_gbs_right), ") .\n\n";
	##print STDOUT '@relv_gbs = (', join (', ', @relv_gbs), ") .\n\n";
	##print STDOUT "\n";

   } # ENd of "if ($TEST_LEVEL_undo_ho_vcp_sets > 2) {...}"

		# MAIN 1: Examine the overlaps between the nearest-neighboring relevant gap-blocks. #
		# & distances between them.

	my @effovlps_distances = (); # $k th element = [$size_ovlp (+1 if the two blocks are of the same type), $dist_left2right (-1 if of the same type), $dist_right2left (-1 if of the same type)] for the $k th and ($k+1) th relevant gap-blocks. #

	my $totsize_effovlp = 0;
	my $tot_max_effdist = 0;

	my @indices_effovlp_pairs = ();
	for (my $k=0; $k <$ct_relv_gbs-1; $k++) { # 1st main  for-loop. #

	    my ($gb1, $lbd1, $rbd1, $rel_w_sbj1) = @{$relv_gbs[$k]};
	    my ($gb2, $lbd2, $rbd2, $rel_w_sbj2) = @{$relv_gbs[$k+1]};

	    my $size_effovlp = $rbd1 - $lbd2 + 1;
	    if ($size_effovlp >= 0) {

		if ($rel_w_sbj1 eq $rel_w_sbj2) { $size_effovlp++; }
		my $effdist = ($size_effovlp >0) ? -1 : 0;
		$effovlps_distances[$k] = [$size_effovlp, $effdist, $effdist];
		push @indices_effovlp_pairs, $k;
		$totsize_effovlp += $size_effovlp;
		next;
	    }

	    my ($effdist_left2right, $effdist_right2left) = compt_inter_gblock_distance ($gb1, $gb2, @{$curr_bds_blks}, @{$inter_block_relations});
	    unless (defined $effdist_left2right) {
		print STDOUT "[ERROR 'undo_hovlp_vcmpl_gbsets_1sbj'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, (\$gb1, \$gb2) = ($gb1, $gb2): $! \n\n";
		return (0, undef);
	    }

	    unless ($effdist_left2right >0) {
		my $size_effovlp = $effdist_left2right;
		if ($rel_w_sbj1 eq $rel_w_sbj2) { $size_effovlp++; }
		my $effdist = ($size_effovlp >0) ? -1 : 0;
		$effovlps_distances[$k] = [$size_effovlp, $effdist, $effdist];
		push @indices_effovlp_pairs, $k;
		$totsize_effovlp += $size_effovlp;
		next;
	    }

	    if ($rel_w_sbj1 eq $rel_w_sbj2) {
		$effdist_left2right--;
		$effdist_right2left--;
	    }
	    $size_effovlp = ($effdist_left2right>0) ? -1 : 0;

	    $effovlps_distances[$k] = [$size_effovlp, $effdist_left2right, $effdist_right2left];

	    my $max_effdist = ($effdist_left2right > $effdist_right2left) ? $effdist_left2right : $effdist_right2left ;
	    $tot_max_effdist += $max_effdist;

	} # End of the 1st main for-loop. #



		# MOVED from near the top of MAIN 2 (on 2019/08/09). #
	my %gb2incr_bl_coord_undo;
	foreach my $info_rlv_gb (@relv_gbs) { $gb2incr_bl_coord_undo{$info_rlv_gb->[0]} = 0; }
		# END of "MOVED from near the top of MAIN 2 (on 2019/08/09)." #


	my $rightmost_margin = $bds_gb_coords->[$relv_rm_gb][1] - $curr_bl_coords->[$relv_rm_gb];
	my $leftmost_margin = $bds_gb_coords->[$relv_lm_gb][0] - $curr_bl_coords->[$relv_lm_gb];
	my $abs_leftmost_margin = - $leftmost_margin; # ADDED on 2019/08/09. #
		# $curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,

	my $ext_tot_max_effdist = $tot_max_effdist + $rightmost_margin + $abs_leftmost_margin; # MODIFIED on 2019/08/09.
	##my $ext_tot_max_effdist = $tot_max_effdist + $right_margin - $left_margin;


    if ($TEST_LEVEL_undo_ho_vcp_sets >2) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "[[ Interim Results after the 1st MAIN for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '$totsize_effovlp = ', $totsize_effovlp, " .\n\n";
	print STDOUT '$tot_max_effdist = ', $tot_max_effdist, " ,\n";
	print STDOUT '($rightmost_margin, $leftmost_margin ) = ', "($rightmost_margin, $leftmost_margin ) ,\n\n";
	print STDOUT '$ext_tot_max_effdist (=$tot_max_effdist + $rightmost_margin - $leftmost_margin)', "\n";
	print STDOUT "                      = $ext_tot_max_effdist  .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >2) {...}"


	unless ($totsize_effovlp <= $ext_tot_max_effdist) {
	    print STDOUT "[CAUTION 'undo_hovlp_vcmpl_gbsets_1sbj'] The hirozontal overlpa will be UNABLE to be UNDONE!!, because \$totsize_effovlp (=$totsize_effovlp) is LARGER THAN \$tot_max_effdist + \$rightmost_margin - \$leftmost_margin (= $ext_tot_max_effdist): $! !\n\n";

	    return (-1, \%gb2incr_bl_coord_undo); # MODIFIED on 2019/08/09. #
	    ##return (-1, undef);
	}

	if ($totsize_effovlp == 0) {
   (print STDOUT "[INFORMATION 'undo_hovlp_vcmpl_gbsets_1sbj'] NO EFFECTIVE OVERLAPS were observed between the nearest-neighboring relevant gap-blocks. ... Thus, I will End the subroutine withOUT \"shift\"ing any gap-blocks vertically equivalent or complementary to \$gb_sbj (=$gb_sbj) ...\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >0);

	    return (+1, \%gb2incr_bl_coord_undo);
	}

	my $ct_effovlp_pairs = @indices_effovlp_pairs;

	##my @indices_effovlp_pairs = ();
	##my @effovlps_distances = (); # $k th element = [$size_ovlp (+1 if the two blocks are of the same type), $dist_left2right (-1 if of the same type), $dist_right2left (-1 if of the same type)] for the $k th and ($k+1) th gap-blocks. #

			### MAIN 2: Virtually "UNDO" the Overlaps. ###

	my @left_margins = my @right_margins = ();
	for (my $gb=0; $gb <$ct_gbs; $gb++) {
	    my ($lbd_coord, $rbd_coord) = @{$bds_gb_coords->[$gb]};
	    my $curr_coord = $curr_bl_coords->[$gb];
	    $left_margins[$gb] = $lbd_coord - $curr_coord;
	    $right_margins[$gb] = $rbd_coord - $curr_coord;
	}
		# $curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,

    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT '[ Possible Ranges of "Shift"s of Each Gap-Block: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'left_margin', 'right_margin'), "\n\n";
	for (my $gb=0; $gb <$ct_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $left_margins[$gb], $right_margins[$gb]), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) { ...}"


			## Main 2-(i)  By "shift"ing to the right.

=MOVED to above the 1st "return (-1, ...)".
	my %gb2incr_bl_coord_undo;
	foreach my $info_rlv_gb (@relv_gbs) { $gb2incr_bl_coord_undo{$info_rlv_gb->[0]} = 0; }
=cut # END of "MOVED to above the 1st "return (-1, ...)"."

	for (my $k = $ct_effovlp_pairs-1; $k>=0; $k--) { # 2nd main for-loop.

		my $indx_curr_pair = $indices_effovlp_pairs[$k];
		my $indx_next_pair = ($#indices_effovlp_pairs > $k) ? $indices_effovlp_pairs[$k+1] : undef;

		my $size_effovlp = $effovlps_distances[$indx_curr_pair]->[0];
		##my ($size_effovlp, $effdist_left2right, $effdist_right2left) = @{$effovlps_distances[$indx_curr_pair]};
		my $size_effovlp_old = $size_effovlp;

		my $indx_rlv_gb_start = $indx_curr_pair +1;
		my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair -1 : $ct_relv_gbs -1;

		for (my $i1 = $indx_rlv_gb_start; $i1 <= $indx_rlv_gb_end; $i1++) { # Middle for-loop.
		    my ($gb1, $lbd1, $rbd1, $rel_w_left1) = @{$relv_gbs[$i1]};
		    my $effovlp_distance1 = ($i1 < $ct_relv_gbs-1) ? $effovlps_distances[$i1] : undef;
		    my $effdist_left2right = (defined $effovlp_distance1) ? $effovlp_distance1->[1] : $rightmost_margin;
		    unless ($effdist_left2right>0) { next; }
		    my $max_shift = ($size_effovlp < $effdist_left2right) ? $size_effovlp : $effdist_left2right;

		    for (my $i2 = $indx_rlv_gb_start; $i2 <= $i1; $i2++) { # 1st Inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    my $prev_incr_coord = $gb2incr_bl_coord_undo{$gb2};
			    if ($prev_incr_coord + $max_shift > $right_margins[$gb2]) {
				$max_shift = $right_margins[$gb2] - $prev_incr_coord;
			    }
		    } # End of the 1st inner for-loop. #
		    unless ($max_shift >0) { next; }

		    if ($i1 < $ct_relv_gbs-1) { # ADDED the condition on 2019/08/10.
			$effovlp_distance1->[1] -= $max_shift;
			$effovlp_distance1->[2] -= $max_shift;
			if ($max_shift == $effdist_left2right) { $effovlp_distance1->[0] = 0; }

		    } else { # ADDED on 2019/08/10.
			$rightmost_margin -= $max_shift;
		    }

		    for (my $i2 = $indx_rlv_gb_start; $i2 <= $i1; $i2++) {  # 2nd inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    $gb2incr_bl_coord_undo{$gb2} += $max_shift;
		    } # End of the 2nd inner for-loop. #

		    $size_effovlp -= $max_shift;

		    if ($size_effovlp == 0) { last; } # End the inspection of the pairs if the overlap of the subject pair is completely removed. #

		} # End of the middle for-loop. #

		if ($size_effovlp == 0) {

		    $effovlps_distances[$indx_curr_pair] = [0, 0, 0];

			# Remove the subject pair from the list of effectively ovelapping pairs. #

		     @indices_effovlp_pairs = @indices_effovlp_pairs[0 .. $k-1, $k+1 .. $ct_effovlp_pairs-1];
		     $ct_effovlp_pairs--;

		} else {
				# Just Update the size of the effective-overlap in the subject pair. #

		    $effovlps_distances[$indx_curr_pair]->[0] = $size_effovlp;
		}

	} # End of the 2nd main for-loop.


    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Interim Results after the 2nd Main for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '%gb2incr_bl_coord_undo = ( # $gb => $incr_bl_coord_undo, ', "\n";

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) {...}"


			## Main 2-(ii) By "shift"ing to the left.

    (print STDOUT "\n{{{  Start of the 3rd main for-loop.  }}}\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >3); # ADDED for DEBUGGING (on 2019/08/10). #

	for (my $k=0; $k < scalar (@indices_effovlp_pairs); $k++) { # 3rd main for-loop.

		my $indx_curr_pair = $indices_effovlp_pairs[$k];
		my $indx_next_pair = ($k > 0) ? $indices_effovlp_pairs[$k-1] : undef;

		my $size_effovlp = $effovlps_distances[$indx_curr_pair]->[0]; # REVISED on 2019/08/10. #
		##my ($size_effovlp, $effdist_left2right, $effdist_right2left) = @{$effovlps_distances[$indx_curr_pair]};
		my $size_effovlp_old = $size_effovlp;

		my $indx_rlv_gb_start = $indx_curr_pair;
		my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair +2 : 0; # A BUG fixed on 2019/08/10.
		##my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair +1 : 0;

    (print STDOUT "\n{{  \$k = $k  (w/ \$indx_curr_pair = $indx_curr_pair, \$size_effovlp = $size_effovlp) }}\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets>=5); # ADDED for DEBUGGING (on 2019/08/10). #

    if ($TEST_LEVEL_undo_ho_vcp_sets >5) { # ADDED for DEBUGGING (on 2019/08/10). #

	print STDOUT '$indx_next_pair = ', ((defined $indx_next_pair) ? $indx_next_pair : 'undef'), " .\n\n";
	print STDOUT '($indx_rlv_gb_start, $indx_rlv_gb_end) = ', "($indx_rlv_gb_start, $indx_rlv_gb_end) .\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >5) {...}"

		for (my $i1 = $indx_rlv_gb_start; $i1 >= $indx_rlv_gb_end; $i1--) { # Middle for-loop.

		    my ($gb1, $lbd1, $rbd1, $rel_w_left1) = @{$relv_gbs[$i1]};


		    my $effovlp_distance1 = ($i1 > 0) ? $effovlps_distances[$i1-1] : undef;

		    my $effdist_right2left = (defined $effovlp_distance1) ? $effovlp_distance1->[2] : - $leftmost_margin; # A BUG fixed on 2019/08/10. #
		    ##my $effdist_righ2left = (defined $effovlp_distance1) ? $effovlp_distance1->[2] : - $leftmost_margin;
    (print STDOUT "\n{ \$i1 = $i1   (w/ \$gb1 = $gb1, \$effdist_right2left = $effdist_right2left) }\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    unless ($effdist_right2left>0) { next; }

		    my $max_shift = ($size_effovlp < $effdist_right2left) ? $size_effovlp : $effdist_right2left;

    (print STDOUT "Before the 1st inner for-loop: \$max_shift = $max_shift .\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    for (my $i2 = $indx_rlv_gb_start; $i2 >= $i1; $i2--) { # 1st Inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    my $prev_incr_coord = $gb2incr_bl_coord_undo{$gb2};
			    if ($prev_incr_coord - $max_shift < $left_margins[$gb2]) {
				$max_shift = $prev_incr_coord - $left_margins[$gb2];
			    }
		    } # End of the 1st inner for-loop. #

    (print STDOUT "After the 1st inner for-loop: \$max_shift = $max_shift .\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    unless ($max_shift >0) { next; }

		    if ($i1>0) { # THe condition ADDED on 2019/08/10.
			$effovlp_distance1->[1] -= $max_shift;
			$effovlp_distance1->[2] -= $max_shift;
			if ($max_shift == $effdist_right2left) { $effovlp_distance1->[0] = 0; }

		    } else { # ADDED on 2019/08/10.
			$leftmost_margin -= $max_shift;
		    }

		    for (my $i2 = $indx_rlv_gb_start; $i2 >= $i1; $i2--) {  # 2nd inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    $gb2incr_bl_coord_undo{$gb2} -= $max_shift;
		    } # End of the 2nd inner for-loop. #

		    $size_effovlp -= $max_shift;

    (print STDOUT "At the end of (\$i1 = $i1) of the middle for-loop: \$size_effovlp = $size_effovlp.\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). 

		    if ($size_effovlp == 0) { last; } # End the inspection of the pairs if the overlap of the subject pair is completely removed. #

		} # End of the middle for-loop. #

		if ($size_effovlp == 0) {

		    $effovlps_distances[$indx_curr_pair] = [0, 0, 0];

			# Remove the subject pair from the list of effectively ovelapping pairs. #

		     @indices_effovlp_pairs = @indices_effovlp_pairs[0 .. $k-1, $k+1 .. $ct_effovlp_pairs-1];
		     $ct_effovlp_pairs--;
		     $k--; # To re-use the current value of $k in the next round (of the for-loop). #

		} else {
				# Just Update the size of the effective-overlap in the subject pair. #

		    $effovlps_distances[$indx_curr_pair]->[0] = $size_effovlp;
		}

	} # End of the 3rd main for-loop. #

    if ($TEST_LEVEL_undo_ho_vcp_sets >2) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Interim Results after the 3rd Main for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '%gb2incr_bl_coord_undo = ( # $gb => $incr_bl_coord_undo, ', "\n";

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_undo_ho_vcp_sets >2) {...}"


	unless ($ct_effovlp_pairs == 0) {
	    print STDOUT "[CAUTION 'undo_hovlp_vcmpl_gbsets_1sbj'] The horizontal overlap will be UNABLE to be UNDONE!!, because ";
	    print STDOUT  '@indices_effovlp_pairs = (', join (',', @indices_effovlp_pairs), ')';
	    print STDOUT " COULD NOT be EMPTIED (w/ $ct_effovlp_pairs elements left): $! !\n\n";

	    return (-1, \%gb2incr_bl_coord_undo);
	}


			### MAIN 3: Actually "UNDO" the Overlaps, via 'serial_shifts_point2point'.. ###

	my @bl_coords_frm = @{$curr_bl_coords};


	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {

			# Update the gap-block coordinates, prior to performing 'serial_shifts_point2point'. #
		$curr_bl_coords->[$gb] += $gb2incr_bl_coord_undo{$gb};
	}

	my @bl_coords_to =  @{$curr_bl_coords};


    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Inputs of 'serial_shifts_point2point' Prepared Here: ]]\n\n";

	print STDOUT '@bl_coords_frm = (', join (', ',  @bl_coords_frm), ") ;\n\n";
	print STDOUT '@bl_coords_to  = (', join (', ', @bl_coords_to), ") .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) {...}"


	my ($fl_ssh_p2p, $tot_incr_ct_null_clms)
	   = serial_shifts_point2point (
		@{$curr_set_clms}, @{$curr_bds_blks}, @bl_coords_frm, @bl_coords_to, @{$bds_gb_coords}, 
		@{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $fl_ssh_p2p) and ($fl_ssh_p2p == 0)) {
	    my $fl_out = (defined $fl_ssh_p2p) ? $fl_ssh_p2p : 'undef';
	    print STDOUT "[ERROR 'undo_hovlp_vcmpl_gbsets_1sbj'] The subroutine 'serial_shifts_point2point' FAILED (w/ \$fl_ssh_p2p = $fl_out): $! !\n\n";
	    return (0, undef);
	}
    (print STDOUT "\n ... The subroutine 'serial_shifts_point2point' Ended (w/ \$fl_ssh_p2p = $fl_ssh_p2p, \$tot_incr_ct_null_clms = $tot_incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >1); # Added on 2019/08/09. #

   (print STDOUT "\n< Subroutine 'undo_hovlp_vcmpl_gbsets_1sbj' Ended (w/ \$flag_undo = 1) />\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets > 0); # ADDED on 2019/08/09. #


	return (1, \%gb2incr_bl_coord_undo);

}	# End of the subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj'. # 


# undo_hovlp_vcmpl_gbsets:
#	This satellite subroutine of 'cmerge_split_grow_gblock_pair'
#	UNDOes the overlaps between vertically equivalent and/or complementary gap-blocks
#	resulting from the main process of \"merge + split\" of a pair of gap-blocks.
#	[CREATED on 2019/08/08.]
#	[DEBUGGED on 2019/08/09-10]:
#	A BUG was found in the 3rd main for-loop;
#	it prevented the "undo"ing of the overlap(s) via the "shift"s to the left.
#	Then, the BUG was fixed.
#	[REVISED on 2019/08/12]:
#	Now, the subroutine can also accept gap-block pairs that are vertical identical to each other.
#
#	[NOTE added on 2019/08/23]:
#	It seems that the current version sometimes "overshoot"s,
#	especially when some vertically identical blocks are nested.
#	In the near future, this problem must be rectified.
#
#	[STRATEGY]
#	(1) First, attempt to undo the overlap(s) by moving the gap-blocks on the right of the overlap(s) further to the left;
#	(2), If the overlap(s) still remain(s), move the gap-blocks on the left of the overlap(s) further to the right;
#	(3) If the overlap(s) still remain(s), return -1.
#
# INPUTS: ($gb_left, $gb_right,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_bds_gb_coords, \@new_curr_bl_coords,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_blks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@new_gblock_sizes
#	), where
#
#	$gb_left and $gb_right are the indices of the gap-blocks on the left and on the right, respectively;
#	They MUST be vertically complementary to each other;
#
#	Other input arguments are the same as those used for 'serial_shifts_point2point',
#	which will be invoked within this subroutine;
#	[For details, see the header of 'serial_shifts_point2point'];
#
# OUTPUTS: ($fl_end, \%gb2incr_bl_coord_undo), where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed,
#		= -1 if the overlap(s) canNOT be removed completely; (in this case, the input gap-block configuration remains intact.)
#
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block-coordinate necessary for 'UNDO'ing the overlaps (incompletely if $fl_end = -1)}, ...) for the gap-blocks serially "shift"ed (or intended to be done so when $fl_end = -1) within this subroutine;
#
sub undo_hovlp_vcmpl_gbsets ($$\@\@\@\@\@\@\@$\@\@\@\@\@) {

	my ($gb_left, $gb_right,
	$curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes
	) = @_;

	my $rel_lr = $inter_block_relations->[$gb_left][$gb_right];

	unless (($rel_lr eq 'Cp') or ($rel_lr eq '=')) { # MODIFIED in REVISION (on 2019/08/12). #
	##unless ($rel_lr eq 'Cp') {
	    print STDOUT "[ERROR 'undo_hovlp_vcmpl_gbsets'] The \$gb_left (=$gb_left) and \$gb_right (=$gb_right) are (vertically) NEITHER COMPLEMENTARY NOR IDENTICAL to each other (, with \$rel_lr = '$rel_lr'): $! !\n\n";
	    return (0, undef);
	}
	unless ($curr_bds_blks->[$gb_left][0] < $curr_bds_blks->[$gb_right][0]) {
	    print STDOUT "[WARNING 'undo_hovlp_vcmpl_gbsets'] The \$gb_left (=$gb_left) is NOT on the LEFT of \$gb_right (=$gb_right)!!\n";
	    my $tmp = $gb_left;
	    $gb_left = $gb_right;
	    $gb_right = $tmp;

	    print STDOUT " ... So, I SWAPPED the INDICES, so that (\$gb_left, \$gb_right) = ($gb_left, $gb_right) ...\n\n";
	}

	my $ct_gbs = @{$curr_bds_blks};

	my $rels_w_left = $inter_block_relations->[$gb_left];
	my ($lbd_left, $rbd_left) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right, $rbd_right) = @{$curr_bds_blks->[$gb_right]};


   (print STDOUT "\n< Subroutine 'undo_hovlp_vcmpl_gbsets' (w/ (\$gb_left, \$gb_right) = ($gb_left, $gb_right), #{gap-blocks}=$ct_gbs, [lbd, rbd]_left = [$lbd_left, $rbd_left], [lbd, rbd]_right = [$lbd_right, $rbd_right]) >\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets > 0); # ADDED on 2019/08/09. #

	my ($rel_left_w_left, $rel_right_w_left) = ('=', $rel_lr); # A potential BUG fixed on 2019/08/13.
	#my ($rel_left_w_left, $rel_right_w_left) = ('=', 'Cp');

	my @relv_gbs_left = my @relv_gbs_in_btwn = my @relv_gbs_right = (); # element = [$indx_gb, $lbd_gb, $rbd_gb, $rel_w_left]

	for (my $gb=0; $gb <$ct_gbs; $gb++) { # Preliminary for-loop (over gap-blocks). #

	    if (($gb == $gb_left) or ($gb == $gb_right)) { next; } # A BUG Fixed on 2019/08/09. #
	    ##unless (($gb == $gb_left) or ($gb == $gb_right)) { next; }

	    my $rel_w_left = $rels_w_left->[$gb];
	    unless (($rel_w_left eq '=') or ($rel_w_left eq 'Cp')) { next; }

	    my ($lbd_gb, $rbd_gb) = @{$curr_bds_blks->[$gb]};

	    if ($lbd_gb < $lbd_left) {
		push @relv_gbs_left, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
		next;
	    } elsif ($rbd_right < $rbd_gb) {
		push @relv_gbs_right, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
		next;
	    }

	    if ($rbd_gb <$rbd_left) {
		push @relv_gbs_left, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    } elsif ($lbd_right < $lbd_gb) {
		push @relv_gbs_right, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    } else {
		push @relv_gbs_in_btwn, [$gb, $lbd_gb, $rbd_gb, $rel_w_left];
	    }

	} # End of the preliminary for-loop (over gap-blocks). #

	if (@relv_gbs_left >1) {
	    my $prprev_rel_w_left = '=';
	    my @std_relv_gbs_left = ();
	    foreach my $element (sort {$b->[2] <=> $a->[2]} @relv_gbs_left) { # 2nd preliminary foreach-loop.
		if (@std_relv_gbs_left <1) {
		    unshift @std_relv_gbs_left, $element;
		    next;
		}
		my ($lbd_prev, $rbd_prev, $prev_rel_w_left) = @{$std_relv_gbs_left[0]}[1,2,3];
		my ($lbd_curr, $rbd_curr, $curr_rel_w_left) = @{$element}[1,2,3];

		if ($rbd_curr < $rbd_prev) {
		    unshift @std_relv_gbs_left, $element;

		} elsif (($prev_rel_w_left eq $prprev_rel_w_left) and ($curr_rel_w_left ne $prprev_rel_w_left)) {  # Attempt to alternate the types of gap-blocks as much as possible.
		    my $prev_element = shift @std_relv_gbs_left;
		    unshift @std_relv_gbs_left, $element;
		    unshift @std_relv_gbs_left, $prev_element;

		} else {
		    unshift @std_relv_gbs_left, $element;
		}
		$prprev_rel_w_left = $prev_rel_w_left;

	    } # End of the 2nd preliminary foreach-loop.

	    @relv_gbs_left = @std_relv_gbs_left;

	} # End of "if (@relv_gbs_left >1) {...}"



	my $rf_relv_gbs_in_btwn = \@relv_gbs_in_btwn;
	my $rf_relv_gbs_right = \@relv_gbs_right;

     foreach my $info ([$rf_relv_gbs_in_btwn, '='], [$rf_relv_gbs_right, $rel_lr]) { # Outer preliminary foreach-loop. # A BUG Fixed (on 2019/08/13), after the Revision on 2019/08/12. #
     ##foreach my $info ([$rf_relv_gbs_in_btwn, '='], [$rf_relv_gbs_right, 'Cp']) { # Outer preliminary foreach-loop.
	my ($rf_relv_gbs, $prprev_rel_w_left) = @{$info};

	if (@{$rf_relv_gbs} >1) {
	    my @std_relv_gbs = ();
	    foreach my $element (sort {$a->[1] <=> $a->[1]} @{$rf_relv_gbs}) { # Inner preliminary foreach-loop.
		if (@std_relv_gbs <1) {
		    push @std_relv_gbs, $element;
		    next;
		}
		my ($lbd_prev, $rbd_prev, $prev_rel_w_left) = @{$std_relv_gbs[$#std_relv_gbs]}[1,2,3];
		my ($lbd_curr, $rbd_curr, $curr_rel_w_left) = @{$element}[1,2,3];

		if ($lbd_prev < $lbd_curr) {
		    push @std_relv_gbs, $element;

		} elsif (($prev_rel_w_left eq $prprev_rel_w_left) and ($curr_rel_w_left ne $prprev_rel_w_left)) {  # Attempt to alternate the types of gap-blocks as much as possible.
		    my $prev_element = pop @std_relv_gbs;
		    push @std_relv_gbs, $element;
		    push @std_relv_gbs, $prev_element;

		} else {
		    push @std_relv_gbs, $element;
		}
		$prprev_rel_w_left = $prev_rel_w_left;

	    } # End of the inner preliminary foreach-loop.

	    @{$rf_relv_gbs} = @std_relv_gbs; # Update the set of relevant gap-blocks. #
	    ##$rf_relv_gbs = \@std_relv_gbs;

	} # End of "if (@{$rf_relv_gbs} >1) {...}"

     } # End of the outer preliminary foreach-loop.

	@relv_gbs_in_btwn = @{$rf_relv_gbs_in_btwn};
	@relv_gbs_right = @{$rf_relv_gbs_right};

	my @relv_gbs = (@relv_gbs_left, [$gb_left, $lbd_left, $rbd_left, '='],
		 @relv_gbs_in_btwn, [$gb_right, $lbd_right, $rbd_right, $rel_lr], @relv_gbs_right); # A BUG Fixed (on 2019/08/13), after the Revision on 2019/08/12. #
		 ## @relv_gbs_in_btwn, [$gb_right, $lbd_right, $rbd_right, 'Cp'], @relv_gbs_right);

	my $ct_relv_gbs = @relv_gbs;
	my $relv_lm_gb = $relv_gbs[0]->[0];
	my $relv_rm_gb = $relv_gbs[$ct_relv_gbs-1]->[0];

   if ($TEST_LEVEL_undo_ho_vcp_sets > 2) { # ADDED in DEBUGGING (on 2019/08/09).

	print STDOUT "\n[ Outputs of the Preliminary Process (#{relevant-blocks}=$ct_relv_gbs): ]\n\n";

	print STDOUT '@relv_gbs_left = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_left], ', "\n";
	foreach my $info_gb (@relv_gbs_left) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";

	print STDOUT '@relv_gbs_in_btwn = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_left], ', "\n";
	foreach my $info_gb (@relv_gbs_in_btwn) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";

	print STDOUT '@relv_gbs_right = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_left], ', "\n";
	foreach my $info_gb (@relv_gbs_right) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";
	print STDOUT "\n";

	print STDOUT '@relv_gbs = ( # element = [$indx_gb, $lbd, $rbd, $rel_w_left], ', "\n";
	foreach my $info_gb (@relv_gbs) { print STDOUT "\t[", join (', ', @{$info_gb}), "],\n"; }
	print STDOUT "\t\t) ;\n\n";
	print STDOUT "\n";

	##print STDOUT '@relv_gbs_left = (', join (', ', @relv_gbs_left), ") ;\n";
	##print STDOUT '@relv_gbs_in_btwn = (', join (', ', @relv_gbs_in_btwn), ") ;\n";
	##print STDOUT '@relv_gbs_right = (', join (', ', @relv_gbs_right), ") .\n\n";
	##print STDOUT '@relv_gbs = (', join (', ', @relv_gbs), ") .\n\n";
	##print STDOUT "\n";

   } # ENd of "if ($TEST_LEVEL_undo_ho_vcp_sets > 2) {...}"

		# MAIN 1: Examine the overlaps between the nearest-neighboring relevant gap-blocks. #
		# & distances between them.

	my @effovlps_distances = (); # $k th element = [$size_ovlp (+1 if the two blocks are of the same type), $dist_left2right (-1 if of the same type), $dist_right2left (-1 if of the same type)] for the $k th and ($k+1) th relevant gap-blocks. #

	my $totsize_effovlp = 0;
	my $tot_max_effdist = 0;

	my @indices_effovlp_pairs = ();
	for (my $k=0; $k <$ct_relv_gbs-1; $k++) { # 1st main  for-loop. #

	    my ($gb1, $lbd1, $rbd1, $rel_w_left1) = @{$relv_gbs[$k]};
	    my ($gb2, $lbd2, $rbd2, $rel_w_left2) = @{$relv_gbs[$k+1]};

	    my $size_effovlp = $rbd1 - $lbd2 + 1;
	    if ($size_effovlp >= 0) {

		if ($rel_w_left1 eq $rel_w_left2) { $size_effovlp++; }
		my $effdist = ($size_effovlp >0) ? -1 : 0;
		$effovlps_distances[$k] = [$size_effovlp, $effdist, $effdist];
		push @indices_effovlp_pairs, $k;
		$totsize_effovlp += $size_effovlp;
		next;
	    }

	    my ($effdist_left2right, $effdist_right2left) = compt_inter_gblock_distance ($gb1, $gb2, @{$curr_bds_blks}, @{$inter_block_relations});
	    unless (defined $effdist_left2right) {
		print STDOUT "[ERROR 'undo_hovlp_vcmpl_gbsets'] The subroutine 'compt_inter_gblock_distance' FAILED for the gap-block pair, (\$gb1, \$gb2) = ($gb1, $gb2): $! \n\n";
		return (0, undef);
	    }

	    unless ($effdist_left2right >0) {
		my $size_effovlp = $effdist_left2right;
		if ($rel_w_left1 eq $rel_w_left2) { $size_effovlp++; }
		my $effdist = ($size_effovlp >0) ? -1 : 0;
		$effovlps_distances[$k] = [$size_effovlp, $effdist, $effdist];
		push @indices_effovlp_pairs, $k;
		$totsize_effovlp += $size_effovlp;
		next;
	    }

	    if ($rel_w_left1 eq $rel_w_left2) {
		$effdist_left2right--;
		$effdist_right2left--;
	    }
	    $size_effovlp = ($effdist_left2right>0) ? -1 : 0;

	    $effovlps_distances[$k] = [$size_effovlp, $effdist_left2right, $effdist_right2left];

	    my $max_effdist = ($effdist_left2right > $effdist_right2left) ? $effdist_left2right : $effdist_right2left ;
	    $tot_max_effdist += $max_effdist;

	} # End of the 1st main for-loop. #



		# MOVED from near the top of MAIN 2 (on 2019/08/09). #
	my %gb2incr_bl_coord_undo;
	foreach my $info_rlv_gb (@relv_gbs) { $gb2incr_bl_coord_undo{$info_rlv_gb->[0]} = 0; }
		# END of "MOVED from near the top of MAIN 2 (on 2019/08/09)." #


	my $rightmost_margin = $bds_gb_coords->[$relv_rm_gb][1] - $curr_bl_coords->[$relv_rm_gb];
	my $leftmost_margin = $bds_gb_coords->[$relv_lm_gb][0] - $curr_bl_coords->[$relv_lm_gb];
	my $abs_leftmost_margin = - $leftmost_margin; # ADDED on 2019/08/09. #
		# $curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,

	my $ext_tot_max_effdist = $tot_max_effdist + $rightmost_margin + $abs_leftmost_margin; # MODIFIED on 2019/08/09.
	##my $ext_tot_max_effdist = $tot_max_effdist + $right_margin - $left_margin;


    if ($TEST_LEVEL_undo_ho_vcp_sets >2) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "[[ Interim Results after the 1st MAIN for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '$totsize_effovlp = ', $totsize_effovlp, " .\n\n";
	print STDOUT '$tot_max_effdist = ', $tot_max_effdist, " ,\n";
	print STDOUT '($rightmost_margin, $leftmost_margin ) = ', "($rightmost_margin, $leftmost_margin ) ,\n\n";
	print STDOUT '$ext_tot_max_effdist (=$tot_max_effdist + $rightmost_margin - $leftmost_margin)', "\n";
	print STDOUT "                      = $ext_tot_max_effdist  .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >2) {...}"


	unless ($totsize_effovlp <= $ext_tot_max_effdist) {
	    print STDOUT "[CAUTION 'undo_hovlp_vcmpl_gbsets'] The hirozontal overlpa will be UNABLE to be UNDONE!!, because \$totsize_effovlp (=$totsize_effovlp) is LARGER THAN \$tot_max_effdist + \$rightmost_margin - \$leftmost_margin (= $ext_tot_max_effdist): $! !\n\n";

	    return (-1, \%gb2incr_bl_coord_undo); # MODIFIED on 2019/08/09. #
	    ##return (-1, undef);
	}

	if ($totsize_effovlp == 0) {
   (print STDOUT "[INFORMATION 'undo_hovlp_vcmpl_gbsets'] NO EFFECTIVE OVERLAPS were observed between the nearest-neighboring relevant gap-blocks. ... Thus, I will End the subroutine withOUT \"shift\"ing any gap-blocks ...\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >0);

	    return (+1, \%gb2incr_bl_coord_undo);
	}

	my $ct_effovlp_pairs = @indices_effovlp_pairs;

	##my @indices_effovlp_pairs = ();
	##my @effovlps_distances = (); # $k th element = [$size_ovlp (+1 if the two blocks are of the same type), $dist_left2right (-1 if of the same type), $dist_right2left (-1 if of the same type)] for the $k th and ($k+1) th gap-blocks. #

			### MAIN 2: Virtually "UNDO" the Overlaps. ###

	my @left_margins = my @right_margins = ();
	for (my $gb=0; $gb <$ct_gbs; $gb++) {
	    my ($lbd_coord, $rbd_coord) = @{$bds_gb_coords->[$gb]};
	    my $curr_coord = $curr_bl_coords->[$gb];
	    $left_margins[$gb] = $lbd_coord - $curr_coord;
	    $right_margins[$gb] = $rbd_coord - $curr_coord;
	}
		# $curr_set_clms, $curr_bds_blks, $bds_gb_coords, $curr_bl_coords,

    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT '[ Possible Ranges of "Shift"s of Each Gap-Block: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', 'left_margin', 'right_margin'), "\n\n";
	for (my $gb=0; $gb <$ct_gbs; $gb++) {
	    print STDOUT join ("\t", $gb, $left_margins[$gb], $right_margins[$gb]), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) { ...}"


			## Main 2-(i)  By "shift"ing to the right.

=MOVED to above the 1st "return (-1, ...)".
	my %gb2incr_bl_coord_undo;
	foreach my $info_rlv_gb (@relv_gbs) { $gb2incr_bl_coord_undo{$info_rlv_gb->[0]} = 0; }
=cut # END of "MOVED to above the 1st "return (-1, ...)"."

	for (my $k = $ct_effovlp_pairs-1; $k>=0; $k--) { # 2nd main for-loop.

		my $indx_curr_pair = $indices_effovlp_pairs[$k];
		my $indx_next_pair = ($#indices_effovlp_pairs > $k) ? $indices_effovlp_pairs[$k+1] : undef;

		my $size_effovlp = $effovlps_distances[$indx_curr_pair]->[0];
		##my ($size_effovlp, $effdist_left2right, $effdist_right2left) = @{$effovlps_distances[$indx_curr_pair]};
		my $size_effovlp_old = $size_effovlp;

		my $indx_rlv_gb_start = $indx_curr_pair +1;
		my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair -1 : $ct_relv_gbs -1;

		for (my $i1 = $indx_rlv_gb_start; $i1 <= $indx_rlv_gb_end; $i1++) { # Middle for-loop.
		    my ($gb1, $lbd1, $rbd1, $rel_w_left1) = @{$relv_gbs[$i1]};
		    my $effovlp_distance1 = ($i1 < $ct_relv_gbs-1) ? $effovlps_distances[$i1] : undef;
		    my $effdist_left2right = (defined $effovlp_distance1) ? $effovlp_distance1->[1] : $rightmost_margin;
		    unless ($effdist_left2right>0) { next; }
		    my $max_shift = ($size_effovlp < $effdist_left2right) ? $size_effovlp : $effdist_left2right;

		    for (my $i2 = $indx_rlv_gb_start; $i2 <= $i1; $i2++) { # 1st Inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    my $prev_incr_coord = $gb2incr_bl_coord_undo{$gb2};
			    if ($prev_incr_coord + $max_shift > $right_margins[$gb2]) {
				$max_shift = $right_margins[$gb2] - $prev_incr_coord;
			    }
		    } # End of the 1st inner for-loop. #
		    unless ($max_shift >0) { next; }

		    if ($i1 < $ct_relv_gbs-1) { # ADDED the condition on 2019/08/10.
			$effovlp_distance1->[1] -= $max_shift;
			$effovlp_distance1->[2] -= $max_shift;
			if ($max_shift == $effdist_left2right) { $effovlp_distance1->[0] = 0; }

		    } else { # ADDED on 2019/08/10.
			$rightmost_margin -= $max_shift;
		    }

		    for (my $i2 = $indx_rlv_gb_start; $i2 <= $i1; $i2++) {  # 2nd inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    $gb2incr_bl_coord_undo{$gb2} += $max_shift;
		    } # End of the 2nd inner for-loop. #

		    $size_effovlp -= $max_shift;

		    if ($size_effovlp == 0) { last; } # End the inspection of the pairs if the overlap of the subject pair is completely removed. #

		} # End of the middle for-loop. #

		if ($size_effovlp == 0) {

		    $effovlps_distances[$indx_curr_pair] = [0, 0, 0];

			# Remove the subject pair from the list of effectively ovelapping pairs. #

		     @indices_effovlp_pairs = @indices_effovlp_pairs[0 .. $k-1, $k+1 .. $ct_effovlp_pairs-1];
		     $ct_effovlp_pairs--;

		} else {
				# Just Update the size of the effective-overlap in the subject pair. #

		    $effovlps_distances[$indx_curr_pair]->[0] = $size_effovlp;
		}

	} # End of the 2nd main for-loop.


    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Interim Results after the 2nd Main for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '%gb2incr_bl_coord_undo = ( # $gb => $incr_bl_coord_undo, ', "\n";

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) {...}"


			## Main 2-(ii) By "shift"ing to the left.

    (print STDOUT "\n{{{  Start of the 3rd main for-loop.  }}}\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >3); # ADDED for DEBUGGING (on 2019/08/10). #

	for (my $k=0; $k < scalar (@indices_effovlp_pairs); $k++) { # 3rd main for-loop.

		my $indx_curr_pair = $indices_effovlp_pairs[$k];
		my $indx_next_pair = ($k > 0) ? $indices_effovlp_pairs[$k-1] : undef;

		my $size_effovlp = $effovlps_distances[$indx_curr_pair]->[0]; # REVISED on 2019/08/10. #
		##my ($size_effovlp, $effdist_left2right, $effdist_right2left) = @{$effovlps_distances[$indx_curr_pair]};
		my $size_effovlp_old = $size_effovlp;

		my $indx_rlv_gb_start = $indx_curr_pair;
		my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair +2 : 0; # A BUG fixed on 2019/08/10.
		##my $indx_rlv_gb_end = (defined $indx_next_pair) ? $indx_next_pair +1 : 0;

    (print STDOUT "\n{{  \$k = $k  (w/ \$indx_curr_pair = $indx_curr_pair, \$size_effovlp = $size_effovlp) }}\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets>=5); # ADDED for DEBUGGING (on 2019/08/10). #

    if ($TEST_LEVEL_undo_ho_vcp_sets >5) { # ADDED for DEBUGGING (on 2019/08/10). #

	print STDOUT '$indx_next_pair = ', ((defined $indx_next_pair) ? $indx_next_pair : 'undef'), " .\n\n";
	print STDOUT '($indx_rlv_gb_start, $indx_rlv_gb_end) = ', "($indx_rlv_gb_start, $indx_rlv_gb_end) .\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >5) {...}"

		for (my $i1 = $indx_rlv_gb_start; $i1 >= $indx_rlv_gb_end; $i1--) { # Middle for-loop.

		    my ($gb1, $lbd1, $rbd1, $rel_w_left1) = @{$relv_gbs[$i1]};


		    my $effovlp_distance1 = ($i1 > 0) ? $effovlps_distances[$i1-1] : undef;

		    my $effdist_right2left = (defined $effovlp_distance1) ? $effovlp_distance1->[2] : - $leftmost_margin; # A BUG fixed on 2019/08/10. #
		    ##my $effdist_righ2left = (defined $effovlp_distance1) ? $effovlp_distance1->[2] : - $leftmost_margin;
    (print STDOUT "\n{ \$i1 = $i1   (w/ \$gb1 = $gb1, \$effdist_right2left = $effdist_right2left) }\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    unless ($effdist_right2left>0) { next; }

		    my $max_shift = ($size_effovlp < $effdist_right2left) ? $size_effovlp : $effdist_right2left;

    (print STDOUT "Before the 1st inner for-loop: \$max_shift = $max_shift .\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    for (my $i2 = $indx_rlv_gb_start; $i2 >= $i1; $i2--) { # 1st Inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    my $prev_incr_coord = $gb2incr_bl_coord_undo{$gb2};
			    if ($prev_incr_coord - $max_shift < $left_margins[$gb2]) {
				$max_shift = $prev_incr_coord - $left_margins[$gb2];
			    }
		    } # End of the 1st inner for-loop. #

    (print STDOUT "After the 1st inner for-loop: \$max_shift = $max_shift .\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). #

		    unless ($max_shift >0) { next; }

		    if ($i1>0) { # THe condition ADDED on 2019/08/10.
			$effovlp_distance1->[1] -= $max_shift;
			$effovlp_distance1->[2] -= $max_shift;
			if ($max_shift == $effdist_right2left) { $effovlp_distance1->[0] = 0; }

		    } else { # ADDED on 2019/08/10.
			$leftmost_margin -= $max_shift;
		    }

		    for (my $i2 = $indx_rlv_gb_start; $i2 >= $i1; $i2--) {  # 2nd inner for-loop. #
			    my $gb2 = $relv_gbs[$i2]->[0];
			    $gb2incr_bl_coord_undo{$gb2} -= $max_shift;
		    } # End of the 2nd inner for-loop. #

		    $size_effovlp -= $max_shift;

    (print STDOUT "At the end of (\$i1 = $i1) of the middle for-loop: \$size_effovlp = $size_effovlp.\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >5); # ADDED for DEBUGGING (on 2019/08/10). 

		    if ($size_effovlp == 0) { last; } # End the inspection of the pairs if the overlap of the subject pair is completely removed. #

		} # End of the middle for-loop. #

		if ($size_effovlp == 0) {

		    $effovlps_distances[$indx_curr_pair] = [0, 0, 0];

			# Remove the subject pair from the list of effectively ovelapping pairs. #

		     @indices_effovlp_pairs = @indices_effovlp_pairs[0 .. $k-1, $k+1 .. $ct_effovlp_pairs-1];
		     $ct_effovlp_pairs--;
		     $k--; # To re-use the current value of $k in the next round (of the for-loop). #

		} else {
				# Just Update the size of the effective-overlap in the subject pair. #

		    $effovlps_distances[$indx_curr_pair]->[0] = $size_effovlp;
		}

	} # End of the 3rd main for-loop. #

    if ($TEST_LEVEL_undo_ho_vcp_sets >2) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Interim Results after the 3rd Main for-loop: ]]\n\n";

	print STDOUT '[ Contents of @effovlps_distances: ]', "\n\n";
	print STDOUT join ("\t", 'Indx_pair', 'gb1', 'gb2', 'eff_size_ovlp', 'eff_dist(left>right)', 'eff_dist(right>left)'),  "\n\n";

	for (my $k=0; $k <@effovlps_distances; $k++) {
	    my $gb1 = $relv_gbs[$k]->[0];
	    my $gb2 = $relv_gbs[$k+1]->[0];
	    ##my ($gb1, $gb2) = @relv_gbs[$k, $k+1];
	    my ($effsize_ovlp, $effdist_l2r, $effdist_r2l) = @{$effovlps_distances[$k]};
	    print STDOUT join ("\t", $k, $gb1, $gb2, $effsize_ovlp, $effdist_l2r, $effdist_r2l), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '@indices_effovlp_pairs = (', join (', ', @indices_effovlp_pairs ), ") .\n\n";

	print STDOUT '%gb2incr_bl_coord_undo = ( # $gb => $incr_bl_coord_undo, ', "\n";

	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_undo_ho_vcp_sets >2) {...}"


	unless ($ct_effovlp_pairs == 0) {
	    print STDOUT "[CAUTION 'undo_hovlp_vcmpl_gbsets'] The horizontal overlap will be UNABLE to be UNDONE!!, because ";
	    print STDOUT  '@indices_effovlp_pairs = (', join (',', @indices_effovlp_pairs), ')';
	    print STDOUT " COULD NOT be EMPTIED (w/ $ct_effovlp_pairs elements left): $! !\n\n";

	    return (-1, \%gb2incr_bl_coord_undo);
	}


			### MAIN 3: Actually "UNDO" the Overlaps, via 'serial_shifts_point2point'.. ###

	my @bl_coords_frm = @{$curr_bl_coords};


	foreach my $gb (sort {$a <=> $b} keys %gb2incr_bl_coord_undo) {

			# Update the gap-block coordinates, prior to performing 'serial_shifts_point2point'. #
		$curr_bl_coords->[$gb] += $gb2incr_bl_coord_undo{$gb};
	}

	my @bl_coords_to =  @{$curr_bl_coords};


    if ($TEST_LEVEL_undo_ho_vcp_sets >3) { # ADDED in DEBUGGING on 2019/08/09. #

	print STDOUT "\n[[ Inputs of 'serial_shifts_point2point' Prepared Here: ]]\n\n";

	print STDOUT '@bl_coords_frm = (', join (', ',  @bl_coords_frm), ") ;\n\n";
	print STDOUT '@bl_coords_to  = (', join (', ', @bl_coords_to), ") .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_undo_ho_vcp_sets >3) {...}"


	my ($fl_ssh_p2p, $tot_incr_ct_null_clms)
	   = serial_shifts_point2point (
		@{$curr_set_clms}, @{$curr_bds_blks}, @bl_coords_frm, @bl_coords_to, @{$bds_gb_coords}, 
		@{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}
		);
	unless ((defined $fl_ssh_p2p) and ($fl_ssh_p2p == 0)) {
	    my $fl_out = (defined $fl_ssh_p2p) ? $fl_ssh_p2p : 'undef';
	    print STDOUT "[ERROR 'undo_hovlp_vcmpl_gbsets'] The subroutine 'serial_shifts_point2point' FAILED (w/ \$fl_ssh_p2p = $fl_out): $! !\n\n";
	    return (0, undef);
	}
    (print STDOUT "\n ... The subroutine 'serial_shifts_point2point' Ended (w/ \$fl_ssh_p2p = $fl_ssh_p2p, \$tot_incr_ct_null_clms = $tot_incr_ct_null_clms) ...\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets >1); # Added on 2019/08/09. #

   (print STDOUT "\n< Subroutine 'undo_hovlp_vcmpl_gbsets' Ended (w/ \$flag_undo = 1) />\n\n") if ($TEST_LEVEL_undo_ho_vcp_sets > 0); # ADDED on 2019/08/09. #


	return (1, \%gb2incr_bl_coord_undo);

}	# End of the subroutine, 'undo_hovlp_vcmpl_gbsets'. # 


# cmerge_split_grow_gblock_pair (NEW VERSION):
#	This new version does essentially the same job as its old version (now renamed as 'cmerge_split_grow_gblock_pair_OLD'),
#	but UNDER a DIFFERENT STRATEGY.
#	NOTE that, accompanying this revision, the portfolio of the OUTPUT data has also CHANGED.
#	[CREATION STARTED on 2019/08/06, ENDED on 2019/08/07.]
#
#	[STRATEGY]:
#	UNLIKE the old version,
#	it attempts to grow the subject pair of gap-blocks
#	while leaving the existing gap-blocks as unmoved as possible,
#	by (i) enumerating the columns available for the "split",
#	and by (ii) actually "split"ting the column one by one,
#	and "merg"ing the resulting gap-blocks to the subject gap-blocks each time a column is "split".
#
#	[REVISED on 2019/08/08]:
#	Now, the subroutine performs 'undo_hovlp_vcmpl_gbsets' before returning the results,
#	in order to undo the horizontal overlaps resulting from the main process of "merge + split".
#	[REVisEd on 2019/08/09]:
#	Now, the subroutine also outputs \%gb2incr_bl_coord_undo,
#	which is one of the main outputs of 'undo_hovlp_vcmpl_gbsets'.
#
#	[REVISED on 2019/08/18]:
#	Corrected a BUG; as a result, the subroutine can now order the gap-blocks CORRECTLY
#	even if the size-growing "split + merge" process changes
#	the intrinsic ranks between some vertically equivalent gap-blocks.
#
# cmerge_split_grow_gblock_pair:
#	This subroutine performs the "merge + split" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, and grows their sizes by a specified number of columns.
#	It also prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'icmerge_gblock_pair'.
#	[CREATED on 2019/08/03.]
#	[DEBUGGED on 2019/08/05]:
#	(i) Now, it should be able to CORRECTLY separate the gap-block boundaries
#	belonging to the columns to be "split".
#	(ii) CORRECTED the use of the output, %{$gb2sbincr_coord}, of 'displace_vincluding_veq_and_onx_to_right_end'.
#	[REVISED on 2019/08/05]:
#	The rules for assigning the coordinate increments for the 2nd (UNDOing) serial "shift"s,
#	so that the final result will look closer to the input alignment.
#
# icmerge_gblock_pair:
#	This subroutine performs the "INcomplete merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other, by a specified number of columns.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'cmerge_gblock_pair'.
#	[CREATED on 2019/08/02.]
#
# cmerge_gblock_pair:
#	This subroutine actually performs the "merge" of a specified pair of gap-blocks
#	vertically COMPLEMENTARY to each other.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'purge_gblock_pair'.
#	[CREATED on 2019/07/19.]
#	[REVISED on 2019/07/21]:
#	Now, 'mk_orgs_bds_gb_coords' is invoked to compute the new coordinate boundaries.
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb1, $gb2, $ct_to_be_merged, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts), where ] ... for the predecessor 'icmerge_gblock_pair'.
#
# INPUTS: ($gb1, $gb2, $incr_gb_size, $char_abs, \@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts
#		# IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# END of "IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (HD 1)" #
#	), where
#
#	$gb1 and $gb2 are the indices of the gap-blocks to be "merge"d;
#	They MUST be (vertically) complementary to each other, and MUST have different sizes;
#
#	$incr_gb_size is the number of columns by which the sizes of $gb1 and $gb2 will increase;
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#		# IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#			# (Added on 2019/07/26.) #
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#			# End of "(Added on 2019/07/26.)" #
#
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (HD 2)" #
#	
#
# OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj, $if_complete, $gb_rplcd, $gb_rmvd_kpt, $ct_rmvd_null_clms, \@new_indices_old_gbs, \@old_indices_new_gbs,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#		# ADDED in creation of 'cvmerge_gblock_pair'. (HD 3) #
#	\@new_list_gblocks
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# END of "IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (HD 2)" #
#
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# [ OUTPUTS: ($flag_end, $bl_sbj, $incr_coord_sbj1, $ct_rmvd_null_clms, $incr_coord_sbj2,
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes), where ] ... for the predecessor 'icmerge_gblock_pair'.
#
# [ OUTPUTS: ($flag_end, $gb_left, $gb_right, $incr_coord_left1, $incr_coord_right1, $incr_coord_left2, $incr_coord_right2, ] ... for the old version, 'cmerge_split_grow_gblock_pair_OLD'.
#
# [ OUTPUTS: ($flag_end, $gb_left, $gb_right, \@records_sessions, ] ... OBSOLETE as of 2019/08/09.
#
# OUTPUTS: ($flag_end, $gb_left, $gb_right, \@records_sessions, \%gb2incr_bl_coord_undo, # REVISED on 2019/08/09. #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	[ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes ] ... OBSOLETE as of 2019/08/18.
#			# IMPORTED from 'ssplit_gblock' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD1) #
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#			# END of "IMPORTED from 'ssplit_gblock' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD1)" #
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \@new_list_gblocks, # ADDED in REVISION on 2019/08/19.
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#	$gb_left and $gb_right are the indices of the gap-blocks that are on the left and on the right, respectively, each chosen between $gb1 and $gb2;
#
#			# ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#	@{$records_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	provides the record of the $k th session, where
#	$session_ID is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1/2 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_right/$gb_left th gap-block;
#			# End of "ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version)." #
#
#			
#			# ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09). #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09)." #
#
#	[[
#	$incr_coord_left1 and $incr_coord_right1 are the coordinate increments
#	for the 1st rounds of serial "shift"s on $gb_left and $gb_right, respectively;
#
#	$incr_coord_left2 and $incr_coord_right2 are the coordinate increments
#	for the 2nd rounds of serial "shift"s on $gb_left and $gb_right, respectively;
#	(NOTE that the 2nd rounds are intended to UNDO the 1st rounds);
#	]] ... for the old version, 'cmerge_split_grow_gblock_pair_OLD'.
#
#	[[
#		# ADDED in revision on 2019/07/15. #
#	$bl_sbj is the index for the gap-block on which 'serial_shifts_sgl_gblock' was performed (= undef if the subroutine failed);
#	$incr_coord_sbj1 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) BEFORE the "incomplete merge";
#		# END of "ADDED in revision on 2019/07/15." #
#
#	$incr_coord_sbj2 is the increment of the coordinate of $bl_sbj that resulted from the 'serial_shifts_sgl_gblock' (= undef if the subroutine failed) AFTER the "incomplete merge";
#
#	$ct_rmvd_null_clms is the count of null-columns that were generated and removed when performing the "merge" (= undef if an error occurs);
#	]] ... for the predecessor 'icmerge_gblock_pair'.
#
#			# IMPORTED from 'ssplit_gblock' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# END of "IMPORTED from 'ssplit_gblock' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD2)" #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair'. (HD)" #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
# [ sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$) { ] ... OBSOLETE as of 2019/08/18.
#
sub cmerge_split_grow_gblock_pair ($$$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\@\%\%$\%) {


	my ($gb1, $gb2, $incr_gb_size, $char_abs,
	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,
		# IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq
		# END of "IMPORTED from 'cvmerge_gblock_pair' in REVISION of 'cmerge_split_grow_gblock_pair'. (IN)" #
	) = @_;

	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

	my $rel = $inter_block_relations->[$gb1][$gb2];
	unless ($rel eq 'Cp') {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The gap-block pair, ($gb1, $gb2), are NOT vertically COMPLEMENTARY to each other (w/ rel = '$rel'): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless ($incr_gb_size>0) { # Added on 2019/08/03. #
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The \$incr_gb_size (=$incr_gb_size) is NOT POSITIVE: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $size1 = $gblock_sizes->[$gb1];
	my $size2 = $gblock_sizes->[$gb2];

	my $gb1_coord_bf = $curr_bl_coords->[$gb1];
	my $gb2_coord_bf = $curr_bl_coords->[$gb2];

	my ($lbd1_bf, $rbd1_bf) = @{$curr_bds_blks->[$gb1]};
	my ($lbd2_bf, $rbd2_bf) = @{$curr_bds_blks->[$gb2]};

    (print STDOUT "< Subroutine 'cmerge_split_grow_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; complementary block-pair to be \"INcomplete-merge\"d: (gb1, gb2) = ($gb1, $gb2); {block-sizes}=($size1, $size2); {coordinates}=($gb1_coord_bf, $gb2_coord_bf); {boundaries}_gb1=($lbd1_bf, $rbd1_bf); {boundaries}_gb2=($lbd2_bf, $rbd2_bf); \$incr_gb_size = $incr_gb_size) >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #

	my ($gb_left, $gb_right) = ($lbd1_bf < $lbd2_bf) ? ($gb1, $gb2) : ($gb2, $gb1);
	my ($size_left, $size_right) = @{$gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$curr_bds_blks->[$gb_right]};

   if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {...}"

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"



		### Preliminary (Out): ###

	my ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

	my $intm1_ct_gbs = $ct_in_gbs +2;
	my $intm2_ct_gbs = $ct_in_gbs +1;

		# Create @intm1_gblock_sizes & @intm1_inter_block_relations. #

	my @cp_curr_gblock_sizes = @{$gblock_sizes};
	my @intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##my @intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);
		# NOTE: The above 3 arrays need be UPDATED in every session. #

	my @intm1_inter_block_relations = my @intm2_inter_block_relations = ();
	for (my $gb=0; $gb <$ct_in_gbs; $gb++) {
	    my @intm1_rels_w_gb = @{$inter_block_relations->[$gb]};
	    ##my @intm2_rels_w_gb = @{$inter_block_relations->[$gb]};

	    if ($gb == $gb_left) {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('=', 'Cp');
		##$intm2_rels_w_gb[$ct_in_gbs] = '=';
	    } elsif ($gb == $gb_right) {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('Cp', '=');
		##$intm2_rels_w_gb[$ct_in_gbs] = 'Cp';
	    } else {
		@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1]
		 = @intm1_rels_w_gb[$gb_left, $gb_right];
		##$intm2_rels_w_gb[$ct_in_gbs] = $intm2_rels_w_gb[$gb_left];
	    }
	    $intm1_inter_block_relations[$gb] = \@intm1_rels_w_gb;
	    ##$intm2_inter_block_relations[$gb] = \@intm2_rels_w_gb;
	}
	my @intm1_rels_w_incr_left = @{$intm1_inter_block_relations[$gb_left]};
	@intm1_rels_w_incr_left[$gb_left, $ct_in_gbs] 
		= @{$intm1_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];

	my @intm1_rels_w_incr_right = @{$intm1_inter_block_relations[$gb_right]};
	@intm1_rels_w_incr_right[$gb_right, $ct_in_gbs+1]
		= @{$intm1_inter_block_relations[$gb_right]}[$ct_in_gbs+1, $gb_right];

	$intm1_inter_block_relations[$ct_in_gbs] = \@intm1_rels_w_incr_left;
	$intm1_inter_block_relations[$ct_in_gbs+1] = \@intm1_rels_w_incr_right;

=Set aside for Later Use.
	my @intm2_rels_w_incr_left = @{$intm2_inter_block_relations[$gb_left]};
	@intm2_rels_w_incr_left[$gb_left, $ct_in_gbs] 
		= @{$intm2_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];
	$intm2_inter_block_relations[$ct_in_gbs] = \@intm2_rels_w_incr_left;
=cut # END of "Set aside for Later Use."


		# Create $intm1_cts_vincluding, etc. #

	my ($intm1_cts_vincluding, $intm1_cts_vincluded, $intm1_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm1_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm1_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm1_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) ;


=Set aside for Later Use.
	my ($intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm2_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm2_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm2_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) ;
=cut # END of "Set aside for Later Use."



		# Create @intm1_rows_affd_by_blks.

	my @intm1_rows_affd_by_blks = (@cp_rows_affd_by_blks, @cp_rows_affd_by_blks[$gb_left, $gb_right]);
	##my @intm2_rows_affd_by_blks = (@cp_rows_affd_by_blks, $cp_rows_affd_by_blks[$gb_left]);



		### MAIN 1 & 2: Iteratively Perform the single-column-growing "Merge + Split"s. ###

	my ($rows_affd_by_left, $rows_affd_by_right) = @cp_rows_affd_by_blks[$gb_left, $gb_right];
	my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];

	my $curr_incr_gb_size = $incr_gb_size +1;
	my $curr_ct_clms = $ct_in_clms;

	my @records_sessions = (); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

     foreach my $session (1 .. $incr_gb_size) { # Outer main foreach-loop (over the sessions). #

	$curr_incr_gb_size--;
	my ($le_wd, $re_wd) = (0, $curr_ct_clms-1);

    (print STDOUT "\n{{ Session $session:  (\$curr_incr_gb_size = $curr_incr_gb_size, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 1);

		# Preliminary (In):  Enumerate the columns available for the "(size-growing) Merge + Split".

	my ($avail_sites_left, $avail_sites_in_btwn, $avail_sites_right)
	    = enum_sites_avail_for_csplit ($gb1, $gb2, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_bds_gb_coords, @{$inter_block_relations}, $le_wd, $re_wd);
	unless (defined $avail_sites_right) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'enum_sites_avail_for_csplit', FAILED on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session: $! !\n\n";

	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'enum_sites_avail_for_csplit', Ended on (\$gb1, \$gb2) = ($gb1, $gb2) for session $session ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	my @avail_sites = (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left}) );
	my $ct_avail_sites = @avail_sites;

	unless ($ct_avail_sites >= $curr_incr_gb_size) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'enum_sites_avail_for_csplit' (for session $session), output ONLY $ct_avail_sites (< \$curr_incr_gb_size (=$curr_incr_gb_size)) sites available for the \"(size-growing) merge + split\": $! !\n\n";
	    ##return (0, undef, undef, undef);
	    return (-1, undef, undef, undef); # REVISED on 2019/08/07. #
	}

	my $sbjct_site = $avail_sites[0];

   if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {

	print STDOUT "\n[ Secondary Outputs of 'enum_sites_avail_for_csplit' (#{available sites}=$ct_avail_sites): ]\n\n";

	print STDOUT '@avail_sites (= (@{$avail_sites_in_btwn}, @{$avail_sites_right}, reverse (@{$avail_sites_left})) )', "\n";
	print STDOUT ' = (', join (', ', @avail_sites), ") .\n\n";
	print STDOUT '$sbjct_site = ', "$sbjct_site .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >3) {...}"


		## MAIN I: Create a new pair of single-column gap-blocks, by "split"ting the $sbjct_site th site. ##

	my @clm_incr_left = @null_clm;
	my @clm_incr_right = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_right}) { # 1st middle foreach.
	    my $tmp = $clm_incr_right[$row];
	    $clm_incr_right[$row] = $clm_incr_left[$row];
	    $clm_incr_left[$row] = $tmp;
	} # End of the 1st middle foreach.

	my @gblks_w_lb_into_left = my @gblks_w_lb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$sbjct_site]}) { # 2nd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_lb_into_left, $gb;
	    } else {
		push @gblks_w_lb_into_right, $gb;
	    }
	} # End of the 2nd middle foreach.
	push @gblks_w_lb_into_left, $ct_in_gbs;
	push @gblks_w_lb_into_right, $ct_in_gbs +1;

	my @gblks_w_rb_into_left = my @gblks_w_rb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$sbjct_site]}) { # 3rd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_rb_into_left, $gb;
	    } else {
		push @gblks_w_rb_into_right, $gb;
	    }
	} # End of the 3rd middle foreach.
	push @gblks_w_rb_into_left, $ct_in_gbs;
	push @gblks_w_rb_into_right, $ct_in_gbs +1;

	my @intm1_curr_set_clms = ();
	my @intm1_curr_gblks_w_spec_lb = my @intm1_curr_gblks_w_spec_rb = ();

	for (my $c=0; $c < $sbjct_site; $c++) { # 1st Middle MAIN for-loop (over sites on the left of $sbjct_site). #
	    $intm1_curr_set_clms[$c] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 1st Middle MAIN for-loop. #

	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = (\@clm_incr_left, \@clm_incr_right);
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_lb_into_left, \@gblks_w_lb_into_right);
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_rb_into_left, \@gblks_w_rb_into_right);

	for (my $c = $sbjct_site+1; $c <$curr_ct_clms; $c++) { # 2nd Middle MAIN for-loop (over sites on the right of $sbjct_site). #
	    $intm1_curr_set_clms[$c+1] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c+1] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c+1] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 2nd Middle MAIN for-loop. #


	my $intm1_ct_clms = @intm1_curr_set_clms;


		# Create @intm1_curr_bds_blks. #

	my @intm1_curr_bds_blks = ();
	for (my $gb=0; $gb < $intm1_ct_gbs; $gb++) { $intm1_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$intm1_ct_clms; $c++) { # 2nd trailing for-loop for Main 1-(i). #
            foreach my $gb (@{$intm1_curr_gblks_w_spec_lb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[0] = $c;
            }
            foreach my $gb (@{$intm1_curr_gblks_w_spec_rb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[1] = $c;
            }
	} # End of the 2nd trailing for-loop for Main 1-(i). #


		## Preliminary (In 2): Create the coordinate frame for the 1st intermediate gap-block configuration. #

	my ($intm1_le_wd, $intm1_re_wd) = (0, $intm1_ct_clms-1);

	my ($intm1_orgs_gb_coords, $intm1_bds_gb_coords, $intm1_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@intm1_curr_bds_blks, $intm1_le_wd, $intm1_re_wd, @intm1_inter_block_relations, $maxct_1way_shifts);
	unless (defined $intm1_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	my @intm1_curr_bl_coords = @{$intm1_orgs_gb_coords};


			# (Added on 2019/08/07. (Main II)) #

   if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_gblock_configs (@intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {...}"

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_remng_attr_gblks (@intm1_inter_block_relations, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm}, @intm1_gblock_sizes, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"


		## MAIN II-(i): Merge the $gb_right and $gb_incr_right th gap-block, via the 1st Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 1st Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_right (=$gb_right) th and \$gb_incr_right (=$gb_incr_right) th gap-blocks ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 1);

	my ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_right, $gb_incr_right, @intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm},
		@intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, @intm1_rows_affd_by_blks, $cnct_null_clm,
		@intm1_inter_block_relations, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_gblock_sizes, $maxct_1way_shifts
		);
	unless ($flag_end_smg1) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 1st round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);



   if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >5) {...}"

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm2_inter_block_relations}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm}, @{$intm2_gblock_sizes}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"



		## MAIN II-(ii): Merge the $gb_left and $gb_incr_left th gap-block, via the 2nd Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 2nd Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_left (=$gb_left) th and \$gb_incr_left (=$gb_incr_left) th gap-blocks ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 1);

	my ($flag_end_smg2, $bl_sbj_smg2, $incr_coord_sbj_smg2, $gb_absbd_smg2, $gb_kpt_smg2, $new_indices_gbs_smg2, $old_indices_gbs_smg2,
	    $intm3_curr_set_clms, $intm3_curr_bds_blks, $intm3_curr_bl_coords, $intm3_bds_gb_coords, $intm3_gb_coords_wd_frm,
	    $intm3_curr_gblks_w_spec_lb, $intm3_curr_gblks_w_spec_rb, $intm3_rows_affd_by_blks,
	    $intm3_inter_block_relations, $intm3_cts_vincluding, $intm3_cts_vincluded, $intm3_cts_onx, $intm3_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_left, $gb_incr_left, @{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm},
		@{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, @{$intm2_rows_affd_by_blks}, $cnct_null_clm,
		@{$intm2_inter_block_relations}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_end_smg2) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The 2nd round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);


   if ($TEST_LEVEL_cmrg_spl_grw_gbp >= 5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm3_curr_set_clms}, @{$intm3_curr_bds_blks}, @{$intm3_curr_bl_coords}, @{$intm3_curr_gblks_w_spec_lb}, @{$intm3_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >= 5) {...}"

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm3_inter_block_relations}, @{$intm3_bds_gb_coords}, @{$intm3_gb_coords_wd_frm}, @{$intm3_gblock_sizes}, @{$intm3_cts_vincluding}, @{$intm3_cts_vincluded}, @{$intm3_cts_onx}, @{$intm3_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 5) {...}"


		## MAIN IIpost: Update Important Data-Sets
		##	that characterize the Updated gap-block configuration. ##

			# Update the set of columns. #
	@cp_curr_set_clms = @{$intm3_curr_set_clms};
	$curr_ct_clms = @cp_curr_set_clms;

			# Update the set of gap-block boundaries & its "inverse-functions". #
	@cp_curr_bds_blks = @{$intm3_curr_bds_blks};
	@cp_curr_gblks_w_spec_lb = @{$intm3_curr_gblks_w_spec_lb};
	@cp_curr_gblks_w_spec_rb = @{$intm3_curr_gblks_w_spec_rb};

			# Update the coordinate frame. #
	@cp_curr_bl_coords = @{$intm3_curr_bl_coords};
	@cp_bds_gb_coords = @{$intm3_bds_gb_coords};
	@cp_gb_coords_wd_frm = @{$intm3_gb_coords_wd_frm};

			# Update the sets of gap-block sizes. #
	@cp_curr_gblock_sizes = @{$intm3_gblock_sizes};
	@intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##@intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);


		## Record the Core information on this Session. ##

	push @records_sessions, [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

			# End of "(Added on 2019/08/07. (Main II))" #

    } # End of the outer MAIN foreach-loop (over the sessions). #


  (print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the $incr_gb_size Sessions !!---\n\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

		### MAIN III: Create the Final Output Data-Sets. ###
		### (Added on 2019/08/07. (MAIN III)) ###

	my @new_curr_set_clms = @cp_curr_set_clms;
	my @new_curr_bds_blks = @cp_curr_bds_blks;

	my @new_curr_bl_coords = @cp_curr_bl_coords;
	my @new_bds_gb_coords = @cp_bds_gb_coords;
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm;

	my @new_curr_gblks_w_spec_lb = @cp_curr_gblks_w_spec_lb;
	my @new_curr_gblks_w_spec_rb = @cp_curr_gblks_w_spec_rb;
	my @new_gblock_sizes = @cp_curr_gblock_sizes;

 		# ADDED in REVISION (on 2019/08/18). #
	my @rows_affd_by_new_blks = @{$rows_affd_by_blks};
	my ($new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
		= ($inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx);

		### END of "(Added on 2019/08/07. (MAIN III))" ###


		### MAIN III': Re-Rank the resulting Gap-Blocks, if necessary. ###
		### (ADDED for DEBUGGING on 2019/08/18. (MAIN III')) ###
		### (=> IMPORTED from 'csplit_gblock' (on 2019/08/18).) ###

	##my @new_indices_old_gbs = @{$new_indices_old_gbs};
	##my @old_indices_new_gbs = @{$old_indices_new_gbs};

		# Fixed a Minor BUG. #
	my $ct_out_gbs = @new_curr_bds_blks;
	my $ct_out_clms = @new_curr_set_clms;
	##my $ct_out_gbs = $ct_in2_gbs;
	##my $ct_out_clms = $ct_in2_clms;

	my @tmp_new_list_gblocks = ();
	## $list_gblocks0, $indx_br, $indx_u_or_d,

=For Reference
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
=cut # END of "For Reference"

	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {

	    my $info_gb = $list_gblocks0->[$gb];
	    my @info_gb = @{$info_gb}[$indx_br, $indx_u_or_d];
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};

			# MODIFIED in DEBUGGING on 2019/08/19. #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);

	    push @info_gb, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows;
	    ##for (my $i=0; $i <@{$info_gb}; $i++) {
	    ##	if (($i == $indx_br) or ($i == $indx_u_or_d) or ($i == 2) or ($i == 3)) { next; }
	    ##	push @info_gb, $info_gb->[$i];
	    ##}
			# END of "MODIFIED in DEBUGGING on 2019/08/19." #

	    $tmp_new_list_gblocks[$gb] = \@info_gb;
	}

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$tmp_new_list_gblocks[$a]}, @{$tmp_new_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1);

	my $if_reordered = 0;
	my @new_indices_old_gbs = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	    if ($ogb != $ngb) { $if_reordered = 1; }
	}

    if ($if_reordered) {

  (print STDOUT "[INRORMATION 'cmerge_split_grow_gblock_pair'] AFTER the \"(size-growing) Merge + Split (complementary types)\", The the gap-blocks are RE-ORDERED (w/ ", '@old_indices_new_gbs = (', join (',', @old_indices_new_gbs), ") !!\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);
  (print STDOUT " ... Accordingly, the output data-sets will be MODIFIED ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

    if ($TEST_LEVEL_cmrg_spl_grw_gbp >2) {

	print STDOUT '@new_indices_old_gbs = (', join (', ', @new_indices_old_gbs), ") .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >2) {...}"

	@new_curr_bds_blks = @cp_curr_bds_blks[@old_indices_new_gbs];
	@new_curr_gblks_w_spec_lb = @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $ngb;
	}


	@new_gblock_sizes = @cp_curr_gblock_sizes[@old_indices_new_gbs];

	@rows_affd_by_new_blks = @{$rows_affd_by_blks}[@old_indices_new_gbs];

			# Re-Create  @{$new_inter_block_relations}. #

	@tmp_new_list_gblocks = @tmp_new_list_gblocks[@old_indices_new_gbs];

	($new_inter_block_relations)
	  = mk_inter_block_relations (@tmp_new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@tmp_new_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@tmp_new_list_gblocks', " ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) ;

			# Re-Create @{$new_cts_vincluding}, etc. #

	($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@{$new_inter_block_relations}', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1) ;


			# Re-Create @new_curr_bl_coords  & @new_bds_gb_coords  & @new_gb_coords_wd_frm . #

	my ($new_le_wd, $new_re_wd) = (0, $ct_out_clms-1);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks after the \"Split (complementary types)\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks after the \"(size-growing) Merge + Split (complementary types)\" ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);

	@new_curr_bl_coords    = @{$new_orgs_gb_coords} ;
	@new_bds_gb_coords     = @{$new_bds_gb_coords} ;
	@new_gb_coords_wd_frm  = @{$new_gb_coords_wd_frm} ;

	$gb_left = $new_indices_old_gbs[$gb_left];
	$gb_right = $new_indices_old_gbs[$gb_right];

    (print STDOUT "NEW (\$gb_left, \$gb_right) = ($gb_left, $gb_right) .\n\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >2);

    } # END of "if ($if_reordered) {...}"


		### END of "(ADDED for DEBUGGING on 2019/08/18. (MAIN III')) ###
		###      (=> IMPORTED from 'csplit_gblock' (on 2019/08/18).)" ###


		### MAIN IIIpost: UNDO the overlaps between the gap-blocks
		###		vertically equivalent to and/or complementary to $gb1 and $gb2. ###
		### (ADDED on 2019/08/08. (MAIN IIIpost)) ###

			# REVISED on 2019/08/19. #
	my ($fl_undo, $gb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets ($gb_left, $gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @rows_affd_by_new_blks, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
=OBSOLETE as of 2019/08/18.
	  = undo_hovlp_vcmpl_gbsets ($gb_left, $gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @new_gblock_sizes
		);
=cut # END of "OBSOLETE as of 2019/08/18."
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'cmerge_split_grow_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets' FAILED on the cofiguration of gap-blocks after the main foreach-loop for \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The the subroutine, 'undo_hovlp_vcmpl_gbsets', Ended on the configuration of gap-blocks immediately after the main foreach-loop for the \"(size-growing) merge + split\" of \$gb_left (=$gb_left) and \$gb_right (=$gb_right) ...\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp >1);


   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) { # ADDED on 2019/08/09. #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets': ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) {...}"


		### END of "(ADDED on 2019/08/08. (MAIN IIIpost))" ###

   if ($TEST_LEVEL_cmrg_spl_grw_gbp >2) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp >2) {...}"

   if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"(size-growing) Merge + Split (complementary types)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

=OBSOLETE as of 2019/08/18.
				# MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02). #
	unless ( br_print_out_remng_attr_gblks (@new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, *STDOUT) ) {
	    print STDOUT "[WARNING 'cmerge_split_grow_gblock_pair'] The subroutine 'br_print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cmerge_split_grow_gblock_pair' ...\n\n";
	}
				# END of "MODIFIED in creation of 'icmerge_gblock_pair' (on 2019/08/02)." #
=cut # END of "OBSOLETE as of 2019/08/18."
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cmrg_spl_grw_gbp > 3) {...}"

			# (Added in creation of 'csplit_gblock'.) (FIN) #
			# (=> IMPORTED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18).) (FIN) #

		# Create @new_list_gblocks. #

	my @new_list_gblocks = ();
	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my @info_gb = @{$tmp_new_list_gblocks[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};
	    $new_list_gblocks[$gb] = \@info_gb;
	}

	my $fl_end = $fl_undo;
			# END of "(Added in creation of 'csplit_gblock'.) (FIN) #
			# (=> IMPORTED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18).) (FIN)" #

	
    (print STDOUT "\n< Subroutine 'cmerge_split_grow_gblock_pair' Ended (w/  \$fl_end = $fl_end, ", '($gb_left, $gb_right) = ', "($gb_left, $gb_right) . >\n\n") if ($TEST_LEVEL_cmrg_spl_grw_gbp > 0);


	## [ return (1, $gb_left, $gb_right, $incr_coord_left1, $incr_coord_right1, $incr_coord_left2, $incr_coord_right2, ] ... for the old version, 'cmerge_split_grow_gblock_pair_OLD'.

	## [ return (1, $gb_left, $gb_right, \@records_sessions,  # REVISED in creation of 'cmerge_split_grow_gblock_pair' (New Version) on 2019/08/07. # ] ... OBSOLETE as of 2019/08/09,

	## return ($fl_undo, $gb_left, $gb_right, 
	return ($fl_end, $gb_left, $gb_right, 
	\@records_sessions,  $gb2incr_bl_coord_undo, # REVISED in REVISION of 'cmerge_split_grow_gblock_pair' (New Version) on 2019/08/09. #

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,

			# MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (FIN) #
	## [ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes ] ... OBSOLETE as of 2019/08/18. #
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_indices_old_gbs, \@old_indices_new_gbs, \@new_list_gblocks,
			# END of "MODIFIED in REVISION of 'cmerge_split_grow_gblock_pair' (on 2019/08/18). (FIN)" #
	);

}	# End of the subroutine, 'cmerge_split_grow_gblock_pair'. #


# vsplit_plus_smerge_gblock_pair:
#	This subroutine "vertically split"s the 1st subject gap-block into a pair of gap-blocks
#	that are its "effective children" & do some adjusting "shift"s, via 'vsplit_gblock';
#	then, the subroutine "merge"s the 2nd subject gap-block into one of the resulting "effective children" that is of the same type as the 2nd subject, via 'smerge_gblock_pair'.
#	It was thus created by combining 'vsplit_gblock' & 'smerge_gblock_pair',
#	and by adding some auxiliary functions.
#	[CREATION STARTED on 2019/09/02, ENDED on 2019/09/05.]
#	[REVISED on 2019/09/06]:
#	Now, the subroutine attempts to UN-NEST the resulting gap-blocks if they are nested after the "vsplit + smerge".
#	 
# vsplit_gblock:
#	This subroutine "vertically split"s the subject gap-block into a pair of gap-blocks that are the "effective children" of the subject,
#	then, it "shift"s one of the resulting child blocks to that they will NOT be nested.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' above.
#	[CREATION STARTED on 2019/08/22, ENDED on 2019/08/23.]
#	
# [ INPUTS: ($gb_sbj, \@info_effch1, \@info_effch2, # MODIFIED in creation of 'vsplit_gblock'. (HD1a) # ] ... for the predecessor, 'vsplit_gblock'. (HD1)
#
# INPUTS: ($gb_sbj_pa, $gb_sbj_ch, # MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1a) #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b) #
#	\@set_all_exnodes, \%nodeid2msa_row
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b)" #
#	), where
#
#		# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a) #
#	$gb_sbj_pa and $gb_sbj_ch are the indices of the subject gap-blocks
#	that are an effective parent and an effective child, respectively, to each other;
#
#	[[
#	$gb_sbj is the index of the gap-block to be "vertically split";
#
#	@info_effch1 and @info_effch2 provide information on the TWO effective children of $gb_sbj;
#	more specifically,
#	@info_effch1/2 = ($br_ch, $up_or_lw_ch, \@exnodes_affd_by_ch, \@rows_affd_by_ch), where
#	the 1st/2nd child is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@exnodes_affd_by_ch stores the indices of external nodes affected by the child,
#	and @rows_affd_by_ch stores the indices of the MSA rows affected by the child.
#	[Typically, \@info_effch1 and \@info_effch2 are parts of the output of 'list_vsplit_candidates' above];
#	]] ... for the predecessor, 'vsplit_gblock'. 
#		# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a)" #
#
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b) #
#	@set_all_exnodes stores the indices of all external nodes;
#
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b)" #
#	
#
# [ OUTPUTS: ($flag_end, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2a) # ] ... for the predecessor, 'vsplit_gblock'. (HD2a) 
#
#			# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a) #
# OUTPUTS: ($flag_end, $gb_ch_left_NEW, $gb_kpt_NEW, 
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, \@intm_indices_old_gbs, \%gb2incr_bl_coord_undo_ovlps_af_vspl,
#	$gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, \@new_indices_intm_gbs,
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_coord_undo_af_smg,
#			# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a)" #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	@new_list_excluded_bin_hrels # ADDED in creation of 'vsplit_gblock'. (HD1a) #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the horizontal overlaps resulting from either 'cvsplit' or 'smerge' are NOT UN-DONE completely;
#
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i) #
#	$gb_ch_left_NEW is the NEW index of the "effective-child" gap-block that resulted from "vertically split"ting $gb_sbj_pa AND that was LEFT unmerged to $gb_sbj_ch;
#	$gb_kpt_NEW is the NEW index of the gap-block resulting from the (horizontal) "merge" of $gb_sbj_ch and one of the products of "vertically split"ting $gb_sbj_pa;
#	[These are two of the MAIN-AUXILIARY outputs of this subroutine];
#	[[
#	$gb_ch1_NEW and $gb_ch2_NEW are the NEW indices of the 1st and 2nd child, respectively, of $gb_sbj;
#	]] ... for the predecessor, 'vsplit_gblock'.
#
#	$gb_shftd_af_vspl_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was "shift"ed to make the product children non-nested;
#	$incr_coord_shftd_af_vspl (= +1 or -1) is the increment of the coordinate of $gb_shftd_af_vspl_intm via the aforementioned "shift"; 
#	@intm_indices_old_gbs stores the "intermediate" indices, after the 'vsplit_gblock', of the "old" gap-blocks before the 'vsplit_gblock';
#	%gb2incr_bl_coord_undo_ovlps_af_vspl = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "vertical split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'vsplit_gblock'];
#	[[
#	$gb_shftd is the index (in the NEW gap-block set) of the child gap-block that was "shift"ed,
#	to make the two children non-nested;
#
#	$incr_coord_shftd (= +1 or -1) is the increment of the coordinate of $gb_shftd via the aforementioned "shift";
#	]] ... for the predecessor, 'vsplit_gblock'.
#
#	$gb_ch_2B_mgd_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was then "merge"d with $gb_sbj_ch (in terms of the old block set);
#	($gb1_intm, $gb2_intm) are just ($gb_ch_2B_mgd_intm, $gb_sbj_ch_intm) sorted in ascending order;
#	[These were created in preparation of 'smerge_gblock_pair'];
#
#	$bl_sbj_ssh4smg_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "shift"ed before their merger;
#	$incr_coord_sbj_ssh4smg is the increment of the coordinate of $bl_sbj_ssh4smg via the aforementioned "shift";
#	$gb_absbd_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "absorted" as a result of the 'smerge_gblock_pair';
#	@new_indices_intm_gbs stores the "new" indices, after the 'smerge_gblock_pair', of the "intermediate" gap-blocks before the "smerge_gblock_pair";
#	[These are the AUXILIARY outputs of 'smerge_gblock_pair'];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i)" #
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vsplit + smerge";
#	More precisely, the gap-block that had been the $ogb th before the "vsplit + smerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "vertical-split"
#	(, and $new_indices_old_gbs[$ogb] is undefined for some $ogb);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vsplit + smerge";
#	More precisely, the gap-block that is the $ngb th after the "vsplit + smerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "vsplit + smerge";
#
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii) #
#	[[
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#	]] ... for the predecessor, 'vsplit_gblock'. 
#
#	%gb2incr_coord_undo_af_smg = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the 'smerge_gblock_pair'}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_smg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "vertical-split";
#
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#
#		# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD3) #
#	[ @new_list_excluded_bin_hrels = ([$gb_ch1_NEW, $gb_ch2_NEW, '=']) ] ... for the predecessor, 'vsplit_gblock'.
#	@new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']);
#		# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD3)" #
#	is the NEW list of PROHIBITED binary horizontal relationships,
#	which conforms to the format prescribed in the header of 'nonrec_mlt_shifts_gapblocks' above;
#
sub vsplit_plus_smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%) {

	# [ my ($gb_sbj, $info_effch1, $info_effch2, ] ... for the predecessor, 'vsplit_gblock'. #
	my ($gb_sbj_pa, $gb_sbj_ch, # Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq,
		# Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) #
	$set_all_exnodes, $nodeid2msa_row
		# End of "Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN)" #
	) = @_; 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	unless (($ct_in_clms>0) and ($ct_in_gbs>0)) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] ILLEGAL INPUT: (#{columns}, #{gap-blocks}) = ($ct_in_clms, $ct_in_gbs) in the input MSA: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $ct_rows = @{$curr_set_clms->[0]};

   (print STDOUT "< Subroutine 'vsplit_plus_smerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; effective parent-child pair of gap-blocks to be \"vsplit + smerge\"d: (\$gb_sbj_pa, \$gb_sbj_ch) = ($gb_sbj_pa, $gb_sbj_ch) ) >\n\n") if ($TEST_LEVEL_vssm_gbp > 0);


   if ($TEST_LEVEL_vssm_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) {...}"


   if ($TEST_LEVEL_vssm_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >3) {...}"


			### Preliminary: Virtually VSplit the $gb_sbj_pa into TWO effective children. ###

	my ($br_sbj_pa, $up_or_lw_sbj_pa) = @{$list_gblocks0->[$gb_sbj_pa]}[$indx_br, $indx_u_or_d];

	my ($set_eff_child_gblks) = vsplit_into_eff_child_gblocks ($br_sbj_pa, $up_or_lw_sbj_pa,
		 %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, @{$set_all_exnodes},
		 %{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw}) ;
	unless (defined $set_eff_child_gblks) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The subroutine 'vsplit_into_eff_child_gblocks' FAILED on (\$br_sbj_pa, \$up_or_lw_sbj_pa) = ($br_sbj_pa, $up_or_lw_sbj_pa), for \$gb_sbj_pa (=$gb_sbj_pa) : $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless (@{$set_eff_child_gblks} == 2) {
	    my $ct_echs = @{$set_eff_child_gblks};
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] SOMEHOW, \$gb_sbj_pa (=$gb_sbj_pa) has $ct_echs (!= 2) effective child sets of msa rows: $! !\n\n";
	    return (0, undef, undef, undef);
	}

   if ($TEST_LEVEL_vssm_gbp >2) { # (Pre) #

	print STDOUT "\n[ Outputs of 'vsplit_into_eff_child_gblocks': ]\n\n";

	print STDOUT join ("\t", 'Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(external,nodes,affected)', '[MSA,rows,affected]'), "\n\n";

	for (my $i=0; $i <2; $i++) {
	    my $cid = $i+1;
	    my ($br, $u_or_l, $exnodes_affd, $rows_affd) = @{$set_eff_child_gblks->[$i]};
	    my $cnct_exnodes_affd = join (',', @{$exnodes_affd});
	    my $cnct_rows_affd = join (',', @{$rows_affd});

	    print STDOUT join ("\t", "Ch_$cid", $br, $u_or_l, "($cnct_exnodes_affd)", "[$cnct_rows_affd]"), "\n";
	} # End of "for (my $i=0; $i <2; $i++) {...}" #

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >2) { # (Pre) # ...}"


		### MAIN 1: "Vertically Split" $gb_sbj_pa. ###

	my ($info_effch1, $info_effch2) = @{$set_eff_child_gblks};


	my ($flag_vspl, $gb_ch1_intm, $gb_ch2_intm, $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $old_indices_intm_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	    $intm_curr_set_clms, $intm_curr_bds_gblks, $intm_curr_bl_coords, $intm_bds_gb_coords, $intm_gb_coords_wd_frm,
	    $intm_curr_gblks_w_spec_lb, $intm_curr_gblks_w_spec_rb, $rows_affd_by_intm_blks,
	    $intm_inter_block_relations, $intm_cts_vincluding, $intm_cts_vincluded, $intm_cts_onx, $intm_gblock_sizes,
	    $intm_list_gblocks, $intm_list_excluded_bin_hrels)
	  = vsplit_gblock (
	     $gb_sbj_pa, @{$info_effch1}, @{$info_effch2},
	     @{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm},
	     @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
	     @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes},  $maxct_1way_shifts,
	     @{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
		);
	unless ($flag_vspl) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The 1st MAIN Subroutine 'vsplit_gblock' FAILED on the \$gb_sbj_pa (=$gb_sbj_pa) th gap-block: $! !\n";
	    return (0, undef, undef, undef);
	}
	unless ($flag_vspl > 0) {
	    print STDOUT "[CAUTION 'vsplit_plus_smerge_gblock_pair'] The 1st MAIN subroutine, 'vsplit_gblock' FAILED on \$gb_sbj_pa (=$gb_sbj_pa), because the resulting horizontal overlaps canNOT be UNDONE completely!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}
   (print STDOUT "\n --- The 1st MAIN Subroutine 'vsplit_gblock' ENDED on the \$gb_sbj_pa (=$gb_sbj_pa) th gap-block (w/ \$flag_vspl = $flag_vspl, \$gb_ch1_intm = $gb_ch1_intm, \$gb_ch2_intm = $gb_ch2_intm, \$gb_shftd_af_vspl_intm = $gb_shftd_af_vspl_intm, \$incr_coord_shftd_af_vspl = $incr_coord_shftd_af_vspl) ...\n\n\n") if ($TEST_LEVEL_vssm_gbp >1);


	my $ct_gbs_intm = @{$intm_curr_bds_gblks};

   if ($TEST_LEVEL_vssm_gbp >2) { # (M1-A) #

	print STDOUT "\n[[ Outputs of 'vsplit_gblock' ]]\n\n";

        print STDOUT "[ Relationships between old and intermediate indices of gap-blocks (#{gap-blocks}_intm=$ct_gbs_intm) : ]\n\n";

        my @out_intm_indices_old_gbs = my @out_old_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_old_gbs, \@out_intm_indices_old_gbs],
                          [$old_indices_intm_gbs, \@out_old_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$intm_indices_old_gbs} = (', join (', ', @out_intm_indices_old_gbs), ") .\n\n";

        print STDOUT '@{$old_indices_intm_gbs} = (', join (', ', @out_old_indices_intm_gbs), ") .\n";

        print STDOUT "\n\n";

	print STDOUT '[ Contents of @{$intm_list_excluded_bin_hrels}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";
	for (my $i=0; $i < @{$intm_list_excluded_bin_hrels}; $i++) {
	    print STDOUT join ("\t", $i, @{$intm_list_excluded_bin_hrels->[$i]}), "\n";
	}
	print STDOUT "\n\n";

			# Added on 2019/09/06. #
	print STDOUT '[ Contents of %{$gb2incr_bl_coord_undo_ovlps_af_vspl}: ]', "\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo_ovlps_af_vspl} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo_ovlps_af_vspl}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo_ovlps_af_vspl->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/09/06." #

   } # End of "if ($TEST_LEVEL_vssm_gbp >2) { # (M1-A) #...}"

   if ($TEST_LEVEL_vssm_gbp >5) { # (M1-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the \"Vertical Split\" ]\n\n";

	unless ( print_out_gblock_configs (@{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >5) { # (M1-B) # ...}"

   if ($TEST_LEVEL_vssm_gbp > 5) { # (M1-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the \"Vertical Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm_inter_block_relations}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm}, @{$intm_gblock_sizes}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$rows_affd_by_intm_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) { # (M1-C) # ...}"

   if ($TEST_LEVEL_vssm_gbp >= 5) { # (M1-D) #

	print STDOUT "\n";
        print STDOUT '[ Contents of @{$intm_list_gblocks}: ]', "\n\n";

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_gbs_intm; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks->[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >= 5) { # (M1-D) #...}"



		### MAIN 2pre: Identify the resulting "effective child" to be "merge"d with $gb_sbj_ch_intm. ### 

	my $gb_sbj_ch_intm = $intm_indices_old_gbs->[$gb_sbj_ch];
	my $rows_affd_by_sbj_ch = $rows_affd_by_intm_blks->[$gb_sbj_ch_intm];

   if ($TEST_LEVEL_vssm_gbp >2) { # (M2Pre-A) #
	print STDOUT "\n";
	print STDOUT "\$gb_sbj_ch_intm = $gb_sbj_ch_intm .\n\n";
	print STDOUT '@{$rows_affd_by_sbj_ch} = (', join (', ', @{$rows_affd_by_sbj_ch}), ") .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_vssm_gbp >2) { # (M2Pre-A) #...}"

	my $gb_ch_2B_mgd_intm;
	foreach my $gb_intm ($gb_ch1_intm, $gb_ch2_intm) {
	    my $rows_affd_by_gb = $rows_affd_by_intm_blks->[$gb_intm];
	    my ($common, $ex1, $ex2) = diff_nmarray (@{$rows_affd_by_sbj_ch}, @{$rows_affd_by_gb});
	    unless (defined $common) {
		print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The subroutine 'diff_nmarray' FAILED on the rows affected by \$gb_sbj_ch_intm (=$gb_sbj_ch_intm) and those affected by \$gb_intm (=$gb_intm): $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    if ((@{$ex1}==0) and (@{$ex2}==0)) { $gb_ch_2B_mgd_intm = $gb_intm; }

	} # End of "foreach my $gb_intm ($gb_ch1_intm, $gb_ch2_intm) {...}". #

	unless (defined $gb_ch_2B_mgd_intm) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The \$gb_ch_2B_mgd_intm is NOT DEFINED: $! !\n";
	    return (0, undef, undef, undef);
	}


	my $gb_ch_left_intm = ($gb_ch_2B_mgd_intm == $gb_ch1_intm) ? $gb_ch2_intm : $gb_ch1_intm ;

   if ($TEST_LEVEL_vssm_gbp >1) { # (M2Pre-B) #
	print STDOUT "\n";
	print STDOUT "(\$gb_ch_2B_mgd_intm, \$gb_ch_left_intm) = ($gb_ch_2B_mgd_intm, $gb_ch_left_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_vssm_gbp >1) { # (M2Pre-B) #...}"


	my ($gb1_intm, $gb2_intm) = ($gb_sbj_ch_intm < $gb_ch_2B_mgd_intm) ?
		($gb_sbj_ch_intm, $gb_ch_2B_mgd_intm) : ($gb_ch_2B_mgd_intm, $gb_sbj_ch_intm) ;

   if ($TEST_LEVEL_vssm_gbp >2) { # (M2Pre-C) #
	print STDOUT '($gb1_intm, $gb2_intm) = ', "($gb1_intm, $gb2_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_vssm_gbp >2) { # (M2Pre-C) #...}" # A BUG fixed (on 2019/09/05).
   #} if ($TEST_LEVEL_vssm_gbp >2) { # (M2Pre-C) #


		### MAIN 2: "Merge" $gb1_intm and $gb2_intm (i.e., $gb_sbj_ch_intm and $gb_ch_2B_mgd_intm). ###

	my ($flag_smg, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $gb_kpt_intm, $new_indices_intm_gbs, $intm_indices_new_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = smerge_gblock_pair (
	     $gb1_intm, $gb2_intm, @{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm},
	     @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, @{$rows_affd_by_intm_blks}, $cnct_null_clm,

	     @{$intm_inter_block_relations}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$intm_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_smg) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The 2nd MAIN Subroutine 'smerge_gblock_pair' FAILED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) in the Configuration AFTER the \"Vertical Split\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n --- The 2nd MAIN Subroutine 'smerge_gblock_pair' ENDED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) AFTER the \"vertical split\" (w/ \$bl_sbj_ssh4smg_intm = $bl_sbj_ssh4smg_intm, \$incr_coord_sbj_ssh4smg = $incr_coord_sbj_ssh4smg, \$gb_absbd_intm = $gb_absbd_intm, \$gb_kpt_intm = $gb_kpt_intm) ...\n\n\n") if ($TEST_LEVEL_vssm_gbp >0);

	my $ct_out_gbs = @{$new_curr_bds_gblks};

	my $ct_out_clms = @{$new_curr_set_clms}; # ADDED in REVISION (on 2019/09/06). #


   if ($TEST_LEVEL_vssm_gbp >2) { # (M2-A) #

	print STDOUT "\n[[ Outputs of 'smerge_gblock_pair' ]]\n\n";

        print STDOUT "[ Relationships between new and intermediate indices of gap-blocks (#{gap-blocks}_new=$ct_out_gbs) : ]\n\n";

        my @out_intm_indices_new_gbs = my @out_new_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_new_gbs, \@out_intm_indices_new_gbs],
                          [$new_indices_intm_gbs, \@out_new_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$new_indices_intm_gbs} = (', join (', ', @out_new_indices_intm_gbs), ") .\n\n";
        print STDOUT '@{$intm_indices_new_gbs} = (', join (', ', @out_intm_indices_new_gbs), ") .\n";
        print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >2) { # (M2-A) # ...}"


   if ($TEST_LEVEL_vssm_gbp >5) { # (M2-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'smerge_gblock_pair': ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >5) { # (M2-B) # ...}"

   if ($TEST_LEVEL_vssm_gbp > 5) { # (M2-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'smerge_gblock_pair': ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) { # (M2-C) # ...}"


		# MAIN 2post: If any overlaps between vertically equivalent and/or complementary gap-blocks result from the 'smerge_gblock_pair', undo them. #

	my $gb_kpt_NEW = $new_indices_intm_gbs->[$gb_kpt_intm]; # The new index for the gap-block resulting from the "smerge". #
	my $gb_ch_left_NEW = $new_indices_intm_gbs->[$gb_ch_left_intm]; # The "effective-child" block that was left unmerged. # MOVED from Main 3 below, in REVISION (on 2019/09/06). #

	my ($fl_undo_af_smg, $gb2incr_coord_undo_af_smg)
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_kpt_NEW,
	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	        @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	        );
	unless ($fl_undo_af_smg) {
	    print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)) on the cofiguration of gap-blocks after the 'smerge_gblock_pair' on (\$gb_sbj_ch_intm, \$gb_ch_2B_mgd_intm) = ($gb_sbj_ch_intm, $gb_ch_2B_mgd_intm): $! !\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)), Ended (w/ \$fl_undo_af_smg = $fl_undo_af_smg) on the configuration of gap-blocks after the 'smerge_gblock_pair' on (\$gb_sbj_ch_intm, \$gb_ch_2B_mgd_intm) = ($gb_sbj_ch_intm, $gb_ch_2B_mgd_intm) ...\n\n") if ($TEST_LEVEL_vssm_gbp >1);


   if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A) #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)): ]\n\n";

	print STDOUT "\$fl_undo_af_smg = $fl_undo_af_smg .\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_smg} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_smg}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_smg->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A) # ...}"

	unless ($fl_undo_af_smg > 0) {
	    print STDOUT "[CAUTION 'vsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

			# ADDED in REVISION (on 2019/09/06). #

	my ($lbd_kpt, $rbd_kpt) = @{$new_curr_bds_gblks->[$gb_kpt_NEW]};
	my ($lbd_ch_left, $rbd_ch_left) = @{$new_curr_bds_gblks->[$gb_ch_left_NEW]}; 

	if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {

		# If $gb_kpt_NEW horizontally includes $gb_ch_left_NEW. #
		# => Attempt to UN-NEST them, by moving $gb_ch_left_NEW. #

   (print STDOUT "\n[INFORMATION 'vsplit_plus_smerge_gblock_pair']  The \$gb_kpt_NEW (=$gb_kpt_NEW), with [lbd, rbd] = [$lbd_kpt, $rbd_kpt], Horizontally INCLUDES  the \$gb_ch_left_NEW (=$gb_ch_left_NEW), with [lbd, rbd] = [$lbd_ch_left, $rbd_ch_left]!!\n") if ($TEST_LEVEL_vssm_gbp > 0);
   (print STDOUT "                      ... I will Attempt to UN-NEST them via \"serial shifts\" of \$gb_ch_left_NEW (=$gb_ch_left_NEW) ...\n\n") if ($TEST_LEVEL_vssm_gbp > 0);

   if ($TEST_LEVEL_vssm_gbp >5) { # (M2post-B') #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >5) { # (M2post-B') # ...}"

	    my $curr_coord_ch_left = $new_curr_bl_coords->[$gb_ch_left_NEW];
	    my ($lbd_coord_ch_left, $rbd_coord_ch_left) = @{$new_bds_gb_coords->[$gb_ch_left_NEW]};

	    my %skipped = ($gb_kpt_NEW => 1); # Because we already know that $gb_kpt_NEW is an "effective sibling" of $gb_ch_left_NEW. 

	    my $ct_nec_shifts_left = ($lbd_kpt > 0) ? ct_necessary_shifts ($lbd_ch_left, $lbd_kpt-1, $gb_ch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1 ;
	    if ($ct_nec_shifts_left > $curr_coord_ch_left - $lbd_coord_ch_left) { $ct_nec_shifts_left = $ct_out_clms +1; }

	    my $ct_nec_shifts_right = ($rbd_kpt < $ct_out_clms-1) ? ct_necessary_shifts ($rbd_ch_left, $rbd_kpt+1, $gb_ch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1;
	    if ($ct_nec_shifts_right > $rbd_coord_ch_left - $curr_coord_ch_left) { $ct_nec_shifts_right = $ct_out_clms +1; }

	    my ($direction, $ct_nec_shifts) = ($ct_nec_shifts_left < $ct_nec_shifts_right) ?
			('L', $ct_nec_shifts_left ) :  ('R', $ct_nec_shifts_right) ;

	    unless ($ct_nec_shifts < $ct_out_clms) {
		print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The \$gb_ch_left_NEW (=$gb_ch_left_NEW) caNNOT be UN-NESTED from \$gb_kpt_NEW (=$gb_kpt_NEW): $! !\n\n";
		return (-1, undef,  undef, undef);
	    }

	    my $incr_coord_ch_left_via_ssh = ($direction eq 'R') ? $ct_nec_shifts : - $ct_nec_shifts ;

   if ($TEST_LEVEL_vssm_gbp >2) {
	print STDOUT "\nIntended increment of the coordinate of \$gb_ch_left_NEW (=$gb_ch_left_NEW): \$incr_coord_ch_left_via_ssh = $incr_coord_ch_left_via_ssh .\n\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >2) {...}"

	    my ($flag_ssh, $incr_ct_null_clms_ssh)
	     = serial_shifts_sgl_gblock (
	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	        $gb_ch_left_NEW, $incr_coord_ch_left_via_ssh, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	    	);
	    unless ((defined $flag_ssh) and ($flag_ssh == 0)) {
	        my $flag_out = (defined $flag_ssh) ? $flag_ssh : 'undef';
	        print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the \$gb_ch_left_NEW (=$gb_ch_left_NEW) th gap-block (w/ \$flag_ssh = $flag_out): $! !\n\n";

	        return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_ch_left_NEW (=$gb_ch_left_NEW) th gap-block (w/ \$incr_ct_null_clms_ssh = $incr_ct_null_clms_ssh) ...\n\n") if ($TEST_LEVEL_vssm_gbp >1);

   if ($TEST_LEVEL_vssm_gbp >5) { # (M2post-B'') #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'serial_shifts_sgl_gblock' on \$gb_ch_left_NEW (=$gb_ch_left_NEW) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >5) { # (M2post-B'') # ...}"

			# Update %{$gb2incr_coord_undo_af_smg}. #
			# (ADDED in revision (on 2019/10/13).) #

	    if (defined $gb2incr_coord_undo_af_smg->{$gb_ch_left_NEW}) {
		$gb2incr_coord_undo_af_smg->{$gb_ch_left_NEW} += $incr_coord_ch_left_via_ssh;
	    } else {
		$gb2incr_coord_undo_af_smg->{$gb_ch_left_NEW} = $incr_coord_ch_left_via_ssh;
	    }
			# End of "(ADDED in revision (on 2019/10/13).)" #

	} # End of "if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {...}" # MOVED from below the 2nd 'undo_hovlp_vcmpl_gbsets_1sbj' (on 2019/10/13). #


	    	## Attempt to UNDO the horizontal overlaps due to the "serial shifts", etc. ##

	my ($fl_undo_af_ssh, $gb2incr_coord_undo_af_ssh)
	   = undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch_left_NEW,
	            @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	            @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	            @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	            );
	unless ($fl_undo_af_ssh) {
	        print STDOUT "[ERROR 'vsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch_left_NEW (=$gb_ch_left_NEW)) on the cofiguration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_ch_left_NEW itself: $! !\n\n";
	        return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch_left_NEW (=$gb_ch_left_NEW)), Ended (w/ \$fl_undo_af_ssh = $fl_undo_af_ssh) on the configuration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_ch_left_NEW itself ...\n\n") if ($TEST_LEVEL_vssm_gbp >1);

   if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A'') #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch_left_NEW (=$gb_ch_left_NEW)): ]\n\n";

	print STDOUT "\$fl_undo_af_ssh = $fl_undo_af_ssh .\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_ssh} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_ssh}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_ssh->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A'') # ...}"

	unless ($fl_undo_af_ssh > 0) {
		print STDOUT "[CAUTION 'vsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_ch_left_NEW (=$gb_ch_left_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
		return (-1, undef, undef, undef);
	}

=DISABLED in revision (on 2019/10/13). 
			# First, Update %{$gb2incr_coord_undo_af_ssh}. #

	    if (defined $gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW}) {
		$gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW} += $incr_coord_ch_left_via_ssh;
	    } else {
		$gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW} = $incr_coord_ch_left_via_ssh;
	    }
=cut # END of "DISABLED in revision (on 2019/10/13)."

			# Then, Update %{$gb2incr_coord_undo_af_smg} again. #

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_ssh}) {
		my $incr_af_ssh = $gb2incr_coord_undo_af_ssh->{$gb};
		if (defined $gb2incr_coord_undo_af_smg->{$gb}) {
		    $gb2incr_coord_undo_af_smg->{$gb} += $incr_af_ssh;
		} else {
		    $gb2incr_coord_undo_af_smg->{$gb} = $incr_af_ssh;
		}
	}

   if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A''') #

	print STDOUT "\n";
	print STDOUT '[ FINAL %{$gb2incr_coord_undo_af_smg}: ]', "\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_smg} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_smg}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_smg->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 5) { # (M2post-A''') # ...}"


	##} # End of "if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {...}" # MOVED to above the 2nd 'undo_hovlp_vcmpl_gbsets_1sbj' (on 2019/10/13). #

			# End of "ADDED in REVISION (on 2019/09/06)." #


   if ($TEST_LEVEL_vssm_gbp >3) { # (M2post-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Vertical Split + Merge (same-type)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp >3) { # (M2post-B) # ...}"

   if ($TEST_LEVEL_vssm_gbp > 3) { # (M2post-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Vertical Split + Merge (same-type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'vsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'vsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 3) { # (M2post-C) # ...}"



		# MAIN 3: Prepare Some Important Additional Sets of Data. #

	my @new_indices_old_gbs = my @old_indices_new_gbs = ();

	for (my $igb=0; $igb < $ct_gbs_intm; $igb++) {
	    my $ogb = $old_indices_intm_gbs->[$igb];
	    my $ngb = $new_indices_intm_gbs->[$igb];
	    if (defined $ogb) { $new_indices_old_gbs[$ogb] = $ngb; }
	    if (defined $ngb) { $old_indices_new_gbs[$ngb] = $ogb; }
	}

	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $igb = $intm_indices_new_gbs->[$ngb];
	    my @info_gb = @{$intm_list_gblocks->[$igb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_gb;
	}

	my @new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']);


   if ($TEST_LEVEL_vssm_gbp > 3) { # (M3) #

	print STDOUT "\n[[ Additional Output Data-Sets: ]]\n\n";

	print STDOUT '($gb_ch_left_NEW, $gb_kpt_NEW) = ', "($gb_ch_left_NEW, $gb_kpt_NEW) .\n\n\n";

	print STDOUT '[ Relationships between new and old indices of the gap-blocks', " (#{gap-blocks}_new=$ct_out_gbs, #{gap-blocks}_old=$ct_in_gbs): ]\n\n";

        my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

        foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
                          [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

        print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

        print STDOUT "\n\n";

	print STDOUT '[ Contents of @new_list_gblocks', " (#{gap-blocks}_new = $ct_out_gbs): ]\n\n";

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_out_gbs; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

	print STDOUT '[ Contents of @new_list_excluded_bin_hrels: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";

	for (my $i=0; $i < @new_list_excluded_bin_hrels; $i++) {
	    print STDOUT join ("\t", $i, @{$new_list_excluded_bin_hrels[$i]}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_vssm_gbp > 3) { # (M3) # ...}"


   (print STDOUT "< Subroutine 'vsplit_plus_smerge_gblock_pair' Ended (w/ \$fl_end = 1, \$gb_ch_left_NEW = $gb_ch_left_NEW, \$gb_kpt_NEW = $gb_kpt_NEW) />\n\n") if ($TEST_LEVEL_vssm_gbp > 0);

	# [ return (1,  $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ] ... for the predecessor, 'vsplit_gblock'. #

			# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT) #
	return (1, $gb_ch_left_NEW, $gb_kpt_NEW, 
	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	$gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	\@new_indices_old_gbs, \@old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
			# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT)" #

	$new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	$new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	\@new_list_gblocks, \@new_list_excluded_bin_hrels
	);

}	# End of the subroutine, 'vsplit_plus_smerge_gblock_pair'. #


# cvsplit_plus_smerge_gblock_pair:
#	This subroutine "vertically split"s the complementary residue-block of the 1st subject gap-block into a pair of residue-blocks
#	that are its "effective children" & do some adjusting "shift"s, via 'cvsplit_gblock';
#	then, the subroutine "merge"s the 2nd subject gap-block into the complementary gap-block of one of the resulting "effective child" residue blocks that is of the same type as the 2nd subject, via 'smerge_gblock_pair'.
#	It was thus created by combining 'cvsplit_gblock' & 'smerge_gblock_pair',
#	and by adding some auxiliary functions.
#	[CREATION STARTED on 2019/09/10, ENDED on 2019/09/dd.]
#
#	[NOTE (on 2019/09/10)]: This subroutine should NOT HAVE TO attempt to UN-NEST the resulting gap-blocks, because they must be in the 'ONCS' relationship to each other.
#
# vsplit_plus_smerge_gblock_pair:
#	This subroutine "vertically split"s the 1st subject gap-block into a pair of gap-blocks
#	that are its "effective children" & do some adjusting "shift"s, via 'vsplit_gblock';
#	then, the subroutine "merge"s the 2nd subject gap-block into one of the resulting "effective children" that is of the same type as the 2nd subject, via 'smerge_gblock_pair'.
#	It was thus created by combining 'vsplit_gblock' & 'smerge_gblock_pair',
#	and by adding some auxiliary functions.
#	[CREATION STARTED on 2019/09/02, ENDED on 2019/09/05.]
#	[REVISED on 2019/09/06]:
#	Now, the subroutine attempts to UN-NEST the resulting gap-blocks if they are nested after the "vsplit + smerge".
#	 
# vsplit_gblock:
#	This subroutine "vertically split"s the subject gap-block into a pair of gap-blocks that are the "effective children" of the subject,
#	then, it "shift"s one of the resulting child blocks to that they will NOT be nested.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of some parts of the framework of 'cvmerge_gblock_pair' above.
#	[CREATION STARTED on 2019/08/22, ENDED on 2019/08/23.]
#	
# [ INPUTS: ($gb_sbj, \@info_effch1, \@info_effch2, # MODIFIED in creation of 'vsplit_gblock'. (HD1a) # ] ... for the predecessor, 'vsplit_gblock'. (HD1)
#
# [ INPUTS: ($gb_sbj_pa, $gb_sbj_ch, # MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1a) # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (HD1a)
#
# INPUTS: ($gb_sbj_ch, $gb_sbj_pa, $char_abs, # MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD1a) #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b) #
#	\@set_all_exnodes, \%nodeid2msa_row
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b)" #
#	), where
#
#		# MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2a) #
#	$gb_sbj_ch and $gb_sbj_pa are the indices of the subject gap-blocks
#	that are an effective child and an effective parent, respectively, to each other;
#
#	$char_abs is a string representing an "absence" state (, i.e., a gap);
#
#		# End of "MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2a)" #
#
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b) #
#	@set_all_exnodes stores the indices of all external nodes;
#
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b)" #
#	
#
# [ OUTPUTS: ($flag_end, $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ## MODIFIED in creation of 'vsplit_gblock' (on 2019/08/22). (HD2a) # ] ... for the predecessor, 'vsplit_gblock'. (HD2a) 
#
#			# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a) #
# [ OUTPUTS: ($flag_end, $gb_ch_left_NEW, $gb_kpt_NEW, 
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, \@intm_indices_old_gbs, \%gb2incr_bl_coord_undo_ovlps_af_vspl,
#	$gb_ch_2B_mgd_intm, ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (HD2a)
#
#			# Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2a) #
# OUTPUTS: ($flag_end, $gb_cch_left_NEW, $gb_kpt_NEW, 
#	\%gb2incr_coord_1way_dsplc_bf_cvspl, \@intm_indices_old_gbs, \%gb2incr_bl_coord_undo_ovlps_af_cvspl,
#	$gb_cch_2B_mgd_intm,
#			# End of "Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2a)" #
#
#	$gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, \@new_indices_intm_gbs,
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_coord_undo_af_smg,
#			# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a)" #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	## [ @new_list_excluded_bin_hrels # ADDED in creation of 'vsplit_gblock'. (HD1a) # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (HD1a)
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the horizontal overlaps resulting from either 'cvsplit' or 'smerge' are NOT UN-DONE completely;
#
#			# MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2b-i) #
#
#	$gb_cch_left_NEW is the NEW index of the complementary gap-block of the "effective-child" residue-block that resulted from "vertically split"ting the residue block complementary to $gb_sbj_ch AND that was LEFT unmerged to $gb_sbj_pa;
#	$gb_kpt_NEW is the NEW index of the gap-block resulting from the (horizontal) "merge" of $gb_sbj_pa and one of the products of "complementarily vertically split"ting $gb_sbj_ch;
#	[These are two of the MAIN-AUXILIARY outputs of this subroutine];
#
#	%gb2incr_coord_1way_dsplc_bf_cvspl = ($indx_gb_intm => {increment of the block coordinate resulting from the DISPLACEMENT prior to the "complementary vertical split"}, ...) for the gap-blocks whose displacement canNOT be undone; (Such gap-blocks MUST BE vertically identical to $gb_sbj_ch); $indx_gb_intm SHOULD BE the index in the INTERMEDIATE gap-block set;

#	@intm_indices_old_gbs stores the "intermediate" indices, after the 'cvsplit_gblock', of the "old" gap-blocks before the 'cvsplit_gblock';
#	%gb2incr_bl_coord_undo_ovlps_af_vspl = ($indx_gb_intm => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "complementary vertical split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'cvsplit_gblock'];
#
#	$gb_cch_2B_mgd_intm is the index (in the INTERMEDIATE gap-block set) of the product of "complementarily vertically split"ting $gb_sbj_pa that was then "merge"d with $gb_sbj_pa (in terms of the old block set);
#	($gb1_intm, $gb2_intm) are just ($gb_cch_2B_mgd_intm, $gb_sbj_pa_intm) sorted in ascending order;
#	[These were created in preparation of 'smerge_gblock_pair'];
#
#			# END of "MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. (HD2b-i)" #
#
#	[[
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i) #
#	$gb_ch_left_NEW is the NEW index of the "effective-child" gap-block that resulted from "vertically split"ting $gb_sbj_pa AND that was LEFT unmerged to $gb_sbj_ch;
#	$gb_kpt_NEW is the NEW index of the gap-block resulting from the (horizontal) "merge" of $gb_sbj_ch and one of the products of "vertically split"ting $gb_sbj_pa;
#	[These are two of the MAIN-AUXILIARY outputs of this subroutine];
#
#	$gb_shftd_af_vspl_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was "shift"ed to make the product children non-nested;
#	$incr_coord_shftd_af_vspl (= +1 or -1) is the increment of the coordinate of $gb_shftd_af_vspl_intm via the aforementioned "shift"; 
#	@intm_indices_old_gbs stores the "intermediate" indices, after the 'vsplit_gblock', of the "old" gap-blocks before the 'vsplit_gblock';
#	%gb2incr_bl_coord_undo_ovlps_af_vspl = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "vertical split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'vsplit_gblock'];
#
#	$gb_ch_2B_mgd_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was then "merge"d with $gb_sbj_ch (in terms of the old block set);
#	($gb1_intm, $gb2_intm) are just ($gb_ch_2B_mgd_intm, $gb_sbj_ch_intm) sorted in ascending order;
#	[These were created in preparation of 'smerge_gblock_pair'];
#	]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
#	$bl_sbj_ssh4smg_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "shift"ed before their merger;
#	$incr_coord_sbj_ssh4smg is the increment of the coordinate of $bl_sbj_ssh4smg via the aforementioned "shift";
#	$gb_absbd_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "absorted" as a result of the 'smerge_gblock_pair';
#	@new_indices_intm_gbs stores the "new" indices, after the 'smerge_gblock_pair', of the "intermediate" gap-blocks before the "smerge_gblock_pair";
#	[These are the AUXILIARY outputs of 'smerge_gblock_pair'];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i)" #
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "cvsplit + smerge";
#	More precisely, the gap-block that had been the $ogb th before the "cvsplit + smerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "cvsplit + smerge"
#	(, and $new_indices_old_gbs[$ogb] is undefined for some $ogb);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "cvsplit + smerge";
#	More precisely, the gap-block that is the $ngb th after the "cvsplit + smerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "cvsplit + smerge";
#
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii) #
#	%gb2incr_coord_undo_af_smg = ($indx_gb_new => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the 'smerge_gblock_pair'}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_smg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "complementary-vertical-split";
#
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#
#	[[
#		# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD3) #
#	@new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']);
#		# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD3)" #
#	is the NEW list of PROHIBITED binary horizontal relationships,
#	which conforms to the format prescribed in the header of 'nonrec_mlt_shifts_gapblocks' above;
#	]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
# [ sub vsplit_plus_smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%) {
#
#	# [ my ($gb_sbj, $info_effch1, $info_effch2, ] ... for the predecessor, 'vsplit_gblock'. #
#	my ($gb_sbj_pa, $gb_sbj_ch, # Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
sub cvsplit_plus_smerge_gblock_pair ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%) {

	my ($gb_sbj_ch, $gb_sbj_pa, $char_abs, # Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (IN) #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq,
		# Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) #
	$set_all_exnodes, $nodeid2msa_row
		# End of "Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN)" #
	) = @_; 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	unless (($ct_in_clms>0) and ($ct_in_gbs>0)) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] ILLEGAL INPUT: (#{columns}, #{gap-blocks}) = ($ct_in_clms, $ct_in_gbs) in the input MSA: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $ct_rows = @{$curr_set_clms->[0]};

   (print STDOUT "< Subroutine 'cvsplit_plus_smerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; effective child-parent pair of gap-blocks to be \"cvsplit + smerge\"d: (\$gb_sbj_ch, \$gb_sbj_pa) = ($gb_sbj_ch, $gb_sbj_pa); \$char_abs = '$char_abs' ) >\n\n") if ($TEST_LEVEL_cvssm_gbp > 0);


   if ($TEST_LEVEL_cvssm_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Complementary Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) {...}"


   if ($TEST_LEVEL_cvssm_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Complementary Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >3) {...}"


			### Preliminary: Virtually VSplit the Residue-Block Complementary to $gb_sbj_ch into TWO effective children. ###

	my ($br_sbj_ch, $up_or_lw_sbj_ch) = @{$list_gblocks0->[$gb_sbj_ch]}[$indx_br, $indx_u_or_d];

	my ($set_ceff_child_rblks) = vsplit_into_cmpl_eff_child_rblocks ($br_sbj_ch, $up_or_lw_sbj_ch,
		 %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, @{$set_all_exnodes},
		 %{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw}) ;
	unless (defined $set_ceff_child_rblks) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'vsplit_into_cmpl_eff_child_rblocks' FAILED on (\$br_sbj_ch, \$up_or_lw_sbj_ch) = ($br_sbj_ch, $up_or_lw_sbj_ch), for \$gb_sbj_ch (=$gb_sbj_ch) : $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless (@{$set_ceff_child_rblks} == 2) {
	    my $ct_cechs = @{$set_ceff_child_rblks};
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] SOMEHOW, \$gb_sbj_ch (=$gb_sbj_ch) has $ct_cechs (!= 2) effective child sets of msa rows: $! !\n\n";
	    return (0, undef, undef, undef);
	}

   if ($TEST_LEVEL_cvssm_gbp >2) { # (Pre) #

	print STDOUT "\n[ Outputs of 'vsplit_into_cmpl_eff_child_rblocks': ]\n\n";

	print STDOUT join ("\t", 'C-Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(cmpl,external,nodes,affected)', '[cmpl,MSA,rows,affected]'), "\n\n";

	for (my $i=0; $i <2; $i++) {
	    my $ccid = $i+1;
	    my ($br, $u_or_l, $cexnodes_affd, $crows_affd) = @{$set_ceff_child_rblks->[$i]};
	    my $cnct_cexnodes_affd = join (',', @{$cexnodes_affd});
	    my $cnct_crows_affd = join (',', @{$crows_affd});

	    print STDOUT join ("\t", "CCh_$ccid", $br, $u_or_l, "($cnct_cexnodes_affd)", "[$cnct_crows_affd]"), "\n";
	} # End of "for (my $i=0; $i <2; $i++) {...}" #

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) { # (Pre) # ...}"


		### MAIN 1: "Vertically Split" the residue-block complementary to $gb_sbj_ch. ###

	my ($info_ceffch1, $info_ceffch2) = @{$set_ceff_child_rblks};


	## [ my ($flag_vspl, $gb_ch1_intm, $gb_ch2_intm, $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $old_indices_intm_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl, ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. ##

	my ($flag_cvspl, $gb_cch1_intm, $gb_cch2_intm, $gb2incr_coord_1way_dsplc_bf_cvspl, 
$intm_indices_old_gbs, $old_indices_intm_gbs, $gb2incr_bl_coord_undo_ovlps_af_cvspl, # MODIFIED in creation of 'cvsplit_plus_smerge_gblock_pair'. #

	    $intm_curr_set_clms, $intm_curr_bds_gblks, $intm_curr_bl_coords, $intm_bds_gb_coords, $intm_gb_coords_wd_frm,
	    $intm_curr_gblks_w_spec_lb, $intm_curr_gblks_w_spec_rb, $rows_affd_by_intm_blks,
	    $intm_inter_block_relations, $intm_cts_vincluding, $intm_cts_vincluded, $intm_cts_onx, $intm_gblock_sizes,
	    $intm_list_gblocks, 
	    ## [ $intm_list_excluded_bin_hrels ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. ##
		)
	  = cvsplit_gblock (
	     $gb_sbj_ch, @{$info_ceffch1}, @{$info_ceffch2}, $char_abs,
	     @{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm},
	     @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
	     @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes},  $maxct_1way_shifts,
	     @{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
		);
	unless ($flag_cvspl) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The 1st MAIN Subroutine 'cvsplit_gblock' FAILED on the \$gb_sbj_ch (=$gb_sbj_ch) th gap-block: $! !\n";
	    return (0, undef, undef, undef);
	}
	unless ($flag_cvspl > 0) {
	    print STDOUT "[CAUTION 'cvsplit_plus_smerge_gblock_pair'] The 1st MAIN subroutine, 'vsplit_gblock' FAILED on \$gb_sbj_ch (=$gb_sbj_ch), because the resulting horizontal overlaps canNOT be UNDONE completely!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}
   (print STDOUT "\n --- The 1st MAIN Subroutine 'cvsplit_gblock' ENDED on the \$gb_sbj_ch (=$gb_sbj_ch) th gap-block (w/ \$flag_cvspl = $flag_cvspl, \$gb_cch1_intm = $gb_cch1_intm, \$gb_cch2_intm = $gb_cch2_intm) ...\n\n\n") if ($TEST_LEVEL_cvssm_gbp >1);


	my $ct_gbs_intm = @{$intm_curr_bds_gblks};

   if ($TEST_LEVEL_cvssm_gbp >2) { # (M1-A) #

	print STDOUT "\n[[ Outputs of 'cvsplit_gblock' ]]\n\n";

        print STDOUT "[ Relationships between old and intermediate indices of gap-blocks (#{gap-blocks}_intm=$ct_gbs_intm) : ]\n\n";

        my @out_intm_indices_old_gbs = my @out_old_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_old_gbs, \@out_intm_indices_old_gbs],
                          [$old_indices_intm_gbs, \@out_old_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$intm_indices_old_gbs} = (', join (', ', @out_intm_indices_old_gbs), ") .\n\n";

        print STDOUT '@{$old_indices_intm_gbs} = (', join (', ', @out_old_indices_intm_gbs), ") .\n";

        print STDOUT "\n\n";

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'.
	print STDOUT '[ Contents of @{$intm_list_excluded_bin_hrels}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";
	for (my $i=0; $i < @{$intm_list_excluded_bin_hrels}; $i++) {
	    print STDOUT join ("\t", $i, @{$intm_list_excluded_bin_hrels->[$i]}), "\n";
	}
	print STDOUT "\n\n";
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'."

			# Added in creation of 'cvsplit_plus_smerge_gblock_pair'. #
	print STDOUT '[ Contents of %{$gb2incr_coord_1way_dsplc_bf_cvspl}: ]', "\n\n";

	print STDOUT '%{$gb2incr_coord_1way_dsplc_bf_cvspl} = ( # $indx_gb_intm => $incr_coord_1way_dsplc,', "\n";
	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_1way_dsplc_bf_cvspl}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_1way_dsplc_bf_cvspl->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";
			# End of "Added in creation of 'cvsplit_plus_smerge_gblock_pair'." #

			# Added on 2019/09/06. #
	print STDOUT '[ Contents of %{$gb2incr_bl_coord_undo_ovlps_af_cvspl}: ]', "\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo_ovlps_af_cvspl} = ( # $indx_gb_intm => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo_ovlps_af_cvspl}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo_ovlps_af_cvspl->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/09/06." #

   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) { # (M1-A) #...}"

   if ($TEST_LEVEL_cvssm_gbp >5) { # (M1-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the \"Complementary Vertical Split\" ]\n\n";

	unless ( print_out_gblock_configs (@{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >5) { # (M1-B) # ...}"

   if ($TEST_LEVEL_cvssm_gbp > 5) { # (M1-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the \"Complementary Vertical Split\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm_inter_block_relations}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm}, @{$intm_gblock_sizes}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$rows_affd_by_intm_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) { # (M1-C) # ...}"

   if ($TEST_LEVEL_cvssm_gbp >= 5) { # (M1-D) #

	print STDOUT "\n";
        print STDOUT '[ Contents of @{$intm_list_gblocks}: ]', "\n\n";

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_gbs_intm; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks->[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >= 5) { # (M1-D) #...}"



		### MAIN 2pre: Identify the resulting "complementary effective child" to be "merge"d with $gb_sbj_pa_intm. ### 

	my $gb_sbj_pa_intm = $intm_indices_old_gbs->[$gb_sbj_pa];
	my $rows_affd_by_sbj_pa = $rows_affd_by_intm_blks->[$gb_sbj_pa_intm];

   if ($TEST_LEVEL_cvssm_gbp >2) { # (M2Pre-A) #
	print STDOUT "\n";
	print STDOUT "\$gb_sbj_pa_intm = $gb_sbj_pa_intm .\n\n";
	print STDOUT '@{$rows_affd_by_sbj_pa} = (', join (', ', @{$rows_affd_by_sbj_pa}), ") .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) { # (M2Pre-A) #...}"

	my $gb_cch_2B_mgd_intm;
	foreach my $gb_intm ($gb_cch1_intm, $gb_cch2_intm) {
	    my $rows_affd_by_gb = $rows_affd_by_intm_blks->[$gb_intm];
	    my ($common, $ex1, $ex2) = diff_nmarray (@{$rows_affd_by_sbj_pa}, @{$rows_affd_by_gb});
	    unless (defined $common) {
		print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'diff_nmarray' FAILED on the rows affected by \$gb_sbj_pa_intm (=$gb_sbj_pa_intm) and those affected by \$gb_intm (=$gb_intm): $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    if ((@{$ex1}==0) and (@{$ex2}==0)) { $gb_cch_2B_mgd_intm = $gb_intm; }

	} # End of "foreach my $gb_intm ($gb_cch1_intm, $gb_cch2_intm) {...}". #

	unless (defined $gb_cch_2B_mgd_intm) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The \$gb_cch_2B_mgd_intm is NOT DEFINED: $! !\n";
	    return (0, undef, undef, undef);
	}


	my $gb_cch_left_intm = ($gb_cch_2B_mgd_intm == $gb_cch1_intm) ? $gb_cch2_intm : $gb_cch1_intm ;

   if ($TEST_LEVEL_cvssm_gbp >1) { # (M2Pre-B) #
	print STDOUT "\n";
	print STDOUT "(\$gb_cch_2B_mgd_intm, \$gb_cch_left_intm) = ($gb_cch_2B_mgd_intm, $gb_cch_left_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_cvssm_gbp >1) { # (M2Pre-B) #...}"


	my ($gb1_intm, $gb2_intm) = ($gb_sbj_pa_intm < $gb_cch_2B_mgd_intm) ?
		($gb_sbj_pa_intm, $gb_cch_2B_mgd_intm) : ($gb_cch_2B_mgd_intm, $gb_sbj_pa_intm) ;

   if ($TEST_LEVEL_cvssm_gbp >2) { # (M2Pre-C) #
	print STDOUT '($gb1_intm, $gb2_intm) = ', "($gb1_intm, $gb2_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) { # (M2Pre-C) #...}" # A BUG fixed (on 2019/09/05).
   #} if ($TEST_LEVEL_cvssm_gbp >2) { # (M2Pre-C) #


		### MAIN 2: "Merge" $gb1_intm and $gb2_intm (i.e., $gb_sbj_pa_intm and $gb_cch_2B_mgd_intm). ###

	my ($flag_smg, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $gb_kpt_intm, $new_indices_intm_gbs, $intm_indices_new_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = smerge_gblock_pair (
	     $gb1_intm, $gb2_intm, @{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm},
	     @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, @{$rows_affd_by_intm_blks}, $cnct_null_clm,

	     @{$intm_inter_block_relations}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$intm_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_smg) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The 2nd MAIN Subroutine 'smerge_gblock_pair' FAILED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) in the Configuration AFTER the \"Complementary Vertical Split\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n --- The 2nd MAIN Subroutine 'smerge_gblock_pair' ENDED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) AFTER the \"complementary vertical split\" (w/ \$bl_sbj_ssh4smg_intm = $bl_sbj_ssh4smg_intm, \$incr_coord_sbj_ssh4smg = $incr_coord_sbj_ssh4smg, \$gb_absbd_intm = $gb_absbd_intm, \$gb_kpt_intm = $gb_kpt_intm) ...\n\n\n") if ($TEST_LEVEL_cvssm_gbp >0);

	my $ct_out_gbs = @{$new_curr_bds_gblks};

	my $ct_out_clms = @{$new_curr_set_clms}; # ADDED in REVISION (on 2019/09/06). #


   if ($TEST_LEVEL_cvssm_gbp >2) { # (M2-A) #

	print STDOUT "\n[[ Outputs of 'smerge_gblock_pair' ]]\n\n";

        print STDOUT "[ Relationships between new and intermediate indices of gap-blocks (#{gap-blocks}_new=$ct_out_gbs) : ]\n\n";

        my @out_intm_indices_new_gbs = my @out_new_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_new_gbs, \@out_intm_indices_new_gbs],
                          [$new_indices_intm_gbs, \@out_new_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$new_indices_intm_gbs} = (', join (', ', @out_new_indices_intm_gbs), ") .\n\n";
        print STDOUT '@{$intm_indices_new_gbs} = (', join (', ', @out_intm_indices_new_gbs), ") .\n";
        print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) { # (M2-A) # ...}"


   if ($TEST_LEVEL_cvssm_gbp >5) { # (M2-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'smerge_gblock_pair': ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >5) { # (M2-B) # ...}"

   if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'smerge_gblock_pair': ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2-C) # ...}"


		# MAIN 2post: If any overlaps between vertically equivalent and/or complementary gap-blocks result from the 'smerge_gblock_pair', undo them. #

	my $gb_kpt_NEW = $new_indices_intm_gbs->[$gb_kpt_intm]; # The new index for the gap-block resulting from the "smerge". #
	my $gb_cch_left_NEW = $new_indices_intm_gbs->[$gb_cch_left_intm]; # The "effective-child" block that was left unmerged. # MOVED from Main 3 below, in REVISION (on 2019/09/06). #

	my ($fl_undo_af_smg, $gb2incr_coord_undo_af_smg)
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_kpt_NEW,
	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	        @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	        );
	unless ($fl_undo_af_smg) {
	    print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)) on the cofiguration of gap-blocks after the 'smerge_gblock_pair' on (\$gb_sbj_pa_intm, \$gb_cch_2B_mgd_intm) = ($gb_sbj_pa_intm, $gb_cch_2B_mgd_intm): $! !\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)), Ended (w/ \$fl_undo_af_smg = $fl_undo_af_smg) on the configuration of gap-blocks after the 'smerge_gblock_pair' on (\$gb_sbj_pa_intm, \$gb_cch_2B_mgd_intm) = ($gb_sbj_pa_intm, $gb_cch_2B_mgd_intm) ...\n\n") if ($TEST_LEVEL_cvssm_gbp >1);


   if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A) #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)): ]\n\n";

	print STDOUT "\$fl_undo_af_smg = $fl_undo_af_smg .\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_smg} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_smg}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_smg->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A) # ...}"

	unless ($fl_undo_af_smg > 0) {
	    print STDOUT "[CAUTION 'cvsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}

			# ADDED in REVISION (on 2019/09/06). #

	my ($lbd_kpt, $rbd_kpt) = @{$new_curr_bds_gblks->[$gb_kpt_NEW]};
	my ($lbd_cch_left, $rbd_cch_left) = @{$new_curr_bds_gblks->[$gb_cch_left_NEW]}; 

	if (($lbd_kpt <= $lbd_cch_left) and ($rbd_cch_left <= $rbd_kpt)) {

		# [NOTE] Actually, This subroutine (i.e., 'cvsplit_plus_smerge_gblock_pair')
		#	 should NEVER step into this region!!

		# If $gb_kpt_NEW horizontally includes $gb_cch_left_NEW. #
		# => Attempt to UN-NEST them, by moving $gb_cch_left_NEW. #

   (print STDOUT "\n[INFORMATION 'cvsplit_plus_smerge_gblock_pair']  The \$gb_kpt_NEW (=$gb_kpt_NEW), with [lbd, rbd] = [$lbd_kpt, $rbd_kpt], Horizontally INCLUDES  the \$gb_cch_left_NEW (=$gb_cch_left_NEW), with [lbd, rbd] = [$lbd_cch_left, $rbd_cch_left]!!\n") if ($TEST_LEVEL_cvssm_gbp > 0);
   (print STDOUT "                      ... I will Attempt to UN-NEST them via \"serial shifts\" of \$gb_cch_left_NEW (=$gb_cch_left_NEW) ...\n\n") if ($TEST_LEVEL_cvssm_gbp > 0);

   if ($TEST_LEVEL_cvssm_gbp >5) { # (M2post-B') #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >5) { # (M2post-B') # ...}"

	    my $curr_coord_cch_left = $new_curr_bl_coords->[$gb_cch_left_NEW];
	    my ($lbd_coord_cch_left, $rbd_coord_cch_left) = @{$new_bds_gb_coords->[$gb_cch_left_NEW]};

	    my %skipped; # $gb_kpt_NEW should be in 'ONCS' relation with $gb_cch_left_NEW. #
	    ##my %skipped = ($gb_kpt_NEW => 1); # Because we already know that $gb_kpt_NEW is an "effective sibling" of $gb_ch_left_NEW. 

	    my $ct_nec_shifts_left = ($lbd_kpt > 0) ? ct_necessary_shifts ($lbd_cch_left, $lbd_kpt-1, $gb_cch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1 ;
	    if ($ct_nec_shifts_left > $curr_coord_cch_left - $lbd_coord_cch_left) { $ct_nec_shifts_left = $ct_out_clms +1; }

	    my $ct_nec_shifts_right = ($rbd_kpt < $ct_out_clms-1) ? ct_necessary_shifts ($rbd_cch_left, $rbd_kpt+1, $gb_cch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1;
	    if ($ct_nec_shifts_right > $rbd_coord_cch_left - $curr_coord_cch_left) { $ct_nec_shifts_right = $ct_out_clms +1; }

	    my ($direction, $ct_nec_shifts) = ($ct_nec_shifts_left < $ct_nec_shifts_right) ?
			('L', $ct_nec_shifts_left ) :  ('R', $ct_nec_shifts_right) ;

	    unless ($ct_nec_shifts < $ct_out_clms) {
		print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The \$gb_cch_left_NEW (=$gb_cch_left_NEW) caNNOT be UN-NESTED from \$gb_kpt_NEW (=$gb_kpt_NEW): $! !\n\n";
		return (-1, undef,  undef, undef);
	    }

	    my $incr_coord_cch_left_via_ssh = ($direction eq 'R') ? $ct_nec_shifts : - $ct_nec_shifts ;

   if ($TEST_LEVEL_cvssm_gbp >2) {
	print STDOUT "\nIntended increment of the coordinate of \$gb_cch_left_NEW (=$gb_cch_left_NEW): \$incr_coord_cch_left_via_ssh = $incr_coord_cch_left_via_ssh .\n\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >2) {...}"

	    my ($flag_ssh, $incr_ct_null_clms_ssh)
	     = serial_shifts_sgl_gblock (
	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	        $gb_cch_left_NEW, $incr_coord_cch_left_via_ssh, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	    	);
	    unless ((defined $flag_ssh) and ($flag_ssh == 0)) {
	        my $flag_out = (defined $flag_ssh) ? $flag_ssh : 'undef';
	        print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the \$gb_cch_left_NEW (=$gb_cch_left_NEW) th gap-block (w/ \$flag_ssh = $flag_out): $! !\n\n";

	        return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_cch_left_NEW (=$gb_cch_left_NEW) th gap-block (w/ \$incr_ct_null_clms_ssh = $incr_ct_null_clms_ssh) ...\n\n") if ($TEST_LEVEL_cvssm_gbp >1);

   if ($TEST_LEVEL_cvssm_gbp >5) { # (M2post-B'') #

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'serial_shifts_sgl_gblock' on \$gb_cch_left_NEW (=$gb_cch_left_NEW) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >5) { # (M2post-B'') # ...}"


	    	## Attempt to UNDO the horizontal overlaps due to the "serial shifts". ##

	    my ($fl_undo_af_ssh, $gb2incr_coord_undo_af_ssh)
	      = undo_hovlp_vcmpl_gbsets_1sbj ($gb_cch_left_NEW,
	            @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	            @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	            @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	            );
	    unless ($fl_undo_af_ssh) {
	        print STDOUT "[ERROR 'cvsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch_left_NEW (=$gb_cch_left_NEW)) on the cofiguration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_cch_left_NEW itself: $! !\n\n";
	        return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch_left_NEW (=$gb_cch_left_NEW)), Ended (w/ \$fl_undo_af_ssh = $fl_undo_af_ssh) on the configuration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_cch_left_NEW itself ...\n\n") if ($TEST_LEVEL_cvssm_gbp >1);

   if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A'') #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_cch_left_NEW (=$gb_cch_left_NEW)): ]\n\n";

	print STDOUT "\$fl_undo_af_ssh = $fl_undo_af_ssh .\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_ssh} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_ssh}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_ssh->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A'') # ...}"

	    unless ($fl_undo_af_ssh > 0) {
		print STDOUT "[CAUTION 'cvsplit_plus_smerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_cch_left_NEW (=$gb_cch_left_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
		return (-1, undef, undef, undef);
	    }

			# First, Update %{$gb2incr_coord_undo_af_ssh}. #

	    if (defined $gb2incr_coord_undo_af_ssh->{$gb_cch_left_NEW}) {
		$gb2incr_coord_undo_af_ssh->{$gb_cch_left_NEW} += $incr_coord_cch_left_via_ssh;
	    } else {
		$gb2incr_coord_undo_af_ssh->{$gb_cch_left_NEW} = $incr_coord_cch_left_via_ssh;
	    }

			# Then, Update %{$gb2incr_coord_undo_af_smg}. #

	    foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_ssh}) {
		my $incr_af_ssh = $gb2incr_coord_undo_af_ssh->{$gb};
		if (defined $gb2incr_coord_undo_af_smg->{$gb}) {
		    $gb2incr_coord_undo_af_smg->{$gb} += $incr_af_ssh;
		} else {
		    $gb2incr_coord_undo_af_smg->{$gb} = $incr_af_ssh;
		}
	    }

   if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A''') #

	print STDOUT "\n";
	print STDOUT '[ FINAL %{$gb2incr_coord_undo_af_smg}: ]', "\n\n";

	print STDOUT '%{$gb2incr_coord_undo_af_smg} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_coord_undo_af_smg}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_coord_undo_af_smg->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 5) { # (M2post-A''') # ...}"


	} # End of "if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {...}"

			# End of "ADDED in REVISION (on 2019/09/06)." #


   if ($TEST_LEVEL_cvssm_gbp >3) { # (M2post-B) #

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Complementary Vertical Split + Merge (same-type)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp >3) { # (M2post-B) # ...}"

   if ($TEST_LEVEL_cvssm_gbp > 3) { # (M2post-C) #

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Complementary Vertical Split + Merge (same-type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'cvsplit_plus_smerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'cvsplit_plus_smerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 3) { # (M2post-C) # ...}"



		# MAIN 3: Prepare Some Important Additional Sets of Data. #

	my @new_indices_old_gbs = my @old_indices_new_gbs = ();

	for (my $igb=0; $igb < $ct_gbs_intm; $igb++) {
	    my $ogb = $old_indices_intm_gbs->[$igb];
	    my $ngb = $new_indices_intm_gbs->[$igb];
	    if (defined $ogb) { $new_indices_old_gbs[$ogb] = $ngb; }
	    if (defined $ngb) { $old_indices_new_gbs[$ngb] = $ogb; }
	}

	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $igb = $intm_indices_new_gbs->[$ngb];
	    my @info_gb = @{$intm_list_gblocks->[$igb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_gb;
	}

	## my @new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']); # For the predecessor, 'vsplit_plus_smerge_gblock_pair'. #


   if ($TEST_LEVEL_cvssm_gbp > 3) { # (M3) #

	print STDOUT "\n[[ Additional Output Data-Sets: ]]\n\n";

	print STDOUT '($gb_cch_left_NEW, $gb_kpt_NEW) = ', "($gb_cch_left_NEW, $gb_kpt_NEW) .\n\n\n";

	print STDOUT '[ Relationships between new and old indices of the gap-blocks', " (#{gap-blocks}_new=$ct_out_gbs, #{gap-blocks}_old=$ct_in_gbs): ]\n\n";

        my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

        foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
                          [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

        print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

        print STDOUT "\n\n";

	print STDOUT '[ Contents of @new_list_gblocks', " (#{gap-blocks}_new = $ct_out_gbs): ]\n\n";

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_out_gbs; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (FIN)
	print STDOUT '[ Contents of @new_list_excluded_bin_hrels: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";

	for (my $i=0; $i < @new_list_excluded_bin_hrels; $i++) {
	    print STDOUT join ("\t", $i, @{$new_list_excluded_bin_hrels[$i]}), "\n";
	}
	print STDOUT "\n\n";
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (FIN)"

   } # End of "if ($TEST_LEVEL_cvssm_gbp > 3) { # (M3) # ...}"


   (print STDOUT "< Subroutine 'cvsplit_plus_smerge_gblock_pair' Ended (w/ \$fl_end = 1, \$gb_cch_left_NEW = $gb_cch_left_NEW, \$gb_kpt_NEW = $gb_kpt_NEW) />\n\n") if ($TEST_LEVEL_cvssm_gbp > 0);

	# [ return (1,  $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ] ... for the predecessor, 'vsplit_gblock'. #

			# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT) #
=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (OUT1)
	return (1, $gb_ch_left_NEW, $gb_kpt_NEW, 
	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	$gb_ch_2B_mgd_intm, 
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (OUT1)"

			# Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (OUT) #
	return (1, $gb_cch_left_NEW, $gb_kpt_NEW, 
	$gb2incr_coord_1way_dsplc_bf_cvspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_cvspl,
	$gb_cch_2B_mgd_intm, 
			# End of "Modified in creation of 'cvsplit_plus_smerge_gblock_pair'. (OUT)" #

	$gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	\@new_indices_old_gbs, \@old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
			# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT)" #

	$new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	$new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	\@new_list_gblocks, 
	## [ \@new_list_excluded_bin_hrels ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (OUT2)
	);

}	# End of the subroutine, 'cvsplit_plus_smerge_gblock_pair'. #


# ssplit_plus_vmerge_gblock_pair:
#	This subroutine "horizontally split"s the 1st subject gap-block into a pair of gap-blocks of the same-type
#	& do some adjusting "shift"s, via 'ssplit_gblock';
#	then, the subroutine "merge"s the 2nd subject gap-block into one of the gap-blocks that resulted from the 'ssplit_gblock', via 'vmerge_gblock_pair'.
#	It was thus created by combining 'ssplit_gblock' & 'vmerge_gblock_pair',
#	and by adding some auxiliary functions,
#	while taking advantage of the framework of 'vsplit_plus_smerge_gblock_pair'. 
#	[CREATION STARTED on 2019/10/05, ENDED on 2019/10/13.]
#
# vsplit_plus_smerge_gblock_pair:
#	This subroutine "vertically split"s the 1st subject gap-block into a pair of gap-blocks
#	that are its "effective children" & do some adjusting "shift"s, via 'vsplit_gblock';
#	then, the subroutine "merge"s the 2nd subject gap-block into one of the resulting "effective children" that is of the same type as the 2nd subject, via 'smerge_gblock_pair'.
#	It was thus created by combining 'vsplit_gblock' & 'smerge_gblock_pair',
#	and by adding some auxiliary functions.
#	[CREATION STARTED on 2019/09/02, ENDED on 2019/09/05.]
#	[REVISED on 2019/09/06]:
#	Now, the subroutine attempts to UN-NEST the resulting gap-blocks if they are nested after the "vsplit + smerge".
#	 
# [ INPUTS: ($gb_sbj_pa, $gb_sbj_ch, # MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1a) # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
# INPUTS: ($gb_sbj1, $gb_sbj2, # MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD1a) #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm, 
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, 
#	\@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, 
#	$maxct_1way_shifts,
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b) #
#	\@set_all_exnodes, \%nodeid2msa_row
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD1b)" #
#	), where
#
#		# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2a) #
#	$gb_sbj1 and $gb_sbj2 are the indices of the subject gap-blocks
#	that are effective siblings to each other;
#	$gb_sbj1 is to be 'ssplit', while $gb_sbj2 will NOT be split;
#
#	[[
#		# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a) #
#	$gb_sbj_pa and $gb_sbj_ch are the indices of the subject gap-blocks
#	that are an effective parent and an effective child, respectively, to each other;
#	[[ ...
#	]]
#
#		# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a)" #
#	]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. (HD2a)
#
#		# End of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2a)" #
#
#	For details on other input arguments up to @gblock_sizes, refer to the header of 'purge_gblock_pair' above;
#
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	[[
#		# ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b) #
#	@set_all_exnodes stores the indices of all external nodes;
#
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#		# End of "ADDED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b)" #
#	]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
#
#		### REVISE the OUTPUTS LATER!! ###
#
#  	[[
#			# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a) #
# OUTPUTS: ($flag_end, $gb_ch_left_NEW, $gb_kpt_NEW, 
#	$gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, \@intm_indices_old_gbs, \%gb2incr_bl_coord_undo_ovlps_af_vspl,
#	$gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, \@new_indices_intm_gbs,
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_coord_undo_af_smg,
#			# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2a)" #
#	]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
#			# Modified in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2a) #
# OUTPUTS: ($flag_end, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
#	$gb_sbj1_left_intm, $gb_sbj1_right_intm, \@intm_indices_old_gbs, \%gb_intm2incr_bl_coord_undo_ovlps_af_sspl, 
#	$if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm,
#	$incr_coord_sbj_shftd_B4_vmg, \%gb_new2incr_coord_1way_dsplc_B4_vmg, \@new_indices_intm_gbs,
#	\@new_indices_old_gbs, \@old_indices_new_gbs, \%gb_new2incr_coord_undo_af_vmg,
#			# End of "Modified in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2a)" #
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#	\@new_list_gblocks,
#	@new_list_excluded_bin_hrels # ADDED in creation of 'vsplit_gblock'. (HD1a) #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed,
#		  = -1 if it failed BECAUSE the horizontal overlaps resulting from either 'ssplit' or 'vmerge' are NOT UN-DONE completely;
#
#			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2b-i) #
#
#	$gb_sbj1_unvmgd_NEW is the NEW index of the fragment of $gb_sbj1 that was left UN-vertical-merged;
#	$gb_mgd_NEW is the NEW index of the gap-block newly created by the "vertical merge" of (a fragment of) $gb_sbj1 and (the entire) $gb_sbj2;
#	[These are two of the MAIN-AUXILIARY outputs of this subroutine];
#	
#	$gb_sbj1_left_intm and $gb_sbj1_right_intm are the indices (in the INTERMEDIATE gap-block set) of the left- and right-fragments that resulted from the 'ssplit_gblock' of $gb_sbj1;
#	@intm_indices_old_gbs stores the "intermediate" indices, after the 'ssplit_gblock', of the "old" gap-blocks before the 'ssplit_gblock';
#	%gb_intm2incr_bl_coord_undo_ovlps_af_sspl = ($indx_gb_intm => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "(horizontal) same-type split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'ssplit_gblock'];
#
#	$if_vmerged_w_lfrag = 1 if $gb_sbj2 was 'vmerge'd with $gb_sbj1_left_intm, = 0 if 'vmerge'd with $gb_sbj_right_intm;
#	$if_shift_to_right = 1 if a (serial) shift was performed to the right when effectively 'vmerge' the two gap-blocks, = 0 if performed to the left;
#	$gb_sbj_fxd_intm is the index (in the INTERMEDIATE gap-block set) of the gap-block to be 'vmerge'd whose coordinate was fixed;
#	$gb_sbj_shftd_intm if the index (in the INTERMEDIATE gap-block set) of the gap-block to be 'vmerge'd which was "shift"ed when effectively performing the 'vmerge';
#	[These were created in preparation of 'vmerge_gblock_pair'];
#
#	$incr_coord_sbj_shftd_B4_vmg is the increment of the coordinate of $gb_sbj_shftd_intm
#	necessary for its 'vmerge' with $gb_sbj_fxd_intm;
#	\%gb_new2incr_coord_1way_dsplc_B4_vmg = ($indx_gb_new => {increment of the block coordinate resulting from the DISPLACEMENT prior to the "complementary vertical split"}, ...) for the gap-blocks whose displacement canNOT be undone; (Such gap-blocks MUST NOT vertically include $gb_mgd_NEW); $indx_gb_new SHOULD BE the index in the NEW gap-block set;
#	@new_indices_intm_gbs stores the "new" indices, after the 'vmerge_gblock_pair', of the "intermediate" gap-blocks before the "vmerge_gblock_pair";
#	[These are the AUXILIARY outputs of 'vmerge_gblock_pair'];
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "ssplit + vmerge";
#	More precisely, the gap-block that had been the $ogb th before the "ssplit + vmerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "(horizonal) same-type-split"
#	(, and $new_indices_old_gbs[$ogb] is undefined for some $ogb);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "ssplit + vmerge";
#	More precisely, the gap-block that is the $ngb th after the "ssplit + vmerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "ssplit + vmerge";
#
#	%gb2incr_coord_undo_af_vmg = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the 'vmerge_gblock_pair', etc.}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_vmg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#
#	[[[
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i) #
#	$gb_ch_left_NEW is the NEW index of the "effective-child" gap-block that resulted from "vertically split"ting $gb_sbj_pa AND that was LEFT unmerged to $gb_sbj_ch;
#	$gb_kpt_NEW is the NEW index of the gap-block resulting from the (horizontal) "merge" of $gb_sbj_ch and one of the products of "vertically split"ting $gb_sbj_pa;
#	[These are two of the MAIN-AUXILIARY outputs of this subroutine];
#
#	$gb_shftd_af_vspl_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was "shift"ed to make the product children non-nested;
#	$incr_coord_shftd_af_vspl (= +1 or -1) is the increment of the coordinate of $gb_shftd_af_vspl_intm via the aforementioned "shift"; 
#	@intm_indices_old_gbs stores the "intermediate" indices, after the 'vsplit_gblock', of the "old" gap-blocks before the 'vsplit_gblock';
#	%gb2incr_bl_coord_undo_ovlps_af_vspl = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "vertical split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[These are the AUXILIARY outputs of 'vsplit_gblock'];
#
#	$gb_ch_2B_mgd_intm is the index (in the INTERMEDIATE gap-block set) of the product of "vertically split"ting $gb_sbj_pa that was then "merge"d with $gb_sbj_ch (in terms of the old block set);
#	($gb1_intm, $gb2_intm) are just ($gb_ch_2B_mgd_intm, $gb_sbj_ch_intm) sorted in ascending order;
#	[These were created in preparation of 'smerge_gblock_pair'];
#
#	$bl_sbj_ssh4smg_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "shift"ed before their merger;
#	$incr_coord_sbj_ssh4smg is the increment of the coordinate of $bl_sbj_ssh4smg via the aforementioned "shift";
#	$gb_absbd_intm is the index (in the INTERMEDIATE gap-block set) of either of $gb1_intm or $gb2_intm that was "absorted" as a result of the 'smerge_gblock_pair';
#	@new_indices_intm_gbs stores the "new" indices, after the 'smerge_gblock_pair', of the "intermediate" gap-blocks before the "smerge_gblock_pair";
#	[These are the AUXILIARY outputs of 'smerge_gblock_pair'];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-i)" #
#
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "vsplit + smerge";
#	More precisely, the gap-block that had been the $ogb th before the "vsplit + smerge"
#	has been assigned $new_indices_old_gbs[$ogb] after the "vertical-split"
#	(, and $new_indices_old_gbs[$ogb] is undefined for some $ogb);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "vsplit + smerge";
#	More precisely, the gap-block that is the $ngb th after the "vsplit + smerge"
#	had been assigned $old_indices_new_gbs[$ngb] before the "vsplit + smerge";
#
#			# MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii) #
#	%gb2incr_coord_undo_af_smg = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the 'smerge_gblock_pair'}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_coord_undo_af_smg is one of the main outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' above];
#			# End of "MODIFIED in creation of 'vsplit_plus_smerge_gblock_pair'. (HD2b-ii)" #
#
#	]]] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
#			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD2b-i)" #
#
#	The other output arguments up to @new_gblock_sizes are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "vertical-split";
#
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#
#		# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD3) #
#	[ @new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']); ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. 
#	@new_list_excluded_bin_hrels = ([$gb_mgd_NEW, $gb_sbj1_unmgd_NEW, '-']);
#		# End of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair'. (HD3)" #
#	is the NEW list of PROHIBITED binary horizontal relationships,
#	which conforms to the format prescribed in the header of 'nonrec_mlt_shifts_gapblocks' above;
#
# [ sub vsplit_plus_smerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%) {
#	my ($gb_sbj_pa, $gb_sbj_ch, # Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
#
# [ sub ssplit_plus_vmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%\@\%) { ] ... OBSOLETE as of 2019/10/14.
sub ssplit_plus_vmerge_gblock_pair ($$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) { # REVISED on 2019/10/14. #


	my ($gb_sbj1, $gb_sbj2, # Modified in creation of 'ssplit_plus_vmerge_gblock_pair'. (IN) #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, 
	$maxct_1way_shifts, 
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
	$node2pa, $node2ch, $topnode_id, $br2eq

	## [[
	## 	# Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN) #
	## , $set_all_exnodes, $nodeid2msa_row
	## 	# End of "Added in creation of 'vsplit_plus_smerge_gblock_pair'. (IN)" #
	## ]] ... for the predecessor, vsplit_plus_smerge_gblock_pair'.

	) = @_; 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	unless (($ct_in_clms>0) and ($ct_in_gbs>0)) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] ILLEGAL INPUT: (#{columns}, #{gap-blocks}) = ($ct_in_clms, $ct_in_gbs) in the input MSA: $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $ct_rows = @{$curr_set_clms->[0]};

			# ADDED on 2019/10/14. (TOP) #
	my $rel_1_2 = $inter_block_relations->[$gb_sbj1][$gb_sbj2];
	unless ($rel_1_2 eq 'S') {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] ILLEGAL INPUT: The \$gb_sbj1(=$gb_sbj1) th gap-block and the \$gb_sbj2(=$gb_sbj2) th gap-block are NOT effective SIBLINGS to each other, but in '$rel_1_2' relationship: $! !\n\n";
	    return (0, undef, undef, undef);
	}
			# End of "ADDED on 2019/10/14. (TOP)" #

			# ADDED on 2019/10/06. (TOP) #
	my $len_sbj1 = $gblock_sizes->[$gb_sbj1];
	my $len_sbj2 = $gblock_sizes->[$gb_sbj2];
	unless ($len_sbj1 > $len_sbj2) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] ILLEGAL INPUT: The \$gb_sbj1(=$gb_sbj1) th gap-block (w/ length $len_sbj1) is NOT LONGER THAN the \$gb_sbj2(=$gb_sbj2) th gap-block (w/ length $len_sbj2): $! !\n\n";
	    return (0, undef, undef, undef);
	}
			# End of "ADDED on 2019/10/06. (TOP)" #

   (print STDOUT "< Subroutine 'ssplit_plus_vmerge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; effective sibling pair of gap-blocks to be \"ssplit + vmerge\"d: (\$gb_sbj1, \$gb_sbj2) = ($gb_sbj1, $gb_sbj2), (len(\$gb_sbj1), len(\$gb_sbj2)) = ($len_sbj1, $len_sbj2) ) >\n\n") if ($TEST_LEVEL_ssvm_gbp > 0);


   if ($TEST_LEVEL_ssvm_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) {...}"


   if ($TEST_LEVEL_ssvm_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Vertical Split + Merge (Same-Type)\" ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >3) {...}"


			### Preliminary: Prepare for the 'ssplit' of $gb_sbj1 (& 'vmerge' afterwards). ###
			### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (Pre) ###

	my ($le_sbj1, $re_sbj1) = @{$curr_bds_blks->[$gb_sbj1]};
	my ($le_sbj2, $re_sbj2) = @{$curr_bds_blks->[$gb_sbj2]};

	my $if_vmerged_w_lfrag; # Whether $gb_sbj2 will be 'vmerge'd with the left-fragment (=1) or the right-frangment (=0) of $gb_sbj1.

	if ($le_sbj2 < $le_sbj1) {
	    $if_vmerged_w_lfrag = 1;

	} elsif ($re_sbj1 < $re_sbj2) {
	    $if_vmerged_w_lfrag = 0;

	} else {
			# In this case, $gb_sbj1 horizontally includes $gb_sbj2. (This is actually rare.) #

	    if ($le_sbj2 - $le_sbj1 < $re_sbj1 - $re_sbj2) { 
		$if_vmerged_w_lfrag = 1;
	    } else {
		$if_vmerged_w_lfrag = 0;
	    }
	}

		# The sizes of the new gap-blocks resulting from the 'ssplit' of $gb_sbj1. #
	my ($new_size_left1, $new_size_right1) = ($if_vmerged_w_lfrag) ?
 		($len_sbj2, $len_sbj1 - $len_sbj2) : ($len_sbj1 - $len_sbj2, $len_sbj2) ;

   if ($TEST_LEVEL_ssvm_gbp >2) { # (Pre) #

	print STDOUT "\n";
	print STDOUT '$if_vmerged_w_lfrag = ', $if_vmerged_w_lfrag, " .\n\n";

	print STDOUT '($new_size_left1, $new_size_right1) = ', "($new_size_left1, $new_size_right1) .\n";
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_ssvm_gbp >2) { # (Pre) # ...}"


			### End of "(ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (Pre)" ###

			## Preliminary 2: Prepare the list of old gap-blocks,
			## conforming to the format employed by the output of 
			## 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"
			## (IMPORTED from 'vsplit_gblock' above in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/14).) (Pre2) ##

	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$curr_bds_blks->[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #

   if ($TEST_LEVEL_ssvm_gbp > 3) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 3) {...}"

			## END of "(IMPORTED from 'vsplit_gblock' above in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/14).) (Pre2)" ##
	
=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (Pre)
			### Preliminary: Virtually VSplit the $gb_sbj_pa into TWO effective children. ###

	my ($br_sbj_pa, $up_or_lw_sbj_pa) = @{$list_gblocks0->[$gb_sbj_pa]}[$indx_br, $indx_u_or_d];

	my ($set_eff_child_gblks) = vsplit_into_eff_child_gblocks ($br_sbj_pa, $up_or_lw_sbj_pa,
		 %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, @{$set_all_exnodes},
		 %{$nodeid2msa_row}, %{$set_msa_rows2br_up_or_lw}) ;
	unless (defined $set_eff_child_gblks) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'vsplit_into_eff_child_gblocks' FAILED on (\$br_sbj_pa, \$up_or_lw_sbj_pa) = ($br_sbj_pa, $up_or_lw_sbj_pa), for \$gb_sbj_pa (=$gb_sbj_pa) : $! !\n\n";
	    return (0, undef, undef, undef);
	}

	unless (@{$set_eff_child_gblks} == 2) {
	    my $ct_echs = @{$set_eff_child_gblks};
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] SOMEHOW, \$gb_sbj_pa (=$gb_sbj_pa) has $ct_echs (!= 2) effective child sets of msa rows: $! !\n\n";
	    return (0, undef, undef, undef);
	}

   if ($TEST_LEVEL_ssvm_gbp >2) { # (Pre) #

	print STDOUT "\n[ Outputs of 'vsplit_into_eff_child_gblocks': ]\n\n";

	print STDOUT join ("\t", 'Child_ID', 'br_ID', 'up(U)_or_lw(L)', '(external,nodes,affected)', '[MSA,rows,affected]'), "\n\n";

	for (my $i=0; $i <2; $i++) {
	    my $cid = $i+1;
	    my ($br, $u_or_l, $exnodes_affd, $rows_affd) = @{$set_eff_child_gblks->[$i]};
	    my $cnct_exnodes_affd = join (',', @{$exnodes_affd});
	    my $cnct_rows_affd = join (',', @{$rows_affd});

	    print STDOUT join ("\t", "Ch_$cid", $br, $u_or_l, "($cnct_exnodes_affd)", "[$cnct_rows_affd]"), "\n";
	} # End of "for (my $i=0; $i <2; $i++) {...}" #

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (Pre) # ...}"

=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (Pre)"

		### MAIN 1: 'ssplit' $gb_sbj1 . ###
		### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (MAIN 1) ###

	my ($flag_sspl, $gb_sbj1_left_intm, $gb_sbj1_right_intm, $intm_indices_old_gbs, $old_indices_intm_gbs, $gb_intm2incr_bl_coord_undo_ovlps_af_sspl,
	# $gb_ch1_intm, $gb_ch2_intm, $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $old_indices_intm_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl, # For 'vsplit_gblock'. #

	    $intm_curr_set_clms, $intm_curr_bds_gblks, $intm_curr_bl_coords, $intm_bds_gb_coords, $intm_gb_coords_wd_frm,
	    $intm_curr_gblks_w_spec_lb, $intm_curr_gblks_w_spec_rb, $rows_affd_by_intm_blks,
	    $intm_inter_block_relations, $intm_cts_vincluding, $intm_cts_vincluded, $intm_cts_onx, $intm_gblock_sizes
	    # $intm_list_gblocks, $intm_list_excluded_bin_hrels # For 'vsplit_gblock'. #
		)
	  = ssplit_gblock (
	     $gb_sbj1, $new_size_right1,
	     @{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm},
	     @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
	     @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes},  $maxct_1way_shifts

	     #@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     #%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq} # These are for 'vsplit_gblock'. #

		);
	unless ($flag_sspl) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The 1st MAIN Subroutine 'ssplit_gblock' FAILED on the \$gb_sbj1 (=$gb_sbj1) th gap-block: $! !\n";
	    return (0, undef, undef, undef);
	}
	unless ($flag_sspl > 0) {
	    print STDOUT "[CAUTION 'ssplit_plus_vmerge_gblock_pair'] The 1st MAIN subroutine, 'ssplit_gblock' FAILED on \$gb_sbj1 (=$gb_sbj1), because the resulting horizontal overlaps canNOT be UNDONE completely!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}
   (print STDOUT "\n --- The 1st MAIN Subroutine 'ssplit_gblock' ENDED on the \$gb_sbj1 (=$gb_sbj1) th gap-block (w/ \$flag_sspl = $flag_sspl, \$gb_sbj1_left_intm = $gb_sbj1_left_intm, \$gb_sbj1_right_intm = $gb_sbj1_right_intm) ...\n\n\n") if ($TEST_LEVEL_ssvm_gbp >1);

		### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (MAIN 1) ###


=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (MAIN 1-a)

		### MAIN 1: "Vertically Split" $gb_sbj_pa. ###

	my ($info_effch1, $info_effch2) = @{$set_eff_child_gblks};


	my ($flag_vspl, $gb_ch1_intm, $gb_ch2_intm, $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $old_indices_intm_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	    $intm_curr_set_clms, $intm_curr_bds_gblks, $intm_curr_bl_coords, $intm_bds_gb_coords, $intm_gb_coords_wd_frm,
	    $intm_curr_gblks_w_spec_lb, $intm_curr_gblks_w_spec_rb, $rows_affd_by_intm_blks,
	    $intm_inter_block_relations, $intm_cts_vincluding, $intm_cts_vincluded, $intm_cts_onx, $intm_gblock_sizes,
	    $intm_list_gblocks, $intm_list_excluded_bin_hrels)
	  = vsplit_gblock (
	     $gb_sbj_pa, @{$info_effch1}, @{$info_effch2},
	     @{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm},
	     @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
	     @{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes},  $maxct_1way_shifts,
	     @{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
		);
	unless ($flag_vspl) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The 1st MAIN Subroutine 'vsplit_gblock' FAILED on the \$gb_sbj_pa (=$gb_sbj_pa) th gap-block: $! !\n";
	    return (0, undef, undef, undef);
	}
	unless ($flag_vspl > 0) {
	    print STDOUT "[CAUTION 'ssplit_plus_vmerge_gblock_pair'] The 1st MAIN subroutine, 'vsplit_gblock' FAILED on \$gb_sbj_pa (=$gb_sbj_pa), because the resulting horizontal overlaps canNOT be UNDONE completely!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}
   (print STDOUT "\n --- The 1st MAIN Subroutine 'vsplit_gblock' ENDED on the \$gb_sbj_pa (=$gb_sbj_pa) th gap-block (w/ \$flag_vspl = $flag_vspl, \$gb_ch1_intm = $gb_ch1_intm, \$gb_ch2_intm = $gb_ch2_intm, \$gb_shftd_af_vspl_intm = $gb_shftd_af_vspl_intm, \$incr_coord_shftd_af_vspl = $incr_coord_shftd_af_vspl) ...\n\n\n") if ($TEST_LEVEL_ssvm_gbp >1);

=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (MAIN 1-a)"


	my $ct_gbs_intm = @{$intm_curr_bds_gblks};

   if ($TEST_LEVEL_ssvm_gbp >2) { # (M1-A) #

	##print STDOUT "\n[[ Outputs of 'vsplit_gblock' ]]\n\n";
	print STDOUT "\n[[ Outputs of 'ssplit_gblock' ]]\n\n"; # MODIFIED on 2019/10/06. #

        print STDOUT "[ Relationships between old and intermediate indices of gap-blocks (#{gap-blocks}_intm=$ct_gbs_intm) : ]\n\n";

        my @out_intm_indices_old_gbs = my @out_old_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_old_gbs, \@out_intm_indices_old_gbs],
                          [$old_indices_intm_gbs, \@out_old_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$intm_indices_old_gbs} = (', join (', ', @out_intm_indices_old_gbs), ") .\n\n";

        print STDOUT '@{$old_indices_intm_gbs} = (', join (', ', @out_old_indices_intm_gbs), ") .\n";

        print STDOUT "\n\n";

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (M1-A)

	print STDOUT '[ Contents of @{$intm_list_excluded_bin_hrels}: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";
	for (my $i=0; $i < @{$intm_list_excluded_bin_hrels}; $i++) {
	    print STDOUT join ("\t", $i, @{$intm_list_excluded_bin_hrels->[$i]}), "\n";
	}
	print STDOUT "\n\n";
=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (M1-A)"

			# Added on 2019/09/06. #
			# => MODIFIED on 2019/10/06. #
	##print STDOUT '[ Contents of %{$gb2incr_bl_coord_undo_ovlps_af_vspl}: ]', "\n\n";
	print STDOUT '[ Contents of %{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl}: ]', "\n\n";

	print STDOUT '%{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl} = ( # $indx_gb_intm => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb_intm2incr_bl_coord_undo_ovlps_af_sspl}) {
	    print STDOUT "\t", $gb, ' => ', $gb_intm2incr_bl_coord_undo_ovlps_af_sspl->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/09/06. #
			# => MODIFIED on 2019/10/06." #

   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (M1-A) #...}"

   if ($TEST_LEVEL_ssvm_gbp >5) { # (M1-B) #

	##print STDOUT "\n[ Gap-Block Configurations AFTER the \"Vertical Split\" ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the \"(Horizontal) Same-Type Split\" ]\n\n"; # Modified on 2019/10/06. #

	unless ( print_out_gblock_configs (@{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >5) { # (M1-B) # ...}"

   if ($TEST_LEVEL_ssvm_gbp > 5) { # (M1-C) #

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the \"Vertical Split\" ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the \"(Horizontal) Same-Type Split\" ]\n\n"; # Modified on 2019/10/06.

	unless ( print_out_remng_attr_gblks (@{$intm_inter_block_relations}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm}, @{$intm_gblock_sizes}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$rows_affd_by_intm_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) { # (M1-C) # ...}"


		# Create @intm_list_gblocks. #
		# (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (M1-D) #
	my @intm_list_gblocks = ();
        for (my $igb=0; $igb< $ct_gbs_intm; $igb++) {

	    my $ogb = $old_indices_intm_gbs->[$igb];
	    unless (defined $ogb) { $ogb = $gb_sbj1; }

	    ##my @info_gb = @{$list_gblocks0->[$ogb]};
	    my @info_gb = @{$list_old_gblocks[$ogb]}; ## CORRECTED in DEBUGGING (on 2019/10/14). ##

	    @info_gb[2,3] = @{$intm_curr_bds_gblks->[$igb]};
	    $intm_list_gblocks[$igb] = \@info_gb;
	}
		# End of "(ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (M1-D)" #

   if ($TEST_LEVEL_ssvm_gbp >= 5) { # (M1-D) #

	print STDOUT "\n";
        ##print STDOUT '[ Contents of @{$intm_list_gblocks}: ]', "\n\n";
        print STDOUT '[ Contents of @intm_list_gblocks: ]', "\n\n"; # Modified in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06). #

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_gbs_intm; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm_list_gblocks[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >= 5) { # (M1-D) #...}"


		### MAIN 2pre: Prepare for the 'vmerge_gblock_pair'. ###
		### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (MAIN 2pre) ###

	my $gb_sbj2_intm = $intm_indices_old_gbs->[$gb_sbj2];

   if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-A) #
	print STDOUT "\n";
	print STDOUT "\$gb_sbj2_intm = $gb_sbj2_intm .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-A) #...}"


	my ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm);  # The gap-blocks to be "vmerge"d, the first to be fixed and the second to be "shift"ed.
	my $if_shift_to_right; # Whether the "shift" before the 'vmerge' is to the right (=1) or to the left (=0). 
	##my $gb_sbj2_unvmgd_intm; # A BUG fixed (on 2019/10/09). #
	my $gb_sbj1_unvmgd_intm; # The fragment of $gb_sbj1 that will NOT be 'vmerge'd with $gb_sbj2. #

	my ($le_sbj2_intm, $re_sbj2_intm) = @{$intm_curr_bds_gblks->[$gb_sbj2_intm]};

	if ($if_vmerged_w_lfrag) { # $gb_sbj2 will be "vmerge"d with the left-fragment of $gb_sbj1.

	    $gb_sbj1_unvmgd_intm = $gb_sbj1_right_intm;
	    $if_shift_to_right = 0;

	    my ($le_lfrg1_intm, $re_lfrg1_intm) = @{$intm_curr_bds_gblks->[$gb_sbj1_left_intm]};

	    ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) = ($re_sbj2_intm < $re_lfrg1_intm) ? 
		($gb_sbj2_intm, $gb_sbj1_left_intm) : ($gb_sbj1_left_intm, $gb_sbj2_intm) ;

	} else { # $gb_sbj2 will be "vmerge"d with the right-fragment of $gb_sbj1.

	    $gb_sbj1_unvmgd_intm = $gb_sbj1_left_intm;
	    $if_shift_to_right = 1;

	    my ($le_rfrg1_intm, $re_rfrg1_intm) = @{$intm_curr_bds_gblks->[$gb_sbj1_right_intm]};

	    ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) = ($le_sbj2_intm > $le_rfrg1_intm) ? 
		($gb_sbj2_intm, $gb_sbj1_right_intm) : ($gb_sbj1_right_intm, $gb_sbj2_intm) ;
	}


   if ($TEST_LEVEL_ssvm_gbp >1) { # (M2Pre-B) #
	print STDOUT "\n";
	print STDOUT '$gb_sbj1_unvmgd_intm = ', $gb_sbj1_unvmgd_intm, " .\n\n";
	print STDOUT '$if_shift_to_right = ', $if_shift_to_right, " .\n\n";
	print STDOUT '($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) = ', "($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_ssvm_gbp >1) { # (M2Pre-B) #...}"


		### END of "(ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/06).) (MAIN 2pre)" ###

=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2pre)
		### MAIN 2pre: Identify the resulting "effective child" to be "merge"d with $gb_sbj_ch_intm. ### 

	my $gb_sbj_ch_intm = $intm_indices_old_gbs->[$gb_sbj_ch];
	my $rows_affd_by_sbj_ch = $rows_affd_by_intm_blks->[$gb_sbj_ch_intm];

   if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-A) #
	print STDOUT "\n";
	print STDOUT "\$gb_sbj_ch_intm = $gb_sbj_ch_intm .\n\n";
	print STDOUT '@{$rows_affd_by_sbj_ch} = (', join (', ', @{$rows_affd_by_sbj_ch}), ") .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-A) #...}"

	my $gb_ch_2B_mgd_intm;
	foreach my $gb_intm ($gb_ch1_intm, $gb_ch2_intm) {
	    my $rows_affd_by_gb = $rows_affd_by_intm_blks->[$gb_intm];
	    my ($common, $ex1, $ex2) = diff_nmarray (@{$rows_affd_by_sbj_ch}, @{$rows_affd_by_gb});
	    unless (defined $common) {
		print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'diff_nmarray' FAILED on the rows affected by \$gb_sbj_ch_intm (=$gb_sbj_ch_intm) and those affected by \$gb_intm (=$gb_intm): $! !\n\n";
		return (0, undef, undef, undef);
	    }

	    if ((@{$ex1}==0) and (@{$ex2}==0)) { $gb_ch_2B_mgd_intm = $gb_intm; }

	} # End of "foreach my $gb_intm ($gb_ch1_intm, $gb_ch2_intm) {...}". #

	unless (defined $gb_ch_2B_mgd_intm) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The \$gb_ch_2B_mgd_intm is NOT DEFINED: $! !\n";
	    return (0, undef, undef, undef);
	}


	my $gb_ch_left_intm = ($gb_ch_2B_mgd_intm == $gb_ch1_intm) ? $gb_ch2_intm : $gb_ch1_intm ;

   if ($TEST_LEVEL_ssvm_gbp >1) { # (M2Pre-B) #
	print STDOUT "\n";
	print STDOUT "(\$gb_ch_2B_mgd_intm, \$gb_ch_left_intm) = ($gb_ch_2B_mgd_intm, $gb_ch_left_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_ssvm_gbp >1) { # (M2Pre-B) #...}"


	my ($gb1_intm, $gb2_intm) = ($gb_sbj_ch_intm < $gb_ch_2B_mgd_intm) ?
		($gb_sbj_ch_intm, $gb_ch_2B_mgd_intm) : ($gb_ch_2B_mgd_intm, $gb_sbj_ch_intm) ;

   if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-C) #
	print STDOUT '($gb1_intm, $gb2_intm) = ', "($gb1_intm, $gb2_intm) .\n";
	print STDOUT "\n\n";
   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-C) #...}" # A BUG fixed (on 2019/09/05).
   #} if ($TEST_LEVEL_ssvm_gbp >2) { # (M2Pre-C) #

=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2pre)"


		### MAIN 2: "VMerge" $gb_sbj_fxd_intm & $gb_sbj_shftd_intm, via 'vmerge_gblock_pair'. ###
		### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09).) (MAIN 2) ###

	## print STDOUT '($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) = ', "($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) .\n";

	my ($flag_vmg, $incr_coord_sbj_shftd_B4_vmg, $gb_new2incr_coord_1way_dsplc_B4_vmg, $gb_mgd_NEW, $new_indices_intm_gbs, $intm_indices_new_gbs,
	    $gb_new2incr_coord_undo_af_vmg, # Added on 2019/10/10. #

	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,

	    $new_list_gblocks # ADDED for 'vmerge_gblock_pair' (OUT). #
		)
	  = vmerge_gblock_pair ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm, $if_shift_to_right,
	     @{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm},
	     @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, @{$rows_affd_by_intm_blks}, $cnct_null_clm,

	     @{$intm_inter_block_relations}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$intm_gblock_sizes}, $maxct_1way_shifts,
			# ADDED for 'vmerge_gblock_pair' (IN). #
	     @intm_list_gblocks, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
			# End of "ADDED for 'vmerge_gblock_pair' (IN)." #
		);
	unless ($flag_vmg) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The 2nd MAIN Subroutine 'vmerge_gblock_pair' FAILED on the gap-block pair: (\$gb_sbj_fxd_intm, \$gb_sbj_shftd_intm) = ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) in the Configuration AFTER the \"(Horizontal) Same-Type Split\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n --- The 2nd MAIN Subroutine 'vmerge_gblock_pair' ENDED on the gap-block pair: (\$gb_sbj_fxd_intm, \$gb_sbj_shftd_intm) = ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) AFTER the \"(horizontal) same-type split\" (w/ \$incr_coord_sbj_shftd_B4_vmg = $incr_coord_sbj_shftd_B4_vmg, \$gb_mgd_NEW = $gb_mgd_NEW) ...\n\n\n") if ($TEST_LEVEL_ssvm_gbp >0);


		### (ADDED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09).) (MAIN 2) ###

=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2)

		### MAIN 2: "Merge" $gb1_intm and $gb2_intm (i.e., $gb_sbj_ch_intm and $gb_ch_2B_mgd_intm). ###

	my ($flag_smg, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $gb_kpt_intm, $new_indices_intm_gbs, $intm_indices_new_gbs,
	    $new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes)
	  = smerge_gblock_pair (
	     $gb1_intm, $gb2_intm, @{$intm_curr_set_clms}, @{$intm_curr_bds_gblks}, @{$intm_curr_bl_coords}, @{$intm_bds_gb_coords}, @{$intm_gb_coords_wd_frm},
	     @{$intm_curr_gblks_w_spec_lb}, @{$intm_curr_gblks_w_spec_rb}, @{$rows_affd_by_intm_blks}, $cnct_null_clm,

	     @{$intm_inter_block_relations}, @{$intm_cts_vincluding}, @{$intm_cts_vincluded}, @{$intm_cts_onx}, @{$intm_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_smg) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The 2nd MAIN Subroutine 'smerge_gblock_pair' FAILED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) in the Configuration AFTER the \"Vertical Split\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n --- The 2nd MAIN Subroutine 'smerge_gblock_pair' ENDED on the gap-block pair: (\$gb1_intm, \$gb2_intm) = ($gb1_intm, $gb2_intm) AFTER the \"vertical split\" (w/ \$bl_sbj_ssh4smg_intm = $bl_sbj_ssh4smg_intm, \$incr_coord_sbj_ssh4smg = $incr_coord_sbj_ssh4smg, \$gb_absbd_intm = $gb_absbd_intm, \$gb_kpt_intm = $gb_kpt_intm) ...\n\n\n") if ($TEST_LEVEL_ssvm_gbp >0);

=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2)"

	my $ct_out_gbs = @{$new_curr_bds_gblks};

	my $ct_out_clms = @{$new_curr_set_clms}; # ADDED in REVISION (on 2019/09/06). #


   if ($TEST_LEVEL_ssvm_gbp >2) { # (M2-A) #

	##print STDOUT "\n[[ Outputs of 'smerge_gblock_pair' ]]\n\n";
	print STDOUT "\n[[ Outputs of 'vmerge_gblock_pair' ]]\n\n"; # Modified (on 2019/10/09). #

        print STDOUT "[ Relationships between new and intermediate indices of gap-blocks (#{gap-blocks}_new=$ct_out_gbs) : ]\n\n";

        my @out_intm_indices_new_gbs = my @out_new_indices_intm_gbs = ();

        foreach my $pair ( [$intm_indices_new_gbs, \@out_intm_indices_new_gbs],
                          [$new_indices_intm_gbs, \@out_new_indices_intm_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@{$new_indices_intm_gbs} = (', join (', ', @out_new_indices_intm_gbs), ") .\n\n";
        print STDOUT '@{$intm_indices_new_gbs} = (', join (', ', @out_intm_indices_new_gbs), ") .\n";
        print STDOUT "\n\n";

				# ADDED in Creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). #
	print STDOUT '[ Contents of %{$gb_new2incr_coord_1way_dsplc_B4_vmg}: ]', "\n\n";

	print STDOUT '%{$gb_new2incr_coord_1way_dsplc_B4_vmg} = (', "\n";
	foreach my $ngb (sort {$a <=> $b} keys %{$gb_new2incr_coord_1way_dsplc_B4_vmg}) {
	    print STDOUT "\t", $ngb, ' => ', $gb_new2incr_coord_1way_dsplc_B4_vmg->{$ngb}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n\n";
				# END of "ADDED in Creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09)." #
	
   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) { # (M2-A) # ...}"


   if ($TEST_LEVEL_ssvm_gbp >5) { # (M2-B) #

	##print STDOUT "\n[ Gap-Block Configurations AFTER the 'smerge_gblock_pair': ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the 'vmerge_gblock_pair': ]\n\n"; # Modified (on 2019/10/09). #

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >5) { # (M2-B) # ...}"

   if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2-C) #

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'smerge_gblock_pair': ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'vmerge_gblock_pair': ]\n\n"; # Modified (on 2019/10/09). #

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2-C) # ...}"


		# [ MAIN 2post: If any overlaps between vertically equivalent and/or complementary gap-blocks result from the 'smerge_gblock_pair', undo them. # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'.
		# MAIN 2post: If any overlaps between vertically equivalent and/or complementary gap-blocks result from the 'ssplit_gblock' and/or 'vmerge_gblock_pair', undo them. # (MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09).) #

	my $gb_sbj1_unvmgd_NEW = $new_indices_intm_gbs->[$gb_sbj1_unvmgd_intm]; # The new index for the fragment of $gb_sbj1 NOT to be 'vmerge'd. # (Added in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09).) #

   if ($TEST_LEVEL_ssvm_gbp > 2) { # (M2post) #

	print STDOUT "\n";
	print STDOUT '$gb_sbj1_unvmgd_NEW  = ', $gb_sbj1_unvmgd_NEW , " .\n";
	print STDOUT "\n\n";
   
   } # End of "if ($TEST_LEVEL_ssvm_gbp > 2) { # (M2post) # ...}"

=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-a)
	my $gb_kpt_NEW = $new_indices_intm_gbs->[$gb_kpt_intm]; # The new index for the gap-block resulting from the "smerge". #
	my $gb_ch_left_NEW = $new_indices_intm_gbs->[$gb_ch_left_intm]; # The "effective-child" block that was left unmerged. # MOVED from Main 3 below, in REVISION (on 2019/09/06). #
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-a)"

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-b) #
=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-b)
	my ($fl_undo_af_smg, $gb2incr_coord_undo_af_smg)
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_kpt_NEW,
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-b)"


=REPLACED with the identical process within 'vmerge_gblock_pair' (on 2019/10/10). (M2post-i)

	my ($fl_undo_af_vmg, $gb2incr_coord_undo_af_vmg)
	  = undo_hovlp_vcmpl_gbsets_1sbj ($gb_mgd_NEW,
			# End of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-b)" #

	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	        @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	        );
	unless ($fl_undo_af_vmg) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)) on the cofiguration of gap-blocks after the 'vmerge_gblock_pair' on (\$gb_sbj_fxd_intm, \$gb_sbj_shftd_intm) = ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm): $! !\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)), Ended (w/ \$fl_undo_af_vmg = $fl_undo_af_vmg) on the configuration of gap-blocks after the 'vmerge_gblock_pair' on (\$gb_sbj_fxd_intm, \$gb_sbj_shftd_intm) = ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm) ...\n\n") if ($TEST_LEVEL_ssvm_gbp >1);

=cut # END of "REPLACED with the identical process within 'vmerge_gblock_pair' (on 2019/10/10). (M2post-i)"


   if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A) #

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-c) #
	##print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)): ]\n\n"; 
	print STDOUT "\n[ An Auxiliary Output of 'vmerge_gblock_pair' (of \$gb_sbj_fxd_intm (=$gb_sbj_fxd_intm) & \$gb_sbj_shftd_intm (=$gb_sbj_shftd_intm): ]\n\n"; # MODIFIED on 2019/10/10. #

	##print STDOUT "\$fl_undo_af_vmg = $fl_undo_af_vmg .\n\n"; ## DISABLED (on 2019/10/10). #

	print STDOUT '%{$gb_new2incr_coord_undo_af_vmg} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $ngb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_vmg}) {
	    print STDOUT "\t", $ngb, ' => ', $gb_new2incr_coord_undo_af_vmg->{$ngb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";
			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-c)" #

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A) # ...}"

=REPLACED with the identical process within 'vmerge_gblock_pair' (on 2019/10/10). (M2post-ii)
	unless ($fl_undo_af_vmg > 0) {
	    print STDOUT "[CAUTION 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
	    return (-1, undef, undef, undef);
	}
=cut # END of "REPLACED with the identical process within 'vmerge_gblock_pair' (on 2019/10/10). (M2post-ii)"


			# ADDED in REVISION (on 2019/09/06). #

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-d) #
	my ($lbd_mgd, $rbd_mgd) = @{$new_curr_bds_gblks->[$gb_mgd_NEW]};
	my ($lbd_sbj1_unmgd, $rbd_sbj1_unmgd) = @{$new_curr_bds_gblks->[$gb_sbj1_unvmgd_NEW]}; 

	my ($dist1, $dist2) = compt_inter_gblock_distance ($gb_mgd_NEW, $gb_sbj1_unvmgd_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations});
	unless (defined $dist2) {
	    print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'compt_inter_gblock_distance' FAILED on (\$gb_mgd_NEW, $gb_sbj1_unvmgd_NEW) (=($gb_mgd_NEW, $gb_sbj1_unvmgd_NEW)) on the cofiguration of gap-blocks after the 'vmerge_gblock_pair' on (\$gb_sbj_fxd_intm, \$gb_sbj_shftd_intm) = ($gb_sbj_fxd_intm, $gb_sbj_shftd_intm): $! !\n";
	    return (0, undef, undef, undef);
	}

	my $dist = $dist2;

=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-c)
	my ($lbd_kpt, $rbd_kpt) = @{$new_curr_bds_gblks->[$gb_kpt_NEW]};
	my ($lbd_ch_left, $rbd_ch_left) = @{$new_curr_bds_gblks->[$gb_ch_left_NEW]}; 
=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-c)"

			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-d)" #

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-e) #

	##if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {
	if ($dist <= 0) {

		# [ If $gb_kpt_NEW horizontally includes $gb_ch_left_NEW. #
		# => Attempt to UN-NEST them, by moving $gb_ch_left_NEW. # ] ... for the predecessor, 'vsplit_plus_smerge_gblock_pair'. 
		# If $gb_mgd_NEW is horizontally NOT SEPARATED from $gb_sbj1_unvmgd_NEW,
		# => Attempt to SEPARATE them, by moving $gb_sbj1_unvmgd_NEW.

   (print STDOUT "\n[INFORMATION 'ssplit_plus_vmerge_gblock_pair']  The \$gb_mgd_NEW (=$gb_mgd_NEW), with [lbd, rbd] = [$lbd_mgd, $rbd_mgd], is Horizontally NOT SEPARATED from the \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW), with [lbd, rbd] = [$lbd_sbj1_unmgd, $rbd_sbj1_unmgd]!!\n") if ($TEST_LEVEL_ssvm_gbp > 0);
   (print STDOUT "                      ... I will Attempt to SEPARATE them via \"serial shifts\" of \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW) ...\n\n") if ($TEST_LEVEL_ssvm_gbp > 0);

   if ($TEST_LEVEL_ssvm_gbp >5) { # (M2post-B') #

	##print STDOUT "\n[ Gap-Block Configurations AFTER the 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_kpt_NEW (=$gb_kpt_NEW)) ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_mgd_NEW (=$gb_mgd_NEW)) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >5) { # (M2post-B') # ...}"

	    my $curr_coord_sbj1_unmgd = $new_curr_bl_coords->[$gb_sbj1_unvmgd_NEW];
	    my ($lbd_coord_sbj1_unmgd, $rbd_coord_sbj1_unmgd) = @{$new_bds_gb_coords->[$gb_sbj1_unvmgd_NEW]};

	    ##my %skipped = ($gb_kpt_NEW => 1); # Because we already know that $gb_kpt_NEW is an "effective sibling" of $gb_ch_left_NEW. 
	    my %skipped = (); # NOTE that $gb_mgd_NEW must NOT be skipped!!

	    ##my $ct_nec_shifts_left = ($lbd_kpt > 0) ? ct_necessary_shifts ($lbd_ch_left, $lbd_kpt-1, $gb_ch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1 ;
	    my $ct_nec_shifts_left = ($lbd_mgd > $len_sbj1 - $len_sbj2) ? ct_necessary_shifts ($rbd_sbj1_unmgd, $lbd_mgd-2, $gb_sbj1_unvmgd_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1 ;
	    ##if ($ct_nec_shifts_left > $curr_coord_ch_left - $lbd_coord_ch_left) { $ct_nec_shifts_left = $ct_out_clms +1; }
	    if ($ct_nec_shifts_left > $curr_coord_sbj1_unmgd - $lbd_coord_sbj1_unmgd) { $ct_nec_shifts_left = $ct_out_clms +1; } # CORRECTED in DEBUGGING on 2019/10/13. #

	    ##my $ct_nec_shifts_right = ($rbd_kpt < $ct_out_clms-1) ? ct_necessary_shifts ($rbd_ch_left, $rbd_kpt+1, $gb_ch_left_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1;
	    my $ct_nec_shifts_right = ($rbd_mgd < $ct_out_clms-1-($len_sbj1-$len_sbj2)) ? ct_necessary_shifts ($lbd_sbj1_unmgd, $rbd_mgd+2, $gb_sbj1_unvmgd_NEW, @{$new_curr_bds_gblks}, @{$new_inter_block_relations}, %skipped) : $ct_out_clms +1;
	    ##if ($ct_nec_shifts_right > $rbd_coord_ch_left - $curr_coord_ch_left) { $ct_nec_shifts_right = $ct_out_clms +1; }
	    if ($ct_nec_shifts_right > $rbd_coord_sbj1_unmgd - $curr_coord_sbj1_unmgd) { $ct_nec_shifts_right = $ct_out_clms +1; } # CORRECTED in DEBUGGING on 2019/10/13. #

	    my ($direction, $ct_nec_shifts) = ($ct_nec_shifts_left < $ct_nec_shifts_right) ?
			('L', $ct_nec_shifts_left ) :  ('R', $ct_nec_shifts_right) ;

	    unless ($ct_nec_shifts < $ct_out_clms) {
		print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW) caNNOT be SEPARATED from \$gb_mgd_NEW (=$gb_mgd_NEW): $! !\n\n";
		return (-1, undef,  undef, undef);
	    }

	    ##my $incr_coord_ch_left_via_ssh = ($direction eq 'R') ? $ct_nec_shifts : - $ct_nec_shifts ;
	    my $incr_coord_sbj1_unmgd_via_ssh = ($direction eq 'R') ? $ct_nec_shifts : - $ct_nec_shifts ;

   if ($TEST_LEVEL_ssvm_gbp >2) {
	print STDOUT "\nIntended increment of the coordinate of \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW): \$incr_coord_sbj1_unmgd_via_ssh = $incr_coord_sbj1_unmgd_via_ssh .\n\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >2) {...}"

	    my ($flag_ssh, $incr_ct_null_clms_ssh)
	     = serial_shifts_sgl_gblock (
	        @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},

	        ##$gb_ch_left_NEW, $incr_coord_ch_left_via_ssh,
	        $gb_sbj1_unvmgd_NEW, $incr_coord_sbj1_unmgd_via_ssh, # MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-f1) #

		@{$rows_affd_by_new_blks}, $cnct_null_clm,
	        @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	    	);
	    unless ((defined $flag_ssh) and ($flag_ssh == 0)) {
	        my $flag_out = (defined $flag_ssh) ? $flag_ssh : 'undef';
	        print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'serial_shifts_sgl_gblock', FAILED on the \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW) th gap-block (w/ \$flag_ssh = $flag_ssh): $! !\n\n";

	        return (0, undef, undef, undef);
	    }
   (print STDOUT "\n ... The subroutine, 'serial_shifts_sgl_gblock', Ended on the \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW) th gap-block (w/ \$incr_ct_null_clms_ssh = $incr_ct_null_clms_ssh) ...\n\n") if ($TEST_LEVEL_ssvm_gbp >1);

   if ($TEST_LEVEL_ssvm_gbp >5) { # (M2post-B'') #

	##print STDOUT "\n[ Gap-Block Configurations AFTER the 'serial_shifts_sgl_gblock' on \$gb_ch_left_NEW (=$gb_ch_left_NEW) ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the 'serial_shifts_sgl_gblock' on \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW) ]\n\n";

	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >5) { # (M2post-B'') # ...}"


			# Update %{$gb_new2incr_coord_undo_af_vmg}: #
			# (Added in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/13).) #

	    if (defined $gb_new2incr_coord_undo_af_vmg->{$gb_sbj1_unvmgd_NEW}) {
		$gb_new2incr_coord_undo_af_vmg->{$gb_sbj1_unvmgd_NEW} += $incr_coord_sbj1_unmgd_via_ssh;
	    } else {
		$gb_new2incr_coord_undo_af_vmg->{$gb_sbj1_unvmgd_NEW} = $incr_coord_sbj1_unmgd_via_ssh;
	    }
			# End of "(Added in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/13).)" #

	} # End of "if ($dist <= 0) { ...}" # MOVED from below the output of "FINAL %{$gb_new2incr_coord_undo_af_vmg}". (on 2019/10/13). #


	    	## Attempt to UNDO the horizontal overlaps due to the "serial shifts", 'ssplit', and/or 'vmerge'. ##

	my ($fl_undo_af_ssh, $gb_new2incr_coord_undo_af_ssh)
	      ##= undo_hovlp_vcmpl_gbsets_1sbj ($gb_ch_left_NEW,
	   = undo_hovlp_vcmpl_gbsets_1sbj ($gb_sbj1_unvmgd_NEW, # MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-f2) #
	            @{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_bds_gb_coords}, @{$new_curr_bl_coords},
	            @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, @{$rows_affd_by_new_blks}, $cnct_null_clm,
	            @{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$new_gblock_sizes}
	            );
	unless ($fl_undo_af_ssh) {
	        print STDOUT "[ERROR 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW)) on the cofiguration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_sbj1_unvmgd_NEW itself: $! !\n\n";
	        return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW)), Ended (w/ \$fl_undo_af_ssh = $fl_undo_af_ssh) on the configuration of gap-blocks after the 'serial_shifts_sgl_gblock' on \$gb_sbj1_unvmgd_NEW itself ...\n\n") if ($TEST_LEVEL_ssvm_gbp >1);

   if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A'') #

	##print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_ch_left_NEW (=$gb_ch_left_NEW)): ]\n\n";
	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets_1sbj' (w.r.t. \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW)): ]\n\n";

	print STDOUT "\$fl_undo_af_ssh = $fl_undo_af_ssh .\n\n";

	print STDOUT '%{$gb_new2incr_coord_undo_af_ssh} = ( # $indx_gb_new => $incr_bl_coord_undo,', "\n";

	foreach my $ngb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_ssh}) {
	    print STDOUT "\t", $ngb, ' => ', $gb_new2incr_coord_undo_af_ssh->{$ngb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A'') # ...}"

	unless ($fl_undo_af_ssh > 0) {
		print STDOUT "[CAUTION 'ssplit_plus_vmerge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets_1sbj' FAILED (w.r.t. \$gb_sbj1_unvmgd_NEW (=$gb_sbj1_unvmgd_NEW), because the horizontal overlaps canNOT be UNDONE!! ... I will STOP HERE ...: $! \n\n";
		return (-1, undef, undef, undef);
	}

=DISABLED in creation of 'ssplit_plus_vmerge_gblock_pair'. (M2post-f3)
			# First, Update %{$gb_new2incr_coord_undo_af_ssh}. #

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-f3) #
	    if (defined $gb_new2incr_coord_undo_af_ssh->{$gb_sbj1_unvmgd_NEW}) {
		$gb_new2incr_coord_undo_af_ssh->{$gb_sbj1_unvmgd_NEW} += $incr_coord_sbj1_unmgd_via_ssh;
	    } else {
		$gb_new2incr_coord_undo_af_ssh->{$gb_sbj1_unvmgd_NEW} = $incr_coord_sbj1_unmgd_via_ssh;
	    }
#=For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-f3)
#	    if (defined $gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW}) {
#		$gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW} += $incr_coord_ch_left_via_ssh;
#	    } else {
#		$gb2incr_coord_undo_af_ssh->{$gb_ch_left_NEW} = $incr_coord_ch_left_via_ssh;
#	    }
#=cut # End of "For the predecessor, 'vsplit_plus_smerge_gblock_pair' (M2post-f3)"

			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-f3)" #
=cut # End of "DISABLED in creation of 'ssplit_plus_vmerge_gblock_pair'. (M2post-f3)"


			# Then, Update %{$gb2incr_coord_undo_af_vmg} again. #

	foreach my $ngb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_ssh}) {
		my $incr_af_ssh = $gb_new2incr_coord_undo_af_ssh->{$ngb};
		if (defined $gb_new2incr_coord_undo_af_vmg->{$ngb}) {
		    $gb_new2incr_coord_undo_af_vmg->{$ngb} += $incr_af_ssh;
		} else {
		    $gb_new2incr_coord_undo_af_vmg->{$ngb} = $incr_af_ssh;
		}
	}

   if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A''') #

	print STDOUT "\n";
	##print STDOUT '[ FINAL %{$gb2incr_coord_undo_af_smg}: ]', "\n\n";
	print STDOUT '[ FINAL %{$gb_new2incr_coord_undo_af_vmg}: ]', "\n\n";

	print STDOUT '%{$gb_new2incr_coord_undo_af_vmg} = ( # $indx_gb_new => $incr_bl_coord_undo,', "\n";

	foreach my $ngb (sort {$a <=> $b} keys %{$gb_new2incr_coord_undo_af_vmg}) {
	    print STDOUT "\t", $ngb, ' => ', $gb_new2incr_coord_undo_af_vmg->{$ngb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 5) { # (M2post-A''') # ...}"


	##} # End of "if ($dist <= 0) { ...}" # MOVED to above 'undo_hovlp_vcmpl_gbsets_1sbj' (on 2019/10/13). #
	##} # End of "if (($lbd_kpt <= $lbd_ch_left) and ($rbd_ch_left <= $rbd_kpt)) {...}"

			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-e)" #

			# End of "ADDED in REVISION (on 2019/09/06)." #


   if ($TEST_LEVEL_ssvm_gbp >3) { # (M2post-B) #

	##print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Vertical Split + Merge (same-type)\" ]\n\n";
	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"(Horizontal) Same-Type Split + Vertical Merge\" ]\n\n"; # MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-B) #


	unless ( print_out_gblock_configs (@{$new_curr_set_clms}, @{$new_curr_bds_gblks}, @{$new_curr_bl_coords}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp >3) { # (M2post-B) # ...}"

   if ($TEST_LEVEL_ssvm_gbp > 3) { # (M2post-C) #

	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Vertical Split + Merge (same-type)\" ]\n\n";
	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"(Horizontal) Same-Type Split + Vertical Merge\" ]\n\n"; # MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M2post-C) #

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$new_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'ssplit_plus_vmerge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'ssplit_plus_vmerge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 3) { # (M2post-C) # ...}"



		# MAIN 3: Prepare Some Important Additional Sets of Data. #

	my @new_indices_old_gbs = my @old_indices_new_gbs = ();

	for (my $igb=0; $igb < $ct_gbs_intm; $igb++) {
	    my $ogb = $old_indices_intm_gbs->[$igb];
	    my $ngb = $new_indices_intm_gbs->[$igb];
	    if (defined $ogb) { $new_indices_old_gbs[$ogb] = $ngb; }
	    if (defined $ngb) { $old_indices_new_gbs[$ngb] = $ogb; }
	}

		# ADDED in BEGUGGING (on 2019/10/14). #
	foreach my $ogb ($gb_sbj1, $gb_sbj2) { $new_indices_old_gbs[$ogb] = undef; }
	foreach my $ngb ($gb_sbj1_unvmgd_NEW, $gb_mgd_NEW) { $old_indices_new_gbs[$ngb] = undef; }
		# END of "ADDED in BEGUGGING (on 2019/10/14)." #

			# MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M3) #

=For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (MAIN 3)
	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $igb = $intm_indices_new_gbs->[$ngb];
	    my @info_gb = @{$intm_list_gblocks->[$igb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_gb;
	}

	my @new_list_excluded_bin_hrels = ([$gb_kpt_NEW, $gb_ch_left_NEW, '>']);
=cut # END of "For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (MAIN 3)"

	my @new_list_gblocks = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my @info_gb = @{$new_list_gblocks->[$ngb]};
	    @info_gb[2,3] = @{$new_curr_bds_gblks->[$ngb]};
	    $new_list_gblocks[$ngb] = \@info_gb;
	}

	my @new_list_excluded_bin_hrels = ([$gb_mgd_NEW, $gb_sbj1_unvmgd_NEW, '-']);

			# END of "MODIFIED in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). (M3)" #

   if ($TEST_LEVEL_ssvm_gbp > 3) { # (M3) #

	print STDOUT "\n[[ Additional Output Data-Sets: ]]\n\n";

	##print STDOUT '($gb_ch_left_NEW, $gb_kpt_NEW) = ', "($gb_ch_left_NEW, $gb_kpt_NEW) .\n\n\n";
	print STDOUT '($gb_sbj1_unvmgd_NEW, $gb_mgd_NEW) = ', "($gb_sbj1_unvmgd_NEW, $gb_mgd_NEW) .\n\n\n"; # Modified in creation of 'ssplit_plus_vmerge_gblock_pair' (on 2019/10/09). #

	print STDOUT '[ Relationships between new and old indices of the gap-blocks', " (#{gap-blocks}_new=$ct_out_gbs, #{gap-blocks}_old=$ct_in_gbs): ]\n\n";

        my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

        foreach my $pair ( [\@new_indices_old_gbs, \@out_new_indices_old_gbs],
                          [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
                ) {
                my ($indices_gbs, $out_indices_gbs) = @{$pair};
                foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx 
: 'undef'); }
        }

        print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

        print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";

        print STDOUT "\n\n";

	print STDOUT '[ Contents of @new_list_gblocks', " (#{gap-blocks}_new = $ct_out_gbs): ]\n\n";

        print STDOUT join ("\t", 'Indx_gb_intm', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

        for (my $gb=0; $gb< $ct_out_gbs; $gb++) {
            my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$new_list_gblocks[$gb]}[0 .. 5];    
            print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
        }
        print STDOUT "\n\n";

	print STDOUT '[ Contents of @new_list_excluded_bin_hrels: ]', "\n\n";

	print STDOUT join ("\t", 'Indx_excluded', 'gb1', 'gb2', 'horizontal_relation'), "\n\n";

	for (my $i=0; $i < @new_list_excluded_bin_hrels; $i++) {
	    print STDOUT join ("\t", $i, @{$new_list_excluded_bin_hrels[$i]}), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_ssvm_gbp > 3) { # (M3) # ...}"


   (print STDOUT "< Subroutine 'ssplit_plus_vmerge_gblock_pair' Ended (w/ \$fl_end = 1, \$gb_sbj1_unvmgd_NEW = $gb_sbj1_unvmgd_NEW, \$gb_mgd_NEW = $gb_mgd_NEW) />\n\n") if ($TEST_LEVEL_ssvm_gbp > 0);

	# [ return (1,  $gb_ch1_NEW, $gb_ch2_NEW, $gb_shftd, $incr_coord_shftd, \@new_indices_old_gbs, \@old_indices_new_gbs, \%gb2incr_bl_coord_undo, ] ... for the predecessor, 'vsplit_gblock'. #


			### RESTART FROM HERE (on 2019/10/10). ###

	return (1, $gb_sbj1_unvmgd_NEW, $gb_mgd_NEW,
	$gb_sbj1_left_intm, $gb_sbj1_right_intm, $intm_indices_old_gbs, $gb_intm2incr_bl_coord_undo_ovlps_af_sspl, 
	$if_vmerged_w_lfrag, $if_shift_to_right, $gb_sbj_fxd_intm, $gb_sbj_shftd_intm,
	$incr_coord_sbj_shftd_B4_vmg, $gb_new2incr_coord_1way_dsplc_B4_vmg, $new_indices_intm_gbs,
	\@new_indices_old_gbs, \@old_indices_new_gbs, $gb_new2incr_coord_undo_af_vmg,

	# [[ (For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (OUT))
	# 		# Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT) #
	#  return (1, $gb_ch_left_NEW, $gb_kpt_NEW, 
	# $gb_shftd_af_vspl_intm, $incr_coord_shftd_af_vspl, $intm_indices_old_gbs, $gb2incr_bl_coord_undo_ovlps_af_vspl,
	# $gb_ch_2B_mgd_intm, $gb1_intm, $gb2_intm, $bl_sbj_ssh4smg_intm, $incr_coord_sbj_ssh4smg, $gb_absbd_intm, $new_indices_intm_gbs,
	# \@new_indices_old_gbs, \@old_indices_new_gbs, $gb2incr_coord_undo_af_smg,
	# 		# End of "Modified in creation of 'vsplit_plus_smerge_gblock_pair'. (OUT)" #
	# ]] ... End of "(For the predecessor, 'vsplit_plus_smerge_gblock_pair'. (OUT))"

	$new_curr_set_clms, $new_curr_bds_gblks, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	$new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	\@new_list_gblocks, \@new_list_excluded_bin_hrels
	);

}	# End of the subroutine, 'ssplit_plus_vmerge_gblock_pair'. #


# initial_revpurge_gblock_pair:
#	This satellite subroutine of 'reverse_purge_gblock_pair' performs the INITIAL "reverse-purge"
#	of the subject candidate region.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by modifying 'initial_csplit_gblock'.
#	[CREATED on 2019/09/15.]
#
# initial_csplit_gblock:
#	This satellite subroutine of 'csplit_gblock' performs the INITIAL "split (complementary types)"
#	of the subject gap-block.
#	And it also prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of a part of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by importing the functions to create a gap-block of a NEW TYPE, from 'cvmerge_gblock_pair'.
#	[CREATED on 2019/08/16.]
#
# csplit_gblock:
#	This subroutine "split"s the subject gap-block into a pair of gap-blocks of complementary types.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by incoporating a NEW satellite subroutine, 'initial_csplit_gblock'.
#	[CREATION STARTED on 2019/08/15, ENDED on 2019/08/16.]
#	
#
# [ INPUTS: ($gb_sbj, $avail_site, $char_abs,  ] ... for the predecessor, 'initial_csplit_gblock'.
#
#		# Modified in creation of 'initial_revpurge_gblock_pair'. (HD1) #
# INPUTS: (\@rows_affd_by_new_sbj, $avail_site, $char_abs, 
#		# End of "Modified in creation of 'initial_revpurge_gblock_pair'. (HD1)" #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm,
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1)" #
#	), where
#
#			# IMPORTED from 'reverse_purge_gblock_pair' in creation of 'initial_revpurge_gblock_pair'. (HD1) #
#	@rows_affd_by_new_sbj stores indices of the MSA rows that are to be affected by the newly created "subject" gap-block;
#			# End of "IMPORTED from 'reverse_purge_gblock_pair' in creation of 'initial_revpurge_gblock_pair'. (HD1)" #
#
#			# ADDED in creation of 'initial_csplit_gblock'. (HD1) #
#	# [ $gb_sbj is the index of the gap-block to be "split" into a pair of vertically complementary blocks; ] ... for the predecessor, 'initial_csplit_gblock'. #
#
#	$avail_site is the index of the site (i.e., column) available for the "Split (complementary types)";
#			# End of "ADDED in creation of 'initial_csplit_gblock'. (HD1)" #
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2)" #
#	
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
# OUTPUTS: ($flag_end, $gb_sbj_NEW, $gb_cmpl_NEW, \@new_indices_old_gbs, \@old_indices_new_gbs, \@record_session, 
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#	[ \@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@new_gblock_sizes ] ... for the predecessor, 'cmerge_split_grow_gblock_pair (NEW VERSION)'.
#			# IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#			# End of "IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD1)" #

#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_list_gblocks
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1)" #
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#			# IMPORTED from 'reverse_purge_gblock_pair' in creation of 'initial_revpurge_gblock_pair'. (HD2) #
#	$gb_sbj_NEW is the new index for the subject gap-block(, which grew in size during this subroutine);
#	$gb_cmpl_NEW is the new index for the vertically complementary gap-block;
#			# End of "IMPORTED from 'reverse_purge_gblock_pair' in creation of 'initial_revpurge_gblock_pair'. (HD2)" #
#
#	[[
#	$gb_sbj_NEW is the new index for the subject gap-block(, which grew in size during this subroutine);
#	$gb_cmpl_NEW is the new index for the vertically complementary gap-block;
#	]] ... for the predecessor, 'initial_csplit_gblock'.
#
#
#			# IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# End of "IMPORTED from 'ssplit_gblock' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#			# ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#			# 	=> MODIFIED in creation of 'initial_csplit_gblock'. #
#	@record_session = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	$session_ID (=1) is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_sbj th gap-block,
#	and $incr_coord_sbj_smg1 = $incr_coord_sbj_smg2 = 'n/a' always holds in this subroutine;
#	[ and $incr_coord_sbj_smg2 = 'n/a' always holds in this subroutine; ] ... for the predecessor, 'initial_csplit_gblock'.
#			# End of "ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#			#	 => MODIFIED in creation of 'initial_csplit_gblock'." #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
# [ sub initial_csplit_gblock ($$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb_sbj, $avail_site, $char_abs, ] ... for the predecessor, 'initial_csplit_gblock'.
#
sub initial_revpurge_gblock_pair (\@$$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

			# MODIFIED in creation of 'initial_revpurge_gblock_pair'. (IN) #
	my ($rows_affd_by_new_sbj, $avail_site, $char_abs,
			# End of "MODIFIED in creation of 'initial_revpurge_gblock_pair'. (IN)" #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,

		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq
		# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock'. (IN)" #
	) = @_;

# INPUTS: ($gb_sbj, $size_cmpl_gb, \@pot_avail_sites, $char_abs, 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

			# MODIFIED in creation of 'initial_revpurge_gblock_pair'. (BEG) #

	my $cnct_rows_affd_by_new_sbj = join (',', @{$rows_affd_by_new_sbj});

	my $sbjct_site = $avail_site;

    (print STDOUT "< Subroutine 'initial_revpurge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; {rows affected by newly created \"subject\"} = ($cnct_rows_affd_by_new_sbj); {available site} (=\$avail_site = \$sbjct_site) = $avail_site) >\n\n") if ($TEST_LEVEL_init_rvprg_gbp > 0);

=For the predecessor, 'initial_csplit_gblock'. (BEG)
			# MODIFIED in creation of 'initial_csplit_gblock'. (BEG) #

	my $size_sbj_bf = $gblock_sizes->[$gb_sbj];
	my $sbj_coord_bf = $curr_bl_coords->[$gb_sbj];

	my ($lbd_sbj_bf, $rbd_sbj_bf) = @{$curr_bds_blks->[$gb_sbj]};

	my $sbjct_site = $avail_site; # ADDED in CORRECTION on 2019/08/17.


    (print STDOUT "< Subroutine 'initial_csplit_gblock' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; gap-block to be \"Split\": \$gb_sbj = $gb_sbj; {subject_block_size}_bf = $size_sbj_bf; {subject_coordinate}_bf = $sbj_coord_bf; {boundaries}_sbj_bf = [$lbd_sbj_bf, $rbd_sbj_bf]; {available site} (=\$avail_site = \$sbjct_site) = $avail_site) >\n\n") if ($TEST_LEVEL_init_cspl_gb > 0);
			# END of "MODIFIED in creation of 'initial_csplit_gblock'. (BEG)" #
=cut # End of "For the predecessor, 'initial_csplit_gblock'. (BEG)"

			# END of "MODIFIED in creation of 'initial_revpurge_gblock_pair'. (BEG)" #

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

		# (Preliminary 1): Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = ();
	my @cp_curr_gblks_w_spec_lb = my @cp_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_in_clms; $c++) { # 1st preliminary for-loop (over the columns). #
	    my @cp_clm = @{$curr_set_clms->[$c]};
	    $cp_curr_set_clms[$c] = \@cp_clm;

	    my @gbs_w_lb_eq_c = @{$curr_gblks_w_spec_lb->[$c]};
	    $cp_curr_gblks_w_spec_lb[$c] = \@gbs_w_lb_eq_c;
	    my @gbs_w_rb_eq_c = @{$curr_gblks_w_spec_rb->[$c]};
	    $cp_curr_gblks_w_spec_rb[$c] = \@gbs_w_rb_eq_c;

	} # End of the 1st preliminary for-loop (over the columns). #

	my @cp_curr_bds_blks = my @cp_bds_gb_coords =  my @cp_gb_coords_wd_frm = ();
	my @cp_rows_affd_by_blks = ();
	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 2nd preliminary for-loop (over the gap-blocks). #
	    my @cp_bds = @{$curr_bds_blks->[$gb]};
	    $cp_curr_bds_blks[$gb] = \@cp_bds;

	    my @cp_bds_coords_for_gb = @{$bds_gb_coords->[$gb]}; 
	    $cp_bds_gb_coords[$gb] = \@cp_bds_coords_for_gb;

	    my @cp_wd_frm_for_gb = @{$gb_coords_wd_frm->[$gb]};
	    $cp_gb_coords_wd_frm[$gb] = \@cp_wd_frm_for_gb;

	    my @cp_rows_affd_by_gb = @{$rows_affd_by_blks->[$gb]};
	    $cp_rows_affd_by_blks[$gb] = \@cp_rows_affd_by_gb;

	} # End of the 2nd preliminary for-loop (over the gap-blocks). #

	my @cp_curr_bl_coords = @{$curr_bl_coords};


   if ($TEST_LEVEL_init_rvprg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of INITIAL \"Reverse-Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@cp_curr_set_clms, @cp_curr_bds_blks, @cp_curr_bl_coords, @cp_curr_gblks_w_spec_lb, @cp_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp >3) {...}"

   if ($TEST_LEVEL_init_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of INITIAL \"Reverse-Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @cp_bds_gb_coords, @cp_gb_coords_wd_frm, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @cp_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 5) {...}"


		# (Preliminary 2)': Create the list of gap-blocks BEFORE the "(incomplete) vertical merge". #
		# (IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock' (on 2019/08/16).) #

	my @list_old_gblocks = (); # element = [$br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows] 

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) { # 3rd preliminary for-loop (over gap-blocks). #
	    my @cp_affd_msa_rows = @{$rows_affd_by_blks->[$gb]};
	    my @affd_exnodes = ();
	    foreach my $row (@cp_affd_msa_rows) { push @affd_exnodes, $nodeids_in_msa->[$row]; }
	    if (@affd_exnodes > 1) { @affd_exnodes = sort {$a <=> $b} @affd_exnodes; }

	    my $cnct_affd_msa_rows = join (',', @cp_affd_msa_rows);
	    my $cnct_affd_exnodes = join (',', @affd_exnodes);
	    $list_old_gblocks[$gb] = [@{$list_gblocks0->[$gb]}[$indx_br, $indx_u_or_d], @{$cp_curr_bds_blks[$gb]}, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@cp_affd_msa_rows];

	} # End of the 3rd preliminary for-loop (over gap-blocks). #
	
# @{$new_list_gblocks[$ngb]} 

   if ($TEST_LEVEL_init_rvprg_gbp > 5) {
	print STDOUT "\n", '( Contents of @list_old_gblocks: )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$ct_in_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$list_old_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 5) {...}"

		# END of "(IMPORTED from 'cvmerge_gblock_pair' in creation of 'initial_csplit_gblock' (on 2019/08/16).)" #


		### Preliminary (Out): ###

	my ($gb_ncmpl_old, $gb_nsbj_old) = ($ct_in_gbs, $ct_in_gbs+1);
	##my ($gb_cmpl_old, $gb_incr_sbj) = ($ct_in_gbs, $ct_in_gbs+1); # For the predecessor, 'initial_csplit_gblock'.
	####my ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1); # For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).

	my $intm1_ct_gbs = $ct_in_gbs +2;
	my $ct_out_gbs = $intm1_ct_gbs;
	##my $ct_out_gbs = $ct_in_gbs +1; # For the predecessor, 'initial_csplit_gblock'.
	####my $intm2_ct_gbs = $ct_in_gbs +1; # For the predecessor 'cmerge_split_grow_gblock_pair' (New Version).

		# Create @intm1_gblock_sizes & @intm1_list_gblocks & @intm1_inter_block_relations. #

	my @cp_curr_gblock_sizes = @{$gblock_sizes};
	my @intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);

				## MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) ##

			# (NOTE: At this point, we don't care whether $lm_clm and $rm_clm are correct or not.)
	my @intm1_list_gblocks = @list_old_gblocks;

			# MODIFIED in creation of 'initial_revpurge_gblock_pair'. (Pre-OUT 1) #
	my @rows_affd_by_nsbj = @{$rows_affd_by_new_sbj};
	my @exnodes_affd_by_nsbj = ();
	foreach my $row (@rows_affd_by_nsbj) { push @exnodes_affd_by_nsbj, $nodeids_in_msa->[$row]; }
	if (@exnodes_affd_by_nsbj > 1) { @exnodes_affd_by_nsbj = sort {$a <=> $b} @exnodes_affd_by_nsbj; }

	my $cnct_rows_affd_by_nsbj = join (',', @rows_affd_by_nsbj);
	my $cnct_exnodes_affd_by_nsbj = join (',', @exnodes_affd_by_nsbj);

	my $br_up_or_lw_nsbj = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_nsbj};
	unless (defined $br_up_or_lw_nsbj) {
	     print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] IMPOSSIBLE!!: NO branch & upper/lower information is ASSOCIATED with the set of MSA rows, ($cnct_rows_affd_by_nsbj), that is supposed to be affected by the newly created \"subject\" gap-block: $! !\n\n";
	     return (0, undef, undef, undef);
	}
	my ($br_nsbj, $up_or_lw_nsbj) = @{$br_up_or_lw_nsbj};

	$intm1_list_gblocks[$gb_nsbj_old] = [$br_nsbj, $up_or_lw_nsbj, $avail_site+1, $avail_site+1, $cnct_exnodes_affd_by_nsbj, $cnct_rows_affd_by_nsbj, \@exnodes_affd_by_nsbj, \@rows_affd_by_nsbj];

=For the predecessor, 'initial_csplit_gblock'. (Pre-OUT 1)
	my @info_gb_incr_sbj = @{$list_old_gblocks[$gb_sbj]};
	@info_gb_incr_sbj[2,3] = ($avail_site+1, $avail_site+1);
	$intm1_list_gblocks[$gb_incr_sbj] = \@info_gb_incr_sbj;
=cut # End of "For the predecessor, 'initial_csplit_gblock'. (Pre-OUT 1)"
			# END of "MODIFIED in creation of 'initial_revpurge_gblock_pair'. (Pre-OUT 1)" #

	my $rows_affd_by_sbj = \@rows_affd_by_nsbj; # Modified in creation of 'initial_revpurge_gblock_pair'. #
	##my $rows_affd_by_sbj = $rows_affd_by_blks->[$gb_sbj];

	my @all_rows = (0 .. $ct_rows-1);
	my ($common, $ex_sbj, $rows_affd_by_cmpl) = diff_nmarray (@{$rows_affd_by_sbj}, @all_rows);
	unless (defined $rows_affd_by_cmpl) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'diff_nmarray', FAILED to create the set of rows affected by the newly created complementary block: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
	my @exnodes_affd_by_cmpl = ();
	foreach my $row (@{$rows_affd_by_cmpl}) { push @exnodes_affd_by_cmpl, $nodeids_in_msa->[$row]; }
	if (@exnodes_affd_by_cmpl > 1) { @exnodes_affd_by_cmpl = sort {$a <=> $b} @exnodes_affd_by_cmpl; }

	my $cnct_rows_affd_by_cmpl = join (',', @{$rows_affd_by_cmpl});
	my $cnct_exnodes_affd_by_cmpl = join (',', @exnodes_affd_by_cmpl);

	my $br_up_or_lw_cmpl = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_cmpl};
	unless (defined $br_up_or_lw_cmpl) {
	     print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] IMPOSSIBLE!!: NO branch & upper/lower information is ASSOCIATED with the set of MSA rows, ($cnct_rows_affd_by_cmpl), that is supposed to be affected by the newly created complementary gap-block: $! !\n\n";
	     return (0, undef, undef, undef);
	}
	my ($br_cmpl, $up_or_lw_cmpl) = @{$br_up_or_lw_cmpl};

	$intm1_list_gblocks[$gb_ncmpl_old] = [$br_cmpl, $up_or_lw_cmpl, $avail_site, $avail_site, $cnct_exnodes_affd_by_cmpl, $cnct_rows_affd_by_cmpl, \@exnodes_affd_by_cmpl, $rows_affd_by_cmpl]; # Modified in creation of 'initial_revpurge_gblock_pair'. #
	##$intm1_list_gblocks[$gb_cmpl_old] = [$br_cmpl, $up_or_lw_cmpl, $avail_site, $avail_site, $cnct_exnodes_affd_by_cmpl, $cnct_rows_affd_by_cmpl, \@exnodes_affd_by_cmpl, $rows_affd_by_cmpl];


   if ($TEST_LEVEL_init_rvprg_gbp > 3) {

	print STDOUT "\n", '( Contents of @intm1_list_gblocks (maybe with incorrect lm_clm and rm_clm): )', "\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_clm', 'rm_clm', '(external,nodes,affected)', '[msa,rows,affected]'), "\n\n";

	for (my $gb=0; $gb<$intm1_ct_gbs; $gb++) {
	    my ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows) = @{$intm1_list_gblocks[$gb]}[0 .. 5];    
	    print STDOUT join ("\t", $gb, $br_id, $up_or_lw, $lm_clm, $rm_clm, "($cnct_affd_exnodes)", "[$cnct_affd_msa_rows]"), "\n";
	}
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 3) {...}"

	my ($intm1_inter_block_relations)
	  = mk_inter_block_relations (@intm1_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $intm1_inter_block_relations) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@intm1_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@intm1_list_gblocks', " ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1) ;

	my @intm1_inter_block_relations = @{$intm1_inter_block_relations};

				## END of "MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out))" ##

		# Create $intm1_cts_vincluding, etc. #

	my ($intm1_cts_vincluding, $intm1_cts_vincluded, $intm1_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm1_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'initial_revpurge_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm1_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm1_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1) ;


		# Create @intm1_rows_affd_by_blks.

				## MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) (2) ##
	my @intm1_rows_affd_by_blks = @cp_rows_affd_by_blks;

				# => Further Modified in creation of 'initial_revpurge_gblock_pair'. (Pre (Out)) (2') #
	$intm1_rows_affd_by_blks[$gb_ncmpl_old] = $rows_affd_by_cmpl;
	$intm1_rows_affd_by_blks[$gb_nsbj_old] = $rows_affd_by_sbj;
	##$intm1_rows_affd_by_blks[$gb_cmpl_old] = $rows_affd_by_cmpl;
	##$intm1_rows_affd_by_blks[$gb_incr_sbj] = $rows_affd_by_sbj;
				# End of "=> Further Modified in creation of 'initial_revpurge_gblock_pair'. (Pre (Out)) (2')" #
	
	###my @intm1_rows_affd_by_blks = (@cp_rows_affd_by_blks, @cp_rows_affd_by_blks[$gb_left, $gb_right]);
	#####my @intm2_rows_affd_by_blks = (@cp_rows_affd_by_blks, $cp_rows_affd_by_blks[$gb_left]);
				## END of "MODIFIED in creation of 'initial_csplit_gblock'. (Pre (Out)) (2)" ##

		### MAIN 1: "Split" the $avail_site th column.
		### [ & MAIN 2: "Merge" the resulting $gb_incr_sbj with $gb_sbj. ] ... for the predecessor, 'initial_csplit_gblock'.

	my $rels_w_sbj = $intm1_inter_block_relations->[$gb_nsbj_old]; # Modified in creation of 'initial_revpurge_gblock_pair'.
	##my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $curr_ct_clms = $ct_in_clms;

		## MAIN I: Create a new pair of single-column gap-blocks, by "split"ting the $sbjct_site th site. ##

			# Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-a) #
	my @clm_gb_ncmpl = @null_clm;
	my @clm_gb_nsbj = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_sbj}) { # 1st middle foreach.
	    my $tmp = $clm_gb_nsbj[$row];
	    $clm_gb_nsbj[$row] = $clm_gb_ncmpl[$row];
	    $clm_gb_ncmpl[$row] = $tmp;
	} # End of the 1st middle foreach.

=For the predecessor, 'initial_csplit_gblock'. (M I-a)
	my @clm_gb_cmpl = @null_clm;
	my @clm_incr_sbj = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_sbj}) { # 1st middle foreach.
	    my $tmp = $clm_incr_sbj[$row];
	    $clm_incr_sbj[$row] = $clm_gb_cmpl[$row];
	    $clm_gb_cmpl[$row] = $tmp;
	} # End of the 1st middle foreach.
=cut # END of "For the predecessor, 'initial_csplit_gblock'. (M I-a)"
			# End of "Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-a)" #

	my @gblks_w_lb_into_sbj = my @gblks_w_lb_into_cmpl = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$sbjct_site]}) { # 2nd middle foreach.
	    my $rel = $rels_w_sbj->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_lb_into_cmpl, $gb;
	    } else {
		push @gblks_w_lb_into_sbj, $gb;
	    }
	} # End of the 2nd middle foreach.
			# Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-b) #
	push @gblks_w_lb_into_sbj, $gb_nsbj_old;
	push @gblks_w_lb_into_cmpl, $gb_ncmpl_old;
	##push @gblks_w_lb_into_sbj, $gb_incr_sbj;
	##push @gblks_w_lb_into_cmpl, $gb_cmpl_old;
			# End of "Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-b)" #

	my @gblks_w_rb_into_sbj = my @gblks_w_rb_into_cmpl = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$sbjct_site]}) { # 3rd middle foreach.
	    my $rel = $rels_w_sbj->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_rb_into_cmpl, $gb;
	    } else {
		push @gblks_w_rb_into_sbj, $gb;
	    }
	} # End of the 3rd middle foreach.
			# Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-c) #
	push @gblks_w_rb_into_sbj, $gb_nsbj_old;
	push @gblks_w_rb_into_cmpl, $gb_ncmpl_old;
	##push @gblks_w_rb_into_sbj, $gb_incr_sbj;
	##push @gblks_w_rb_into_cmpl, $gb_cmpl_old;
			# End of "Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-c)" #

	my @intm1_curr_set_clms = ();
	my @intm1_curr_gblks_w_spec_lb = my @intm1_curr_gblks_w_spec_rb = ();

	for (my $c=0; $c < $sbjct_site; $c++) { # 1st Middle MAIN for-loop (over sites on the left of $sbjct_site). #
	    $intm1_curr_set_clms[$c] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 1st Middle MAIN for-loop. #

			# Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-d) #
	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = (\@clm_gb_ncmpl, \@clm_gb_nsbj) ;
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_lb_into_cmpl, \@gblks_w_lb_into_sbj) ;
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_rb_into_cmpl, \@gblks_w_rb_into_sbj) ;
=For the predecessor, 'initial_csplit_gblock'. (M I-d)
	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@clm_incr_sbj, \@clm_gb_cmpl) : (\@clm_gb_cmpl, \@clm_incr_sbj) ;
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@gblks_w_lb_into_sbj, \@gblks_w_lb_into_cmpl) : (\@gblks_w_lb_into_cmpl, \@gblks_w_lb_into_sbj) ;
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = ($rbd_sbj_bf < $avail_site) ? (\@gblks_w_rb_into_sbj, \@gblks_w_rb_into_cmpl) : (\@gblks_w_rb_into_cmpl, \@gblks_w_rb_into_sbj) ;;
=cut # End of "For the predecessor, 'initial_csplit_gblock'. (M I-d)"
			# End of "Modified in creation of 'initial_revpurge_gblock_pair'. (MAIN I-d)" #

	for (my $c = $sbjct_site+1; $c <$curr_ct_clms; $c++) { # 2nd Middle MAIN for-loop (over sites on the right of $sbjct_site). #
	    $intm1_curr_set_clms[$c+1] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c+1] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c+1] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 2nd Middle MAIN for-loop. #


	my $intm1_ct_clms = @intm1_curr_set_clms;


		# Create @intm1_curr_bds_blks. #

	my @intm1_curr_bds_blks = ();
	for (my $gb=0; $gb < $intm1_ct_gbs; $gb++) { $intm1_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$intm1_ct_clms; $c++) { # 2nd trailing for-loop for Main 1-(i). #
            foreach my $gb (@{$intm1_curr_gblks_w_spec_lb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[0] = $c;
            }
            foreach my $gb (@{$intm1_curr_gblks_w_spec_rb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[1] = $c;
            }
	} # End of the 2nd trailing for-loop for Main 1-(i). #


		## Preliminary (In 2): Create the coordinate frame for the 1st intermediate gap-block configuration. #

	my ($intm1_le_wd, $intm1_re_wd) = (0, $intm1_ct_clms-1);

	my ($intm1_orgs_gb_coords, $intm1_bds_gb_coords, $intm1_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@intm1_curr_bds_blks, $intm1_le_wd, $intm1_re_wd, @intm1_inter_block_relations, $maxct_1way_shifts);
	unless (defined $intm1_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column : $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1);

	my @intm1_curr_bl_coords = @{$intm1_orgs_gb_coords};


			# (Added on 2019/08/07. (Main II)) #

   if ($TEST_LEVEL_init_rvprg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_gblock_configs (@intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp >5) {...}"

   if ($TEST_LEVEL_init_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_remng_attr_gblks (@intm1_inter_block_relations, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm}, @intm1_gblock_sizes, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 5) {...}"

		## MAIN II-(i): To Ease the Modification from 'init_csplit_gblock' to 'init_revpurge_gblock_pair',
		##		Create a MOCK set of "2nd intermediate" Data-Sets. 
		## (ADDED in creation of 'initial_revpurge_gblock_pair'. (MAIN II-(i))) ##

	my $intm2_curr_set_clms = \@intm1_curr_set_clms;
	my $intm2_curr_bds_blks = \@intm1_curr_bds_blks;

	##my $intm2_curr_bl_coords = \@intm1_curr_bl_coords;
	##my $intm2_bds_gb_coords = $intm1_bds_gb_coords;
	##my $intm2_gb_coords_wd_frm = $intm1_gb_coords_wd_frm;

	##my $intm2_curr_gblks_w_spec_lb = \@intm1_curr_gblks_w_spec_lb;
	##my $intm2_curr_gblks_w_spec_rb = \@intm1_curr_gblks_w_spec_rb;

	my $intm2_gblock_sizes = \@intm1_gblock_sizes;
	my $intm2_rows_affd_by_blks = \@intm1_rows_affd_by_blks;

		## End of "(ADDED in creation of 'initial_revpurge_gblock_pair'. (MAIN II-(i)))" ##

=For the predecessor, 'initial_csplit_gblock'. (MAIN II-(i))

		## MAIN II-(i): Merge the $gb_sbj and $gb_incr_sbj th gap-block, via 'smerge_gblock_pair'. ##

	## ($gb_cmpl_old, $gb_incr_sbj) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the subroutine 'smerge_gblock_pair', to merge the \$gb_sbj (=$gb_sbj) th and \$gb_incr_sbj (=$gb_incr_sbj) th gap-blocks ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp > 1);

	my ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_sbj, $gb_incr_sbj, @intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm},
		@intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, @intm1_rows_affd_by_blks, $cnct_null_clm,
		@intm1_inter_block_relations, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_gblock_sizes, $maxct_1way_shifts
		);
	unless ($flag_end_smg1) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1);



   if ($TEST_LEVEL_init_rvprg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 'smerge_gblock_pair' (to \"merge\" \$gb_sbj (=$gb_sbj) and \$gb_incr_sbj (=$gb_incr_sbj) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp >5) {...}"

   if ($TEST_LEVEL_init_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 'smerge_gblock_pair' for (to \"merge\" \$gb_sbj (=$gb_sbj) and \$gb_incr_sbj (=$gb_incr_sbj) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm2_inter_block_relations}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm}, @{$intm2_gblock_sizes}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 5) {...}"

=cut # END of "For the predecessor, 'initial_csplit_gblock'. (MAIN II-(i))"


		## Record the Core information on this Session. ##

	my @record_session = (1, $sbjct_site, 'n/a', 'n/a'); # Further MODIFIED in creation of 'initial_revpurge_gblock_pair'. #
	##my @record_session = (1, $sbjct_site, $incr_coord_sbj_smg1, 'n/a'); # MODIFIED in creation of 'initial_csplit_gblock'. #
	####push @records_sessions, [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

			# End of "(Added on 2019/08/07. (Main II))" #


		### MAIN IIIpre:  Create the correspondence between the old & new indices of the gap-blocks. ###
		### (ADDED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN IIIpost) ##


	my @intm2_list_gblocks = ();
	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my @info_gb = @{$intm1_list_gblocks[$gb]};
	    @info_gb[2,3] = @{$intm2_curr_bds_blks->[$gb]};
	    $intm2_list_gblocks[$gb] = \@info_gb;
	}

	my @old_indices_new_gbs = sort { cmp_gblocks3 (@{$intm2_list_gblocks[$a]}, @{$intm2_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1);
	my @new_indices_old_gbs = ();
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my $ogb = $old_indices_new_gbs[$ngb];
	    $new_indices_old_gbs[$ogb] = $ngb;
	}

			# MODIFIED in creation of '
	my $gb_sbj_NEW = $new_indices_old_gbs[$gb_nsbj_old];
	my $gb_cmpl_NEW = $new_indices_old_gbs[$gb_ncmpl_old];
	##my $gb_sbj_NEW = $new_indices_old_gbs[$gb_sbj];
	##my $gb_cmpl_NEW = $new_indices_old_gbs[$gb_cmpl_old];

   if ($TEST_LEVEL_init_rvprg_gbp > 2) {

	print STDOUT "\n[ Relationships between old and new indices of gap-blocks (#{gap-blocks}_out=$ct_out_gbs) : ]\n\n";

	my @out_new_indices_old_gbs = my @out_old_indices_new_gbs = ();

	foreach my $pair ([\@new_indices_old_gbs, \@out_new_indices_old_gbs],
			  [\@old_indices_new_gbs, \@out_old_indices_new_gbs]
		) {
		my ($indices_gbs, $out_indices_gbs) = @{$pair};
		foreach my $indx (@{$indices_gbs}) { push @{$out_indices_gbs}, ((defined $indx) ? $indx : 'undef'); }
	}

	print STDOUT '@new_indices_old_gbs = (', join (', ', @out_new_indices_old_gbs), ") .\n\n";

	print STDOUT '@old_indices_new_gbs = (', join (', ', @out_old_indices_new_gbs), ") .\n";
	print STDOUT "\n\n";

	print STDOUT '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 2) {...}"


		### END of "(ADDED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN IIIpost)" ##

		### MAIN III: Create the Final Output Data-Sets. ###
		### (Added on 2019/08/07. (MAIN III)) ###
		### (=> MODIFIED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN III) ###

	my @new_curr_set_clms = @{$intm2_curr_set_clms};
	my $ct_out_clms = @new_curr_set_clms;

	my @new_curr_bds_blks = @{$intm2_curr_bds_blks}[@old_indices_new_gbs];

	my @new_gblock_sizes = @{$intm2_gblock_sizes}[@old_indices_new_gbs];
	my @rows_affd_by_new_blks = @{$intm2_rows_affd_by_blks}[@old_indices_new_gbs];

	# ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	#    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	#    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	#    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes

	my @new_curr_gblks_w_spec_lb = my @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $ngb=0; $ngb <$ct_out_gbs; $ngb++) {
	    my ($lbd, $rbd) = @{$new_curr_bds_blks[$ngb]};
	    push @{$new_curr_gblks_w_spec_lb[$lbd]}, $ngb;
	    push @{$new_curr_gblks_w_spec_rb[$rbd]}, $ngb;
	}

			# Create  @new_list_gblocks & @{$new_inter_block_relations}. #

	my @new_list_gblocks = @intm2_list_gblocks[@old_indices_new_gbs];

	my ($new_inter_block_relations)
	  = mk_inter_block_relations (@new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@new_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@new_list_gblocks', " ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1) ;

			# Create @{$new_cts_vincluding}, etc. #

	my ($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'initial_revpurge_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@{$new_inter_block_relations}', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1) ;


			# Create @new_curr_bl_coords  & @new_bds_gb_coords  & @new_gb_coords_wd_frm . #

	my ($new_le_wd, $new_re_wd) = (0, $ct_out_clms-1);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'initial_revpurge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks after Whole Process of INITIAL \"Split (complementary types)\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks after the Whole Process of the INITIAL \"Split (complementary types)\" ...\n\n") if ($TEST_LEVEL_init_rvprg_gbp >1);

	my @new_curr_bl_coords    = @{$new_orgs_gb_coords} ;
	my @new_bds_gb_coords     = @{$new_bds_gb_coords} ;
	my @new_gb_coords_wd_frm  = @{$new_gb_coords_wd_frm} ;

	##my @new_curr_bl_coords = @{$intm2_curr_bl_coords}[@old_indices_new_gbs];
	##my @new_bds_gb_coords = @{$intm2_bds_gb_coords}[@old_indices_new_gbs];
	##my @new_gb_coords_wd_frm = @{$intm2_gb_coords_wd_frm}[@old_indices_new_gbs];

		### END of "(=> MODIFIED in creation of 'initial_csplit_gblock' (on 2019/08/16).) (MAIN III)" ###


   if ($TEST_LEVEL_init_rvprg_gbp >2) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of INITIAL \"Reverse-Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp >2) {...}"

   if ($TEST_LEVEL_init_rvprg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of INITIAL \"Reverse-Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @rows_affd_by_new_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_revpurge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_revpurge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_init_rvprg_gbp > 3) {...}"

	
    (print STDOUT "\n< Subroutine 'initial_revpurge_gblock_pair' Ended (w/ ", '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) . >\n\n") if ($TEST_LEVEL_init_rvprg_gbp > 0);


	return (1, $gb_sbj_NEW, $gb_cmpl_NEW, \@new_indices_old_gbs, \@old_indices_new_gbs, \@record_session,  
	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks);

}	# End of the subroutine, 'initial_revpurge_gblock_pair'. #


# reverse_purge_gblock_pair:
#	This subroutine "reverse"s the supposed "purge" of a gap-block pair,
#	by "splitting" the column at each site in a given set into vertically complementary columns
#	and by 'smerg'ing the resulting columns into a pair of mutually complementary gap-blocks.
#	The subroutine was created by modifying 'csplit_gblock'.
#	[CREATION STARTED on 2019/09/13, ENDED on 2019/09/15.]
#
# csplit_gblock:
#	This subroutine "split"s the subject gap-block into a pair of gap-blocks of complementary types.
#	And it prepares some important data sets for the resulting local alignment.
#	It was created by taking advantage of the framework of 'cmerge_split_grow_gblock_pair' (NEW VERSION) below,
#	and by incoporating a NEW satellite subroutine, 'initial_csplit_gblock'.
#	[CREATION STARTED on 2019/08/15, ENDED on 2019/08/16.]
#	[REVISED on 2019/08/17-19]:
#	Corrected a BUG; as a result, the subroutine can now order the gap-blocks CORRECTLY
#	even if the size-growing "split + merge" process changes
#	the intrinsic ranks between some vertically equivalent gap-blocks.
#
#	... OMITTED ...
#
# purge_gblock_pair:
#	This subroutine actually perform the "purge" of a specified pair of gap-blocks.
#	And prepares some important data sets for the resulting local alignment.
#	[CREATION STARTED on 2019/07/11, ENDED on 2019/07/14.]
#	[REVISED on 2019/07/15]:
#	Now, the subroutine also returns $bl_sbj and $incr_coord_sbj,
#	both which are the key inputs of 'serial_shifts_sgl_gblock'.
#
# [ INPUTS: ($gb_sbj, $size_cmpl, \@pot_avail_sites, $char_abs,  ] ... for the predecessor, 'csplit_gblock'.
#
#			# MODIFIED in creation of 'reverse_purge_gblock_pair'. (HD1) #
# INPUTS: (\@rows_affd_by_new_sbj, $size_rvprg, \@sites_in_pc_rg, $char_abs, 
#			# End of "MODIFIED in creation of 'reverse_purge_gblock_pair'. (HD1)" #
#
#	\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@bds_gb_coords, \@gb_coords_wd_frm,
#	\@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \@rows_affd_by_gblks, $cnct_null_clm,
#	\@inter_block_relations, \@cts_vincluding, \@cts_vincluded, \@cts_onx, \@gblock_sizes, $maxct_1way_shifts,
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1) #
#	\@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa, # Added on 2019/07/23.
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq # (This line was added on 2019/07/26.) #
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 1)" #

#	), where
#
#			# ADDED in creation of 'reverse_purge_gblock_pair'. (HD2) #
#	@rows_affd_by_new_sbj stores indices of the MSA rows that are to be affected by the newly created "subject" gap-block;
#
#	$size_rvprg is the size of the "reverse-purge", that is, the size that the newly created "subject" gap-block will have at the end of this subroutine;
#
#	@sites_in_mpc stores all the sites in the "purge" candidate region that is currently used;
#	[Typically, @rows_affd_by_new_sbj and @sites_in_mpc are components of a main output of 'readin_list_mgd_pcands' in "MyANEX_IO.pm"];
#	[Typically, $size_rvprg is an element of the output of 'list_revpurge_cand_sizes' above];
#			# End of "ADDED in creation of 'reverse_purge_gblock_pair'. (HD2)" #
#
#	[[
#			# ADDED in creation of 'csplit_gblock'. (HD1) #
#	$gb_sbj is the index of the gap-block to be "split" into a pair of vertically complementary blocks;
#
#	$size_cmpl is the final size of the newly created block vertically complementary to $gb_sbj;
#
#	@pot_avail_sites stores the indices of the sites (i.e., columns) POTENTIALLY available for the "Split (complementary types)", in the order to be used;
#			# End of "ADDED in creation of 'csplit_gblock'. (HD1)" #
#	]] ... for the predecessor, 'csplit_gblock'.
#
#	$char_abs is the character representing the "absence" state (i.e., a gap);
#
#		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2) #
#	@list_gblocks0 stores information on the gap-blocks. More precisely,
#	($br, $up_or_lw) = @{$list_gbloks0[$gb]}[$indx_br, $indx_u_or_d] give information on the $gb th gap-block, where
#	$br is the ID of the branch separating the $gb th block, and
#	$up_or_lw = 'U'/'L' indicates that the $gp th block is on the upper/lower-end of $br;
#	[Typically, @list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of, e.g., 'mk_set_msa_rows2br_up_or_lw_dp' in "MyTreeMap_sppl.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	[Typically, \@nodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows', both in "MyTreeMap_sppl.pm"];
#
#       %node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#       %node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#       $topnode_id is the ID of the top node (i.e., root);
#
#       %br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#       [Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#		# End of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (HD 2)" #
#	
#	For details on other input arguments, refer to the header of 'purge_gblock_pair' above.
#
#
#
# OUTPUTS: ($flag_end, $gb_sbj_NEW, $gb_cmpl_NEW, $gb_left, $gb_right, \@new_indices_old_gbs, \@old_indices_new_gbs,
#	   \@records_sessions, \%gb2incr_bl_coord_undo,
#
#	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
#			# IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD1) #
#	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, \@rows_affd_by_new_blks, 
#	\@new_inter_block_relations, \@new_cts_vincluding, \@new_cts_vincluded, \@new_cts_onx, \@new_gblock_sizes,
#			# End of "IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD1)" #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1) #
#	\@new_list_gblocks,
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD1)" #
#	\@med_indices_new_gbs # ADDED in revision on 2019/08/19.
#	), where
#
#	$flag_end = 1 if the subroutine ended without any remarkable problems,
#	          = 0 if it failed;
#		  = -1 if it failed BECAUSE the number of columns available for the "split" is smaller than the increment of gap-block sizes at some session; (ADDED on 2019/08/07.)
#
#			# ADDED in creation of 'csplit_gblock'. (HD) #
			# => Modified the meaning in creation of 'reverse_purge_gblock_pair'. (HD) #
#	$gb_sbj_NEW is the new index for the NEW subject gap-block;
#	$gb_cmpl_NEW is the new index for the NEW vertically complementary gap-block;
#	[[
#	$gb_sbj_NEW is the new index for the subject gap-block(, which grew in size during this subroutine);
#	$gb_cmpl_NEW is the new index for the vertically complementary gap-block;
#	]] ... for the predecessor, 'csplit_gblock'.
#			# END of "ADDED in creation of 'csplit_gblock'. (HD) #
#			# => Modified the meaning in creation of 'reverse_purge_gblock_pair'. (HD)" #
#
#	$gb_left and $gb_right are the indices of the gap-blocks that are on the left and on the right, respectively, each chosen between $gb_sbj_NEW and $gb_cmpl_NEW;
#
#			# IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD2) #
#	@new_indices_old_gbs stores the new indices assigned to the gap-blocks AFTER the "merge";
#	More precisely, the gap-block that had been the $ogb th before the "merge"
#	has been assigned $new_indices_gbs[$ogb] after the "merge"
#	(, and $new_indices_gbs[$gb_sbj] is undefined);
#
#	@old_indices_new_gbs stores the old indices assigned to the gap-blocks BEFORE the "merge";
#	More precisely, the gap-block that is the $ngb th after the "merge"
#	had been assigned $old_indices_gbs[$ngb] before the "merge"
#	(, and $old_indices_new_gbs[$new_gb_left] and $old_indices_new_gbs[$new_gb_right] are undefined);
#			# End of "IMPORTED from 'ssplit_gblock' in creation of 'csplit_gblock'. (HD2)" #
#
#			# ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version). #
#	@{$records_sessions[$k]} = ($session_ID, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2)
#	provides the record of the $k th session, where
#	$session_ID is the ID of the session (integer),
#	$sbjct_site is the index of the column that is "split" in the session,
#	$incr_coord_sbj_smg1/2 is the increment of the coordinate of the gap-blocks
#	created by the "split" of the subject column necessary for them to be "absorb"ed to the $gb_right/$gb_left th gap-block;
#			# End of "ADDED in creation of 'cmerge_split_grow_gblock_pair' (New Version)." #
#
#			
#			# ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09). #
#	%gb2incr_bl_coord_undo = ($indx_gb => {increment of the block coordinate resulting from 'UNDO'ing the horizontal overlaps generated by the "merge + split"}, ...);
#	(if $flag_end = -1, these 'UNDO'ing process is NOT actually executed);
#	[\%gb2incr_bl_coord_undo is one of the main outputs of 'undo_hovlp_vcmpl_gbsets' above];
#			# End of "ADDED in revision of 'cmerge_split_grow_gblock_pair' (New Version) (on 2019/08/09)." #
#
#			# IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2) #
#	@{$new_list_gblocks[$ngb]} = ($br_id, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows) provides information on the $ngb th new gap-block, where
#	the gap-block is separated by, and on the $up_or_lw side of, branch $br_id,
#	and $lm_clm and $rm_clm are the indices of the left-most and right-most columns of the gap-block,
#	@affd_exnodes stores the IDs of the external nodes affected by the gap-block,
#	@affd_msa_rows stores the indices of the MSA rows affected by the gap-block,
#	$cnct_affd_exnodes = join (',', @affd_exnodes),
#	$cnct_affd_msa_rows = join (',', @affd_msa_rows);
#	[It conforms to the format employed by the output of 'readin_list_gblks' and 'revise_list_gblocks', both in "MyANEX_IO.pm"];
#			# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of '(initial_)csplit_gblock'. (HD2)" #
#
#			# ADDED in REVISION (on 2019/08/19). #
#	@med_indices_new_gbs records the re-ordering of the new gap-blocks
#	resulting from growing the complementary gap-block pair by accruing the products of "split"ting the available sites;
#	Specifically, $med_indices_new_gbs[$ngb] gives the index that
#	the $ngb th gap-block (in the final output data-sets) had immediately after the subroutine 'initial_csplit_gblock';
#			# END of "ADDED in REVISION (on 2019/08/19)." #
#
#	The other output arguments are the copies of the input arguments, which are modified accordingly during the processes of "shift"s and "merge";
#
#
# [ sub csplit_gblock ($$\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {
#	my ($gb_sbj, $size_cmpl, $pot_avail_sites, $char_abs, ] ... for the predecessor, 'csplit_gblock '.
#
sub reverse_purge_gblock_pair (\@$\@$\@\@\@\@\@\@\@\@$\@\@\@\@\@$\@$$\%\@\%\%$\%) {

		# MODIFIED in creation of 'reverse_purge_gblock_pair'. (IN) #
	my ($rows_affd_by_new_sbj, $size_rvprg, $sites_in_pc_rg, $char_abs, 
		# End of "MODIFIED in creation of 'reverse_purge_gblock_pair'. (IN)" #

	$curr_set_clms, $curr_bds_blks, $curr_bl_coords, $bds_gb_coords, $gb_coords_wd_frm, 
	$curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $rows_affd_by_blks, $cnct_null_clm,
	$inter_block_relations, $cts_vincluding, $cts_vincluded, $cts_onx, $gblock_sizes, $maxct_1way_shifts,

		# IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (IN) #
	$list_gblocks0, $indx_br, $indx_u_or_d, $set_msa_rows2br_up_or_lw, $nodeids_in_msa,
		# \@list_gblocks0, $indx_br, $indx_u_or_d, \%set_msa_rows2br_up_or_lw, \@nodeids_in_msa
	$node2pa, $node2ch, $topnode_id, $br2eq
		# END of "IMPORTED from 'cvmerge_gblock_pair' in creation of 'csplit_gblock'. (IN)" #
	) = @_;

# INPUTS: (\@rows_affd_by_new_sbj, $size_rvprg, \@sites_in_pc_rg, $char_abs, 


	my $ct_in_clms = @{$curr_set_clms};
	my $ct_in_gbs = @{$curr_bds_blks};

	my $ct_rows = @{$curr_set_clms->[0]};

			# MODIFIED in creation of 'csplit_gblock'. (BEG) #
	my $ct_sites_in_pcrg = @{$sites_in_pc_rg};

	unless ($size_rvprg <= $ct_sites_in_pcrg) {
	    my $ct = $ct_sites_in_pcrg;
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The number of sites in the \"purge\"-candidate region (=$ct) is SMALLER THAN the final size of the Newly created gap-blocks (=$size_rvprg): $! !\n\n";
	    return (0, undef, undef, undef);
	}

	my $indx_beg_site = int (($ct_sites_in_pcrg - $size_rvprg)/2);
	my $indx_end_site = $indx_beg_site + $size_rvprg -1;

	my @avail_sites = @{$sites_in_pc_rg}[$indx_beg_site .. $indx_end_site];

	my $cnct_rows_affd_by_new_sbj = join (',', @{$rows_affd_by_new_sbj});

    (print STDOUT "< Subroutine 'reverse_purge_gblock_pair' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; {MSA rows to be affected by the newly created \"subject\"} = ($cnct_rows_affd_by_new_sbj); {final size of the newly created gap-blocks} = $size_rvprg; {available sites} = (", join (',', @avail_sites), "). ) >\n\n") if ($TEST_LEVEL_rvprg_gbp > 0);

=For the predecessor, 'csplit_gblock'.
	my $size_sbj_bf = $gblock_sizes->[$gb_sbj];
	my $sbj_coord_bf = $curr_bl_coords->[$gb_sbj];
	my ($lbd_sbj_bf, $rbd_sbj_bf) = @{$curr_bds_blks->[$gb_sbj]};

    (print STDOUT "< Subroutine 'csplit_gblock' (#{columns}_in=$ct_in_clms; #{rows}=$ct_rows; #{gap-blocks}_in=$ct_in_gbs; gap-block to be \"Split\": \$gb_sbj = $gb_sbj; {subject_block_size}_bf = $size_sbj_bf; {subject_coordinate}_bf = $sbj_coord_bf; {boundaries}_sbj_bf = [$lbd_sbj_bf, $rbd_sbj_bf]; {complementary_block_size}_fin = $size_cmpl; {available sites} = (", join (',', @avail_sites), "). ) >\n\n") if ($TEST_LEVEL_cspl_gb > 0);
=cut # End of "For the predecessor, 'csplit_gblock'."

   if ($TEST_LEVEL_rvprg_gbp >3) {

	print STDOUT "\n[ Gap-Block Configurations BEFORE the Whole Process of \"Reverse-Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >3) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks BEFORE the Whole Process of \"Reverse-Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$inter_block_relations}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 5) {...}"

			# END of "MODIFIED in creation of 'csplit_gblock'. (BEG)" #


		### [ (MAIN I): Perform the Initial "csplit", using the 1st available site. ] ... for the predecessor, 'csplit_gblock'.
		### (MAIN I): Perform the Initial "reverse-purge", using the 1st available site.

	my $avail_site_1st = shift @avail_sites;
	my $ct_curr_avail_sites = @avail_sites;
	for (my $i=0; $i <$ct_curr_avail_sites; $i++) {
	    if ($avail_site_1st < $avail_sites[$i]) { $avail_sites[$i]++; }
	}

			# MODIFIED in creation of 'reverse_purge_gblock_pair'. (MAIN1) #

    (print STDOUT "\n ... First, Perform the 'initial_revpurge_gblock_pair' (w/ {rows affected by newly created subject} = ($cnct_rows_affd_by_new_sbj), \$avail_site_1st = $avail_site_1st) ...\n\n") if ($TEST_LEVEL_rvprg_gbp>1);

	my ($flag_init_rvprg, $gb_sbj_NEW, $gb_cmpl_NEW, $new_indices_old_gbs, $old_indices_new_gbs,
	   $record_init_session,
	$new1_curr_set_clms, $new1_curr_bds_blks, $new1_curr_bl_coords, $new1_bds_gb_coords, $new1_gb_coords_wd_frm,
	$new1_curr_gblks_w_spec_lb, $new1_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new1_gblock_sizes,
	$new_list_gblocks
			)
	    = initial_revpurge_gblock_pair (
		@{$rows_affd_by_new_sbj}, $avail_site_1st, $char_abs,
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, 
		@{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, $maxct_1way_shifts,

		@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
		%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
			);
	unless ($flag_init_rvprg) {
	    print STDERR "[ERROR 'reverse_purge_gblock_pair'] The subroutine 'initial_revpurge_gblock_pair' (w/ {rows affected by newly created subject} = ($cnct_rows_affd_by_new_sbj), \$avail_site_1st = $avail_site_1st) FAILED on the input gap-block configuration: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'initial_revpurge_gblock_pair' (w/ {rows affected by newly created subject} = ($cnct_rows_affd_by_new_sbj), \$avail_site_1st  = $avail_site_1st) Ended on the input gap-block configuration (w/ \$flag_init_rvprg = $flag_init_rvprg, \$gb_sbj_NEW = $gb_sbj_NEW, \$gb_cmpl_NEW = $gb_cmpl_NEW) ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1) ;

=For the predecessor, 'csplit_gblock'. (MAIN1)

    (print STDOUT "\n ... First, Perform the 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st = $avail_site_1st) ...\n\n") if ($TEST_LEVEL_rvprg_gbp>1);

	my ($flag_init_cs, $gb_sbj_NEW, $gb_cmpl_NEW, $new_indices_old_gbs, $old_indices_new_gbs,
	   $record_init_session,
	$new1_curr_set_clms, $new1_curr_bds_blks, $new1_curr_bl_coords, $new1_bds_gb_coords, $new1_gb_coords_wd_frm,
	$new1_curr_gblks_w_spec_lb, $new1_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new1_gblock_sizes,
	$new_list_gblocks
			)
	    = initial_csplit_gblock (
		$gb_sbj, $avail_site_1st, $char_abs,
		@{$curr_set_clms}, @{$curr_bds_blks}, @{$curr_bl_coords}, @{$bds_gb_coords}, @{$gb_coords_wd_frm}, 
		@{$curr_gblks_w_spec_lb}, @{$curr_gblks_w_spec_rb}, @{$rows_affd_by_blks}, $cnct_null_clm,
		@{$inter_block_relations}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx}, @{$gblock_sizes}, $maxct_1way_shifts,

		@{$list_gblocks0}, $indx_br, $indx_u_or_d, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
		%{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}
			);
	unless ($flag_init_cs) {
	    print STDERR "[ERROR 'reverse_purge_gblock_pair'] The subroutine 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st = $avail_site_1st) FAILED on the input gap-block configuration: $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'initial_csplit_gblock' (w/ \$gb_sbj = $gb_sbj, \$avail_site_1st  = $avail_site_1st) Ended on the input gap-block configuration (w/ \$flag_init_cs = $flag_init_cs, \$gb_sbj_NEW = $gb_sbj_NEW, \$gb_cmpl_NEW = $gb_cmpl_NEW) ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1) ;

=cut # END of "For the predecessor, 'csplit_gblock'. (MAIN1)"

			# END of "MODIFIED in creation of 'reverse_purge_gblock_pair'. (MAIN1)" #


	my $ct_in2_clms = @{$new1_curr_set_clms};
	my $ct_in2_gbs = @{$new1_curr_bds_blks};


   if ($TEST_LEVEL_rvprg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations output by the subroutine, 'initial_revpurge_gblock_pair' ]\n\n"; # Modified in creation of 'reverse_purge_gblock_pair'.
	##print STDOUT "\n[ Gap-Block Configurations output by the subroutine, 'initial_csplit_gblock' ]\n\n";

	unless ( print_out_gblock_configs (@{$new1_curr_set_clms}, @{$new1_curr_bds_blks}, @{$new1_curr_bl_coords}, @{$new1_curr_gblks_w_spec_lb}, @{$new1_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >5) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks by the subroutine, 'initial_revpurge_gblock_pair' ]\n\n"; # Modified in creation of 'reverse_purge_gblock_pair'.
	##print STDOUT "\n[ Remaining Attributes of Gap-Blocks by the subroutine, 'initial_csplit_gblock' ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @{$new1_bds_gb_coords}, @{$new1_gb_coords_wd_frm}, @{$new1_gblock_sizes}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 5) {...}"


		### [ (MAIN II): Grow the subject and the newly created v-complementary gap-blocks, ] ... for the predecessor, 'csplit_gblock'.
		### (MAIN II): Grow the newly created subject and v-complementary gap-blocks,
		###	by "csplit"ting the available sites one by one.

	my @null_clm = ();
	for (1 .. $ct_rows) { push @null_clm, $char_abs; }

		# (Preliminary 0): Determine which will be removed and which will be kept. #
		#  (ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1)) #
		# (=> MODIFIED in DEBUGGING 'csplit_gblock' (on 2019/08/17). (Pre 0)) #

	my ($gb_left, $gb_right) = ($new1_curr_bds_blks->[$gb_sbj_NEW][0] < $new1_curr_bds_blks->[$gb_cmpl_NEW][0]) ? ($gb_sbj_NEW, $gb_cmpl_NEW) : ($gb_cmpl_NEW, $gb_sbj_NEW); # MODIFIED in creation of 'csplit_gblock'. #

	my ($size_left, $size_right) = @{$new1_gblock_sizes}[$gb_left, $gb_right];

	my ($lbd_left_bf, $rbd_left_bf) = @{$new1_curr_bds_blks->[$gb_left]};
	my ($lbd_right_bf, $rbd_right_bf) = @{$new1_curr_bds_blks->[$gb_right]};

		# END of "(=> MODIFIED in DEBUGGING 'csplit_gblock' (on 2019/08/17). (Pre 0))" #


   if ($TEST_LEVEL_rvprg_gbp >1) {
	print STDOUT "\n[ Secondary Inputs ]\n\n";
	print STDOUT '($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT '($size_left, $size_right) = ', "($size_left, $size_right) .\n\n";
	print STDOUT '($lbd_left_bf, $rbd_left_bf) = ', "($lbd_left_bf, $rbd_left_bf) .\n";
	print STDOUT '($lbd_right_bf, $rbd_right_bf) = ', "($lbd_right_bf, $rbd_right_bf) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >1) {...}"

		#  End of "(ADDED in creation of 'cmerge_gblock_pair' on 2019/07/19. (1))" #


			# (MODIFIED in creation of 'csplit_gblock'. (Pre 1)) #

		# (Preliminary 1): NOMINALLY Copy some arrays to represent the local MSA and the gap-blocks #
		#		   (, in order to preserve the input ones). 			 #

	my @cp_curr_set_clms = @{$new1_curr_set_clms};

	my @cp_curr_gblks_w_spec_lb = @{$new1_curr_gblks_w_spec_lb};
	my @cp_curr_gblks_w_spec_rb = @{$new1_curr_gblks_w_spec_rb};

	my @cp_curr_bds_blks = @{$new1_curr_bds_blks};
	my @cp_bds_gb_coords = @{$new1_bds_gb_coords};
	my @cp_gb_coords_wd_frm = @{$new1_gb_coords_wd_frm};
	my @cp_rows_affd_by_blks = @{$rows_affd_by_new_blks};

	my @cp_curr_bl_coords = @{$new1_curr_bl_coords};

	# $new1_curr_set_clms, $new1_curr_bds_blks, $new1_curr_bl_coords, $new1_bds_gb_coords, $new1_gb_coords_wd_frm,
	# $new1_curr_gblks_w_spec_lb, $new1_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	# $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new1_gblock_sizes


			# END of "(MODIFIED in creation of 'csplit_gblock'. (Pre 1))" #


		### Preliminary (Out): ###

	my ($gb_incr_left, $gb_incr_right) = ($ct_in2_gbs, $ct_in2_gbs+1);
	##my ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

	my $intm1_ct_gbs = $ct_in2_gbs +2;
	my $intm2_ct_gbs = $ct_in2_gbs +1;
	##my $intm1_ct_gbs = $ct_in_gbs +2;
	##my $intm2_ct_gbs = $ct_in_gbs +1;

		# Create @intm1_gblock_sizes & @intm1_inter_block_relations. #

	my @cp_curr_gblock_sizes = @{$new1_gblock_sizes};
	##my @cp_curr_gblock_sizes = @{$gblock_sizes};
	my @intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##my @intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);
		# NOTE: The above 3 arrays need be UPDATED in every session. #

	my @intm1_inter_block_relations = my @intm2_inter_block_relations = ();
	for (my $gb=0; $gb <$ct_in2_gbs; $gb++) {
	##for (my $gb=0; $gb <$ct_in_gbs; $gb++) {
	    my @intm1_rels_w_gb = @{$new_inter_block_relations->[$gb]};
	    ##my @intm1_rels_w_gb = @{$inter_block_relations->[$gb]};
	    ####my @intm2_rels_w_gb = @{$inter_block_relations->[$gb]};

	    if ($gb == $gb_left) {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1] = ('=', 'Cp');
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('=', 'Cp');
		####$intm2_rels_w_gb[$ct_in_gbs] = '=';
	    } elsif ($gb == $gb_right) {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1] = ('Cp', '=');
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1] = ('Cp', '=');
		####$intm2_rels_w_gb[$ct_in_gbs] = 'Cp';
	    } else {
		@intm1_rels_w_gb[$ct_in2_gbs, $ct_in2_gbs+1]
		##@intm1_rels_w_gb[$ct_in_gbs, $ct_in_gbs+1]
		 = @intm1_rels_w_gb[$gb_left, $gb_right];
		####$intm2_rels_w_gb[$ct_in_gbs] = $intm2_rels_w_gb[$gb_left];
	    }
	    $intm1_inter_block_relations[$gb] = \@intm1_rels_w_gb;
	    ##$intm2_inter_block_relations[$gb] = \@intm2_rels_w_gb;
	}
	my @intm1_rels_w_incr_left = @{$intm1_inter_block_relations[$gb_left]};
	@intm1_rels_w_incr_left[$gb_left, $ct_in2_gbs] 
		= @{$intm1_inter_block_relations[$gb_left]}[$ct_in2_gbs, $gb_left];
	##@intm1_rels_w_incr_left[$gb_left, $ct_in_gbs] 
	##	= @{$intm1_inter_block_relations[$gb_left]}[$ct_in_gbs, $gb_left];

	my @intm1_rels_w_incr_right = @{$intm1_inter_block_relations[$gb_right]};
	@intm1_rels_w_incr_right[$gb_right, $ct_in2_gbs+1]
		= @{$intm1_inter_block_relations[$gb_right]}[$ct_in2_gbs+1, $gb_right];
	##@intm1_rels_w_incr_right[$gb_right, $ct_in_gbs+1]
	##	= @{$intm1_inter_block_relations[$gb_right]}[$ct_in_gbs+1, $gb_right];

	$intm1_inter_block_relations[$ct_in2_gbs] = \@intm1_rels_w_incr_left;
	##$intm1_inter_block_relations[$ct_in_gbs] = \@intm1_rels_w_incr_left;

	$intm1_inter_block_relations[$ct_in2_gbs+1] = \@intm1_rels_w_incr_right;
	##$intm1_inter_block_relations[$ct_in_gbs+1] = \@intm1_rels_w_incr_right;


		# Create $intm1_cts_vincluding, etc. #

	my ($intm1_cts_vincluding, $intm1_cts_vincluded, $intm1_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@intm1_inter_block_relations);
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'reverse_purge_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@intm1_inter_block_relations', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@intm1_inter_block_relations', " ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1) ;


		# Create @intm1_rows_affd_by_blks.

	my @intm1_rows_affd_by_blks = (@cp_rows_affd_by_blks, @cp_rows_affd_by_blks[$gb_left, $gb_right]);
	##my @intm2_rows_affd_by_blks = (@cp_rows_affd_by_blks, $cp_rows_affd_by_blks[$gb_left]);


		### MAIN 1 & 2: Iteratively Perform the single-column-growing "Merge + Split"s. ###

	my ($rows_affd_by_left, $rows_affd_by_right) = @cp_rows_affd_by_blks[$gb_left, $gb_right];
	my ($rels_w_left, $rels_w_right) = @{$new_inter_block_relations}[$gb_left, $gb_right];
	##my ($rels_w_left, $rels_w_right) = @{$inter_block_relations}[$gb_left, $gb_right];

	my $curr_size_rvprg = 1;
	##my $curr_size_cmpl = 1;
	####my $curr_incr_gb_size = $incr_gb_size +1;

	my $curr_ct_clms = $ct_in2_clms;
	##my $curr_ct_clms = $ct_in_clms;

	my @records_sessions = ($record_init_session); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2]; # MODIFIED in creation of 'csplit_gblock'. #
	##my @records_sessions = (); # element = [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

     foreach my $session (2 .. $size_rvprg) { # Outer main foreach-loop (over the sessions). #
     ##foreach my $session (2 .. $size_cmpl) { # Outer main foreach-loop (over the sessions). #
     ####foreach my $session (1 .. $incr_gb_size) { # Outer main foreach-loop (over the sessions). #

	##$curr_incr_gb_size--;
	my ($le_wd, $re_wd) = (0, $curr_ct_clms-1);

    (print STDOUT "\n{{ Session $session:  (\$curr_size_rvprg = $curr_size_rvprg, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_rvprg_gbp > 1); # MODIFIED in creation of 'reverse_purge_gblock_pair'. #
    ##(print STDOUT "\n{{ Session $session:  (\$curr_size_cmpl = $curr_size_cmpl, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_rvprg_gbp > 1); # MODIFIED in creation of 'csplit_gblock'. #
    ####(print STDOUT "\n{{ Session $session:  (\$curr_incr_gb_size = $curr_incr_gb_size, \$curr_ct_clms = $curr_ct_clms,  (\$le_wd, \$re_wd) = ($le_wd, $re_wd)) }}\n\n") if ($TEST_LEVEL_rvprg_gbp > 1);


			# MODIFIED in creation of 'csplit_gblock' (on 2019/08/16). (Pre (In)) #

		# Preliminary (In):  Extract the subject site, and update @avail_sites. 

	my $sbjct_site = shift @avail_sites;
	my $ct_curr_avail_sites = @avail_sites;
	for (my $i=0; $i <$ct_curr_avail_sites; $i++) {
	    if ($sbjct_site < $avail_sites[$i]) { $avail_sites[$i]++; }
	}

   if ($TEST_LEVEL_rvprg_gbp >3) {

	print STDOUT "[ Current sets of available sites: ]\n\n";

	print STDOUT "\$sbjct_site = $sbjct_site .\n\n";
	print STDOUT 'Current @avail_sites = (', join (', ', @avail_sites), ") .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >3) {...}"

			# END of "MODIFIED in creation of 'csplit_gblock' (on 2019/08/16). (Pre (In))" #


		## MAIN I: Create a new pair of single-column gap-blocks, by "split"ting the $sbjct_site th site. ##

	my @clm_incr_left = @null_clm;
	my @clm_incr_right = @{$cp_curr_set_clms[$sbjct_site]};
	foreach my $row (@{$rows_affd_by_right}) { # 1st middle foreach.
	    my $tmp = $clm_incr_right[$row];
	    $clm_incr_right[$row] = $clm_incr_left[$row];
	    $clm_incr_left[$row] = $tmp;
	} # End of the 1st middle foreach.

	my @gblks_w_lb_into_left = my @gblks_w_lb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_lb[$sbjct_site]}) { # 2nd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_lb_into_left, $gb;
	    } else {
		push @gblks_w_lb_into_right, $gb;
	    }
	} # End of the 2nd middle foreach.
	push @gblks_w_lb_into_left, $ct_in2_gbs;
	push @gblks_w_lb_into_right, $ct_in2_gbs +1;
	##push @gblks_w_lb_into_left, $ct_in_gbs;
	##push @gblks_w_lb_into_right, $ct_in_gbs +1;

	my @gblks_w_rb_into_left = my @gblks_w_rb_into_right = ();
	foreach my $gb (@{$cp_curr_gblks_w_spec_rb[$sbjct_site]}) { # 3rd middle foreach.
	    my $rel = $rels_w_right->[$gb];

	    if (($rel eq '>') or ($rel eq '>(ch)')) { # A BUG Fixed on 2019/08/05. #
	    ##if ($rel eq '>') {
		push @gblks_w_rb_into_left, $gb;
	    } else {
		push @gblks_w_rb_into_right, $gb;
	    }
	} # End of the 3rd middle foreach.
	push @gblks_w_rb_into_left, $ct_in2_gbs;
	push @gblks_w_rb_into_right, $ct_in2_gbs +1;
	##push @gblks_w_rb_into_left, $ct_in_gbs;
	##push @gblks_w_rb_into_right, $ct_in_gbs +1;

	my @intm1_curr_set_clms = ();
	my @intm1_curr_gblks_w_spec_lb = my @intm1_curr_gblks_w_spec_rb = ();

	for (my $c=0; $c < $sbjct_site; $c++) { # 1st Middle MAIN for-loop (over sites on the left of $sbjct_site). #
	    $intm1_curr_set_clms[$c] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 1st Middle MAIN for-loop. #

	@intm1_curr_set_clms[$sbjct_site, $sbjct_site+1] = (\@clm_incr_left, \@clm_incr_right);
	@intm1_curr_gblks_w_spec_lb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_lb_into_left, \@gblks_w_lb_into_right);
	@intm1_curr_gblks_w_spec_rb[$sbjct_site, $sbjct_site+1] = (\@gblks_w_rb_into_left, \@gblks_w_rb_into_right);

	for (my $c = $sbjct_site+1; $c <$curr_ct_clms; $c++) { # 2nd Middle MAIN for-loop (over sites on the right of $sbjct_site). #
	    $intm1_curr_set_clms[$c+1] = $cp_curr_set_clms[$c];
	    $intm1_curr_gblks_w_spec_lb[$c+1] = $cp_curr_gblks_w_spec_lb[$c];
	    $intm1_curr_gblks_w_spec_rb[$c+1] = $cp_curr_gblks_w_spec_rb[$c];
	} # End of the 2nd Middle MAIN for-loop. #


	my $intm1_ct_clms = @intm1_curr_set_clms;


		# Create @intm1_curr_bds_blks. #

	my @intm1_curr_bds_blks = ();
	for (my $gb=0; $gb < $intm1_ct_gbs; $gb++) { $intm1_curr_bds_blks[$gb] = []; }

	for (my $c=0; $c <$intm1_ct_clms; $c++) { # 2nd trailing for-loop for Main 1-(i). #
            foreach my $gb (@{$intm1_curr_gblks_w_spec_lb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[0] = $c;
            }
            foreach my $gb (@{$intm1_curr_gblks_w_spec_rb[$c]}) {
                $intm1_curr_bds_blks[$gb]->[1] = $c;
            }
	} # End of the 2nd trailing for-loop for Main 1-(i). #


		## Preliminary (In 2): Create the coordinate frame for the 1st intermediate gap-block configuration. #

	my ($intm1_le_wd, $intm1_re_wd) = (0, $intm1_ct_clms-1);

	my ($intm1_orgs_gb_coords, $intm1_bds_gb_coords, $intm1_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@intm1_curr_bds_blks, $intm1_le_wd, $intm1_re_wd, @intm1_inter_block_relations, $maxct_1way_shifts);
	unless (defined $intm1_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);

	my @intm1_curr_bl_coords = @{$intm1_orgs_gb_coords};


			# (Added on 2019/08/07. (Main II)) #

   if ($TEST_LEVEL_rvprg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_gblock_configs (@intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >5) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks IMMEDIATELY AFTER the \"Split (complementary types)\" of the \$sbject_site (=$sbjct_site) th column ]\n\n";

	unless ( print_out_remng_attr_gblks (@intm1_inter_block_relations, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm}, @intm1_gblock_sizes, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_rows_affd_by_blks, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 5) {...}"


		## MAIN II-(i): Merge the $gb_right and $gb_incr_right th gap-block, via the 1st Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 1st Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_right (=$gb_right) th and \$gb_incr_right (=$gb_incr_right) th gap-blocks ...\n\n") if ($TEST_LEVEL_rvprg_gbp > 1);

	my ($flag_end_smg1, $bl_sbj_smg1, $incr_coord_sbj_smg1, $gb_absbd_smg1, $gb_kpt_smg1, $new_indices_gbs_smg1, $old_indices_gbs_smg1,
	    $intm2_curr_set_clms, $intm2_curr_bds_blks, $intm2_curr_bl_coords, $intm2_bds_gb_coords, $intm2_gb_coords_wd_frm,
	    $intm2_curr_gblks_w_spec_lb, $intm2_curr_gblks_w_spec_rb, $intm2_rows_affd_by_blks,
	    $intm2_inter_block_relations, $intm2_cts_vincluding, $intm2_cts_vincluded, $intm2_cts_onx, $intm2_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_right, $gb_incr_right, @intm1_curr_set_clms, @intm1_curr_bds_blks, @intm1_curr_bl_coords, @{$intm1_bds_gb_coords}, @{$intm1_gb_coords_wd_frm},
		@intm1_curr_gblks_w_spec_lb, @intm1_curr_gblks_w_spec_rb, @intm1_rows_affd_by_blks, $cnct_null_clm,
		@intm1_inter_block_relations, @{$intm1_cts_vincluding}, @{$intm1_cts_vincluded}, @{$intm1_cts_onx}, @intm1_gblock_sizes, $maxct_1way_shifts
		);
	unless ($flag_end_smg1) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The 1st round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 1st round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"split\" of the \$sbjct_site (=$sbjct_site) th column (in session $session) ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);


   if ($TEST_LEVEL_rvprg_gbp >5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >5) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 1st Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm2_inter_block_relations}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm}, @{$intm2_gblock_sizes}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 5) {...}"



		## MAIN II-(ii): Merge the $gb_left and $gb_incr_left th gap-block, via the 2nd Round of 'smerge_gblock_pair'. ##

	## ($gb_incr_left, $gb_incr_right) = ($ct_in_gbs, $ct_in_gbs+1);

   (print STDOUT "\n ... Perform the 2nd Round of 'smerge_gblock_pair' for session $session, to merge the \$gb_left (=$gb_left) th and \$gb_incr_left (=$gb_incr_left) th gap-blocks ...\n\n") if ($TEST_LEVEL_rvprg_gbp > 1);

	my ($flag_end_smg2, $bl_sbj_smg2, $incr_coord_sbj_smg2, $gb_absbd_smg2, $gb_kpt_smg2, $new_indices_gbs_smg2, $old_indices_gbs_smg2,
	    $intm3_curr_set_clms, $intm3_curr_bds_blks, $intm3_curr_bl_coords, $intm3_bds_gb_coords, $intm3_gb_coords_wd_frm,
	    $intm3_curr_gblks_w_spec_lb, $intm3_curr_gblks_w_spec_rb, $intm3_rows_affd_by_blks,
	    $intm3_inter_block_relations, $intm3_cts_vincluding, $intm3_cts_vincluded, $intm3_cts_onx, $intm3_gblock_sizes
		)
	   = smerge_gblock_pair (
		$gb_left, $gb_incr_left, @{$intm2_curr_set_clms}, @{$intm2_curr_bds_blks}, @{$intm2_curr_bl_coords}, @{$intm2_bds_gb_coords}, @{$intm2_gb_coords_wd_frm},
		@{$intm2_curr_gblks_w_spec_lb}, @{$intm2_curr_gblks_w_spec_rb}, @{$intm2_rows_affd_by_blks}, $cnct_null_clm,
		@{$intm2_inter_block_relations}, @{$intm2_cts_vincluding}, @{$intm2_cts_vincluded}, @{$intm2_cts_onx}, @{$intm2_gblock_sizes}, $maxct_1way_shifts
		);
	unless ($flag_end_smg2) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The 2nd round of the subroutine, 'smerge_gblock_pair' FAILED on the cofiguration of gap-blocks after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session: $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The 2nd round of the subroutine, 'smerge_gblock_pair', Ended on the configuration of gap-blocks immediately after the \"merge\" of \$gb_right (=$gb_right) and \$gb_incr_right (=$gb_incr_right), in session $session ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);


   if ($TEST_LEVEL_rvprg_gbp >= 5) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_gblock_configs (@{$intm3_curr_set_clms}, @{$intm3_curr_bds_blks}, @{$intm3_curr_bl_coords}, @{$intm3_curr_gblks_w_spec_lb}, @{$intm3_curr_gblks_w_spec_rb}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >= 5) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 5) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the 2nd Round of 'smerge_gblock_pair' for session $session, (to \"merge\" \$gb_left (=$gb_left) and \$gb_incr_left (=$gb_incr_left) ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$intm3_inter_block_relations}, @{$intm3_bds_gb_coords}, @{$intm3_gb_coords_wd_frm}, @{$intm3_gblock_sizes}, @{$intm3_cts_vincluding}, @{$intm3_cts_vincluded}, @{$intm3_cts_onx}, @{$intm3_rows_affd_by_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'reverse_purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 5) {...}"


		## MAIN IIpost: Update Important Data-Sets
		##	that characterize the Updated gap-block configuration. ##

			# Update the set of columns. #
	@cp_curr_set_clms = @{$intm3_curr_set_clms};
	$curr_ct_clms = @cp_curr_set_clms;

			# Update the set of gap-block boundaries & its "inverse-functions". #
	@cp_curr_bds_blks = @{$intm3_curr_bds_blks};
	@cp_curr_gblks_w_spec_lb = @{$intm3_curr_gblks_w_spec_lb};
	@cp_curr_gblks_w_spec_rb = @{$intm3_curr_gblks_w_spec_rb};

			# Update the coordinate frame. #
	@cp_curr_bl_coords = @{$intm3_curr_bl_coords};
	@cp_bds_gb_coords = @{$intm3_bds_gb_coords};
	@cp_gb_coords_wd_frm = @{$intm3_gb_coords_wd_frm};

			# Update the sets of gap-block sizes. #
	@cp_curr_gblock_sizes = @{$intm3_gblock_sizes};
	@intm1_gblock_sizes = (@cp_curr_gblock_sizes, 1, 1);
	##@intm2_gblock_sizes = (@cp_curr_gblock_sizes, 1);


		## Record the Core information on this Session. ##

	push @records_sessions, [$session, $sbjct_site, $incr_coord_sbj_smg1, $incr_coord_sbj_smg2];

			# End of "(Added on 2019/08/07. (Main II))" #

    } # End of the outer MAIN foreach-loop (over the sessions). #


  (print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the Sessions,  2 .. $size_rvprg !!---\n\n\n") if ($TEST_LEVEL_rvprg_gbp >1);
  ##(print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the Sessions,  2 .. $size_cmpl !!---\n\n\n") if ($TEST_LEVEL_rvprg_gbp >1);
  ####(print STDOUT "\n ---!! END of the MAIN Foreach-Loop over the $incr_gb_size Sessions !!---\n\n\n") if ($TEST_LEVEL_cspl_gb >1);


		### MAIN III: Create the Final Output Data-Sets. ###
		### (Added on 2019/08/07. (MAIN III)) ###

	my @new_curr_set_clms = @cp_curr_set_clms;
	my @new_curr_bds_blks = @cp_curr_bds_blks;

	my @new_curr_bl_coords = @cp_curr_bl_coords;
	my @new_bds_gb_coords = @cp_bds_gb_coords;
	my @new_gb_coords_wd_frm = @cp_gb_coords_wd_frm;

	my @new_curr_gblks_w_spec_lb = @cp_curr_gblks_w_spec_lb;
	my @new_curr_gblks_w_spec_rb = @cp_curr_gblks_w_spec_rb;
	my @new_gblock_sizes = @cp_curr_gblock_sizes;

		### END of "(Added on 2019/08/07. (MAIN III))" ###


		### MAIN III': Re-Rank the resulting Gap-Blocks, if necessary. ###
		### (ADDED for DEBUGGING on 2019/08/18. (MAIN III')) ###

	##my @new_indices_old_gbs = @{$new_indices_old_gbs};
	##my @old_indices_new_gbs = @{$old_indices_new_gbs};

		# Fixed a Minor BUG. #
	my $ct_out_gbs = @new_curr_bds_blks;
	my $ct_out_clms = @new_curr_set_clms;
	##my $ct_out_gbs = $ct_in2_gbs;
	##my $ct_out_clms = $ct_in2_clms;

	my @tmp_new_list_gblocks = ();
	for (my $gb=0; $gb <$ct_out_gbs; $gb++) {
	    my @info_gb = @{$new_list_gblocks->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};
	    $tmp_new_list_gblocks[$gb] = \@info_gb;
	}

	my @new_indices_new2_gbs = sort { cmp_gblocks3 (@{$tmp_new_list_gblocks[$a]}, @{$tmp_new_list_gblocks[$b]}, 2, 3, 6) } (0 .. $ct_out_gbs -1);

	my $if_reordered = 0;
	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {
	    if ($new_indices_new2_gbs[$n2gb] != $n2gb) {
		$if_reordered = 1;
		last;
	    }
	}

    if ($if_reordered) {

  (print STDOUT "[INRORMATION 'reverse_purge_gblock_pair'] AFTER the \"Reverse-Purge\", The the gap-blocks are RE-ORDERED (w/ ", '@new_indices_new2_gbs = (', join (',', @new_indices_new2_gbs), ") !!\n\n") if ($TEST_LEVEL_rvprg_gbp> 0);
  (print STDOUT " ... Accordingly, the output data-sets will be UPDATED ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);

	my @new2_indices_new_gbs = ();
	my @new2_indices_old_gbs = my @old_indices_new2_gbs = ();

	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {

	    my $ngb = $new_indices_new2_gbs[$n2gb];
	    $new2_indices_new_gbs[$ngb] = $n2gb;

	    my $ogb = $old_indices_new_gbs->[$ngb];
	    $new2_indices_old_gbs[$ogb] = $n2gb;
	    $old_indices_new2_gbs[$n2gb] = $ogb;
	}
	$new_indices_old_gbs = \@new2_indices_old_gbs;
	$old_indices_new_gbs = \@old_indices_new2_gbs;

    if ($TEST_LEVEL_rvprg_gbp >2) {

	print STDOUT 'Updated @{$new_indices_old_gbs} = (', join (', ', @{$new_indices_old_gbs}), ") ;\n";
	print STDOUT 'Updated @{$old_indices_new_gbs} = (', join (', ', @{$old_indices_new_gbs}), ") .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_rvprg_gbp >2) {...}"


			# Update the KEY gap-block indices. #

	$gb_sbj_NEW = $new2_indices_new_gbs[$gb_sbj_NEW];
	$gb_cmpl_NEW = $new2_indices_new_gbs[$gb_cmpl_NEW];
	$gb_left = $new2_indices_new_gbs[$gb_left];
	$gb_right = $new2_indices_new_gbs[$gb_right];

    if ($TEST_LEVEL_rvprg_gbp >2) {

	print STDOUT 'Updated ($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW) ;\n";
	print STDOUT 'Updated ($gb_left, $gb_right) = ', "($gb_left, $gb_right) .\n";
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_rvprg_gbp >2) {...}"

			# Update the data-sets. #

	@new_curr_bds_blks = @cp_curr_bds_blks[@new_indices_new2_gbs];
	@new_curr_gblks_w_spec_lb = @new_curr_gblks_w_spec_rb = ();
	for (my $c=0; $c <$ct_out_clms; $c++) {
	    $new_curr_gblks_w_spec_lb[$c] = [];
	    $new_curr_gblks_w_spec_rb[$c] = [];
	}
	for (my $n2gb=0; $n2gb <$ct_out_gbs; $n2gb++) {
	    my ($lb, $rb) = @{$new_curr_bds_blks[$n2gb]};
	    push @{$new_curr_gblks_w_spec_lb[$lb]}, $n2gb;
	    push @{$new_curr_gblks_w_spec_rb[$rb]}, $n2gb;
	}


	@new_gblock_sizes = @cp_curr_gblock_sizes[@new_indices_new2_gbs];

	my @rows_affd_by_new2_blks = @{$rows_affd_by_new_blks}[@new_indices_new2_gbs];
	$rows_affd_by_new_blks = \@rows_affd_by_new2_blks;


			# Re-Create  @{$new_inter_block_relations}. #

	@tmp_new_list_gblocks = @tmp_new_list_gblocks[@new_indices_new2_gbs];
	$new_list_gblocks = \@tmp_new_list_gblocks;

	($new_inter_block_relations)
	  = mk_inter_block_relations (@tmp_new_list_gblocks, 0, 1, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $new_inter_block_relations) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The subroutine, 'mk_inter_block_relations', FAILED on ", '@tmp_new_list_gblocks', " : $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'mk_inter_block_relations' Ended on ", '@tmp_new_list_gblocks', " ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1) ;


			# Re-Create @{$new_cts_vincluding}, etc. #

	($new_cts_vincluding, $new_cts_vincluded, $new_cts_onx)
	     = cts_vincluding_vincluded_onx_gblks (@{$new_inter_block_relations});
	unless (defined $cts_onx) {
	    print STDERR "[ERROR 'reverse_purge_gblock_pair'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on ", '@{$new_inter_block_relations}', ": $! !\n\n";
	    return (0, undef, undef, undef, undef);
	}
    (print STDOUT " ... The subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on ", '@{$new_inter_block_relations}', " ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1) ;


			# Re-Create @new_curr_bl_coords  & @new_bds_gb_coords  & @new_gb_coords_wd_frm . #

	my ($new_le_wd, $new_re_wd) = (0, $ct_out_clms-1);

	my ($new_orgs_gb_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@new_curr_bds_blks, $new_le_wd, $new_re_wd, @{$new_inter_block_relations}, $maxct_1way_shifts);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The subroutine, 'mk_orgs_bds_gb_coords' FAILED on the cofiguration of gap-blocks after the \"Reverse-Purge\": $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The subroutine, 'mk_orgs_bds_gb_coords', Ended on the configuration of gap-blocks after the \"Reverse-Purge\" ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);

	@new_curr_bl_coords    = @{$new_orgs_gb_coords} ;
	@new_bds_gb_coords     = @{$new_bds_gb_coords} ;
	@new_gb_coords_wd_frm  = @{$new_gb_coords_wd_frm} ;

    } # END of "if ($if_reordered) {...}"

		### END of "(ADDED for DEBUGGING on 2019/08/18. (MAIN III'))" ###


		### MAIN IIIpost: UNDO the overlaps between the gap-blocks
		###		vertically equivalent to and/or complementary to $gb1 and $gb2. ###
		### (ADDED on 2019/08/08. (MAIN IIIpost)) ###

	my ($fl_undo, $gb2incr_bl_coord_undo) 
	  = undo_hovlp_vcmpl_gbsets ($gb_left, $gb_right,
		@new_curr_set_clms, @new_curr_bds_blks, @new_bds_gb_coords, @new_curr_bl_coords,
		@new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, @{$rows_affd_by_new_blks}, $cnct_null_clm,
		@{$new_inter_block_relations}, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @new_gblock_sizes
		);
	unless ($fl_undo) {
	    print STDOUT "[ERROR 'reverse_purge_gblock_pair'] The subroutine, 'undo_hovlp_vcmpl_gbsets' FAILED on the cofiguration of gap-blocks after the main foreach-loop for \"Reverse-Purge\" creating \$gb_left (=$gb_left) and \$gb_right (=$gb_right): $! !\n\n";
	    return (0, undef, undef, undef);
	}
   (print STDOUT "\n ... The the subroutine, 'undo_hovlp_vcmpl_gbsets', Ended on the configuration of gap-blocks immediately after the main foreach-loop for the \"Reverse-Purge\" creating \$gb_left (=$gb_left) and \$gb_right (=$gb_right) ...\n\n") if ($TEST_LEVEL_rvprg_gbp >1);


   if ($TEST_LEVEL_rvprg_gbp > 3) { # ADDED on 2019/08/09. #

	print STDOUT "\n[ Outputs of 'undo_hovlp_vcmpl_gbsets': ]\n\n";

	print STDOUT "\$fl_undo = $fl_undo .\n\n";

	print STDOUT '%{$gb2incr_bl_coord_undo} = ( # $indx_gb => $incr_bl_coord_undo,', "\n";

	foreach my $gb (sort {$a <=> $b} keys %{$gb2incr_bl_coord_undo}) {
	    print STDOUT "\t", $gb, ' => ', $gb2incr_bl_coord_undo->{$gb}, " ,\n";
	}
	print STDOUT "\t\t) .\n";

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 3) {...}"


		### END of "(ADDED on 2019/08/08. (MAIN IIIpost))" ###

   if ($TEST_LEVEL_rvprg_gbp >2) {

	print STDOUT "\n[ Gap-Block Configurations AFTER the Whole Process of \"Reverse-Purge\" ]\n\n";

	unless ( print_out_gblock_configs (@new_curr_set_clms, @new_curr_bds_blks, @new_curr_bl_coords, @new_curr_gblks_w_spec_lb, @new_curr_gblks_w_spec_rb, *STDOUT) ) {
	    print STDOUT "[WARNING 'reverse_purge_gblock_pair'] The subroutine 'print_out_gblock_configs' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'purge_gblock_pair' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp >2) {...}"

   if ($TEST_LEVEL_rvprg_gbp > 3) {

	print STDOUT "\n[ Remaining Attributes of Gap-Blocks AFTER the Whole Process of \"Reverse-Purge\" ]\n\n";

	unless ( print_out_remng_attr_gblks (@{$new_inter_block_relations}, @new_bds_gb_coords, @new_gb_coords_wd_frm, @new_gblock_sizes, @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, @{$rows_affd_by_new_blks}, *STDOUT) ) {
	    print STDOUT "[WARNING 'initial_csplit_gblock'] The subroutine 'print_out_remng_attr_gblks' FAILED: $! ! ... just ignore it, because it is not directly related to the results of 'initial_csplit_gblock' ...\n\n";
	}
	print STDOUT "\n";

   } # End of "if ($TEST_LEVEL_rvprg_gbp > 3) {...}"

			# (Added in creation of 'csplit_gblock'.) (FIN) #

		# Create @new_list_gblocks. #

	my @new_list_gblocks = ();
	for (my $gb=0; $gb <$ct_in2_gbs; $gb++) {
	    my @info_gb = @{$new_list_gblocks->[$gb]};
	    @info_gb[2,3] = @{$new_curr_bds_blks[$gb]};
	    $new_list_gblocks[$gb] = \@info_gb;
	}

	my $med_indices_new_gbs = \@new_indices_new2_gbs; # ADDED in revision on 2019/08/19.

	my $fl_end = $fl_undo;
			# END of "(Added in creation of 'csplit_gblock'.) (FIN)" #

    (print STDOUT "\n< Subroutine 'reverse_purge_gblock_pair' Ended (w/ \$fl_end = $fl_end, ", '($gb_sbj_NEW, $gb_cmpl_NEW) = ', "($gb_sbj_NEW, $gb_cmpl_NEW)", '($gb_left, $gb_right) = ', "($gb_left, $gb_right) . >\n\n") if ($TEST_LEVEL_rvprg_gbp > 0);

	return ($fl_end, $gb_sbj_NEW, $gb_cmpl_NEW, $gb_left, $gb_right, 
	$new_indices_old_gbs, $old_indices_new_gbs, \@records_sessions, $gb2incr_bl_coord_undo,

	\@new_curr_set_clms, \@new_curr_bds_blks, \@new_curr_bl_coords, \@new_bds_gb_coords, \@new_gb_coords_wd_frm,
	\@new_curr_gblks_w_spec_lb, \@new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks, 
	$new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, \@new_gblock_sizes,
	\@new_list_gblocks, $med_indices_new_gbs); # ADDED "$med_indices_new_gbs" on 2019/08/19.

}	# End of the subroutine, 'reverse_purge_gblock_pair'. #



# printout_multidim_storage:
#	This subroutine print outs a specified component of the contents of
#	the storage for some information (e.g., the increment of the log-probability
#	of the alternative alignment compared to the initial alignment)
#	on each point in the coordinate space.
#	[CREATED on 2019/03/04; BUGs fixed on 2019/03/06.]
#
#	[DEBUGGED on 2020/04/15]:
#	Thus far, the subroutine COULD run into an infinite-loop when (#{gap-blocks} < 2).
#	Now, the subroutine should END NORMALLY even if (#{gap-blocks} < 2).
#
#	NOTE:
#	When printing out, each coordinate (for the $bl th block) will be SHIFT-ed back by $org_bl_coords[$bl], SO THAT the coordinate origin will get back to ZERO.
#
# INPUT: (\@org_bl_coords, \@bds_bl_coords, \@multidim_storage, $indx_po, \*${$fh}, $string_undef), where
#
#	$org_bl_coords[$bl] is the position of the "origin" of the coordinate of the $bl th block.
#	(Although the "origin" should actually be ZERO, it is SHIFTED to this value, SO THAT the range of the coordinate will become entirely non-negaitve.);
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds.]
#
#	@storage_for_pnt = @{$mltdim_storage[$k_0]->...->[$k_{B-1}]} (where B = scalar (@bds_bl_coords))
#	is the storage for some information on the point with the coordinate,
#	($k_0, ..., $k_{B-1}) within the hyperrectangle in question.
#	$storage_for_pnt[$indx_po] will be printed out.
#
#	$fh is the reference to the file-handle for the "file" to which the information components will be printed out;
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub printout_multidim_storage (\@\@\@$\*$) {

	my ($org_bl_coords, $bds_bl_coords, $mltdim_storage, $indx_po, $fh, $string_undef) = @_;

	my $ct_blks = @{$bds_bl_coords}; 
	unless (@{$org_bl_coords} == $ct_blks) {
	    my $ct_orgs = @{$org_bl_coords};
	    print STDERR "[ERROR 'printout_multidim_storage'] The number of 'origin's (=$ct_orgs) CLASHES with the number of block-coordinate-boundary-pairs (=$ct_blks): $! !\n\n";
	    return 0;
	}
    (print STDOUT "< Subroutine 'printout_multidim_storage' (#{coordinates} = $ct_blks) >\n\n") if ($TEST_LEVEL_po_md_strg>0);

	if ($ct_blks == 0) { # (ADDED in DEBUGGING on 2020/04/15.) (1) #
	    print STDOUT "[INFORMATION 'printout_multidim_storage'] The number of input blocks is ZERO!! ... I will end this subroutine withOUT doing anything ...\n\n";
	    return 1;
	}

		# Examine the coordinate consistencies. #

	my @errors1 = my @errors2 = ();
	for (my $k=0; $k < $ct_blks; $k++) { # Preliminary for-loop (over blocks).
	    my ($lb, $rb) = @{$bds_bl_coords->[$k]};

	    if ($lb < 0) { push @errors1, $k; }
		
	    if ($rb < $lb) { push @errors2, $k; }
	
	} # End of the preliminary for-loop (over blocks).

	if ((@errors1>0) or (@errors2>0)) {
	    print STDERR "[ERROR 'printout_multidim_storage'] The blocks (", join (', ', @errors1), ") has NEGATIVE left-boundaries, and the blocks (", join (', ', @errors2), ") has their right-bound ON THE LEFT OF their left-bounds: $! !\n\n";
	    return 0;
	}

		# Initialize the current parent. #
	my @curr_parent = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { $curr_parent[$k] = $bds_bl_coords->[$k][0]; }

		# & Initialize the stack (as a nested series of single anonymous arrays specified by the coordinates).

	my $sbj_storage0 = $mltdim_storage;
	my @stack = ($sbj_storage0);
	for (my $k=0; $k < $ct_blks-1; $k++) { # 2nd for-loop (over blocks).
	    my $coord = $curr_parent[$k];
	    my $child_storage = $sbj_storage0->[$coord];
	    push @stack, $child_storage;
	    $sbj_storage0 = $child_storage;
	} # End of the 2nd for-loop (over blocks).
	my $parent_storage0 = pop @stack;

	my ($lb_last, $rb_last) = @{$bds_bl_coords->[$ct_blks-1]};
	my $org_last = $org_bl_coords->[$ct_blks-1];


		# Print out the Header. #

	my @labels_corner = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { push @labels_corner, "X$k"; }
	my $labels_corner = join (':', @labels_corner);
	$labels_corner .= '\X';
	$labels_corner .= $ct_blks-1;
	my @last_coords = ();
	for (my $k=$lb_last; $k <= $rb_last; $k++) { push @last_coords, $k - $org_last; }
	print $fh join ("\t", $labels_corner, @last_coords), "\n\n";

		# Print out the subject components at the starting points. #

		# A BUG FIXED on 2019/03/06. (c)
	my @curr_pa_coords0 = ();
	for (my $i=0; $i < $ct_blks -1; $i++) {
	    $curr_pa_coords0[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	}
	my $cnct_curr_pa_coords0 = join (':', @curr_pa_coords0);
	##my $cnct_curr_parent = join (':', @curr_parent);

	my @components0 = ();
	for (my $k = $lb_last; $k <= $rb_last; $k++) {
	    my $component_sbj = $parent_storage0->[$k][$indx_po];
	    unless (defined $component_sbj) { $component_sbj = $string_undef; }
	    push @components0, $component_sbj;
	}
		# A BUG FIXED on 2019/03/06. (d)
	print $fh join ("\t", $cnct_curr_pa_coords0, @components0), "\n";
	##print $fh join ("\t", $cnct_curr_parent, @components0), "\n";

	if ($ct_blks <2) {
		# This is necessary for the subroutine to end normally even if ($ct_blks <2). #
		# (ADDED in DEBUGGING on 2020/04/15.) (2) #
	    return 1;
	}

		# Main: Print out the subject components from the entire output storage. #

	my $indx_curr_crd = $ct_blks -2;

	while (1) { # Main while-loop (to construct the output).

	    my $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$indx_curr_crd]};

	    if ($curr_sbj_coord == $rbd) {
		if ($indx_curr_crd == 0) { last; }

		$curr_parent[$indx_curr_crd] = $lbd;
		$indx_curr_crd--;
		pop @stack;
		next;
	    }

		# A BUG fixed (on 2019/02/22).
	    $curr_parent[$indx_curr_crd]++;
	    $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    ##$curr_sbj_coord = $curr_parent[$indx_curr_crd]++;

	    my $grand_parent_storage = $stack[$#stack];

	    my $parent_storage = $grand_parent_storage->[$curr_sbj_coord];

	    while ($indx_curr_crd < $ct_blks - 2) { # Inner while-loop.
		push @stack, $parent_storage;
		$indx_curr_crd++;
		$curr_sbj_coord = $curr_parent[$indx_curr_crd];
		$parent_storage = $parent_storage->[$curr_sbj_coord];

	    } # End of the inner while-loop.

		# Print out the components at the current points. #

		# A BUG FIXED on 2019/03/06. (a)
	    my @curr_pa_coords = ();
	    for (my $i = 0; $i < $ct_blks -1; $i++) {
		$curr_pa_coords[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	    }
	    my $cnct_curr_pa_coords = join (':', @curr_pa_coords);
	    ##my $cnct_curr_parent = join (':', @curr_parent);

	    my @components = ();
	    for (my $k = $lb_last; $k <= $rb_last; $k++) {
		my $component_sbj = $parent_storage->[$k][$indx_po];
		unless (defined $component_sbj) { $component_sbj = $string_undef; }
		push @components, $component_sbj;
	    }

		# A BUG FIXED on 2019/03/06. (b)
	    print $fh join ("\t", $cnct_curr_pa_coords, @components), "\n";
	    ##print $fh join ("\t", $cnct_curr_parent, @components), "\n";


	} # End of the main while-loop (to construct the output).


    (print STDOUT "\n< Subroutine 'printout_multidim_storage' Ended />\n\n") if ($TEST_LEVEL_po_md_strg>0);

	return 1;

}	# End of the subroutine 'printout_multidim_storage'. #


# print_out_gblock_configs:
#	This satellite subroutine of 'serial_shifts_sgl_gblock' and 'serial_shifts_point2point'
#	prints out the gap-block configurations.
#	[CREATED on 2019/07/09.]
#
# INPUTS: (\@curr_set_clms, \@curr_bds_blks, \@curr_bl_coords, \@curr_gblks_w_spec_lb, \@curr_gblks_w_spec_rb, \*${$fh_out}), where
#
#	The first 5 arrays are main inputs of 'serial_shifts_sgl_gblock';
#	(For details, see the header of 'serial_shifts_sgl_gblock');
#
#	$fh_out is the reference to an output file-handle;
#
# OUTPUT: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_gblock_configs (\@\@\@\@\@\*) {

	my ($curr_set_clms, $curr_bds_blks, $curr_bl_coords, $curr_gblks_w_spec_lb, $curr_gblks_w_spec_rb, $fh_out) = @_;

	my $ct_gblks = @{$curr_bds_blks};
	my $ct_clms = @{$curr_set_clms};
	my $ct_rows = ($ct_clms>0) ? @{$curr_set_clms->[0]} : 0;

	print $fh_out '( Contents of @{$curr_set_clms}: )', "\n\n";
	print $fh_out join ("\t", 'Column\Row:', 0 .. $ct_rows -1), "\n\n";

	for (my $c = 0; $c < $ct_clms; $c++) {
	    print $fh_out join ("\t", $c, @{$curr_set_clms->[$c]}), "\n"; # A BUG Fixed on 2019/07/10. #
	    ##print $fh_out join ("\t", $c, @{$curr_set_clms->[0]}), "\n";
	}
	print $fh_out "\n\n";

	print $fh_out '( Contents of @{$curr_bds_blks} & @{$curr_bl_coords}: )', "\n\n";
	print $fh_out join ("\t", 'Indx_gb', 'le_gb_curr', 're_gb_curr', 'gb_coord_curr'), "\n\n";

	for (my $gb = 0; $gb < $ct_gblks; $gb++) {
	    print $fh_out join ("\t", $gb, @{$curr_bds_blks->[$gb]}, $curr_bl_coords->[$gb]), "\n";
	}
	print $fh_out "\n\n";

	print $fh_out '( Contents of @{$curr_gblks_w_spec_lb} & @{$curr_gblks_w_spec_rb} )', "\n\n";
	print $fh_out join ("\t", 'Indx_clm(=c)', '[gap,blocks,w,le=c]', '[gap,blocks,w,re=c]'), "\n\n";
	
	for (my $c = 0; $c < $ct_clms; $c++) {
	    my $cnct_gbs_w_le_eq_c = join (',', @{$curr_gblks_w_spec_lb->[$c]});
	    my $cnct_gbs_w_re_eq_c = join (',', @{$curr_gblks_w_spec_rb->[$c]});

	    print $fh_out join ("\t", $c, "[$cnct_gbs_w_le_eq_c]", "[$cnct_gbs_w_re_eq_c]"), "\n";
	}
	print $fh_out "\n\n";

	return 1;

}	# End of the subroutine, 'print_out_gblock_configs'. #


# print_out_remng_attr_gblks:
#	This satellite subroutine of 'purge_gblock_pair', etc. prints out
#	some remaining attributes of the gap-blocks that 'print_out_gblock_configs' does NOT print out.
#	[CREATED on 2019/07/14.]
#
# INPUTS: (\@inter_block_relations, \@bds_gb_coords, \@gb_coords_wd_frm, \@gblock_sizes, \@cts_vincluding, \@cts_vincluded}, \@cts_onx, \@rows_affd_by_blks, \*${$fh_out}), where
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]}
#	give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@bds_gb_coords and \@gb_coords_wd_frm are outputs of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[\@cts_vincluding, \@cts_vincluded & \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub print_out_remng_attr_gblks (\@\@\@\@\@\@\@\@\*) {

	my ($inter_block_relations, $bds_gb_coords, $gb_coords_wd_frm, $gblock_sizes, $cts_vincluding, $cts_vincluded, $cts_onx, $rows_affd_by_blks, $fh_out) = @_;

	my $ct_gblks = @{$gblock_sizes};

	print $fh_out '( Contents of @{$inter_block_relations}', "  (#{gap-blocks}=$ct_gblks): )\n\n";
	print $fh_out join ("\t", 'Gb1\Gb2:', 0 .. $ct_gblks -1), "\n\n";

	for (my $gb1=0; $gb1 <$ct_gblks; $gb1++) { # 1st main for-loop (over 1st gap-blocks). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];
	    my @out_rels_w_gb1 = ();
	    for (my $gb2=0; $gb2 <$ct_gblks; $gb2++) { # Inner for-loop (over 2nd gap-blocks). #
		my $rel = $rels_w_gb1->[$gb2];
		unless (defined $rel) { $rel = 'undef'; }
	        $out_rels_w_gb1[$gb2] = $rel;

	    } # End of the inner for-loop (over 2nd gap-blocks). #

	    print $fh_out join ("\t", $gb1, @out_rels_w_gb1), "\n";

	} # End of the 1st main for-loop (over 1st gap-blocks). #
	print $fh_out "\n\n";

	
	print $fh_out '( Contents of @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx} & @{$rows_affd_by_blks}', "  (#{gap-blocks}=$ct_gblks): )\n\n"; # MODIFIED on 2019/07/14. #
	##print $fh_out '( Contents of @{$gblock_sizes}, @{$cts_vincluding}, @{$cts_vincluded}, @{$cts_onx} & @{$rows_affd_by_blks}', "  (#{gap-blocks}=$ct_gblks): )\n\n";

	print $fh_out join ("\t", 'Indx_gb', 'lb_coord', 'rb_coord', 'coord_le_wd', 'coord_re_wd', '{block_size}', '#{vincluding_gblks}', '#{vincluded_gblks}', '#{gblks_w_onx_rels}', '[msa,rows,affected,by,gb]'), "\n\n"; # CORRECTED on 2019/07/15. #
	##print $fh_out join ("\t", 'Indx_gb', '{block_size}', 'lb_coord', 'rb_coord', 'coord_le_wd', 'coord_re_wd', '#{vincluding_gblks}', '#{vincluded_gblks}', '#{gblks_w_onx_rels}', '[msa,rows,affected,by,gb]'), "\n\n"; # MODIFIED on 2019/07/14. #
	####print $fh_out join ("\t", 'Indx_gb', '{block_size}', '#{vincluding_gblks}', '#{vincluded_gblks}', '#{gblks_w_onx_rels}', '[msa,rows,affected,by,gb]'), "\n\n";

	for (my $gb=0; $gb <$ct_gblks; $gb++) { # 2nd main for-loop (over gap-blocks). #

	    my $cnct_rows_affd = join (',', @{$rows_affd_by_blks->[$gb]});

	    print $fh_out join ("\t", $gb, @{$bds_gb_coords->[$gb]}, @{$gb_coords_wd_frm->[$gb]}, $gblock_sizes->[$gb], $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb], "[$cnct_rows_affd]"), "\n"; # A BUG Fixed on 2019/07/15. #
	    ##print $fh_out join ("\t", $gb, @{$bds_gb_coords->[$gb]}, @{$gb_coords_wd_frm}, $gblock_sizes->[$gb], $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb], "[$cnct_rows_affd]"), "\n"; # MODIFIED on 2019/07/14. #
	    ####print $fh_out join ("\t", $gb, $gblock_sizes->[$gb], $cts_vincluding->[$gb], $cts_vincluded->[$gb], $cts_onx->[$gb], "[$cnct_rows_affd]"), "\n";

	} # End of the 2nd main for-loop (over gap-blocks). #

	print $fh_out "\n\n";

	return 1;

}	# End of the subroutine, 'print_out_remng_attr_gblks'. #


# br_print_out_remng_attr_gblks:
#	This reduced version of 'print_out_remng_attr_gblks' prints out
#	ONLY those attributes that could change even if the "vertical" (or phylogenetic)
#	attributes of all the gap-blocks remain unchanged.
#	[CREATED on 2019/07/31.]
#
# print_out_remng_attr_gblks:
#	This satellite subroutine of 'purge_gblock_pair', etc. prints out
#	some remaining attributes of the gap-blocks that 'print_out_gblock_configs' does NOT print out.
#	[CREATED on 2019/07/14.]
#
# INPUTS: (\@bds_gb_coords, \@gb_coords_wd_frm, \@gblock_sizes, \*${$fh_out}), where
#
#	[[
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(See the header of 'mk_inter_block_relations' for more details);
#	[Typically, \@inter_block_relations is a main output of 'mk_inter_block_relations' or 'extensive_exam_inter_block_relations', both in "MyGapPatternTopology.pm"];
#	]] ... For the predecessor, 'print_out_remng_attr_gblks'.
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block;
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]}
#	give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window;
#	[Typically, \@bds_gb_coords and \@gb_coords_wd_frm are outputs of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm"];
#
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block;
#	[\@gblock_sizes is the output of 'mk_list_gblock_sizes' in "MyGapPatternTopology.pm"]; 
#
#	[[
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block;
#	[\@cts_vincluding, \@cts_vincluded & \@cts_onx are the outputs of 'cts_vincluding_vincluded_onx_gblks' in "MyGapPatternTopology.pm"];
#
#	@{$rows_affd_by_gblks[$gb]} lists the indices of the rows in the input MSA affected (i.e., given gaps) by the $gb th gap-block;
#	[ \@rows_affd_by_gblks is an output of, e.g., 'extract_br_info_gblocks' in "MyANEX_Supple.pm".]
#	]] ... For the predecessor, 'print_out_remng_attr_gblks'.
#
# OUTPUTS: $fl_end, where
#
#	$fl_end = 1 if the subroutine ended without any remarkable problems,
#		= 0 if it failed.
#
sub br_print_out_remng_attr_gblks (\@\@\@\*) {

	my ($bds_gb_coords, $gb_coords_wd_frm, $gblock_sizes, $fh_out) = @_;

	my $ct_gblks = @{$gblock_sizes};

	print $fh_out '( Contents of @{$bds_gb_coords}, @{$gb_coords_wd_frm}, @{$gblock_sizes}', "  (#{gap-blocks}=$ct_gblks): )\n\n"; # MODIFIED on 2019/07/14. #

	print $fh_out join ("\t", 'Indx_gb', 'lb_coord', 'rb_coord', 'coord_le_wd', 'coord_re_wd', '{block_size}'), "\n\n";

	for (my $gb=0; $gb <$ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    print $fh_out join ("\t", $gb, @{$bds_gb_coords->[$gb]}, @{$gb_coords_wd_frm->[$gb]}, $gblock_sizes->[$gb]), "\n";

	} # End of the main for-loop (over gap-blocks). #

	print $fh_out "\n\n";

	return 1;

}	# End of the subroutine, 'br_print_out_remng_attr_gblks'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #



