#!/usr/bin/perl
#
# screen_complex_errors_via_indel_cts.pl:
#	This script reads in a list of "Complex" errors, and
#	outputs a new list of errors, after filtering out those "Complex" errors
#	each of which involve only a small number of indels.
#	[CREATED on July 19 (Tue), 2016.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);



	# If #{indels} < $UB_CT_INDELS for BOTH reference and reconstructed MSAs,
	# the error is regarded as "non-complex".
our $UB_CT_INDELS = 4;

our @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Tot#{insertions}_Dpars.ref', 'Tot#{insertions}_Dpars.rec', 'Tot#{deletions}_Dpars.ref', 'Tot#{deletions}_Dpars.rec');

our @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID');




	# ADDED in creation of 'summary_examine_purge.raw.pl'. #
#my $input_msa_type = '12primates';
#my $input_msa_type = '15mammals';
my $input_msa_type = '9fe_mammals';


	# Input/Output files. #

my $infile = "../Tbl_Complex/tbl_errsegs_w_complex.$input_msa_type.txt.gz";


##my $auxfile = "../Cts_Dpars_Indels/Summary_tbl/TESTs/Test1c.brief/tbl_cmp_ref_vs_rec_cts_indels.txt.gz"; # For TESTING. #
my $auxfile = "../Cts_Dpars_Indels/Summary_tbl/$input_msa_type/tbl_cmp_ref_vs_rec_cts_indels.txt.gz"; # For a full-scale analysis. #

#my $outdir = "TESTs/Test1c_1"; # For TESTING. #
my $outdir = "."; # For a full-scale analyis. #
###my $outdir = "Wo_ud$UB_CT_INDELS";
my $outfile_name = "tbl_errsegs_w_complex.$input_msa_type.txt";

my $outfile = "$outdir/$outfile_name";


	# Check the input/output files. #

foreach my $file ($infile, $auxfile) {
    unless ( ( -f $file ) and ( -r $file ) ) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a READABLE input file '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}

foreach my $dir ($outdir) {
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
    }
    unless (( -r $dir ) and ( -w $dir )) {
	unless (0 == system ("chmod u+rw $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<< Program name: $program_name (path: $program_path) >>\n\n";

print STDOUT '$input_msa_type = ', $input_msa_type , " .\n\n"; # ADDED in creation of 'summary_examine_purge.raw.pl'. #

print STDOUT "Input file:  $infile .\n\n";

print STDOUT "Auxiliary input files:\n\n";

print STDOUT "    (for the counts of indels)  $auxfile .\n\n";

print STDOUT "Output file:   {out_dir}/$outfile_name ,\n";
print STDOUT "           where {out_dir}  = $outdir .\n\n\n";

print STDOUT "#  An error is regarded as \"Non-Complex\" (and thus filtered out)\n";
print STDOUT "#  if it has #{indels} < $UB_CT_INDELS for both reference and reconstructed MSAs.\n\n\n";


	### Read in the auxiliary file. ###

$/ = "\n"; # Use the newline character as the input-record-separator. #

my $fh_aux = open_input_file2 ($auxfile);
unless (defined $fh_aux) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the auxiliary file '$auxfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $aux_header = <$fh_aux>;

my $indices_fld_aux = tfm_labels2indices_gen (@TAGS_AUX_FLD, $aux_header, "\t");
unless (defined $indices_fld_aux) {
	$aux_header =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the auxiliary header:\n";
	print STDOUT " '$aux_header' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx_aux = 0;
foreach my $indx (@{$indices_fld_aux}) { ($max_indx_aux <$indx) and ($max_indx_aux = $indx); }

my %combo_segid2if_cmplx_cand; # = ("$id_class:$session_id:$seg_id" => 1, ...) for the segments with #{indels} >= $UB_CT_INDELS for either reference or reconstructed MSA.

while (my $line = <$fh_aux>) { # Preliminary inner while-loop (over unmatched segments). #

	my @aux_flds = split /\s+/, $line;
	(@aux_flds > $max_indx_aux) or next;

	my ($id_class, $session_id, $seg_id, $ct_ins_ref, $ct_ins_rec, $ct_del_ref, $ct_del_rec) = @aux_flds[@{$indices_fld_aux}];

	# @TAGS_AUX_FLD = ('ID_Class', 'Session_ID', 'Segment_ID', 'Tot#{insertions}_Dpars.ref', 'Tot#{insertions}_Dpars.rec', 'Tot#{deletions}_Dpars.ref', 'Tot#{deletions}_Dpars.rec');

	((uc($ct_ins_ref) eq 'N/A') or ($ct_ins_ref eq 'undef')) and next; # In this case, the segment is likely "TOO-LONG" to be analyzed. #

	my $totct_indels_ref = $ct_ins_ref + $ct_del_ref;
	my $totct_indels_rec = $ct_ins_rec + $ct_del_rec;

	(($totct_indels_ref < $UB_CT_INDELS) and ($totct_indels_rec < $UB_CT_INDELS)) and next;

	my $combo_segid = join (':', $id_class, $session_id, $seg_id);

	$combo_segid2if_cmplx_cand{$combo_segid} = 1;

}			# End of the preliminary inner while-loop (over unmatched segments). #
close ($fh_aux);

print STDOUT "... Fetched ", scalar (keys %combo_segid2if_cmplx_cand), " candidates of \"Complex\" errors ...\n\n\n";


	### Create a new list of "Complex" errors by filtering out those with only a few indels. ###


my $fh_in = open_input_file2 ($infile);
unless (defined $fh_in) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


my $in_header = <$fh_in>;

print $fh_out $in_header, "\n";

my $indices_fld = tfm_labels2indices_gen (@TAGS_IN_FLD, $in_header, "\t");
unless (defined $indices_fld) {
	$in_header =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the input header:\n";
	print STDOUT " '$in_header' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx = 0;
foreach my $indx (@{$indices_fld}) { ($max_indx <$indx) and ($max_indx = $indx); }

my $ct_old_cmplx = my $ct_filtered_out = my $ct_new_cmplx = 0;

while (my $in_line = <$fh_in>) { # Main while-loop (over input lines). #

	my @in_flds = split /\s+/, $in_line;
	(@in_flds > $max_indx) or next;

	my ($id_class, $session_id, $seg_id) = @in_flds[@{$indices_fld}]; 

		# @TAGS_IN_FLD = ('ID_Class', 'Session_ID', 'Segment_ID');

	### ($session_id eq '0000003') and last; ### For TESTING. ###

	$ct_old_cmplx++;

	my $combo_segid = join (':', $id_class, $session_id, $seg_id);

	if (defined $combo_segid2if_cmplx_cand{$combo_segid}) {

	    $ct_new_cmplx++;
	    print $fh_out $in_line;

	} else {

	    $ct_filtered_out++;
	}

}			# End of the main while-loop (over input lines). #

close ($fh_in);
close ($fh_out);


print STDOUT "\n\n === The Entire Process ENDED !! ===\n\n\n";

print STDOUT "#{Original \"Complex\" errors}  = $ct_old_cmplx .\n";
print STDOUT "    (#{those filtered out}       = $ct_filtered_out )\n\n";
print STDOUT "#{Retained \"Complex\" errors}  = $ct_new_cmplx .\n\n\n";




