package MyGapPatternTopology;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyGapPatternTopology.pm:
#	This package contains some subroutines that deal with problems
#	concerning or related to the topologies of gap-patterns
#	in multiple sequence alignments.
#	[CREATION STARTED on 2019/03/07, ENDED on 2019/03/dd.]	
#
#	[REVISED on 2019/05/17]:
#	(1) In the subroutines, 'compt_aln_degeneracy' and 'chk_ovlp_bw_vequiv_gblks',
#	we replaced the subroutine component, 'ct_necessary_shifts',
#	with 'smlt2_extend_left_end_to_left' or 'smlt2_extend_right_end_to_right',
#	whenever possible;
#	(2) A new subroutine, 'chk_excluded_bin_hrels', was added.
#
#	[REVISED on 2019/06/05]:
#	The subroutine 'compt_aln_degeneracy' has been revised
#	so that it will return $degeneracy = 0
#	whenever it encounters a chain of mutually adjacent collectively complementary gap-blocks
#	even if its components do NOT cause degeneracy.
#

use strict;
use warnings;


	## IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (1) ##

use MyArrayManipu_num qw( diff_nmarray union_nmarray ); # ADDED on 2019/02/05.

	# MODIFIED in creation of "MyGapPatternTopology.pm" (on 2019/03/08). #
use MyTaxonomy qw(
	fetch_common_ancestors
	);
##use MyTaxonomy qw(
##	fetch_ext_offsprings_nr
##	); # RESURRECTED on 2019/01/29.
	## fetch_ancestry 
	## fetch_last_common_ancestor_ov2
	## fetch_common_ancestors
	## fetch_all_offsprings_nr
	## fetch_ext_offsprings fetch_all_offsprings

use MyTreeManipu_II qw(
	);
	## input_tree_file read_tree_file br_output_tree
	## make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

=DISABLED in creation of "MyGapPatternTopology.pm" (on 2019/03/07).

use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	); # RESURRECTED on after 'FAIL3' of 'test_partit_into_eff_indep_indel_blocks.pl', on March 21st (Tue), 2017. #
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	create_initial_set_indels_spt_odr
	);
	## enum_psm_indel_scns_for_msa

=cut # END of "DISABLED in creation of "MyGapPatternTopology.pm" (on 2019/03/07)."

=set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );


use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);

=cut # END of "set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')"

	## END of "IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (1)" ##


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	mk_inter_block_relations
	mk_inter_block_relations2
	mk_list_interfering_blocks
	mk_br2up_down2blocks

	mk_br2up_down2blocksets
	mk_list_interfering_blocksets
	mk_list_cmpl_interfering_blocksets
	extensive_exam_inter_block_relations

	compt_size_gblock
	enum_sites_contribute_to_gblk
	mk_list_gblock_sizes
	three_smallest_sizes
	mk_list_coll_cmpl_gblocks
	mk_gblock2coll_cmpl
	search_for_coll_cmpl_gblks
	cmp_nmarrays
	compt_aln_degeneracy

	ct_necessary_shifts
	compt_distance_bw_gblk_and_pt
	compt_inter_gblock_distance

	mk_lists_gblocks_w_spec_bounds

	cts_vincluding_vincluded_onx_gblks

	vextend_right_end_sbj_gblk
	vextend_left_end_sbj_gblk
	vextend_rear_ends_rflnk_vincluded_gblks
	vextend_rear_ends_lflnk_vincluded_gblks

	vretract_left_end_sbj_gblk
	vretract_right_end_sbj_gblk
	vretract_right_ends_vincluded_gblks
	vretract_left_ends_vincluded_gblks

	mk_sets_vert_equiv_gblks
	mk_sets_swappable_gblks
	chk_ovlp_bw_vequiv_gblks
	chk_pos_orders_amg_swpbl_gblks

	chk_chng_gptrn_topology_psm_R
	chk_chng_gptrn_topology_psm_L
	smlt_extend_left_end_to_left
	smlt_extend_right_end_to_right
	encode_gptrn_topology_psm

	smlt2_extend_left_end_to_left
	smlt2_extend_right_end_to_right

	examine_if_gbs_on_msa_ends
	chk_excluded_bin_hrels

	expand_list_x_bhrels
	);


sub mk_inter_block_relations (\@$$\%\%$\%);
sub mk_inter_block_relations2 (\@$\%); # ADDED on 2019/08/22.
sub mk_list_interfering_blocks (\@);
sub mk_br2up_down2blocks (\@$$);

sub mk_br2up_down2blocksets (\@$$\%\%$\%);
sub mk_list_interfering_blocksets (\@$$\%\%$\%\%);
sub mk_list_cmpl_interfering_blocksets (\@);
sub extensive_exam_inter_block_relations (\@$$\%\%$\%\%);

sub compt_size_gblock ($\@$$\@);
sub enum_sites_contribute_to_gblk ($\@$$\@); # ADDED on 2019/07/30.
sub mk_list_gblock_sizes (\@$$\@);

sub three_smallest_sizes (\@\@);
sub mk_list_coll_cmpl_gblocks (\%$\%\%\@);
sub mk_gblock2coll_cmpl ($\@);
sub search_for_coll_cmpl_gblks (\@$$$\%\%\@);
sub cmp_nmarrays (\@\@); 
sub compt_aln_degeneracy (\@\@\@\@\@\@\@); # Major revision on 2019/04/04.
##sub compt_aln_degeneracy (\@\@\@\@);

sub ct_necessary_shifts ($$$\@\@\%);
sub compt_distance_bw_gblk_and_pt ($$\@\@);
sub compt_inter_gblock_distance ($$\@\@);

sub mk_lists_gblocks_w_spec_bounds ($$\@);
sub vextend_right_end_sbj_gblk ($\@\@\@\@\@);
sub vextend_left_end_sbj_gblk ($\@\@\@\@\@);
sub vextend_rear_ends_rflnk_vincluded_gblks ($\@\@\@);
sub vextend_rear_ends_lflnk_vincluded_gblks ($\@\@\@);

sub vretract_left_end_sbj_gblk ($\@\@\@\@);
sub vretract_right_end_sbj_gblk ($\@\@\@\@);
sub vretract_right_ends_vincluded_gblks ($\@\@\@\@);
sub vretract_left_ends_vincluded_gblks ($\@\@\@\@);

sub mk_sets_vert_equiv_gblks (\%\%);
sub mk_sets_swappable_gblks (\@\@);
sub br_chk_ovlp_bw_vequiv_gblks (\@\@); # MODIFIED on 2019/04/03.
sub chk_ovlp_bw_vequiv_gblks (\@\@\@\@\@); # ADDED on 2019/04/03.
sub chk_pos_orders_amg_swpbl_gblks (\@\@);

		# Added from 2019/04/07 to 2019/04/09. #
sub chk_chng_gptrn_topology_psm_R ($$$\@\@\@\@\@\@\@);
sub chk_chng_gptrn_topology_psm_L ($$$\@\@\@\@\@\@\@);
sub smlt_extend_left_end_to_left ($$\@\@\@);
sub smlt_extend_right_end_to_right ($$\@\@\@);
sub encode_gptrn_topology_psm (\@\@\@\@\@\@);

		# Added on 2019/04/11. #
sub smlt2_extend_left_end_to_left ($$\@\@\@);
sub smlt2_extend_right_end_to_right ($$\@\@\@);

sub examine_if_gbs_on_msa_ends (\@\@$$); # Added on 2019/05/23. #
sub chk_excluded_bin_hrels (\@\@\@\@\@); # Added on 2019/05/27. #

sub expand_list_x_bhrels (\@\@); # Added on 2020/05/17.

our $SOPHIST_MVBLTY_TEST_FOR_CAD = 0;
#our $SOPHIST_MVBLTY_TEST_FOR_CAD = 1;
	# =1 if the 'compt_aln_degeneracy' employs a sophisticated version of the movability test,
	# = 0 otherwise.
	# (ADDED in MANOR REVISION of 'compt_aln_degeneracy' on 2019/04/05.)


	## IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (2) ##

	# This specifies the level of tests on the subroutines, 
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

##		# For the subroutine, 'cal_init_psm_cands_for_sgl_msa':
##		# (IMPORTED from 'classify_msa_errors_via_mblks.2.pl' in creation of 'MyTreeMap_indels_spt_odr2.pm'.)
##our $TEST_LEVEL_cipc2 = 0;
####our $TEST_LEVEL_cipc2 = 5;

	## END of "IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (2)" ##

	# For the subroutine, 'extensive_exam_inter_block_relations':
our $TEST_LEVEL_exex_iblk_rltns = 0;
##our $TEST_LEVEL_exex_iblk_rltns = 1;
##our $TEST_LEVEL_exex_iblk_rltns = 5;


	# For the subroutine, 'search_for_coll_cmpl_gblks'. #
our $TEST_LEVEL_sch4ccgb = 0;
#our $TEST_LEVEL_sch4ccgb = 1;
#our $TEST_LEVEL_sch4ccgb = 5;

	# For the subroutine 'compt_aln_degeneracy'.
our $TEST_LEVEL_cad = 0;
#our $TEST_LEVEL_cad = 1;
#our $TEST_LEVEL_cad = 5;

	# For the subroutine, 'vextend_right_end_sbj_gblk'. #
our $TEST_LEVEL_vx_re_sgb = 0;
#our $TEST_LEVEL_vx_re_sgb = 5;

	# For the subroutine, 'vextend_left_end_sbj_gblk'. #
our $TEST_LEVEL_vx_le_sgb = 0;
#our $TEST_LEVEL_vx_le_sgb = 5;

	# For the subroutines 'chk_chng_gptrn_topology_psm_R' & 'chk_chng_gptrn_topology_psm_L'. #
our $TEST_LEVEL_ccgpt_p = 0;
#our $TEST_LEVEL_ccgpt_p = 1;
#our $TEST_LEVEL_ccgpt_p = 3;


	## IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (3) ##

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.

	## END of "IMPORTED from "MyTreeMap_indels_spt_odr2.pm" (on 2019/03/07)). (3)" ##


# mk_inter_block_relations:
#	This subroutine examines the relations between blocks,
#	and records them to a 2D array, @inter_block_relations.
#	[CREATED on 2019/03/08.]
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, \%node2pa, \%node2ch, $topnode_id, \%br2eq), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#
# OUTPUT: (\@inter_block_relations), where
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1.
#	$relations can be:
#
#	'NIF' (for "non-interfering"),
#	'S' (for "effective sibling"),
#	'Cp' (for "complementary"),
#	'ONN' (for "overlapping yet non-nesting" NOR "(effective-)complementary-sibling",
#	'ONCS' (for "overlapping yet non-nesting" BUT "(effective-)complementary-sibling",
#	'>' (if the $gb2 th block is "vertically included" in the $gb1 th block),
#	'>(ch)' (if the $gb2 th block is "vertically included" in, and an "effective child" of, the $gb1 th block),
#	'<' (if the $gb2 th block vertically includes the $gb1 th block),
#	'<(pa)' (if the $gb2 th block vertically includes, and is an "effective parent" of, the $gb1 th block),
#	'=' (for "vertically identical").
#
#	[NOTE: The "effective sibling" includes the 'parent-child' relationship.
#	      Likewise, the "complementary effective sibling" includes the 'complementary parent-child' relationship.]
#
sub mk_inter_block_relations (\@$$\%\%$\%) {

	my ($list_gblocks, $indx_br, $indx_u_or_d, $node2pa, $node2ch, $topnode_id, $br2eq) = @_;

# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, \%node2pa, \%node2ch, $topnode_id, \%br2eq), where

	my @inter_block_relations = ();

	my $ct_gblks = @{$list_gblocks};

	for (my $gb1 = 0; $gb1 < $ct_gblks; $gb1++) { # Main outer for-loop (over $gb1).

	    my ($br1, $u_or_d1) = @{$list_gblocks->[$gb1]}[$indx_br, $indx_u_or_d];
	    my $eq_br1 = $br2eq->{$br1};
	    my $pa1 = $node2pa->{$br1};
	    my $sibs1 = $node2ch->{$pa1};

	    my @relations_w_gb1 = ();

	    for (my $gb2 = 0; $gb2 < $ct_gblks; $gb2++) { # Main middle for-loop (over $gb2).

		if ($gb2 == $gb1) { # (0) Identical.

		    $relations_w_gb1[$gb2] = undef;
		    next;
		}

		my ($br2, $u_or_d2) = @{$list_gblocks->[$gb2]}[$indx_br, $indx_u_or_d];
		my $eq_br2 = $br2eq->{$br2};
		my $pa2 = $node2pa->{$br2};
		my $sibs2 = $node2ch->{$pa2};

			# (1) (vertically) equivalent or complementary. #

		if ($br1 == $br2) {
		    $relations_w_gb1[$gb2] = ($u_or_d1 eq $u_or_d2) ? '=' : 'Cp';
		    next;
		} elsif ((defined $eq_br2) and ($br1 == $eq_br2)) {
		    $relations_w_gb1[$gb2] = ($u_or_d1 eq $u_or_d2) ? 'Cp' : '=';
		    next;
		}

			# (2) $br2 is the parent of $br1, or its equivalent. #

		if ($br2 == $pa1) {

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} == 2) {
				$rel = '>(ch)';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = '>';
			    }
			} else { # if ($u_or_d2 eq 'L')
			    if (@{$sibs1} == 2) {
				$rel = 'ONCS';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'ONN';
			    }
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} == 2) {
				$rel = 'S';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'NIF';
			    }
			} else { # if ($u_or_d2 eq 'L')
			    if (@{$sibs1} == 2) {
				$rel = '<(pa)';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = '<';
			    }
			}
		    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} elsif ((defined $eq_br2) and ($eq_br2 == $pa1)) {

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} == 2) {
				$rel = 'ONCS';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'ONN';
			    }
			} else { # if ($u_or_d2 eq 'L')
			    if (@{$sibs1} == 2) {
				$rel = '>(ch)';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = '>';
			    }
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} == 2) {
				$rel = '<(pa)';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = '<';
			    }
			} else { # if ($u_or_d2 eq 'L')
			    if (@{$sibs1} == 2) {
				$rel = 'S';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'NIF';
			    }
			}
		    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} # End of "if ($br2 == $pa1) {...} elsif ((defined $eq_br2) and ($eq_br2 == $pa1)) {...}".

			# (3) $br1 is the parent of $br2, or its equivalent. #

		if ($br1 == $pa2) {

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = '<(pa)';

			    } else { # ASSUME [(@{$sibs1} > 2) ... WRONG!!] (@{$sibs2} > 2) ... CORRECTED.
				$rel = '<';
			    }
			} else { # if ($u_or_d2 eq 'L')

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = 'S';

			    } else { # ASSUME [(@{$sibs1} > 2) ... WRONG!!] (@{$sibs2} > 2) ... CORRECTED.
				$rel = 'NIF';
			    }
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = 'ONCS';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'ONN';
			    }
			} else { # if ($u_or_d2 eq 'L')

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = '>(ch)';

			    } else { # ASSUME [(@{$sibs1} > 2) ... WRONG!!]  (@{$sibs2}>2) ... CORRECTED.
				$rel = '>';
			    }
			}
		    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} elsif ((defined $eq_br1) and ($eq_br1 == $pa2)) {

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = 'ONCS';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = 'ONN';
			    }
			} else { # if ($u_or_d2 eq 'L')

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = '>(ch)';

			    } else { # ASSUME [(@{$sibs1} > 2) ... WRONG!!] (@{$sibs2} > 2) ... CORRECTED.
				$rel = '>';
			    }
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = '<(pa)';
			    } else { # ASSUME (@{$sibs1} > 2).
				$rel = '<';
			    }
			} else { # if ($u_or_d2 eq 'L')

			    if (@{$sibs2} == 2) { # A BUG fixed on 2019/03/12.
			    ##if (@{$sibs1} == 2) {

				$rel = 'S';
			    } else { # ASSUME [(@{$sibs1} > 2) ... WRONG!!] (@{$sibs2} > 2) ... CORRECTED.
				$rel = 'NIF';
			    }
			}
		    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} # End of "if ($br1 == $pa2) {...} elsif ((defined $eq_br1) and ($eq_br1 == $pa2)) {...}".


			# (4) $br1 is a sibling of $br2.

		if ($pa1 == $pa2) {
		    my $rel;

		    if (($pa1 == $topnode_id) and (@{$sibs1} == 2)) {
			# In this case, $br1 and $br2 are actually equivalent to each other.
			$relations_w_gb1[$gb2] = ($u_or_d1 eq $u_or_d2) ? 'Cp' : '=';
			next;

		    } elsif ( (($pa1 == $topnode_id) and (@{$sibs1} == 3))
			or (($pa1 != $topnode_id) and (@{$sibs1} == 2)) ) {
			# In the normal case (of bifurcating tree).

			if ($u_or_d1 eq 'U') {
			    if ($u_or_d2 eq 'U') {
				$rel = 'ONCS';
			    } else { # if ($u_or_d2 eq 'L')
				$rel = '>(ch)';
			    }
			} else { # if ($u_or_d1 eq 'L')
			    if ($u_or_d2 eq 'U') {
				$rel = '<(pa)';
			    } else { # if ($u_or_d2 eq 'L')
				$rel = 'S';
			    }
			}  # End of "if ($u_or_d1 eq 'U') {...} else {...}"

		    } else {

			# $br1 and $br2 have more siblings than normal. (ASSUME that no node has a single child.)

			if ($u_or_d1 eq 'U') {
			    if ($u_or_d2 eq 'U') {
				$rel = 'ONN';
			    } else { # if ($u_or_d2 eq 'L')
				$rel = '>';
			    }
			} else { # if ($u_or_d1 eq 'L')
			    if ($u_or_d2 eq 'U') {
				$rel = '<';
			    } else { # if ($u_or_d2 eq 'L')
				$rel = 'NIF';
			    }
			}  # End of "if ($u_or_d1 eq 'U') {...} else {...}"

		    } # End of "if (($pa1 == $topnode_id) and (@{$sibs1} == 2)) {...} elsif (...) {...} else {...}".

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} # End of "if ($pa_br1 == $pa_br2) {...}".


			# (5) Other cases. #

		my ($common_anc, $div1, $div2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});

		if (@{$div1} == 0) { # $br1 is an ancestor of $br2.

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    $rel = '<';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = 'NIF';
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    $rel = 'ONN';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = '>';
			}
		    }

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} elsif (@{$div2} == 0) { # $br2 is an ancestor of $br1.

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    $rel = '>';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = 'ONN';
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    $rel = 'NIF';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = '<';
			}
		    }

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} else { # $br1 and $br2 diverged from each other at some node (possibly the root) in the tree.

		    my $rel;

		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    $rel = 'ONN';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = '>';
			}
		    } else { # if ($u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    $rel = '<';
			} else { # if ($u_or_d2 eq 'L')
			    $rel = 'NIF';
			}
		    }

		    $relations_w_gb1[$gb2] = $rel;
		    next;

		} # End of "if (@{$div1} == 0) {...} elsif (@{$div2} == 0) {...} else {...}".

	    }  # End of the main middle for-loop (over $gb2).

	    $inter_block_relations[$gb1] = \@relations_w_gb1;

	}  # End of the main outer for-loop (over $gb1).

	return \@inter_block_relations;

}	# End of the subroutine 'mk_inter_block_relations'. #


# mk_inter_block_relations2:
#	This subroutine does the same job as the predecessor, 'mk_inter_block_relations',
#	via a totally different method.
#	Unlike the predecessor, it can also handle a tree with mono-furcated non-root nodes.
#	[CREATED on 2019/08/22.]
#
# mk_inter_block_relations:
#	This subroutine examines the relations between blocks,
#	and records them to a 2D array, @inter_block_relations.
#	[CREATED on 2019/03/08.]
#
# INPUTS: (\@rows_affd_by_gblks, $ct_rows, \%set_msa_rows2br_up_or_lw), where
#
#	@{$rows_aff_by_gblks[$gb]} lists the indices of the rows of the MSA affected by the $gb th gap-block;
#	[Typically, \@rows_affd_by_gblks is an input of 'extract_br_info_gblocks' in "MyANEX_Supple.pm"];
#
#	$ct_rows is the number of rows in the input MSA;
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id;
#	[Typically, \%set_msa_rows2br_up_or_lw is the output of 'mk_set_msa_rows2br_up_or_lw(_dp)' in "MyTreeMap_sppl.pm"];
#
# OUTPUTS: (\@inter_block_relations), where
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1;
#	(For details on the relations, see the header of 'mk_inter_block_relations' above.)
#
sub mk_inter_block_relations2 (\@$\%) {

	my ($rows_affd_by_gblks, $ct_rows, $set_msa_rows2br_up_or_lw) = @_;

	my $ct_gblks = @{$rows_affd_by_gblks};

	my @inter_block_relations = ();
	for (my $gb=0; $gb <$ct_gblks; $gb++) {
	    $inter_block_relations[$gb] = [];
	    $inter_block_relations[$gb]->[$gb] = undef;
	}

	for (my $gb1 = 0; $gb1 < $ct_gblks-1; $gb1++) { # Main outer for-loop (over $gb1).

	    my $rows_affd1 = $rows_affd_by_gblks->[$gb1];
	    my $ct_rows_affd1 = @{$rows_affd1};

	    for (my $gb2 = $gb1+1; $gb2 < $ct_gblks; $gb2++) { # Main middle for-loop (over $gb2).

		my $rows_affd2 = $rows_affd_by_gblks->[$gb2];
		my $ct_rows_affd2 = @{$rows_affd2};

		my ($shared, $ex1, $ex2) = diff_nmarray (@{$rows_affd1}, @{$rows_affd2});

		my $ct_shared = @{$shared};
		my $ct_ex1 = @{$ex1};
		my $ct_ex2 = @{$ex2};

		if (($ct_ex1 == 0) and ($ct_ex2 == 0)) {
			# $gb1 and $gb2 affect the identical set of rows. #

		    $inter_block_relations[$gb1]->[$gb2] = '=';
		    $inter_block_relations[$gb2]->[$gb1] = '=';
		    next;
		} # End of "if (($ct_ex1 == 0) and ($ct_ex2 == 0)) {...}"


		if ($ct_shared ==0) { # $gb1 and $gb2 do NOT overlap vertically to each other. #

		   my $union_1_2 = union_nmarray (@{$rows_affd1}, @{$rows_affd2});
		   my $cnct_union_1_2 = join (',', @{$union_1_2});

		   if ($ct_ex1 + $ct_ex2 == $ct_rows) {

				# The sets of rows affected by $gb1 and $gb2 are complementary to each other. #

			$inter_block_relations[$gb1]->[$gb2] = 'Cp';
			$inter_block_relations[$gb2]->[$gb1] = 'Cp';

		   } elsif (defined $set_msa_rows2br_up_or_lw->{$cnct_union_1_2}) {

				# The union of rows affected by $gb1 and $gb2 is separated by a ("parent") branch. #

			$inter_block_relations[$gb1]->[$gb2] = 'S';
			$inter_block_relations[$gb2]->[$gb1] = 'S';

		   } else {
				# $gb1 and $gb2 are vertically "non-interfering". #

			$inter_block_relations[$gb1]->[$gb2] = 'NIF';
			$inter_block_relations[$gb2]->[$gb1] = 'NIF';
		   }
		   next;

		} # End of "if ($ct_shared ==0) {...}"


			## if ($ct_shared > 0) ## $gb1 and $gb2 overlap vertically to each other. #

		my $cnct_ex1 = join (',', @{$ex1});
		my $cnct_ex2 = join (',', @{$ex2});

		if ($ct_ex1 == 0) { # $gb2 includes $gb1. #

		    if (defined $set_msa_rows2br_up_or_lw->{$cnct_ex2}) {

				# Effective "parent-child" relationship. #

			$inter_block_relations[$gb1]->[$gb2] = '<(pa)';
			$inter_block_relations[$gb2]->[$gb1] = '>(ch)';
		    } else {
			$inter_block_relations[$gb1]->[$gb2] = '<';
			$inter_block_relations[$gb2]->[$gb1] = '>';
		    }
		    next;
		} # End of "if ($ct_ex1 == 0) {...}"

		if ($ct_ex2 == 0) { # $gb1 includes $gb2. #

		    if (defined $set_msa_rows2br_up_or_lw->{$cnct_ex1}) {

				# Effective "parent-child" relationship. #

			$inter_block_relations[$gb1]->[$gb2] = '>(ch)';
			$inter_block_relations[$gb2]->[$gb1] = '<(pa)';
		    } else {
			$inter_block_relations[$gb1]->[$gb2] = '>';
			$inter_block_relations[$gb2]->[$gb1] = '<';
		    }
		    next;
		} # End of "if ($ct_ex2 == 0) {...}"

			## if (($ct_shared >0) and ($ct_ex1 >0) and ($ct_ex2 >0)) ##

			# NOTE: In this case (with an underlying tree),
			#	@{$shared} is the complement of union (@{$ex1}, @{$ex2}),
			#     & @{$ex1} is the complement of @{$rows_affd2}, 
			#     & @{$ex2} is the complement of @{$rows_affd1}.

		my $cnct_shared = join (',', @{$shared});

		if (defined $set_msa_rows2br_up_or_lw->{$cnct_shared}) {

			# union (@{$ex1}, @{$ex2}) is separated by a branch. #
		    
			$inter_block_relations[$gb1]->[$gb2] = 'ONCS';
			$inter_block_relations[$gb2]->[$gb1] = 'ONCS';
		} else {
			$inter_block_relations[$gb1]->[$gb2] = 'ONN';
			$inter_block_relations[$gb2]->[$gb1] = 'ONN';

		} # End of "if (defined $set_msa_rows2br_up_or_lw->{$cnct_shared}) {...} else {...}".

	    }  # End of the main middle for-loop (over $gb2).

	}  # End of the main outer for-loop (over $gb1).

	return \@inter_block_relations ;

}	# End of the subroutine, 'mk_inter_block_relations2'. #


# mk_list_interfering_blocks:
#	This subroutine creates a list of interfering blocks (@interfering_blocks)
#	out of the set of iter-block relations (@inter_block_relations),
#	which is the output of 'mk_inter_block_relations' above.
#	[CREATED on 2019/03/08.]
#
#	NOTE: Here, we will only include the following relations:
#
#	'S' (for "effective sibling"),
#	'Cp' (for "complementary"),
#	'>(ch)' (if the $gb2 th block is "vertically included" in, and an "effective child" of, the $gb1 th block),
#	'<(pa)' (if the $gb2 th block vertically includes, and is an "effective parent" of, the $gb1 th block),
#	'=' (for "vertically identical").
#
#	This is because we are now concerned ONLY with changes in the gap-pattern topology (in terms of parsimonious ancestral gap-states).
#
#	[Supplementary notes]
#	(1) & (2) Blocks with the 'ONN' or 'ONCS' relations cause NO topological changes, as far as we employ the current coordinate system (and as far as we restrict ourselves to the parsimonious ancestral gap states);
#	(3) As far as we rectrict ourselves to the parsimonious ancestral gap states, 
#	blocks with the '>' (but NOT '>(ch)') or '<' (but NOT '<(pa)') cause NO topological changes.
#
# INPUTS: (\@inter_block_relations), where
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	for the relation of $gb2 against $gb1.
#	As for the values that $relation can take,
#	see the header of 'mk_inter_block_relations' above.
#
# OUTPUTS: (\@interfering_blocks), where
#
#	@{$interfering_blocks[$gb]} = (\@blocks, \@relations)
#	stores information on blocks that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	$blocks[$k] is the index (or rank) of the $k th interfering block,
#	$relations[$k] is the relation of the $k th interfering block against the $gb th block.
#
sub mk_list_interfering_blocks (\@) {

	my ($inter_block_relations) = @_;

	my @interfering_blocks = ();

	my %keep = (
		'S' => 1,
		'Cp' => 1,
		'=' => 1,
		'>(ch)' => 1,
		'<(pa)' => 1,
		);

	my $ct_blks = @{$inter_block_relations};

	for (my $gb=0; $gb < $ct_blks; $gb++) { # Outer for-loop (over $gb). #

	    my $rel_w_gb = $inter_block_relations->[$gb];

	    my @blocks = my @relations = ();

	    for (my $gb2 =0; $gb2 < $ct_blks; $gb2++) { # Inner for-loop (over $gb2).

		if ($gb2 == $gb) { next; }

		my $rel = $rel_w_gb->[$gb2];

		if (defined $keep{$rel}) {
		    push @blocks, $gb2;
		    push @relations, $rel;
		}

	    } # End of the inner for-loop (over $gb2).

	    $interfering_blocks[$gb] = (@blocks>0) ? [\@blocks, \@relations] : [] ;

	} # End of the outer for-loop (over $gb). #

	return \@interfering_blocks;

}	# End of the subroutine 'mk_list_interfering_blocks'. #

# mk_br2up_down2blocks:
#	This quite simple but useful subroutine creates a hash of hash, %br2up_down2blocks,
#	using only a main list of gap-blocks.
#	[CREATED on 2019/03/13.]
#
#	NOTE: It should NOT be confused with the similarly named subroutine, 'mk_br2up_down2blocksets', below.
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
# OUTPUTS: (\%br2up_down2blocks), where
#
#	%br2up_down2blocks = ($branch => {$up_or_down => \@blocks, ...}, ...),
#	where @blocks lists the gap-blocks each of which is separated by, and on the $up_or_down side of,
#	the branch with ID $branch.
#
sub mk_br2up_down2blocks (\@$$) {

	my ($list_gblocks, $indx_br, $indx_u_or_d) = @_;

	my $ct_gblks = @{$list_gblocks};

	my %br2up_down2blocks;
	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($br, $u_or_d) =  @{$list_gblocks->[$gb]};

	    my $up_down2blocks = $br2up_down2blocks{$br};
	    unless (defined $up_down2blocks) {
		$up_down2blocks = $br2up_down2blocks{$br} = {};
	    }
	    my $blocks = $up_down2blocks->{$u_or_d};
	    unless (defined $blocks) {
		$blocks = $up_down2blocks->{$u_or_d} = [];
	    }
	    push @{$blocks}, $gb;

	} # End of the main for-loop (over gap-blocks). #

	return \%br2up_down2blocks;

}	# END of the subroutine 'mk_br2up_down2blocks'. #


# mk_br2up_down2blocksets:
#	This satellite subroutine of 'mk_list_interfering_blocksets'
#	creates a hash of hash, %br2up_down2blocksets,
#	using a list of gap-blocks and main information on the phylogenetic tree.
#	[CREATED on 2019/03/09.]
#
#	NOTE: It should NOT be confused with the similarly named subroutine, 'mk_br2up_down2blocks', above.
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, \%node2pa, \%node2ch, $topnode_id), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[Typically, \%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
# OUTPUTS: (\%br2up_down2blocksets), where
#
#	%br2up_down2blocksets = ($branch => {$up_or_down => \@blocksets, ...}, ...),
#	where @{$blocksets[$bs]} lists the gap-blocks constituting the $bs th block-set that is separated by, and on the $up_or_down side of, the branch with ID $branch.
#
#	[NOTE that each block-set forms a (complementary-)monophyletic group.]
#
sub mk_br2up_down2blocksets (\@$$\%\%$\%) {

	my ($list_gblocks, $indx_br, $indx_u_or_d, $node2pa, $node2ch, $topnode_id, $node2depth) = @_;

	my $ct_gblks = @{$list_gblocks}; ## A BUG fixed on 2019/03/12!!
	##my $ct_gblks = 0;

		# (o) Initialization. #

	my %pre_br2up_down2blocksets;
	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Preliminary for-loop (over gap-blocks). #
	    my ($br, $u_or_d) =  @{$list_gblocks->[$gb]};

	    my $up_down2blocksets = $pre_br2up_down2blocksets{$br};
	    unless (defined $up_down2blocksets) {
		$up_down2blocksets = $pre_br2up_down2blocksets{$br} = {};
	    }
	    my $blocksets = $up_down2blocksets->{$u_or_d};
	    unless (defined $blocksets) {
		$blocksets = $up_down2blocksets->{$u_or_d} = [];
	    }
	    push @{$blocksets}, [$gb];
	} # End of the preliminary for-loop (over gap-blocks). #


		# (a) The bottom-up traversal. #

	foreach my $br (sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2pa}) { # 1st outer foreach-loop (over branches in descending order of their depths).

	    my $children = $node2ch->{$br};

	    unless ((defined $children) and (@{$children}>0)) { next; } # Skip external nodes.

	    my $ct_children = @{$children};

	    my @sets_d_blksets_ch = ();
	    for (my $i=0; $i < $ct_children; $i++) { # 1st middle for-loop (over children).
		my $ch = $children->[$i];
		my $u_or_d2blksets_ch = $pre_br2up_down2blocksets{$ch};
		unless (defined $u_or_d2blksets_ch) { last; }
		my $d_blksets_ch = $u_or_d2blksets_ch->{'L'};
		unless (defined $d_blksets_ch) { last; }
		push @sets_d_blksets_ch, $d_blksets_ch;

	    }  # End of the 1st middle for-loop (over children).

	    unless (@sets_d_blksets_ch == $ct_children) { next; }

	    my @blocksets = ([]);

	    foreach my $d_blksets_ch (@sets_d_blksets_ch) { # 2nd middle for-loop (over block-sets on children).

		my @new_blocksets = ();
		foreach my $old_blkset (@blocksets) { # Inner foreach-loop (over old block-sets).
		    foreach my $add_blkset (@{$d_blksets_ch}) { # Innermost foreach-loop (over block-sets to be added).
			my @new_blkset = (@{$old_blkset}, @{$add_blkset});
			push @new_blocksets, \@new_blkset;

		    } # End of the innermost foreach-loop (over block-sets to be added).
		} # End of the inner foreach-loop (over old block-sets).

		if (@new_blocksets>0) {
		    @blocksets = @new_blocksets; # Update the list of block-sets.
		} else {
		    @blocksets = ();
		    last;
		}

	    } # End of the 2nd middle for-loop (over block-sets on children).

	   unless (@blocksets > 0) { next; }

	   my $pre_up_down2blocksets = $pre_br2up_down2blocksets{$br};
	   unless (defined $pre_up_down2blocksets) {
	       $pre_up_down2blocksets = $pre_br2up_down2blocksets{$br} = {};
	   }
	   my $pre_d_blocksets = $pre_up_down2blocksets->{'L'};
	   unless (defined $pre_d_blocksets) {
	       $pre_d_blocksets = $pre_up_down2blocksets->{'L'} = [];
	   }
	   push @{$pre_d_blocksets}, @blocksets;
	    
	} # End of the 1st outer foreach-loop (over branches in descending order of their depths). #

		# (c) The top-down traversal. #

	foreach my $br (sort {$node2depth->{$a} <=> $node2depth->{$b} } keys %{$node2pa}) { # 2nd outer foreach-loop (over branches in ascending order of their depths).

	    my $pa = $node2pa->{$br};
	    my $sibs = $node2ch->{$pa};

	    my $ct_src_nodes = scalar (@{$sibs}) -1;
	    my @sets_src_blocksets = ();

	    if (defined $node2pa->{$pa}) { # The parent node represents a branch as well.
		$ct_src_nodes++;
		my $up_down2blocksets_pa = $pre_br2up_down2blocksets{$pa};
		unless (defined $up_down2blocksets_pa) { next; }
		my $pre_u_blocksets_pa = $up_down2blocksets_pa->{'U'};
		unless (defined $pre_u_blocksets_pa) { next; }

		push @sets_src_blocksets, $pre_u_blocksets_pa;
	    }

	    foreach my $sib (@{$sibs}) { # 1st middle foreach-loop (over siblings).

		if ($sib == $br) { next; }
		my $up_down2blocksets = $pre_br2up_down2blocksets{$sib};
		unless (defined $up_down2blocksets) { last; }
		my $pre_d_blocksets_sib = $up_down2blocksets->{'L'};
		unless (defined $pre_d_blocksets_sib) { last; }

		push @sets_src_blocksets, $pre_d_blocksets_sib;

	    } # End of the 1st middle foreach-loop (over siblings).

	    unless (@sets_src_blocksets == $ct_src_nodes) { next; }

	    my @blocksets = ([]);

	    foreach my $src_blksets (@sets_src_blocksets) { # 2nd middle foreach-loop (over block-sets on siblings and parent (if at all)).

		my @new_blocksets = ();
		foreach my $old_blkset (@blocksets) { # Inner foreach-loop (over old block-sets).
		    foreach my $add_blkset (@{$src_blksets}) { # Innermost foreach-loop (over block-sets to be added).
			my @new_blkset = (@{$old_blkset}, @{$add_blkset});
			push @new_blocksets, \@new_blkset;

		    } # End of the innermost foreach-loop (over block-sets to be added).
		} # End of the inner foreach-loop (over old block-sets).

		if (@new_blocksets>0) {
		    @blocksets = @new_blocksets; # Update @blocksets.
		} else {
		    @blocksets = ();
		    last;
		}

	    } # End of the 2nd middle foreach-loop (over block-sets on siblings and parent (if at all)).

	    unless (@blocksets >0) { next; }

	    my $pre_up_down2blocksets = $pre_br2up_down2blocksets{$br};
	    unless (defined $pre_up_down2blocksets) {
		$pre_up_down2blocksets = $pre_br2up_down2blocksets{$br} = {};
	    }

	    my $pre_u_blocksets = $pre_up_down2blocksets->{'U'};
	    unless (defined $pre_u_blocksets) {
		$pre_u_blocksets = $pre_up_down2blocksets->{'U'} = [];
	    }
	    push @{$pre_u_blocksets}, @blocksets;

	} # End of the 2nd outer foreach-loop (over branches in ascending order of their depths).

		# (d) Cleaning-up the output, by removing trivial block-sets (each consisting of a single block). #

	my %br2up_down2blocksets;

	foreach my $br (keys %pre_br2up_down2blocksets) { # 3rd outer foreach-loop (over branches).

	    my $up_down2pre_blksets = $pre_br2up_down2blocksets{$br};
	    my %up_down2blocksets;

	    foreach my $u_or_d (keys %{$up_down2pre_blksets}) { # Middle foreach-loop (over 'U' and 'L').
		my $pre_blksets = $up_down2pre_blksets->{$u_or_d};
		my @blocksets = ();
		foreach my $blkset (@{$pre_blksets}) { # Inner foreach-loop (over block-sets).
		    if (@{$blkset}>1) { push @blocksets, $blkset; }
		} # End of the inner foreach-loop (over block-sets).

		if (@blocksets>0) { $up_down2blocksets{$u_or_d} = \@blocksets; }

	    } # End of the middle foreach-loop (over 'U' and 'L').

	    if (0 < scalar keys %up_down2blocksets) {
		$br2up_down2blocksets{$br} = \%up_down2blocksets;
	    }

	} # End of the 3rd outer foreach-loop (over branches).


	return \%br2up_down2blocksets;

}	# End of the subroutine 'mk_br2up_down2blocksets'. #


# mk_list_interfering_blocksets:
#	This subroutine constructs @interfering_blocksets,
#	whose $gb th elements gives information on the (complementary-)monophyletic sets of gap-blocks
#	that can interfere with the $bl th gap-block.	
#	[CREATED on 2019/03/09.]
#
#	NOTE: Here, for the moment, we will only consider the following relations:
#	'S' (for "effective sibling"),
#	'>(ch)' (if the blockset is "vertically included" in, and an "effective child" of, the $gb th block).
#
#	This is because we are concerned only with changes in the gap-pattern topology (besides, especially in this subroutine, regarding parsimonious ancestral gap states).
#
#	[Supplementary notes]
#	(5) As for the blocksets with '>' (but NOT '>(ch)'), the same arguments apply as that for (3) (in "mk_list_interfering_blocks" above);
#	(6) Each blockset with the 'Cp' relation likely involves a single block with the 'S' relation;
#	(7) Likewise, '=' likely involves a single block with the '>(ch)' relation;
#	(8) '<(pa)' likely involves a single block with the '=', '>', and/or the 'S' relations.
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, \%node2pa, \%node2ch, $topnode_id, \%br2eq, \%node2depth), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[Typically, \%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
# OUTPUTS: (\@interfering_blocksets, \%br2up_down2blocksets), where
#
#	@{$interfering_blocksets[$gb]} = (\@blocksets, \@relations) 
#	stores information on block-SETs that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	@{$blocksets[$k]} lists the indices (or ranks) of the gap-blocks constiruting the $k th interfering block-set,
#	$relations[$k] is the relation of the $k th interfering block-set against the $gb th block;
#	[NOTE: The gap-blocks in each @{$blocksets[$k]} ALWAYS form a (complementary-)monophyletic group];
#
#	%br2up_dwon2blocksets is the output of 'mk_br2up_down2blocksets' above;
#	[See the header of 'mk_br2up_down2blocksets' for more details.]
#
sub mk_list_interfering_blocksets (\@$$\%\%$\%\%) {

	my ($list_gblocks, $indx_br, $indx_u_or_d, $node2pa, $node2ch, $topnode_id, $br2eq, $node2depth) = @_;

	my $ct_gblks = @{$list_gblocks};

	my $br2up_down2blocksets = mk_br2up_down2blocksets (@{$list_gblocks}, $indx_br, $indx_u_or_d, %{$node2pa}, %{$node2ch}, $topnode_id, %{$node2depth});
	unless (defined $br2up_down2blocksets) {
	    print STDERR "[ERROR 'mk_list_interfering_blocksets'] Subroutine 'mk_br2up_down2blocksets' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my @interfering_blocksets = ();

	unless (0 < scalar keys %{$br2up_down2blocksets}) {
		# There should be NO interfering block-sets, because there are NO non-trivial block-sets (each consisting of two or more blocks). #

	    for (my $gb = 0; $gb < $ct_gblks; $gb++) {
		$interfering_blocksets[$gb] = [];
	    }

	    return (\@interfering_blocksets, $br2up_down2blocksets);
	}

		# MAIN: Construct non-trivial @interfering_blocksets. #

	for (my $gb1 = 0; $gb1 < $ct_gblks; $gb1++) { # Main outer for-loop (over $gb1).

	    my @blocksets = my @relations = ();

	    my ($br1, $u_or_d1) = @{$list_gblocks->[$gb1]}[$indx_br, $indx_u_or_d];

	    my $eq_br1 = $br2eq->{$br1};
	    my $pa1 = $node2pa->{$br1};
	    my $children1 = $node2ch->{$br1};
	    my $eq_children1 = (defined $eq_br1) ? $node2ch->{$eq_br1} : [];
	    my $sibs1 = $node2ch->{$pa1};

		# (a) Examine the parent of $br1. #

	    my $up_down2blocksets_pa = $br2up_down2blocksets->{$pa1};
	    if (defined $up_down2blocksets_pa) {
		foreach my $u_or_d2 (keys %{$up_down2blocksets_pa}) { # 1st inner foreach-loop (over $u_or_d2).
		    my $blsets_pa = $up_down2blocksets_pa->{$u_or_d2};
		    my $rel;
		    if ($u_or_d1 eq 'U') {
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} == 2) { $rel = '>(ch)'; }
			} else { # if ($u_or_d2 eq 'L')
			    next; # $rel = 'NNCS'.
			}
		    } else { # if (u_or_d1 eq 'L')
			if ($u_or_d2 eq 'U') {
			    if (@{$sibs1} ==2) { $rel = 'S'; }
			} else { # if ($u_or_d2 eq 'L')
			    next; # $rel = '<' or '<(pa)'.
			}
		    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".
		    if (defined $rel) {
			foreach my $blset (@{$blsets_pa}) {
			    push @blocksets, $blset;
			    push @relations, $rel;
			}
		    }
		} # End of the 1st inner foreach-loop (over $u_or_d2).
		
	    } # End of "if (defined $up_down2blocksets_pa) {...}".

		# (b) Examine the children of $br1. #

	    if ((defined $children1) and (@{$children1} == 2)) {

		foreach my $ch (@{$children1}) { # 2nd middle foreach-loop (over children).
		    my $up_down2blocksets_ch = $br2up_down2blocksets->{$ch};
		    if (defined $up_down2blocksets_ch) {
			foreach my $u_or_d2 (keys %{$up_down2blocksets_ch}) { # 2nd inner foreach-loop (over $u_or_d2).
			    my $blsets_ch = $up_down2blocksets_ch->{$u_or_d2};
			    my $rel;
			    if ($u_or_d1 eq 'U') {
				if ($u_or_d2 eq 'U') {
				    next; # $rel = '<(pa)'.
				} else { # if ($u_or_d2 eq 'L')
				    $rel = 'S';
				}
			    } else { # if ($u_or_d1 eq 'L')
				if ($u_or_d2 eq 'U') {
				    next; # $rel = 'NNCS'.
				} else { # if ($u_or_d2 eq 'L')
				    $rel = '>(ch)';
				}
			    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".
			    if (defined $rel) {
				foreach my $blset (@{$blsets_ch}) {
				    push @blocksets, $blset;
				    push @relations, $rel;
				}
			    } # End of "if (defined $rel) {...}".

			} # End of the 2nd inner foreach-loop (over $u_or_d2).

		    } # End of "if (defined $up_down2blocksets_ch) {...}".

		} # End of the 2nd middle foreach-loop (over children).

	    } # End of "if ((defined $children1) and (@{$children1} == 2)) {...}".

		# (c) Examine the children of $eq_br1.

	    if ((defined $eq_children1) and (@{$eq_children1} == 2)) {
		foreach my $eq_ch (@{$eq_children1}) { # 3rd middle foreach-loop (over children of $eq_br1). #
		    my $up_down2blocksets_eq_ch = $br2up_down2blocksets->{$eq_ch};
		    if (defined $up_down2blocksets_eq_ch) {
			foreach my $u_or_d2 (keys %{$up_down2blocksets_eq_ch}) { # 3rd inner foreach-loop (over $u_or_d2).
			    my $blsets_eq_ch = $up_down2blocksets_eq_ch->{$u_or_d2};
			    my $rel;
			    if ($u_or_d1 eq 'U') {
				if ($u_or_d2 eq 'U') {
				    next; # $rel = 'NNCS'.
				} else { # if ($u_or_d2 eq 'L')
				    $rel = '>(ch)';
				}
			    } else { # if ($u_or_d1 eq 'L')
				if ($u_or_d2 eq 'U') {
				    next; # $rel = '<(pa)'.
				} else { # if ($u_or_d2 eq 'L')
				    $rel = 'S';
				}
			    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".
			    if (defined $rel) {
				foreach my $blset (@{$blsets_eq_ch}) {
				    push @blocksets, $blset;
				    push @relations, $rel;
				}
			    } # End of "if (defined $rel) {...}".

			} # End of the 3rd inner foreach-loop (over $u_or_d2).
		    } # End of "if (defined $up_down2blocksets_eq_ch) {...}".

		} # End of the 3rd middle foreach-loop (over children of $eq_br1).

	    } # End of "if ((defined $eq_children) and (@{$eq_children1} == 2)) {...}".

			# (d) Examine the siblings of $br1. #

	    if ( (($pa1 == $topnode_id) and (@{$sibs1} == 3)) or (($pa1 != $topnode_id) and (@{$sibs1} == 2)) ) {
		foreach my $sib (@{$sibs1}) { # 5th middle foreach-loop (over siblings).
		    if ($sib == $br1) { next; }
		    my $up_down2blocksets_sib = $br2up_down2blocksets->{$sib};
		    if (defined $up_down2blocksets_sib) {
			foreach my $u_or_d2 (keys %{$up_down2blocksets_sib}) { # 5th inner foreach-loop (over $u_or_d2).
			    my $blsets_sib = $up_down2blocksets_sib->{$u_or_d2};
			    my $rel;
			    if ($u_or_d1 eq 'U') {
				if ($u_or_d2 eq 'U') {
				    next; # $rel = 'NNCS'.
				} else { # if ($u_or_d2 eq 'L')
				    $rel = '>(ch)';
				}
			    } else { # if ($u_or_d1 eq 'L')
				if ($u_or_d2 eq 'U') {
				    next; # $rel = '<(pa)'
				} else { # if ($u_or_d2 eq 'L')
				    $rel = 'S';
				}
			    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".
			    if (defined $rel) {
				foreach my $blset (@{$blsets_sib}) {
				    push @blocksets, $blset;
				    push @relations, $rel;
				}
			    } # End of "if (defined $rel) {...}".

			} # End of the 5th inner foreach-loop (over $u_or_d2).
		    } # End of "if (defined $up_down2blocksets_sib) {...}".

		} # End of the 5th middle foreach-loop (over siblings).
	    } # End of "if ( (($pa1 == $topnode_id) and (@{$sibs1} == 3)) or (($pa1 != $topnode_id) and (@{$sibs1} == 2)) ) {...}".

	    $interfering_blocksets[$gb1] = (@blocksets>0) ? [\@blocksets, \@relations] : [] ;

	} # End of the main outer for-loop (over $gb1).

	return (\@interfering_blocksets, $br2up_down2blocksets);

}	# END of the subroutine 'mk_list_interfering_blocksets'. #


# mk_list_cmpl_interfering_blocksets:
#	This subroutine receives \@interfering_blocksets,
#	which is the output of 'mk_list_interfering_blocksets' above,
#	and creates its "inverse" list,
#	consisting of the "complement" of each interfering blockset.
#	[CREATED on 2019/03/10.]
#
# INPUTS: (\@interfering_blocksets), where
#
#	@{$interfering_blocksets[$gb]} = (\@blocksets, \@relations) 
#	stores information on block-SETs that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	@{$blocksets[$k]} lists the indices (or ranks) of the gap-blocks constiruting the $k th interfering block-set,
#	$relations[$k] is the relation of the $k th interfering block-set against the $gb th block.
#	(NOTE: The gap-blocks in each @{$blocksets[$k]} ALWAYS form a (complementary-)monophyletic group.)
#
# OUTPUTS: (\@cmpl_interfering_blocksets), where
#
#	@{$cmpl_interfering_blocksets[$gb]} = (\@cmpl_blocksets, \@relations)
#	records block-blockset pairs in @interfering_blocksets in which
#	the $gb th block a constituent of @blockset = @{$interfering_blocksets[$gb2]->[0][$k2]} for some $gb2 and $k2.
#	(But it is empty if the $gb th block is NOT a constituent of ANY block-sets.)
#	We will use the following convension:
#	@{$cmpl_blocksets[$k]} = ($gb2, @blockset above with $gb removed)
#	and
#	$relations[$k] = the "complement" of $interfering_blocksets[$gb2]->[1][$k2].
#	Thus, the relation should be:
#	'S' if the corresponding relation in @interfering_blockset is 'S',
#	'<(pa)' if the corresponding relation in @interfering_blockset is '>(ch)'.
#		
sub mk_list_cmpl_interfering_blocksets (\@) {

	my ($interfering_blocksets) = @_;

	my $ct_gblks = @{$interfering_blocksets};

		# (o) Initialization. #

	my @cmpl_interfering_blocksets = ();
	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Preliminary for-loop.
	    $cmpl_interfering_blocksets[$gb] = [ [], [] ];
	} # End of the preliminary for-loop.


		# (i) Raw construction.

	for (my $gb2 = 0; $gb2 < $ct_gblks; $gb2++) { # Main outer for-loop (over $gb2).

	    my ($blocksets2, $relations2) = @{$interfering_blocksets->[$gb2]};
	    unless ((defined $blocksets2) and (@{$blocksets2}>0)) { next; } # Skip if $gb2 has NO interfering blocksets.

	    my $ct_blksets2 = @{$blocksets2};
	    for (my $k2=0; $k2 < $ct_blksets2; $k2++) { # Middle for-loop (over block-sets).
		my $blkset2 = $blocksets2->[$k2];
		my $ct_blks2 = @{$blkset2};
		my $rel2 = $relations2->[$k2];
		my $rel = ($rel2 eq 'S') ? 'S' : '<(pa)';

		my @blkset_1st_half = ($gb2);
		my @blkset_2nd_half = @{$blkset2};
		for (my $i=0; $i<$ct_blks2; $i++) { # Inner for-loop (over blocks in the block-set).
		    my $gb = shift @blkset_2nd_half;
		    my @blkset = (@blkset_1st_half, @blkset_2nd_half);

		    my ($cmpl_blocksets, $relations) = @{$cmpl_interfering_blocksets[$gb]};
		    push @{$cmpl_blocksets}, \@blkset;
		    push @{$relations}, $rel;

			# Prepare for the next round. #
		    push @blkset_1st_half, $gb;

		} # End of the inner for-loop (over blocks in the block-set).

	    } # End of the middle for-loop (over block-sets).

	} # End of the main outer for-loop (over $gb2).

		# (ii) Clearing up.

	for (my $gb=0; $gb < $ct_gblks; $gb++) { # Trailing for-loop (over $gb).
	    my ($cmpl_blocksets, $relations) = @{$cmpl_interfering_blocksets[$gb]};
	    unless ((defined $cmpl_blocksets) and (@{$cmpl_blocksets}>0)) {
		$cmpl_interfering_blocksets[$gb] = [];
	    } 

	} # End of the trailing for-loop (over $gb).


	return \@cmpl_interfering_blocksets;

}	# END of the subroutine 'mk_list_cmpl_interfering_blocksets'. #


# extensive_exam_inter_block_relations:
#	This subroutine extensively examines
#	relationships between gap-blocks.
#	Actually, it is a conglomerate of the above four main subroutines, that is,
#	'mk_inter_block_relations', 'mk_list_interfering_blocks',
#	'mk_list_interfering_blocksets', and 'mk_list_cmpl_interfering_blocksets'.
#	[CREATED on 2019/03/10.]
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, \%node2pa, \%node2ch, $topnode_id, \%br2eq, \%node2depth), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[Typically, \%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
# OUTPUTS: (\@inter_block_relations, \@interfering_blocks, \@interfering_blocksets, \@cmpl_interfering_blocksets, \%br2up_down2blocksets), where
#
#	\@inter_block_relations is the output of 'mk_inter_block_relations' above;
#	[See the header of 'mk_inter_block_relations' for details];
#
#	\@interfering_blocks is the output of 'mk_list_interfering_blocks' above;
#	[See the header of 'mk_list_interfering_blocks' for details];
#
#	\@interfering_blocksets is an output of 'mk_list_interfering_blocksets' above;
#	[See the header of 'mk_list_interfering_blocksets' for details];
#
#	\@cmpl_interfering_blocksets is the output of 'mk_list_cmpl_interfering_blocksets' above;
#	[See the header of 'mk_list_cmpl_interfering_blocksets' for details];
#
#	\%br2up_dwon2blocksets is the output of 'mk_br2up_down2blocksets' above, which is the satellite subroutine of 'mk_list_interfering_blocksets';
#	[See the header of 'mk_br2up_down2blocksets' for details];
#
sub extensive_exam_inter_block_relations (\@$$\%\%$\%\%) {

	my ($list_gblocks, $indx_br, $indx_u_or_d, $node2pa, $node2ch, $topnode_id, $br2eq, $node2depth) = @_;

	my $ct_gblks = @{$list_gblocks};

    (print STDOUT "<< Subroutine 'extensive_exam_inter_block_relations' (#{gap-blocks} = $ct_gblks) >>\n\n") if ($TEST_LEVEL_exex_iblk_rltns>0);


	my ($inter_block_relations) = mk_inter_block_relations (@{$list_gblocks}, $indx_br, $indx_u_or_d, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq});
	unless (defined $inter_block_relations) {
	    print STDERR "[ERROR 'extensive_exam_inter_block_relations'] Subroutine 'mk_inter_block_relations' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_inter_block_relations' Ended ...\n\n") if ($TEST_LEVEL_exex_iblk_rltns>1);

    if ($TEST_LEVEL_exex_iblk_rltns>2) {

	print STDOUT "\n", '[ Contents of @{$inter_block_relations} ]', "\n\n";
	print STDOUT join ("\t", 'Gb1\gb2:', (0 .. $ct_gblks-1) ), "\n\n";

	for (my $gb1  = 0; $gb1 < $ct_gblks; $gb1++) {
	    my @out_inter_block_rels = @{$inter_block_relations->[$gb1]};
	    $out_inter_block_rels[$gb1] = 'undef';
	    print STDOUT join ("\t", $gb1, @out_inter_block_rels), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_exex_iblk_rltns>2) {...}".


	my ($interfering_blocks) = mk_list_interfering_blocks (@{$inter_block_relations});
	unless (defined $interfering_blocks) {
	    print STDERR "[ERROR 'extensive_exam_inter_block_relations'] Subroutine 'mk_list_interfering_blocks' FAILED: $! !\n\n";
	    return ($inter_block_relations, undef, undef, undef, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_list_interfering_blocks' Ended ...\n\n") if ($TEST_LEVEL_exex_iblk_rltns>1);

    if ($TEST_LEVEL_exex_iblk_rltns>2) {

	print STDOUT "\n", '[ Contents of @{$interfering_blocks} ]', "\n\n";
	print STDOUT join ("\t", 'Gb', 'Intfr_gb(rel)_1', 'Intfr_gb(rel)_2', '...'), "\n\n";

	for (my $gb =0; $gb < $ct_gblks; $gb++) {
	    my ($blocks, $rels) = @{$interfering_blocks->[$gb]};

	    unless (defined $blocks) { # ADDED for DEBUGGING (on 2019/04/15). #
		print STDOUT join ("\t", $gb, 'NONE'), "\n";
		next;
	    }

	    my $ct_blks = @{$blocks};
	    my @out = ();
	    for (my $k=0; $k <$ct_blks; $k++) {
		my $gb2 = $blocks->[$k];
		my $rel = $rels->[$k];
		push @out, "$gb2($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_exex_iblk_rltns>2) {...}".


	my ($interfering_blocksets, $br2up_down2blocksets) = mk_list_interfering_blocksets (@{$list_gblocks}, $indx_br, $indx_u_or_d, %{$node2pa}, %{$node2ch}, $topnode_id, %{$br2eq}, %{$node2depth});
	unless (defined $interfering_blocksets) {
	    print STDERR "[ERROR 'extensive_exam_inter_block_relations'] Subroutine 'mk_list_interfering_blocksets' FAILED: $! !\n\n";
	    return ($inter_block_relations, $interfering_blocks, undef, undef, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_list_interfering_blocksets' Ended ...\n\n") if ($TEST_LEVEL_exex_iblk_rltns>1);

    if ($TEST_LEVEL_exex_iblk_rltns>3) {

	print STDOUT '[ Contents of %{$br2up_down2blocksets} ]', "\n\n";
	print STDOUT join ("\t", 'Br_ID', 'up(U)_or_down(L)', '[gap,block,set]_1', '[gap,block,set]_2', '...'), "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocksets}) {
	    my $up_down2blocksets = $br2up_down2blocksets->{$br};
	    foreach my $u_or_d (sort keys %{$up_down2blocksets}) {
		my $blocksets = $up_down2blocksets->{$u_or_d};
		my @out_blksets = ();
		foreach my $blkset (@{$blocksets}) {
		    my $cnct_blkset = join (',', @{$blkset});
		    push @out_blksets, "[$cnct_blkset]"; 
		}
		print STDOUT join ("\t", $br, $u_or_d, @out_blksets), "\n\n";
	    }
	}
	print STDOUT "\n";
	
    } # End of "if ($TEST_LEVEL_exex_iblk_rltns>3) {...}".

    if ($TEST_LEVEL_exex_iblk_rltns>2) {

	print STDOUT "\n", '[ Contents of @{$interfering_blocksets} ]', "\n\n";
	print STDOUT join ("\t", 'Gb', '[interfering,gap,block,set](rel)_1', '[interfering,gap,block,set](rel)_2', '...'), "\n\n";

	for (my $gb=0; $gb < $ct_gblks; $gb++) {
	    my ($blocksets, $rels) = @{$interfering_blocksets->[$gb]};

	    unless (defined $blocksets) { # ADDED for debugging (on 2019/03/12). #
		print STDOUT join ("\t", $gb, 'NONE'), "\n\n";
		next;
	    }

	    my $ct_blksets = @{$blocksets};
	    my @out = ();
	    for (my $k=0; $k < $ct_blksets; $k++) {
		my $cnct_blkset = join (',', @{$blocksets->[$k]});
		my $rel = $rels->[$k];
		push @out, "[$cnct_blkset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out), "\n\n";
	}
	print STDOUT "\n";
	
    } # End of "if ($TEST_LEVEL_exex_iblk_rltns>2) {...}".


	my $cmpl_interfering_blocksets = mk_list_cmpl_interfering_blocksets (@{$interfering_blocksets});
	unless (defined $cmpl_interfering_blocksets) {
	    print STDERR "[ERROR 'extensive_exam_inter_block_relations'] Subroutine 'mk_list_cmpl_interfering_blocksets' FAILED: $! !\n\n";
	    return ($inter_block_relations, $interfering_blocks, $interfering_blocksets, undef, $br2up_down2blocksets);
	}
    (print STDOUT "\n ... Subroutine 'mk_list_cmpl_interfering_blocksets' Ended ...\n\n") if ($TEST_LEVEL_exex_iblk_rltns>1);

    if ($TEST_LEVEL_exex_iblk_rltns>2) {

	print STDOUT "\n", '[ Contents of @{$cmpl_interfering_blocksets} ]', "\n\n";
	print STDOUT join ("\t", 'Gb', '[complementary,interfering,gap,block,set](rel)_1', '[complementary,interfering,gap,block,set](rel)_2', '...'), "\n\n";

	for (my $gb=0; $gb < $ct_gblks; $gb++) {
	    my ($cmpl_blocksets, $rels) = @{$cmpl_interfering_blocksets->[$gb]};

	    unless (defined $cmpl_blocksets) { # ADDED for debugging (on 2019/03/12).
		print STDOUT join ("\t", $gb, 'NONE'), "\n\n";
		next;
	    }

	    my $ct_cblksets = @{$cmpl_blocksets};
	    my @out = ();
	    for (my $k=0; $k < $ct_cblksets; $k++) {
		my $cnct_cmpl_blkset = join (',', @{$cmpl_blocksets->[$k]});
		my $rel = $rels->[$k];
		push @out, "[$cnct_cmpl_blkset]($rel)";
	    }
	    print STDOUT join ("\t", $gb, @out), "\n\n";
	}
	print STDOUT "\n";
	
    } # End of "if ($TEST_LEVEL_exex_iblk_rltns>2) {...}".


    (print STDOUT "\n<< Subroutine 'extensive_exam_inter_block_relations' ENDED />/>\n\n") if ($TEST_LEVEL_exex_iblk_rltns>0);

	return ($inter_block_relations, $interfering_blocks, $interfering_blocksets, $cmpl_interfering_blocksets, $br2up_down2blocksets);

}	# END of the subroutine 'extensive_exam_inter_block_relations'. #


# compt_size_gblock:
#	This satellite subroutine of 'mk_list_gblock_sizes'
#	computes the size of a specified gap-block.
#	[CREATED on 2019/03/13.]
#
# INPUTS: ($gb, \@list_gblocks, $indx_lbd, $indx_rbd, \@inter_block_relations), where
#
#	$gb is the index (within @list_gblocks) of the gap-block whose size you want to compute;
#
#	($lbd, $rbd) = @{$list_gblocks[$gb]}[$indx_lbd, $indx_rbd]
#	give the left- and right-boundaries of the $gb th gap-block (both boundaries are included in the block);
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relationship of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is the output of, e.g., 'mk_inter_block_relations' above];
#
# OUTPUTS: $gb_size, which is the size of the $gb th gap-block.
#
sub compt_size_gblock ($\@$$\@) {

	my ($gb, $list_gblocks, $indx_lbd, $indx_rbd, $inter_block_relations) = @_;

	my $ct_gblks = @{$list_gblocks};

	my ($lbd, $rbd) = @{$list_gblocks->[$gb]}[$indx_lbd, $indx_rbd];
	my $relations_w_gb = $inter_block_relations->[$gb];

	my %clm2straddled;

	for (my $gb2 = 0; $gb2 < $ct_gblks; $gb2++) { # Outer for-loop (over $gb2).

	    if ($gb2 == $gb) { next; } # Skip the subject itself.

	    my $rel = $relations_w_gb->[$gb2];

	    unless ( ($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb2 < $gb))
		or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; } # NOTE: Normally, gap-blocks with $rel eq 'ONN' or 'ONCS' should NOT be straddled by $gb; However, I included them just in case. (In any case, including them should be harmless.)

	    my ($lbd2, $rbd2) = @{$list_gblocks->[$gb2]}[$indx_lbd, $indx_rbd];

	    if (($rbd < $lbd2) or ($rbd2 < $lbd)) { next; }

	    my $lbd_ovlp = ($lbd > $lbd2) ? $lbd : $lbd2;
	    my $rbd_ovlp = ($rbd < $rbd2) ? $rbd : $rbd2;

	    for (my $c = $lbd_ovlp; $c <= $rbd_ovlp; $c++) { # Inner for-loop (over columns in the overlap between $gb and $gb2).
		my $straddled = $clm2straddled{$c};
		unless (defined $straddled) { $straddled = $clm2straddled{$c} = []; }

		push @{$straddled}, $gb2; # Corrected on 2019/05/22. (... although with NO effect on the output.)
		##push @{$straddled}, $c;

	    } # End of the inner for-loop (over columns in the overlap between $gb and $gb2).

	} # End of the outer for-loop (over $gb2).

	my $size = $rbd - $lbd + 1;
	$size -= scalar (keys %clm2straddled);

	return $size;

}	# End of the subroutine 'compt_size_gblock'. #


# enum_sites_contribute_to_gblk:
#	This subroutine enumerates the sites that contribute to a specified gap-block.
#	It was created based on 'compt_size_gblock', by modifying ONLY its last part.
#	[CREATED on 2019/07/30.]
#
# compt_size_gblock:
#	This satellite subroutine of 'mk_list_gblock_sizes'
#	computes the size of a specified gap-block.
#	[CREATED on 2019/03/13.]
#
# INPUTS: ($gb, \@list_gblocks, $indx_lbd, $indx_rbd, \@inter_block_relations), where
#
#	$gb is the index (within @list_gblocks) of the gap-block whose size you want to compute;
#
#	($lbd, $rbd) = @{$list_gblocks[$gb]}[$indx_lbd, $indx_rbd]
#	give the left- and right-boundaries of the $gb th gap-block (both boundaries are included in the block);
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relationship of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is the output of, e.g., 'mk_inter_block_relations' above];
#
# OUTPUTS: (\@sites_cntrbt_to_gb), where
#
#	@sites_cntrbt_to_gb stores the indices of the sites (i.e, columns)
#	contributing to $gb.
#
sub enum_sites_contribute_to_gblk ($\@$$\@) {

	my ($gb, $list_gblocks, $indx_lbd, $indx_rbd, $inter_block_relations) = @_;

	my $ct_gblks = @{$list_gblocks};

	my ($lbd, $rbd) = @{$list_gblocks->[$gb]}[$indx_lbd, $indx_rbd];
	my $relations_w_gb = $inter_block_relations->[$gb];

	my %clm2straddled;

	for (my $gb2 = 0; $gb2 < $ct_gblks; $gb2++) { # Outer for-loop (over $gb2).

	    if ($gb2 == $gb) { next; } # Skip the subject itself.

	    my $rel = $relations_w_gb->[$gb2];

	    unless ( ($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb2 < $gb))
		or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; } # NOTE: Normally, gap-blocks with $rel eq 'ONN' or 'ONCS' should NOT be straddled by $gb; However, I included them just in case. (In any case, including them should be harmless.)

	    my ($lbd2, $rbd2) = @{$list_gblocks->[$gb2]}[$indx_lbd, $indx_rbd];

	    if (($rbd < $lbd2) or ($rbd2 < $lbd)) { next; }

	    my $lbd_ovlp = ($lbd > $lbd2) ? $lbd : $lbd2;
	    my $rbd_ovlp = ($rbd < $rbd2) ? $rbd : $rbd2;

	    for (my $c = $lbd_ovlp; $c <= $rbd_ovlp; $c++) { # Inner for-loop (over columns in the overlap between $gb and $gb2).
		my $straddled = $clm2straddled{$c};
		unless (defined $straddled) { $straddled = $clm2straddled{$c} = []; }

		push @{$straddled}, $gb2; # Corrected on 2019/05/22. (... although with NO effect on the output.)
		##push @{$straddled}, $c;

	    } # End of the inner for-loop (over columns in the overlap between $gb and $gb2).

	} # End of the 1st outer for-loop (over $gb2).


		# MODIFIED in creation of 'enum_sites_contribute_to_gblk' (on 2019/07/30). #

	my @sites_cntrbt_to_gb = ();

	for (my $c=$lbd; $c <= $rbd; $c++) { # 2nd outer for-loop (over sites).
	    unless (defined $clm2straddled{$c}) { push @sites_cntrbt_to_gb, $c; }
	} # End of the 2nd outer for-loop (over sites).

	return \@sites_cntrbt_to_gb;

=for the predecessor 'compt_size_gblock'.
	my $size = $rbd - $lbd + 1;
	$size -= scalar (keys %clm2straddled);
	return $size;
=cut # End of "for the predecessor 'compt_size_gblock'."

		# End of "MODIFIED in creation of 'enum_sites_contribute_to_gblk' (on 2019/07/30)." #

}	# End of the subroutine 'enum_sites_contribute_to_gblk'. #


# mk_list_gblock_sizes:
#	This subroutine creats @gblock_sizes, which stores the (horizontal) sizes of the gap-blocks,
#	each which is defined as the number of columns actually affected by each gap-block.
#	[CREATED on 2019/03/13.]
#
# INPUTS: (\@list_gblocks, $indx_lbd, $indx_rbd, \@inter_block_relations), where
#
#	($lbd, $rbd) = @{$list_gblocks[$gb]}[$indx_lbd, $indx_rbd]
#	give the left- and right-boundaries of the $gb th gap-block (both boundaries are included in the block);
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relationship of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is the output of, e.g., 'mk_inter_block_relations' above];
#
# OUTPUTS: \@gblock_sizes, where
#	$gblock_sizes[$gb] gives the (horizontal) block-size of the $gb th gap-block.
#
sub mk_list_gblock_sizes (\@$$\@) {

	my ($list_gblocks, $indx_lbd, $indx_rbd, $inter_block_relations) = @_;

	my $ct_gblks = @{$list_gblocks};

	my @gblock_sizes = ();

	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Main for-loop (over $gb).

	    my $gb_size = compt_size_gblock ($gb, @{$list_gblocks}, $indx_lbd, $indx_rbd, @{$inter_block_relations});
	    unless (defined $gb_size) {
		print STDERR "[ERROR 'mk_list_gblock_sizes'] Subroutine 'compt_size_gblock' FAILED on the $gb th gap-block: $! !\n\n";
		return undef;
	    }

	    $gblock_sizes[$gb] = $gb_size;

	} # End of the main for-loop (over $gb).

	return \@gblock_sizes;

}	# END of the subroutine 'mk_list_gblock_sizes'. #


# three_smallest_sizes:
#	This satellite subroutine of 'mk_list_coll_cmpl_gblocks'
#	finds the three smallest sizes out of the list of sizes.
#	(NOTE: When two or more entries have a same size, the size is counted ONLY ONCE.)
#	[CREATED on 2019/03/13.]
#
# INPUTS: (\@list_sizes, \@indices_rlv), where
#
#	$list_sizes[$k] is the size of the $k th entry;
#
#	@indices_rlv stores the indices of the relevant entries.
#
# OUTPUTS: ($size1st, $size2nd, $size3rd), where
#
#	$size1st, $size2nd, $size3rd are the 1st, 2nd, and 3rd SMALLEST values among the relevant sizes.
#	(The latter two outputs could be undef if the relevant entries have only one or two sizes.)
#
sub three_smallest_sizes (\@\@) {

	my ($list_sizes, $indices_rlv) = @_;

	my $ct_rlv = @{$indices_rlv};

	my $size1st = $list_sizes->[$indices_rlv->[0]];
	my ($size2nd, $size3rd);

	for (my $k = 1; $k < $ct_rlv; $k++) { # Main for-loop (over relevant entries).

	    my $indx = $indices_rlv->[$k];
	    my $size = $list_sizes->[$indx];

	    if ($size < $size1st) {
		if (defined $size2nd) { $size3rd = $size2nd; }
		$size2nd = $size1st;
		$size1st = $size;
		next;
	    } elsif ($size == $size1st) {
		next;
	    }

		# $size1st < $size. #

	    unless (defined $size2nd) {
		$size2nd = $size;
		next;
	    }

	    if ($size < $size2nd) {
		$size3rd = $size2nd;
		$size2nd = $size;
		next;
	    } elsif ($size == $size2nd) {
		next;
	    }

		# $size2nd < $size.

	    unless (defined $size3rd) {
		$size3rd = $size;
		next;
	    }

	    if ($size < $size3rd) {
		$size3rd = $size;
	    }

	} # End of the main for-loop (over relevant entries).

	return ($size1st, $size2nd, $size3rd);

} 	# End of the subroutine 'three_smallest_sizes'. #


# mk_list_coll_cmpl_gblocks:
#	This subroutine creates a list of sets of (collectively) complementary gap-blocks.
#	Each such set consists of 2 or more gap-blocks that are (collectively) complementary to one another.
#	[CREATED on 2019/03/13; CORRECTED on 2019/03/17, 2019/03/19.]
#
# INPUTS: (\%br2up_down2blocks, $topnode_id, \%node2ch, \%node2depth, \@gblock_sizes), where
#
#	%br2up_down2blocks = ($branch => {$up_or_down => \@blocks, ...}, ...),
#	where @blocks lists the gap-blocks each of which is separated by, and on the $up_or_down side of,
#	the branch with ID $branch;
#	($up_or_down = 'U'/'L' if the block is on the upper/lower-side of the branch);
#	[Typically, it is an output of, e.g, 'mk_br2up_down2blocks' above];
#
#	$topnode_id is the ID of the top-node, i.e., root;
#
#	%node2ch = ($node_id => \@children, ...) for all internal nodes;
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[Typically, \%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is an output of, e.g., 'mk_list_gblock_sizes' above];
#
# OUTPUTS: (\@coll_cmpl_gblocks), where
#
#	@{$coll_cmpl_gblocks[$cc]} stores the indices of the gap-blocks constituting the $cc th collectively complementary gap-block (in ascending order of the block-size);
# 
sub mk_list_coll_cmpl_gblocks (\%$\%\%\@) {

	my ($br2up_down2blocks, $topnode_id, $node2ch, $node2depth, $gblock_sizes) = @_;

		# Initialize the output array. #
	my @coll_cmpl_gblocks = ();

		# (i) Bottom-up traversal. #

	my %br2cmpos_blks_lower;
		# = ($br => \@cmpos_blks_lower, ...), where
		# @cmpos_blks_lower stores the (composite) gap-blocks that are on the 'lower'-side of $br.

	foreach my $br (sort {$node2depth->{$b} <=> $node2depth->{$a}} keys %{$node2depth}) { # 1st outer foreach-loop (over $br (including $topnode_id)).

	    my @cmpos_blks_lower = ();

		# First, examine single gap-blocks on the lower-side of $br.
	    my $up_down2blocks = $br2up_down2blocks->{$br};
	    if (defined $up_down2blocks) {
		my $blocks_lower = $up_down2blocks->{'L'};
		if (defined $blocks_lower) {
		    foreach my $gb (@{$blocks_lower}) { push @cmpos_blks_lower, [$gb]; }
		}
	    }

	    my $children = $node2ch->{$br};

		## A BUG fixed on 2019/03/19. (B) ##
	 if ((defined $children) and (@{$children}>0)) {

			# $br is an internal node. #

	    ##unless ((defined $children) and (@{$children}>0)) {
	    ##
	    ##		# $br is an external node.
	    ##
	    ##	if (@cmpos_blks_lower>0) {
	    ##	    $br2cmpos_blks_lower{$br} = \@cmpos_blks_lower;
	    ##
=SET ASIDE just in case. 
			# ADDED for fixing a BUG (that ignores complementary pairs along external branches) (on 2019/03/19). #

		    my $blocks_upper = $up_down2blocks->{'U'};
		    unless ((defined $blocks_upper) and (@{$blocks_upper})) { next; }

		    foreach my $cblk_lw (@cmpos_blks_lower) { # Additional middle foreach-loop (over blocks on the lower-side of $br).
			foreach my $blk_up (@{$blocks_upper}) { # Inner foreach-loop (over blocks on the upper-side of $br).
			    my @ccblock = (@{$cblk_lw}, $blk_up);
			    my @std_ccblock = sort {$gblock_sizes->[$a] <=> $gblock_sizes->[$b]} @ccblock;
			    push @coll_cmpl_gblocks, \@std_ccblock;
			    ##push @coll_cmpl_gblocks, \@ccblock;

			}  # End of the inner foreach-loop (over blocks on the upper-side of $br).

		    } # End of the additional middle foreach-loop (over blocks on the lower-side of $br).

			# END of "ADDED for fixing a BUG (that ignores complementary pairs along external branches) (on 2019/03/19)." #
=cut # END of "SET ASIDE just in case."

	    ##	}
	    ##	next;
	    ##}

		# Then, examine the (composite-)gap-blocks on the lower-side of the children of $br. #

	    my @sets_cblocks_lch = ();
	    foreach my $ch (@{$children}) { # 1st middle foreach-loop (over children).
		my $cblks_lower_ch = $br2cmpos_blks_lower{$ch};
		unless ((defined $cblks_lower_ch) and (@{$cblks_lower_ch}>0)) { last; }
		push @sets_cblocks_lch, $cblks_lower_ch;
	    } # End of the 1st middle foreach-loop (over children).

	    my $ct_children = @{$children};


		# A BUG fixed on 2019/03/19. (A)
	   if (@sets_cblocks_lch == $ct_children) {
	    ##if (@sets_cblocks_lch < $ct_children) {
	    ##
	    ##	# There can be NO composite gap-blocks on the lower-side of $br.
	    ##
	    ##	if (@cmpos_blks_lower>0) {
	    ##	    $br2cmpos_blks_lower{$br} = \@cmpos_blks_lower;
	    ##	}
	    ##	next;
	    ##}

	    my @new_cblocks = ([]);
	    foreach my $set_cblks_lch (@sets_cblocks_lch) { # 2nd middle foreach-loop (over children).

		my @newnew_cblocks = ();

		foreach my $new_cblk (@new_cblocks) { # Inner foreach-loop (over incomplete composite-blocks).
		    foreach my $cblk_lch (@{$set_cblks_lch}) { # Inner-most foreach-loop (over composite-blocks on the lower-side of the child).
			my @newnew_cblk = (@{$new_cblk}, @{$cblk_lch});
			push @newnew_cblocks, \@newnew_cblk; 
		    } # End of the inner-most foreach-loop (over composite-blocks on the lower-side of the child).
		} # End of the inner foreach-loop (over incomplete composite-blocks).

		@new_cblocks = @newnew_cblocks; # Update @new_cblocks.

	    } # End of the 2nd middle foreach-loop (over children).

	    if (@new_cblocks>0) {
		push @cmpos_blks_lower, @new_cblocks;
	    }

	   } # END of "if (@sets_cblocks_lch == $ct_children) {...}"
		## END of "A BUG fixed on 2019/03/19. (A)" ##

	 } # END of "if ((defined $children) and (@{$children}>0)) {...}"
		# END of "A BUG fixed on 2019/03/19. (B)" ##

	    unless (@cmpos_blks_lower>0) { next; }

	    $br2cmpos_blks_lower{$br} = \@cmpos_blks_lower;


		# Finally, attempt to pair each gap-block on the upper-side of $br
		# with each (composite-)gap-block on the lower-side of $br.

	    unless (defined $up_down2blocks) { next; }
	       # $up_down2blocks = $br2up_down2blocks->{$br};

	    my $blocks_upper  = $up_down2blocks->{'U'};
	    unless (defined $blocks_upper) { next; }
		
	    foreach my $cblock_lower (@cmpos_blks_lower) { # 3rd middle foreach-loop (over composite-blocks on the lower-side of $br).
		foreach my $block_upper (@{$blocks_upper}) { # Inner foreach-loop (over gap-blocks on the upper-side of $br).

		    my @ccblock = (@{$cblock_lower}, $block_upper);

		    my ($size1st, $size2nd, $size3rd) = three_smallest_sizes (@{$gblock_sizes}, @ccblock);

		    if ((defined $size3rd) and ($size3rd < $size1st + $size2nd)) { next; } # Such a set of "composite-complementary blocks" can NEVER yield degenerate configurations.

		    my @std_ccblock = sort {$gblock_sizes->[$a] <=> $gblock_sizes->[$b]} @ccblock;
		    push @coll_cmpl_gblocks, \@std_ccblock;

		    ## push @coll_cmpl_gblocks, \@ccblock; # If this is better than the above, replace it.

		} # End of the inner foreach-loop (over gap-blocks on the upper-side of $br).
	    } # End of the 3rd middle foreach-loop (over composite-blocks on the lower-side of $br).

	} # End of the 1st outer foreach-loop (over $br (including $topnode_id)). #


		# (ii) Final examination at the root. #

	my $cmpos_blks_lower_r = $br2cmpos_blks_lower{$topnode_id};

	if ((defined $cmpos_blks_lower_r) and (@{$cmpos_blks_lower_r}>0)) { # FURTHER complementary composite blocks MAY be constructed.

	    foreach my $ccblock (@{$cmpos_blks_lower_r}) { # 2nd outer foreach-loop (over composite blocks on the "lower"-side of the root).

		my ($size1st, $size2nd, $size3rd) = three_smallest_sizes (@{$gblock_sizes}, @{$ccblock});

		if ((defined $size3rd) and ($size3rd < $size1st + $size2nd)) { next; } # Such a set of "composite-complementary blocks" can NEVER yield degenerate configurations.

		my @std_ccblock = sort {$gblock_sizes->[$a] <=> $gblock_sizes->[$b]} @{$ccblock};
		push @coll_cmpl_gblocks, \@std_ccblock;

		## push @coll_cmpl_gblocks, $ccblock; # If this is better than the above, replace it.

	    } # End of the 2nd outer foreach-loop (over composite blocks on the "lower"-side of the root).

	} # END of "if ((defined $cmpos_blks_lower_r) and (@{$cmpos_blks_lower_r}>0)) {...}".

		# (iii) ADDITIONAL examination at the root (ADDED for CORRECTION on 2019/03/17). #

	my $children_root = $node2ch->{$topnode_id};
	if ((defined $children_root) and (scalar (@{$children_root}) == 2)) {

	    my @sets_blks_uchr = ();
	    foreach my $ch_r (@{$children_root}) {
		my $up_down2blocks_ch_r = $br2up_down2blocks->{$ch_r};
		unless (defined $up_down2blocks_ch_r) { last; }
		my $blocks_upper_ch_r = $up_down2blocks_ch_r->{'U'};
		unless (defined $blocks_upper_ch_r) { last; }
		push @sets_blks_uchr, $blocks_upper_ch_r;
	    }
	    if (scalar (@sets_blks_uchr) == 2) {

		my @new_cblocks = ([]);
		foreach my $blks_uchr (@sets_blks_uchr) { # 3rd outer foreach-loop (over children of the root).

		    my @newnew_cblocks = ();

		    foreach my $new_cblk (@new_cblocks) { # Middle foreach-loop (over incomplete composite-blocks).
			foreach my $blk_uchr (@{$blks_uchr}) { # Inner foreach-loop (over blocks on the upper-side of the child).
			    my @newnew_cblk = (@{$new_cblk}, $blk_uchr);

			    push @newnew_cblocks, \@newnew_cblk; 

			} # End of the inner foreach-loop (over blocks on the upper-side of the child).
		    } # End of the middle foreach-loop (over incomplete composite-blocks).

		    @new_cblocks = @newnew_cblocks; # Update @new_cblocks.

		} # End of the 3rd outer foreach-loop (over children of the root).

		foreach my $set_ccblks (@new_cblocks) { # 5th outer foreach-loop (over new sets of compposite-complementary blocks).
		    my ($size1st, $size2nd, $size3rd) = three_smallest_sizes (@{$gblock_sizes}, @{$set_ccblks});

		    if ((defined $size3rd) and ($size3rd < $size1st + $size2nd)) { next; } # Such a set of "composite-complementary blocks" can NEVER yield degenerate configurations.

			# MODIFIED on 2019/03/19. #
		    my @std_set_ccblks = sort {$gblock_sizes->[$a] <=> $gblock_sizes->[$b]} @{$set_ccblks};
		    push @coll_cmpl_gblocks, \@std_set_ccblks;
		    ##push @coll_cmpl_gblocks, $set_ccblks;

		} # End of the 5th outer foreach-loop (over new sets of composite-complementary blocks).

	    } # End of "if (scalar (@sets_cblks_uchr) == 2) {...}".
	} # END of "if ((defined $children_root) and (scalar (@{$children_root}) == 2)) {...}".

		# END of "(iii) ADDITIONAL examination at the root (ADDED for CORRECTION on 2019/03/17)." #


	return \@coll_cmpl_gblocks;

}	# END of the subroutine 'mk_list_coll_cmpl_gblocks'. #

# mk_gblock2coll_cmpl:
#	This subroutine creates @gblock2coll_cmpl, which associates each gap-block with the sets of collectively complementary gap-blocks it belongs to,
#	out of @coll_cmpl_gblocks, the output of 'mk_list_coll_cmpl_gblocks' above,
#	which lists the sets of collectively complementary gap-blocks.
#	[CREATED on 2019/03/14.]
#
# INPUTS: ($ct_gblks, \@coll_cmpl_gblocks), where
#
#	$ct_gblks is the number of gap-blocks (within the window you deal with);
#
#	@{$coll_cmpl_gblocks[$cc]} stores the indices of the gap-blocks constituting the $cc th collectively complementary gap-block;
#	[Typically, \@coll_cmpl_gblocks is an output of ''mk_list_coll_cmpl_gblocks' agbove];
#	
# OUTPUTS: (\@gblock2coll_cmpl), where
#
#	@{$gblock2coll_cmpl[$gb]} lists the sets of
#	(the indices in @coll_cmpl_gblocks of) collectively complementary gap-blocks that the $gb th gap-block belongs to;
#	It is empty if there is no such set.
#
sub mk_gblock2coll_cmpl ($\@) {

	my ($ct_gblks, $coll_cmpl_gblocks) = @_;

		# Initialize the output. #
	my @gblock2coll_cmpl = ();
	for (my $gb=0; $gb < $ct_gblks; $gb++) { $gblock2coll_cmpl[$gb] = []; }

		# Fill in the output. #

	my $ct_ccgbs = @{$coll_cmpl_gblocks};
	for (my $i = 0; $i < $ct_ccgbs; $i++) { # Main for-loop (over sets of collective complementary gap-blocks).

	    my $set_ccgbs = $coll_cmpl_gblocks->[$i];

	    foreach my $gb (@{$set_ccgbs}) { # Inner foreach-loop (over gap-blocks in the set).
		push @{$gblock2coll_cmpl[$gb]}, $i;
	    } # End of the inner foreach-loop (over gap-blocks in the set).

	} # End of the main for-loop  (over sets of collective complementary gap-blocks).

	return \@gblock2coll_cmpl;

}	# END of the subroutine 'mk_gblock2coll_cmpl'. #


# search_for_coll_cmpl_gblks:
#	This subroutine (exhaustively) searches for
#	the sets of collectively complementary gap-blocks within a window.
#	Actually, the subroutine merely performs 3 subroutines,
#	'mk_br2up_down2blocks', 'mk_list_coll_cmpl_gblocks', and 'mk_gblock2coll_cmpl',
#	one after another.
#	[CREATED on 2019/03/14.]
#
# INPUTS: (\@list_gblocks, $indx_br, $indx_u_or_d, $topnode_id, \%node2ch, \%node2depth, \@gblock_sizes), where
#
#	($branch, $up_or_down) = @{$list_gblocks[$gb]}[$indx_br, $indx_u_or_d]
#	gives information on the $gb th gap-block, where
#	$branch is the ID of the branch separating the block,
#	and $up_or_down = 'U'/'L' if the block is on the "upper"/"lower"-side of $branch;
#	[Typically, @list_gblocks =  @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	$topnode_id is the ID of the top-node, i.e., root;
#
#	%node2ch = ($node_id => \@children, ...) for all internal nodes;
#
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[Typically, \%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is an output of, e.g., 'mk_list_gblock_sizes' above];
#
# OUTPUTS: (\%br2up_down2blocks, \@coll_cmpl_gblocks, \@gblock2coll_cmpl), where
#
#	\%br2up_down2blocks is the main output of 'mk_br2up_down2blocks' above;
#	[See 'mk_br2up_down2blocks' for details];
#
#	\@coll_cmpl_gblocks is the main output of 'mk_list_coll_cmpl_gblocks' above;
#	[See 'mk_list_coll_cmpl_gblocks' for details];
#	
#	\@gblock2coll_cmpl is the main output of 'mk_gblock2coll_cmpl' above;
#	[See 'mk_gblock2coll_cmpl' for details];
#	
sub search_for_coll_cmpl_gblks (\@$$$\%\%\@) {

	my ($list_gblocks, $indx_br, $indx_u_or_d, $topnode_id, $node2ch, $node2depth, $gblock_sizes) = @_;

	my $ct_gblks = @{$list_gblocks};

    (print STDOUT "<< Subroutine 'search_for_coll_cmpl_gblks' (#{gap-blocks} = $ct_gblks) >>\n\n") if ($TEST_LEVEL_sch4ccgb > 0);


	my $br2up_down2blocks = mk_br2up_down2blocks (@{$list_gblocks}, $indx_br, $indx_u_or_d);
	unless (defined $br2up_down2blocks) {
	    print STDERR "[ERROR 'search_for_coll_cmpl_gblks'] Subroutine 'mk_br2up_down2blocks' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_br2up_down2blocks' Ended ...\n\n") if ($TEST_LEVEL_sch4ccgb > 1);


    if ($TEST_LEVEL_sch4ccgb >2) {

	print STDOUT "\n", '[ Contents of %{$br2up_down2blocks} ]', "\n\n";

	print STDOUT join ("\t", 'Br', 'up(U)_or_down(L)', 'relevant,gap,blocks'), "\n\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocks}) {

	    my $up_down2blocks = $br2up_down2blocks->{$br};

	    foreach my $up_down (sort keys %{$up_down2blocks}) {
		my $cnct_blocks = join (',', @{$up_down2blocks->{$up_down}});
		print STDOUT join ("\t", $br, $up_down, $cnct_blocks), "\n";
	    }
	}
	print STDOUT "\n\n";

#	print STDOUT join ("\t", 'Gb', 'Intfr_gb(rel)_1', 'Intfr_gb(rel)_2', '...'), "\n\n";
#
#	for (my $gb =0; $gb < $ct_gblks; $gb++) {
#	    my ($blocks, $rels) = @{$interfering_blocks->[$gb]};
#	    print STDOUT join ("\t", $gb, @out), "\n";
#	}
#	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_sch4ccgb >2) {...}".


	my $coll_cmpl_gblocks = mk_list_coll_cmpl_gblocks (%{$br2up_down2blocks}, $topnode_id, %{$node2ch}, %{$node2depth}, @{$gblock_sizes});
	unless (defined $coll_cmpl_gblocks) {
	    print STDERR "[ERROR 'search_for_coll_cmpl_gblks'] Subroutine 'mk_list_coll_cmpl_gblocks' FAILED: $! !\n\n";
	    return ($br2up_down2blocks, undef, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_list_coll_cmpl_gblocks' Ended ...\n\n") if ($TEST_LEVEL_sch4ccgb > 1);

    if ($TEST_LEVEL_sch4ccgb >2) {

	my $ct_ccgbs = @{$coll_cmpl_gblocks};
	print STDOUT "\n", '[ Contents of @{$coll_cmpl_gblocks} ', "(#{sets} = $ct_ccgbs) ]\n\n";
	print STDOUT join ("\t", 'Indx_set', '[gap,blocks,in,the,set]'), "\n\n";

	for (my $i = 0; $i < $ct_ccgbs; $i++) { 
	    my $cnct_gblks_in_set = join (',', @{$coll_cmpl_gblocks->[$i]});
	    $cnct_gblks_in_set = "[$cnct_gblks_in_set]";
	    print STDOUT join ("\t", $i, $cnct_gblks_in_set), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_sch4ccgb >2) {...}".


	my $gblock2coll_cmpl = mk_gblock2coll_cmpl ($ct_gblks, @{$coll_cmpl_gblocks});
	unless (defined $gblock2coll_cmpl) {
	    print STDERR "[ERROR 'search_for_coll_cmpl_gblks'] Subroutine 'mk_gblock2coll_cmpl' FAILED: $! !\n\n";
	    return ($br2up_down2blocks, $coll_cmpl_gblocks, undef);
	}
    (print STDOUT "\n ... Subroutine 'mk_gblock2coll_cmpl' Ended ...\n\n") if ($TEST_LEVEL_sch4ccgb > 1);

    if ($TEST_LEVEL_sch4ccgb >2) {

	print STDOUT "\n", '[ Contents of @{$gblock2coll_cmpl} ]', "\n\n";
	print STDOUT join ("\t", 'Indx_gb', '(indices,sets,containing,the,gb)'), "\n\n";

	for (my $gb = 0; $gb < $ct_gblks; $gb++) {

	    my $cnct_indices_sets = join (',', @{$gblock2coll_cmpl->[$gb]});
	    $cnct_indices_sets = "($cnct_indices_sets)";

	    print STDOUT join ("\t", $gb, $cnct_indices_sets), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_sch4ccgb >2) {...}".



    (print STDOUT "\n<< Subroutine 'search_for_coll_cmpl_gblks' ENDED />/>\n\n\n") if ($TEST_LEVEL_sch4ccgb > 0);

	return ($br2up_down2blocks, $coll_cmpl_gblocks, $gblock2coll_cmpl);

}	# END of the subroutine 'search_for_coll_cmpl_gblks'. #


# cmp_nmarrays:
#	This satellite subroutine of 'compt_aln_degeneracy' compares
#	two simple arrays, whose elements are simple scalars representing numbers.
#	[CREATED on 2019/04/02.]
#
#	NOTE: ASSUME that the two arrays are already sorted in the same way.
#
# INPUTS: (\@array1, \@array2), where
#
#	@array1 and @array2 are two simple arrays to be compared.
#
# OUTPUTS: $sgn_cmp, where
#
#	$sgn_cmp = 0 if the two arrays are identical,
#		 = -1 if @array1 is "smaller than" @array2,
#		 = +1 if @array1 is "larger than" @array2.
#
sub cmp_nmarrays (\@\@) {

	my ($array1, $array2) = @_;

	my $size1 = @{$array1};
	my $size2 = @{$array2};

	if ($size1 < $size2) {
	    return -1;
	} elsif ($size1 > $size2) {
	    return +1;
	}

		# if ($size1 == $size2)

	for (my $k=0; $k < $size1; $k++) { # Main for-loop (over the array elements).
	    my $ele1 = $array1->[$k];
	    my $ele2 = $array2->[$k];

	    if ($ele1 < $ele2) {
		return -1;
	    } elsif ($ele1 > $ele2) {
		return +1;
	    }
	} # End of the main for-loop (over the array elements).

	return 0;

}	# End of the subroutine 'cmp_nmarrays'. #


# compt_aln_degeneracy:
#	This subroutine computes the degree of degeneracy of the alignment.
#	[CREATION STARTED on 2019/04/01, ENDED on 2019/04/03.]
#
#	[MAJOR REVISION on 2019/04/04:]
#	Now, the subroutine can handle cases where a pair of collectively complementary block-sets 
#	is intervened by one or more blocks vertically including one of the pair.
#
#	[MAJOR REVISION on 2019/04/05.]
#	Now, if $SOPHIST_MVBLTY_TEST_FOR_CAD = 1,
#	the subroutine employs a sophisticated version of the test for the movability of the left- and right- complex blocks.
#
#	[MAJOR REVISION on 2019/04/14.]
#	Now, the subroutine also outputs a string formed by concatenating the list of sets of collectively complementary gap-blocks causing the degeneracy.
#
#	[MAJOR REVISION on 2019/05/27.]
#	We replaced the subroutine component, 'ct_necessary_shifts',
#	with 'smlt2_extend_left_end_to_left' or 'smlt2_extend_right_end_to_right',
#	whenever possible.
#
#	[MAJOR REVISION on 2019/06/05]:
#	Now, the subroutine returns $degeneracy = 0
#	whenever it encounters a chain of mutually adjacent collectively complementary gap-blocks
#	even if its components do NOT cause degeneracy (as long as they effectively touch one another).
#
# [ INPUTS: (\@coll_cmpl_gblocks, \@bds_gblocks, \@bl_coords, \@bds_bl_coord), where ] ... OBSOLETE as of 2019/04/04.
# INPUTS: (\@coll_cmpl_gblocks, \@bds_gblocks, \@bl_coords, \@bds_bl_coord, \@inter_block_relations, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	@{$coll_cmpl_gblocks[$cc]} stores the indices of the gap-blocks constituting the $cc th collectively complementary gap-block (in ascending order of the block-size);
#	[Typically, \@coll_cmpl_gblocks is an output of 'mk_list_coll_cmpl_gblocks' above, or of 'search_for_coll_cmpl_gblks' above];
#
#	@{$bds_gblocks[$gb]} = ($lbd, $rbd) give the left- and right-boundaries of the $gb th gap-block;
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$bl_coords[$gb] gives the current coordinate of the $gb th gap-block;
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coord[$gb]} give the boundaries of the coordinate of the $gb th gap-block.
#
#		# ADDED on 2019/04/04. #
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#		# END of "ADDED on 2019/04/04." #
#	
# [ OUTPUTS:  ($degeneracy), where ] ... OBSOLETE as of 2019/04/14.
# OUTPUTS:  ($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy), where
#
#	$degeneracy is the degree of degeneracy, i.e., how many alignments equivalent to the subject one are in the coordinate space currently used.
#	Especially,
#	$degeneracy = 1 either if no collectively complementary gap-blocks exist or if some collectively complementary gap-blocks exist but none of them causes degeneracy;
#	$degeneracy = 0 if the alignment SHOULD be ignored within this coordinate space(, and thus the subsequent processes SHOULD be SKIPPED).
#
#	$cnct_sets_ccgbs_cs_dgnrcy is a string formed by concatenating
#	"($cnct_left)($cnct_right)($cnct_including)"
#	where $cnct_left = join (',', @left), $cnct_right = join (',', @right), and $cnct_including = join (',', @including), from each set of collectively complementary gap-blocks.
#	Here, @left, @right, and @including are the sets of gap-blocks on the left, on the right, and including both left and right blocks, respectively;
#	$cnct_sets_ccgbs_cs_dgnrcy = 'NONE' if there are no such sets of collectively complementary gap-blocks.
#
# [ sub compt_aln_degeneracy (\@\@\@\@) {
#	my ($coll_cmpl_gblocks, $bds_gblocks, $bl_coords, $bds_bl_coord) = @_; ] ... OBSOLETE as of 2019/04/04.

sub compt_aln_degeneracy (\@\@\@\@\@\@\@) {

	my ($coll_cmpl_gblocks, $bds_gblocks, $bl_coords, $bds_bl_coord, $inter_block_relations, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my $ct_ccgbs = @{$coll_cmpl_gblocks};

    (print STDOUT "<< Subroutine 'compt_aln_degeneracy' (#{sets of collectively complementary gap-blocks} = $ct_ccgbs) >>\n\n") if ($TEST_LEVEL_cad >0);

	unless ($ct_ccgbs >0) {
		# NO collectively complementary gap-blocks, which can potentially cause degeneracy;
		# in this case, the "degree of degeneracy" = 1.

    (print STDOUT "\n ...  There are NO Collectively Complementary Gap-Blocks ...\n\n") if ($TEST_LEVEL_cad >1);

    (print STDOUT "\n<< Subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy = 1) />/>\n\n") if ($TEST_LEVEL_cad >0);

	    return (1, 'NONE'); # REVISED on 2019/04/14.
	    ##return 1;
	}


    (print STDOUT "{ Main Process (1): Find the sets of collectively complementary gap-blocks that causes degeneracy of the subject alignment. }\n\n") if ($TEST_LEVEL_cad > 1);

	my @degenerate_sets_ccbs = (); # element = [$lbd_left, $rbd_left, $lbd_right, $rbd_right, \@left, \@right, \@including], where @left/right is the set of blocks on the left/right, @including is the set of blocks (horizontally) including both.
	my @cnct_deg_sets_ccbs = (); # element = "($cnct_left)($cnct_right)($cnct_including)", where $cnct_left = join (',', @left), $cnct_right = join (',', @right), and $cnct_including = join (',', @including). Each element has a counterpart element in @degenerate_sets_ccbs. # ADDED on 2019/04/14.
		# REVISED on 2019/06/05: Now, @degenerate_sets_ccbs, and @cnct_deg_sets_ccbs contains the sets EVEN IF they do not satisfy the boundary conditions.

		# ADDED in MAJOR REVISION (on 2019/06/05). (a) #
	my @indices_really_degenerate = (); # Stores indices of the elements of @degenerate_sets_ccbs that also satisfy the boundary conditions.
	my $indx = 0;
		# End of "ADDED in MAJOR REVISION (on 2019/06/05). (a)" #

	foreach my $set_cc_gblks (@{$coll_cmpl_gblocks}) { # 1st MAIN outer foreach-loop (over the sets of collectively complementary gap-blocks).

	    my $gb1 = $set_cc_gblks->[0];
	    my ($lbd1, $rbd1) = @{$bds_gblocks->[$gb1]};

	    my $ct_ccbs = @{$set_cc_gblks}; 

	    my @identical = my @including = my @left_flanking = my @right_flanking = my @overlapping = my @separated = ();

		# ADDED in major revision (on 2019/04/04). #
	    my %skipped;
	    foreach my $gb2 (@{$set_cc_gblks}) { $skipped{$gb2} = 1; }
		# END of "ADDED in major revision (on 2019/04/04)." #

	    for (my $k = 1; $k < $ct_ccbs; $k++) { # 1st inner for-loop (over constituent gap-blocks in the set).

		my $gb2 = $set_cc_gblks->[$k];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};

			# MODIFIED in major revision (on 2019/04/04). #

		if ($rbd2 + 1 < $lbd1) {
		    if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd1-1]}>0)) {

				# MODIFIED in major revision (on 2019/05/27). (1) #

			my ($if_reached, $lbd2_new, $rbd2_new) = smlt2_extend_right_end_to_right ($gb2, $lbd1-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			unless (defined $if_reached) {
			    print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb2 th gap-block, which is on the left of $gb1: $! !\n\n";
			    return (undef, undef);
			}

			if ($if_reached) { # $gb2 is virtually left-flanking $gb1.
			    push @left_flanking, $gb2;
			    next;
			}

=OBSOLETE as of 2019/05/27. (1)
			my $ct_shifts = ct_necessary_shifts ($rbd2, $lbd1 -1, $gb2, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			unless (defined $ct_shifts) {
			    print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block, which is on the left of $gb1: $! !\n\n";
			    return (undef, undef); # REVISED on 2019/04/14.
			    ##return undef;
			}

			if ($ct_shifts == 0) { # $gb2 is virtually left-flanking $gb1.
			    push @left_flanking, $gb2;
			    next;
			} 
=cut # End of "OBSOLETE as of 2019/05/27. (1)" #

				# End of "MODIFIED in major revision (on 2019/05/27). (1)" #
		    }

		    push @separated, $gb2;
		    last;

		} elsif ($rbd1 + 1 < $lbd2) {
		    if ((@{$gblks_w_spec_lb->[$rbd1+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {

				# MODIFIED in major revision (on 2019/05/27). (2) #

			my ($if_reached, $lbd2_new, $rbd2_new) = smlt2_extend_left_end_to_left ($gb2, $rbd1+1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});

			unless (defined $if_reached) {
			    print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb2 th gap-block, which is on the right of $gb1: $! !\n\n";
			    return (undef, undef);
			}

			if ($if_reached) { # $gb2 is virtually right-flanking $gb1.
			    push @right_flanking, $gb2;
			    next;
			} 

=OBSOLETE as of 2019/05/27. (2)
			my $ct_shifts = ct_necessary_shifts ($lbd2, $rbd1 +1, $gb2, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			unless (defined $ct_shifts) {
			    print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block, which is on the right of $gb1: $! !\n\n";
			    return (undef, undef); # REVISED on 2019/04/14.
			    ##return undef;
			}

			if ($ct_shifts == 0) { # $gb2 is virtually right-flanking $gb1.
			    push @right_flanking, $gb2;
			    next;
			} 
=cut # End of "OBSOLETE as of 2019/05/27. (2)" #

				# End of "MODIFIED in major revision (on 2019/05/27). (2)" #

		    }

		    push @separated, $gb2;
		    last;

		##if (($rbd2 + 1 < $lbd1) or ($rbd1 + 1 < $lbd2)) {
		##
		##    push @separated, $gb2;
		##    last;

			# END of "MODIFIED in major revision (on 2019/04/04)." #

		} elsif  ($rbd1 + 1 == $lbd2) {
		    push @right_flanking, $gb2;
		    next;

		} elsif ($rbd2 + 1 == $lbd1) {
		    push @left_flanking, $gb2;
		    next;

		} elsif ($lbd1 < $lbd2) {
		    push @overlapping, $gb2;
		    last;

		} elsif ($lbd2 < $lbd1) {
		    if ($rbd2 < $rbd1) {
			push @overlapping, $gb2;
			last;

		    } else { # if ($rbd1 <= $rbd2)
			push @including, $gb2;
			next;
		    }
		} else { # if ($lbd1 == $lbd2)
		    if ($rbd1 < $rbd2) {
			push @including, $gb2;
			next;

		    } elsif ($rbd2 < $rbd1) {
			push @overlapping, $gb2; # This must NOT happen.
			last;

		    } else { # if ($rbd1 == $rbd2)
			push @identical, $gb2;
			next;
		    }
		}
	    } # End of the 1st inner for-loop (over constituent gap-blocks in the set).
    
	    if ((@separated > 0) or (@overlapping > 0)) { next;
	    } elsif ((@left_flanking == 0) and (@right_flanking == 0)) { next;
	    } elsif ((@left_flanking > 0) and (@right_flanking > 0)) { next;
	    }

	    my $ref_flanking = (@left_flanking>0) ? \@left_flanking : \@right_flanking;
	    my ($lbd_fl, $rbd_fl) = @{$bds_gblocks->[$ref_flanking->[0]]};
	    my $ct_flanking = @{$ref_flanking};
	    my $if_staggered = 0;
	    for (my $i=0; $i < $ct_flanking; $i++) { # 2nd inner for-loop (over flanking gap-blocks).

		my ($lbd3, $rbd3) = @{$bds_gblocks->[$ref_flanking->[$i]]};
		if (($lbd3 != $lbd_fl) or ($rbd3 != $rbd_fl)) {
		    $if_staggered = 1;
		    last;
		}
	    } # End of the 2nd inner for-loop (over flanking gap-blocks).

	    if ($if_staggered) { next; } # Usually, the set is degenerated ONLY IF the flanking blocks have identical boundaries.

	    my $lbd_union = ($lbd1 < $lbd_fl) ? $lbd1 : $lbd_fl;
	    my $rbd_union = ($rbd1 > $rbd_fl) ? $rbd1 : $rbd_fl;

	    if (@including > 0) {
		my $if_accommodating = 1;
		foreach my $gb3 (@including) { # 3rd inner foreach-loop.
		    my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
		    if (($lbd_union < $lbd3) or ($rbd3 < $rbd_union)) {
			$if_accommodating = 0;
			last;
		    }
		} # End of the 3rd inner foreach-loop.

		unless ($if_accommodating) { next; } # The set is degenerated ONLY IF ALL blocks accommodating the subject also accommodate the flanking blocks as well.
	    }

	    my @std_identical = sort {$a <=> $b} ($gb1, @identical);
	    my @std_flanking = sort {$a <=> $b} @{$ref_flanking};
	    my @std_including = sort {$a <=> $b} @including;

	    my ($lbd_left, $rbd_left, $lbd_right, $rbd_right, $std_left, $std_right, $std_including) =
	    my @degenerate_set = (@left_flanking > 0) ?
		($lbd_fl, $rbd_fl, $lbd1, $rbd1, \@std_flanking, \@std_identical, \@std_including) :
		($lbd1, $rbd1, $lbd_fl, $rbd_fl, \@std_identical, \@std_flanking, \@std_including) ;

		## Finally, examine whether the collective blocks are really swappable or not. ##

	    my $if_movable = 1;
	    foreach my $gb5 (@{$std_left}) { # 5th inner foreach-loop (over blocks belonging to the left collective block).
		# Check if the block is movable to the position of the right collective block.
		my $coord_gb5 = $bl_coords->[$gb5];
		my $rbd_coord5 = $bds_bl_coord->[$gb5][1];

			## MODIFIED in MAJOR REVISION on 2019/04/05. (L) ##

		if ($SOPHIST_MVBLTY_TEST_FOR_CAD) {

		    my $ct_shifts = ct_necessary_shifts ($rbd_left, $rbd_right, $gb5, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
		    unless (defined $ct_shifts) {
			print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'ct_necessary_shifts' FAILED on the $gb5 th gap-block, which belongs to the left complex block: $! !\n\n";
			return (undef, undef); # REVISED on 2019/04/14.
			##return undef;
		    }

		    if ($rbd_coord5 - $coord_gb5 < $ct_shifts) {
			$if_movable = 0;
			last;
		    }

		} else {

		    if ($rbd_coord5 - $coord_gb5 < $rbd_right - $rbd_left) {
			$if_movable = 0;
			last;
		    }
		}
		##if ($rbd_coord5 - $coord_gb5 < $rbd_right - $rbd_left) {
		##    $if_movable = 0;
		##    last;
		##	# NOTE: although the test could be made more rigorous using, e.g., 'ct_necessary_shifts' below, we employed a simpler test here, because the latter should be much faster, and because making the test perfect requires considering this issue more throughly anyway.
		##}
			## END of "MODIFIED in MAJOR REVISION on 2019/04/05. (L)" ##

	    } # End of the 5th inner foreach-loop (over blocks belonging to the left collective block).

	    if ($if_movable) {
		foreach my $gb5 (@{$std_right}) { # 6th inner foreach-loop (over blocks belonging to the right collective block).
			# Check if the block is movable to the position of the left collective block.
		    my $coord_gb5 = $bl_coords->[$gb5];
		    my $lbd_coord5 = $bds_bl_coord->[$gb5][0];

			## MODIFIED in MAJOR REVISION on 2019/04/05. (R) ##

		    if ($SOPHIST_MVBLTY_TEST_FOR_CAD) {

			my $ct_shifts = ct_necessary_shifts ($lbd_right, $lbd_left, $gb5, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			unless (defined $ct_shifts) {
			    print STDERR "[ERROR 'compt_aln_degeneracy'] Subroutine 'ct_necessary_shifts' FAILED on the $gb5 th gap-block, which belongs to the left complex block: $! !\n\n";
			    return (undef, undef); # REVISED on 2019/04/14.
			    ##return undef;
			}

		        if ($coord_gb5 - $lbd_coord5 < $ct_shifts) {
			    $if_movable = 0;
			    last;
		        }

		    } else {

			if ($coord_gb5 - $lbd_coord5 < $lbd_right - $lbd_left) {
			    $if_movable = 0;
			    last;
			}
		    }
		    ##if ($coord_gb5 - $lbd_coord5 < $lbd_right - $lbd_left) {
		    ##	$if_movable = 0;
		    ##	last;
		    ##}
			## END of "MODIFIED in MAJOR REVISION on 2019/04/05. (R)" ##

		} # End of the 6th inner foreach-loop (over blocks belonging to the right collective block).
	    }

		# MODIFIED in MAJOR REVISION (on 2019/06/05). (b) #
	    if ($if_movable) { push @indices_really_degenerate, $indx; }
	    $indx++;
	    ##unless ($if_movable) { next; }
		# END of "MODIFIED in MAJOR REVISION (on 2019/06/05). (b)" #

	    push @degenerate_sets_ccbs, \@degenerate_set;


			# ADDED in MAJOR REVISION (on 2019/04/14).
	    my $cnct_left = join (',', @{$std_left});
	    my $cnct_right = join (',', @{$std_right});
	    my $cnct_including = join (',', @{$std_including});
	    push @cnct_deg_sets_ccbs, "($cnct_left)($cnct_right)($cnct_including)";
			# END of "ADDED in MAJOR REVISION (on 2019/04/14)."

	} # End of the 1st MAIN outer foreach-loop (over the sets of collectively complementary gap-blocks).

	my $ct_degsets = @degenerate_sets_ccbs;

    if ($TEST_LEVEL_cad > 3) {

	print STDOUT "\n\n", '[ Interim Report:  The content of @degenerate_sets_ccbs ', "(#{elements} = $ct_degsets): ]\n\n";
	print STDOUT join ("\t", 'Indx_set', 'lbd_left', 'rbd_left', 'lbd_right', 'rbd_right',
			'[gap,blocks,on,the,left]', '[gap,blocks,on,the,right]', '[gap,blocks,including,both,left,and,right]'), "\n\n"; 

	for (my $k = 0; $k <$ct_degsets; $k++) { 

	    my ($lbd_left, $rbd_left, $lbd_right, $rbd_right, $left, $right, $including) = @{$degenerate_sets_ccbs[$k]};
	    my $cnct_left = join (',', @{$left});
	    my $cnct_right = join (',', @{$right});
	    my $cnct_including = join (',', @{$including});

	    print STDOUT join ("\t", $k, $lbd_left, $rbd_left, $lbd_right, $rbd_right, "[$cnct_left]", "[$cnct_right]", "[$cnct_including]"), "\n";
	}
	print STDOUT "\n\n";

	## @degenerate_sets_ccbs = (); # element = [$lbd_left, $rbd_left, $lbd_right, $rbd_right, \@left, \@right, \@including], where @left/right is the set of blocks on the left/right, @including is the set of blocks (horizontally) including both.

		# Added in major revision (on 2019/06/05). (c) #
	print STDOUT '@indices_really_degenerate = (', join (', ', @indices_really_degenerate), ") .\n";
	print STDOUT "\n\n";
		# End of "Added in major revision (on 2019/06/05). (c)" #

    } # End of "if ($TEST_LEVEL_cad > 3) {...}"

	unless ($ct_degsets >0) {
		# NO collectively complementary gap-blocks cause degeneracy,
		# which means the "degree of degeneracy" = 1.

    (print STDOUT "\n ...  None of the $ct_ccgbs Collectively Complementary Gap-Blocks will cause degeneracy of the current alignment ...\n\n") if ($TEST_LEVEL_cad >1);

    (print STDOUT "\n<< Subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy = 1) />/>\n\n") if ($TEST_LEVEL_cad >0);

	    return (1, 'NONE'); # REVISED on 2019/04/14.
	    ##return 1;
	}

	my $cnct_sets_ccgbs_cs_dgnrcy = join (';', @cnct_deg_sets_ccbs); # ADDED in REVISION (on 2019/04/14).
	my $cnct_sets_ccgbs_cs_dgnrcy_true = (@indices_really_degenerate>0) ? join (';', @cnct_deg_sets_ccbs[@indices_really_degenerate]) : 'NONE'; # ADDED in MAJOR REVISION (on 2019/06/05).


		## Then, examine whether some degenerate sets can be chained or not. ##

    (print STDOUT "{ Main Process (2): Examine whether some of the $ct_degsets sets can be chained or not. }\n\n") if ($TEST_LEVEL_cad > 1);

	my @std_deg_sets = sort {$a->[0] <=> $b->[0]} @degenerate_sets_ccbs;
	@degenerate_sets_ccbs = @std_deg_sets;

	my @chains_deg_sets = ();
	my $ct_cmplx_chains = 0;

	while (@degenerate_sets_ccbs>0) { # 2nd MAIN outer while-loop.

	    my $chain_tail = shift @degenerate_sets_ccbs;
	    my @chain = ($chain_tail);
	    my @remainder = ();

	    while (my $subject = shift @degenerate_sets_ccbs) { # Inner while-loop.

		if ($chain_tail->[2] < $subject->[0]) {
		    @remainder = (@remainder, $subject, @degenerate_sets_ccbs);
		    last;
		} elsif ($subject->[0] < $chain_tail->[2]) {
		    push @remainder, $subject;
		    next;
		}

		   # if ($chain_tail->[2] == $subject->[0])

		if ($subject->[1] == $chain_tail->[3]) { # Now, the left-blockset of the subject is positionally identical with the right-blockset of the tail of the chain.

		    if (0 == cmp_nmarrays (@{$chain_tail->[5]}, @{$subject->[4]}) ) { # The aforementioned two blocksets are identical.
			if (0 == cmp_nmarrays (@{$chain_tail->[6]}, @{$subject->[6]}) ) { # The "including" blocksets in the two sets are also identical.
			    push @chain, $subject;
			    $chain_tail = $subject;
			    last;
			}

		    } # End of "if (0 == cmp_nmarrays (@{$chain_tail->[5]}, @{$subject->[4]}) ) {...}".

		} # End of "if ($subject->[1] == $chain_tail->[3]) {...}".

	# @degenerate_sets_ccbs = (); # element = [$lbd_left, $rbd_left, $lbd_right, $rbd_right, \@left, \@right, \@including], where @left/right is the set of blocks on the left/right, @including is the set of blocks (horizontally) including both.

		push @remainder, $subject;

	    } # End of the inner while-loop.

	    push @chains_deg_sets, \@chain;

	    if (@chain >1) {
		$ct_cmplx_chains++;
		last;
	    }

	    @degenerate_sets_ccbs = @remainder; # Update the sets of collectively complementary blocks each of which causes degeneracy.

	} # End of the 2nd MAIN outer while-loop.

	if ($ct_cmplx_chains>0) { # This case should actually be dealt with in a different coordinate space, where the @left in the left ccb and the @right in the right ccb are merged.

    if ($TEST_LEVEL_cad >1) {

	print STDOUT "\n\n ... The Following Sets were Found to be CHAINED!!:\n\n";

	print STDOUT join ("\t", 'Indx_in_chain', 'lbd_left', 'rbd_left', 'lbd_right', 'rbd_right',
			'[gap,blocks,on,the,left]', '[gap,blocks,on,the,right]', '[gap,blocks,including,both,left,and,right]'), "\n\n"; 

	my $chain = pop @chains_deg_sets;
	my $size_chain = @{$chain};

	for (my $k = 0; $k <$size_chain; $k++) {

	    my $deg_set = $chain->[$k];
	    my ($lbd_left, $rbd_left, $lbd_right, $rbd_right, $left, $right, $including) = @{$deg_set};
	    my $cnct_left = join (',', @{$left});
	    my $cnct_right = join (',', @{$right});
	    my $cnct_including = join (',', @{$including});

	    print STDOUT join ("\t", $k, $lbd_left, $rbd_left, $lbd_right, $rbd_right, "[$cnct_left]", "[$cnct_right]", "[$cnct_including]"), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_cad >1) {...}"

    (print STDOUT "\n<< Subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy = 0) />/>\n\n") if ($TEST_LEVEL_cad >0);

	    return (0, $cnct_sets_ccgbs_cs_dgnrcy); # REVISED on 2019/04/14.
	    ##return 0; # Degeneracy = 0 means INFENITE DEGENECARY, i.e., NO CONTRIBUTION to the alignment probability distribution here.
			# => SKIP the subsequent processes!!

	} # End of "if ($ct_cmplx_chains>0) {...}".

    (print STDOUT "\n\n ... None of the $ct_degsets Sets were Chained ...\n\n") if ($TEST_LEVEL_cad>1);


	my $degeneracy = 1;

		# MODIFIED in MAJOR REVISION (on 2019/06/05). (d) #
	for (1 .. scalar (@indices_really_degenerate) ) { $degeneracy *= 2; } # Each degeneracy set doubles the degeneracy.
	##for (1 .. scalar (@chains_deg_sets) ) { $degeneracy *= 2; } # Each degeneracy set doubles the degeneracy.
		# End of "MODIFIED in MAJOR REVISION (on 2019/06/05). (d)" #

    (print STDOUT "\n<< Subroutine 'compt_aln_degeneracy' Ended (w/ \$degeneracy = $degeneracy) />/>\n\n") if ($TEST_LEVEL_cad >0);

	return ($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy_true); # Further MODIFIED in MAJOR REVISION (on 2019/06/05). (e)
	##return ($degeneracy, $cnct_sets_ccgbs_cs_dgnrcy); # REVISED on 2019/04/14.
	####return $degeneracy;

}	# END of the subroutine 'compt_aln_degeneracy'. #


# ct_necessary_shifts:
#	This is the core satellite subroutine of
#	'compt_inter_gblock_distance' and 'compt_distance_bw_gblk_and_pt' below.
#	It counts the number of "shifts" necessary for moving a specified boundary of a subject gap-block
#	to (or beyond) a specified place (i.e., column).
#	[CREATED on 2019/03/20.]
#
#	NOTE: This subroutine ASSUMES that the input alignment contains NO null columns.
#	Thus, if the alignment contains some null columns, the output COULD differ from your expectation.
#
# INPUTS: ($bd_curr, $bd_goal, $gb_sbj, \@bds_gblocks, \@inter_block_relations, \%skipped), where
#
#	$bd_curr is the current position of the specified boundary of the subject gap-block;
#	$bd_goal is the position to (or beyond) which the subject boundary will be "shift"ed;
#	$gb_sbj is the index for the subject gap-block;
#
#	@{$bds_gblocks[$gb]} = ($lbd, $rbd) give the left- and right-boundaries of the $gb th gap-block;
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] is the relation of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above; See 'mk_inter_block_relations' for details];
#
#	%skipped = ($indx_gb => {some defined value}, ...) for the gap-blocks to be skipped in this analysis;
#
# OUTPUTS: $ct_shifts, which is the number of necessary "shift"s.
#
sub ct_necessary_shifts ($$$\@\@\%) {

	my ($bd_curr, $bd_goal, $gb_sbj, $bds_gblocks, $inter_block_relations, $skipped) = @_;

	if ($bd_curr == $bd_goal) { return 0; }

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_gblks = @{$rels_w_sbj};

	my ($lb_mv, $rb_mv) = ($bd_curr < $bd_goal) ? ($bd_curr +1, $bd_goal) : ($bd_goal, $bd_curr -1);

	my %clm2passable;

	for (my $gb3 = 0; $gb3 < $ct_gblks; $gb3++) { # Main for-loop (over $gb3).

	    if (($gb3 == $gb_sbj) or (defined $skipped->{$gb3})) { next; }

	    my $rel = $rels_w_sbj->[$gb3];
	    unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb3 < $gb_sbj))
			or ($rel eq 'ONN') or ($rel eq 'ONCS')) { next; }

	    my ($lb3, $rb3) = @{$bds_gblocks->[$gb3]};
	    if (($rb_mv < $lb3) or ($rb3 < $lb_mv)) { next; }

	    my $lb_intrsct = ($lb3 > $lb_mv) ? $lb3 : $lb_mv;
	    my $rb_intrsct = ($rb3 < $rb_mv) ? $rb3 : $rb_mv;

	    for (my $c = $lb_intrsct; $c <= $rb_intrsct; $c++) { # Inner for-loop (over columns within the intersection).

		my $passable = $clm2passable{$c};
		unless (defined $passable) { $passable = $clm2passable{$c} = []; }
		push @{$passable}, $gb3;

	    } # End of the inner for-loop (over columns within the intersection).

	} # End of the main for-loop (over $gb3).

	my $ct_shifts = $rb_mv - $lb_mv +1;

	for (my $c = $lb_mv; $c <= $rb_mv; $c++) { # Trailing for-loop (over columns within the region of the boundary-move).
	    if (defined $clm2passable{$c}) { $ct_shifts--; }

	} # End of the trailing for-loop (over columns within the region of the boundary-move).


	return $ct_shifts;

}	# END of the subroutine 'ct_necessary_shifts'. #


# compt_distance_bw_gblk_and_pt:
#	This subroutine computes the distance between a gap-block and a point (i.e., column) in the input alignment.
#	[CREATED on 2019/03/20.]
#
#	NOTE;
#	Here, the "distance" IS DEFINED AS the number of alignment columns (EXCLUDING those of gap-blocks vertically including or overlapping yet non-nested with the subject gap-block) IN BETWEEN the subject gap-block and the point.
#	(In other words, the "distance" is the number of "shift"s necessary for the subject gap-block to touch the point.)
#	The "distance" IS DEFINED AS -1 if the point is (horizontally) included in the gap-block.
#	
# INPUTS: ($point, $gb_sbj, \@bds_gblocks, \@inter_block_relations), where
#
#	$point is the position of the point (i.e., column) in question;
#
#	$gb_sbj is the index for the subject gap-block;
#
#	@{$bds_gblocks[$gb]} = ($lbd, $rbd) give the left- and right-boundaries of the $gb th gap-block;
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] is the relation of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above; See 'mk_inter_block_relations' for details];
#
# OUTPUTS: $distance, which is the distance computed.
#
sub compt_distance_bw_gblk_and_pt ($$\@\@) {

	my ($point, $gb_sbj, $bds_gblocks, $inter_block_relations) = @_;

	my ($lb_sbj, $rb_sbj) = @{$bds_gblocks->[$gb_sbj]};

	if (($lb_sbj <= $point) and ($point <= $rb_sbj)) { return -1; } # $point is horizontally included in the gap-block.

	if (($point == $lb_sbj-1) or ($point == $rb_sbj +1)) { return 0; } # $point already touches the gap-block.

	my ($bd_curr, $bd_goal);
	if ($point < $lb_sbj) {

	    ($bd_curr, $bd_goal) = ($lb_sbj, $point +1);
	
	} else { # if ($rb_sbj < $point)

	    ($bd_curr, $bd_goal) = ($rb_sbj, $point -1);
	}

	my %skipped = ($gb_sbj => 1);

	my $distance = ct_necessary_shifts ($bd_curr, $bd_goal, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
	unless (defined $distance) {
	    print STDERR "[ERROR 'compt_distance_bw_gblk_and_pt'] The subroutine 'ct_necessary_shifts ' FAILED on the $gb_sbj th gap-block (with current boundary = $bd_curr) and the goal $bd_goal: $! !\n\n";
	    return undef;
	}

	return $distance;

}	# END of the subroutine 'compt_distance_bw_gblk_and_pt'. #



# compt_inter_gblock_distance:
#	This MAIN subroutine computes the distanceS between two gap-blocks,
#	each of which IS DEFINED AS the number of "shift"s necessary for
#	one gap-block to touch (or overlap) the other.
#	The "distance" IS DEFINED AS -1 if the two gap-blocks overlap horizontally.
#	[CREATED on 2019/03/20.]
#
# INPUTS: ($gb1, $gb2, \@bds_gblocks, \@inter_block_relations), where
#
#	$gb1 and $gb2 are the indices for the two subject gap-blocks;
#
#	@{$bds_gblocks[$gb]} = ($lbd, $rbd) give the left- and right-boundaries of the $gb th gap-block;
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] is the relation of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above; See 'mk_inter_block_relations' for details];
#
# OUTPUTS: ($dist1, $dist2), where
#
#	$dist1 is the number of "shifts" for the $gb1 th gap-block to touch (or overlap) the $gb2 th gap-block;
#	$dist2 is the number of "shifts" for the $gb2 th gap-block to touch (or overlap) the $gb1 th gap-block;
#	$dist1 = $dist2 = -1 if $gb1 and $gb2 horizontally overlap each other.
#
sub compt_inter_gblock_distance ($$\@\@) {

	my ($gb1, $gb2, $bds_gblocks, $inter_block_relations) = @_;

	my ($lb1, $rb1) = @{$bds_gblocks->[$gb1]};
	my ($lb2, $rb2) = @{$bds_gblocks->[$gb2]};

	if (($rb1 + 1 == $lb2) or ($rb2 + 1 == $lb1)) { return (0, 0); } # $gb1 and $gb2 touch each other.

	if (($lb2 <= $rb1) and ($lb1 <= $rb2)) { return (-1, -1); } # $gb1 and $gb2 overlap each other.

	my %skipped = ($gb1 => 1, $gb2 => 1);


	if ($rb1 < $lb2) { # $gb1 is on the left of $gb2.

	    my ($bd_curr1, $bd_goal1) = ($rb1, $lb2-1);

	    my $dist1 = ct_necessary_shifts ($bd_curr1, $bd_goal1, $gb1, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
	    unless (defined $dist1) {
		print STDERR "[ERROR 'compt_inter_gblock_distance'] The subroutine 'ct_necessary_shifts ' FAILED on the $gb1 th gap-block (with current boundary = $bd_curr1) and the goal $bd_goal1: $! !\n\n";
		return (undef, undef);
	    }

	    my ($bd_curr2, $bd_goal2) = ($lb2, $rb1+1);

	    my $dist2 = ct_necessary_shifts ($bd_curr2, $bd_goal2, $gb2, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
	    unless (defined $dist2) {
		print STDERR "[ERROR 'compt_inter_gblock_distance'] The subroutine 'ct_necessary_shifts ' FAILED on the $gb2 th gap-block (with current boundary = $bd_curr2) and the goal $bd_goal2: $! !\n\n";
		return (undef, undef);
	    }

	    return ($dist1, $dist2);

	} else { # if ($rb2 < $lb1) # $gb2 is on the left of $gb1.

	    my ($bd_curr1, $bd_goal1) = ($lb1, $rb2+1);

	    my $dist1 = ct_necessary_shifts ($bd_curr1, $bd_goal1, $gb1, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
	    unless (defined $dist1) {
		print STDERR "[ERROR 'compt_inter_gblock_distance'] The subroutine 'ct_necessary_shifts ' FAILED on the $gb1 th gap-block (with current boundary = $bd_curr1) and the goal $bd_goal1: $! !\n\n";
		return (undef, undef);
	    }

	    my ($bd_curr2, $bd_goal2) = ($rb2, $lb1-1);

	    my $dist2 = ct_necessary_shifts ($bd_curr2, $bd_goal2, $gb2, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
	    unless (defined $dist2) {
		print STDERR "[ERROR 'compt_inter_gblock_distance'] The subroutine 'ct_necessary_shifts ' FAILED on the $gb2 th gap-block (with current boundary = $bd_curr2) and the goal $bd_goal2: $! !\n\n";
		return (undef, undef);
	    }

	    return ($dist1, $dist2);

	} # End of "if ($rb1 < $lb2) {...} else {...}".



}	# END of the subroutine 'compt_inter_gblock_distance'. #


# mk_lists_gblocks_w_spec_bounds:
#	This subroutine creates TWO sets of lists,
#	one is the set of lists of gap-blocks with specific left-boundaries,
#	the other is the set of lists of gap-blocks with specific right-boundaries.
#	The outputs of this subroutine will be used IN MANY OCCASIONS
#	in VARIOUS MOVES of the gap-blocks, including the "shift"s.
#	[CREATED on 2019/03/22; DEBUGGED on 2019/03/24.]
#
# INPUTS: ($le_wd, $re_wd, \@bds_gblocks), where
#
#	$le_wd and $re_wd, respectively, are the indexes
#	specifying the columns at the left-end and right-end of the input MSA window;
#	(0 <= $le_wd <= $re_wd must hold);
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
# OUTPUTS: (\@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column.
#
sub mk_lists_gblocks_w_spec_bounds ($$\@) {

	my ($le_wd, $re_wd, $bds_gblocks) = @_;

	unless ((0 <= $le_wd) and ($le_wd <= $re_wd)) {
	    print STDERR "[ERROR 'mk_lists_gblocks_w_spec_bounds'] Illegal inputs: ", '($le_wd, $re_wd) = ', "($le_wd, $re_wd): $! !\n\n";
	    return (undef, undef);
	}

		# (o) Initialize the outputs. #

	my @gblks_w_spec_lb = my @gblks_w_spec_rb = ();
	for (my $c = $le_wd; $c <= $re_wd; $c++) { # Preliminary for-loop (over columns in the window).

		$gblks_w_spec_lb[$c] = [];
		$gblks_w_spec_rb[$c] = [];

	} # End of the preliminary for-loop (over columns in the window).

		# (i) Fill in the outputs. #

	my $ct_gblks = @{$bds_gblocks}; # A BUG fixed on 2019/03/24.
	#my $ct_gblks = ${$bds_gblocks};
	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Main for-loop (over gap-blocks).

	    my ($lbd, $rbd) = @{$bds_gblocks->[$gb]};

	    push @{$gblks_w_spec_lb[$lbd]}, $gb;
	    push @{$gblks_w_spec_rb[$rbd]}, $gb;

	} # End of the main for-loop (over gap-blocks).

	return (\@gblks_w_spec_lb, \@gblks_w_spec_rb);

}	# END of the subroutine 'mk_lists_gblocks_w_spec_bounds'. #


# cts_vincluding_vincluded_onx_gblks:
#	This subroutine counts the numbers of gap-blocks
#	that vertically includes, that are vertically included in, and that are in the 'ONN' or 'ONCS' relationship with, each gap-block.
#	The output will be used to judge whether the subroutines named 'vextend_xxx' or 'vretract_xxx'
#	should be invoked or not.
#	[CREATED on 2019/05/21.]
#
# INPUTS: (\@inter_block_relations), where
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
# OUTPUTS: (\@cts_vincluding, \@cts_vincluded, \@cts_onx), where
#
#	$cts_vincluding[$gb] is the number of gap-blocks vertically including the $gb th gap-block;
#	$cts_vincluded[$gb] is the number of gap-blocks vertically included in the $gb th gap-block;
#	$cts_onx[$gb] is the number of gap-blocks that are in the 'ONN' or 'ONCS' relationship with the $gb th gap-block.
#
sub cts_vincluding_vincluded_onx_gblks (\@) {

	my ($inter_block_relations) = @_;

	my $ct_gbs = @{$inter_block_relations};

	my @cts_vincluding = my @cts_vincluded = my @cts_onx = ();

	for (my $gb1=0; $gb1<$ct_gbs; $gb1++) { # Main outer for-loop (over gap-blocks). #

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];

	    my $ct_vincluding = my $ct_vincluded = my $ct_onx = 0;

	    for (my $gb2=0; $gb2 < $ct_gbs; $gb2++) { # Inner for-loop (over gap-blocks). #
		if ($gb2 == $gb1) { next; }

		my $rel = $rels_w_gb1->[$gb2];
		if (($rel eq '<') or ($rel eq '<(pa)')) { # $gb2 vertically includes $gb1.

		    $ct_vincluding++;

		} elsif (($rel eq '>') or ($rel eq '>(ch)')) { # $gb2 is vertically included in $gb1.

		    $ct_vincluded++;

		} elsif ($rel eq '=') {

		    if ($gb2 < $gb1) { # $gb2 "vertically includes" $gb1.

			$ct_vincluding++;

		    } else { # if ($gb2 > $gb1), that is, $gb2 "is vertically included in" $gb1.

			$ct_vincluded++;
		    }

		} elsif (($rel eq 'ONN') or ($rel eq 'ONCS')) { # $gb2 is in the 'ONX' relationship with $gb1.
		    $ct_onx++;
		}

	    } # End of the inner for-loop (over gap-blocks). #

	    $cts_vincluding[$gb1] = $ct_vincluding;
	    $cts_vincluded[$gb1]  = $ct_vincluded;
	    $cts_onx[$gb1] = $ct_onx;

	} # End of the main outer for-loop (over gap-blocks).


	return (\@cts_vincluding, \@cts_vincluded, \@cts_onx);

}	# End of the subroutine 'ct_vincluding_vincluded_onx_gblks'. #


# vextend_right_end_sbj_gblk:
#	This subroutine keeps virtually extending the right-end of the "subject" gap-block you specify,
#	until the gap-block EITHER encounters a "solid" obstacle on its immediate right OR reaches the right-end of the window.
#	[CREATION STARTED on 2019/03/22, ENDED on 2019/03/23; REVISED on 2019/03/24; DEBUGGED on 2019/03/25.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vextend_left_end_sbj_gblk', below.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@set_columns, \@inter_block_relations, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@column = @{$set_columns[$c]} specifies the $c th column in the local MSA at present. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the (initial) input alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_columns is initialized by 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#	
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#
# [ OUTPUTS: $if_vextended, where
#
#	$if_vextended	= 1 if the right-end of the subject gap-block was virtually extended,
#			= 0 if it was NOT extended while the subroutine ended successfully, ] ... OBSOLETE as of 2019/03/24.
#
# OUTPUTS: $ct_vextensions, where
#
#	$ct_vextensions	= {The number of times that the left-end has been virtually extended}, 
#			= undef if the subroutine failed.
#
sub vextend_right_end_sbj_gblk ($\@\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $set_columns, $inter_block_relations, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my $ct_clms = @{$set_columns};
	if (@{$gblks_w_spec_lb} > $ct_clms) {
	    my $size = @{$gblks_w_spec_lb};
	    print STDERR "[ERROR 'vextend_right_end_sbj_gblk'] Illegal input: the size of ", '@{$gblks_w_spec_lb}', " (=$size) is LARGER THAN that of ", '@{$set_columns}', " (=$ct_clms): $! !\n\n";
	    return undef;
	}
	if (@{$gblks_w_spec_rb} > $ct_clms) {
	    my $size = @{$gblks_w_spec_rb};
	    print STDERR "[ERROR 'vextend_right_end_sbj_gblk'] Illegal input: the size of ", '@{$gblks_w_spec_rb}', " (=$size) is LARGER THAN that of ", '@{$set_columns}', " (=$ct_clms): $! !\n\n";
	    return undef;
	}

	my $ct_gblks = @{$bds_gblocks};

    (print STDOUT "< Subroutine 'vextend_right_end_sbj_gblk' (", '$gb_sbj = ', "$gb_sbj, #{columns} = $ct_clms, #{gap-blocks} = $ct_gblks) >\n\n") if ($TEST_LEVEL_vx_re_sgb > 0); # Added on 2019/03/24.

	my $right_end_wd = $ct_clms -1;

	my $bds_gb_sbj = $bds_gblocks->[$gb_sbj];
	my ($prev_lbd_sbj, $prev_rbd_sbj) = @{$bds_gb_sbj};

	my $rels_to_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vextensions = 0; # REVISED on 2019/03/24.
	##my $if_vextended = 0;

	while ($prev_rbd_sbj < $right_end_wd) { # Outer-outer while-loop (until the subject gap-block reaches the right-end of the window).

	    my $right_neighbors = $gblks_w_spec_lb->[$prev_rbd_sbj+1];
	    unless ((defined $right_neighbors) and (@{$right_neighbors} > 0)) { last; } # REVISED on 2019/03/25.
	    ##unless (@{$right_neighbors} > 0) { last; }

	    my $if_rlv = 0;

	    foreach my $gb2 (@{$right_neighbors}) { # Outer foreach-loop (over blocks ($gb2) that are immediate right-neighbors of $gb_sbj).

		my $bds_gb2 = $bds_gblocks->[$gb2];
		my ($lbd2, $rbd2) = @{$bds_gb2};

		my $rel = $rels_to_sbj->[$gb2];

		if (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb_sbj > $gb2)) ) { # A BUG fixed on 2019/03/25.
		##if (($rel eq '<') and ($rel eq '<(pa)') and (($rel eq '=') and ($gb_sbj > $gb2)) ) {

		    $if_rlv = 1;

			# (i) When $gb2 vertically "include"s $gb_sbj,
			#     extend the right-end of $gb_sbj.

			## Added on 2019/03/23. (1) ##

				# (1) Remove $gb_sbj from the set of gap-blocks with right-boundary = $prev_rbd_sbj.

		    my @new_gblks_w_prev_rbd_sbj = ();
		    foreach my $gb3 (@{$gblks_w_spec_rb->[$prev_rbd_sbj]}) { # 1st middle foreach-loop.
			if ($gb3 != $gb_sbj) { push @new_gblks_w_prev_rbd_sbj, $gb3; }
		    } # End of the 1st middle foreach-loop.
		    $gblks_w_spec_rb->[$prev_rbd_sbj] = \@new_gblks_w_prev_rbd_sbj;

				# (2) Virtually extend the right-boundary.

		    $prev_rbd_sbj = $bds_gb_sbj->[1] = $rbd2;

				# (3) Add $gb_sbj to the set of gap-blocks with right-boundary = $rbd2.

		     push @{$gblks_w_spec_rb->[$rbd2]}, $gb_sbj;

		    $ct_vextensions++; # REVISED on 2019/03/24.
		    ##$if_vextended = 1;

			## End of "Added on 2019/03/23. (1)" ##

		    last; # Leave the outer foreach-loop.

		} elsif (($rel eq 'ONN') or ($rel eq 'ONCS')) {

		    $if_rlv = 1;

			# (ii) When $gb2  vertically overlaps, yet is non-nested with, $gb_sbj,
			#	horizontally swap the minimal segment including $gb_sbj
			#	with that including $gb2.

			## Added on 2019/03/23. (2) ##

		    my $size_gb_sbj = $prev_rbd_sbj - $prev_lbd_sbj + 1;
		    my $size_gb2 = $rbd2 - $lbd2 + 1;

			# (1) Swap the alignment columns.

		    my @reservoir_clms = @{$set_columns}[$prev_lbd_sbj .. $prev_rbd_sbj];
		    for (my $c = $lbd2; $c <= $rbd2; $c++) { $set_columns->[$c-$size_gb_sbj] = $set_columns->[$c]; }
		    for (my $c = 0; $c < $size_gb_sbj; $c++) { $set_columns->[$rbd2 - $c] = $reservoir_clms[$size_gb_sbj - 1 - $c]; }

			# (2) & (3) Swap the left- and right-boundaries of the relevant gap-blocks.

		    if ($ct_gblks < $size_gb_sbj + $size_gb2) { # If the number of gap-blocks is relatively small.

    (print STDOUT " ... Swap-Type A ...\n\n") if ($TEST_LEVEL_vx_re_sgb > 1); # Added on 2019/03/24.

			for (my $gb3 = 0; $gb3 < $ct_gblks; $gb3++) { # 5th middle for-loop (over gap-blocks).

			    my $bds_gb3 = $bds_gblocks->[$gb3];

			    foreach my $indx (0, 1) { # Inner foreach-loop (dealing with left- and right-boundaries).
				my $bd = $bds_gb3->[$indx];
				if (($prev_lbd_sbj <= $bd) and ($bd <= $prev_rbd_sbj)) {
				    $bds_gb3->[$indx] += $size_gb2;
				} elsif (($lbd2 <= $bd) and ($bd <= $rbd2)) {
				    $bds_gb3->[$indx] -= $size_gb_sbj;
				}
			    } # End of the inner foreach-loop (dealig with left- and right-boundaries).

			} # End of the 5th middle for-loop (over gap-blocks).

		    } else { # If the number of involved alignment columns is relatively small.

    (print STDOUT " ... Swap-Type B ...\n\n") if ($TEST_LEVEL_vx_re_sgb > 1); # Added on 2019/03/24.

			for (my $c = $prev_lbd_sbj; $c <= $prev_rbd_sbj; $c++) { # 6th middle for-loop (over columns horizontally included in $gb_sbj).

			    foreach my $gb3 (@{$gblks_w_spec_lb->[$c]}) { $bds_gblocks->[$gb3][0] += $size_gb2; } # Move the left-bounds.

			    foreach my $gb3 (@{$gblks_w_spec_rb->[$c]}) { $bds_gblocks->[$gb3][1] += $size_gb2; } # Move the right-bounds.

			} # End of the 6th middle for-loop (over columns horizontally included in $gb_sbj).

			for (my $c = $lbd2; $c <= $rbd2; $c++) { # 7th middle for-loop (over columns horizontally included in $gb2).

			    foreach my $gb3 (@{$gblks_w_spec_lb->[$c]}) { $bds_gblocks->[$gb3][0] -= $size_gb_sbj; } # Move the left-bounds.

			    foreach my $gb3 (@{$gblks_w_spec_rb->[$c]}) { $bds_gblocks->[$gb3][1] -= $size_gb_sbj; } # Move the right-bounds.

			} # End of the 7th middle for-loop (over columns horizontally included in $gb2).

		    } # End of "if ($ct_gblks < $size_gb_sbj + $size_gb2) {...} else {...}".


			# (5) Swap the relevant elements in @{$gblks_w_spec_lb} and @{$gblks_w_spec_rb}.

		    my @reservoir_w_spec_lb = @{$gblks_w_spec_lb}[$prev_lbd_sbj .. $prev_rbd_sbj];
		    my @reservoir_w_spec_rb = @{$gblks_w_spec_rb}[$prev_lbd_sbj .. $prev_rbd_sbj];

		    for (my $c = $lbd2; $c <= $rbd2; $c++) { # 8th middle for-loop (over columns horizontally included in $gb2).
			$gblks_w_spec_lb->[$c - $size_gb_sbj] = $gblks_w_spec_lb->[$c];
			$gblks_w_spec_rb->[$c - $size_gb_sbj] = $gblks_w_spec_rb->[$c];

		    } # End of the 8th middle for-loop (over columns horizontally included in $gb2).

		    for (my $c =0; $c < $size_gb_sbj; $c++) { # 9th middle for-loop (over columns horizontally included in $gb_sbj).
			$gblks_w_spec_lb->[$rbd2 - $c] = $reservoir_w_spec_lb[$size_gb_sbj -1 - $c];
			$gblks_w_spec_rb->[$rbd2 - $c] = $reservoir_w_spec_rb[$size_gb_sbj -1 - $c];

		    } # End of the 9th middle for-loop (over columns horizontally included in $gb_sbj).

			# Finally, update the boundaries of $gb_sbj. #

		    ($prev_lbd_sbj, $prev_rbd_sbj) = @{$bds_gb_sbj};


		    $ct_vextensions++; # REVISED on 2019/03/24.
		    ##$if_vextended = 1;

			## End of "Added on 2019/03/23. (2)" ##

		    last; # Leave the outer foreach-loop.
		}

	    } # End of the outer foreach-loop (over blocks ($gb2) that are immediate right-neighbors of $gb_sbj).

	    unless ($if_rlv) { last; } # If no relevant gap-blocks are in the immediate right-neighbourhood of $gb_sbj, END the virtual extension of the right-boundary.

	} # End of the outer-outer while-loop (until the subject gap-block reaches the right-end of the window.)
	

    (print STDOUT "< Subroutine 'vextend_right_end_sbj_gblk' Ended />\n\n") if ($TEST_LEVEL_vx_re_sgb > 0); # Added on 2019/03/24.

	return $ct_vextensions; # REVISED on 2019/03/24.
	##return $if_vextended;

}	# END of the subroutine 'vextend_right_end_sbj_gblk'. #



# vextend_left_end_sbj_gblk:
#	This subroutine keeps virtually extending the left-end of the "subject" gap-block you specify,
#	until the gap-block EITHER encounters a "solid" obstacle on its immediate left OR reaches the left-end of the window.
#	[CREATION STARTED on 2019/03/22, ENDED on 2019/03/23; REVISED on 2019/03/24; DEBUGGED on 2019/03/25.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vextend_right_end_sbj_gblk', above.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@set_columns, \@inter_block_relations, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@column = @{$set_columns[$c]} specifies the $c th column in the local MSA at present. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the (initial) input alignment that accommodates the corresponding site of the $k th "sequence";
#	[Typically, \@set_columns is initialized by 'mk_init_set_columns0' in "MyANEX_Supple.pm"];
#	
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#
# [ OUTPUTS: $if_vextended, where
#
#	$if_vextended	= 1 if the left-end of the subject gap-block was virtually extended,
#			= 0 if it was NOT extended while the subroutine ended successfully,] ... OBSOLETE as of 2019/03/24.
#
# OUTPUTS: $ct_vextensions, where
#
#	$ct_vextensions	= {The number of times that the left-end has been virtually extended}, 
#			= undef if the subroutine failed.
#
sub vextend_left_end_sbj_gblk ($\@\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $set_columns, $inter_block_relations, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my $ct_clms = @{$set_columns};
	if (@{$gblks_w_spec_lb} > $ct_clms) {
	    my $size = @{$gblks_w_spec_lb};
	    print STDERR "[ERROR 'vextend_left_end_sbj_gblk'] Illegal input: the size of ", '@{$gblks_w_spec_lb}', " (=$size) is LARGER THAN that of ", '@{$set_columns}', " (=$ct_clms): $! !\n\n";
	    return undef;
	}
	if (@{$gblks_w_spec_rb} > $ct_clms) {
	    my $size = @{$gblks_w_spec_rb};
	    print STDERR "[ERROR 'vextend_left_end_sbj_gblk'] Illegal input: the size of ", '@{$gblks_w_spec_rb}', " (=$size) is LARGER THAN that of ", '@{$set_columns}', " (=$ct_clms): $! !\n\n";
	    return undef;
	}

			## Restarted on 2019/03/23. ##

	my $ct_gblks = @{$bds_gblocks};

    (print STDOUT "< Subroutine 'vextend_left_end_sbj_gblk' (", '$gb_sbj = ', "$gb_sbj, #{columns} = $ct_clms, #{gap-blocks} = $ct_gblks) >\n\n") if ($TEST_LEVEL_vx_le_sgb > 0); # Added on 2019/03/24.

	my $left_end_wd = 0;

	my $bds_gb_sbj = $bds_gblocks->[$gb_sbj];
	my ($prev_lbd_sbj, $prev_rbd_sbj) = @{$bds_gb_sbj};

	my $rels_to_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vextensions = 0; # REVISED on 2019/03/24.
	##my $if_vextended = 0;

	while ($left_end_wd < $prev_lbd_sbj) { # Outer-outer while-loop (until the subject gap-block reaches the left-end of the window).

	    my $left_neighbors = $gblks_w_spec_rb->[$prev_lbd_sbj -1];
	    unless ((defined $left_neighbors) and (@{$left_neighbors} > 0)) { last; } # REVISED on 2019/03/25.
	    ##unless (@{$left_neighbors} > 0) { last; }

	    my $if_rlv = 0;

	    foreach my $gb2 (@{$left_neighbors}) { # Outer foreach-loop (over blocks ($gb2) that are immediate left-neighbors of $gb_sbj).

		my $bds_gb2 = $bds_gblocks->[$gb2];
		my ($lbd2, $rbd2) = @{$bds_gb2};

		my $rel = $rels_to_sbj->[$gb2];

		if (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb_sbj > $gb2)) ) { # DEBUGGED on 2019/03/25.
		##if (($rel eq '<') and ($rel eq '<(pa)') and (($rel eq '=') and ($gb_sbj > $gb2)) ) {

		    $if_rlv = 1;

			# (i) When $gb2 vertically "include"s $gb_sbj,
			#     extend the left-end of $gb_sbj.

				# (1) Remove $gb_sbj from the set of gap-blocks with left-boundary = $prev_lbd_sbj.

		    my @new_gblks_w_prev_lbd_sbj = ();
		    foreach my $gb3 (@{$gblks_w_spec_lb->[$prev_lbd_sbj]}) { # 1st middle foreach-loop.
			if ($gb3 != $gb_sbj) { push @new_gblks_w_prev_lbd_sbj, $gb3; }
		    } # End of the 1st middle foreach-loop.
		    $gblks_w_spec_lb->[$prev_lbd_sbj] = \@new_gblks_w_prev_lbd_sbj;

				# (2) Virtually extend the left-boundary.

		    $prev_lbd_sbj = $bds_gb_sbj->[0] = $lbd2;

				# (3) Add $gb_sbj to the set of gap-blocks with left-boundary = $lbd2.

		     push @{$gblks_w_spec_lb->[$lbd2]}, $gb_sbj;

		    $ct_vextensions++; # REVISED on 2019/03/24.
		    ##$if_vextended = 1;

		    last; # Leave the outer foreach-loop.

		} elsif (($rel eq 'ONN') or ($rel eq 'ONCS')) {

		    $if_rlv = 1;

			# (ii) When $gb2  vertically overlaps, yet is non-nested with, $gb_sbj,
			#	horizontally swap the minimal segment including $gb_sbj
			#	with that including $gb2.

		    my $size_gb_sbj = $prev_rbd_sbj - $prev_lbd_sbj + 1;
		    my $size_gb2 = $rbd2 - $lbd2 + 1;

			# (1) Swap the alignment columns.

		    my @reservoir_clms = @{$set_columns}[$prev_lbd_sbj .. $prev_rbd_sbj];
		    for (my $c = $rbd2; $c >= $lbd2; $c--) { $set_columns->[$c +$size_gb_sbj] = $set_columns->[$c]; }
		    for (my $c = 0; $c < $size_gb_sbj; $c++) { $set_columns->[$lbd2 + $c] = $reservoir_clms[$c]; }

			# (2) & (3) Swap the left- and right-boundaries of the relevant gap-blocks.

		    if ($ct_gblks < $size_gb_sbj + $size_gb2) { # If the number of gap-blocks is relatively small.

    (print STDOUT " ... Swap-Type A ...\n\n") if ($TEST_LEVEL_vx_le_sgb > 1); # Added on 2019/03/24.

			for (my $gb3 = 0; $gb3 < $ct_gblks; $gb3++) { # 5th middle for-loop (over gap-blocks).

			    my $bds_gb3 = $bds_gblocks->[$gb3];

			    foreach my $indx (0, 1) { # Inner foreach-loop (dealing with left- and right-boundaries).
				my $bd = $bds_gb3->[$indx];
				if (($prev_lbd_sbj <= $bd) and ($bd <= $prev_rbd_sbj)) {
				    $bds_gb3->[$indx] -= $size_gb2;
				} elsif (($lbd2 <= $bd) and ($bd <= $rbd2)) {
				    $bds_gb3->[$indx] += $size_gb_sbj;
				}
			    } # End of the inner foreach-loop (dealig with left- and right-boundaries).

			} # End of the 5th middle for-loop (over gap-blocks).

		    } else { # If the number of involved alignment columns is relatively small.

    (print STDOUT " ... Swap-Type B ...\n\n") if ($TEST_LEVEL_vx_le_sgb > 1); # Added on 2019/03/24.

			for (my $c = $prev_lbd_sbj; $c <= $prev_rbd_sbj; $c++) { # 6th middle for-loop (over columns horizontally included in $gb_sbj).

			    foreach my $gb3 (@{$gblks_w_spec_lb->[$c]}) { $bds_gblocks->[$gb3][0] -= $size_gb2; } # Move the left-bounds.

			    foreach my $gb3 (@{$gblks_w_spec_rb->[$c]}) { $bds_gblocks->[$gb3][1] -= $size_gb2; } # Move the right-bounds.

			} # End of the 6th middle for-loop (over columns horizontally included in $gb_sbj).

			for (my $c = $lbd2; $c <= $rbd2; $c++) { # 7th middle for-loop (over columns horizontally included in $gb2).

			    foreach my $gb3 (@{$gblks_w_spec_lb->[$c]}) { $bds_gblocks->[$gb3][0] += $size_gb_sbj; } # Move the left-bounds.

			    foreach my $gb3 (@{$gblks_w_spec_rb->[$c]}) { $bds_gblocks->[$gb3][1] += $size_gb_sbj; } # Move the right-bounds.

			} # End of the 7th middle for-loop (over columns horizontally included in $gb2).

		    } # End of "if ($ct_gblks < $size_gb_sbj + $size_gb2) {...} else {...}".


			# (5) Swap the relevant elements in @{$gblks_w_spec_lb} and @{$gblks_w_spec_rb}.

		    my @reservoir_w_spec_lb = @{$gblks_w_spec_lb}[$prev_lbd_sbj .. $prev_rbd_sbj];
		    my @reservoir_w_spec_rb = @{$gblks_w_spec_rb}[$prev_lbd_sbj .. $prev_rbd_sbj];

		    for (my $c = $rbd2; $c >= $lbd2; $c--) { # 8th middle for-loop (over columns horizontally included in $gb2).
			$gblks_w_spec_lb->[$c + $size_gb_sbj] = $gblks_w_spec_lb->[$c];
			$gblks_w_spec_rb->[$c + $size_gb_sbj] = $gblks_w_spec_rb->[$c];

		    } # End of the 8th middle for-loop (over columns horizontally included in $gb2).

		    for (my $c =0; $c < $size_gb_sbj; $c++) { # 9th middle for-loop (over columns horizontally included in $gb_sbj).
			$gblks_w_spec_lb->[$lbd2 + $c] = $reservoir_w_spec_lb[$c];
			$gblks_w_spec_rb->[$lbd2 + $c] = $reservoir_w_spec_rb[$c];

		    } # End of the 9th middle for-loop (over columns horizontally included in $gb_sbj).

			# Finally, update the boundaries of $gb_sbj. #

		    ($prev_lbd_sbj, $prev_rbd_sbj) = @{$bds_gb_sbj};


		    $ct_vextensions++; # REVISED on 2019/03/24.
		    ##$if_vextended = 1;

		    last; # Leave the outer foreach-loop.
		}

	    } # End of the outer foreach-loop (over blocks ($gb2) that are immediate left-neighbors of $gb_sbj).

	    unless ($if_rlv) { last; } # If no relevant gap-blocks are in the immediate left-neighbourhood of $gb_sbj, END the virtual extension of the left-boundary

	} # End of the outer-outer while-loop (until the subject gap-block reaches the left-end of the window.)
	

    (print STDOUT "< Subroutine 'vextend_left_end_sbj_gblk' Ended />\n\n") if ($TEST_LEVEL_vx_le_sgb > 0); # Added on 2019/03/24.

	return $ct_vextensions; # REVISED on 2019/03/24.
	##return $if_vextended;

}	# END of the subroutine 'vextend_left_end_sbj_gblk'. #


# vextend_rear_ends_rflnk_vincluded_gblks:
#	This subroutine first examines whether there are gap-blocks
#	that are right-flanking and also vertically included in the subject gap-block,
#	and, if so, virtually extends their "rear-ends" (i.e., left-ends) to the left-end of the subject gap-block.
#	Typically, it is performed after 'vextend_right_end_sbj_gblk' above.
#	[CREATED on 2019/03/23; Soft-DEBUGGED on 2019/03/25.]	
#
#	NOTE: This subroutine is paired with 'vextend_rear_ends_lflnk_vincluded_gblks' below.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblks_w_spec_lb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vextended, which is the number of gap-blocks whose "rear-ends" are virtually extended.
#	   It is undef if the subroutine FAILED.
#
sub vextend_rear_ends_rflnk_vincluded_gblks ($\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblks_w_spec_lb) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};

		# ADDED to prevent an unexpected bug (on 2019/03/25). #
	my $re_wd = scalar (@{$gblks_w_spec_lb}) -1;
	unless ($rbd_sbj < $re_wd) { return 0; }
		# END of "ADDED to prevent an unexpected bug (on 2019/03/25)." #

	my $rflanking_gblks = $gblks_w_spec_lb->[$rbd_sbj +1];

	unless ((defined $rflanking_gblks) and (@{$rflanking_gblks}>0)) { return 0; }

	my $rels_to_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vextended = 0;
	my @new_rflanking_gblks = ();
	foreach my $gb2 (@{$rflanking_gblks}) { # Main foreach-loop (over right-flanking gap-blocks).

	    my $rel = $rels_to_sbj->[$gb2];

	    unless (($rel eq '>') or ($rel eq '>(ch)') or (($rel eq '=') and ($gb2 > $gb_sbj)) ) {

		push @new_rflanking_gblks, $gb2; # Its left-bound remains unchanged.
		next;
	    }
		# Virtually extend its left-bound to $lbd_sbj.

	    $bds_gblocks->[$gb2][0] = $lbd_sbj;

	    push @{$gblks_w_spec_lb->[$lbd_sbj]}, $gb2;

	    $ct_vextended++;

	} # End of the main foreach-loop (over right-flanking gap-blocks).

	if ($ct_vextended>0) { $gblks_w_spec_lb->[$rbd_sbj+1] = \@new_rflanking_gblks; } # Update the set of right-flanking gap-blocks.

	return  $ct_vextended;

}	# END of the subroutine 'vextend_rear_ends_rflnk_vincluded_gblks'. #


# vextend_rear_ends_lflnk_vincluded_gblks:
#	This subroutine first examines whether there are gap-blocks
#	that are left-flanking and also vertically included in the subject gap-block,
#	and, if so, virtually extends their "rear-ends" (i.e., right-ends) to the right-end of the subject gap-block.
#	Typically, it is performed after 'vextend_left_end_sbj_gblk' above.
#	[CREATED on 2019/03/23; DEBUGGED on 2019/03/25.]	
#
#	NOTE: This subroutine is paired with 'vextend_rear_ends_rflnk_vincluded_gblks' above.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblks_w_spec_rb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vextended, which is the number of gap-blocks whose "rear-ends" are virtually extended.
#	   It is undef if the subroutine FAILED.
#
sub vextend_rear_ends_lflnk_vincluded_gblks ($\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblks_w_spec_rb) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};

		# ADDED to FIX a BUG (on 2019/03/25). #
	my $le_wd = 0;
	unless ($le_wd < $lbd_sbj) { return 0; }
		# END of "ADDED to FIX a BUG (on 2019/03/25)." #

	my $lflanking_gblks = $gblks_w_spec_rb->[$lbd_sbj -1];

	unless ((defined $lflanking_gblks) and (@{$lflanking_gblks}>0)) { return 0; }

	my $rels_to_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vextended = 0;
	my @new_lflanking_gblks = ();
	foreach my $gb2 (@{$lflanking_gblks}) { # Main foreach-loop (over left-flanking gap-blocks).

	    my $rel = $rels_to_sbj->[$gb2];

	    unless (($rel eq '>') or ($rel eq '>(ch)') or (($rel eq '=') and ($gb2 > $gb_sbj)) ) {

		push @new_lflanking_gblks, $gb2; # Its right-bound remains unchanged.
		next;
	    }
		# Virtually extend its right-bound to $rbd_sbj.

	    $bds_gblocks->[$gb2][1] = $rbd_sbj;

	    push @{$gblks_w_spec_rb->[$rbd_sbj]}, $gb2;

	    $ct_vextended++;

	} # End of the main foreach-loop (over left-flanking gap-blocks).

	if ($ct_vextended>0) { $gblks_w_spec_rb->[$lbd_sbj -1] = \@new_lflanking_gblks; } # Update the set of left-flanking gap-blocks.

	return  $ct_vextended;

}	# END of the subroutine 'vextend_rear_ends_lflnk_vincluded_gblks'. #


# vretract_left_end_sbj_gblk:
#	This subroutine attempts to retract the left-end of the subject gap-block whenever possible,
#	that is, whenever it coincides with the left-end of a gap-block vertically including the subject.
#	Typically, the subroutine is used AFTER a "shift" (to the right) of a gap-block(, which is going to be the "subject").
#	[CREATED on 2019/03/27.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vretract_right_end_sbj_gblk', below.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the output of 'mk_list_gblock_sizes' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vretracting, which is the number of times that the "left-end" is virtually retracted.
#	   It is undef if the subroutine FAILED.
#
sub vretract_left_end_sbj_gblk ($\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblock_sizes, $gblks_w_spec_lb) = @_;

	my $bds_sbj = $bds_gblocks->[$gb_sbj];
	my ($lbd1, $rbd1) = @{$bds_sbj};
	my $size1 = $gblock_sizes->[$gb_sbj];

	unless ($lbd1 + $size1 -1 < $rbd1)  { return 0; }

	my $gblks_w_lb_eq_lbd1 = $gblks_w_spec_lb->[$lbd1];

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vretracting = 0;

	while ((defined $gblks_w_lb_eq_lbd1) and (@{$gblks_w_lb_eq_lbd1}>1)) { # Main while-loop (while the set contains at least one non-subject gap-block).

	    my $if_rlv = 0;
	    foreach my $gb2 (@{$gblks_w_lb_eq_lbd1}) { # Middle foreach-loop (over gap-blocks whose left-bounds equal $lbd1).

		if ($gb2 == $gb_sbj) { next; }

		my $rel = $rels_w_sbj->[$gb2];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb2 < $gb_sbj)) ) { next; }

		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};

		unless ($rbd2 < $rbd1) {
		    print STDERR "[ERROR 'vretract_left_end_sbj_gblk'] The right-bound, $rbd1, of the subject (the $gb_sbj th gap-block) is NOT on the right of the right-bound, $rbd2, of the $gb2 th block, which vertically includes the subject: $! !!\n\n";
		    return undef;
		}

			# Update the set of gap-blocks w/ left-bound = $lbd1 (before retracting). #
		my @gblks_w_lb_eq_old_lbd1 = ();
		foreach my $gb3 (@{$gblks_w_lb_eq_lbd1}) { # Inner foreach-loop).
		    if ($gb3 != $gb_sbj) { push @gblks_w_lb_eq_old_lbd1, $gb3; }
		} # End of the inner foreach-loop.
		$gblks_w_spec_lb->[$lbd1] = \@gblks_w_lb_eq_old_lbd1;

			# Update $lbd1 and $gblks_w_lb_eq_lbd1.
		$lbd1 = $rbd2 +1;
		$bds_sbj->[0] = $lbd1;
		$gblks_w_lb_eq_lbd1 = $gblks_w_spec_lb->[$lbd1];

			# ADD $gb_sbj to the NEW @{$gblks_w_lb_eq_lbd1}.
		push @{$gblks_w_lb_eq_lbd1}, $gb_sbj;

		$if_rlv = 1;
		$ct_vretracting++;
		last;

	    } # End of the middle foreach-loop (over gap-blocks whose left-bounds equal $lbd1).

	    unless ($if_rlv and ($lbd1 + $size1 -1 < $rbd1) ) { last; } # If $lbd1 was NOT retracted in this round OR if the subject is already tight, do NOT attempt to retract FURTHER.

	} # END of the main while-loop (while the set contains at least one non-subject gap-block).

	return $ct_vretracting;

}	# END of the subroutine 'vretract_left_end_sbj_gblk'. #


# vretract_right_end_sbj_gblk:
#	This subroutine attempts to retract the right-end of the subject gap-block whenever possible,
#	that is, whenever it coincides with the right-end of a gap-block vertically including the subject.
#	Typically, the subroutine is used AFTER a "shift" (to the left) of a gap-block(, which is going to be the "subject").
#	[CREATED on 2019/03/27.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vretract_left_end_sbj_gblk', above.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_rb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the output of 'mk_list_gblock_sizes' above];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vretracting, which is the number of times that the "right-end" is virtually retracted.
#	   It is undef if the subroutine FAILED.
#
sub vretract_right_end_sbj_gblk ($\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblock_sizes, $gblks_w_spec_rb) = @_;

	my $bds_sbj = $bds_gblocks->[$gb_sbj];
	my ($lbd1, $rbd1) = @{$bds_sbj};
	my $size1 = $gblock_sizes->[$gb_sbj];

	unless ($lbd1 + $size1 -1 < $rbd1)  { return 0; }

	my $gblks_w_rb_eq_rbd1 = $gblks_w_spec_rb->[$rbd1];

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vretracting = 0;

	while ( (defined $gblks_w_rb_eq_rbd1) and (@{$gblks_w_rb_eq_rbd1}>1)) { # Main while-loop (while the set contains at least one non-subject gap-block).

	    my $if_rlv = 0;
	    foreach my $gb2 (@{$gblks_w_rb_eq_rbd1}) { # Middle foreach-loop (over gap-blocks whose right-bounds equal $rbd1).

		if ($gb2 == $gb_sbj) { next; }

		my $rel = $rels_w_sbj->[$gb2];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb2 < $gb_sbj)) ) { next; }

		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};

		unless ($lbd1 < $lbd2) {
		    print STDERR "[ERROR 'vretract_right_end_sbj_gblk'] The left-bound, $lbd1, of the subject (the $gb_sbj th gap-block) is NOT on the left of the left-bound, $lbd2, of the $gb2 th block, which vertically includes the subject: $! !!\n\n";
		    return undef;
		}

			# Update the set of gap-blocks w/ right-bound = $rbd1 (before retracting). #
		my @gblks_w_rb_eq_old_rbd1 = ();
		foreach my $gb3 (@{$gblks_w_rb_eq_rbd1}) { # Inner foreach-loop).
		    if ($gb3 != $gb_sbj) { push @gblks_w_rb_eq_old_rbd1, $gb3; }
		} # End of the inner foreach-loop.
		$gblks_w_spec_rb->[$rbd1] = \@gblks_w_rb_eq_old_rbd1;

			# Update $lbd1 and $gblks_w_rb_eq_rbd1.
		$rbd1 = $lbd2 -1;
		$bds_sbj->[1] = $rbd1;
		$gblks_w_rb_eq_rbd1 = $gblks_w_spec_rb->[$rbd1];

			# ADD $gb_sbj to the NEW @{$gblks_w_rb_eq_rbd1}.
		push @{$gblks_w_rb_eq_rbd1}, $gb_sbj;

		$if_rlv = 1;
		$ct_vretracting++;
		last;

	    } # End of the middle foreach-loop (over gap-blocks whose right-bounds equal $rbd1).

	    unless ($if_rlv and ($lbd1 + $size1 -1 < $rbd1) ) { last; } # If $lbd1 was NOT retracted in this round OR if the subject is already tight, do NOT attempt to retract FURTHER.

	} # END of the main while-loop (while the set contains at least one non-subject gap-block).

	return $ct_vretracting;

}	# END of the subroutine 'vretract_right_end_sbj_gblk'. #


# vretract_right_ends_vincluded_gblks:
#	This subroutine first looks for gap-blocks whose right-ends coincides with the right-end of the subject gap-block,
#	and, if there are any such gap-blocks,
#	it attempts to retract the right-end of each of such gap-blocks as much as possible.
#	Typically, it is performed AFTER the shift (to the right) of a gap-block(, which is going to be the "subject"),
#	and (preferably) BEFORE performing 'vretract_left_end_sbj_gblk'.
#	[CREATED on 2019/03/27.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vretract_left_ends_vincluded_gblks', below.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_rb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the output of 'mk_list_gblock_sizes' above];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vretracted, which is the number of gap-blocks whose "rightd-ends" are virtually retracted.
#	   It is undef if the subroutine FAILED.
#
sub vretract_right_ends_vincluded_gblks ($\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblock_sizes, $gblks_w_spec_rb) = @_;

	my $bds_sbj = $bds_gblocks->[$gb_sbj];
	my ($lbd1, $rbd1) = @{$bds_sbj};

	my $gblks_w_rb_eq_rbd1 = $gblks_w_spec_rb->[$rbd1];
	unless ( (defined $gblks_w_rb_eq_rbd1) and (@{$gblks_w_rb_eq_rbd1}>1) ) { return 0; } # If the set contains NO non-subject gap-blocks, NO right-ends can be retracted anyway.

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vretracted = 0;

	my @new_gblks_w_rb_eq_rb1 = ();
	foreach my $gb2 (@{$gblks_w_rb_eq_rbd1}) { # Main foreach-loop (over gap-blocks whose right-bounds equal $rb1).
	    if ($gb2 == $gb_sbj) {
		push @new_gblks_w_rb_eq_rb1, $gb2;
		next;
	    }

	    my $rel = $rels_w_sbj->[$gb2];

	    unless (($rel eq '>') or ($rel eq '>(ch)') or (($rel eq '=') and ($gb_sbj < $gb2)) ) {
			# Here, we are NOT interested in this type of gap-block. #
		push @new_gblks_w_rb_eq_rb1, $gb2;
		next;
	    }

	    my $bds_gb2 = $bds_gblocks->[$gb2];
	    my ($lbd2, $rbd2) = @{$bds_gb2};

	    unless ($lbd2 < $lbd1) {
		print STDERR "[ERROR 'vretract_right_ends_vincluded_gblks'] The left-bound, $lbd1, of the subject (the $gb_sbj th gap-block) is NOT on the right of the left-bound, $lbd2, of the $gb2 th block, which is vertically included in the subject: $! !!\n\n";
		return undef;
	    }

	    $ct_vretracted++;

	    	# Update $rbd2.
	    $rbd2 = $lbd1 -1;
	    $bds_gb2->[1] = $rbd2;
	    push @{$gblks_w_spec_rb->[$rbd2]}, $gb2;  

	    my $size2 = $gblock_sizes->[$gb2];

	    unless ($lbd2 + $size2 - 1 < $rbd2) { next; } # The right-enf of $gb2 canNOT be retracted FURTHER.

		# Attempt to FURTHER retract the right-end of $gb2.
			# (NOTE added on 2019/03/29:
			#    Come to think of it, the subsequent part may be superfluous.
			#    Just in case, however, I will leave it as it is.)

	    my $ct_vretracting2 = vretract_right_end_sbj_gblk ($gb2, @{$bds_gblocks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$gblks_w_spec_rb});
	    unless (defined $ct_vretracting2) {
		print STDERR "[ERROR 'vretract_right_ends_vincluded_gblks'] The subroutine 'vretract_right_end_sbj_gblk' FAILED on the $gb2 th gap-block, which is vertically included in the subject (i.e., the $gb_sbj th gap-block): $! !!\n\n";
		return undef;
	    }

	} # End of the main foreach-loop (over gap-blocks whose right-bounds equal $rb1).

	if ($ct_vretracted >0) {
		# Update the set whose right-bounds equal $rbd1.
	    $gblks_w_spec_rb->[$rbd1] = \@new_gblks_w_rb_eq_rb1;
	}

	return $ct_vretracted;

}	# END of the subroutine 'vretract_right_ends_vincluded_gblks'. #


# vretract_left_ends_vincluded_gblks:
#	This subroutine first looks for gap-blocks whose left-ends coincides with the left-end of the subject gap-block,
#	and, if there are any such gap-blocks,
#	it attempts to retract the left-end of each of such gap-blocks as much as possible.
#	Typically, it is performed AFTER the shift (to the left) of a gap-block(, which is going to be the "subject"),
#	and (preferably) BEFORE performing 'vretract_right_end_sbj_gblk'.
#	[CREATED on 2019/03/27.]
#
#	NOTE: This subroutine is paired with its own MIRROR IMAGE, 'vretract_right_ends_vincluded_gblks', above.
#
# INPUTS: ($gb_sbj, \@bds_gblocks, \@inter_block_relations, \@gblock_sizes, \@gblks_w_spec_lb), where 
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the output of 'mk_list_gblock_sizes' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $ct_vretracted, which is the number of gap-blocks whose "rightd-ends" are virtually retracted.
#	   It is undef if the subroutine FAILED.
#
sub vretract_left_ends_vincluded_gblks ($\@\@\@\@) {

	my ($gb_sbj, $bds_gblocks, $inter_block_relations, $gblock_sizes, $gblks_w_spec_lb) = @_;

	my $bds_sbj = $bds_gblocks->[$gb_sbj];
	my ($lbd1, $rbd1) = @{$bds_sbj};

	my $gblks_w_lb_eq_lbd1 = $gblks_w_spec_lb->[$lbd1];
	unless ( (defined $gblks_w_lb_eq_lbd1) and (@{$gblks_w_lb_eq_lbd1}>1) ) { return 0; } # If the set contains NO non-subject gap-blocks, NO left-ends can be retracted anyway.

	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];

	my $ct_vretracted = 0;

	my @new_gblks_w_lb_eq_lb1 = ();
	foreach my $gb2 (@{$gblks_w_lb_eq_lbd1}) { # Main foreach-loop (over gap-blocks whose left-bounds equal $lb1).
	    if ($gb2 == $gb_sbj) {
		push @new_gblks_w_lb_eq_lb1, $gb2;
		next;
	    }

	    my $rel = $rels_w_sbj->[$gb2];

	    unless (($rel eq '>') or ($rel eq '>(ch)') or (($rel eq '=') and ($gb_sbj < $gb2)) ) {
			# Here, we are NOT interested in this type of gap-block. #
		push @new_gblks_w_lb_eq_lb1, $gb2;
		next;
	    }

	    my $bds_gb2 = $bds_gblocks->[$gb2];
	    my ($lbd2, $rbd2) = @{$bds_gb2};

	    unless ($rbd1 < $rbd2) {
		print STDERR "[ERROR 'vretract_left_ends_vincluded_gblks'] The right-bound, $rbd1, of the subject (the $gb_sbj th gap-block) is NOT on the left of the right-bound, $rbd2, of the $gb2 th block, which is vertically included in the subject: $! !!\n\n";
		return undef;
	    }

	    $ct_vretracted++;

	    	# Update $lbd2.
	    $lbd2 = $rbd1 +1;
	    $bds_gb2->[0] = $lbd2;
	    push @{$gblks_w_spec_lb->[$lbd2]}, $gb2;  

	    my $size2 = $gblock_sizes->[$gb2];

	    unless ($lbd2 + $size2 - 1 < $rbd2) { next; } # The left-enf of $gb2 canNOT be retracted FURTHER.

		# Attempt to FURTHER retract the right-end of $gb2.
			# (NOTE added on 2019/03/29:
			#    Come to think of it, the subsequent part may be superfluous.
			#    Just in case, however, I will leave it as it is.)

	    my $ct_vretracting2 = vretract_left_end_sbj_gblk ($gb2, @{$bds_gblocks}, @{$inter_block_relations}, @{$gblock_sizes}, @{$gblks_w_spec_lb});
	    unless (defined $ct_vretracting2) {
		print STDERR "[ERROR 'vretract_left_ends_vincluded_gblks'] The subroutine 'vretract_left_end_sbj_gblk' FAILED on the $gb2 th gap-block, which is vertically included in the subject (i.e., the $gb_sbj th gap-block): $! !!\n\n";
		return undef;
	    }

	} # End of the main foreach-loop (over gap-blocks whose left-bounds equal $lb1).

	if ($ct_vretracted >0) {
		# Update the set whose left-bounds equal $lbd1.
	    $gblks_w_spec_lb->[$lbd1] = \@new_gblks_w_lb_eq_lb1;
	}

	return $ct_vretracted;

}	# END of the subroutine 'vretract_left_ends_vincluded_gblks'. #


# mk_sets_vert_equiv_gblks:
#	This subroutine creates a list of sets of vertically equivalent gap-blocks.
#	[CREATED on 2019/03/30.]
#
# INPUTS: (\%br2up_down2blocks, \%br2eq), where
#
#	%br2up_down2blocks = ($branch => {$up_or_down => \@blocks, ...}, ...),
#	where @blocks lists the gap-blocks each of which is separated by, and on the $up_or_down (= 'U' or 'L') side of,
#	the branch with ID $branch;
#	[Typically, \%br2up_down2blocks is an output of, e.g., 'mk_br2up_down2blocks' above or 'search_for_coll_cmpl_gblks' above];
#
#	%br2eq = ($branch_ID => $equivalent_branch_ID, ...) for the branches with their equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
# OUTPUTS: \@sets_vequiv_gblks, where
#
#	@{$sets_vequiv_gblks[$k]} lists the indices of the gap-blocks in the $k th set of vertically equivalent gap-blocks.
#
sub mk_sets_vert_equiv_gblks (\%\%) {

	my ($br2up_down2blocks, $br2eq) = @_;

	my %skipped; # = ($br => 1, ...) for the branches that MUST be SKIPPED.

	my @pre_sets_vequiv_gblks = ();

	foreach my $br (sort {$a <=> $b} keys %{$br2up_down2blocks}) { # Main outer foreach-loop (over branches).

	    if (defined $skipped{$br}) { next; }

	    my $up_down2blocks_br = $br2up_down2blocks->{$br};

	    my $blocks_up_br = $up_down2blocks_br->{'U'};
	    my @blocks1 = (defined $blocks_up_br) ? @{$blocks_up_br} : ();
	    my $blocks_dw_br = $up_down2blocks_br->{'L'};
	    my @blocks2 = (defined $blocks_dw_br) ? @{$blocks_dw_br} : ();

		# Examine the existence of the equivalent branch. #

	    my $eq_br = $br2eq->{$br};
	    if (defined $eq_br) {

		$skipped{$eq_br} = 1;

		my $up_down2blocks_eq = $br2up_down2blocks->{$eq_br};

		if (defined $up_down2blocks_eq) {

		    my $blocks_up_eq = $up_down2blocks_eq->{'U'};
		    if (defined $blocks_up_eq) { push @blocks2, @{$blocks_up_eq}; }

		    my $blocks_dw_eq = $up_down2blocks_eq->{'L'};
		    if (defined $blocks_dw_eq) { push @blocks1, @{$blocks_dw_eq}; }

		} # End of "if (defined $up_down2blocks_eq) {...}".
	    } # End of "if (defined $eq_br) {...}".

	    foreach my $set_blks (\@blocks1, \@blocks2) { # Main inner foreach-loop (over the sets of gap-blocks).

		if (@{$set_blks} < 2) { next; }

		my @std_set_blks = sort {$a <=> $b} @{$set_blks}; # This should effectively sort the gap-blocks in ascending order of their initial left-ends.

		push @pre_sets_vequiv_gblks, \@std_set_blks;

	    }  # End of the main inner foreach-loop (over the sets of gap-blocks).

	}	# End of the main outer foreach-loop (over branches).

	my @sets_vequiv_gblks = sort {$a->[0] <=> $b->[0]} @pre_sets_vequiv_gblks;

	return \@sets_vequiv_gblks;

}	# END of the subroutine 'mk_sets_vert_equiv_gblks'. #


# mk_sets_swappable_gblks:
#	This subroutine creates a lists of sets of "swappable" gap-blocks.
#	The members of each set are vertically equivalent to each other and have the same size.
#	[CREATED on 2019/03/30.]
#
# INPUTS: (\@sets_vequiv_gblks, \@gblock_sizes), where
#
#	@{$sets_vequiv_gblks[$k]} lists the indices of the gap-blocks
#	in the $k th set of vertically equivalent gap-blocks;
#	[Typically, \@sets_vequiv_gblks is an output of 'mk_sets_vert_equiv_gblks' above];
#	
#	$gblock_sizes[$gb] gives the size of the $gb th gap-block;
#	[Typically, \@gblock_sizes is the output of 'mk_list_gblock_sizes' above];
#
# OUTPUTS: (\@sets_swappable_gblks), where
#
#	@{$sets_swappable_gblks[$j]} lists the indices of the gap-blocks
#	in the $j th set of swappable gap-blocks.
#
sub mk_sets_swappable_gblks (\@\@) {

	my ($sets_vequiv_gblks, $gblock_sizes) = @_;

	my @sets_swappable_gblks = ();

	foreach my $set_veq_gblks (@{$sets_vequiv_gblks}) { # Main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	    my %size2gblks;
	    foreach my $gb (@{$set_veq_gblks}) { # 1st inner foreach-loop (over constituent gap-blocks).

		my $size = $gblock_sizes->[$gb];
		my $gblks = $size2gblks{$size};
		unless (defined $gblks) { $gblks = $size2gblks{$size} = []; }
		push @{$gblks}, $gb;

	    } # End of the 1st inner foreach-loop (over constituent gap-blocks).

	    my @sb_sets = ();
	    foreach my $size (sort {$a <=> $b} keys %size2gblks) { # 2nd inner foreach-loop (over gap-sizes).

		my $gblks =  $size2gblks{$size};
		if (@{$gblks}>1) { push @sb_sets, $gblks; }

	    } # End of the 2nd inner foreach-loop (over gap-sizes).

	    my @std_sb_sets = sort {$a->[0] <=> $b->[0]} @sb_sets;
	    push @sets_swappable_gblks, @std_sb_sets;

	} # End of the main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	return \@sets_swappable_gblks;

}	# END of the subroutine 'mk_sets_swappable_gblks'. #



# br_chk_ovlp_bw_vequiv_gblks:
#	This subroutine BRIEFLY examines whether some vertically equivalent gap-blocks overlap one another or not,
#	using @sets_vequiv_gblks, which is an output of 'mk_sets_vert_equiv_gblks' above.
#	[CREATED on 2019/03/30; RENAMED from 'chk_ovlp_bw_vequiv_gblks' on 2019/04/03.]
#
#	NOTE1: Here, the "overlapping" includes "being immediately adjacent" to, i.e., "touching", each other.
#
#	NOTE2: This subroutine does NOT CARE if swappable gap-blocks are in the right order or not.
#		It is the job of 'chk_pos_orders_amg_swpbl_gblks' below to examine that.
#
#	NOTE3: This subroutine IGNORES the effects of gap-blocks vertically including the subject blocks.
#		The 'chk_ovlp_bw_vequiv_gblks' below takes account of such gap-blocks as well.
#
# INPUTS: (\@sets_vequiv_gblks, \@bds_gblocks), where
#
#	@{$sets_vequiv_gblks[$k]} lists the indices of the gap-blocks
#	in the $k th set of vertically equivalent gap-blocks;
#	[Typically, \@sets_vequiv_gblks is an output of 'mk_sets_vert_equiv_gblks' above];
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#	
# OUTPUTS: ($if_overlap), where
#
#	$if_overlap = 1 if any pair of vertically equivalent gap-blocks overlap, or touch, each other;
#		    = 0 if no gap-blocks overlap any vertically equivalent gap-blocks.

sub br_chk_ovlp_bw_vequiv_gblks (\@\@) {

	my ($sets_vequiv_gblks, $bds_gblocks) = @_;

	foreach my $veq_gblks (@{$sets_vequiv_gblks}) { # The main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	    my @std_veq_gblks = sort {$bds_gblocks->[$a][0] <=> $bds_gblocks->[$b][0]} @{$veq_gblks};

	    my $ct_veqs = @{$veq_gblks};

	    my $gb1 = $std_veq_gblks[0];
	    my $rbd1 = $bds_gblocks->[$gb1][1];

	    for (my $k=1; $k < $ct_veqs; $k++) { # Inner for-loop (over constituent gap-blocks).

		my $gb2 = $std_veq_gblks[$k];
		my $bds2 = $bds_gblocks->[$gb2];
		my $lbd2 = $bds2->[0];

		if ($rbd1 + 1 < $lbd2) {

		    $rbd1 = $bds2->[1]; # Update $rbd1.
		} else {
			# $gb1 overlaps, or touches, $gb2
		    return 1;

		} # End of "if ($rbd1 + 1 < $lbd2) {...} else {...}".

	    } # End of the inner for-loop (over constituent gap-blocks).

	} # End of the main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	return 0;

}	# END of the subroutine 'br_chk_ovlp_bw_vequiv_gblks'. #


# chk_ovlp_bw_vequiv_gblks:
#	This subroutine examines whether some vertically equivalent gap-blocks overlap one another or not,
#	using @sets_vequiv_gblks, which is an output of 'mk_sets_vert_equiv_gblks' above.
#	It was created by AUGMENTING its predecessor, 'br_chk_ovlp_bw_vequiv_gblks'.
#	[CREATED on 2019/04/03.]
#
#	NOTE1: Here, the "overlapping" includes "being immediately adjacent" to, i.e., "touching", each other.
#
#	NOTE2: This subroutine does NOT CARE if swappable gap-blocks are in the right order or not.
#		It is the job of 'chk_pos_orders_amg_swpbl_gblks' below to examine that.
#
#	NOTE3: This subroutine takes account of the effects of gap-blocks vertically including the subject blocks. (The predecessor, 'br_chk_ovlp_bw_vequiv_gblks', does NOT.)
#
#	[MAJOR REVISION on 2019/05/27.]
#	We replaced the subroutine component, 'ct_necessary_shifts',
#	with 'smlt2_extend_left_end_to_left' or 'smlt2_extend_right_end_to_right',
#	whenever possible.
#
# INPUTS: (\@sets_vequiv_gblks, \@bds_gblocks, \@inter_block_relations, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	@{$sets_vequiv_gblks[$k]} lists the indices of the gap-blocks
#	in the $k th set of vertically equivalent gap-blocks;
#	[Typically, \@sets_vequiv_gblks is an output of 'mk_sets_vert_equiv_gblks' above];
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#	
# OUTPUTS: ($if_overlap), where
#
#	$if_overlap = 1 if any pair of vertically equivalent gap-blocks overlap, or touch, each other;
#		    = 0 if no gap-blocks overlap any vertically equivalent gap-blocks.

sub chk_ovlp_bw_vequiv_gblks (\@\@\@\@\@) {

	my ($sets_vequiv_gblks, $bds_gblocks, $inter_block_relations, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	foreach my $veq_gblks (@{$sets_vequiv_gblks}) { # The main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	    my %skipped;
	    foreach my $gb (@{$veq_gblks}) { $skipped{$gb} = 1; }

	    my @std_veq_gblks = sort {$bds_gblocks->[$a][0] <=> $bds_gblocks->[$b][0]} @{$veq_gblks};

	    my $ct_veqs = @{$veq_gblks};

	    my $gb1 = $std_veq_gblks[0];
	    my $rbd1 = $bds_gblocks->[$gb1][1];

	    for (my $k=1; $k < $ct_veqs; $k++) { # Inner for-loop (over constituent gap-blocks).

		my $gb2 = $std_veq_gblks[$k];
		my $bds2 = $bds_gblocks->[$gb2];
		my $lbd2 = $bds2->[0];

		if ($rbd1 + 1 < $lbd2) {

			# ADDED in augmentation (on 2019/04/03).
		    my $gblks_w_le_med = $gblks_w_spec_lb->[$rbd1+1];
		    my $gblks_w_re_med = $gblks_w_spec_rb->[$lbd2-1];
		    if ((@{$gblks_w_le_med}>0) and (@{$gblks_w_re_med}>0)) {

				# Modified in major revision on 2019/05/27. (1) #

			my ($if_reached, $lbd2_new, $rbd2_new) = smlt2_extend_left_end_to_left ($gb2, $rbd1+1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			unless (defined $if_reached) {
			    print STDERR "[ERROR 'chk_ovlp_bw_vequiv_gblks'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb2 th gap-block: $! !\n\n";
			    return undef;
			}

			if ($if_reached) { return 1; } # $gb2 virtually touches $gb1.

=OBSOLETE as of 2019/05/27. (1)
			my $ct_shifts = ct_necessary_shifts ($lbd2, $rbd1+1, $gb2, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			unless (defined $ct_shifts) {
			    print STDERR "[ERROR 'chk_ovlp_bw_vequiv_gblks'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block: $! !\n\n";
			    return undef;
			}

			if ($ct_shifts == 0) { return 1; } # $gb2 virtually touches $gb1.
=cut # End of "OBSOLETE as of 2019/05/27. (1)"

				# End of "Modified in major revision on 2019/05/27. (1)" #

		    } # End of "if ((@{$gblks_w_le_med}>0) and (@{$gblks_w_re_med}>0)) {...}".

		    $gb1 = $gb2; # Update $gb1.

			# END of "ADDED in augmentation (on 2019/04/03)."

		    $rbd1 = $bds2->[1]; # Update $rbd1.
		} else {
			# $gb1 overlaps, or touches, $gb2
		    return 1;

		} # End of "if ($rbd1 + 1 < $lbd2) {...} else {...}".

	    } # End of the inner for-loop (over constituent gap-blocks).

	} # End of the main outer foreach-loop (over the sets of vertically equivalent gap-blocks).

	return 0;

}	# END of the subroutine 'chk_ovlp_bw_vequiv_gblks'. #


# chk_pos_orders_amg_swpbl_gblks:
#	This subroutine checks the positional orders among swappable gap-blocks,
#	using @sets_swappable_gblks, which is an output of 'mk_sets_swappable_gblks'.
#	[CREATED on 2019/03/30.]
#
#	NOTE: This subroutine does NOT CARE if the gap-blocks overlap each other;
#		It is the job of 'chk_ovlp_bw_vequiv_gblks' above to examine the overlap (between vertically equivalent gap-blocks in general).
#
# INPUTS: (\@sets_swappable_gblks, \@bds_gblocks), where
#
#	@{$sets_swappable_gblks[$j]} lists the indices of the gap-blocks
#	in the $j th set of swappable gap-blocks;
#	[Typically, \@sets_swappable_gblks is an output of 'mk_sets_swappable_gblks' above];
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#	
# OUTPUTS: $if_right_order, where
#
#	$if_right_order = 1 if ALL swappable gap-blocks are in the right positional order in EVERY set,
#			= 0 if the right positional order is violated in AT LEAST ONE set.
#
sub chk_pos_orders_amg_swpbl_gblks (\@\@) {

	my ($sets_swappable_gblks, $bds_gblocks) = @_;

	foreach my $swpbls (@{$sets_swappable_gblks}) { # Main outer foreach-loop (over the sets of swappable gap-blocks).

	    my $ct_swpbls = @{$swpbls};
	    my $gb1 = $swpbls->[0];
	    my $lbd1 = $bds_gblocks->[$gb1][0];
	    for (my $i=1; $i < $ct_swpbls; $i++) { # Inner foreach-loop (over the constituent gap-blocks).

		my $gb2 = $swpbls->[$i];
		my $lbd2 = $bds_gblocks->[$gb2][0];

		if ($lbd1 < $lbd2) {
		    $lbd1 = $lbd2; # Update $lbd1.
		} else {
		    return 0;
		}

	    } # End of the inner foreach-loop (over the constituent gap-blocks).

	} # End of the main outer foreach-loop (over the sets of swappable gap-blocks).

	return 1;

}	# END of the subroutine 'chk_pos_orders_amg_swpbl_gblks'. #


# chk_chng_gptrn_topology_psm_R:
#	This subroutine examines whether or not the current "shift" TO THE RIGHT (i.e., w/ $sh > 0)
#	has changed the gap-pattern topology of the alignment
#	(concerning parsimonious ancestral gap-states).
#	[CREATION STARTED on 2019/04/07, ENDED on 2019/04/08.]
#
#	[MAJOR REVISION on 2019/04/12:] In each place where possible, 'ct_necessary_shifts' was replaced with 'smlt2_extend_right_end_to_right' or 'smlt2_extend_left_end_to_left', to somewhat speed up the process. Besides, the subroutine also regards the topology as changed when the subject block now touches its "complement" on its left. (The same pattern will NOT be regarded as a topological change when $rel = '=', '>(ch)' or '<(pa)'.)
#
#	[MAJOR REVISION on 2019/04/14:] Now, the subroutine can output some diagnostic outputs if necessary.
#
#
#	NOTE A: This subroutine is paired with 'chk_chng_gptrn_topology_psm_L' below.
#
#	NOTE B: "Topology" of the gap-pattern is considered as "changed" if the resulting sets of
#	ancestral gap-states under consideration undergoes larger changes than mere positional shifts of some "insertion"s/"deletion"s.
#
#	NOTE1: Blocks with 'ONN' relations will cause NO topological change, as long as we employ the current coordinate system;
#	NOTE2: Likewise, blocks with 'ONCS' relations will cause NO topological change, as long as we employ the current coordinate system, and as far as parsimonious ancestral gap-states are concerned;
#	NOTE3: As far as parsimonious ancestral gap-states are concerned, blocks with '>' or '<' but NOT with '>(ch)' or '<(pa)' relations will cause NO topological change.
#
# INPUTS: ($gb_sbj, $lbd_sbj_bf, $rbd_sbj_bf, \@bds_gblocks, \@inter_block_relations, \@interfering_blocks, \@interfering_blocksets, \@cmpl_interfering_blocksets, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd_sbj_bf, $rbd_sbj_bf) give the left- and right-bounds of the subject gap-block BEFORE the current "shift";
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$interfering_blocks[$gb]} = (\@blocks, \@relations)
#	stores information on blocks that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	$blocks[$k] is the index (or rank) of the $k th interfering block,
#	$relations[$k] is the relation of the $k th interfering block against the $gb th block;
#	[Typically, \@interfering_blocks is the output of 'mk_list_interfering_blocks' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$interfering_blocksets[$gb]} = (\@blocksets, \@relations) 
#	stores information on block-SETs that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	@{$blocksets[$k]} lists the indices (or ranks) of the gap-blocks constiruting the $k th interfering block-set,
#	$relations[$k] is the relation of the $k th interfering block-set against the $gb th block;
#	[NOTE: The gap-blocks in each @{$blocksets[$k]} ALWAYS form a (complementary-)monophyletic group];
#	[Typically, \@interfering_blocksets is an output of 'mk_list_interfering_blocksets' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$cmpl_interfering_blocksets[$gb]} = (\@cmpl_blocksets, \@relations)
#	records block-blockset pairs in @interfering_blocksets in which
#	the $gb th block a constituent of @blockset = @{$interfering_blocksets[$gb2]->[0][$k2]} for some $gb2 and $k2.
#	(But it is empty if the $gb th block is NOT a constituent of ANY block-sets.)
#	We will use the following convension:
#	@{$cmpl_blocksets[$k]} = ($gb2, @blockset above with $gb removed)
#	and
#	$relations[$k] = the "complement" of $interfering_blocksets[$gb2]->[1][$k2].
#	Thus, the relation should be:
#	'S' if the corresponding relation in @interfering_blockset is 'S',
#	'<(pa)' if the corresponding relation in @interfering_blockset is '>(ch)'.
#	[Typically, \@cmpl_interfering_blocksets is the output of 'mk_list_cmpl_interfering_blocksets' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $if_change	= 1 if the "shift" has changed the topology,
#			= 0  otherwise.
#
sub chk_chng_gptrn_topology_psm_R ($$$\@\@\@\@\@\@\@) {

	my ($gb_sbj, $lbd_sbj_bf, $rbd_sbj_bf, $bds_gblocks, $inter_block_relations, $interfering_blocks, $interfering_blocksets, $cmpl_interfering_blocksets, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my ($lbd_sbj_af, $rbd_sbj_af) = @{$bds_gblocks->[$gb_sbj]};
	my $rlv_interfering_blocks = $interfering_blocks->[$gb_sbj];
	my $rlv_intrfr_blksets = $interfering_blocksets->[$gb_sbj];
	my $rlv_cmpl_intrfr_blksets = $cmpl_interfering_blocksets->[$gb_sbj];

    (print STDOUT "< Subroutine 'chk_chng_gptrn_topology_psm_R' on \$gb_sbj = $gb_sbj, w/ (lbd, rbd)_bf = ($lbd_sbj_bf, $rbd_sbj_bf) & (lbd, rbd)_af = ($lbd_sbj_af, $rbd_sbj_af) >\n\n") if ($TEST_LEVEL_ccgpt_p >0); # ADDED in REVISION on 2019/04/14.


	my $if_change = 0; # ADDED in REVISION on 2019/04/14.

	if (@{$rlv_interfering_blocks}>0) { # (1) Examine the interfering blocks.


	    my ($intrfr_blks, $relations) = @{$rlv_interfering_blocks};
	    my $ct_intrfr_blks = @{$intrfr_blks};

    (print STDOUT "\n... There are $ct_intrfr_blks gap-blocks interfering with $gb_sbj ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISION on 2019/04/14.
    (print STDOUT join ("\t", 'Intrfr_gblk', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.

	    for (my $k=0; $k < $ct_intrfr_blks; $k++) { # 1st outer for-loop (over interfering blocks).

		my $gb2 = $intrfr_blks->[$k];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
		my $rel = $relations->[$k];

		if ($rel eq 'S') { # effective siblings

		    if (($lbd2 < $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block is now non-nested with $gb2.
			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'a'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			} elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'b'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}
		    } elsif (($lbd_sbj_af == $lbd2) and ($rbd_sbj_af <= $rbd2)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'c'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    } elsif (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af == $rbd2)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'd'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} elsif ( ($rel eq 'Cp') # $gb2 is complementary to $gb_sbj.
			or ($rel eq '>(ch)') # $gb2 is an "effective" child of $gb_sbj.
			or ($rel eq '<(pa)') # $gbs is an "effective" parent of $gb_sbj.
			or ($rel eq '=') # $gb2 is vertically equivalent to $gb_sbj.
			) {

		    if ($rbd_sbj_af + 1 == $lbd2) { # The subject block, which was separated from $gb2 by one column, now touches the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'e'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    } elsif ($rbd2 + 1 == $lbd_sbj_bf) { # The subject block, which touched $gb2, is now separated from the latter by one column.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'f'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif ( ($rel eq 'Cp') and (($rbd2 == $lbd_sbj_bf) or ($rbd2 + 1 == $lbd_sbj_af)) ) { # The subject block now touches its complement, $gb2, which is on the left. # ADDED on 2019/04/12. (1)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'g'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif ($rbd_sbj_af + 1 < $lbd2) { # The subject block now appears separated from $gb2, which is on the right.

			if ((@{$gblks_w_spec_lb->[$rbd_sbj_af+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {

				### MODIFIED on 2019/04/12. (1) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_right_end_to_right ($gb_sbj, $lbd2-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb_sbj th gap-block against $gb2 (A): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'A'), "\n";
			        $if_change = 1;
			        next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }

=REPLACED with a better method. (1)
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_af, $lbd2-1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (A): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { return 1; }
=cut # END of "REPLACED with a better method. (1)"
				### END of "MODIFIED on 2019/04/12. (1)" ###
			}

		    } elsif ($rbd2 + 1 < $lbd_sbj_bf) { # The subject block now appeared separated from $gb2, which is on the left. 

			if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_bf-1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_bf, $rbd2+1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (B): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'B'), "\n";
			        $if_change = 1;
			        next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }
			}

		    } # End of "if ($rbd_sbj_af + 1 == $lbd2) {...} elsif (...) {...} .... {...}"

		} # End of "if ($rel eq 'S') {...} elsif (($rel eq 'Cp') or ...) {...}".

	    } # End of the 1st outer for-loop (over interfering blocks).

	} # End of "if (@{$rlv_interfering_blocks}>0) {...}".


	if (@{$rlv_intrfr_blksets}>0) { # (2) Examine the interfering block-sets.

	    my ($intrfr_blksets, $relations2) = @{$rlv_intrfr_blksets};
	    my $ct_intrfr_blksets = @{$intrfr_blksets};

    (print STDOUT "\n... There are $ct_intrfr_blksets gap-block-sets interfering with $gb_sbj ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISIOn on 2019/04/14.
    (print STDOUT join ("\t", 'Intrfr,gap,blks', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.

	    for (my $k = 0; $k < $ct_intrfr_blksets; $k++) { # 2nd outer for-loop (over interfering block-sets).
		my $blset = $intrfr_blksets->[$k];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$blset->[0]]};
		my $size_blset = @{$blset};
		my $rel = $relations2->[$k];

		my $cnct_blset = join (',', @{$blset}); # ADDED in REVISION on 2019/04/14.

		if ($rel eq 'S') {

		    for (my $j=1; $j <$size_blset; $j++) { # 1st inner for-loop (over constituent blocks in the block-set).
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$blset->[$j]]};
			if ($lbd2 < $lbd3) { $lbd2 = $lbd3; }
			if ($rbd3 < $rbd2) { $rbd2 = $rbd3; }
			if ($rbd2 < $lbd2) { last; } # Added on 2019/04/08.

		    } # End of the 1st inner for-loop (over constituent blocks in the block-set).

		    if ($rbd2 < $lbd2) { next; }

		    my $if_test = 0;
		    if ( (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) 
			or (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) ) {
			$if_test = 1;
		    } else {

			my $if_min_blks = 0;
			for (my $j = 0; $j < $size_blset; $j++) { # 2nd inner for-loop (over constituent blocks in the block-set).
			    my $gb3 = $blset->[$j];
			    my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
			    if (($lbd3 == $lbd2) and ($rbd3 == $rbd2)) {
				$if_min_blks =1;
				last;
			    }
			} # End of the 2nd inner for-loop (over constituent blocks in the block-set).

			if ($if_min_blks) { $if_test = 1; }

		    } # End of "if ( (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) or (...)) {...} else {...}".

		  if ($if_test) {

			# Do the identical tests as for "if ($rel eq 'S') {...}" in (1). #

		    if (($lbd2 < $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block is now non-nested with $gb2.
			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'h'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			} elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'i'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}
		    } elsif (($lbd_sbj_af == $lbd2) and ($rbd_sbj_af <= $rbd2)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'j'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af == $rbd2)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'k'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		  } # End of "if ($if_test) {...}".

		} elsif ($rel eq '>(ch)') {

		    my $ct_topo_chngs = my $ct_including = 0;

		    for (my $j=0; $j < $size_blset; $j++) { # 3rd inner for-loop (over constituent blocks).
			my $gb3 = $blset->[$j];
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};

			if (($lbd3 <= $rbd_sbj_bf +1) and ($lbd_sbj_af <= $rbd3 + 1)) { # During the "shift", $gb3 keeps touching $gb_sbj.
			    $ct_including++;

			} elsif ($rbd_sbj_af +1 == $lbd3) { # The subject block, which was separated from $gb3 by one column, now touches the latter.
			    $ct_topo_chngs++;
			} elsif ($rbd3 +1 == $lbd_sbj_bf) { # The subject block, which touched $gb3, is now separated from the latter by one column.
			    $ct_topo_chngs++;
			} elsif ($rbd_sbj_af + 1 < $lbd3) { # The subject block now appears separated from $gb3, which is on the right.
			  if ((@{$gblks_w_spec_lb->[$rbd_sbj_af+1]}>0) and (@{$gblks_w_spec_rb->[$lbd3-1]}>0)) {
				### MODIFIED on 2019/04/12. (2) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_right_end_to_right ($gb_sbj, $lbd3-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb_sbj th gap-block against $gb3 (C): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) { 
				$ct_topo_chngs++;
			    } else {
				last;
			    }
=REPLACED with a better method. (2)
			    my %skipped = ($gb_sbj => 1, $gb3 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_af, $lbd3-1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block, which belongs to the $k th interfering block-set (C): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts > 0) { 
				last;
			    } else {
				$ct_topo_chngs++;
			    }
=cut # END of "REPLACED with a better method. (2)"
			  } else {
				last;
			  } # End of "if ((@{$gblks_w_spec_lb->[$rbd_sbj_af+1]}>0) and (@{$gblks_w_spec_rb->[$lbd3-1]}>0)) {...} else {...}".

			} elsif ($rbd3 +1 < $lbd_sbj_bf) { # The subject block appeared separated from $gb3, which is on the left.
			  if ((@{$gblks_w_spec_lb->[$rbd3+1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_bf-1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb3 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_bf, $rbd3+1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block, which belongs to the $k th interfering block-set (D): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts > 0) {
				last;
			    } else {
				$ct_topo_chngs++;
			    }
			  } else {
				last;
			  } # End of "if ((@{$gblks_w_spec_lb->[$rbd3+1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_bf-1]}>0)) {...} else {...}".

			} else {
			    last; # Because we are sure that there should be NO topological change between the block-set and $gb_sbj, leave the loop here.
			}

		    } # End of the 3rd inner for-loop (over constituent blocks).

		    if (($ct_topo_chngs>0) and ($ct_topo_chngs + $ct_including == $size_blset)) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'l'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} # End of "if ($rel eq 'S') {...} elsif ($rel eq '>(ch)') {...}".

	    } # End of the 2nd outer for-loop (over interfering block-sets).

	} # End of "if (@{$rlv_intrfr_blksets}>0) {...}".


	if (@{$rlv_cmpl_intrfr_blksets}>0) { # (3) Examine the complements of the interfering block-sets.

			# Restarted on 2019/04/08. #

		## (The following is from p.49 of "suppl3_blueprint1_ANEX.xxx.pdf".) ##

	    my ($intrfr_cblksets, $relations3) = @{$rlv_cmpl_intrfr_blksets};
	    my $ct_intrfr_cbss = @{$intrfr_cblksets};

    (print STDOUT "\n... There are $ct_intrfr_cbss gap-block-sets including $gb_sbj and with an interfering partner ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISIOn on 2019/04/14.
    (print STDOUT join ("\t", 'Cmpl,intrfr,gap,blks', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.


	    for (my $k =0; $k <$ct_intrfr_cbss; $k++) { # 3rd outer for-loop (over complements of interfering block-sets).

		my $cblkset = $intrfr_cblksets->[$k];
		my $size_cblkset = @{$cblkset};
		my $gb2 = $cblkset->[0];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
		my $rel = $relations3->[$k];

		my $cnct_cblkset = join (',', @{$cblkset}); # ADDED in REVISION on 2019/04/14.

		if ($rel eq 'S') { # $gb2 is an "effective sibling" of the block-set.

		    my ($lbd3, $rbd3) = @{$bds_gblocks->[$cblkset->[1]]};
		    for (my $j=2; $j <$size_cblkset; $j++) { # 1st inner for-loop (over constituent blocks).
			my ($lbd5, $rbd5) = @{$bds_gblocks->[$cblkset->[$j]]};
			if ($lbd3 < $lbd5) { $lbd3 = $lbd5; }
			if ($rbd5 < $rbd3) { $rbd3 = $rbd5; }
			if ($rbd3 < $lbd3) { last; }
		    } # End of the 1st inner for-loop (over constituent blocks).

		    if ($rbd3 < $lbd3) { next; } # Topology should NOT change!!

		    my ($max_sbj_lbd, $min_sbj_lbd) = ($lbd_sbj_af < $lbd_sbj_bf) ?
				($lbd_sbj_af, $lbd_sbj_bf) : ($lbd_sbj_bf, $lbd_sbj_af);

		    my ($max_sbj_rbd, $min_sbj_rbd) = ($rbd_sbj_af > $rbd_sbj_bf) ?
			        ($rbd_sbj_af, $rbd_sbj_bf) : ($rbd_sbj_bf, $rbd_sbj_af);

		    if (($lbd3 <= $lbd2) and ($rbd2 <= $rbd3)) { # $gb2 is horizontally included in the block-set (excluding $gb_sbj). => Focus on the relationship between $gb2 and $gb_sbj. (We do NOT know yet, which of $gb_sbj and $gb2 horizontally includes the other.)

			# Do tests identical to those above for ($rel eq 'S') in (1). #

			if (($lbd2 < $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block is now non-nested with $gb2.
			    if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-1'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;

			    } elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-2'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }
			} elsif (($lbd_sbj_af == $lbd2) and ($rbd_sbj_af <= $rbd2)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-3'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			} elsif (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af == $rbd2)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-5'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $max_sbj_lbd) and ($max_sbj_rbd <= $rbd3)) { # BOTH before and after the change, $gb_sbj is horizontally included in the block-set (excluding $gb_sbj). => Focus on the relationship between $gb2 and $gb_sbj. (In this case, $gb_sbj will NOT include $gb2 horizontally (because $gb2 did NOT satisfy the previous conditions).)

			if (($lbd2 < $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # $gb_sbj is now non-nested with $gb2.
			    if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # $gb_sbj was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'n-1'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }
			} elsif (($lbd_sbj_af == $lbd2) and ($rbd_sbj_af <= $rbd2)) { # $gb_sbj, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'n-2'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd3)) { # After the change, $gb_sbj is horizontally included in the block-set (excluding $gb_sbj), but it was NOT before the change.

			if (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) { # $gb2 also includes $gb_sbj afer the change. (We don't care about their relation before the change.)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'o'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd3)) { # Before the change, $gb_sbj was horizontally included in the block-set (excluding $gb_sbj), but it is NOT after the change.

			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # $gb2 also included $gb_sbj before the change. (We don't care about their relation after the change.)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'p'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($min_sbj_lbd <= $lbd3) and ($rbd3 <= $min_sbj_rbd)) { # Both before and after the change, $gb_sbj horizontally includes the intersection of the block-set (excluding $gb_sbj). => Topology will NEVER change, no matter what.

			next;

		    } elsif ( (($lbd_sbj_af <= $lbd3) and ($rbd3 <= $rbd_sbj_af)) # After the change, $gb_sbj horizontally includes the intersection of the block-set (excluding $gb_sbj), but it did NOT before the change.
				or (($lbd_sbj_bf <= $lbd3) and ($rbd3 <= $rbd_sbj_bf)) # Before the change, $gb_sbj horizontally included the intersection of the block-set (excluding $gb_sbj), but it does NOT after the change.
		    ) {

			unless (($lbd2 <= $lbd3) and ($rbd3 <= $rbd2)) { next; } # $gb2 does NOT include the intersection of the block-set (excluding $gb_sbj). => Topology will NOT change!!

			my $if_min_blks=0;
			for (my $j=1; $j < $size_cblkset; $j++) { # 2nd inner for-loop (over constituent blocks).
			    my ($lbd5, $rbd5) = @{$bds_gblocks->[$cblkset->[$j]]};
			    if (($lbd3 == $lbd5) and ($rbd3 == $rbd5)) {
				$if_min_blks=1;
				last;
			    }
			} # End of the 2nd inner for-loop (over constituent blocks).

			if ($if_min_blks) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'q'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } # End of "if (($lbd3 <= $lbd2) and ($rbd2 <= $rbd3)) {...} elsif (...) ... elsif ((($lbd_sbj_af <= $lbd3) and ($rbd3 <= $rbd_sbj_af)) or (($lbd_sbj_bf <= $lbd3) and ($rbd3 <= $rbd_sbj_bf)) ) {...}".

		} elsif ($rel eq '<(pa)') { # $gb2 is an "effective parent" of the block-set.

			# Do tests identical to those above for ($rel eq '<(pa)') in (1).

		    my $if_topo_chng = 0;
		    if ($rbd_sbj_af + 1 == $lbd2) { # The subject block, which was separated from $gb2 by one column, now touches the latter.
			$if_topo_chng = 1;
		    } elsif ($rbd2 + 1 == $lbd_sbj_bf) { # The subject block, which touched $gb2, is now separated from the latter by one column.
			$if_topo_chng = 1;
		    } elsif ($rbd_sbj_af + 1 < $lbd2) { # The subject block now appears separated from $gb2, which is on the right.

			if ((@{$gblks_w_spec_lb->[$rbd_sbj_af+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {
				### MODIFIED on 2019/04/12. (3) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_right_end_to_right ($gb_sbj, $lbd2-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb_sbj th gap-block against $gb2 (E): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) { $if_topo_chng = 1; }
=REPLACED with a better method. (3)
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_af, $lbd2-1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (E): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { $if_topo_chng = 1; }
=cut # END of "REPLACED with a better method. (3)"
			}

		    } elsif ($rbd2 + 1 < $lbd_sbj_bf) { # The subject block now appeared separated from $gb2, which is on the left. 

			if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_bf-1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_bf, $rbd2+1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (F): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { $if_topo_chng = 1; }
			}
		    } # End of "if ($rbd_sbj_af + 1 == $lbd2) {...} elsif (...) {...} .... {...}"

		    unless ($if_topo_chng) { next; } # NO topological change in conjunction with this block-set!!

			# Check if all the remaining blocks in the block-set are horizontally "connected with" $gb2. #

		    my %skipped = ($gb_sbj => 1);
		    foreach my $gb3 (@{$cblkset}) { $skipped{$gb3} = 1; }

		    for (my $j=1; $j<$size_cblkset; $j++) {  # 3rd inner for-loop (over constituent blocks).

			my $gb3 = $cblkset->[$j]; # A BUG fixed on 2019/04/17!!
			##my $gb3 = $cblkset->[$k];
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};

			if ($rbd3 < $lbd2-1) { # $gb3 appears separated from $gb2 on its right. 

			    if ((@{$gblks_w_spec_lb->[$rbd3+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {
				### MODIFIED on 2019/04/12. (5) ###
				my ($if_reached, $lbd3_new, $rbd3_new) = smlt2_extend_right_end_to_right ($gb3, $lbd2-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
				unless (defined $if_reached) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb3 th gap-block against $gb2 (E): $! !\n\n";
				    return undef;
				}
				if ($if_reached) { next; } # Actually NOT separated.
=REPLACED with a better method. (5)
				my $ct_shifts = ct_necessary_shifts ($rbd3, $lbd2-1, $gb3, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
				unless (defined $ct_shifts) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block (G): $! !\n\n";
				    return undef;
				}
				if ($ct_shifts == 0) { next; } # Actually NOT separated.
=cut # END of "REPLACED with a better method. (5)"
			    }
			    $if_topo_chng = 0;
			    last;

			} elsif ($rbd2+1 < $lbd3) { # $gb3 appears separated from $gb2 on its left.

			    if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd3-1]}>0)) {
				### MODIFIED on 2019/04/12. (6) ###
				my ($if_reached, $lbd3_new, $rbd3_new) = smlt2_extend_left_end_to_left ($gb3, $rbd2+1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
				unless (defined $if_reached) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb3 th gap-block against $gb2 (E): $! !\n\n";
				    return undef;
				}
				if ($if_reached) { next; } # Actually NOT separated.
=REPLACED with a better method. (6)
				my $ct_shifts = ct_necessary_shifts ($lbd3, $rbd2+1, $gb3, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
				unless (defined $ct_shifts) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block (H): $! !\n\n";
				    return undef;
				}
				if ($ct_shifts == 0) { next; } # Actually NOT separated.
=cut # END of "REPLACED with a better method. (6)"
			    }

			    $if_topo_chng = 0;
			    last;

			} # End of "if ($rbd3 < $lbd2-1) {...} elsif ($rbd2+1 < $lbd3) {...}".

		    } # End of the 3rd inner for-loop (over constituent blocks).

		    if ($if_topo_chng) { # Topology DID change in conjunction with this block-set!!

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'r'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} # End of "if ($rel eq 'S') {...} elsif ($rel eq '<(pa)') {...}".

	    } # End of the 3rd outer for-loop (over complements of interfering block-sets).

	} # End of "if (@{$rlv_cmpl_intrfr_blksets}>0) {...}".


			### RESTART FROM HERE (1) ###

    if ($TEST_LEVEL_ccgpt_p>2) {
	print STDOUT "\n< Subroutine 'chk_chng_gptrn_topology_psm_R' Ended (w/ \$if_change = $if_change) />\n\n\n";
	return $if_change;
    }

    (print STDOUT "\n< Subroutine 'chk_chng_gptrn_topology_psm_R' Ended />\n\n\n") if ($TEST_LEVEL_ccgpt_p >0);

	return 0; # NO topological change.

}	# END of the subroutine 'chk_chng_gptrn_topology_psm_R'. #


# chk_chng_gptrn_topology_psm_L:
#	This subroutine examines whether or not the current "shift" TO THE LEFT (i.e., w/ $sh < 0)
#	has changed the gap-pattern topology of the alignment
#	(concerning parsimonious ancestral gap-states).
#	The subroutine was created from 'chk_chng_gptrn_topology_psm_R' by swapping the roles
#	of the left- and right-ends, and by flipping some inequalities.
#	[CREATION STARTED on 2019/04/08, ENDED on 2019/04/16; 
#	MODIFIED on 2019/04/12; REVISED (i.e.,  added test outputs) on 2019/04/14; DEBUGGED on 2019/04/17.]
#
#	NOTE A: This subroutine is paired with 'chk_chng_gptrn_topology_psm_R' above.
#
#	NOTE B: "Topology" of the gap-pattern is considered as "changed" if the resulting sets of
#	ancestral gap-states under consideration undergoes larger changes than mere positional shifts of some "insertion"s/"deletion"s.
#
#	NOTE1: Blocks with 'ONN' relations will cause NO topological change, as long as we employ the current coordinate system;
#	NOTE2: Likewise, blocks with 'ONCS' relations will cause NO topological change, as long as we employ the current coordinate system, and as far as parsimonious ancestral gap-states are concerned;
#	NOTE3: As far as parsimonious ancestral gap-states are concerned, blocks with '>' or '<' but NOT with '>(ch)' or '<(pa)' relations will cause NO topological change.
#
# INPUTS: ($gb_sbj, $lbd_sbj_bf, $rbd_sbj_bf, \@bds_gblocks, \@inter_block_relations, \@interfering_blocks, \@interfering_blocksets, \@cmpl_interfering_blocksets, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	$gb_sbj is the index specifying the subject gap-block;
#
#	($lbd_sbj_bf, $rbd_sbj_bf) give the left- and right-bounds of the subject gap-block BEFORE the current "shift";
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	For the descriptions and formats of the other input variables,
#	see the header of 'chk_chng_gptrn_topology_psm_R' above.
#
# OUTPUTS: $if_change	= 1 if the "shift" has changed the topology,
#			= 0  otherwise.
#
sub chk_chng_gptrn_topology_psm_L ($$$\@\@\@\@\@\@\@) {

	my ($gb_sbj, $lbd_sbj_bf, $rbd_sbj_bf, $bds_gblocks, $inter_block_relations, $interfering_blocks, $interfering_blocksets, $cmpl_interfering_blocksets, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

	my ($lbd_sbj_af, $rbd_sbj_af) = @{$bds_gblocks->[$gb_sbj]};
	my $rlv_interfering_blocks = $interfering_blocks->[$gb_sbj];
	my $rlv_intrfr_blksets = $interfering_blocksets->[$gb_sbj];
	my $rlv_cmpl_intrfr_blksets = $cmpl_interfering_blocksets->[$gb_sbj];

		# The following are COPIED from 'chk_chng_gptrn_topology_psm_R' and MODIFIED on 2019/04/16. #

    (print STDOUT "< Subroutine 'chk_chng_gptrn_topology_psm_L' on \$gb_sbj = $gb_sbj, w/ (lbd, rbd)_bf = ($lbd_sbj_bf, $rbd_sbj_bf) & (lbd, rbd)_af = ($lbd_sbj_af, $rbd_sbj_af) >\n\n") if ($TEST_LEVEL_ccgpt_p >0); # ADDED in REVISION on 2019/04/14.


	my $if_change = 0; # ADDED in REVISION on 2019/04/14.

	if (@{$rlv_interfering_blocks}>0) { # (1) Examine the interfering blocks.


	    my ($intrfr_blks, $relations) = @{$rlv_interfering_blocks};
	    my $ct_intrfr_blks = @{$intrfr_blks};

    (print STDOUT "\n... There are $ct_intrfr_blks gap-blocks interfering with $gb_sbj ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISION on 2019/04/14.
    (print STDOUT join ("\t", 'Intrfr_gblk', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.

	    for (my $k=0; $k < $ct_intrfr_blks; $k++) { # 1st outer for-loop (over interfering blocks).

		my $gb2 = $intrfr_blks->[$k];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
		my $rel = $relations->[$k];

		if ($rel eq 'S') { # effective siblings

		    if (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af < $rbd2)) { # The subject block is now non-nested with $gb2.
			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'a'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			} elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'b'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd2 <= $lbd_sbj_af) and ($rbd2 == $rbd_sbj_af)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'c'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif (($lbd2 == $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'd'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} elsif ( ($rel eq 'Cp') # $gb2 is complementary to $gb_sbj.
			or ($rel eq '>(ch)') # $gb2 is an "effective" child of $gb_sbj.
			or ($rel eq '<(pa)') # $gbs is an "effective" parent of $gb_sbj.
			or ($rel eq '=') # $gb2 is vertically equivalent to $gb_sbj.
			) {

		    if ($rbd2 + 1 == $lbd_sbj_af) { # The subject block, which was separated from $gb2 by one column, now touches the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'e'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif ($rbd_sbj_bf + 1 == $lbd2) { # The subject block, which touched $gb2, is now separated from the latter by one column.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'f'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif ( ($rel eq 'Cp') and (($rbd_sbj_bf == $lbd2) or ($rbd_sbj_af +1 == $lbd2)) ) { # The subject block now touches its complement, $gb2, which is on the right. # ADDED on 2019/04/12. (1)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'g'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif ($rbd2 + 1 < $lbd_sbj_af) { # The subject block now appears separated from $gb2, which is on the left.

			if ((@{$gblks_w_spec_lb->[$rbd2 +1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_af -1]}>0)) {

				### MODIFIED on 2019/04/12. (1) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_left_end_to_left ($gb_sbj, $rbd2 +1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb_sbj th gap-block against $gb2 (A): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'A'), "\n";
			        $if_change = 1;
			        next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }

=REPLACED with a better method. (1)
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_af, $rbd2 +1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (A): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { return 1; }
=cut # END of "REPLACED with a better method. (1)"
				### END of "MODIFIED on 2019/04/12. (1)" ###
			}

		    } elsif ($rbd_sbj_bf +1 < $lbd2) { # The subject block now appeared separated from $gb2, which is on the right. 

			if ((@{$gblks_w_spec_lb->[$rbd_sbj_bf +1]}>0) and (@{$gblks_w_spec_rb->[$lbd2 -1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_bf, $lbd2 -1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (B): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $gb2, $rel, 'B'), "\n";
			        $if_change = 1;
			        next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }
			}

		    } # End of "if ($rbd2 +1 == $lbd_sbj_af) {...} elsif (...) {...} .... {...}"

		} # End of "if ($rel eq 'S') {...} elsif (($rel eq 'Cp') or ...) {...}".

	    } # End of the 1st outer for-loop (over interfering blocks).

	} # End of "if (@{$rlv_interfering_blocks}>0) {...}".


	if (@{$rlv_intrfr_blksets}>0) { # (2) Examine the interfering block-sets.

	    my ($intrfr_blksets, $relations2) = @{$rlv_intrfr_blksets};
	    my $ct_intrfr_blksets = @{$intrfr_blksets};

    (print STDOUT "\n... There are $ct_intrfr_blksets gap-block-sets interfering with $gb_sbj ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISIOn on 2019/04/14.
    (print STDOUT join ("\t", 'Intrfr,gap,blks', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.

	    for (my $k = 0; $k < $ct_intrfr_blksets; $k++) { # 2nd outer for-loop (over interfering block-sets).
		my $blset = $intrfr_blksets->[$k];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$blset->[0]]};
		my $size_blset = @{$blset};
		my $rel = $relations2->[$k];

		my $cnct_blset = join (',', @{$blset}); # ADDED in REVISION on 2019/04/14.

		if ($rel eq 'S') {

		    for (my $j=1; $j <$size_blset; $j++) { # 1st inner for-loop (over constituent blocks in the block-set).
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$blset->[$j]]};
			if ($lbd2 < $lbd3) { $lbd2 = $lbd3; }
			if ($rbd3 < $rbd2) { $rbd2 = $rbd3; }
			if ($rbd2 < $lbd2) { last; } # Added on 2019/04/08.

		    } # End of the 1st inner for-loop (over constituent blocks in the block-set).

		    if ($rbd2 < $lbd2) { next; }

		    my $if_test = 0;
		    if ( (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) 
			or (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) ) {
			$if_test = 1;
		    } else {

			my $if_min_blks = 0;
			for (my $j = 0; $j < $size_blset; $j++) { # 2nd inner for-loop (over constituent blocks in the block-set).
			    my $gb3 = $blset->[$j];
			    my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
			    if (($lbd3 == $lbd2) and ($rbd3 == $rbd2)) {
				$if_min_blks =1;
				last;
			    }
			} # End of the 2nd inner for-loop (over constituent blocks in the block-set).

			if ($if_min_blks) { $if_test = 1; }

		    } # End of "if ( (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) or (...)) {...} else {...}".

		  if ($if_test) {

			# Do the identical tests as for "if ($rel eq 'S') {...}" in (1). #

		    if (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af < $rbd2)) { # The subject block is now non-nested with $gb2.
			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'h'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			} elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'i'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd2 <= $lbd_sbj_af) and ($rbd2 == $rbd_sbj_af)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'j'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;

		    } elsif (($lbd2 == $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'k'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		  } # End of "if ($if_test) {...}".

		} elsif ($rel eq '>(ch)') {

		    my $ct_topo_chngs = my $ct_including = 0;

		    for (my $j=0; $j < $size_blset; $j++) { # 3rd inner for-loop (over constituent blocks).
			my $gb3 = $blset->[$j];
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};

			if (($lbd3 <= $rbd_sbj_af +1) and ($lbd_sbj_bf <= $rbd3 + 1)) { # During the "shift", $gb3 keeps touching $gb_sbj.
			    $ct_including++;

			} elsif ($rbd3 +1 == $lbd_sbj_af) { # The subject block, which was separated from $gb3 by one column, now touches the latter.
			    $ct_topo_chngs++;
			} elsif ($rbd_sbj_bf +1 == $lbd3) { # The subject block, which touched $gb3, is now separated from the latter by one column.
			    $ct_topo_chngs++;
			##} elsif ($rbd_sbj_af + 1 < $lbd3) { # The subject block now appears separated from $gb3, which is on the right.
			} elsif ($rbd3 +1 < $lbd_sbj_af) { # The subject block now appears separated from $gb3, which is on the left.
			  if ((@{$gblks_w_spec_lb->[$rbd3 +1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_af -1]}>0)) {
				### MODIFIED on 2019/04/12. (2) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_left_end_to_left ($gb_sbj, $rbd3 +1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb_sbj th gap-block against $gb3 (C): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) { 
				$ct_topo_chngs++;
			    } else {
				last;
			    }
=REPLACED with a better method. (2)
			    my %skipped = ($gb_sbj => 1, $gb3 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_af, $rbd3 +1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block, which belongs to the $k th interfering block-set (C): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts > 0) { 
				last;
			    } else {
				$ct_topo_chngs++;
			    }
=cut # END of "REPLACED with a better method. (2)"

			  } else {
				last;
			  } # End of "if ((@{$gblks_w_spec_lb->[$rbd3 +1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_af -1]}>0)) {...} else {...}".

			} elsif ($rbd_sbj_bf + 1 < $lbd3) { # The subject block appeared separated from $gb3, which is on the right.
			  if ((@{$gblks_w_spec_lb->[$rbd_sbj_bf +1]}>0) and (@{$gblks_w_spec_rb->[$lbd3 -1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb3 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_bf, $lbd3 -1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block, which belongs to the $k th interfering block-set (D): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts > 0) {
				last;
			    } else {
				$ct_topo_chngs++;
			    }
			  } else {
				last;
			  } # End of "if ((@{$gblks_w_spec_lb->[$rbd_sbj_bf +1]}>0) and (@{$gblks_w_spec_rb->[$lbd3 -1]}>0)) {...} else {...}".

			} else {
			    last; # Because we are sure that there should be NO topological change between the block-set and $gb_sbj, leave the loop here.
			}

		    } # End of the 3rd inner for-loop (over constituent blocks).

		    if (($ct_topo_chngs>0) and ($ct_topo_chngs + $ct_including == $size_blset)) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_blset, $rel, 'l'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} # End of "if ($rel eq 'S') {...} elsif ($rel eq '>(ch)') {...}".

	    } # End of the 2nd outer for-loop (over interfering block-sets).

	} # End of "if (@{$rlv_intrfr_blksets}>0) {...}".


	if (@{$rlv_cmpl_intrfr_blksets}>0) { # (3) Examine the complements of the interfering block-sets.

			# Restarted on 2019/04/08. #

		## (The following is from p.49 of "suppl3_blueprint1_ANEX.xxx.pdf".) ##

	    my ($intrfr_cblksets, $relations3) = @{$rlv_cmpl_intrfr_blksets};
	    my $ct_intrfr_cbss = @{$intrfr_cblksets};

    (print STDOUT "\n... There are $ct_intrfr_cbss gap-block-sets including $gb_sbj and with an interfering partner ...\n\n") if ($TEST_LEVEL_ccgpt_p>1); # ADDED in REVISIOn on 2019/04/14.
    (print STDOUT join ("\t", 'Cmpl,intrfr,gap,blks', 'relation', 'Type_of_topology_change'), "\n\n") if ($TEST_LEVEL_ccgpt_p>2); # ADDED in REVISION on 2019/04/14.


	    for (my $k =0; $k <$ct_intrfr_cbss; $k++) { # 3rd outer for-loop (over complements of interfering block-sets).

		my $cblkset = $intrfr_cblksets->[$k];
		my $size_cblkset = @{$cblkset};
		my $gb2 = $cblkset->[0];
		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
		my $rel = $relations3->[$k];

		my $cnct_cblkset = join (',', @{$cblkset}); # ADDED in REVISION on 2019/04/14.

		if ($rel eq 'S') { # $gb2 is an "effective sibling" of the block-set.

		    my ($lbd3, $rbd3) = @{$bds_gblocks->[$cblkset->[1]]};
		    for (my $j=2; $j <$size_cblkset; $j++) { # 1st inner for-loop (over constituent blocks).
			my ($lbd5, $rbd5) = @{$bds_gblocks->[$cblkset->[$j]]};
			if ($lbd3 < $lbd5) { $lbd3 = $lbd5; }
			if ($rbd5 < $rbd3) { $rbd3 = $rbd5; }
			if ($rbd3 < $lbd3) { last; }
		    } # End of the 1st inner for-loop (over constituent blocks).

		    if ($rbd3 < $lbd3) { next; } # Topology should NOT change!!

		    my ($max_sbj_lbd, $min_sbj_lbd) = ($lbd_sbj_af < $lbd_sbj_bf) ?
				($lbd_sbj_af, $lbd_sbj_bf) : ($lbd_sbj_bf, $lbd_sbj_af);

		    my ($max_sbj_rbd, $min_sbj_rbd) = ($rbd_sbj_af > $rbd_sbj_bf) ?
			        ($rbd_sbj_af, $rbd_sbj_bf) : ($rbd_sbj_bf, $rbd_sbj_af);

		    if (($lbd3 <= $lbd2) and ($rbd2 <= $rbd3)) { # $gb2 is horizontally included in the block-set (excluding $gb_sbj). => Focus on the relationship between $gb2 and $gb_sbj. (We do NOT know yet, which of $gb_sbj and $gb2 horizontally includes the other.)

			# Do tests identical to those above for ($rel eq 'S') in (1). #

			if (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af < $rbd2)) { # The subject block is now non-nested with $gb2.
			    if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # The subject block was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-1'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;

			    } elsif (($lbd_sbj_bf <= $lbd2) and ($rbd2 <= $rbd_sbj_bf)) { # The subject block horizontally included $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-2'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }

			} elsif (($lbd2 <= $lbd_sbj_af) and ($rbd2 == $rbd_sbj_af)) { # The subject block, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-3'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;

			##} elsif (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af == $rbd2)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.
			} elsif (($lbd2 == $lbd_sbj_af) and ($rbd2 < $rbd_sbj_af)) { # The subject block, which overlapped but was horizontally non-nested with $gb2, now horizontally includes the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'm-5'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $max_sbj_lbd) and ($max_sbj_rbd <= $rbd3)) { # BOTH before and after the change, $gb_sbj is horizontally included in the block-set (excluding $gb_sbj). => Focus on the relationship between $gb2 and $gb_sbj. (In this case, $gb_sbj will NOT include $gb2 horizontally (because $gb2 did NOT satisfy the previous conditions).)

			if (($lbd_sbj_af < $lbd2) and ($rbd_sbj_af < $rbd2)) { # $gb_sbj is now non-nested with $gb2.
			    if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # $gb_sbj was horizontally included in $gb2.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'n-1'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

				return 1;
			    }

			} elsif (($lbd2 <= $lbd_sbj_af) and ($rbd2 == $rbd_sbj_af)) { # $gb_sbj, which overlapped but was non-nested with $gb2, now is horizontally included in the latter.

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'n-2'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd3)) { # After the change, $gb_sbj is horizontally included in the block-set (excluding $gb_sbj), but it was NOT before the change.

			if (($lbd2 <= $lbd_sbj_af) and ($rbd_sbj_af <= $rbd2)) { # $gb2 also includes $gb_sbj afer the change. (We don't care about their relation before the change.)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'o'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($lbd3 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd3)) { # Before the change, $gb_sbj was horizontally included in the block-set (excluding $gb_sbj), but it is NOT after the change.

			if (($lbd2 <= $lbd_sbj_bf) and ($rbd_sbj_bf <= $rbd2)) { # $gb2 also included $gb_sbj before the change. (We don't care about their relation after the change.)

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'p'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } elsif (($min_sbj_lbd <= $lbd3) and ($rbd3 <= $min_sbj_rbd)) { # Both before and after the change, $gb_sbj horizontally includes the intersection of the block-set (excluding $gb_sbj). => Topology will NEVER change, no matter what.

			next;

		    } elsif ( (($lbd_sbj_af <= $lbd3) and ($rbd3 <= $rbd_sbj_af)) # After the change, $gb_sbj horizontally includes the intersection of the block-set (excluding $gb_sbj), but it did NOT before the change.
				or (($lbd_sbj_bf <= $lbd3) and ($rbd3 <= $rbd_sbj_bf)) # Before the change, $gb_sbj horizontally included the intersection of the block-set (excluding $gb_sbj), but it does NOT after the change.
		    ) {

			unless (($lbd2 <= $lbd3) and ($rbd3 <= $rbd2)) { next; } # $gb2 does NOT include the intersection of the block-set (excluding $gb_sbj). => Topology will NOT change!!

			my $if_min_blks=0;
			for (my $j=1; $j < $size_cblkset; $j++) { # 2nd inner for-loop (over constituent blocks).
			    my ($lbd5, $rbd5) = @{$bds_gblocks->[$cblkset->[$j]]};
			    if (($lbd3 == $lbd5) and ($rbd3 == $rbd5)) {
				$if_min_blks=1;
				last;
			    }
			} # End of the 2nd inner for-loop (over constituent blocks).

			if ($if_min_blks) {

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'q'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			    return 1;
			}

		    } # End of "if (($lbd3 <= $lbd2) and ($rbd2 <= $rbd3)) {...} elsif (...) ... elsif ((($lbd_sbj_af <= $lbd3) and ($rbd3 <= $rbd_sbj_af)) or (($lbd_sbj_bf <= $lbd3) and ($rbd3 <= $rbd_sbj_bf)) ) {...}".

		} elsif ($rel eq '<(pa)') { # $gb2 is an "effective parent" of the block-set.

			# Do tests identical to those above for ($rel eq '<(pa)') in (1).

		    my $if_topo_chng = 0;
		    if ($rbd2 + 1 == $lbd_sbj_af) { # The subject block, which was separated from $gb2 by one column, now touches the latter.
			$if_topo_chng = 1;
		    } elsif ($rbd_sbj_bf +1 == $lbd2) { # The subject block, which touched $gb2, is now separated from the latter by one column.
			$if_topo_chng = 1;
		    } elsif ($rbd2 + 1 < $lbd_sbj_af) { # The subject block now appears separated from $gb2, which is on the left.

			if ((@{$gblks_w_spec_lb->[$rbd2 +1]}>0) and (@{$gblks_w_spec_rb->[$lbd_sbj_af -1]}>0)) {
				### MODIFIED on 2019/04/12. (3) ###
			    my ($if_reached, $lbd_sbj_af_new, $rbd_sbj_af_new) = smlt2_extend_left_end_to_left ($gb_sbj, $rbd2 +1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			    unless (defined $if_reached) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb_sbj th gap-block against $gb2 (E): $! !\n\n";
				return undef;
			    }
			    if ($if_reached) { $if_topo_chng = 1; }
=REPLACED with a better method. (3)
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($lbd_sbj_af, $rbd2 +1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (E): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { $if_topo_chng = 1; }
=cut # END of "REPLACED with a better method. (3)"
			}

		    } elsif ($rbd_sbj_bf + 1 < $lbd2) { # The subject block now appeared separated from $gb2, which is on the right. 

			if ((@{$gblks_w_spec_lb->[$rbd_sbj_bf +1]}>0) and (@{$gblks_w_spec_rb->[$lbd2 -1]}>0)) {
			    my %skipped = ($gb_sbj => 1, $gb2 => 1);
			    my $ct_shifts = ct_necessary_shifts ($rbd_sbj_bf, $lbd2 -1, $gb_sbj, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
			    unless (defined $ct_shifts) {
				print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb2 th gap-block (F): $! !\n\n";
				return undef;
			    }
			    if ($ct_shifts == 0) { $if_topo_chng = 1; }
			}
		    } # End of "if ($rbd2 +1 == $lbd_sbj_af) {...} elsif (...) {...} .... {...}"

		    unless ($if_topo_chng) { next; } # NO topological change in conjunction with this block-set!!

			# Check if all the remaining blocks in the block-set are horizontally "connected with" $gb2. #

		    my %skipped = ($gb_sbj => 1);
		    foreach my $gb3 (@{$cblkset}) { $skipped{$gb3} = 1; }

		    for (my $j=1; $j<$size_cblkset; $j++) {  # 3rd inner for-loop (over constituent blocks).

			my $gb3 = $cblkset->[$j]; # A BUG FIXED on 2019/04/17!!
			##my $gb3 = $cblkset->[$k];
			my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};

			if ($rbd3 < $lbd2-1) { # $gb3 appears separated from $gb2 on its right. 

			    if ((@{$gblks_w_spec_lb->[$rbd3+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {
				### MODIFIED on 2019/04/12. (5) ###
				my ($if_reached, $lbd3_new, $rbd3_new) = smlt2_extend_right_end_to_right ($gb3, $lbd2-1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
				unless (defined $if_reached) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'smlt2_extend_right_end_to_right' FAILED on the $gb3 th gap-block against $gb2 (E): $! !\n\n";
				    return undef;
				}
				if ($if_reached) { next; } # Actually NOT separated.
=REPLACED with a better method. (5)
				my $ct_shifts = ct_necessary_shifts ($rbd3, $lbd2-1, $gb3, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
				unless (defined $ct_shifts) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_R'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block (G): $! !\n\n";
				    return undef;
				}
				if ($ct_shifts == 0) { next; } # Actually NOT separated.
=cut # END of "REPLACED with a better method. (5)"
			    }
			    $if_topo_chng = 0;
			    last;

			} elsif ($rbd2+1 < $lbd3) { # $gb3 appears separated from $gb2 on its left.

			    if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd3-1]}>0)) {
				### MODIFIED on 2019/04/12. (6) ###
				my ($if_reached, $lbd3_new, $rbd3_new) = smlt2_extend_left_end_to_left ($gb3, $rbd2+1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
				unless (defined $if_reached) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'smlt2_extend_left_end_to_left' FAILED on the $gb3 th gap-block against $gb2 (E): $! !\n\n";
				    return undef;
				}
				if ($if_reached) { next; } # Actually NOT separated.
=REPLACED with a better method. (6)
				my $ct_shifts = ct_necessary_shifts ($lbd3, $rbd2+1, $gb3, @{$bds_gblocks}, @{$inter_block_relations}, %skipped);
				unless (defined $ct_shifts) {
				    print STDERR "[ERROR 'chk_chng_gptrn_topology_psm_L'] Subroutine 'ct_necessary_shifts' FAILED on the $gb3 th gap-block (H): $! !\n\n";
				    return undef;
				}
				if ($ct_shifts == 0) { next; } # Actually NOT separated.
=cut # END of "REPLACED with a better method. (6)"
			    }

			    $if_topo_chng = 0;
			    last;

			} # End of "if ($rbd3 < $lbd2-1) {...} elsif ($rbd2+1 < $lbd3) {...}".

		    } # End of the 3rd inner for-loop (over constituent blocks).

		    if ($if_topo_chng) { # Topology DID change in conjunction with this block-set!!

    if ($TEST_LEVEL_ccgpt_p>2) { # ADDED in REVISION on 2019/04/14.
	print STDOUT join ("\t", $cnct_cblkset, $rel, 'r'), "\n";
			    $if_change = 1;
			    next;
    } # End of "if ($TEST_LEVEL_ccgpt_p>2) {...}"

			return 1;
		    }

		} # End of "if ($rel eq 'S') {...} elsif ($rel eq '<(pa)') {...}".

	    } # End of the 3rd outer for-loop (over complements of interfering block-sets).

	} # End of "if (@{$rlv_cmpl_intrfr_blksets}>0) {...}".


			### RESTART FROM HERE (1) ###

    if ($TEST_LEVEL_ccgpt_p>2) {
	print STDOUT "\n< Subroutine 'chk_chng_gptrn_topology_psm_L' Ended (w/ \$if_change = $if_change) />\n\n\n";
	return $if_change;
    }

    (print STDOUT "\n< Subroutine 'chk_chng_gptrn_topology_psm_L' Ended />\n\n\n") if ($TEST_LEVEL_ccgpt_p >0);

		# END of "The following are COPIED from 'chk_chng_gptrn_topology_psm_R' and MODIFIED on 2019/04/16." #

	return 0; # NO topological change.

}	# END of the subroutine 'chk_chng_gptrn_topology_psm_L'. #


# smlt_extend_left_end_to_left:
#	This satellite subroutine of 'encode_gptrn_topology_psm'
#	simulates an extention of the left-end of a subject gap-block further to the left
#	UNTIL it touches a "reference" gap-block.
#	[CREATED on 2019/04/09; DEBUGGED on 2019/04/17]
#
#	NOTE1: This subroutine DIFFERS from 'vextend_left_end_sbj_gblk' above in the following manner:
#		the former just examines what will happen to the boundaries without actually moving them,
#		 whereas the latter actually moves the boundaries.
#
#	NOTE2: It is paired with 'smlt_extend_right_end_to_right' below.
#
# INPUTS: ($gb_sbj, $gb_ref, \@bds_gblocks, \@gblks_w_spec_rb, \@inter_block_relations), where 
#
#	$gb_sbj is the index for the subject gap-block, whose left-end is to be extended;
#	$gb_ref is the index for the reference gap-block, which serves as the "destination";
#	[NOTE: The left-end of $gb_sbj is SUPPOSED TO be on the right of the right-end of $gb_ref.]
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
# OUTPUTS: ($if_separated, $lbd_sbj_new, $rbd_sbj_new), where
#
#	$if_separated	= 1 if $gb_sbj and $gb_ref remains separated after the simulated extension,
#			= 0 if they either overlap or touch then;
#
#	($lbd_sbj_new, $rbd_sbj_new) are the boundaries of $gb_sbj after the simulated extension.
#
sub smlt_extend_left_end_to_left ($$\@\@\@) {

	my ($gb_sbj, $gb_ref, $bds_gblocks, $gblks_w_spec_rb, $inter_block_relations) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};
	my ($lbd_ref, $rbd_ref) = @{$bds_gblocks->[$gb_ref]};

	unless ($rbd_ref + 1 < $lbd_sbj) {
	    print STDERR "[ERROR 'smlt_extend_left_end_to_left'] ILLEGAL INPUT: \$rbd_ref (=$rbd_ref) is NOT on the left of \$lbd_sbj (=$lbd_sbj): $! !\n\n";
	    return (undef, undef, undef);
	}

	my ($lbd_sbj_new, $rbd_sbj_new) = ($lbd_sbj, $rbd_sbj);

	my $if_separated = 1;
	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];
	my $left_flanking_gblks = $gblks_w_spec_rb->[$lbd_sbj-1];

	while (@{$left_flanking_gblks}>0) { # Attempt to "extend" the left-end of $gb_sbj to the left.

	    my $if_rlv = 0;
	    foreach my $gb3 (@{$left_flanking_gblks}) { # Middle foreach-loop (over left-flanking gap-blocks).
		my $rel = $rels_w_sbj->[$gb3];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb3 < $gb_sbj))
			or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; }

		my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
		$lbd_sbj_new = $lbd3;
		if (($rel eq 'ONN') or ($rel eq 'ONCS')) { $rbd_sbj_new -= $rbd3 - $lbd3 +1; }
		$if_rlv = 1;
		last;
	    } # End of the middle foreach-loop (over left-flanking gap-blocks).

	    unless ($if_rlv) { last; }

	    if ($rbd_ref + 1 < $lbd_sbj_new) {
		$left_flanking_gblks = $gblks_w_spec_rb->[$lbd_sbj_new-1]; # Update the list of left-flanking blocks.
	    } else {
		if ($lbd_ref <= $rbd_sbj_new +1) { $if_separated = 0; }
		last;
	    }

	} # End of "while (@{$left_flanking_gblks}>0) {...}".


	return ($if_separated, $lbd_sbj_new, $rbd_sbj_new);

}	# End of the subroutine 'smlt_extend_left_end_to_left'. #


# smlt_extend_right_end_to_right:
#	This satellite subroutine of 'encode_gptrn_topology_psm'
#	simulates an extention of the right-end of a subject gap-block further to the right
#	UNTIL it touches a "reference" gap-block.
#	[CREATED on 2019/04/09.]
#
#	NOTE1: This subroutine DIFFERS from 'vextend_right_end_sbj_gblk' above in the following manner:
#		the former just examines what will happen to the boundaries without actually moving them,
#		 whereas the latter actually moves the boundaries.
#
#	NOTE2: It is paired with 'smlt_extend_left_end_to_left' above.
#
# INPUTS: ($gb_sbj, $gb_ref, \@bds_gblocks, \@gblks_w_spec_lb, \@inter_block_relations), where 
#
#	$gb_sbj is the index for the subject gap-block, whose left-end is to be extended; #	$gb_ref is the index for the reference gap-block, which serves as the "destination";
#	[NOTE: The right-end of $gb_sbj is SUPPOSED TO be on the left of the left-end of $gb_ref.]
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
# OUTPUTS: ($if_separated, $lbd_sbj_new, $rbd_sbj_new), where
#
#	$if_separated	= 1 if $gb_sbj and $gb_ref remains separated after the simulated extension,
#			= 0 if they either overlap or touch then;
#
#	($lbd_sbj_new, $rbd_sbj_new) are the boundaries of $gb_sbj after the simulated extension.
#
sub smlt_extend_right_end_to_right ($$\@\@\@) {

	my ($gb_sbj, $gb_ref, $bds_gblocks, $gblks_w_spec_lb, $inter_block_relations) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};
	my ($lbd_ref, $rbd_ref) = @{$bds_gblocks->[$gb_ref]};

	unless ($rbd_sbj + 1 < $lbd_ref) {
	    print STDERR "[ERROR 'smlt_extend_right_end_to_right'] ILLEGAL INPUT: \$lbd_ref (=$lbd_ref) is NOT on the right of \$rbd_sbj (=$rbd_sbj): $! !\n\n";
	    return (undef, undef, undef);
	}

	my ($lbd_sbj_new, $rbd_sbj_new) = ($lbd_sbj, $rbd_sbj);

	my $if_separated = 1;
	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];
	my $right_flanking_gblks = $gblks_w_spec_lb->[$rbd_sbj +1];

	while (@{$right_flanking_gblks}>0) { # Attempt to "extend" the right-end of $gb_sbj to the right.

	    my $if_rlv = 0;
	    foreach my $gb3 (@{$right_flanking_gblks}) { # Middle foreach-loop (over right-flanking gap-blocks).
		my $rel = $rels_w_sbj->[$gb3];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb3 < $gb_sbj))
			or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; }

		my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
		$rbd_sbj_new = $rbd3;
		if (($rel eq 'ONN') or ($rel eq 'ONCS')) { $lbd_sbj_new += $rbd3 - $lbd3 +1; }
		$if_rlv = 1;
		last;
	    } # End of the middle foreach-loop (over right-flanking gap-blocks).

	    unless ($if_rlv) { last; }

	    if ($rbd_sbj_new + 1 < $lbd_ref) {
		$right_flanking_gblks = $gblks_w_spec_lb->[$rbd_sbj_new +1]; # Update the list of right-flanking blocks.
	    } else {
		if ($lbd_sbj_new <= $rbd_ref +1) { $if_separated = 0; }
		last;
	    }

	} # End of "while (@{$right_flanking_gblks}>0) {...}".


	return ($if_separated, $lbd_sbj_new, $rbd_sbj_new);

}	# End of the subroutine 'smlt_extend_right_end_to_right'. #


# encode_gptrn_topology_psm:
#	This subroutine encodes the gap-pattern topology (regarding the parsimonious ancestral gap-states)
#	of a window of a multiple sequence alignment (MSA)
#	in a manner specified here.
#	[CREATION STARTED on 2019/04/08, ENDED on 2019/04/09.]
#
#	[NOTE1] This subroutine is based on the ASSUMPTION that the collection of all relevait binary relations (between the gap-blocks) can uniquely determine the gap-pattern topology.
#		... This assumption MUST be proven later.
#
#	[NOTE2] CODES for the Horizontal Binary Relations are DEFINED as follows:
#	"$gb1|$gb2" denotes that $gb1 and $gb2 are "separated",
#	"$gb1-$gb2" denotes that $gb1 and $gb2 are "connected" (i.e., overlapping or touching),
#	"$gb1^$gb2" denotes that $gb1 and $gb2 are "non-nested" (When $gb1 and $gb2 are effective siblings, we don't care whether they are "connected" or not),
#	"$gb1>$gb2" denotes that $gb1 horizontally includes $gb2,
#	"$gb1<$gb2" denotes that $gb1 is horizontally included in $gb2,
#	"$gb1=$gb2" denotes that $gb1 is horizontally identical with $gb2.
#
#
# INPUTS: (\@bds_gblocks, \@inter_block_relations, \@interfering_blocks, \@interfering_blocksets, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$interfering_blocks[$gb]} = (\@blocks, \@relations)
#	stores information on blocks that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	$blocks[$k] is the index (or rank) of the $k th interfering block,
#	$relations[$k] is the relation of the $k th interfering block against the $gb th block;
#	[Typically, \@interfering_blocks is the output of 'mk_list_interfering_blocks' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$interfering_blocksets[$gb]} = (\@blocksets, \@relations) 
#	stores information on block-SETs that can interfere with the $gb th gap-block (or empty if it has no such blocks), where
#	@{$blocksets[$k]} lists the indices (or ranks) of the gap-blocks constiruting the $k th interfering block-set,
#	$relations[$k] is the relation of the $k th interfering block-set against the $gb th block;
#	[NOTE: The gap-blocks in each @{$blocksets[$k]} ALWAYS form a (complementary-)monophyletic group];
#	[Typically, \@interfering_blocksets is an output of 'mk_list_interfering_blocksets' above, or 'extensive_exam_inter_block_relations' above];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_lb and \@gblks_w_spec_rb are the outputs of 'mk_lists_gblocks_w_spec_bounds' above];
#
# OUTPUTS: $code_topology, where
#
#	$code_topology is a string encoding the gap-pattern topology,
#	which is actually a concatenation of all binary relationships, each between two gap-blocks, that are relevant to the gap-pattern topology.
#
sub encode_gptrn_topology_psm (\@\@\@\@\@\@) {

	my ($bds_gblocks, $inter_block_relations, $interfering_blocks, $interfering_blocksets, $gblks_w_spec_lb, $gblks_w_spec_rb) = @_;

# INPUTS: (\@bds_gblocks, \@inter_block_relations, \@interfering_blocks, \@interfering_blocksets, \@gblks_w_spec_lb, \@gblks_w_spec_rb), where

			## (From p.93 of "suppl3_blueprint1_ANEX.draft9_Final.pdf".) ##

	my $ct_gblks = @{$bds_gblocks};

	my @binary_relations = ();

	for (my $gb1 =0; $gb1 < $ct_gblks; $gb1++) { # Outermost for-loop (over $gb1).

	    ##unless (@{$interfering_blocks->[$gb1]}>0) { next; } # DISABLED for DEBUGGING (on 2019/04/15).

	    my ($intrfr_blks, $relations) = @{$interfering_blocks->[$gb1]};

	    my ($lbd1, $rbd1) = @{$bds_gblocks->[$gb1]};

	    my $ct_intrfr_blks = (defined $intrfr_blks) ? @{$intrfr_blks} : 0; # MODIFIED for DEBUGGING (on 2019/04/15).
	    ##my $ct_intrfr_blks = @{$intrfr_blks};

	    my $rels_w_gb1 = $inter_block_relations->[$gb1];

	    for (my $k=0; $k <$ct_intrfr_blks; $k++) { # 1st middle for-loop (over interfering blocks).

		my $gb2 = $intrfr_blks->[$k];
		unless ($gb1 < $gb2) { next; } # To avoid redundancy.

		my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
		my $rel = $relations->[$k];
		my $rels_w_gb2 = $inter_block_relations->[$gb2];

		if ($rel eq '=') {
		    next; # $gb1 and $gb2 must always be separated when this subroutine is run.
			# This means that, in this case, the positional relation between $gb1 and $gb2 will NEVER affect the gap-pattern topology.

		} elsif ($rel eq 'Cp') {
			# $gb1 and $gb2 must NEVER ovelap (horizontally) when this subroutine is run.

		    my $if_separated = 1;

		    if ($rbd1 + 1 < $lbd2) { # $gb1 appears separated from $gb2, which is on the right.

			($if_separated, my $lbd2_new, my $rbd2_new) = smlt_extend_left_end_to_left ($gb2, $gb1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			unless (defined $if_separated) {
			    print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_left_end_to_left' FAILED on the $gb2 th gap-block against the $gb1 th gap-block as reference (A-L): $! !\n\n";
			    return undef;
			}

		    } elsif ($rbd2 + 1 < $lbd1) { # $gb1 appears separated from $gb2, which is on the left.

			($if_separated, my $lbd2_new, my $rbd2_new) = smlt_extend_right_end_to_right ($gb2, $gb1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			unless (defined $if_separated) {
			    print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_right_end_to_right' FAILED on the $gb2 th gap-block against the $gb1 th gap-block as reference (A-R): $! !\n\n";
			    return undef;
			}

		    } else {
				# $gb1 obviously overlaps or touches $gb2. #
			$if_separated = 0;
		    }

		    my $bin_rel = ($if_separated) ? join ('|', $gb1, $gb2) : join ('-', $gb1, $gb2) ;

		    push @binary_relations, $bin_rel; # ADDED in DEBUGGING (on 2019/04/15).

		} elsif ($rel eq 'S') {

		    my $bin_rel;
		    if ($lbd1 < $lbd2) {
			if ($rbd1 < $rbd2) { # Non-nested. (We don't care whether they are "connected" or not.)
			    $bin_rel = join ('^', $gb1, $gb2);
			} else { # Nested ($gb1 horizontally includes $gb2).
			    $bin_rel = join ('>', $gb1, $gb2);
			}

		    } elsif ($lbd2 < $lbd1) {

			if ($rbd2 < $rbd1) { # Non-nested. (We don't care whether they are "connected" or not.) 
			    $bin_rel = join ('^', $gb1, $gb2);
			} else { # Nested ($gb1 is horizontally included in $gb2).
			    $bin_rel = join ('<', $gb1, $gb2);
			}

		    } else { # if ($lbd1 == $lbd2)

			if ($rbd1 <$rbd2) { # $gb1 is horizontally included in $gb2.
			    $bin_rel = join ('<', $gb1, $gb2);
			} elsif ($rbd2 < $rbd1) { # $gb1 horizontally includes $gb2.
			     $bin_rel = join ('>', $gb1, $gb2);
			} else { # if ($rbd1 == $rbd2) # $gb1 and $gb2 are horizontally identical.
			    $bin_rel = join ('=', $gb1, $gb2);
			}

		    } # End of "if ($lbd1 < $lbd2) {...} elsif ($lbd2 < $lbd1) {...} else {...}".

		    push @binary_relations, $bin_rel;

		} elsif (($rel eq '>(ch)') or ($rel eq '<(pa)')) { # In this case, the binary relation itself is, after all, (effective) "parent-child", but the roles of $gb1 and $gb2 can be swapped.

		    my ($gb_ch, $gb_pa) = ($rel eq '>(ch)') ? ($gb2, $gb1) : ($gb1, $gb2);
		    my ($lbd_ch, $rbd_ch) = @{$bds_gblocks->[$gb_ch]};
		    my ($lbd_pa, $rbd_pa) = @{$bds_gblocks->[$gb_pa]};

		    my $if_separated = 1;

		    if ($rbd_pa + 1 < $lbd_ch) { # $gb_ch appears separated from $gb_pa, which is on the left.

			($if_separated, my $lbd_ch_new, my $rbd_ch_new) = smlt_extend_left_end_to_left ($gb_ch, $gb_pa, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			unless (defined $if_separated) {
			    print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_left_end_to_left' FAILED on the $gb_ch th gap-block against the $gb_pa th gap-block as reference (B-L): $! !\n\n";
			    return undef;
			}

		    } elsif ($rbd_ch + 1 < $lbd_pa) { # $gb_ch appears separated from $gb_pa, which is on the right.

			($if_separated, my $lbd_ch_new, my $rbd_ch_new) = smlt_extend_right_end_to_right ($gb_ch, $gb_pa, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			unless (defined $if_separated) {
			    print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_right_end_to_right' FAILED on the $gb_ch th gap-block against the $gb_pa th gap-block as reference (B-R): $! !\n\n";
			    return undef;
			}

		    } else {
				# $gb_ch and $gb_pa are obviously "connected" (i.e., "overlapping" or "touching").
			$if_separated = 0;

		    } # End of "if ($rbd_pa + 1 < $lbd_ch) {...} elsif ($rbd_ch + 1 < $lbd_pa) {...} else {...}".

		    my $bin_rel = ($if_separated) ? join ('|', $gb1, $gb2) : join ('-', $gb1, $gb2);
		    push @binary_relations, $bin_rel;

		} # End of "if ($rel eq '=') {...} elsif ($rel eq 'Cp') {...} elsif ($rel eq 'S') {...} elsif (($rel eq '>(ch)') or ($rel eq '<(pa)')) {...}".

	    } # End of the 1st middle for-loop (over interfering blocks).


		# Examine the relationships between $gb1 and block-sets interfering with it.

	    unless (@{$interfering_blocksets->[$gb1]}>0) { next; }
	    my ($intrfr_blksets, $relations2) = @{$interfering_blocksets->[$gb1]};
	    my $ct_intrfr_blksets = @{$intrfr_blksets};

	    for (my $k=0; $k < $ct_intrfr_blksets; $k++) { # 2nd middle for-loop (over interfering block-sets).

		my $blkset = $intrfr_blksets->[$k];
		my $rel = $relations2->[$k];

		my $cnct_blkset = join (',', @{$blkset});

		if ($rel eq 'S') {

			# NOTE (added on 2019/01/21): The situation gets extremely complicated if we consider the extension of the boundaries of the relevant blocks.
			# Thus, FOR THE MOMENT, we will OMIT the boundary extension. (And we will REVISIT the issue when we have time.)
			# (I have a hunch, however, that extending the boundaries will NOT affect the results (as in the case of single-block siblings, $gb1 and $gb2).

		    my $if_nesting = 1; # Whether the block-set "nest"s $gb1 or not.
		    my $if_equal = 1; # Whether all constituents of the block-set share the same boundaries.
		    my ($lbd_min, $rbd_min) = my ($lbd_sh, $rbd_sh) = @{$bds_gblocks->[$blkset->[0]]};

		    foreach my $gb2 (@{$blkset}) { # 1st inner foreach-loop (over the constituents, $gb2).
			my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};

			if ($lbd_min < $lbd2) { $lbd_min = $lbd2; }
			if ($rbd2 < $rbd_min) { $rbd_min = $rbd2; }

			if ($rbd_min < $lbd_min) {
			    $if_nesting = $if_equal = 0;
			    last;
			}

			if (($lbd1 < $lbd2) or ($rbd2 < $rbd1)) { $if_nesting = 0; }
			unless (($lbd_sh == $lbd2) and ($rbd_sh == $rbd2)) { $if_equal = 0; }

		    } # End of the 1st inner foreach-loop (over the constituents, $gb2).

		    my $bin_rel ;
				# MODIFIED on 2019/04/12. (a) #
		    if ($if_equal and ($lbd_sh == $lbd1) and ($rbd_sh == $rbd1)) {
			$bin_rel = join ('=', $gb1, $cnct_blkset);
		    } elsif ($if_nesting) {
		    ##if ($if_nesting) {
				# END of "MODIFIED on 2019/04/12. (a)" #

			$bin_rel = join ('<', $gb1, $cnct_blkset);

		    } elsif ($if_equal) {

				# MODIFIED on 2019/04/12. (b) #
			##if (($lbd_sh == $lbd1) and ($rbd_sh == $rbd1)) {
			## $bin_rel = join ('=', $gb1, $cnct_blkset);
			##} elsif (($lbd1 <= $lbd_sh) and ($rbd_sh <= $rbd1)) {
			if (($lbd1 <= $lbd_sh) and ($rbd_sh <= $rbd1)) {
				# END of "MODIFIED on 2019/04/12. (b)" #

			    $bin_rel = join ('>', $gb1, $cnct_blkset);

			} else {
			    $bin_rel = join ('^', $gb1, $cnct_blkset); # Non-nested (we don't care whether they are "connected" or not).
			} # End of "if (($lbd1 <= $lbd_sh) and ($rbd_sh <= $rbd1)) {...} else {...}".
			##} # End of "if (($lbd_sh == $lbd1) and ($rbd_sh == $rbd1)) {...} elsif (...) {...} else {...}".

		    } elsif (($lbd_min <= $rbd_min) and ($lbd1 <= $lbd_min) and ($rbd_min <= $rbd1)) {

			my @min_gbs = ();
			foreach my $gb2 ($gb1, @{$blkset}) { # 2nd inner foreach-loop (over $gb1 and constituents of the block-set).
			    my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};
			    if (($lbd2 == $lbd_min) and ($rbd2 == $rbd_min)) {
				push @min_gbs, $gb2;
			    }
			} #  End of the 2nd inner foreach-loop (over $gb1 and constituents of the block-set).

			$bin_rel = (@min_gbs >0) ?
				join ('', $gb1, ',', $cnct_blkset, '>', join (',', @min_gbs) ) :
				join ('^', $gb1, $cnct_blkset) ;

		    } else {
			$bin_rel = join ('^', $gb1, $cnct_blkset);

		    } # End of "if ($if_nesting) {...} elsif ($if_equal) {...} elsif (($lbd_min <= $rbd_min) and ($lbd1 <= $lbd_min) and ($rbd_min <= $rbd1)) {...} else {...}".

		    push @binary_relations, $bin_rel;

		} elsif ($rel eq '>(ch)') {

		    my $gb_pa = $gb1;
		    my ($lbd_pa, $rbd_pa) = ($lbd1, $rbd1);

		    my $if_separated = 0;

		    foreach my $gb_ch (@{$blkset}) { # 3rd inner foreach-loop (over constituents of the block-set.)

			my ($lbd_ch, $rbd_ch) = @{$bds_gblocks->[$gb_ch]};

			my $if_separated2 = 1;
			if ($rbd_pa +1 < $lbd_ch) { # $gb_pa appears separated from $gb_ch, which is on the right.
			    ($if_separated2, my $lbd_ch_new, my $rbd_ch_new) = smlt_extend_left_end_to_left ($gb_ch, $gb_pa, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			    unless (defined $if_separated2) {
				print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_left_end_to_left' FAILED on the $gb_ch th gap-block against the $gb_pa th gap-block as reference (C-L): $! !\n\n";
				return undef;
			    }

			} elsif ($rbd_ch +1 < $lbd_pa) { # $gb_pa appears separated from $gb_ch, which is on the left.
			    ($if_separated2, my $lbd_ch_new, my $rbd_ch_new) = smlt_extend_right_end_to_right ($gb_ch, $gb_pa, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			    unless (defined $if_separated2) {
				print STDERR "[ERROR 'encode_gptrn_topology_psm'] Subroutine 'smlt_extend_right_end_to_right' FAILED on the $gb_ch th gap-block against the $gb_pa th gap-block as reference (D-R): $! !\n\n";
				return undef;
			    }

			} else {
				# $gb_pa and $gb_ch are obviously "connecting". #
			    $if_separated2 = 0;
			}

			if ($if_separated2) {
			    $if_separated = 1;
			    last;
			}

		    } # End of the 3rd inner foreach-loop (over constituents of the block-set).

		    my $bin_rel = ($if_separated) ? join ('|', $gb1, $cnct_blkset) : join ('-', $gb1, $cnct_blkset) ;
		    push @binary_relations, $bin_rel ;

		} # End of "if ($rel eq 'S') {...} elsif ($rel eq '>(ch)') {...}".


	    } # End of the 2nd middle for-loop (over interfering block-sets).

	} # End of the outermost for-loop (over $gb1).

	my $code_topology = (@binary_relations>0) ? join (';', @binary_relations) : 'NO_INTERFERENCE' ;

	return $code_topology ;

}	# END of the subroutine 'encode_gptrn_topology_psm'. #


			### RESTART FROM HERE (2) ###


# smlt2_extend_left_end_to_left:
#	This subroutine simulates an extention of the left-end of a subject gap-block further to the left
#	UNTIL its left-end reaches a "destination" point.
#	It was created by slightly modifying 'smlt_extend_left_end_to_left', a satellite subroutine of 'encode_gptrn_topology_psm'.
#	[CREATED on 2019/04/11.]
#
#	NOTE: It is paired with 'smlt2_extend_right_end_to_right' below.
#
# INPUTS: ($gb_sbj, $destination, \@bds_gblocks, \@gblks_w_spec_rb, \@inter_block_relations), where 
#
#	$gb_sbj is the index for the subject gap-block, whose left-end is to be extended;
#
#	$destination is the index for the column the subject gap-block aims to reach;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
# OUTPUTS: ($if_reached, $lbd_sbj_new, $rbd_sbj_new), where
#
#	$if_reached	= 1 if the left-end of $gb_sbj reaches (or overshoots) $distination,
#			= 0 otherwise;
#
#	($lbd_sbj_new, $rbd_sbj_new) are the boundaries of $gb_sbj after the simulated extension.
#
sub smlt2_extend_left_end_to_left ($$\@\@\@) {

	my ($gb_sbj, $destination, $bds_gblocks, $gblks_w_spec_rb, $inter_block_relations) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};

	if ($lbd_sbj <= $destination) { return (1, $lbd_sbj, $rbd_sbj); }

	my ($lbd_sbj_new, $rbd_sbj_new) = ($lbd_sbj, $rbd_sbj);

	my $if_reached = 0;
	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];
	my $left_flanking_gblks = $gblks_w_spec_rb->[$lbd_sbj-1];

	while (@{$left_flanking_gblks}>0) { # Attempt to "extend" the left-end of $gb_sbj to the left.

	    my $if_rlv = 0;
	    foreach my $gb3 (@{$left_flanking_gblks}) { # Middle foreach-loop (over left-flanking gap-blocks).
		my $rel = $rels_w_sbj->[$gb3];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb3 < $gb_sbj))
			or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; }

		my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
		$lbd_sbj_new = $lbd3;
		if (($rel eq 'ONN') or ($rel eq 'ONCS')) { $rbd_sbj_new -= $rbd3 - $lbd3 +1; }
		$if_rlv = 1;
		last;
	    } # End of the middle foreach-loop (over left-flanking gap-blocks).

	    unless ($if_rlv) { last; }

	    if ($destination < $lbd_sbj_new) {
		$left_flanking_gblks = $gblks_w_spec_rb->[$lbd_sbj_new-1]; # Update the list of left-flanking blocks.
	    } else {
		$if_reached = 1;
		last;
	    }

	} # End of "while (@{$left_flanking_gblks}>0) {...}".


	return ($if_reached, $lbd_sbj_new, $rbd_sbj_new);

}	# End of the subroutine 'smlt2_extend_left_end_to_left'. #


# smlt2_extend_right_end_to_right:
#	This subroutine simulates an extention of the right-end of a subject gap-block further to the right
#	UNTIL its right-end reaches a "destination" point.
#	It was created by slightly modifying 'smlt_extend_right_end_to_right', a satellite subroutine of 'encode_gptrn_topology_psm'.
#	[CREATED on 2019/04/11.]
#
#	NOTE: It is paired with 'smlt2_extend_left_end_to_left' above.
#
# INPUTS: ($gb_sbj, $destination, \@bds_gblocks, \@gblks_w_spec_lb, \@inter_block_relations), where 
#
#	$gb_sbj is the index for the subject gap-block, whose left-end is to be extended;
#
#	$destination is the index for the column the subject gap-block aims to reach;
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
# OUTPUTS: ($if_reached, $lbd_sbj_new, $rbd_sbj_new), where
#
#	$if_reached	= 1 if the right-end of $gb_sbj reaches (or overshoots) $distination,
#			= 0 otherwise;
#
#	($lbd_sbj_new, $rbd_sbj_new) are the boundaries of $gb_sbj after the simulated extension.
#
sub smlt2_extend_right_end_to_right ($$\@\@\@) {

	my ($gb_sbj, $destination, $bds_gblocks, $gblks_w_spec_lb, $inter_block_relations) = @_;

	my ($lbd_sbj, $rbd_sbj) = @{$bds_gblocks->[$gb_sbj]};

	if ($destination <= $rbd_sbj) { return (1, $lbd_sbj, $rbd_sbj); }

	my ($lbd_sbj_new, $rbd_sbj_new) = ($lbd_sbj, $rbd_sbj);

	my $if_reached = 0;
	my $rels_w_sbj = $inter_block_relations->[$gb_sbj];
	my $right_flanking_gblks = $gblks_w_spec_lb->[$rbd_sbj +1];

	while (@{$right_flanking_gblks}>0) { # Attempt to "extend" the right-end of $gb_sbj to the right.

	    my $if_rlv = 0;
	    foreach my $gb3 (@{$right_flanking_gblks}) { # Middle foreach-loop (over right-flanking gap-blocks).
		my $rel = $rels_w_sbj->[$gb3];
		unless (($rel eq '<') or ($rel eq '<(pa)') or (($rel eq '=') and ($gb3 < $gb_sbj))
			or ($rel eq 'ONN') or ($rel eq 'ONCS') ) { next; }

		my ($lbd3, $rbd3) = @{$bds_gblocks->[$gb3]};
		$rbd_sbj_new = $rbd3;
		if (($rel eq 'ONN') or ($rel eq 'ONCS')) { $lbd_sbj_new += $rbd3 - $lbd3 +1; }
		$if_rlv = 1;
		last;
	    } # End of the middle foreach-loop (over right-flanking gap-blocks).

	    unless ($if_rlv) { last; }

	    if ($rbd_sbj_new  < $destination) {
		$right_flanking_gblks = $gblks_w_spec_lb->[$rbd_sbj_new +1]; # Update the list of right-flanking blocks.
	    } else {
		$if_reached = 1;
		last;
	    }

	} # End of "while (@{$right_flanking_gblks}>0) {...}".


	return ($if_reached, $lbd_sbj_new, $rbd_sbj_new);

}	# End of the subroutine 'smlt2_extend_right_end_to_right'. #



# examine_if_gbs_on_msa_ends:
#	This subroutine examines whether each of the gap-blocks
#	are (effectively) on the left-end or the right-end of the input MSA.
#	[CREATED on 2019/05/23.]
#
# INPUTS: (\@bl_coords, \@gb_coords_wd_frm, $if_on_le_msa, $if_on_re_msa), where
#
#	$bl_coords[$gb] is the current coordinate of the $gb th gap-block;
#
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give
#	the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the current window;
#
#	$if_on_le/re_msa = 1 if the current window is on the left/right-end of the input MSA;
#
# OUTPUTS: (\@if_gbs_on_msa_ends), where
#
#	$if_gbs_on_msa_ends[$gb] = 'N', 'L', 'R', 'B', if the $gb th gap-block is
#	on neither end, only on the left-end, only on the right-end, and on both ends, respectively, 
#	of the input MSA.
#
sub examine_if_gbs_on_msa_ends (\@\@$$) {

	my ($bl_coords, $gb_coords_wd_frm, $if_on_le_msa, $if_on_re_msa) = @_;

	my $ct_gbs = @{$bl_coords};

	my @if_gbs_on_msa_ends = ();

	unless ($if_on_le_msa or $if_on_re_msa) {
		# The window is neither on the left-end nor on the right-end of the MSA. #
	    for (my $k=0; $k<$ct_gbs; $k++) { $if_gbs_on_msa_ends[$k] = 'N'; }
	    return \@if_gbs_on_msa_ends ;
	}

	for (my $k=0; $k<$ct_gbs; $k++) { # Main for-loop (over gap-blocks). #

	    my $status = 'N'; # By default, the block is assumed to be on Neither MSA-end.

	    my $coord = $bl_coords->[$k];
	    my ($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm->[$k]};

	    if ($if_on_le_msa and ($coord == $gb_coord_le_wd)) { $status = 'L'; }
	    if ($if_on_re_msa and ($coord == $gb_coord_re_wd)) { $status = ($status eq 'L') ? 'B' : 'R' ; }

	    $if_gbs_on_msa_ends[$k] = $status;

	}	# End of the Main for-loop (over gap-blocks). #

	return \@if_gbs_on_msa_ends ;

}	# End of the subroutine 'examine_if_gbs_on_msa_ends'. #


# chk_excluded_bin_hrels:
#	This subroutine receives a list of binary horizontal relationships
#	that will cause the current alignment to be excluded from the analysis.
#	Then, it examines whether or not any of the relationships indeed hold.
#	[CREATED on 
#
# INPUTS: (\@list_excluded_bin_hrels, \@bds_gblocks, \@gblks_w_spec_lb, \@gblks_w_spec_rb, \@inter_block_relations), where
#
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined;
#	Currently, the following codes can possibly be in the list:
#	'-', indicating that the gap-blocks are "connected" (overlapping or touching);
#	'>', indicating that $gb1 horizontally includes $gb2;
#	'<', indicating that $gb1 is horizontally included in $gb2;
#	'=', indicating that $gb1 is horizontally identical to $gb2;
#	(NOTE that the relationship assignment is the opposite of that for the vertical relationships);
#	(See the header of 'encode_gptrn_topology_psm' for detailed descriptions of all codes);
#
#	($lbd, $rbd) = @{$bds_gblocks[$gb]} give the indexes of the columns
#	at the left- and right-boundaries of the $gb th gap-block;
#	(The boundaries are included in the block);
#	[Typically, \@bds_gblocks is an output of 'extract_br_info_gblocks' in "MyANEX_Supple.pm";
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	@{$gblks_w_spec_lb[$lb]} lists the gap-blocks whose left-bounds are at the $lb th column;
#	[Typically, \@gblks_w_spec_lb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	@{$gblks_w_spec_rb[$rb]} lists the gap-blocks whose right-bounds are at the $rb th column;
#	[Typically, \@gblks_w_spec_rb is an output of 'mk_lists_gblocks_w_spec_bounds' above];
#
#	$inter_block_relations[$gb1]->[$gb2] = $relation
#	gives the relation of $gb2 against $gb1;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' above, or 'extensive_exam_inter_block_relations' above];
#
# OUTPUT: ($if_excluded), where
#
#	$if_excluded	= 1 if any of the relationships hold,
#			= 0 if none of them is true (including the cases where @list_excluded_bin_hrels is empty).
#
sub chk_excluded_bin_hrels (\@\@\@\@\@) {

	my ($list_excluded_bin_hrels, $bds_gblocks, $gblks_w_spec_lb, $gblks_w_spec_rb, $inter_block_relations) = @_;

	my $ct_x_bhrels = @{$list_excluded_bin_hrels};
	unless ($ct_x_bhrels>0) {
	    return 0; # The list is empty.
	}

	for (my $hr=0; $hr<$ct_x_bhrels; $hr++) { # Main for-loop (over binary relationships in the list).

	    my ($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels->[$hr]};

	    if ($gb1 > $gb2) {
		# Swap the gap-blocks, so that ($gb1 < $gb2) always holds.  #

		my $tmp = $gb1;
		$gb1 = $gb2;
		$gb2 = $tmp;

		if ($hrel eq '<') {
		    $hrel = '>';
		} elsif ($hrel eq '>') {
		    $hrel = '<';
		}
	    } # End of "if ($gb1 > $gb2) {...}"

	    my ($lbd1, $rbd1) = @{$bds_gblocks->[$gb1]};
	    my ($lbd2, $rbd2) = @{$bds_gblocks->[$gb2]};

	    if ($hrel eq '=') { # Horizontally identical.

		if (($lbd1 == $lbd2) and ($rbd1 == $rbd2)) { return 1; }

	    } elsif ($hrel eq '>') { # $gb1 horizontally includes $gb2.

		if (($lbd1 <= $lbd2) and ($rbd2 <= $rbd1)) { return 1; }

			# NOTE: The above condition actually includes '=';
			#	we are being a bit too lenient here.

	    } elsif ($hrel eq '<') { # $gb1 is horizontally included in $gb2.

		if (($lbd2 <= $lbd1) and ($rbd1 <= $rbd2)) { return 1; }

			# NOTE: The above condition actually includes '=';
			#	we are being a bit too lenient here.

	    } elsif ($hrel eq '-') { # $gb1 and $gb2 are "connected".

		if ($rbd1 +1 < $lbd2) { # $gb2 appears separated on the right of $gb1.

		    if ((@{$gblks_w_spec_lb->[$rbd1+1]}>0) and (@{$gblks_w_spec_rb->[$lbd2-1]}>0)) {
			my ($if_reached, $lbd2_new, $rbd2_new) = smlt2_extend_left_end_to_left ($gb2, $rbd1 +1, @{$bds_gblocks}, @{$gblks_w_spec_rb}, @{$inter_block_relations});
			unless (defined $if_reached) {
			    print STDOUT "[ERROR 'chk_excluded_bin_hrels'] The subroutine, 'smlt2_extend_left_end_to_left' FAILED on the $gb2 th gap-block, while attempting to get it to connect $gb1: $! !\n\n";
			    return undef;
			}
			if ($if_reached) { return 1; } # $gb2 is actually "connected" to $gb1.
		    }

		} elsif ($rbd2 < $lbd1 -1) { # $gb2 appears separated on the left of $gb1.

		    if ((@{$gblks_w_spec_lb->[$rbd2+1]}>0) and (@{$gblks_w_spec_rb->[$lbd1-1]}>0)) {
			my ($if_reached, $lbd2_new, $rbd2_new) = smlt2_extend_right_end_to_right ($gb2, $lbd1 -1, @{$bds_gblocks}, @{$gblks_w_spec_lb}, @{$inter_block_relations});
			unless (defined $if_reached) {
			    print STDOUT "[ERROR 'chk_excluded_bin_hrels'] The subroutine, 'smlt2_extend_right_end_to_right' FAILED on the $gb2 th gap-block, while attempting to get it to connect $gb1: $! !\n\n";
			    return undef;
			}
			if ($if_reached) { return 1; } # $gb2 is actually "connected" to $gb1.
		    }

		} else { # $gb1 and $gb2 are clearly "connected".
		    return 1;
		}

	    } else {
		print STDOUT "[ERROR 'chk_excluded_bin_hrels'] An UNEXPECTED horizontal relationship, '$hrel', is specified between $gb1 and $gb2: $! !\n\n";
		return undef;
	    }

	} # End of the main for-loop (over binary relationships in the list).

	return 0; # None of the relationships are true.

}	# End of the subroutine 'chk_excluded_bin_hrels'. #



# expand_list_x_bhrels: 
#	This subroutine attempts to expand the list of "excluded" binary horizontal relationships,
#	so that all swappable equivalents of the relevant gap-blocks will be monitored.
#	[CREATED on 2020/05/17.]
#
# INPUTS: (\@list_excluded_bin_hrels, \@sets_swappable_gblks), where
#
#	($gb1, $gb2, $hrel) = @{$list_excluded_bin_hrels[$hr]} provide information on the $hr th horizontal relationship in the list, where
#	$gb1 and $gb2 are indices of the pair of gap-blocks in question,
#	and $hrel codes the horizontal relationships to be examined.
#	[Some specific codes are described in the header of 'chk_excluded_bin_hrels' above.]
#
#	@{$sets_swappable_gblks[$j]} lists the indices of the gap-blocks
#	in the $j th set of swappable gap-blocks.
#	[Typically, \@sets_swappable_gblks is a main output of 'mk_sets_swappable_gblks' above.]
#
# OUTPUTS: (\@exp_list_excluded_bin_hrels), where
#
#	@exp_list_excluded_bin_hrels is the expanded list of "excluded" binary horizontal relationships;
#	It follows the same format as the input @list_excluded_bin_hrels.
#
sub expand_list_x_bhrels (\@\@) {

	my ($list_excluded_bin_hrels, $sets_swappable_gblks) = @_;

	my $ct_x_bhrels = @{$list_excluded_bin_hrels};
	my $ct_sets_swpbl = @{$sets_swappable_gblks};

	unless (($ct_x_bhrels>0) and ($ct_sets_swpbl>0)) {

			# There is no chance of expansion. #

	    return $list_excluded_bin_hrels;
	}

			# Preliminary: Associate the constituent gap-blocks with the swappable sets they belong to. #

	my %indx_gb2set_swpbl;
	for (my $s=0; $s <$ct_sets_swpbl; $s++) { # Preliminary for-loop. #

	    my $set_swpbl = $sets_swappable_gblks->[$s];
	    foreach my $indx_gb (@{$set_swpbl}) {
		$indx_gb2set_swpbl{$indx_gb} = $set_swpbl;
	    }
	} # End of the preliminary for-loop. #

			# MAIN: Attempt to expand the listed binary horizontal relationships. #

	my @exp_list_excluded_bin_hrels = ();

	for (my $hr =0; $hr < $ct_x_bhrels; $hr++) { # Main outer for-loop (over binary relationships). #

	    my ($gb10, $gb20, $hrel) = @{$list_excluded_bin_hrels->[$hr]};

	    my $set_swpbl1 = $indx_gb2set_swpbl{$gb10};
	    unless (defined $set_swpbl1) { $set_swpbl1 = [$gb10]; }

	    my $set_swpbl2 = $indx_gb2set_swpbl{$gb20};
	    unless (defined $set_swpbl2) { $set_swpbl2 = [$gb20]; }


	    foreach my $gb1 (@{$set_swpbl1}) { # Middle for-loop (over swappable equivalents of $gb10). #
		foreach my $gb2 (@{$set_swpbl2}) { # Inner for-loop (over swappable equivalents of $gb20). #

		    push @exp_list_excluded_bin_hrels, [$gb1, $gb2, $hrel];

		} # End of the inner for-loop (over swappable equivalents of $gb20). #
	    } # End of the middle for-loop (over swappable equivalents of $gb10). #

	} # End of the main outer for-loop (over binary relationships). #


	return \@exp_list_excluded_bin_hrels;

}	# End of 'expand_list_x_bhrels'. #



=Set aside for later use (on 2019/03/09).
			    my $rel;
			    if ($u_or_d1 eq 'U') {
				if ($u_or_d2 eq 'U') {
				} else { # if ($u_or_d2 eq 'L')
				}
			    } else { # if ($u_or_d1 eq 'L')
				if ($u_or_d2 eq 'U') {
				} else { # if ($u_or_d2 eq 'L')
				}
			    } # End of "if ($u_or_d1 eq 'U') {...} else {...}".
			    if (defined $rel) {
			    } # End of "if (defined $rel) {...}".

=cut # End of "Set aside for later use (on 2019/03/09)."



1;	# a "spell" which makes the module "use"-able and/or "require"-able #



