package MyTreeMap_indels_ML;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_indels_ML.pm:
#	This package calculates an approximate log-likelihood of
#	the gap pattern of an input MSA,
#	using the set of parsimonious indel scenarios realizing the gap pattern.
#
#	[ REVISION from 'MyTreeMap_indels_ML_alpha.pm' ]
#	Whenever possible, 
#	this version uses the logarithms of the (relative) frequencies of indel events,
#	to prevent the subroutines from crashing unexpectedly.
#	
#	(CREATION STARTED on June ?? (???), 2012, ENDED on July ?? (???), 2012)
#
#	[ REVISED on Mar 2 (Sun), 2014]
#	The subroutine 'apprx1_tot_lnlk_gpattern_msa' can now continue its analysis
#	even if some of the gap-block clusters have empty sets for parsimonious scenarios,
#	which indicates that there were some abnormalities,
#	while skipping the calculation of likelihoods for such clusters
#	(and setting their likelihood 'N/A'). 
#
#	[ REVISED on Mar 6 (Thu), 2014 ]
#	The subroutine 'apprx1_tot_lnlk_gpattern_msa' can now continue its analysis
#	even if its satellite subroutine, 'cal_lnlk1_indel_scn_cls', returns undefined outputs,
#	while registering the likelihood in question as 'undef'.
#
#	[ REVISED on Mar 10 (Mon), 2014 ]
#	Accompanying the revision of 'MyTreeMap_indels.pm',
#	the subroutine 'apprx1_tot_lnlk_gpattern_msa' was slightly modified
#	in order to handle exceptional cases which probably involves (a) long gap(s).
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

=set aside for later use
use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	log_freq_indel_len
	cal_log_normal_fac_distri
	cal_log_normal_facs_indel_distri
	cal_delta_exp_del_5p
	log_dbl_sum_freq_indel
	log_compos_freq_indel_len
	dbl_tinteg_exp_indel
	search_paired_indels
	search_leftmost_events
	search_rightmost_events
	cal_lnlk1_sgl_indel_cls
	cal_lnlk1_pair_indel_cls
	cal_lnlk1_indel_br_cls
	cal_lnlk1_indel_scn_cls
	apprx1_tot_lnlk_gpattern_msa
	);
	## enum_psm_indel_scns_for_msa


	# This specifies the level of tests on the main subroutines,
	# 'apprx1_tot_lnlk_gpattern_msa', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
our $TEST_LEVEL = 5;

	# This specifies the level of tests on the primary satellite subroutines,
	# 'cal_lnlk1_indel_scn_cls', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 3;
our $TEST_LEVEL1 = 5;

	# This specifies the level of tests on the sub-primary satellite subroutine,
	# 'cal_lnlk1_indel_br_cls'.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL1a = 1;
##our $TEST_LEVEL1a = 2;
##our $TEST_LEVEL1a = 3;
##our $TEST_LEVEL1a = 5;

	# This specifies the level of tests on the important secondary satellite subroutine,
	# 'cal_lnlk1_sgl_indel_cls'.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL2a = 0;
##our $TEST_LEVEL2a = 1;
##our $TEST_LEVEL2a = 2;
##our $TEST_LEVEL2a = 3;
##our $TEST_LEVEL2a = 5;

	# This specifies the level of tests on the important secondary satellite subroutine,
	# 'cal_lnlk1_pair_indel_cls'.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL2b = 0;
##our $TEST_LEVEL2b = 1;
##our $TEST_LEVEL2b = 2;
##our $TEST_LEVEL2b = 3;
##our $TEST_LEVEL2b = 5;

	# This specifies the level of tests on the satellite subroutine 'log_compos_freq_indel_len'.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL2c = 0;
##our $TEST_LEVEL2c = 1;
##our $TEST_LEVEL2c = 2;
##our $TEST_LEVEL2c = 3;
##our $TEST_LEVEL2c = 5;

	# This specifies the level of tests on the satellite subroutine 'search_paired_indels'.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL2s = 0;
##our $TEST_LEVEL2s = 1;
##our $TEST_LEVEL2s = 2;
##our $TEST_LEVEL2s = 3;
##our $TEST_LEVEL2s = 5;

	# This specifies the level of tests on the elementary subroutines.
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL3 = 0;
##our $TEST_LEVEL3 = 1;
##our $TEST_LEVEL3 = 2;
##our $TEST_LEVEL3 = 3;
##our $TEST_LEVEL3 = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %node2pars_stats = ($node_id => \@stat_info, ...) #
	#
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	# This parameter is important to decide whether we should employ
	# the direct summation over indel lengths or its complement,
	# when calculating a 'composite' frequency.

our $MID_INDEL_LEN = 20; # An indel longer than this is regarded as 'long' => direct summation will be calculated.
## our $MID_INDEL_LEN = 10;

	## This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	# (ADDED on Mar 10 (Mon), 2014.)
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = 150;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	# (ADDED on Mar 11 (Tue), 2014.):w
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# create_initial_set_indels:
#	This satellite subroutine of 'enum_psm_indel_scns_in_cls_blks' creates
#	the initial 'tentative' candidate of the parsimonious indel scenario,
#	out of the raw output of the Dollo parsimony algorithm on the cluster of alignment blocks.
#	Each indel event is made by merging the neighboring (or "quasi-neighboring") events of the same type along the same branch.
#	The "quasi_neighboring" events are indel events NOT separated by (a) block(s) with $char_pres at BOTH ends of the branch in question.
#

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# log_freq_indel_len:
#	This revised subroutine calculates the logarithm of the frequency.
#
# freq_indel_len:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa'
#	calculates the relative frequency of insertions/deletions of a given length among all insertions/deletions.
#	This version only implements a geometric distribution:
#	freq_indel ($l) = $q**($l-1) / $nfac_indel,
#	 and a zeta power-law distribution:
#	freq_indel ($l) = $l**(-$alpha) / $nfac_indel.
#
# [ INPUT: ($distri_indel, $prm_indel, $l, $nfac_indel), where ] -- OBSOLETE as of July 11 (Wed), 2012. 
# INPUT: ($distri_indel, $prm_indel, $l, $log_nfac_indel), where
#	$distri_indel specifies the type of the distribution, with
#		= 'G' designating a geometric distribution, and
#		= 'P' designating a zeta power-law distribution;
#	$prm_indel = $q above (0<=$q<1) if $distri_indel = 'G',
#		   = $alpha above (>= 1) if $distri_indel = 'P';
#	$l is the length of the insertion/deletion;
#	$log_nfac_indel is the logarithm of the normalization factor, $nfac_indel,
#	used in the above definition.
#
# [ OUTPUT: $freq_len, which is freq_indel ($l) given above. ] --- OBSOLETE as of July 11 (Wed), 2012. 
# OUTPUT: $log_freq_len, which is the logarithm of freq_indel ($l) given above.
#
sub log_freq_indel_len ($$$$) {

	my ($distri_indel, $prm_indel, $l, $log_nfac_indel) = @_;

	unless ($l>=1) {
	    print STDOUT "[ERROR 'log_freq_indel_len'] ILLEGAL INPUT: UNEXPECTED INDEL LENGTH (=$l): $! !\n";
	    return undef;
	}

	my $log_freq_len ;
	if ($distri_indel eq 'G') {
	    unless (($prm_indel>= 0) and ($prm_indel<1)) {
		print STDOUT "[ERROR 'log_freq_indel_len'] ILLEGAL INPUT: UNEXPECTED PARAMETER for the geometric distribution (\$q = $prm_indel NOT in [0, 1)): $! !\n";
		return undef;
	    }
	    if ($prm_indel > 0) {
		$log_freq_len = ($l-1) * log($prm_indel) ;
		##$freq_len = exp (($l-1)*log($prm_indel)) ;
	    } else {
		if ($l > 1) {
		    print STDOUT "[ERROR 'log_freq_indel_len'] ILLEGAL INPUT: ", '$prm_indel = ', "$prm_indel and yet the indel length = $l (>1), which is impossible: $! !\n\n";
		    return undef;
		} else {
		    $log_freq_len = 0;
		}
		## $freq_len = ($l > 1) ? 0 : 1;
	    }
		
	} elsif ($distri_indel eq 'P') {
	    unless ($prm_indel >= 1) {
		print STDOUT "[ERROR 'log_freq_indel_len'] ILLEGAL INPUT: UNEXPECTED PARAMETER for the power-law distribution (\$alpha = $prm_indel LESS THAN 1): $! !\n";
		return undef;
	    }
	    $log_freq_len = - $prm_indel * log ($l) ;
	    ##$freq_len = exp (- $prm_indel * log ($l)) ;

	} else {
	    print STDOUT "[ERROR 'log_freq_indel_len'] ILLEGAL INPUT: UNEXPECTED DISTRIBUTION TYPE (= '$distri_indel'): $! !\n";
	    return undef;
	}

	$log_freq_len -= $log_nfac_indel;
	##$freq_len /= $nfac_indel;

	return $log_freq_len ;

}	# End of the subroutine 'log_freq_indel_len' #


# cal_log_normal_fac_distri:
#	This satellite subroutine of 'cal_log_normal_facs_indel_distri' calculates
#	the logarithm of the normalization factor for a single discrete distribution.
#
# INPUT: ($distri, $param, $ub), where
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	$ub is the upper-bound of the discrete variable $l.
#	they should satisfy $lb <= $ub,
#	where $lb is the lower-bound, which SHOULD ALWAYS BE ONE!!
#
# OUTPUT: $log_nfac, where
#	$log_nfac = log { Sum_{$l=$lb}^{$ub} distri ($l) }
#	is the logarithm of the normalization factor.
#
sub cal_log_normal_fac_distri ($$$) {

	my ($distri, $param, $ub) = @_;

	my $lb = 1; # The lower-bound should always be 1.

	unless ($lb <= $ub) {
	    print STDOUT "[ERROR 'cal_log_normal_fac_distri'] ILLEGAL INPUT: The lower-bound (=$lb) of the variate range EXCEEDS the upper-bound (=$ub): $! !\n";
	    return undef;
	}

	if ($distri eq 'G') {
	    unless (($param>= 0) and ($param <1)) {
		print STDOUT "[ERROR 'cal_log_normal_fac_distri'] ILLEGAL INPUT: UNEXPECTED PARAMETER for the geometric distribution (\$q = $param NOT in [0, 1)): $! !\n";
		return undef;
	    }
	    if ($param ==  0) {
		print STDOUT "[WARNING 'cal_log_normal_fac_distri'] The distribution is geometric with \$q == 0, so I will make the upper-bound, which used to be $ub, equal to the lower-bound (=$lb) ... !\n";
		$ub = $lb ;
	    }
		
	} elsif ($distri eq 'P') {
	    unless ($param >= 1) {
		print STDOUT "[ERROR 'cal_log_normal_fac_distri'] ILLEGAL INPUT: UNEXPECTED PARAMETER for the power-law distribution (\$alpha = $param LESS THAN 1): $! !\n";
		return undef;
	    }
	} else {
	    print STDOUT "[ERROR 'cal_log_normal_fac_distri'] ILLEGAL INPUT: UNEXPECTED DISTRIBUTION TYPE (= '$distri'): $! !\n";
	    return undef;
	}

	my @terms_log_nfac = ();
	for (my $l=$lb; $l <= $ub; $l++) { # Main for-loop (over the range of the discrete variate) #
	    my $term = log_freq_indel_len ($distri, $param, $l, 0);
	    unless (defined $term) {
		print STDOUT "[ERROR 'cal_log_normal_fac_distri'] THe subroutine 'log_freq_indel_len' FAILED for the length $l, with \$distri = $distri and \$param = $param: $! !\n";
		return undef;
	    }
	    push @terms_log_nfac, $term;

	}		# End of the main for-loop (over the range of the discrete variate) #

	my $log_nfac = log_sum_exp (@terms_log_nfac);

	return $log_nfac;

}	# End of the subroutine 'cal_log_normal_fac_distri' #


# cal_log_normal_facs_indel_distri:
#	This subroutine calculates the logarithms of the normalization factors.
#
# cal_normal_facs_indel_distri:
#	This subroutine calculates normalization factors.
#
#	This subroutine was introduced mainly because the normalization factors
#	are universal, once the distributions (and cut-off lengths) are fixed.
#
# INPUT: (\@params_ins, \@params_del, $cutoff_5p, $cutoff_3p), where
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	$cutoff_5p/3p is the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region;
#
# [ OUTPUT: \@nfacs, where
#	@nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p), with ] --- OBSOLETE as of July 11 (Wed), 2012. 
# OUTPUT: \@log_nfacs, where
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p), with
#	$log_nfac_ins/del_5p/3p = log { Sum_{$l=1}^{$cutoff_5p/3p} distr_ins/del ($l) }
#
sub cal_log_normal_facs_indel_distri (\@\@$$) {

	my ($params_ins, $params_del, $cutoff_5p, $cutoff_3p) = @_;

	my ($totrate_ins, $distr_ins, $prm_ins) = @{$params_ins};
	my ($totrate_del, $distr_del, $prm_del) = @{$params_del};

	my $log_nfac_ins_5p = cal_log_normal_fac_distri ($distr_ins, $prm_ins, $cutoff_5p);
	unless (defined $log_nfac_ins_5p) {
	    print STDOUT "[ERROR 'cal_log_normal_facs_indel_distri'] The subroutine 'cal_log_normal_fac_distri' FAILED for the insertions extending into the 5'-end: $! !\n";
	    return undef;
	}
	my $log_nfac_del_5p = cal_log_normal_fac_distri ($distr_del, $prm_del, $cutoff_5p);
	unless (defined $log_nfac_del_5p) {
	    print STDOUT "[ERROR 'cal_log_normal_facs_indel_distri'] The subroutine 'cal_log_normal_fac_distri' FAILED for the deletions extending into the 5'-end: $! !\n";
	    return undef;
	}

	##my $nfac_ins_5p = my $nfac_del_5p = 0;
	##for (my $l=1; $l<= $cutoff_5p; $l++) { # 3rd preliminary for-loop #
	##    $nfac_ins_5p += freq_indel_len ($distr_ins, $prm_ins, $l, 1);
	##    $nfac_del_5p += freq_indel_len ($distr_del, $prm_del, $l, 1);
	##}			# End of the 3rd preliminary for-loop #

	my $log_nfac_ins_3p = cal_log_normal_fac_distri ($distr_ins, $prm_ins, $cutoff_3p);
	unless (defined $log_nfac_ins_3p) {
	    print STDOUT "[ERROR 'cal_log_normal_facs_indel_distri'] The subroutine 'cal_log_normal_fac_distri' FAILED for the insertions possibly extending into the 3'-end: $! !\n";
	    return undef;
	}
	my $log_nfac_del_3p = cal_log_normal_fac_distri ($distr_del, $prm_del, $cutoff_3p);
	unless (defined $log_nfac_del_3p) {
	    print STDOUT "[ERROR 'cal_log_normal_facs_indel_distri'] The subroutine 'cal_log_normal_fac_distri' FAILED for the deletions possibly extending into the 3'-end: $! !\n";
	    return undef;
	}

	##my $nfac_ins_3p = my $nfac_del_3p = 0;
	##for (my $l=1; $l<= $cutoff_3p; $l++) { # 4th preliminary foreach-loop #
	##    $nfac_ins_3p += freq_indel_len ($distr_ins, $prm_ins, $l, 1);
	##    $nfac_del_3p += freq_indel_len ($distr_del, $prm_del, $l, 1);
	##}			# End of the 4th preliminary foreach-loop #

	my @log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p);
	##my @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p);

	return \@log_nfacs;

}		# End of the subroutine 'cal_log_normal_facs_indel_distri' #


# cal_delta_exp_del_5p:
#	This subroutine calculates an excessive exponential factor for the likelihood of indel
#	scenarios (due to (lack of) deletions extending to the 5'-end):
#	$delta_del_5p = Sum_{$l=1}^{$cutoff_5p} ($l-1) * distri_del ($l),
#	where distri_del ($l) is the relative frequency of deletions with length $l.
#	(The actual exponential factor is:
#	 - $delta_del_5p * {total deletion rate} * {total tree length}.) 
#
#	This was introduced mainly because this factor is shared by any MSAs,
#	irrespective of phylogenetic trees. 
#
# [ INPUT: (\@params_del, $cutoff_5p, $nfac_del_5p), where ] --- OBSOLETE as of July 12 (Wed), 2012.
# INPUT: (\@params_del, $cutoff_5p, $log_nfac_del_5p), where
#	@params_del = ($totrate_del, $distr_del, $prm_del), with
#	$totrate_del being the total rate of deletions (relative to the substitution rate),
#	$distr_del = 'G' specifying the geometric distribution,
#		     'P' specifying the (zeta) power-law distribution,
#	$prm_del   = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_del = 'G',
#		   = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_del = 'P';
#
#	$cutoff_5p is the cut-off length common for the deletions on the 5'-side of the aligned region;
#
#	[ $nfac_del_5p (= Sum_{$l=1}^{$cutoff_5p} distr_del ($l)) ]
#	$log_nfac_del_5p (= log { Sum_{$l=1}^{$cutoff_5p} distr_del ($l) })
#	is the logarithm of the normalization factor for the deletions extending into the 5'-end.
#
# OUTPUT: $delta_del_5p, which is defined above.
#
sub cal_delta_exp_del_5p (\@$$) {

	my ($params_del, $cutoff_5p, $log_nfac_del_5p) = @_; 

	my ($totrate_del, $distr_del, $prm_del) = @{$params_del};
	if (($distr_del eq 'G') and ($prm_del == 0)) {
	   print STDOUT "[WARNING 'cal_delta_exp_del_5p'] The distribution os geometric with q = 0, so the cutoff (=$cutoff_5p) will be made 1 !\n";
	   $cutoff_5p = 1;
	}

	my $delta_del_5p = 0;
	for (my $l=1; $l<= $cutoff_5p; $l++) { # 5th preliminary foreach-loop #
	    my $log_freq = log_freq_indel_len ($distr_del, $prm_del, $l, $log_nfac_del_5p);
	    $delta_del_5p += ($l-1)* exp ($log_freq);
	    ##my $freq = freq_indel_len ($distr_del, $prm_del, $l, $nfac_del_5p);
	    ##$delta_del_5p += ($l-1)* $freq;
	}			# End of the 5th preliminary foreach-loop # 

	return $delta_del_5p;

}		# End of the subroutine 'cal_delta_exp_del_5p' #


# log_dbl_sum_freq_indel:
#	This revised subroutine calculates the logarithm of the double summation.
#
# dbl_sum_freq_indel:
#	This subroutine calculates a double summation of the indel frequency f(l):
#
#	$dbl_sum = Sum_{$x=0}^{$M5-int($l/2)} Sum_{$y=$l}^{$M3+int($l/2)} f($x+$y)
#	= Sum_{$z=$l}^{$M5+$M3} f($z) * min [$z-$l+1, $M5 - int($l/2)+1, $M5+$M3 - $z+1].
#	Note that f($z) is NOT normalized yet.
#
# INPUT: ($distri, $param, $max_5, $max_3, $l), where
#
#	$distri specifies the type of the distribution, with
#		= 'G' designating a geometric distribution, and
#		= 'P' designating a zeta power-law distribution;
#	$param = $q above (0<=$q<1) if $distri_indel = 'G',
#		   = $alpha above (>= 1) if $distri_indel = 'P';
#	$max_5/$max_3 is $M5/$M3 in the above description;
#	$l is $l above.
#
# [ OUTPUT: $dbl_sum, which is double sum just as defined above. ] --- OBSOLETE as of July 11 (Wed), 2012.
# OUTPUT: $log_dbl_sum, which is the logarithm of the double sum just as defined above.
#
sub log_dbl_sum_freq_indel ($$$$$) {

	my ($distri, $param, $max_5, $max_3, $l) = @_;

	unless (($max_5>0) and ($max_3>0)) {
	    print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] ILLEGAL INPUT: either \$max_5 (=$max_5) or $max_3 (=$max_3) is NON-POSITIVE: $! !\n";
	    return undef;
	}
	unless ($l >0) {
	   print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] ILLEGAL INPUT: The indel length (=$l) is NOT POSITIVE: $! !\n";
	   return undef;
	}
	my $ub_5 = $max_5 - int ($l/2) + 1;
	unless ($ub_5>0) {
	    print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] ILLEGAL INPUT: ", '$ub_5 (= $max_5 - int ($l/2) + 1)', " = $ub_5 is NOT POSITIVE: $! !\n";
	    return undef;
	}
	## my $ub_3 = $max_3 + int ($l/2) + 1;
	my $ub_z = $max_5 + $max_3 + 1;
	unless ($l < $ub_z) {
	    print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] ILLEGAL INPUT: The indel length (=$l) is NOT LESS THAN ", '$ub_z (= $max_5 + $max_3 + 1)', " = $ub_z: $! !\n";
	    return undef;
	}

		# A Bug-fixation measure on July 15 (Sun), 2012. #
	my $width_3 = $max_3 + int ($l/2) - $l + 1;
	unless ($width_3>0) {
	    print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] ILLEGAL INPUT: The number of summands in the 3'-direction (=$width_3) is NOT POSITIVE: $! !\n";
	    return undef;
	}

	if (($distri eq 'G') and ($param == 0)) { # An exceptional case. #
	    unless ($l == 1) {
		print STDOUT "[ERROR 'log_dbl_sum_freq_indel] ILLEGAL INPUT: The geometric distribution with q = 0, and yet the indel length (=$l) is LONGER THAN 1: $! !\n";
		return undef;
	    }
	    my $log_dbl_sum = log_freq_indel_len ($distri, $param, 1, 0);
	    print STDOUT "[WARNING 'log_dbl_sum_freq_indel'] The geometric distribution with q = 0 (and length = 1) should ALWAYS HAVE \$log_dbl_sum = $log_dbl_sum ...\n";
	    return $log_dbl_sum;
	}

	my @terms_log_dbl_sum = ();
	##my $dbl_sum = 0;
	for (my $z = $l; $z<$ub_z; $z++) { # Main for-loop #

	    my $coeff = $z-$l+1;
	    ($coeff > $ub_5) and ($coeff = $ub_5);
	    ($coeff > $ub_z - $z) and ($coeff = $ub_z - $z);

		# A Bug-fixation measure on July 15 (Sun), 2012. #
	    ($coeff > $width_3) and ($coeff = $width_3);

	    my $log_freq = log_freq_indel_len ($distri, $param, $z, 0);
	    ##my $freq = freq_indel_len ($distri, $param, $z, 1);
	    unless (defined $log_freq) {
		print STDOUT "[ERROR 'log_dbl_sum_freq_indel'] The subroutine 'freq_indel_len' FAILED for the indel length $z: $! !\n";
		return undef;
	    }

	    $log_freq += log $coeff;

	    push @terms_log_dbl_sum, $log_freq; 
	    ##$dbl_sum += $coeff * $freq;

	}			# End of the main for-loop #

	my $log_dbl_sum = log_sum_exp (@terms_log_dbl_sum);

	return $log_dbl_sum ;
	##return $dbl_sum ;

}	# End of the subroutine 'log_dbl_sum_freq_indel' #


# log_compos_freq_indel_len:
#	This revised subroutine calculates the logarithms of the composite frequency.
#
# compos_freq_indel_len:
#	This satellite subroutine of 'cal_lnlk1_sgl_indel_cls' and of 'cal_lnlk1_pair_indel_cls'
#	calculates the composite frequency of indel lengths (when $opt_bd_indels = 1).
#
# [ INPUT: ($if_5p_end_indel, $if_3p_end_indel, $distri, $param, $cutoff_5p, $cutoff_3p, $nfac_5p, $nfac_3p, $len_indel), where ] --- OBSOLETE as of July 11 (Wed), 2012.
# INPUT: ($if_5p_end_indel, $if_3p_end_indel, $distri, $param, $cutoff_5p, $cutoff_3p, $log_nfac_5p, $log_nfac_3p, $len_indel), where
#	$if_5p/3p_end_indel = 1 if the indel event extends into the 5'-/3'-end of the alignment;
#	$distri specifies the type of the indel length distribution (= 'G' for geometric, = 'P' for power-law);
#	$param specifies an important parameter of the distribution;
#	$cutoff_5p/3p specifies the cut-off length into the 5'/3'-direction;
#	[ $nfac_5p/3p is the normalization factor for the distribution extended into the 5'-/3'-region; ] --- OBSOLETE as of July 11 (Wed), 2012.
#	$log_nfac_5p/3p is the logarithm of the normalization factor for the distribution extended into the 5'-/3'-region;
#	$len_indel is the length of the subject insertion/deletion event.
#
# [ OUTPUT: $freq, ] --- OBSOLETE as of July 11 (Wed), 2012.
# OUTPUT: $log_freq,
#	which is actually the logarithm of the composite frequency (for the indel events longer than $len_indel).
#	
sub log_compos_freq_indel_len ($$$$$$$$$) {

	my ($if_5p_end_indel, $if_3p_end_indel, $distri, $param, $cutoff_5p, $cutoff_3p, $log_nfac_5p, $log_nfac_3p, $len_indel) = @_;

	unless (($cutoff_5p>0) and ($cutoff_3p>0)) {
	    print STDOUT "[ERROR 'log_compos_freq_indel_len'] ILLEGAL INPUT: either \$cutoff_5p (=$cutoff_5p) or $cutoff_3p (=$cutoff_3p) is NON-POSITIVE: $! !\n";
	    return undef;
	}
	unless ($len_indel >0) {
	   print STDOUT "[ERROR 'log_compos_freq_indel_len'] ILLEGAL INPUT: The indel length (=$len_indel) is NOT POSITIVE: $! !\n";
	   return undef;
	}

    (print STDOUT "\n", ' ' x 3, "< Subroutine 'log_compos_freq_indel_len' >\n\n") if ($TEST_LEVEL2c>0);

	if (($distri eq 'G') and ($param == 0)) { # An exceptional case. #
	    unless ($len_indel == 1) {
		print STDOUT "[ERROR 'log_compos_freq_indel_len'] ILLEGAL INPUT: The geometric distribution with q = 0, and yet the indel length (=$len_indel) is LONGER THAN 1: $! !\n";
		return undef;
	    }
	    my $log_freq = log_freq_indel_len ($distri, $param, 1, $log_nfac_3p);
	    print STDOUT "[WARNING 'log_compos_freq_indel_len'] The geometric distribution with q = 0 (and length = 1) should ALWAYS HAVE \$log_freq = $log_freq ...\n";
	    return $log_freq;
	}


	my $log_freq;

	if ($if_5p_end_indel) {
	    if ($if_3p_end_indel) {
			# The indel event extends into BOTH ends. #

    (print STDOUT ' ' x 3, "(5'-end, 3'-end) = (Yes, Yes)\n\n") if ($TEST_LEVEL2c>1);

		my $log_numer = log_dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, $len_indel);
		##my $numer = dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, $len_indel);
		unless (defined $log_numer) {
			print STDOUT "[ERROR 'log_compos_freq_indel_len'] The subroutine 'log_dbl_sum_freq_indel' FAILED to calculate the numerator: $! !\n";
			return undef;
		}
		my $log_denom = log_dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, 1);
		##my $denom = dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, 1);
		unless (defined $log_denom) {
			print STDOUT "[ERROR 'log_compos_freq_indel_len'] The subroutine 'log_dbl_sum_freq_indel' FAILED to calculate the denominator: $! !\n";
			return undef;
		}
		$log_freq = $log_numer - $log_denom;
		##$freq = $numer/$denom;

	    } else {
			# The indel event extends into the 5'-end. #

    (print STDOUT ' ' x 3, "(5'-end, 3'-end) = (Yes, No)\n\n") if ($TEST_LEVEL2c>1);

		unless ($len_indel <= $cutoff_5p) {
		    print STDOUT "[ERROR 'log_compos_freq_indel_len'] The indel length (=$len_indel) EXCEEDS the cutoff length (\$cutoff_5p=$cutoff_5p): $! !\n";
		    return undef;
		}

		##$freq = 1;
		if (($len_indel < $MID_INDEL_LEN) and ($len_indel < $cutoff_5p /2)) { # REVISED on July 15 (Sun), 2012.
				# Fixed a potential bug on July 11 (Wed), 2012. #
		##if (($len_indel < $MID_INDEL_LEN) and ($len_indel < $cutoff_5p)) { 

    (print STDOUT ' ' x 5, "... Indirect calculation ...\n\n") if ($TEST_LEVEL2c>1);

		    my @terms_excess = ();
		    for (my $l=1; $l<$len_indel; $l++) { # 1st for-loop #
		    ##for (my $l=0; $l<$len_indel; $l++) { # 1st for-loop #
			my $log_excess = log_freq_indel_len ($distri, $param, $l, $log_nfac_5p);
			push @terms_excess, $log_excess;
			##my $exess = freq_indel_len ($distri, $param, $l, $nfac_5p);
			##$freq -= $exess;
		    }			# End of the 1st for-loop #

			# Bug fixed on July 16 (Mon), 2012. #
		    my $log_sum_excess = (@terms_excess>0)? log_sum_exp (@terms_excess) : undef;
		    $log_freq = (@terms_excess>0) ? log (1 - exp ($log_sum_excess)) : 0; # Absence of excess terms means that the composite frequency must be unity.
		    ##my $log_sum_excess = log_sum_exp (@terms_excess);
		    ##$log_freq = log (1 - exp ($log_sum_excess));

		} else {

    (print STDOUT ' ' x 5, "... Direct calculation ...\n\n") if ($TEST_LEVEL2c>1);

		    my @terms_log_freq = ();
		    for (my $l = $len_indel; $l<=$cutoff_5p; $l++) { # 1.5 th for-loop #
			my $term = log_freq_indel_len ($distri, $param, $l, $log_nfac_5p);
			push @terms_log_freq, $term;
		    }			# End of the 1.5 th for-loop #
		    $log_freq = log_sum_exp (@terms_log_freq);
		}

	    }
	} else {

	    if ($if_3p_end_indel) {
			# The indel event extends into the 3'-end. #

    (print STDOUT ' ' x 3, "(5'-end, 3'-end) = (No, Yes)\n\n") if ($TEST_LEVEL2c>1);

		unless ($len_indel <= $cutoff_3p) {
		    print STDOUT "[ERROR 'log_compos_freq_indel_len'] The indel length (=$len_indel) EXCEEDS the cutoff length (\$cutoff_3p=$cutoff_3p): $! !\n";
		    return undef;
		}

		##$freq = 1;
		if (($len_indel < $MID_INDEL_LEN) and ($len_indel < $cutoff_3p /2)) { # REVISED on July 15 (Sun), 2012.(
		##if (($len_indel < $MID_INDEL_LEN) and ($len_indel < $cutoff_3p)) {

    (print STDOUT ' ' x 5, "... Indirect calculation ...\n\n") if ($TEST_LEVEL2c>1);

				# Fixed a potential bug on July 11 (Wed), 2012. #
		    my @terms_excess = ();
		    for (my $l=1; $l<$len_indel; $l++) { # 2nd for-loop #
		    ##for (my $l=0; $l<$len_indel; $l++) { # 2nd for-loop #
			my $log_excess = log_freq_indel_len ($distri, $param, $l, $log_nfac_3p);
			push @terms_excess, $log_excess;
			##my $exess = freq_indel_len ($distri, $param, $l, $nfac_3p);
			##$freq -= $exess;
		    }			# End of the 2nd for-loop #

			# Bug fixed on July 16 (Mon), 2012. #
		    my $log_sum_excess = (@terms_excess>0)? log_sum_exp (@terms_excess) : undef;
		    $log_freq = (@terms_excess>0) ? log (1 - exp ($log_sum_excess)) : 0; # Absence of excess terms means that the composite frequency must be unity.
		    ##my $log_sum_excess = log_sum_exp (@terms_excess);
		    ##$log_freq = log (1 - exp ($log_sum_excess));

		} else {

    (print STDOUT ' ' x 5, "... Direct calculation ...\n\n") if ($TEST_LEVEL2c>1);

		    my @terms_log_freq = ();
		    for (my $l = $len_indel; $l<=$cutoff_3p; $l++) { # 2.5 th for-loop #
			my $term = log_freq_indel_len ($distri, $param, $l, $log_nfac_3p);
			push @terms_log_freq, $term;
		    }			# End of the 1.5 th for-loop #
		    $log_freq = log_sum_exp (@terms_log_freq);

		}
	    } else {
			# The indel event does NOT extend into EITHER end,

    (print STDOUT ' ' x 3, "(5'-end, 3'-end) = (No, No)\n\n") if ($TEST_LEVEL2c>1);

	    	$log_freq = log_freq_indel_len ($distri, $param, $len_indel, $log_nfac_3p);
	    }
	}

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'log_compos_freq_indel_len' Ended --\n\n") if ($TEST_LEVEL2c>0);

	return $log_freq;

}		# End of the subroutine 'log_compos_freq_indel_len' #


# dbl_tinteg_exp_indel:
#	This subroutine calculates a double time-integral
#	that provides an important factor for the likelihood of
#	a pair of overlapping insertion and deletion events:
#
#	$dbl_t_integ = ($mid_exp1 - $mid_exp2)/$sum_totrates /$len_del, where
#
#	$mid_exp1 = (1 - exp (-$sum_totrates*$br_len*($len_ins-$len_del)))/$sum_totrates/($len_sins-$len_del) if $len_ins != $len_del,
#		= $br_len if $len_ins == $len_del;
#	$mid_exp2 = (1 - exp (-$sum_totrates * $br_len * $len_ins))/$sum_totrates/$len_ins.
#
# INPUT: ($sum_totrates, $len_ins, $len_del, $br_len), where
#	$sum_totrates = {total rate of insertions} + {total rate of deletions};
#	$len_ins/del is the length of the insertion/deletion;
#	$br_len is the branch length, which is regarded as 'time';
#
# OUTPUT: $dbl_t_integ, which is defined above.
#
sub dbl_tinteg_exp_indel ($$$$) {

	my ($sum_totrates, $len_ins, $len_del, $br_len) = @_;

	my $mid_exp1 = ($len_ins == $len_del) ? $br_len :
		(1 - exp (-$sum_totrates * $br_len * ($len_ins - $len_del)))/$sum_totrates/($len_ins-$len_del) ;
	my $mid_exp2 = (1 - exp (-$sum_totrates * $br_len * $len_ins))/$sum_totrates/$len_ins;

	my $dbl_t_integ = ($mid_exp1 - $mid_exp2)/$sum_totrates/$len_del;

	return $dbl_t_integ;

}		# End of the subroutine 'dbl_tinteg_exp_indel' #


# search_paired_indels:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls' searches
#	a set of indel events along a branch for pairs of events,
#	each of which consists of an insertion and a subsequent deletion
#	and results in blocks with both ends showing the 'absence' state.
#
# INPUT: (\@indels_br, \@blwise_gapstats_ue, \@blwise_gapstats_le, $char_abs), where
#	@{$indels_br[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th block (in the specified cluster) at the upper-/lower-end node of the specified branch;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
# OUTPUT: (\@sgl_indels, \@paired_indels), where
#
#	$sgl_indels[$sg] gives the index (in @indels_br) of
#	the $sg th 'single' indel event (, which does NOT overlap with other indel events);
#
#	@{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
#	specifies the $pr th pair of an insertion and a deletion,
#	with $indx_ins/del being the index in @indels_br,
#	and @indices_ovlp_blk storing the indices of blocks where the two events overlap.
#
sub search_paired_indels (\@\@\@$) {

	my ($indels_br, $blwise_gstats_ue, $blwise_gstats_le, $char_abs) = @_;

	my $ct_indels = @{$indels_br};
	my $ct_blks = @{$blwise_gstats_ue};

   (print STDOUT ' ' x 3, "< Subroutine 'search_paired_indels' (#{blocks} = $ct_blks\; #{indels} = $ct_indels) >\n\n") if ($TEST_LEVEL2s>0);

    if ($TEST_LEVEL2s>2) {
	print STDOUT ' ' x 3, "[ Input Data ('absence' state = $char_abs) ]\n\n";

	print STDOUT ' ' x 3, "Block-wise states (#{blocks} = $ct_blks):\n\n";

	print STDOUT ' ' x 3, join ("\t", 'Block_ID', 'state_upper_end', 'state_lower_end'), "\n\n";
	for (my $b=0; $b<$ct_blks; $b++) {
	    print STDOUT ' ' x 3, join ("\t", $b, $blwise_gstats_ue->[$b], $blwise_gstats_le->[$b]), "\n";
	}

	print STDOUT ' ' x 3, "\nIndel events (#{events} = $ct_indels):\n\n";

	print STDOUT ' ' x 3, join ("\t", 'Event_ID', 'state_before', 'state_after', 'IDs,of,involved,blocks'), "\n\n";

	for (my $e=0; $e<$ct_indels; $e++) {
	    my ($stat_bf, $stat_af, $indices_blk) = @{$indels_br->[$e]};
	    my $cnct_indices = join (',', @{$indices_blk});
	    print STDOUT ' ' x 3, join ("\t", $e, $stat_bf, $stat_af, $cnct_indices), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT ' ' x 3, ".. Now, start searching for single indel events ..\n\n";
    }

	my %indx_blk2pair; # ($indx_blk => [$indx_ins, $indx_del], ...)
	my @sgl_indels = ();

	for (my $id=0; $id<$ct_indels; $id++) { # 1st main outer for-loop (over indel events) #

	    my ($stat_ue, $stat_le, $indices_blk) = @{$indels_br->[$id]};
	    unless ($stat_ue ne $stat_le) {
		print STDOUT "[ERROR 'search_paired_indels'] ILLEGAL INPUT: The $id th indel event has the IDENTICAL STATE (= '$stat_ue') at BOTH ENDS: $! !\n";
		return (undef, undef);
	    }
	    my $ct_ovlp=0;
	    foreach my $indx_blk (@{$indices_blk}) { # Inner foreach-loop (over blocks involved) #

		my $gstat_ue = $blwise_gstats_ue->[$indx_blk];
		my $gstat_le = $blwise_gstats_le->[$indx_blk];

			# A potential bug fixed on July 11 (Wed), 2012. #
		(($gstat_ue eq $stat_ue) and ($gstat_le eq $stat_le)) and next; # Skip blocks showing the normal end-states. #
		##(($gstat_ue eq $stat_ue) and ($gstat_ue eq $stat_le)) and next; # Skip blocks showing the normal end-states. #

		unless ($gstat_ue eq $gstat_le) {
		    print STDOUT "[ERROR 'search_paired_indels'] ILLEGAL INPUT: The $indx_blk th block, which is involved in the $id th indel event, has different gap states", ' ($gstat_ue, $gstat_le) = ', "($gstat_ue, $gstat_le) at the ends of the branch, yet somehow the states do NOT coincide with those of the event itself", ' ($stat_ue, $stat_le) = ', "($stat_ue, $stat_le) : $! !\n";
		    return (undef, undef);
		}
		unless ($gstat_ue eq $char_abs) {
		    print STDOUT "[ERROR 'search_paired_indels'] ILLEGAL INPUT: The $indx_blk th block, which is involved in the $id th indel event, has the 'presence' state at BOTH ENDS: $! !\n";
		    return (undef, undef);
		}

			# Hereafter, both ends should have the 'absence' state,
			# which means that the $id th event must have
			# (at least) an overlapping event.
		$ct_ovlp++;

		my $pair = $indx_blk2pair{$indx_blk};
		(defined $pair) or ($pair = $indx_blk2pair{$indx_blk} = []);

		if ($stat_ue eq $char_abs) { # An insertion. #
		    if (defined $pair->[0]) {
			print STDOUT "[ERROR 'search_paired_indels'] The $indx_blk th block ALREADY ACCOMODATES an INSERTION (= the ", $pair->[0], " th event): $! !\n";
			return (undef, undef);
		    }
		    $pair->[0] = $id;
		} else { # A deletion. #
		    if (defined $pair->[1]) {
			print STDOUT "[ERROR 'search_paired_indels'] The $indx_blk th block ALREADY ACCOMODATES a DELETION (= the ", $pair->[1], " th event): $! !\n";
			return (undef, undef);
		    }
		    $pair->[1] = $id;
		}

	    }				# End of the inner foreach-loop (over blocks involved) #

	    ($ct_ovlp>0) or (push @sgl_indels, $id); # No block show signs of overlapping events, which means that the event should be 'single'. #

	}			# End of the 1st main outer for-loop (over indel events) #

	my $ct_sgls = @sgl_indels;
   (print STDOUT "\n", ' ' x 3, ".. $ct_sgls single events were found ..\n\n") if ($TEST_LEVEL2s>1);
   (print STDOUT ' ' x 5, "(Indices:" , join (', ', @sgl_indels), ")\n\n") if ($TEST_LEVEL2s>2);

		# Make a concise list of overlapping indel pairs. #

	my %indx_indel2ovlp; # = ($indx_indel => \@indices_overlapping_events, ...)
	my %indx_ins_del2indx_pair; # = ("$indx_ins:$indx_del" => $indx_pair, ...)

	my @paired_indels = (); # @{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
	## my %indx_blk2pair; # ($indx_blk => [$indx_ins, $indx_del], ...)
	foreach my $indx_blk (sort {$a <=> $b} keys %indx_blk2pair) { # 2nd main outer foreach-loop (over blocks where indel events overlap) #

	    my ($indx_ins, $indx_del) = @{$indx_blk2pair{$indx_blk}};
	    unless ((defined $indx_ins) and (defined $indx_del)) {
		(defined $indx_ins) or ($indx_ins = 'undef');
		(defined $indx_del) or ($indx_del = 'undef');
		print STDOUT "[ERROR 'search_paired_indels'] The $indx_blk th block involves at least an UNDEFINED INDEL EVENT: ", '($indx_ins, $indx_del) = ', "($indx_ins, $indx_del): $! !\n";
		return (undef, undef);
	    }

	    my $indx_pair = $indx_ins_del2indx_pair{"$indx_ins:$indx_del"};
	    if (defined $indx_pair) {
			# The pair already occurred before. #
		push @{$paired_indels[$indx_pair]->[2]}, $indx_blk;
		next;
	    }

		# The pair occurs for the first time. #

	    $indx_pair = @paired_indels;
	    $indx_ins_del2indx_pair{"$indx_ins:$indx_del"} = $indx_pair;
	    $paired_indels[$indx_pair] = [$indx_ins, $indx_del, [$indx_blk]];

	    if (defined $indx_indel2ovlp{$indx_ins}) {
		push @{$indx_indel2ovlp{$indx_ins}}, $indx_del;
	    } else {
		$indx_indel2ovlp{$indx_ins} = [$indx_del];
	    }

	    if (defined $indx_indel2ovlp{$indx_del}) {
		push @{$indx_indel2ovlp{$indx_del}}, $indx_ins;
	    } else {
		$indx_indel2ovlp{$indx_del} = [$indx_ins];
	    }

	}				# End of the 2nd main outer foreach-loop (over blocks where indel events overlap) #

	my $ct_pairs = @paired_indels; 
    (print STDOUT "\n", ' ' x 3, ".. $ct_pairs pairs of overlapping indels were found ..\n\n") if ($TEST_LEVEL2s>1);
    if ($TEST_LEVEL2s>3) {
	print STDOUT ' ' x 3, join ("\t", 'Indx_ins', 'Indx_del', 'indices,of,overlapping,blocks'), "\n\n";
	foreach my $pair (@paired_indels) {
	    my ($indx_ins, $indx_del, $ovlp_blocks) = @{$pair};
	    my $cnct_ovlp_blks = join (',', @{$ovlp_blocks});
	    print STDOUT ' ' x 3, join ("\t", $indx_ins, $indx_del, $cnct_ovlp_blks), "\n";
	}
	print STDOUT "\n";
    }

		# Examine whether there are any 'multiply-overlapping' events. #

    (print STDOUT "\n", ' ' x 3, ".. Now I will check if there are multiply-overlapping events..\n\n") if ($TEST_LEVEL2s>1);

	my @list_events_w_mlt_ovlps; # element = [$indx_indel, \@indices_overlapping_events] for those events with > 1 overlapping event. #

	foreach my $indx_indel (sort {$a <=> $b} keys %indx_indel2ovlp) { # Trailing foreach-loop (over indel events overlapping others) #
	    my $ovlp = $indx_indel2ovlp{$indx_indel};
	    (@{$ovlp}>1) and (push @list_events_w_mlt_ovlps, [$indx_indel, $ovlp]);
	}			# End of the trailing foreach-loop (over indel events overlapping others) #

	if (@list_events_w_mlt_ovlps>0) { # AT this point, the algorithm canNOT handle such cases. #
	    print STDOUT "[ERROR 'search_paired_indels'] Found ", scalar (@list_events_w_mlt_ovlps), " indel events, each of which overlaps with more than one other event: $! !\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Indx_indel', 'indices,overlapping,events'), "\n\n";

	    foreach my $event_w_mlt_ovlps (@list_events_w_mlt_ovlps) {
		my ($indx_event, $indices_ovlp) = @{$event_w_mlt_ovlps};
		my $cnct_indices_ovlp = join (',', @{$indices_ovlp});
		print STDOUT ' ' x 3, join ("\t", $indx_event, $cnct_indices_ovlp), "\n";
	    }
	    print STDOUT "\n";
	    return (undef, undef);
	}

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'search_paired_indels' Ended --\n\n") if ($TEST_LEVEL2s>0);

	return (\@sgl_indels, \@paired_indels) ;

}	# End of the subroutine 'search_paired_indels' #


=This subroutine became OBSOLETE as of July 22 (Sun), 2012.
# search_leftmost_block:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls'
#	searches a cluster of gap-blocks for
#	the 'leftmost' block, which is defined as the leftmost block of the following:
#	(i) blocks, in each of which both ends are 'present';
#	(ii) blocks involved in an overlapping pair of an insertion followed by a deletion;
#	(iii) blocks involved in an indel event.
#	As an exception, if all blocks of type (i) and (ii) are on the right of
#	an non-overlapping pair of an insertion and a deletion
#	that occur leftmost and second leftmost,
#	the leftmost block of the second leftmost event is returned.
#
# INPUT: (\@blwise_gapstats_ue, \@blwise_gapstats_le, \@indels_br, $char_abs), where
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th block (in the specified cluster) at the upper-/lower-end node of the specified branch;
#
#	@{$indels_br[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
# OUTPUT: $indx_lm_blk,
#	which is the index specifying the leftmost block.
#
sub search_leftmost_block (\@\@\@$) {

	my ($blwise_gstats_ue, $blwise_gstats_le, $indels_br, $char_abs) = @_;

	my $ct_gblks = @{$blwise_gstats_ue};
	my %gblk2if_involved; # ($indx_gblock => 1, ...), for the gap-blocks involved in indels.
	foreach my $event (@{$indels_br}) { # Preliminary outer foreach-loop #
	    foreach my $indx_blk (@{$event->[2]}) { $gblk2if_involved{$indx_blk} = 1; }
	}			# End of the preliminary outer foreach-loop #

	my ($cand_indx_lm_blk, $cand_lm_gstat_ue);

	for (my $bl=0; $bl<$ct_gblks; $bl++) { # Main for-loop (over gap blocks) #

	    my $gstat_ue = $blwise_gstats_ue->[$bl];
	    my $gstat_le = $blwise_gstats_le->[$bl];

	    if (($gstat_ue ne $char_abs) and ($gstat_le ne $char_abs)) {
			# Both ends are 'present'.
			# => Return the left-most block of the already-found indel.
			#  If there is no such indel event, return the current block.

		return (defined $cand_indx_lm_blk) ? $cand_indx_lm_blk : $bl;

	    } elsif (($gstat_ue eq $char_abs) and ($gstat_le eq $char_abs)) {
			# Both ends are 'absent'. #

		(defined $gblk2if_involved{$bl}) or next; # Pass through the block unless it is involved in any indel events.

			# => Return the left-most block of the already-found indel.
			#  If there is no such indel event, return the current block.

		return (defined $cand_indx_lm_blk) ? $cand_indx_lm_blk : $bl;
	    }
		# One end is 'present' and the other end is 'absent'. #

	    if (!(defined $cand_indx_lm_blk)) {
			# The leftmost block of the first indel event encountered. #
		($cand_indx_lm_blk, $cand_lm_gstat_ue) = ($bl, $gstat_ue);
	    } elsif ($cand_lm_gstat_ue ne $gstat_ue) {
			# The leftmost block of the second 'genuinely leftmost' indel event. #
			# => return this block.
		return $bl;
	    } else {
		# The 2nd or later block of the first indel event.
		# => Do nothing.
	    }

	}			# End of the main for-loop (over gap blocks) #

		# All the blocks are scanned. #

	unless (defined $cand_indx_lm_blk) {
	    print STDOUT "[ERROR 'search_leftmost_block'] Encountered NEITHER indel events NOR blocks with both ends 'present': $! !\n";
	    return undef;
	}

	return $cand_indx_lm_blk;

}		# End of the subroutine 'search_leftmost_block' #
=cut


# search_leftmost_events:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls'
#	searches for the 'leftmost' indel events, which are defined as the leftmost 
#	events either among all insertions or among all deletions
#	that are on the left of the leftmost block with both upper- and lower-ends having 'presence' states.
#
#	The smaller of the overlapping events will not be counted as such events
#	unless its leftmost block is identical to that of the larger event.
#
#	If the leftmost insertion and the leftmost deletion incompletely overlap with each other,
#	only the event to the left is counted in.
#
#	If an insertion and a deletion completely overlap with each other,
#	they will be counted in if there is no (non-overlapping or incompletely-overlapping)
#	event of the same type on the left.
#
# INPUT: (\@blwise_gapstats_ue, \@blwise_gapstats_le, \@indels_br, $char_abs, \@sgl_indels, \@paired_indels), where
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th block (in the specified cluster) at the upper-/lower-end node of the specified branch;
#
#	@{$indels_br[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	$sgl_indels[$sg] gives the index (in @indels_br) of
#	the $sg th 'single' indel event (, which does NOT overlap with other indel events);
#
#	@{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
#	specifies the $pr th pair of an insertion and a deletion,
#	with $indx_ins/del being the index in @indels_br,
#	and @indices_ovlp_blk storing the indices of blocks where the two events overlap.
#
# OUTPUT: \%indx_indel2if_lm, where
#
#	%indx_indel2if_lm = ($indx_indel => 1, ...) for the 'left-most' indel events.
#
sub search_leftmost_events (\@\@\@$\@\@) {

	my ($blwise_gstats_ue, $blwise_gstats_le, $indels_br, $char_abs, $sgl_indels, $paired_indels) = @_;

	my $ct_gblks = @{$blwise_gstats_ue};
	my $ct_indels =  @{$indels_br};
	my $ct_pairs = @{$paired_indels};

    (print STDOUT ' ' x 5, "< Subroutine 'search_leftmost_events' ('absence' = $char_abs, #{blocks} = $ct_gblks, #{indels} = $ct_indels, #{pairs} = $ct_pairs) >\n\n") if ($TEST_LEVEL3>0);

    if ($TEST_LEVEL3>2) {
	print STDOUT ' ' x 5, "[ Input Data ]\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Block_ID', 'gstat_ue', 'gstat_le'), "\n\n";
	for (my $b=0; $b<$ct_gblks; $b++) { print STDOUT ' ' x 5, join ("\t", $b, $blwise_gstats_ue->[$b], $blwise_gstats_le->[$b]), "\n"; }
	print STDOUT "\n", ' ' x 5, join ("\t", 'Indel_ID', 'stat_bf', 'stat_af', 'indices,involved,blocks'), "\n\n";
	for (my $e=0; $e<$ct_indels; $e++) {
	    my ($stat_bf, $stat_af, $indices_blks) = @{$indels_br->[$e]};
	    my $cnct_indices_blks = join (',', @{$indices_blks});
	    print STDOUT ' ' x 5, join ("\t", $e, $stat_bf, $stat_af, $cnct_indices_blks), "\n";
	}
	print STDOUT "\n\n";
	print STDOUT ' ' x 5, '@{$sgl_indels} = (', join (', ', @{$sgl_indels}), ") .\n\n\n";
	print STDOUT ' ' x5 , 'Contents of @{$paired_indels}:', "\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Indx_pair', 'indx_ins', 'indx_del', 'indices,ovlp,blocks'), "\n\n";
	for (my $p = 0; $p<$ct_pairs; $p++) {
	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$p]};
	    my $cnct_indices_ovlp = join (',', @{$indices_ovlp});
	    print STDOUT ' ' x 5, join ("\t", $p, $indx_ins, $indx_del, $cnct_indices_ovlp), "\n";
	}
	print STDOUT "\n\n";
    }

    (print STDOUT ' ' x 5, "[ Process I: Determine the right boundary block ]\n\n") if ($TEST_LEVEL3>1);
	my $rbound = $ct_gblks; # The leftmost block with both upper- and lower-ends having 'presence' states.
	for (my $b = 0; $b<$ct_gblks; $b++) { # 1st for-loop (over gap-blocks) #

	    (($blwise_gstats_ue->[$b] eq $char_abs) or ($blwise_gstats_le->[$b] eq $char_abs)) and next; 
	    $rbound = $b;
	    last;
	}			# End of the 1st for-loop (over gap-blocks) #

    (print STDOUT "\n", ' ' x 5, "Right boundary: \$rbound = $rbound .\n\n") if ($TEST_LEVEL3>2);

    (print STDOUT "\n", ' ' x 5, "[ Process II: Search for the leftmost single indel events ]\n\n") if ($TEST_LEVEL3>1);

	my $indx_lm_ins_blk = my $indx_lm_del_blk = $rbound;
	my ($indx_lm_sgl_ins, $indx_lm_sgl_del);

	foreach my $indx_id (@{$sgl_indels}) { # 2nd foreach-loop (over single indel events) #

	    my ($stat_bf, $stat_af, $indices_blk) = @{$indels_br->[$indx_id]};
	    my $indx_lcl_lm = $indices_blk->[0];

	    if ($stat_bf eq $char_abs) {
		# Insertion #
		if ($indx_lcl_lm < $indx_lm_ins_blk) {
			# The event is to the left of the current leftmost insertion. #
		    $indx_lm_ins_blk = $indx_lcl_lm;
		    $indx_lm_sgl_ins = $indx_id;
		}
	    } else {
		# Deletion #
		if ($indx_lcl_lm < $indx_lm_del_blk) {
			# The event is to the left of the current leftmost deletion. #
		    $indx_lm_del_blk = $indx_lcl_lm;
		    $indx_lm_sgl_del = $indx_id;
		}
	    }
	}			# End of the 2nd foreach-loop (over single indel events) #

    if ($TEST_LEVEL3>2) {
	print STDOUT "\n", ' ' x 5, '($indx_lm_ins_blk, $indx_lm_del_blk) = ', "($indx_lm_ins_blk, $indx_lm_del_blk) ;\n\n";
	my @out_indx_lm = ();
	foreach my $indx ($indx_lm_sgl_ins, $indx_lm_sgl_del) { push @out_indx_lm, ((defined $indx) ? $indx : 'undef'); }
	print STDOUT ' ' x 5, '($indx_lm_sgl_ins, $indx_lm_sgl_del) = (', join (', ', @out_indx_lm), ") .\n\n"; 
    }

    (print STDOUT "\n", ' ' x 5, "[ Process III: Search for the leftmost pair of overlapping indel events ]\n\n") if ($TEST_LEVEL3>1);

	my @cmpl_ovlp_pairs = (); # Completely overlapping pairs to the left of either the current leftmost insertion or the current leftmost deletion.
	my ($indx_lm_pr_ins, $indx_lm_pr_del); # Index of the pair that contains the leftmost insertion/deletion.

	for (my $pr=0; $pr<$ct_pairs; $pr++) { # 3rd for-loop (over paired events) #

	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$pr]};
	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];

	    my $indx_lcl_lm_ins = $indices_blk_ins->[0];
	    my $indx_lcl_lm_del = $indices_blk_del->[0];

	    my ($common, $ext_ins, $ext_del) = diff_nmarray (@{$indices_blk_ins}, @{$indices_blk_del});
	    unless ((@{$ext_ins}>0) or (@{$ext_del}>0)) {
		# Completely overlapping. #
		(($indx_lcl_lm_ins < $indx_lm_ins_blk) or ($indx_lcl_lm_del < $indx_lm_del_blk)) and (push @cmpl_ovlp_pairs, $pr); # To the left of either the current leftmost insertion or the current leftmost deletion.
		next;
	    }
	    if (@{$ext_ins}>0) {
			# Insertion sticks out. #
		if ($indx_lcl_lm_ins < $indx_lm_ins_blk) {
		    # The insertion part is to the left of the current leftmost insertion. #
		    $indx_lm_ins_blk = $indx_lcl_lm_ins;
		    $indx_lm_pr_ins = $pr;
		}
	    }
	    if (@{$ext_del}>0) {
			# Deletion sticks out. #
		if ($indx_lcl_lm_del < $indx_lm_del_blk) {
		    # The deletion part is to the left of the current leftmost deletion. #
		    $indx_lm_del_blk = $indx_lcl_lm_del;
		    $indx_lm_pr_del = $pr;
		}
	    }
#	@{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)

	}			# End of the 3rd for-loop (over paired events) #

    if ($TEST_LEVEL3>2) {
	print STDOUT "\n", ' ' x 5, '($indx_lm_ins_blk, $indx_lm_del_blk) = ', "($indx_lm_ins_blk, $indx_lm_del_blk) ;\n\n";
	my @out_indx_lm = ();
	foreach my $indx ($indx_lm_pr_ins, $indx_lm_pr_del) { push @out_indx_lm, ((defined $indx) ? $indx : 'undef'); }
	print STDOUT ' ' x 5, '($indx_lm_pr_ins, $indx_lm_pr_del) = (', join (', ', @out_indx_lm), ") ;\n\n"; 
	print STDOUT ' ' x 5, '@cmpl_ovlp_pairs = (', join (', ', @cmpl_ovlp_pairs), ") .\n\n";
    }

		# Finishing process. #

    (print STDOUT "\n", ' ' x 5, "[ Final Process: Searching for the genuine 'leftmost' indel events ]\n\n") if ($TEST_LEVEL3>1);

	my %indx_indel2if_lm;

	if ((defined $indx_lm_pr_ins) and (defined $indx_lm_pr_del) and ($indx_lm_pr_ins == $indx_lm_pr_del)) {
	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_lm_pr_ins]};

	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];
	    my $indx_lcl_lm_ins = $indices_blk_ins->[0];
	    my $indx_lcl_lm_del = $indices_blk_del->[0];

	    if ($indx_lcl_lm_ins < $indx_lcl_lm_del) {
		$indx_indel2if_lm{$indx_ins} = 1;
	    } elsif ($indx_lcl_lm_del < $indx_lcl_lm_ins) {
		$indx_indel2if_lm{$indx_del} = 1;
	    } else {
		$indx_indel2if_lm{$indx_ins} = 1;
		$indx_indel2if_lm{$indx_del} = 1;
	    }
	  

	} else {

	    if (defined $indx_lm_pr_ins) {
			# The leftmost insertion is contained in a pair. #
		my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_lm_pr_ins]};
		$indx_indel2if_lm{$indx_ins} = 1;

		my $indices_blk_ins = $indels_br->[$indx_ins][2];
		my $indices_blk_del = $indels_br->[$indx_del][2];
		my $indx_lcl_lm_ins = $indices_blk_ins->[0];
		my $indx_lcl_lm_del = $indices_blk_del->[0];

		($indx_lcl_lm_ins == $indx_lcl_lm_del) and ($indx_indel2if_lm{$indx_del} = 1);

	    } elsif (defined $indx_lm_sgl_ins) {
			# The leftmost insertion is a single event. #
		$indx_indel2if_lm{$indx_lm_sgl_ins} = 1;
	    }

	    if (defined $indx_lm_pr_del) {
			# The leftmost deletion is contained in a pair. #
		my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_lm_pr_del]};
		$indx_indel2if_lm{$indx_del} = 1;

		my $indices_blk_ins = $indels_br->[$indx_ins][2];
		my $indices_blk_del = $indels_br->[$indx_del][2];
		my $indx_lcl_lm_ins = $indices_blk_ins->[0];
		my $indx_lcl_lm_del = $indices_blk_del->[0];

		($indx_lcl_lm_ins == $indx_lcl_lm_del) and ($indx_indel2if_lm{$indx_ins} = 1);

	    } elsif (defined $indx_lm_sgl_del) {
			# The leftmost deletion is a single event. #
		$indx_indel2if_lm{$indx_lm_sgl_del} = 1;
	    }

	}

	foreach my $pr (@cmpl_ovlp_pairs) { # Trailing foreach-loop (over completely overlapping pairs) #

	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$pr]};
	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];

	    my $indx_lcl_lm_ins = $indices_blk_ins->[0];
	    my $indx_lcl_lm_del = $indices_blk_del->[0];

	    ($indx_lcl_lm_ins < $indx_lm_ins_blk) and ($indx_indel2if_lm{$indx_ins} = 1);
	    ($indx_lcl_lm_del < $indx_lm_del_blk) and ($indx_indel2if_lm{$indx_del} = 1);

	}		# End of the trailing foreach-loop (over completely overlapping pairs) #

    (print STDOUT "\n", ' ' x 5, "- Subroutine 'search_leftmost_events' ended -\n\n") if ($TEST_LEVEL3>0);

	return \%indx_indel2if_lm;

	# %indx_indel2if_lm = ($indx_indel => 1, ...) for the 'left-most' indel events.

}		# End of the subroutine 'search_leftmost_events' #


# search_rightmost_events:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls'
#	was created by flipping the search direction of 'search_leftmost_events'
#	so that it can search for the 'rightmost' indel events.
#
# search_leftmost_events:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls'
#	searches for the 'leftmost' indel events, which are defined as the leftmost 
#	events either among all insertions or among all deletions
#	that are on the left of the leftmost block with both upper- and lower-ends having 'presence' states.
#
#	The smaller of the overlapping events will not be counted as such events
#	unless its leftmost block is identical to that of the larger event.
#
#	If the leftmost insertion and the leftmost deletion incompletely overlap with each other,
#	only the event to the left is counted in.
#
#	If an insertion and a deletion completely overlap with each other,
#	they will be counted in if there is no (non-overlapping or incompletely-overlapping)
#	event of the same type on the left.
#
# INPUT: (\@blwise_gapstats_ue, \@blwise_gapstats_le, \@indels_br, $char_abs, \@sgl_indels, \@paired_indels), where
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th block (in the specified cluster) at the upper-/lower-end node of the specified branch;
#
#	@{$indels_br[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	$sgl_indels[$sg] gives the index (in @indels_br) of
#	the $sg th 'single' indel event (, which does NOT overlap with other indel events);
#
#	@{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
#	specifies the $pr th pair of an insertion and a deletion,
#	with $indx_ins/del being the index in @indels_br,
#	and @indices_ovlp_blk storing the indices of blocks where the two events overlap.
#
# OUTPUT: \%indx_indel2if_rm, where
#
#	%indx_indel2if_rm = ($indx_indel => 1, ...) for the 'right-most' indel events.
#
sub search_rightmost_events (\@\@\@$\@\@) {

	my ($blwise_gstats_ue, $blwise_gstats_le, $indels_br, $char_abs, $sgl_indels, $paired_indels) = @_;

	my $ct_gblks = @{$blwise_gstats_ue};
	my $ct_indels =  @{$indels_br};
	my $ct_pairs = @{$paired_indels};

    (print STDOUT ' ' x 5, "< Subroutine 'search_rightmost_events' ('absence' = $char_abs, #{blocks} = $ct_gblks, #{indels} = $ct_indels, #{pairs} = $ct_pairs) >\n\n") if ($TEST_LEVEL3>0);

    if ($TEST_LEVEL3>2) {
	print STDOUT ' ' x 5, "[ Input Data ]\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Block_ID', 'gstat_ue', 'gstat_le'), "\n\n";
	for (my $b=0; $b<$ct_gblks; $b++) { print STDOUT ' ' x 5, join ("\t", $b, $blwise_gstats_ue->[$b], $blwise_gstats_le->[$b]), "\n"; }
	print STDOUT "\n", ' ' x 5, join ("\t", 'Indel_ID', 'stat_bf', 'stat_af', 'indices,involved,blocks'), "\n\n";
	for (my $e=0; $e<$ct_indels; $e++) {
	    my ($stat_bf, $stat_af, $indices_blks) = @{$indels_br->[$e]};
	    my $cnct_indices_blks = join (',', @{$indices_blks});
	    print STDOUT ' ' x 5, join ("\t", $e, $stat_bf, $stat_af, $cnct_indices_blks), "\n";
	}
	print STDOUT "\n\n";
	print STDOUT ' ' x 5, '@{$sgl_indels} = (', join (', ', @{$sgl_indels}), ") .\n\n\n";
	print STDOUT ' ' x5 , 'Contents of @{$paired_indels}:', "\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Indx_pair', 'indx_ins', 'indx_del', 'indices,ovlp,blocks'), "\n\n";
	for (my $p = 0; $p<$ct_pairs; $p++) {
	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$p]};
	    my $cnct_indices_ovlp = join (',', @{$indices_ovlp});
	    print STDOUT ' ' x 5, join ("\t", $p, $indx_ins, $indx_del, $cnct_indices_ovlp), "\n";
	}
	print STDOUT "\n\n";
    }

    (print STDOUT ' ' x 5, "[ Process I: Determine the left boundary block ]\n\n") if ($TEST_LEVEL3>1);
	my $lbound = -1; # The rightmost block with both upper- and lower-ends having 'presence' states.
	for (my $b = $ct_gblks-1; $b>=0; $b--) { # 1st for-loop (over gap-blocks) #

	    (($blwise_gstats_ue->[$b] eq $char_abs) or ($blwise_gstats_le->[$b] eq $char_abs)) and next; 
	    $lbound = $b;
	    last;
	}			# End of the 1st for-loop (over gap-blocks) #

    (print STDOUT "\n", ' ' x 5, "Left boundary: \$lbound = $lbound .\n\n") if ($TEST_LEVEL3>2);

    (print STDOUT "\n", ' ' x 5, "[ Process II: Search for the rightmost single indel events ]\n\n") if ($TEST_LEVEL3>1);

	my $indx_rm_ins_blk = my $indx_rm_del_blk = $lbound;
	my ($indx_rm_sgl_ins, $indx_rm_sgl_del);

	foreach my $indx_id (@{$sgl_indels}) { # 2nd foreach-loop (over single indel events) #

	    my ($stat_bf, $stat_af, $indices_blk) = @{$indels_br->[$indx_id]};
	    my $indx_lcl_rm = $indices_blk->[$#{$indices_blk}];

	    if ($stat_bf eq $char_abs) {
		# Insertion #
		if ($indx_rm_ins_blk < $indx_lcl_rm) {
			# The event is to the right of the current rightmost insertion. #
		    $indx_rm_ins_blk = $indx_lcl_rm;
		    $indx_rm_sgl_ins = $indx_id;
		}
	    } else {
		# Deletion #
		if ($indx_rm_del_blk < $indx_lcl_rm) {
			# The event is to the right of the current rightmost deletion. #
		    $indx_rm_del_blk = $indx_lcl_rm;
		    $indx_rm_sgl_del = $indx_id;
		}
	    }
	}			# End of the 2nd foreach-loop (over single indel events) #

    if ($TEST_LEVEL3>2) {
	print STDOUT "\n", ' ' x 5, '($indx_rm_ins_blk, $indx_rm_del_blk) = ', "($indx_rm_ins_blk, $indx_rm_del_blk) ;\n\n";
	my @out_indx_rm = ();
	foreach my $indx ($indx_rm_sgl_ins, $indx_rm_sgl_del) { push @out_indx_rm, ((defined $indx) ? $indx : 'undef'); }
	print STDOUT ' ' x 5, '($indx_rm_sgl_ins, $indx_rm_sgl_del) = (', join (', ', @out_indx_rm), ") .\n\n"; 
    }

    (print STDOUT "\n", ' ' x 5, "[ Process III: Search for the rightmost pair of overlapping indel events ]\n\n") if ($TEST_LEVEL3>1);

	my @cmpl_ovlp_pairs = (); # Completely overlapping pairs to the right of either the current rightmost insertion or the current rightmost deletion.
	my ($indx_rm_pr_ins, $indx_rm_pr_del); # Index of the pair that contains the rightmost insertion/deletion.

	for (my $pr=0; $pr<$ct_pairs; $pr++) { # 3rd for-loop (over paired events) #

	    # @{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)

	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$pr]};
	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];

	    my $indx_lcl_rm_ins = $indices_blk_ins->[$#{$indices_blk_ins}];
	    my $indx_lcl_rm_del = $indices_blk_del->[$#{$indices_blk_del}];

	    my ($common, $ext_ins, $ext_del) = diff_nmarray (@{$indices_blk_ins}, @{$indices_blk_del});
	    unless ((@{$ext_ins}>0) or (@{$ext_del}>0)) {
		# Completely overlapping. #
		(($indx_rm_ins_blk < $indx_lcl_rm_ins) or ($indx_rm_del_blk < $indx_lcl_rm_del)) and (push @cmpl_ovlp_pairs, $pr); # To the right of either the current rightmost insertion or the current rightmost deletion.
		next;
	    }
	    if (@{$ext_ins}>0) {
			# Insertion sticks out. #
		if ($indx_rm_ins_blk < $indx_lcl_rm_ins) {
		    # The insertion part is to the right of the current rightmost insertion. #
		    $indx_rm_ins_blk = $indx_lcl_rm_ins;
		    $indx_rm_pr_ins = $pr;
		}
	    }
	    if (@{$ext_del}>0) {
			# Deletion sticks out. #
		if ($indx_rm_del_blk < $indx_lcl_rm_del) {
		    # The deletion part is to the right of the current rightmost deletion. #
		    $indx_rm_del_blk = $indx_lcl_rm_del;
		    $indx_rm_pr_del = $pr;
		}
	    }

	}			# End of the 3rd for-loop (over paired events) #

    if ($TEST_LEVEL3>2) {
	print STDOUT "\n", ' ' x 5, '($indx_rm_ins_blk, $indx_rm_del_blk) = ', "($indx_rm_ins_blk, $indx_rm_del_blk) ;\n\n";
	my @out_indx_rm = ();
	foreach my $indx ($indx_rm_pr_ins, $indx_rm_pr_del) { push @out_indx_rm, ((defined $indx) ? $indx : 'undef'); }
	print STDOUT ' ' x 5, '($indx_rm_pr_ins, $indx_rm_pr_del) = (', join (', ', @out_indx_rm), ") ;\n\n"; 
	print STDOUT ' ' x 5, '@cmpl_ovlp_pairs = (', join (', ', @cmpl_ovlp_pairs), ") .\n\n";
    }

		# Finishing process. #

    (print STDOUT "\n", ' ' x 5, "[ Final Process: Searching for the genuine 'rightmost' indel events ]\n\n") if ($TEST_LEVEL3>1);

	my %indx_indel2if_rm;

	if ((defined $indx_rm_pr_ins) and (defined $indx_rm_pr_del) and ($indx_rm_pr_ins == $indx_rm_pr_del)) {
	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_rm_pr_ins]};

	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];
	    my $indx_lcl_rm_ins = $indices_blk_ins->[$#{$indices_blk_ins}];
	    my $indx_lcl_rm_del = $indices_blk_del->[$#{$indices_blk_del}];

	    if ($indx_lcl_rm_ins < $indx_lcl_rm_del) {
		$indx_indel2if_rm{$indx_del} = 1;
	    } elsif ($indx_lcl_rm_del < $indx_lcl_rm_ins) {
		$indx_indel2if_rm{$indx_ins} = 1;
	    } else {
		$indx_indel2if_rm{$indx_ins} = 1;
		$indx_indel2if_rm{$indx_del} = 1;
	    }
	  

	} else {

	    if (defined $indx_rm_pr_ins) {
			# The rightmost insertion is contained in a pair. #
		my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_rm_pr_ins]};
		$indx_indel2if_rm{$indx_ins} = 1;

		my $indices_blk_ins = $indels_br->[$indx_ins][2];
		my $indices_blk_del = $indels_br->[$indx_del][2];
		my $indx_lcl_rm_ins = $indices_blk_ins->[$#{$indices_blk_ins}];
		my $indx_lcl_rm_del = $indices_blk_del->[$#{$indices_blk_del}];

		($indx_lcl_rm_ins == $indx_lcl_rm_del) and ($indx_indel2if_rm{$indx_del} = 1);

	    } elsif (defined $indx_rm_sgl_ins) {
			# The leftmost insertion is a single event. #
		$indx_indel2if_rm{$indx_rm_sgl_ins} = 1;
	    }

	    if (defined $indx_rm_pr_del) {
			# The rightmost deletion is contained in a pair. #
		my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$indx_rm_pr_del]};
		$indx_indel2if_rm{$indx_del} = 1;

		my $indices_blk_ins = $indels_br->[$indx_ins][2];
		my $indices_blk_del = $indels_br->[$indx_del][2];
		my $indx_lcl_rm_ins = $indices_blk_ins->[$#{$indices_blk_ins}];
		my $indx_lcl_rm_del = $indices_blk_del->[$#{$indices_blk_del}];

		($indx_lcl_rm_ins == $indx_lcl_rm_del) and ($indx_indel2if_rm{$indx_ins} = 1);

	    } elsif (defined $indx_rm_sgl_del) {
			# The leftmost deletion is a single event. #
		$indx_indel2if_rm{$indx_rm_sgl_del} = 1;
	    }

	}

	foreach my $pr (@cmpl_ovlp_pairs) { # Trailing foreach-loop (over completely overlapping pairs) #

	    my ($indx_ins, $indx_del, $indices_ovlp) = @{$paired_indels->[$pr]};
	    my $indices_blk_ins = $indels_br->[$indx_ins][2];
	    my $indices_blk_del = $indels_br->[$indx_del][2];

	    my $indx_lcl_rm_ins = $indices_blk_ins->[$#{$indices_blk_ins}];
	    my $indx_lcl_rm_del = $indices_blk_del->[$#{$indices_blk_del}];

	    ($indx_rm_ins_blk < $indx_lcl_rm_ins) and ($indx_indel2if_rm{$indx_ins} = 1);
	    ($indx_rm_del_blk < $indx_lcl_rm_del) and ($indx_indel2if_rm{$indx_del} = 1);

	}		# End of the trailing foreach-loop (over completely overlapping pairs) #

    (print STDOUT "\n", ' ' x 5, "- Subroutine 'search_rightmost_events' ended -\n\n") if ($TEST_LEVEL3>0);

	return \%indx_indel2if_rm;

	# %indx_indel2if_rm = ($indx_indel => 1, ...) for the 'right-most' indel events.

}		# End of the subroutine 'search_rightmost_events' #


# cal_lnlk1_sgn_indel_cls:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls' calculates
#	the 1st approximation of the log-likelihood of a single indel event
#	along a specified branch in a specified gap-block cluster.
#
#	[Major revision on July 22 (Sun), 2012 ]
#	The last two arguments are now the indicators of left/right-most indel events,
#	instead of the indices of the left/right-most blocks.
#
# [ INPUT: (\@id_event, \@cls_gblocks, $br_len, $char_abs, \@nfacs, \@params_ins, \@params_del, \@params_boundaries, $indx_lm_blk, $indx_rm_blk), where ] --- OBSOLETE as of July 11 (Wed), 2012. 
# [ INPUT: (\@id_event, \@cls_gblocks, $br_len, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries, $indx_lm_blk, $indx_rm_blk), where ] --- OBSOLETE as of July 22 (Sun), 2012.
# INPUT: (\@id_event, \@cls_gblocks, $br_len, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries, $if_lm, $if_rm), where
#
#	@id_event = ($stat_parent, $stat_node, \@indices_blk) describing the input single indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#	
#	@{$cls_gblocks[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	$br_len is the length of the specified branch;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	[ @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p); ] --- OBSOLETE as of July 11 (Wed), 2012.
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p);
#
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	@params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end), with
#
#	$cutoff_5p/3p being the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region,
#	$opt_bd_indels = 0 if you want to ignore the indels extending beyond the alignment boundary,
#			= 1 if you want to only account for the rate of indels (but not the entire integrand) extending beyond the boundary,
#			= 2 if you want to fully account for the effect of indels extending beyond the boundary
#	(Theoretically, option '2' is the most reasonable but it takes long),
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise);
#
#	[ $indx_lm/rm_blk is the left-/right-most block with the 'presence' state on either end of the specified branch. ] --- OBSOLETE as of July 22 (Sun), 2012.
#
#	$if_lm/rm is the indicator of whether the input event is leftmost/rightmost
#	in the gap-block cluster.
#
# OUTPUT: ($lnlk1_sgl_event_cls, $fl_smpl), where
#	$lnlk1_sgl_event_cls is the 1st approximation of the log-likelihood contribution
#	of the input single indel event;
#	$fl_smpl = 1 if the calculation is simplified,
#		 = 0 otherwise.
#
sub cal_lnlk1_sgl_indel_cls (\@\@$$\@\@\@\@$$) {

	##my ($id_event, $cls_gblocks, $br_len, $char_abs, $nfacs, $params_ins, $params_del, $params_bds, $indx_lm_blk, $indx_rm_blk) = @_; --- OBSOLETE as of July 11 (Wed), 2012.
	##my ($id_event, $cls_gblocks, $br_len, $char_abs, $log_nfacs, $params_ins, $params_del, $params_bds, $indx_lm_blk, $indx_rm_blk) = @_; --- OBSOLETE as of July 22 (Sun), 2012.
	my ($id_event, $cls_gblocks, $br_len, $char_abs, $log_nfacs, $params_ins, $params_del, $params_bds, $if_lm, $if_rm) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'cal_lnlk1_sgl_indel_cls' >\n\n") if ($TEST_LEVEL2a>0);

	my ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p) = @{$log_nfacs};
	##my ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p) = @{$nfacs}; --- OBSOLETE as of July 11 (Wed), 2012.
	##my ($totrate_ins, $distr_ins, $prm_ins) = @{$params_ins};
	##my ($totrate_del, $distr_del, $prm_del) = @{$params_del};
	my ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end) = @{$params_bds};

	my ($stat_ue, $stat_le, $indices_blk) = @{$id_event};

    if ($TEST_LEVEL2a>2) {
	print STDOUT '#  ', "[ Input Data ]\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$br_len = ', $br_len, "\n";
	print STDOUT '#  ', '$char_abs = ', $char_abs, "\n";
	##print STDOUT '#  ', '($indx_lm_blk, $indx_rm_blk) = ', "($indx_lm_blk, $indx_rm_blk)\n";
	print STDOUT '#  ', 'Leftmost event  ? ', ($if_lm ? 'Yes' : 'No'), ",\n";
	print STDOUT '#  ', 'Rightmost event ? ', ($if_rm ? 'Yes' : 'No'), " .\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '@{$cls_gblocks} = (# [$indx_gp, $len_gb, $start, $end]', "\n";
	foreach my $gblk (@{$cls_gblocks}) {
	    print STDOUT '#  ', '                  [', join (', ', @{$gblk}), "],\n";
	}
	print STDOUT '#  ', "                  )\n";
	print STDOUT "#\n";
	print STDOUT '#  ', 'Content of @{$id_event}:', "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($stat_ue, $stat_le) = ', "($stat_ue, $stat_le)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '@{$indices_blk} = (', join (', ', @{$indices_blk}), ")\n";
	print STDOUT "#\n"; 

	print STDOUT '#  ', '($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p)', "\n";
	print STDOUT '#  ', '= (', join (', ', @{$log_nfacs}), ")\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($totrate_ins, $distr_ins, $prm_ins) = (', join (', ', @{$params_ins}), ")\n";
	print STDOUT '#  ', '($totrate_del, $distr_del, $prm_del) = (', join (', ', @{$params_del}), ")\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end)', "\n";
	print STDOUT '#  ', '= (', join (', ', @{$params_bds}), ")\n";
	print STDOUT "#\n" x 2;
    }

	  # See if the indel event itself extends into the 5'- or 3'-end of the alignment. #
	my ($if_5p_end_indel, $if_3p_end_indel) = ($if_5p_end, $if_3p_end);

		# MODIFIED accompanying the introduction of 'search_left/rightmost_events' on July 22 (Sun), 2012. #
	($if_5p_end_indel and !($if_lm)) and ($if_5p_end_indel = 0);
	($if_3p_end_indel and !($if_rm)) and ($if_3p_end_indel = 0);
	##($if_5p_end_indel and ($indx_lm_blk < $indices_blk->[0])) and ($if_5p_end_indel = 0);
	##($if_3p_end_indel and ($indices_blk->[$#{$indices_blk}] < $indx_rm_blk)) and ($if_3p_end_indel = 0);

    if ($TEST_LEVEL2a>1) {
    ##if ($TEST_LEVEL2a>2) {
	print STDOUT '#  ', '[ Secondary Inputs ]', "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($if_5p_end_indel, $if_3p_end_indel) = ', "($if_5p_end_indel, $if_3p_end_indel)\n";
	print STDOUT "#\n";
    }

	my $fl_smpl = 0;
	##if (($opt_bd_indels>1) and $if_5p_end_indel and $if_3p_end_indel) {
	##   print STDOUT "[WARNING 'cal_lnlk1_sgl_indel_cls'] The event extends to BOTH 5'- and 3'- ends of the alignment. So I will lower the indel boundary handling mode from $opt_bd_indels to 1...\n";
	##   $fl_smpl = 1;
	##   $opt_bd_indels = 1;
	##} ==> DISABLED because I noticed that this simplification does NOT significantly reduce the computation time.

	my $len_indel = 0;
	foreach my $indx (@{$indices_blk}) { # Preliminary foreach-loop #
	    my ($indx_gp, $len_gb, $start, $end)  = @{$cls_gblocks->[$indx]};
	    $len_indel += $len_gb;
	}			# End of the preliminary foreach-loop #

	unless ($len_indel <= $cutoff_3p) {
	    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The indel length (=$len_indel) is LONGER THAN the regular CUTOFF (", '$cutoff_3p=', "$cutoff_3p): $! !\n";
	    return (undef, undef);
	}

	my $if_ins = ($stat_ue eq $char_abs) ? 1 : 0;

	my ($log_nfac_5p, $log_nfac_3p) = $if_ins ?
		($log_nfac_ins_5p, $log_nfac_ins_3p) : ($log_nfac_del_5p, $log_nfac_del_3p) ;
	##my ($nfac_5p, $nfac_3p) = $if_ins ? ($nfac_ins_5p, $nfac_ins_3p) : ($nfac_del_5p, $nfac_del_3p) ; -- OBSOLETE as of July 11 (Wed), 2012.
	my ($totrate, $distri, $param) = $if_ins ?  @{$params_ins} : @{$params_del} ;
	## my @indel_parameters = ($totrate, $distri, $param, $nfac_5p, $nfac_3p);

	my $sum_totrates = $params_ins->[0] + $params_del->[0]; # {total insertion rate} + {total deletion rate}

    if ($TEST_LEVEL2a>2) {
	print STDOUT '#  ', '$len_indel = ', $len_indel, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', 'if_ins = ', $if_ins, "\n"; 
	print STDOUT "#\n";
	print STDOUT '#  ', '($log_nfac_5p, $log_nfac_3p) = ', "($log_nfac_5p, $log_nfac_3p)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($totrate, $distri, $param) = ', "($totrate, $distri, $param)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$sum_totrates = ', $sum_totrates, "\n";
	print STDOUT "#\n" x 2;
	print STDOUT '#  ', '[ Main Process ]', "\n";
	print STDOUT "#\n";
    }

	my $lnlk1_sgl_indel_cls = log ($totrate);

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', 'Initial $lnlk1_sgl_indel_cls (= log ($totrate)) = ', $lnlk1_sgl_indel_cls, "\n";
	print STDOUT "#\n";
    }

	if (($opt_bd_indels<2) or !($if_5p_end_indel or $if_3p_end_indel)) {

    (print STDOUT ' ' x 3, "... Outer Block I (single or approximate calculation) ...\n\n") if ($TEST_LEVEL2a>1);
   
		# First, calculate the time integral of the exponential term. #
	    my $delta_l = $if_ins ? $len_indel : - $len_indel;
	    my $t_integ = (1 - exp (- $sum_totrates * $delta_l * $br_len))/$sum_totrates/$delta_l;

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', '$t_integ = ', $t_integ, ' , log ($t_integ) = ', log ($t_integ), "\n";
	print STDOUT "#\n";
    }

	    $lnlk1_sgl_indel_cls += log ($t_integ);

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', 'Second $lnlk1_sgl_indel_cls = ', $lnlk1_sgl_indel_cls, "\n";
	print STDOUT "#\n";
    }

	    my $log_freq;
	    ##my $freq;
	    if ($opt_bd_indels>0) {

    (print STDOUT ' ' x 5, ".. Inner Block A (composite freq) ..\n\n") if ($TEST_LEVEL2a>1);

		$log_freq = log_compos_freq_indel_len ($if_5p_end_indel, $if_3p_end_indel, $distri, $param, $cutoff_5p, $cutoff_3p, $log_nfac_5p, $log_nfac_3p, $len_indel);
		## $freq = compos_freq_indel_len ($if_5p_end_indel, $if_3p_end_indel, $distri, $param, $cutoff_5p, $cutoff_3p, $nfac_5p, $nfac_3p, $len_indel);
		unless (defined $log_freq) {
		    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The subroutine 'log_compos_freq_indel_len' FAILED: $! !\n";
		    return (undef, undef);
		}

	    } else {

    (print STDOUT ' ' x 5, ".. Inner Block B (single freq) ..\n\n") if ($TEST_LEVEL2a>1);

		    # $opt_bd_indels = 0 (i.e. we ignore the boundary effects altogether). #

	    	 $log_freq = log_freq_indel_len ($distri, $param, $len_indel, $log_nfac_3p);
	    	 ## $freq = freq_indel_len ($distri, $param, $len_indel, $nfac_3p);
	    }

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', '$log_freq = ', $log_freq, "\n";
	print STDOUT "#\n";
    }

	    $lnlk1_sgl_indel_cls += $log_freq;
	    ##$lnlk1_sgl_indel_cls += log $freq;

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', 'Final $lnlk1_sgl_indel_cls = ', $lnlk1_sgl_indel_cls, "\n";
	print STDOUT "#\n";
    }

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'cal_lnlk1_sgl_indel_cls' BRIEFLY Ended --\n\n") if ($TEST_LEVEL2a>0);

	    return ($lnlk1_sgl_indel_cls, $fl_smpl);
	}

		## The rest is for $opt_bd_indels = 2
		## (i.e. we seriously calculate the boundary effects).

    (print STDOUT ' ' x 3, "... Outer Block II (full calculation when extending into either end) ...\n\n") if ($TEST_LEVEL2a>1);

	my @terms_logsum_freq_tint = ();
	##my $sum_freq_tinteg_exp = 0;

	if ($if_5p_end_indel and $if_3p_end_indel) {
			# The indel event extends into BOTH ends. #

    (print STDOUT ' ' x 5, ".. Inner Block A (5'- and 3'-ends) ..\n\n") if ($TEST_LEVEL2a>1);

	    my $log_denom = log_dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, 1);
	    ##my $log_denom = log_dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, 0); # A potential bug fixed on July 15 (Sun), 2012.
	    ## my $denom = dbl_sum_freq_indel ($distri, $param, $cutoff_5p, $cutoff_3p, 1);
	    unless (defined $log_denom) {
		print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The subroutine 'log_dbl_sum_freq_indel' FAILED to calculate the denominator for an indel events extending into BOTH ends: $! !\n";
		return (undef, undef);
	    }

	    my $ub_5 = $cutoff_5p - int ($len_indel/2) + 1;
	    unless ($ub_5 >0) {
		print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The indel length (=$len_indel) is SO LARGE that ", ' $ub_5 ', "(=$ub_5) is NONPOSITIVE: $! !\n";
		return (undef, undef);
	    }
	    my $ub_z = $cutoff_5p + $cutoff_3p + 1;

		# A Bug-fixation measure on July 15 (Sun), 2012. #
	    my $width_3 = $cutoff_3p + int ($len_indel/2) - $len_indel + 1;
	    unless ($width_3>0) {
		print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] ILLEGAL INPUT: The number of summands in the 3'-direction (\$width_3=$width_3) is NOT POSITIVE: $! !\n";
		return (undef, undef);
	    }

    if ($TEST_LEVEL2a>2) {
	print STDOUT '#  ', '$log_denom = ', $log_denom, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($ub_z, $ub_5, $width_3) = ', "($ub_z, $ub_5, $width_3)\n";
	print STDOUT "#\n";
    }
    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'coeff', '(log coeff)', 't_integ', '(log t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

	    for (my $z = $len_indel; $z<$ub_z; $z++) { # 1st inner for-loop (over indel lengths) #

		my $coeff = $z-$len_indel+1;
		($coeff > $ub_5) and ($coeff = $ub_5);
		($coeff > $ub_z - $z) and ($coeff = $ub_z - $z);

		# A Bug-fixation measure on July 15 (Sun), 2012. #
		($coeff > $width_3) and ($coeff = $width_3);

		my $log_freq0 = my $log_freq = log_freq_indel_len ($distri, $param, $z, $log_denom);
		##my $freq = freq_indel_len ($distri, $param, $z, $denom);

		my $delta_l = $if_ins ? $z : - $z;
		my $t_integ = (1 - exp (- $sum_totrates * $delta_l * $br_len))/$sum_totrates/$delta_l;

		$log_freq += log ($coeff * $t_integ);
		push @terms_logsum_freq_tint, $log_freq;
		##$sum_freq_tinteg_exp += $coeff * $freq * $t_integ;

    if ($TEST_LEVEL2a>3) {
	my $log_coeff = log $coeff;
	my $log_t_integ = log $t_integ;
	print STDOUT '#  ', join ("\t", $z, $log_freq0, $coeff, "($log_coeff)", $t_integ, "($log_t_integ)", $log_freq), "\n";
    }

	    }			# End of the 1st inner for-loop (over indel lengths) #

    if ($TEST_LEVEL2a>3) { print STDOUT "#\n"; }

	} elsif ($if_5p_end_indel) {
			# The indel event extends into the 5'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block B (5'-end) ..\n\n") if ($TEST_LEVEL2a>1);

	    unless ($len_indel <= $cutoff_5p) {
		print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The indel length (=$len_indel) is LONGER THAN the CUTOFF (", '$cutoff_5p=', "$cutoff_5p): $! !\n";
		return (undef, undef);
	    }

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 't_integ', '(log t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

	    for (my $l=$len_indel; $l<= $cutoff_5p; $l++) { # 2nd inner for-loop (over indel lengths) #

		my $delta_l = $if_ins ? $l : - $l;
		my $t_integ = (1 - exp (- $sum_totrates * $delta_l * $br_len))/$sum_totrates/$delta_l;
		my $log_freq0 = my $log_freq = log_freq_indel_len ($distri, $param, $l, $log_nfac_5p);
		##my $freq = freq_indel_len ($distri, $param, $l, $nfac_5p);

		$log_freq += log ($t_integ);
		push @terms_logsum_freq_tint, $log_freq;
		##$sum_freq_tinteg_exp += $freq * $t_integ;

    if ($TEST_LEVEL2a>3) {
	my $log_t_integ = log $t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $t_integ, "($log_t_integ)", $log_freq), "\n";
    }

	    }			# End of the 2nd inner for-loop (over indel lengths) #

    if ($TEST_LEVEL2a>3) { print STDOUT "#\n"; }

	} elsif ($if_3p_end_indel) {
			# The indel event extends into the 3'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block C (3p-end) ..\n\n") if ($TEST_LEVEL2a>1);

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 't_integ', '(log t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

	    for (my $l=$len_indel; $l<= $cutoff_3p; $l++) { # Inner for-loop (over indel lengths) #

		my $delta_l = $if_ins ? $l : - $l;
		my $t_integ = (1 - exp (- $sum_totrates * $delta_l * $br_len))/$sum_totrates/$delta_l;
		my $log_freq0 = my $log_freq = log_freq_indel_len ($distri, $param, $l, $log_nfac_3p);
		##my $freq = freq_indel_len ($distri, $param, $l, $nfac_3p);

		$log_freq += log ($t_integ);
		push @terms_logsum_freq_tint, $log_freq;
		##$sum_freq_tinteg_exp += $freq * $t_integ;

    if ($TEST_LEVEL2a>3) {
	my $log_t_integ = log $t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $t_integ, "($log_t_integ)", $log_freq), "\n";
    }

	    }			# End of the inner for-loop (over indel lengths) #

    if ($TEST_LEVEL2a>3) { print STDOUT "#\n"; }

	} else {

    (print STDOUT ' ' x 5, ".. Inner Block D (Neither end) ..\n\n") if ($TEST_LEVEL2a>1);

		# This canNOT occur. #
	    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] Now we are in an IMPOSSIBLE ZONE: $! !\n";
	    return (undef, undef);
	}

	my $log_sum_freq_tinteg_exp = log_sum_exp (@terms_logsum_freq_tint);

    if ($TEST_LEVEL2a>3) {
	print STDOUT '#  ', '$log_sum_freq_tinteg_exp = ', $log_sum_freq_tinteg_exp, "\n";
	print STDOUT "#\n";
    }

	$lnlk1_sgl_indel_cls += $log_sum_freq_tinteg_exp ;
	##$lnlk1_sgl_indel_cls += log $sum_freq_tinteg_exp ;

    if ($TEST_LEVEL2a>2) {
	print STDOUT '#  ', 'Final $lnlk1_sgl_indel_cls = ', $lnlk1_sgl_indel_cls, "\n";
	print STDOUT "#\n";
    }

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'cal_lnlk1_sgl_indel_cls' Ended --\n\n") if ($TEST_LEVEL2a>0);

	return ($lnlk1_sgl_indel_cls, $fl_smpl);

}		# End of the subroutine 'cal_lnlk1_sgl_indel_cls' #


# cal_lnlk1_pair_indel_cls:
#	This satellite subroutine of 'cal_lnlk1_indel_br_cls' calculates
#	the 1st approximation of the log-likelihood of a pair of overlapping indel events
#	along a specified branch in a specified gap-block cluster.
#	[ NOTE that the order of the events is always: insertion -> deletion
#	(, because deletion followed by insertion will never be aligned
#	in the same block(s) in principle). ]
#
#	[ Major revision on July 22 (Sun), 2012]
#	The last two input arguments are now \%indx_indel2if_lm/rm,
#	which indicates whether the indel event in question is 'left/right-most' or not.
#
# [ INPUT: (\@info_pair, \@indels_br, \@cls_gblocks, $br_len, $char_abs, \@nfacs, \@params_ins, \@params_del, \@params_boundaries, $indx_lm_blk, $indx_rm_blk), where ] --- OBSOLETE as of July 11 (Wed), 2012.
# [ INPUT: (\@info_pair, \@indels_br, \@cls_gblocks, $br_len, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries, $indx_lm_blk, $indx_rm_blk), where ] --- OBSOLETE as of July 22 (Sun), 2012.
# INPUT: (\@info_pair, \@indels_br, \@cls_gblocks, $br_len, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries, \%indx_indel2if_lm, \%indx_indel2if_rm), where
#
#	@info_pair = ($indx_ins, $indx_del, \@indices_ovlp_blk) provides information on the pair, with
#	$indx_ins/del being the index of the subject insertion/deletion event (in @indels_br),
#	and @indices_ovlp_blk listing the indices (in @cls_gblocks) of the gap-blocks where the indel events overlap;
#
#	@{$indels_br[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event along the specified branch, with $stat_ue/le being the state at the upper-/lower-end of the branch, and with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	@{$cls_gblocks[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	$br_len is the length of the specified branch;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	[ @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p); ] --- OBSOLETE as of July 11 (Wed), 2012.
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p);
#
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	@params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end), with
#
#	$cutoff_5p/3p being the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region,
#	$opt_bd_indels = 0 if you want to ignore the indels extending beyond the alignment boundary,
#			= 1 if you want to only account for the rate of indels (but not the entire integrand) extending beyond the boundary,
#			= 2 if you want to fully account for the effect of indels extending beyond the boundary
#	(Theoretically, option '2' is the most reasonable but it takes long),
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise);
#
#	[ $indx_lm/rm_blk is the left-/right-most block with the 'presence' state on either end of the specified branch. ] --- OBSOLETE as of July 22 (sun), 2012.
#	%indx_indel2if_lm/rm = ($indx_indel => 1, ...) for the indel events
#	that are 'leftmost/rightmost'.
#
# OUTPUT: ($lnlk1_pair_indel_cls, $fl_smpl), where
#	$lnlk1_pair_indel_cls is the 1st approximation of the log-likelihood contribution
#	of the input pair of insertion and deletion events;
#	$fl_smpl = 1 if the calculation is simplified,
#		 = 0 otherwise.
#
##sub cal_lnlk1_pair_indel_cls (\@\@\@$$\@\@\@\@$$) { -- OBSOLETE as of July 22 (Sun), 2012.
#
sub cal_lnlk1_pair_indel_cls (\@\@\@$$\@\@\@\@\%\%) {

	##my ($info_pair, $indels_br, $cls_gblocks, $br_len, $char_abs, $nfacs, $params_ins, $params_del, $params_bds, $indx_lm_blk, $indx_rm_blk) = @_; -- OBSOLETE as of July 11 (Wed), 2012.
	## my ($info_pair, $indels_br, $cls_gblocks, $br_len, $char_abs, $log_nfacs, $params_ins, $params_del, $params_bds, $indx_lm_blk, $indx_rm_blk) = @_; --- OBSOLETE as of July 22 (Sun), 2012.
	my ($info_pair, $indels_br, $cls_gblocks, $br_len, $char_abs, $log_nfacs, $params_ins, $params_del, $params_bds, $indx_indel2if_lm, $indx_indel2if_rm) = @_;

    (print STDOUT '#  ', "< Subroutine 'cal_lnlk1_pair_indel_cls' >\n\n") if ($TEST_LEVEL2b>0);

	my ($indx_ins, $indx_del, $indices_ovlp_blk) = @{$info_pair};

	my ($stat_ue_ins, $stat_le_ins, $indices_blk_ins) = @{$indels_br->[$indx_ins]};
	unless ($stat_ue_ins eq $char_abs) {
	    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] ILLEGAL INPUT: The state of the insertion event (index: $indx_ins) at the upper-end of the branch (=$stat_ue_ins) is NOT the 'absence' state (=$char_abs): $! !\n";
	    return (undef, undef);
	}
	my ($stat_ue_del, $stat_le_del, $indices_blk_del) = @{$indels_br->[$indx_del]};
	unless ($stat_le_del eq $char_abs) {
	    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] ILLEGAL INPUT: The state of the deletion event (index: $indx_del) at the lower-end of the branch (=$stat_le_del) is NOT the 'absence' state (=$char_abs): $! !\n";
	    return (undef, undef);
	}

	my ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p) = @{$log_nfacs};
	##my ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p) = @{$nfacs}; --- OBSOLETE as of July 11 (Wed), 2012.
	my ($totrate_ins, $distr_ins, $prm_ins) = @{$params_ins};
	my ($totrate_del, $distr_del, $prm_del) = @{$params_del};
	my ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end) = @{$params_bds};

		# ADDED accompanying the introduction of 'search_left/rightmost_events' (on July 22 (Sun), 2012)
	my $if_lm_ins = $indx_indel2if_lm->{$indx_ins};
	my $if_lm_del = $indx_indel2if_lm->{$indx_del};
	my $if_rm_ins = $indx_indel2if_rm->{$indx_ins};
	my $if_rm_del = $indx_indel2if_rm->{$indx_del};

    if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', "[ Raw Input Data ]\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$br_len = ', $br_len, "\n";
	print STDOUT '#  ', '$char_abs = ', $char_abs, "\n";
	my @out_if = ();
		# MODIFIED accompanying the introduction of 'search_left/rightmost_events' (on July 22 (Sun), 2012)
	foreach my $if ($if_lm_ins, $if_lm_del, $if_rm_ins, $if_rm_del) { push @out_if, ((defined $if_lm_ins) ? $if_lm_ins : 'undef'); }
	print STDOUT '#  ', '($if_lm_ins, $lf_lm_del, $if_rm_ins, $if_rm_del) = (', join (', ', @out_if), ")\n";
	##print STDOUT '#  ', '($indx_lm_blk, $indx_rm_blk) = ', "($indx_lm_blk, $indx_rm_blk)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '@{$cls_gblocks} = (# [$indx_gp, $len_gb, $start, $end]', "\n";
	foreach my $gblk (@{$cls_gblocks}) {
	    print STDOUT '#  ', '                  [', join (', ', @{$gblk}), "],\n";
	}
	print STDOUT '#  ', "                  )\n";
	print STDOUT "#\n";
	print STDOUT '#  ', 'Insertion/deletion events:', "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', join ("\t", 'Ins/Del  ', 'index', 'state_before', 'state_after', 'indices,of,blocks,involved'), "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', join ("\t", 'Insertion', $indx_ins, $stat_ue_ins, $stat_le_ins, join (',', @{$indices_blk_ins}) ), "\n";
	print STDOUT '#  ', join ("\t", 'Deletion ', $indx_del, $stat_ue_del, $stat_le_del, join (',', @{$indices_blk_del}) ), "\n";
	print STDOUT "#\n";
	print STDOUT "#\n"; 

	print STDOUT '#  ', '($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p)', "\n";
	print STDOUT '#  ', '= (', join (', ', @{$log_nfacs}), ")\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($totrate_ins, $distr_ins, $prm_ins) = (', join (', ', @{$params_ins}), ")\n";
	print STDOUT '#  ', '($totrate_del, $distr_del, $prm_del) = (', join (', ', @{$params_del}), ")\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end)', "\n";
	print STDOUT '#  ', '= (', join (', ', @{$params_bds}), ")\n";
	print STDOUT "#\n" x 2;
    }

	my $sum_totrates = $totrate_ins + $totrate_del;


	  # See if the indel event itself extends into the 5'- or 3'-end of the alignment. #
	my $if_5p_end_ins = my $if_5p_end_del = $if_5p_end;
	my $if_3p_end_ins = my $if_3p_end_del = $if_3p_end;

		# MODIFIED accompanying the introdcution of 'search_left/rightmost_events' (on July 22 (Sun), 2012) #
	($if_5p_end_ins and !($if_lm_ins)) and ($if_5p_end_ins = 0);
	($if_5p_end_del and !($if_lm_del)) and ($if_5p_end_del = 0);
	($if_3p_end_ins and !($if_rm_ins)) and ($if_3p_end_ins = 0);
	($if_3p_end_del and !($if_rm_del)) and ($if_3p_end_del = 0);
	##($if_5p_end_ins and ($indx_lm_blk < $indices_blk_ins->[0])) and ($if_5p_end_ins = 0);
	##($if_5p_end_del and ($indx_lm_blk < $indices_blk_del->[0])) and ($if_5p_end_del = 0);
	##($if_3p_end_ins and ($indices_blk_ins->[$#{$indices_blk_ins}] < $indx_rm_blk)) and ($if_3p_end_ins = 0);
	##($if_3p_end_del and ($indices_blk_del->[$#{$indices_blk_del}] < $indx_rm_blk)) and ($if_3p_end_del = 0);

	my $if_end_ins = ($if_5p_end_ins or $if_3p_end_ins) ? 1 : 0;
	my $if_end_del = ($if_5p_end_del or $if_3p_end_del) ? 1 : 0;

    if ($TEST_LEVEL2b>1) {
    ##if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', '[ Secondary Inputs ]', "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($if_5p_end_ins, $if_3p_end_ins) = ', "($if_5p_end_ins, $if_3p_end_ins)\n";
	print STDOUT '#  ', '($if_5p_end_del, $if_3p_end_del) = ', "($if_5p_end_del, $if_3p_end_del)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($if_end_ins, $if_end_del) = ', "($if_end_ins, $if_end_del)\n";
	print STDOUT "#\n" x 2;
    }

	my $fl_smpl = 0;
	if (($opt_bd_indels>1) and $if_end_ins and $if_end_del) {
	   print STDOUT "[WARNING 'cal_lnlk1_pair_indel_cls'] BOTH insertion and deletion extend to EITHER 5'-end or 3'-end of the alignment. So I will lower the indel boundary handling mode from $opt_bd_indels to 1...\n";
	   $fl_smpl = 1;
	   $opt_bd_indels = 1;
	}

	my $len_ins = my $len_del = my $len_ovlp = 0;

	foreach my $indx (@{$indices_blk_ins}) { # 1st preliminary foreach-loop #
	    my ($indx_gp, $len_gb, $start, $end)  = @{$cls_gblocks->[$indx]};
	    $len_ins += $len_gb;
	}			# End of the 1st preliminary foreach-loop #

	foreach my $indx (@{$indices_blk_del}) { # 2nd preliminary foreach-loop #
	    my ($indx_gp, $len_gb, $start, $end)  = @{$cls_gblocks->[$indx]};
	    $len_del += $len_gb;
	}			# End of the 2nd preliminary foreach-loop #

	##foreach my $indx (@{$indices_ovlp_blk}) { # 3rd preliminary foreach-loop #
	##    my ($indx_gp, $len_gb, $start, $end)  = @{$cls_gblocks->[$indx]};
	##    $len_ovlp += $len_gb;
	##}			# End of the 3rd preliminary foreach-loop #

	##my @indel_parameters = ($totrate, $distri, $param, $nfac_5p, $nfac_3p);

	unless ($len_ins <= $cutoff_3p) {
	    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The insertion length (=$len_ins) is LONGER THAN the regular CUTOFF (", '$cutoff_3p=', "$cutoff_3p): $! !\n";
	    return (undef, undef);
	}

	unless ($len_del <= $cutoff_3p) {
	    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The deletion length (=$len_del) is LONGER THAN the regular CUTOFF (", '$cutoff_3p=', "$cutoff_3p): $! !\n";
	    return (undef, undef);
	}


    if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', '($len_ins, $len_del) = ', "($len_ins, $len_del)\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$sum_totrates = ', $sum_totrates, "\n";
	print STDOUT "#\n" x 2;
	print STDOUT '#  ', '[ Main Process ]', "\n";
	print STDOUT "#\n";
    }

	my $lnlk1_pair_indel_cls = log ($totrate_ins) + log ($totrate_del);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', 'Initial $lnlk1_pair_indel_cls (= log ($totrate_ins) + log ($totrate_del) ) = ', $lnlk1_pair_indel_cls, "\n";
	print STDOUT "#\n";
    }

	if (($opt_bd_indels<2) or !($if_end_ins or $if_end_del)) {
   
    (print STDOUT ' ' x 3, "... Outer Block I (single or approximate) ...\n\n") if ($TEST_LEVEL2b>1);
   
		# First, calculate the time integral of the exponential term. #
	    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $len_ins, $len_del, $br_len);
	    unless (defined $dbl_t_integ) {
		print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral: $! !\n";
		return (undef, undef);
	    }

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', '$dbl_t_integ = ', $dbl_t_integ, ' , log ($dbl_t_integ) = ', log ($dbl_t_integ), "\n";
	print STDOUT "#\n";
    }

	    $lnlk1_pair_indel_cls += log ($dbl_t_integ);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', 'Second $lnlk1_pair_indel_cls = ', $lnlk1_pair_indel_cls, "\n";
	print STDOUT "#\n";
    }

	    my ($log_freq_ins, $log_freq_del);
	    ##my ($freq_ins, $freq_del);
	    if ($opt_bd_indels>0) {

    (print STDOUT ' ' x 5, ".. Inner Block A (composite indel freqs) ..\n\n") if ($TEST_LEVEL2b>1);

		$log_freq_ins = log_compos_freq_indel_len ($if_5p_end_ins, $if_3p_end_ins, $distr_ins, $prm_ins, $cutoff_5p, $cutoff_3p, $log_nfac_ins_5p, $log_nfac_ins_3p, $len_ins);
		##$freq_ins = compos_freq_indel_len ($if_5p_end_ins, $if_3p_end_ins, $distr_ins, $prm_ins, $cutoff_5p, $cutoff_3p, $nfac_ins_5p, $nfac_ins_3p, $len_ins);
		unless (defined $log_freq_ins) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'log_compos_freq_indel_len' FAILED on the insertion frequency: $! !\n";
		    return (undef, undef);
		}

		$log_freq_del = log_compos_freq_indel_len ($if_5p_end_del, $if_3p_end_del, $distr_del, $prm_del, $cutoff_5p, $cutoff_3p, $log_nfac_del_5p, $log_nfac_del_3p, $len_del);
		##$freq_del = compos_freq_indel_len ($if_5p_end_del, $if_3p_end_del, $distr_del, $prm_del, $cutoff_5p, $cutoff_3p, $nfac_del_5p, $nfac_del_3p, $len_del);
		unless (defined $log_freq_del) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'log_compos_freq_indel_len' FAILED on the deletion frequency: $! !\n";
		    return (undef, undef);
		}

	    } else {

		    # $opt_bd_indels = 0 (i.e. we ignore the boundary effects altogether). #

    (print STDOUT ' ' x 5, ".. Inner Block B (single indel freqs) ..\n\n") if ($TEST_LEVEL2b>1);

	    	$log_freq_ins = log_freq_indel_len ($distr_ins, $prm_ins, $len_ins, $log_nfac_ins_3p);
		$log_freq_del = log_freq_indel_len ($distr_del, $prm_del, $len_del, $log_nfac_del_3p);
	    	##$freq_ins = freq_indel_len ($distr_ins, $prm_ins, $len_ins, $nfac_ins_3p);
		##$freq_del = freq_indel_len ($distr_del, $prm_del, $len_del, $nfac_del_3p);
	    }

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', '($log_freq_ins, $log_freq_del) = ', "($log_freq_ins, $log_freq_del)\n";
	print STDOUT "#\n";
    }

	    $lnlk1_pair_indel_cls += $log_freq_ins + $log_freq_del;
	    ##$lnlk1_pair_indel_cls += log ($freq_ins) + log ($freq_del);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', 'Final $lnlk1_pair_indel_cls = ', $lnlk1_pair_indel_cls, "\n";
	print STDOUT "#\n";
    }

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'cal_lnlk1_pair_indel_cls' BRIEFLY Ended --\n\n") if ($TEST_LEVEL2b>0);

	    return ($lnlk1_pair_indel_cls, $fl_smpl);
	}

		## The rest is for $opt_bd_indels == 2
		## (i.e. we seriously calculate the boundary effects).
		## Due to the preceding processes,
		## either $if_end_ins or $if_end_del, but NOT BOTH, should be true. 

	if ($if_end_ins) {
		# The insertion extends into either end (or both ends). #

    (print STDOUT ' ' x 3, "... Outer Block II (full calculation when the INSERTION extends into either end) ...\n\n") if ($TEST_LEVEL2b>1);

	    my @terms_logsum_freq_tint_ins = ();
	    ##my $sum_freq_tinteg_ins = 0;

	    if ($if_5p_end_ins and $if_3p_end_ins) {
			# The insertion event extends into BOTH ends. #

    (print STDOUT ' ' x 5, ".. Inner Block A (5'- and 3'-ends) ..\n\n") if ($TEST_LEVEL2b>1);

		my $log_denom = log_dbl_sum_freq_indel ($distr_ins, $prm_ins, $cutoff_5p, $cutoff_3p, 1);
		##my $log_denom = log_dbl_sum_freq_indel ($distr_ins, $prm_ins, $cutoff_5p, $cutoff_3p, 0); # A potential bug fixed on July 15 (Sun), 2012.
		## my $denom = dbl_sum_freq_indel ($distr_ins, $prm_ins, $cutoff_5p, $cutoff_3p, 1);
		unless (defined $log_denom) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'log_dbl_sum_freq_indel' FAILED to calculate the denominator for an insertion events extending into BOTH ends: $! !\n";
		    return (undef, undef);
		}

		my $ub_5 = $cutoff_5p - int ($len_ins/2) + 1;

		unless ($ub_5 >0) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The insertion length (=$len_ins) is SO LARGE that ", ' $ub_5 ', "(=$ub_5) is NONPOSITIVE: $! !\n";
		    return (undef, undef);
		}

		# A Bug-fixation measure on July 15 (Sun), 2012. #
		my $width_3 = $cutoff_3p + int ($len_ins/2) - $len_ins + 1;
		unless ($width_3>0) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] ILLEGAL INPUT: The number of summands in the 3'-direction (\$width_3=$width_3) for the insertion is NOT POSITIVE: $! !\n";
		    return (undef, undef);
		}

		my $ub_z = $cutoff_5p + $cutoff_3p + 1;

    if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', '$log_denom = ', $log_denom, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($ub_z, $ub_5, $width_3) = ', "($ub_z, $ub_5, $width_3)\n";
	print STDOUT "#\n";
    }
    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'coeff', '(log coeff)', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $z = $len_ins; $z<$ub_z; $z++) { # Inner for-loop (over insertion lengths) #

		    my $coeff = $z - $len_ins +1;
		    ($coeff > $ub_5) and ($coeff = $ub_5);
		    ($coeff > $ub_z - $z) and ($coeff = $ub_z - $z);

			# A Bug-fixation measure on July 15 (Sun), 2012. #
		    ($coeff > $width_3) and ($coeff = $width_3);

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_ins, $prm_ins, $z, $log_denom);
		    ##my $freq = freq_indel_len ($distr_ins, $prm_ins, $z, $denom);

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $z, $len_del, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the insertion length $z (with insertion extending into both ends): $! !\n";
			return (undef, undef);
		    }
		    $log_freq += log ($coeff * $dbl_t_integ);
		    push @terms_logsum_freq_tint_ins, $log_freq;
		    ##$sum_freq_tinteg_ins += $coeff * $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_coeff = log $coeff;
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $z, $log_freq0, $coeff, "($log_coeff)", $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over insertion lengths) #

    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }

	    } elsif ($if_5p_end_ins) {
		    	# The insertion event extends into the 5'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block B (5'-end) ..\n\n") if ($TEST_LEVEL2b>1);

		unless ($len_ins <= $cutoff_5p) {
		    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The insertion length (=$len_ins) is LONGER THAN the CUTOFF (", '$cutoff_5p=', "$cutoff_5p): $! !\n";
		    return (undef, undef);
		}

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $l=$len_ins; $l<= $cutoff_5p; $l++) { # Inner for-loop (over insertion lengths) #

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $l, $len_del, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the insertion length $l (with insertion extending into the 5'-end): $! !\n";
			return (undef, undef);
		    }

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_ins, $prm_ins, $l, $log_nfac_ins_5p);
		    ## my $freq = freq_indel_len ($distr_ins, $prm_ins, $l, $nfac_ins_5p);

		    $log_freq += log ($dbl_t_integ);
		    push @terms_logsum_freq_tint_ins, $log_freq;
		    ##$sum_freq_tinteg_ins += $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over insertion lengths) #

    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }

	    } elsif ($if_3p_end_ins) {
		    	# The insertion event extends into the 3'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block C (3'-end) ..\n\n") if ($TEST_LEVEL2b>1);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $l=$len_ins; $l<= $cutoff_3p; $l++) { # Inner for-loop (over insertion lengths) #

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $l, $len_del, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the insertion length $l (with insertion extending into the 3'-end): $! !\n";
			return (undef, undef);
		    }

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_ins, $prm_ins, $l, $log_nfac_ins_3p);
		    ##my $freq = freq_indel_len ($distr_ins, $prm_ins, $l, $nfac_ins_3p);

		    $log_freq += log ($dbl_t_integ);
		    push @terms_logsum_freq_tint_ins, $log_freq;
		    ##$sum_freq_tinteg_ins += $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over insertion lengths) #
    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }

	    }		# End of the middle if-elsif-else block. #

	    my $log_freq_del = log_freq_indel_len ($distr_del, $prm_del, $len_del, $log_nfac_del_3p);
	    ##my $freq_del = freq_indel_len ($distr_del, $prm_del, $len_del, $nfac_del_3p);

	    my $log_sum_freq_tinteg_ins = log_sum_exp (@terms_logsum_freq_tint_ins);
	    $lnlk1_pair_indel_cls += $log_sum_freq_tinteg_ins + $log_freq_del;
	    ##$lnlk1_pair_indel_cls += log ($sum_freq_tinteg_ins) + log ($freq_del);
	    
    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', '$log_freq_del = ', $log_freq_del, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$log_sum_freq_tinteg_ins = ', $log_sum_freq_tinteg_ins, "\n";
	print STDOUT "#\n";
    }

	} elsif ($if_end_del) {
		# The deletion extends into either end (or both ends). #

    (print STDOUT ' ' x 3, "... Outer Block III (full calculation when the DELETION extends into either end) ...\n\n") if ($TEST_LEVEL2b>1);

	    my $log_freq_ins = log_freq_indel_len ($distr_ins, $prm_ins, $len_ins, $log_nfac_ins_3p);
	    ##my $freq_ins = freq_indel_len ($distr_ins, $prm_ins, $len_ins, $nfac_ins_3p);

	    my @terms_logsum_freq_tint_del = ();
	    ##my $sum_freq_tinteg_del = 0;

	    if ($if_5p_end_del and $if_3p_end_del) {
			# The deletion event extends into BOTH ends. #

    (print STDOUT ' ' x 5, ".. Inner Block A (5'- and 3'-ends) ..\n\n") if ($TEST_LEVEL2b>1);

		my $log_denom = log_dbl_sum_freq_indel ($distr_del, $prm_del, $cutoff_5p, $cutoff_3p, 1);
		## my $log_denom = log_dbl_sum_freq_indel ($distr_del, $prm_del, $cutoff_5p, $cutoff_3p, 0); # A potential bug fixed on July 15 (Sun), 2012.
		##my $denom = dbl_sum_freq_indel ($distr_del, $prm_del, $cutoff_5p, $cutoff_3p, 1);
		unless (defined $log_denom) {
		    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The subroutine 'log_dbl_sum_freq_indel' FAILED to calculate the denominator for a deletion events extending into BOTH ends: $! !\n";
		    return (undef, undef);
		}

		my $ub_5 = $cutoff_5p - int ($len_del/2) + 1;

		unless ($ub_5 >0) {
		    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The deletion length (=$len_del) is SO LARGE that ", ' $ub_5 ', "(=$ub_5) is NONPOSITIVE: $! !\n";
		    return (undef, undef);
		}

		# A Bug-fixation measure on July 15 (Sun), 2012. #
		my $width_3 = $cutoff_3p + int ($len_del/2) - $len_del + 1;
		unless ($width_3>0) {
		    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] ILLEGAL INPUT: The number of summands in the 3'-direction (\$width_3=$width_3) for the deletion is NOT POSITIVE: $! !\n";
		    return (undef, undef);
		}

		my $ub_z = $cutoff_5p + $cutoff_3p + 1;

    if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', '$log_denom = ', $log_denom, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '($ub_z, $ub_5, $width_3) = ', "($ub_z, $ub_5, $width_3)\n";
	print STDOUT "#\n";
    }
    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'coeff', '(log coeff)', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $z = $len_del; $z<$ub_z; $z++) { # Inner for-loop (over deletion lengths) #

		    my $coeff = $z - $len_del +1;
		    ($coeff > $ub_5) and ($coeff = $ub_5);
		    ($coeff > $ub_z - $z) and ($coeff = $ub_z - $z);

			# A Bug-fixation measure on July 15 (Sun), 2012. #
		    ($coeff > $width_3) and ($coeff = $width_3);

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_del, $prm_del, $z, $log_denom);
		    ##my $freq = freq_indel_len ($distr_del, $prm_del, $z, $denom);

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $len_ins, $z, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the deletion length $z (with deletion extending into both ends): $! !\n";
			return (undef, undef);
		    }

		    $log_freq += log ($coeff * $dbl_t_integ);
		    push @terms_logsum_freq_tint_del, $log_freq;
		    ##$sum_freq_tinteg_del += $coeff * $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_coeff = log $coeff;
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $z, $log_freq0, $coeff, "($log_coeff)", $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over deletion lengths) #

    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }


	    } elsif ($if_5p_end_del) {
		    	# The deletion event extends into the 5'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block B (5'-end) ..\n\n") if ($TEST_LEVEL2b>1);

		unless ($len_del <= $cutoff_5p) {
		    print STDOUT "[ERROR 'cal_lnlk1_sgl_indel_cls'] The deletion length (=$len_del) is LONGER THAN the CUTOFF (", '$cutoff_5p=', "$cutoff_5p): $! !\n";
		    return (undef, undef);
		}

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $l=$len_del; $l<= $cutoff_5p; $l++) { # Inner for-loop (over deletion lengths) #

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $len_ins, $l, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the deletion length $l (with deletion extending into the 5'-end): $! !\n";
			return (undef, undef);
		    }

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_del, $prm_del, $l, $log_nfac_del_5p);
		    ##my $freq = freq_indel_len ($distr_del, $prm_del, $l, $nfac_del_5p);

		    $log_freq += log ($dbl_t_integ);
		    push @terms_logsum_freq_tint_del, $log_freq;
		    ##$sum_freq_tinteg_del += $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over deletion lengths) #
    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }

	    } elsif ($if_3p_end_del) {
		    	# The deletion event extends into the 3'-end. #

    (print STDOUT ' ' x 5, ".. Inner Block C (3'-end) ..\n\n") if ($TEST_LEVEL2b>1);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', join ("\t", 'len', 'log_freq', 'dbl_t_integ', '(log dbl_t_integ)', 'final_term'), "\n";
	print STDOUT "#\n";
    }

		for (my $l=$len_del; $l<= $cutoff_3p; $l++) { # Inner for-loop (over deletion lengths) #

		    my $dbl_t_integ = dbl_tinteg_exp_indel ($sum_totrates, $len_ins, $l, $br_len);
		    unless (defined $dbl_t_integ) {
			print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] The subroutine 'dbl_tinteg_exp_indel' FAILED to calculate the sole double time integral for the deletion length $l (with deletion extending into the 3'-end): $! !\n";
			return (undef, undef);
		    }

		    my $log_freq0 = my $log_freq = log_freq_indel_len ($distr_del, $prm_del, $l, $log_nfac_del_3p);

		    $log_freq += log ($dbl_t_integ);
		    push @terms_logsum_freq_tint_del, $log_freq;
		    ##$sum_freq_tinteg_del += $freq * $dbl_t_integ;

    if ($TEST_LEVEL2b>3) {
	my $log_dbl_t_integ = log $dbl_t_integ;
	print STDOUT '#  ', join ("\t", $l, $log_freq0, $dbl_t_integ, "($log_dbl_t_integ)", $log_freq), "\n";
    }

		}			# End of the inner for-loop (over deletion lengths) #

    if ($TEST_LEVEL2b>3) { print STDOUT "#\n"; }

	    }		# End of the middle if-elsif-else block. #

	    my $log_sum_freq_tinteg_del = log_sum_exp (@terms_logsum_freq_tint_del);

	    $lnlk1_pair_indel_cls += $log_freq_ins + $log_sum_freq_tinteg_del;
	    ##$lnlk1_pair_indel_cls += log ($freq_ins) + log ($sum_freq_tinteg_del);

    if ($TEST_LEVEL2b>3) {
	print STDOUT '#  ', '$log_freq_ins = ', $log_freq_ins, "\n";
	print STDOUT "#\n";
	print STDOUT '#  ', '$log_sum_freq_tinteg_del = ', $log_sum_freq_tinteg_del, "\n";
	print STDOUT "#\n";
    }

	} else {

    (print STDOUT ' ' x 5, ".. Inner Block D (Neither Insertion Nor Deletion extends) ..\n\n") if ($TEST_LEVEL2b>1);

	    print STDOUT "[ERROR 'cal_lnlk1_pair_indel_cls'] This is IMPOSSIBLE! NEITHER the insertion NOR the deletion extends into any of the 5'-end or the 3'-end of the alignment, and yet ", '$opt_bd_indels = ', "$opt_bd_indels (>1), such a case should have been processed already: $! !\n";
	    return (undef, undef);

	} 	# End of the outer if-elsif-else block. #

    if ($TEST_LEVEL2b>2) {
	print STDOUT '#  ', 'Final $lnlk1_pair_indel_cls = ', $lnlk1_pair_indel_cls, "\n";
	print STDOUT "#\n";
    }

    (print STDOUT "\n", ' ' x 3, "-- Subroutine 'cal_lnlk1_pair_indel_cls' Ended --\n\n") if ($TEST_LEVEL2b>0);


	return ($lnlk1_pair_indel_cls, $fl_smpl);

}		# End of the subroutine 'cal_lnlk1_pair_indel_cls' #


# cal_lnlk1_indel_br_cls:
#	This satellite subroutine of 'cal_lnlk1_indel_scn_cls'
#	calculates the 1st approximation of the log-likelihood of
#	indel events along a specified branch in a specified gap-block cluster.
#
# [ INPUT: (\@cls_gblocks, $br_len, \@indels_br, \@blwise_gstats_ue, \@blwise_gstats_le, $char_abs, \@nfacs, \@params_ins, \@params_del, \@params_boundaries), where ] --- OBSOLETE as of July 11 (Wed), 2012.
# INPUT: (\@cls_gblocks, $br_len, \@indels_br, \@blwise_gstats_ue, \@blwise_gstats_le, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries), where
#
#	@{$cls_gblocks[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	$br_len is the length of the specified branch;
#
#	@{$indels_br[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gblocks)
#	of the blocks involved in the indel event;
#
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th block (in the specified cluster) at the upper-/lower-end node of the specified branch;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	[ @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p); ] --- OBSOLETE as of July 11 (Wed), 2012
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p);
#
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	@params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end), with
#
#	$cutoff_5p/3p being the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region,
#	$opt_bd_indels = 0 if you want to ignore the indels extending beyond the alignment boundary,
#			= 1 if you want to only account for the rate of indels (but not the entire integrand) extending beyond the boundary,
#			= 2 if you want to fully account for the effect of indels extending beyond the boundary
#	(Theoretically, option '2' is the most reasonable but it takes long),
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise).
#
# OUTPUT: ($lnlk1_indel_br_cls, $ct_smpl), where
#	$lnlk1_indel_br_cls is the 1st approximation of the log-likelihood of the input
#	indel scenario along the specified branch in the specified gap-block cluster;
#	$ct_smpl is the number of simplifications that had to be done during the course of this subroutine.
# 
sub cal_lnlk1_indel_br_cls (\@$\@\@\@$\@\@\@\@) {

	##my ($cls_gblocks, $br_len, $indels_br, $blwise_gstats_ue, $blwise_gstats_le, $char_abs, $nfacs, $params_ins, $params_del, $params_bds) = @_; --- OBSOLETE as of July 11 (Wed), 2012.

	my ($cls_gblocks, $br_len, $indels_br, $blwise_gstats_ue, $blwise_gstats_le, $char_abs,
		$log_nfacs, $params_ins, $params_del, $params_bds) = @_;

    (print STDOUT ' ' x 2, "< Subroutine 'cal_lnlk1_indel_br_cls' (branch length = $br_len; character for 'absence' = $char_abs) >\n\n") if ($TEST_LEVEL1a>0);

	my ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p) = @{$log_nfacs};
	##my ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p) = @{$nfacs}; --- OBSOLETE as of July 11 (Wed), 2012.
	my ($totrate_ins, $distr_ins, $prm_ins) = @{$params_ins};
	my ($totrate_del, $distr_del, $prm_del) = @{$params_del};
	my ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end) = @{$params_bds};
	my $ct_gblks = @{$cls_gblocks};
	my $ct_indels = @{$indels_br};

    if ($TEST_LEVEL1a>2) {
	print STDOUT ' ' x 2, "[ Raw Input variables (#{gap-blocks} = $ct_gblks; #{indels} = $ct_indels) ]\n\n";

	print STDOUT ' ' x 2, '@{$cls_gblocks} = ( # [$indx_gp, $len_gb, $start, $end]', "\n";
	foreach my $gblock (@{$cls_gblocks}) {
	    print STDOUT ' ' x 2, '                 [', join (', ', @{$gblock}), "],\n";
	}
	print STDOUT ' ' x 2, '                  );', "\n\n";

	print STDOUT ' ' x 2, '@{$indels_br} = ( # [$stat_before, $stat_after, indices/involved/blocks]', "\n";
	foreach my $event (@{$indels_br}) {
	    my ($stat_bf, $stat_af, $indices_blks) = @{$event};
	    my $cnct_indices_blks = join ('/', @{$indices_blks});
	    print STDOUT ' ' x 2, '               [', join (', ', $stat_bf, $stat_af, $cnct_indices_blks), "],\n";
	}
	print STDOUT ' ' x 2, '                );', "\n\n";

	print STDOUT ' ' x 2, 'Contents of @{$blwise_gstats_ue} and @{$blwise_gstats_le}:', "\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Indx_blk', 'gstat_upper', 'gstat_lower'), "\n\n";
	for (my $i=0; $i<$ct_gblks; $i++) { print STDOUT ' ' x 2, join ("\t", $i, $blwise_gstats_ue->[$i], $blwise_gstats_le->[$i]), "\n"; }
	print STDOUT "\n\n";

	print STDOUT ' ' x 2, '($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p)', "\n";
	print STDOUT ' ' x 2, ' = (', join (', ', @{$log_nfacs}), ")\n\n";
	print STDOUT ' ' x 2, '($totrate_ins, $distr_ins, $prm_ins) = (', join (', ', @{$params_ins}), ")\n";
	print STDOUT ' ' x 2, '($totrate_del, $distr_del, $prm_del) = (', join (', ', @{$params_del}), ")\n\n";;
	print STDOUT ' ' x 2, '($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end)', "\n";
	print STDOUT ' ' x 2, '= (', join (', ', @{$params_bds}), ")\n\n";

    }

		## Initialize the output ##

    (print STDOUT ' ' x 2, '[ Initialize the output variable: $lnlk1_indel_br_cls ]', "\n\n") if ($TEST_LEVEL1a>1);

	##my $lnlk1_indel_br_cls = 0;
	my $totlen_init_blks = 0;
	for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st preliminary for-loop (over gap blocks) #
	    ($blwise_gstats_ue->[$bl] eq $char_abs) and next;

    (print STDOUT ' ' x 2, "( The $bl th block has a 'presence' state at the upper-end )\n\n") if ($TEST_LEVEL1a>2);

	    my ($indx_gp, $len_bl, $start, $end) = @{$cls_gblocks->[$bl]};
	    ##$lnlk1_indel_br_cls -= $len_bl;
	    $totlen_init_blks += $len_bl;

    if ($TEST_LEVEL1a>3) {
	print STDOUT ' ' x 2, "Block length = $len_bl ,\n";
	print STDOUT ' ' x 2, "Total initial block length so far = $totlen_init_blks .\n\n";
    }

	}			# End of the 1st preliminary for-loop (over gap blocks) #

	##$lnlk1_indel_br_cls *= $br_len * ($totrate_ins + $totrate_del);
	my $lnlk1_indel_br_cls = - $br_len * ($totrate_ins + $totrate_del) * $totlen_init_blks;

    if ($TEST_LEVEL1a>3) {
	print STDOUT ' ' x 2, "Total initial block length (final) = $totlen_init_blks ,\n";
	print STDOUT ' ' x 2, 'Branch length = ', $br_len, " ,\n";
	print STDOUT ' ' x 2, '$totrate_ins + $totrate_del = ', $totrate_ins + $totrate_del, " .\n\n";
    }
    (print STDOUT ' ' x 2, 'Initial $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " .\n\n") if ($TEST_LEVEL1a>2);

	unless (@{$indels_br}) {
		# NO indel event is assigned to the branch. #
	    return ($lnlk1_indel_br_cls, 0);
	}

    (print STDOUT ' ' x 2, "[ Preliminary Process (i): Sort out $ct_indels indel events into single and paired events ]\n\n") if ($TEST_LEVEL1a>1);

	# @{$paired_indels[$pr]} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
	# specifies the $pr th pair of an insertion and a deletion,
	# with $indx_ins/del being the index in @indels_br,
	# and @indices_ovlp_blk storing the indices of blocks where the two events overlap.
	#
	my ($sgl_indels, $paired_indels) = search_paired_indels (@{$indels_br}, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs);
	unless (defined $sgl_indels) {
	    print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'search_paired_indels' FAILED: $! !\n";
	    return (undef, undef);
	}
	my $ct_sgls = @{$sgl_indels};
	my $ct_pairs = @{$paired_indels};

    (print STDOUT "\n", ' ' x 2, "... 'search_paired_indels' Ended, finding $ct_sgls single events and $ct_pairs pairs of events...\n\n") if ($TEST_LEVEL1a>1);
    if ($TEST_LEVEL1a>2) {
	print STDOUT ' ' x 2, '@{$sgl_indels} = (', join (', ', @{$sgl_indels}), ");\n\n";
	print STDOUT ' ' x 2, '@{$paired_indels} = ( # [$indx_ins, $indx_del, $indices/overlapping/blocks]', "\n";
	foreach my $pair (@{$paired_indels}) {
	    my ($indx_ins, $indx_del, $indices_ovlp_blks) = @{$pair};
	    my $cnct_indices = join ('/', @{$indices_ovlp_blks}); 
	    print STDOUT ' ' x 2, '                    [', join (', ', $indx_ins, $indx_del, $cnct_indices), '],', "\n";
	}
	print STDOUT ' ' x 2, '                    );', "\n\n";
    }


		# NEWLY INTRODUCED in revision on July 22 (Sun), 2012. #

	my $indx_indel2if_rm = {};
	my $indx_indel2if_lm = {};

	if ($if_5p_end and ($opt_bd_indels>0)) {
		# A BUG fixed on July 23 (Mon), 2012. #
	    $indx_indel2if_lm = search_leftmost_events (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @{$indels_br}, $char_abs, @{$sgl_indels}, @{$paired_indels});
	    ##my $indx_indel2if_lm = search_leftmost_events (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @{$indels_br}, $char_abs, @{$sgl_indels}, @{$paired_indels});
	    unless (defined $indx_indel2if_lm) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'search_leftmost_events' FAILED: $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "\n", ' ' x 2, "... 'search_leftmost_events' Ended...\n\n") if ($TEST_LEVEL1a>1);
    if ($TEST_LEVEL1a>2) {
	print STDOUT "\n", ' ' x 2, '[ Contents of %{$indx_indel2if_lm} ]', "\n\n";
	print STDOUT ' ' x 2, "Indices of 'leftmost' indel events:  ", join (', ', sort {$a <=> $b} keys %{$indx_indel2if_lm}), " .\n\n";
    }
	}

	if ($if_3p_end and ($opt_bd_indels>0)) {
	    $indx_indel2if_rm = search_rightmost_events (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @{$indels_br}, $char_abs, @{$sgl_indels}, @{$paired_indels});
	    unless (defined $indx_indel2if_rm) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'search_rightmost_events' FAILED: $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "\n", ' ' x 2, "... 'search_rightmost_events' Ended...\n\n") if ($TEST_LEVEL1a>1);
    if ($TEST_LEVEL1a>2) {
	print STDOUT "\n", ' ' x 2, '[ Contents of %{$indx_indel_if_rm} ]', "\n\n";
	print STDOUT ' ' x 2, "Indices of 'rightmost' indel events: ", join (', ', sort {$a <=> $b} keys %{$indx_indel2if_rm}), " .\n\n";
    }
	}


=This part, as well as the involved subroutine, was made OBSOLETE on July 22 (Sun), 2012.
		# A Bug fixed on July 20 (Sat), 2012 #
	my ($indx_lm_blk, $indx_rm_blk) = ($ct_gblks-1, 0);
	if (($if_5p_end) and ($opt_bd_indels>0)) {
	    $indx_lm_blk = search_leftmost_block (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @{$indels_br}, $char_abs);
	    unless (defined $indx_lm_blk) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'search_leftmost_block' FAILED: $! !\n";
		return (undef, undef);
	    }
	}
	if (($if_3p_end) and ($opt_bd_indels>0)) {
	    $indx_rm_blk = search_rightmost_block (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @{$indels_br}, $char_abs);
	    unless (defined $indx_rm_blk) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'search_rightmost_block' FAILED: $! !\n";
		return (undef, undef);
	    }
	}
=cut

=A BUG fixed on July 20 (Sat), 2012.
	my ($indx_lm_blk, $indx_rm_blk);
	if (($if_5p_end) and ($opt_bd_indels>0)) {
	    for (my $bl=0; $bl<$ct_gblks; $bl++) { # 2nd preliminary for-loop (over gap blocks) #
		(($blwise_gstats_ue->[$bl] eq $char_abs) and ($blwise_gstats_le->[$bl] eq $char_abs)) and next;
		$indx_lm_blk = $bl;
		last;
	    }			# End of the 2nd preliminary for-loop (over gap blocks) #
	}
	if (($if_3p_end) and ($opt_bd_indels>0)) {
	    for (my $bl=$ct_gblks-1; $bl>=0; $bl--) { # 3rd preliminary for-loop (over gap blocks) #
		(($blwise_gstats_ue->[$bl] eq $char_abs) and ($blwise_gstats_le->[$bl] eq $char_abs)) and next;
		$indx_rm_blk = $bl;
		last;
	    }			# End of the 3rd preliminary for-loop (over gap blocks) #
	}
=cut

=OBSOLETE as of July 22 (Sun), 2012. #
    if ($TEST_LEVEL1a>2) {
	##my $out_indx_lm_blk = (defined $indx_lm_blk) ? $indx_lm_blk : 'undef';
	##my $out_indx_rm_blk = (defined $indx_rm_blk) ? $indx_rm_blk : 'undef';
	print STDOUT "\n", ' ' x 2, '($indx_lm_blk, $indx_rm_blk) = ', "($out_indx_lm_blk, $out_indx_rm_blk).\n\n";
    }
=cut

	my $ct_smpl = 0;

		## Main Process 1: calculate the contributions from single indel events. ##

    (print STDOUT "\n", ' ' x 2, "[ Main Process 1: Contributions from $ct_sgls single indel events ]\n\n") if ($TEST_LEVEL1a>1);

	for (my $sg = 0; $sg < $ct_sgls; $sg++) { # 1st main outer for-loop (over single indel events) #

	    my $indx_id = $sgl_indels->[$sg];
	    my $id_event = $indels_br->[$indx_id];

		# ADDED accompanying the introduction of 'search_left/rightmost_events' (on July 22 (Sun), 2012) #
	    my $if_lm = $indx_indel2if_lm->{$indx_id};
	    my $if_rm = $indx_indel2if_rm->{$indx_id};

    (print STDOUT ' ' x 2, "{ The $sg th event (indel index = $indx_id) }\n\n") if ($TEST_LEVEL1a>2);
    if ($TEST_LEVEL1a>3) {
	my ($stat_bf, $stat_af, $indices_blks) = @{$id_event};
	my $cnct_indices_blks = join ('/', @{$indices_blks});
	print STDOUT ' ' x 2, '($stat_before, $stat_after, indices/involved/blocks) = ', "($stat_bf, $stat_af, $cnct_indices_blks)\n\n";
	print STDOUT ' ' x 2, 'Leftmost  ? = ', ($if_lm ? 'Yes' : 'No'), " ,\n"; 
	print STDOUT ' ' x 2, 'Rightmost ? = ', ($if_rm ? 'Yes' : 'No'), " .\n\n";
    }


		# MODIFIED accompanying the introdcution of 'search_left/rightmost_events' (on July 22 (Sun), 2012) #
	    my ($lnlk1_sgl_event_cls, $fl_smpl) = cal_lnlk1_sgl_indel_cls (@{$id_event}, @{$cls_gblocks}, $br_len, $char_abs, @{$log_nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, $if_lm, $if_rm);
	    ##my ($lnlk1_sgl_event_cls, $fl_smpl) = cal_lnlk1_sgl_indel_cls (@{$id_event}, @{$cls_gblocks}, $br_len, $char_abs, @{$log_nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, $indx_lm_blk, $indx_rm_blk);
	    ## my ($lnlk1_sgl_event_cls, $fl_smpl) = cal_lnlk1_sgl_indel_cls (@{$id_event}, @{$cls_gblocks}, $br_len, $char_abs, @{$nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, $indx_lm_blk, $indx_rm_blk);
	    unless (defined $lnlk1_sgl_event_cls) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'cal_lnlk1_sgl_indel_cls' FAILED on the $sg th single indel event (= the $indx_id th among all the input events): $! !\n";
		return (undef, undef);
	    }
    (print STDOUT ' ' x 2, " ... 'cal_lnlk1_sgl_indel_cls' Ended ", '($lnlk1_sgl_event_cls = ', $lnlk1_sgl_event_cls, ") ...\n\n") if ($TEST_LEVEL1a>2);
	    if ($fl_smpl) {
		print STDOUT "[WARNING 'cal_lnlk1_indel_br_cls'] The subroutine 'cal_lnlk1_sgl_indel_cls' was SIMPLIFIED on the $sg th single indel event (= the $indx_id th among all the input events): $! !\n";
		$ct_smpl++ ;
	    }

	    $lnlk1_indel_br_cls += $lnlk1_sgl_event_cls;

    (print STDOUT ' ' x 2, 'Current $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " .\n\n") if ($TEST_LEVEL1a>3);


	}			# End of the 1st main outer for-loop (over single indel events) #

    if ($TEST_LEVEL1a>2) {
	print STDOUT ' ' x 2, 'Interim $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " ,\n";
	print STDOUT ' ' x 2, 'Interim $ct_smpl = ', $ct_smpl, " .\n\n";
    }

		## Main Process 2: calculate the contributions from paired indel events. ##

    (print STDOUT "\n", ' ' x 2, "[ Main Process 2: Contributions from $ct_pairs paired indel events ]\n\n") if ($TEST_LEVEL1a>1);

	for (my $pr=0; $pr < $ct_pairs; $pr++) { # 2nd main outer for-loop (over paired indel events) #

	    # @{$info_pair} = ($indx_ins, $indx_del, \@indices_ovlp_blk)
	    # with $indx_ins/del being the index in @indels_br,
	    # and @indices_ovlp_blk storing the indices of blocks where the two events overlap.
	    my $info_pair = $paired_indels->[$pr];
	    my ($indx_id1, $indx_id2, $indices_ovlp) = @{$info_pair};

    if ($TEST_LEVEL1a>2) {
	my $cnct_indices_ovlp = join (',', @{$indices_ovlp});
        print STDOUT ' ' x 2, "{ The $pr th pair (insertion index = $indx_id1; deletion index = $indx_id2; indices,of,overlapping,blocks =  $cnct_indices_ovlp) }\n\n";
    }
    if ($TEST_LEVEL1a>3) {
	my ($stat_bf_ins, $stat_af_ins, $indices_blks_ins) = @{$indels_br->[$indx_id1]};
	my $cnct_indices_blks_ins = join ('/', @{$indices_blks_ins});
	my ($stat_bf_del, $stat_af_del, $indices_blks_del) = @{$indels_br->[$indx_id2]};
	my $cnct_indices_blks_del = join ('/', @{$indices_blks_del});
	print STDOUT ' ' x 2, join ("\t", 'Ins/Del  ', 'stat_before', 'stat_after', 'indices/involved/blocks'), "\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Insertion', $stat_bf_ins, $stat_af_ins, $cnct_indices_blks_ins), "\n";
	print STDOUT ' ' x 2, join ("\t", 'Deletion ', $stat_bf_del, $stat_af_del, $cnct_indices_blks_del), "\n";
	print STDOUT "\n\n";
    }

		# MODIFIED accompanying the introdcution of 'search_left/rightmost_events' (on July 22 (Sun), 2012) #
	    my ($lnlk1_pr_id_cls, $fl_smpl) = cal_lnlk1_pair_indel_cls (@{$info_pair}, @{$indels_br}, @{$cls_gblocks}, $br_len, $char_abs, @{$log_nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, %{$indx_indel2if_lm}, %{$indx_indel2if_rm});
	    ##my ($lnlk1_pr_id_cls, $fl_smpl) = cal_lnlk1_pair_indel_cls (@{$info_pair}, @{$indels_br}, @{$cls_gblocks}, $br_len, $char_abs, @{$log_nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, $indx_lm_blk, $indx_rm_blk);
	    ## my ($lnlk1_pr_id_cls, $fl_smpl) = cal_lnlk1_pair_indel_cls (@{$info_pair}, @{$indels_br}, @{$cls_gblocks}, $br_len, $char_abs, @{$nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds}, $indx_lm_blk, $indx_rm_blk);
	    unless (defined $lnlk1_pr_id_cls) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_br_cls'] The subroutine 'cal_lnlk1_pair_indel_cls' FAILED on the $pr th pair of indels (= the $indx_id1 th and the $indx_id2 th among all the input events): $! !\n";
		return (undef, undef);
	    }
    (print STDOUT ' ' x 2, " ... 'cal_lnlk1_pair_indel_cls' Ended ", '($lnlk1_pr_id_cls = ', $lnlk1_pr_id_cls, ") ...\n\n") if ($TEST_LEVEL1a>2);
	    if ($fl_smpl) {
		print STDOUT "[WARNING 'cal_lnlk1_indel_br_cls'] The subroutine 'cal_lnlk1_pair_indel_cls' was SIMPLIFIED on the $pr th pair of indels (= the $indx_id1 th and the $indx_id2 th among all the input events): $! !\n";
		$ct_smpl++ ;
	    }

	    $lnlk1_indel_br_cls += $lnlk1_pr_id_cls;

    (print STDOUT ' ' x 2, 'Current $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " .\n\n") if ($TEST_LEVEL1a>3);

	}				# End of the 2nd main outer for-loop (over paired indel events) #

    if ($TEST_LEVEL1a>2) {
	print STDOUT ' ' x 2, 'Final $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " ,\n"; 
	print STDOUT ' ' x 2, 'Final $ct_smpl = ', $ct_smpl, " .\n\n";
    }

    (print STDOUT "\n", ' ' x 2, "--- Subroutine 'cal_lnlk1_indel_br_cls' Ended ---\n\n\n") if ($TEST_LEVEL1a>0);

	return ($lnlk1_indel_br_cls, $ct_smpl);

}		# End of the subroutine 'cal_lnlk1_indel_br_cls' #


# cal_lnlk1_indel_scn_cls:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa' calculates
#	the first approximation of the log-likelihood of an indel scenario
#	in a specified gap-block cluster on a tree.
#
# [ INPUT: (\@cls_gblocks, \%br2attr, \%node2pa, \%bl2indels, \%node2blwise_gstats, $char_abs, \@nfacs, \@params_ins, \@params_del, \@params_boundaries), where ] --- OBSOLETE as of July 11 (Wed), 2012.
# INPUT: (\@cls_gblocks, \%br2attr, \%node2pa, \%bl2indels, \%node2blwise_gstats, $char_abs, \@log_nfacs, \@params_ins, \@params_del, \@params_boundaries), where
#
#	@{$cls_gblocks[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event;
#
#	%node2blwise_gapstats = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#
#	[ @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p); ] --- OBSOLETE as of July 11 (Wed), 2012.
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p);
#
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	@params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end), with
#
#	$cutoff_5p/3p being the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region,
#	$opt_bd_indels = 0 if you want to ignore the indels extending beyond the alignment boundary,
#			= 1 if you want to only account for the rate of indels (but not the entire integrand) extending beyond the boundary,
#			= 2 if you want to fully account for the effect of indels extending beyond the boundary
#	(Theoretically, option '2' is the most reasonable but it takes long),
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise).
#
# OUTPUT: ($lnlk1_indel_scn_cls, \%br2lnlk1_indel_cls), where
#	$lnlk1_indel_scn_cls is the 1st approximation of the log-likelihood of the input
#	indel scenario in the specified gap-block cluster;
#	%br2lnlk1_indel_cls = ($br_id => $lnlk1_indel_br_cls, ...) is the 1st approximation of the
#	log-likelihood of the input indel scenario along the branch $br_id in the specified gap-block cluster.
#
sub cal_lnlk1_indel_scn_cls (\@\%\%\%\%$\@\@\@\@) {

	## my ($cls_gblks, $br2attr, $node2pa, $br2indels, $node2blwise_gstats, $char_abs, $nfacs, $params_ins, $params_del, $params_bds) = @_; --- OBSOLETE as of July 11 (Wed), 2012.
	my ($cls_gblks, $br2attr, $node2pa, $br2indels, $node2blwise_gstats, $char_abs, $log_nfacs, $params_ins, $params_del, $params_bds) = @_;

	my $ct_gblks = @{$cls_gblks};
	my @br_ids = sort {$a <=> $b} keys %{$br2attr};
	my $ct_branches = @br_ids;

    (print STDOUT "\n < Subroutine 'cal_lnlk1_indel_scn_cls' ('Absence' = $char_abs\; #{gap-blocks in the cluster} = $ct_gblks\; #{branches in the tree} = $ct_branches) >\n\n") if ($TEST_LEVEL1>0);

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n [ Input Parameters ]\n\n";
	print STDOUT ' ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p) = (', join (', ', @{$log_nfacs}), ");\n\n";
	print STDOUT ' ($totrate_ins, $distr_ins, $prm_ins) = (', join (', ', @{$params_ins}), "),\n";
	print STDOUT ' ($totrate_del, $distr_del, $prm_del) = (', join (', ', @{$params_del}), ");\n\n";
	print STDOUT ' ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end) = (', join (', ', @{$params_bds}), ");\n\n";

	print STDOUT ' [ Content of @{$cls_gblks} ]', "\n\n";

	print STDOUT ' ' , join ("\t", 'Indx_blk', 'indx_gpattern', 'length_blk', 'start', 'end'), "\n\n";
	for (my $g=0; $g<$ct_gblks; $g++) { print STDOUT ' ' , join ("\t", $g, @{$cls_gblks->[$g]}), "\n"; }
	print STDOUT "\n";

    }

    if ($TEST_LEVEL1>3) {
	my @nodes = sort {$a <=> $b} keys %{$node2blwise_gstats};
	my $ct_nodes = @nodes;
	print STDOUT "\n [ There are $ct_nodes nodes ]\n\n";
	print STDOUT ' ' , join ("\t", 'Node_ID', "'{gap-states on the blocks}'"), "\n\n";
	foreach my $node (@nodes) {
	    my $cnct_blwise_gstats = join ('', @{$node2blwise_gstats->{$node}});
	    print STDOUT ' ' , join ("\t", $node, "'$cnct_blwise_gstats'"), "\n";
	}
	print STDOUT "\n";

	print STDOUT "\n [  There are $ct_branches branches ]\n\n";
	print STDOUT ' ', join ("\t", 'Branch_ID', 'parent_ID', 'branch_length', '#{indels}', '[stat_ue/stat_le/indices,involved,blks]_event1', '[...]_event2', '...'), "\n\n";

	foreach my $br (@br_ids) {
	    my $br_len = $br2attr->{$br}[$INDX_LENGTH];
	    my $pa = $node2pa->{$br};
	    my $list_indels = $br2indels->{$br};
	    (defined $list_indels) or ($list_indels = []);
	    my $ct_indels = @{$list_indels};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
		my $cnct_indices_blk = join (',', @{$indices_blk});
		my $out_indel = join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		push @out_indels, "[$out_indel]";
	    }
	    print STDOUT ' ' , join ("\t", $br, $pa, $br_len, $ct_indels, @out_indels), "\n";
	}
	print STDOUT "\n";

	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
    }

		### RESTART FROM HERE ###

    (print STDOUT "\n [  Calculation of \$lnlk1_indel_scn_cls  ]\n\n") if ($TEST_LEVEL1>1);

	my $lnlk1_indel_scn_cls = 0;
	my %br2lnlk1_indel_cls;

	foreach my $br (@br_ids) { # Main foreach-loop (over branches) #
	##foreach my $br (sort keys %{$br2attr}) { # Main foreach-loop (over branches) #

	    my $attr = $br2attr->{$br};
	    my $br_len = $attr->[$INDX_LENGTH];
	    my ($node_le, $node_ue) = ($br, $node2pa->{$br});

		# Indel events along the branch #

	    my $indels_br = $br2indels->{$br};
	    my @indels_br = ();
	    if (defined $indels_br) {
			# A measure to prevent a bug in the upstream processes. #
		foreach my $event (@{$indels_br}) { (@{$event}>0) and (push @indels_br, $event); }
	    }
	    my $ct_indels = @indels_br;

    (print STDOUT "\n { Branch ID: $br (length = $br_len, upper-end = $node_ue, lower-end = $node_le, #{indel events} = $ct_indels) }\n\n") if ($TEST_LEVEL1>1);

		# Gap states on the lower-end and upper-end nodes. #

	    my $blwise_gstats_le = $node2blwise_gstats->{$node_le};
	    my $blwise_gstats_ue = $node2blwise_gstats->{$node_ue};

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n ( Gap states on the lower- and upper-ends )\n\n";
	print STDOUT ' ' , join ("\t", 'Block_ID', 'state_upper', 'state_lower'), "\n\n";
	for (my $b=0; $b<$ct_gblks; $b++) { print STDOUT ' ' , join ("\t", $b, $blwise_gstats_ue->[$b], $blwise_gstats_le->[$b]), "\n"; }

	print STDOUT "\n\n ",  "( There are $ct_indels indel events )\n\n";
	print STDOUT ' ' , join ("\t", 'Indel_indx', 'stat_bf', 'stat_af', 'indices,involved,blocks'), "\n\n";

	for (my $id=0; $id<$ct_indels; $id++) {
	    my ($stat_bf, $stat_af, $indices_involved) = @{$indels_br[$id]};
	    my $cnct_indices = join (',', @{$indices_involved});
	    print STDOUT ' ' , join ("\t", $id, $stat_bf, $stat_af, $cnct_indices), "\n";
	}
	print STDOUT "\n\n";

	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
	
    }

		# Calculate the log-likelihood of the input indel scenario along a specified branch. #

	    my ($lnlk1_indel_br_cls, $ct_smpl) = cal_lnlk1_indel_br_cls (@{$cls_gblks}, $br_len, @{$indels_br}, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs, @{$log_nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds});
	    ##my ($lnlk1_indel_br_cls, $ct_smpl) = cal_lnlk1_indel_br_cls (@{$cls_gblks}, $br_len, @{$indels_br}, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs, @{$nfacs}, @{$params_ins}, @{$params_del}, @{$params_bds});
	    unless (defined $lnlk1_indel_br_cls) {
		print STDOUT "[ERROR 'cal_lnlk1_indel_scn_cls'] The subroutine 'cal_lnlk1_indel_br_cls' FAILED along the branch $br: $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "\n", ' ' x 3, "--- Subroutine 'cal_lnlk1_indel_br_cls' Ended ...\n\n") if ($TEST_LEVEL1>1);

	    if ($ct_smpl>0) {
		print STDOUT "[WARNING 'cal_lnlk1_indel_scn_cls'] The subroutine 'cal_lnlk1_indel_br_cls' experienced $ct_smpl simplifications ...\n\n";
	    }

	    $lnlk1_indel_scn_cls += $lnlk1_indel_br_cls;
	    $br2lnlk1_indel_cls{$br} = $lnlk1_indel_br_cls;

    if ($TEST_LEVEL1>2) {
	print STDOUT "\n ( Outputs of 'cal_lnlk1_indel_scn_cls' )\n\n"; 
	print STDOUT ' $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, ",\n";
	print STDOUT ' $ct_smpl = ', $ct_smpl, " .\n\n";
	print STDOUT ' Current $lnlk1_indel_scn_cls = ', $lnlk1_indel_scn_cls, " .\n\n";
    }

	}				# End of the main foreach-loop (over branches) #

    (print STDOUT "\n", ' Final $lnlk1_indel_scn_cls = ' , $lnlk1_indel_scn_cls, " .\n\n") if ($TEST_LEVEL1>0);

    (print STDOUT "\n --- End of the subroutine 'cal_lnlk1_indel_scn_cls' ---\n\n") if ($TEST_LEVEL1>0);

	return ($lnlk1_indel_scn_cls, \%br2lnlk1_indel_cls);

}		# End of the subroutine 'cal_lnlk1_indel_scn_cls' #


# apprx1_tot_lnlk_gpattern_msa:
#	This subroutine calculates the 1st approximation of
#	the total log-likelihood of the gap-pattern of an input MSA
#	by using the set of parsimonious indel scenarios
#	that realize the gap pattern (, which are output by the subroutine
#	'enum_psm_indel_scns_for_msa' in the module 'MyTreeMap_indels.pm').
#
#	MINOR REVISION on June 10 (Tue), 2012:
#	The normalization factors, stored in @nfacs, and $delta_del_5p, 
#	both of which were calculated and output by this subroutine,
#	are now input from outside, to potentially save computational time (in many cases).
#	MINOR REVISION on June 11 (Wed), 2012:
#	The old input @nfacs has now been replaced by @log_nfacs.
#
#	REVISED on Mar 2 (Sun), 2014:
#	The subroutine continues the calculation even if the input includes an empty 
#	@{$set_psm_scns_for_cls[$cl]->[$ps]}, which indicates that there is an abnormality in this scenario,
#	just while setting the likelihood of the scenario to be 'N/A'.
#
#	REVISED on Mar 6 (Sun), 2014:
#	The subroutine continues the calculation even if its satellite subroutine, 'cal_lnlk1_indel_scn_cls',
#	returns undef, while registering the likelihood in question as 'undef'.
#
#
# [ INPUT: (\%br2attr, \%node2pa, $len_msa, $char_abs, \@set_cls_gblocks, \@sets_psm_scns_for_cls, \@params_ins, \@params_del, $cutoff_5p, $cutoff_3p, $opt_bd_indels), where ] --- OBSOLETE as of July 10 (Tue), 2012. 
# [ INPUT: (\%br2attr, \%node2pa, $len_msa, $char_abs, \@set_cls_gblocks, \@sets_psm_scns_for_cls, \@params_ins, \@params_del, $cutoff_5p, $cutoff_3p, \@nfacs, $delta_del_5p, $opt_bd_indels), where ] --- OBSOLETE as of July 11 (Wed), 2012.
# INPUT: (\%br2attr, \%node2pa, $len_msa, $char_abs, \@set_cls_gblocks, \@sets_psm_scns_for_cls, \@params_ins, \@params_del, $cutoff_5p, $cutoff_3p, \@log_nfacs, $delta_del_5p, $opt_bd_indels), where
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention.
#
#	@{$set_psm_scns_for_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#
#	NOTE that @{$set_psm_scns_for_cls[$cl]->[$ps]} = () indicates that the lieklihood of the parsimonious scenario cannot be calculated due to some abnormality. (ADDED on Mar 2 (Sun), 2014)
#
#	@params_ins/del = ($totrate_ins/del, $distr_ins/del, $prm_ins/del), with
#	$totrate_ins/del being the total rate of insertions/deletions (relative to the substitution rate),
#	$distr_ins/del = 'G' specifying the geometric distribution,
#			'P' specifying the (zeta) power-law distribution,
#	$prm_ins/del = the 'persistence' rate (i.e. q in f(l) = q**(l-1)/N(q,M)) if $distr_ins/del = 'G',
#		     = the power (i.e. a in f(l) = l**(-a)/N(a,M)) if $distr_ins/del = 'P';
#
#	$cutoff_5p/3p is the cut-off length common for the insertions and deletions on the 5'-side/3'-side of the aligned region;
#
#	[ @nfacs = ($nfac_ins_5p, $nfac_del_5p, $nfac_ins_3p, $nfac_del_3p), with
#	$nfac_ins/del_5p/3p = Sum_{$l=1}^{$cutoff_5p/3p} distr_ins/del ($l),
#	which can be calculated via 'cal_normal_facs_indel_distri'; ] --- OBSOLETE as of July 11 (Wed), 2012.
#	@log_nfacs = ($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p), with
#	$log_nfac_ins/del_5p/3p = log { Sum_{$l=1}^{$cutoff_5p/3p} distr_ins/del ($l) },
#	which can be calculated via 'cal_log_normal_facs_indel_distri';
#
#	$delta_del_5p = Sum_{$l=1}^{$cutoff_5p}($l-1)*distr_del ($l)/$nfac_del_5p,
#	which can be calculated via 'cal_delta_exp_del_5p', is the excessive exponential factor
#	due to deletions extending into the 5p end;
#
#	$opt_bd_indels = 0 if you want to ignore the indels extending beyond the alignment boundary,
#			= 1 if you want to only account for the rate of indels (but not the entire integrand) extending beyond the boundary,
#			= 2 if you want to fully account for the effect of indels extending beyond the boundary.
#	(Theoretically, option '2' is the most reasonable but it takes long.)
#
#
# [OUTPUT: ($tot_lnlk_gp, $shared_fac1, $shared_fac2, \@set_clwise_tot_lnlk_gp, $tot_div_tree, $delta_del_5p, $totlen_ng, \@nfacs, \@setset_lnlk_gp_cl_scn), where ] --- Obsolete as of June 10 (Tue), 2012.
# OUTPUT: ($tot_lnlk_gp, $shared_fac1, $shared_fac2, \@set_clwise_tot_lnlk_gp, $tot_div_tree, $totlen_ng, \@setset_lnlk_gp_cl_scn), where
#
#	$tot_lnlk_gp is the total log-likelihood of the gap-pattern in the 1st approximation
#	(= 'N/A' if any of the parsimonious scenarios contain such abnormalities that hamper the calculation of the likelihoods ... ADDED on Mar 2 (Sun), 2014);
#
#	$shared_fac1 = - $tot_div_tree * ($totrate_del * $delta_del_5p + $totrate_ins);
#	$shared_fac2 = - ($totrate_ins+$totrate_del) * $tot_div_tree * $totlen_ng;
#	$set_clwise_tot_lnlk_gp[$i] is the total log-likelihood of the gap pattern in the ($i+1) th gap-block cluster, which is log (Sum_{$m} exp ($setset_lnlk_gp_cl_scn[$i]->[$m]))
#	(= 'N/A' if any of the parsimonious scenarios in the cluster contain such abnormalities that hamper the calculation of the likelihoods ... ADDED on Mar 2 (Sun), 2014);
#
#	$tot_div_tree is the total divergence of the tree, which is the summation of the branch lengths over the entire tree;
#
#	$setset_lnlk_gp_cl_scn[$i]->[$m] is the log-likelihood of the $m th indel scenario
#	that parsimoniously give the ($i+1) th gap-block cluster
#	(= 'N/A' if the parsimonious scenario contains such abnormalities that hamper the calculation of the likelihood ... ADDED on Mar 2 (Sun), 2014).
#
# [ sub apprx1_tot_lnlk_gpattern_msa (\%\%$$\@\@\@\@$$$) {
#	my ($br2attr, $node2pa, $len_msa, $char_abs, $set_cls_gblocks, $sets_psm_scns_for_cls,
#		$params_ins, $params_del, $cutoff_5p, $cutoff_3p, $opt_bd_indels) = @_; ] --- OBSOLETE as of June 10 (Tue), 2012.
#
sub apprx1_tot_lnlk_gpattern_msa (\%\%$$\@\@\@\@$$\@$$) {

	##my ($br2attr, $node2pa, $len_msa, $char_abs, $set_cls_gblocks, $sets_psm_scns_for_cls, $params_ins, $params_del, $cutoff_5p, $cutoff_3p, $nfacs, $delta_del_5p, $opt_bd_indels) = @_; --- OBSOLETE as of July 11 (Wed), 2012.

	my ($br2attr, $node2pa, $len_msa, $char_abs, $set_cls_gblocks, $sets_psm_scns_for_cls,
		$params_ins, $params_del, $cutoff_5p, $cutoff_3p, $log_nfacs, $delta_del_5p, $opt_bd_indels) = @_;

	my $ct_cls = @{$set_cls_gblocks};
	my ($totrate_ins, $distr_ins, $prm_ins) = @{$params_ins};
	my ($totrate_del, $distr_del, $prm_del) = @{$params_del};

	my $cutoff_min = ($cutoff_3p < $cutoff_5p) ? $cutoff_3p : $cutoff_5p; # ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014. #

    (print STDOUT "\n<< Subroutine 'apprx1_tot_lnlk_gpattern_msa' ('Absence' = $char_abs, #{gap-block clusters} = $ct_cls, length(MSA) = $len_msa) >>\n\n") if ($TEST_LEVEL>0);

    if ($TEST_LEVEL>2) {

	print STDOUT "[[ Primary Input Data ]]\n\n";
	print STDOUT '($cutoff_5p, $cutoff_3p, $opt_bd_indels) = ', "($cutoff_5p, $cutoff_3p, $opt_bd_indels) .\n\n";
	print STDOUT '($totrate_ins, $distr_ins, $prm_ins) = (', join (', ', @{$params_ins}), ") ,\n";
	print STDOUT '($totrate_del, $distr_del, $prm_del) = (', join (', ', @{$params_del}), ") .\n\n";
	print STDOUT '($log_nfac_ins_5p, $log_nfac_del_5p, $log_nfac_ins_3p, $log_nfac_del_3p)', "\n";
	print STDOUT ' = (', join (', ', @{$log_nfacs}), ") .\n\n";
	print STDOUT '$delta_del_5p = ', $delta_del_5p, " .\n";
	print STDOUT "\n\n";
    }

    if ($TEST_LEVEL>3) {
	print STDOUT "[ Input tree ]\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'parent_ID', 'branch_length'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2attr}) {
	     my $len = $br2attr->{$br}[$INDX_LENGTH];
	     my $pa_id = $node2pa->{$br};
	     print STDOUT join ("\t", $br, $pa_id, $len), "\n";
	}
	print STDOUT "\n\n";
	print STDOUT "[ There are $ct_cls gap-block clusters ]\n\n";
	for (my $cl=0; $cl<$ct_cls; $cl++) {

	    my $cls_gblks = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$cls_gblks};

	    print STDOUT "{ The $cl th cluster (#{blocks} = $ct_gblks) }\n\n";
	    print STDOUT join ("\t", 'Indx_blk', 'indx_gpattern', 'length_blk', 'start', 'end'), "\n\n";

	    for (my $bl=0; $bl<$ct_gblks; $bl++) { print STDOUT join ("\t", $bl, @{$cls_gblks->[$bl]}), "\n"; }
	    print STDOUT "\n\n";
	}

	# @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is
    }

		### RESTART FROM HERE ###

	if ( (($distr_ins eq 'G') and  ($prm_ins == 0)) 
	    or (($distr_del eq 'G') and ($prm_del == 0)) ) {
	     print STDOUT "[WARNING 'apprx1_tot_lnlk_gpattern_msa'] Because the length distribution (for insertions or deletions) is geometric with ZERO persistence, I will set the length cut-offs, \$cutoff_5p = $cutoff_5p and \$cutoff_3p = $cutoff_3p, to 1 ...\n\n";
	    $cutoff_3p = $cutoff_5p = 1;
	}

	##my @nfacs = @{$nfacs};
	my @log_nfacs = @{$log_nfacs};

		## Prepare auxiliary variables. ##

    if ($TEST_LEVEL>3) {
	print STDOUT '[ Calculating $tot_div_tree ]', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'br_len', '$tot_div_tree(current)'), "\n\n";
    }

	my $tot_div_tree = 0;
	foreach my $br (keys %{$br2attr}) { # 1st preliminary foreach-loop #
	    my $attr = $br2attr->{$br};
	    unless (defined $attr) {
		print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa'] The branch '$br' has NO ATTRIBUTE : $! !\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    my $br_len = $attr->[$INDX_LENGTH];
	    $tot_div_tree += $br_len;

    (print STDOUT join ("\t", $br, $br_len, $tot_div_tree), "\n") if ($TEST_LEVEL>3);

	}			# End of the 1st preliminary foreach-loop #

    if ($TEST_LEVEL>3) {
	print STDOUT "\n\n";
	print STDOUT '[ Calculating $totct_gpsites ]', "\n\n";
	print STDOUT join ("\t", 'Cluster_ID', 'beg_cls', 'end_cls', 'len_cls', '$totct_gpsites(current)'), "\n\n";
    }

	my $totct_gpsites = 0;
	for (my $cl=0; $cl<$ct_cls; $cl++) { # 2nd preliminary for-loop #
	    #  @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
	    my $cls_gb = $set_cls_gblocks->[$cl];
	    my $first_gb = $cls_gb->[0];
	    my $last_gb = $cls_gb->[$#{$cls_gb}];
	    my $beg_cls = $first_gb->[2];
	    my $end_cls = $last_gb->[3];
	    my $len_cls = $end_cls - $beg_cls + 1;
	    $totct_gpsites += $len_cls;

    (print STDOUT join ("\t", $cl, $beg_cls, $end_cls, $len_cls, $totct_gpsites), "\n") if ($TEST_LEVEL>3);

	}			# End of the 2nd preliminary for-loop #

    (print STDOUT "\n\n") if ($TEST_LEVEL>3);

	my $totlen_ng = $len_msa - $totct_gpsites;

	my $shared_fac1 = - $tot_div_tree * ($totrate_del * $delta_del_5p + $totrate_ins);
	my $shared_fac2 = - ($totrate_ins+$totrate_del) * $tot_div_tree * $totlen_ng;

    if ($TEST_LEVEL>2) {
	print STDOUT "[[ Secondary Input Data ]]\n\n";
	print STDOUT '$tot_div_tree = ', $tot_div_tree, " .\n\n";
	print STDOUT '$len_msa = ', $len_msa, " ,\n";
	print STDOUT '$totct_gpsites = ', $totct_gpsites, " ,\n";
	print STDOUT '$totlen_ng = ', $totlen_ng, " .\n\n";
	print STDOUT '$shared_fac1 (= - $tot_div_tree * ($totrate_del * $delta_del_5p + $totrate_ins) = ', $shared_fac1, " ,\n";
	print STDOUT '$shared_fac2 (= - ($totrate_ins + $totrate_del) * $tot_div_tree * $totlen_ng) = ' , $shared_fac2, " .\n\n";
    }

	   ## Main process to calculate log-likelihood for each gap-block cluster. ##

    (print STDOUT "\n[[ Main Process ]]\n\n") if ($TEST_LEVEL>1);

	my $tot_lnlk_gp = $shared_fac1 + $shared_fac2;
	my @set_clwise_tot_lnlk_gp = my @setset_lnlk_gp_cl_scn = ();

    (print STDOUT 'Initial $tot_lnlk_gp (= $shared_fac1 + $shared_fac2) = ', $tot_lnlk_gp, " .\n\n") if ($TEST_LEVEL>2);

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over gap-block clusters) #

	    #  @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)

	    my $cls_gb = $set_cls_gblocks->[$cl];
	    my $ct_gblocks = @{$cls_gb};

    (print STDOUT "\n[ The $cl th cluster (#{gap-blocks} = $ct_gblocks) ]\n\n") if ($TEST_LEVEL>1);

	    my $first_gb = $cls_gb->[0];
	    my $last_gb = $cls_gb->[$ct_gblocks-1];
	    my $beg_cls = $first_gb->[2];
	    my $end_cls = $last_gb->[3];

	    my $if_5p_end = ($beg_cls == 0) ? 1 : 0;
	    my $if_3p_end = ($end_cls == $len_msa-1) ? 1 : 0;

	    my @params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end);

	    my $psm_scns_in_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_in_cls};

	    ##my $clwise_tot_lnlk1_gp = 0;
	    my @set_lnlk_gp_scns = ();

    if ($TEST_LEVEL>2) {
	print STDOUT '($beg_cls, $end_cls) = ', "($beg_cls, $end_cls) .\n\n";
	print STDOUT '($if_5p_end, $if_3p_end) = ', "($if_5p_end, $if_3p_end) .\n\n";
	print STDOUT '#{parsimonious scenarios} = ', $ct_psm_scns, " .\n\n";
    }


		## Handle an exceptional case, which probably contains some long gaps. ##
		## (ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) ##
		## (=> Further MODIFIED accompanying the further revision of 'MyTreeMap_indels.pm' on Mar 11 (Tue), 2014.)

	    my $len_cls = $end_cls - $beg_cls + 1;
	    unless ( !( ($ct_psm_scns == 0) and (($ct_gblocks> 1) or ($len_cls > $UPPER_BOUND_ANAL_GAPLEN) or ($len_cls > $UPPER_BOUND_ANAL_LEN_GBCLS) ) ) ) {
	    ##unless ( !( ($ct_psm_scns == 0) and (($ct_gblocks> 1) or ($len_cls > $UPPER_BOUND_ANAL_GAPLEN) ) ) ) {...}

		print STDOUT "[CAUTION 'apprx1_tot_lnlk_gpattern_msa'] The $cl th cluster (#{gap-blocks} = $ct_gblocks \; region = \[$beg_cls\, $end_cls\] \; length = $len_cls) has an EMPTY SET of parsimonious scenarios, probably because it contains (a) long gap(s): $! !\n\n";
		$tot_lnlk_gp = 'N/A';
		$set_clwise_tot_lnlk_gp[$cl] = 'N/A';
		$setset_lnlk_gp_cl_scn[$cl] = [];

		next;
	    }


	    my $ct_anom_psns0 = my $ct_anom_psns = 0;

	    for (my $ps =0; $ps<$ct_psm_scns; $ps++) {  # Main inner for-loop (over parsimonious indel scenarios) #

		unless (@{$psm_scns_in_cls->[$ps]}>0) {
		    # The scenario information is empty. #
		    # (ADDED in revision on Mar 2 (Sun), 2014). #

		    $ct_anom_psns++;
		    $set_lnlk_gp_scns[$ps] = 'N/A';
		    next;
		}

		my ($br2indels, $node2blwise_gstats) = @{$psm_scns_in_cls->[$ps]};

    (print STDOUT "{ The $ps th indel scenario }\n\n") if ($TEST_LEVEL>2);

		my ($lnlk1_indel_scn_cls, $br2lnlk1_indel_cls) = cal_lnlk1_indel_scn_cls (@{$cls_gb}, %{$br2attr}, %{$node2pa}, %{$br2indels}, %{$node2blwise_gstats}, $char_abs, @log_nfacs, @{$params_ins}, @{$params_del}, @params_boundaries);
		## my ($lnlk1_indel_scn_cls, $br2lnlk1_indel_cls) = cal_lnlk1_indel_scn_cls (@{$cls_gb}, %{$br2attr}, %{$node2pa}, %{$br2indels}, %{$node2blwise_gstats}, $char_abs, @nfacs, @{$params_ins}, @{$params_del}, @params_boundaries); --- OBSOLETE as of July 11 (Wed), 2012.
		unless (defined $lnlk1_indel_scn_cls) {

			# MODIFIED on Mar 6 (Thu), 2014. # 
		    print STDOUT "[CAUTION 'apprx1_tot_lnlk_gpattern_msa'] The subroutine 'cal_lnlk1_indel_scn_cls' FAILED on the $ps th parsimonious indel scenario in the $cl th cluster: $! !\n\n";
		    $ct_anom_psns0++;
		    $set_lnlk_gp_scns[$ps] = 'undef';
		    next;
		    ##print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa'] The subroutine 'cal_lnlk1_indel_scn_cls' FAILED on the $ps th parsimonious indel scenario in the $cl th cluster: $! !\n";
		    ##return (undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

    (print STDOUT " --- The subroutine 'cal_lnlk1_indel_scn_cls' ENDED (with ", '$lnlk1_indel_scn_cls = ', $lnlk1_indel_scn_cls, ") ...\n\n") if ($TEST_LEVEL>1);

		##$clwise_tot_lnlk1_gp += $lnlk1_indel_scn_cls;
		$set_lnlk_gp_scns[$ps] = $lnlk1_indel_scn_cls;

	    }		# End of the main inner for-loop (over parsimonious indel scenarios) #

	    	# FURTHER MODIFIED in revision on Mar 6 (Thu), 2014. #
	    my $clwise_tot_lnlk1_gp = ($ct_anom_psns>0) ? 'N/A' : (($ct_anom_psns0>0) ? 'undef' : (($ct_psm_scns == 1) ? $set_lnlk_gp_scns[0] : log_sum_exp (@set_lnlk_gp_scns))) ;
	    ##	# MODIFIED in revision on Mar 2 (Sun), 2014. #
	    ##my $clwise_tot_lnlk1_gp = ($ct_anom_psns>0) ? 'N/A' : (($ct_psm_scns == 1) ? $set_lnlk_gp_scns[0] : log_sum_exp (@set_lnlk_gp_scns)) ;
	    ####my $clwise_tot_lnlk1_gp = ($ct_psm_scns == 1) ? $set_lnlk_gp_scns[0] : log_sum_exp (@set_lnlk_gp_scns) ;

	    	# FURTHER MODIFIED in revision on Mar 6 (Thu), 2014. #
	    if ($ct_anom_psns>0) {
		$tot_lnlk_gp = 'N/A';
	    } elsif ($ct_anom_psns0>0) {
		$tot_lnlk_gp = 'undef';
	    } else {
		(($tot_lnlk_gp eq 'N/A') or ($tot_lnlk_gp eq 'undef')) or ($tot_lnlk_gp += $clwise_tot_lnlk1_gp);
		##$tot_lnlk_gp += $clwise_tot_lnlk1_gp; # A BUG fixed on Mar 6 (Thu), 2014. #
	    }
	    ##	# MODIFIED in revision on Mar 2 (Sun), 2014. #
	    ##unless ($tot_lnlk_gp eq 'N/A') {
	    ##	($ct_anom_psns >0 ) ? ($tot_lnlk_gp = 'N/A') : ($tot_lnlk_gp += $clwise_tot_lnlk1_gp);
	    ##}
	    ####$tot_lnlk_gp += $clwise_tot_lnlk1_gp;
	    $set_clwise_tot_lnlk_gp[$cl] = $clwise_tot_lnlk1_gp;
	    $setset_lnlk_gp_cl_scn[$cl] = \@set_lnlk_gp_scns;

    if ($TEST_LEVEL>2) {
	print STDOUT "\n{ Final results for the $cl th cluster }\n\n";
	print STDOUT '$clwise_tot_lnlk1_gp = ', $clwise_tot_lnlk1_gp, " ,\n";
	print STDOUT '$tot_lnlk_gp (current) = ', $tot_lnlk_gp, " .\n\n";
    }

	}			# End of the main outer for-loop (over gap-block clusters) #

    if ($TEST_LEVEL>2) {
	print STDOUT "\n[ Final Result of the Main Process ]\n\n";
	print STDOUT '$tot_lnlk1_gp = ', $tot_lnlk_gp, " .\n\n";
    }

    (print STDOUT "\n=== END of the Subroutine 'apprx1_tot_lnlk_gpattern_msa' ===\n\n") if ($TEST_LEVEL>0);

	return ($tot_lnlk_gp, $shared_fac1, $shared_fac2, \@set_clwise_tot_lnlk_gp, $tot_div_tree, $totlen_ng, \@setset_lnlk_gp_cl_scn);

	## [ return ($tot_lnlk_gp, $shared_fac1, $shared_fac2, \@set_clwise_tot_lnlk_gp, $tot_div_tree, $delta_del_5p, $totlen_ng, \@nfacs, \@setset_lnlk_gp_cl_scn); ] --- OBSOLETE as of June 10 (Tue), 2012. 

}	# End of the subroutine 'apprx1_tot_lnlk_gpattern_msa' #


	## my ($br2attr, $node2pa, $len_msa, $char_abs, $set_cls_gblocks, $sets_psm_scns_for_cls,
	##	$params_ins, $params_del, $cutoff_5p, $cutoff_3p, $opt_bd_indels) = @_;

# INPUT: (\%br2attr, \%node2pa, $len_msa, $char_abs, \@set_cls_gblocks, \@sets_psm_scns_for_cls, \@params_ins, \@params_del, $cutoff_5p, $cutoff_3p, $opt_bd_indels), where

#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


