#!/usr/bin/perl
#
# extract_shiftshift.pl:
#	This Perl script reads in a table of errors and only extracts the error pairs of the "Shift + Shift" type.
#	[CREATED on May 16 (Mon), 2016.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( compare_arrays );

sub extract_seg_attributes (\$); # MODIFIED in creation of 'tabulate_classify_msa_errors.pl'. #
##sub extract_seg_scores_ref_rec (\$);

$| = 1; # Flush the buffer everytime it receives information to be written through the standard out.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


our @IN_LABELS = ( 'Types;of;associated;errors' );
## our @IN_LABELS = ( '#{Equiv_indels}', 'Tot#{composite-blocks}', '#{cblks}_skipped', '#{cblks}_no_assoc', '#{cblks}_noneff', '#{cblks}_eff', 'Tot_len{cblks}_eff', 'Types;of;associated;errors', 'Lengths,of,effective,composite,blocks'); # IMPORTED FROM 'tabulate_classify_msa_errors.pl'. #


	### Input/Output files/directories ###

##my $input_msa_type = '12primates';
##my $input_msa_type = '15mammals';
my $input_msa_type = '9fe_mammals';


my $indir = "../$input_msa_type";

my $infile_name = 'tbl_classify_msa_errors.2.txt.gz';

my $infile = "$indir/$infile_name";


my $outdir = $input_msa_type;
my $outfile_name = 'tbl_errsegs_w_shiftshift.txt';
my $outfile = "$outdir/$outfile_name";


	### Check the input/output files/directories ###

unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT find a readable input file '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDERR "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	die "$! .\n\n";
    }
}


print STDOUT "\n<< Program name: $program_name (path: $program_path) >>\n\n";

print STDOUT "Input MSA type: $input_msa_type .\n\n";

print STDOUT "Input file: $infile .\n\n";

print STDOUT "Output file: $outfile .\n\n\n";


		### Main Process ###

$/ = "\n"; # Use the newline character as the input-record-separator.

my $fh_in = open_input_file2 ($infile);
unless (defined $fh_in) {
	print STDOUT "[ERROR '$program_name'] 'open_input_file2' FAILED on the input file '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $in_header = <$fh_in>;
my $indices_fld = tfm_labels2indices_gen (@IN_LABELS, $in_header, "\t");
unless (defined $indices_fld) {
	$in_header =~ s/\s+$//; 
	print STDOUT "[ERROR '$program_name'] 'tfm_labels2indices_gen' FAILED on the input header:\n";
	print STDOUT "[ERROR '$program_name'] '$in_header' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
my $max_indx = 0;
foreach my $indx (@{$indices_fld}) { ($max_indx <$indx) and ($max_indx = $indx); }

my $fh_out = open_output_file ($outfile, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out $in_header;
print $fh_out "\n";

my $ct_entries = my $ct_relv = my $ct_irrelv = 0;
my $ct_anom = 0;

while (my $inline = <$fh_in>) { # Main while-loop (over the entries (=erroneous segments)). #

	($inline =~ m/^\S/) or next; # Skip blank lines. #

	my @fields = split /\s+/, $inline;

	unless (@fields > $max_indx) {
	    $inline =~ s/\s+$//;
	    print STDOUT "[WARNING '$program_name'] Found an anomalous input line: $! !\n";
	    print STDOUT "[WARNING '$program_name'] '$inline' .\n\n";
	    $ct_anom++;
	    next;
	}

	$ct_entries++;

	my ($types) = @fields[@{$indices_fld}];

	my @types = split /\;/, $types;

	my $if_shiftshift = 0;

##=For extracting nearly all segments each containing a 'Shift+Shift'.
	foreach my $type (@types) {
	    ($type =~ m/^Paired/) or next;
	    ($type =~ m/\[\s*Shift/) or next;
	    ($type =~ m/\-\>\s*Shift/) or next;
	    $if_shiftshift = 1;
	    last;
	}

	unless ($if_shiftshift) {
	    $ct_irrelv++;
	    next;
	}
##=cut

=For extracting only segments each containing a 'Shift + Shift' alone
	unless ((@types==2) and ($types[0] =~ m/^Paired/)) {
	    $ct_irrelv++;
	    next;
	}

	unless (($types[0] =~ m/\[\s*Shift/) and ($types[0] =~ m/\-\>\s*Shift/)) {
	##unless (($types[0] =~ m/\[Shift/) and ($types[0] =~ m/\-\>Shift/)) {...}
	    $ct_irrelv++;
	    next;
	}
=cut

	$ct_relv++;

	print $fh_out $inline;

}			# End of the main while-loop (over the entries (=erroneous segments)). #
close ($fh_in);
close ($fh_out);

print STDOUT "#{Anomalous lines} = $ct_anom .\n\n";
print STDOUT "#{Entries} = $ct_entries .\n";
print STDOUT "    #{Relevant}   = $ct_relv ,\n";
print STDOUT "    #{Irrelevant} = $ct_irrelv .\n\n";




