#!/usr/bin/perl
#
# test_cmpt_log_total_qnt_in_multidim_storage.pl:
#	This script examines the behaviors of the subroutines,
#	'readin_coord_frame', 'readin_sgl_qnt_on_multidim_space',
#	'find_maxvalue_in_multidim_storage', & 'cmpt_log_total_qnt_in_multidim_storage',
#	in the Perl package, "MyANEX_IO2.pm".
#	The script was created by taking advantage of the backbone of "test_nonrec_mlt_shifts_noninf_gapblocks.pl".
#	[CREATION STARTED on 2020/04/16, ENDED on 2020/04/17.]
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	);

use MyANEX_Main qw(
	printout_multidim_storage
	);
	# nonrec_mlt_shifts_noninf_gapblocks

use MyANEX_IO qw(
	print_out_coord_frame
	);

use MyANEX_IO2 qw(
	readin_coord_frame
	readin_sgl_qnt_on_multidim_space
	find_maxvalue_in_multidim_storage
	cmpt_log_total_qnt_in_multidim_storage
	);


$| = 1; # Flush the input buffer as soon as it receives data.

my ($plscript_name) = ($0 =~ m!([^/]+)$!);

my $program_name = $plscript_name;

	### Glogal parameter setting ###

##our $STRING_UNDEF = 'undef'; # Added on 2019/03/04. #
our $STRING_UNDEF = 'n/a'; # Added on 2019/03/04. #
##our $STRING_UNDEF = 'N/A'; # Added on 2019/03/04. #


our $INDX_DGN = 0; # Index (for each terminal element of  @multidim_storage) for the degree of degeneracy.
our $INDX_LQNT = 1; # Index (for each terminal element of  @multidim_storage) for the log-quantity.


	# This specifies the level of tests hard-wired to this script.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

#our $TEST_LEVEL = 0;
#our $TEST_LEVEL = 1;
our $TEST_LEVEL = 5;
	
	# For 'find_maxvalue_in_multidim_storage'. #

$MyANEX_IO2::TEST_LEVEL_fnd_mxv_mdms = 0;
##$MyANEX_IO2::TEST_LEVEL_fnd_mxv_mdms = 5;

	# For 'cmpt_log_total_qnt_in_multidim_storage'. #

#$MyANEX_IO2::TEST_LEVEL_cmpt_ltot_mdms = 0;
$MyANEX_IO2::TEST_LEVEL_cmpt_ltot_mdms = 5;

	# For 'printout_multidim_storage'. #

$MyANEX_Main::TEST_LEVEL_po_md_strg = 0;
##$MyANEX_Main::TEST_LEVEL_po_md_strg = 6;


	### Input/Output files/directories. ###


##my $indir = 'Inputs/Example1';
##my $outdir = 'Outputs/Example1';
my $indir = 'Inputs/Example2';
my $outdir = 'Outputs/Example2';


my $fname_cf = 'coord_frame.txt';
my $fname_dgn = 'degrees_of_degeneracy.txt';
my $fname_lqnt = 'log_quantities.txt';

my $infile_cf = "$indir/$fname_cf";
my $infile_dgn = "$indir/$fname_dgn";
my $infile_lqnt = "$indir/$fname_lqnt";

my $outfile_cf = "$outdir/$fname_cf";
my $outfile_dgn = "$outdir/$fname_dgn";
my $outfile_lqnt = "$outdir/$fname_lqnt";



	### Check the Input/Output files/directories. ###

unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT DIRECTORY, '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}

foreach my $infile ($infile_cf, $infile_dgn, $infile_lqnt) {
    unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE, '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
    }
}

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
        print STDOUT "[ERROR '$program_name'] FAILED to create a TOP OUTPUT DIRECTORY '$outdir': $? !\n";
        print STDERR "[ERROR '$program_name'] $!\n\n";
        die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
        print STDOUT "[ERROR '$program_name'] FAILED to CHANGE the permission MODES of the top output directory '$outdir': $? !\n";
        print STDERR "[ERROR '$program_name'] $!\n\n";
        die ;
    }
}



	### Declare the beginning of the script ###

print STDOUT "\n<<<< Program name: $plscript_name (path: $0) >>>>\n\n";

print STDOUT "<< Input/Output Directories >>\n\n";

print STDOUT "Input Files:\n";

print STDOUT "  (For the coordinate frame) $infile_cf ,\n";
print STDOUT "  (For degrees of degeneracy) $infile_dgn ,\n";
print STDOUT "  (For log-quantities) $infile_lqnt .\n\n";

print STDOUT "Output Files:\n";

print STDOUT "  (For the coordinate frame) $outfile_cf ,\n";
print STDOUT "  (For degrees of degeneracy) $outfile_dgn ,\n";
print STDOUT "  (For log-quantities) $outfile_lqnt .\n\n\n";


print STDOUT "<< Glogal Parameters >>\n\n";

print STDOUT '$STRING_UNDEF = ', $STRING_UNDEF, " .\n\n"; # Added on 2019/03/04. #

print STDOUT '($INDX_DGN, $INDX_LQNT) = ', "($INDX_DGN, $INDX_LQNT) .\n\n\n";


print STDOUT '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyANEX_IO2::TEST_LEVEL_fnd_mxv_mdms = ', $MyANEX_IO2::TEST_LEVEL_fnd_mxv_mdms , " .\n\n";

print STDOUT '$MyANEX_IO2::TEST_LEVEL_cmpt_ltot_mdms = ', $MyANEX_IO2::TEST_LEVEL_cmpt_ltot_mdms, " .\n\n";

print STDOUT '$MyANEX_Main::TEST_LEVEL_po_md_strg = ', $MyANEX_Main::TEST_LEVEL_po_md_strg, " .\n\n";
print STDOUT "\n";


print STDOUT "\n<<<  (Pre 1) Read in the Coordinate Frame from the input file, \"$indir/$fname_cf\" ... >>\n\n\n";

my ($bds_gb_coords_raw, $orgs_gb_coords_offset, $bds_gb_coords_offset)
   = readin_coord_frame ($infile_cf);
unless (defined $bds_gb_coords_offset) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_coord_frame', FAILED on the input file, \"$infile_cf\": $? !\n";
        print STDERR "[ERROR '$program_name'] $!\n\n";
        die ;
}

my $ct_gbs = @{$bds_gb_coords_raw};


if ($TEST_LEVEL>1) { # (Pre 1) #

	print STDOUT "\n[ Outputs of 'readin_coord_frame' (#{gap-blocks} = $ct_gbs): ]\n\n";

	print STDOUT join ("\t", 'Indx_gb', 'lbd_raw', 'rbd_raw', 'org_offset', 'lbd_offset', 'rbd_offset'), "\n\n";

	for (my $gb=0; $gb < $ct_gbs; $gb++) { # 1st testing for-loop (over gap-blocks). #

	    print STDOUT join ("\t", $gb, @{$bds_gb_coords_raw->[$gb]}, $orgs_gb_coords_offset->[$gb], @{$bds_gb_coords_offset->[$gb]}), "\n";

	} # End of the 1st testing for-loop (over gap-blocks). #
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL>1) { # (Pre 1) # ...}"


print STDOUT "\n<<< (Pre Test 1) Print the Coordinate Frame into the output file \"$outdir/$fname_cf\" ... >>>\n\n";

unless ( print_out_coord_frame ($outdir, $fname_cf, @{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'print_out_coord_frame', FAILED on the output file, \"$outdir/$fname_cf\": $? !\n";
        print STDERR "[ERROR '$program_name'] $!\n\n";
        die ;
}


my @multidim_storage = ();

foreach my $info ([2, 'Degree of Degeneracy', $fname_dgn, $INDX_DGN, 1],
		  [3, 'Log-Quantity', $fname_lqnt, $INDX_LQNT, 1],
		) { # Preliminary Outer foreach-loop. #

	my ($num, $name_sbj, $fname_sbj, $indx_sbj, $if_skip_1st_header) = @{$info};

	my $infile_sbj = "$indir/$fname_sbj";
	my $outfile_sbj = "$outdir/$fname_sbj";


	print STDOUT "\n<<<  (Pre $num) Read in the values of '$name_sbj' at the points in ${ct_gbs}-Dimensional Lattice from the input file, \"$infile_sbj\" ... >>\n\n\n";

	my $fh_in = open_input_file2 ($infile_sbj);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAINED on the input file, \"$infile_sbj\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	unless ( readin_sgl_qnt_on_multidim_space (*${$fh_in}, @{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}, @multidim_storage, $indx_sbj, $STRING_UNDEF, $if_skip_1st_header) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_sgl_qnt_on_multidim_space', FAILED on the file-handle to the input file, \"$infile_sbj\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}


	print STDOUT "\n<<<  (Pre Test $num) Print the values of '$name_sbj' at the ${ct_gbs}-Dimensional Lattice-points into the output file, \"$outfile_sbj\" ... >>\n\n\n";

	my $fh_out = open_output_file ($outfile_sbj, 0);
	unless (defined $fh_out) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'open_output_file', FAINED on the output file, \"$outfile_sbj\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	unless ( printout_multidim_storage (@{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}, @multidim_storage, $indx_sbj, *${$fh_out}, $STRING_UNDEF) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'printout_multidim_storage', FAILED on the file-handle to the output file, \"$outfile_sbj\": $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

} # End of the Preliminary Outer foreach-loop. #



print STDOUT "\n<<<<< (MAIN 1) Find the Maximum of 'Log-Quantity' among Values at the ${ct_gbs}-Dimensional Lattice-Points ... >>>>>\n\n\n";

##my ($fl_end1, $max_lqnt) = find_maxvalue_in_multidim_storage (@{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}, @multidim_storage, $INDX_LQNT, $STRING_UNDEF);
my ($fl_end1, $max_lqnt) = find_maxvalue_in_multidim_storage (@{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}, @multidim_storage, $INDX_LQNT, $INDX_DGN, $STRING_UNDEF); # REVISED in accompany with the revision of "MyANEX_IO2.pm" (on 2020/04/18). #
unless ($fl_end1) {
    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_LQNT th element of the terminal arrays in the ", '@multidim_storage', ": $? !\n";
    print STDERR "[ERROR '$program_name'] $!\n\n";
    die ;
}

print STDOUT "\n[ Output of 'find_maxvalue_in_multidim_storage': ]\n\n";

my $out_max_lqnt = (defined $max_lqnt) ? $max_lqnt : "'UNDEF'";

print STDOUT '$max_lqnt = ', "$out_max_lqnt .\n\n\n";


print STDOUT "\n<<<<< (MAIN 2) Compute the Total Value of exp ('Log-Quantity') among those at the ${ct_gbs}-Dimensional Lattice-Points ... >>>>>\n\n\n";

my ($fl_end2, $log_totalval_qnt, $total_wgts)
    = cmpt_log_total_qnt_in_multidim_storage (@{$orgs_gb_coords_offset}, @{$bds_gb_coords_offset}, @multidim_storage, $INDX_LQNT, $INDX_DGN, $STRING_UNDEF, $max_lqnt);
unless ($fl_end2) {
    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_LQNT th element of each terminal array in the ", '@multidim_storage', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array: $? !\n";
    print STDERR "[ERROR '$program_name'] $!\n\n";
    die ;
}


print STDOUT "\n[ Outputs of 'cmpt_log_total_qnt_in_multidim_storage': ]\n\n";

my $out_log_totalval_qnt = (defined $log_totalval_qnt) ? $log_totalval_qnt : "'UNDEF'" ;
my $out_total_wgts = (defined $total_wgts) ? $total_wgts : "'UNDEF'" ;

print STDOUT '$log_totalval_qnt = ', "$out_log_totalval_qnt .\n\n";
print STDOUT '$total_wgts = ', "$out_total_wgts .\n\n\n";
 

print STDOUT "\n\n === The script, '$program_name' ENDED !! ===\n\n\n";

		### RESTART FROM HERE. ###






