###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#	Version 0.7: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# anex.ver0.7_reduced.pl:
#	This script was created from "anex.ver0.7.pl" by REMOVING the for-loop for the 1st sliding-window analyses,
#	to test the 2nd sliding-window analyses (for reverse-purges).
#	[CREATED on 2020/05/21.]
#
# anex.ver0.7.pl:
#	This script is the UPGRADED version of 'anex.ver0.5.pl'.
#	The following MAJOR REVISIONS were made:
#
#	(1) The script can now take full advantage of the main outputs of 'lastpiece.xxx.pl' in the program package, "LASTPIECE_P.verNNN",
#	and thus the computed MSA probabilities can now be MUCH MORE ACCURATE than computed by the predecessor (ver 0.5);
#	To enable this revision,
#	  (1-i) the block for checking files of multiplication factors can now accept symbolic links as well, in addition to normal files as the predecessor (ver 0.5) does;
#
#	  (1-ii) the subroutines, 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm") (& their satellite subroutines) & 'compt_log_msa_prob_plus' & 'nonrec_mlt_shifts_gapblocks' (the latter two are in "MyANEX_Main.pm"), were substantially REVISED
#		(& renamed the 1st one as 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm")),
#		especially to take advantage of multiplication factors for case-(i) "gapped segment"s;
#
#	   (1-iii) the subroutine, 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm"), was REVISED, so that it can now read in the input file(s) recording multiplication factors for case-(i) "gapped segment"s;
#
#	(2) REVISED the rules (in the subroutine 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm") and its satellite subroutines) to "extrapolate" the multiplication factors (for case-(iv) and case-(i)) to those w/ the time-lapses smaller than the minimum time-lapse recorded in the input file (of the relevant multiplication factors).
#
#	[MAJOR REVISION STARTED on 2020/04/23; ENDED on 2020/04/dd.]
#
#	[Further REVISION on 2020/05/09]:
#	Now, this script uses the standard package, "Env", and its associated hash %ENV.
#
#
# anex.ver0.5.pl:
#	This is the 1st prototype of "ANEX" (for the "Alignment Neighborhood EXplorer")
#	written in Perl.
#	It was created by putting together useful portions of a number of test scripts.
#	More precisely, the sources are:
#	(1) Some near-top preparatory parts of 'fa_lolipog_hs.alpha2.pl' (in the package, LOLIPOG ver 0.6.1.7);
#	(2) Some near-top preparatory parts of 'test_bch_br_parsimonize_dollo_indel_histories.pl' (at $RIGHT_NOW_WORKING_CHICHIBU2);
#	(3) The main and pre-main parts of 'test_prepare_sliding_windows.pl' (at $RIGHT_NOW_WORKING_CHICHIBU3A);
#	(5) The main and other parts of 'test_wrapper_detect_cmplx_error_cand_acls0.pl' (at $RIGHT_NOW_WORKING_CHICHIBU7B);
#
#	Then, other necessary parts were either imported from other scripts
#	or created de novo.
#	[CREATION STARTED on 2019/06/08, ENDED (drafting) on 2019/06/24.]
#
#	[MAJOR REVISION on 2019/06/27]:
#	CHANGED the source of the subroutine, 'associate_nodeids_w_msa_rows',
#	from "MyANEX_Supple.pm" to "MyTreeMap_sppl.pm".
#
#	[REVISED on 2019/07/03]:
#	Now, the script also outputs the purge-candidate regions within each window,
#	not only those within each sub-window as before the revision.
#
#	[MAJOR REVISION on 2019/07/04]:
#	ADDED a 2nd main for-loop, where the "reverse-purge" (aka. "ex-nihilo") will be performed
#	(via "anex_rev_purge_1wd.verXXX.pl").
#
#	[MAJOR REVISION on 2019/07/07]:
#	MOVED (nearly) ALL the subroutines from within this script to the package, "MyANEX_IO.pm".
#
#	[REVISED on 2019/07/08]:
#	(1) MOVED the block to take in the command-line arguments down, to below the part to specify global parameters;
#	(2) ADDED the block to extract command-line options.
#	[The portions concerning the command-line options were IMPORTED from "detect_purge_cands.ver0.5.pl" & "anex_for_sgl_wd.ver0.5.pl".]
#
#	[REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/11/14]:
#	Now, TWO new global variables, $IF_TEST_W_COMPACT_MSS & $HF_WDTH_SHIFT_4_CMSS, are available,
#	for enabling the test with a "compact" multiple-shift space (whose half-width is given by $HF_WDTH_SHIFT_4_CMSS).
#
#	[REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' on 2019/11/15]:
#	Now, TWO new global variables, $XLEVEL_1WD & $XLEVEL_REV_PRG_1WD, are available,
#	for specifying the LEVELs of the MSA Space Explorations for the scripts,
#	$path_anex_1wd & $path_anex_rev_prg_1wd, respectively.
#

use strict;
use warnings;

use Env; # ADDED in further REVISION on 2020/05/09.

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyMalnIO qw(
	extract_alignment
	pr_ODEN_input
	);
	# extract_fasta
	# pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyTreeManipu_II qw(
	input_tree_file output_tree
	br_output_tree
	);
	# extract_sgltree_from_file
	# make_node2parent

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	br_output_tree_dawg
		) ;
	# extract_sgltree_from_file_dawg


	# MODIFIED as of 2019/02/04.
use MyTreeMap_indels_spt_odr2 qw(
	cal_init_psm_cands_for_sgl_msa
	bch_br_parsimonize_dollo_indel_histories
	);

use MyTreeMap_indels_spt_odr_finer qw(
	enum_psm_indel_hsts_for_msa_finer_PRE
	);
	# partit_into_gblocks_seqblocks
	# if_idb_includes_root
	# partit_into_eff_indep_indel_blocks
	# find_invlvd_in_eff_indep_indel_blocks
	# mk_node2blwise_gstat

use MyReadAlnProbIngredients qw(
	readin_delta_del
	assign_br2log_mfacs
	);

	# Added on 2019/06/14. #
use MyDetect_cmplx_error_cands qw(
	wrapper_detect_cmplx_error_cand_acls0
	);
	# mk_self_nn_events
use MyANEX_Supple qw(
	prepare_sliding_windows
	mk_pcc_sliding_windows
	);
	# Added 'mk_pcc_sliding_windows' in REVISION of 'anex.ver0.5.pl' (on 2019/07/05).

	# End of "Added on 2019/06/14." #

	# Added on 2019/06/20. #
use MyArrayManipu_num qw(
	diff_nmarray
	nr_nmarray
	);
	# Added 'nr_nmarray' on 2019/07/03.

	# End of "Added on 2019/06/20." #

=Set Aside for Later Use.
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_III qw(
	make_node2abs_depth
	make_node2ct_exoffs
	);

use MyTreeManipu_IV qw(
	bi2multifurcated_top
	construct_neat_subtree
	construct_skeleton_subtree
	);

	# ADDED in REVISION (on 2019/06/27). #
	# (=> MOVED to below the "Set Aside for Later Use" on 2020/05/07.)
use MyTreeMap_sppl qw(
	associate_nodeids_w_msa_rows
	);
	# End of "ADDED in REVISION (on 2019/06/27)." #

use MyTreeMap_indels_spt_odr_hs qw(
	enum_psm_indel_scns_for_msa_hs
	);
##use MyTreeMap_indels qw(
##	enum_psm_indel_scns_for_msa
##	);
##	## extract_cls_gblocks msa_rep_gpatterns

use MyTreeMap_indels_sppl qw(
	create_br2eqbr
	create_br2nn
	);

use MyTreeMap_indels_ML_hs_hs qw(
	apprx1_tot_lnlk_gpattern_msa2_hs
	);
	## cal_delta_exp_del_5p2
	## cal_log_normal_facs_indel_distri
	## cal_lnlk1_indel_scn_cls2

=cut # End of "Set Aside for Later Use."

	# ADDED in REVISION (on 2019/06/27). #
	# => MOVED out of the "Set Aside for Later Use" above on 2020/05/07.
use MyTreeMap_sppl qw(
	associate_nodeids_w_msa_rows
	);
	# End of "ADDED in REVISION (on 2019/06/27). #
	# => MOVED out of the "Set Aside for Later Use" above on 2020/05/07."


	### ADDED in MAJOR REVISION on 2019/07/07. ###
use MyANEX_IO qw(
	read_dawg_ctrl_file_sbst 
	read_dawg_ctrl_file3 
	readin_list_mgd_pcands 

	print_out_input_tree 
	print_out_br2log_mfacs 
	print_out_tree_profile 
	print_out_cal_init_psm_cands_for_sgl_msa 
	print_out_bch_br_parsimonize_dollo_indel_histories 
	print_out_alignment_psmnzd 
	print_out_wrapper_detect_cmplx_error_cand_acls0 
	print_out_prepare_sliding_windows 
	print_out_list_mgd_pcands 
	print_out_chk_ovlp_pcands_vs_sliding_wds 
	print_out_set_clms 
	print_out_list_gblocks_in_wd 
	print_out_mk_pcc_sliding_windows 

	select_sgn_wds_in_pcands 
	chk_ovlp_pcands_vs_sliding_wds 
	determine_limits_window 
	);
	### End of "ADDED in MAJOR REVISION on 2019/07/07." ###


	## ADDED in CORRECTION on 2020/05/08. ##
use MyANEX_Supple qw(
	mk_init_set_columns0
	);
	## End of "ADDED in CORRECTION on 2020/05/08." ##


$| = 1; # Flush the input buffer as soon as it receives data.

my $plscript_path = $0;
my ($plscript_name) = ($plscript_path =~ m/([^\/]+)$/); 
my $program_name = $plscript_name; # Added on 2019/06/10, to reduce the number of error messages.

##(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;
our $CURR_WORK_DIR = $ENV{'PWD'}; # Modified in further REVISION on 2020/05/09. (GL) #

my $exec_perl = "/usr/bin/perl"; # Perl executable to be used. (Added on 2019/06/19.) 


	### File names storing important parameters. ###

our $FILENAME_DELTA_DEL = 'delta_del.txt'; # The name of the file storing '$delta_del', which is important for calculating the exit rate of the entire sequence. #

our @FILENAMES_LOG_MFACS = ( # Names of the files storing logarithmic multiplication factors, which are in $indir_mfacs. #

		# In the bulk #
	'lmfacs_asites_vs_dgaps.txt', # For the cases where there are some ancestral sites and no descendant sites in between preserved ancestral sites (PASs). #
	'lmfacs_agaps_vs_dsites.txt', # For the cases where there are no ancestral sites and some descendant sites in between PASs.
	'lmfacs_asites_nhom_dsites.txt', # For the cases where there are some ancestral sites and some descendant sites in between PASs.

		# On either boundary (i.e., either PWA end). "bd_$filename" corresponds to $filename above. #
	'bd_lmfacs_asites_vs_dgaps.txt',
	'bd_lmfacs_agaps_vs_dsites.txt',
	'bd_lmfacs_asites_nhom_dsites.txt',

		# For case-(i) "gapped segment"s. #
		# (ADDED in creation of 'anex.ver0.7.pl' on 2020/04/24). (GL1) #
	'lmfacs_nosites.txt',
	'bd_lmfacs_nosites.txt',
		# End of "(ADDED in creation of 'anex.ver0.7.pl' on 2020/04/24). (GL1)" #
	);

	### End of "File names storing important parameters." ###


	##### Global Parameter Settings #####

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;


	##### User-Specifiable Parameters concerning the Behaviors of the Program #####

our $IF_ARCHIVE_WD_RESULTS = 1; # = 1 if archive (via gzip) the working directory, = 0 otherwise. # (Added on 2019/06/24.)

our $DIS_LEN = 50; # The maximum length of the (sub)sequence displayed in a line.
#our $DIS_LEN = 60; # The maximum length of the (sub)sequence displayed in a line.

$MyANEX_IO::DIS_LEN = $DIS_LEN; # ADDED in REVISION (on 2019/07/07). #


our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

=A RELIC of some predecessor scripts. (1)
	# Characters indicating that the site is missing or that its orthology is questionable. 
our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);
=cut # End of "A RELIC of some predecessor scripts. (1)"

our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.

	# Parameters for the relative probabilities of Indel Histories. #
our %INDEL_MODEL2DISTRI = ( # = (indel model => distribution type, ...)
	'zipf' => 'P',
	'"PL"' => 'P',
	'"NB"' => 'NB', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	'"US"' => 'US', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	);

		# MODIFIED in creation of 'anex.ver0.5.pl' (on 2019/06/25). #

$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS = 1; # Ignore null columns(, if at all).

$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = 0; # NOT analyze null-clusters (=clusters consisting only of null-columns), if at all.

=OBSOLETE in creation of 'anex.ver0.5.pl'. (GL)
	# In this analysis of an MSA simulated by Dawg,
	# null columns will play important roles.
$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS = 0;

	# Null-clusters (=clusters consisting only of a null-block each) will be analysed.
$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = 1;
=cut # END of "OBSOLETE in creation of 'anex.ver0.5.pl'. (GL)"
		# End of "MODIFIED in creation of 'anex.ver0.5.pl' (on 2019/06/25)." #

=DISABLED because they look superfluous (on 2019/07/03). (GL)

	# This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = 120;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 150;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;

=cut # End of "DISABLED because they look superfluous (on 2019/07/03). (GL)"

		# (Added on 2019/06/14.) #
	# Parameters for creating and sorting artificial clusters and detecting candidate regions of complex-errors. # 

##our $IF_LEN = 0; # to use 'artf_cluster_gapped_segs'. #
our $IF_LEN = 1; # to use 'artf_cluster_gapped_segs_len'. #

##our ($UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH, $LIST_INDEL_CTS_FOR_NCMPLX) = @MyDetect_cmplx_error_cands::PARAMS_ACGS_15MAMM_NCMPLX90;
our ($UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH, $LIST_INDEL_CTS_FOR_NCMPLX) = @MyDetect_cmplx_error_cands::PARAMS_ACGS_LEN_15MAMM_NCMPLX80a;
#our ($UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH, $LIST_INDEL_CTS_FOR_NCMPLX) = @MyDetect_cmplx_error_cands::PARAMS_ACGS_LEN_15MAMM_NCMPLX80b;

	# End of "Parameters for creating and sorting artificial clusters and detecting candidate regions of complex-errors." # 

	# Parameters for defining sliding windows. #
our $WDTH_MARGIN = 5; # The sliding windows MUST BE at least $WDTH_MARGIN columns apart from any artificial clusters inferred to be likely involved in complex errors.
our $HF_WDTH_SHIFT = 10; # Each gap-block can "shift" up to (and including) $HF_WDTH_SHIFT columns in each direction.
our $NUM_MAX = 4; # The maximum number of gap-blocks each window is allowed to hold.
	# End of "Parameters for defining sliding windows." #
		# End of "(Added on 2019/06/14.)" #

		# Parameters for detecting purge-candidates. (Added on 2019/06/19.) #
our $MIN_WD_SIZE_PCAND =3; # Minimum number of sites in each window for detecting purge-candidates.
our $MAX_WD_SIZE_PCAND =6; # Maximum number of sites in each window for detecting purge-candidates.
our $THRSH_PV_PCAND = 0.05; # P-value measuring the deviation from the specified substitution model.
our $IF_MODPV_PCAND = 0; # = 1 when using a modified P-value, = 0 otherwise.
our $THRSH_PV0_PCAND = 0.2; # P-value measuring the deviation from the random-match model.
		# End of "Parameters for detecting purge-candidates. (Added on 2019/06/19.)" #

		# REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (GL) #
	# The levels of MSA space explorations.

		# (i) For $path_anex_1wd:
	# (1: performing only pure multiple-"shift"s;
	#  2: also performing ;
	#  3: also performing ;
	#  5: performing all the programmed moves.)
#our $XLEVEL_1WD = 1; 
our $XLEVEL_1WD = 3; 
#our $XLEVEL_1WD = 5; 

		# (ii) For $path_anex_rev_prg_1wd:
	# (1: performing only pure multiple-"shift"s;
	#  2: also performing ;
	#  3: also performing ;
	#  5: performing all the programmed moves.) # [NOTE]: This description of levels NEEDS TO BE CORRECTED!!
#our $XLEVEL_REV_PRG_1WD = 1; 
our $XLEVEL_REV_PRG_1WD = 3; 
#our $XLEVEL_REV_PRG_1WD = 5; 
		# END of "REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (GL)" #

	# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (GL) #
our $IF_TEST_W_COMPACT_MSS = 0; # Whether to perform tests with "compact" multiple-shift spaces(=1) or not(=0).
our $HF_WDTH_SHIFT_4_CMSS  = 2;
	# END of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (GL)" #

	##### End of "User-Specifiable Parameters concerning the Behaviors of the Program" #####


	### Controlling the Outputs for TESTING the script. ###

	# Levels of the Tests. #
	#  (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

our $TEST_LEVEL_MAIN = 5; # Testing the inputs/outputs of the processes going through this MAIN program.

$MyANEX_IO::TEST_LEVEL_MAIN = $TEST_LEVEL_MAIN; # Added in REVISION (on 2019/07/07). #


$MyReadAlnProbIngredients::TEST_LEVEL1  = 0;

		## Added on 2019/06/14. ##
	# For 'mk_ingredients_acls_gpd_segs'. #
$MyDetect_cmplx_error_cands::TEST_LEVEL_migacgss = 0;
##$MyDetect_cmplx_error_cands::TEST_LEVEL_migacgss = 5;

	# For 'artf_cluster_gapped_segs'. #
$MyDetect_cmplx_error_cands::TEST_LEVEL_acgs = 0;
##$MyDetect_cmplx_error_cands::TEST_LEVEL_acgs = 5;

	# For 'artf_cluster_gapped_segs_len'. #
$MyDetect_cmplx_error_cands::TEST_LEVEL_acgsl = 0;
##$MyDetect_cmplx_error_cands::TEST_LEVEL_acgsl = 5;

	# For 'detect_cmplx_error_cand_acls0'. #
##$MyDetect_cmplx_error_cands::TEST_LEVEL_dceca = 0;
$MyDetect_cmplx_error_cands::TEST_LEVEL_dceca = 2;
##$MyDetect_cmplx_error_cands::TEST_LEVEL_dceca = 5;
		## End of "Added on 2019/06/14." ##


	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_spt_odr_hs.pm" for details.)
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL = 3;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1 = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1hf = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL2 = 0;


	# Testing 'br_parsimonize_dollo_indel_histories'.
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = 0;
#$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = 3;

	# Testing 'bch_br_parsimonize_dollo_indel_histories'.
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = 0;
#$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = 3;

	# Testing 'enum_psm_indel_hsts_for_msa_finer_PRE'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 0;
#$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 3;

	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_ML_hs_hs.pm" for details.)
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL = 1;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1 = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1a = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL2a = 0;

	### END of "Controlling the Outputs for TESTING the script." ###

	##### END of "Global Parameter Settings" #####


	##### Take in the Command-Line Arguments. ##### (MOVED from above the global-parameter settings, & MODIFIED, in REVISION on 2019/07/08.) #

unless (@ARGV > 5) {
	print STDERR "\nUsage:  [perl] $program_name {input_msa_file} {dawg_ctrl_file (incl. input tree)} {dir_mfac_data} {top_output_dir} {path/to/'anex_for_sgl_wd.verXX.pl'} {path/to/'detect_purge_cands.verXX.pl'} [ options ]\n";

			# Mostly IMPORTED from "detect_purge_cands.ver0.5.pl" (on 2019/07/08). #
	print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";

	print STDERR "           --PATH_REV_PURGE_SCT={path/to/'anex_rev_purge_for_sgl_wd.verXXX.pl'}     if you want to perform \"reverse-purge\"s (aka. \"ex-nihilo\"s) as well   (default=undef);\n"; # Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #
	print STDERR "           --NUM_MAX=integer           for the maximum number of gap-blocks each window can contain (default=$NUM_MAX);\n"; # Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #

	print STDERR "           --HF_WDTH_SHIFT=integer     for the maximum number of \"shift\"s of each gap-block in one direction (default=$HF_WDTH_SHIFT);\n";  # Imported from "anex_for_sgl_wd.ver0.5.pl" (on 2019/07/08). #

	print STDERR "            --MIN_WD_SIZE_PC=integer      for the minimum window size when detecting purge-candidates (default=$MIN_WD_SIZE_PCAND);\n";
	print STDERR "            --MAX_WD_SIZE_PC=integer      for the maximum window size when detecting purge-candidates (default=$MAX_WD_SIZE_PCAND);\n";
	print STDERR "            --THRSH_PV_PC=real(<1)     for the threshold value for the P-value measuring the deviation from the specified substitution model when detecting purge-candidates (default=$THRSH_PV_PCAND);\n";
	print STDERR "            --IF_MODPV_PC=(0 or 1)     to determine whether to use a modified P-value(=1) or not(=0) when detecting purge-candidate (default=$IF_MODPV_PCAND);\n";
	print STDERR "            --THRSH_PV0_PC=real(<1)    for the threshold value for the P-value measuring the deviation from the random-match model when detecting purge-candidate (default=$THRSH_PV0_PCAND);\n\n";
			# End of "Mostly IMPORTED from "detect_purge_cands.ver0.5.pl" (on 2019/07/08)." #

		# REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (ARGV) #
	print STDERR "           --XLEVEL_1WD=integer       for the level of MSA space exploration via 'anex_for_sgl_wd.verXX.pl' (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL_1WD);\n";
	print STDERR "           --XLEVEL_REV_PRG_1WD=integer       for the level of MSA space exploration via 'anex_rev_purge_for_sgl_wd.verXXX.pl' (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL_REV_PRG_1WD);\n"; ### [NOTE]: The description of the levels NEEDS TO BE CORRECTED!!! ###
		# END of "REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (ARGV)" #

		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (ARGV) #
	print STDERR "           --IF_TEST_W_COMPACT_MSS=boolean     for whether tests with \"compact\" multiple-shift spaces are performed (=1) or not (=0) (default=$IF_TEST_W_COMPACT_MSS);\n";
	print STDERR "           --HF_WDTH_SHIFT_4_CMSS=integer      for the maximum number of \"shift\"s of each gap-block in one direction WHEN \$IF_TEST_W_COMPACT_MSS=1  (default=$HF_WDTH_SHIFT_4_CMSS);\n";
		# END of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (ARGV)" #

	##print STDERR "\nUsage:  [perl] $program_name {input_msa_file} {dawg_ctrl_file (incl. input tree)} {dir_mfac_data} {top_output_dir} {path to 'anex_for_sgl_wd.verXX.pl'} {path to 'detect_purge_cands.verXX.pl'}\n";

	die "$!\n\n";
}
my ($in_msa_file, $dawg_ctrl_file, $indir_mfacs, $toptop_out_dir, $path_anex_1wd, $path_det_pcands) = @ARGV[0 .. 5];

=OBSOLETE as of 2019/06/19.
unless (@ARGV > 3) {
	print STDERR "\nUsage:  [perl] $program_name {input_msa_file} {dawg_ctrl_file (incl. input tree)} {dir_mfac_data} {top_output_dir}\n";
	die "$!\n\n";
}
my ($in_msa_file, $dawg_ctrl_file, $indir_mfacs, $toptop_out_dir) = @ARGV[0 .. 3];
=cut # END of "OBSOLETE as of 2019/06/19."

	# NOTE: The output file system will be specified & prepared below. #

	##### END of "Take in the Command-Line Arguments." #####


		### ADDED in REVISION on 2019/07/08. ###

	##### Extract the Command-Line Options, if at all. ##### (Mostly imported from "detect_purge_cands.ver0.5.pl".)

our $PATH_REV_PURGE_SCT; # Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #

for (my $k=6; $k <= $#ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'MIN_WD_SIZE_PC') {
	    $MIN_WD_SIZE_PCAND = $value;
	} elsif ($option eq 'MAX_WD_SIZE_PC') {
	    $MAX_WD_SIZE_PCAND = $value;
	} elsif ($option eq 'THRSH_PV_PC') {
	    $THRSH_PV_PCAND = $value;
	} elsif ($option eq 'IF_MODPV_PC') {
	    $IF_MODPV_PCAND = $value;
	} elsif ($option eq 'THRSH_PV0_PC') {
	    $THRSH_PV0_PCAND = $value;

	} elsif ($option eq 'HF_WDTH_SHIFT') { # Imported from "anex_for_sgl_wd.ver0.5.pl".
	    $HF_WDTH_SHIFT = $value;

		# Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #

	} elsif ($option eq 'NUM_MAX') {
	    $NUM_MAX = $value;

	} elsif ($option eq 'PATH_REV_PURGE_SCT') {
	    $PATH_REV_PURGE_SCT = $value;

		# End of "Added in revision of 'anex.ver0.5.pl' (on 2019/07/08)." #

		# REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (OP) #
	} elsif ($option eq 'XLEVEL_1WD') {
	    $XLEVEL_1WD = $value;

	} elsif ($option eq 'XLEVEL_REV_PRG_1WD') {
	    $XLEVEL_REV_PRG_1WD = $value;
		# End of "REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (OP)" #

		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (OP) #
	} elsif ($option eq 'IF_TEST_W_COMPACT_MSS') {
	    $IF_TEST_W_COMPACT_MSS = $value;

	} elsif ($option eq 'HF_WDTH_SHIFT_4_CMSS') {
	    $HF_WDTH_SHIFT_4_CMSS = $value;
		# End of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (OP)" #

	} else {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

}	# End of the preliminary for-loop (over additional command-line arguments. #

my $path_anex_rev_prg_1wd = $PATH_REV_PURGE_SCT; # Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #

	##### END of "Extract the Command-Line Options, if at all." ##### 

		### END of "ADDED in REVISION on 2019/07/08." ###


	##### Specify & Check Input/Output files. #####


my $intree_file = $dawg_ctrl_file;
my $infile_delta_del = "$indir_mfacs/$FILENAME_DELTA_DEL"; # This file stores '$delta_del', which is important for calculating the exit rate of the entire sequence. #

my $out_in_chk = "$toptop_out_dir/Inputs_chk";
my $out_interim = "$toptop_out_dir/Interim_Data";

##my $top_outputs = "$toptop_out_dir/Sliding_Window_Analysis"; # MODIFIED (on 2019/06/25). # DISABLED on 2019/07/06.
my $top_outputs = "$toptop_out_dir/Output_Data"; # RESURRECTED on 2019/07/06.

	# MOVED UP, from immediately above the MAIN outer for-loop. (on 2019/06/24) #
##my $top_outdir_main = $top_outputs; # MODIFIED (on 2019/06/25). # DISABLED on 2019/07/06.
my $top_outdir_main = "$top_outputs/Sliding_Window_Analysis"; # RESURRECTED on 2019/07/06.

			# REVISED in DEBUGGING on 2020/05/09. (I/O-F1) #
##my $logdir = "$top_outdir_main/LogFiles"; # Added on 2019/06/24. #
my $logdir_name = 'LogFiles';
my $logdir = "$top_outdir_main/$logdir_name"; # Added on 2019/06/24. #
			# End of "REVISED in DEBUGGING on 2020/05/09. (I/O-F1)" #

		# Added on 2019/07/06. #
my $top_outdir_main2 = "$top_outputs/PCC_Sliding_Window_Analysis";

			# REVISED in DEBUGGING on 2020/05/09. (I/O-F2) #
##my $logdir2 = "$top_outdir_main2/LogFiles";
my $logdir2 = "$top_outdir_main2/$logdir_name";
			# End of "REVISED in DEBUGGING on 2020/05/09. (I/O-F2)" #
		# End of "Added on 2019/07/06." #

	# End of "MOVED UP, from immediately above the MAIN outer for-loop. (on 2019/06/24)" #


		# Added in revision (on 2019/06/19). (Chk 1) #
	## Check Perl exacutable & satellite scripts. ##

foreach my $file ($path_anex_1wd, $path_det_pcands) {
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable SCRIPT '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

if (defined $path_anex_rev_prg_1wd) { # Added in REVISION (on 2019/07/08). #
    my $file = $path_anex_rev_prg_1wd;
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable SCRIPT '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of "Added in REVISION (on 2019/07/08)." #

foreach my $file ($exec_perl) {
    unless (( -f $file ) and ( -x $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND an executable, '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}
	## End of "Check Perl exacutable & satellite scripts." ##
		# End of "Added in revision (on 2019/06/19). (Chk 1)" #

	## Check input files/directories. ##

unless (( -d $indir_mfacs ) and ( -r $indir_mfacs )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY, '$indir_mfacs', that stores data files for the multiplication factors !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

foreach my $file ($in_msa_file, $dawg_ctrl_file, $intree_file, $infile_delta_del) {
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

foreach my $filename (@FILENAMES_LOG_MFACS) {
    my $file = "$indir_mfacs/$filename";
    ##unless (( -f $file ) and ( -r $file )) {
    unless ( (( -f $file ) or ( -l $file )) and ( -r $file )) { # MODIFIED in creation of 'anex.ver0.7.pl' (on 2020/04/24). (CHK) #

	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}
	## End of "Check input files/directories." ##

	## Check output directiries. ##

# $out_in_chk = "$toptop_out_dir/Inputs_chk";
# $out_interim = "$toptop_out_dir/Interim_Data";
# $top_outputs = "$toptop_out_dir/Sliding_Window_Analysis"; # MODIFIED (on 2019/06/25). #
# $top_outdir_main = $top_outputs; # MODIFIED (on 2019/06/25). #
# $logdir = "$top_outdir_main/LogFiles"; # Added on 2019/06/24. #

		# Added on 2019/07/06. #
# $top_outdir_main2 = "$top_outputs/PCC_Sliding_Window_Analysis";
# $logdir2 = "$top_outdir_main2/LogFiles";
		# End of "Added on 2019/07/06." #

foreach my $dir ($toptop_out_dir, $out_in_chk, $out_interim, $top_outputs, $top_outdir_main, $logdir, $top_outdir_main2, $logdir2) { # Preliminary foreach-loop (over output directories). # Further Modified on 2019/07/06.
##foreach my $dir ($toptop_out_dir, $out_in_chk, $out_interim, $top_outputs, $top_outdir_main, $logdir) { # Preliminary foreach-loop (over output directories). # Modified on 2019/06/24. # [OBSOLETE as of 2019/07/06.] ...}
####foreach my $dir ($toptop_out_dir, $out_in_chk, $out_interim, $top_outputs) { # Preliminary foreach-loop (over output directories). # [OBSOLETE as of 2019/06/24.] ...}
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the preliminary foreach-loop (over output directories). #

	## End of "Check output directiries." ##


	##### END of "Specify & Check Input/Output files." #####


	##### Declare the Beginning of the Script #####

if ($TEST_LEVEL_MAIN>0) { # (1) #

	print STDOUT "\n<<<< Program name: $plscript_name (path: $plscript_path) >>>>\n\n";

		# Added in revision (on 2019/06/19). (PR 2) #
	print STDOUT "Perl executable:   $exec_perl .\n\n";

	print STDOUT "<< Satellite Scripts >>\n\n";
	print STDOUT "For detecting purge-candidate regions: $path_det_pcands ,\n";
	print STDOUT "For exploring within each window: $path_anex_1wd ,\n";
	print STDOUT "For performing \"reverse-purge\"s in each specially prepared window: ", ((defined $path_anex_rev_prg_1wd) ?  $path_anex_rev_prg_1wd : 'undef'), " .\n"; # Added in revision of 'anex.ver0.5.pl' (on 2019/07/08). #
	print STDOUT "\n\n";
		# End of "Added in revision (on 2019/06/19). (PR 2)" #

  if ($TEST_LEVEL_MAIN>1) {

	print STDOUT "<< Input/Output files/directories >>\n\n";

	print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

	print STDOUT "Input MSA file: $in_msa_file .\n\n";

	print STDOUT "Control file: $dawg_ctrl_file .\n\n";

	print STDOUT "Input tree and indel parameters: $intree_file .\n\n";

	print STDOUT "Input file recording ", '$delta_del', ": $infile_delta_del .\n\n";

	print STDOUT "Input directory storing (log-)multiplication factors: $indir_mfacs ,\n\n";
	print STDOUT "   which holds the following input files:\n\n";
	print STDOUT '     ', join (', ', @FILENAMES_LOG_MFACS), " .\n";
	print STDOUT "\n\n";

	print STDOUT "Output File System:\n\n";

	print STDOUT ' ' x 3, "{Super-Top directory} = $toptop_out_dir ,\n";
	print STDOUT ' ' x 3, "{Directory for checking the input data} = $out_in_chk ,\n";
	print STDOUT ' ' x 3, "{Directory for interim output data}     = $out_interim ,\n";
	print STDOUT ' ' x 3, "{Top directory for final output data}   = $top_outputs ,\n";
	print STDOUT ' ' x 3, "{Top directory for the sliding-window analyses} = $top_outdir_main ,\n"; # Added on 2019/06/24.
	print STDOUT ' ' x 3, "{Directory for the log-files on the sliding-window analyses} = $logdir ,\n"; # Added on 2019/06/24.
	print STDOUT ' ' x 3, "{Top directory for the PCC sliding-window analyses} = $top_outdir_main2 ,\n"; # Added on 2019/07/06.
	print STDOUT ' ' x 3, "{Directory for the log-files on the PCC sliding-window analyses} = $logdir2 .\n"; # Added on 2019/07/06.
	print STDOUT "\n\n";


	print STDOUT "<< User-specified Parameters >>\n\n";

		# Added in REVISION on 2019/07/07. #
	print STDOUT '$DIS_LEN = ', $DIS_LEN, " ,\n";
	print STDOUT '$MyANEX_IO::DIS_LEN = ', $MyANEX_IO::DIS_LEN, " .\n\n";
		# End of "Added in REVISION on 2019/07/07." #

	print STDOUT 'Whether to archive the working directories or not: $IF_ARCHIVE_WD_RESULTS = ', $IF_ARCHIVE_WD_RESULTS, " .\n\n\n"; # (Added on 2019/06/24.)

=A RELIC of some predecessor scripts. (2)
	print STDOUT "Characters indicating that the site is missing or that its orthology is questionable:\n\n";
	print STDOUT ' ' x 5, join (', ', sort keys %CHAR_MISSING), " .\n";
	print STDOUT "\n\n";
=cut # End of "A RELIC of some predecessor scripts. (2)"

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS =' , $MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS, " ,\n";
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = ', $MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS, " ,\n\n";
=DISABLED because they look superfluous (on 2019/07/03). (PR)
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_GAPLEN = ', $UPPER_BOUND_ANAL_GAPLEN, " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " .\n";
	print STDOUT "\n\n";
=cut # End of "DISABLED because they look superfluous (on 2019/07/03). (PR)"

			# Added on 2019/06/14. (PR 1) #

	print STDOUT '< Parameters for detecting candidate regions of complex-errors & defining sliding windows >', "\n\n";

	print STDOUT '$IF_LEN = ', $IF_LEN, " .\n\n";

	print STDOUT '($UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH) = ', "($UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH) .\n\n";

	print STDOUT '@{$LIST_INDEL_CTS_FOR_NCMPLX} = ( # element = [#{ins} + #{dels}, #{ins} - #{dels}],', "\n";
	foreach my $indel_cts (@{$LIST_INDEL_CTS_FOR_NCMPLX}) {
		print STDOUT "\t[", join (', ', @{$indel_cts}), "],\n";
	}
	print STDOUT "\t\t) .\n";
	print STDOUT "\n\n";

	print STDOUT '($WDTH_MARGIN, $HF_WDTH_SHIFT, $NUM_MAX) = ', "($WDTH_MARGIN, $HF_WDTH_SHIFT, $NUM_MAX) .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/06/14. (PR 1)" #

		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (PR) #
	print STDOUT "< Options regarding the tests with \"compact\" multiple-shift spaces. >\n\n";
	print STDOUT '$IF_TEST_W_COMPACT_MSS = ', $IF_TEST_W_COMPACT_MSS, " .\n";
	if ($IF_TEST_W_COMPACT_MSS) {
	    print STDOUT '$HF_WDTH_SHIFT_4_CMSS  = ', $HF_WDTH_SHIFT_4_CMSS, " .\n";
	}
	print STDOUT "\n\n";
		# End of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (PR)" #

			# Added on 2019/06/19. (PR 3) #
	print STDOUT "< Parameters for detecting purge-candidates. >\n\n";
	print STDOUT '($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) = ', "($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) .\n\n";
	print STDOUT '($THRSH_PV_PCAND, $IF_MODPV_PCAND, $THRSH_PV0_PCAND) = ', "($THRSH_PV_PCAND, $IF_MODPV_PCAND, $THRSH_PV0_PCAND) .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/06/19. (PR 3)" #

		# REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (DC) #
	print STDOUT "< The LEVELs of MSA Space Explorations: >\n\n";
	print STDOUT "(for 'anex_for_sgl_wd.verXX.pl')            \$XLEVEL_1WD         = ", $XLEVEL_1WD, " ,\n";
	print STDOUT "(for 'anex_rev_purge_for_sgl_wd.verXX.pl')  \$XLEVEL_REV_PRG_1WD = ", $XLEVEL_REV_PRG_1WD, " .\n";
	print STDOUT "\n";
		# End of "REVISION IMPORTED w/ modifications from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/15). (DC)" #

	print STDOUT "< Parameters specifying How verbose the Test Outputs will be: >\n\n";

	print STDOUT '$TEST_LEVEL_MAIN = ', $TEST_LEVEL_MAIN, " ,n";
	print STDOUT '$MyANEX_IO::TEST_LEVEL_MAIN = ', $MyANEX_IO::TEST_LEVEL_MAIN, " .\n\n"; # Added in REVISION (on 2019/07/07). #

	#print STDOUT '$MyDetect_cmplx_error_cands::TEST_LEVEL_migacgss = ', $MyDetect_cmplx_error_cands::TEST_LEVEL_migacgss, " ,\n";
	#print STDOUT '$MyDetect_cmplx_error_cands::TEST_LEVEL_acgs = ', $MyDetect_cmplx_error_cands::TEST_LEVEL_acgs, " ,\n";
	#print STDOUT '$MyDetect_cmplx_error_cands::TEST_LEVEL_acgsl = ', $MyDetect_cmplx_error_cands::TEST_LEVEL_acgsl, " ,\n";
	print STDOUT '$MyDetect_cmplx_error_cands::TEST_LEVEL_dceca = ', $MyDetect_cmplx_error_cands::TEST_LEVEL_dceca, " .\n\n";


	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs , " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs, " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = ', $MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE, " .\n\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) {...}"

} # END of "if ($TEST_LEVEL_MAIN>0) { # (1) # ...}"


	### Set an Option-String for the tests w/ Compact Multiple-Shift Spaces. ###
	### (ADDED on 2019/11/14.) ###

our $OPTION_4_TEST_W_CMSS = ($IF_TEST_W_COMPACT_MSS) ?
		" --IF_TEST_W_COMPACT_MSS=1  --HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS " :
		" --IF_TEST_W_COMPACT_MSS=0 " ;


if ($TEST_LEVEL_MAIN>0) {
	print STDOUT "\n<< Set an Option-String for the tests w/ Compact Multiple-Shift Spaces: >>\n\n";
	print STDOUT "\$OPTION_4_TEST_W_CMSS = $OPTION_4_TEST_W_CMSS .\n";
	print STDOUT "\n\n";
} # End of "if ($TEST_LEVEL_MAIN>0) {...}"

	### End of "(ADDED on 2019/11/14.)" ###


	##### Read in the Control file & Set Some Model Parameters. #####

		# Added on 2019/06/11. #

print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for substitution model parameters. >>>\n\n\n";

my ($model_sbst, $nt_freqs, $sbst_params) = read_dawg_ctrl_file_sbst ($dawg_ctrl_file);
unless (defined $model_sbst) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $MODEL_SBST = $model_sbst;

	# MODIFIED on 2019/06/14. #

our @NT_FREQS = @{$nt_freqs}[0,2,1,3]; # Re-arrange the frequencies: from (ACGT) in Dawg to (AGCT) in this package.
##our @NT_FREQS = @{$nt_freqs};

our @SBST_PARAMS = @{$sbst_params};

if ($TEST_LEVEL_MAIN>0) { # (2A) #
	## Print out the Parameters for Substitution Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Substitution Probabilities ]]\n\n";

	print STDOUT '$MODEL_SBST = ', $MODEL_SBST, " .\n\n";

	print STDOUT '@NT_FREQS = (', join (', ', @NT_FREQS), ") .\n\n";

	print STDOUT '@SBST_PARAMS = (', join (', ', @SBST_PARAMS), ") .\n\n";

	print STDOUT "\n";

} # End of "if ($TEST_LEVEL_MAIN>0) { # (2A) # ...}"

		# End of "Added on 2019/06/11." #


	# Translate the @SBST_PARAMS into the Parameters for the Tamura-Nei model. #
	# (Added on 2019/06/14.) #

our ($ALPHA1, $ALPHA2, $BETA);
	# $ALPHA1 = rate for a transitional substitution (A <-> G); 
	# $ALPHA2 = rate for a transitional substitution (C <-> T);
	# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).

if (($MODEL_SBST eq 'JC') or ($MODEL_SBST eq 'F81')) {

	($ALPHA1, $ALPHA2, $BETA) = (1, 1, 1);

} elsif (($MODEL_SBST eq 'K2P') or ($MODEL_SBST eq 'HKY')) {

	($ALPHA1, $BETA) = @SBST_PARAMS;
	 $ALPHA2 = $ALPHA1;
	
} elsif ($MODEL_SBST eq 'TN') {

	($ALPHA1, $ALPHA2, $BETA) = @SBST_PARAMS;
} else {
	print STDOUT "[ERROR '$program_name'] The current version does NOT HANDLE the substitution model, '$MODEL_SBST'! ... USE either 'JC', 'K2P', 'HKY', 'F81' or 'TN'!!\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL_MAIN>0) { # (2A') #
	print STDOUT '[ Translation of @SBST_PARAMS into Parameters of Tamura-Nei model ]', "\n\n";
	print STDOUT '($ALPHA1, $ALPHA2, $BETA) = ', "($ALPHA1, $ALPHA2, $BETA) .\n\n";
	print STDOUT "# NOTE:\n";
	print STDOUT '# $ALPHA1 = rate for a transitional substitution (A <-> G);', "\n";
	print STDOUT '# $ALPHA2 = rate for a transitional substitution (C <-> T);', "\n";
	print STDOUT '# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).', "\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>0) { # (2A') # ...}"

	# End of "(Added on 2019/06/14.)" #

print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for indel model parameters and phylogenetic tree. >>>\n\n\n";

my ($tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del, $gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate)
	= read_dawg_ctrl_file3 ($dawg_ctrl_file);
unless (defined $tree_string) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $INS_MODEL = $gapmodel_ins;
our $DEL_MODEL = $gapmodel_del;
our $DISTRI_INS = $INDEL_MODEL2DISTRI{$INS_MODEL};
our $DISTRI_DEL = $INDEL_MODEL2DISTRI{$DEL_MODEL};

	# Total rates of insertions & deletions at a site relative to the substitution rate.
our $TOTRATE_INS = $ins_rate;
our $TOTRATE_DEL = $del_rate;

 	# Exponents for the power-law distributions of insertion/deletion lengths (when $INDEL_MODEL = 'zipf').
our $PARAM_INS = $gap_param_ins;
our $PARAM_DEL = $gap_param_del;

	# Cut-off insertion/deletion lengths.
our $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
our $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL); 


if ($TEST_LEVEL_MAIN>0) { # (2B) #

	## Print out the Parameters for Indel Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Indel Probabilities ]]\n\n";

	print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
	print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";

	print STDOUT '$CUTOFF_INS_LEN = ', "$CUTOFF_INS_LEN ,\n";
	print STDOUT '$CUTOFF_DEL_LEN = ', "$CUTOFF_DEL_LEN .\n";
	print STDOUT "\n\n";

	## End of "Print out the Parameters for Indel Probabilities." ##

} # End of "if ($TEST_LEVEL_MAIN>0) { # (2B) # ...}"

	##### End of "Read in the Control file." #####

	##### Read in the $delta_del. #####

(print STDOUT "\n<<< Read in the multiplication factor '\$delta_del' for a 'universal' term of log-lilekihood ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $delta_del = readin_delta_del ($infile_delta_del);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_delta_del' FAILED on the input file, '$infile_delta_del' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}
(print STDOUT "\n ... The subroutine, 'readin_delta_del' Ended (w/ \$delta_del = $delta_del) ...\n\n") if ($TEST_LEVEL_MAIN> 0);

	##### End of "Read in the $delta_del." #####

	##### Read in the INPUT TREE. #####

(print STDOUT "\n<<< Read in the Input Tree >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
# [my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches. ] --- for the predecessor 'br_infer_psm_indel_events.pl'. #
my %br_attr; # = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches.
my %name2id_intnode; # = ($node_name => $node_id, ...) for all interior nodes. #
my $topnode_name; # The name of the top-node (root).

unless ( input_tree_file_dawg (${$tree_string}, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $topnode_id = $name2id{$TOP_NODE};


if ($TEST_LEVEL_MAIN>0) { # (3) #

	print STDOUT "\n ... To check the input tree, perform the subroutine, 'print_out_input_tree' ...\n\n";

	unless ( print_out_input_tree ($out_in_chk, $topnode_id, %node2ch, %node2pa, %id2name, %br_attr, $topnode_name) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_input_tree' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_input_tree' Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>0) { # (3) # ...}"

	##### End of "Read in the INPUT TREE." #####


	##### Assign Log-Multiplication Factors to Each Branch. #####

(print STDOUT "\n<<< Assign the logarithmic multiplication factors to each branch, based on the files in '$indir_mfacs' ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $br2log_mfacs = assign_br2log_mfacs (%br_attr, $TOTRATE_INS + $TOTRATE_DEL, $indir_mfacs, @FILENAMES_LOG_MFACS);
unless (defined $br2log_mfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'assign_br2log_mfacs' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'assign_br2log_mfacs', Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN> 3) { # (5) #

	print STDOUT "\n ... To check the outputs of 'assign_br2log_mfacs', Run the subroutine, 'print_out_br2log_mfacs' ...\n\n";

# $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
# $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

	unless ( print_out_br2log_mfacs ($out_in_chk, %{$br2log_mfacs}, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2log_mfacs' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_br2log_mfacs' Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN> 3) { # (5) # ...}"

	##### End of "Assign Log-Multiplication Factors to Each Branch." #####


	##### Read in the INPUT MSA. #####

(print STDOUT "\n<<< Read in the Input MSA, via the subroutine, 'extract_alignment' >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my ($in_msa, $seqnames, $seq_lens) = extract_alignment ($in_msa_file);
unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from the file '$in_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_seqs = @{$seqnames};

unless ($ct_seqs>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] NO SEQUENCES in the input MSA from the file '$in_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_aln = length ${$in_msa->[0]} ;

unless ($totlen_aln>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] ZERO-LENGTH SEQUENCES in the input MSA from the file '$in_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=SET ASIDE for potential use.
	# REMOVE the Ancestral Sequences from the Input MSA. #
	# (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my @msa_extseqs = my @ext_seqnames = ();
for (my $i=0; $i<$ct_seqs; $i++) { # Preliminary for-loop (over sequences in the input MSA). #

	my $seqid = $seqnames->[$i];
	($seqid  =~ m/^anc/) and next;

	push @ext_seqnames, $seqid;
	push @msa_extseqs, $in_msa->[$i];

}			# End of the preliminary for-loop (over sequences in the input MSA). #

@{$in_msa} = @msa_extseqs;
@{$seqnames} = @ext_seqnames;
$ct_seqs = @{$seqnames};

=cut # End of "SET ASIDE for potential use."

(print STDOUT "\n ... The subroutine, 'extract_alignment', Ended (w/ #{(external) sequences}=$ct_seqs, #{columns}=$totlen_aln) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (6) #

	my $chk_imsa_file = "$out_in_chk/in_msa.txt";
	my $fh_ci_msa = open_output_file ($chk_imsa_file, 0);
	unless (defined $fh_ci_msa) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}

	print STDOUT "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";
	print $fh_ci_msa "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";

	unless ( pr_ODEN_input (@{$seqnames}, @{$in_msa}, $DIS_LEN, *${$fh_ci_msa}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_ODEN_input'  FAILED to print out the input MSA into \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print $fh_ci_msa "\n";
	print STDOUT "\n ... the input MSA has been written into the file, \"$chk_imsa_file\" ...\n\n";

	close ($fh_ci_msa);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (6) # ...}"

	##### End of "Read in the INPUT MSA." #####


	####### Perform some Preliminary Processes #######

(print STDOUT "\n<<<<< Perform some Preliminary Processes >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

	### (Pre 1) 'enum_psm_indel_hsts_for_msa_finer_PRE' ###

(print STDOUT '<<< Preliminary 1: Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, ', "by performing 'enum_psm_indel_hsts_for_msa_finer_PRE' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

(print STDOUT "\n (If you want to know what's going on, SWITCH ON the global variable, \$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE .)\n\n") if ($TEST_LEVEL_MAIN>1);

my ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes)
	= enum_psm_indel_hsts_for_msa_finer_PRE ($topnode_id, %node2ch, %node2pa);
unless (defined $br2nn) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (5) #

	print STDOUT "\n ... Check the outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE', by performing 'print_out_tree_profile' ...\n\n";
	unless ( print_out_tree_profile ($out_interim, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, %{$br2nn}, @{$set_all_nodes}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_tree_profile' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_tree_profile', Ended ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (5) # ...}"

	### End of "(Pre 1) 'enum_psm_indel_hsts_for_msa_finer_PRE'." ###


	### (Pre 1.5) Copy the input MSA. ###

my @cp_in_msa = ();
foreach my $seq (@{$in_msa}) {
	my $cp_seq = ${$seq};
	push @cp_in_msa, \$cp_seq;
}

(print STDOUT "\n", ' ... To Prepare for multiple processes, Copied the @{$in_msa} into an array, @cp_in_msa ...', "\n\n") if ($TEST_LEVEL_MAIN>0);

	### End of "(Pre 1.5) Copy the input MSA." ###

	## (Pre 2) Chop the input MSA into gapped and gapless segments. ##
	## & (Pre 3) Create an initial candidate indel history based on Dollo-parsimonious gap states. ##

if ($TEST_LEVEL_MAIN>1) {
   print STDOUT "<<< Preliminary 2: Chop the input MSA into gapped and gapless segments >>>\n";
   print STDOUT "& <<< Preliminary 3: Create an initial candidate indel history based on Dollo-parsimonious gap states >>>\n";
   print STDOUT "   (, BOTH by performing the subroutine, 'cal_init_psm_cands_for_sgl_msa'.)\n\n\n";
} # End of "if ($TEST_LEVEL_MAIN>1) {...}"

my @info_gpatterns0 = ([], [], []); # A dummy initial set of information on gap-patterns of MSA columns. #

my ($info_init_psm_hst, $info_gpatterns) = cal_init_psm_cands_for_sgl_msa (%node2ch, %node2pa, %{$node2depth}, %{$node2ct_exoffs}, %name2id, @{$seqnames}, @{$in_msa}, $CHAR_ABS, $CHAR_PRES, @info_gpatterns0);
unless (defined $info_init_psm_hst) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'cal_init_psn_cands_for_sgl_msa' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}

my ($set_cls_gpbs, $init_totcts_indels_for_cls, $init_psm_hsts_for_cls) = @{$info_init_psm_hst} ;
my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns} ;

my $ct_gpd_segments = @{$set_cls_gpbs};
my $ct_gp = @{$set_gpatterns};

(print STDOUT "\n ... The subroutine, 'cal_init_psm_cands_for_sgl_msa', Ended (w/ #{gap-patterns}=$ct_gp, #{gapped segments}=$ct_gpd_segments) ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN > 3) { # (6) #

	print STDOUT "\n ... Check the outputs of 'cal_init_psm_cands_for_sgl_msa', by performing the subroutine, 'print_out_cal_init_psm_cands_for_sgl_msa' ...\n\n";
	unless ( print_out_cal_init_psm_cands_for_sgl_msa ($out_interim, @{$info_gpatterns}, @{$info_init_psm_hst}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_cal_init_psm_cands_for_sgl_msa' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_cal_init_psm_cands_for_sgl_msa', Ended ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN > 3) { # (6) # ...}"

	## END of "(Pre 2) Chop the input MSA into gapped and gapless segments. ##
	## & (Pre 3) Create an initial candidate indel history based on Dollo-parsimonious gap states." ##


	## (Pre 5) Briefly Attempt to Parsimonize the Dollo parsimony-based indel history for each gapped segment. ##

(print STDOUT "\n\n<<< Preliminary 5: Briefly Attempt to Parsimonize the Dollo parsimony-based indel history for each gapped segment, via 'bch_br_parsimonize_dollo_indel_histories' >>>\n\n\n") if ($TEST_LEVEL_MAIN>1);

	# Prepare an additional input. #
my @set_exnodes = ();
foreach my $node (@{$set_all_nodes}) { 
	unless (defined $node2ch{$node}) { push @set_exnodes, $node; } 
}
(print STDOUT 'An Additional Input: @set_exnodes = (', join (', ', @set_exnodes), ") .\n\n") if ($TEST_LEVEL_MAIN>2);
	# End of "Prepare an additional input." #

my ($new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places)
   = bch_br_parsimonize_dollo_indel_histories (@cp_in_msa, @{$info_init_psm_hst}, @set_exnodes, $topnode_id, %node2pa, %node2ch, $CHAR_ABS, $CHAR_PRES, %{$br2eq});
unless (defined $new_info_init_psm_hst) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'bch_br_parsimonize_dollo_indel_histories' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'bch_br_parsimonize_dollo_indel_histories', Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

my ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;

unless (@{$new_set_cls_gpbs} == $ct_gpd_segments) {
	print STDOUT "[ERROR '$plscript_name'] The number of gapped segments AFTER the brief parsimonization (=", scalar (@{$new_set_cls_gpbs}), ") CLASHES that BEFORE the brief parsimonization (=$ct_gpd_segments) !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}

		# Added in revision (on 2019/06/17). #
my $outdir_name_psmnzd = "Init_psm_cands_parsimonized";
my $outdir_psmnzd = "$out_interim/$outdir_name_psmnzd";
my $new_aln_filename = "new_alignment.txt";
my $out_new_aln = "$outdir_psmnzd/$new_aln_filename"; # This is very IMPORTANT!!
		# End of "Added in revision (on 2019/06/17)." #

if ($TEST_LEVEL_MAIN > 3) { # (7) #

	print STDOUT "\n ... Check the outputs of 'bch_br_parsimonize_dollo_indel_histories', by performing the subroutine, 'print_out_bch_br_parsimonize_dollo_indel_histories' ...\n\n";

		# REVISED on 2019/06/17. #
		# => Further REVISED on 2019/07/07. #
	unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_interim, $outdir_name_psmnzd, $new_aln_filename, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places}, @{$set_gpatterns})  ) {
	##unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_interim, $outdir_name_psmnzd, $new_aln_filename, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places})  ) {
	####unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_interim, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places})  ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_bch_br_parsimonize_dollo_indel_histories' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_bch_br_parsimonize_dollo_indel_histories', Ended ...\n\n";

		# MODIFIED in revision (on 2019/06/17). (Pre 5) #
} else {

	print STDOUT "\n ... Print out the new alignment output by 'bch_br_parsimonize_dollo_indel_histories', via 'print_out_alignment_psmnzd' ...\n\n";

	unless ( print_out_alignment_psmnzd ($outdir_psmnzd, $new_aln_filename, @{$seqnames},  @{$new_in_msa}) ) { 
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_alignment_psmnzd' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_alignment_psmnzd', Ended ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN > 3) { # (7) # ...} else {...}"
##} # End of "if ($TEST_LEVEL_MAIN > 3) { # (7) # ...}"
		# End of "Modified in revision (on 2019/06/17). (Pre 5)" #

	## END of "(Pre 5) Briefly Attempt to Parsimonize the Dollo parsimony-based indel history for each gapped segment." ##


	## (Pre 6) Detect 'candidate' artificial clusters, which are likely to be involved in complex errors. ## (Added on 2019/06/12&14.)

(print STDOUT "\n\n<<< Preliminary 6: Detect 'candidate' artificial clusters, which are likely to be involved in complex errors, via 'wrapper_detect_cmplx_error_cand_acls0'. >>>\n\n\n") if ($TEST_LEVEL_MAIN>1);


# ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;

my ($cmplx_cands, $ncmplx_cands, $artf_cls, $gs2prev_nn_events, $set_lists_gblocks)
   = wrapper_detect_cmplx_error_cand_acls0 (@{$new_set_cls_gpbs}, @{$new_init_psm_hsts_for_cls}, @set_exnodes, $CHAR_ABS, %node2ch, %node2pa, %{$br2eq}, %{$br2nn}, $IF_LEN, $UB_SPSIZE, $THRSH_SPSIZE0, $MAG_THRSH, @{$LIST_INDEL_CTS_FOR_NCMPLX}, *STDOUT);
unless (defined $cmplx_cands) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'wrapper_detect_cmplx_error_cand_acls0' FAILED on the input tree & initial indel history candidate !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_ccands = @{$cmplx_cands};
my $ct_nccands = @{$ncmplx_cands};

(print STDOUT "\n ... The subroutine, 'wrapper_detect_cmplx_error_cand_acls0', Ended (w/ #{complex_candidates}=$ct_ccands, #{non-complex_candidates)=$ct_nccands) ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN > 3) { # (8) # (Added on 2019/06/14)

	print STDOUT "\n ... Check the outputs of 'wrapper_detect_cmplx_error_cand_acls0', by performing the subroutine, 'print_out_wrapper_detect_cmplx_error_cand_acls0' ...\n\n";

	unless ( print_out_wrapper_detect_cmplx_error_cand_acls0 ($out_interim, @{$cmplx_cands}, @{$ncmplx_cands}, @{$artf_cls}, %{$gs2prev_nn_events}, @{$set_lists_gblocks}, @{$new_set_cls_gpbs}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_wrapper_detect_cmplx_error_cand_acls0' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_wrapper_detect_cmplx_error_cand_acls0', Ended ...\n\n";

} # END of "if ($TEST_LEVEL_MAIN > 3) { # (8) # ...}"

	## END of "(Pre 6) Detect 'candidate' artificial clusters, which are likely to be involved in complex errors." ##


	## (Pre 7) Prepare Sliding Windows. ## (Added on 2019/06/14.)

(print STDOUT "\n\n<<< Preliminary 7:  Prepare Sliding Windows to be used for Exploration of Neighborhoods of Local Alignments, via 'prepare_sliding_windows'. >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1);

my $ct_exnodes = @set_exnodes;
my ($beg_in_msa, $end_in_msa) = (0, $totlen_aln -1);

my ($window_frames, $prohibited_zones, $set_windows, $set_lists_gblocks_in_wds)
	= prepare_sliding_windows ($beg_in_msa, $end_in_msa, @{$new_set_cls_gpbs}, @{$artf_cls}, @{$cmplx_cands}, @{$set_lists_gblocks}, $ct_exnodes, $WDTH_MARGIN, $HF_WDTH_SHIFT, $NUM_MAX);
unless (defined $set_lists_gblocks_in_wds) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'prepare_sliding_windows' FAILED for Max#{gap-blocks} = $NUM_MAX !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'prepare_sliding_windows' Ended for Max#{gap-blocks}=$NUM_MAX ...\n\n") if ($TEST_LEVEL_MAIN >1);


if ($TEST_LEVEL_MAIN > 2) { # (9) #

	print STDOUT "\n ... Check the outputs of 'prepare_sliding_windows' for Max#{gap-blocks} = $NUM_MAX, by performing the subroutine, 'print_out_prepare_sliding_windows' ...\n\n";
	my $outdir_name = sprintf ("Sliding_Windows_Max%d", $NUM_MAX) ;

	unless ( print_out_prepare_sliding_windows ($out_interim, $outdir_name, @{$window_frames}, @{$prohibited_zones}, @{$set_windows}, @{$set_lists_gblocks_in_wds}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_prepare_sliding_windows' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_prepare_sliding_windows', Ended ...\n\n";

} # END of "if ($TEST_LEVEL_MAIN > 2) { # (9) # ...}"


	## END of "(Pre 7) Prepare Sliding Windows." ##


	## (Pre 7') Prepare Sliding Sub-Windows. ## (Added on 2019/06/15.)

(print STDOUT "\n\n<<< Preliminary 7':  Prepare Sliding Sub-Windows to be used for Exploration of Neighborhoods of Local Alignments, via 'prepare_sliding_windows'. >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1);

our $NUM_SBMAX = $NUM_MAX-1;

my ($window_frames2, $prohibited_zones2, $set_subwindows, $set_lists_gblocks_in_sbwds)
	= prepare_sliding_windows ($beg_in_msa, $end_in_msa, @{$new_set_cls_gpbs}, @{$artf_cls}, @{$cmplx_cands}, @{$set_lists_gblocks}, $ct_exnodes, $WDTH_MARGIN, $HF_WDTH_SHIFT, $NUM_SBMAX);
unless (defined $set_lists_gblocks_in_sbwds) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'prepare_sliding_windows' FAILED for Max#{gap-blocks} = $NUM_SBMAX !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'prepare_sliding_windows' Ended for Max#{gap-blocks}=$NUM_SBMAX ...\n\n") if ($TEST_LEVEL_MAIN >1);


if ($TEST_LEVEL_MAIN > 2) { # (9') #

	print STDOUT "\n ... Check the outputs of 'prepare_sliding_windows' for Max#{gap-blocks} = $NUM_SBMAX, by performing the subroutine, 'print_out_prepare_sliding_windows' ...\n\n";
	my $outdir_name = sprintf ("Sliding_Windows_Max%d", $NUM_SBMAX) ;

	unless ( print_out_prepare_sliding_windows ($out_interim, $outdir_name, @{$window_frames2}, @{$prohibited_zones2}, @{$set_subwindows}, @{$set_lists_gblocks_in_sbwds}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_prepare_sliding_windows' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_prepare_sliding_windows', Ended ...\n\n"; 

} # END of "if ($TEST_LEVEL_MAIN > 2) { # (9') # ...}"

	## END of "(Pre 7') Prepare Sliding Sub-Windows." ##


	## (Pre 8) Detect Purge-Candidate Regions. ## (Added on 2019/06/19.)

(print STDOUT "\n\n<<< Preliminary 8:  Detect Purge-Candidate Regions, via the satellite script, \"$path_det_pcands\" . >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1);

my $outdir_pcands = "$out_interim/Purge_cands";
my $name_ofile_pc = 'list_purge_candidates.txt';
my $outfile_pcands = "$outdir_pcands/$name_ofile_pc";

my $log_det_pcands = "$out_interim/log_detect_pcands.txt";
my $cnct_nt_freqs = join (',', @NT_FREQS);
my $cnct_TN_params = join (',', $ALPHA1, $ALPHA2, $BETA);

my $options_det_pcands = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND --MAX_WD_SIZE=$MAX_WD_SIZE_PCAND --THRSH_PV=$THRSH_PV_PCAND --IF_MODPV=$IF_MODPV_PCAND --THRSH_PV0=$THRSH_PV0_PCAND --BASE_FREQS=$cnct_nt_freqs  --TN_PARAMS=$cnct_TN_params";

if ($TEST_LEVEL_MAIN >1) { # (Pre-8) #

	print STDOUT "The options to be used:\n\n";
	print STDOUT "    $options_det_pcands .\n";
	print STDOUT "\n\n";

	print STDOUT "The output files will be put into: {Out_dir} = $outdir_pcands .\n";
	print STDOUT " ... Especially, the main output will be recorded in: {Out_dir}/$name_ofile_pc .\n\n";
	print STDOUT "The progress of the script will be logged in: $log_det_pcands .\n";
	print STDOUT "\n\n";

} # END of "if ($TEST_LEVEL_MAIN >1) { # (Pre-8) # ...}"

# my ($in_msa_file, $dawg_ctrl_file, $indir_mfacs, $toptop_out_dir, $path_anex_1wd, $path_det_pcands) = @ARGV[0 .. 5];
# my $intree_file = $dawg_ctrl_file;

unless (0 == system ("(time $exec_perl $path_det_pcands $out_new_aln $intree_file $outdir_pcands $options_det_pcands) > $log_det_pcands 2>&1") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to run the script, \"$path_det_pcands\" via the executable '$exec_perl' !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... Finished Running the script, \"$path_det_pcands\" via the executable '$exec_perl' ...\n\n") if ($TEST_LEVEL_MAIN>1);

my ($list_mgd_pcands, $maxspan_mgd_pcand) = readin_list_mgd_pcands ($outfile_pcands);
unless (defined $list_mgd_pcands) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_mgd_pcands' FAILED on the input file, \"$outfile_pcands\" !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'readin_list_mgd_pcands' Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);

	## END of "(Pre 8) Detect Purge-Candidate Regions." ##


	## (Pre 8') Select Significant Windows within Purge-Candidate Regions. ## (Added on 2019/06/20.)

##(print STDOUT "\n\n<<< Preliminary 8': Select among Significant Windows within each Purge-Candidate Region, via '' . >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1);
(print STDOUT "\n\n<<< Preliminary 8': Select among Significant Windows (for detecting purge-candidates) within each merged Purge-Candidate Region, via 'select_sgn_wds_in_pcand' . >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1); # CORRECTED on 2020/05/07.

unless ( select_sgn_wds_in_pcands (@{$list_mgd_pcands}) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'select_sgn_wds_in_pcands' FAILED on the input file, \"$outfile_pcands\" !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'select_sgn_wds_in_pcands' Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);


if ($TEST_LEVEL_MAIN >2) { # (Pre-8') #

	my $name_ofile_pc_red = 'list_purge_candidates_reduced.txt';
	my $outfile_pcands_red = "$outdir_pcands/$name_ofile_pc_red";

	print STDOUT "\n", '[[ Contents of @{$list_mgd_pcands} after ', "'select_sgn_wds_in_pcands' ]]\n\n";
	print STDOUT " ... will be recorded in '$outfile_pcands_red', via 'print_out_list_mgd_pcands' ...\n\n";

	unless ( print_out_list_mgd_pcands ($outdir_pcands, $name_ofile_pc_red, @{$list_mgd_pcands}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_list_mgd_pcands', Ended ...\n";
	print STDOUT "\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre-8') # ...}"

	## END of "(Pre 8') Select Significant Windows within Purge-Candidate Regions." ##

	## (Pre 9) Examine WHICH Sliding Sub-Windows overlap each (merged) Purge-Candidate Region. ## (Added on 2019/06/20&21.)

(print STDOUT "\n\n<<< Preliminary 9:  Examine WHICH Sliding Sub-Windows overlap each (merged) Purge-Candidate Region, via 'chk_ovlp_pcands_vs_sliding_wds'.  >>>\n\n\n")  if ($TEST_LEVEL_MAIN>1);

my ($list_mgd_pcands_ovlp_sbwds, $off_sbwds_but_ovlp_wds, $off_wds_but_ovlp_wframes, $off_wframes) 
  = chk_ovlp_pcands_vs_sliding_wds (@{$list_mgd_pcands}, 0, 1, $maxspan_mgd_pcand, @{$window_frames}, 0, 1, @{$set_windows}, @{$set_subwindows}, 0, 1);
unless (defined $list_mgd_pcands_ovlp_sbwds) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'chk_ovlp_pcands_vs_sliding_wds' FAILED !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'chk_ovlp_pcands_vs_sliding_wds' Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);


if ($TEST_LEVEL_MAIN > 2) { # (Pre 9) #

	my $outdir_ovlp = "$outdir_pcands/Ovlp_w_Sliding_Wds";

	print STDOUT "\n[[ The Outputs of 'chk_ovlp_pcands_vs_sliding_wds' ]]\n\n";
	print STDOUT " ... will be recorded under the directory, \"$outdir_ovlp/\",\n";
	print STDOUT "       via 'print_out_chk_ovlp_pcands_vs_sliding_wds' ...\n\n";

	unless ( print_out_chk_ovlp_pcands_vs_sliding_wds ($outdir_ovlp, @{$list_mgd_pcands_ovlp_sbwds}, @{$off_sbwds_but_ovlp_wds}, @{$off_wds_but_ovlp_wframes}, @{$off_wframes}) ) { 
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_chk_ovlp_pcands_vs_sliding_wds' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_chk_ovlp_pcands_vs_sliding_wds', Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN > 2) { # (Pre 9) # ...}"


	## END of "(Pre 9) Examine WHICH Sliding Sub-Windows overlap each (merged) Purge-Candidate Region." ##


	## (Pre 10) Compute some auxiliary data necessary for the MAIN process. ## (ADDED on 2019/06/23.)

my ($nodeids_in_msa, $nodeid2msa_row) = associate_nodeids_w_msa_rows (@{$seqnames}, %name2id);
unless (defined $nodeid2msa_row) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'associate_nodeids_w_msa_rows' FAILED !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'associate_nodeids_w_msa_rows' Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);

if ($TEST_LEVEL_MAIN >2) { # (Pre 10) #

	print STDOUT "\n[[ The Outputs of 'associate_nodeids_w_msa_rows' ]]\n\n";
	print STDOUT '@{$nodeids_in_msa} = (', join (', ', @{$nodeids_in_msa}), ") .\n\n";
	print STDOUT '%{$nodeid2msa_row} = ( # $node_id => $msa_row ,', "\n";
	foreach my $nodeid (sort {$a <=> $b} keys %{$nodeid2msa_row}) {
		print STDOUT "\t", $nodeid, ' => ', $nodeid2msa_row->{$nodeid}, " ,\n";
	}
	print STDOUT "\t) .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 10) # ...}"

	## END of "(Pre 10) Compute some auxiliary data necessary for the MAIN process." ##

	## (Pre 11) Copy the new input MSA. ## (Added in major revision on 2019/07/04.)

my @cp_new_in_msa = ();
foreach my $row (@{$new_in_msa}) {
	my $cp_row = ${$row};
	push @cp_new_in_msa, \$cp_row;
}
(print STDOUT "\n ... Copied the new input MSA ...\n\n") if ($TEST_LEVEL_MAIN >1);

	## End of "(Pre 11) Copy the new input MSA." ##


	##### (MAIN 1) Perform the 1st Sliding-Window Analysis. #####

my $ct_wds = @{$set_windows};
my $ct_sbwds = @{$set_subwindows};

(print STDOUT "\n\n<<<<< MAIN 1:  Perform the 1st Sliding-Window Analysis (w/ #{windows}=$ct_wds, #{sub-windows}=$ct_sbwds).  >>>>>\n\n\n")  if ($TEST_LEVEL_MAIN >0); # Modified on 2019/07/04.
##(print STDOUT "\n\n<<<<< MAIN:  Perform the Sliding-Window Analysis (w/ #{windows}=$ct_wds, #{sub-windows}=$ct_sbwds).  >>>>>\n\n\n")  if ($TEST_LEVEL_MAIN >0);

=MOVED UP, to the block specifying output directories.
my $top_outdir_main = $top_outputs; # MODIFIED (on 2019/06/25). #
##my $top_outdir_main = "$top_outputs/Sliding_Window_Analysis";
my $logdir = "$top_outdir_main/LogFiles"; # Added on 2019/06/24. #
=cut # End of "MOVED UP, to the block specifying output directories."



print STDOUT "\n\n";

print STDOUT "#\n" x 3;
print STDOUT "# [INFORMATION '$program_name'] This script DOES SKIP the MAIN 1 Process ALTOGETHER !!\n";
print STDOUT "#\n" x 3;
print STDOUT "\n\n";



my $indx_sbwd = 0; # Index of the sub-window to be handled now. #

my $beg_prev_wd = $beg_in_msa;

=DISABLED in creation of 'anex.ver0.7_reduced.pl' (on 2020/05/21). (MAIN 1)

for (my $wd=0; $wd < $ct_wds; $wd++) { # 1st MAIN outer for-loop (over the windows). #

		# (A) Basic information on the window. #

	my ($beg_wd, $end_wd, $invlvd_gpd_segs, $ct_gbs_in_wd, $fr_accom_wd) = @{$set_windows->[$wd]};
	my $list_gblocks_in_wd = $set_lists_gblocks_in_wds->[$wd];
	my $ct_gblocks = @{$list_gblocks_in_wd};
	my ($beg_fr, $end_fr) = @{$window_frames->[$fr_accom_wd]};

    (print STDOUT "{{{ The $wd th window ([beg, end] = [$beg_wd, $end_wd], #{gap-blocks}=$ct_gblocks) (within the $fr_accom_wd th frame, i.e., [$beg_fr, $end_fr]) }}}\n\n") if ($TEST_LEVEL_MAIN >1);

	my @mpcands_ovlp_wd = @{$off_sbwds_but_ovlp_wds->[$wd]}; # Added in REVISION (on 2019/07/03).

		# (B) Create a working directory. #

	my $name_odir_main = sprintf ("Window%05d", $wd);
	my $outdir_main = "$top_outdir_main/$name_odir_main";
	my $outdir_main_in = "$outdir_main/Inputs";

	unless ( -d $outdir_main_in )  {
	  unless (0 == system ("mkdir -p $outdir_main_in") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory \"$outdir_main_in\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir_main_in ) {
	  unless (0 == system (" chmod -R u+w $outdir_main_in " ) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory \"$outdir_main_in\": $! !\n\n";
	    return 0;
	  }
	}
    (print STDOUT "\n ... The output directory, \"$outdir_main\", is READY!! ...\n\n") if ($TEST_LEVEL_MAIN >1);

		## RESTART FROM HERE ##

		# (C) Find & handle sub-windows that belong to the window, but that have not been handled yet. #

	while (($indx_sbwd < $ct_sbwds) and ($set_subwindows->[$indx_sbwd][0] < $beg_wd)) { # 1st middle while-loop. #
	    print STDOUT "[WARNING '$program_name'] The $indx_sbwd th sub-window does NOT belong to ANY windows, hence will be SKIPPED: $! !\n\n";
	    $indx_sbwd++;
	} # End of the 1st middle while-loop. #

	my @sbjct_sbwds = ();
	while (($indx_sbwd < $ct_sbwds) and ($set_subwindows->[$indx_sbwd][1] <= $end_wd)) { # 2nd middle while-loop. #
	    push @sbjct_sbwds, $indx_sbwd;
	    $indx_sbwd++;
	} # End of the 2nd middle while-loop. #


   (print STDOUT "The sub-windows to be handled along with this window: ", join (', ', @sbjct_sbwds), " .\n\n") if ($TEST_LEVEL_MAIN >1);

#=MOVED to below "foreach my $sbwd (@sbjct_sbwds) {...}" (on 2019/07/03).
#	my ($left_limit_wd, $right_limit_wd) = determine_limits_window ($wd, @{$set_windows}, @{$window_frames}, @{$new_set_cls_gpbs});
#	unless (defined $right_limit_wd) {
#		print STDOUT "[ERROR '$program_name'] The subroutine, 'determine_limits_window', FAILED on the $wd th window: $! !\n";
#		print STDERR "[ERROR '$plscript_name'] $! !\n";
#		die ;
#	}
#   (print STDOUT "\n ... Subroutine, 'determine_limits_window', Ended on the $wd th window (w/ (\$left_limit_wd, \$right_limit_wd) = ($left_limit_wd, $right_limit_wd)) ...\n\n") if ($TEST_LEVEL_MAIN >2);
#
#=cut # END of "MOVED to below "foreach my $sbwd (@sbjct_sbwds) {...}" (on 2019/07/03)."

	foreach my $sbwd (@sbjct_sbwds) { # 1st middle foreach-loop (over subject sub-windows). #

	    my ($beg_sbwd, $end_sbwd, $invlvd_gpd_segs2, $ct_gbs_in_sbwd, $fr_accom_sbwd) = @{$set_subwindows->[$sbwd]};
	    my $list_gblocks_in_sbwd = $set_lists_gblocks_in_sbwds->[$sbwd];
	    my $ct_gblocks2 = @{$list_gblocks_in_sbwd};
	    my ($beg_fr2, $end_fr2) = @{$window_frames->[$fr_accom_sbwd]};

    (print STDOUT ' ' x 3, "{{ The $sbwd th sub-window ([beg, end] = [$beg_sbwd, $end_sbwd], #{gap-blocks}=$ct_gblocks2) (within the $fr_accom_sbwd th frame, i.e., [$beg_fr2, $end_fr2]) }}\n\n") if ($TEST_LEVEL_MAIN >1);


			# (a) Create a working sub-directory. #

	    my $name_odir_sbmain = sprintf ("Sbwd%05d", $sbwd);
	    my $outdir_sbmain = "$outdir_main/$name_odir_sbmain";
	    my $outdir_sbmain_in = "$outdir_sbmain/Inputs";

	    unless ( -d $outdir_sbmain_in )  {
	      unless (0 == system ("mkdir -p $outdir_sbmain_in") ) {
	        print STDOUT "[ERROR '$program_name'] FAILED to create the directory \"$outdir_sbmain_in\": $! !\n\n";
	        return 0;
	      }
	    }
	    unless ( -w $outdir_sbmain_in ) {
	      unless (0 == system (" chmod -R u+w $outdir_sbmain_in " ) ) {
	        print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory \"$outdir_sbmain_in\": $! !\n\n";
	        return 0;
	      }
	    }
    (print STDOUT "\n ... The output sub-directory, \"$outdir_sbmain\", is READY!! ...\n\n") if ($TEST_LEVEL_MAIN >1);


	    my ($left_limit_sbwd, $right_limit_sbwd) = determine_limits_window ($sbwd, @{$set_subwindows}, @{$window_frames}, @{$new_set_cls_gpbs});
	    unless (defined $right_limit_sbwd) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'determine_limits_window', FAILED on the $sbwd th sub-window: $! !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	    }
   (print STDOUT ' ' x 3, "\n ... Subroutine, 'determine_limits_window', Ended on the $sbwd th sub-window (w/ (\$left_limit_sbwd, \$right_limit_sbwd) = ($left_limit_sbwd, $right_limit_sbwd)) ...\n\n") if ($TEST_LEVEL_MAIN >2);


			# (b) Examine the (merged) purge-candidate regions that overlap the sub-window. #

	    my @mpcands_ovlp_sbwd = @{$list_mgd_pcands_ovlp_sbwds->[$sbwd]};

	    push @mpcands_ovlp_wd, @mpcands_ovlp_sbwd; # Added in REVISION (on 2019/07/03). #

    (print STDOUT ' ' x 3,  "The purge-candidate regions overlapping this sub-window: ", join (', ', @mpcands_ovlp_sbwd), " .\n\n") if ($TEST_LEVEL_MAIN >2);

	    my @mpcands_incorporated_sbwd = ();

	    foreach my $pc (@mpcands_ovlp_sbwd) { # Inner foreach-loop (over overlapping purge-candidate regions). #

		my ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_included) = @{$list_mgd_pcands->[$pc]};
		my $ct_cnst_sites = @{$constituent_sites};

    (print STDOUT ' ' x 5, "{ The $pc th purge-candidate region ([beg, end] = [$beg_pc, $end_pc], {constituent_sites}=(", join (',', @{$constituent_sites}), ") (#{sites}=$ct_cnst_sites)) }\n\n") if ($TEST_LEVEL_MAIN >2); 

		my @sites_wi_limits = ();
		foreach my $site (@{$constituent_sites}) { # Innermost foreach-loop (over constituent sites). #
		    if (($left_limit_sbwd <= $site) and ($site <= $right_limit_sbwd)) {
			push @sites_wi_limits, $site;
		    }
		} # End of the innermost foreach-loop (over constituent sites). #

		my $ct_wi_limits = @sites_wi_limits;

    (print STDOUT ' ' x 5, "{The constituent sites within the limits}= (", join (',', @sites_wi_limits), ") (#{sites}=$ct_wi_limits) .\n\n") if ($TEST_LEVEL_MAIN >2);

		if (($TEST_LEVEL_MAIN>1) and ($ct_wi_limits < $ct_cnst_sites/2)) {
		    print STDOUT ' ' x 5, "[WARNING '$program_name'] Only less than half(=$ct_wi_limits) of the constituent sites (#=$ct_cnst_sites) are within the limit of the $sbwd th sub-window, hence IGNORE this ($pc th) purge-candidate region  ...\n\n";
		    next;
		}

		push @mpcands_incorporated_sbwd, $pc;

			# If the current interval of the window (or the sub-window) does NOT accommodate all @sites_wi_limits,
			# expand the window (or the sub-window).

		my ($lm_site_wi, $rm_site_wi) = @sites_wi_limits[0, $#sites_wi_limits];

		if ($lm_site_wi < $beg_wd) {
		    print STDOUT "[WARNING '$program_name'] The left-most constituent site within limits (=$lm_site_wi) is on the left of \$beg_wd(=$beg_wd) !! ... Thus, RESET: \$beg_wd (of the $wd th window) = \$beg_sbwd (of the $sbwd th sub-window) = $lm_site_wi ...\n\n";
		    $beg_wd = $set_windows->[$wd][0] = $lm_site_wi;
		    $beg_sbwd = $set_subwindows->[$sbwd][0] = $lm_site_wi;

		} elsif ($lm_site_wi < $beg_sbwd) {
		    print STDOUT "[WARNING '$program_name'] The left-most constituent site within limits (=$lm_site_wi) is on the left of \$beg_sbwd(=$beg_sbwd) !! ... Thus, RESET: \$beg_sbwd (of the $sbwd th sub-window) = $lm_site_wi ...\n\n";
		    $beg_sbwd = $set_subwindows->[$sbwd][0] = $lm_site_wi;

		} # End of "if ($lm_site_wi < $beg_wd) {...} elsif (...) {...}"

		if ($end_wd < $rm_site_wi) {
		    print STDOUT "[WARNING '$program_name'] The right-most constituent site within limits (=$rm_site_wi) is on the right of \$end_wd(=$end_wd) !! ... Thus, RESET: \$end_wd (of the $wd th window) = \$end_sbwd (of the $sbwd th sub-window) = $rm_site_wi ...\n\n";
		    $end_wd = $set_windows->[$wd][1] = $rm_site_wi;
		    $end_sbwd = $set_subwindows->[$sbwd][1] = $rm_site_wi;

		} elsif ($end_sbwd < $rm_site_wi) {
		    print STDOUT "[WARNING '$program_name'] The right-most constituent site within limits (=$rm_site_wi) is on the right of \$end_sbwd(=$end_sbwd) !! ... Thus, RESET: \$end_sbwd (of the $sbwd th sub-window) = $rm_site_wi ...\n\n";
		    $end_sbwd = $set_subwindows->[$sbwd][1] = $rm_site_wi;

		} # End of "if ($end_wd < $rm_site_wi) {...} elsif (...) {...}"


#       @{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#       $beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#       $br is the ID of the branch across which the "purge" is suspected within the region,
#       @msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#       @constituent_sites stores the sites that constitute the region,
#       $pv is the P-value measuring the deviation from the specified substutution model,
#       $pv0 is the P-value measuring the deviation from the random-match model,
#       and @sgn_wds_included is the significant windows included in the region;
#       more precisely, 
#       @{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#       @sites_in_wd stores the sites in the window,
#       $pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;

	    } # End of the inner foreach-loop (over overlapping purge-candidate regions). #


			# (c) Record necessary information on the (merged) purge-candidates to be incorporated. # (Added on 2019/06/22.)

    (print STDOUT "The sub-window will incorporate the merged purge-candidates:", join (',', @mpcands_incorporated_sbwd), " .\n\n") if ($TEST_LEVEL_MAIN >2);

		# ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_included) = @{$list_mgd_pcands->[$pc]};
		# push @mpcands_incorporated, $pc;

	    my @list_mpcands_icptd_sbwd = @{$list_mgd_pcands}[@mpcands_incorporated_sbwd];
	    my $name_ofile_pc_rlv = 'list_rlv_pcands.txt';

	    unless ( print_out_list_mgd_pcands ($outdir_sbmain_in, $name_ofile_pc_rlv, @list_mpcands_icptd_sbwd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the $sbwd th sub-window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	    }
    (print STDOUT "\n ... The subroutine, 'print_out_list_mgd_pcands', Ended RECORDING the purge-candidate regions (", join (',', @mpcands_incorporated_sbwd), "), which are to be incorporated into the $sbwd th sub-window ...\n") if ($TEST_LEVEL_MAIN >1);
    ##(print STDOUT " ... into the file, \"$outdir_sbmain/$name_ofile_pc_rlv\" ...\n") if ($TEST_LEVEL_MAIN >2);
    (print STDOUT " ... into the file, \"$outdir_sbmain_in/$name_ofile_pc_rlv\" ...\n") if ($TEST_LEVEL_MAIN >2); # CORRECTED on 2020/05/08. #
    (print STDOUT "\n\n") if ($TEST_LEVEL_MAIN >1);


			# (d) Prepare & Record information on the sub-window necessary for the sliding-window analysis. # (Added on 2019/06/23.)

	    my ($set_clms0_sbwd) = mk_init_set_columns0 (@{$nodeids_in_msa}, $CHAR_ABS, $beg_sbwd, $end_sbwd, @{$invlvd_gpd_segs2}, @{$new_set_cls_gpbs}, @{$new_init_psm_hsts_for_cls});
	    unless (defined $set_clms0_sbwd) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_init_set_columns0' FAILED for the $sbwd th sub-window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	    }
    (print STDOUT "\n ... The subroutine, 'mk_init_set_columns0'. Ended for the $sbwd th sub-window ...\n\n") if ($TEST_LEVEL_MAIN >1);

	    	# my ($beg_sbwd, $end_sbwd, $invlvd_gpd_segs2, $ct_gbs_in_sbwd, $fr_accom_sbwd) = @{$set_subwindows->[$sbwd]};

	    my $if_on_le_msa2 = ($beg_sbwd == $beg_in_msa) ? 1 : 0;
	    my $if_on_re_msa2 = ($end_sbwd == $end_in_msa) ? 1 : 0;
	    my $fname_set_clms0 = 'set_columns0.txt';

	    unless ( print_out_set_clms ($outdir_sbmain_in, $fname_set_clms0, @{$set_clms0_sbwd}, $beg_sbwd, $end_sbwd, $if_on_le_msa2, $if_on_re_msa2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the $sbwd th sub-window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	    }
    (print STDOUT "\n ... The subroutine, 'print_out_set_clms', Ended RECORDING information (on the $sbwd th sub-window) into the output file, \"$outdir_sbmain_in/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	    my $fname_list_gbs0 = 'list_gblocks0.txt';

	    unless ( print_out_list_gblocks_in_wd ($outdir_sbmain_in, $fname_list_gbs0, @{$list_gblocks_in_sbwd}, %{$nodeid2msa_row}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblocks_in_wd' FAILED for the $sbwd th sub-window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	    }
    (print STDOUT "\n ... The subroutine, 'print_out_list_gblocks_in_wd', Ended RECORDING information (on the $sbwd th sub-window) into the output file, \"$outdir_sbmain_in/$fname_list_gbs0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	    # my $list_gblocks_in_sbwd = $set_lists_gblocks_in_sbwds->[$sbwd];


    (print STDOUT "\n --- Finished Processing the $sbwd th sub-window ---\n\n") if ($TEST_LEVEL_MAIN >1);

	    ## my $outdir_sbmain = "$outdir_main/$name_odir_sbmain";
	    ## my $outdir_sbmain_in = "$outdir_sbmain/Inputs";

	} # End of the 1st middle foreach-loop (over subject sub-windows). #


		# (D) Examine the purge-candidate regions overlapping the subject window,
		#	and RESET the window boundaries, if necessary.  # (ADDED in REVISION on 2019/07/03.) #

			# MOVED from above "foreach my $sbwd (@sbjct_sbwds) {...}" (on 2019/07/03).

	my ($left_limit_wd, $right_limit_wd) = determine_limits_window ($wd, @{$set_windows}, @{$window_frames}, @{$new_set_cls_gpbs});
	unless (defined $right_limit_wd) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'determine_limits_window', FAILED on the $wd th window: $! !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
   (print STDOUT "\n ... Subroutine, 'determine_limits_window', Ended on the $wd th window (w/ (\$left_limit_wd, \$right_limit_wd) = ($left_limit_wd, $right_limit_wd)) ...\n\n") if ($TEST_LEVEL_MAIN >2);

			# END of "MOVED from above "foreach my $sbwd (@sbjct_sbwds) {...}" (on 2019/07/03)."

			# ADDED in REVISION (on 2019/07/03), based on the "foreach my $pc (@mpcands_ovlp_sbwd) {...}" above. #

	my $nr_mpcands_ovlp_wd = nr_nmarray (@mpcands_ovlp_wd);
	 @mpcands_ovlp_wd = @{$nr_mpcands_ovlp_wd};

    (print STDOUT "The purge-candidate regions overlapping this ($wd th) window: ", join (', ', @mpcands_ovlp_wd), " .\n\n") if ($TEST_LEVEL_MAIN >2);

	my @mpcands_incorporated_wd = ();

	foreach my $pc (@mpcands_ovlp_wd) { # 2nd middle foreach-loop (over overlapping purge-candidate regions). #

	    my ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_included) = @{$list_mgd_pcands->[$pc]};
	    my $ct_cnst_sites = @{$constituent_sites};

    (print STDOUT ' ' x 2, "{ The $pc th purge-candidate region ([beg, end] = [$beg_pc, $end_pc], {constituent_sites}=(", join (',', @{$constituent_sites}), ") (#{sites}=$ct_cnst_sites)) }\n\n") if ($TEST_LEVEL_MAIN >2); 

	    my @sites_wi_limits = ();
	    foreach my $site (@{$constituent_sites}) { # Inner foreach-loop (over constituent sites). #
		if (($left_limit_wd <= $site) and ($site <= $right_limit_wd)) {
			push @sites_wi_limits, $site;
		}
	    } # End of the inner foreach-loop (over constituent sites). #

	    my $ct_wi_limits = @sites_wi_limits;

    (print STDOUT ' ' x 2, "{The constituent sites within the limits}= (", join (',', @sites_wi_limits), ") (#{sites}=$ct_wi_limits) .\n\n") if ($TEST_LEVEL_MAIN >2);


	    if (($TEST_LEVEL_MAIN>1) and ($ct_wi_limits < $ct_cnst_sites/2)) {
		print STDOUT ' ' x 2, "[WARNING '$program_name'] Only less than half(=$ct_wi_limits) of the constituent sites (#=$ct_cnst_sites) are within the limit of the $wd th window, hence IGNORE this ($pc th) purge-candidate region  ...\n\n";
		next;
	    }

	    push @mpcands_incorporated_wd, $pc;

			# If the current interval of the window (or the sub-window) does NOT accommodate all @sites_wi_limits,
			# expand the window.

	    my ($lm_site_wi, $rm_site_wi) = @sites_wi_limits[0, $#sites_wi_limits];

	    if ($lm_site_wi < $beg_wd) {
		    print STDOUT "[WARNING '$program_name'] The left-most constituent site within limits (=$lm_site_wi) is on the left of \$beg_wd(=$beg_wd) !! ... Thus, RESET: \$beg_wd (of the $wd th window) = $lm_site_wi ...\n\n";
		    $beg_wd = $set_windows->[$wd][0] = $lm_site_wi;

	    } # End of "if ($lm_site_wi < $beg_wd) {...}"

	    if ($end_wd < $rm_site_wi) {
		    print STDOUT "[WARNING '$program_name'] The right-most constituent site within limits (=$rm_site_wi) is on the right of \$end_wd(=$end_wd) !! ... Thus, RESET: \$end_wd (of the $wd th window) = $rm_site_wi ...\n\n";
		    $end_wd = $set_windows->[$wd][1] = $rm_site_wi;

	    } # End of "if ($end_wd < $rm_site_wi) {...}"

	} # End of the 2nd middle foreach-loop (over overlapping purge-candidate regions). #

			# End of "ADDED in REVISION (on 2019/07/03), based on the "foreach my $pc (@mpcands_ovlp_sbwd) {...}" above." #
			# ADDED in REVISION (on 2019/07/03), based of the "(c) Record necessary information on the (merged) purge-candidates to be incorporated" above.

    (print STDOUT "The ($wd th) window will incorporate the merged purge-candidates:", join (',', @mpcands_incorporated_wd), " .\n\n") if ($TEST_LEVEL_MAIN >2);

	my @list_mpcands_icptd_wd = @{$list_mgd_pcands}[@mpcands_incorporated_wd];
	my $name_ofile_pc_rlv = 'list_rlv_pcands.txt';

	unless ( print_out_list_mgd_pcands ($outdir_main_in, $name_ofile_pc_rlv, @list_mpcands_icptd_wd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the $wd th window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_list_mgd_pcands', Ended RECORDING the purge-candidate regions (", join (',', @mpcands_incorporated_wd), "), which are to be incorporated into the $wd th window ...\n") if ($TEST_LEVEL_MAIN >1);
    ##(print STDOUT " ... into the file, \"$outdir_main/$name_ofile_pc_rlv\" ...\n") if ($TEST_LEVEL_MAIN >2);
    (print STDOUT " ... into the file, \"$outdir_main_in/$name_ofile_pc_rlv\" ...\n") if ($TEST_LEVEL_MAIN >2); # CORRECTED on 2020/05/08. #
    (print STDOUT "\n\n") if ($TEST_LEVEL_MAIN >1);

			# END of "ADDED in REVISION (on 2019/07/03), based of the "(c) Record necessary information on the (merged) purge-candidates to be incorporated" above."


		# [ (D) Print out the segment of the new input MSA that corresponds to the current window. # (Added on 2019/06/24.) ] ... OBSOLETE as of 2019/07/03.
		# (E) Print out the segment of the new input MSA that corresponds to the current window. # (REVISED on 2019/07/03.)

	my @sub_in_msa = ();
	for (my $k=0; $k <$ct_seqs; $k++) {
	    substr (${$new_in_msa->[$k]}, 0, $beg_wd - $beg_prev_wd, ''); # Chop off the left-margin.
	    my $sbseq = substr (${$new_in_msa->[$k]}, 0, $end_wd - $beg_wd +1);
	    $sub_in_msa[$k] = \$sbseq;
	}
	my $fname_sbaln = "msa_in_wd.txt";
	my $ofile_sbaln = "$outdir_main_in/$fname_sbaln";

	my $fh_sbaln = open_output_file ($ofile_sbaln, 0);
	unless (defined $fh_sbaln) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_sbaln\", to write the new MSA (within the $wd th window) into !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}

	unless ( pr_ODEN_input (@{$seqnames}, @sub_in_msa, $DIS_LEN, *${$fh_sbaln}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_ODEN_input'  FAILED to print out the new MSA (within the $wd th window) into \"$ofile_sbaln\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... 'pr_ODEN_input' Ended, to write the new MSA (within the $wd th window) into the file, \"$ofile_sbaln\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

	close ($fh_sbaln);

	$beg_prev_wd = $beg_wd; # Update $beg_prev_wd.


		# [ (E) Prepare & Record other information on the window necessary for the sliding-window analysis. # (Added on 2019/06/24.) ] ... OBSOLETE as of 2019/07/03.
		# (F) Prepare & Record other information on the window necessary for the sliding-window analysis. # (REVISED on 2019/07/03.)

	my ($set_clms0_wd) = mk_init_set_columns0 (@{$nodeids_in_msa}, $CHAR_ABS, $beg_wd, $end_wd, @{$invlvd_gpd_segs}, @{$new_set_cls_gpbs}, @{$new_init_psm_hsts_for_cls});
	    unless (defined $set_clms0_wd) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_init_set_columns0' FAILED for the $wd th window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'mk_init_set_columns0'. Ended for the $wd th window ...\n\n") if ($TEST_LEVEL_MAIN >1);

	## my ($beg_wd, $end_wd, $invlvd_gpd_segs, $ct_gbs_in_wd, $fr_accom_wd) = @{$set_windows->[$wd]};


	my $if_on_le_msa = ($beg_wd == $beg_in_msa) ? 1 : 0;
	my $if_on_re_msa = ($end_wd == $end_in_msa) ? 1 : 0;
	my $fname_set_clms0 = 'set_columns0.txt';

	unless ( print_out_set_clms ($outdir_main_in, $fname_set_clms0, @{$set_clms0_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the $wd th window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_set_clms', Ended RECORDING information (on the $wd th window) into the output file, \"$outdir_main_in/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	my $fname_list_gbs0 = 'list_gblocks0.txt';

	unless ( print_out_list_gblocks_in_wd ($outdir_main_in, $fname_list_gbs0, @{$list_gblocks_in_wd}, %{$nodeid2msa_row}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblocks_in_wd' FAILED for the $wd th window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_list_gblocks_in_wd', Ended RECORDING information (on the $wd th window) into the output file, \"$outdir_main_in/$fname_list_gbs0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	## my $list_gblocks_in_wd = $set_lists_gblocks_in_wds->[$wd];


		# [ (F) Perform a single-window analysis, via the script, "$path_anex_1wd". # (Added on 2019/06/24.) ] ... OBSOLETE as of 2019/07/03.
		# (G) Perform a single-window analysis, via the script, "$path_anex_1wd". # (REVISED on 2019/07/03.)

	my $log_anex_1wd = "$logdir/log_anex_1wd.$name_odir_main.txt";

	my $options_anex_1wd = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND --HF_WDTH_SHIFT=$HF_WDTH_SHIFT "; # MODIFIED on 2019/06/29.
	##my $options_anex_1wd = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND ";

	$options_anex_1wd .= " --XLEVEL=$XLEVEL_1WD "; # ADDED in REVISION (on 2019/11/15). #

	$options_anex_1wd .= $OPTION_4_TEST_W_CMSS; # ADDED in REVISION (on 2019/11/14). #

## my $options_det_pcands = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND --MAX_WD_SIZE=$MAX_WD_SIZE_PCAND --THRSH_PV=$THRSH_PV_PCAND --IF_MODPV=$IF_MODPV_PCAND --THRSH_PV0=$THRSH_PV0_PCAND --BASE_FREQS=$cnct_nt_freqs  --TN_PARAMS=$cnct_TN_params"; ## None of them seems necessary for $path_anex_1wd...

	unless (0 == system ("(time $exec_perl $path_anex_1wd $outdir_main $dawg_ctrl_file $indir_mfacs $options_anex_1wd ) > $log_anex_1wd 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to run the script, \"$path_anex_1wd\" (on the $wd th window), via the executable '$exec_perl' (w/ options=\"$options_anex_1wd\") !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
    (print STDOUT "\n ... Finished Running the script, \"$path_anex_1wd\" for the $wd th window, via the executable '$exec_perl' (w/ options= \"$options_anex_1wd\") ...\n\n") if ($TEST_LEVEL_MAIN >1);


   if ($IF_ARCHIVE_WD_RESULTS) { # Added on 2019/06/24. #

		# (G) (Optional) Archive the Working Directory, $outdir_main. #

     (print STDOUT " ... Because \$IF_ARCHIVE_WD_RESULTS = $IF_ARCHIVE_WD_RESULTS, Archive the Working Directory, \"$outdir_main/\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

			# MAJOR REVISION in DEBUGGING on 2020/05/09. (1) #

	my $current_dir = $ENV{'PWD'};

	unless ( chdir ($top_outdir_main) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the current directory to \"$top_outdir_main\", via 'chdir' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Changed the current directory to: \"$top_outdir_main\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

	##my $log_archive = "$logdir_name/log_gzip.$name_odir_main.txt";
	my $log_archive = "$logdir_name/log_tar_cpzf.$name_odir_main.txt"; # Corrected on 2020/05/10.

	unless (0 == system ("(tar -cpzf $name_odir_main.tgz $name_odir_main) > $log_archive 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to archive the directory, \"$name_odir_main\", via 'tar -cpzf' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
     (print STDOUT "\n ... Finished archiving the directory, \"$name_odir_main\", via 'tar -cpzf' (& logged in \"$log_archive\")...\n\n") if ($TEST_LEVEL_MAIN >1);

	# my $logdir = "$top_outdir_main/$logdir_name"; # Added on 2019/06/24. #
	# my $outdir_main = "$top_outdir_main/$name_odir_main";

	unless ( chdir ($current_dir) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the current directory back to \"$current_dir\", via 'chdir' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Changed the current directory back to: \"$current_dir\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

#=A BUG found on 2020/05/09.
#	my $log_archive = "$logdir/log_gzip.$name_odir_main.txt";
#	unless (0 == system ("(gzip $outdir_main) > $log_archive 2>&1") ) {
#	
#	    print STDOUT "[ERROR '$program_name'] FAILED to archive the directory, \"$outdir_main\", via 'gzip' !!\n";
#	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
#	    die ;
#	}
#     (print STDOUT "\n ... Finished archiving the directory, \"$outdir_main\", via 'gzip' ...\n\n") if ($TEST_LEVEL_MAIN >1);
#=cut # END of "A BUG found on 2020/05/09."

			# End of "MAJOR REVISION in DEBUGGING on 2020/05/09. (1)" #


			# Added on 2019/07/03. #
	my $log_rm_rf = "$logdir/log_rm_rf.$name_odir_main.txt";
	unless (0 == system ("(rm -rf $outdir_main) > $log_rm_rf 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to remove the directory, \"$outdir_main\", via 'rm -rf' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
     (print STDOUT "\n ... Finished removing the directory, \"$outdir_main\", via 'rm -rf' ...\n\n") if ($TEST_LEVEL_MAIN >1);
			# End of "Added on 2019/07/03." #
   
   } # END of "if ($IF_ARCHIVE_WD_RESULTS) { # Added on 2019/06/24. # ...}"
   ## [ } # End of "if ($IF_ARCHIVE_MAIN_RESULTS) {...}" ] ... A Mistake?

#=DISABLED because it is NO LONGER necessary (on 2019/07/03).
#    if ($TEST_LEVEL_MAIN >2) { # Added on 2019/06/24.
#
#	my $unprocessed_pcands = $off_sbwds_but_ovlp_wds->[$wd];
#	if (@{$unprocessed_pcands}>0) {
#	    print STDOUT "\n[INFORMATION] The purge-candidates, (", join (', ', @{$unprocessed_pcands}), "), OVERLAP this window, but were NOT processed because they do NOT OVERLAP any sub-windows.\n\n";
#	}
#    } # End of "if ($TEST_LEVEL_MAIN >2) {...}"
#=cut # End of "DISABLED because it is NO LONGER necessary (on 2019/07/03)."

    (print STDOUT "\n === Finished Processing the $wd th Window ==\n\n\n") if ($TEST_LEVEL_MAIN >1);

} # End of the 1st MAIN outer for-loop (over the windows). #


if ($indx_sbwd < $ct_sbwds) {
	my @unanalyzed_sbwds = ($indx_sbwd .. $ct_sbwds-1);
	my $ct_unanalyzed = @unanalyzed_sbwds;
	print STDOUT "[WARNING '$program_name'] The following $ct_unanalyzed sub-windows do NOT belong to any windows, hence WERE NOT ANALYZED: ", join (', ', @unanalyzed_sbwds), " : $! !\n";
	print STDOUT "\n\n";
} # End of "if ($indx_sbwd < $ct_sbwds) {...}"


(print STDOUT "\n\n ----- The 1st Sliding-Window Analysis Finished !!! -----\n\n") if ($TEST_LEVEL_MAIN>0); # Modified on 2019/07/04.
##(print STDOUT "\n\n ----- The Sliding-Window Analysis Finished !!! -----\n\n") if ($TEST_LEVEL_MAIN>0);

=cut # END of "DISABLED in creation of 'anex.ver0.7_reduced.pl' (on 2020/05/21). (MAIN 1)"



		### Added in revision on 2019/07/08. ###
unless (defined $path_anex_rev_prg_1wd) {

	print STDOUT "[INFORMATION '$program_name'] The \$path_anex_rev_prg_1wd (specifiable via the command-line option, \"--PATH_REV_PURGE_SCT={path/to/the/script}\") is UNDEFINED !! ... Thus, I will FINISH This SCRIPT withOUT performing \"Reverse-Purge\"s ...\n\n\n";

	(print STDOUT "\n\n ===== ALL the Processes ENDED (w/o \"Reverse-Purge\"s) !!!!! =====\n\n") if ($TEST_LEVEL_MAIN >0);

	exit 0;
}
		### End of "Added in revision on 2019/07/08." ###


	# (Pre 12) Construct Sliding-Windows for "Reverse-Purge"s, each associated with a (merged) purge-candidate region. # (Added on 2019/07/04-05.)

my ($set_pcc_windows, $set_indices_mpcands, $set_lists_gblocks_in_pcc_wds) = mk_pcc_sliding_windows (@{$list_mgd_pcands},0, 1,  @{$window_frames}, 0, 1, @{$new_set_cls_gpbs}, @{$set_lists_gblocks}, $HF_WDTH_SHIFT, $NUM_MAX);
unless (defined $set_lists_gblocks_in_pcc_wds) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_pcc_sliding_windows' FAILED !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}

my $ct_pwds = @{$set_pcc_windows};
(print STDOUT "\n ... The subroutine 'mk_pcc_sliding_windows' Ended (w/ #{PCC windows}=$ct_pwds) ...\n\n") if ($TEST_LEVEL_MAIN >1);

if ($TEST_LEVEL_MAIN > 2) { # (Pre 12) #

	print STDOUT "\n ... Check the outputs of 'mk_pcc_sliding_windows' for Max#{gap-blocks} = $NUM_MAX -2, by performing the subroutine, 'print_out_mk_pcc_sliding_windows' ...\n\n";
	my $outdir_name = "PCC_Sliding_Windows";

	unless ( print_out_mk_pcc_sliding_windows ($out_interim, $outdir_name, @{$window_frames}, @{$set_indices_mpcands}, @{$set_pcc_windows}, @{$set_lists_gblocks_in_pcc_wds}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_mk_pcc_sliding_windows' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_mk_pcc_sliding_windows', Ended ...\n\n";

} # END of "if ($TEST_LEVEL_MAIN > 2) { # (Pre 12) # ...}"


	# End of "(Pre 12) Construct Sliding-Windows for "Reverse-Purge"s, each associated with a (merged) purge-candidate region." #



	##### (MAIN 2) Perform the 2nd Sliding-Window Analysis, exclusively for "Reverse-Purge"s. ##### (Added on 2019/07/04)

(print STDOUT "\n\n<<<<< MAIN 2:  Perform the 2nd Sliding-Window Analysis, exclusively for \"Reverse-Purge\"s (w/ #{PCC windows}=$ct_pwds).  >>>>>\n\n\n")  if ($TEST_LEVEL_MAIN >0); # (Added on 2019/07/04.)

		# Added on 2019/07/06. #
# $top_outdir_main2 = "$top_outputs/PCC_Sliding_Window_Analysis";
# $logdir2 = "$top_outdir_main2/LogFiles";
		# End of "Added on 2019/07/06." #

		# COPIED with substantial MODIFICATIONS from the 1st MAIN for-loop above (on 2019/07/06). #

my $beg_prev_pwd = $beg_in_msa;

for (my $pwd=0; $pwd < $ct_pwds; $pwd++) { # 2nd MAIN outer for-loop (over the PCC windows). #

		# (A) Basic information on the PCC window. #

	my ($beg_pwd, $end_pwd, $invlvd_gpd_segs, $ct_gbs_in_pwd, $fr_accom_pwd) = @{$set_pcc_windows->[$pwd]};
	my $list_gblocks_in_pwd = $set_lists_gblocks_in_pcc_wds->[$pwd];
	my $ct_gblocks = @{$list_gblocks_in_pwd};
	my ($beg_fr, $end_fr) = @{$window_frames->[$fr_accom_pwd]};

	my $indx_mpcand = $set_indices_mpcands->[$pwd]; # Added IN REVISION (on 2019/07/06). #
	my @mpcands_incorporated_pwd = ($indx_mpcand);

#  ($set_pcc_windows, $set_indices_mpcands, $set_lists_gblocks_in_pcc_wds) = mk_pcc_sliding_windows (@{$list_mgd_pcands},0, 1,  @{$window_frames}, 0, 1, @{$new_set_cls_gpbs}, @{$set_lists_gblocks}, $HF_WDTH_SHIFT, $NUM_MAX);


    (print STDOUT "{{{ The $pwd th PCC window ([beg, end] = [$beg_pwd, $end_pwd], #{gap-blocks}=$ct_gblocks) (within the $fr_accom_pwd th frame, i.e., [$beg_fr, $end_fr]) }}}\n\n") if ($TEST_LEVEL_MAIN >1);

		# (B) Create a working directory. #

	my $name_odir_main = sprintf ("PCC_Window%05d", $pwd);
	my $outdir_main = "$top_outdir_main2/$name_odir_main";
	my $outdir_main_in = "$outdir_main/Inputs";

	unless ( -d $outdir_main_in )  {
	  unless (0 == system ("mkdir -p $outdir_main_in") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory \"$outdir_main_in\": $! !\n\n";
	    return 0;
	  }
	}
	unless ( -w $outdir_main_in ) {
	  unless (0 == system (" chmod -R u+w $outdir_main_in " ) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory \"$outdir_main_in\": $! !\n\n";
	    return 0;
	  }
	}
    (print STDOUT "\n ... The output directory, \"$outdir_main\", is READY!! ...\n\n") if ($TEST_LEVEL_MAIN >1);



			# ADDED in REVISION (on 2019/07/03), based of the "(c) Record necessary information on the (merged) purge-candidates to be incorporated" above.

    (print STDOUT "The ($pwd th) PCC window will incorporate the merged purge-candidate(s):", join (',', @mpcands_incorporated_pwd), " .\n\n") if ($TEST_LEVEL_MAIN >2);

	my @list_mpcands_icptd_pwd = @{$list_mgd_pcands}[@mpcands_incorporated_pwd];
	my $name_ofile_pc_rlv = 'list_rlv_pcands.txt';

	unless ( print_out_list_mgd_pcands ($outdir_main_in, $name_ofile_pc_rlv, @list_mpcands_icptd_pwd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the $pwd th PCC window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_list_mgd_pcands', Ended RECORDING the purge-candidate region(s) (", join (',', @mpcands_incorporated_pwd), "), which are to be incorporated into the $pwd th PCC window ...\n") if ($TEST_LEVEL_MAIN >1);
    (print STDOUT " ... into the file, \"$outdir_main/$name_ofile_pc_rlv\" ...\n") if ($TEST_LEVEL_MAIN >2);
    (print STDOUT "\n\n") if ($TEST_LEVEL_MAIN >1);

			# END of "ADDED in REVISION (on 2019/07/03), based of the "(c) Record necessary information on the (merged) purge-candidates to be incorporated" above."


		# (E) Print out the segment of the new input MSA that corresponds to the current PCC window. #

# my @cp_new_in_msa = ();

	my @sub_in_msa = ();
	for (my $k=0; $k <$ct_seqs; $k++) {
	    substr (${$cp_new_in_msa[$k]}, 0, $beg_pwd - $beg_prev_pwd, ''); # Chop off the left-margin.
	    my $sbseq = substr (${$cp_new_in_msa[$k]}, 0, $end_pwd - $beg_pwd +1);
	    $sub_in_msa[$k] = \$sbseq;
	}
	my $fname_sbaln = "msa_in_pcc_wd.txt";
	my $ofile_sbaln = "$outdir_main_in/$fname_sbaln";

	my $fh_sbaln = open_output_file ($ofile_sbaln, 0);
	unless (defined $fh_sbaln) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$ofile_sbaln\", to write the new MSA (within the $pwd th PCC window) into !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}

	unless ( pr_ODEN_input (@{$seqnames}, @sub_in_msa, $DIS_LEN, *${$fh_sbaln}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_ODEN_input'  FAILED to print out the new MSA (within the $pwd th PCC window) into \"$ofile_sbaln\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... 'pr_ODEN_input' Ended, to write the new MSA (within the $pwd th PCC window) into the file, \"$ofile_sbaln\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

	close ($fh_sbaln);

	$beg_prev_pwd = $beg_pwd; # Update $beg_prev_pwd.


		# (F) Prepare & Record other information on the PCC window necessary for the sliding-window analysis. #

	my ($set_clms0_pwd) = mk_init_set_columns0 (@{$nodeids_in_msa}, $CHAR_ABS, $beg_pwd, $end_pwd, @{$invlvd_gpd_segs}, @{$new_set_cls_gpbs}, @{$new_init_psm_hsts_for_cls});
	    unless (defined $set_clms0_pwd) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_init_set_columns0' FAILED for the $pwd th PCC window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'mk_init_set_columns0'. Ended for the $pwd th PCC window ...\n\n") if ($TEST_LEVEL_MAIN >1);

	## ($beg_pwd, $end_pwd, $invlvd_gpd_segs, $ct_gbs_in_pwd, $fr_accom_pwd) = @{$set_pcc_windows->[$pwd]};


	my $if_on_le_msa = ($beg_pwd == $beg_in_msa) ? 1 : 0;
	my $if_on_re_msa = ($end_pwd == $end_in_msa) ? 1 : 0;
	my $fname_set_clms0 = 'set_columns0.txt';

	unless ( print_out_set_clms ($outdir_main_in, $fname_set_clms0, @{$set_clms0_pwd}, $beg_pwd, $end_pwd, $if_on_le_msa, $if_on_re_msa) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the $pwd th PCC window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_set_clms', Ended RECORDING information (on the $pwd th PCC window) into the output file, \"$outdir_main_in/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	my $fname_list_gbs0 = 'list_gblocks0.txt';

	unless ( print_out_list_gblocks_in_wd ($outdir_main_in, $fname_list_gbs0, @{$list_gblocks_in_pwd}, %{$nodeid2msa_row}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblocks_in_wd' FAILED for the $pwd th PCC window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... The subroutine, 'print_out_list_gblocks_in_wd', Ended RECORDING information (on the $pwd th PCC window) into the output file, \"$outdir_main_in/$fname_list_gbs0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

	## my $list_gblocks_in_pwd = $set_lists_gblocks_in_pcc_wds->[$pwd];


		# (G) Perform a single-window analysis, via the script, "$path_anex_rev_purge_1wd". # 

	my $log_anex_rev_prg_1wd = "$logdir2/log_anex_rev_purge_1wd.$name_odir_main.txt";

	my $options_anex_rev_prg_1wd = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND --HF_WDTH_SHIFT=$HF_WDTH_SHIFT ";

	$options_anex_rev_prg_1wd .= " --XLEVEL=$XLEVEL_REV_PRG_1WD "; # ADDED in REVISION (on 2019/11/15). #

	$options_anex_rev_prg_1wd .= $OPTION_4_TEST_W_CMSS ; ## ADDED in REVISION (on 2019/11/14). ##

## my $options_det_pcands = " --MIN_WD_SIZE=$MIN_WD_SIZE_PCAND --MAX_WD_SIZE=$MAX_WD_SIZE_PCAND --THRSH_PV=$THRSH_PV_PCAND --IF_MODPV=$IF_MODPV_PCAND --THRSH_PV0=$THRSH_PV0_PCAND --BASE_FREQS=$cnct_nt_freqs  --TN_PARAMS=$cnct_TN_params"; ## None of them seems necessary for $path_anex_1wd...

	unless (0 == system ("(time $exec_perl $path_anex_rev_prg_1wd $outdir_main $dawg_ctrl_file $indir_mfacs $options_anex_rev_prg_1wd ) > $log_anex_rev_prg_1wd 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to run the script, \"$path_anex_rev_prg_1wd\" (on the $pwd th PCC window), via the executable '$exec_perl' (w/ options=\"$options_anex_rev_prg_1wd\") !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
    (print STDOUT "\n ... Finished Running the script, \"$path_anex_rev_prg_1wd\" for the $pwd th PCC window, via the executable '$exec_perl' (w/ options= \"$options_anex_rev_prg_1wd\") ...\n\n") if ($TEST_LEVEL_MAIN >1);


   if ($IF_ARCHIVE_WD_RESULTS) {

		# (G) (Optional) Archive the Working Directory, $outdir_main. #

     (print STDOUT " ... Because \$IF_ARCHIVE_WD_RESULTS = $IF_ARCHIVE_WD_RESULTS, Archive the Working Directory, \"$outdir_main/\" ...\n\n") if ($TEST_LEVEL_MAIN >1);


			# MAJOR REVISION in DEBUGGING on 2020/05/09. (2) #

	my $current_dir = $ENV{'PWD'};

	unless ( chdir ($top_outdir_main2) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the current directory to \"$top_outdir_main2\", via 'chdir' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Changed the current directory to: \"$top_outdir_main2\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

	##my $log_archive = "$logdir_name/log_gzip.$name_odir_main.txt";
	my $log_archive = "$logdir_name/log_tar_cpzf.$name_odir_main.txt"; # Corrected on 2020/05/10.

	unless (0 == system ("(tar -cpzf $name_odir_main.tgz $name_odir_main) > $log_archive 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to archive the directory, \"$name_odir_main\", via 'tar -cpzf' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
     (print STDOUT "\n ... Finished archiving the directory, \"$name_odir_main\", via 'tar -cpzf' (& logged in \"$log_archive\")...\n\n") if ($TEST_LEVEL_MAIN >1);

	# my $logdir2 = "$top_outdir_main2/$logdir_name";
	# my $outdir_main = "$top_outdir_main2/$name_odir_main";

	unless ( chdir ($current_dir) ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the current directory back to \"$current_dir\", via 'chdir' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n ... Changed the current directory back to: \"$current_dir\" ...\n\n") if ($TEST_LEVEL_MAIN >1);

=A BUG found on 2020/05/09. (2)
	my $log_archive = "$logdir2/log_gzip.$name_odir_main.txt";
	unless (0 == system ("(gzip $outdir_main) > $log_archive 2>&1") ) {
	
	    print STDOUT "[ERROR '$program_name'] FAILED to archive the directory, \"$outdir_main\", via 'gzip' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
     (print STDOUT "\n ... Finished archiving the directory, \"$outdir_main\", via 'gzip' ...\n\n") if ($TEST_LEVEL_MAIN >1);
=cut # End of "A BUG found on 2020/05/09. (2)"

			# END of "MAJOR REVISION in DEBUGGING on 2020/05/09. (2)" #


	my $log_rm_rf = "$logdir2/log_rm_rf.$name_odir_main.txt";
	unless (0 == system ("(rm -rf $outdir_main) > $log_rm_rf 2>&1") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to remove the directory, \"$outdir_main\", via 'rm -rf' !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
     (print STDOUT "\n ... Finished removing the directory, \"$outdir_main\", via 'rm -rf' ...\n\n") if ($TEST_LEVEL_MAIN >1);
   
   } # End of "if ($IF_ARCHIVE_WD_RESULTS) { ...}"
   ## [ } # End of "if ($IF_ARCHIVE_MAIN_RESULTS) {...}" ] ... A Mistake?


    (print STDOUT "\n === Finished Processing the $pwd th PCC Window ==\n\n\n") if ($TEST_LEVEL_MAIN >1);

} # End of the 2nd MAIN outer for-loop (over the PCC windows). #

		# END of "COPIED with substantial MODIFICATIONS from the 1st MAIN for-loop above (on 2019/07/06)." #

		### RESTART FROM HERE (on 2019/07/06). ###


(print STDOUT "\n\n ----- The 2nd Sliding-Window Analysis Finished !!! -----\n\n") if ($TEST_LEVEL_MAIN>0); # (Added on 2019/07/04.)

(print STDOUT "\n\n ===== ALL the Processes ENDED !!!!! =====\n\n") if ($TEST_LEVEL_MAIN >0);



	####### DEFINITIONS of the SUBROUTINES #######


	### NOTE: The Subroutines that had been here were MOVED to "MyANEX_IO.pm" in MAJOR REVISION (on 2019/07/07). ###



