###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.7: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# anex_rev_purge_for_sgl_wd.ver0.7.pl:
#	This script is the UPGRADED version of 'anex_rev_purge_for_sgl_wd.ver0.5.pl'.
#	The following MAJOR REVISIONS were made:
#
#	(1) The script can now take full advantage of the main outputs of 'lastpiece.xxx.pl' in the program package, "LASTPIECE_P.verNNN",
#	and thus the computed MSA probabilities can now be MUCH MORE ACCURATE than computed by the predecessor (ver 0.5);
#	To enable this revision,
#	  (1-i) the block for checking files of multiplication factors can now accept symbolic links as well, in addition to normal files as the predecessor (ver 0.5) does;
#
#	  (1-ii) the subroutines, 'apprx1_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs.pm") (& their satellite subroutines) & 'compt_log_msa_prob_plus' & 'nonrec_mlt_shifts_gapblocks' (the latter two are in "MyANEX_Main.pm"), were substantially REVISED
#		(& renamed the 1st one as 'apprx1_wlp_tot_lnlk_gpattern_msa2_hs' (in "MyTreeMap_indels_ML_hs_hs_wLP.pm")),
#		especially to take advantage of multiplication factors for case-(i) "gapped segment"s;
#
#	   (1-iii) the subroutine, 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm"), was REVISED, so that it can now read in the input file(s) recording multiplication factors for case-(i) "gapped segment"s;
#
#	(2) REVISED the rules (in the subroutine 'assign_br2log_mfacs' (in "MyReadAlnProbIngredients.pm") and its satellite subroutines) to "extrapolate" the multiplication factors (for case-(iv) and case-(i)) to those w/ the time-lapses smaller than the minimum time-lapse recorded in the input file (of the relevant multiplication factors).
#
#	[MAJOR REVISION STARTED on 2020/04/24; ENDED on 2020/04/dd.]
#
#
# anex_rev_purge_for_sgl_wd.ver0.5.pl:
#	This Perl script is the 1st prototype (written in Perl) of the Sub-Main ENGINE
#	(second to "anex_for_sgl_wd.verXXX.pl")	of 'anex.ver0.5.pl'.
#	Unlike the main engine (i.e., "anex_for_sgl_wd.verXXX.pl"), which performs a wide variety of moves,
#	this script EXCLUSIVELY performs the "Reverse-Purge" (aka. "Ex-Nihilo")
#	for some candidate regions (all included in a merged region), followed by Multiple-"Shift"s,
#	on a single window specially prepared for this purpose.
#	It was created by substantially modifying (especially removing many parts of)
#	"anex_for_sgl_wd.ver0.5.pl".
#	[CREATION STARTED on 2019/07/06, ENDED on 2019/09/18.]
#
#	NOTE: This script ASSUMES the following structure of the working directory:
#
#	{Top_Working_Directory}/ .
#
#	For the window:
#	{Main_Input_Directory}/ = "{Top_Working_Directory}/Inputs/" .
#	{Input_file_for_MSA_in_the_window} = "{Main_Input_Directory}/msa_in_pcc_wd.txt" ,
#	{Input_file_for_pcands} = "{Main_Input_Directory}/list_rlv_pcands.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Main_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Main_Input_directory}/list_gblocks0.txt" .
#
#
# anex_for_sgl_wd.ver0.5.pl:
#	This Perl script is the 1st prototype (written in Perl) of the MAIN ENGINE of 'anex.ver0.5.pl'.
#	It explores the neighborhood of an input MSA within a window,
#	by first performing the pure multiple-"shift"s the gap-blocks,
#	then by performing other elementary changes of the MSA, followed by multiple-"shift"s.
#	It was created by putting together useful portions of a number of scripts.
#	More precisely, the main sources are:
#	(1) "anex.ver0.5.pl", especially its 1st half;
#	(2) "detect_purge_cands.ver0.5.pl", especially its 1st half;
#	(3) The main & some other parts of 'test_nonrec_mlt_shifts_gapblocks.pl' (at $RIGHT_NOW_WORKING_CHICHIBU6M).
#
#	Then, other necessary parts were either imported from other scripts
#	or created de novo.
#	[CREATION STARTED on 2019/06/24, ENDED on 2019/mm/dd.]
#
#	NOTE: This script ASSUMES the following structure of the working directory:
#
#	{Top_Working_Directory}/ .
#
#	For the main window:
#	{Main_Input_Directory}/ = "{Top_Working_Directory}/Inputs/" .
#	{Input_file_for_MSA_in_the_window} = "{Main_Input_Directory}/msa_in_wd.txt" ,
#	{Input_file_for_pcands} = "{Main_Input_Directory}/list_rlv_pcands.txt" , # Added on 2019/07/03.
#	{Input_file_for_initial_set_of_columns} = "{Main_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Main_Input_directory}/list_gblocks0.txt" .
#
#	For the sub-windows belonging to the window:
#	{Sub_Working_Directory}/ = "{Top_Working_Directory}/Sbwd{sub-window-ID}/" .
#	{Sub_Input_Directory}/   = "{Sub_Working_Directory}/Inputs/" .
#	{Input_file_for_pcands} = "{Sub_Input_Directory}/list_rlv_pcands.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Sub_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Sub_Input_Directory}/list_gblocks0.txt" .
#
#	NOTE2 (added on 2019/07/04):
#	This script does NOT perform "Reverse-Purge" (aka, "Ex-Nihilo") + Multiple-"Shift"s;
#	such moves are performed via the script, 'anex_rev_purge_1wd.verXXX.pl'.
#
#	[REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/11/14]:
#	Now, TWO new global variables, $IF_TEST_W_COMPACT_MSS & $HF_WDTH_SHIFT_4_CMSS, are available,
#	for enabling the test with a "compact" multiple-shift space (whose half-width is given by $HF_WDTH_SHIFT_4_CMSS).
#
#	[REVISED on 2020/04/22]:
#	Now, two subroutines, 'find_maxvalue_in_multidim_storage' & 'cmpt_log_total_qnt_in_multidim_storage',
#	in the Perl package, "MyANEX_IO2.pm", are incorporated.
#	And the script now outputs a new "synopsis" file,
#	which summarizes the results of ALL runs of multiple-"shift"s performed (& their inputs).
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	);

use MyMalnIO qw(
	extract_alignment
	pr_ODEN_input
	);
	# extract_fasta
	# pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyTreeManipu_II qw(
	input_tree_file output_tree
	br_output_tree
	);
	# extract_sgltree_from_file
	# make_node2parent

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	br_output_tree_dawg
		) ;
	# extract_sgltree_from_file_dawg


use MyTreeMap_indels_spt_odr2 qw(
	cal_init_psm_cands_for_sgl_msa
	bch_br_parsimonize_dollo_indel_histories
	);

use MyTreeMap_indels_spt_odr_finer qw(
	enum_psm_indel_hsts_for_msa_finer_PRE
	);
	# partit_into_gblocks_seqblocks
	# if_idb_includes_root
	# partit_into_eff_indep_indel_blocks
	# find_invlvd_in_eff_indep_indel_blocks
	# mk_node2blwise_gstat

use MyReadAlnProbIngredients qw(
	readin_delta_del
	assign_br2log_mfacs
	);

use MyANEX_Supple qw(
	prepare_sliding_windows
	mk_orgs_bds_gb_coords
	);
	# Added 'mk_orgs_bds_gb_coords' in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29).

use MyArrayManipu_num qw(
	diff_nmarray
	);

use MyTreeMap_indels_spt_odr_hs qw(
	enum_psm_indel_scns_for_msa_hs_mrounds
	);
	## enum_psm_indel_scns_for_msa_hs

use MyTreeMap_indels_ML_hs_hs qw(
	apprx1_tot_lnlk_gpattern_msa2_hs
	);
	## cal_delta_exp_del_5p2
	## cal_log_normal_facs_indel_distri
	## cal_lnlk1_indel_scn_cls2

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26). #
use MyTreeMap_sppl qw(
	wrapper_mk_node2exoffs_msa_rows
	mk_set_msa_rows2br_up_or_lw
	mk_set_msa_rows2br_up_or_lw_dp
	);
	# Added 'mk_set_msa_rows2br_up_or_lw_dp' on 2019/07/03.

	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/26)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27). #
use MyLikelihoods_respattern_msa_dna qw(
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);
	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/27)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29). #
use MyGapPatternTopology qw(
	mk_inter_block_relations
	mk_list_gblock_sizes
	mk_lists_gblocks_w_spec_bounds
	cts_vincluding_vincluded_onx_gblks
	);
	    # Added 'mk_list_gblock_sizes' & 'mk_lists_gblocks_w_spec_bounds' & 'cts_vincluding_vincluded_onx_gblks' on 2019/09/18. #

	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/06/29)." #

	# Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02). #
use MyANEX_Main qw(
	nonrec_mlt_shifts_gapblocks
	compt_log_msa_prob_plus

	list_revpurge_cand_sizes
	reverse_purge_gblock_pair

	printout_multidim_storage
	);
	# Added 'compt_log_msa_prob_plus' on 2019/07/04.
	# Added 'list_revpurge_cand_sizes' & 'reverse_purge_gblock_pair' on 2019/09/16.

	# End of "Added in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/02)." #

=Set Asite for Later Use. (PM)
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_III qw(
	make_node2abs_depth
	make_node2ct_exoffs
	);

use MyTreeManipu_IV qw(
	bi2multifurcated_top
	construct_neat_subtree
	construct_skeleton_subtree
	);

##use MyTreeMap_indels qw(
##	enum_psm_indel_scns_for_msa
##	);
##	## extract_cls_gblocks msa_rep_gpatterns

use MyTreeMap_indels_sppl qw(
	create_br2eqbr
	create_br2nn
	);

use MyDetect_cmplx_error_cands qw(
	wrapper_detect_cmplx_error_cand_acls0
	);
	# mk_self_nn_events

=cut # End of "Set Asite for Later Use. (PM)"


	### ADDED in MAJOR REVISION on 2019/07/07. ###
use MyANEX_IO qw(

	read_dawg_ctrl_file_sbst 
	read_dawg_ctrl_file3 
	readin_list_mgd_pcands 
	readin_set_clms 
	readin_list_gblks 
	revise_list_gblocks 

	print_out_input_tree 
	print_out_br2log_mfacs 
	print_out_set_clms 

	print_out_readin_list_gblks 

	print_out_tree_profile 
	print_out_wrapper_mk_node2exoffs_msa_rows 

	print_out_mk_set_msa_rows2br_up_or_lw 
	print_out_br2tr_mtrx 
	print_out_node2ntfreqs 
	print_out_node2set_sw_cond_probs 
	print_out_br2set_sw_jnt_probs 
	print_out_set_clm_probs 
	print_out_orgs_bds_gb_coords 
	print_out_row_clm_info_gblocks 
	print_out_inter_block_relations 
	print_out_list_mgd_pcands 
	print_out_features_gblks 
	print_out_x_bin_hrels 
	print_in_nr_mshifts_glks 
	print_out_ln_prb_plus_msa 
	print_out_coord_frame 
	print_out_nr_mshifts_gblks 
	print_out_list_gblock_sizes

	print_out_revpurge_cand_sizes

	print_out_gblks_w_spec_bds
	print_out_cts_vincluding_vincluded_onx_gblks

	print_out_info_chngs_by_split

	perform_mshifts_and_print_inputs_outputs

	relative_coords_for_list_mgd_pcands 
	);
	   # Added 'print_out_revpurge_cand_sizes' on 2019/09/16. #
	   # Added 'print_out_list_gblock_sizes' & 'print_out_gblks_w_spec_bds' & 'print_out_cts_vincluding_vincluded_onx_gblks' & 'print_out_info_chngs_by_split' & 'perform_mshifts_and_print_inputs_outputs' on 2019/09/18. #

	### End of "ADDED in MAJOR REVISION on 2019/07/07." ###

	### ADDED in REVISION on 2020/04/22. (US) ##
use MyANEX_IO2 qw(
	find_maxvalue_in_multidim_storage
	cmpt_log_total_qnt_in_multidim_storage
	);
	### End of "ADDED in REVISION on 2020/04/22. (US)" ##


$| = 1; # Flush the input buffer as soon as it receives data.

my $plscript_path = $0;
my ($plscript_name) = ($plscript_path =~ m/([^\/]+)$/); 
my $program_name = $plscript_name; # Added on 2019/06/10, to reduce the number of error messages.

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

my $exec_perl = "/usr/bin/perl"; # Perl executable to be used. (Added on 2019/06/19.) 


	### File names storing important parameters. ###

our $FILENAME_DELTA_DEL = 'delta_del.txt'; # The name of the file storing '$delta_del', which is important for calculating the exit rate of the entire sequence. #

our @FILENAMES_LOG_MFACS = ( # Names of the files storing logarithmic multiplication factors, which are in $indir_mfacs. #

		# In the bulk #
	'lmfacs_asites_vs_dgaps.txt', # For the cases where there are some ancestral sites and no descendant sites in between preserved ancestral sites (PASs). #
	'lmfacs_agaps_vs_dsites.txt', # For the cases where there are no ancestral sites and some descendant sites in between PASs.
	'lmfacs_asites_nhom_dsites.txt', # For the cases where there are some ancestral sites and some descendant sites in between PASs.

		# On either boundary (i.e., either PWA end). "bd_$filename" corresponds to $filename above. #
	'bd_lmfacs_asites_vs_dgaps.txt',
	'bd_lmfacs_agaps_vs_dsites.txt',
	'bd_lmfacs_asites_nhom_dsites.txt',

		# For case-(i) "gapped segment"s. #
		# (ADDED in creation of 'anex_rev_purge_for_sgl_wd.ver0.7.pl' on 2020/04/24). (GL1) #
	'lmfacs_nosites.txt',
	'bd_lmfacs_nosites.txt',
		# End of "(ADDED in creation of 'anex_rev_purge_for_sgl_wd.ver0.7.pl' on 2020/04/24). (GL1)" #
	);

	### End of "File names storing important parameters." ###

	##### Glogal parameter setting #####

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

our $STRING_UNDEF = 'n/a'; # This is used for printing out some elements or quantities that are undefined for some reasons. (Added on 2019/07/02.)

			## Indexes in the terminal array of a multi-dimensional storage (for multiple-"shift"s). ##
			## (Added in REVISION on 2020/04/22.) (GL) ##
our $INDX_DIFF_LN_PRB_MSA = 3; # For the difference in the log-probability of the MSA (under a model w/ both substitutions & indels.)
our $INDX_DGN = 1;  # For the degree of degeneracy of the MSA.
			## End of "(Added in REVISION on 2020/04/22.) (GL)" ##


		# Added in revision (on 2019/06/18). #
#our $DIS_LEN = 50; # The maximum length of the (sub)sequence displayed in a line.
our $DIS_LEN = 60; # The maximum length of the (sub)sequence displayed in a line.
		# End of "Added in revision (on 2019/06/18)." #

$MyANEX_IO::DIS_LEN = $DIS_LEN;  # Added in REVISION (on 2019/07/07). #

		# IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/09/13). #
	# The level of MSA space exploration.
	# (1: performing only pure multiple-"shift"s;
	#  2: also performing ;
	#  3: also performing ;
	#  5: performing all the programmed moves.)
#our $XLEVEL = 1; 
our $XLEVEL = 3; 
#our $XLEVEL = 5; 
		# END of "IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/09/13)." #

	# Minimum and maximum sizes of windows used for detecting purge-candidate regions. #
our $MIN_WD_SIZE_PCAND = 3;
our $MAX_WD_SIZE_PCAND = 6;

	# Each gap-block can "shift" up to (and including) $HF_WDTH_SHIFT columns in each direction. # (Added on 2019/06/29.)
our $HF_WDTH_SHIFT = 10;


	# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (GL) #
our $IF_TEST_W_COMPACT_MSS = 0; # Whether to perform tests with "compact" multiple-shift spaces(=1) or not(=0).
our $HF_WDTH_SHIFT_4_CMSS  = 2;
	# END of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (GL)" #

our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

=Set Aside for Later Use. (GL 1)
	# Characters indicating that the site is missing or that its orthology is questionable. 
our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);
=cut # END of "Set Aside for Later Use. (GL 1)"

	# Parameters for the relative probabilities of Indel Histories. #
our %INDEL_MODEL2DISTRI = ( # = (indel model => distribution type, ...)
	'zipf' => 'P',
	'"PL"' => 'P',
	'"NB"' => 'NB', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	'"US"' => 'US', # ADDED in creation of 'fa_lolipog_hs.alpha2.pl'. #
	);


our @RESIDUES = our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
##our @RESIDUES = our @NUCLEOTIDES = @MyPRANK_Scores::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our $SV_RATIO_R = 1; # Ratio of purine-type transition rate to transversion rate.
our $SV_RATIO_Y = 1; # Ratio of pyrimidine-type transition rate to transversion rate.

	# Equilibrium residue frequencies. #
my @root_resfreqs = (0.25, 0.25, 0.25, 0.25);
##my @root_resfreqs = (0.1, 0.2, 0.3, 0.4);

my @labels_resfreqs = ();
foreach my $res (@NUCLEOTIDES) { push @labels_resfreqs, "f($res)"; }


	### ReSet parameters defined in some Perl packages. ###

$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS = 1; # Ignore null columns(, if at all).

$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = 0; # NOT analyze null-clusters (=clusters consisting only of null-columns), if at all.

	# This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = 120;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 150;

$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;
$MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;



	### ReSet Parameters to adjust the verboseness of the test outputs. ###
	### (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)

	# For this script. #
our $TEST_LEVEL = 5;
##our $TEST_LEVEL = 1;

	# Testing the inputs/outputs of the processes going through this MAIN program.
our $TEST_LEVEL_MAIN = 5;

$MyANEX_IO::TEST_LEVEL_MAIN = $TEST_LEVEL_MAIN;  # Added in REVISION (on 2019/07/07). #

	# For some subroutines defined in this script. #
	# (e.g., 'print_in_nr_mshifts_glks', 'print_out_nr_mshifts_gblks'). # (Added on 2019/07/02.)
our $TEST_LEVEL_SUB = 2;

$MyANEX_IO::TEST_LEVEL_SUB = $TEST_LEVEL_SUB;  # Added in REVISION (on 2019/07/07). #


	# Testing some primary satellite subroutines of 'readin_delta_del' & 'assign_br2log_mfacs'. #
$MyReadAlnProbIngredients::TEST_LEVEL1  = 0;


        # For 'calc_prob_descresstats_cond_ancstat2'. #
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 5;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 6;

	# For 'calc_joint_prob_ancresstats_and_ancstat2'. #
	# (ADDED in creation of 'test_calc_jnt_prob_ancresstats2.pl'.)
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 6;

	# For 'enum_psm_indel_hsts_for_msa_finer_PRE'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 0;
#$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 3;

	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_spt_odr_hs.pm" for details.)
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL = 3;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1 = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL1hf = 5;
#$MyTreeMap_indels_spt_odr_hs::TEST_LEVEL2 = 0;

	# Set the following (and some others) if necessary.
	# (See "MyTreeMap_indels_ML_hs_hs.pm" for details.)
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL = 1;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1 = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL1a = 0;
#$MyTreeMap_indels_ML_hs_hs::TEST_LEVEL2a = 0;

=Set Aside for Later Use. (GL 2) 
	# For 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_1ST = 0;

	# For 'br_parsimonize_dollo_indel_history'. 
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = 0;

	# For 'bch_br_parsimonize_dollo_indel_histories'.
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = 0;

=cut # End of "Set Aside for Later Use. (GL 2)"

	##### END of "Glogal parameter setting" #####


	####### Take in the Command-Line Arguments. #######

unless (@ARGV > 2) {
	print STDERR "\nUsage:  [perl] $program_name {top_working_directory} {dawg_ctrl_file (incl. input tree)} {dir_mfac_data} [ options ] ,\n";
	print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";
	print STDERR "            --MIN_WD_SIZE=integer      for the minimum window size for detecting purge-candidates (default=$MIN_WD_SIZE_PCAND);\n";
	## print STDERR "            --MAX_WD_SIZE=integer      for the maximum window size for detecting purge-candidates (default=$MAX_WD_SIZE_PCAND);\n";
	print STDERR "           --HF_WDTH_SHIFT=integer     for the maximum number of \"shift\"s of each gap-block in one direction (default=$HF_WDTH_SHIFT);\n";  # Added on 2019/06/29.
	print STDERR "           --XLEVEL=integer       for the level of MSA space exploration (from =1 (performing only multiple-\"shift\"s) to =5 (performing all programed moves) (default=$XLEVEL);\n"; # IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/13.

	# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (ARGV) #
	print STDERR "           --IF_TEST_W_COMPACT_MSS=boolean     for whether tests with \"compact\" multiple-shift spaces are performed (=1) or not (=0) (default=$IF_TEST_W_COMPACT_MSS);\n";
	print STDERR "           --HF_WDTH_SHIFT_4_CMSS=integer      for the maximum number of \"shift\"s of each gap-block in one direction WHEN \$IF_TEST_W_COMPACT_MSS=1  (default=$HF_WDTH_SHIFT_4_CMSS);\n";
	# END of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (ARGV)" #

	die "$!\n\n";
}
my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	# NOTE: The input/output file system will be specified & prepared below. #

	####### END of "Take in the Command-Line Arguments." #######


	##### Check the Command-Line Options. #####

for (my $k=3; $k <= $#ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

	if ($option eq 'MIN_WD_SIZE') {
	    $MIN_WD_SIZE_PCAND = $value;

	##} elsif ($option eq 'MAX_WD_SIZE') {
	##    $MAX_WD_SIZE_PCAND = $value;

		# Added on 2019/06/29. (OP) #
	} elsif ($option eq 'HF_WDTH_SHIFT') {
	    $HF_WDTH_SHIFT = $value;
		# End of "Added on 2019/06/29. (OP)" #

		# IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/13. (OP) #
	} elsif ($option eq 'XLEVEL') {
	    $XLEVEL = $value;
		# End of "IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/13. (OP)" #

		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (OP) #
	} elsif ($option eq 'IF_TEST_W_COMPACT_MSS') {
	    $IF_TEST_W_COMPACT_MSS = $value;

	} elsif ($option eq 'HF_WDTH_SHIFT_4_CMSS') {
	    $HF_WDTH_SHIFT_4_CMSS = $value;
		# End of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (OP)" #

	} else {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

}	# End of the preliminary for-loop (over additional command-line arguments. #

	##### End of "Check the Command-Line Options." #####`


	##### Specify & Check Input/Output Files/Directories. #####

## my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	### Input files/directories. ###

my $intree_file = $dawg_ctrl_file;
my $infile_delta_del = "$indir_mfacs/$FILENAME_DELTA_DEL"; # This file stores '$delta_del', which is important for calculating the exit rate of the entire sequence. #

my $maindir_in = "$top_work_dir/Inputs";

my $infile_msa_in_pwd = "$maindir_in/msa_in_pcc_wd.txt"; # MODIFIED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl'. #
##my $infile_msa_in_wd = "$maindir_in/msa_in_wd.txt"; # For the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. #

my $infile_set_clms0 = "$maindir_in/set_columns0.txt";
my $infile_list_gblks0  = "$maindir_in/list_gblocks0.txt";
my $infile_pcands0 = "$maindir_in/list_rlv_pcands.txt"; # Added in REVISION (on 2019/07/03). #

#	{Main_Input_Directory}/ = "{Top_Working_Directory}/Inputs/" .
#	{Input_file_for_MSA_in_the_window} = "{Main_Input_Directory}/msa_in_pcc_wd.txt" ,
#	{Input_file_for_initial_set_of_columns} = "{Main_Input_Directory}/set_columns0.txt" ,
#	{Input_file_for_initial_gap-blocks} = "{Main_Input_directory}/list_gblocks0.txt" .


	# Check input files/directories. #

unless (( -d $indir_mfacs ) and ( -r $indir_mfacs )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY, '$indir_mfacs', that stores data files for the multiplication factors !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

foreach my $filename (@FILENAMES_LOG_MFACS) { # 1st preliminary foreach-loop (over input data files for multiplication factors). #
    my $file = "$indir_mfacs/$filename";
    ##unless (( -f $file ) and ( -r $file )) {
    unless ( (( -f $file ) or ( -l $file )) and ( -r $file )) { # MODIFIED in creation of 'anex_rev_purge_for_sgl_wd.ver0.7.pl' (on 2020/04/24). (CHK) #
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 1st preliminary foreach-loop (over input data files for multiplication factors). #

foreach my $file ($dawg_ctrl_file, $intree_file, $infile_delta_del) { # 2nd preliminary foreach-loop (over some key input files). #
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 2nd preliminary foreach-loop (over some key input files). #

foreach my $rf_dir (\$top_work_dir, \$maindir_in) { # 3rd preliminary foraech-loop (over some key directories). # Modified in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07). #
## [ foreach my $rf_dir (\$top_work_dir, \$maindir_in, @subdirs_in) { # 3rd preliminary foraech-loop (over some key directories). # ] ... for the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. # ...}
    my $dir = ${$rf_dir};
    unless (( -d $dir ) and ( -r $dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY '$dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 3rd preliminary foreach-loop (over some key directories). #

foreach my $rf_file (\$infile_msa_in_pwd, \$infile_set_clms0, \$infile_list_gblks0, \$infile_pcands0) { # 5th preliminary foreach-loop (over some input files). # Modified in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07). #
## [ foreach my $rf_file (\$infile_msa_in_wd, \$infile_set_clms0, \$infile_list_gblks0, \$infile_pcands0,
##		@infiles_pcands, @infiles_sbset_clms0, @infiles_sblist_gblks0) { # 5th preliminary foreach-loop (over some input files). # ] ... for the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. # ...}

		# Added \$infile_pcands0 on 2019/07/03.

    my $file = ${$rf_file};
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
} # End of the 5th preliminary foreach-loop (over some input files). #

	### Output directories. ###

my $maindir_in_chk = "$top_work_dir/Check_Inputs";
my $maindir_interim = "$top_work_dir/Interim_Data";
my $maindir_out = "$top_work_dir/Outputs";

foreach my $rf_dir (\$maindir_in_chk, \$maindir_interim, \$maindir_out) { # 6th preliminary foreach-loop (over putput directories). # Modified in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07). #
## [ foreach my $rf_dir (\$maindir_in_chk, \$maindir_interim, \$maindir_out, @subdirs_in_chk, @subdirs_interim, @subdirs_out) { # 6th preliminary foreach-loop (over putput directories). # ] ... for the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. # ...}
    my $dir = ${$rf_dir};
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the 6th preliminary foreach-loop (over output directories). #


		## "Synopsis" Output File, which summarizes the results of ALL multiple-"shift"s performed. ##
		## (ADDED in REVISION on 2020/04/22.) (I/O-F/D) ##

my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";

		## End of "(ADDED in REVISION on 2020/04/22.) (I/O-F/D)" ##


	##### End of "Specify & Check Input/Output Files/Directories." #####


	##### Declare the Beginning of the Script #####

if ($TEST_LEVEL_MAIN>0) { # (1) #

	print STDOUT "\n<<<< Program name: $plscript_name (path: $plscript_path) >>>>\n\n";

  if ($TEST_LEVEL_MAIN>1) {

	print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

	print STDOUT "<< Input/Output Files/Directories >>\n\n";

## my ($top_work_dir, $dawg_ctrl_file, $indir_mfacs) = @ARGV[0 .. 2];

	print STDOUT "< Key Input Files: >\n\n";

	print STDOUT "Control file: $dawg_ctrl_file .\n\n";

	print STDOUT "Input tree and indel parameters: $intree_file .\n\n";

	print STDOUT "Input file recording ", '$delta_del', ": $infile_delta_del .\n\n";

	print STDOUT "Input directory storing (log-)multiplication factors: $indir_mfacs ,\n\n";
	print STDOUT "   which holds the following input files:\n\n";
	print STDOUT '     ', join (', ', @FILENAMES_LOG_MFACS), " .\n";
	print STDOUT "\n\n";

	print STDOUT "< Working File System: >\n\n";

	print STDOUT "{Top Working Directory}: $top_work_dir .\n\n";

	print STDOUT "< Input Files/Directories: >\n\n";

	print STDOUT "[ For the PCC Window: ]\n\n"; # Modified in creation of 'anex_for_sgl_wd.ver0.5.pl' (on 2019/07/07). #
	##print STDOUT "[ For the full Window: ]\n\n";
	print STDOUT "{Main Input Directory}:       $maindir_in .\n";
	print STDOUT "{Input MSA in the window}:    $infile_msa_in_pwd ,\n";
	print STDOUT "{Initial set of columns}:     $infile_set_clms0 ,\n";
	print STDOUT "{Initial list of gap-blocks}: $infile_list_gblks0 ,\n";
	print STDOUT "{List of purge-candidates}:   $infile_pcands0 .\n\n"; # Added in REVISION (on 2019/07/03). #

	print STDOUT "\n";

				# (ADDED in REVISION on 2020/04/22.) (DC2) #
	print STDOUT "< \"Synopsis\" Output File: >\n\n";

	print STDOUT "$outfile_synopsis .\n\n\n";

# my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";
				# End of "(ADDED in REVISION on 2020/04/22.) (DC2)" #

	print STDOUT "< Output Directories: >\n\n";

	print STDOUT "[ For the full Window: ]\n\n";
	print STDOUT "{Directory to Check the Inputs}: $maindir_in_chk ,\n";
	print STDOUT "{Directory for Interim Data}:    $maindir_interim ,\n";
	print STDOUT "{Directory for the Outputs}:     $maindir_out .\n\n";

	print STDOUT "\n\n";


	print STDOUT "<< User-specified Parameters >>\n\n";

	print STDOUT '$STRING_UNDEF = ', $STRING_UNDEF, " .\n\n"; # Added on 2019/07/04. #

	print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/22. (DC1) #

	print STDOUT '$CHAR_ABS  =  $GAP  = ', $GAP,  " ,\n";
	print STDOUT '$CHAR_PRES =  $NGAP = ', $NGAP, " .\n\n";

	print STDOUT "{A displayed alignment will be chopped into chunks of length $DIS_LEN, & a shorter remainder.}\n\n";
	print STDOUT '$MyANEX_IO::DIS_LEN = ', $MyANEX_IO::DIS_LEN, " .\n\n";  # Added in REVISION (on 2019/07/07). #

=Set Aside for Later Use. (PR 1)
	print STDOUT "Characters indicating that the site is missing or that its orthology is questionable:\n\n";
	print STDOUT ' ' x 5, join (', ', sort keys %CHAR_MISSING), " .\n";
	print STDOUT "\n\n";
=cut # END of "Set Aside for Later Use. (PR 1)"

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS =' , $MyTreeMap_indels_spt_odr_hs::FL_IGNORE_NULL_COLUMNS, " ,\n";
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS = ', $MyTreeMap_indels_spt_odr_hs::FL_HANDLE_NULL_CLS, " ,\n\n";
	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML_hs_hs::UPPER_BOUND_ANAL_GAPLEN', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_GAPLEN = ', $UPPER_BOUND_ANAL_GAPLEN, " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_hs::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $MyTreeMap_indels_ML::UPPER_BOUND_ANAL_LEN_GBCLS', "\n";
	print STDOUT ' = $UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " .\n";
	print STDOUT "\n\n";

		# IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/13. (PR) #
	print STDOUT "< The LEVEL of MSA Space Exploration. >\n\n";
	print STDOUT '$XLEVEL = ', $XLEVEL, " .\n";
	print STDOUT "\n";
		# End of "IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/13. (PR)" #

	print STDOUT "< Range of sizes of windows for detecting purge-candidates. >\n\n";
	print STDOUT '($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) = ', "($MIN_WD_SIZE_PCAND, $MAX_WD_SIZE_PCAND) .\n";
	print STDOUT "\n";

			# Added on 2019/06/29. (PR 3) #
	print STDOUT "< Half-width of the coordinate of each gap-block when performing multiple-\"shift\"s. >\n\n";
	print STDOUT '$HF_WDTH_SHIFT = ', "$HF_WDTH_SHIFT .\n";
	print STDOUT "\n\n";
			# End of "Added on 2019/06/29. (PR 3)" #

		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (PR) #
	print STDOUT "< Options regarding the tests with \"compact\" multiple-shift spaces. >\n\n";
	print STDOUT '$IF_TEST_W_COMPACT_MSS = ', $IF_TEST_W_COMPACT_MSS, " .\n";
	if ($IF_TEST_W_COMPACT_MSS) {
	    print STDOUT '$HF_WDTH_SHIFT_4_CMSS  = ', $HF_WDTH_SHIFT_4_CMSS, " .\n";
	}
	print STDOUT "\n\n";
		# End of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (PR)" #

	print STDOUT "< Parameters specifying How verbose the Test Outputs will be: >\n\n";

	print STDOUT '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

	print STDOUT '$TEST_LEVEL_MAIN = ', $TEST_LEVEL_MAIN, " ,\n";
	print STDOUT '$MyANEX_IO::TEST_LEVEL_MAIN = ', $MyANEX_IO::TEST_LEVEL_MAIN, " .\n\n";  # Added in REVISION (on 2019/07/07). #

	print STDOUT '$TEST_LEVEL_SUB  = ', $TEST_LEVEL_SUB , " ,\n"; # Added on 2019/07/02.
	print STDOUT '$MyANEX_IO::TEST_LEVEL_SUB  = ', $MyANEX_IO::TEST_LEVEL_SUB, " .\n\n";  # Added in REVISION (on 2019/07/07). #

	print STDOUT '$MyReadAlnProbIngredients::TEST_LEVEL1 = ', $MyReadAlnProbIngredients::TEST_LEVEL1, " .\n\n";

	print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca, " ,\n\n";
	print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa, " .\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = ', $MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE, " .\n\n";

=Set Aside for Later Use. (PR 2)
	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs , " ,\n\n";

	print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs =  ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs, " ,\n\n";
=cut # END of "Set Aside for Later Use. (PR 2)"

	print STDOUT "\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) {...}"

} # END of "if ($TEST_LEVEL_MAIN>0) { # (1) # ...}"

	##### End of "Declare the Beginning of the Script" #####


	##### Read in the Control file & Set Some Model Parameters. #####

print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for substitution model parameters. >>>\n\n\n";

my ($model_sbst, $nt_freqs, $sbst_params) = read_dawg_ctrl_file_sbst ($dawg_ctrl_file);
unless (defined $model_sbst) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $MODEL_SBST = $model_sbst;

our @NT_FREQS = @{$nt_freqs}[0,2,1,3]; # Re-arrange the frequencies: from (ACGT) in Dawg to (AGCT) in this package.
##our @NT_FREQS = @{$nt_freqs};

@root_resfreqs = @NT_FREQS; # ADDED on 2019/06/27.

our @SBST_PARAMS = @{$sbst_params};

if ($TEST_LEVEL_MAIN>2) { # (2A) #

	## Print out the Parameters for Substitution Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Substitution Probabilities ]]\n\n";

	print STDOUT '$MODEL_SBST = ', $MODEL_SBST, " .\n\n";

	##print STDOUT '@NT_FREQS = (', join (', ', @NT_FREQS), ") .\n\n";
	print STDOUT '@NT_FREQS (=@root_resfreqs) = (', join (', ', @NT_FREQS), ") .\n\n"; # MODIFIED on 2019/06/27.

	print STDOUT '@SBST_PARAMS = (', join (', ', @SBST_PARAMS), ") .\n\n";

	print STDOUT "\n";

} # End of "if ($TEST_LEVEL_MAIN>2) { # (2A) # ...}"


	# Translate the @SBST_PARAMS into the Parameters for the Tamura-Nei model. #

our ($ALPHA1, $ALPHA2, $BETA);
	# $ALPHA1 = rate for a transitional substitution (A <-> G); 
	# $ALPHA2 = rate for a transitional substitution (C <-> T);
	# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).

if (($MODEL_SBST eq 'JC') or ($MODEL_SBST eq 'F81')) {

	($ALPHA1, $ALPHA2, $BETA) = (1, 1, 1);

} elsif (($MODEL_SBST eq 'K2P') or ($MODEL_SBST eq 'HKY')) {

	($ALPHA1, $BETA) = @SBST_PARAMS;
	 $ALPHA2 = $ALPHA1;
	
} elsif ($MODEL_SBST eq 'TN') {

	($ALPHA1, $ALPHA2, $BETA) = @SBST_PARAMS;
} else {
	print STDOUT "[ERROR '$program_name'] The current version does NOT HANDLE the substitution model, '$MODEL_SBST'! ... USE either 'JC', 'K2P', 'HKY', 'F81' or 'TN'!!\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL_MAIN>1) { # (2A') #
	print STDOUT '[ Translation of @SBST_PARAMS into Parameters of Tamura-Nei model ]', "\n\n";
	print STDOUT '($ALPHA1, $ALPHA2, $BETA) = ', "($ALPHA1, $ALPHA2, $BETA) .\n\n";
	print STDOUT "# NOTE:\n";
	print STDOUT '# $ALPHA1 = rate for a transitional substitution (A <-> G);', "\n";
	print STDOUT '# $ALPHA2 = rate for a transitional substitution (C <-> T);', "\n";
	print STDOUT '# $BETA = rate for a transvertional substitution (A <-> C, A <-> T, G <-> C, G <-> T).', "\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (2A') # ...}"


print STDOUT "\n<<< Read the Dawg Control File, '$dawg_ctrl_file', for indel model parameters and phylogenetic tree. >>>\n\n\n";

my ($tree_string, $gapmodel_ins, $gapmodel_del, $gap_param_ins, $gap_param_del, $gaplen_cutoff_ins, $gaplen_cutoff_del, $ins_rate, $del_rate)
	= read_dawg_ctrl_file3 ($dawg_ctrl_file);
unless (defined $tree_string) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file3' FAILED on the control file, '$dawg_ctrl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $INS_MODEL = $gapmodel_ins;
our $DEL_MODEL = $gapmodel_del;
our $DISTRI_INS = $INDEL_MODEL2DISTRI{$INS_MODEL};
our $DISTRI_DEL = $INDEL_MODEL2DISTRI{$DEL_MODEL};

	# Total rates of insertions & deletions at a site relative to the substitution rate.
our $TOTRATE_INS = $ins_rate;
our $TOTRATE_DEL = $del_rate;

 	# Exponents for the power-law distributions of insertion/deletion lengths (when $INDEL_MODEL = 'zipf').
our $PARAM_INS = $gap_param_ins;
our $PARAM_DEL = $gap_param_del;

	# Cut-off insertion/deletion lengths.
our $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
our $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL); 


if ($TEST_LEVEL_MAIN>1) { # (2B) #

	## Print out the Parameters for Indel Probabilities. ##

	print STDOUT "\n[[ Input Parameters for Indel Probabilities ]]\n\n";

	print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI_INS, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
	print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI_DEL, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";

	print STDOUT '$CUTOFF_INS_LEN = ', "$CUTOFF_INS_LEN ,\n";
	print STDOUT '$CUTOFF_DEL_LEN = ', "$CUTOFF_DEL_LEN .\n";
	print STDOUT "\n\n";

	## End of "Print out the Parameters for Indel Probabilities." ##

} # End of "if ($TEST_LEVEL_MAIN>1) { # (2B) # ...}"

	##### End of "Read in the Control file." #####


	##### Read in the $delta_del. #####

(print STDOUT "\n<<< Read in the multiplication factor '\$delta_del' for a 'universal' term of log-lilekihood ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $delta_del = readin_delta_del ($infile_delta_del);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_delta_del' FAILED on the input file, '$infile_delta_del' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}
(print STDOUT "\n ... The subroutine, 'readin_delta_del' Ended (w/ \$delta_del = $delta_del) ...\n\n") if ($TEST_LEVEL_MAIN> 0);

	##### End of "Read in the $delta_del." #####


	##### Read in the INPUT TREE. #####

(print STDOUT "\n<<< Read in the Input Tree >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
# [my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches. ] --- for the predecessor 'br_infer_psm_indel_events.pl'. #
my %br_attr; # = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches.
my %name2id_intnode; # = ($node_name => $node_id, ...) for all interior nodes. #
my $topnode_name; # The name of the top-node (root).

unless ( input_tree_file_dawg (${$tree_string}, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $topnode_id = $name2id{$TOP_NODE};


if ($TEST_LEVEL_MAIN>2) { # (3) #

	print STDOUT "\n ... To check the input tree, perform the subroutine, 'print_out_input_tree' ...\n\n";

	unless ( print_out_input_tree ($maindir_in_chk, $topnode_id, %node2ch, %node2pa, %id2name, %br_attr, $topnode_name) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_input_tree' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_input_tree' Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN>2) { # (3) # ...}"

	##### End of "Read in the INPUT TREE." #####


	##### Assign Log-Multiplication Factors to Each Branch. #####

(print STDOUT "\n<<< Assign the logarithmic multiplication factors to each branch, based on the files in '$indir_mfacs' ... >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my $br2log_mfacs = assign_br2log_mfacs (%br_attr, $TOTRATE_INS + $TOTRATE_DEL, $indir_mfacs, @FILENAMES_LOG_MFACS);
unless (defined $br2log_mfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'assign_br2log_mfacs' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'assign_br2log_mfacs', Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN >= 5) { # (5) #

	print STDOUT "\n ... To check the outputs of 'assign_br2log_mfacs', Run the subroutine, 'print_out_br2log_mfacs' ...\n\n";

# $CUTOFF_INS_LEN = $gaplen_cutoff_ins ;
# $CUTOFF_DEL_LEN = $gaplen_cutoff_del ;

	unless ( print_out_br2log_mfacs ($maindir_in_chk, %{$br2log_mfacs}, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2log_mfacs' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine 'print_out_br2log_mfacs' Ended ...\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL_MAIN >= 5) { # (5) # ...}"

	##### End of "Assign Log-Multiplication Factors to Each Branch." #####


	##### Read in the INPUT MSA. #####

(print STDOUT "\n<<< Read in the Input MSA, via the subroutine, 'extract_alignment' >>>\n\n") if ($TEST_LEVEL_MAIN>0);

my ($in_msa, $seqnames, $seq_lens) = extract_alignment ($infile_msa_in_pwd);
unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from the file '$infile_msa_in_pwd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_seqs = @{$seqnames};

unless ($ct_seqs>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] NO SEQUENCES in the input MSA from the file '$infile_msa_in_pwd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_aln = length ${$in_msa->[0]} ;

unless ($totlen_aln>0) { # Added on 2019/06/14.
	print STDOUT "[ERROR '$program_name'] ZERO-LENGTH SEQUENCES in the input MSA from the file '$infile_msa_in_pwd' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=SET ASIDE for potential use.
	# REMOVE the Ancestral Sequences from the Input MSA. #
	# (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my @msa_extseqs = my @ext_seqnames = ();
for (my $i=0; $i<$ct_seqs; $i++) { # Preliminary for-loop (over sequences in the input MSA). #

	my $seqid = $seqnames->[$i];
	($seqid  =~ m/^anc/) and next;

	push @ext_seqnames, $seqid;
	push @msa_extseqs, $in_msa->[$i];

}			# End of the preliminary for-loop (over sequences in the input MSA). #

@{$in_msa} = @msa_extseqs;
@{$seqnames} = @ext_seqnames;
$ct_seqs = @{$seqnames};

=cut # End of "SET ASIDE for potential use."

(print STDOUT "\n ... The subroutine, 'extract_alignment', Ended (w/ #{(external) sequences}=$ct_seqs, #{columns}=$totlen_aln) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (6) #

	my $chk_imsa_file = "$maindir_in_chk/in_msa.txt";
	my $fh_ci_msa = open_output_file ($chk_imsa_file, 0);
	unless (defined $fh_ci_msa) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the file, \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}

	print STDOUT "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";
	print $fh_ci_msa "[[ Input MSA (#{sequences} = $ct_seqs; gap = '$GAP') ]]\n\n";

	unless ( pr_ODEN_input (@{$seqnames}, @{$in_msa}, $DIS_LEN, *${$fh_ci_msa}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'pr_ODEN_input'  FAILED to print out the input MSA into \"$chk_imsa_file\" !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print $fh_ci_msa "\n";
	print STDOUT "\n ... the input MSA has been written into the file, \"$chk_imsa_file\" ...\n\n";

	close ($fh_ci_msa);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (6) # ...}"

	##### End of "Read in the INPUT MSA." #####


	##### Read in the Initial Set of Columns. #####

## my $infile_set_clms0 = "$maindir_in/set_columns0.txt";

my ($beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, $set_clms0_in_wd)
	= readin_set_clms ($infile_set_clms0);
unless (defined $set_clms0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_set_clms', FAILED on the input file, \"$infile_set_clms0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
my $ct_clms0_in_wd = @{$set_clms0_in_wd};

(print STDOUT "\n ... The subroutine, 'readin_set_clms' Ended (w/ \$ct_clms0_in_wd = $ct_clms0_in_wd, [\$beg_wd, \$end_wd] = [$beg_wd, $end_wd], [\$if_on_le_msa, \$if_on_re_msa] = [$if_on_le_msa, $if_on_re_msa]) ...\n\n") if ($TEST_LEVEL_MAIN>0);

if ($TEST_LEVEL_MAIN>1) { # (7) #

	my $fname_set_clms0 = 'set_columns0.txt';

	unless ( print_out_set_clms ($maindir_in_chk, $fname_set_clms0, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING, the subroutine, 'print_out_set_clms', Ended RECORDING information (on the main window) into the output file, \"$maindir_in_chk/$fname_set_clms0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (7) # ...}"

	##### End of "Read in the Initial Set of Columns." #####


	##### Read in the Initial List of Gap-Blocks. #####

## my $infile_list_gblks0  = "$maindir_in/list_gblocks0.txt";

my ($list_gblks0_in_wd) = readin_list_gblks ($infile_list_gblks0);
unless (defined $list_gblks0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_list_gblks', FAILED on the input file, \"$infile_list_gblks0\" !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
my $ct_gblks0_in_wd = @{$list_gblks0_in_wd};

(print STDOUT "\n ... The subroutine, 'readin_list_gblks' Ended (w/ \$ct_gblks0_in_wd = $ct_gblks0_in_wd) ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (8) #

	my $fname_list_gblks0 = 'list_gblocks0_org.txt';

	unless ( print_out_readin_list_gblks ($maindir_in_chk, $fname_list_gblks0, @{$list_gblks0_in_wd}, 'org') ) { # REVISED on 2019/06/27.
	##unless ( print_out_readin_list_gblks ($maindir_in_chk, $fname_list_gblks0, @{$list_gblks0_in_wd}) ) { ...}
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING 'readin_list_gblks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'original' gap-blocks in the window) into the output file, \"$maindir_in_chk/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (8) # ...}"


	## NOTE: Because the node (branch) ID assignments may have changed from the environment where $infile_list_gblks0 was output to the current environment,
	##	we MUST examine the node (branch) IDs, and, if necessary, REVISE them in "(Pre 3)"!!

	##### End of "Read in the Initial List of Gap-Blocks." #####



	####### Perform some Preliminary Processes #######

(print STDOUT "\n<<<<< Perform some Preliminary Processes >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


	## (Pre 1) Assign indices of MSA rows under each node to the node. ##

(print STDOUT '<<< Preliminary 1: Compute %{$node2exoffs_msa_rows}, %{$node2depth2}, @{$nodeids_in_msa} & %{$nodeid2msa_row}, ', "by performing 'wrapper_mk_node2exoffs_msa_rows' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my ($node2exoffs_msa_rows, $node2depth2, $nodeids_in_msa, $nodeid2msa_row)
	= wrapper_mk_node2exoffs_msa_rows ($topnode_id, %node2ch, @{$seqnames}, %name2id);
unless (defined $node2exoffs_msa_rows) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'wrapper_mk_node2exoffs_msa_rows', FAILED on the input tree !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'wrapper_mk_node2exoffs_msa_rows' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (Pre 1) #

	my $ofile_name = 'out_wrapper_mk_node2exoffs_msa_rows.txt';

	print STDOUT "\n ... Check the outputs of 'wrapper_mk_node2exoffs_msa_rows', by performing 'print_out_wrapper_mk_node2exoffs_msa_rows' ...\n\n";
	unless ( print_out_wrapper_mk_node2exoffs_msa_rows ($maindir_interim, $ofile_name, %{$node2exoffs_msa_rows}, %{$node2depth2}, @{$nodeids_in_msa}, %{$nodeid2msa_row}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_wrapper_mk_node2exoffs_msa_rows' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_wrapper_mk_node2exoffs_msa_rows', Ended ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 1) # ...}"

	## End of "(Pre 1) Assign indices of MSA rows under each node to the node." ##


	## (Pre 2) Using %{$node2exoffs_msa_rows}, assosiate some sets of MSA rows to the upper- & lower-sides of the branches. ## (Added on 2019/06/25.)

(print STDOUT '<<< Preliminary 2: Compute %{$set_msa_rows2br_up_or_lw}, ', "by performing 'mk_set_msa_rows2br_up_or_lw_dp' >>>\n\n") if ($TEST_LEVEL_MAIN>1); # Modified on 2019/07/03.
##(print STDOUT '<<< Preliminary 2: Compute %{$set_msa_rows2br_up_or_lw}, ', "by performing 'mk_set_msa_rows2br_up_or_lw' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my @set_all_msa_rows = (0 .. $ct_seqs -1);

my ($set_msa_rows2br_up_or_lw) = mk_set_msa_rows2br_up_or_lw_dp (%{$node2exoffs_msa_rows}, @set_all_msa_rows, %{$node2depth2}); # Modified on 2019/07/03.
##my ($set_msa_rows2br_up_or_lw) = mk_set_msa_rows2br_up_or_lw (%{$node2exoffs_msa_rows}, @set_all_msa_rows);
unless (defined $set_msa_rows2br_up_or_lw) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_set_msa_rows2br_up_or_lw_dp' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'mk_set_msa_rows2br_up_or_lw_dp' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (Pre 2) #

	my $ofile_name = 'out_mk_set_msa_rows2br_up_or_lw.txt';

	print STDOUT "\n ... Check the outputs of 'mk_set_msa_rows2br_up_or_lw_dp', by performing 'print_out_mk_set_msa_rows2br_up_or_lw' ...\n\n"; # Modified on 2019/07/03.
	##print STDOUT "\n ... Check the outputs of 'mk_set_msa_rows2br_up_or_lw', by performing 'print_out_mk_set_msa_rows2br_up_or_lw' ...\n\n";
	unless ( print_out_mk_set_msa_rows2br_up_or_lw ($maindir_interim, $ofile_name, %{$set_msa_rows2br_up_or_lw}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_mk_set_msa_rows2br_up_or_lw' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_mk_set_msa_rows2br_up_or_lw', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 2) # ...}"


	## End of "(Pre 2) Using %{$node2exoffs_msa_rows}, assosiate some sets of MSA rows to the upper- & lower-sides of the branches." ##


	## (Pre 3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary. ## (Added on 2019/06/26-27)

(print STDOUT '<<< Preliminary 3: Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary,', " via 'revise_list_gblocks' >>>\n\n") if ($TEST_LEVEL_MAIN>1);


unless ( revise_list_gblocks (@{$list_gblks0_in_wd}, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa}) ) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'revise_list_gblocks' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'revise_list_gblocks' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>1) { # (Pre 3) #

	my $fname_list_gblks0 = 'list_gblocks0_rvsd.txt';

	unless ( print_out_readin_list_gblks ($maindir_interim, $fname_list_gblks0, @{$list_gblks0_in_wd}, 'rvsd') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_readin_list_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    (print STDOUT "\n ... For the purpose of CHECKING 'revise_list_gblocks', the subroutine, 'print_out_readin_list_gblks', Ended RECORDING information (on the 'revised' gap-blocks in the window) into the output file, \"$maindir_interim/$fname_list_gblks0\" ... \n\n") if ($TEST_LEVEL_MAIN >1);

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 3) # ...}"

	## End of "(Pre 3) Using %{$set_msa_rows2br_up_or_lw} & @{$nodeids_in_msa}, REVISE @{$list_gblks0_in_wd}, if necessary." ##


	### (Pre 5)  Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, via 'enum_psm_indel_hsts_for_msa_finer_PRE'. ### (Added on 2019/06/25)

(print STDOUT '<<< Preliminary 5: Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, ', "by performing 'enum_psm_indel_hsts_for_msa_finer_PRE' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

(print STDOUT "\n (If you want to know what's going on, SWITCH ON the global variable, \$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE .)\n\n") if ($TEST_LEVEL_MAIN>1);

my ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes)
	= enum_psm_indel_hsts_for_msa_finer_PRE ($topnode_id, %node2ch, %node2pa);
unless (defined $br2nn) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' Ended ...\n\n") if ($TEST_LEVEL_MAIN>1);

if ($TEST_LEVEL_MAIN>3) { # (Pre 5) #

	print STDOUT "\n ... Check the outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE', by performing 'print_out_tree_profile' ...\n\n";

	unless ( print_out_tree_profile ($maindir_interim, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, %{$br2nn}, @{$set_all_nodes}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_tree_profile' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_tree_profile', Ended (w/ records under \"$maindir_interim/\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>3) { # (Pre 5) # ...}"

	### End of "(Pre 5)  Compute %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq} & %{$br2nn}, via 'enum_psm_indel_hsts_for_msa_finer_PRE'." ###


	### (Pre 6) Compute %{$br2tr_mtrx}, via 'calc_br2tr_mtrx_TN'. ### (Added on 2019/06/27)

(print STDOUT "<<< Preliminary 6: Assign a substitution transition matrix to each branch, via 'calc_br2tr_mtrx_TN' >>>\n\n") if ($TEST_LEVEL_MAIN>1);

# @root_resfreqs = @NT_FREQS; # ADDED on 2019/06/27.
# our ($ALPHA1, $ALPHA2, $BETA);

my $br2tr_mtrx = calc_br2tr_mtrx_TN (%br_attr, @root_resfreqs, $ALPHA1, $ALPHA2, $BETA);
unless (defined $br2tr_mtrx) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'calc_br2tr_mtrx_TN', FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_br2tr_mtrx_TN', Ended ...\n\n") if ($TEST_LEVEL_MAIN >1);

if ($TEST_LEVEL_MAIN >2) { # (Pre 6) #

	my $ofile_name = 'br2tr_mtrx.txt';

	print STDOUT "\n ... Check the outputs of 'calc_br2tr_mtrx_TN', by performing 'print_out_br2tr_mtrx' ...\n\n";
	unless ( print_out_br2tr_mtrx ($maindir_interim, $ofile_name, %{$br2tr_mtrx}, @RESIDUES) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2tr_mtrx' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2tr_mtrx', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 6) # ...}"

	### End of "(Pre 6) Compute %{$br2tr_mtrx}, via 'calc_br2tr_mtrx_TN'." ### 


	### (Pre 6') Compute the residue frequences at ALL nodes. ### (Added on 2019/06/27)

(print STDOUT "<<< Preliminary 6': Compute the residue frequences at ALL nodes, via 'calc_node2ntfreqs_td'. >>>\n\n") if ($TEST_LEVEL_MAIN>1);

my $node2ntfreqs = calc_node2ntfreqs_td (@root_resfreqs, %{$br2tr_mtrx}, %node2ch, $topnode_id);

unless (defined $node2ntfreqs) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine, 'calc_node2ntfreqs_td', FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_node2ntfreqs_td', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL_MAIN >2) { # (Pre 6') #

	my $ofile_name = 'node2ntfreqs.txt';

	print STDOUT "\n ... Check the outputs of 'calc_node2ntfreqs_td', by performing 'print_out_node2ntfreqs' ...\n\n";
	unless ( print_out_node2ntfreqs ($maindir_interim, $ofile_name, %{$node2ntfreqs}, @labels_resfreqs) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2ntfreqs' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2ntfreqs', Ended (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 6') # ...}"

	# End of "(Pre 6') Compute the residue frequences at ALL nodes." #


	### (Pre 7) Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0}, via 'calc_prob_descresstats_cond_ancstat2'. ### (Added on 2019/06/27.)
	# (That is, the probabilities of the residue states of extant sequences on the descendant side of each branch, conditioned on the residue state at either end of the branch.)

(print STDOUT '<<< Preliminary 7: Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0},', " via 'calc_prob_descresstats_cond_ancstat2'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($node2set_sw_cond_probs0, $node2set_sw_extcond_probs0)
 = calc_prob_descresstats_cond_ancstat2 (@{$in_msa}, @{$seqnames}, %name2id, %node2ch, %{$br2tr_mtrx}, $GAP);

unless (defined $node2set_sw_cond_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_prob_descresstats_cond_ancstat2', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);


if ($TEST_LEVEL_MAIN >3) { # (Pre 7) #

	my $ofile_name1 = 'node2set_sw_cond_probs0.txt';
	my $ofile_name2 = 'node2set_sw_extcond_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_prob_descresstats_cond_ancstat2', by performing 'print_out_node2set_sw_cond_probs' TWICE ...\n\n";
	unless ( print_out_node2set_sw_cond_probs ($maindir_interim, $ofile_name1, %{$node2set_sw_cond_probs0}, @NUCLEOTIDES, '%{$node2set_sw_cond_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2set_sw_cond_probs' FAILED for '", '%{$node2set_sw_cond_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2set_sw_cond_probs', Ended on '", '%{$node2set_sw_cond_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name1\") ...\n\n";

	unless ( print_out_node2set_sw_cond_probs ($maindir_interim, $ofile_name2, %{$node2set_sw_extcond_probs0}, @NUCLEOTIDES, '%{$node2set_sw_extcond_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_node2set_sw_cond_probs' FAILED for '", '%{$node2set_sw_extcond_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_node2set_sw_cond_probs', Ended on '", '%{$node2set_sw_extcond_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name2\") ...\n\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 7) # ...}"

	# END of "(Pre 7) Compute %{$node2set_sw_cond_probs0} & %{$node2set_sw_extcond_probs0}, via 'calc_prob_descresstats_cond_ancstat2'." #

###=The Following Preliminary Processes May be UNnecessary; if so, DISABLE them.

	### (Pre 8) Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0}, via 'calc_joint_prob_ancresstats_and_ancstat2'. ### (Added on 2019/06/28.)
	### (That is, the joint probabilities of the residue state of the upper-end of each branch and the residue states of extant sequences on the ancestor side of the branch.)

(print STDOUT '<<< Preliminary 8: Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0},', " via 'calc_joint_prob_ancresstats_and_ancstat2'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($br2set_sw_jnt_probs0, $br2set_sw_extjnt_probs0)
    = calc_joint_prob_ancresstats_and_ancstat2 (@{$in_msa}, @{$seqnames}, %name2id, %node2ch, %{$br2tr_mtrx}, $GAP, @root_resfreqs, %{$node2set_sw_extcond_probs0}) ;

unless (defined $br2set_sw_jnt_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_joint_prob_ancresstats_and_ancstat2', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);


if ($TEST_LEVEL_MAIN >3) { # (Pre 8) #

	my $ofile_name1 = 'br2set_sw_jnt_probs0.txt';
	my $ofile_name2 = 'br2set_sw_extjnt_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_joint_prob_ancresstats_and_ancstat2', by performing 'print_out_br2set_sw_jnt_probs' TWICE ...\n\n";
	unless ( print_out_br2set_sw_jnt_probs ($maindir_interim, $ofile_name1, %{$br2set_sw_jnt_probs0}, @NUCLEOTIDES, '%{$br2set_sw_jnt_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2set_sw_jnt_probs' FAILED for '", '%{$br2set_sw_jnt_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2set_sw_jnt_probs', Ended on '", '%{$br2set_sw_jnt_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name1\") ...\n\n";

	unless ( print_out_br2set_sw_jnt_probs ($maindir_interim, $ofile_name2, %{$br2set_sw_extjnt_probs0}, @NUCLEOTIDES, '%{$br2set_sw_extjnt_probs0}') ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_br2set_sw_jnt_probs' FAILED for '", '%{$br2set_sw_extjnt_probs0}', "' in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_br2set_sw_jnt_probs', Ended on '", '%{$br2set_sw_extjnt_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name2\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 8) # ...}"

	### End of "(Pre 8) Compute %{$br2set_sw_jnt_probs0} & %{$br2set_sw_extjnt_probs0}, via 'calc_joint_prob_ancresstats_and_ancstat2'." ###


	### (Pre 9) Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 9: Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my $set_clm_probs0 = calc_clm_probs (@root_resfreqs, %{$node2set_sw_cond_probs0}, $topnode_id);
unless (defined $set_clm_probs0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_clm_probs', Ended ...\n\n") if ($TEST_LEVEL>0);

my $ct_clms0 = @{$set_clm_probs0};
unless ($ct_clms0 == $ct_clms0_in_wd) {
	print STDOUT "[ERROR '$program_name'] The number of columns according to 'calc_clm_probs' (=$ct_clms0) CLASHES with that according to 'readin_set_clms' (=$ct_clms0_in_wd) !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}

if ($TEST_LEVEL_MAIN >3) { # (Pre 9) # 

	my $ofile_name = 'set_clm_probs0.txt';

	print STDOUT "\n ... Check the outputs of 'calc_clm_probs', by performing 'print_out_set_clm_probs' ...\n\n";
	unless ( print_out_set_clm_probs ($maindir_interim, $ofile_name, @{$set_clm_probs0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clm_probs' FAILED on ", '@{$set_clm_probs0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_set_clm_probs', Ended on '", '@{$set_clm_probs0}', "' (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >3) { # (Pre 9) #  ...}"

	### End of "(Pre 9) Compute the probability of each column (under the given substitution model), via 'calc_clm_probs'." ###

###=cut # END of "The Following Preliminary Processes May be UNnecessary; if so, DISABLE them."


	### (Pre 10) Prepare @{$inter_block_relations0}, via 'mk_inter_block_relations'. ### (Added on 2019/06/29.) 

(print STDOUT '<<< Preliminary 10: Compute @{$inter_block_relations0}, ', " via 'mk_inter_block_relations'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

# @{$list_gblks[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

		# MODIFIED on 2019/07/02. #
our ($INDX_BR, $INDX_U_OR_D) = (0, 1);
(print STDOUT 'Manually Set: ($INDX_BR, $INDX_U_OR_D) = ', "($INDX_BR, $INDX_U_OR_D) .\n\n") if ($TEST_LEVEL_MAIN>1);
##my ($indx_br, $indx_u_or_d) = (0, 1);
##(print STDOUT 'Manually Set: ($indx_br, $indx_u_or_d) = ', "($indx_br, $indx_u_or_d) .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($inter_block_relations0) = mk_inter_block_relations (@{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %node2pa, %node2ch, $topnode_id, %{$br2eq}); # Modified on 2019/07/02. #
##my ($inter_block_relations0) = mk_inter_block_relations (@{$list_gblks0_in_wd}, $indx_br, $indx_u_or_d, %node2pa, %node2ch, $topnode_id, %{$br2eq});
unless ($inter_block_relations0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_inter_block_relations' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_inter_block_relations', Ended ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN>1) { # (Pre 10) #

	my $ofile_name = 'inter_block_relations0.txt';

	print STDOUT "\n ... Check the outputs of 'mk_inter_block_relations', by performing 'print_out_inter_block_relations' ...\n\n";
	unless ( print_out_inter_block_relations ($maindir_interim, $ofile_name, @{$inter_block_relations0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_inter_block_relations' FAILED on ", '@{
$inter_block_relations0}', " in the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_inter_block_relations', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 10) # ...}"

	### End of "(Pre 10) Prepare @{$inter_block_relations0}, via 'mk_inter_block_relations'." ###  


	### (Pre 11) MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd. ### (Added on 2019/06/29.)

(print STDOUT '<<< Preliminary 11: MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd >>>', "\n\n") if ($TEST_LEVEL_MAIN >0);

# @{$list_gblks[$indx_gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, \@affd_exnodes, \@affd_msa_rows),

my @bds_gblocks0_in_wd = ();
my @rows_affd_by_blks0_in_wd = ();
for (my $gb=0; $gb <$ct_gblks0_in_wd; $gb++) {

	my ($br, $up_or_lw, $lm_clm, $rm_clm, $cnct_affd_exnodes, $cnct_affd_msa_rows, $set_affd_exnodes, $set_affd_msa_rows) = @{$list_gblks0_in_wd->[$gb]};

	$bds_gblocks0_in_wd[$gb] = [$lm_clm - $beg_wd, $rm_clm - $beg_wd];
	$rows_affd_by_blks0_in_wd[$gb] = $set_affd_msa_rows;
}

if ($TEST_LEVEL_MAIN>1) { # (Pre 11) #

	my $ofile_name = 'row_clm_info_gblocks0.txt';

	print STDOUT "\n", ' ... Check the contents of @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd, ', "by performing 'print_out_row_clm_info_gblocks' ...\n\n";
	unless ( print_out_row_clm_info_gblocks ($maindir_interim, $ofile_name, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_row_clm_info_gblocks' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_row_clm_info_gblocks', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 11) # ...}"

	### End of "(Pre 11) MANUALLY prepare @bds_gblocks0_in_wd & @rows_affd_by_blks0_in_wd." ### 


	### (Pre 12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 12: Prepare the coordinate frame for the multiple-\"shifts\" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'.  >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($orgs_gb0_coords, $bds_gb0_coords, $gb0_coords_wd_frm)
	= mk_orgs_bds_gb_coords (@bds_gblocks0_in_wd, 0, $end_wd - $beg_wd, @{$inter_block_relations0}, $HF_WDTH_SHIFT);
unless (defined $gb0_coords_wd_frm) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_orgs_bds_gb_coords', Ended ...\n\n") if ($TEST_LEVEL_MAIN >0);

if ($TEST_LEVEL_MAIN>1) { # (Pre 12) #

	my $ofile_name = 'orgs_bds_gblocks0_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords', by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($maindir_interim, $ofile_name, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the main window (w/ records in \"$maindir_interim/$ofile_name\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 12) # ...}"

	### End of "(Pre 12) Prepare the coordinate frame for the multiple-"shifts" from the initial alignment of the window, via 'mk_orgs_bds_gb_coords'." ###


	### (Pre 13: IMPORTANT!!) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_wd}. ### (Added on 2019/06/29.)

(print STDOUT "<<< Preliminary 13 (IMPORTANT!!): Manually SUBTRACT \$beg_wd (=$beg_wd) from ALL the site numbers in ", '@{$set_clms0_in_wd}', ". >>>\n\n") if ($TEST_LEVEL_MAIN >0);

for (my $c=0; $c <$ct_clms0_in_wd; $c++) { # Outer for-loop (over columns). #

	my $column = $set_clms0_in_wd->[$c];
	my $ct_rows = @{$column};

	for (my $r=0; $r <$ct_rows; $r++) { # Inner for-loop (over rows). #
	    unless ($column->[$r] eq $CHAR_ABS) { $column->[$r] -= $beg_wd; }
	} # End of the inner for-loop (over rows). #

} # End of the outer for-loop (over columns). #

if ($TEST_LEVEL_MAIN>1) { # (Pre 13) #

        my $fname_set_clms0 = 'set_columns0_af_coord_chg.txt';

        ##unless ( print_out_set_clms ($maindir_interim, $fname_set_clms0, @{$set_clms0_in_wd}, 0, $end_wd - $beg_wd, $if_on_le_msa, $if_on_re_msa) ) {
        unless ( print_out_set_clms ($maindir_interim, $fname_set_clms0, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa) ) { # CORRECTED on 2020/05/17. #
                print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_set_clms' FAILED for the main window !\n";
                print STDERR "[ERROR '$plscript_name'] $! !\n";
                die ;
        }
    print STDOUT "\n ... For the purpose of CHECKING the result of  SUBTRACTING \$beg_wd (=$beg_wd) from the sites in the columns, the subroutine, 'print_out_set_clms', Ended RECORDING ", '@{$set_clms0_in_wd}', " (AFTER the SUBTRACTION) into the output file, \"$maindir_interim/$fname_set_clms0\" ... \n\n";

} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre 13) # ...}"

	### End of "(Pre 13) Manually SUBTRACT $beg_wd from ALL the site numbers in @{$set_clms0_in_wd}." ###


	### (Pre 14) Read in the list of (merged) purge-candidate regions, and SUBTRACT $beg_wd from ALL positions. ### (ADDED on 2019/07/03.)

(print STDOUT "<<< Preliminary 14: Read in the list of (merged) purge-candidate regions, and manually SUBTRACT \$beg_wd (=$beg_wd) from ALL positions. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($list_mgd_pcands0, $maxspan_mgd_pcand0) = readin_list_mgd_pcands ($infile_pcands0);
unless (defined $list_mgd_pcands0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_list_mgd_pcands' FAILED on the input file, \"$infile_pcands0\" !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
my $ct_mpcands0 = @{$list_mgd_pcands0};
(print STDOUT "\n ... The subroutine 'readin_list_mgd_pcands' Ended for the main window (w/ #{purge-candidates}=$ct_mpcands0, Max{span}=$maxspan_mgd_pcand0) ...\n\n") if ($TEST_LEVEL_MAIN >1);


unless ( relative_coords_for_list_mgd_pcands (@{$list_mgd_pcands0}, $beg_wd) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'relative_coords_for_list_mgd_pcands' FAILED for ", '@{$list_mgd_pcands0}', " of the main window !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine 'relative_coords_for_list_mgd_pcands' Ended for ", '@{$list_mgd_pcands0}', " of the main window ...\n\n") if ($TEST_LEVEL_MAIN >1);


if ($TEST_LEVEL_MAIN >1) { # (Pre 14) #

	my $ofname_pcands = 'list_rlv_pcands.txt';

	unless ( print_out_list_mgd_pcands ($maindir_interim, $ofname_pcands, @{$list_mgd_pcands0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_mgd_pcands' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... For the purpose of CHECKING the result of SUBTRACTING \$beg_wd (=$beg_wd) from the positions regarding the purge-candidate regions, the subroutine, 'print_out_list_mgd_pcands', Ended RECORDING ", '@{$list_mgd_pcands0}', " (AFTER the SUBTRACTION) into the output file, \"$maindir_interim/$ofname_pcands\" ... \n\n";

} # End of "if ($TEST_LEVEL_MAIN >1) { # (Pre 14) # ...}"

	### End of "(Pre 14) Read in the list of (merged) purge-candidate regions, and manually SUBTRACT $beg_wd from ALL positions." ###


		## IMPORTED (w/ some modifications) from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/18. (Pre 15 & 16 & 17) ##

	### (Pre 15) Prepare @{$gblock_sizes0}, via 'mk_list_gblock_sizes'. ### (ADDED on 2019/07/16.)

(print STDOUT '<<< Preliminary 15: Compute @{$gblock_sizes0}, ', " via 'mk_list_gblock_sizes'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

our ($INDX_LBD, $INDX_RBD) = (0, 1);
(print STDOUT 'Manually Set: ($INDX_LBD, $INDX_RBD) = ', "($INDX_LBD, $INDX_RBD) .\n\n") if ($TEST_LEVEL_MAIN>1); 

my ($gblock_sizes0) = mk_list_gblock_sizes (@bds_gblocks0_in_wd, $INDX_LBD, $INDX_RBD, @{$inter_block_relations0});
unless (defined $gblock_sizes0) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'mk_list_gblock_sizes' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'mk_list_gblock_sizes', Ended for the main window ...\n\n") if ($TEST_LEVEL_MAIN>0);

if ($TEST_LEVEL_MAIN >2) { # (Pre 15) #

	my $ofname = 'gblock_sizes0.txt';

	unless ( print_out_list_gblock_sizes ($maindir_interim, $ofname, @{$gblock_sizes0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_list_gblock_sizes' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_list_gblock_sizes', Ended on the main window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 15) # ...}"

	### End of "(Pre 15) Prepare @{$gblock_sizes0}, via 'mk_list_gblock_sizes'." ###


	### (Pre 16) Prepare @{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}, via 'mk_lists_gblocks_w_spec_bounds'. ### (ADDED on 2019/07/16-17.) #

(print STDOUT "\n\n<<< Preliminary 16: Compute ", '@{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}', ", via 'mk_lists_gblocks_w_spec_bounds'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($le_wd, $re_wd) = (0, $ct_clms0_in_wd-1);

(print STDOUT 'Manually Set: ($le_wd, $re_wd) = ', "($le_wd, $re_wd) .\n\n") if ($TEST_LEVEL_MAIN>1);

my ($gblks_w_spec_lb0, $gblks_w_spec_rb0) = mk_lists_gblocks_w_spec_bounds ($le_wd, $re_wd, @bds_gblocks0_in_wd);
unless (defined $gblks_w_spec_rb0) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_lists_gblocks_w_spec_bounds' FAILED on the subject window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n --- Subroutine 'mk_lists_gblocks_w_spec_bounds' Ended on the subject window ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN >2) { # (Pre 16) #

	my $ofname = 'gblks_w_spec_bds0.txt';

	unless ( print_out_gblks_w_spec_bds ($maindir_interim, $ofname, @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_gblks_w_spec_bds' FAILED for the subject window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_gblks_w_spec_bds', Ended on the subject window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";


} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 16) # ...}"

	### End of "(Pre 16) Prepare @{$gblks_w_spec_lb0} & @{$gblks_w_spec_rb0}, via 'mk_lists_gblocks_w_spec_bounds'." ###


	### (Pre 17) Prepare @{$cts_vincluding0}, @{$cts_vincluded0} & @{$cts_onx0},', " via 'cts_vincluding_vincluded_onx_gblks'. ### (ADDED on 2019/07/16-17.) #

(print STDOUT "\n\n", '<<< Preliminary 17: Compute @{$cts_vincluding0}, @{$cts_vincluded0} & @{$cts_onx0},', " via 'cts_vincluding_vincluded_onx_gblks'. >>>\n\n") if ($TEST_LEVEL_MAIN >0);

my ($cts_vincluding0, $cts_vincluded0, $cts_onx0) = cts_vincluding_vincluded_onx_gblks (@{$inter_block_relations0});
unless (defined $cts_onx0) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'cts_vincluding_vincluded_onx_gblks' FAILED on the main window !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT "\n --- Subroutine 'cts_vincluding_vincluded_onx_gblks' Ended on the main window ...\n\n") if ($TEST_LEVEL_MAIN>0);


if ($TEST_LEVEL_MAIN >2) { # (Pre 17) #

	my $ofname = 'cts_vincluding_vincluded_onx_gblks0.txt';

	unless ( print_out_cts_vincluding_vincluded_onx_gblks ($maindir_interim, $ofname, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks' FAILED for the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
    print STDOUT "\n ... The subroutine 'print_out_cts_vincluding_vincluded_onx_gblks', Ended on the main window (w/ records in \"$maindir_interim/$ofname\") ...\n\n";

} # End of "if ($TEST_LEVEL_MAIN >2) { # (Pre 17) # ...}"

	### End of "(Pre 17) Prepare @{$cts_vincluding}, @{$cts_vincluded} & @{$cts_onx},', " via 'cts_vincluding_vincluded_onx_gblks'." ###


		## END of "IMPORTED (w/ some modifications) from 'anex_for_sgl_wd.ver0.5.pl' on 2019/09/18. (Pre 15 & 16 & 17)" ##


	### (Pre-MAIN) Prepare the remaining input arguments for 'nonrec_mlt_shifts_gapblocks'. ### (Added on 2019/06/29.)

(print STDOUT "\n<<< Pre-MAIN: Prepare some input arguments for 'nonrec_mlt_shifts_gapblocks'. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

our @SET_INGREDIENTS_PRB_MSA_INDEL = ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, $br2log_mfacs);

my %clm2ln_prb_sbst;  # = ( join (':', @column_in_site_numbers) => $log_probability_under_substitution_model, ...) .
my @set_ingredients_indel_hsts = ({}, [], [], []); # = (\%gpattern2indx, \@set_gpatterns, \@set_node2dparsstat, \@set_br2change) .

if ($TEST_LEVEL_MAIN>1) { # (Pre-MAIN) #

	print STDOUT "[ The Input Arguments for 'nonrec_mlt_shifts_gapblocks' Manually Set Here: ]\n\n";
	print STDOUT '@SET_INGREDIENTS_PRB_MSA_INDEL (= ($TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, $br2log_mfacs))', "\n";
	print STDOUT ' = (', join (', ', @SET_INGREDIENTS_PRB_MSA_INDEL), ") .\n\n";
	print STDOUT '%clm2ln_prb_sbst = ()', " .\n\n";
	print STDOUT '@set_ingredients_indel_hsts = ({}, [], [], [])', " .\n\n\n";
	
} # End of "if ($TEST_LEVEL_MAIN>1) { # (Pre-MAIN) # ...}"

	### End of "(Pre-MAIN) Prepare the remaining input arguments for 'nonrec_mlt_shifts_gapblocks'." ###


	### (Pre-MAIN2) Open a file-handle for the "Synopsis" Output File ($outfile_synopsis). ### (Added in REVISION on 2020/04/22.)

# my $outfile_synopsis = "$top_work_dir/synopsis_results_mshifts.txt";

(print STDOUT "\n<<< Pre-MAIN2: Open a file-handle for the \"Synopsis\" Output File (=\"$outfile_synopsis\").  >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

my $fh_osyn = open_output_file ($outfile_synopsis, 0);
unless (defined $fh_osyn) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'open_output_file' FAILED to  open a filehandle for the \"Synopsis\" Output File, \"$outfile_synopsis\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}

print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

my $ln_prb_msa_init;

	# perform_mshifts_and_print_inputs_outputs

	### End of "(Pre-MAIN2) Open a file-handle for the "Synopsis" Output File ($outfile_synopsis)." ### 


			### NOTE added in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07):
			###	UP TO HERE, the program is almost IDENTICAL to that for 'anex_for_sgl_wd.ver0.5.pl'
			###	(EXCEPT that some processes dedicated for the sub-windows are removed).
			###	In the Following, though, The MAJOR REVISIONS WILL BE MADE!!


			### ADDED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07-). ###

	##### (MAIN III): Explore the Neighborhoods of the Input Alignment via "Reverse-Purge" (aka. "Ex-Nihilo") + Multiple-"Shift"s. #####

(print STDOUT "\n<<<<< MAIN (III): Explore the Neighborhoods of the Input Alignment via \"Reverse-Purge\" (aka. \"Ex-Nihilo\") + Multiple-\"Shift\"s. >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


	# (MAIN III-0) Compute the Log-Probabilities & some Other key Quantities of the Initial MSA. #

if (1) { # (MAIN III-0) #

   (print STDOUT "\n<<< MAIN III-0: Compute the Log-Probabilities & some Other key Quantities of the Initial MSA. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

	my @list_excluded_bin_hrels0 = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels0 = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);

			# (i) Print out specific Input Data. #

	my $subdir = "$maindir_out/III_0.Initial_MSA";
	my $subdir_in = "$subdir/In";

	my @ofnames = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/09/13.

	unless (print_in_nr_mshifts_glks ($subdir_in, @ofnames, @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, @list_excluded_bin_hrels0) ) { # REVISED on 2019/09/13.
	##unless (print_in_nr_mshifts_glks ($subdir_in, @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd, @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @{$set_clms0_in_wd}, $beg_wd, $end_wd, $if_on_le_msa, $if_on_re_msa, @list_excluded_bin_hrels0) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_in_nr_mshifts_gblks' FAILED for MAIN III-0 on the input MSA !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'print_in_nr_mshifts_gblks', Ended for MAIN III-0 ...\n\n") if ($TEST_LEVEL_MAIN >0);


			# (ii) Perform 'compt_log_msa_prob_plus'. #

	my ($if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $grand_code_topology0)
	= compt_log_msa_prob_plus (
	    @{$orgs_gb0_coords}, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$set_clms0_in_wd}, @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, @bds_gblocks0_in_wd, @rows_affd_by_blks0_in_wd,
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels0
		);
	unless ($if_success) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'compt_log_msa_prob_plus' FAILED for MAIN III-0 on the input MSA !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'compt_log_msa_prob_plus', Ended for MAIN III-0 (w/ \$grand_code_topology0 = '$grand_code_topology0') ...\n\n") if ($TEST_LEVEL_MAIN >0);


			# (iii) Print out the Output Data. #

	my $subdir_out = "$subdir/Out";
	my $ofname = 'ln_prb_plus_msa.txt';

	unless ( print_out_ln_prb_plus_msa ($subdir_out, $ofname, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $STRING_UNDEF) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_ln_prb_plus_msa' FAILED for MAIN III-0 on the input MSA !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT " ... Subroutine, 'print_out_ln_prb_plus_msa', Ended for MAIN III-0 ...\n\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT " ... Subroutine, 'print_out_ln_prb_plus_msa', Ended for MAIN III-0 (w/ records in \"$subdir_out/$ofname\") ...\n\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/19.

	$ln_prb_msa_init = $ln_prb_msa0; # ADDED in REVISION on 2020/04/22. (MAIN III-0) #

} # End of "if (1) { # (MAIN III-0) # ...}"

	# End of "(MAIN III-0) Compute the Log-Probabilities & some Other key Quantities of the Initial MSA." #


my $subdir_tgt_cands = "$maindir_out/Target_Candidates"; # Added on 2019/09/16. #

	# (MAIN III-1) Actually Perform "Reverse-Purge" (aka. "Ex-Nihilo") + Multiple-"Shift"s. #####

(print STDOUT "\n<<< MAIN III-I: Actually Perform the \"Reverse-Purge\" (aka. \"Ex-Nihilo\") + Multiple-\"Shift\"s. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

my $subdir_MIII_1 = "$maindir_out/III_1.RevPurge_plus_MltShifts"; # Added on 2019/09/16. #


		# IMPORTED from Main I-2 (ii) of "anex_for_sgl_wd.ver0.5.pl" (on 2019/09/18). #
my @null_column = ();
for (1 .. $ct_seqs) { push @null_column, $GAP; }
my $cnct_null_clm = join (':', @null_column);

if ($TEST_LEVEL_MAIN>1) {
	print STDOUT "[ An Additional Input: ]\n\n";
	print STDOUT '$cnct_null_clm = ', $cnct_null_clm, " .\n";
	print STDOUT "\n\n";
} 
		# END of "IMPORTED from Main I-2 (ii) of "anex_for_sgl_wd.ver0.5.pl" (on 2019/09/18)." #

		# ADDED on 2019/09/13. (1a) #

our $INDX_SITES_IN_SGN_WD = 0;
our $max_rpc_size_for_xlevel2 = $MIN_WD_SIZE_PCAND -1;

(print STDOUT 'Manually Set: $max_rpc_size_for_xlevel2 = ', $max_rpc_size_for_xlevel2 , " .\n\n\n") if ($TEST_LEVEL_MAIN >1); # (Added on 2019/09/16.) #
		# End of "ADDED on 2019/09/13. (1a)" #

		# ADDED on 2019/09/13. (1b) #


	# NOTE: ASSUME that only A SINGLE merged purge-candidate region is involved in this PCC window. #

for my $i (0) { # Main OUTER for-loop (over the (merged) \"purge\"-candidates). # ADDED on 2019/09/18. #

	my $session = 'MAIN III-1'; # Added on 2019/09/18. #

	my $sbsbdir_tgt_cands = "$subdir_tgt_cands/MPCand_$i"; # Added on 2019/09/18. #

    ##my ($beg_mpc, $end_mpc, $br_mpc, $set_msa_rows_under_br_mpc, $set_constituent_sites_mpc, $pv_mpc, $pv0_mpc, $set_sgn_wds_included) = @{$list_mgd_pcands0->[0]};
    my ($beg_mpc, $end_mpc, $br_mpc, $set_msa_rows_under_br_mpc, $set_constituent_sites_mpc, $pv_mpc, $pv0_mpc, $set_sgn_wds_included) = @{$list_mgd_pcands0->[$i]};

   (print STDOUT "\n{{ The $i th \"purge\"-candidate  (w/ [beg, end] = [$beg_mpc, $end_mpc], branch = $br_mpc) }}\n\n") if ($TEST_LEVEL_MAIN>0);


   my ($rvprg_cand_sizes) = list_revpurge_cand_sizes (@{$set_constituent_sites_mpc}, @{$set_sgn_wds_included}, $INDX_SITES_IN_SGN_WD, $XLEVEL, $max_rpc_size_for_xlevel2);
  unless (defined $rvprg_cand_sizes) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'list_revpurge_cand_sizes' FAILED for ", '@{$list_mgd_pcands0->[$i]}', " (w/ \$i = $i) of the subject window !!\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
   }
   (print STDOUT "\n ... The subroutine 'list_revpurge_cand_sizes' Ended for ", '@{$list_mgd_pcands0->[$i]}', " (w/ \$i = $i) of the subject window ...\n\n") if ($TEST_LEVEL_MAIN >1);

		# END of "ADDED on 2019/09/13. (1b)" #

   if (1) { # (Main III-1 (i) # (Added on 2019/09/16.)

	my $ofile_name1 = 'info_revpurge_cand_sizes.txt';

	print STDOUT "\n ... Record the outputs of 'list_revpurge_cand_sizes', by performing 'print_out_revpurge_cand_sizes' ...\n\n";

	unless ( print_out_revpurge_cand_sizes ($sbsbdir_tgt_cands, $ofile_name1, @{$rvprg_cand_sizes}, $XLEVEL, $max_rpc_size_for_xlevel2) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_revpurge_cand_sizes' FAILED on ", '@{$rvprg_cand_sizes}', " in the subject window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	##print STDOUT "\n ... The subroutine, 'print_out_revpurge_cand_sizes', Ended on ", '@{$rvprg_cand_sizes}', "in the main window (w/ records in \"$subdir_tgt_cands/$ofile_name1\") ...\n\n";
	print STDOUT "\n ... The subroutine, 'print_out_revpurge_cand_sizes', Ended on ", '@{$rvprg_cand_sizes}', "in the main window (w/ records in \"$sbsbdir_tgt_cands/$ofile_name1\") ...\n\n"; # CORRECTED on 2020/05/19.

  } # End of "if (1) { # (Main III-1 (i) # ...}"


=Set aside for a Potential Future Use. (on 2019/09/12)

		# In any case, the significant windows could be REDUNDANT if their sizes are identical,
		# considering that Multiple-"Shifts" will follow...
		# => It may be BETTER first to LIST the sizes of "reverse-purge"s to attempt,
		#	and then to ACTUALLY PERFORM the "reverse-purge" for each of the sizes listed.

    foreach my $info_pc ([$set_constituent_sites_mpc, $pv_mpc, $pv0_mpc], @{$set_sgn_wds_included}) { # MAIN foreach-loop (over the significant purge-candidate regions within the merged region). #

	my ($sites_in_pc, $pv_pc, $pv0_pc) = @{$info_pc};
	my $ct_sites_in_pc = @{$sites_in_pc};
	my ($beg_pc, $end_pc) = @{$sites_in_pc}[0, $ct_sites_in_pc-1];
	my $cnct_sites_in_pc = join (',', @{$sites_in_pc});

		### Actual Processes will follow. ###

    } # End of the MAIN foreach-loop (over the significant purge-candidate regions within the merged region). #

=cut # End of "Set aside for a Potential Future Use. (on 2019/09/12)"


		# ADDED on 2019/09/16 & 18. #

   #my $subdir_MIII_1 = "$maindir_out/III_1.RevPurge_plus_MltShifts"; # Added on 2019/09/16. #

   my $sbsbdir = "$subdir_MIII_1/MPCand_$i"; # Modified on 2019/09/18. #
   ##my $sbsbdir = "$subdir_MIII_1/MPCand_0";
   my $sbsbdir_info = "$sbsbdir/Info";


   foreach my $rpc_size (@{$rvprg_cand_sizes}) { # Middle foreach-loop (over $rpc_size). #

	(print STDOUT "\n{{ \$rpc_size = $rpc_size }}\n\n") if ($TEST_LEVEL_MAIN>0);

	my $sbsbdir_out = "$sbsbdir/Out/rpc_size$rpc_size";
	my $sbsbdir_out0 = "$sbsbdir_out/Origin";
	my $sbsbdir_out_ms = "$sbsbdir_out/MShifts";

		### RESTART FROM HERE (on 2019/09/13). ###

	my @curr_gb_coords = @{$orgs_gb0_coords};
	my ($rows_under_brpc, $sites_in_mpc) = ($set_msa_rows_under_br_mpc, $set_constituent_sites_mpc); # Added on 2019/09/18. #

	my ($flag_rvprg, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, $new_indices_old_gbs, $old_indices_new_gbs, $records_sessions, $gb2incr_bl_coord_undo,
	    $new_curr_set_clms_in_wd, $new_curr_bds_gblks_in_wd, $new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm,
	    $new_curr_gblks_w_spec_lb, $new_curr_gblks_w_spec_rb, $rows_affd_by_new_blks_in_wd,
	    $new_inter_block_relations, $new_cts_vincluding, $new_cts_vincluded, $new_cts_onx, $new_gblock_sizes,
	    $new_list_gblocks_in_wd, $med_indices_new_gbs)

		# MODIFIED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/09/18). #
	  # [ = csplit_gblock ( $gb_sbj, $new_size_cmpl, @{$pot_avail_sites}, $CHAR_ABS, ] ... for (MAIN II-S4b) of "anex_for_sgl_wd.ver0.5.pl". #
	  = reverse_purge_gblock_pair (@{$rows_under_brpc}, $rpc_size, @{$sites_in_mpc}, $CHAR_ABS,
		# END of "MODIFIED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/09/18)." #

	     @{$set_clms0_in_wd}, @bds_gblocks0_in_wd, @curr_gb_coords, @{$bds_gb0_coords}, @{$gb0_coords_wd_frm},
	     @{$gblks_w_spec_lb0}, @{$gblks_w_spec_rb0}, @rows_affd_by_blks0_in_wd, $cnct_null_clm,
	     @{$inter_block_relations0}, @{$cts_vincluding0}, @{$cts_vincluded0}, @{$cts_onx0}, @{$gblock_sizes0},  $HF_WDTH_SHIFT,
	     @{$list_gblks0_in_wd}, $INDX_BR, $INDX_U_OR_D, %{$set_msa_rows2br_up_or_lw}, @{$nodeids_in_msa},
	     %node2pa, %node2ch, $topnode_id, %{$br2eq}
		);
	unless ($flag_rvprg) {
	    print STDOUT "[ERROR '$plscript_name'] The MAIN Subroutine 'reverse_purge_gblock_pair' FAILED on the $i th (merged) \"purge\"-candidate for $session (w/ \$rpc_size = $rpc_size) with the subject window !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT "\n --- The MAIN Subroutine 'reverse_purge_gblock_pair' ENDED on the $i th (merged) \"purge\"-candidate  (w/ \$rpc_size = $rpc_size) for $session with the subject window (w/ \$flag_rvprg = $flag_rvprg, \$gb_sbj_NEW = $gb_sbj_NEW, \$gb_cmpl_NEW = $gb_cmpl_NEW, \$new_gb_left = $new_gb_left, \$new_gb_right = $new_gb_right) ...\n\n\n") if ($TEST_LEVEL_MAIN>0);


		# REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (MAIN III-1) #
    if ($IF_TEST_W_COMPACT_MSS) { 

  (print STDOUT "\n ... RE-Define the new coordinate frame as a \"Compact\" Multiple-Shift Space, using \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS ...\n\n\n") if ($TEST_LEVEL_MAIN >0);

	my ($le_new_wd, $re_new_wd) = (0, $#{$new_curr_set_clms_in_wd});

	($new_curr_bl_coords, $new_bds_gb_coords, $new_gb_coords_wd_frm)
	    = mk_orgs_bds_gb_coords (@{$new_curr_bds_gblks_in_wd}, $le_new_wd, $re_new_wd, @{$new_inter_block_relations}, $HF_WDTH_SHIFT_4_CMSS);
	unless (defined $new_gb_coords_wd_frm) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'mk_orgs_bds_gb_coords' FAILED on the $i th (merged) \"purge\"-candidate for $session (w/ \$rpc_size = $rpc_size) on the input MSA (w/ \$HF_WDTH_SHIFT_4_CMSS = $HF_WDTH_SHIFT_4_CMSS)!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}

  if ($TEST_LEVEL_MAIN>1) { # (For CMSS) #

	my $sbsbdir_out4CMSS = "$sbsbdir_out0/Compact_MSS";

	my $ofile_name = 'orgs_bds_gblocks_coords.txt';

	print STDOUT "\n ... Check the outputs of 'mk_orgs_bds_gb_coords' (w/ \$HF_WDTH_SHIFT_4_CMSS=$HF_WDTH_SHIFT_4_CMSS), by performing 'print_out_orgs_bds_gb_coords' ...\n\n";
	unless ( print_out_orgs_bds_gb_coords ($sbsbdir_out4CMSS, $ofile_name, @{$new_curr_bl_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}) ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_orgs_bds_gb_coords' FAILED on the main window !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_orgs_bds_gb_coords', Ended on the $i th (merged) \"purge\"-candidate for $session (w/ \$rpc_size = $rpc_size) the subject window (w/ records in \"$sbsbdir_out4CMSS/$ofile_name\") ...\n\n";

  } # End of "if ($TEST_LEVEL_MAIN>1) { # (For CMSS) # ...}"

    } # End of "if ($IF_TEST_W_COMPACT_MSS) {...}"
		# END of "REVISION IMPORTED from 'anex_for_sgl_wd.ver0.5.pl' (on 2019/11/14). (MAIN III-1)" #


		# (i') Prepare Some Additional Data-Sets. #

	my $gb_sbj = 'None'; # ADDED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/09/18). #

	my $ct_out_gbs = @{$new_curr_bds_gblks_in_wd};

	my @new_list_gblks_in_wd = @{$new_list_gblocks_in_wd};

	my @list_excluded_bin_hrels = (); # NO binary horizontal-relations should be excluded (EXCEPT those specified by default).

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);


		# (ii-0) Print out the Changes Resulting from the "Reverse-Purge". #

	##my $sbsbdir_info = "$sbsbdir/Info";
	my $sbsbsbdir_info = "$sbsbdir_info/rpc_size$rpc_size";

	my @ofnames0 = ('basic_info.txt', 'old_new_indices_gblks.txt', 'rec_split_sessions.txt', 'rec_undo_hovlps.txt');

	# $flag_sspl, $new_gb_left, $new_gb_right, $new_inices_old_gbs, $old_indices_new_gbs, $gb2incr_bl_coord_undo,

	my @split_sessions = @{$records_sessions};
	##my @split_sessions = (); # A dummy list.

	unless ( print_out_info_chngs_by_split ($sbsbsbdir_info, @ofnames0, $flag_rvprg, $gb_sbj, $gb_sbj_NEW, $gb_cmpl_NEW, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @{$med_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) { # MODIFIED on 2019/08/20.
	## unless ( print_out_info_chngs_by_split ($sbsbsbdir_info, @ofnames0, $flag_sspl, $gb_sbj, $new_gb_left, $new_gb_right, @{$new_indices_old_gbs}, @{$old_indices_new_gbs}, @split_sessions, %{$gb2incr_bl_coord_undo}) ) { 
	    print STDOUT "[ERROR '$program_name'] The subroutine 'print_out_info_chngs_by_split' FAILED on the $i th (merged) \"purge\"-candidate region for $session with the subject window (w/ \$rpc_size = $rpc_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   ##(print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th (merged) \"purge\"-candidate region for $session with the subject window (w/ \$rpc_size = $rpc_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);
   (print STDOUT "\n ... Subroutine, 'print_out_info_chngs_by_split', Ended on the $i th (merged) \"purge\"-candidate region for $session with the subject window (w/ \$rpc_size = $rpc_size) (w/ records under the directory, \"$sbsbsbdir_info/\") ...\n\n") if ($TEST_LEVEL_MAIN >0); # Modified on 2020/05/19.


		# (ii) Print out the Gap-Configurations at the "Origin", which resulted from the  "Reverse-Purge". #

		# & (iii) Perform the Multiple-"Shift"s.
		# & (iv) Print out the Results of the Multiple-"Shift"s.

	my @ofnames1 = ('init_features_gblocks.txt', 'orgs_bds_gb_coords.txt', 'init_set_columns.txt', 'excluded_bin_hrels.txt'); # ADDED on 2019/07/18.
	my @ofnames2 = ('inter_block_relations.txt', 'gblock_sizes.txt', 'gblks_w_spec_bds.txt', 'cts_vincluding_vincluded_onx_gblks.txt'); # ADDED on 2019/07/18.
	my @ofnames3 = ('ln_prb_plus_msa_at_origin.txt', 'coord_frame.txt', 'ct_null_clms.txt', 'degeneracy.txt', 'diff_minct_indels.txt', 'diff_ln_prb_msa.txt', 'diff_ln_prb_sbst.txt', 'diff_ln_prb_indel.txt'); # ADDED on 2019/07/18.

	my $new_orgs_gb_coords = $new_curr_bl_coords;
	my @new_list_gblks = @new_list_gblks_in_wd;
	##my @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}];


    if ($TEST_LEVEL_MAIN>2) {
	print STDOUT "\n";
	print STDOUT 'Manually Set: @{$new_orgs_gb_coords} (=@{$new_curr_bl_coords}) = (', join (', ', @{$new_orgs_gb_coords}), ") .\n\n";
	print STDOUT 'Manually Set: @new_list_gblks (= @new_list_gblks_in_wd)', " .\n";
	##print STDOUT 'Manually Set:  @new_list_gblks = @{$list_gblks0_in_wd}[@{$old_indices_gbs}]', " .\n";
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_MAIN>2) {...}"


	    ## $new_curr_set_clms_in_sbwd, $new_curr_bds_gblks_in_sbwd, $new_curr_bl_coords_sbwd, $new_bds_gb_coords_sbwd, $new_gb_coords_sbwd_frm,
	    ## $new_curr_gblks_w_spec_lb_sbwd, $new_curr_gblks_w_spec_rb_sbwd, $rows_affd_by_new_blks_in_sbwd,
	    ## $new_inter_block_relations_sbwd, $new_cts_vincluding_sbwd, $new_cts_vincluded_sbwd, $new_cts_onx_sbwd, $new_gblock_sizes_sbwd)

	my ($fl_end, $if_success, $ct_null_clms0, $degeneracy0, $minct_indels0, $ln_prb_msa0, $ln_prb_sbst0, $ln_prb_indel0, $mdimarray_diffs_ln_prbs_plus)
	  = perform_mshifts_and_print_inputs_outputs (
	    $session, $sbsbdir_out0, $sbsbdir_out_ms, @ofnames1, @ofnames2, @ofnames3,
	    @{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$new_gb_coords_wd_frm}, @{$nodeids_in_msa}, $CHAR_ABS, $CHAR_PRES,
	    @{$new_curr_set_clms_in_wd}, @new_list_gblks, $INDX_BR, $INDX_U_OR_D, @{$new_curr_bds_gblks_in_wd}, @{$rows_affd_by_new_blks_in_wd},
            %clm2ln_prb_sbst, $topnode_id, %node2ch, %node2pa, %br_attr, %{$node2depth}, %{$node2ct_exoffs}, %{$br2eq}, 
	    @root_resfreqs, %{$node2set_sw_extcond_probs0}, %{$br2tr_mtrx}, 
            @set_ingredients_indel_hsts, @SET_INGREDIENTS_PRB_MSA_INDEL, $if_on_le_msa, $if_on_re_msa,
            @list_excluded_bin_hrels,
	    @{$new_inter_block_relations}, @{$new_gblock_sizes}, @{$new_curr_gblks_w_spec_lb}, @{$new_curr_gblks_w_spec_rb},
	    @{$new_cts_vincluding}, @{$new_cts_vincluded}, @{$new_cts_onx}, $STRING_UNDEF, $beg_wd, $end_wd
		);
	unless ($fl_end) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'perform_mshifts_and_print_inputs_outputs' FAILED for $session on the input MSA of the subject window, after the $i th (merged) \"purge\"-candidate region had undergone \"reverse-purge\" (w/ \$rpc_size = $rpc_size) !\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'perform_mshifts_and_print_inputs_outputs', Ended for $session with the subject window, after the $i th (merged) \"purge\"-candidate region had undergone \"reverse-purge\" (w/ \$rpc_size = $rpc_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);



			## Compute the Maximum & Total out of ln_prb_msa's, and record them. ##
			## (ADDED in REVISION on 2020/04/22.) (MAIN III-1) ##

	## $ln_prb_msa_init = $ln_prb_msa0;

# print STDOUT '($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) = ', "($INDX_DIFF_LN_PRB_MSA, $INDX_DGN) .\n\n"; # Added in REVISION on 2020/04/20. (DC1) #

	my ($fl_end1, $max_diff_ln_prb_msa)
	     = find_maxvalue_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF);
	unless ($fl_end1) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'find_maxvalue_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of the terminal arrays in the ", '@{$mdimarray_diffs_ln_prbs_plus}', ", for $session w/ the $i th \"purge\"-candidate \"reverse-purge\"d (w/ \$rpc_size = $rpc_size) : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'find_maxvalue_in_multidim_storage', Ended for $session w/ the $i th \"purge\"-candidate \"reverse-purge\"d (w/ \$rpc_size = $rpc_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my ($fl_end2, $total_diff_ln_prb_msa, $total_wgts)
	    = cmpt_log_total_qnt_in_multidim_storage (@{$new_orgs_gb_coords}, @{$new_bds_gb_coords}, @{$mdimarray_diffs_ln_prbs_plus}, $INDX_DIFF_LN_PRB_MSA, $INDX_DGN, $STRING_UNDEF, $max_diff_ln_prb_msa);
	unless ($fl_end2) {
	    print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_log_total_qnt_in_multidim_storage', FAILED on the $INDX_DIFF_LN_PRB_MSA th element of each terminal array in the ", '@{$mdimarray_diffs_ln_prbs_plus}', " w/ 'Degree of Degeneracy' given as the $INDX_DGN th element of each same terminal array, for $session w/ the $i th \"purge\"-candidate \"reverse-purge\"d (w/ \$rpc_size = $rpc_size) : $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
   (print STDOUT " ... Subroutine, 'cmpt_log_total_qnt_in_multidim_storage', Ended for $session w/ the $i th \"purge\"-candidate \"reverse-purge\"d (w/ \$rpc_size = $rpc_size) ...\n\n") if ($TEST_LEVEL_MAIN >0);


	my $rel_ln_prb_msa0 = (defined $ln_prb_msa0) ? $ln_prb_msa0 - $ln_prb_msa_init : $STRING_UNDEF ;
	my $rel_max_ln_prb_msa = (defined $max_diff_ln_prb_msa) ? $max_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;
	my $rel_total_ln_prb_msa = (defined $total_diff_ln_prb_msa) ? $total_diff_ln_prb_msa + $rel_ln_prb_msa0 : $STRING_UNDEF ;

	(my $session_out = $session) =~ s/\ /_/g;

	print $fh_osyn join ("\t", 'PCC', $session_out, $sbsbdir_out, $ln_prb_msa_init, $rel_ln_prb_msa0, $rel_max_ln_prb_msa, $total_wgts, $rel_total_ln_prb_msa), "\n"; 

# print $fh_osyn join ("\t", 'Window', 'Session_ID', 'Top_of_Output_File_System', 'ln(prob_msa_init)', 'ln(prob_msa_origin)-ln(prob_msa_init)', 'ln(Max[prob_msa])-ln(prob_msa_init)', 'Total[weight]', 'ln(Total[prob_msa])-ln(prob_msa_init)'), "\n\n"; 

			## End of "(ADDED in REVISION on 2020/04/22.) (MAIN III-1)" ##


   } # End of the Middle foreach-loop (over $rpc_size). #

} # END of "for my $i (0) { # Main OUTER for-loop (over the (merged) \"purge\"-candidates). # ADDED on 2019/09/18. # ...}"

		# END of "ADDED on 2019/09/16 & 18." #

	# End of "(MAIN III-1) Actually Perform "Reverse-Purge" (aka. "Ex-Nihilo") + Multiple-"Shift"s." #####

	##### END of "(MAIN III): Explore the Neighborhoods of the Input Alignment via "Reverse-Purge" (aka. "Ex-Nihilo") + Multiple-"Shift"s." #####

			### END of "ADDED in creation of 'anex_rev_purge_for_sgl_wd.ver0.5.pl' (on 2019/07/07-)." ###


=For the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. (MAIN I)

	##### (MAIN I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}. #####

(print STDOUT "\n<<<<< MAIN (I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}. >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);


	# (MAIN I-1) Perform Pure Multiple-"Shift"s. #

if (1) { # (MAIN I-1) #

   (print STDOUT "\n<<< MAIN I-1: Perform Pure Multiple-\"Shift\"s, via 'nonrec_mlt_shifts_gapblocks'. >>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

	my @list_excluded_bin_hrels0 = (); # NO binary horizontal-relations should be excluded.

   (print STDOUT 'Manually Set: @list_excluded_bin_hrels0 = ()', " .\n\n") if ($TEST_LEVEL_MAIN>1);

			# (i) Print out specific Input Data. #

	# ... (omitted the concrete process) ... #

			# (ii) Perform the Multiple-"Shift"s. #

	# ... (omitted the concrete process) ... #

			# (iii) Print out the Output Data. #

	# ... (omitted the concrete process) ... #

} # End of "if (1) { # (MAIN I-1) # ...}"

	# End of "(MAIN I-1) Perform Pure Multiple-"Shift"s." #


	# (MAIN I-2) Perform "Purge" + Multiple-"Shift"s. #
	# (See SM-5 (ii) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

	# End of "(MAIN I-2) Perform "Purge" + Multiple-"Shift"s." #

	# ... (omitted the other processes) ... #

	# (MAIN-I-37b) Perform "Horizontal Merge + (incomplete) Vertical-Split" (into complementary-sibling blocks) + Multiple-"Shift"s. #
	# (See SM-6 (iii-vii-b) of "suppl5_blueprint1_ANEX.draft9.odt" as a reference.)

	# End of "(MAIN-I-37b) Perform "Horizontal Merge + (incomplete) Vertical-Split" (into complementary-sibling blocks) + Multiple-"Shift"s." #

	##### END of "(MAIN I): Explore the Neighborhoods of the Input Alignment via Moves NOT Increasing #{gap-blocks}." #####

=cut # END of "For the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. (MAIN I)"


=For the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. (MAIN II)

	##### (MAIN II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}. #####

(print STDOUT "\n<<<<< MAIN (II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}. >>>>>\n\n\n") if ($TEST_LEVEL_MAIN>0);

for (my $sw=0; $sw <$ct_sbwds; $sw++) { # OUTER for-loop for MAIN (II) (over the relevant sub-windows). #

    my $sbwd_ID = $sbwd_IDs->[$sw];

  (print STDOUT "\n{{{ Sub-window: '$sbwd_ID' }}}\n\n") if ($TEST_LEVEL_MAIN>0);

	### ... (omitted the bulk of the OUTER for-loop) ... ###

  (print STDOUT " === ALL Processes Ended for the sub-window, '$sbwd_ID' ===\n\n") if ($TEST_LEVEL_MAIN>0);

} # END of the "OUTER for-loop for MAIN (II) (over the relevant sub-windows)." #

	##### END of "(MAIN II): Explore the Neighborhoods of the Input Alignment via Moves INCREASING #{gap-blocks}." #####

=cut # END of "For the predecessor, 'anex_for_sgl_wd.ver0.5.pl'. (MAIN II)"

(print STDOUT "\n ===== ALL THE PROCESSED ENDED !!! =====\n\n") if ($TEST_LEVEL_MAIN>0);



	####### DEFINITIONS of the SUBROUTINES #######

	### NOTE: The Subroutines that had been here were MOVED to "MyANEX_IO.pm" in MAJOR REVISION (on 2019/07/07). ###



