package MyANEX_Supple;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyANEX_Supple.pm:
#	This package is a collection of subroutines
#	that supplement the main processes of ANEX.
#	[CREATION STARTED on 2019/02/09, ENDED on 2019/mm/dd.]
#
#	[REVISED on 2019/05/19]:
#	The subroutine 'mk_orgs_bds_gb_coords' was MODIFIED to output
#	the gap-block coordinates corresponding to the window frame as well.
#
#	[MAJOR REVISION on 2019/06/27]:
#	MOVED the subroutine 'associate_nodeids_w_msa_rows' to the package, "MyTreeMap_sppl.pm",
#	for A BETTER DEPENDENCY STRUCTURE.
#
# MyDetect_cmplx_error_cands.pm:
#	This package is a collection of subroutines
#	necessary for detecting the candidates of regions containing complex errors.
#	[CREATION STARTED on December 10th (Sat), 2016, ENDED on Mmm dd (www), yyyy.]
#

use strict;
use warnings;

use MyArrayManipu_num qw(
	diff_nmarray
	union_nmarray
	binsearch_num
	); # Added on 2019/02/16.
	# Added 'binsearch_num' on 2019/07/05.

use MyArrayManipu qw(
	multiple_sort
	); # Added on 2019/07/05.

	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (1) #

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTaxonomy qw(
	fetch_common_ancestors
	);
	# if_ancestor fetch_ancestry calculate_abs_depth
	# fetch_ext_offsprings_nr fetch_all_offsprings_nr

use MyLikelihoods_respattern_msa_dna qw( );
	# calc_node2ntfreqs_td
	# calc_prob_descresstats_cond_ancstat2
	# calc_joint_prob_ancresstats_and_ancstat2
	# calc_clm_probs


=Set aside for potential use (I)

use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw( binomial_cum_via_betai );

=cut # END of "Set aside for potential use (I)" #

	# END of "IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (1)" #


use MyTreeMap_indels_spt_odr2 qw(
	cmp_gblocks
	); # ADDED on 2019/02/13. #

	# ADDED in creation of 'mk_orgs_bds_gb_coords'. #
use MyGapPatternTopology qw(
	ct_necessary_shifts
	);
	# END of "ADDED in creation of 'mk_orgs_bds_gb_coords'." #

	# ADDED in MAJOR REVISION (on 2019/06/27). #
use MyTreeMap_sppl qw(
	associate_nodeids_w_msa_rows
	);
	# END of "ADDED in MAJOR REVISION (on 2019/06/27)." #


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	mk_set_bounds_gpd_segs
	define_window_frames
	define_sliding_windows
	prepare_sliding_windows
	extract_br_info_gblocks
	mk_init_set_columns0
	mk_orgs_bds_gb_coords
	mk_pcc_sliding_windows

	check_consistency_bw_msa_and_clmset
	convert_clmset2msa
	);
	## associate_nodeids_w_msa_rows # EXCLUDED in MAJOR REVISION (on 2019/06/27). #
	# Added 'mk_pcc_sliding_windows' on 2019/07/05.

sub mk_set_bounds_gpd_segs (\@);
sub define_window_frames ($$\@$$\@\@$);
sub define_sliding_windows (\@$$\@$\@$$);
sub extract_br_info_gblocks ($\@\%); # ADDED on 2019/02/23.
## sub associate_nodeids_w_msa_rows (\@\%); # ADDED on 2019/02/26. # EXCLUDED in MAJOR REVISION (on 2019/06/27). #
sub mk_init_set_columns0 (\@$$$\@\@\@);
sub mk_orgs_bds_gb_coords (\@$$\@$); # ADDED on 2019/04/04.

sub check_consistency_bw_msa_and_clmset (\@\@$$); # ADDED on  2019/09/27. #

sub convert_clmset2msa (\@\@$$); # ADDED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/05).

	# IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2) #

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;


	# END of "IMPORTED from 'MyDetect_cmplx_error_cands.pm' (on 2019/02/09). (2)" #



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.



# mk_set_bounds_gpd_segs:
#	This subroutine creates a set of boundaries of gapped segments.
#	[The function was IMPORTED from 'test_detect_cmplx_error_cand_acls0.pl' on 2019/02/13.]
#
#	NOTE: The output will be used as an input, e.g., of 'artf_cluster_gapped_segs(_len)' (in "MyDetect_cmplx_error_cands.pm"), and also of 'define_window_frames' and 'define_sliding_windows' (both in "MyANEX_Supple.pm", i.e., this subroutine).
#
# INPUT: (\@set_cls_gpblocks), where
#	@{$set_cls_gpblocks[$gs]->[$gpb]} = ($indx_gp, $len_gpb, $start_gpb, $end_gpb) is for the $gpb th gap-pattern block in the $gs th gapped segment in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#		and $start_gpb/$end_gpb being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#	(\@set_cls_gpblocks is an output of, e.g., 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm".);

#
# OUTPUT: (\@set_bds_gpd_segs), where
#	@{$set_bds_gpd_segs[$gs]} = ($beg_gs, $end_gs) is for the $gs th gapped-segment,
#	where $beg_gs/$end_gs is the beginning/ending coordinate of the gapped segment (in the same convention as for $start_gpb/$end_gpb in @set_cls_gpblocks in the input).
#
sub mk_set_bounds_gpd_segs (\@) {

	my ($set_cls_gpblocks) = @_;

	my $ct_gpd_segs = @{$set_cls_gpblocks};

	my @set_bds_gpd_segs = ();

	for (my $gs = 0; $gs < $ct_gpd_segs; $gs++) { # Main for-loop (over gapped segments).

	    my $cls_gpblocks = $set_cls_gpblocks->[$gs];
	    my $ct_gpbs = @{$cls_gpblocks};
	    my $beg_gs = $cls_gpblocks->[0][2];
	    my $end_gs = $cls_gpblocks->[$ct_gpbs -1][3];

	    $set_bds_gpd_segs[$gs] = [$beg_gs, $end_gs];

	}	# End of the main for-loop (over gapped segments).

	return \@set_bds_gpd_segs;

}	# End of the subroutine 'mk_set_bounds_gpd_segs'. #


#	my $gpd_segs_in_msa = mk_set_bounds_gpd_segs (@{$set_cls_gpblocks});



# define_window_frames:
#	This subroutine defines the frames, i.e., the sets of columns in an input alignment,
#	within which the window analysis should be performed,
#	using the results of inferences on the regions suffering from complex errors.
#	[CREATED on 2019/02/09; MODIFIED on 2019/02/10.]
#
#	[REVISED in DEBUGGING 'anex.verxxx.pl' on 2020/05/24]:
#	So far, this subroutine defined the prohibited zones REGARDLESS of whether there are some INCOMPLETELY INCLUDED gapped segments or not;
#	Now, when a prohibited zone candidate INCOMPLETELY INCLUDES a gapped segment,
#	the subroutine EXTENDS the zone so that it will COMPLETELY INCLUDE the segment (+ a flanking column).
#
# INPUTS: ($beg_in_msa, $end_in_msa, \@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@artf_cls, \@cmplx_cands, $wdth_margin), where
#
#	$beg_in_msa and $end_in_msa are the coordinates (or the indices of the alignment columns) at the beginning and the end of (the subject region of) the input MSA, respectively (both are included);
#
#	($beg_gs, $end_gs) = @{$gpd_segs_in_msa[$gs]}[$indx_beg_gs, $indx_end_gs] are the coordinates (or the indices of the alignment columns) at the beginning and the end of the $gs th gapped segment, respectively (both are included);
#	(\@set_gpd_segs_in_msa is an input of , e.g., 'artf_cluster_gapped_segs' or 'artf_cluster_gapped_segs_len' in "MyDetect_cmplx_error_cands.pm");
#
#	@{$artf_cls[$acl]} stores the indices of the gapped segments that constitute the $acl th artificial cluster ;
#	(\@artf_cls is an output of, e.g., 'artf_cluster_gapped_segs' or 'artf_cluster_gapped_segs_len' in "MyDetect_cmplx_error_cands.pm");
#
#	@{$cmplx_cands[$cc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $cc th artificial cluster likely to be involved in complex errors, where
#	$indx_ac is the index in @artf_cluster, $totct_indels = $totct_ins + $totct_dels, and  $diffct_indels = $totct_ins - $totct_dels in the artificial cluster;
#	(\@cmplx_cands is an output, e.g., of 'detect_cmplx_error_cand_acls0' in "MyDetect_cmplx_error_cands.pm");
#
#	$wdth_margin defines the "margin" for the "prohibited-zones", that is,
#	[ the sliding windows MUST BE at least ($wdth_margin +1) columns apart from any artificial clusters inferred to be likely involved in complex errors; ] ... OBSOLETE as of 2019/02/15.
#	the sliding windows MUST BE at least $wdth_margin columns apart from any artificial clusters inferred to be likely involved in complex errors;
#
# [ OUTPUTS: \@window_frames, where ] ... OBSOLETE as of 2019/02/10.
# OUTPUTS: (\@window_frames, \@prohibited_zones), where
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.)
#
#	@{$prohibited_zones[$pz]} = ($start_pz, $end_pz)
#	defines the $pz th prohibited zone,
#	that is, the zone starts at the $start_pz th column and ends at the $end_pz th column.
#	(Both are included in the zone.) (ADDED on 2019/02/10.)
#
sub define_window_frames ($$\@$$\@\@$) {

	my ($beg_in_msa, $end_in_msa, $gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, $artf_cls, $cmplx_cands, $wdth_margin) = @_;

	my $ct_ccands = @{$cmplx_cands};

	my $ct_gsegs = @{$gpd_segs_in_msa}; # ADDED in REVISION on 2020/05/24. #

	##my $dbl_wdth_margin = $wdth_margin * 2; # Disabled on 2019/02/15.

		# First, compute the prohibited zones. #

	my @prohibited_zones = ();

	for (my $cc = 0; $cc < $ct_ccands; $cc++) { # 1st main for-loop (over complex candidate regions). #

	    my $indx_ac = $cmplx_cands->[$cc][0];
	    my $acl = $artf_cls->[$indx_ac];
	    my ($lm_gs, $rm_gs) = @{$acl}[0, $#{$acl}];
	    my $beg_acl = $gpd_segs_in_msa->[$lm_gs][$indx_beg_gs];
	    my $end_acl = $gpd_segs_in_msa->[$rm_gs][$indx_end_gs];

		# "Extend" the artificial cluster, by taking account of margins. #
	    my $beg_ex_acl = $beg_acl - $wdth_margin;
	    my $end_ex_acl = $end_acl + $wdth_margin;


		# Examine whether there is an INCOMPLETELY INCLUDED gapped segment on each side of the zone;
		# if so, extend the zone so that it will COMPLETELY INCLUDE the segment.
		# (ADDED in DEBUGGING on 2020/05/24.)

	    if ($lm_gs > 0) {
		my $gs_ii_left;
		for (my $gs = $lm_gs-1; $gs >= 0; $gs--) {
		    my ($beg_gs, $end_gs) = @{$gpd_segs_in_msa->[$gs]}[$indx_beg_gs, $indx_end_gs];
		    if ($end_gs < $beg_ex_acl) { last ; }
		    if ($beg_gs <= $beg_ex_acl) { $gs_ii_left = $gs; }
		}

		if (defined $gs_ii_left) { $beg_ex_acl = $gpd_segs_in_msa->[$gs_ii_left][$indx_beg_gs] -1; }

	    } # End of "if ($lm_gs > 0) {...}"

	    if ($rm_gs < $ct_gsegs-1) {
		my $gs_ii_right;
		for (my $gs = $rm_gs+1; $gs < $ct_gsegs; $gs++) {
		    my ($beg_gs, $end_gs) = @{$gpd_segs_in_msa->[$gs]}[$indx_beg_gs, $indx_end_gs];
		    if ($end_ex_acl < $beg_gs) { last; }
		    if ($end_ex_acl <= $end_gs) { $gs_ii_right = $gs; }
		}

		if (defined $gs_ii_right) { $end_ex_acl = $gpd_segs_in_msa->[$gs_ii_right][$indx_end_gs] + 1; }

	    } # End of "if ($rm_gs < $ct_gsegs-1) {...}"

		# END of "(ADDED in DEBUGGING on 2020/05/24.)"


	    if ($end_in_msa < $beg_ex_acl) { last; }
	    if ($end_ex_acl < $beg_in_msa) { next; }

	    if ($beg_ex_acl < $beg_in_msa) { $beg_ex_acl = $beg_in_msa; }
	    if ($end_in_msa < $end_ex_acl) { $end_ex_acl = $end_in_msa; }

	    if (@prohibited_zones == 0) {
		push @prohibited_zones, [$beg_ex_acl, $end_ex_acl];
		next;
	    }

	    my $last_pz = $prohibited_zones[$#prohibited_zones];

	    my $end_last_pz = $last_pz->[1];

	    if ($end_last_pz + 1 < $beg_ex_acl) {
		push @prohibited_zones, [$beg_ex_acl, $end_ex_acl]; # Define a new prohibited zone. #
	    } else {

		$last_pz->[1] = $end_ex_acl; # Extend the last prohibited zone. #
	    }

	}	# End of the 1st main for-loop (over complex candidate regions). #


		# Second, compute the window frames. #

	my @window_frames = ();

	#if (@prohibited_zones == 0) {
	#	# In this case, the entire subject region should constitute a single frame. #
	#    push @window_frames,[$beg_in_msa, $end_in_msa];
	#    return (\@window_frames, \@prohibited_zones);
	#} # (DISABLED as of 2019/02/10.)

	my $beg_fr = $beg_in_msa;

	foreach my $pz (@prohibited_zones) {

	    my ($beg_pz, $end_pz) = @{$pz};

	    my $end_fr = $beg_pz-1;
	    if ($beg_fr <= $end_fr) {
		push @window_frames, [$beg_fr, $end_fr];
	    }

	    $beg_fr = $end_pz+1;
	}

	if ($beg_fr <= $end_in_msa) {
	    push @window_frames, [$beg_fr, $end_in_msa];
	}

		# MODIFIED on 2019/02/10. #
	return (\@window_frames, \@prohibited_zones) ;
	# return \@window_frames ;
		# END of "MODIFIED on 2019/02/10." #

}	# END of the subroutine 'define_window_frames'. #


# define_sliding_windows:
#	This subroutine defines "sliding windows" that will be used for the main analyses.
#	[ For the detailed rules to define the windows, refer to "simultaneous_moves_of_multiple_blocks_METH.draft9.pdf".] ... OBSOLETE as of 2019/02/17.
#	For the detailed rules to define the windows, refer to "simultaneous_moves_of_multiple_blocks_METH.draft10.pdf".
#	Each window is defined by the columns at its left- and right-ends,
#	as well as the gapped segments it contains.
#	[Creation STARTED on 2019/02/10, ENDED on 2019/02/11; MODIFIED on 2019/02/17.]
#
# INPUTS: (\@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@set_lists_gblocks, $ct_exnodes, \@window_frames, $hf_wdth_shift, $num_max), where

#	($beg_gs, $end_gs) = @{$gpd_segs_in_msa[$gs]}[$indx_beg_gs, $indx_end_gs] are the coordinates (or the indices of the alignment columns) at the beginning and the end of the $gs th gapped segment, respectively (both are included);
#	(\@set_gpd_segs_in_msa is an input of , e.g., 'artf_cluster_gapped_segs' or 'artf_cluster_gapped_segs_len' in "MyDetect_cmplx_error_cands.pm");
#
#	@{$set_lists_gblocks[$gs]} is @list_gblocks, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' or 'br_list_gblocks', both in "MyTreeMap_indels_spt_odr2.pm", for the $gs th gapped segment;
#	Here,
#	@{$list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block resides on the upper/lower side of the branch,
#	$lm/rm_gpb is the index of the left-most/right-most gap-pattern block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block;
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#
#	$ct_exnodes is the number of all external nodes (i.e., sequences in the alignment);
#
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.)
#	(\@window_frames is an output of, e.g., 'define_window_frames' in "MyANEX_Supple.pm" (i.e., this package).);
#
#	$hf_wdth_shift (corresponding to "W_{M}" in the aforementioned reference) is the "width" of the "shift"s of each gap-block to one direction,
#	that is, each gap-block can "shift" up to (and including) $hf_wdth_shift columns in each direction (i.e., to the left or to the right)(, if there is no obstacle);
#
#	$num_max (corresponding to "N_{M}" in the aforementioned reference) the maximum number of gap-blocks each window is allowed to hold.
#
# OUTPUT: (\@set_windows), where
#
#	@{$set_windows[$wd]} = ($beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $wd th sliding window, where
#	the window starts at the $beg_wd th column and ends at the $end_wd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to.
#
sub define_sliding_windows (\@$$\@$\@$$) {

# INPUTS: (\@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@set_lists_gblocks, $ct_exnodes, \@window_frames, $hf_wdth_shift, $num_max), where

	my ($gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, $set_lists_gblocks, $ct_exnodes, $window_frames, $hf_wdth_shift, $num_max) = @_;

	my $ct_gpd_sgs = @{$gpd_segs_in_msa};
	unless (@{$set_lists_gblocks} == $ct_gpd_sgs) {
	    my $ct_lsts = @{$set_lists_gblocks};
	    print STDERR "[ERROR 'define_sliding_windows'] The number of lists of gap-blocks (=$ct_lsts) CLASHES with the number of gapped segments (=$ct_gpd_sgs): $! !\n\n";
	    return (undef);
	}

	my @set_windows = ();

	my $ct_frms = @{$window_frames};
	my $curr_gs = 0;
	for (my $fr = 0; $fr < $ct_frms; $fr++) { # Main outer for-loop (over window-frames).

	    my ($beg_fr, $end_fr) = @{$window_frames->[$fr]};
	    my @ingr_cand_gss = (); # Gapped segments that are candidates of the ingredients of the windows within this frame. #

	    while ($curr_gs < $ct_gpd_sgs) { # 1st middle while-loop (over the remaining gapped segments).

		my ($beg_gs, $end_gs) = @{$gpd_segs_in_msa->[$curr_gs]}[$indx_beg_gs, $indx_end_gs];

		if ($end_fr < $end_gs) { last; } # In this case, the gapped segment will be set aside for the next window-frame.
		if ($beg_fr <= $beg_gs) { push @ingr_cand_gss, $curr_gs; } # In this case, the gapped segment is completely within the window-frame. #

		$curr_gs++;
	    } # End of the 1st middle while-loop (over the remaining gapped segments). #

	    #unless (@ingr_cand_gss>0) { # Found NO candidates for ingredients.
	    #	if ($ct_gpd_sgs <= $curr_gs) {
	    #	    last; # The gapped segments RAN OUT.
	    #	} else {
	    #	    next;
	    #	}
	    #}

	    my @new_windows = (); # element = [$beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs]
	    ## my @exnodes_aff_by_prev_rm_gbs = (); # Lists the external nodes affected by any right-most gap-block in the previous window. # OBSOLETE as of 2019/02/17.

	    my $end_prev_gpd_seg = $beg_fr -2; # MODIFIED on 2019/02/17.
	    #my $end_prev_gpd_seg = $beg_fr -1;

	    my @records_if_v_ovlp = (); # This stores $if_v_ovlp between contiguous gapped segments within the current window. (Added on 2019/02/11) #
	    while (@ingr_cand_gss>0) {  # 2nd middle while-loop (over gapped segments that are candidates of the ingredients).

		my $gs = shift @ingr_cand_gss;
		my ($beg_gs, $end_gs) = @{$gpd_segs_in_msa->[$gs]}[$indx_beg_gs, $indx_end_gs];

		my $list_gbs = $set_lists_gblocks->[$gs];
		my $ct_gbs = @{$list_gbs};

		my ($min_lm_gpb, $max_rm_gpb, $aff_exnodes0) = @{$list_gbs->[0]}[2,3,5];
		      # @{$list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.

=OBSOLETE_1a # as of 2019/02/17
		my @exnodes_aff_by_lm_gbs = ($aff_exnodes0);
		my @exnodes_aff_by_rm_gbs = ($aff_exnodes0);

		for (my $k=1; $k < $ct_gbs; $k++) { # 1st inner for-loop (over gap-blocks). #
		    my ($lm_gpb, $rm_gpb, $aff_exnodes) = @{$list_gbs->[$k]}[2,3,5];
		    if ($lm_gpb < $min_lm_gpb) {
			$min_lm_gpb = $lm_gpb;
			@exnodes_aff_by_lm_gbs = ($aff_exnodes);
		    } elsif (!($lm_gpb > $min_lm_gpb)) {
			push @exnodes_aff_by_lm_gbs, $aff_exnodes;
		    }
		    if ($max_rm_gpb < $rm_gpb) {
			$max_rm_gpb = $rm_gpb;
			@exnodes_aff_by_rm_gbs = ($aff_exnodes);
		    } elsif (!($max_rm_gpb > $rm_gpb)) {
			push @exnodes_aff_by_rm_gbs, $aff_exnodes;
		    }
		} # End of the 1st inner for-loop (over gap-blocks). #

			# (Added on 2019/02/11.) (1)
		my $exnodes_aff_by_rm_gbs = shift @exnodes_aff_by_rm_gbs;
		while (@exnodes_aff_by_rm_gbs>0) { # 2nd inner while-loop. #
		    my $additional = shift @exnodes_aff_by_rm_gbs;
		    $exnodes_aff_by_rm_gbs = union_nmarray (@{$exnodes_aff_by_rm_gbs}, @{$additional});
		} # End of the 2nd inner while-loop. 

			# END of "(Added on 2019/02/11.) (1)"
=cut # END of "OBSOLETE_1a # as of 2019/02/17"

		if ($ct_gbs > $num_max) { # The current gapped segment contains too many gap-blocks.

		    # @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.
		    $end_prev_gpd_seg = $end_gs;

		    if (scalar (@new_windows) >0) { # (Added on 2019/02/11)
			my $prev_window = $new_windows[$#new_windows];
			if ($beg_gs -2 < $prev_window->[1]) { $prev_window->[1] = $beg_gs -2; }
		    }
		    @records_if_v_ovlp = (-1); # (Added on 2019/02/11.)
		    next;
		}

		if (@new_windows == 0) {
			# A new window MUST be created.

		    my $beg_wd = $beg_gs - $hf_wdth_shift;

			# MODIFIED on 2019/02/17. #
		    if ($beg_wd < $end_prev_gpd_seg + 2) { $beg_wd = $end_prev_gpd_seg +2; }
		    #if ($beg_wd < $beg_fr) { $beg_wd = $beg_fr; }
		    #if ((scalar (@records_if_v_ovlp) == 1) and ($records_if_v_ovlp[0] == -1)) {
		    #	if ($beg_wd < $end_prev_gpd_seg + 2) { $beg_wd = $end_prev_gpd_seg +2; }
		    #}

		    my $end_wd = $end_gs + $hf_wdth_shift;
		    if ($end_fr < $end_wd) { $end_wd = $end_fr; }
		    push @new_windows, [$beg_wd, $end_wd, [$gs], $ct_gbs, $fr];

		    #@exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.

		    $end_prev_gpd_seg = $end_gs;
		    @records_if_v_ovlp = (); # (Added on 2019/02/11.)
		    next;
		}

		if ((scalar (@records_if_v_ovlp) == 1) and ($records_if_v_ovlp[0] == -1)) {
			# In this case, too, A new window MUST be created. (Added on 2019/02/11) (2).
		    my $beg_wd = $beg_gs - $hf_wdth_shift;
		    if ($beg_wd < $end_prev_gpd_seg +2) { $beg_wd = $end_prev_gpd_seg +2; }
		    my $end_wd = $end_gs + $hf_wdth_shift;
		    if ($end_fr < $end_wd) { $end_wd = $end_fr; }
		    push @new_windows, [$beg_wd, $end_wd, [$gs], $ct_gbs, $fr];

		    # @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.
		    $end_prev_gpd_seg = $end_gs;
		    @records_if_v_ovlp = (); # (Added on 2019/02/11.)
		    next;

		}

		my $ub_beg_gs = $end_prev_gpd_seg + 1 + 2 * $hf_wdth_shift;

		if ($ub_beg_gs < $beg_gs) {
			# In this case, the previous window should NEVER overlap the current extended gapped segment. #
			# Thus, a new window MUST be created.

		    my $end_wd = $end_gs + $hf_wdth_shift;
		    if ($end_fr < $end_wd) { $end_wd = $end_fr; }
		    push @new_windows, [$beg_gs - $hf_wdth_shift, $end_wd, [$gs], $ct_gbs, $fr];

		    # @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.

		    $end_prev_gpd_seg = $end_gs;
		    @records_if_v_ovlp = (); # (Added on 2019/02/11.)
		    next;

		} 

			# Restarted on 2019/02/11. #

=OBSOLETE_1b # as of 2019/02/17
		my $exnodes_aff_by_lm_gbs = shift @exnodes_aff_by_lm_gbs;
		while (@exnodes_aff_by_lm_gbs>0) { # 3rd inner while-loop.
		    my $additional = shift @exnodes_aff_by_lm_gbs;
		    $exnodes_aff_by_lm_gbs = union_nmarray (@{$exnodes_aff_by_lm_gbs}, @{$additional});
		}	# End of the 3rd inner while-loop.

		my $if_v_ovlp = 0;
		if ($ct_exnodes <= scalar (@{$exnodes_aff_by_lm_gbs}) + scalar (@exnodes_aff_by_prev_rm_gbs) ) {
		    $if_v_ovlp = 1;
		} else {
		    my ($common, $ex1, $ex2) = diff_nmarray (@{$exnodes_aff_by_lm_gbs}, @exnodes_aff_by_prev_rm_gbs);
		    if (@{$common}>0) { $if_v_ovlp = 1; }
		}
=cut # END of "OBSOLETE_1b # as of 2019/02/17"

=OBSOLETE_2 # as of 2019/02/17
		# This case will NEVER occur, because $if_v_ovlp should always be REGARDED as 1 (or -1).
		# (NOTE: Under the new rule (effective on 2019/02/17), at least a column MUST ALWAYS separate a window and its neighboring gap-block(s)).

		if (($beg_gs == $ub_beg_gs) and !($if_v_ovlp)) {

			# In this case, either, the previous window is not regarded as overlapping the current extended gapped segment. #
			# Thus, a new window MUST be created.

		    my $end_wd = $end_gs + $hf_wdth_shift;
		    if ($end_fr < $end_wd) { $end_wd = $end_fr; }
		    push @new_windows, [$beg_gs - $hf_wdth_shift, $end_wd, [$gs], $ct_gbs, $fr];
		    @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment.
		    $end_prev_gpd_seg = $end_gs;
		    @records_if_v_ovlp = ();
		    next;
		}
=cut # END of "OBSOLETE_2 # as of 2019/02/17"

			# The previous window is regarded as overlapping the current extended gapped segment. #

		my $if_v_ovlp = 1; # ADDED on 2019/02/17.

		my $end_wd = $end_gs + $hf_wdth_shift;
		if ($end_fr < $end_wd) { $end_wd = $end_fr; }

		my $prev_window = $new_windows[$#new_windows];
		my ($beg_prev_wd, $end_prev_wd, $invlvd_gs_prev, $ct_gbs_prev) = @{$prev_window};

		if ($ct_gbs + $ct_gbs_prev <= $num_max) {
			# Just add the current gapped segment to the previous window. #

		    $prev_window->[1] = $end_wd;
		    push @{$prev_window->[2]}, $gs;
		    $prev_window->[3] += $ct_gbs;

		    # @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.

		    $end_prev_gpd_seg = $end_gs;
		    push @records_if_v_ovlp, $if_v_ovlp;
		    next;
		}

			# Finalize the right-end of the previous window. #

		if ($beg_gs -1 - $if_v_ovlp < $end_prev_wd) {
		    $prev_window->[1] = $beg_gs -1 - $if_v_ovlp;
		}

			# Create a new window.

		my $end_wd_new = $end_wd;
		my @invlvd_gs_new = ($gs);
		my $ct_gbs_new = $ct_gbs;
		my @cp_invlvd_gs_prev = @{$invlvd_gs_prev};
		my @src_if_v_ovlp = (@records_if_v_ovlp, $if_v_ovlp);
		@records_if_v_ovlp = ();

		while (@cp_invlvd_gs_prev >0) {	 # 5th inner while-loop.

		    my $gs2 = pop @cp_invlvd_gs_prev;
		    my $if_v_ovlp2 =  pop @src_if_v_ovlp;

		    my $list_gbs2 = $set_lists_gblocks->[$gs2];
		    my $ct_gbs2 = @{$list_gbs2};

		    if ($ct_gbs_new + $ct_gbs2 > $num_max) { 
			push @cp_invlvd_gs_prev, $gs2;
			push @src_if_v_ovlp, $if_v_ovlp2;
			last;
		    }

		    $ct_gbs_new += $ct_gbs2;
		    unshift @invlvd_gs_new, $gs2;
		    unshift @records_if_v_ovlp, $if_v_ovlp2;


		}   # End of the 5th inner while-loop.

		my $gs2 = $invlvd_gs_new[0];
		my ($beg_gs2, $end_gs2) = @{$gpd_segs_in_msa->[$gs2]}[$indx_beg_gs, $indx_end_gs];

		my $if_v_ovlp2 = pop @src_if_v_ovlp;
		my $prev_gs = pop @cp_invlvd_gs_prev;
		$end_prev_gpd_seg = $gpd_segs_in_msa->[$prev_gs][$indx_end_gs];

		my $beg_wd_new = $beg_gs2 - $hf_wdth_shift;
		if ($beg_wd_new < $end_prev_gpd_seg +1 + $if_v_ovlp2) {
		    $beg_wd_new = $end_prev_gpd_seg + 1 + $if_v_ovlp2;
		}

		push @new_windows, [$beg_wd_new, $end_wd_new, \@invlvd_gs_new, $ct_gbs_new, $fr];

		# @exnodes_aff_by_prev_rm_gbs = @{$exnodes_aff_by_rm_gbs}; # Reset the set of external nodes affected by previous right-most gapped segment. # OBSOLETE as of 2019/02/17.

		$end_prev_gpd_seg = $end_gs;


	    } # End of the 2nd middle while-loop (over gapped segments that are candidates of the ingredients).

	    if (@new_windows>0) { push @set_windows, @new_windows; }

	    if ($ct_gpd_sgs <= $curr_gs) { last; } # The gapped segments RAN OUT.
	
	}	# End of the main outer for-loop (over window-frames).


	return \@set_windows;

}	# END of the subroutine 'define_sliding_windows'. #


# prepare_sliding_windows:
#	This subroutine is essentially a wrapper of the subroutine, 'define_sliding_windows'.
#	It prepares the ingredients for defining sliding windows,
#	actually defines them,
#	and it also prepares a sorted list of gap-blocks in each sliding window.
#	The subroutine does its jobs by combining two major subroutines,
#	i.e., 'define_window_frames' and 'define_sliding_windows', in this package.
#	[CREATION STARTED on 2019/02/11, ENDED on 2019/02/13.]
#	
# [ INPUTS: ($beg_in_msa, $end_in_msa, \@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@artf_cls, \@cmplx_cands, \@set_lists_gblocks, $ct_exnodes, $wdth_margin, $hf_wdth_shift, $num_max), where ] ... OBSOLETE as of 2019/02/13.
# INPUTS: ($beg_in_msa, $end_in_msa, \@set_cls_gpblocks, \@artf_cls, \@cmplx_cands, \@set_lists_gblocks, $ct_exnodes, $wdth_margin, $hf_wdth_shift, $num_max), where
#
#	$beg_in_msa and $end_in_msa are the coordinates (or the indices of the alignment columns) at the beginning and the end of (the subject region of) the input MSA, respectively (both are included);
#
#	[ ($beg_gs, $end_gs) = @{$gpd_segs_in_msa[$gs]}[$indx_beg_gs, $indx_end_gs] are the coordinates (or the indices of the alignment columns) at the beginning and the end of the $gs th gapped segment, respectively (both are included);
#	(\@set_gpd_segs_in_msa is an input of , e.g., 'artf_cluster_gapped_segs' or 'artf_cluster_gapped_segs_len' in "MyDetect_cmplx_error_cands.pm"); ] ... OBSOLETE as of 2019/02/13.
#
#	@{$set_cls_gpblocks[$gs]->[$gpb]} = ($indx_gp, $len_gpb, $start_gpb, $end_gpb) is for the $gpb th gap-pattern block in the $gs th gapped segment in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#		and $start_gpb/$end_gpb being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#	(\@set_cls_gpblocks is an output of, e.g., 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm".);

#	@{$artf_cls[$acl]} stores the indices of the gapped segments that constitute the $acl th artificial cluster ;
#	(\@artf_cls is an output of, e.g., 'artf_cluster_gapped_segs' or 'artf_cluster_gapped_segs_len' in "MyDetect_cmplx_error_cands.pm");
#
#	@{$cmplx_cands[$cc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $cc th artificial cluster likely to be involved in complex errors, where
#	$indx_ac is the index in @artf_cluster, $totct_indels = $totct_ins + $totct_dels, and  $diffct_indels = $totct_ins - $totct_dels in the artificial cluster;
#	(\@cmplx_cands is an output, e.g., of 'detect_cmplx_error_cand_acls0' in "MyDetect_cmplx_error_cands.pm");
#
#	@{$set_lists_gblocks[$gs]} is @list_gblocks, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' or 'br_list_gblocks', both in "MyTreeMap_indels_spt_odr2.pm", for the $gs th gapped segment;
#	Here,
#	@{$list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block resides on the upper/lower side of the branch,
#	$lm/rm_gpb is the index of the left-most/right-most gap-pattern block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block;
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#
#	$ct_exnodes is the number of all external nodes (i.e., sequences in the alignment);
#
#	$wdth_margin defines the "margin" for the "prohibited-zones", that is,
#	[ the sliding windows MUST BE at least ($wdth_margin +1) columns apart from any artificial clusters inferred to be likely involved in complex errors; ] ... OBSOLETE as of 2019/02/15.
#	the sliding windows MUST BE at least $wdth_margin columns apart from any artificial clusters inferred to be likely involved in complex errors;
#
#	$hf_wdth_shift (corresponding to "W_{M}" in the aforementioned reference) is the "width" of the "shift"s of each gap-block to one direction,
#	that is, each gap-block can "shift" up to (and including) $hf_wdth_shift columns in each direction (i.e., to the left or to the right)(, if there is no obstacle);
#
#	$num_max (corresponding to "N_{M}" in the aforementioned reference) the maximum number of gap-blocks each window is allowed to hold.
#
# OUTPUTS: (\@window_frames, \@prohibited_zones, \@set_windows, \@set_lists_gblocks_in_wds), where
#
#	@{$window_frames[$fr]} = ($start_fr, $end_fr)
#	defines the $fr th frame for the window analysis,
#	that is, the frame starts at the $start_fr th column and ends at the $end_fr th column.
#	(Both are included in the frame.)
#
#	@{$prohibited_zones[$pz]} = ($start_pz, $end_pz)
#	defines the $pz th prohibited zone,
#	that is, the zone starts at the $start_pz th column and ends at the $end_pz th column.
#	(Both are included in the zone.) (ADDED on 2019/02/10.)
#
#	@{$set_windows[$wd]} = ($beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $wd th sliding window, where
#	the window starts at the $beg_wd th column and ends at the $end_wd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to;
#
#	@{$set_lists_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block resides on the upper/lower side of the branch,
#	$lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks (within the $gs th gapped segment) involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment accommodating the subject gap-block,
#	$gb_in_gs is the index of the subject gap-block within the $gs th gapped segment.
#
#[ sub prepare_sliding_windows ($$\@$$\@\@\@$$$$) {
#	my ($beg_in_msa, $end_in_msa, $gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, $artf_cls, $cmplx_cands, $set_lists_gblocks, $ct_exnodes, $wdth_margin, $hf_wdth_shift, $num_max) = @_; ] ... OBSOLETE as of 2019/02/13.
#
		# MODIFIED on 2019/02/13. #
sub prepare_sliding_windows ($$\@\@\@\@$$$$) {

	my ($beg_in_msa, $end_in_msa, $set_cls_gpblocks, $artf_cls, $cmplx_cands, $set_lists_gblocks, $ct_exnodes, $wdth_margin, $hf_wdth_shift, $num_max) = @_;

	my $ct_gpd_segs = @{$set_cls_gpblocks};
	unless (@{$set_lists_gblocks} == $ct_gpd_segs) {
	    my $ct_lists = @{$set_lists_gblocks};
	    print STDERR "[ERROR 'prepare_sliding_windows'] The number of lists of gap-blocks (=$ct_lists) CLASHES with the number of gapped segments (=$ct_gpd_segs): $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	# @{$gpd_segs_in_msa->[$gs]} = ($beg_gs, $end_gs) for the $gs th gapped segment.
	#
	my $gpd_segs_in_msa = mk_set_bounds_gpd_segs (@{$set_cls_gpblocks});

	unless (defined $gpd_segs_in_msa) {
	     print STDERR "[ERROR 'prepare_sliding_windows'] The subroutine 'mk_set_bounds_gpd_segs' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my ($indx_beg_gs, $indx_end_gs) = (0, 1);

		# END of "MODIFIED on 2019/02/13." #


	# ($beg_in_msa, $end_in_msa, $set_cls_gpblocks, $artf_cls, $cmplx_cands, $set_lists_gblocks, $ct_exnodes, $wdth_margin, $hf_wdth_shift, $num_max) = @_;

	my ($window_frames, $prohibited_zones) 
	  =  define_window_frames ($beg_in_msa, $end_in_msa, @{$gpd_segs_in_msa}, $indx_beg_gs, $indx_end_gs, @{$artf_cls}, @{$cmplx_cands}, $wdth_margin) ;
	unless (defined $window_frames) {
	     print STDERR "[ERROR 'prepare_sliding_windows'] The subroutine 'define_window_frames' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my $set_windows
	    = define_sliding_windows (@{$gpd_segs_in_msa}, $indx_beg_gs, $indx_end_gs, @{$set_lists_gblocks}, $ct_exnodes, @{$window_frames}, $hf_wdth_shift, $num_max) ;
	unless (defined $set_windows) {
	     print STDERR "[ERROR 'prepare_sliding_windows'] The subroutine 'define_sliding_windows' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my $ct_wds = @{$set_windows};

	my @set_lists_gblocks_in_wds = ();

	for (my $wd = 0; $wd < $ct_wds; $wd++) { # Main outer for-loop (over windows). #

	    my $invlvd_gpd_sgs = $set_windows->[$wd][2];

		# @{$set_windows[$wd]} = ($beg_wd, $end_wd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $wd th sliding window, where
		# the window ... contains gapped segments listed in @invlvd_gpd_sgs,...

	    my @gblocks_in_wd = ();

	    foreach my $gs (@{$invlvd_gpd_sgs}) { # Middle foreach-loop (over gapped segments in the window). #

		my $cls_gpblocks = $set_cls_gpblocks->[$gs];

		# @{$set_cls_gpblocks[$gs]->[$gpb]} = ($indx_gp, $len_gpb, $start_gpb, $end_gpb) is for the $gpb th gap-pattern block in the $gs th gapped segment in the input MSA, with
		# 	... $start_gpb/$end_gpb being the starting/ending coordinate of the block (in the 0-based full-closed convention);

		my $list_gblocks = $set_lists_gblocks->[$gs];

		my $ct_gbs_in_gs = @{$list_gblocks};
		for (my $gb_in_gs = 0; $gb_in_gs < $ct_gbs_in_gs; $gb_in_gs++) { # Inner for-loop (over gap-blocks). #

		    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $aff_exnodes, $e) = @{$list_gblocks->[$gb_in_gs]};
	 		# @{$set_lists_gblocks[$gs]} is @list_gblocks,...
	 		# Here,
	 		# @{$list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
	 		# ... $lm/rm_gpb is the index of the left-most/right-most gap-pattern block,...

		    my $lm_clm = $cls_gpblocks->[$lm_gpb][2];
		    my $rm_clm = $cls_gpblocks->[$rm_gpb][3];

		    push @gblocks_in_wd, [$br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs];

			# @{$set_list_gblocks_in_wds[$wd]->[$gb_in_wd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_wd th gap-block in the $wd th window.
			# ... $lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,...
			# ... $gs is the index of the gapped segment accommodating the subject gap-block,
			# $gb_in_gs is the index of the subject gap-block within the $gs th gapped segment.

		}	# End of the inner for-loop (over gap-blocks). #

	    }	# End of the middle foreach-loop (over gapped segments in the window). #

	    my @std_gblocks_in_wd = sort { cmp_gblocks (@{$a}, @{$b})  } @gblocks_in_wd; 


	    $set_lists_gblocks_in_wds[$wd] = \@std_gblocks_in_wd ;

	}	# End of the main outer for-loop (over windows). #

	return ($window_frames, $prohibited_zones, $set_windows, \@set_lists_gblocks_in_wds) ;

}	# END of the subroutine 'prepare_sliding_windows'. #


# extract_br_info_gblocks:
#	This subroutine receives @{$set_lists_gblocks_in_wds[$wd]}, which is an element of an output of 'prepare_sliding_windows' above,
#	and extracts some brief pieces of information on the gap-blocks.
#	[CREATED on 2019/02/23.]
#
# INPUTS: ($left_end_wd, \@list_gblocks_in_wd, \%nodeid2row_indx),  Where
#
#	$left_end_wd is the index (or column number) of the left-end of the window (included in it);
#
#	@list_gblocks_in_wd = @{$set_lists_gblocks_in_wds[$wd]}, where
#	\@set_lists_gblocks_in_wds is an output of 'prepare_sliding_windows'. 
#	More precisely, @{$list_gblocks_in_wd[$gb]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb th gap-block in the $wd th window.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block resides on the upper/lower side of the branch,
#	$lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks (within the $gs th gapped segment) involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment accommodating the subject gap-block,
#	$gb_in_gs is the index of the subject gap-block within the $gs th gapped segment.
#
#	%nodeid2row_indx = ($node_id => $indx_row, ...) for the "external" nodes, where
#	$node_id is the ID of an "external" node,
#	$indx_row is the index specifying a row of the input MSA.
#
# OUTPUTS: (\@bds_gblocks, \@rows_aff_by_blks), where
#
#	@{$bds_gblocks[$gb]} = ($lbd_blk, $rbd_blk) gives the positions (within the window, 0-based) of the boundaries (included) of the $gb th gap-block;
#	@{$rows_aff_by_blks[$gb]} lists the indices of the rows of the MSA affected by the $gb th gap-block.
#
sub extract_br_info_gblocks ($\@\%) {

	my ($left_end_wd, $list_gblocks_in_wd, $nodeid2row_indx) = @_;

	my $ct_gblocks = @{$list_gblocks_in_wd};

	my @bds_gblocks = my @rows_aff_by_blks = ();

	for (my $gb = 0; $gb < $ct_gblocks; $gb++) { # Main for-loop (over gap-blocks).

	    my ($br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $affected_exnodes, $e, $gs, $gb_in_gs) = @{$list_gblocks_in_wd->[$gb]};

	    $bds_gblocks[$gb] = [$lm_clm - $left_end_wd, $rm_clm - $left_end_wd];
	    my @affd_rows = ();
	    foreach my $node (@{$affected_exnodes}) {
		my $row_indx = $nodeid2row_indx->{$node};
		unless (defined $row_indx) {
		    print STDERR "[ERROR 'extract_br_info_gblocks'] The \"external\" node, $node, does NOT CORRESPOND to any row in the MSA: $! !!\n\n";
		    return (undef, undef);
		}
		push @affd_rows, $row_indx;
	    }

	    my @std_affd_rows = sort {$a <=> $b} @affd_rows;

	    $rows_aff_by_blks[$gb] = \@std_affd_rows;

	} # End of the main for-loop (over gap-blocks).

	return (\@bds_gblocks, \@rows_aff_by_blks);

}	# END of the subroutine 'extract_br_info_gblocks'. #


=MOVED to "MyTreeMap_sppl.pm" (on 2019/06/27). #

# associate_nodeids_w_msa_rows:
#	This subroutine associates the IDs of the (usually external) nodes (of a tree)
#	with the rows of the MSA (associated with the tree).
#	[CREATED on 2019/02/26.]
#
# INPUTS: (@seqnames, \%name2id), where
#
#	$seqnames[$k] is the name (or ID) of the sequence at the $k th row of the MSA;
#	%name2id = ($node_name  => $node_ID, ...) for nodes with names,
#	where $node_name corresponds to $seqnames[$k] for external nodes;
#
# OUTPUTS: (\@nodeids_in_msa, \%nodeid2msa_row), where
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $now_ID.
#
sub associate_nodeids_w_msa_rows (\@\%) {

	my ($seqnames, $name2id) = @_;

	my $ct_rows = @{$seqnames};

	my @nodeids_in_msa = ();
	my %nodeid2msa_row;

	for (my $row = 0; $row < $ct_rows; $row++) { # Main for-loop (over MSA rows).

	    my $name = $seqnames->[$row];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'associate_nodeids_w_msa_rows'] The $row th row (w/ name = '$name') does NOT correspond to ANY node in the tree: $! !!\n\n";
		return (undef, undef);
	    }

	    $nodeids_in_msa[$row] = $node_id;
	    $nodeid2msa_row{$node_id} = $row;

	} # End of the main for-loop (over MSA rows).


	return (\@nodeids_in_msa, \%nodeid2msa_row);

}	# End of the subroutine 'associate_nodeids_w_msa_rows'. #

=cut # END of "MOVED to "MyTreeMap_sppl.pm" (on 2019/06/27)." #


# mk_init_set_columns0:
#	This subroutine creates an INITIAL VALUE of @set_columns0,
#	which is an important input of the subroutine, 'nonrec_mlt_shifts(_noninf)_gapblocks', in "MyANEX_Main.pm".
#	[CREATED on 2019/02/27.]
#
#	[REVISED in DEBUGGING on 2020/05/10]:
#	Before this revision, a BUG caused the output (i.e., @init_set_columns0)
#	NOT to begin with the 0th element when $beg_wd > 0.
#	Now, this BUG has been fixed.
#
# INPUTS: (\@exnodes_in_msa, $char_abs, $beg_wd, $end_wd, \@invlvd_gpd_sgs, \@set_cls_gpbs, \@init_psm_hst_for_cls), where
#
#	$exnodeids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#	[ [Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' above in this package.] ] ... OBSOLETE as of 2019/06/27.
#	[Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyTreeMap_sppl.pm"]; # Revised on 2019/06/27.
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#
#	$beg_wd and $end_wd specify the columns at which the window starts and ends, respectively (both are included in the window);
#	@invlvd_gpd_sgs lists the indices of the gapped segments involved in the window;
#	[NOTE: $beg_wd, $end_wd, and @invlvd_gpd_sgs are parts of an output of the subroutine, 'prepare_sliding_windows', above in this package];
#
#
#	@{$set_cls_gpbs[$cl]->[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gpb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention);
#	
#	@{$init_psm_hst_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	[NOTE: \@set_cls_gpbs and \@init_psm_hst_for_cls are important parts of the outputs of, e.g., 'cal_init_psm_cands_for_sgl_msa' and 'bch_br_parsimonize_dollo_indel_histories', both in "MyTreeMap_indels_spt_odr2.pm", and 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' in "MyTreeMap_indels_spt_odr_finer.pm".]
#
# OUTPUTS: (\@set_columns0), where
#
#	@column = @{$set_columns0[$c]} specifies the $c th column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment that accommodates the corresponding site of the $k th "sequence".
#
sub mk_init_set_columns0 (\@$$$\@\@\@) {

	my ($exnodes_in_msa, $char_abs, $beg_wd, $end_wd, $invlvd_gpd_sgs, $set_cls_gpbs, $init_psm_hst_for_cls) = @_;

# INPUTS: (\@exnodes_in_msa, $char_abs, $beg_wd, $end_wd, \@invlvd_gpd_sgs, \@set_cls_gpbs, \@init_psm_hst_for_cls), where

	my $ct_rows = @{$exnodes_in_msa};

		# Initialize the output. #
	my @set_columns0 = ();

	for (my $c = $beg_wd; $c <= $end_wd; $c++) { # Preliminary for-loop (over columns in the window). #
	    my @clm = ();
	    for (my $r = 0; $r < $ct_rows; $r++) { $clm[$r] = $c; }

			# Fixed a Serious BUG (on 2020/05/10)!! #
	    ##$set_columns0[$c] = \@clm;
	    $set_columns0[$c - $beg_wd] = \@clm;
			# End of "Fixed a Serious BUG (on 2020/05/10)!!" #

	} # End of the preliminary for-loop (over columns in the window). #

=This was a WRONG solution!!
			# Fixed a Serious BUG (on 2020/05/10)!! #
	##for (my $c = $beg_wd; $c <= $end_wd; $c++) { # Preliminary for-loop (over columns in the window). #
	for (my $c = 0; $c <= $end_wd - $beg_wd; $c++) { # Preliminary for-loop (over columns in the window). #
			# End of "Fixed a Serious BUG (on 2020/05/10)!!" #
	    my @clm = ();
	    for (my $r = 0; $r < $ct_rows; $r++) { $clm[$r] = $c; }

	    $set_columns0[$c] = \@clm;

	} # End of the preliminary for-loop (over columns in the window). #
=cut # End of "This was a WRONG solution!!"

		# Main Process. #

	foreach my $cl (@{$invlvd_gpd_sgs}) { # Main outer foreach-loop (over gapped segments involved). #

	    my $gpbs_in_cl = $set_cls_gpbs->[$cl];
	    my $ct_gpbs = @{$gpbs_in_cl};
	    my ($bl2indels, $node2blwise_gstats) = @{$init_psm_hst_for_cls->[$cl]};

	    for (my $r = 0; $r < $ct_rows; $r++) { # Middle for-loop (over MSA rows).
		my $node = $exnodes_in_msa->[$r];
		my $blwise_gstats = $node2blwise_gstats->{$node};

		for (my $gpb = 0; $gpb < $ct_gpbs; $gpb++) { # Inner for-loop (over gap-pattern blocks in the segment).
		    unless ($blwise_gstats->[$gpb] eq $char_abs) { next; } # Skip if the block is in the "presence" state.

		    my ($indx_gp, $len_gpb, $beg_gpb, $end_gpb) = @{$gpbs_in_cl->[$gpb]};

			# Shift the "coordinate origin" from the beginning of the MSA to the beginning of the window.
		    $beg_gpb -= $beg_wd;

		    $end_gpb -= $beg_wd;
		    ##$end_gpb -= $end_wd; # A BUG fixed on 2019/02/28. #

			# @{$set_cls_gpbs[$cl]->[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with

		    for (my $c = $beg_gpb; $c <= $end_gpb; $c++) { # Innermost for-loop (over columns in $gpb).
		    ##for (my $c = $beg_gpb; $c < $end_gpb; $c++) { # Innermost for-loop (over columns in $gpb). # A BUG fixed on 2019/02/28.

			$set_columns0[$c]->[$r] = $char_abs;

		    } # End of the innermost for-loop (over columns in $gpb).

		} # End of the inner for-loop (over gap-pattern blocks in the segment). #

	    } # End of the middle for-loop (over MSA rows).

	} # End of the main outer foreach-loop (over gapped segments involved). #


	return \@set_columns0;

}	# End of the subroutine 'mk_init_set_columns0'. #


# mk_orgs_bds_gb_coords:
#	This subroutine creates the origins and the boundaries
#	for the coordinates of the gap-blocks.
#	[CREATED on 2019/04/04; CORRECTED on 2019/04/05.]
#
#	[MAJOR REVISION on 2019/05/19]:
#	Now, the subroutine outputs a 3rd array,
#	which stores gap-block coordinates corresponding to the findow frame.
#
# INPUTS: (\@bds_gblks0, $le_wd, $re_wd, \@inter_block_relations, $maxct_1way_shifts), where
#	($lbd0, $rbd0) = @{$bds_gblks0[$gb]} give the initial values of the left- and right-boundaries of (and included in) the $gb th gap-block;
#	[Typically, \@bds_gblocks0 is an output of 'extract_br_info_gblocks' above;
#	in this case, $le_wd MUST BE 0 (zero)];
#
#	$le_wd and $re_wd give the left-end and the right-end of the window we are dealing with;
#
#	$inter_block_relations[$gb1]->[$gb2] gives the relation of the $gb2 th gap-block against the $gb1 th gap-block;
#	[Typically, \@inter_block_relations is an output of 'mk_inter_block_relations' in "MyGapPatternTopology.pm"];
#
#	$maxct_1way_shifts specifies the upper-bound of the number of shifts in 1 way, i.e., either to the left or to the right;
#
# [ OUTPUTS: (\@orgs_gb_coords, \@bds_gb_coords), where ] ... OBSOLETE as of 2019/05/19.
# OUTPUTS: (\@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm), where
#
#	$org_coord = $orgs_gb_coords[$gb] gives the origin of the coordinate for the $gb th gap-block;
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords[$gb]} give the (accessible) left- and right-boundaries of the coordinate for the $gb th gap-block.
#
#	NOTE: 0 <= $lbd_coord <= $org_coord <= $rbd_coord MUST always hold.
#	      (Here, we choose the convention of 0 == $lbd_coord.)
#
#		# ADDED in MAJOR REVISION (on 2019/05/19). #
#	($gb_coord_le_wd, $gb_coord_re_wd) = @{$gb_coords_wd_frm[$gb]} give the coordinates for the $gb th gap-block corresponding to the left-end and right-end of the window.
#	NOTE 2: $gb_coord_le_wd <= $lbd_coord <= $rbd_coord <= $gb_coord_re_wd MUST always hold.
#		(However, $gb_coord_le_wd may be negative.) 
#		# End of "ADDED in MAJOR REVISION (on 2019/05/19)." #
#
sub mk_orgs_bds_gb_coords (\@$$\@$) {

	my ($bds_gblks0, $le_wd, $re_wd, $inter_block_relations, $maxct_1way_shifts) = @_;

	my $ct_gblks = @{$bds_gblks0};

	my @orgs_gb_coords = my @bds_gb_coords = ();
	my @gb_coords_wd_frm = (); # ADDED in MAJOR REVISION on 2019/05/19.

	my %skipped;

	for (my $gb = 0; $gb < $ct_gblks; $gb++) { # Main for-loop (over gap-blocks).

	    my ($lbd0, $rbd0) = @{$bds_gblks0->[$gb]};

	    my $ct_shifts_left = ct_necessary_shifts ($lbd0, $le_wd, $gb, @{$bds_gblks0}, @{$inter_block_relations}, %skipped);
	    unless (defined $ct_shifts_left) {
		print STDERR "[ERROR 'mk_orgs_bds_gb_coords'] Subroutine 'ct_necessary_shifts' FAILED on the left-side of the $gb th gap-block: $! !\n\n";
		return (undef, undef);
	    }

	    my $ct_shifts_right = ct_necessary_shifts ($rbd0, $re_wd, $gb, @{$bds_gblks0}, @{$inter_block_relations}, %skipped);
	    unless (defined $ct_shifts_left) {
		print STDERR "[ERROR 'mk_orgs_bds_gb_coords'] Subroutine 'ct_necessary_shifts' FAILED on the right-side of the $gb th gap-block: $! !\n\n";
		return (undef, undef);
	    }

		# MODIFIED in MAJOR REVISION on 2019/05/19. (1) #
	    my $dist_lbd = ($ct_shifts_left < $maxct_1way_shifts) ? $ct_shifts_left : $maxct_1way_shifts ;
	    my $dist_rbd = ($ct_shifts_right < $maxct_1way_shifts) ? $ct_shifts_right : $maxct_1way_shifts ;
	    $orgs_gb_coords[$gb] = $dist_lbd;
	    $bds_gb_coords[$gb] = [0, $dist_rbd + $dist_lbd];
	    $gb_coords_wd_frm[$gb] = [-$ct_shifts_left + $dist_lbd, $ct_shifts_right + $dist_lbd];
	    ##if ($ct_shifts_left > $maxct_1way_shifts) { $ct_shifts_left = $maxct_1way_shifts; }
	    ##if ($ct_shifts_right > $maxct_1way_shifts) { $ct_shifts_right = $maxct_1way_shifts; }
	    ##$orgs_gb_coords[$gb] = $ct_shifts_left;
	    ##$bds_gb_coords[$gb] = [0, $ct_shifts_left + $ct_shifts_right];
		# End of "MODIFIED in MAJOR REVISION on 2019/05/19. (1)" #

	} # End of the main for-loop (over gap-blocks).

		# Modified in MAJOR REVISION on 2019/05/19. (2) #
	return (\@orgs_gb_coords, \@bds_gb_coords, \@gb_coords_wd_frm);
	##return (\@orgs_gb_coords, \@bds_gb_coords); # A BUG fixed on 2019/04/05.
	####return (\@orgs_gb_coords, @bds_gb_coords);
		# End of "Modified in MAJOR REVISION on 2019/05/19. (2)" #

}	# End of the subroutine 'mk_orgs_bds_gb_coords'. #


# mk_pcc_sliding_windows:
#	This subroutine constructs purge-candidate-containing (PCC) sliding-windows,
#	each of which is associated with, and COMPLETELY contains, a (merged) purge-candidate region,
#	and contains at most ($num_max-2) gap-blocks.
#	Also, each window MUST be within one of the window-frames.
#	[CREATION Started on 2019/07/04, Ended on 2019/07/06.]
#
#	[RULEs to Construct the PCC Sliding Windows]
#	For each (meged) purge-candidate region:
#	(0) If the region is NOT contained in a window-frame completely (i.e., if it overlaps any "prohibited zones"), dismiss the region;
#	(1) Search for gapped segments overlapping the region and/or the “neighboring” segments within $hf_wdth_shift columns from the region;
#	(2) If the total number of gap-blocks in the overlapping segments exceeds ($max_num-2), dismiss the region;
#	(3) Otherwise, compose a "seed" of a PCC window from the region and the overlapping gapped segments;
#	(4) As long as the total number of gap-blocks is < $max_num -2, attempt to expand the window
#	    by incorporating then nearest-neighboring gapped segment into the window;
#	(5a) If the total number is expected to be > $max_num -2, stop expanding the window BEFORE incorporating THE gapped segment,
#	(5b) OR, if the “neighboring” segments run out, stop expanding the window EVEN if the total number is < $max_num -2;
#	(6) Determine the boundaries of the window so that all constituent segments & the purge-candidate region will be contained, but so that no other segments will be contained.
#	(The boundaries MUST always be within $hf_wdth_shift columns from the outermost boundaries among those of the region and the segments.)
#
#	[NOTE: The "overlapping" (the region) above is supposed to include "touching" (the region).]
#
#	[DEBUGGED on 2020/05/24]:
#	So far, the subroutine set WRONGLY the boundaries of the PCC Sliding Window, when some "superfluous" gapped segments
#	are within the un-finalized version of the window.
#	Now, this BUG has been FIXED.
#
# [ INPUTS: (\@list_mgd_pcands, \@window_frames, $indx_beg_fr, $indx_end_fr, \@set_gpd_segs, \@set_lists_gblocks, $hf_wdth_shift, $num_max), where ] ... OBSOLETE as of 2019/07/05.
# INPUTS: (\@list_mgd_pcands, $indx_beg_pc, $indx_end_pc, \@window_frames, $indx_beg_fr, $indx_end_fr, \@set_gpd_segs, \@set_lists_gblocks, $hf_wdth_shift, $num_max), where
#
#	[
#	@{$list_mgd_pcands[$pc]} = ($beg_pc, $end_pc, $br, \@msa_rows_under_br, \@constituent_sites, $pv, $pv0, \@sgn_wds_included) provides information on the $pc th (merged) purge-candidate region, where
#	$beg_pc & $end_pc are the sites at the beginning & end, respectively, of the region,
#	$br is the ID of the branch across which the "purge" is suspected within the region,
#	@msa_rows_under_br stores the indices of the MSA rows corresponding to the external offsprings of $br,
#	@constituent_sites stores the sites that constitute the region,
#	$pv is the P-value measuring the deviation from the specified substutution model,
#	$pv0 is the P-value measuring the deviation from the random-match model,
#	and @sgn_wds_included is the significant windows included in the region;
#	more precisely, 
#	@{$sgn_wds_included[$swd]} = (\@sites_in_wd, $pv_wd, $pv0_wd) provies information on the $swd th significant window, where
#	@sites_in_wd stores the sites in the window,
#	$pv_wd and $pv0_wd are the P-values for the window, measuring the same types of deviations as the aforementioned $pv and $pv0, respectively;
#	] ... OBSOLETE as of 2019/07/05.
#	($beg_pc, $end_pc) = @{$list_mgd_pcands[$pc]}[$indx_beg_pc, $indx_end_pc] is the sites at the beginning & end of the $pc th (merged) purge-candidate region; # MODIFIED on 2019/07/05.
#	[Typically, \@list_mgd_pcands is an output of 'readin_list_mgd_pcands' in "anex.verXXX.pl" or "anex_for_sgl_wd.verXXX.pl"];
#
#	($beg_fr, $end_fr) = @{$window_frames[$fr]}[$indx_beg_fr, $indx_end_fr] is the beginning & end coordinates of the $fr th window-frame;
#	[Typically, \@window_frames is an output of 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	@set_gpd_segs (=@set_cls_gpblocks) stores information on the gapped-segments, i.e., clusters of gap-pattern-blocks.
#	More precisely,
#	($indx_gp, $len, $beg, $end) = @{$set_gpd_segs->[$gs][$gpb]} are information on the $gpb th gap-pattern-block in the $gs th gapped segment, where
#	$indx_gp is the index of the gap-pattern,
#	$len, $beg, $end are the length, beginning coordinate, and end coordinate, respectively,
#	 of the gap-pattern-block;
#	[Typically, \@set_gpd_segs (=\@set_cls_gpblocks) is an important input of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
#	@{$set_lists_gblocks[$gs]->[$gb]} provides information on the $gb th gap-block in the $gs th gapped segment. More precisely, 
#	($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@aff_exnodes, $e) = @{$set_lists_gblocks[$cl]->[$gb], where
#	the gap-block is separated by, and on the $up_or_lw (='U' or 'L') side of, branch $br,
#	$lm_gpb and $rm_gpb are the indices of the left-most and right-most gap-pattern blocks,
#	@invlvd_gpbs lists the gap-pattern-blocks involved;
#	@aff_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#
#	[Typically, @br_list_gblocks = @{$set_lists_gblocks[$gs]} is the output of 'br_list_gblocks' (in "MyTreeMap_indels_spt_odr2.pm"), or 'wrapper_detect_cmplx_error_cand_acls0' in "MyDetect_cmplx_error_cands.pm"];
#
# OUTPUTS: (\@set_pcc_windows, \@set_indices_mpcands, \@set_lists_gblocks_in_pcc_wds), where
#
#	@{$set_pcc_windows[$pwd]} = ($beg_pwd, $end_pwd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $pwd th PCC sliding window, where
#	the window starts at the $beg_pwd th column and ends at the $end_pwd th column,
#	contains gapped segments listed in @invlvd_gpd_sgs,
#	$ct_gbs is the count of gap-blocks in the window,
#	and $fr is the index of the frame the window belongs to;
#
#	$set_indices_mpcands[$pwd] = $indx_mpcand, which is the index of the merged purge-candidate region
#	associated with the $pwd th PCC window;
#
#	@{$set_lists_gblocks_in_pcc_wds[$pwd]->[$gb_in_pwd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_pwd th gap-block in the $pwd th PCC window.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block resides on the upper/lower side of the branch,
#	$lm/rm_clm is the index of the left-most/right-most column affected by the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks (within the $gs th gapped segment) involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}},
#	$gs is the index of the gapped segment accommodating the subject gap-block,
#	$gb_in_gs is the index of the subject gap-block within the $gs th gapped segment.
#
#	NOTE: The PCC windows will be sorted in ascending order of $beg_pwd -> $end_pwd -> $indx_mpcand.
#
# [ sub mk_pcc_sliding_windows (\@\@$$\@\@$$) {
#	my ($list_mgd_pcands, $window_frames, $indx_beg_fr, $indx_end_fr, $set_gpd_segs, $set_lists_gblocks, $hf_wdth_shift, $num_max) = @_; ] ... OBSOLETE as of 2019/07/05.
#
sub mk_pcc_sliding_windows (\@$$\@$$\@\@$$) {

	my ($list_mgd_pcands, $indx_beg_pc, $indx_end_pc, $window_frames, $indx_beg_fr, $indx_end_fr, $set_gpd_segs, $set_lists_gblocks, $hf_wdth_shift, $num_max) = @_;

	my $ct_mpcands = @{$list_mgd_pcands};
	my $ct_frames = @{$window_frames};
	my $ct_gpd_sgs = @{$set_gpd_segs};

	my @indices_all_frs = (0 .. $ct_frames-1);
	my @indices_all_gss = (0 .. $ct_gpd_sgs-1); 

	my @set_bds_gpd_segs = ();
	for (my $gs=0; $gs<$ct_gpd_sgs; $gs++) { # Preliminary for-loop (over the gapped segments). #
	    my $list_gblks_in_gs = $set_gpd_segs->[$gs];
	    my $beg_gs = $list_gblks_in_gs->[0][2];
	    my $end_gs = $list_gblks_in_gs->[$#{$list_gblks_in_gs}][3];

		# ($indx_gp, $len, $beg, $end) = @{$set_gpd_segs->[$gs][$gpb]} are information on the $gpb th gap-pattern-block in the $gs th gapped segment, where

	    $set_bds_gpd_segs[$gs] = [$beg_gs, $end_gs];
	} # End of the preliminary for-loop (over the gapped segments). #


	my @src_set_pcc_windows = (); # element = [$beg_pwd, $end_pwd, $indx_mpcand, \@invlvd_gpd_sgs, $ct_gbs, $fr, \@list_gblocks_in_pcc_wd].

	for (my $pc = 0; $pc <$ct_mpcands; $pc++) { # Main Outer for-loop (over the (merged) purge-candidate regions). #

		# All we need here is the positions of the both boundaries of the subject purge-candidate. #

	    my ($beg_pc, $end_pc) = @{$list_mgd_pcands->[$pc]}[$indx_beg_pc, $indx_end_pc];
	    ## [ my ($beg_pc, $end_pc, $br, $msa_rows_under_br, $constituent_sites, $pv, $pv0, $sgn_wds_included) = @{$list_mgd_pcands->[$pc]}; ] ... OBSOLETE as of 2019/07/05.

		# (1) Examine whether or not the purge-candidate is completely within a window-frame. #

	    my ($fl_exist_fr, $indx2_fr, $indx_fr) = binsearch_num ($beg_pc, @{$window_frames}, $indx_beg_fr, @indices_all_frs);
	    unless (defined $fl_exist_fr) {
		print STDOUT "[ERROR 'mk_pcc_sliding_windows'] The subroutine, 'binsearch_num', FAILED on ", '@{$window_frames}', " against $beg_pc, i.e., the beginning of the $pc th purge-candidate region: $! !\n\n";
		return (undef, undef, undef);
	    }
	    if ($indx_fr < 0) { next; } # The purge-candidate is on the left of ALL window-frames.

	    my ($beg_fr, $end_fr) = @{$window_frames->[$indx_fr]}[$indx_beg_fr, $indx_end_fr];

	    if ($end_fr < $end_pc) { next; } # This purge-candidate is NOT completely within the window-frame.


		# (2) Search for the gapped segments overlapping or neighboring the purge-candidate.

	    my @overlapping_gsegs = ();
	    my @neighboring_gsegs = (); # element = [$indx_gs, $distance, $ct_gblocks, $beg_gs]

	    my ($fl_exist_gs, $indx2_gs, $indx_gs) = binsearch_num ($beg_pc, @set_bds_gpd_segs, 1, @indices_all_gss);
	    unless (defined $fl_exist_gs) {
		print STDOUT "[ERROR 'mk_pcc_sliding_windows'] The subroutine, 'binsearch_num', FAILED on ", '@set_bds_gpd_segs', " against $beg_pc, i.e., the beginning of the $pc th purge-candidate region: $! !\n\n";
		return (undef, undef, undef);
	    }


	    for (my $gs = $indx_gs; $gs >= 0; $gs--) { # 1st inner for-loop (examining gapped segments on the left). #
		my ($beg_gs, $end_gs) = @{$set_bds_gpd_segs[$gs]};

		if ($beg_gs < $beg_fr) { last; } # The gapped-segment is NOT within the window-frame.

		my $distance = $beg_pc - $end_gs -1;
		if ($distance <= 0) {
		    push @overlapping_gsegs, $gs;
		} elsif ($distance <= $hf_wdth_shift) {
			# Modified on 2019/07/06. #
		    my $ct_gblocks = @{$set_lists_gblocks->[$gs]};
		    push @neighboring_gsegs, [$gs, $distance, $ct_gblocks, $beg_gs];
		    ##push @neighboring_gsegs, [$gs, $distance, $beg_gs];
		} else { # if ($distance > $hf_wdth_shift)
		    last;
		}
	    } # End of the 1st inner for-loop.

	    for (my $gs = $indx_gs+1; $gs < $ct_gpd_sgs; $gs++) { # 2nd inner for-loop (examining gapped segments on the right). #
		my ($beg_gs, $end_gs) = @{$set_bds_gpd_segs[$gs]};

		if ($end_fr < $end_gs) { last; } # The gapped-segment is NOT within the window-frame.

		my $distance = $beg_gs - $end_pc -1;
		if ($distance <= 0) {
		    push @overlapping_gsegs, $gs;
		} elsif ($distance <= $hf_wdth_shift) {
			# Modified on 2019/07/06. #
		    my $ct_gblocks = @{$set_lists_gblocks->[$gs]};
		    push @neighboring_gsegs, [$gs, $distance, $ct_gblocks, $beg_gs];
		    ##push @neighboring_gsegs, [$gs, $distance, $beg_gs];
		} else { # if ($distance > $hf_wdth_shift)
		    last;
		}
	    } # End of the 2nd inner for-loop.

		# (3) Examine the total number of gap-blocks in the gapped segments overlapping the purge-candidate.

	    my $totct_gblocks = 0;
	    foreach my $gs (@overlapping_gsegs) { $totct_gblocks += scalar (@{$set_lists_gblocks->[$gs]}) ; }

	    if ($totct_gblocks > $num_max-2) { next; }  # In this case, give up creating a window containing the purge-candidate.

	    my @invlvd_gpd_sgs = @overlapping_gsegs;

		# (5) Attempt to gradually expand the PCC window by incorporating neighboring gap-blocks one by one into the PCC window.

	    my @indices_srt1 = (1, 2, 3); my @methods_srt1 = ('+', '+', '+'); # Modified on 2019/07/06.
	    ##my @indices_srt1 = (1, 2); my @methods_srt1 = ('+', '+');
	    my $std_neighboring_gsegs = (@neighboring_gsegs>1) ? multiple_sort (@neighboring_gsegs, @indices_srt1, @methods_srt1) : \@neighboring_gsegs ;
	    unless (defined $std_neighboring_gsegs) {
		print STDOUT "[ERROR 'mk_pcc_sliding_windows'] The subroutine 'multiple_sort' FAILED on ", '@neighboring_gsegs', " for the $pc th purge-candidate: $! !\n\n";
		return (undef, undef, undef);
	    }

	    foreach my $info_gs (@{$std_neighboring_gsegs}) { # 3rd inner foreach-loop (over neighboring gapped-segments). #
		my ($gs, $distance, $ct_gblocks, $beg_gs) = @{$info_gs};
		if ($totct_gblocks + $ct_gblocks > $num_max -2) { last; }
		$totct_gblocks += $ct_gblocks;
		push @invlvd_gpd_sgs, $gs;
	    } # End of the 3rd inner foreach-loop. #

	    if (@invlvd_gpd_sgs == 0) {

		# (A) In this case, only the purge-candidate region constitutes a PCC window. #

			# (i) Determine the boundaries of the PCC window. #

		my $beg_pwd = ($beg_fr > $beg_pc - $hf_wdth_shift) ? $beg_fr : $beg_pc - $hf_wdth_shift ;
		my $end_pwd = ($end_fr < $end_pc + $hf_wdth_shift) ? $end_fr : $end_pc + $hf_wdth_shift ;

		if ($indx_gs >=0) { # In this case, $indx_gs MUST be the left-flanking gapped-segment.

		    my ($beg_lf_gs, $end_lf_gs) = @{$set_bds_gpd_segs[$indx_gs]};
		    ##if ($beg_pwd < $end_lf_gs + 2) { $beg_pwd = $end_lf_gs; }
		    if ($beg_pwd < $end_lf_gs + 2) { $beg_pwd = $end_lf_gs +2; } # CORRECTED on 2020/05/24.
		}
		   
		# my ($beg_gs, $end_gs) = @{$set_bds_gpd_segs[$gs]};

		if ($indx_gs +1 < $ct_gpd_sgs) { # In this case, ($indx_gs +1) MUST be the right-flanking gapped-segment.
		    my ($beg_rf_gs, $end_rf_gs) = @{$set_bds_gpd_segs[$indx_gs +1]};

		    ##if ($end_rf_gs -2 < $end_pwd) { $end_pwd = $end_rf_gs-2; } 
		    if ($beg_rf_gs -2 < $end_pwd) { $end_pwd = $beg_rf_gs-2; } # REVISED in DEBUGGING on 2020/05/24!!
		}

			# (ii) Create a(n empty) list of gap-blocks in the PCC window. #
		my @list_gblocks_in_pcc_wd = (); # It follows that this PCC window contains NO gap-blocks.

		push @src_set_pcc_windows, [$beg_pwd, $end_pwd, $pc, \@invlvd_gpd_sgs, $totct_gblocks, $indx_fr, \@list_gblocks_in_pcc_wd].

		next;

	    } # End of "if (@invlvd_gpd_sgs == 0) {...}"


		# (B) At least one gapped segment will be contained in the PCC window. #

	    if (@invlvd_gpd_sgs>1) { @invlvd_gpd_sgs = sort {$a <=> $b} @invlvd_gpd_sgs; }

			# (i) Determine the boundaries of the PCC window. #

	    my ($indx_lm_gseg, $indx_rm_gseg) = @invlvd_gpd_sgs[0, $#invlvd_gpd_sgs];
	    my $beg_lm_gseg = $set_bds_gpd_segs[$indx_lm_gseg]->[0];
	    my $end_rm_gseg = $set_bds_gpd_segs[$indx_rm_gseg]->[1];

	    my $lm_bulk = ($beg_pc < $beg_lm_gseg) ? $beg_pc : $beg_lm_gseg ;
	    my $rm_bulk = ($end_pc > $end_rm_gseg) ? $end_pc : $end_rm_gseg ;

	    my $beg_pwd = ($beg_fr > $lm_bulk - $hf_wdth_shift) ? $beg_fr : $lm_bulk - $hf_wdth_shift ;
	    my $end_pwd = ($end_fr < $rm_bulk + $hf_wdth_shift) ? $end_fr : $rm_bulk + $hf_wdth_shift ;

	   if ($indx_lm_gseg >0) { # In this case, there MUST be a left-flanking gapped-segment.

		my ($beg_lf_gs, $end_lf_gs) = @{$set_bds_gpd_segs[$indx_lm_gseg -1]};
		##if ($beg_pwd < $end_lf_gs + 2) { $beg_pwd = $end_lf_gs; }
		if ($beg_pwd < $end_lf_gs + 2) { $beg_pwd = $end_lf_gs +2; } # CORRECTED on 2020/05/24.
	   }

	   if ($indx_rm_gseg +1 < $ct_gpd_sgs) { # In this case, there MUST be a right-flanking gapped-segment.

		my ($beg_rf_gs, $end_rf_gs) = @{$set_bds_gpd_segs[$indx_rm_gseg +1]};

		##if ($end_rf_gs -2 < $end_pwd) { $end_pwd = $end_rf_gs-2; }
		if ($beg_rf_gs -2 < $end_pwd) { $end_pwd = $beg_rf_gs-2; } # REVISED in DEBUGGING on 2020/05/24!!
	   }

			# (ii) Create a list of gap-blocks in the PCC window. #

	   my @gblocks_in_pcc_wd = ();
	   foreach my $gs (@invlvd_gpd_sgs) { # 5th inner foreach-loop (over gapped-segments involved). #

			## IMPORTED from 'prepare_sliding_windows' above in this package (on 2019/07/06). (1) ##

		my $cls_gpblocks = $set_gpd_segs->[$gs];

	# @{$set_cls_gpblocks[$gs]->[$gpb]} = ($indx_gp, $len_gpb, $start_gpb, $end_gpb) is for the $gpb th gap-pattern block in the $gs th gapped segment in the input MSA, with

		my $list_gblocks = $set_lists_gblocks->[$gs];

	# ($list_mgd_pcands, $indx_beg_pc, $indx_end_pc, $window_frames, $indx_beg_fr, $indx_end_fr, $set_gpd_segs, $set_lists_gblocks, $hf_wdth_shift, $num_max) = @_;

		my $ct_gbs_in_gs = @{$list_gblocks};
		for (my $gb_in_gs = 0; $gb_in_gs < $ct_gbs_in_gs; $gb_in_gs++) { # Innermost for-loop (over gap-blocks). #

		    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $aff_exnodes, $e) = @{$list_gblocks->[$gb_in_gs]};
	 		# @{$set_lists_gblocks[$gs]} is @list_gblocks,...
	 		# Here,
	 		# @{$list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
	 		# ... $lm/rm_gpb is the index of the left-most/right-most gap-pattern block,...

		    my $lm_clm = $cls_gpblocks->[$lm_gpb][2];
		    my $rm_clm = $cls_gpblocks->[$rm_gpb][3];

		    push @gblocks_in_pcc_wd, [$br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs]; # MODIFIED in creation of 'mk_pcc_sliding_windows' (on 2019/07/06). #
		    ##push @gblocks_in_wd, [$br, $up_or_lw, $lm_clm, $rm_clm, $invlvd_gpbs, $aff_exnodes, $e, $gs, $gb_in_gs];

		# @{$set_lists_gblocks_in_pcc_wds[$pwd]->[$gb_in_pwd]} = ($br, $up_or_lw, $lm_clm, $rm_clm, \@invlvd_gpbs, \@affected_exnodes, $e, $gs, $gb_in_gs) gives information on the $gb_in_pwd th gap-block in the $pwd th PCC window.

		}  # End of the innermost for-loop (over gap-blocks). #

			## END of "IMPORTED from 'prepare_sliding_windows' above in this package (on 2019/07/06). (1)" ##
	   } # End of the 5th inner foreach-loop. #

	   my @std_gblocks_in_pcc_wd = sort { cmp_gblocks (@{$a}, @{$b})  } @gblocks_in_pcc_wd;  # IMPORTED with modification from 'prepare_sliding_windows' above in this package (on 2019/07/06). (2)

	   push @src_set_pcc_windows, [$beg_pwd, $end_pwd, $pc, \@invlvd_gpd_sgs, $totct_gblocks, $indx_fr, \@std_gblocks_in_pcc_wd];

	#  @src_set_pcc_windows = (); # element = [$beg_pwd, $end_pwd, $indx_mpcand, \@invlvd_gpd_sgs, $ct_gbs, $fr, \@list_gblocks_in_pcc_wd].

	} # End of the Main Outer for-loop (over the (merged) purge-candidate regions). #

	### RESTART FROM HERE (on 2019/07/06). ###


	my @indices_srt = (0, 1, 2); my @methods_srt = ('+', '+', '+');
	my $std_src_set_pcc_windows = (@src_set_pcc_windows>1) ? multiple_sort (@src_set_pcc_windows, @indices_srt, @methods_srt) : \@src_set_pcc_windows ;
	unless (defined $std_src_set_pcc_windows) {
	    print STDOUT "[ERROR 'mk_pcc_sliding_windows'] The subroutine 'multiple_sort' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}

	my @set_pcc_windows = my @set_indices_mpcands = my @set_lists_gblocks_in_pcc_wds = ();
	foreach my $info_pcc_window (@{$std_src_set_pcc_windows}) { # Trailing foreach-loop (over the PCC windows). #

	    my ($beg_pwd, $end_pwd, $indx_mpcand, $invlvd_gpd_sgs, $ct_gbs, $fr, $list_gblocks_in_pcc_wd) = @{$info_pcc_window};
		# @{$set_pcc_windows[$pwd]} = ($beg_pwd, $end_pwd, \@invlvd_gpd_sgs, $ct_gbs, $fr) is for the $pwd th PCC sliding window, where
	    push @set_pcc_windows, [$beg_pwd, $end_pwd, $invlvd_gpd_sgs, $ct_gbs, $fr];
	    push @set_indices_mpcands, $indx_mpcand;
	    push @set_lists_gblocks_in_pcc_wds, $list_gblocks_in_pcc_wd;

	} # End of the trailing foreach-loop (over the PCC windows). #


	return (\@set_pcc_windows, \@set_indices_mpcands, \@set_lists_gblocks_in_pcc_wds);

}	# End of the subroutine, 'mk_pcc_sliding_windows'. #


# check_consistency_bw_msa_and_clmset:
#	This subroutine examines the consistency between
#	an MSA (multiple sequence alignment) and a set of columns.
#	[CREATED on 2019/09/27.]
#
#	NOTE: Currently, this subroutine assumes that the elements of @msa are of an identical length,
#			and that the elements of @set_clms are of an identical size.
#
# INPUTS: (\@msa, \@set_clms, $char_abs_msa, $char_abs_sc), where
#
#	@msa stores an MSA; More specifically,
#	${$msa[$k]} is a string representing the $k th sequence embedded in the MSA;
#
#	@set_clms stores the set of all columns in the MSA,
#	and @column = @{$set_clms[$c]} resresents the $c th column in the MSA;
#	More specifically,
#	$column[$r] = $char_abs_sc if the $r th row of the column is occupied by an "absence" state (i.e., a gap),
#	otherwise,
#	$column[$r] = a number (either a residue number of a site number), which uniquely represent the residue occupying the $r th row;
#
#	$char_abs_msa and $char_abs_sc are the character representing an "absence" state
#	in @msa and @set_clms, respectively;
#
# OUTPUTS: ($if_consistent, \@inconsistent_cells), where
#
#	$if_consistent = 1 if @msa and @set_clms are consistent with each other,
#			 = 0 if they are INconsistent with each other,
#			 = undef if the subroutine failed.
#
#	@inconsistent_cells stores inconsistent cells;
#	More specifically,
#	@{$inconsistent_cells[$ic]} = ($column, $row) specifies the position of the $ic th inconsistent cell;
#
sub check_consistency_bw_msa_and_clmset (\@\@$$) {

	my ($msa, $set_clms, $char_abs_msa, $char_abs_sc) = @_;

	my $ct_rows = @{$msa};
	my $ct_clms = @{$set_clms};

		# (0) Handle special cases (w/ either #{rows}=0 or #{columns}=0) . #
	unless ($ct_rows>0) {
	    if ($ct_clms > 0) {
		print STDOUT "[WARNING 'check_consistency_bw_msa_and_clmset'] #{columns} (=$ct_clms) is NONZERO while #{rows} is ZERO !!\n\n";
		return (0, []);
	     } else {
		return (1, []);
	     }
	}
	unless ($ct_clms>0) {
	    if ($ct_rows>0) {
		print STDOUT "[WARNING 'check_consistency_bw_msa_and_clmset'] #{rows} (=$ct_rows) is NONZERO while #{columns} is ZERO !!\n\n";
		return (0, []);
	    }
	}

		# (1) Handle Normal, but Clearly INconsistent Cases. #

	unless (length (${$msa->[0]}) == $ct_clms) {
	    my $ct_clms2 = length (${$msa->[0]}) ;
	    print STDOUT "[WARNING 'check_consistency_bw_msa_and_clmset'] #{columns} based on ", '@{$set_clms}', " (=$ct_clms) CLASHES with that based on ", '@{$msa}', " (=$ct_clms2) !!\n\n";
	    return (0, []);
	}
	unless (scalar (@{$set_clms->[0]}) == $ct_rows) {
	    my $ct_rows2 = @{$set_clms->[0]};
	    print STDOUT "[WARNING 'check_consistency_bw_msa_and_clmset'] #{rows} based on ", '@{$set_clms}', " (=$ct_rows2) CLASHES with that based on ", '@{$msa}', " (=$ct_rows) !!\n\n";
	    return (0, []);
	}

		# (2) Handle Normal Cases. #

	my @inconsistent_cells = ();
	for (my $r=0; $r <$ct_rows; $r++) { # Main Outer for-loop (over the rows). #

	    my $cp_row_msa = ${$msa->[$r]};

	    for (my $c=0; $c <$ct_clms; $c++) { # Main Inner for-loop (over the columns). #

		my $cell_msa = substr ($cp_row_msa, 0, 1, '');
		my $if_abs_msa = ($cell_msa eq $char_abs_msa) ? 1 : 0;
		my $if_abs_sc = ($set_clms->[$c][$r] eq $char_abs_sc) ? 1 : 0;

		if ($if_abs_msa != $if_abs_sc) { 
		    push @inconsistent_cells, [$c, $r];
		}

	    }	# End of the Main Inner for-loop (over the columns). #

	} # End of the Main Outer for-loop (over the rows). #

	my $if_consistent = (@inconsistent_cells>0) ? 0 : 1 ;

	return ($if_consistent, \@inconsistent_cells);

}	# End of the subroutine, 'check_consistency_bw_msa_and_clmset'. #


# convert_clmset2msa:
#	This subroutine creates an MSA that corresponds to a set of columns,
#	with the guidance of the "original" MSA,
#	from which the set of columns was made.
#	The subroutine was created taking advantage of the top part of
#	'check_consistency_bw_msa_and_clmset'.
#	[CREATED in creation of 'coordinate_point_lcl_msa.ver0.7.pl' (on 2020/06/05).]
#
# check_consistency_bw_msa_and_clmset:
#	This subroutine examines the consistency between
#	an MSA (multiple sequence alignment) and a set of columns.
#	[CREATED on 2019/09/27.]
#
#	NOTE: Currently, this subroutine assumes that the elements of @msa are of an identical length,
#			and that the elements of @set_clms are of an identical size.
#
# INPUTS: (\@msa0, \@set_clms_new, $char_abs_msa, $char_abs_sc), where
#
#	@msa0 stores the "original" MSA; More specifically,
#	${$msa0[$k]} is a string representing the $k th sequence embedded in the MSA;
#
#	@set_clms_new stores the set of columns in the "new" MSA, which is likely modified from the "original".
#	And @column = @{$set_clms_new[$c]} resresents the $c th column in the "new" MSA;
#	More specifically,
#	$column[$r] = $char_abs_sc if the $r th row of the column is occupied by an "absence" state (i.e., a gap),
#	otherwise,
#	$column[$r] = a number that uniquely represent the residue occupying the $r th row;
#	More precisely, $column[$r] is the index of the column IN THE "ORIGINAL" MSA (, i.e., @msa0,)
#	accommodating the residue in question.
#
#	$char_abs_msa and $char_abs_sc are the character representing an "absence" state
#	in @msa and @set_clms, respectively;
#
# OUTPUTS: (\@new_msa), where
#
#	@new_msa is the MSA corresponding to the @set_clms_new;
#	It follows the same format as @msa0.
#
sub convert_clmset2msa (\@\@$$) {

	my ($msa0, $set_clms_new, $char_abs_msa, $char_abs_sc) = @_;

	my $ct_rows = @{$msa0};
	my $ct_clms_new = @{$set_clms_new};

		# (0) Handle special cases (w/ either #{rows}=0 or #{columns}_new=0) . #

	unless ($ct_rows>0) {
	    if ($ct_clms_new > 0) {
		print STDOUT "[ERROR 'convert_clmset2msa'] #{columns}_new (=$ct_clms_new) is NONZERO while #{rows} is ZERO !!\n\n";
		return undef ;
	     } else {
		return [];
	     }
	}

	unless ($ct_clms_new>0) {
	    if ($ct_rows>0) {
		print STDOUT "[WARNING 'convert_clmset2msa'] #{rows} (=$ct_rows) is NONZERO while #{columns} is ZERO !!\n\n";
		my @new_msa = ();
		for (my $r=0; $r< $ct_rows; $r++) {
		    my $seq = '';
		    $new_msa[$r] = \$seq;
		}
		return \@new_msa;
	    }
	}

	my $ct_rows_new = @{$set_clms_new->[0]};
	unless ($ct_rows_new == $ct_rows) {
	    print STDOUT "[ERROR 'convert_clmset2msa'] #{rows}_new (=$ct_rows_new) is NOT EQUAL to #{rows}_0 (= $ct_rows) !!\n\n";
	    return undef;
	}

=DISABLED, because it turned out to be UNnecessary (& even FATAL sometimes), on 2020/06/06.
	my $ct_clms0 = length ${$msa0->[0]};
	unless ($ct_clms0 >= $ct_clms_new) {
	    print STDOUT "[ERROR 'convert_clmset2msa'] #{columns}_new (=$ct_clms_new) is LARGER THAN #{columns}_0 (= $ct_clms0): $! !!\n\n";
	    return undef;
	}
=cut # END of "DISABLED, because it turned out to be UNnecessary (& even FATAL sometimes), on 2020/06/06."


		# (1) Main Process. #

	my @new_msa = ();

	for (my $r=0; $r < $ct_rows; $r++) { # Main outer for-loop (over rows). #

	    my @chars_in_row0 = split //, ${$msa0->[$r]} ;

	    my $new_seq = '';

	    for (my $c=0; $c < $ct_clms_new; $c++) { # Main inner for-loop (over new columns). #

		my $new_sindx = $set_clms_new->[$c][$r];

		if ($new_sindx eq $char_abs_sc) {
		    $new_seq .= $char_abs_msa;
		    next;
		}

		my $new_char = $chars_in_row0[$new_sindx];

		unless (defined $new_char) {
		    print STDOUT "[ERROR 'convert_clmset2msa'] The '$new_sindx', which is at the $r th row and the $c th column of the new column-set, CORRESPONDS to NO CHARACTER in the $r th row of the \"ORIGINAL\" MSA: $! !\n\n";
		    return undef;
		}

		$new_seq .= $new_char;

	    } # End of the main inner for-loop (over new columns). #

	    $new_msa[$r] = \$new_seq;

	} # End of the main outer for-loop (over rows). #


	return \@new_msa ;

}	# End of the subroutine, 'convert_clmset2msa'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


