package MyCluster_distribution_II;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyCluster_distribution_II.pm:
#	This package consists of some subroutines
#	that extend the functions of those in 'MyCluster_distribution.pm'.
#
#	[ CREATED on Dec 25 (Sun), 2016.]
#
# MyCluster_distribution.pm: a package to construct and print distribution tables of clusters,
#				according to the cluster size.
#
#


use strict;
use warnings;

use MyCluster_distribution qw(
	cluster_distributor_wgt
	);


use Exporter;
our @ISA = qw(Exporter);
##our @EXPORT = qw( );

our @EXPORT_OK = qw(
	distributor_2dim_wgt
	);


		### DEFINITIONS of the Subroutines. ###


# distributor_2dim_wgt:
#	This extended version of the subroutine 'distributor_2dim' (in "MyCluster_distribution.pm")
#	builds a weighted 2-dimensional distribution.
#
#	[CREATED on Dec 25 (Sun), 2016.]
#
# distributor_2dim: This subroutine  distributes a pair of quantities
#				into a 2dimensinal histogram.
#
sub distributor_2dim_wgt ($$\@\@\@$) {

	my ($x, $y, $rf_bins, $rf_abscissa, $rf_ordinate, $wgt) = @_;

	my $ct_abs = @{$rf_abscissa};
	unless (@{$rf_bins} == $ct_abs +2) {
		print STDERR "Illegal input for \'distributor_2dimensional_wgt\'.\n";
		return 0;
	}

	for (my $i=0; $i<$ct_abs; $i++) {

		if ($x < $rf_abscissa->[$i]) {
			cluster_distributor_wgt ($y, @{$rf_bins->[$i]}, @{$rf_ordinate}, $wgt);
			cluster_distributor_wgt ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate}, $wgt) ;
			##cluster_distributor($y, @{$rf_bins->[$i]}, @{$rf_ordinate});
			##cluster_distributor ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate} ) ;
			return 1;
		}
	}					# end of the for #

	unless ($x >= $rf_abscissa->[$ct_abs-1] ) {
		print STDERR "Some unsusual thing has happened for \'distributor_2dimensional\'.\n";
		return 0;
	}

	cluster_distributor_wgt ($y, @{$rf_bins->[$ct_abs]}, @{$rf_ordinate}, $wgt);
	cluster_distributor_wgt ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate}, $wgt);
	##cluster_distributor ($y, @{$rf_bins->[$ct_abs]}, @{$rf_ordinate});
	##cluster_distributor ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate});

	return 1;
}			# end of the subroutine 'distributor_2dim_wgt' #


1;			# a "spell" necessary for the package to be "use"-able or "require"-able #


