package MyPosShiftMap;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019? Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyPosShiftMap.pm:
#	This package is a collection of subroutines
#	necessary for processing a Position-Shift Map,
#	which maps "position-shift"s, each of which is the difference in the position of a particular site of a particular sequence between two sequence alignments, onto one of the alignment.
#	(A Position-Shift Map is typically created by the subroutine, 'map_shifts_respos_bw_2msas', in the package, "MyMalnComparison.pm" (in "../MyPerlModules_ComplLiMment/").)
#	[CREATION Started on 2020/03/13, ENDED on 2020/03/19.]
#
#	[NOTE]:
#	Although some subroutines in the package, "MyMalnComparison.pm", also process Position-Shift Maps,
#	the subroutines in this package are based on different strategies.
#	Briefly, the subroutines in "MyMalnComparison.pm" first chop each sequence into "segments" of identical position-shifts,
#	then they attempt to virtically cluster the segments from different (but related) sequences if the segments share the same horizontal boundaries in the alignment.
#	In contrast, the subroutines in this package first virtically cluster the position-shifts in each alignment column,
#	then, they attempt to lump together the effectively contiguous clusters sharing the same position-shifts.
#


use strict;
use warnings;

		# (ADDED on 2020/03/17.) #
use MyArrayManipu_num qw(
	diff_nmarray
	);
	# union_nmarray
	# binsearch_num
		# End of "(ADDED on 2020/03/17.)" #


use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

=Set aside for potential use. (1)

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyLikelihoods_respattern_msa_dna qw(
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw(
	binomial_cum_via_betai
	);

use MyArrayManipu qw(
	merge_integer_intervals
	);

=cut # End of "Set aside for potential use. (1)"

use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeMap_II_xgap qw(
	map_parsim_states_sankoff_xgap_1clm
	locate_parsim_changes_xgap
	);


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	chop_psmap_into_pspblocks
	enum_ps_miniclass_seeds
	enum_psm_clusterings_for_new_psps
	select_pcls_for_psp_blks
	mk_node2parsstat_sep_mcs
	mk_set_miniclasses_shift_respos
	identify_trivial_miniclasses
	list_minicls_shift_respos_clm_based
	);

sub chop_psmap_into_pspblocks (\@$$\%\@);
sub enum_ps_miniclass_seeds ($\%\%\%\%$$); # ADDED on 2020/03/16.

##sub enum_psm_clusterings_for_new_psps (\@\@$\%\%\%\@$$);
sub enum_psm_clusterings_for_new_psps (\@\@$\%\%\%\@\%$$); # => MODIFIED on 2020/03/16.

sub select_pcls_for_psp_blks (\@\@$\%\%$);

sub mk_node2parsstat_sep_mcs (\@$\%\%$$); # ADDED on 2020/03/17.
sub mk_set_miniclasses_shift_respos (\@\@\@\@$\%\@$$); # ADDED on 2020/03/17.

sub identify_trivial_miniclasses (\@\@\@$$$$$$); # ADDED on 2020/03/19.

##sub list_minicls_shift_respos_clm_based (\@$$$\%\%\%\@$$\@); # A BUG Fixed on 2020/03/20.
sub list_minicls_shift_respos_clm_based (\@$$$$$\%\%\%\@$$\@); # => CORRECTED on 2020/03/20.


	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

=Set aside for potential use. (2)
our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }
=cut # End of "Set aside for potential use. (2)"


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'list_minicls_shift_respos_clm_based'. #
our $TEST_LEVEL_lmc = 0;
#our $TEST_LEVEL_lmc = 5;
##our $TEST_LEVEL_lmc = 6;

	# For 'enum_psm_clusterings_for_new_psps'. # (ADDED in DEBUGGING on 2020/03/22.) 
our $TEST_LEVEL_epcfnp = 0;

	# For 'select_pcls_for_psp_blks'. # (ADDED in DEBUGGING on 2020/03/21.)
our $TEST_LEVEL_spfpb = 0;
#our $TEST_LEVEL_spfpb = 5;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
#


		##### DEFINITIONS of the Subroutines. #####

# chop_psmap_into_pspblocks:
#	This satellite subroutine of 'list_minicls_shift_respos_clm_based'
#	chops an input Position-Shift Map into a set of position-shift-pattern blocks.
#	[CREATED on 2020/03/14.]
#
# INPUTS: (\@map_shifts_respos, $beg_sbj_rg, $end_sbj_rg, \%pspattern2indx, \@set_pspatterns), where
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $char_abs if occupied by a gap;
#
#	$beg_sbj_rg/$end_sbj_rg is the beginning/end of the subject region in the input MSA (in the 0-based full-closed convention);
#
#	%pspattern2indx = (join (':', @pspattern) => $index, ...), where $index is assigned in the order that the position-shift patterns appear in the map (from left to right),
#
#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,
#
#	%pspattern2indx and @set_pspatterns will grow in the course of this subroutine.
#
# OUTPUTS: (\@set_pspattern_blocks), where
#
#	${$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#
sub chop_psmap_into_pspblocks (\@$$\%\@) {

	my ($map_shifts_respos, $beg_sbj_rg, $end_sbj_rg, $pspattern2indx, $set_pspatterns) = @_;

	my $ct_seqs = @{$map_shifts_respos};
	my $ct_psps = @{$set_pspatterns};

	my @set_pspattern_blocks = ();

	for (my $s = $beg_sbj_rg; $s <= $end_sbj_rg; $s++) { # Main outer for-loop (over sites). #

	    my @pspattern = ();

	    for (my $k = 0; $k < $ct_seqs; $k++) { $pspattern[$k] = $map_shifts_respos->[$k][$s]; }

	    my $cnct_pspattern = join (':', @pspattern);

	    my $indx_psp = $pspattern2indx->{$cnct_pspattern}; 
	    unless (defined $indx_psp) {
		$indx_psp = $ct_psps;
		$pspattern2indx->{$cnct_pspattern} = $indx_psp;
		push @{$set_pspatterns}, \@pspattern;
		$ct_psps++;
	    }

	    my $last_block = (@set_pspattern_blocks>0) ? $set_pspattern_blocks[$#set_pspattern_blocks]: undef;

	    if ((defined $last_block) and ($last_block->[3] == $indx_psp)) {

		$last_block->[1] = $s;
		$last_block->[2]++;

	    } else {

		push @set_pspattern_blocks, [$s, $s, 1, $indx_psp];
	    }


	} # End of the main outer for-loop (over sites). #

	return (\@set_pspattern_blocks);

}	# End of the subroutine, 'chop_psmap_into_pspblocks'. #


# enum_ps_miniclass_seeds:
#	This satellite subroutine of 'enum_psm_clusterings_for_new_psps'
#	receives information on a parsimonious clustering of a position-shift pattern,
#	and enumerates the "seeds" of the position-shift miniclasses resulting from the clustering.
#	[CREATED on 2020/03/16.]
#
#	[NOTE]:
#	This subroutine uses a modification of the algorithm for 'fetch_ext_offsprings_nr' in "MyTaxonomy.pm".
#
# INPUTS: ($topnode_id, \%node2ch, \%node2row_indx, \%br2change, \%node2parsstat, $indx_stat, $char_abs), where
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@children, ...) for the interior nodes;
#	%node2row_indx = ($node_id => $row_indx, ...) for the nodes corresponding to the rows of the input MSA;
#
#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id);
#
#	%node2parsstat = ($node_id => $state_info), where
#	$state_info->[$indx_stat] = $char if $indx_stat is defined,
#	or $state_info = $char if $indx_stat is UNdefined,
#	where $char is the character on the node.
#
#	$char_abs is a character representing the "absence" state;
#
# OUTPUTS: (\@set_ps_miniclass_seeds), where
#
#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node for the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.
#
sub enum_ps_miniclass_seeds ($\%\%\%\%$$) {

	my ($topnode_id, $node2ch, $node2row_indx, $br2change, $node2parsstat, $indx_stat, $char_abs) = @_;

		# Preliminary process (1): Create %{$node2depth}. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'enum_ps_miniclass_seeds'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

		# Preliminary process (2): Enumerate the top-nodes for the miniclass seeds.

	my @set_topnodes_mcs = ();

	my $state_info_top = $node2parsstat->{$topnode_id} ;
	my $posshift_top = (defined $indx_stat) ? $state_info_top->[$indx_stat] : $state_info_top ;
	##if ($posshift_top ne $char_abs) { push @set_topnodes_mcs; } # A BUG fixed on 2020/03/18.
	if ($posshift_top ne $char_abs) { push @set_topnodes_mcs, $topnode_id; }

	foreach my $br (sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$br2change}) {
	    my ($state_ue, $state_le) = @{$br2change->{$br}};
	    if ($state_le ne $char_abs) { push @set_topnodes_mcs, $br; }
	}

	my $ct_mcss = @set_topnodes_mcs;

		# Main Process: Characterize each of the miniclass seeds.

	my @set_ps_miniclass_seeds = ();

	for (my $mcs = 0; $mcs < $ct_mcss; $mcs++) { # Main outer for-loop (over miniclass-seeds). #

	    my $topnode_mcs = $set_topnodes_mcs[$mcs];

	    my $shift;
	    my %br2rlv_change;
 	    if ($topnode_mcs == $topnode_id) {
		$shift = $posshift_top ;
	    } else {
		my ($state_ue, $state_le) = @{$br2change->{$topnode_mcs}}; 
		$shift = $state_le;
		$br2rlv_change{$topnode_mcs} = [$state_ue, $state_le];
	    }

	    ##my @seqs_affectedd = (); # A BUG fixed on 2020/03/18.
	    my @seqs_affected = ();
	    my @subjects = ($topnode_mcs);

	    while (1) { # Main middle while-loop (while subject nodes exist). #

		my @new_sbjcts = ();

		foreach my $sbj (@subjects) { # The inner foreach-loop (over subject nodes). #

		    my $row_indx_sbj = $node2row_indx->{$sbj};
		    if (defined $row_indx_sbj) {

				# The node has a corresponding row of the MSA. #
			push @seqs_affected, $row_indx_sbj;

				# Check the consistency. #

			my $state_info_sbj = $node2parsstat->{$sbj};
			my $posshift_sbj = (defined $indx_stat) ? $state_info_sbj->[$indx_stat] : $state_info_sbj ;
			unless ($posshift_sbj eq $shift) {
			    print STDOUT "[ERROR ''] For the $mcs th mini-class seed (w/ \$shift = $shift), the constituent node $sbj (row index = $row_indx_sbj) has the position-shift, \$posshift_sbj = $posshift_sbj, which CLASHES with the position-shift of the mini-class (i.e., $shift): $! !\n\n";
			    return undef;
			}
		    }

		    my $children = $node2ch->{$sbj};
		    unless (defined $children) { next; } # $sbj is an external node. #

		    foreach my $ch (@{$children}) { # The innermost foreach-loop (over child nodes). #

			my $change_ch = $br2change->{$ch};
			if (defined $change_ch) {
				# The child SHOULD NOT belong to this miniclass seed. #
			    ##$br2rlv_change{$ch} = $change; # A BUG fixed on 2020/03/18.
			    $br2rlv_change{$ch} = $change_ch;
			} else {
				# The child SHOULD still belong to this miniclass seed.
			    push @new_sbjcts, $ch;
			}
		    } # End of the innermost foreach-loop (over child nodes). #

		}  # End of the inner foreach-loop (over subject nodes). #

		(@new_sbjcts > 0) or last; # There will be NO subject nodes for the next round. #

			# Update @subjects. #

		@subjects = @new_sbjcts;

	    }  # End of the Main middle while-loop (while subject nodes exist). #

	    my @std_seqs_affected = sort {$a <=> $b} @seqs_affected;

	    $set_ps_miniclass_seeds[$mcs] = [$shift, $topnode_mcs, \@std_seqs_affected, \%br2rlv_change];

	}  # End of the Main outer for-loop (over miniclass-seeds). #


	return \@set_ps_miniclass_seeds ;

} 	# End of the subroutine, 'enum_ps_miniclass_seeds'. #


# enum_psm_clusterings_for_new_psps:
#	This satellite subroutine of 'list_minicls_shift_respos_clm_based'
#	enumerates the parsimonious clusterings for each of the newly listed position-shift patterns.
#	[CREATED on 2020/03/14; REVISED on 2020/03/16]
#
# [ INPUTS: (\@set_pspatterns, \@setset_psm_clusterings,
#	$topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, $char_abs, $char_pres), where ] ... OBSOLETE as of 2020/03/16.
#
# INPUTS: (\@set_pspatterns, \@setset_psm_clusterings,
#	$topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \%node2row_indx, $char_abs, $char_pres), where
#
#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence.
#
#	[ @{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where ] ... OBSOLETE as of 2020/03/16.
#
#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where # MODIFIED on 2020/03/16. #
#
#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs;
#
#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id);
#
#		# (ADDED on 2020/03/16.) (a) #
#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node for the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.
#		# End of "(ADDED on 2020/03/16.) (a)" #

#	[NOTE: During the course of this subroutine,
#		@set_pspatterns will remain unchanged, whereas @set_psm_clusterings will grow.]
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#
#	%node2row_indx = ($node_id => $row_indx, ...) for the nodes corresponding to the rows of the input MSA; # (ADDED on 2020/03/16.) (b) #
#
#	$char_abs/$char_pres is a character representing the "absence"/"presence" state;
#
#
# OUTPUT: $fl_end, where
#
#	$fl_end	= 1 if the subroutine ended without any particular problem,
#		= 0 if it failed.
#
# [ sub enum_psm_clusterings_for_new_psps (\@\@$\%\%\%\@$$) {
#	my ($set_pspatterns, $setset_psm_clusterings,
#	   $topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $char_abs, $char_pres) = @_; ] ... OBSOLETE as of 2020/03/16.
#
sub enum_psm_clusterings_for_new_psps (\@\@$\%\%\%\@\%$$) {

	my ($set_pspatterns, $setset_psm_clusterings,
	   $topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $node2row_indx, $char_abs, $char_pres) = @_; # MODIFIED on 2020/03/16.

	my $ct_psps = @{$set_pspatterns};
	my $ct_psps0 = @{$setset_psm_clusterings};

	unless ($ct_psps > $ct_psps0) {
	    print STDOUT "[INFORMATION 'enum_psm_clusterings_for_new_psps'] The current number of position-shift patterns (=$ct_psps) is NOT LARGER THAN its previous number (=$ct_psps0)!! ... I will END this subroutine without doing anything ...\n\n";
	    return 1;
	}

   (print STDOUT "< Subroutine 'enum_psm_clusterings_for_new_psps' (#{position-shift-patterns}_OLD=$ct_psps0, #{position-shift-patterns}_NEW=$ct_psps) >\n\n") if ($TEST_LEVEL_epcfnp>0); # ADDED in DEBUGGING on 2020/03/22. (1) #

	my $indx_stat = 0; # Following the output format of 'map_parsim_states_sankoff_xgap_1clm'.
	my $if_ignore_gaps = 0; # Include the changes from or to gaps. #


	for (my $psp = $ct_psps0; $psp < $ct_psps; $psp++) { # Main outer for-loop (over NEW position-shift patterns). #

   (print STDOUT "{ For the $psp th position-shift-pattern }\n\n") if ($TEST_LEVEL_epcfnp>1); # ADDED in DEBUGGING on 2020/03/22. (2) #

		# Create @clm. #

	    my $pspattern = $set_pspatterns->[$psp];
	    my @clm = @{$pspattern};

		# Create @pslist. #

	    my %ps2ct;
	    foreach my $ps (@{$pspattern}) {  # 1st middle foreach-loop. #
		if ($ps eq $char_abs) { next; } # Skip gaps. #
		unless (defined $ps2ct{$ps}) { $ps2ct{$ps} = 0; }
		$ps2ct{$ps}++;
	    }
	    my @pslist = sort {$a <=> $b} keys %ps2ct ;

	        # Create @pen_mtrx. #
	    my $ct_pss = @pslist;

	    my @pen_mtrx = ();
	    my @base_row = ();
	    for (my $i =0; $i < $ct_pss; $i++) { $base_row[$i] = 1; }
	    for (my $i =0; $i < $ct_pss; $i++) {  # 3rd middle for-loop. #
		my @cp_row = @base_row;
		$cp_row[$i] = 0;
		$pen_mtrx[$i] = \@cp_row;
	    }

   if ($TEST_LEVEL_epcfnp>2) { # ADDED in DEBUGGING on 2020/03/22. (3)

	print STDOUT '@clm = (', join (', ', @clm), ") .\n\n";

   } # End of "if ($TEST_LEVEL_epcfnp>2) {...}"

	    my $set_node2parsstat = map_parsim_states_sankoff_xgap_1clm ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @clm, $char_abs, $char_pres, @pslist, @pen_mtrx, 1);
	    unless (defined $set_node2parsstat) {
		print STDOUT "[ERROR 'enum_psm_clusterings_for_new_psps'] The subroutine, 'map_parsim_states_sankoff_xgap_1clm', FAILED for the $psp th position-shift pattern: $! !\n";
		return 0;
	    }

	    my $ct_pcls = @{$set_node2parsstat};

	    my @set_psm_clusterings = ();

	    for (my $pcl = 0; $pcl < $ct_pcls; $pcl++) { # (3+1) th middle for-loop (over parsimonious clusterings). #

		my $node2parsstat = $set_node2parsstat->[$pcl];

		my ($br2change, $ct_changes) =  locate_parsim_changes_xgap (%{$node2pa}, %{$node2parsstat}, $indx_stat, $char_abs, $if_ignore_gaps);
		unless (defined $br2change) {
		    print STDOUT "[ERROR 'enum_psm_clusterings_for_new_psps'] The subroutine, 'locate_parsim_changes_xgap', FAILED on the $pcl th parsimonious clustering for the $psp th position-shift pattern: $! !\n";
		    return 0;
		}

				# (ADDED in REVISION on 2020/03/16.) #
		my ($set_ps_miniclass_seeds) = enum_ps_miniclass_seeds ($topnode_id, %{$node2ch}, %{$node2row_indx}, %{$br2change}, %{$node2parsstat}, $indx_stat, $char_abs);
		unless (defined $set_ps_miniclass_seeds) {
		    print STDOUT "[ERROR 'enum_psm_clusterings_for_new_psps'] The subroutine, 'enum_ps_miniclass_seeds', FAILED on the $pcl th parsimonious clustering for the $psp th position-shift pattern: $! !\n";
		    return 0;
		}
				# END of "(ADDED in REVISION on 2020/03/16.)" #

		##$set_psm_clusterings[$pcl] = [$node2parsstat, $br2change];
		$set_psm_clusterings[$pcl] = [$node2parsstat, $br2change, $set_ps_miniclass_seeds]; # MODIFIED on 2020/03/16. #

	    } # End of the (3+1) th middle for-loop (over parsimonious clusterings). #

	    ##$setset_psm_clusterings->[$psp] = \@set_psp_clusterings; # Grow @{$setset_psm_clustrings}. # # A BUG fixed on 2020/03/18.
	    $setset_psm_clusterings->[$psp] = \@set_psm_clusterings; # Grow @{$setset_psm_clustrings}. #

	} # END of the Main outer for-loop (over NEW position-shift patterns). #


   (print STDOUT "< Subroutine 'enum_psm_clusterings_for_new_psps' Ended />\n\n") if ($TEST_LEVEL_epcfnp>0); # ADDED in DEBUGGING on 2020/03/22. (3+1)

	return 1;

}	# End of the subroutine, 'enum_psm_clusterings_for_new_psps'. #


# select_pcls_for_psp_blks:
#	This subroutine selects some "optimum" combinations of parsimonious clusterings
#	for the positon-shift-pattern blocks.
#	[CREATION STARTED on 2020/03/14, ENDED on 2020/03/15.]
#
#	[STRATEGY for the Selection]:
#	(1) First, for the position-shift pattern blocks with only one way of clustering each,
#		choose that unique clustering;
#	(2) For those blocks flanked by those blocks with uniquely selected clustering(s),
#		choose the clustering(s) with the largest number of changes consistent with the flanking unique clustering(s);
#	(3) Repeat (2) until the clusterings cannot be narrowed down any longer;
#	(3+1) If some remaining blocks are contiguous to each other,
#		examine, from left to right, the consistency of the changes,
#		and most consistent pairs will be chosen,
#		and this selection will be done until such contiguous remaining blocks are exhausted.
#	(5) If there are still more than one combination of the clusterings,
#	    further select according to the following criteria (in this order):
#		(i) the largest sum of the total depths of the branches along which changes occur, over all position-shift-pattern blocks,
#		(ii) the largest sum of the minimum depths of change-occuring-branches over all position-shift-pattern blocks.
#	(6) If there are still more than one combination of the clusterings, keep them as equally optimum choices.
#
#
# INPUTS: (\@set_pspattern_blocks, \@setset_psm_clusterings, $topnode_id, \%node2ch, \%node2pa, $char_abs), where
#
#	${$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#	[\@set_pspattern_blocks is the main output of 'chop_psmap_into_pspblocks'.]
#
#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where
#
#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs.
#
#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id).
#
#	[\@setset_psm_clusterings is a growing input of 'list_minicls_shift_respos_clm_based'.]
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@children, ...) for the interior nodes;
#	%node2pa = ($node_id => $parent_id, ...) for the non-top nodes;
#
#	$char_abs is a character representing the "absence" state;
#
#
# OUTPUTS: (\@sets_pcls_assoc_w_psp_blks), where
#
#	@pcls_assoc_w_psp_blks = @{$sets_pcls_assoc_w_psp_blks[$s]} is the $s th set of
#	"optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$pcls_assoc_w_psp_blks[$bl] is the index (in @{$setset_psm_clusterings[$indx_psp]} where $indx_psp = $set_pspattern_blocks->[$bl][3]) of the parsimonious clusterings "optimum" for the $psp th position-shift-pattern block.
#
sub select_pcls_for_psp_blks (\@\@$\%\%$) {

	my ($set_pspattern_blocks, $setset_psm_clusterings,
		$topnode_id, $node2ch, $node2pa, $char_abs) = @_;

	my $ct_blks = @{$set_pspattern_blocks};

   (print STDOUT "< Subroutine 'select_pcls_for_psp_blks' (#{psp-blocks}=$ct_blks) >\n\n") if ($TEST_LEVEL_spfpb >0);

	# my ($set_pscls_assoc_w_psp_blks) = select_pcls_for_psp_blks (@{$set_pspattern_blocks}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, %{$node2pa}, $char_abs);

		# Preliminary process (1): create %{$node2depth}. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'select_pcls_for_psp_blks'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

		# Main Process (1): Initialize @pcls_selected_for_blks & %bl2new #
		# (i.e., identify the position-shift-pattern blocks each of which has a UNIQUE parsimonious clustering from the beginning.)

	my @pcls_selected_for_blks = (); # The $bl th element = [indexes of parsimonious clusterings selected for the $bl th position-shift-pattern block].
	my %bl_new_unq; # = ($bl => 1, ...) for ALL blocks whose UNIQUE "optimum" parsimonious clustering has been NEWLY selected.

	for (my $bl =0; $bl <$ct_blks; $bl++) { # 1st Main for-loop (over the blocks). #
	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]};
	    my $ct_pcls = @{$setset_psm_clusterings->[$indx_psp]};
	    my @indexes_pcls = (0 .. $ct_pcls-1);
	    $pcls_selected_for_blks[$bl] = \@indexes_pcls;
	    if ($ct_pcls == 1) { $bl_new_unq{$bl} = 1; }
	} # End of the 1st Main for-loop (over the blocks). #

   if ($TEST_LEVEL_spfpb >2) { # (M1) #

	print STDOUT "\n[ Results of the Main Process (1) ]\n\n";

	print STDOUT '( Contents of @pcls_selected_for_blks: )', "\n\n";
	print STDOUT join ("\t", 'Block_ID', '(indexes,of,psm-clusterings)'), "\n\n";
	for (my $bl=0; $bl <$ct_blks; $bl++) {
	    my $cnct_indexes_pcls = join (',', @{$pcls_selected_for_blks[$bl]});
	    print STDOUT join ("\t", $bl, "($cnct_indexes_pcls)"), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '%bl_new_unq = ( # $bl => 1 (for ALL blocks with NEW unique "optimum" psm-clusterings', "\n\n";
	foreach my $bl (sort {$a <=> $b} keys %bl_new_unq) {
		print STDOUT "\t", $bl , ' => ', $bl_new_unq{$bl}, " ,\n";
	}
	print STDOUT "\t\t) .\n\n\n";

   } # End of "if ($TEST_LEVEL_spfpb >2) { # (M1) # ...}"


		# Main Process (2) & (3): 
		# (2) Choose the clustering(s) with the largest number of changes consistent with the flanking unique clustering(s);
		# & (3) Repeat (2) until the clusterings cannot be narrowed down any longer;

	while (0 < scalar (keys %bl_new_unq) ) { # 2nd Main Outermost while-loop (while unique clustering(s) is(are) newly found). #

	    my %new_bl_new_unq;

	    for (my $bl =0; $bl <$ct_blks; $bl++) { # Main outer for-loop (over the blocks). #

		my $curr_opt_pcls = $pcls_selected_for_blks[$bl];
		my $ct_curr_opt_pcls = @{$curr_opt_pcls};
		if ($ct_curr_opt_pcls < 2) { next; } # Skip if the block already has a unique clustering.

		my @to_be_consistent = (); # A list of flanking blocks whose consistency with $bl MUST be examined.
		foreach my $bl2 ($bl-1, $bl+1) { # Preliminary middle foreach-loop. #
		    if (defined $bl_new_unq{$bl2}) { push @to_be_consistent, $bl2; }
		}

		if (@to_be_consistent == 0) { next; } # Skip if no blocks are found to check the consistency with $bl.

		my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]};
		my $set_pcls = $setset_psm_clusterings->[$indx_psp];

		my @new_curr_opt_pcls = ();
		my $max_consistency = 0;

		foreach my $indx_pcl (@{$curr_opt_pcls}) { # Main middle foreach-loop (over the current "optimum" parsimonious clusterings for $bl). #

		    my ($node2parsstat, $br2change) = @{$set_pcls->[$indx_pcl]};
		    my $consistency = 0;

		    foreach my $bl2 (@to_be_consistent) { # Main inner foreach-loop (over the blocks whose consistency with $bl MUST be checked.) #

			my $curr_opt_pcls2 = $pcls_selected_for_blks[$bl2];
			my $indx_pcl2 = $curr_opt_pcls2->[0];

			my ($beg_bl2, $end_bl2, $len_bl2, $indx_psp2) = @{$set_pspattern_blocks->[$bl2]};
			my $set_pcls2 = $setset_psm_clusterings->[$indx_psp2];

			my ($node2parsstat2, $br2change2) = @{$set_pcls2->[$indx_pcl2]};

			foreach my $br (keys %{$br2change}) { # Main inner-inner foreach-loop (over the branches where the position-shifts change). #
			    my $change2 = $br2change2->{$br};
			    unless (defined $change2) { next; }

			    my $change1 = $br2change->{$br};

			    foreach my $i (0, 1) { # Main innermost foreach-loop (examining upper- and lower-ends). #
				my $posshift1 = $change1->[$i];
				if ($posshift1 eq $char_abs) { next; }
				my $posshift2 = $change2->[$i];
				if ($posshift2 eq $char_abs) { next; }
				if ($posshift1 == $posshift2) { $consistency++; }
			    } # End of the main innermost foreach-loop (examining upper- and lower-ends). #

			}  # End of the main inner-inner foreach-loop (over the branches where the position-shifts change). #
		    } # End of the main inner foreach-loop (over the blocks whose consistency with $bl MUST be checked.) #

		    if ($consistency < $max_consistency) { next; }

			# $indx_pcl is currently the most consistent. #

		    if ($consistency > $max_consistency) {

			$max_consistency = $consistency;
			@new_curr_opt_pcls = (); # Start the set from scratch.
		    }

		    push @new_curr_opt_pcls, $indx_pcl;

#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where


		} # End of the main middle foreach-loop (over the current "optimum" parsimonious clusterings for $bl). #

			# Update the $bl th element of @pcls_selected_for_blks. #

		$pcls_selected_for_blks[$bl] = \@new_curr_opt_pcls;

		if (@new_curr_opt_pcls < 2) { $new_bl_new_unq{$bl} = 1; }

	    } # End of the main outer for-loop (over the blocks). #

		# Update %bl_new_unq.

	    %bl_new_unq = %new_bl_new_unq;

	} # End of the 2nd Main Outermost while-loop (while unique clustering(s) is(are) newly found). #


   if ($TEST_LEVEL_spfpb >2) { # (M2&3) #

	print STDOUT "\n[ Results of the Main Processes (2&3) ]\n\n";

	print STDOUT '( Contents of @pcls_selected_for_blks: )', "\n\n";
	print STDOUT join ("\t", 'Block_ID', '(indexes,of,psm-clusterings)'), "\n\n";
	for (my $bl=0; $bl <$ct_blks; $bl++) {
	    my $cnct_indexes_pcls = join (',', @{$pcls_selected_for_blks[$bl]});
	    print STDOUT join ("\t", $bl, "($cnct_indexes_pcls)"), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT '%bl_new_unq = ( # $bl => 1 (for ALL blocks with NEW unique "optimum" psm-clusterings', "\n\n";
	foreach my $bl (sort {$a <=> $b} keys %bl_new_unq) {
		print STDOUT "\t", $bl , ' => ', $bl_new_unq{$bl}, " ,\n";
	}
	print STDOUT "\t\t) .\n\n\n";

   } # End of "if ($TEST_LEVEL_spfpb >2) { # (M2&3) # ...}"



		# Main Process (3+1): Enumerate all "most pairwise-neighboring-consistent" sets of parsimonious clusterings, by growing them from the selected sest of clusterings for the $bl = 0 th block.

# 	(3+1) If some remaining blocks are contiguous to each other,
#		examine, from left to right, the consistency of the changes,
#		and most consistent pairs will be chosen,
#		and this selection will be done until such contiguous remaining blocks are exhausted.

		# Initialize the set of sets of "pairwise optimum" parsimonious clusterings.

	my @prepresets_pcls_assoc_w_psp_blks = ();

	my $pcls_selected_for_leftmost = $pcls_selected_for_blks[0];

	foreach my $indx_pcl (@{$pcls_selected_for_leftmost}) {
	    push @prepresets_pcls_assoc_w_psp_blks, [$indx_pcl];
	}

	##my $if_unique_lf = (@{$pcls_selected_for_leftmost < 2) ? 1 : 0; # A BUG fixed on 2020/03/18.
	my $if_unique_lf = (@{$pcls_selected_for_leftmost} < 2) ? 1 : 0;

	for (my $bl=1; $bl < $ct_blks; $bl++) { # 3rd Main Outermost for-loop (over the blocks except 0). # 

	   my $pcls_selected_for_sbj = $pcls_selected_for_blks[$bl];

	   if (@{$pcls_selected_for_sbj} < 2) {

		# The subject block already has a unique "optimum" parsimonous clustering.
		# => Just incorporate it into all the sets of clusterings associated with blocks.

		my $indx_pcls = $pcls_selected_for_sbj->[0];

		foreach my $pcls_assoc_w_psp_blks (@prepresets_pcls_assoc_w_psp_blks) {
		    $pcls_assoc_w_psp_blks->[$bl] = $indx_pcls;
		}

			# Update $if_unique_lf. #

		$if_unique_lf = 1;

		next;

	   } # End of "if (@{$pcls_selected_for_sbj} < 2) {...}"

	    if ($if_unique_lf) {

		# The left-flanking block already had a unique "optimum" parsimonous clustering.
		# =>  incorporate all the selected clusterings for the subject block into all the sets of clusterings associated with blocks, without doing any consistency checks(, which should have no effects).

		my @new_prepresets = ();
		foreach my $pcls_assoc_w_psp_blks (@prepresets_pcls_assoc_w_psp_blks) { # Preliminary outer foreach-loop. #

		   foreach my $indx_pcls (@{$pcls_selected_for_sbj}) { # Preliminary inner foreach-loop. #

			my @cp_pcls_assoc_w_psp_blks = @{$pcls_assoc_w_psp_blks};
			$cp_pcls_assoc_w_psp_blks[$bl] = $indx_pcls;
			push @new_prepresets, \@cp_pcls_assoc_w_psp_blks;

		   } # End of the preliminary inner foreach-loop. #

		} # End of the preliminary outer foreach-loop. #

			# Update @prepresets_pcls_assoc_w_psp_blks. #

		@prepresets_pcls_assoc_w_psp_blks = @new_prepresets;

			# Update $if_unique_lf. #

		$if_unique_lf = 0;

		next;

	   } # End of "if ($if_unique_lf) {...}"


			## Otherwise, Check the consistency between the $bl th and ($bl-1) th blocks. ##

	    my ($beg_sbj, $end_sbj, $len_sbj, $indx_psp_sbj) = @{$set_pspattern_blocks->[$bl]};
	    my $set_pcls_sbj = $setset_psm_clusterings->[$indx_psp_sbj];

	    my ($beg_lf, $end_lf, $len_lf, $indx_psp_lf) = @{$set_pspattern_blocks->[$bl-1]};
	    my $set_pcls_lf = $setset_psm_clusterings->[$indx_psp_lf];

	    my @pw_optimun_sets = ();
	    my $max_consistency = 0;

	    my @new_prepresets = ();
	    foreach my $pcls_assoc_w_psp_blks (@prepresets_pcls_assoc_w_psp_blks) { # Main outer foreach-loop. #

		my $indx_pcl_lf = $pcls_assoc_w_psp_blks->[$bl-1];
		my ($node2parsstat_lf, $br2change_lf) = @{$set_pcls_lf->[$indx_pcl_lf]};

		foreach my $indx_pcl_sbj (@{$pcls_selected_for_sbj}) { # Main middle foreach-loop. #

		    my ($node2parsstat_sbj, $br2change_sbj) = @{$set_pcls_sbj->[$indx_pcl_sbj]};

		    my $consistency = 0;

		    foreach my $br (keys %{$br2change_sbj}) { # Main inner-inner foreach-loop (over the branches where the position-shifts change). #
		        my $change_lf = $br2change_lf->{$br};
		        unless (defined $change_lf) { next; }

		        my $change_sbj = $br2change_sbj->{$br};

		        foreach my $i (0, 1) { # Main innermost foreach-loop (examining upper- and lower-ends). #
				my $posshift_lf = $change_lf->[$i];
				if ($posshift_lf eq $char_abs) { next; }
				my $posshift_sbj = $change_sbj->[$i];
				if ($posshift_sbj eq $char_abs) { next; }
				if ($posshift_lf == $posshift_sbj) { $consistency++; }
		        } # End of the main innermost foreach-loop (examining upper- and lower-ends). #

		    }  # End of the main inner-inner foreach-loop (over the branches where the position-shifts change). #

		    if ($consistency < $max_consistency) { next; }

			# The combination is currently the most consistent. #

		    my @new_pcls_assoc_w_psp_blks = @{$pcls_assoc_w_psp_blks};
		    ##$new_pcls_assoc_w_psp_blks[$bl] = $indx_pcls_sbj; # A BUG fixed on 2020/03/18.
		    $new_pcls_assoc_w_psp_blks[$bl] = $indx_pcl_sbj;

		    if ($max_consistency < $consistency) {

			$max_consistency = $consistency;
			@new_prepresets = (); # Dump all the combinations that were the most consistent thus far.
		    }

		    push @new_prepresets, \@new_pcls_assoc_w_psp_blks;

		} # End of the Main middle foreach-loop. #

	    } # End of the Main outer foreach-loop. #


	    	# Update @prepresets_pcls_assoc_w_psp_blks. #

	    @prepresets_pcls_assoc_w_psp_blks = @new_prepresets;

	    	# Update $if_unique_lf. #

	    $if_unique_lf = 0; # Because we already know that (@{$pcls_selected_for_sbj} >= 2) .

=Probably WRONG.
	    my %indx2ct;
	    foreach my $new_pcls_assoc_w_psp_blks (@new_prepresets) {
		my $indx_pcl_sbj = $new_pcls_assoc_w_psp_blks[$bl];
		unless (defined $indx2ct{$indx_pcl_sbj}) { $indx2ct{$indx_pcl_sbj} = 0; }
		$indx2ct{$indx_pcl_sbj}++;
	    }

	    $if_unique_lf = (scalar (keys %indx2ct) < 2) 1 : 0;
=cut # END of "Probably WRONG."

	}  # End of the 3rd Main Outermost for-loop (over the blocks except 0). # 


   if ($TEST_LEVEL_spfpb >2) { # (M3+1) #

	print STDOUT "\n[ Results of the Main Processes (3+1) ]\n\n";

	my $ct_sets = @prepresets_pcls_assoc_w_psp_blks;

	print STDOUT '( Contents of @prepresets_pcls_assoc_w_psp_blks ', "(#{sets}=$ct_sets): )\n\n";

	print STDOUT join ("\t", 'Indx', '(indexes,of,psm-clusterings,assoc,w/,psp-blocks)'), "\n\n";

	for (my $i =0; $i < $ct_sets; $i++) {
	    my $pcls_assoc_w_psp_blks = $prepresets_pcls_assoc_w_psp_blks[$i];
	    my @out = ();
	    for (my $bl=0; $bl <$ct_blks; $bl++) {
		my $pcl = $pcls_assoc_w_psp_blks->[$bl];
		unless (defined $pcl) { $pcl = 'undef'; }
		$out[$bl] = $pcl;
	    }
	    my $cnct_out = join (',', @out);

	    print STDOUT join ("\t", $i, "($cnct_out)"), "\n";
	}
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_spfpb >2) { # (M3+1) # ...}"


		# Main Process (5)-(i):  Select those sets of clusterings with the largest sum of the total depths of the position-shift-changing-branches, over all position-shift-pattern blocks.

#	(5) If there are still more than one combination of the clusterings,
#	    further select according to the following criteria (in this order):
#		(i) the largest sum of the total depths of the branches along which changes occur, over all position-shift-pattern blocks,

	my @presets_pcls_assoc_w_psp_blks = ();
	my $maxsum_totdepths = -1;

	foreach my $pcls_assoc_w_psp_blks (@prepresets_pcls_assoc_w_psp_blks) { # The (3+1) th Main Outermost foreach-loop (over the prepre-optimum sets of clusterings associated with psp-blocks). #

	    my $sum_totdepths = 0;
	    for (my $bl=0; $bl < $ct_blks; $bl++) { # Main outer for-loop (over the psp-blocks). #

		my $indx_pcl_bl = $pcls_assoc_w_psp_blks->[$bl];

		my ($beg_bl, $end_bl, $len_bl, $indx_psp_bl) = @{$set_pspattern_blocks->[$bl]};
		my $set_pcls_bl = $setset_psm_clusterings->[$indx_psp_bl];

		my ($node2parsstat_bl, $br2change_bl) = @{$set_pcls_bl->[$indx_pcl_bl]};

		my $totdepth = 0;
		foreach my $br (keys %{$br2change_bl}) { $totdepth += $node2depth->{$br}; }

		$sum_totdepths += $totdepth;

	    }  # End of the Main outer for-loop (over the psp-blocks). #

	    if ($maxsum_totdepths > $sum_totdepths) { next; }

			# This set of clusterings has the current largest sum of the total depths. #

	    if ($maxsum_totdepths < $sum_totdepths) {
		$maxsum_totdepths = $sum_totdepths;
		@presets_pcls_assoc_w_psp_blks = (); # Start the set from scratch. #
	    }

	    push @presets_pcls_assoc_w_psp_blks, $pcls_assoc_w_psp_blks;

	}  # End of The (3+1) th Main Outermost foreach-loop (over the prepre-optimum sets of clusterings associated with psp-blocks). #


   if ($TEST_LEVEL_spfpb >2) { # (M5-i) #

	print STDOUT "\n[ Results of the Main Processes (5-(i)) ]\n\n";

	my $ct_sets = @presets_pcls_assoc_w_psp_blks;

	print STDOUT '( Contents of @presets_pcls_assoc_w_psp_blks ', "(#{sets}=$ct_sets): )\n\n";

	print STDOUT join ("\t", 'Indx', '(indexes,of,psm-clusterings,assoc,w/,psp-blocks)'), "\n\n";

	for (my $i =0; $i < $ct_sets; $i++) {
	    my $pcls_assoc_w_psp_blks = $presets_pcls_assoc_w_psp_blks[$i];
	    my @out = ();
	    for (my $bl=0; $bl <$ct_blks; $bl++) {
		my $pcl = $pcls_assoc_w_psp_blks->[$bl];
		unless (defined $pcl) { $pcl = 'undef'; }
		$out[$bl] = $pcl;
	    }
	    my $cnct_out = join (',', @out);

	    print STDOUT join ("\t", $i, "($cnct_out)"), "\n";
	}
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_spfpb >2) { # (M5-i) # ...}"


		# Main Process (5)-(ii):  Select those sets of clusterings with the largest sum of the minimum depths of the position-shift-changing-branches, over all position-shift-pattern blocks.

#		(ii) the largest sum of the minimum depths of change-occuring-branches over all position-shift-pattern blocks.

	my @sets_pcls_assoc_w_psp_blks = ();
	my $maxsum_mindepths = -1;

	foreach my $pcls_assoc_w_psp_blks (@presets_pcls_assoc_w_psp_blks) { # The 5 th Main Outermost foreach-loop (over the pre-optimum sets of clusterings associated with psp-blocks). #

	    my $sum_mindepths = 0;
	    for (my $bl=0; $bl < $ct_blks; $bl++) { # Main outer for-loop (over the psp-blocks). #

		my $indx_pcl_bl = $pcls_assoc_w_psp_blks->[$bl];

		my ($beg_bl, $end_bl, $len_bl, $indx_psp_bl) = @{$set_pspattern_blocks->[$bl]};
		my $set_pcls_bl = $setset_psm_clusterings->[$indx_psp_bl];

		my ($node2parsstat_bl, $br2change_bl) = @{$set_pcls_bl->[$indx_pcl_bl]};

		my $mindepth;
		foreach my $br (keys %{$br2change_bl}) {
		    my ($ps_ue, $ps_le) = @{$br2change_bl->{$br}};
		    if (($ps_ue eq $char_abs) or ($ps_le eq $char_abs)) { next; } # Skip if the change is to or from the gap.

		    my $depth += $node2depth->{$br};
		    if (!(defined $mindepth) or ($depth < $mindepth)) {
			$mindepth = $depth;
		    }
		}

		if (defined $mindepth) { $sum_mindepths += $mindepth; }

	    }  # End of the Main outer for-loop (over the psp-blocks). #

	    if ($maxsum_mindepths > $sum_mindepths) { next; }

			# This set of clusterings has the current largest sum of the minimum depths. #

	    if ($maxsum_mindepths < $sum_mindepths) {
		$maxsum_mindepths = $sum_mindepths;
		@sets_pcls_assoc_w_psp_blks = (); # Start the set from scratch. #
	    }

	    push @sets_pcls_assoc_w_psp_blks, $pcls_assoc_w_psp_blks;

	}  # End of The 5 th Main Outermost foreach-loop (over the pre-optimum sets of clusterings associated with psp-blocks). #

   if ($TEST_LEVEL_spfpb >2) { # (M5-ii) #

	print STDOUT "\n[ Results of the Main Processes (5-(ii)) ]\n\n";

	my $ct_sets = @sets_pcls_assoc_w_psp_blks;

	print STDOUT '( Contents of @sets_pcls_assoc_w_psp_blks ', "(#{sets}=$ct_sets): )\n\n";

	print STDOUT join ("\t", 'Indx', '(indexes,of,psm-clusterings,assoc,w/,psp-blocks)'), "\n\n";

	for (my $i =0; $i < $ct_sets; $i++) {
	    my $pcls_assoc_w_psp_blks = $sets_pcls_assoc_w_psp_blks[$i];
	    my @out = ();
	    for (my $bl=0; $bl <$ct_blks; $bl++) {
		my $pcl = $pcls_assoc_w_psp_blks->[$bl];
		unless (defined $pcl) { $pcl = 'undef'; }
		$out[$bl] = $pcl;
	    }
	    my $cnct_out = join (',', @out);

	    print STDOUT join ("\t", $i, "($cnct_out)"), "\n";
	}
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_spfpb >2) { # (M5-ii) # ...}"


#	(6) If there are still more than one combination of the clusterings, keep them as equally optimum choices.


   (print STDOUT "< Subroutine 'select_pcls_for_psp_blks' Ended />\n\n") if ($TEST_LEVEL_spfpb >0);

	return (\@sets_pcls_assoc_w_psp_blks);

}	# End of the subroutine, 'select_pcls_for_psp_blks'. #


# mk_node2parsstat_sep_mcs:
#	This satellite subroutine of 'mk_set_miniclasses_shift_respos'
#	creats %node2parsstat_sep_mcs, which shows the nodes involved in the subject MINI-class seed.
#	[CREATED on 2020/03/17.]
#
#	[NOTE]:
#	This was created by slightly modifying the main algorithm for 'enum_ps_miniclass_seeds' above.
#
# INPUTS: (\@set_all_nodes, $topnode_mcs, \%br2rlv_change, \%node2ch, $char_abs, $char_pres), where
#
#	@set_all_nodes stores the IDs of ALL nodes;
#
#	$topnode_mcs is the top-node of the sub-tree of the nodes involved in the subject MINI-class seed;
#
#	%br2rlv_change = ($br_id => [$state_ue, $state_le], ...) for all the branches along which relevant changes occur;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#
#	$char_abs/$char_pres is a character representing the "absence"/"presence" state;
#
# OUTPUTS: (\%node2parsstat_sep_mcs), where
#
#	%node2parsstat_sep_mcs shows the states of nodes 
#	under the set of "parsimonious" changes (i.e,, %br2rlv_change)
#	that separates the subject MINI-class seed from the rest.
#	More specifically,
#	$node2parsstat_sep_mcs{$node}	= $char_pres if the node is involved in the subject MINI-class seed,
#					= $char_abs otherwise.
#
sub mk_node2parsstat_sep_mcs (\@$\%\%$$) {

	my ($set_all_nodes, $topnode_mcs, $br2rlv_change, $node2ch, $char_abs, $char_pres) = @_;

		# Preliminary: Initialize the output. #
	my %node2parsstat_sep_mcs;
	foreach my $node (@{$set_all_nodes}) { $node2parsstat_sep_mcs{$node} = $char_abs; }


		# MAIN: exhaustively search for the nodes involved in this MINI-class seed. #

	my @subjects = ($topnode_mcs);

	while (1) { # Main Outer while-loop (while subject nodes exist). #

	    my @new_sbjcts = ();

	    foreach my $sbj (@subjects) { # The middle foreach-loop (over subject nodes). #

		##$node2parsstat{$sbj} = $char_pres; # A BUG fixed on 2020/03/18.
		$node2parsstat_sep_mcs{$sbj} = $char_pres;

		my $children = $node2ch->{$sbj};
		unless (defined $children) { next; } # $sbj is an external node. #

		foreach my $ch (@{$children}) { # The inner foreach-loop (over child nodes). #

		    my $change_ch = $br2rlv_change->{$ch};
		    if (defined $change_ch) {
		    	# The child SHOULD NOT belong to this miniclass seed. #
			# [ $br2rlv_change{$ch} = $change; ] ... OBSOLETE.
			next;

		    } else {
		    	# The child SHOULD still belong to this miniclass seed.
			push @new_sbjcts, $ch;
		    }

		} # End of the inner foreach-loop (over child nodes). #

	    }  # End of the middle foreach-loop (over subject nodes). #

	    (@new_sbjcts > 0) or last; # There will be NO subject nodes for the next round. #

			# Update @subjects. #

	    @subjects = @new_sbjcts;

	}  # End of the Main Outer while-loop (while subject nodes exist). #

	##return \%node2parsstat; # A BUG fixed on 2020/03/18.
	return \%node2parsstat_sep_mcs;

}	# End of the subroutine, 'mk_node2parsstat_sep_mcs'. #


# mk_set_miniclasses_shift_respos:
#	This subroutine creates a set of MINI-classes of runs of shifts of residue-positions, 
#	under a particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks,
#	by merging effectively neighboring mini-class seeds having the IDENTICAL shift and the IDENTICAL set of affected MSA rows. 
#	[CREATION STARTED on 2020/03/16, ENDED on 2020/03/17.]
#
# [ INPUTS: (\@pcls_assoc_w_psp_blks, \@set_paspattern_blocks, \@set_pspatterns, \@setset_psm_clusterings, $topnode_id, \%node2ch, $char_abs, $char_pres), where ] ... OBSOLETE as of 2020/03/17.
#
# [ INPUTS: (\@pcls_assoc_w_psp_blks, \@set_paspattern_blocks, \@set_pspatterns, \@setset_psm_clusterings, $topnode_id, \%node2ch, \@nodes_in_msa, $char_abs, $char_pres), where # REVISED on 2020/03/17. ]
#
# INPUTS: (\@pcls_assoc_w_psp_blks, \@set_pspattern_blocks, \@set_pspatterns, \@setset_psm_clusterings, $topnode_id, \%node2ch, \@nodes_in_msa, $char_abs, $char_pres), where # => CORRECTED on 2020/03/18.
#
#	@pcls_assoc_w_psp_blks specifies a particular set of "optimum" block-associated parsimonious clusterings,
#	more precisely, $pcls_assoc_w_psp_blks[$bl] is the index of the parsimonious clustering associated with the $bl th block;
#	[\@pcls_assoc_w_psp_blks is an element of the main output of 'select_pcls_for_psp_blks' above.]
#
#	@{$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#	[ \@set_pspattern_blocks is the main output of 'chop_psmap_into_pspblocks' above.]
#
#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,
#
#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where 

#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs;
#
#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id);
#
#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node fot the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@children, ...) for the interior nodes;
#	$nodes_in_msa[$k] is the node ID corresponding to the $k th row of the input MSA; # Added on 2020/03/17.
#
#	$char_abs/$char_pres is a character representing the "absence"/"presence" state;
#
#
# OUTPUTS: (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses), where
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_miniclasses[$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#
#
# [ sub mk_set_miniclasses_shift_respos (\@\@\@\@$\%$$) {
#	my ($pcls_assoc_w_psp_blks, $set_paspattern_blocks, $set_pspatterns, $setset_psm_clusterings, $topnode_id, $node2ch, $char_abs, $char_pres) = @_; ] ... OBSOLETE as of 2020/03/17.
#
sub mk_set_miniclasses_shift_respos (\@\@\@\@$\%\@$$) {

	##my ($pcls_assoc_w_psp_blks, $set_paspattern_blocks, $set_pspatterns, $setset_psm_clusterings, $topnode_id, $node2ch, $nodes_in_msa, $char_abs, $char_pres) = @_; # Revised on 2020/03/17.
	my ($pcls_assoc_w_psp_blks, $set_pspattern_blocks, $set_pspatterns, $setset_psm_clusterings, $topnode_id, $node2ch, $nodes_in_msa, $char_abs, $char_pres) = @_; # => DEBUGGED on 2020/03/18.


	my $ct_psp_blks = @{$set_pspattern_blocks};

	unless (@{$pcls_assoc_w_psp_blks} == $ct_psp_blks) {
	    my $ct2 = @{$pcls_assoc_w_psp_blks};
	    print STDOUT "[ERROR 'mk_set_miniclasses_shift_respos'] ILLEGAL INPUT: The number of psp-blocks in ", '@{pcls_assoc_w_psp_blks}', " (=$ct2) CLASHES with that in ", '@{$set_pspattern_blocks}', " (=$ct_psp_blks): $! !\n\n";
	    return (undef, undef);
	}

	unless ($ct_psp_blks > 0) {
	    print STDOUT "[INFORMATION 'mk_set_miniclasses_shift_respos'] There are NO position-shift-pattern blocks in the input, ", '@{$set_pspattern_blocks}', ": $! !\n\n";
	    return ([], []);
	}

		# Preliminary (1): Create @set_pspatterns_for_blks, @set_node2parsstat_for_blks, @set_br2change_for_blks, and @setset_ps_miniclass_seeds_for_blks.

	my @set_pspatterns_for_blks = my @set_node2parsstat_for_blks = ();
	my @set_br2change_for_blks = my @setset_ps_miniclass_seeds_for_blks = ();

	for (my $bl=0; $bl <$ct_psp_blks; $bl++) { # 1st preliminary for-loop (over position-shift-pattern blocks). #


	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]} ;

	    my $pspattern = $set_pspatterns->[$indx_psp];

	    my $indx_pcl = $pcls_assoc_w_psp_blks->[$bl];

	    ##my ($node2parsstat, $br2change, $set_ps_miniclass_seeds) = @{$setset_psm_clusterings->[$indx_psp][$indx_pcl]; # A BUG fixed on 2020/03/18.
	    my ($node2parsstat, $br2change, $set_ps_miniclass_seeds) = @{$setset_psm_clusterings->[$indx_psp][$indx_pcl]};

	    ##$set_pspattern_for_blks[$bl] = $pspattern; # A BUG fixed on 2020/03/18.
	    $set_pspatterns_for_blks[$bl] = $pspattern;

	    $set_node2parsstat_for_blks[$bl] = $node2parsstat;
	    $set_br2change_for_blks[$bl] = $br2change;
	    $setset_ps_miniclass_seeds_for_blks[$bl] = $set_ps_miniclass_seeds;

#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,
#
#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds)
#	 is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where 

	} # End of the 1st preliminary for-loop (over position-shift-pattern blocks). #


		# Preliminary (2): Create @set_all_nodes. #

	my @set_all_nodes = sort {$a <=> $b} keys %{$set_node2parsstat_for_blks[0]};

		# Preliminary (3): Initialize @set_already. #

	my @set_already = (); # %{$set_already[$bl]} = ($mcs => 1, ...) for the MINI-class seeds for the $bl th block that are ALREADY incorporated into a MINI-class.
	for (my $bl=0; $bl <$ct_psp_blks; $bl++) { $set_already[$bl] = {}; }


		### MAIN: Create the Main Outputs, @set_miniclasses_shift_respos  & @set_sep_patterns_for_miniclasses. ###

	my @set_miniclasses_shift_respos = my @set_sep_patterns_for_miniclasses = ();

	for (my $bl=0; $bl <$ct_psp_blks; $bl++) { # MAIN Outermost for-loop (over position-shift-pattern blocks). #

	    my $already_for_bl = $set_already[$bl];

	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]} ;

	    ##my $pspattern_le = $set_pspattern_for_blks[$bl]; # A BUG fixed on 2020/03/18.
	    my $pspattern_le = $set_pspatterns_for_blks[$bl];
	    my $set_ps_miniclass_seeds_le = $setset_ps_miniclass_seeds_for_blks[$bl];
	    my $ct_mcss_le = @{$set_ps_miniclass_seeds_le};

	    for (my $mcs = 0; $mcs < $ct_mcss_le; $mcs++) { # Main Outer for-loop (over miniclass seeds in the $bl th psp-block). #
		if (defined $already_for_bl->{$mcs}) { next; }

			# Initialize the elements of the two MAIN outputs. #

		my ($shift, $topnode_mcs, $seqs_affected, $br2rlv_change) = @{$set_ps_miniclass_seeds_le->[$mcs]};

#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node fot the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.
#

		my ($beg_site, $end_site) = ($beg_bl, $end_bl);
		my @affected_regions = ([$beg_bl, $end_bl]);
		my @seqs_affected = @{$seqs_affected};

		my $topnode_mc = $topnode_mcs;
		my %br2change = %{$br2rlv_change};
		my $ct_changes = scalar (keys %br2change);

		my @rlv_brs = keys %br2change;

		my $node2parsstat_sep_mcs = mk_node2parsstat_sep_mcs (@set_all_nodes, $topnode_mcs, %{$br2rlv_change}, %{$node2ch}, $char_abs, $char_pres);
		unless (defined $node2parsstat_sep_mcs) {
		    print STDOUT "[ERROR 'mk_set_miniclasses_shift_respos'] The subroutine, 'mk_node2parsstat_sep_mcs', FAILED: $! !\n";
		    return (undef, undef);
		}

		my %node2parsstat = %{$node2parsstat_sep_mcs};


		my @miniclass_shift_respos = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected);
		##my @sep_pattern_for_miniclass = (\%node2parsstat, \%br2change, $ct_changes; $topnode_mc); # A BUG fixed on 2020/03/18.
		my @sep_pattern_for_miniclass = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc);

		push @set_miniclasses_shift_respos, \@miniclass_shift_respos;
		push @set_sep_patterns_for_miniclasses, \@sep_pattern_for_miniclass;

#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#	@{$set_sep_patterns_for_miniclasses[$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.


			# Attempt to extend the MINI-class. #

		for (my $bl2=$bl+1; $bl2 <$ct_psp_blks; $bl2++) { # Main middle for-loop (over blocks on the right of $bl). #
	
		    my $already_for_bl2 = $set_already[$bl2];
	
		    my ($beg_bl2, $end_bl2, $len_bl2, $indx_psp2) = @{$set_pspattern_blocks->[$bl2]} ;
	
		    ##my $pspattern2 = $set_pspattern_for_blks[$bl2]; # A BUG fixed on 2020/03/18.
		    my $pspattern2 = $set_pspatterns_for_blks[$bl2];

		    my $set_ps_miniclass_seeds2 = $setset_ps_miniclass_seeds_for_blks[$bl2];
		    my $ct_mcss2 = @{$set_ps_miniclass_seeds2};

				# Examine whether or not the current block hampers the expansion of the MINI-class AT ALL. #
		    my $if_all_abs = 1;
		    foreach my $indx_seq (@seqs_affected) { # Preliminary inner foreach-loop. #
			if ($pspattern2->[$indx_seq] ne $char_abs) {
			    $if_all_abs = 0;
			    last;
			}
		    } # End of the preliminary inner foreach-loop. #

		    if ($if_all_abs) { next; } # SKIP if the current block will NOT hamper the MINI-class expansion AT ALL.


		    my $if_extended = 0;
		    for (my $mcs2 = 0; $mcs2 < $ct_mcss2; $mcs2++) { # Main inner for-loop (over miniclass seeds in the $bl2 th psp-block). #
			if (defined $already_for_bl2->{$mcs2}) { next; }
	
			my ($shift2, $topnode_mcs2, $seqs_affected2, $br2rlv_change2) = @{$set_ps_miniclass_seeds2->[$mcs2]};
			unless (($shift2 == $shift) and ($topnode_mcs2 == $topnode_mcs)) { next; }

			##my ($common, $ex1, $ex2) = diff_nmarrays (@seqs_affected, @{$seqs_affected2});
			my ($common, $ex1, $ex2) = diff_nmarray (@seqs_affected, @{$seqs_affected2});

			if ((@{$ex1}>0) or (@{$ex2}>0)) { next; }


			if (1) {
				# NOTE: This condition may be somewhat strict. #
				#       If not necessary, disable it.

			    my @rlv_brs2 = keys %{$br2rlv_change2};

			    ##my ($common, $ex1, $ex2) = diff_nmarrays (@rlv_brs, @rlv_brs2);
			    my ($common, $ex1, $ex2) = diff_nmarray (@rlv_brs, @rlv_brs2); # CORRECTED on 2020/03/19.

			    if ((@{$ex1}>0) or (@{$ex2}>0)) { next; }

			} # End of "if (1) {...}"

				# The current seed will be incorporated into the current MINI-class. #

			if ($end_site + 1 < $beg_bl2) {
					# The current block provides a new affected region. #
 			    push @affected_regions, [$beg_bl2, $end_bl2];
			} else {
					# The current block extends the last affected region. #
			    my $last_rg = $affected_regions[$#affected_regions];
			    $last_rg->[1] = $end_bl2;
			}

			$miniclass_shift_respos[2] = $end_site = $end_bl2;

			$if_extended = 1;
			$already_for_bl2->{$mcs2} = 1;
			last;

		    } # Main inner for-loop (over miniclass seeds in the $bl2 th psp-block). #


		    unless ($if_extended) { last; } # This means that $bl2 DID OBSTRUCT the extension of the current MINI-class. #

		} # End of the main middle for-loop (over blocks on the right of $bl). #

	    }  # End of the Main Outer for-loop (over miniclass seeds in the $bl th psp-block). #

	} # END of the MAIN Outermost for-loop (over position-shift-pattern blocks). #


	return (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses) ;

}	# End of the subroutine, 'mk_set_miniclasses_shift_respos'. #


# identify_trivial_miniclasses:
#	This subroutine identifies "trivial" MINI-classes out of a set of MINI-classes of position-shifts.
#	It was created by referring to a similar subroutine, 'find_trivial_blocks',
#	in "MyMalnComparison.pm" (in "../MyPerlModules_ComplLiMment/").
#	[CREATION STARTED on 2020/03/18, ENDED on 2020/03/19.]
#
#	[DEFINITION]
#	A MINI-class is regarded as "trivial" if the following two conditions are satisfied:
#	(i) the position-shift of the class is the commoner flanking position-shift;
#	(ii) the class either touches another "trivial" class (or the relevant MSA end) (and shares some affected sequences with it)
#		or separated from it by a run of gaps in at least one sequence affected by both classes.
#
# INPUTS: (\@set_miniclasses_shift_respos, \@set_pspattern_blocks, \@set_pspatterns, $char_abs,
#	 $commoner_shift_flank, $shift_lf, $shift_rf, $beg_sbj_rg, $end_sbj_rg), where
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@pcls_assoc_w_psp_blks specifies a particular set of "optimum" block-associated parsimonious clusterings,
#	more precisely, $pcls_assoc_w_psp_blks[$bl] is the index of the parsimonious clustering associated with the $bl th block;
#	[\@pcls_assoc_w_psp_blks is an element of the main output of 'select_pcls_for_psp_blks' above.]
#
#	@{$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp)
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#	[ \@set_pspattern_blocks is the main output of 'chop_psmap_into_pspblocks' above.]
#
#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence.
#	[\@set_pspatterns is an element of an input of 'list_minicls_shift_respos_clm_based' that grows during the course of the subroutine.]
#
#	$char_abs is the character to represent an "absence" state;
#
#	$commoner_shift_flank is whichever is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#	$shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
#	$shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default;
#
#	$beg_sbj_rg/$end_sbj_rg is the beginning/end of the subject region in the input MSA (in the 0-based full-closed convention);
#
# OUTPUTS: (\%miniclass2trivial), where
#
#	%miniclass2trivial = ($mc => 1, ...) for all "trivial" MINI-classes.
#
#
sub identify_trivial_miniclasses (\@\@\@$$$$$$) {

	my ($set_miniclasses_shift_respos, $set_pspattern_blocks, $set_pspatterns, $char_abs,
	 	$commoner_shift_flank, $shift_lf, $shift_rf, $beg_sbj_rg, $end_sbj_rg) = @_;

	my $ct_mcs = @{$set_miniclasses_shift_respos};
	my $ct_blks = @{$set_pspattern_blocks};


		# Preliminary 1: Create some auxiliary arrays & hashes. #

	my @pspatterns_for_blks = (); # the $bl th elemnent = the position-shift pattern for the $bl th block.
	my %beg2bl; # = ($beg_bl => $bl, ...)
	##my %end_bl; # = ($end_bl => $bl, ...)
	my %end2bl; # = ($end_bl => $bl, ...) # CORRECTED on 2020/03/19.

	for (my $bl =0; $bl <$ct_blks; $bl++) { # 1st preliminary for-loop (over psp-blocks). #

	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]} ;


	    $beg2bl{$beg_bl} = $bl;
	    $end2bl{$end_bl} = $bl;
	    $pspatterns_for_blks[$bl] = $set_pspatterns->[$indx_psp];

#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,

	} # End of the 1st preliminary for-loop (over psp-blocks). #


		# Preliminary 2: Select the candidates for "trivial" MINI-classes. #
		# (By default, the resulting set will be sorted in ascending order of their beginning-sites.)

	my @trivial_cands_std_by_beg = ();
	for (my $mc =0; $mc <$ct_mcs; $mc++) { # 2nd preliminary for-loop (over MINI-classes). #

	    my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]} ;

	    if ($shift == $commoner_shift_flank) {
		push @trivial_cands_std_by_beg, $mc;
	    }

	} # End of the 2nd preliminary for-loop (over MINI-classes). #


		# Preliminary 3: Sort the indexes of the "trivial" candidates in ascending order of their end-sites.

	my @trivial_cands_std_by_end = sort { $set_miniclasses_shift_respos->[$b][2] <=> $set_miniclasses_shift_respos->[$a][2] } @trivial_cands_std_by_beg ;


		# MAIN: Actually IDENTIFY "trivial" MINI-classes. #

	my %miniclass2trivial;

	if ($commoner_shift_flank == $shift_lf) {

			# Examine whether or not the "trivial" candidates can reach the left-end of the MSA. #

	    my @trivial_mcs_revstd_by_beg = ();

	    foreach my $mc (@trivial_cands_std_by_beg) { # 1st Outer Main foreach-loop (over candidates of "trivial" MINI-classes). #

		my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]} ;

		if ($beg_site == $beg_sbj_rg) {

			# (1) The MINI-class is on the left-end of the subject region. #

		    $miniclass2trivial{$mc} = 1;
		    unshift @trivial_mcs_revstd_by_beg, $mc;
		    next;
		}


		my $if_connected = 0;
		my $remaining_seqs_affected = $seqs_affected;

			# (2) Examine whether or not the present MINI-class EITHER touches a "trivial" MINI-class OR is mediated by a run of gaps from one. 

		foreach my $mc2 (@trivial_mcs_revstd_by_beg) { # 1st middle main foreach-loop (over "trivial" MINI-classes on the left of $mc). #
		    my ($shift2, $beg_site2, $end_site2, $affected_regions2, $seqs_affected2) = @{$set_miniclasses_shift_respos->[$mc2]} ;

		    if ($beg_site <= $end_site2) { next; } # The two MINI-classes overlap.

		    if ($beg_site == $end_site2 +1) {
			$if_connected = 1;
			last;
		    }

			# if ($end_site2 +1 < $beg_site) #

		    my ($common, $ex1, $ex2) = diff_nmarray (@{$remaining_seqs_affected}, @{$seqs_affected2});
		    unless (@{$common}>0) { next; }

		    my @intervening_blks = ();
		    for (my $beg_site3 = $end_site2+1; $beg_site3 < $beg_site; ) { # 1st inner for-loop (over intervening blocks). #
			my $bl3 = $beg2bl{$beg_site3};
			push @intervening_blks, $bl3;
			my ($beg_bl3, $end_bl3, $len_bl3, $indx_psp3) = @{$set_pspattern_blocks->[$bl3]} ;
			$beg_site3 = $end_bl3+1;

		    } # End of the 1st inner for-loop (over intervening blocks). #

		    foreach my $r (@{$common}) {  # 2nd inner for-loop (over shared MSA rows). #
			my $if_connected2 = 1;
			foreach my $bl3 (@intervening_blks) { # Innermost foreach-loop (over intervening blocks). #
			    if ($pspatterns_for_blks[$bl3]->[$r] ne $char_abs) {
				$if_connected2 = 0;
				last;
			    }
			} # End of the innermost foreach-loop (over intervening blocks). #
			if ($if_connected2) {
			    $if_connected = 1;
			    last;
			}
		    }  # End of the 2nd inner for-loop (over shared MSA rows). #

		    if ($if_connected) { last; }

		    $remaining_seqs_affected = $ex1; # We already know that rows in @{$common} contain intervening NON-gaps. #

		    if (@{$remaining_seqs_affected} == 0) { last; } # We've run out of affected sequences to examine. #

		}  # End of the 1st middle main foreach-loop (over "trivial" MINI-classes on the left of $mc). #


		if (!($if_connected) and (@{$remaining_seqs_affected}>0)) {

			# (3) Examine whether or not the present MINI-class is connected to the left-end of the subject region via a run of gaps. #

		    my @intervening_blks = ();
		    for (my $beg_site3 = $beg_sbj_rg; $beg_site3 < $beg_site; ) { # 2nd middle for-loop (over intervening blocks). #
			my $bl3 = $beg2bl{$beg_site3};
			push @intervening_blks, $bl3;
			my ($beg_bl3, $end_bl3, $len_bl3, $indx_psp3) = @{$set_pspattern_blocks->[$bl3]} ;
			$beg_site3 = $end_bl3+1;

		    } # End of the 2nd middle for-loop (over intervening blocks). #

		    foreach my $r (@{$remaining_seqs_affected}) {  # 3rd middle for-loop (over remaining "affected" sequences). #
			my $if_connected2 = 1;
			foreach my $bl3 (@intervening_blks) { # Inner foreach-loop (over intervening blocks). #
			    if ($pspatterns_for_blks[$bl3]->[$r] ne $char_abs) {
				$if_connected2 = 0;
				last;
			    }
			} # End of the inner foreach-loop (over intervening blocks). #
			if ($if_connected2) {
			    $if_connected = 1;
			    last;
			}
		    }  # End of the 3rd middle for-loop (over remaining "affected" sequences). #

		} # End of "if (!($if_connected) and (@{$remaining_seqs_affected}>0)) {...}"

		if ($if_connected) {
		    $miniclass2trivial{$mc} = 1;
		    unshift @trivial_mcs_revstd_by_beg, $mc;
		}

	    } # End of the 1st Outer Main foreach-loop (over candidates of "trivial" MINI-classes). #


	} # END of "if ($commoner_shift_flank == $shift_lf) { ... }"


	if ($commoner_shift_flank == $shift_rf) {

			# Examine whether or not the "trivial" candidates can reach the right-end of the MSA. #

	    my @trivial_mcs_revstd_by_end = ();

	    foreach my $mc (@trivial_cands_std_by_end) { # 2nd Outer Main foreach-loop (over candidates of "trivial" MINI-classes). #

		if ($miniclass2trivial{$mc}) {
			# We already know that the MINI-class is "trivial". #
		    unshift @trivial_mcs_revstd_by_end, $mc;
		    next;
		} 

		my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected) = @{$set_miniclasses_shift_respos->[$mc]} ;

		if ($end_site == $end_sbj_rg) {

			# (1) The MINI-class is on the right-end of the subject region. #

		    $miniclass2trivial{$mc} = 1;
		    unshift @trivial_mcs_revstd_by_end, $mc;
		    next;
		}


		my $if_connected = 0;
		my $remaining_seqs_affected = $seqs_affected;

			# (2) Examine whether or not the present MINI-class EITHER touches a "trivial" MINI-class OR is mediated by a run of gaps from one. 

		foreach my $mc2 (@trivial_mcs_revstd_by_end) { # 1st middle main foreach-loop (over "trivial" MINI-classes on the right of $mc). #

		    my ($shift2, $beg_site2, $end_site2, $affected_regions2, $seqs_affected2) = @{$set_miniclasses_shift_respos->[$mc2]} ;

		    if ($beg_site2 <= $end_site) { next; } # The two MINI-classes overlap.

		    if ($end_site + 1 == $beg_site2) {
			$if_connected = 1;
			last;
		    }

		### RESTART FROM HERE ###

			# if ($end_site +1 < $beg_site2) #

		    my ($common, $ex1, $ex2) = diff_nmarray (@{$remaining_seqs_affected}, @{$seqs_affected2});
		    unless (@{$common}>0) { next; }

		    my @intervening_blks = ();
		    for (my $beg_site3 = $end_site+1; $beg_site3 < $beg_site2; ) { # 1st inner for-loop (over intervening blocks). #
			my $bl3 = $beg2bl{$beg_site3};
			push @intervening_blks, $bl3;
			my ($beg_bl3, $end_bl3, $len_bl3, $indx_psp3) = @{$set_pspattern_blocks->[$bl3]} ;
			$beg_site3 = $end_bl3+1;

		    } # End of the 1st inner for-loop (over intervening blocks). #

		    foreach my $r (@{$common}) {  # 2nd inner for-loop (over shared MSA rows). #
			my $if_connected2 = 1;
			foreach my $bl3 (@intervening_blks) { # Innermost foreach-loop (over intervening blocks). #
			    if ($pspatterns_for_blks[$bl3]->[$r] ne $char_abs) {
				$if_connected2 = 0;
				last;
			    }
			} # End of the innermost foreach-loop (over intervening blocks). #
			if ($if_connected2) {
			    $if_connected = 1;
			    last;
			}
		    }  # End of the 2nd inner for-loop (over shared MSA rows). #

		    if ($if_connected) { last; }

		    $remaining_seqs_affected = $ex1; # We already know that rows in @{$common} contain intervening NON-gaps. #

		    if (@{$remaining_seqs_affected} == 0) { last; } # We've run out of affected sequences to examine. #

		}  # End of the 1st middle main foreach-loop (over "trivial" MINI-classes on the right of $mc). #


		if (!($if_connected) and (@{$remaining_seqs_affected}>0)) {

			# (3) Examine whether or not the present MINI-class is connected to the left-end of the subject region via a run of gaps. #

		    my @intervening_blks = ();
		    for (my $end_site3 = $end_sbj_rg; $end_site3 > $end_site; ) { # 2nd middle for-loop (over intervening blocks). #
			my $bl3 = $end2bl{$end_site3};
			push @intervening_blks, $bl3;
			my ($beg_bl3, $end_bl3, $len_bl3, $indx_psp3) = @{$set_pspattern_blocks->[$bl3]} ;
			$end_site3 = $beg_bl3 -1;

		    } # End of the 2nd middle for-loop (over intervening blocks). #

		    foreach my $r (@{$remaining_seqs_affected}) {  # 3rd middle for-loop (over remaining "affected" sequences). #
			my $if_connected2 = 1;
			foreach my $bl3 (@intervening_blks) { # Inner foreach-loop (over intervening blocks). #
			    if ($pspatterns_for_blks[$bl3]->[$r] ne $char_abs) {
				$if_connected2 = 0;
				last;
			    }
			} # End of the inner foreach-loop (over intervening blocks). #
			if ($if_connected2) {
			    $if_connected = 1;
			    last;
			}
		    }  # End of the 3rd middle for-loop (over remaining "affected" sequences). #

		} # End of "if (!($if_connected) and (@{$remaining_seqs_affected}>0)) {...}"

		if ($if_connected) {
		    $miniclass2trivial{$mc} = 1;
		    unshift @trivial_mcs_revstd_by_end, $mc;
		}

	    } # End of the 2nd Outer Main foreach-loop (over candidates of "trivial" MINI-classes). #


	} # END of "if ($commoner_shift_flank == $shift_rf) { ... }"


	return (\%miniclass2trivial);

}	# End of the subroutine, 'identify_trivial_miniclasses'. #


# list_minicls_shift_respos_clm_based:
#	This subroutine creates a list of MINI-classes of position-shifts
#	in a column-based manner. (See the header for brief explanation.)
#	[CREATION STARTED on 2020/03/13, ENDED on 2020/03/17.]
#
#	[REVISED on 2020/03/18-19]:
#	Now, the subroutine also outputs a set of sets of "trivial" mini-classes.
#
#	[DEBUGGED on 2020/03/20]: Corrected the ERROR on the subroutine prototype.
#
# INPUTS: (\@map_shifts_respos, $shift_lf, $shift_rf, $beg_sbj_rg, $end_sbj_rg,
#	$topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, $char_abs, $char_pres, \@set_info_pspatterns), where
#
#	$map_shifts_respos[$k]->[$s] is the shift in the residue position from the reference MSA to the reconstructed one
#	for the $s th site of the $k th sequence, in the reconstructed MSA.
#	It is $char_abs if occupied by a gap;
#
#	$shift_lf is the shift at the left-flanking column of the reconstructed MSA. It must be 0 (zero) by default;
#	$shift_rf is the shift at the right-flanking column of the reconstructed MSA. It must be (len(reconstructed) - len(reference)) by default;
#
#	[Typically, \@map_shifts_respos & $shift_lf & $shift_rf are the outputs of 'map_shifts_respos_bw_2msas' (in "MyMalnComparison.pm").]
#
#
#	$beg_sbj_rg/$end_sbj_rg is the beginning/end of the subject region in the input MSA (in the 0-based full-closed convention);
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#
#	$char_abs/$char_pres is a character representing the "absence"/"presence" state;
#
#	@set_info_pspatterns = (\%pspattern2indx, \@set_pspatterns, \@setset_psm_clusterings), where
#
#	%pspattern2indx = (join (':', @pspattern) => $index, ...), where $index is assigned in the order that the position-shift patterns appear in the map (from left to right),
#
#	@${set_pspatterns[$psp]} = @pspattern is the $psp th position-shift pattern,
#	 where $pspattern[$k] is the position-shift of the $k th sequence,
#
#	[ @{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where ] ... OBSOLETE as of 2020/03/16.
#
#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where # MODIFIED on 2020/03/16. #
#
#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs;
#
#	%br2change = ($branch_id => [$stat_ue, $stat_le], ...) for all branches along which the changes occur, where
#	$stat_ue/$stat_le is the state at the upper-end/lower-end of the branch ($branch_id);
#
#			# (ADDED on 2020/03/16) #
#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node fot the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.
#			# End of "(ADDED on 2020/03/16)" #
#
#	@set_info_pspatterns starts with the initial value, ({}, [], []),
#	and grows in the course of the process,
#
# [ OUTPUTS: (\@set_miniclasses_shift_respos, \@set_sep_patterns_for_classes, $commoner_shift_flank,
#	   \@set_pspattern_blocks), where ] ... OBSOLETE as of 2020/03/16.
#
# [ OUTPUTS: (\@setset_miniclasses_shift_respos, \@setset_sep_patterns_for_classes, $commoner_shift_flank,
#	   \@set_pspattern_blocks, \@sets_pcls_assoc_w_psp_blks), where # MODIFIED on 2020/03/16. # ] ... OBSOLETE as of 2020/03/18.
#
# OUTPUTS: (\@setset_miniclasses_shift_respos, \@setset_sep_patterns_for_classes, $commoner_shift_flank,
#	   \@set_pspattern_blocks, \@sets_pcls_assoc_w_psp_blks, \@set_miniclass2trivial), where # => Further MODIFIED on 2020/03/18. #
#
#	@{$setset_miniclasses_shift_respos[$s]->[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$setset_sep_patterns_for_miniclasses[$s]->[$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#
=OBSOLETE as of 2020/03/16. (HD1)
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#	[The following may be subject to change:]
#	@{$set_sep_patterns_for_miniclasses[$cl]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes)
#	is for the $pt th pattern separating the sequences affected by the $cl th class of position shifts, where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.
#	[END of "The following may be subject to change:"]
=cut # END of "OBSOLETE as of 2020/03/16. (HD1)"
#
#	$commoner_shift_flank is whichever is commoner between $shift_lf and $shift_rf (or $shift_lf if they tie).
#
#	[ ${$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp) ] ... WRONG!!
#	@{$set_pspattern_blocks[$bl]} = ($beg_bl, $end_bl, $len_bl, $indx_psp) # CORRECTED on 2020/03/20. #
#	 is for the $bl th position-shift-pattern block, where
#	$beg_bl/$end_bl is the starting/ending coordinate of the block (in the 0-based convention),
#	$len_bl is the length (in terms of #{columns}) of the block,
#	and $indx_psp is the index of the position-shift pattern.
#	[ \@set_pspattern_blocks is the main output of 'chop_psmap_into_pspblocks' above.]
#
#		# (ADDED on 2020/03/16.) #
#	@pcls_assoc_w_psp_blks = @{$sets_pcls_assoc_w_psp_blks[$s]} is the $s th set of
#	"optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$pcls_assoc_w_psp_blks[$bl] is the index (in @{$setset_psm_clusterings[$indx_psp]} where $indx_psp = $set_pspattern_blocks->[$bl][3]) of the parsimonious clusterings "optimum" for the $psp th position-shift-pattern block.
#	[ \@sets_pcls_assoc_w_psp_blks is the main output of 'select_pcls_for_psp_blks' above.]
#		# End of "(ADDED on 2020/03/16.)" #
#
#		# (ADDED on 2020/03/18.) #
#	%{$set_miniclass2trivial[$s]} = ($miniclass_id => 1, ...) for all "trivial" MINI-classes
#	under the $s th set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks.
#	[\%miniclass2trivial is the main output of 'identify_trivial_miniclasses' above.]
#		# End of "(ADDED on 2020/03/18.)" #
#
# [sub list_minicls_shift_respos_clm_based (\@$$$\%\%\%\@$$\@) { ] ... A BUG Fixed on 2020/03/20.
#
sub list_minicls_shift_respos_clm_based (\@$$$$$\%\%\%\@$$\@) { # => CORRECTED on 2020/03/20.

	my ($map_shifts_respos, $shift_lf, $shift_rf, $beg_sbj_rg, $end_sbj_rg,
	$topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $char_abs, $char_pres, $set_info_pspatterns) = @_;

	my $ct_seqs = @{$map_shifts_respos};
	my $minlen_row = @{$map_shifts_respos->[0]};
	for (my $k=1; $k < $ct_seqs; $k++) {
	    my $len_row = @{$map_shifts_respos->[$k]};
	    if ($len_row < $minlen_row) { $minlen_row = $len_row; }
	}

	unless ($beg_sbj_rg >= 0) {
	    print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] ILLEGAL INPUT: \$beg_sbj_rg (=$beg_sbj_rg) is NEGATIVE: $! !\n";
	    return (undef, undef, undef);
	}
	##unless ($milen_row > $end_sbj_rg) { # (Opening) A BUG fixed on 2020/03/18. 
	unless ($minlen_row > $end_sbj_rg) {
	    print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] ILLEGAL INPUT: minimum row-length (=$minlen_row) is NOT LARGER THAN \$end_sbj_rg (=$end_sbj_rg): $! !\n";
	    return (undef, undef, undef);
	}
	##} # (Closing) A BUG fixed on 2020/03/18.

   (print STDOUT "\n< Subtoutine 'list_minicls_shift_respos_clm_based' (", '$shift_lf=', "$shift_lf ; ", '$shift_rf = ', "$shift_rf ; ", '$char_abs = ', "'$char_abs'; ", '$char_pres = ', "'$char_pres'; ", '[$beg_sbj_rg, $end_sbj_rg] = ', "[$beg_sbj_rg, $end_sbj_rg]) >\n\n") if ($TEST_LEVEL_lmc>0);


		## Preliminary (1): count the flanking shifts. ##

	my $freq_shift_lf = my $freq_shift_rf = 0;

	for (my $k=0; $k<$ct_seqs; $k++) { # Preliminary outer for-loop (over sequences). #

	    my $row_shifts = $map_shifts_respos->[$k];
	    for (my $s=$beg_sbj_rg; $s<=$end_sbj_rg; $s++) { # Preliminary inner for-loop (over sites). #
		my $shift = $row_shifts->[$s];
		($shift eq $char_abs) and next;
		if ($shift == $shift_lf) { $freq_shift_lf++;
		} elsif ($shift == $shift_rf) { $freq_shift_rf++;
		}
	    }			# End of the preliminary inner for-loop (over sites). #
	}		# End of the preliminary outer for-loop (over sequences). #

	my $commoner_shift_flank = ($freq_shift_lf >= $freq_shift_rf) ? $shift_lf : $shift_rf ;

   if ($TEST_LEVEL_lmc>2) {
	print STDOUT "[ Results of preliminary process ]\n\n";
	print STDOUT "#{Sequences} = $ct_seqs ,\n";
	print STDOUT '$commoner_shift_flank = ', $commoner_shift_flank, " .\n\n\n";
   }

		# [ Preliminary (2): Create %node2row_indx. # ] ... OBSOLETE as of 2020/03/17.
		# Preliminary (2): Create %node2row_indx & @nodes_in_msa  #
		# (ADDED on 2020/03/16; MODIFIED on 2020/03/17.) #

	my %node2row_indx;
	my @nodes_in_msa = (); # Added on 2020/03/17.
	for (my $k=0; $k < $ct_seqs; $k++) { # Preliminary for-loop (over MSA rows). #
	    my $seqname = $seqnames->[$k];
	    my $node_id = $name2id->{$seqname};
	    unless (defined $node_id) {
		print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] NO NODE ID corresponds to the sequence name '$seqname' (in the $k th row of the alignment) : $! !\n";
		return (undef, undef, undef);
	    }
	    $node2row_indx{$node_id} = $k;
	    $nodes_in_msa[$k] = $node_id; # Added on 2020/03/17.

	}  # End of the Preliminary for-loop (over MSA rows). #

		# End of "(ADDED on 2020/03/16; MODIFIED on 2020/03/17.)" #


		# (1) Chop the position-shift map into blocks of contiguous columns,
		# each of which is characterized by the pattern of position-shifts the constituent columns share.

	my ($pspattern2indx, $set_pspatterns, $setset_psm_clusterings) = @{$set_info_pspatterns};
	my $ct_psps0 = @{$set_pspatterns};

	my ($set_pspattern_blocks) = chop_psmap_into_pspblocks (@{$map_shifts_respos}, $beg_sbj_rg, $end_sbj_rg, %{$pspattern2indx}, @{$set_pspatterns});
	unless (defined $set_pspattern_blocks) {
		print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] The subroutine, 'chop_psmap_into_pspblocks', FAILED: $! !\n";
		return (undef, undef, undef);
	}

	my $ct_psp_blocks = @{$set_pspattern_blocks};
	my $ct_psps = @{$set_pspatterns};

   if ($TEST_LEVEL_lmc>3) {

	##print STDOUT "\n[ Output of 'chop_psmap_into_pspblocks', i.e., contents of ", '@{$set_pspattern_blocks}', ": ]\n\n" # A BUG fixed on 2020/03/18.
	print STDOUT "\n[ Output of 'chop_psmap_into_pspblocks', i.e., contents of ", '@{$set_pspattern_blocks}', ": ]\n\n";

	print STDOUT join ("\t", 'Block_ID', 'beg_bl', 'end_bl', 'len_bl', 'Indx_psp', "'position:shift:pattern'"), "\n\n";

	for (my $bl = 0; $bl < $ct_psp_blocks; $bl++) {
	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]};
	    my $cnct_psp = join (':', @{$set_pspatterns->[$indx_psp]});

	    print STDOUT join ("\t", $bl, $beg_bl, $end_bl, $len_bl, $indx_psp, "'$cnct_psp'"), "\n";
	}

	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_lmc>3) {...}"


		# (2) If necessary, 
		# Enumerate parsimonious clusterings for the new position-shift patterns.

	if ($ct_psps > $ct_psps0) {

	    my $ct_new_psps = $ct_psps - $ct_psps0;

   (print STDOUT "\n ... The current #{position-shift-patterns} (=$ct_psps) is LARGER THAN its input value (=$ct_psps0) !!\n") if ($TEST_LEVEL_lmc >1);
   ##(print STDOUT   "     => enumerate parsimonious clusterings for the $ct_new_psps new position-shift patterns ...\n\n") if ($TEST_LEVEL_lcm > 1); # A BUG fixed on 2020/03/18.
   (print STDOUT   "     => enumerate parsimonious clusterings for the $ct_new_psps new position-shift patterns ...\n\n") if ($TEST_LEVEL_lmc > 1);

	    ## unless ( enum_psm_clusterings_for_new_psps (@{$set_pspatterns}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, $char_abs, $char_pres) ) { 
	    unless ( enum_psm_clusterings_for_new_psps (@{$set_pspatterns}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, %node2row_indx, $char_abs, $char_pres) ) { # MODIFIED on 2020/03/16.
		print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] The subroutine, 'enum_psm_clusterings_for_new_psps', FAILED: $! !\n";
		return (undef, undef, undef);
	    }

	} # END of "if ($ct_psps > $ct_psps0) {...}"


#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where # MODIFIED on 2020/03/16. #
#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node for the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.


   if ($TEST_LEVEL_lmc>=5) {

	print STDOUT "\n[[ Relevant Contents of ", '@{$setset_psm_clusterings}', ": ]]\n\n";

	my %already;
	for (my $bl = 0; $bl < $ct_psp_blocks; $bl++) { # Outer testing for-loop (over position-shift-pattern blocks). #

	    my ($beg_bl, $end_bl, $len_bl, $indx_psp) = @{$set_pspattern_blocks->[$bl]};


	    print STDOUT "{{ For the $bl th Position-Shift-Pattern Block (w/ [beg, end] = [$beg_bl, $end_bl]; indx_psp = $indx_psp) }}\n\n";

	    my $prev_bl = $already{$indx_psp};

	    if (defined $prev_bl) {
		print STDOUT " ... The position-shift pattern is the same as that for the $prev_bl th block !!\n\n";
		next;
	    }
	    $already{$indx_psp} = $bl;

	    my $set_psm_clusterings = $setset_psm_clusterings->[$indx_psp];
	    my $ct_pcls = @{$set_psm_clusterings};

	    print STDOUT "#{Ways of parsimonious clusterings} = $ct_pcls .\n\n\n";

	    for (my $pcl = 0; $pcl < $ct_pcls; $pcl++) { # Middle testing for-loop (over parsiminous clusterings). #

		print STDOUT "{ For the $pcl th parsimonious clustering: }\n\n\n";

		##my ($node2parsstat, $br2change) = @{$set_psm_clusterings->[$pcl]};
		my ($node2parsstat, $br2change, $set_ps_miniclass_seeds) = @{$set_psm_clusterings->[$pcl]}; # MODIFIED on 2020/03/16.

#	@{$setset_psm_clusterings[$psp]->[$pcl]} = (\%node2parsstat, \%br2change, \@set_ps_miniclass_seeds) is for the $pcl th parsimonious clustering of the $psp th position-shift pattern, where # MODIFIED on 2020/03/16. #

#	%node2parsstat = ($node_id => \@state_info), where
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs.

		print STDOUT '[ Contents of %{$node2parsstat}: ]', "\n\n";

		print STDOUT join ("\t", 'Node_ID', 'pos-shift', 'sum_penal'), "\n\n";


	 	foreach my $node (sort {$a <=> $b} keys %{$node2parsstat}) {
			print STDOUT join ("\t", $node, @{$node2parsstat->{$node}}), "\n";
		}
		print STDOUT "\n\n\n";

		my $ct_changes = scalar (keys %{$br2change});

		print STDOUT '[ Contents of %{$br2change} ', "(w/ #{changes} = $ct_changes): ]\n\n";

		print STDOUT join ("\t", 'Br_ID', 'pos-shift_ue', 'pos-shift_le'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2change}) {
			print STDOUT join ("\t", $br, @{$br2change->{$br}}), "\n";
		}
		print STDOUT "\n\n\n";

		my $ct_mcss = @{$set_ps_miniclass_seeds};

		print STDOUT '[ Contents of @{$set_ps_miniclass_seeds} ', "(w/ #{seeds} = $ct_mcss): ]\n\n";

		print STDOUT join ("\t", 'Seed_ID', 'shift', 'topnode_mcs', '(affected,MSA,row,IDs)', '#{relevant_changes}', '{br_ID:state_ue,state_le}_1', '{br_ID:state_ue,state_le}_2', '...'), "\n\n";

		for (my $mcs =0; $mcs <$ct_mcss; $mcs++) {
		
		    my ($shift, $topnode_mcs, $seqs_affected, $br2rlv_change) = @{$set_ps_miniclass_seeds->[$mcs]} ;

		    my $cnct_row_ids = join (',', @{$seqs_affected});
		    my $ct_rlv_chngs = scalar keys %{$br2rlv_change};
		    my @out_printed = ($shift, $topnode_mcs, "($cnct_row_ids)", $ct_rlv_chngs);
		    foreach my $br (sort {$a <=> $b} keys %{$br2rlv_change}) {
			my ($state_ue, $state_le) = @{$br2rlv_change->{$br}};
			push @out_printed, "$br\:$state_ue\,$state_le";
		    }

		    print STDOUT join ("\t", $mcs, @out_printed), "\n";
		}
		print STDOUT "\n\n\n";

#	@{$set_ps_miniclass_seeds[$mcs]} = ($shift, $topnode_mcs, \@seqs_affected, \%br2rlv_change)
#	 is for the $mcs th seed of a position-shift miniclass, where
#	$shift is the position-shift characterizing the miniclass seed,
#	$topnode_mcs is the top-node fot the miniclass seed,
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts, and
#	%br2rlv_change = ($branch_id => [$state_upper_end, $state_lower_end], ...) ONLY for the branches that accommodate changes relevant to this miniclass seed.


	    } # End of the Middle testing for-loop (over parsimonious clusterings). #

	} # End of the Outer testing for-loop (over position-shift-pattern blocks). #

   } # End of "if ($TEST_LEVEL_lmc>=5) {...}"


		# (3) Select the "optimum" combination(s) of
		#   parsimonious clusterings for the position-shift-pattern blocks. #

	my ($sets_pscls_assoc_w_psp_blks) = select_pcls_for_psp_blks (@{$set_pspattern_blocks}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, %{$node2pa}, $char_abs);
	unless (defined $sets_pscls_assoc_w_psp_blks) {
	    print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] The subroutine, 'select_pcls_for_psp_blks', FAILED: $! !\n";
	    return (undef, undef, undef);
	}

	my $ct_sets_pcls_for_blks = @{$sets_pscls_assoc_w_psp_blks};

   if ($TEST_LEVEL_lmc >3) {

	print STDOUT "\n[[ Output of 'select_pcls_for_psp_blks': Contents of ", '@{$sets_pscls_assoc_w_psp_blks)}', " (#{sets} = $ct_sets_pcls_for_blks) ]]\n\n";

	my @indices_psps_for_blks = ();
	##for (my $bl = 0; $bl < $ct_blks; $bl++) { $indices_psps_for_blks[$bl] = $set_pspattern_blocks->[$bl][3]; } # A BUG fixed on 2020/03/18.
	for (my $bl = 0; $bl < $ct_psp_blocks; $bl++) { $indices_psps_for_blks[$bl] = $set_pspattern_blocks->[$bl][3]; }
		# my $ct_psp_blocks = @{$set_pspattern_blocks};

	print STDOUT "[ For reference, indices of the position-shift-patterns for the blocks are:\n\n";
	print STDOUT " (", join (', ', @indices_psps_for_blks), ") . ]\n\n";

	print STDOUT join ("\t", 'Set_no.', 'indx_pcl(bl=0)', 'indx_pcl(bl=1)', '...'), "\n\n";

	for (my $s = 0; $s < $ct_sets_pcls_for_blks; $s++) {
		print STDOUT join ("\t", $s, @{$sets_pscls_assoc_w_psp_blks->[$s]}), "\n";
	}
#
#	@pcls_assoc_w_psp_blks = @{$sets_pcls_assoc_w_psp_blks[$s]} is the $s th set of
#	"optimum" parsimonious clusterings associated with the position-shift-pattern blocks, where
#	$pcls_assoc_w_psp_blks[$bl] is the index (in @{$setset_psm_clusterings[$indx_psp]} where $indx_psp = $set_pspattern_blocks->[$bl][3]) of the parsimonious clusterings "optimum" for the $psp th position-shift-pattern block.


	print STDOUT "\n\n";

   }  # End of "if ($TEST_LEVEL_lmc >3) {...}"



		# (3+1) For each "optimum" combination of
		#  parsimonious clusterings for the position-shift-pattern blocks,
		#  create a set of MINI-classes of position-shifts,
		#  and provide each MINI-class with separating pattern.
		# [ADDED on 2020/03/18] Then, identify "trivial" MINI-classes.

   ##(print STDOUT "\n{{  MAIN PROCESS (3+1): For each \"optimum\" combination of parsimonious clusterings, Create a set of MINI-classes of position-shifts }}\n\n") if ($TEST_LEVEL_lmc >0);
   (print STDOUT "\n{{  MAIN PROCESS (3+1): For each \"optimum\" combination of parsimonious clusterings, Create a set of MINI-classes of position-shifts, then, identify \"trivial\" MINI-classes }}\n\n") if ($TEST_LEVEL_lmc >0); # Modified in REVISION on 2020/03/18.

	my @setset_miniclasses_shift_respos = ();
	my @setset_sep_patterns_for_miniclasses = ();

	##my @setset_trivial_miniclasses = (); # ADDED in REVISION on 2020/03/18.
	my @set_miniclass2trivial = (); # => Further REVISED on 2020/03/19.

	# my $ct_sets_pcls_for_blks = @{$sets_pscls_assoc_w_psp_blks}; # Same command already exists above.

	for (my $s = 0; $s <$ct_sets_pcls_for_blks; $s++) { # Main Outer for-loop (over "optimum" combinations of parsimonious clusterings associated with psp-blocks). #

	    my $pscls_assoc_w_psp_blks = $sets_pscls_assoc_w_psp_blks->[$s];

   (print STDOUT "{ For the $s th \"optimum\" combination of parsimonious clusterings }\n\n") if ($TEST_LEVEL_lmc>1);


#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th MINI-class of the run of shifts of residue positions 
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#	@{$set_sep_patterns_for_miniclasses[$mc]} = (\%node2parsstat, \%br2change, $ct_changes, $topnode_mc)
#	is for the pattern separating the sequences affected by the $mc th MINI-class of the run of shifts of residue positions
#	(under the particular set of "optimum" parsimonious clusterings associated with the position-shift-pattern blocks), where
#	%node2parsstat = ($node_id => [$state], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest.

	    my ($set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses)
		## [  = mk_set_miniclasses_shift_respos (@{$pcls_assoc_w_psp_blks}, @{$set_pspattern_blocks}, @{$set_pspatterns}, @{$setset_psm_clusterings}, $topnode_id, $char_abs); ] ... OBSOLETE as of 2020/03/17.
		##= mk_set_miniclasses_shift_respos (@{$pcls_assoc_w_psp_blks}, @{$set_pspattern_blocks}, @{$set_pspatterns}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, @nodes_in_msa, $char_abs, $char_pres); # REVISED on 2020/03/17.
		= mk_set_miniclasses_shift_respos (@{$pscls_assoc_w_psp_blks}, @{$set_pspattern_blocks}, @{$set_pspatterns}, @{$setset_psm_clusterings}, $topnode_id, %{$node2ch}, @nodes_in_msa, $char_abs, $char_pres); # => DEBUGGED on 2020/03/18.
	    unless (defined $set_sep_patterns_for_miniclasses) {
		print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] The subroutine, 'mk_set_miniclasses_shift_respos', FAILED: $! !\n";
		return (undef, undef, undef);
	    }

	    $setset_miniclasses_shift_respos[$s]  = $set_miniclasses_shift_respos;
	    $setset_sep_patterns_for_miniclasses[$s] = $set_sep_patterns_for_miniclasses;


			# (ADDED in REVISION on 2020/03/18.) #

	    my ($miniclass2trivial) = identify_trivial_miniclasses (@{$set_miniclasses_shift_respos}, @{$set_pspattern_blocks}, @{$set_pspatterns}, $char_abs, $commoner_shift_flank, $shift_lf, $shift_rf, $beg_sbj_rg, $end_sbj_rg) ;
	    unless (defined $miniclass2trivial) {
		print STDOUT "[ERROR 'list_minicls_shift_respos_clm_based'] The subroutine, 'identify_trivial_miniclasses', FAILED: $! !\n";
		return (undef, undef, undef);
	    }


	    $set_miniclass2trivial[$s] = $miniclass2trivial;

			# End of "(ADDED in REVISION on 2020/03/18.)" #

	}  # Main Outer for-loop (over "optimum" combinations of parsimonious clusterings associated with psp-blocks). #


			# TWO BUGS fixed on 2020/03/18. #
	#### return (\@setset_miniclasses_shift_respos, \@setset_sep_patterns_for_classes, $commoner_shift_flank,
	####   $set_pspattern_blocks, $sets_pcls_assoc_w_psp_blks);
	##return (\@setset_miniclasses_shift_respos, \@setset_sep_patterns_for_miniclasses, $commoner_shift_flank,
	##   $set_pspattern_blocks, $sets_pscls_assoc_w_psp_blks);
			# End of "TWO BUGS fixed on 2020/03/18." #
	return (\@setset_miniclasses_shift_respos, \@setset_sep_patterns_for_miniclasses, $commoner_shift_flank,
	   $set_pspattern_blocks, $sets_pscls_assoc_w_psp_blks, \@set_miniclass2trivial); # => REVISED on 2020/03/18.

}	# End of the subroutine, 'list_minicls_shift_respos_clm_based'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able # (ADDED on 2020/03/10.)!! #


