package MyTreeMap_indels_sppl;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_indels_sppl.pm
#	This package is a collection of subroutines
#	that supplement the functions of the package, "MyTreeMap_indels.pm".
#
#	[ CREATION STARTED on Dec 13 (Tue), 2016, ENDED on Mmm dd (Www), yyyy. ]
#


use strict;
use warnings;

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

=Set aside for later use

use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray union_nmarray );

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

=cut # END of "Set aside for later use"


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	counts_indels
	counts_totlens_indels
	create_br2eqbr
	create_br2nn
	replace_w_equiv_indels
	replace_w_smaller_equiv_indels
	mk_eventset
	);


	# This specifies the level of tests on the subroutine, 
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###


# counts_indels:
#	This subroutine counts insertions and deletions in %br2indels.
#
#	[CREATED on May 10 (Wed), 2017.]
#
# INPUT: (\%br2indels, $char_abs), where
#
#	%br2indels = ($branc_id => \@indels, ...) is the map of the indel events onto branches,
#	with @{$indels[$ev]} = ($state_ue, $state_le, \@invlvd_gpbs).
#
# OUTPUT: ($ct_ins, $ct_del), where
#	$ct_ins/del is the number of insertions/deletions.
#
sub counts_indels (\%$) {

	my ($br2indels, $char_abs) = @_;

	my $ct_ins = my $ct_del = 0;

	foreach my $br (keys %{$br2indels})  { # Main foreach-loop (over branches). #

	    my $indels = $br2indels->{$br};
	    (defined $indels) or next;

	    my $ct_events = @{$indels};

	    for (my $ev = 0; $ev < $ct_events; $ev++) { # Inner for-loop (over events along $br). #

		my ($state_ue, $state_le, $invlvd_gpbs) = @{$indels->[$ev]};

		if ($state_ue eq $char_abs) { # An insertion. #
		    $ct_ins++;

		} else { # A deletion. #
		    $ct_del++;
		}
	    }					# End of the inner for-loop (over events along $br). #

	}				# End of the main foreach-loop (over branches). #

	return ($ct_ins, $ct_del);

}	# End of the subroutine 'counts_indels'. #


# counts_totlens_indels:
#	This subroutine counts the numbers of insertions and deletions
#	in a set of indels. It also calculates the total lengths of indels.
#
#	[ CREATED on Dec 14 (Wed), 2016. ]
#
# INPUT: (\%br2indels, $char_abs, \@blks_in_cls, $indx_blk_len), where
#
#	%br2indels = ($branch_id => \@list_indels, ...) for the initial set of indels, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices of blocks (in @blks_in_cls) involved in the event;
#	$sbj_br and $eq_br are the IDs of the subject branch and the equivalent branch, respectively;
#
#	$char_abs is the character representing the absence of the site;
#
#	@blks_in_cls is a list of blocks in the relevant cluster (i.e., gapped segment),
#	with $blks_in_cls[$bl]->[$indx_blk_len] giving the length of the $bl th block.
#	($indx_blk_len = undef if the length information is not directly available.)
#
# OUTPUT: ($ct_ins, $ct_del, $totlen_ins, $totlen_del), where
#
#	$ct_ins and $ct_del are the numbers of insertions and deletions, respectively;
#
#	$totlen_ins and $totlen_del are the total lengths of insertions and deletions, respectively.
#	(They are undef if $indx_blk_len = undef.)
#
sub counts_totlens_indels (\%$\@$) {

	my ($br2indels, $char_abs, $blks_in_cls, $indx_blk_len) = @_;

	my $ct_ins = my $ct_del = my $totlen_ins = my $totlen_del = 0;

	(defined $indx_blk_len) or ($totlen_ins = $totlen_del = undef);

	foreach my $br (keys %{$br2indels}) { # Main foreach-loop (over branches). #

	    my $list_indels = $br2indels->{$br};
	    (defined $list_indels) or next; # ADDED as an error-proof measure (on May 10, 2017). #

	    my $ct_events = @{$list_indels};

	    for (my $ev=0; $ev<$ct_events; $ev++) { # Inner for-loop (over events). #

		my ($stat_pa, $stat_node, $indices_blk) = @{$list_indels->[$ev]};

		my $len_indel = 0;
		if (defined $indx_blk_len) {
		    foreach my $indx_blk (@{$indices_blk}) { $len_indel += $blks_in_cls->[$indx_blk][$indx_blk_len]; }
		} else {
		    $len_indel = undef;
		}

		if ($stat_pa eq $char_abs) { # Insertion. 
		##if ($stat_pa = $char_abs) { # Insertion.  # A BUG fixed on May 10, 2017. #

		    $ct_ins++;
		    (defined $len_indel) and ($totlen_ins += $len_indel);

		} else {		    # Deletion.

		    $ct_del++;
		    (defined $len_indel) and ($totlen_del += $len_indel);
		}

	    }				# End of the inner for-loop (over events). #

	}			# End of the main foreach-loop (over branches). #

	return ($ct_ins, $ct_del, $totlen_ins, $totlen_del);

}	# End of the subroutine 'counts_totlens_indels'. #


# create_br2eqbr:
#	This subroutine creates a hash, %br2eq_br,
#	which associates a branch with its "equivalent" branch.
#	The pair of mutually equivalent branches occur only when
#	the top-node (i.e., root) of the tree has only two child nodes.
#
#	[ CREATED on December 13 (Tue), 2016. ]
#
# INPUT: ($topnode_id, \%node2ch), where
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, where
#	\@children stores the IDs of the child nodes of $node_id.
#
# OUTPUT: \%br2eq_br, where
#	%br2eq_br = ($br_id => $equiv_br_id, ...) for the branches with their equivalents. 
#
sub create_br2eqbr ($\%) {

	my ($topnode_id, $node2ch) = @_; 

	my $children_top = $node2ch->{$topnode_id};

	unless (@{$children_top} == 2) {
	    return {};
	}

	my ($ch1, $ch2) = @{$children_top};

	my %br2eq_br = ($ch1 => $ch2, $ch2 => $ch1);

	return \%br2eq_br;

}	# END of the subroutine 'create_br2eqbr'. #


# create_br2nn:
#	This subroutine creates a hash, %br2nn,
#	which associates each branch with a set of its 'nearest neighboring' branches.
#
#	[ CREATED on Dec. 16 (Fri), 2016.]
#
# INPUT: (\%br_attr, \%node2ch, \%node2pa, \%br2eq_br), where
#
#	%br_attr = ($branch_id => \@attributes, ...) for all branches;
#	%node2ch = ($node_id => \@child_ids, ...) for all internal branches;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but root;
#	%br2eq_br = ($branch_id => $equiv_br_id, ...) for branches with their equivalents.
#
# OUTPUT: \%br2nn, where
#
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...).
#
sub create_br2nn (\%\%\%\%) {

	my ($br_attr, $node2ch, $node2pa, $br2eq_br) = @_;

	my %br2nn;

	foreach my $br (keys %{$br_attr}) { # Main foreach-loop (over branches). #

	    my %nn; # = ($id_nearest_neighbor => 1, ...)
	    $br2nn{$br} = \%nn;

	    my $children = $node2ch->{$br};
	    if (defined $children) {
	    	foreach my $ch (@{$children}) { $nn{$ch} = 1; }
	    }

	    my $eq_br = $br2eq_br->{$br};
	    if (defined $eq_br) { # The upper-end is the top node.
	    	my $children_eq = $node2ch->{$eq_br};
	    	if (defined $children_eq) {
	    	    foreach my $ch (@{$children_eq}) { $nn{$ch} = 1; }
	    	}
	    	next;
	    }

	    my $pa = $node2pa->{$br};
	    if (defined $br_attr->{$pa}) {
	    	$nn{$pa} = 1;
	    	my $eq_pa = $br2eq_br->{$pa};
	    	if (defined $eq_pa) { $nn{$eq_pa} = 1; }
	    }
	    my $sibls = $node2ch->{$pa};
	    foreach my $sbl (@{$sibls}) { ($sbl == $br) or ($nn{$sbl} =1); }

	}			# End of the main foreach-loop (over branches). #

	return \%br2nn;

}		# END of the subroutiner 'create_br2nn'. #


# replace_w_equiv_indels:
#	This subroutine works only on the child branches of a top-node with only two children.
#	The subroutine replaces the set of indels on one child branch
#	with the set of "equivalent indels" on the equivalent branch.
#
#	[ CREATED on December 13 (Tue), 2016. ]
#
# [ INPUT: (\%br2indels, $sbj_br, $eq_br, $char_pres, $char_abs), where ] --- OBSOLETE
# INPUT: (\%br2indels, $sbj_br, $eq_br), where
#
#	%br2indels = ($branch_id => \@list_indels, ...) for the initial set of indels, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices of blocks involved in the event;
#	$sbj_br and $eq_br are the IDs of the subject branch and the equivalent branch, respectively;
#
#	[ $char_pres and $char_abs are the characters representing the presence and the absence of the sub-sequence, respectively. ] --- They turned out to be unnecessary while creating the subroutine.
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 otherwise.
#
##sub replace_w_equiv_indels (\%$$$$) {
##	my ($br2indels, $sbj_br, $eq_br, $char_pres, $char_abs) = @_;
#
sub replace_w_equiv_indels (\%$$) {

	my ($br2indels, $sbj_br, $eq_br) = @_;

	my $list_indels_sbj = $br2indels->{$sbj_br};

	unless (defined $list_indels_sbj) {
	    return 1; # There is nothing to do.
	}

	delete $br2indels->{$sbj_br};

	unless (@{$list_indels_sbj}>0) {
	    return 1; # There is nothing else to do.
	}

	my $list_indels_eq = $br2indels->{$eq_br};
	(defined $list_indels_eq) or ($list_indels_eq = $br2indels->{$eq_br} = []);

	foreach my $indel_sbj (@{$list_indels_sbj}) {

	    my ($stat_pa, $stat_node, $indices_blk) = @{$indel_sbj};

	    my @indel_eq = ($stat_node, $stat_pa, $indices_blk) ;

	    push @{$list_indels_eq}, \@indel_eq;
	}

	return 1;

}	# END of the subroutine 'replace_w_equiv_indels'. #


# replace_w_smaller_equiv_indels:
#	When the top-node (i.e., root) of the tree has only two children,
#	the subroutine replaces the set of indels on the child branch with more external offsprings
#	with the set of "equivalent indels" on the equivalent branch (with less external offsprings).
#
#	[ CREATED on December 14 (Wed), 2016. ]
#
# INPUT: (\%br2indels, \%br2eq_br, \%node2ct_exoffs), where
#
#	%br2indels = ($branch_id => \@list_indels, ...) for the initial set of indels, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices of blocks involved in the event;
#	$sbj_br and $eq_br are the IDs of the subject branch and the equivalent branch, respectively;
#
#	%br2eq_br = ($branch_id => $equivalent_branch_id, ...);
#
#	%node2ct_exoffs = ($node_id => $ct_external_offsprings, ...);
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 otherwise.
#
sub replace_w_smaller_equiv_indels (\%\%\%) {

	my ($br2indels, $br2eq_br, $node2ct_exoffs) = @_;

	my ($sbj_br, $eq_br);

	foreach my $br (keys %{$br2eq_br}) { # Main foreach-loop (over branches with equivalents). #

	    my $equiv = $br2eq_br->{$br};

	    ($node2ct_exoffs->{$br} > $node2ct_exoffs->{$equiv}) or next;

	    $sbj_br = $br;
	    $eq_br = $equiv;

	    last;

	}		# End of the main foreach-loop (over branches with equivalents). #

	(defined $sbj_br) or (return 1); # Do nothing if there is no branch with its equivalent. #

	unless ( replace_w_equiv_indels (%{$br2indels}, $sbj_br, $eq_br) ) {

	    print STDERR "[ERROR 'replace_w_smaller_equiv_indels'] The subroutine, 'replace_w_equiv_indels', FAILED: $! !\n\n";

	    return 0;

	}

	return 1;

}	# END of the subroutine 'replace_w_smaller_equiv_indels'. #


# mk_eventset:
#	This subroutine creates a set of indel events in a gapped segment,
#	using the output of the subroutine, 'extract_cls_globks', in "MyTreeMap_indels(_xxx).pm",
#	and a hash associating branches with the indel events along them.
#
#	[CREATED on Dec 16 (Fri), 2016.]
#
# INPUT: (\%br2indels, \@gblks_in_gseg, $indx_gb_start, $indx_gb_end, $indx_gb_len), where
#
#	%br2indels = ($branch_id => \@list_indels, ...) is an output of e.g., 'create_initial_set_indels(_xxx)', with
#	@{$list_indels[$id]} = ($stat_ue, $stat_le, \@indices_blk), which are for the $id th indel event along $branch_id. $stat_pa and $stat_node are the states on the upper-end node and the lower-end node of the branch, and @indices_blk stores the indices of involved blocks in @gblks_in_gseg;

#	@gblks_in_gseg is an element of an array, @set_cls_gblocks, which is an output of the subroutine, 'extract_cls_gblocks'.
#	And ($gb_start, $gb_end, $gb_len) = @{$gblks_in_gseg[$gb]}[$indx_gb_start, $indx_gb_end, $indx_gb_len]
#	is the start coordinate, end coordinate, and length, respectively,
#	of the $gb th gap-pattern block in the gapped segment.
#	(If $indx_gb_len = undef, the block length will be calculated as: $gb_len = $gb_end - $gb_start +1.)
#
# OUTPUT: (\@eventset), where
#
#	@{$eventset[$ev]} = ($branch_id, $stat_ue, $start_ev, $end_ev, $len_ev) is for the $ev th event.
#	$start_ev, $end_ev, and $len_ev are the start coordinate, end coordinate, and length, respectively, of the inserted/deleted subsequence.
#
#
sub mk_eventset (\%\@$$$) {

	my ($br2indels, $gblks_in_gseg, $indx_gb_start, $indx_gb_end, $indx_gb_len) = @_;

	my @eventset = ();

	foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # Outer foreach-loop (over branches w/ indels). #

	    my $list_indels = $br2indels->{$br};

	    foreach my $indel (@{$list_indels}) { # Middle foreach-loop (over indels along $br). #

		my ($stat_ue, $stat_le, $indices_blk) = @{$indel};

		my ($lm_blk, $rm_blk) = @{$indices_blk}[0, $#{$indices_blk}];
		my $start_ev = $gblks_in_gseg->[$lm_blk][$indx_gb_start];
		my $end_ev   = $gblks_in_gseg->[$rm_blk][$indx_gb_end];

		my $len_ev = 0;

		foreach my $indx_gb (@{$indices_blk}) { # Inner foreach-loop (over gap-pattern blocks involved). #

		    my $info_gb = $gblks_in_gseg->[$indx_gb];

		    if (defined $indx_gb_len) {

			$len_ev += $info_gb->[$indx_gb_len];
			
		    } else {
			my ($start_gb, $end_gb) = @{$info_gb}[$indx_gb_start, $indx_gb_end];
			$len_ev += $end_gb - $start_gb + 1;
		    }
		}				# End of the inner foreach-loop (over gap-pattern blocks involved). #

		push @eventset, [$br, $stat_ue, $start_ev, $end_ev, $len_ev];

	    }				# End of the middle foreach-loop (over indels along $br). #

	}					# End of the outer foreach-loop (over branches w/ indels). #

	return \@eventset;

}	# END of the subroutine 'mk_eventset'. #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


