package MyTreeMap_sppl;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_sppl.pm:
#	This package is a collection of subroutines for some tree mapping
#	that may be handy in some applications.
#	[CREATION STARTED on 2019/06/18, ENDED on 2019/mm/dd.]
#
#	[MAJOR REVISION on 2019/06/27]:
#	MOVED the subroutine 'associate_nodeids_w_msa_rows' from "MyANEX_Supple.pm" to "MyTreeMap_sppl.pm".
#

use strict;
use warnings;


use MyArrayManipu_num qw( diff_nmarray union_nmarray ); # Enabled on 2019/06/26.

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth
	);
	# make_node2ct_exoffs

=DISABLED in MAJOR REVISION (on 2019/06/27).
use MyANEX_Supple qw(
	associate_nodeids_w_msa_rows
	);
=cut # END of "DISABLED in MAJOR REVISION (on 2019/06/27)."


		# ENABLED in creation of 'vsplit_into_eff_child_gblocks' (on 2019/08/21). #
use MyTaxonomy qw(
	fetch_ext_offsprings_nr
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ancestry 
	## fetch_last_common_ancestor_ov2
	## fetch_common_ancestors
		# END of "ENABLED in creation of 'vsplit_into_eff_child_gblocks' (on 2019/08/21)." #


=Set aside for later use

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

=cut # END of "Set aside for later use"


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	associate_nodeids_w_msa_rows
	mk_node2ex_offsprings
	mk_node2exoffs_msa_rows
	wrapper_mk_node2exoffs_msa_rows
	mk_set_msa_rows2br_up_or_lw
	mk_set_msa_rows2br_up_or_lw_dp
	vsplit_into_eff_child_gblocks
	vsplit_into_cmpl_eff_child_rblocks
	);
	# ADDED 'associate_nodeids_w_msa_rows' in MAJOR REVISION (on 2019/06/27). #
	# ADDED 'mk_set_msa_rows2br_up_or_lw_dp' on 2019/07/03.
	# ADDED 'vsplit_into_eff_child_gblocks' on 2019/08/21.
	# ADDED 'vsplit_into_cmpl_eff_child_rblocks' on 2019/08/25.


	# This specifies the level of tests on the subroutine, 
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# associate_nodeids_w_msa_rows:
#	This subroutine associates the IDs of the (usually external) nodes (of a tree)
#	with the rows of the MSA (associated with the tree).
#	[CREATED on 2019/02/26 (in "MyANEX_Supple.pm").]
#	[=> MOVED from "MyANEX_Supple.pm" to "MyTreeMap_sppl.pm" (on 2019/06/27).]
#
# INPUTS: (@seqnames, \%name2id), where
#
#	$seqnames[$k] is the name (or ID) of the sequence at the $k th row of the MSA;
#	%name2id = ($node_name  => $node_ID, ...) for nodes with names,
#	where $node_name corresponds to $seqnames[$k] for external nodes;
#
# OUTPUTS: (\@nodeids_in_msa, \%nodeid2msa_row), where
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $now_ID.
#
sub associate_nodeids_w_msa_rows (\@\%) {

	my ($seqnames, $name2id) = @_;

	my $ct_rows = @{$seqnames};

	my @nodeids_in_msa = ();
	my %nodeid2msa_row;

	for (my $row = 0; $row < $ct_rows; $row++) { # Main for-loop (over MSA rows).

	    my $name = $seqnames->[$row];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'associate_nodeids_w_msa_rows'] The $row th row (w/ name = '$name') does NOT correspond to ANY node in the tree: $! !!\n\n";
		return (undef, undef);
	    }

	    $nodeids_in_msa[$row] = $node_id;
	    $nodeid2msa_row{$node_id} = $row;

	} # End of the main for-loop (over MSA rows).


	return (\@nodeids_in_msa, \%nodeid2msa_row);

}	# End of the subroutine 'associate_nodeids_w_msa_rows'. #


# mk_node2ex_offsprings:
#	This subroutine creates, in a bottom-up manner, a hash that associates
#	each node (of a tree) to the list of its external offsprings.
#	[CREATED on 2019/06/18.]
#
# INPUTS: (\%node2ch, \%node2depth), where
#
#	%node2ch = ($node_id => \@child_node_ids, ...) for all internal nodes;
#
#	%node2depth = ($node_id => $abs_depth, ...) for all nodes;
#	[Typically, \%node2depth is an output of 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
# OUTPUT: (\%node2exoffs), where
#
#	%node2exoffs = ($node_id => \@external_offspring_ids, ...) for all nodes.
#
sub mk_node2ex_offsprings (\%\%) {

	my ($node2ch, $node2depth) = @_;

	my %node2exoffs;

	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a}} keys %{$node2depth}) { # Main foreach-loop (over all nodes). #

	    my $children = $node2ch->{$node};
	    unless ((defined $children) and (@{$children}>0)) {
		# $node is an external node.
		$node2exoffs{$node} = [$node];
		next;
	    }

		# $node is an internal node.
	    my @exoffs = ();
	    foreach my $ch (@{$children}) {
		push @exoffs, @{$node2exoffs{$ch}};
	    }
	    my @std_exoffs = sort {$a <=> $b} @exoffs;

	   $node2exoffs{$node} = \@std_exoffs;

	} # End of the main foreach-loop (over all nodes). #

	return \%node2exoffs;

}	# End of the subroutine, 'mk_node2ex_offsprings'. #


# mk_node2exoffs_msa_rows:
#	This subroutine creates a hash that associates each node with a list of MSA rows
#	corresponding to the external offsprings of the node.
#	It was created by slightly modifying the subroutine, 'mk_node2ex_offsprings', above.
#	[CREATED on 2019/06/18.]
#
# INPUTS: (\%node2ch, \%node2depth, \%nodeid2msa_row), where
#
#	%node2ch = ($node_id => \@child_node_ids, ...) for all internal nodes;
#
#	%node2depth = ($node_id => $abs_depth, ...) for all nodes;
#	[Typically, \%node2depth is an output of 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $now_ID.
#	[ [Typically, \%nodeid2msa_row is an output of 'associate_nodeids_w_msa_rows' in "MyANEX_Supple.pm"]; ] ... OBSOLETE as of 2019/06/27.
#	[Typically, \%nodeid2msa_row is an output of 'associate_nodeids_w_msa_rows' above]; # Revised on 2019/06/27.
#
# OUTPUT (\%node2exoffs_msa_rows), where
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#
sub mk_node2exoffs_msa_rows (\%\%\%) {

	my ($node2ch, $node2depth, $nodeid2msa_row) = @_;

	my %node2exoffs_msa_rows;

	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a}} keys %{$node2depth}) { # Main foreach-loop (over all nodes). #

	    my $children = $node2ch->{$node};
	    unless ((defined $children) and (@{$children}>0)) {
		# $node is an external node.
		my $msa_row = $nodeid2msa_row->{$node};
		unless (defined $msa_row) {
		    print STDOUT "[ERROR 'mk_node2exoffs_msa_rows'] NO MSA row is associated with the 'external' node, $node: $! !\n\n";
		    return undef;
		}
		$node2exoffs_msa_rows{$node} = [$msa_row];
		next;
	    }

		# $node is an internal node.
	    my @exoffs_msa_rows = ();
	    foreach my $ch (@{$children}) {
		push @exoffs_msa_rows, @{$node2exoffs_msa_rows{$ch}};
	    }
	    my @std_exoffs_msa_rows = sort {$a <=> $b} @exoffs_msa_rows;

	   $node2exoffs_msa_rows{$node} = \@std_exoffs_msa_rows;

	} # End of the main foreach-loop (over all nodes). #

	return \%node2exoffs_msa_rows;

}	# End of the subroutine, 'mk_node2exoffs_msa_rows'. #


# wrapper_mk_node2exoffs_msa_rows:
#	This subroutine is a wrapper of the subroutine, 'mk_node2exoffs_msa_rows', above.
#	It enables the creation of %node2exoffs_msa_rows from its raw ingredients.
#	[CREATED on 2019/06/18.]
#
# INPUTS: ($topnode_id, \%node2ch, @seqnames, \%name2id), where
#
#	$topnode_id is the ID of the top-node, i.e., root;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all internal nodes;
#
#	$seqnames[$k] is the name (or ID) of the sequence at the $k th row of the MSA;
#	%name2id = ($node_name  => $node_ID, ...) for nodes with names,
#	where $node_name corresponds to $seqnames[$k] for external nodes;
#
# OUTPUT (\%node2exoffs_msa_rows, \%node2depth, \@nodeids_in_msa, \%nodeid2msa_row), where
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#	[This \%node2exoffs_msa_rows is the MAIN OUTPUT, which is also the output of 'mk_node2exoffs_msa_rows'];
#
#	%node2depth = ($node_id => $abs_depth, ...) for all nodes;
#	[Actually, \%node2depth is an output of 'make_node2abs_depth' in "MyTreeManipu_III.pm"];
#
#	$nodeids_in_msa[$k] is the ID of the (usually external) node that corresponds to the $k th row of the MSA;
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[ [Actually, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' in "MyANEX_Supple.pm"]; ] ... OBSOLETE as of 2019/06/27.
#	[Actually, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' above]; # Revised on 2019/06/27.
#
sub wrapper_mk_node2exoffs_msa_rows ($\%\@\%) {

	my ($topnode_id, $node2ch, $seqnames, $name2id) = @_;

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'wrapper_mk_node2exoffs_msa_rows'] The subroutine, 'make_node2abs_depth', FAILED: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my ($nodeids_in_msa, $nodeid2msa_row) = associate_nodeids_w_msa_rows (@{$seqnames}, %{$name2id});
	unless (defined $nodeid2msa_row) {
	    print STDOUT "[ERROR 'wrapper_mk_node2exoffs_msa_rows'] The subroutine, 'associate_nodeids_w_msa_rows', FAILED: $! !\n\n";
	    return (undef, $node2depth, undef, undef);
	}

	my $node2exoffs_msa_rows = mk_node2exoffs_msa_rows (%{$node2ch}, %{$node2depth}, %{$nodeid2msa_row});
	unless (defined $node2exoffs_msa_rows) {
	    print STDOUT "[ERROR 'wrapper_mk_node2exoffs_msa_rows'] The subroutine, 'mk_node2exoffs_msa_rows', FAILED: $! !\n\n";
	    return (undef, $node2depth, $nodeids_in_msa, $nodeid2msa_row);
	}

	return ($node2exoffs_msa_rows, $node2depth, $nodeids_in_msa, $nodeid2msa_row);

}	# End of the subroutine, 'wrapper_mk_node2exoffs_msa_rows'. #


# mk_set_msa_rows2br_up_or_lw:
#	This subroutine associates some sets of MSA rows to the upper- & lower-sides of the branches,
#	using the main output, \%node2exoffs_msa_rows, of 'wrapper_mk_node2exoffs_msa_rows'.
#	[CREATED on 2019/06/26.]
#
#	NOTE: We ASSUME that the indices in @exoffs_msa_rows are sorted in ascending order.
#
#	[REVISED on 2019/07/03]:
#	Now, the subroutine prioritizes the value with $up_or_lw = 'U'
#	if a single set of rows is realized by multiple values.
#	This way, the association will reflect Dollo parsimony.
#	(When the tree has some monofurcated nodes, however,
#	the association may not reflect Dollo parsimony.)
#
# INPUTS: (\%node2exoffs_msa_rows, \@set_rlv_msa_rows), where
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#	[This \%node2exoffs_msa_rows is the MAIN OUTPUT of 'wrapper_mk_node2exoffs_msa_rows';
#	it is also the output of 'mk_node2exoffs_msa_rows'];
#
#	@set_rlv_msa_rows stores the indices of the MSA rows that are under consideration.
#
# OUTPUTS: (\%set_msa_rows2br_up_or_lw), where
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
# 
sub mk_set_msa_rows2br_up_or_lw (\%\@) {

	my ($node2exoffs_msa_rows, $set_rlv_msa_rows) = @_;

	my %set_msa_rows2br_up_or_lw;

	foreach my $node (keys %{$node2exoffs_msa_rows}) {
	    my $br = $node;

	    my $sbj_msa_rows = $node2exoffs_msa_rows->{$node};

	    my ($common, $ex_sbj, $cmpl_msa_rows) = diff_nmarray (@{$sbj_msa_rows}, @{$set_rlv_msa_rows});
	    unless (defined $cmpl_msa_rows) {
		print STDOUT "[ERROR 'mk_set_msa_rows2br_up_or_lw'] The subroutine, 'diff_nmarray', FAILED: $! !\n";
		return undef;
	    }
	    if (@{$ex_sbj}>0) {
		print STDOUT "[ERROR 'mk_set_msa_rows2br_up_or_lw'] The MSA rows under the branch $br, that is, (", join (',', @{$sbj_msa_rows}), "), contains some EXTRA rows: (", join (',', @{$ex_sbj}), ") !\n";
		return undef;
	    }

	    my $cnct_sbj_msa_rows = join (',', @{$sbj_msa_rows});
	    my $cnct_cmpl_msa_rows = join (',', @{$cmpl_msa_rows});

			# REVISED on 2019/07/03. #
			# (If the set is already associated with 'U', leave it untouched;
			# however, if the set is already associated with 'L', override it.)
	    unless ((defined $set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows})
		    and ($set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows}->[1] eq 'U')) {
		$set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows} = [$br, 'L'];
	    }
	    ##$set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows} = [$br, 'L'];

			# (Override the previous association, no matter what it was.)
	    $set_msa_rows2br_up_or_lw{$cnct_cmpl_msa_rows} = [$br, 'U'];
	}

	return \%set_msa_rows2br_up_or_lw;

}	# End of the subroutine, 'mk_set_msa_rows2br_up_or_lw'. #


# mk_set_msa_rows2br_up_or_lw_dp:
#	This modified version of 'mk_set_msa_rows2br_up_or_lw'
#	uses the node depths to make the association completely reflect the Dollo parsimony.
#	[CREATED on 2019/07/03.]
#
# mk_set_msa_rows2br_up_or_lw:
#	This subroutine associates some sets of MSA rows to the upper- & lower-sides of the branches,
#	using the main output, \%node2exoffs_msa_rows, of 'wrapper_mk_node2exoffs_msa_rows'.
#	[CREATED on 2019/06/26.]
#
#	NOTE: We ASSUME that the indices in @exoffs_msa_rows are sorted in ascending order.
#
#	[REVISED on 2019/07/03]:
#	Now, the subroutine prioritizes the value with $up_or_lw = 'U'
#	if a single set of rows is realized by multiple values.
#	This way, the association will reflect Dollo parsimony.
#	(When the tree has some monofurcated nodes, however,
#	the association may not reflect Dollo parsimony.)
#
# INPUTS: (\%node2exoffs_msa_rows, \@set_rlv_msa_rows, \%node2depth), where
#
#	%node2exoffs_msa_rows = ($node_id => \@exoffs_msa_rows, ...) for all nodes,
#	with @exoffs_msa_rows stores the indices of all MSA rows corresponding to the external offsprings of $node_id;
#	[This \%node2exoffs_msa_rows is the MAIN OUTPUT of 'wrapper_mk_node2exoffs_msa_rows';
#	it is also the output of 'mk_node2exoffs_msa_rows'];
#
#	@set_rlv_msa_rows stores the indices of the MSA rows that are under consideration.
#
#	%node2depth = ($node_id => $depth, ...) for all nodes;
#
# OUTPUTS: (\%set_msa_rows2br_up_or_lw), where
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
# 
sub mk_set_msa_rows2br_up_or_lw_dp (\%\@\%) {

	my ($node2exoffs_msa_rows, $set_rlv_msa_rows, $node2depth) = @_;

	my %set_msa_rows2br_up_or_lw;

	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a}} keys %{$node2exoffs_msa_rows}) { # The bottom-up traversal (i.e., from leaves to the root). # MODIFIED in creation of 'mk_set_msa_rows2br_up_or_lw_dp' (on 2019/07/03). #
	##foreach my $node (keys %{$node2exoffs_msa_rows}) { ...} 
	    my $br = $node;

	    my $sbj_msa_rows = $node2exoffs_msa_rows->{$node};

	    my ($common, $ex_sbj, $cmpl_msa_rows) = diff_nmarray (@{$sbj_msa_rows}, @{$set_rlv_msa_rows});
	    unless (defined $cmpl_msa_rows) {
		print STDOUT "[ERROR 'mk_set_msa_rows2br_up_or_lw'] The subroutine, 'diff_nmarray', FAILED: $! !\n";
		return undef;
	    }
	    if (@{$ex_sbj}>0) {
		print STDOUT "[ERROR 'mk_set_msa_rows2br_up_or_lw_dp'] The MSA rows under the branch $br, that is, (", join (',', @{$sbj_msa_rows}), "), contains some EXTRA rows: (", join (',', @{$ex_sbj}), ") !\n";
		return undef;
	    }

	    my $cnct_sbj_msa_rows = join (',', @{$sbj_msa_rows});
	    my $cnct_cmpl_msa_rows = join (',', @{$cmpl_msa_rows});

			# REVISED on 2019/07/03. #
			# (If the set is already associated with 'U', leave it untouched;
			# however, if the set is already associated with 'L', override it.)
	    unless ((defined $set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows})
		    and ($set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows}->[1] eq 'U')) {
		$set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows} = [$br, 'L'];
	    }
	    ##$set_msa_rows2br_up_or_lw{$cnct_sbj_msa_rows} = [$br, 'L'];

			# REVISED in creation of 'mk_set_msa_rows2br_up_or_lw_dp' (on 2019/07/03). #
			# (If the set is already associated with 'U', leave it untouched;
			# however, if the set is already associated with 'L', override it.)
	    unless ((defined $set_msa_rows2br_up_or_lw{$cnct_cmpl_msa_rows})
		   and ($set_msa_rows2br_up_or_lw{$cnct_cmpl_msa_rows}->[1] eq 'U')) {
		$set_msa_rows2br_up_or_lw{$cnct_cmpl_msa_rows} = [$br, 'U'];
	    }
	    ##$set_msa_rows2br_up_or_lw{$cnct_cmpl_msa_rows} = [$br, 'U'];
	}

	return \%set_msa_rows2br_up_or_lw;

}	# End of the subroutine, 'mk_set_msa_rows2br_up_or_lw_dp'. #


# vsplit_into_eff_child_gblocks:
#	This satellite subroutine of 'list_vsplit_candidates' (in "MyANEX_Main.pm")
#	attempts to vertually split an input gap-block into
#	its "effective child" gap-blocks.
#	[CREATED on 2019/08/21.]
#
#	NOTE: This subroutine can also handle a tree that contains nono-furcated non-root nodes,
#	and/or a bifurcated root.
#
# INPUTS: ($br_sbj0, $up_or_lw_sbj0,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq, \@set_all_exnodes,
#	\%nodeid2msa_row, \%set_msa_rows2br_up_or_lw), where
#
#	$br_sbj0 is the ID of the branch separating the subject gap-block, and
#	$up_or_lw_sbj0 = 'U'/'L' indicates that the subject gap-block is on the upper/lower-end of $br_sbj,
#	BOTH initially assigned to the subject;
#	[Typically, $br_sbj and $up_or_lw_sbj are stored in
#	@list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	@set_all_exnodes stores the indices of all external nodes;
#	
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
#	[Typcally, \%set_msa_rows2br_up_or_lw is an output of 'mk_set_msa_rows2br_up_or_lw_dp' above];
# 
# OUTPUTS: (\@set_eff_child_gblks), where
#
#	@{$set_eff_child_gblks[$ch]} = ($br_ch, $up_or_lw_ch, \@exnodes_affd_by_ch, \@rows_affd_by_ch)
#	provides information on the $ch th "effective child" of the subject gap-block, where
#	the child is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@exnodes_affd_by_ch stores the indices of external nodes affected by the child,
#	and @rows_affd_by_ch stores the indices of the MSA rows affected by the child.
#
#	@set_eff_child_gblks is empty if the gap-block affects only a single MSA row (or external node);
#
sub vsplit_into_eff_child_gblocks ($$\%\%$\%\@\%\%) {

	my ($br_sbj0, $up_or_lw_sbj0,
	    $node2pa, $node2ch, $topnode_id, $br2eq, $set_all_exnodes,
	    $nodeid2msa_row, $set_msa_rows2br_up_or_lw) = @_;

	my ($br_sbj, $up_or_lw_sbj) = ($br_sbj0, $up_or_lw_sbj0);

	my $pa = $node2pa->{$br_sbj};
	my $children = $node2ch->{$br_sbj};

	if ($up_or_lw_sbj eq 'U') {

		# Go up the tree through uni-furcated non-root nodes, if any. #

	    while ((defined $pa) and ($pa != $topnode_id) and (@{$node2ch->{$pa}} ==1)) {
		$br_sbj = $pa;
		$pa = $node2pa->{$br_sbj};
	    }

	    my $eq_sbj = $br2eq->{$br_sbj};
	    if (defined $eq_sbj) {
			# Go beyond the bifurcated root. #
		$br_sbj = $eq_sbj;
		$up_or_lw_sbj = 'L';
	    }

	    $children = $node2ch->{$br_sbj}; # Update the set of child gap-blocks. #
	}

	my @preset_eff_child_gblks = ();

	if ($up_or_lw_sbj eq 'L') {

		# Go down the tree through uni-furcated non-root nodes, if any. #
	    while ((defined $children) and (@{$children} ==1)) {
		$br_sbj = $children->[0];
	        $children = $node2ch->{$br_sbj};
	    }

 	    unless ((defined $children) and (@{$children}>0)) {
		return []; # The subject effectively affects a single node. #
	    }

	    foreach my $ch (@{$children}) {
		push @preset_eff_child_gblks, [$ch, 'L'];
	    }

	} else { # if ($up_or_lw_sbj eq 'U') #

	    if ($pa != $topnode_id) { push @preset_eff_child_gblks, [$pa, 'U']; }

	    foreach my $sib (@{$node2ch->{$pa}}) {
		if ($sib == $br_sbj) { next; }
		push @preset_eff_child_gblks, [$sib, 'L'];
	    }
	}

		# Complete the output. #
	my @set_eff_child_gblks = ();

	foreach my $info_eff_ch (@preset_eff_child_gblks) { # Main foreach-loop (over effective children). #

	    my ($br_ch0, $up_or_lw_ch0) = @{$info_eff_ch};
	    my $exnodes_affd_by_ch = fetch_ext_offsprings_nr ($br_ch0, %{$node2ch});

	    if ($up_or_lw_ch0 eq 'U') {
		my ($common, $must_be_empty, $cmpl_exnodes) = diff_nmarray (@{$exnodes_affd_by_ch}, @{$set_all_exnodes});
		unless (defined $cmpl_exnodes) {
		    print STDOUT "[ERROR 'vsplit_into_eff_child_gblocks'] The subroutine, 'diff_nmarray', FAILED: $! !\n";
		    return undef;
		}
		$exnodes_affd_by_ch = $cmpl_exnodes;
	    }

	    my @rows_affd_by_ch = ();
	    foreach my $node (@{$exnodes_affd_by_ch}) { push @rows_affd_by_ch, $nodeid2msa_row->{$node}; }
	    if (@rows_affd_by_ch >1) { @rows_affd_by_ch = sort {$a <=> $b} @rows_affd_by_ch; }

	    my $cnct_rows_affd_by_ch = join (',', @rows_affd_by_ch);
	    my $br_up_or_lw = $set_msa_rows2br_up_or_lw->{$cnct_rows_affd_by_ch};
	    unless (defined $br_up_or_lw) {
		print STDOUT "[ERROR 'vsplit_into_eff_child_gblocks'] Somehow, the set of rows, ($cnct_rows_affd_by_ch), affected by (\$br_ch0, \$up_or_lw_ch0) = ($br_ch0, $up_or_lw_ch0), is NOT ASSOCIATED with ANYTHING through ", '%{$set_msa_rows2br_up_or_lw}', "!!: $! !\n\n";
		return undef;
	    }

	    my ($br_ch, $up_or_lw_ch) = @{$br_up_or_lw};

	    push @set_eff_child_gblks, [$br_ch, $up_or_lw_ch, $exnodes_affd_by_ch, \@rows_affd_by_ch];

	} # End of the Main foreach-loop (over effective children). #


	return \@set_eff_child_gblks;

}	# End of the subroutine, 'vsplit_into_eff_child_gblocks'. #


# vsplit_into_cmpl_eff_child_rblocks:
#	This satellite subroutine of 'list_cvsplit_candidates' (in "MyANEX_Main.pm")
#	attempts to vertually split the COMPLEMENTARY RESIDUE-BLOCK of an input gap-block into
#	its "effective child" residue-blocks.
#	It was created by modifying its predecessor, 'vsplit_into_eff_child_gblocks'.
#	[CREATED on 2019/08/25.]
#
#	[NOTE] The subroutine returns information on the complementary GAP-BLOCKS of the "effective child" residue-blocks.
#
# vsplit_into_eff_child_gblocks:
#	This satellite subroutine of 'list_vsplit_candidates' (in "MyANEX_Main.pm")
#	attempts to vertually split an input gap-block into
#	its "effective child" gap-blocks.
#	[CREATED on 2019/08/21.]
#
#	NOTE: This subroutine can also handle a tree that contains nono-furcated non-root nodes,
#	and/or a bifurcated root.
#
# INPUTS: ($br_sbj0, $up_or_lw_sbj0,
#	\%node2pa, \%node2ch, $topnode_id, \%br2eq, \@set_all_exnodes,
#	\%nodeid2msa_row, \%set_msa_rows2br_up_or_lw), where
#
#	$br_sbj0 is the ID of the branch separating the subject gap-block, and
#	$up_or_lw_sbj0 = 'U'/'L' indicates that the subject gap-block is on the upper/lower-end of $br_sbj,
#	BOTH initially assigned to the subject;
#	[Typically, $br_sbj and $up_or_lw_sbj are stored in
#	@list_gblocks = @{$set_lists_gblocks_in_wds[$wd]},
#	where \@set_lists_gblocks_in_wds is an output of, e.g., 'prepare_sliding_windows' in "MyANEX_Supple.pm"];
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the root;
#	%node2ch = ($node_id => \@child_nodes, ...) for all internal nodes;
#	$topnode_id is the ID of the top node (i.e., root);
#
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for all branches with their own equivalents;
#	[Typically, \%br2eq is an output of, e.g., 'create_br2eqbr' in "MyTreeMap_indels_sppl.pm", or of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm"];
#
#	@set_all_exnodes stores the indices of all external nodes;
#	
#	%nodeid2msa_row = ($node_ID => $indx_row, ...), where
#	$indx_row specifies the row of the MSA that corresponds to the node with ID = $node_ID.
#	[Typically, \@nodeids_in_msa & \%nodeid2msa_row are the outputs of 'associate_nodeids_w_msa_rows' or 'wrapper_mk_node2exoffs_msa_rows' above];
#
#	%set_msa_rows2br_up_or_lw = ($cnct_set_msa_rows => [$br_id, $up_or_lw], ...), with
#	$cnct_set_msa_rows = join (',', @set_msa_rows), where @set_msa_rows stores some MSA rows,
#	$br_id is the branch separating the @set_msa_rows to the rest,
#	and $up_or_lw = 'U'/'L' if @set_msa_rows is on the upper-/lower-side of $br_id.
#	[Typcally, \%set_msa_rows2br_up_or_lw is an output of 'mk_set_msa_rows2br_up_or_lw_dp' above];
# 
# OUTPUTS: (\@set_cmpl_eff_child_rblks), where
#
#	@{$set_cmpl_eff_child_rblks[$cch]} = ($br_cch, $up_or_lw_cch, \@cexnodes_affd_by_cch, \@crows_affd_by_cch)
#	provides information on the $cch th "effective child" of the COMPLEMENTARY RESIDUE-BLOCK of the subject gap-block, where
#	the COMPLEMENTARY gap-block of the child residue-block is on the $up_or_lw_ch side of the branch with ID $br_ch ,
#	@cexnodes_affd_by_ch stores the indices of external nodes affected by the COMPLEMENTARY gap-block the child residue-block,
#	and @crows_affd_by_ch stores the indices of the MSA rows affected by the COMPLEMENTARY gap-block of the child residue-block.
#
#	@set_cmpl_eff_child_rblks is empty if the COMPLEMENTARY RESIDUE-BLOCK of the gap-block affects only a single MSA row (or external node);
#
# [ sub vsplit_into_eff_child_gblocks ($$\%\%$\%\@\%\%) { ] ... for the predecessor, 'vsplit_into_eff_child_gblocks'.
#
sub vsplit_into_cmpl_eff_child_rblocks ($$\%\%$\%\@\%\%) {

	my ($br_sbj0, $up_or_lw_sbj0,
	    $node2pa, $node2ch, $topnode_id, $br2eq, $set_all_exnodes,
	    $nodeid2msa_row, $set_msa_rows2br_up_or_lw) = @_;

	my ($br_sbj, $up_or_lw_sbj) = ($br_sbj0, $up_or_lw_sbj0);

	my $pa = $node2pa->{$br_sbj};
	my $children = $node2ch->{$br_sbj};

	if ($up_or_lw_sbj eq 'L') {

		# Go up the tree through uni-furcated non-root nodes, if any. #

	    while ((defined $pa) and ($pa != $topnode_id) and (@{$node2ch->{$pa}} ==1)) {
		$br_sbj = $pa;
		$pa = $node2pa->{$br_sbj};
	    }

	    my $eq_sbj = $br2eq->{$br_sbj};
	    if (defined $eq_sbj) {
			# Go beyond the bifurcated root. #
		$br_sbj = $eq_sbj;
		$up_or_lw_sbj = 'U';
	    }

	    $children = $node2ch->{$br_sbj}; # Update the set of child gap-blocks. #
	}

	my @preset_cmpl_eff_child_rblks = ();

	if ($up_or_lw_sbj eq 'U') {

		# Go down the tree through uni-furcated non-root nodes, if any. #
	    while ((defined $children) and (@{$children} ==1)) {
		$br_sbj = $children->[0];
	        $children = $node2ch->{$br_sbj};
	    }

 	    unless ((defined $children) and (@{$children}>0)) {
		return []; # The complementary residue-block of the subject effectively affects a single node. #
	    }

	    foreach my $ch (@{$children}) {
		push @preset_cmpl_eff_child_rblks, [$ch, 'U']; # Store the seed information on the COMPLEMENTARY gap-block of the child residue-block. #
	    }

	} else { # if ($up_or_lw_sbj eq 'L') #

	    if ($pa != $topnode_id) { push @preset_cmpl_eff_child_rblks, [$pa, 'L']; } # Store the seed informatin on the COMPLEMENTARY gap-block of the "effective-child" residue-block. 

	    foreach my $sib (@{$node2ch->{$pa}}) {
		if ($sib == $br_sbj) { next; }
		push @preset_cmpl_eff_child_rblks, [$sib, 'U']; # Store the seed information on the COMPLEMENTARY gap-block of the "effective-child" residue-block.
	    }
	}

		# Complete the output. #
	my @set_cmpl_eff_child_rblks = ();

	foreach my $info_ceff_ch (@preset_cmpl_eff_child_rblks) { # Main foreach-loop (over "complementary-effective children"). #

	    my ($br_cch0, $up_or_lw_cch0) = @{$info_ceff_ch};
	    my $cexnodes_affd_by_cch = fetch_ext_offsprings_nr ($br_cch0, %{$node2ch});

	    if ($up_or_lw_cch0 eq 'U') {
		my ($common, $must_be_empty, $cmpl_cexnodes) = diff_nmarray (@{$cexnodes_affd_by_cch}, @{$set_all_exnodes});
		unless (defined $cmpl_cexnodes) {
		    print STDOUT "[ERROR 'vsplit_into_cmpl_eff_child_rblocks'] The subroutine, 'diff_nmarray', FAILED: $! !\n";
		    return undef;
		}
		$cexnodes_affd_by_cch = $cmpl_cexnodes;
	    }

	    my @crows_affd_by_cch = ();
	    foreach my $node (@{$cexnodes_affd_by_cch}) { push @crows_affd_by_cch, $nodeid2msa_row->{$node}; }
	    if (@crows_affd_by_cch >1) { @crows_affd_by_cch = sort {$a <=> $b} @crows_affd_by_cch; }

	    my $cnct_crows_affd_by_cch = join (',', @crows_affd_by_cch);
	    my $br_up_or_lw = $set_msa_rows2br_up_or_lw->{$cnct_crows_affd_by_cch};
	    unless (defined $br_up_or_lw) {
		print STDOUT "[ERROR 'vsplit_into_cmpl_eff_child_rblocks'] Somehow, the set of rows, ($cnct_crows_affd_by_cch), affected by (\$br_cch0, \$up_or_lw_cch0) = ($br_cch0, $up_or_lw_cch0), is NOT ASSOCIATED with ANYTHING through ", '%{$set_msa_rows2br_up_or_lw}', "!!: $! !\n\n";
		return undef;
	    }

	    my ($br_cch, $up_or_lw_cch) = @{$br_up_or_lw};

	    push @set_cmpl_eff_child_rblks, [$br_cch, $up_or_lw_cch, $cexnodes_affd_by_cch, \@crows_affd_by_cch];

	} # End of the Main foreach-loop (over "complementary-effective children"). #

	return \@set_cmpl_eff_child_rblks;

}	# End of the subroutine, 'vsplit_into_cmpl_eff_child_rblocks'. #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #



