package MyBlastp_parser_05;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyBlastp_parser_05.pm: (created on Sep 11 (Sun), 2005)
#	A revised version of the module 'MyBlastp_parser_04.pm',
#	which parses the output of blastp.
#	The MAJOR REVISION is that this version can handle hits with
#	regions masked by 'x's.
#	More precisely, the script calculates the total length,
#	#{identitcal aas}, #{gaps}, and the total raw score IN THE X-MASKED REGIONS. 
#	This takes a refinement of the program which can slow down
#	the search significantly.
#	This module also extracts information which is necessary when we convert
#	a raw score into the corresponding bit score,
#	which had been neglected in the previous versions.
#
# REVISED on Sunday, May 28, 2006:
#	The module can now process the output of PSI-BLAST
#	by retrieving only the essential output with 'trimmed_psi_output'.
#	
#
# MyBlastp_parser_04.pm:
#	a module to parse the output of blastp version 0.4
#
# Two inportant BUG FIXES on Jan 26(Sun), 2003:
#	In the module "parse_blastp", two bugs associated with
#	the changes of returning value type
#	(from an array to an array reference) were fixed.
#
# Two  minor revisions on Jan 24(Fri), 2003:
#	For a while,I am going to  disable the capability
#	of the subroutine 'parse_align' to extract
#	 the aligned subject sequence $se_seq.
#	( To keep compatibility, I'm going to  return an empty sequence as $se_seq.)
#	I also decided to return the REFERENCE of an array,
#	instead of the array itself.
#
# a minor revision on Jan 12(Sun), 2003:
#	I changed the filtering by means of E-values so that
#	the sequences on the boundary ( namely, those with
#	its E-value = threshold) are included in the homology sets.
#
# REVISED on Jan 10(Fri), 2003:
#	the module now recieves as an additional filtering parameter,
#	named $flag_ab_rel which determines the length filtering mode:
#	when $flag_ab_rel = 1, filtering via absolute length
#			  = 0, 		 via relative length.
#
# REVISED from version 0.2 on Jan 9(Thu), 2003:
#	(i) In the modile homolog_filter, the filtering criterion concerning the aligned length
#		changed from using relative value to using absolute length.
#	(ii) In the module homolog_filter, I explicitly added the filtering
#		in terms of the E-value.
#	(iii) I enabled the subroutines parse_align and parse_align_br to change the
#		E-value of the form "e-???" to "1e-???".
#
# a minor revision on Jan. 7(Tue), 2003: In the module homolog_filter, I quit storing a list of
#					 "redundant sequences" which are to be
#					ommitted from the future query.
#					(This might save the memory this module consumes.)
#
# REVISED on some day in Dec, 2002: the filtering criterion of the module homolog_filter somewhat improved.
#		( see the header of the module for details )
#

use strict;
use warnings;
use Exporter;

our @ISA = qw(Exporter);
##our @EXPORT = qw(parse_blastp homolog_filter parsed_output_remover);
##our @EXPORT = qw(parse_blastp homolog_filter parsed_output_remover);
##our @EXPORT_OK = qw(parse_blastp examine_masked_regions_pw );
our @EXPORT_OK = qw(
	trimmed_psi_output
	parse_blastp separate_masked_aln score_pairalign 
	examine_unmasked_regions_pw
		);

sub parse_align ($\@);
sub parse_align_br ($\@);
sub extract_auxiliary_info (\@);
sub initialize_subalign (\@\@$);
sub finalize_subalign (\@\@$);
sub score_pairalign (\$\$\%$$$$);

=obsolete
# parsed_output_remover(\@): a subroutine to free the memory allocated to an array
#			 	which stores parsed BLAST output.
#
# ---- virtually useless unless we can deal with extremely large BLAST outputs ----
#
sub parsed_output_remover(\@) {

	my ($rf_bl_output) =@_;

	for(my $i = 0; $i < 2; $i++) {
		unless(shift @{$rf_bl_output}) {
			undef @{$rf_bl_output};
			return 1;
		}
	} 

	while ( my $rf_sb_info = pop(@{$rf_bl_output}) ) { # the outer while #

		for (my $i = 0; $i < 2; $i++) {
			unless(shift @{$rf_sb_info}) {
				#undef @{$rf_sb_info};
				last;
			}
		}					# end of for #

		while ( my $rf_hit = pop(@{$rf_sb_info}) ) { undef @{$rf_hit}; }

		undef @{$rf_sb_info};
	}						# end of the outer while #

	undef @{$rf_bl_output};
	return 1;
}				# end of the subroutine parsed_output_remover #
=cut

=obsolete
# homolog_fileter($\@\@\%): a subroutine which recieves the reference of @parsed_output
#			(,which is a raw output of the subroutine parse_blastp),
#			reference of @filer_parameters, and the reference of %omitted,
#			then, add elements to %omitted, if necessary,
#			and returns the reference of a list of homologues which passed the filter,
#				              that of homologues which are too short,
#				          and that of homologues which are too long.
#
# a minor revision on Jan 12(Sun), 2003: I changed the filtering condition on the E-value
#					so that sequences with E-value equaling to the threshold
#					will not be filtered out.
#
# REVISION on Jan 10(Fri), 2003: I added a new filtering parameter $flag_ab_rel,
#				which decides whether we filter entries by mefans of
#				absolute length($flag_ab_rel =1), or relative length(=0).
# REVISED the filtering criteria (on Jan 9(Thu), 2003):
#	(i) Explicit addition of filtering via E-value, which accompanies
#		the revision of "homolog_search_v2.pl" to "homolog_search_(a/b)_v3.pl".
#	(ii) The criterion regarding the alignment length changed
#	from looking at (aligned length excluding gaps) / min(query length, subject length)
#	to the absolute length of the aligned part of the subject sequence (gaps excluded).
#
# REVISED the filtering process( in Dec, 2002):
#	so far, I used the total average of sequece identity ratio as a criterion.
#	in the new version, I examine the average of THE HIGHEST identities whose
#	sequence length sum exceeds some cutoff length.
#
# ADDTITIONAL FEATURE: removing the whole structure of @parsed_output ( output of parse_blastp )
#			---- disabled later as it turned out to be useless (or dangerous),
#				at least within the Land of Perl        ---
#
sub homolog_filter(\@\@\%$) {

	my ($ref_parsed_output, $ref_filter_parameters, $ref_omitted, $cluster_num) = @_;
	(defined $cluster_num) or ($cluster_num = "UK");
	my ($cutoff_identity, $cutoff_length,
		$omission_identity, $omission_length, $omission_align,
		$caution_too_short, $caution_too_long,
		 $e_threshold, $flag_ab_rel)	= @{$ref_filter_parameters};
	my $qu_id = shift @{$ref_parsed_output};
	my $qu_len = shift @{$ref_parsed_output};
	my @homologues = ();
	my @too_short = ();
	my @too_long = ();

	OUTER:
	foreach my $sb_inf (@{$ref_parsed_output}) {		# the outer foreach #
	#while ( my $sb_inf = pop(@{$ref_parsed_output}) ) {
	   my $sb_id = shift @{$sb_inf};
	   if ( my ($tmp) = ($sb_id =~ m/^\s*(\S+)/) ) {
	   	$sb_id = $tmp;
	   } else {
		#if (shift @{$sb_inf}) {
		#	while (my $hit = pop(@{$sb_inf}) ) { undef @{$hit};}
		#}
		#undef @{$sb_inf};
		next;
	   }
	   my $sb_len = shift @{$sb_inf};

	   my @fragments = ();
	   foreach my $hit (@{$sb_inf}) {	# the inner foreach #
	   #while ( my $hit = pop(@{$sb_inf}) ) {
		my ($score, $identities, $e_value, $gaps, $sb_init, $sb_end) = @{$hit};
		#undef @{$hit};
		if($e_value eq 'UK') {
			print "Query=$qu_id, Subject=$sb_id:\tE-value is missing!!\n";
			next;
		}
		($e_value > $e_threshold) and next;	# revison on Jan 12(Sun), 2003 #
		#($e_value < $e_threshold) or next;	# revision on Jan 9(Thu), 2003 #
		my ($num_frag, $den_frag, $percent_id) = ($identities =~ m!(\d*)/(\d*)\s*\((\d*)!) or next;
		my $len_frag = ($sb_end > $sb_init) ? ($sb_end - $sb_init + 1) : ($sb_init - $sb_end + 1);
		push @fragments, [$percent_id,$len_frag,$num_frag, $den_frag];
	   #}					# end of the inner while loop #
	   }					# end of the inner foreach loop #
	   #undef @{$sb_inf};

	   my $min_len = ($qu_len < $sb_len) ? $qu_len : $sb_len;
						# revision made on Jan 10(fri), 2003 #
	   my $ref_len = $flag_ab_rel ? $cutoff_length : ($min_len * $cutoff_length / 100);
	   my $aligned_len = 0;
	   my $id_numerator = 0;
	   my $id_denominator = 0;
	   foreach my $frag (sort {${$b}[0] <=> ${$a}[0]} @fragments) {	# the 2nd inner foreach #

		my ($percent_id,$len_frag,$num_frag,$den_frag) = @{$frag};
		$aligned_len += $len_frag;
		$id_numerator += $num_frag;
		$id_denominator += $den_frag;
	   	#$id_denominator or next;
	   	(my $id_prop = 100 * $id_numerator/$id_denominator) < $cutoff_identity and next OUTER;
	   	#if ( (my $aligned_prop = 100 * $aligned_len/$min_len) > $cutoff_length ) { # outer if #
							# revision made on Jan 9(Thu), 2003 #
	   	#if (  $aligned_len  > $cutoff_length ) { # outer if #
							# revision made on Jan 10(fri), 2003 #
	   	if (  $aligned_len  > $ref_len ) { # outer if #
	   		push @homologues, $sb_id;
	   		#$homolog_nos{$sb_id} = $cluster_num;
	   
		   	my ($tmp2) = ($qu_id =~ m/^\s*(\S+)/);
		   	($sb_id eq $tmp2) and next OUTER;
		   	my $su_qu_ratio = 100 * $sb_len/$qu_len;

			# The following comment-outs are the minor revision made on Jan 7(Tue), 2003 #
		   	#if (($id_prop > $omission_identity) and ($su_qu_ratio < $omission_length)  and ($aligned_prop > $omission_align) ) {
			    #${$ref_omitted}{$sb_id} = $cluster_num;	# for general purpose #
			    #($sb_id =~ m/^ref/) and ${$ref_omitted}{$sb_id} = $cluster_num; # for query=Rat #
		   	#}

		   	if ($su_qu_ratio < $caution_too_short) {
			    push @too_short, "$sb_id:\($sb_len\/$qu_len\)";
		    	} elsif ($su_qu_ratio > $caution_too_long) {
			    push @too_long, "$sb_id:\($sb_len\/$qu_len\)";
		    	}							# end of if-elsif #
			last;
		}								# end of the outer if #
	   }						# end of the 2nd inner foreach #

	} 					# end of the outer while loop #
	#} 					# end of the outer foreach loop #

	#undef @{$ref_parsed_output};
	return (\@homologues, \@too_short, \@too_long);

}				# end of the subroutine homolog_filter #
=cut


# trimmed_psi_output:
#	This subroutine reads in a raw output of PSI-BLAST, or 'blastpgp' to be more precise,
#	then skips the interim rounds
#	and just retrieves the result of the final round.
#
# INPUT: (\*${$fh}, $maxnum_rounds), where
#	$fh is the reference to the filehandle of the raw PSI-BLAST output file;
#	$maxnum_rounds is the maximum number of the iterative rounds (the argument of '-j').
#
# OUTPUT: \$bl_output,
#	which contains query information as well as information on the final round
#	of the PSI-BLAST search.
#	The subroutine returns undef if the raw output runs out of the information.
#
sub trimmed_psi_output (\*$) {

	my ($fh, $maxnum_rounds) = @_;
	local $/ = '';   # Use blank lines as input-record-separators #

	my $in = <$fh>;
	(defined $in) or (return undef);

	while ($in !~ m/^Query\s*\=\s*(\S+)/ ) {
	    $in = <$fh>;
	    (defined $in) or (return undef);
	}

	my $bl_output = $in . "\n";

		## Skip superfluous information
		## -- Actually this is the only difference from
		## 'homolog_info05_pipe_II.pl'

	while ($in !~  m/^CONVERGED\!/ ) {
	    my ($round) = ($in =~ m/^Results\s+from\s+round\s+(\d+)/ );

	    if ((defined $round) and ($round == $maxnum_rounds)) {
		while ($in = <$fh>) {
		    (($in =~ m/^CONVERGED\!/ )  or ($in =~ m/^\>/))
			 and last;
		}
		unless (defined $in) {
		    print STDERR "[ERROR 'trimmed_psi_output'] ILLEGAL INPUT for the header:\n";
		    print STDERR $bl_output;
		    return undef;
		}
		last;
	    } elsif ($in =~ m/^\s+\*+\s+No\s+hits\s+found/) {
		last;
	    } else {
		##$in = <STDIN>; ## A BUG FIXED on Sunday, Oct 16, 2006
		$in = <$fh>;
		(defined $in) and next;

		print STDERR "[ERROR 'trimmed_psi_output'] We've RUN OUT OF THE INPUT for the header:\n";
		print STDERR $bl_output;
		return undef;
	    }
	}

	$in =~ s/^CONVERGED\!\s+//;
	$bl_output .= $in;

	$/ = "\nBLAST";
	$bl_output .= <$fh>;

	return \$bl_output;

}		# end of the subroutine 'trimmed_psi_output'


# subroutine parse_blastp:
#	take an output of blastp (for the 1st argument = 0) or
#	tblastn (for the 1st argument = 1)  as an argument,
#	parse it, and return the result in the following list:
#	($query : identity of the query, $query_length,
#	\@subject_info_1, \@subject_info_2, ...),
#	where @subject_info =
#	   ($subject : idenity of the subject, $subject_length, \@hit_1, \@hit_2, ...),
#	where  @hit =
#	   ($score, $e_value, $identities, $gaps, [$frame for tblastn,] $qu_init, $qu_end, $sb_init, $sb_end, \$qu_seq, \$sb_seq)
#	   ##($score, $e_value, $identities, $gaps, [$frame for tblastn,] $qu_init, $qu_end, $sb_init, $sb_end, $sb_seq) ## REVISED on Sep 10 (Sat), 2005 
#	when the 2nd argument is 1 (full-parsing mode).
#	In the brief mode (2nd argument = 0), 
#	   @hit = ($score, $identities, $e_value, $gaps, [$frame for tblastn,] $sb_init, $sb_end).
#
# MAJOR REVISION on Sep 10 (Sat), 2005:
#	The subroutine now returns the auxiliary information
#	which is essential for the conversion of a raw score into the bit score.
#
sub parse_blastp ($$\$) {

	my ($p_or_tn, $option, $ref) = @_;

	my $query = undef;
	my $query_length;
	my @output = ();
	my $tmp;
	my @array = reverse split (/\n\s*/, ${$ref});
	undef ${$ref};

	while ($tmp = pop  @array){	# the 1st while #
		($query) = ($tmp =~ m/^Query=\s*(.*)/) and last;
	}				# end of the 1st while #

	#while ($tmp = pop @array or return ($query)) { # an important bug fix on Jan 26 (sun), 2003 #
	while ($tmp = pop @array) {
	   if( ($query_length) = ($tmp =~ m/^\((\S+)\s*letters\)/) ) {
		$query_length =~ s/,//g;
		last;
	   }
	   $query .= " $tmp";
	}
	$query or ($query = "UK");
	$query_length or ($query_length = "UK");
	#$tmp = pop @array or return ($query);
	#($query_length) = ($tmp =~ m/^\((\S+)\s*letters\)/) or ($query_length = "UK");

	@output = ($query, $query_length);

	while ($tmp = pop @array) {		# the 2nd while #
	    if ($tmp =~ m/^>.+/) {
		push @array, $tmp;
		last;
	    } elsif ($tmp =~ m/^\*+\s+No\s+hits\s+found/) { # Revised on Sunday, May 28, 2006
		push @array, $tmp;
		last;
	    }
	}						# end of the 2nd while #

	#@array or return @output;		# an important bug fix on Jan 26 (Sun), 2003 #
	@array or return \@output;

	# we have to choose a parsing-mode from 2 options #
	# if $option = 1, then use the full version #
	# if $option = 0, use the abbreviated version #
	my $rf_temp = $option ? parse_align($p_or_tn,@array) : parse_align_br($p_or_tn,@array);
	# end of the choice #

	# @{$aux_info} = ($lambda, $K, $H, $gpd_lambda, $gpd_K, $gpd_H)
	my $aux_info = extract_auxiliary_info (@array);

	undef @array;
	while (my $tmp = shift @{$rf_temp}) {
	push @output, $tmp;
	}
	undef @{$rf_temp};
	return (\@output, $aux_info);		# a minor revison on Jan 24 (Fri), 2003 #
}					# end of the subroutine "parse_blastp" #



# subroutine parse_align ($\@) :
#	parse the alignment sections of the blastp (for $_[0] =0) or
#	the tblastn (for $_[0] =1)  output,
#	and return the result in the format:
#	  (\@subject_info_1, \@subject_info_2, ...),
#	where @subject_info = 
#	  ($subject, $subject_length, \@hit_1, \@hit_2, ...),
#	where @hit =
#	   ($score, $identities, $e_value, $gaps, [$frame for tblastn,] $qu_init,$qu_end,$sb_init,$sb_end, \$qu_seq, \$sb_seq).
#	   ###($score, $identities, $e_value, $gaps, [$frame for tblastn,] $qu_init,$qu_end,$sb_init,$sb_end,$sb_seq). 
#
# MAJOR REVISION on Sep 10 (Sat), 2005:
#	The subroutine now extracts the aligned query sequence as well as the subject sequence.
#
# Two  minor revisions on Jan 24(Fri), 2003:
#	For a while,I am going to  disable the capability to extract
#	 the aligned subject sequence $se_seq.
#	( To keep compatibility, I'm going to  return an empty sequence as $se_seq.)
#	I also decided to return the REFERENCE of an array, instead of the array itself.
#
sub parse_align ($\@) {

	my ($p_or_tn, $ref) = @_;
	my $subject = '';
	my $subject_length = 0;
	my $tmp;
	my @output = ();

	((@{$ref} == 0) or ($ref->[$#{$ref}] =~ m/^\*+\s+No\s+hits\s+found/))
		and (return []); # Revised on Sunday, May 28, 2006

   OUTER:
   while ( $tmp = pop @{$ref}) {				# the outer while #


	( ($subject) = ($tmp =~ m/^>(.+)/) ) or next; 
	   while ($tmp = pop @{$ref} ) {			# the 1st while #
	      if (($subject_length)  = ($tmp =~ m/^Length\s*=\s*(\S*)/) ) {
		$subject_length =~ s/,//g;
		last;
	      }
	      $subject .= " $tmp";
	   }							# end of the 1st while #
	   my @subject_info = ($subject, $subject_length);

	MIDDLE:
	while ($tmp = pop @{$ref}) {		# the middle while #
	   my ($score)  = ($tmp =~ m/^Score\s*=\s*([^,]*)/); 
	   #($score eq ''  or $e_value eq '')  and next;
	   $score eq ''  and next;
	   $score =~ s/\s+bits\s+//g;
	   my @hit = ($score);
	   my $e_value;
	   ($e_value) = ($tmp =~ m/Expect\S*\s*=\s*(\S*)/);	# a minor revision #
	    ($e_value eq '')  and ($e_value = "UK");
	    $e_value =~ s/^e/1e/;
	   $tmp = pop @{$ref};
	   my ($gaps, $identities, $frame) = ();
	   if ( ($identities) = ($tmp =~ m/^Identities\s*=\s*([^,]*)/)) {
		$identities =~ s/\s+//g;
	      ($gaps) = ($tmp =~ m/Gaps\s*=\s*(.*)/) or ($gaps = 0);
		$gaps =~ s/\s+//g;
	      if ( $p_or_tn and $tmp = pop @{$ref}) {
		($frame) = ($tmp =~ m/^Frame\s*=\s*(.*)/);
		($frame =~ m/\S/) or ($frame = 'UK');
	      }
	   } else {
	      $identities = "UK";
	      $gaps = "UK";
	      push(@{$ref}, $tmp);
	   }
	   push @hit, ($identities, $e_value, $gaps);
	   $p_or_tn and (push @hit, $frame);

	   my ($qu_init,$qu_end,$sb_init,$sb_end,$qu_seq, $sb_seq) = (0,0,0,0,'', '');

	   while ( $tmp = pop @{$ref}) {	# the inner while #

		## A minor revision on Sep 10 (Sat), 2005
		##if (my ($init, $end) = ($tmp =~ m/^Query:\s*(\d+)\s*\S+\s+(\d+)/)) {
		if (my ($init, $seq, $end) = ($tmp =~ m/^Query:\s*(\d+)\s*(\S+)\s+(\d+)/)) {
			($qu_init) or  ($qu_init = $init);
			$qu_end = $end;
			$qu_seq .= $seq;
			next;

		# A minor revision on Jan 24 (Fri), 2003 #
		#} elsif ( my ($init2, $end2) =($tmp =~m/^Sbjct:\s*(\d+)\s*\S+\s+(\d+)/)) {
		# --> The function resurrected on Sep 10 (Sat), 2005 #
		} elsif ( my ($init2, $seq2, $end2) =($tmp =~m/^Sbjct:\s*(\d+)\s*(\S+)\s+(\d+)/)) {
			$sb_init or ($sb_init = $init2);
			$sb_end = $end2;
			$sb_seq .= $seq2;
			next;
		} elsif ($tmp =~ m/^Score\s*=/) {	
			#$sb_seq =~ s/\W+//g;
			##push @hit, ($qu_init,$qu_end,$sb_init,$sb_end,$sb_seq);
			push @hit, ($qu_init,$qu_end,$sb_init,$sb_end, \$qu_seq, \$sb_seq);
			push @subject_info, \@hit;
			push @{$ref}, $tmp;
			last;
		} elsif ($tmp =~ m/^>.+/) {
			#$sb_seq =~ s/\W+//g;
			##push @hit, ($qu_init,$qu_end,$sb_init,$sb_end,$sb_seq);
			push @hit, ($qu_init,$qu_end,$sb_init,$sb_end, \$qu_seq, \$sb_seq);
			push @subject_info, \@hit;
			push @output, \@subject_info;
			push @{$ref}, $tmp;
			last MIDDLE;
		} elsif ($tmp =~ m/^Database/) {
			#$sb_seq =~ s/\W+//g;
			##push @hit, ($qu_init,$qu_end,$sb_init,$sb_end,$sb_seq);
			push @hit, ($qu_init,$qu_end,$sb_init,$sb_end, \$qu_seq, \$sb_seq);
			push @subject_info, \@hit;
			push @output, \@subject_info;
			last OUTER;
		} elsif ($tmp =~ m/^Lambda/ ) { # Revised on Jan 30 (Mon), 2006 
				# so that it can also handle an output of 'bl2seq'.
			push @{$ref}, $tmp;
			push @hit, ($qu_init,$qu_end,$sb_init,$sb_end, \$qu_seq, \$sb_seq);
			push @subject_info, \@hit;
			push @output, \@subject_info;
			last OUTER;
		}				# end of the if-elsif-...-elsif #
	   }					# end of the inner while #
	}					# end of the middle while #
   }							# end of the outer while #

   return \@output;			# A minor revision on Jan 24 (Fri), 2003 #

}							# end of the subroutine parse_align #


# subroutine parse_align_br ($\@) : parse the alignment sections of the blastp(for $_[0] = 0) or tblastn(for $_[0] = 1) output
#				and return the result in the abbreviated format:
#				(\@subject_info_1, \@subject_info_2, ...),
#				where @subject_info = 
#				($subject, $subject_length, \@hit_1, \@hit_2, ...),
#				where @hit =
#				($score, $identities, $e_value, $gaps, [$frame for tblastn,]  $sb_init,$sb_end).
#
sub parse_align_br ($\@) {

	my ($p_or_tn, $ref) = @_;
	my $subject = '';
	my $subject_length = 0;
	my $tmp;
	my @output = ();

	((@{$ref} == 0) or ($ref->[$#{$ref}] =~ m/^\*+\s+No\s+hits\s+found/))
		and (return []); # Revised on Sunday, May 28, 2006.

   OUTER:
   while ( $tmp = pop @{$ref}) {				# the outer while #

	( ($subject) = ($tmp =~ m/^>(.+)/) ) or next; 
	   while ($tmp = pop @{$ref} ) {			# the 1st while #
	      if (($subject_length)  = ($tmp =~ m/^Length\s*=\s*(\S*)/) ) {
		$subject_length =~ s/,//g;
		last;
	      }
	      $subject .= " $tmp";
	   }							# end of the 1st while #
	   my @subject_info = ($subject, $subject_length);

	MIDDLE:
	while ($tmp = pop @{$ref}) {		# the middle while #
	   my @hit = my ($score)  = ($tmp =~ m/^Score\s*=\s*([^,]*)/); 
	   #($score eq ''  or $e_value eq '')  and next;
	   $score eq ''  and next;
	   my $e_value;
	   ($e_value) = ($tmp =~ m/Expect\s*=\s*(\S*)/);
	   ($e_value eq '')  and ($e_value = "UK");
	    $e_value =~ s/^e/1e/;
	   $tmp = pop @{$ref};
	   my ($gaps, $identities, $frame) = ();
	   if ( ($identities) = ($tmp =~ m/^Identities\s*=\s*([^,]*)/)) {
	      ($gaps) = ($tmp =~ m/Gaps\s*=\s*(.*)/) or ($gaps = 0);
		if ($p_or_tn and $tmp = pop @{$ref}) {
			($frame) = ($tmp =~ m/^Frame\s*=\s*(.*)/);
			($frame =~ m/\S/) or ($frame = 'UK');
		}
	   } else {
	      $identities = "UK";
	      $gaps = "UK";
	      push(@{$ref}, $tmp);
	   }
	   push @hit, ($identities, $e_value, $gaps);
	   $p_or_tn and push(@hit, $frame);

	   my ($sb_init,$sb_end) = (0,0);

	   while ( $tmp = pop @{$ref}) {	# the inner while #

		if ( my ($init2, $end2) =($tmp =~m/^Sbjct:\s*(\d+)\s*\S+\s+(\d+)/)) {
			$sb_init or $sb_init = $init2;
			$sb_end = $end2;
			next;
		} elsif ($tmp =~ m/^Score\s*=/) {	
			push @hit, ($sb_init,$sb_end);
			push @subject_info, \@hit;
			push @{$ref}, $tmp;
			last;
		} elsif ($tmp =~ m/^>.+/) {
			push @hit, ($sb_init,$sb_end);
			push @subject_info, \@hit;
			push @output, \@subject_info;
			push @{$ref}, $tmp;
			last MIDDLE;
		} elsif ($tmp =~ m/^Database/) {
			push @hit, ($sb_init,$sb_end);
			push @subject_info, \@hit;
			push @output, \@subject_info;
			last OUTER;
		}				# end of the if-elsif-...-elsif #
	   }					# end of the inner while #
	}					# end of the middle while #
   }							# end of the outer while #
   return \@output;	# a minor revision on Jan 24 (Fri), 2003 #

}				# end of the subroutine parse_align #


# extract_auxiliary_info:
#	The subroutine receives the tail of BLAST output,
#	and extracts information which is essential for the conversion
#	of the raw score into the bit score.
#
# INPUT: \@remainder, 
#	where @remainder is the BLAST output which are left with after
#	extracting the alignment information.
#	The lines are stored in the reverse order.
#
# OUTPUT: \@aux_info, where
#	@aux_info = ($lambda, $K, $H, $gpd_lambda, $gpd_K, $gpd_H).
#
sub extract_auxiliary_info (\@) {

	my ($remainder) = @_;


	while (my $line = pop @{$remainder}) { ($line =~ m/^Lambda/) and last; }

	my $line = pop @{$remainder};
	unless ($line ) {
	   print STDERR "ILLEGAL INPUT for the subroutine 'extract_auxiliary_info'.\n";
	   return undef;
	}

	my ($lambda, $K, $H) = split ' ', $line;

	unless ($H) {
	   print STDERR "ILLEGAL INPUT for the subroutine 'extract_auxiliary_info'.\n";
	   return undef;
	}

	my @output = ($lambda, $K, $H);

	unless (@{$remainder} > 2) { 
	    print STDERR "NO AUXILIARY INFORMATION ON GAPPED ALIGNMENTS, nor other information either, for the subroutine 'extract_auxiliary_info'.\n";
	    return \@output;
	}

	$line = pop @{$remainder};

	if ($line =~ m/^Gapped/ ) {
	   $line = pop @{$remainder};
	   unless ($line =~ m/^Lambda/ ) {
		print STDERR "NO AUXILIARY INFORMATION ON GAPPED ALIGNMENTS for the subroutine 'extract_auxiliary_info'.\n";
		return \@output;
	   }
	   $line = pop @{$remainder};

	} elsif ($line =~ m/^Lambda/ ) {
	   $line = pop @{$remainder};
	} else {
	    print STDERR "NO SIGNS OF AUXILIARY INFORMATION ON GAPPED ALIGNMENTS for the subroutine 'extract_auxiliary_info'.\n";
	    return \@output;
	}

	my ($gpd_lambda, $gpd_K, $gpd_H) = split ' ' , $line;

	unless (defined $gpd_H) {
	    print STDERR "ILLEGAL AUXILIARY INFORMATION ON GAPPED ALIGNMENTS for the subroutine 'extract_auxiliary_info'.\n";
	    return \@output;
	}

	push @output, $gpd_lambda, $gpd_K, $gpd_H;

	return \@output;

}			# end of the subroutine "extract_auxiliary_info"


# initialize_subalign:
#	A satellite subroutine of 'separate_masked_aln'
#	which initializes the record for a subalignment.
#
# INPUT: (\@residues, \@res_nums, $init_frame), where
#	@residues = (residues (or gaps) in the initiation site),
#	@res_nums = (residue numbers, or the number of the left closest residue if the site is a gap, of the initiation site),
#	$init_frame = the coordinate of the initiation site
#		viewed in the frame of the alignment.
#
# OUTPUT: \@initiated_info, where
#	@initiated_info = (\@sub_align, \@init_res, [], $init_frame, $init_frame).
#
sub initialize_subalign (\@\@$) {

	my ($residues, $res_nums, $init_frame) = @_;

	my @subalign = my @end_res = ();
	foreach my $res (@{$residues}) { push @subalign, \$res; }
	my @init_res = @{$res_nums};

	my @init_info = (\@subalign, \@init_res, [], $init_frame, $init_frame);

	return \@init_info;

}		# end of the subroutine initialize_subalign #


# finalize_subalign:
#	This subroutine gives information on end points of
#	the sub-alignment the program has been processing so far.
#
# INPUT: (\@list_subalign, \@res_nums, $final_frame), where
#	@{$list_subalign[m-1]} = (\@align, \@init_res, \@end_res, $init_frame, $end_frame) 
#	for the m-th subalignment.
#	@res_nums = (residue numbers of the right neighbourhood in the sequences).
#	$final_frame is the coordinate (0-based) of the ending point
#	of the subalignment viewed in the frame of the alignment.
#
# OUTPUT: NULL
#	The program completes the array @{$list_subalign[$#list_subalign]}
#	by filling in its \@end_res, and $end_frame.
#
sub finalize_subalign (\@\@$) {

	my ($list_subalign, $res_nums, $final_frame) = @_;

	my @end_res = ();
	foreach my $resnum (@{$res_nums}) { push @end_res, $resnum - 1; }

	my $current_sbaln = $list_subalign->[$#{$list_subalign}];
	$current_sbaln->[2] = \@end_res;
	$current_sbaln->[4] = $final_frame;

	return 1;

}		# end of the subroutine finalize_subalign #


# separate_masked_aln:
#	This subroutine receives a multiple alignment and
#	separates the alignment into masked and unmasked regions.
#
# INPUT: (\@align, \@init_coords, $char_mask, $char_gap), where
#	${$align->[m-1]} is the m-th sequence embedded in the alignment
#	$init_coords->[m-1] is the initial coordinate (residue number)
#	 of the alignment for the m-th sequence.
#	$char_mask is the character representing masked regions.
#
# OUTPUT: (\@unmasked, \@masked), where
#	@{$(un)masked->[m-1]} =
#	(\@subalign, \@init_res, \@end_res, $init_frame, $end_frame) 
#	for the m-th (un)masked region.
#	@init_res and @end_res are start and end residue numbers
#	in the subalignment.
#	$init_frame and $end_frame are coordinates (1-based)
#	of the start and end points of the region.
#
# CAUTION: This subroutine chews out the alignment file,
#	so you will be left with just remnants of the alignment, if any,
#	after the subroutine.
#
sub separate_masked_aln (\@\@$$) {

	my ($align, $init_coords, $char_mask, $char_gap) = @_;

	$char_mask = uc $char_mask;
	$char_gap  = uc $char_gap;
	my $ct_seq = @{$align};

	unless (@{$init_coords} == $ct_seq) {
	    printf STDERR "The number of initial coordinates CLASHES with that of sequences:  %d vs $ct_seq.\n", scalar (@{$init_coords});
	    return (undef, undef);
	}
	($ct_seq > 0) or (return ([], []) );

	my $len_aln = length ${$align->[0]}; # take the mimimum length
	for (my $k=1; $k<$ct_seq; $k++) {
	    my $len = length ${$align->[$k]};
	   ($len_aln > $len) and ($len_aln = $len);
	}

	my $fl_lmasked =0; # whether the left neighborhood is masked or not.
	my @masked = my @unmasked = ();

	my @res_nums = @{$init_coords}; # The residue number of the current site,
		# or of the right nearest residue if the current site is a gap.

	for (my $l=0; $l<$len_aln; $l++) {

	    my @residues = ();
	    my $fl_masked = 0;
	    for (my $k=0; $k<$ct_seq; $k++) {
		my $res = uc substr (${$align->[$k]}, 0, 1, '');
		($char_mask eq $res) and ($fl_masked = 1);
		push @residues, $res;
	    }

	    if ($fl_masked > 0) { # the current site is masked 

		if ($l==0) { # the beginning of the alignment
		    my $sub_aln_info
			= initialize_subalign (@residues, @res_nums, $l+1);
		    push @masked, $sub_aln_info;

		} elsif ($fl_lmasked > 0) { # the status is continuing
		    my $curr_sbaln = $masked[$#masked]->[0];
		    for (my $k=0; $k<$ct_seq; $k++) {
			${$curr_sbaln->[$k]} .= $residues[$k];
		    }

		} else {
		    my $sub_aln_info
			= initialize_subalign (@residues, @res_nums, $l+1);
		    push @masked, $sub_aln_info;
		    
		   finalize_subalign (@unmasked, @res_nums, $l);
		}


	    } else { # the current site is NOT masked

		if ($l==0) { # the beginning of the alignment
		    my $sub_align_info
			= initialize_subalign (@residues, @res_nums, $l+1);
		    push @unmasked, $sub_align_info;

		} elsif ($fl_lmasked > 0) { # the left neighbour is masked
		    my $sub_align_info
			= initialize_subalign (@residues, @res_nums, $l+1);
		    push @unmasked, $sub_align_info;

		    finalize_subalign (@masked, @res_nums, $l);

		} else { # the status is continuing

		   my $curr_sbaln = $unmasked[$#unmasked]->[0];
		   for (my $k=0; $k<$ct_seq; $k++) {
			${$curr_sbaln->[$k]} .= $residues[$k];
		   }
		}
	    }

	    for (my $k=0; $k<$ct_seq; $k++) {
		($residues[$k] eq $char_gap) or ($res_nums[$k]++);
	    }
	    $fl_lmasked = $fl_masked;
	}

	if ($fl_lmasked > 0) { # the last site is masked
	    finalize_subalign (@masked, @res_nums, $len_aln);

	} else {
	   finalize_subalign (@unmasked, @res_nums, $len_aln);
	}

	return (\@unmasked, \@masked);

}		# end of the subroutine separate_masked_aln #


# score_pairalign:
#	This subroutine calculates the score of a pairwise alignment.
#
# INPUT: (\$qu_seq, \$sb_seq, \%score_matrix,
#		$char_delim, $gapopen, $gapext, $char_gap), where
#	$qu_seq and $sb_seq are query and subject (sub)sequences in the alignment;
#	%score_matrix = ( $residue1 . $char_delim . $residue2 => $score, ...)
#	is a score matrix used for the homology search;
#	$gapopen and $gapext are the gap opening and gap extension penalties,
#	We will use the scoring scheme:
#	   {gap penalty} = $gapopen + $gap_length * $gapext .
#	   ##{gap penalty} = $gapopen + ($gap_length -1) * $gapext .
#	$char_gap is the character representing a gap.
#
# OUTPUT: \@info_align, where
#	@info_align = (#{total_sites}, #{res}_qu, #{res}_sb, 
#	   #{identical_res}, #{gaps}, total_score) for the pairwise alignment.
#	Here the total_score is the total raw score of the alignment.
#
sub score_pairalign (\$\$\%$$$$) {

	my ($qu_seq, $sb_seq, 
		$score_matrix, $delim, $gapopen, $gapext, $char_gap) = @_;

	$char_gap = uc $char_gap;
	my $cp_qu = uc ${$qu_seq};
	my $cp_sb = uc ${$sb_seq};
	my $len_qu = length ($cp_qu);
	my $len_sb = length ($cp_sb);

	unless ($len_qu == $len_sb) {
	    print STDERR "[CAUTION] query and subject sequences embedded in the alignment are NOT of EQUAL LENGTH.\n";
	}

	my $len_aln = ($len_qu < $len_sb) ? $len_qu : $len_sb; # take the shorter one 

	# whether a gap is the left neighborhood of the current site or not.
	my $fl_qu_lgap = my $fl_sb_lgap = 0;

	my $tot_sites = my $ct_res_qu = my $ct_res_sb
	= my $ct_id = my $ct_gap = my $tot_score = 0;

	for (my $l=0; $l<$len_aln; $l++) { # the outer for-loop #

	   my $qu_char = substr ($cp_qu, 0, 1, '');
	   my $sb_char = substr ($cp_sb, 0, 1, ''); 
	   my $pair = $qu_char . $delim . $sb_char ;

		# Examine the status of the current site #
	   my $fl_qugap = ($qu_char eq $char_gap) ? 1 : 0;
	   my $fl_sbgap = ($sb_char eq $char_gap) ? 1 : 0;

	   if ($fl_qugap >0) { # the query site is occupied with a gap

		if ($fl_sbgap > 0) { # the subject site is also occupied with a gap

		    print STDERR "[ATEENSION] An UNEXPECTED SITE CONFIGURATION: both the query and the subject are occupied with gaps !!\n";
		    next;  # skip the site without doing anything

		} else {
		   $ct_res_sb++;
		   $ct_gap++;
		   $tot_score -= ($fl_qu_lgap) ?  $gapext : $gapopen + $gapext ;
		}

	   } else {
		$ct_res_qu++;

		if ($fl_sbgap > 0) { # the subject site is occupied with a gap
		    $ct_gap++;
		    $tot_score -= ($fl_sb_lgap) ? $gapext : $gapopen + $gapext ;

		} else {
		    $ct_res_sb++;
		    $tot_score += $score_matrix->{$pair};
		    ($qu_char eq $sb_char) and ($ct_id++);
		}
	   }

	   $tot_sites++;
	   $fl_qu_lgap = $fl_qugap;
	   $fl_sb_lgap = $fl_sbgap;

	} 			# end of the outer for-loop #

	return [$tot_sites, $ct_res_qu, $ct_res_sb, $ct_id, $ct_gap, $tot_score];

}		# end of the subroutine score_pairalign #


# examine_unmasked_regions_pw:
#	This subroutine examines regions in a BLAST hit
#	which remain unmasked by other programs or subroutines.
#
# INPUT: (\$qu_seq, \$sb_seq, $qu_init, $sb_init, $char_mask, $char_gap,
#		\%score_matrix, $char_delim, $gapopen, $gapext), where
#	$qu_seq and $sb_seq are query and subject (sub)sequences in the alignment;
#	$char_mask and $char_gap are characters representing masked residues and gaps;
#	%score_matrix = ( $residue1 . $char_delim . $residue2 => $score, ...)
#	is a score matrix used for the homology search;
#	$gapopen and $gapext are the gap opening and gap extension penalties.
#	We will use the scoring scheme:
#	   {gap penalty} = $gapopen + $gap_length * $gapext .
#	   #{gap penalty} = $gapopen + ($gap_length -1) * $gapext .
#
# OUTPUT: \@info_unmasked_align, where
#	@info_unmasked_align = (total_length, #{res}_qu, #{res}_sb, 
#	  #{identical_res}, #{gaps}, total_score) for the unmasked regions.
#	The total_score here is the total raw score of the alignment
#	made by concatenating UNmasked regions.
#
sub  examine_unmasked_regions_pw (\$\$$$$$\%$$$) {

	my ($qu_seq, $sb_seq, $qu_init, $sb_init,$char_mask, $char_gap,
		$score_matrix, $delim, $gapopen, $gapext) = @_;

	my @aln = ($qu_seq, $sb_seq);
	my @inits = ($qu_init, $sb_init);

	# @{${un}masked_alns->[m-1]}
	# = (\@align, \@init_res, \@end_res, $init_frame, $end_frame)
	# for the m-th (un)masked block.
	#
	my ($unmasked_alns, $masked_alns)
		= separate_masked_aln (@aln, @inits, $char_mask, $char_gap);

	unless (defined $unmasked_alns) {
	    print STDERR "The subroutine 'separate_masked_aln' FAILED.\n";
	    return undef;
	}

	my @frames = my @qu_intervals = my @sb_intervals = (); # element = "$init:$end"
	my $qu_unmasked = my $sb_unmasked = '';

	foreach my $block_info (@{$unmasked_alns}) {
	    my ($aln, $init_res, $end_res, $init_frame, $end_frame)
		= @{$block_info};
	    $qu_unmasked .= ${$aln->[0]};
	    $sb_unmasked .= ${$aln->[1]};

	    push @frames, ($init_frame . ':' . $end_frame );
	    push @qu_intervals, ($init_res->[0] . ':' . $end_res->[0]);
	    push @sb_intervals, ($init_res->[1] . ':' . $end_res->[1]);
	}

	# @info_unmasked_align = (total_length, #{res}_qu, #{res}_sb, 
	#  #{identical_res}, #{gaps}, total_score) for the unmasked regions.
	#
	my $info_unmasked_aln
	  = score_pairalign ($qu_unmasked, $sb_unmasked,
		%{$score_matrix}, $delim, $gapopen, $gapext, $char_gap);

	unless (defined $info_unmasked_aln) {
	    print STDERR "The subroutine 'score_pairaln' FAILED.\n";
	    return undef;
	}

	return $info_unmasked_aln ;
	##return ($info_unmasked_aln, $unmasked_alns, $masked_alns) ; ### for testing the script ###

}		# end of the subroutine examine_unmasked_regions_pw #


=on construction
# examine_masked_regions_pw:
#	This subroutine examines only regions in a BLAST hit
#	which are suspected to have been masked by other programs or subroutines.
#
# INPUT: (\$qu_seq, \$sb_seq, $char_mask, $char_gap,
#		\%score_matrix, $char_delim, $gapopen, $gapext), where
#	$qu_seq and $sb_seq are query and subject (sub)sequences in the alignment;
#	$char_mask and $char_gap are characters representing masked residues and gaps;
#	%score_matrix = ( $residue1 . $char_delim . $residue2 => $score, ...)
#	is a score matrix used for the homology search;
#	$gapopen and $gapext are the gap opening and gap extension penalties.
#	We will use the scoring scheme:
#	   {gap penalty} = $gapopen + ($gap_length -1) * $gapext .
#
# OUTPUT: \@info_masked_regions, where
#	@info_masked_regions = (total_length, #{res}_qu, #{res}_sb, 
#		#{identical_res}, #{gaps}, total_score) for the masked regions.
#	It calculates the total_score so that it gives the difference
#	between the total (raw) score of the whole alignment
#	and the total (raw) score of the alignment which is a concatenation
#	of the NON-masked regions.
#
sub examine_masked_regions_pw (\$\$$$\%$$$) {

	my ($qu_seq, $sb_seq, $char_mask, $char_gap,
		$score_matrix, $delim, $gapopen, $gapext) = @_;

	$char_mask = uc $char_mask;
	$char_gap = uc $char_gap;
	my $cp_qu = uc ${$qu_seq};
	my $cp_sb = uc ${$sb_seq};
	my $len_qu = length ($cp_qu);
	my $len_sb = length ($cp_sb);

	my $len_aln = ($len_qu < $len_sb) ? $len_qu : $len_sb; # take the shorter one 

	# whether a masked residue is the left neighborhood of the current site or not.
	my $fl_qu_lmask = my $fl_sb_lmask =0;

	# whether a gap the left neighborhood of the current site or not.
	my $fl_qu_lgap = my $fl_sb_lgap = 0;

	# whether the gap is extending from the upstream flanking site of the masked region.
	my $fl_qu_extgap = my $fl_sb_extgap = 0;

	my $tot_len = my $ct_res_qu = my $ct_res_sb
	= my $ct_id = my $ct_gap = my $tot_score = 0;

	for (my $l=0; $l<$len_aln; $l++) { # the outer for-loop #

	   my $qu_char = substr ($cp_qu, 0, 1, '');
	   my $sb_char = substr ($cp_sb, 0, 1, ''); 
	   my $pair = $qu_char . $delim . $sb_char ;

		# Examine the status of the current site #
	   my $fl_qugap = ($qu_char eq $char_gap) ? 1 : 0;
	   my $fl_qumask = ($qu_char eq $char_mask) ? 1 : 0;
	   my $fl_sbgap = ($sb_char eq $char_gap) ? 1 : 0;
	   my $fl_sbmask = ($sb_char eq $char_mask) ? 1 : 0;

	   if ($fl_qugap >0) {
		if ($fl_sbgap > 0) {
		    print STDERR "An UNEXPECTED CONFIGURATION: both the query and the subject are filled with gaps !!\n";
		} elsif ($fl_sbmask > 0) {
		   $ct_gap++;
		   $tot_score -= ($fl_qu_lgap) ?  $gapext : $gapopen ;
		}

	   } elsif ($fl_qumask > 0) {
		if ($fl_sbgap > 0) {
		    $ct_gap++;
		    $tot_score -= ($fl_sb_lgap) ? $gapext : $gapopen ;

		} elsif ($fl_sbmask > 0) {
		    $tot_score += $score_matrix->{$pair};
		    $ct_id++;
		} else { # subject site is an ordinari residue  
		    $tot_score += $score_matrix->{$pair};
		}

	   } else { # query site is an ordinary residue

		if ($fl_sbmask > 0) {
		    $tot_score += $score_matrix->{$pair};
		}
	   }

	   $fl_qu_lgap = $fl_qugap;
	   $fl_qu_lmask = $fl_qumask;
	   $fl_sb_lgap = $fl_sbgap;
	   $fl_sb_lmask = $fl_sbmask;

	   if ( ($fl_qumask == 0) and ($fl_sbmask == 0) ) {
		# The currernt site is  NOT MASKED.

		# --> can serve as an upstream site of the next masked region.
		$fl_qu_upgap = $fl_qugap;
		$fl_sb_upgap = $fl_sbgap;
	   }

	} 			# end of the outer for-loop #

}		# end of the subroutine masked_regions_pw #
=cut

1;			# a spell to enable the module to be used or required
