package MyInferAncStates_parser_dawg_v1;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyInferAncStates_parser_dawg_v1.pm:
#	This slightly modified version of 'MyInferAncStates_parser_dawg.pm'
#	administers a 'first-aid' to cope with the old feature of the output of 'br_infer_psm_indel_events_dawg.pl'
#	that was made obsolete in the new version, 'br_infer_psm_indel_events(_spt_odr)_dawg.v2.pl'. 
#	(CREATED on Mar 25 (Tue), 2014)
#
# MyInferAncStates_parser_dawg.pm:
#	This package provides modules to parse the outputs of the program 'br_infer_psm_indel_events_dawg.pl',
#	which infers the ancestral states of the sequences simulated by Dawg.
#	It was created by modifying some modules in the package 'MyInferAncStates_parser.pm'.
#	(CREATED on Feb 28 (Fri), 2014)
#
#	TEMPORARILY REVISED on Mar 6 (Thu), 2014:
#	The subroutine 'parse_brief_output_infer_psm_indel_events_dawg' can now tentively skip
#	the gap-block cluster with "#{scenarios} = 0" and "Min_Total#{indel_events} = 0",
#	which is likely to be due to the cluster consisting ONLY OF a NULL BLOCK,
#	which cannot be properly handled by the current version of 'MyTreeMap_indel_ML.pm'.
#	
#
# MyInferAncStates_parser.pm:
#	This package provides modules to parse the outputs of programs inferring the ancestral states
#	of the sequences, such as those in PAML and the Perl script, 'br_infer_psm_indel_events.pl'.
#	(CREATION STARTED on Nov 23 (Sat), 2013, ENDED on Nov 26 (Tue), 2013)
#
#	MODIFIED on Dec 31 (Tue), 2013: 
#	A SERIOUS BUG, which concerns the complementary bases, was fixed in the subroutine 'revcom_anc_base_state_distr'.
#


use strict;
use warnings;

use MyArrayManipu qw( diff_array union_array );
use MyFile_manipu qw( open_input_file2 );
use MyTaxonomy qw( fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
		 fetch_ext_offsprings fetch_all_offsprings );

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg read_tree_file_dawg input_subtree_dawg
	);

=for the predecessor 'MyInferAncStates_parser.pm'
use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file read_tree_file input_subtree
	input_tree_with_nodeids
	make_node2children
	);
	## copy_tree
	## output_tree draw_subtree
	## br_output_tree br_draw_subtree topo_subtree
	## br_output_tree_sep br_draw_subtree_sep
	## make_node2parent

use MyTreeManipu_II_spe qw(
	replace_nodeids
	input_ext_tree_file
	);
	## read_ext_tree_file input_ext_subtree
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	parse_brief_output_infer_psm_indel_events_dawg
	);
	# map_treenode2treenode
	# revcom_distr_psm_indel_events
	# linear_order_indel_state_distributions

=for the predecessor 'MyInferAncStates_parser.pm'
sub reconstruct_rst_tree ($$$);
=cut


our $TOP_NODE = $MyTreeManipu_II_dawg::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II_dawg::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

=for the package 'MyTreeManipu_II.pm'
our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.
=cut



# parse_brief_output_infer_psm_indel_events_dawg:
#	This subroutine was created by modifying 'parse_brief_output_infer_psm_indel_events'
#	so that it will be cut out for the output of the Perl script 'br_infer_psm_indel_events_dawg.pl'.
#	(CREATED on Feb 28 (Fri), 2014)
#
#	TEMPORARILY REVISED on Mar 6 (Thu), 2014:
#	The subroutine now copes with a gap-block cluster with "Min_Total#{indel_events} = 0; #{scenarios} = 0",
#	which is likely because the cluster consists ONLY of a NULL BLOCK,
#	by assigning an EMPTY SCENARIO DISTRIBUTION.
#
# parse_brief_output_infer_psm_indel_events:
#	This subroutine parses the brief output of the Perl script 'br_infer_psm_indel_events.pl',
#	which calculates the 'probability distribution' of ancestral gap states (and indel scenarios),
#	based on the parsimonious indel scenarios.
#
# INPUT: (\$infile, $fl_full_nodes), where
#	$infile is the path to the input file, which is an brief output of 'br_infer_psm_indel_events.pl';
#	$fl_full_nodes determines on what nodes the ancestral states should be stored in the output:
#	$fl_full_nodes = 0 when you need only the state distribution on the top node,
#	               = 1 when you need those on all interior nodes,
#		       = 2 when you need those on all nodes (including exterior nodes).
#
# OUTPUT: (\@parameters, \@info_segments, \@indel_state_distributions), where
#	@parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with
#	[ @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr) for the full MSA, ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
#	@full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name) for the full MSA,
#	@seqids storing the sequence IDs,
#	$ct_segments = scalar (@info_segments) = scalar (@indel_state_distributions);
#
#	[ @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, \@red_tree), with ] --- replaced with:
#	[ @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
#	@{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_name), with 
#	$seg_start/end in the 0-based half-open convention,
#	[ @red_tree = ($rtopnode_id, \%rnode2pa, \%rnode2ch, \%rname2id, \%rid2name, \%rbr_attr) for the MSA reduced
#	according to the usability pattern; ] --- obsolete.
#	[ $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
#	 ($red_topnode_id = undef for the segment skipped to infer ancestral states) ; ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
#	$red_topnode_name is the name of the top-node (root) of the tree reduced according to the usability pattern
#	 ($red_topnode_name = undef for the segment skipped to infer ancestral states) ;
#
#	@sw_indel_state_distributons = @{$indel_state_distributons[$s]}
#	stores the indel state distributions for the $s th segment, and
#
#	@{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
#	for the gapless block left-flanking the $cl th gap-block cluster
#	(or on the right-end of the segment (with $cl = #{clusters})),
#	or (always with $cl = 0) for the skipped segment due to inadquate information.
#	$state	= 'X' for a gapless block,
#		= '?' for a skipped segment due to inadequate information.
#
#	@{$sw_indel_state_distributions[2*$cl + 1]}
#	= ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
#	is for the $cl th gap-block cluster, where
#
#	@{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and
#
#	@{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
#	 is for the $ps th parsimonious indel scenario, with
#	
#	[ %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
#	This subroutine parses the brief output of the Perl script 'br_infer_psm_indel_events.pl',
#	%node2bw_gap_states = ($node_name => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
#	[ %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
#	%br2indel_events = ($branch_lower_end_name => \@events_on_branch, ...) only for branches experiencing indel events, with
#	@{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.
#
sub parse_brief_output_infer_psm_indel_events_dawg (\$$) {

	my ($infile, $fl_full_nodes) = @_;
	local $/ = ''; # Use blank lines as input-record-separators. #

	my $fh_in = open_input_file2 (${$infile});
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] 'open_input_file2' FAILED on the input file '${$infile}': $! !\n";
	    return (undef, undef, undef); 
	}

		## Read the input sequence tree. ##

	my $top_node_name0;
	##my $top_node_id0;

	while (my $in = <$fh_in>) {
	    ($in =~ m/^\[\[\s*Input\s+Tree/) or next;
	    ##($in =~ m/^\[\[\s*Input\s+tree/) or next; # A BUG fixed on Nov 26 (Tue), 2013. #

	    ($top_node_name0) = ($in =~ m/top_node_name\s*\=\s*([^\s\(\)]+)/);
	    ##($top_node_id0) = ($in =~ m/top_node_ID\s*\=\s*([^\s\(\)]+)/);
	    last;
	}
	unless (defined $top_node_name0) {
	##unless (defined $top_node_id0) {...}
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT FIND the top-node name: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $full_tree = <$fh_in>;

	my %fnode2pa; my %fnode2ch; my %fname2id; my %fid2name; my %fbr_attr;
	my %fname2id_intnode; my $ftopnode_name;

	unless ( input_tree_file_dawg ($full_tree, %fnode2pa, %fnode2ch, %fname2id, %fid2name, %fbr_attr, %fname2id_intnode, $ftopnode_name) ) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The subroutine 'input_tree_file_dawg' FAILED to extract the full tree: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}
=for the predecessor 'parse_brief_output_infer_psm_indel_events'.
	unless ( input_ext_tree_file ($full_tree, %fnode2pa, %fnode2ch, %fname2id, %fid2name, %fbr_attr, 1) ) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The subroutine 'input_ext_tree_file' FAILED to extract the full tree: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}
=cut

	my $ftopnode_id = $fname2id{$TOP_NODE};
	unless ($ftopnode_name eq $top_node_name0) {
	##unless ($ftopnode_id eq $top_node_id0) {...} #For the predecessor 'parse_brief_output_infer_psm_indel_events'.
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The top-node name in the label (=$top_node_name0) CONFLICTS with the name displayed in the tree (=$ftopnode_name): $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

		# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #

	my @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name);

		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name) for the full MSA,

=for the predecessor 'parse_brief_output_infer_psm_indel_events'.
	my @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr);
		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr) for the full MSA,
=cut

		## Read the MSA information ##

	my $label_msa = <$fh_in>;

	my ($ct_seqs0) = ($label_msa =~ m/\(\#\{sequences\}\s*\=\s*(\d+)\)/);
	unless (defined $ct_seqs0) {
	    $label_msa =~ s/\s+$//;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of sequences canNOT be extracted from the label: $!\n";
	    print STDOUT "'$label_msa' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	(my $list_seqids = <$fh_in>) =~ s/\s*\.\s+$//; 
	##(my $list_seqids = <$fh_in>) =~ s/\.\s+$//; # A BUG fixed on Dec 19 (Thu), 2013. #
	$list_seqids =~ s/^Sequences\:\s+//;
	my @seqids = split /\,\s*/, $list_seqids;

	unless (@seqids == $ct_seqs0) {
	    my $ct_seqids = @seqids;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of sequences in the label (=$ct_seqs0) CONFLICTS with $ct_seqids, which is the number of IDs in the list: $! !\n";
	    print STDOUT ' ' x 3, join (', ', @seqids), " .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_msa_len = <$fh_in>;
	my ($msa_len) = ($descr_msa_len =~ m/^\#\{columns\s+in\s+MSA\}\s*\=\s*(\d+)/);
	unless (defined $msa_len) {
	    $descr_msa_len =~ s/\s+$//;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The length of the MSA canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_msa_len' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_tgt_indx = <$fh_in>;
	my ($tgt_seqid, $indx_tgt) = ($descr_tgt_indx =~ m/^\{Index\s+for\s+the\s+target\s+sequence\s+\'([^\s\']+)\'\}\s*\=\s*(\S+)/ );
	unless (defined $indx_tgt) {
	    $descr_tgt_indx =~ s/\s+$//; 
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The index of the target sequence canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_tgt_indx' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_ct_segs = <$fh_in>;
	my ($ct_segments) = ($descr_ct_segs =~ m/^\#\{MSA\s+Segments\}\s*\=\s*(\d+)/);
	unless (defined $ct_segments) {
	    $descr_ct_segs =~ s/\s+$//; 
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of MSA segments canNOT be extracted from the line: $! !\n";
	    print STDOUT "'$descr_ct_segs' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my @parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments);

		# @parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with


		### MAIN PROCESS: Read the cluster-wise distributions of the indel scenarios
		### (and the associated ancestral gap states), which were bundled into segments.

	my @info_segments = my @indel_state_distributions = ();

	while (my $head_seg = <$fh_in>) { # Main outer while-loop (over segments) #

	    my ($seg_no) = ($head_seg =~ m/^\<\<\<\<\s+The\s+(\d+)\s+th\s+Segment/) ;
	    unless (defined $seg_no) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment no. from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my ($seg_start, $seg_end) = ($head_seg =~ m/Region\s+\=\s+\[(\d+)\,\s*(\d+)\)/);
	    unless (defined $seg_end) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my ($us_pattern) = ($head_seg =~ m/Usability\s+pattern\s+\=\s*\'([^\s\']+)\'/);
	    unless (defined $us_pattern) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my @seg_info = ($seg_start, $seg_end, $us_pattern);
	    $info_segments[$seg_no] = \@seg_info;

		# @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
		# $seg_start/end in the 0-based half-open convention,
		# $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
		#  ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;

	    my $second_chunk = <$fh_in>;

	    if ($second_chunk =~ s/^\[\[\s+Reduced\s+Tree\s+//) {

			# Normal case. #

			# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
		my ($red_topnode_name) = ($second_chunk =~ m/\(top_node_name\s*\=\s*([^\s\(\)]+)\)/);
		##my ($red_topnode_id) = ($second_chunk =~ m/\(top_node_ID\s*\=\s*([^\s\(\)]+)\)/);
		unless (defined $red_topnode_name) {
		##unless (defined $red_topnode_id) {...}
		    $second_chunk =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the top-node name for the reduced tree from the second chunk: $! !\n";
		    print STDOUT "'$second_chunk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}
		push @seg_info, $red_topnode_name;
		##push @seg_info, $red_topnode_id;

			# Go ahead with reading the indel scenario distribution.

	    } elsif ($second_chunk =~ m/^SKIPPED\s+INFERRING/) {
	    ##} elsif ($second_chunk =~ m/^SKIPPED\s+INDERRING/) { # A BUG fixed on Nov 26 (Tue), 2013. #

			# The ancestral state inference was skipped for this segment. #

		my @sw_indel_state_distributions = ( [$seg_start, $seg_end, '?'] );
		$indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		next;	# Go ahead to the next segment.
		
	    } else {
			# An Unexpected Case. #

		$second_chunk =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: an ANOMALOUS SECOND CHUNK for the $seg_no th segment: $! !\n";
		print STDOUT "'$second_chunk' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my $red_topnode_name = $seg_info[3];
	    ##my $red_topnode_id = $seg_info[3];

	    my $red_tree = <$fh_in>; # Emit the chunk for the reduced tree.
	    my $heading_main = <$fh_in>; # Emit the heading that tells that the main description begins. 

	    my @sw_indel_state_distributions = ();

		# A FIRST-AID administered in creation of 'MyInferAncStates_parser_dawg_v1.pm' (on Mar 25 (Tue), 2014). #
	    my $reservoir;
	    while (my $head_cls = (defined $reservoir) ? $reservoir : <$fh_in>) { # Main middle while-loop (over clusters) #
		 (defined $reservoir) and ($reservoir = undef);
	    ##while (my $head_cls = <$fh_in>) { ...} # Main middle while-loop (over clusters) #

		if ($head_cls =~ s/^\<\s+Gapless\s+block\s+// ) {

			# Gapless block. #

		    my ($bl_start, $bl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		    unless (defined $bl_end) {
			$head_cls =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the block-end coordinate from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			print STDOUT "'< Gapless block $head_cls' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d+)/) ;
		    ##my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d)/) ; # A BUG fixed on Nov 26 (Tue), 2013. #

		    unless (defined $rfl_cl_no) {
			if ($head_cls =~ m/^on\s+the\s+right\-end/) {

				# The right-end of the segment. #

			    push @sw_indel_state_distributions, [$bl_start, $bl_end, 'X'];

			    last; # End the middle while-loop.

			} else {
				# An unexpected case. #

			    $head_cls =~ s/\s+$//;
			    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the right-flanking cluster number from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			    print STDOUT "'< Gapless block $head_cls' .\n\n";
			    close ($fh_in);
			    return (undef, undef, undef);
			}

		    } # END of "unless (defined $right_cl_no) {...}"

		    unless (@sw_indel_state_distributions == 2* $rfl_cl_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ") is NOT EQUAL to twice the number of the right-flanking cluster (=$rfl_cl_no): $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    $sw_indel_state_distributions[2*$rfl_cl_no] = [$bl_start, $bl_end, 'X'];

		    next; # Go ahead to the gap-block cluster on the right. 

			# @{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
			# for the gapless block left-flanking the $cl th gap-block cluster
			# (or on the right-end of the segment (with $cl = #{clusters})),
			# or (always with $cl = 0) for the skipped segment due to inadquate information.
			# $state	= 'X' for the gapless block,
			# 	= '?' for the skipped segment due to inadequate information.

		} # END of "if ($head_cls =~ s/^\<\s+Gapless\s+block\s+/) {...}"


			## Read information on the gap-block cluster. ##

		my ($cl_no) = ($head_cls =~ m/^\<\s+The\s+(\d+)\s+th\s+cluster/ );
		unless (defined $cl_no) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the cluster number from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		unless (@sw_indel_state_distributions == 2 * $cl_no + 1) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The temporary size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ') is NOT EQUAL to 2 * $cl_no + 1, where $cl_no = ', "$cl_no, for the $seg_no th segment: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($cl_start, $cl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		unless (defined $cl_end) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the end-coordinate of the cluster from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($ct_gblks) = ($head_cls =~ m/\#\{gap\-blocks\}\s+\=\s+(\d+)/ );
		unless (defined $ct_gblks) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the number of gap-blocks from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read information on the gap-blocks. #

		my $head_gblocks = <$fh_in>;
		my $fld_tags_gblocks = <$fh_in>;
		my $info_gblocks = <$fh_in>;

		my @gap_blocks = ();

		foreach my $gblock (split /\n+/, $info_gblocks) { # 1st inner foreach-loop (over gap-blocks) #
		     my ($bl_no, $bl_len, $bl_start, $bl_end, $indx_gp, $gpattern) = split /\s+/, $gblock;
		     push @gap_blocks, [$bl_start, $bl_end, $indx_gp, $gpattern];

			# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and

		}					# End of the 1st inner foreach-loop (over gap-blocks) #

		unless (@gap_blocks == $ct_gblks) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@gap_blocks (=', scalar (@gap_blocks), ") CONFLICTS with the number of gap-blocks (=$ct_gblks) described in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read the distribution of parsimonious scenarios for a gap-block cluster. #

		my $heading_list_psm_scns = <$fh_in>;

			# MODIFIED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) #
		my ($min_totct_indels) = ($heading_list_psm_scns =~ m/Min_Total\#\{indel_events\}\s+\=\s+([^\;\s]+)/ );
		##my ($min_totct_indels) = ($heading_list_psm_scns =~ m/Min_Total\#\{indel_events\}\s+\=\s+(\d+)/ );
		unless (defined $min_totct_indels) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the minimum total count of indel events from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# MODIFIED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 11 (Tue), 2014.) #
		my ($ct_psm_scns) = ($heading_list_psm_scns =~ m/\#\{scenarios\}\s+\=\s+([^\)\s]+)/ );
		##my ($ct_psm_scns) = ($heading_list_psm_scns =~ m/\#\{scenarios\}\s+\=\s+(\d+)/ );
		unless (defined $ct_psm_scns) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the count of parsimonious indel scenarios from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			## When encountering ($min_totct_indels eq 'N/A'),
			## assign an empty scenario distribution to the cluster.
			## (TEMPORARILY ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) ##

		if ($min_totct_indels eq 'N/A') { # Further Modified on Mar 11 (Tue), 2014. #
		##if (($min_totct_indels eq 'N/A') and ($ct_psm_scns == 0)) {...} 
		    print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] The $cl_no th gap-block cluster in the $seg_no th segment has Min_Total#{indel_events} = $min_totct_indels and #{scenarios} = $ct_psm_scns, probably because it contains (a) long gap(s): $! !\n\n";
		    my @scenario_distribution = (); # An empty scenario distribution !!
		    $sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];
		    my $info = <$fh_in>;
			# A FIRST-AID administered in creation of 'MyInferAncStates_parser_dawg_v1.pm' on Mar 25 (Tue), 2014. #
		    if ($info =~ m/^\<\s+Gapless\s+block/) {
			$reservoir = $info;
			next;
		    }
		    unless ($info =~ m/^\[INFORMATION\]/) {
			$info  =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] Found an UNEXPECTED CHUNK: $! !\n";
			print STDOUT "'$info' .\n\n";
		 	close ($fh_in);
			return (undef, undef, undef);
		    }
		    next;
		}

			## When encountering ($min_totct_indels == 0) and ($ct_psm_scns == 0),
			## assign an empty scenario distribution to the cluster.
			## (TEMPORARILY ADDED (as a first-aid) on Mar 6 (Thu), 2014.) ##

		if (($min_totct_indels == 0) and ($ct_psm_scns == 0)) {
		    if (($ct_gblks == 1) and ($gap_blocks[0]->[3] eq '-' x $ct_seqs0)) {
			print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] The $cl_no th gap-block cluster in the $seg_no th segment consists ONLY of a NULL BLOCK: $! !\n";
		    } else {
			print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] Somehow, the $cl_no th gap-block cluster in the $seg_no th segment has Min_Total#{indel_events} = 0 and #{scenarios} = 0: $! !\n";
		    }
		    my @scenario_distribution = (); # An empty scenario distribution !!
		    $sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];
		    next;
		}

		my @scenario_distribution = ();

		while (my $head_scn = <$fh_in>) { # 2nd inner whlie-loop (over parsimonious scenarios). #

		    my ($scn_no, $rel_prob) = ($head_scn =~ m/^\{\s+Scenario\s+no\.\s+(\d+)\s*\;\s+rel_prob\.\s*\=\s*([^\s\{\}]+)/ );
		    unless (defined $rel_prob) {
			$head_scn =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the relative probability from the following heading of a parsimonious scenario for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			print STDOUT "'$head_scn' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }
		    unless (@scenario_distribution == $scn_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The temporary size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the parsimonious scenario no. (=$scn_no) for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

			# Read block-wise gap states mapped onto the nodes. #

		    my $head_bw_gstates = <$fh_in>;
		    my $fld_tags_bw_gstates = <$fh_in>;
		    my $info_bw_gstates = <$fh_in>;

		    my %node2bw_gap_states;
		    my $ct_rlv_nodes = my $ct_gaponly = 0; # A measure to avoid a buggy behavior. (introduced on Nov 26 (Tue), 2013). #

		    foreach my $line (split /\n+/, $info_bw_gstates) { # 1st inner-inner foreach-loop (over nodes). #
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			my ($node_name, $cnct_bw_gstats) = split /\s+/, $line;
			##my ($node_id, $cnct_bw_gstats) = split /\s+/, $line;

				# Introduced (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
			$ct_rlv_nodes++;
			($cnct_bw_gstats =~ m/X/) or ($ct_gaponly++);

			if ($fl_full_nodes <= 0) {
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			    ($node_name eq $red_topnode_name) or next; # Skip except the top node.
			    ##($node_id eq $red_topnode_id) or next; # Skip except the top node.
			} elsif ($fl_full_nodes < 2) {
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			    (defined $fname2id_intnode{$node_name}) or next; # Skip an exterior node. #
			    ##my $children = $fnode2ch{$node_id};
			    ##((defined $children) and (@{$children}>0)) or next; # Skip an exterior node.
			}
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			$node2bw_gap_states{$node_name} = $cnct_bw_gstats;
			##$node2bw_gap_states{$node_id} = $cnct_bw_gstats;

				# %node2bw_gap_states = ($node_name => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
				# [ %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.

		    }						# End of the 1st inner-inner foreach-loop (over nodes). #

			# Read indel events mapped onto the branches. #

		    my $head_indels = <$fh_in>;
		    my $fld_tags_indels = <$fh_in>;

				# FURTHER MODIFIED (on Mar 9 (Sun), 2014) to cope with the new revision of 'MyTreeMap_indels.pm' that handles gap-only columns. #
		    my $info_indels = ($min_totct_indels > 0) ? <$fh_in> : '';
		    ##		#  MODIFIED (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
		    ##my $info_indels = ($ct_gaponly < $ct_rlv_nodes) ? <$fh_in> : '';
		    ####my $info_indels = <$fh_in>;

		    my %br2indel_events;

		    foreach my $line (split /\n+/, $info_indels) { # 2nd inner-inner foreach-loop (over indel-undergoing branches). #

			my @fields = split /\s+/, $line;
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			my $br_le_name = shift @fields;
			##my $br_id = shift @fields;

			my @events_on_branch = ();

			foreach my $fld (@fields) { # Inner-most foreach-loop (over indel events on the branch). # 
			    my ($stat_ue, $stat_le, $cnct_blks) = split /\//, $fld;
			    my @involved_blocks = split /\,/, $cnct_blks;
			    push @events_on_branch, [$stat_ue, $stat_le, \@involved_blocks];

			}		# End of the inner-most foreach-loop (over indel events on the branch). #

				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			$br2indel_events{$br_le_name} = \@events_on_branch;
			##$br2indel_events{$br_id} = \@events_on_branch;

			    # [ %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
			    # %br2indel_events = ($branch_lower_end_name => \@events_on_branch, ...) only for branches experiencing indel events, with
			    # @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event

		    }				# End of the 2nd inner-inner foreach-loop (over indel-undergoing branches). #

		    $scenario_distribution[$scn_no] = [$rel_prob, \%node2bw_gap_states, \%br2indel_events];


			# @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
			#  is for the $ps th parsimonious indel scenario, with

		    ($scn_no == $ct_psm_scns - 1) and last; # We know that this is the last scenario. #


		}			# End of the 2nd inner while-loop (over parsimonious scenarios). #

		unless (@scenario_distribution == $ct_psm_scns) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the number of parsimonious scenarios (=$ct_psm_scns) recorded in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}


		$sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];

			# @{$sw_indel_state_distributions[2*$cl + 1]}
			# = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
			# is for the $cl th gap-block cluster, where

	    }			# End of the main middle while-loop (over clusters) #

	    $indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		# @sw_indel_state_distributons = @{$indel_state_distributons[$s]}
		# stores the indel state distributions for the $s th segment, and

	}		# End of the main outer while-loop (over segments) #

	close ($fh_in);

=An earlier attempt (TO BE REMOVED if all are successful).
	push @indel_state_distributions, \@sw_indel_state_distributions;

	unless ((defined $seg_no) and ($seg_no == $ct_segments - 1)) {
	    (defined $seg_no) or ($seg_no = 'undef');
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The final segment number (=$seg_no) CLASHES with the number of MSA segments (=$ct_segments) minus one: $! !\n";
	    return (undef, undef, undef);
	}
=cut

	unless (@info_segments == $ct_segments) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@info_segments', " (=", scalar (@info_segments), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}
	unless (@indel_state_distributions == $ct_segments) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@indel_state_distributions', " (=", scalar (@indel_state_distributions), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}

	return (\@parameters, \@info_segments, \@indel_state_distributions);

}	# End of the subroutine 'parse_brief_output_infer_psm_indel_events_dawg'


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


