package MyPAML_parser;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPAML_parser.pm:
#	This module manipulates various inputs/outputs of programs
#	in the 'PAML' package.
#

use strict;
use warnings;

use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);

use MyFile_manipu qw(
	open_input_file open_output_file
	);
use MyArrayManipu qw(
	compare_interval_sets
	);
	## diff_array union_array nr_array

use MyTreeManipu_II qw( br_output_tree );
	## input_tree_file read_tree_file input_subtree copy_tree
	## make_node2children make_node2parent
	## move_topnode segregate_subset_nodes
	## resolve_multipartite
use MyMalnIO qw(
	pr_phylip_aln pr_ODEN_input_lite
	);
	## extract_clw pr_clustal_like_lite_aa
use MyMalnManipu_suppl qw(
	extract_gap_blocks pad_concate_blocks
	gap_indicator splice_blocks_in_aln
	);
	## rm_spec_sites 
use MyTreeMap qw(
	bch_infer_gapstat_on_nodes
	);
	## fractionate_by_ndepth assign_node2closest_exoffs
	## initialize_node2inout assign_node2bival_states

=set aside for later use
use MyArrayManipu_num qw( union_nmarray );
use MyTaxonomy qw(
	fetch_ext_offsprings fetch_all_offsprings
	);
	## fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
use MyTreeManipu_IV qw(
	remove_trivial_nodes
	);
	## rm_external_node construct_neat_subtree
use MyTreeManipu_II_spe qw(
	input_ext_tree_file
	);
use MyMalnManipu_pep_spe qw( init_consv_hashes );
=cut

our @EXPORT_OK = qw(
	readin_evol_model
	retrieve_aa_freqs preprocess_aln_paml
	prepare_paml_ctrl extract_ancestral_paml
	extract_topseq_paml pr_ancestral_paml
	reconstruct_topnode_paml
	parse_yn00_output
	) ;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	## Amino acids in the order of being recorded in the data file of a substitution model ##
our @AMINO_ACIDS = qw( A R N D C Q E G H I L K M F P S T W Y V );
our $CT_AAS = @AMINO_ACIDS;

# readin_evol_model:
#	This subroutine reads in an evolutionary model
#	stored in the "dat/" sub-directory of the PAML package.
#
# INPUT: \$datafile.
#
# OUTPUT: (\@aa_freqs, \@pre_score), where
#	$aa_freqs[$i] is the frequency of $AMINO_ACiDS[$i];
#	$pre_score->[$i][$j] (0 <= $j < $i < $CT_AAs) is
#	the ($i, $j) th component of the symmetric matrix S_ij (= S_ji).
#
sub readin_evol_model (\$) {

	my ($datafile) = @_;
	local $/ = "\n"; # Use the newline character as an input-record separator #

	my $fh_tmp = open_input_file (${$datafile});
	unless (defined $fh_tmp) {
	    print STDERR "[ERROR 'readin_evol_model'] 'open_input_file' FAILED on the data file '${$datafile}': $! !\n";
	    return (undef, undef);
	}

###=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'readin_evol_model' >\n\n";
	print STDERR "[ The 1st while-loop ]\n\n";
###=cut

	my @pre_score = ([]);
	<$fh_tmp>;# Skip the 1st line
	while (my $in = <$fh_tmp>) { # The 1st while-loop #
###=for DEBUGGING the subroutine
	    print STDERR "Input:\n";
	    print STDERR $in, "\n";
###=cut
		# Revised for DEBUGGING #
	    ($in =~ m/\S/) or last;
	    ##$in =~ s/\n$//;
	    ##($in =~ m/^$/) and last;

	    my @row = split ' ', $in; # Get all non-white-space fields separated by white spaces.
	    push @pre_score, \@row;
	}		# End of the 1st while-loop #

###=for DEBUGGING the subroutine
	print STDERR "\n[ The 2nd while-loop ]\n\n";
###=cut

	my @aa_freqs = ();
	while (my $in = <$fh_tmp>) { # The 2nd while-loop #
###=for DEBUGGING the subroutine
	    print STDERR "Input:\n";
	    print STDERR $in, "\n";
###=cut
	    ($in =~ m!^//!) and last; # End of the data part #
	    my @elements = split ' ', $in;
	    (@elements>0) or next;
	    push @aa_freqs, @elements;
	} 		 # End of the 2nd while-loop #

###=for DEBUGGING the subroutine
	print STDERR "\n\n --- 'readin_evol_model' ENDED ---\n\n";
###=cut

	return (\@aa_freqs, \@pre_score);

}	# End of the subroutine 'readin_evol_model' #


# retrieve_aa_freqs:
#	This subroutine retrieves the frequencies of the amino acids
#	from the PAML data file of a substitution model.
#
#	According to the format of PAML data file,
#	the 2nd chunk provides the amino-acid frequencies
#	in a specified order. So this subroutine fetches them.
#
# INPUT: (\$datafile, \%aa2freq), where
#	%aa2freq = ($amino_acid (a single captal letter) => $frequency, ...)
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub retrieve_aa_freqs (\$\%) {

	my ($datafile, $aa2freq) = @_;
	local $/ = "\n"; # Use blank lines as input-record-separators

	my $fh_tmp = open_input_file (${$datafile});
	unless (defined $fh_tmp) {
	    print STDERR "[ERROR 'retrieve_aa_freqs'] 'open_input_file' FAILED on the data file '${$datafile}': $! !\n";
	    return 0;
	}

	(my $matrix = <$fh_tmp>) =~ s/^\s+//;
	while (my $line = <$fh_tmp>) { # 1st while-loop #
	    ($line =~ m/\S/) or last;
	    $matrix .= $line;
	}		# End of the 1st while-loop #

	my $list_aa_freqs = '';
	while (my $line = <$fh_tmp>) { # 2nd while-loop #
	    ($line =~ m/\S/) or last;
	    $list_aa_freqs .= $line;
	}		# End of the 2nd while-loop #
	
	close ($fh_tmp);

=for DEBUGGING the subroutine
	print STDERR "\n{ subroutine 'retrieve_aa_freqs' }\n\n";
	print STDERR "\$matrix = \n";
	print STDERR "'$matrix' .\n\n";
	print STDERR " \$list_aa_freqs = \n";
	print STDERR "'$list_aa_freqs' .\n\n";
=cut

	my @aa_freqs = split /\s+/, $list_aa_freqs;
	unless (@aa_freqs == $CT_AAS) {
	    print STDERR "[ERROR 'retrieve_aa_freqs'] The number of frequencies (", scalar (@aa_freqs), ") CLASHES with the number of amino acids ($CT_AAS): $! !\n";
	    return 0;
	}

	for (my $i = 0; $i < $CT_AAS; $i++) { $aa2freq->{$AMINO_ACIDS[$i]} = $aa_freqs[$i]; }

	return 1;

}	# End of the subroutine 'retrieve_aa_freqs' #


# preprocess_aln_paml:
#	This subroutine pre-processes the input alignment for PAML
#	so as to reduce the times where the program falls into 'error's.
#	NOTE that this subroutine chews up the input sequences.
#
# INPUT: (\@seqs, \%replace), where
#	${$seqs[$k]} is the $k th sequence in the alignment;
#	%replace = ($to_be_replaced => $to_replace, ...) .
#
# OUTPUT: \@prepro_seqs, where
#	${$prepro_seqs[$k]} is the $k th sequence after the preprocessing.
#
sub preprocess_aln_paml (\@\%) {

	my ($seqs, $replace) = @_;

	my @prepro_seqs = ();

	foreach my $seq (@{$seqs}) { # The outer foreach-loop #
	    my $pp_seq = '';
	    my $len = length ${$seq};
	    for (my $l=0; $l<$len; $l++) { # The inner for-loop #
		my $char = substr (${$seq}, 0, 1, '');
		my $repl = $replace->{$char};
		$pp_seq .= (defined $repl) ? $repl : $char;
	    }			# End of the inner for-loop #

	    push @prepro_seqs, \$pp_seq;

	}			# End of the outer foreach-loop #

	return \@prepro_seqs;

}	# End of the subroutine 'preprocess_aln_paml' #


# prepare_paml_ctrl:
#	This subroutine prepares a control file for a program in the PAML package,
#	by revising a template in a manner specified by the input hash.
#
# INPUT: (\$template, \$outfile, \%revisions), where
#	$template is the path pointing to a template of the control file;
#	$outfile is the path of the control file to be used;
#	%revisions = ($variable => $value, ...)
#	lists variables you want to modify compared to the template.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub prepare_paml_ctrl (\$\$\%) {

	my ($template, $outfile, $revisions) = @_;
	local $/ = "\n";

		# If there is no revision, just copy the template #
	unless (0 < scalar keys %{$revisions}) {
	   if (system ("cp ${$template} ${$outfile}") ) {
	        return 1;
	   } else {
		print STDERR "[ERROR 'prepare_paml_ctrl'] 'system (cp ${$template} ${$outfile})' FAILED !\n";
		return 0;
	   }
	}

	my $fh_in = open_input_file (${$template});
	unless (defined $fh_in) {
	    print STDERR "[ERROR 'prepare_paml_ctrl'] FAILED to open the template:\n";
	    print STDERR "    '${$template}' !\n\n";
	    return 0;
	}
	my $fh_out = open_output_file (${$outfile}, 0);
	unless (defined $fh_out) {
	   close ($fh_in);
	   print STDERR "[ERROR 'prepare_paml_ctrl'] FAILED to open the output file:\n";
	   print STDERR "     '${$outfile}' !\n\n";
	   return 0;
	}

	my %already; # ( $variable => 1, ...) for the variables already modified #

	while (my $line = <$fh_in>) { # The main while-loop #

	    if ($line =~ m/^\s*\*/) { # keep comment lines
		print $fh_out $line ;
		next;
	    }
	    my ($spaces, $variable, $value) = ($line =~ m/^(\s*)(\S+)\s+\=\s+(\S+)/ );

		# leave the line intact if no change is required #
	    unless ((defined $variable) and (defined $value)
		   and (defined $revisions->{$variable})) {
		print $fh_out $line;
		next;
	    }

	    $already{$variable} = 1;

		# leave the line intact if the variable is expected to remain the same #
	    if ($revisions->{$variable} eq $value) {
		print $fh_out $line;
		next;
	    }

		# Modify the parameter as specified #
	    print $fh_out '*', $line;
	    print $fh_out $spaces, $variable, ' = ', $revisions->{$variable}, "\n";

	}			# End of the main while-loop #

	close ($fh_in);

		## Specify the remaining variables ##

	print $fh_out "\n", '* The remaining variables to be fixed', "\n\n"; 

	foreach my $variable (sort keys %{$revisions}) { # The subsidiary foreach-loop #
	    (defined $already{$variable}) and next;
	    
	     printf $fh_out "%12s = %s\n", $variable, $revisions->{$variable}; 

	}			# End of the subsidiary foreach-loop #

	close ($fh_out);

	return 1;

}	# end of the subroutine 'prepare_paml_ctrl' #


# extract_ancestral_paml:
#	This subroutine extracts information on ancestral states
#	from an 'rst' file output by 'baseml' or 'codeml'.
#	This subroutine reads in output for an input tree.
#
# INPUT: (\*${$fh_in}, $fl_full_anc), where
#	$fh_in is the filehandle for inputting
#	an 'rst' file produced by 'baseml' or 'codeml' of PAML;
#	$fl_full_anc = 1 if you want to extract distributions for ALL ancestral sequences,
#		     = 0 if you want to extract only distributions for the TOP sequence.
#	
# OUTPUT: (\$viewtree, \%prob_distri, \%reconst_seq, $next_tree_signpost), where
#	$viewtree is the input tree in the format readable by TreeView;
#	%prob_distri = ($node_label => \$prob_distri, ...) for the internal nodes,
#	with $prob_distri being a string (chunk) representing probability distributions of states on the node for sites;
#	%reconst_seq = ($node_label => \$seq, ...) for the internal nodes,
#	 with $seq being a DNA (or peptide) sequence reconstructed for the node;
#	$next_tree_signpost = undef if we are reading the final tree.
#
sub extract_ancestral_paml (\*$) {

	my ($fh_in, $fl_full_anc) = @_;
	local $/ = "\n";

	while (my $in = <$fh_in>) { # The 1st while-loop #
	    ($in =~ m/^tree\s+with\s+node\s+labels\s+for\s+Rod\s+Page/ ) and last;
	}		# End of the 1st while-loop #

		# Read in the input tree with ancestral nodes labelled #
	my $viewtree = '';
	while (my $in = <$fh_in>) { # The 2nd while-loop #
	   ($in =~ m/^\s*$/) and last;
	    $viewtree .= $in;
	}		# End of the 2nd while-loop #

	my ($top_label) = ($viewtree =~ m/(\d+)\s*\;\s*$/);
	unless (defined $top_label) {
	    print STDERR "[ERROR 'extract_ancestral_paml'] Found NO LABEL for TOP NODE !\n";
	    my $next_tree_signpost;
	    while (my $in = <$fh_in>) { # The innerwhile-loop #
		if ($in =~ m/^TREE\s+\#/) {
		    $next_tree_signpost = $in;
		    last;
		}
	    }			# End of the inner while-loop #
	    return ($viewtree, undef, undef, $next_tree_signpost);
	}

	while (my $in = <$fh_in>) { # The 3rd while-loop #
	   ($in =~ m/^\(1\)\s+Marginal\s+reconstruction\s+of\s+ancestral\s+sequences/) and last;
	}		# End of the 3rd while-loop #


		# Read in the probability distributions #
	my %prob_distri;
	while (my $in = <$fh_in>) { # The 4th while-loop #

	   ($in =~ m/^List\s+of\s+extant\s+and\s+reconstructed\s+sequences/) and last;

	   my ($nlabel) = ($in =~ m/^Prob\s+distribution\s+at\s+node\s+([^\s\,]+)\,\s+by\s+site/ ) ;
	    (defined $nlabel) or next;
	    ($fl_full_anc or ($nlabel eq $top_label)) or next;

	    for (1..3) { <$fh_in>; } # Skip the next 3 lines

	    my $pr_dist = '';
	    while (my $in = <$fh_in>) {
		($in =~ m/^\s*$/) and last;
		$pr_dist .= $in;
	    }
	    $prob_distri{$nlabel} = \$pr_dist;

	}		# End of the 4th while-loop #

	<$fh_in>; # skip a line

		# Read in the reconstructed sequences #
	my %reconst_seq;
	while (my $in = <$fh_in>) { # The 5th while-loop #
	    ($in =~ m/^\s*$/) and last;

	     my ($nlabel) = ($in =~ m/^node\s+\#(\d+)/ );
	     (defined $nlabel) or next;

	     $in =~ s/^node\s+\#\d+\s+//; # Remove header part
	     $in =~ s/\s+//g;	# Remove white spaces #

	     $reconst_seq{$nlabel} = \$in;

	}		# End of the 5th while-loop #

		# Read in the signpost for the next tree #
	my $next_tree_signpost;
	while (my $in = <$fh_in>) { # The 6th while-loop #
	    if ($in =~ m/^TREE\s+\#/) {
		$next_tree_signpost = $in;
		last;
	    }
	}			# End of the 6th while-loop #

	return (\$viewtree, \%prob_distri, \%reconst_seq, $next_tree_signpost);

}	# end of the subroutine 'extract_ancestral_paml' #


# extract_topseq_paml:
#	This subroutine extracts the topnode sequence reconstructed by
#	a PAML program.
#
# INPUT: (\$viewtree, \%reconstruct_seq), where
#	$viewtree is the input tree in the format readable by TreeView;
#	%reconst_seq = ($node_label => \$seq, ...) for the internal nodes,
#	 with $seq being a DNA (or peptide) sequence reconstructed for the node.
#
# OUTPUT: ($topnode_label, \$seq_topnode), where
#	$seq_topnode is the reconstructed topnode sequence.
#
sub extract_topseq_paml (\$\%) {

	my ($viewtree, $reconst_seq) = @_;

	my ($topnode_label) = (${$viewtree} =~ m/(\d+)\s*\;\s*$/);
	unless (defined $topnode_label) {
	    print STDERR "[ERROR 'extract_topseq_paml'] NO TOPNODE LABEL found in the tree:\n";
	    print STDERR ${$viewtree}, "\n";
	}

	my $seq_topnode = $reconst_seq->{$topnode_label};

	return ($topnode_label, $seq_topnode);

}	# end of the subroutine 'extract_topseq_paml' #


# pr_ancestral_paml:
#	This subroutine prints out information on ancestral states
#	extracted by 'extract_ancestral_states' following the format
#	of the original 'rst' file.
#
# INPUT: (\*${$fh_out}, $tree_signpost, \$viewtree, \%prob_distri, \%reconst_seq), where
#	$viewtree is the input tree in the format readable by TreeView;
#	%prob_distri = ($node_label => \$prob_distri, ...) for the internal nodes,
#	with $prob_distri being a string (chunk) representing probability distributions of states on the node for sites;
#	%reconst_seq = ($node_label => \$seq, ...) for the internal nodes,
#	 with $seq being a DNA (or peptide) sequence reconstructed for the node;
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub pr_ancestral_paml (\*$\$\%\%) {

	my ($fh_out, $tree_signpost, $viewtree, $prob_distri, $reconst_seq) = @_;

	print $fh_out $tree_signpost, "\n";
	print $fh_out "tree with node labels for Rod Page's TreeView\n";
	print $fh_out ${$viewtree}, "\n";

	print $fh_out "(1) Marginal reconstruction of ancestral sequences\n\n";
	print $fh_out "Prob distribs at nodes, those with p < 0.001 not listed\n\n";

	foreach my $nlabel (sort {$a <=> $b} keys %{$prob_distri}) { # The 1st foreach-loop #
	    print $fh_out "Prob distribution at node $nlabel, by site\n\n";
	    printf $fh_out "%7s%7s%s:\n\n", 'Site', 'Freq', 'Data';
	    unless (defined $prob_distri->{$nlabel}) {
		print STDERR "[CAUTION 'pr_ancestral_paml'] UNDEFINED distribution for the ancestral node '$nlabel' !\n";
	        printf $fh_out "\n\n";
		next;
	    }
	    printf $fh_out ${$prob_distri->{$nlabel}};
	    printf $fh_out "\n\n";
	}		# End of the 1st foreach-loop #

	print $fh_out "List of extant and reconstructed sequences\n\n";

	foreach my $nlabel (sort {$a <=> $b} keys %{$reconst_seq}) { # The 2nd foreach-loop #
	     printf $fh_out "node \#%-12s", $nlabel;
	     my $seq = $reconst_seq->{$nlabel};
	     unless (defined $seq) {
		print STDERR "[ERROR 'pr_ancestral_paml'] UNDEFINED sequence for the ancestral node '$nlabel' !\n";
		print $fh_out  "\n";
		next;
	     }
	     print $fh_out ${$seq};
	     print $fh_out  "\n";
	}		# End of the 2nd foreach-loop #

	print $fh_out "\n\n";

	return 1;

}	# end of the subroutine 'pr_ancestral_paml' #


# reconstruct_topnode_paml:
#	This subroutine reconstructs the sequence on the top node
#	of a gene tree from the multiple alignment of the external nodes.
#	It infers the gap/nongap state by using parsimony principle,
#	and infers the resdue state using the PAML package.
#
# INPUT: (\@seqids, \@seqs, \@tree, \$workdir, $fl_clock, \@paml_params, \@aln_params), where
#	$seqids[$k] is the ID of the $k th sequence;
#	${$seqs[$k]} is the $k th sequence;
#	@tree = (\%node2chi, \%node2pa, \%id2name, \%name2id, \%br_attr), with
#	%node2chi = ($node_id => [children], ...) for the internal nodes;
#	%node2pa = ($node_id => $parent_id, ...) for the nodes;
#	%id2name = ($node_id => $node_name, ...) for the external nodes,
#	NOTE that the $node_name should be identical to one of the IDs stored in @seqids;
#	%name2id = ($node_name => $node_id, ...) for the external nodes;
#	%br_attr = ($node_id => \@attr, ...) with @attr is the attribute of the branch just above the node, and
#	$attr[$INDX_LENGTH] is its branch length;
#
#	$workdir is the path of the directory where the program(s) of PAML is(are) conducted;
#	$fl_clock = 1 if the molecular clock is assumed,
#		  = 0 otherwise;
#
#	@paml_params = ($fl_peptide, $fl_full_anc, \$path_matrix,
#	\$template, \%common_revisions, \%replace, \$exec_paml), with
#	    $fl_peptide = 1 if dealing with peptide sequences,
#		    = 0 if nucleotide sequences are handled;
#	    $fl_full_anc = 1 if print out the distributions of all ancestral states,
#			 = 0 if print out only the distributions of the top state;
#	    $template is the path to the template of the control file;
#	    %common_revisions is the revisions that are common
#		in this sesseion of the program;
#	    %replace = ($to_be_replaced => $to_replace, ...)
#		for characters to be taken care of
#		before feeding the alignment into the PAML program;
#	    $exec_paml is the path to the executable
#		PAML program to be used;
#
#	@aln_params = ($dis_len, $gap, $non_gap), with
#	    $dis_len is the maximum sequence length to be displayed in one line;
#	    $gap and $non_gap are charecters to represent a gap and a non-gap state, respectively.
#
# OUTPUT: (\$seq_topnode, $tmp_log, $abbr_reconstruct, \$distri_topnode), where
#	$seq_topnode is the inferred sequence of the top node;
#	$tmp_log is the name of the log file for the PAML program;
#	$abbr_reconstruct is the name of the file storing abbreviated output of 
#	the ancestral reconstruction;
#	$distri_topnode is a string (chunk) giving the state distributions
#	for all sites dominated by non-gap residues.
#
sub reconstruct_topnode_paml (\@\@\@\$$\@\@) {

	my ($seqids, $seqs, $tree, $workdir, $fl_clock,
		$paml_params, $aln_params) = @_;

	my ($node2chi, $node2pa, $id2name, $name2id, $br_attr) = @{$tree};
	my ($fl_pep, $fl_full_anc, $path_matrix, $template,
		$common_revisions, $replace, $exec_paml) = @{$paml_params};
	my ($dis_len, $gap, $non_gap) = @{$aln_params};

	my $ct_seqs = @{$seqs};
	($ct_seqs>0) or (return (undef, undef, undef, undef));

	my $aln_len = length (${$seqs->[0]});
	my @cp_seqs = ();
	foreach my $seq (@{$seqs}) { my $cp = ${$seq}; push @cp_seqs, \$cp; }

	my $topnode_id = $name2id->{$TOP_NODE};
	my $tmp_inaln = 'input.aln';
	my $tmp_intree = 'intree.ph';
	my $tmp_control = $fl_pep ? 'codeml.ctl' : 'baseml.ctl';
	my $tmp_log = 'log_paml.txt';
	my $in_reconst = 'rst';
	my $abbr_reconst = 'abbr_reconst.txt';

	my %revisions = %{$common_revisions}; # Used for revising the control file of PAML
	$revisions{'seqfile'} = $tmp_inaln;
	$revisions{'treefile'} = $tmp_intree;
	($fl_clock>0) and ($revisions{'clock'} = 1);

		## Infer the gap/nongap state of the top node ##

	my @dummy = ();
	# @{$gap_blocks->[$k]} = ($start, $end, \@blstart_seq, \@blend_seq, \@pattern, $sbalign)
	#  for the $k th block of the sites with the same phylogenetic distribution of gaps.
	#
	my $gap_blocks = extract_gap_blocks (@cp_seqs, 1, @dummy, $gap);
	my $ct_blocks = @{$gap_blocks};

	# %{$gapstat_wt_on_nodes->[$k]} = ($node_id => \@status, ...) for the $k th block,
	# with @{$status[$p]} = [$state, $weight, $min_change, $ct_chi] for the $p th state.
	#
	my $gapstat_wt_on_nodes = bch_infer_gapstat_on_nodes (@{$gap_blocks}, @{$seqids},
		%{$node2chi}, %{$node2pa}, %{$name2id}, $gap, $non_gap);

	my @top_gap_blcs = (); # element = [$start, $end] of the block where the top node is occupied by gaps.

	for (my $k=0; $k<$ct_blocks; $k++) { # The 2nd for-loop #
	    my ($start, $end, $pattern) = @{$gap_blocks->[$k]}[0,1,4];
	    my $fl_gap = 0;
	    if (2 * scalar (@{$pattern}) >= $ct_seqs) { # If a majority of the genes are occupied by gaps # 
		$fl_gap = 1;
	    } else { # If the gap is likely to be the parsimonious state of the top node

		my $top_status = $gapstat_wt_on_nodes->[$k]{$topnode_id};
		my %stat2minch;
		foreach my $info (@{$top_status}) { # The inner foreach-loop #
		    my ($stat, $minch) = @{$info}[0,2];
		    $stat2minch{$stat} = $minch;
		}				# End of the inner foreach-loop #

		if (defined (my $gap_minch = $stat2minch{$gap})) {
		    if (defined (my $ng_minch = $stat2minch{$non_gap})) {
		        ($gap_minch <= $ng_minch) and ($fl_gap = 1);
		    } else {
		        $fl_gap =1;
		    }
		}
	    }
	    $fl_gap or next;

	    my $last_block = (@top_gap_blcs>0) ? $top_gap_blcs[$#top_gap_blcs] : undef;
	    if ((defined $last_block) and ($last_block->[1] + 1 >= $start)) {
		($last_block->[1] < $end) and ($last_block->[1] = $end);
	    } else {
		push @top_gap_blcs, [$start, $end];
	    }
	}			# End of the 2nd for-loop #

	my @frame = ([1, $aln_len]);

	(undef, undef, my $top_ng_blcs) = compare_interval_sets (@top_gap_blcs, @frame);

	my ($spl_aln_ng_blcs, $aln_gap_blcs) = splice_blocks_in_aln (@{$seqs}, @{$top_ng_blcs}, 1);

	$spl_aln_ng_blcs = preprocess_aln_paml (@{$spl_aln_ng_blcs}, %{$replace});

=for TESTING the subroutine
	print STDERR "\n<Alignments of gap blocks for '${$workdir}' >\n\n";
	for (my $k=0; $k< @{$aln_gap_blcs}; $k++) {
	    printf STDERR "{ The %s th gap: \[%d, %d\] }\n\n", $k+1, @{$top_gap_blcs[$k]};
	    my $aln = $aln_gap_blcs->[$k];
	    pr_ODEN_input_lite (@{$seqids}, @{$aln}, $dis_len, *STDERR);
	    print STDERR "\n";
	}
	print STDERR "\n\n";
=cut

		## Make a working directory ##
	(my $org_dir = ` pwd `) =~ s/\s+$//;
	unless ( -d ${$workdir} ) {
	    system ("mkdir -p ${$workdir}");
	}
	unless (( -r ${$workdir} ) and ( -w ${$workdir} )) {
	    system ("chmod u+rw ${$workdir}");
	}

		## Copy the substitution matrix into the working directory ##
	unless ( system ("cp ${$path_matrix} ${$workdir}") == 0) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't copy the matrix '${$path_matrix}' into the working directory '${$workdir}': $?\n";
	    return (undef, undef, undef);
	}
		## Go into the working directory ##
	unless (chdir ${$workdir}) {
	   print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '${$workdir}': $!\n";
	    return (undef, undef, undef, undef);
	}

		
		## Print out (the nongap part of) the input alignment ##
	my $fh_inaln = open_output_file ($tmp_inaln, 0);
	print $fh_inaln ' ' x 5, $ct_seqs, ' ' x 5 , length (${$spl_aln_ng_blcs->[0]}),
		' ' x 5, 'I',  "\n\n"; # Interleaved format #
	unless ( pr_phylip_aln (@{$seqids}, @{$spl_aln_ng_blcs}, $dis_len, *${$fh_inaln}) ) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] FAILED to print out the input alignment !\n";
	    close ($fh_inaln);
	    (chdir $org_dir) or ( print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, undef, undef, undef);
	}
	close ($fh_inaln);

		## Make a hash that only stores branch lengths ##
	my %br_lens;
	foreach my $node (keys %{$br_attr}) { # The 1st foreach-loop #
	    my $len = $br_attr->{$node}[$INDX_LENGTH]; 
	    (defined $len) or next;
	    ($len < 0) and ($len = 0);
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = $len;
	    $br_lens{$node} = \@tmp;
	}		# End of the 1st foreach-loop #

		## Print out the input gene tree ##
	my $fh_intree = open_output_file ($tmp_intree, 0);
	unless ( br_output_tree ($topnode_id, %{$node2chi}, %{$id2name}, %br_lens, *${$fh_intree}, 0,0) ) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] FAILED to print out the input tree !\n";
	    close ($fh_intree);
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, undef, undef, undef);
	}
	close  ($fh_intree);

		## Prepare the control file necessary to run a PAML program ##
	unless ( prepare_paml_ctrl (${$template}, $tmp_control, %revisions) ) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] 'prepare_paml_ctrl' FAILED !\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, undef, undef, undef);
	}

		## Run a PAML program ##
	unless ( system ("${$exec_paml} > $tmp_log 2>&1") == 0) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] 'system ( \"${$exec_paml}\") FAILED: $?\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, $tmp_log, undef, undef);
	}

	unless ( ( -f $in_reconst) and ( -r $in_reconst )) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] Found NO OUTPUT FILE of ANCESTRAL RSTORATION '$in_reconst'!\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, $tmp_log, undef, undef);
	}

		## Extract main information on the reconstructed states
		## from the PAML output

	local $/ = "\n"; # Use a Line Feed as an input record separator
	my $fh_inrst = open_input_file ($in_reconst);
	my $tree_signpost;
	while (my $in = <$fh_inrst>) {
	    ($in =~ m/^TREE\s+\#/) or next;
	    $tree_signpost = $in;
	    last;
	}
	my ($viewtree, $prob_distri, $reconst_seq, $next_tree_signpost)
	    = extract_ancestral_paml (*${$fh_inrst}, $fl_full_anc);
	close ($fh_inrst);
	unless (defined $viewtree) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] NO ENTRIES found in the PAML reconstruction file '$in_reconst' !\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, $tmp_log, undef, undef);
	}

		## Extract the (truncated) sequence on the top node ##
	my ($label_topnode, $trct_seq_topnode)
		 = extract_topseq_paml (${$viewtree}, %{$reconst_seq});
	my $distri_topnode = $prob_distri->{$label_topnode};
	##my $out_distri = $fl_full_anc ? $prob_distri : {$label_topnode => $distri_topnode};
	my $out_distri = $prob_distri;

		## Output the extracted information ##
	my $fh_abrst = open_output_file ($abbr_reconst, 0);
	unless ( pr_ancestral_paml (*${$fh_abrst}, $tree_signpost, ${$viewtree}, %{$out_distri}, %{$reconst_seq}) ) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] 'pr_ancestral_paml' FAILED on the output file '$abbr_reconst'!\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, $tmp_log, undef, $distri_topnode);
	}
	close ($fh_abrst);


=for TESTING the subroutine
	print STDERR "\n{The Topnode State}\n";
	print STDERR ${$trct_seq_topnode}, "\n\n";
=cut

=for TESTING the subroutine
	my $gap_topnode = gap_indicator (@top_gap_blcs, 1, $aln_len, $gap, $non_gap);
		## Return to the original directory ##
	unless (chdir $org_dir) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n";
	    (chdir $org_dir) or (print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n");
	    return (undef, $tmp_log, $abbr_reconst, $distri_topnode);
	}
	return ($gap_topnode, $tmp_log, undef, $distri_topnode);
=cut

		## Return to the original directory ##
	unless (chdir $org_dir) {
	    print STDERR "[ERROR 'reconstruct_topnode_paml'] Can't cd to '$org_dir': $!\n";
	    return (undef, $tmp_log, $abbr_reconst);
	}

	   ## Put a Finishing Touch
	   ## by padding spaces between alignment blocks with gaps

	my $cp_trct_seq_topnode = ${$trct_seq_topnode};
	my $full_seq_topnode
	   = pad_concate_blocks ($cp_trct_seq_topnode, @{$top_ng_blcs}, 1, $aln_len, $gap);

	return ($full_seq_topnode, $tmp_log, $abbr_reconst, $distri_topnode);

}		# End of the subroutine 'reconstruct_topnode_paml' #


# parse_yn00_output:
#	This subroutine reads in an main output file of the program 'yn00'
#	in the PAML package, then it extracts 2 sets of information on the
#	dN dS values for all pairs, one estimated by Nei & Gojobori's method
#	and the other by Yang & Nielsen's method.
#	In order to save as much memory as possible,
#	it will just read in the lower triangular part
#	of the matrix.
#
# INPUT: (\*${$fh_in}, $fl_init), where
#	$fh_in is the filehandle to the main output file of 'yn00'.
#	$fl_init = 1 if the input file records more than 1 alignment
#	   AND if you have just opened the input file,
#	       = 0 otherwise.
#
# OUTPUT: (\@seqnames, \@dnds_ng, \@dnds_yn, $fl_eof), where
#	$seqnames[$i] is the name of the $i th sequence;
#	@{$dnds_ng[$i]->[$j]} = ($dN_dS_ration, $dN, $dS),
#	with (0 <= $j < $i < #{sequences}),
#	 is the dN dS information between the $i th and the $j th sequences
#	 that is estimated via the Nei-Gojobori method,
#	@{$dnds_yn[$i]->[$j]} = ($ct_syn, $ct_non, $t, $kappa, $omega, $dN, $se_dN, $dS, $se_dS),
#	with (0 <= $j < $i < #{sequences}),
#	is the dN dS information between the $i th and the $j th sequences
#	that is estimated via the Yang-Nielsen method.
#	Here I will give definitions of some unfamiliar quantities:
#	$t = #{nucleotide substitutions}/#{codons},
#	$kappa = {transition rate}/{transversion rate},
#	$omega = $dN/$dS;
#	$fl_eof = 1 if we reached the END OF FILE,
#		= 0 if we have not read in the entire file yet.
#
#	NOTE that $dists_ng[0] and $dists_yn[0] are references to empty arrays.
#
sub parse_yn00_output (\*$) {

	my ($fh_in, $fl_init) = @_;
	local $/ = "\n";

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'parse_yn00_output' >\n\n"; 
=cut

	my $in ;
	if ($fl_init) {
		# Skip the header part #
	    while ($in = <$fh_in>) { ($in =~ m/^Data\s+set/) and last; }
	    (defined $in) or (return (undef, undef, undef, 1));
=for DEBUGGING the subroutine
	    $in =~ s/\s+$//;
	    print STDERR "We are now at:\n";
	    print STDERR "'$in' .\n\n";
=cut
	}

	my @seqnames = my @dists_ng = my @dists_yn = ();

	   ## Look for the title line of the estimations via Nei & Gojobori's method ##
	while ($in = <$fh_in>) { # The 2nd while-loop #
	    if ($in =~ m/^Data\s+set/) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the end of the entry while looking for dN dS estimates via Nei & Gojobori's method.\n\n";
		return (undef, undef, undef, 0);
	    }
	    ($in =~ m/^Nei\s+\&\s+Gojobori/) and last;
	}		# End of the 2nd while-loop #
	unless (defined $in) {
	    print STDERR "[ERROR 'parse_yn00_output'] We have reached the END OF FILE while looking for dN dS estimates via Nei & Gojobori's method.\n\n";
	    return (undef, undef, undef, 1);
	}

=for DEBUGGING the subroutine
	$in =~ s/\s+$//;
	print STDERR "We are now at:\n";
	print STDERR "'$in' .\n\n";
=cut

		# Read in Nei & Gojobori's estimates #
	my $ct_seqs = 0;
	while ($in = <$fh_in>) {  # The 3rd while-loop #
	    ($in =~ m/\S/) or last;
	    if ($in =~ m/^Data\s+set/) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the end of the entry while reading in dN dS estimates via Nei & Gojobori's method.\n\n";
		return (undef, undef, undef, 0);
	    }
	    ($in =~ m/^NOTE:/) and next; # Skip the note #

	    $in =~ s/\s+$//;
	    $in =~ s/\)$//;
	    my ($seqname, $values) = split /\s+/, $in, 2;
	    push @seqnames, $seqname;
	    $ct_seqs++;
	    unless ((defined $values) and ($values =~ m/\S/)) {
		push @dists_ng, [];
		next;
	    }

	    my @values = split /\)\s+/, $values;
	    my @dnds = ();
	    foreach my $item (@values) {	# The 1st inner foreach-loop #
		(@dnds<$ct_seqs) or last;

		my ($ratio, $dn, $ds) = split /[\(\s]+/, $item;
		unless (defined $ds) {
		    print STDERR "[ 'parse_yn00_output'] I've encountered an ANOMALOSU RECORD: $item !\n";
		    push @dnds, [$ratio, $dn, 'NA'];
		    next;
		}
		push @dnds, [$ratio, $dn, $ds];
	    }				# End of the 1st inner foreach-loop #
	    push @dists_ng, \@dnds;
	}			# End of the 3rd while-loop #
	unless (defined $in) {
	    print STDERR "[ERROR 'parse_yn00_output'] We have reached the END OF FILE before we finish reading in dN dS estimates via Nei & Gojobori's method.\n\n";
	    return (undef, undef, undef, 1);
	}

=for DEBUGGING the subroutine
	print STDERR " --- Finished reading in the estimates via Nei & Gojobori's method --\n\n";
	print STDERR "[ dN dS estimates ]\n\n";
	print STDERR join ("\t", 'Sequence1', 'Sequence2', 'dN/dS', 'dN', 'dS'), "\n\n";
	for (my $i=0; $i<$ct_seqs; $i++) { # the outer temporary for-loop #
	    my $seq1 = $seqnames[$i];
	    for (my $j=0; $j<$i; $j++) { # the inner temporary for-loop #
		my $seq2 = $seqnames[$j];
		my $values = $dists_ng[$i]->[$j];
		print STDERR join ("\t", $seq1, $seq2, @{$values}), "\n";
	    }			# end of the inner temporary for-loop #
	    print STDERR "\n";
	}		# end of the outer temporary for-loop #
	print STDERR "\n";
=cut

	   ## Look for the title line of the estimations via Yang & Nielsen's method ##
	while ($in = <$fh_in>) { # The 4th while-loop #
	    if ($in =~ m/^Data\s+set/) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the end of the entry while looking for the title of dN dS estimates via Yang & Nielsen's method.\n\n";
		return (undef, undef, undef, 0);
	    }
	    ($in =~ m/^Estimation\s+by\s+the\s+method\s+of\s+Yang\s+\&\s+Nielsen/) and last;
	}		# End of the 4th while-loop #

=for DEBUGGING the subroutine
	if (defined $in) {
	    $in =~ s/\s+$//;
	    print STDERR "We are now at:\n";
	    print STDERR "'$in' .\n\n"; 
	}
=cut
	while ($in = <$fh_in>) { # The 5th while-loop #
	    if ($in =~ m/^Data\s+set/) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the end of the entry while looking for labels of dN dS estimates via Yang & Nielsen's method.\n\n";
		return (undef, undef, undef, 0);
	    }
	    ($in =~ m/^seq\.\s+seq\./) and last;
	}		# End of the 5th while-loop #
	unless (defined $in) {
	    print STDERR "[ERROR 'parse_yn00_output'] We have reached the END OF FILE while looking for the header parts for the dN dS estimates via Yang & Nielsen's method.\n\n";
	    return (undef, undef, undef, 1);
	}
=for DEBUGGING the subroutine
	$in =~ s/\s+$//;
	print STDERR "We are now at:\n";
	print STDERR "'$in' .\n\n";
=cut
	$in = <$fh_in>; # skip a blank line separating the labels and main contents

	@dists_yn = ([]);
	while ($in = <$fh_in>) { # The 6th while-loop #
	    ($in =~ m/\S/) or last;
	    if ($in =~ m/^Data\s+set/) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the end of the entry while reading in dN dS estimates via Yang & Nielsen's method.\n\n";
		return (undef, undef, undef, 0);
	    }
	    my @fields = split ' ', $in;
	    (@fields<13) and next;
	    my $indx1 = $fields[0] -1;
	    my $indx2 = $fields[1] -1;
	    my $size = @dists_yn;
	    unless ($size>$indx1) {
		for (my $k=$size; $k<=$indx1; $k++) { $dists_yn[$k] = []; }
	    }
	    my @info = @fields[2..7, 9,10,12];
	    $dists_yn[$indx1]->[$indx2] = \@info;
	}		# End of the 6th while-loop #

	unless (defined $in) {
	   unless ((@dists_yn == $ct_seqs) and (@{$dists_yn[$ct_seqs-1]} == $ct_seqs-1)) {
		print STDERR "[ERROR 'parse_yn00_output'] We have reached the END OF FILE before we finish reading in dN dS estimates via Nei & Gojobori's method.\n\n";
		return (undef, undef, undef, 1);
	   }
	}

		# Discard the trailing information #
	while ($in = <$fh_in>) { ($in =~ m/^Data\s+set/) and last; }
	my $fl_eof = (defined $in) ? 0 : 1;

=for DEBUGGING the subroutine
	print STDERR "\n=== the subroutine 'parse_yn00_output' FINISHED ===\n\n";
=cut

	return (\@seqnames, \@dists_ng, \@dists_yn, $fl_eof);

}		# End of the subroutine 'parse_yn00_output' #

1;	# a "spell" which makes the module "use"-able and/or "require"-able #


