package MyPBS_UH;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPBS_UH.pm:
#	This module produces shell script to be submitted via PBS in the UH cluster systerm,
#	and manipulates various inputs/outputs of the PBS-submitted scripts.
#	It was made by revising the module 'MyGridEngine.pm'.
#
# MyGridEngine.pm:
#	This module produces shell script to be submitted via Grid Engine,
#	and manipulates various inputs/outputs of the GridEngine-submitted scripts.
#

use strict;
use warnings;

use Exporter;

our @ISA = qw(Exporter);

use MyFile_manipu qw(
	open_input_file open_output_file
	);
use MyArrayManipu qw(
	compare_interval_sets
	);
	## diff_array union_array nr_array

=probably unnecessary
use MyTreeManipu_II qw( br_output_tree );
	## input_tree_file read_tree_file input_subtree copy_tree
	## make_node2children make_node2parent
	## move_topnode segregate_subset_nodes
	## resolve_multipartite
use MyMalnIO qw(
	pr_phylip_aln pr_ODEN_input_lite
	);
	## extract_clw pr_clustal_like_lite_aa
use MyMalnManipu_suppl qw(
	extract_gap_blocks pad_concate_blocks
	gap_indicator splice_blocks_in_aln
	);
	## rm_spec_sites 
use MyTreeMap qw(
	bch_infer_gapstat_on_nodes
	);
	## fractionate_by_ndepth assign_node2closest_exoffs
	## initialize_node2inout assign_node2bival_states
=cut

=keep the following just in case
use MyArrayManipu_num qw( union_nmarray );
use MyTaxonomy qw(
	fetch_ext_offsprings fetch_all_offsprings
	);
	## fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
use MyTreeManipu_IV qw(
	remove_trivial_nodes
	);
	## rm_external_node construct_neat_subtree
use MyTreeManipu_II_spe qw(
	input_ext_tree_file
	);
use MyMalnManipu_pep_spe qw( init_consv_hashes );
=cut

our @EXPORT_OK = qw(
	combine_pbs_options prepare_pbs_script
	) ;


##our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
##our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;


# combine_pbs_options:
#	This subroutine collects PBS options from the template,
#	and combines them with the new PBS options to be imposed.
#	Basically, original options are kept as they are,
#	as long as they don't clash with the new settings.
#	When they do, the latter replaces the former.
#	Finally, new options not specified in the old setting
#	are appended.
#
#	The subroutine was made by revising 'combine_ge_options'
#	in the module 'MyGridEngine.pm'.
#
# INPUT: (\$template, \@new_options), where
#	$template is the path to the template of a shell script for PBS ;
#	@new_options = ([$option (with a hyphen prepended), $value], ...)
#	to be used for the submitted script. If the option is just a flag
#	and takes no value, $value is an empty string;
#
# OUTPUT: \@cmb_options, which follows the same format as @new_options,
#	stores the result of combining old and new options. 
#
sub combine_pbs_options (\$\@) {

	my ($template, $new_options) = @_;
	local $/ = "\n";

	my %new_options; # ($option => $new_value, ...) for the new options
	my %replaced; # ($option => $old_value, ...) for the options replaced

	foreach my $element (@{$new_options}) { $new_options{$element->[0]} = $element->[1]; }

		# Compare the old options with the new ones #

	my @cmb_options = ();
	my $fh_in = open_input_file (${$template});
	while (my $in = <$fh_in>) { # The main while-loop #

		## REVISED in revision to 'combine_pbs_options' ##
	    ($in =~ s/^\#PBS\s+//) or next; 
	    ##($in =~ s/^\#\$\s+//) or next; 

	    $in =~ s/\s+$//;
	    my ($option, $value) = split /\s+/, $in, 2;
	    (defined $value) or ($value = '');
	    my $new_value = $new_options{$option};

	    if (defined $new_value) {
		if (defined $replaced{$option}) {
		    print STDERR "[CAUTION 'combine_pbs_options'] The option '$option' has already been used !\n";
		    next;
		}
		push @cmb_options, [$option, $new_value];
		$replaced{$option} = $value;

	    } elsif (($option eq '-j') and (defined $new_options{'-e'})) {
		if (defined $replaced{'-e'}) {
		    print STDERR "[CAUTION 'combine_pbs_options'] The option '-e', which clashes with '-j', has already been used !\n";
		    next;
		}
		push @cmb_options, ['-e', $new_options{'-e'}];
		$replaced{'-e'} = "-j $value";

	    } elsif (($option eq '-e') and (defined $new_options{'-j'})) {
		if (defined $replaced{'-j'}) {
		    print STDERR "[CAUTION 'combine_pbs_options'] The option '-j', which clashes with '-e', has already been used !\n";
		    next;
		}
		push @cmb_options, ['-j', $new_options{'-j'}];
		$replaced{'-j'} = "-e $value";

	    } else {
		push @cmb_options, [$option, $value];
	    }
	}			# End of the main while-loop #
	close ($fh_in);

		# Append the new options not replaced the old one yet #

	foreach my $element (@{$new_options}) {
	    (defined $replaced{$element->[0]}) and next;
	    push @cmb_options, $element;
	}

	return \@cmb_options;

}			# End of the subroutine 'combine_pbs_options'


# prepare_pbs_script:
#	This subroutine prepares a shell script to be submitted via PBS,
#	by revising a template in a manner specified by the input hash and array.
#
#	The subroutine was made by revising 'prepare_ge_script' in the module 'MyGridEngine.pm'.
#
# prepare_ge_script:
#	This subroutine prepares a shell script to be submitted via GridEngine,
#	by revising a template in a manner specified by the input hash and array.
#
# INPUT: (\$template, \$outfile, \@options, \%revisions), where
#	$template is the path pointing to a template of the control file;
#	$outfile is the path of the control file to be used;
#	@options = ( [$option (with a hyphen prepended), $value], ...)
#	to be used for the script.
#	Basically, the options in the template are kept as long as they clash with
#	what are contained in @options;
#	%revisions = ($variable => $value, ...)
#	lists variables you want to modify compared to the template.
#	We ASSUME that the assignment lines do NOT begin with white spaces,
#	but simply takes the form '{variable}=value' from the beginning of the line.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub prepare_pbs_script (\$\$\@\%) {

	my ($template, $outfile, $options, $revisions) = @_;
	local $/ = "\n";

	my $cmb_pbs_options = combine_pbs_options (${$template}, @{$options});

	my $fh_in = open_input_file (${$template});
	unless (defined $fh_in) {
	    print STDERR "[ERROR 'prepare_pbs_script'] FAILED to open the template:\n";
	    print STDERR "    '${$template}' !\n\n";
	    return 0;
	}

	my $fh_out = open_output_file (${$outfile}, 0);
	unless (defined $fh_out) {
	   close ($fh_in);
	   print STDERR "[ERROR 'prepare_pbs_script'] FAILED to open the output file:\n";
	   print STDERR "     '${$outfile}' !\n\n";
	   return 0;
	}

		# Print out the header part #
	my $reserved;
	while (my $in = <$fh_in>) { # The 1st while-loop # 
	    unless ($in =~ m/^\#/) {
		($in =~ m/\S/) and ($reserved = $in);
		last;
	    }
		## REVISED in revision to 'prepare_pbs_script' ##
	    if ($in =~ m/^\#PBS/) {
		my ($option)= ($in =~ m/^\#PBS\s+(\S+)/);
		if ((defined $option) and ($option eq $cmb_pbs_options->[0][0])) {
			# Replace the old option with the new one, if available
		    my $new_option = shift @{$cmb_pbs_options};
		    print $fh_out join (' ', '#PBS', @{$new_option}), "\n";
		}
		next; # Skip the old option #
	    }
	    ##($in =~ m/^\#\$/) and next; # Skip the old option #

	    print $fh_out $in;
	}			# End of the 1st while-loop #

		## REVISED in revision to 'prepare_pbs_script' ##
	print $fh_out '#', "\n";
	##print $fh_out "\n";

		# Print out combined options #
	foreach my $info (@{$cmb_pbs_options}) { # The 2nd foreach-loop #
	    my ($option, $value) = @{$info};
	    ($value ne '') and ($option .= ' ' . $value );

		## REVISED in revision to 'prepare_pbs_script' ##
	    print $fh_out '#PBS' . ' '  . $option, "\n";
	    ##print $fh_out '#$' . ' '  . $option, "\n";
	}			# End of the 2nd foreach-loop #

	print $fh_out "\n"; ## ADDED in revision to 'prepare_pbs_script' ##


		# Modify the parameter section #

	my %already; # ( $variable => 1, ...) for the variables already modified #

	while (1) { # The main while-loop #
	    my $line;
	    if (defined $reserved) {
		$line = $reserved;
		$reserved = undef;
	    } else {
		$line = <$fh_in>;
		(defined $line) or last;
	    }

		## REVISED in revision to 'prepare_pbs_script' ##
	    ($line =~ m/^\#PBS/) and next; ## Skip the old PBS options #
	    ## ($line =~ m/^\#\$/) and next; ## Skip the old GE options #

	    if (uc ($line) =~ m/END\s+OF\s+(?:THE\s+)?PARAMETER\s+SETTING\s+BY\s+THE\s+USER/) {
		$reserved = $line;
		last;
	    }

	    my ($variable, $value) = ($line =~ m/^(\w+)\=(\S*)/ );

		# leave the line intact if no change is required #
	    unless ((defined $variable) and (defined $value)
		   and (defined $revisions->{$variable})) {
		print $fh_out $line;
		next;
	    }

	    $already{$variable} = 1;

	    if ( (($value =~ m/^\'/) and ($value =~ m/\'$/)) or
		 (($value =~ m/^\"/) and ($value =~ m/\"$/)) ) {
		substr ($value, 0, 1, '');
		substr ($value, -1, 1, '');
	    }

	    my $new_value = $revisions->{$variable};

		# leave the line intact if the variable is expected to remain the same #
	    if ($new_value eq $value) {
		print $fh_out $line;
		next;
	    }

		# Modify the parameter as specified #
	    print $fh_out '##', ' ' ,  $line;
	    ($new_value =~ m/^\d+$/) or ($new_value = "'$new_value'");
	    print $fh_out $variable, '=', $new_value, "\n\n";

	}			# End of the main while-loop #

		## Specify the remaining variables ##

	print $fh_out "\n", '    # The remaining variables to be fixed', "\n\n"; 

	my @additional = ();
	foreach my $variable (sort keys %{$revisions}) { # The subsidiary foreach-loop #
	    (defined $already{$variable}) and next;
	    
	     printf $fh_out "%s=%s\n", $variable, $revisions->{$variable}; 
	     push @additional, $variable;

	}			# End of the subsidiary foreach-loop #
	print $fh_out "\n";

	if (defined $reserved) {
	    print $fh_out $reserved;
	} elsif (@additional>0) {
	    print STDERR "[WARNING 'prepare_pbs_script'] It may have been too late specifying the additional variables :\n";
	    print STDERR ' ' x 5, join (', ', @additional), " .\n\n";
	}

		## Print out the remaining part until the end of the script file ##
	while (my $line = <$fh_in>) { print $fh_out $line; }

	close ($fh_in);
	close ($fh_out);

	return 1;

}	# end of the subroutine 'prepare_pbs_script' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


